/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - Classification.h                                                                                           */
/* - ****************                                                                                           */
/* -                                                                                                            */
/* - Schnittstelle der Klassifikationsroutinen fuer SUEM.                                                       */
/* -                                                                                                            */
/* - 14.03.96 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#ifndef CLASSIFICATION_H
#define CLASSIFICATION_H


#include "TermOperationen.h"
#include "general.h"
#include "RUndEVerwaltung.h"
#include "InfoStrings.h"
#include "SUEBasis.h"

#ifdef __cplusplus
extern "C" {
#endif

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Initialize ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void C_ParamInfo(ParamInfoArtT what, int whichModule, 
                 InfoStringT clasDEF, InfoStringT clasUSR, 
                 InfoStringT reclasDEF, InfoStringT reclasUSR, 
                 InfoStringT cgDEF, InfoStringT cgUSR);
/* whichModule: 1 - CLAS
                2 - RECLAS
                3 - CGCLAS
*/

void C_InitApriori(void);

BOOLEAN C_INITIALIZE(InfoStringT clasDEF, InfoStringT clasUSR, 
                     InfoStringT reclasDEF, InfoStringT reclasUSR, 
                     InfoStringT cgDEF, InfoStringT cgUSR, 
                     int *WeightEntryType);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Classify --------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

typedef void (* ClassifyProcT) (WeightEntryT *, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT, UTermT);
typedef int (* ReClassifyProcT) (WeightEntryT *, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT);
typedef void (* PrepareProcT) (void);

extern ClassifyProcT   C_ClassifyProcAR [4], C_CG_ClassifyProc;
extern ReClassifyProcT C_ReClassifyProc;
extern PrepareProcT    C_PrepareProc;

#define C_PREPARE() ((* C_PrepareProc)())

#define C_CLASSIFY(wtPtr, lhs, rhs, actualParent, otherParent, actualPosition, history)\
( (* C_ClassifyProcAR[history]) (wtPtr, lhs, rhs, actualParent, otherParent, actualPosition))

#define C_CGCLASSIFY(wtPtr, lhs, rhs, actualParent, otherParent, actualPosition)\
( (* C_CG_ClassifyProc) (wtPtr, lhs, rhs, actualParent, otherParent, actualPosition))

#define C_RECLASSIFY(wtPtr, lhs, rhs, actualParent, otherParent)\
( (* C_ReClassifyProc) (wtPtr, lhs, rhs, actualParent, otherParent))
/* returns 0 iff the classification has not changed
           1 iff the classification has become better
           2 iff the classification has become worse
*/

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Output ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* C_PrintProc) (WeightEntryT *, RegelOderGleichungsT, RegelOderGleichungsT);

#define C_PrintWeight(wtPtr, actualParent, otherParent)\
((* C_PrintProc) (wtPtr, actualParent, otherParent))

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Statistics ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


void C_Statistics(InfoStringT IS);

extern void (* C_WeightStatisticsStartProc) (InfoStringT);
extern void (* C_WeightStatisticsEndProc) (void);
extern PQ_WatchingWalkthroughProcedureT C_WeightStatProc;

#define C_WeightStatisticsStart(StatisticsIS) ((* C_WeightStatisticsStartProc) (StatisticsIS))
#define C_WeightStatisticsEnd() ((* C_WeightStatisticsEndProc) ())


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Additional Comparison -------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern int (* C_AdditionalComparisonProc) (long int, long int, void *, TermpairT, RegelOderGleichungsT, RegelOderGleichungsT,
                                           long int, long int, void *, TermpairT, RegelOderGleichungsT, RegelOderGleichungsT);

#define C_AdditionalComparison(w11, w12, p1, tp1, aP1, oP1, w21, w22, p2, tp2, aP2, oP2)\
( (* C_AdditionalComparisonProc) (w11, w12, p1, tp1, aP1, oP1, w21, w22, p2, tp2, aP2, oP2))
 /* Ergebnis 0 : gleich
             1 : 1. Element ist groesser
             2 : 2. Element ist groesser
  */

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Deletion of Weight-Entries --------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* C_KillAdditionalReferenceProc) (void *);

#define C_KillAdditionalReference(addRef) ( (* C_KillAdditionalReferenceProc) (addRef))

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ CLEANING WOMAN ---------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void C_CleanUp(void);

#ifdef __cplusplus
}
#endif

#endif

