/*=================================================================================================================
  =                                                                                                               =
  =  DomaenenMemorium.h                                                                                           =
  =                                                                                                               =
  =  Kapselung mmap-Speicherzugriff                                                                               =
  =                                                                                                               =
  =  19.02.97 Thomas: aus StrahlMemorium.h ausgelagert                                                            =
  =                                                                                                               =
  =================================================================================================================*/


#ifndef DOMAENENMEMORIUM_H
#define DOMAENENMEMORIUM_H

#include "InfoStrings.h"

#ifdef __cplusplus
extern "C" {
#endif



/*=================================================================================================================*/
/*= I. Aufteilung des virtuellen Adressraums                                                                      =*/
/*=================================================================================================================*/

/* Jeder Domaene muss bei der Einrichtung ein Adressraum zugewiesen werden. Die Angabe erfolgt in der Form (a,b)
   mit 0x0 <= a <= b <= 0xF und bezieht sich auf den Adressraum [a*0x10000000,b*0x10000000+0x0FFFFFFF].
   Beim Einrichten wird ueberprueft, ob a und b sinnvolle Werte haben und der beanspruchte Adressraum nicht schon
   belegt ist. Waechst eine Domaene ueber ihre Grenze hinaus, so wird ein Systemabbruch ausgeloest. Nach dem Frei-
   geben einer Domaene steht ihr Adressraum wieder zur Verfuegung. 

   Adressraumaufteilung nach Konzept 19.02.97:
   -Bereich  0x0:         1/4 GByte Adressraum fuer Programmkode, statische und dynamische Variablen (UNIX-malloc)
   *Bereiche 0x1 bis 0x4:  1  GByte Adressraum fuer DOMAENE SpeicherManager (+)
   -Bereich  0x5:         1/4 GByte Adressraum fuer zukuenftige Erweiterungen
   *Bereiche 0x6 bis 0x9:  1  GByte Adressraum fuer DOMAENE Heap of Heaps 

   Ergaenzung 14.03.97 fuer Waldmeister-TW:
   -Bereich  0x5:         1/4 GByte Adressraum fuer elegante Versendeoperationen


   Diese Bereiche sind als #define-Konstanten eingetragen in das Modul SpeicherParameter.h;
   etwaige Aenderungen sind also genau dort vorzunehmen.
   (+) Unter Linux ist der Bereich 0x4 reserviert, der Adressraum fuer die Domaene 
   SpeicherManager reicht dann nur bis 0x3. */



/*=================================================================================================================*/
/*= II. Initialisierung; Allgemeines                                                                              =*/
/*=================================================================================================================*/

extern size_t DM_Seitengroesse;                                                                /* SW-Seitengroesse */
#define /*size_t*/ DM_Seitengroesse(/*void*/)                                                                       \
  DM_Seitengroesse

void DM_InitApriori(void);

void DM_NieMehrDomaenen_SagNiemalsNie(void);      /* Falls kein anonymes Mapping, die Datei /dev/zero/ schliessen. */

#define DM_STATISTIK 0
/* Alternativen: 0 fuer aus, 1 fuer an */

#if DM_STATISTIK
  #define if_sst(Action)  Action
  #define if_sstF(Action) (Action,0)
#elif !DM_STATISTIK
  #define if_sst(Action)
  #define if_sstF(Action) (void) 0
#else
  #error "DM_STATISTIK muss als 0 oder 1 definiert sein."
#endif




/*=================================================================================================================*/
/*= IV. Domaenen                                                                                                  =*/
/*=================================================================================================================*/

typedef struct DomaenenTS *DomaenenT;

typedef enum {Strahldomaene, Managerdomaene} DomaenenartenT;

typedef void (*DomaenenhakenT)(void *Kollision);

struct DomaenenTS {
  const char *Name;
  void *Start,
       *Ende,
       *Frei,
       *Obergrenze;
  void *Sonstiges; /* Stoepsel fuer Spezialisierungen */
  DomaenenartenT Domaenenart;
  DomaenenhakenT Haken;
  volatile DomaenenT Vorg,
                     Nachf;
  if_sst(
    unsigned long int AnzahlDomaenenausdehnungen;
    unsigned long int AnzahlGefroreneStrahlen;
    unsigned long int AnzahlGeloeschteStrahlen;
    unsigned long int AnzahlGemapptePages;
    unsigned long int AnzahlGeunmapptePages;
    unsigned long int AnzahlMaxGemapptePages;)
};

/* Start, Ende, Frei, Obergrenze sind Startadressen von Seiten. Mittels mmap sind genau die Seiten [Start,Ende)
   belegt worden. Genau die Seite Ende-1 ist mit mprotect geschuetzt worden. Ist bei Verwendung als Strahldomaene 
   ein Strahl eingefroren worden, so sind genau die Seiten [Frei,Ende) noch nicht benutzt. Die Domaene darf maximal 
   die Ausdehnung [Start,Obergrenze) haben. Inkrement ist Vielfaches der Seitengroesse. */


DomaenenT DM_NeueDomaene(DomaenenartenT Domaenenart, unsigned int Start, unsigned int Obergrenze,  
  DomaenenhakenT Haken, const char *Name);

void DM_AmEndeGesperrtAllokieren(DomaenenT Domaene, unsigned long int Delta, void *Hakendatum);

void DM_DomaeneLeeren(volatile DomaenenT Domaene);      /* alle Seiten weggeben, aber Verwaltungsstruktur behalten */

void DM_DomaeneLoeschen(volatile DomaenenT Domaene);                          /* auch Verwaltungsstruktur weggeben */



/*=================================================================================================================*/
/*= III. Mapping, Unmapping, Seitenschutz                                                                         =*/
/*=================================================================================================================*/

void DM_mmap(void *Start, void *Ende);              /* Seiten [Start,Ende) holen. Kein Test, ob Ausdehnung erlaubt.*/

void DM_munmap(void *Start, void *Ende);                                          /* Seiten [Start,Ende) weggeben. */




/*=================================================================================================================*/
/*= V. Statistik                                                                                                  =*/
/*=================================================================================================================*/

void DM_Statistik(InfoStringT IS);


#ifdef __cplusplus
}
#endif

#endif
