/* Datei: error_handling.h
 * Zweck: Stellt vier verschiedene Behandlungsm"oglichkeiten f"ur Fehler zur Verf"ugung
 * Autor: Arnim Buch
 * Stand: 28.02.94
 * Ersterstellung: 28.02.94.  ACHTUNG: Bisher geben alle Prozeduren nur die Meldungen aus
 *                                     und verabschieden sich.
 * Aenderungen:
 * 
 * 
 * 
 * 
 * 
 * 
 */

#ifndef FEHLERBEHANDLUNG_H
#define FEHLERBEHANDLUNG_H


#include "general.h"

#ifdef __cplusplus
extern "C" {
#endif


/* -------------------------------------------------------------------------------*/
/* --------------- Interrupt-Kontrolle -------------------------------------------*/
/* -------------------------------------------------------------------------------*/

extern BOOLEAN volatile FB_IRFlag; 
extern BOOLEAN volatile FB_Usr1IRFlag;

#define FB_InterruptflagSetzen() FB_IRFlag = TRUE
#define FB_InterruptflagLoeschen() FB_IRFlag = FALSE
#define FB_InterruptflagGesetzt() FB_IRFlag

#define FB_Usr1InterruptflagSetzen() FB_Usr1IRFlag = TRUE
#define FB_Usr1InterruptflagLoeschen() FB_Usr1IRFlag = FALSE
#define FB_Usr1InterruptflagGesetzt() FB_Usr1IRFlag

/* -------------------------------------------------------------------------------*/
/* --------------- Allgemeine Fehlermeldungen  -----------------------------------*/
/* -------------------------------------------------------------------------------*/

void our_assertion_failed (const char *hoppala);
/* Bringt zum Ausdruck, wenn eine eigentlich unm"ogliche Konstellation erreicht wird.
   (Implementierungs-Fehler). */

void our_warning(const char *autsch);
/* Gibt einfach den uebergebenen Text aus und f"ahrt im Programm fort. */

void our_error(const char *aua);
/* Gibt die angegebene Fehlermeldung aus und gibt die Macht an den Benutzer zurueck.*/

void our_fatal_error(const char *aechhh);
/* Gibt die Fehlermeldung aus und verabschiedet sich aus dem Programm. */

void not_yet_implemented(const char *functionName);
/* Gibt "Not yet implemeted: <fuction_name> aus, und TSCHUEEESSSSSSSSS */

/* -------------------------------------------------------------------------------*/
/* --------------- Fehlersuche und -beseitigung ----------------------------------*/
/* -------------------------------------------------------------------------------*/

/* Es werden drei Arten von Fehlertest bereitgestellt:
   1. Unbedingter Fehlertest.
   2. Von bis zu vier Zaehlern abhaengiger Fehlertest, der bei Erreichen/Ueberschreiten der zuvor entweder
      per Aufruf-Option oder durch expliziten Prozeduraufruf eingestellten Marken aktiviert wird.
      Die Zaehler koennen mit den vier angegebenen INC-Makros erhoeht werden. Der Fehlertest selbst
      ist NICHT Interrupt-Gesteuert, muss also an die entsprechenden Code-Stellen eingebaut werden.
      Werden die AutoInc-Zaehl-Makros verwendet, so erfolgt der Fehlertest sofort nach der Erhoehung
      des Zaehlers.
      Werden die Aufruf-Optionen zur Besetzung der Marken herangezogen, so wird dabei die zusaetzliche
      Option NoDebug herangezogen, bei deren Angabe kein Abbruch bei Erreichen der Marken erfolgt.
      Dies ermoeglicht bei eincompilierten Abbruch-Aufrufen normale Laeufe.
   3. Von Signal abhaengiger Fehlertest, dessen Aufruf zwar vom Setzen des Signals abhaengig,
      zeitlich aber davon entkoppelt ist.
   Alle Fehlertests loeschen saemtliche zugaengliche Systemstrukturen und aktivieren den Speichertest.
   Ferner wird die Anzahl der noch auf dem Zeitstack liegenden Zeiten angegeben.
   Nach Ablauf des Tests wird das Programm verlassen.
*/


void FB_Fehlertest(void);
/* Nicht von Bedingungen abhaengiger Fehlertest */

extern unsigned long FB_Z[10];

#define FB_IncZ(i) (FB_Z[i]++)

#define FB_AutoIncZ(i) (FB_Z[i]++,FB_FehlertestBeiErreichterMarke())

#define FB_AktionIncZ(i, /*Block*/ Aktion)\
{\
    FB_Z[i]++;\
    FB_AktionBeiErreichterMarke(/*Block*/ Aktion);\
}

#define FB_ZaehlerSetzen(i, wert) (FB_Z[i]=wert)

void FB_MarkeSetzen(unsigned long m1,unsigned long m2,unsigned long m3,unsigned long m4,
                    unsigned long m5,unsigned long m6,unsigned long m7,unsigned long m8,
                    unsigned long m9,unsigned long m0);
/* Loescht bei gesetztem Stopp-Signal alle zugreifbaren System-Strukturen,
   so dass im fehlerfreien Falle keine Diskrepanzen in der Speicherstatistik auftreten
   duerften.*/

void FB_DebugOptionAlsMarke(void);
/* Setzt die Stopp-Signale auf die Werte der Debug-Option 0,...9 */

BOOLEAN FB_MarkeErreicht(void);
/* Liefert TRUE, wenn die eingestellte Marke erreicht wurde und nicht NoDebug gewaehlt wurde. */

void FB_FehlertestBeiErreichterMarke(void);

#define FB_AktionBeiErreichterMarke(/*Block*/ Aktion) {if(FB_MarkeErreicht()) Aktion;}

void FB_DruckeAktuelleZaehler(void);
/* Gibt die aktuellen Zaehlerstaende aus */

void FB_FehlerSignalLoeschen(void);
/* Initialisieren der Fehlermarke. Ist nicht unbedingt erforderlich, da das Signal mit FALSE initialisiert wird.*/

void FB_FehlerSignalSetzen(void);
/* Setzt ein Fehlerflag,... */

BOOLEAN FB_SignalGesetzt(void);
/* Liefert TRUE, wenn das Signal gesetzt wurde. */

void FB_FehlertestBeiSignal(void);
/* auf das diese Prozedur mit dem Abraeumen aller Strukturen und einem Speichertest reagiert.*/

#define FB_AktionBeiGesetztemSignal(/*Block*/ Aktion) {if (FB_SignalGesetzt()) Aktion;}


#ifdef __cplusplus
}
#endif

#endif /* FEHLERBEHANDLUNG_H */
