/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - KPPairing.h                                                                                                */
/* - ***********                                                                                                */
/* -                                                                                                            */
/* - Enthaelt die Operationen zur Verwaltung der Kritischen-Paar-Menge als Paring-Heap, aber nicht den Pairing- */
/* - Heap selbst.                                                                                               */
/* -                                                                                                            */
/* - Diese Datenstruktur wird verwendet in den Vervollstaendigungsbeispielen (completion, proof), wenn          */
/* - komplexeren Optionen wie Waisenmord oder Interreduktion der KP-Menge eingestellt sind.                     */
/* -                                                                                                            */
/* - 09.03.95 Arnim. Angelegt in Aufspaltung von KPVerwaltung.h                                                 */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#ifndef KPPAIRING_H
#define KPPAIRING_H


#include "general.h"
#include "InfoStrings.h"

#include "Classification.h"
#include "TermpairRepresentation.h"
#include "SUEBasis.h"

#include "RUndEVerwaltung.h"

#ifdef __cplusplus
extern "C" {
#endif

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Initialize ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

BOOLEAN KPP_INITIALIZE(InfoStringT DEF, InfoStringT USR, int WeightEntryType);


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Insert ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* KPP_INSERT)   (TermpairT, WeightEntryT, RegelOderGleichungsT, RegelOderGleichungsT);
extern void (* KPP_INSERTImmortal)   (TermpairT, WeightEntryT, RegelOderGleichungsT, RegelOderGleichungsT);

extern BOOLEAN (* KPP_POSTPONE) (TermpairT, WeightEntryT, RegelOderGleichungsT, RegelOderGleichungsT);


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Select ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern BOOLEAN (* KPP_SELECT) (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *, void **);

extern BOOLEAN (* KPP_SELECTINITIAL) (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *, void **);


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Parents ---------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void KPP_NEWPARENT(RegelOderGleichungsT re);

void KPP_KILLPARENT(RegelOderGleichungsT re);

void KPP_REDIRECTPARENT(void);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Walkthrough -----------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

unsigned long int KPP_AnzKPsToGo(void);

extern void (* KPP_WALKTHROUGH) (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms);

void KPP_WATCHINGWALKTHROUGH(PQ_WatchingWalkthroughProcedureT AP);

void KPP_TermVerfolgung(void (* OperationAufTermen) (TermT, TermT));

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Statistics ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void KPP_STATISTICS(InfoStringT is);
/* output    -- Ausgabe der KP-Menge
   limit=... -- dieser bis hoechstens Gewicht ...
*/
void KPP_ActSize(unsigned long int *AktKPZahl, unsigned long int *AktGueltigeKPZahl);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Clean up --------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void KPP_SoftAbraeumen(void);
void KPP_Abraeumen(void);
/* Alles loeschen*/


#ifdef __cplusplus
}
#endif

#endif
