#ifdef __cplusplus
extern "C" {
#endif

#ifndef NFADHOC_H
#define NFADHOC_H
/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFAdhoc.h                                                                                                  */
/* - *********                                                                                                  */
/* -                                                                                                            */
/* - Enthaelt die Adhoc-Version der Normalform-Bildung, dh. Termdurchlauf ohne jede Pfadverwaltung.             */
/* - Es wird auch nur zwischen zwei Varianten unterschieden, naemlich zwischen 'nochmal' und 'direkt'.          */
/* - Bei 'nochmal' wird eine reduzierte Stelle zunaechst fertigreduziert, bevor weitergelaufen wird;            */
/* - Bei 'direkt' wird auch nach einer Reduktion weitergelaufen.                                                */
/* -                                                                                                            */
/* - Strategie: Der zu reduzierende Term wird von vorne nach hinten durchlaufen. Dabei wird jede Stelle auf     */
/* -            Reduzierbarkeit geprueft und gegebenenfalls reduziert. War waehrend eines solchen Durchlaufes   */
/* -            eine Reduktion moeglich, so wird der Durchlauf wiederholt, andernfalls ist der Term jetzt       */
/* -            auf Normalform.                                                                                 */
/* -                                                                                                            */
/* - 02.03.95 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/



#include "TermOperationen.h"    /* wegen TermT */
#include "NFBasis.h"

void NFA_Init(void);

BOOLEAN NFA_NormalformZuRegelnDirekt(TermT Term, ObjektT objekt);
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach erfolgter Reduktion wird direkt weitergegangen.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/

BOOLEAN NFA_NormalformZuRegelnNochmal(TermT Term, ObjektT objekt);
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach erfolgter Reduktion wird zunaechst die betroffene Stelle fertigreduziert.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/

BOOLEAN NFA_NormalformZuGleichungenDirekt(TermT Term, ObjektT objekt);
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach erfolgter Reduktion wird direkt weitergegangen.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/

BOOLEAN NFA_NormalformZuGleichungenNochmal(TermT Term, ObjektT objekt);
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regelsystems.
   Nach erfolgter Reduktion wird zunaechst die betroffene Stelle fertigreduziert.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/

BOOLEAN NFA_NormalformBildenDirekt(TermT Term, ObjektT objekt);
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regel- und Gleichungssystems.
   Nach erfolgter Reduktion wird direkt weitergegangen.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/

BOOLEAN NFA_NormalformBildenNochmal(TermT Term, ObjektT objekt);
/* Bildet destruktiv! die Normalform zu Term bezueglich des aktuellen Regel- und Gleichungssystems.
   Nach erfolgter Reduktion wird zunaechst die betroffene Stelle fertigreduziert.
   Rueckgabe ist, ob Reduktionen stattgefunden haben.*/

#endif

#ifdef __cplusplus
}
#endif
