#ifdef __cplusplus
extern "C" {
#endif

#ifndef NFOUTERMOSTRE_H
#define NFOUTERMOSTRE_H
/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFOutermostRE.h                                                                                            */
/* - ***************                                                                                            */
/* -                                                                                                            */
/* - Enthaelt die Operationen zur Normalform-Bildung (Leftmost-Outermost, Fassung mit Gleichberechtigung)       */
/* -                                                                                                            */
/* - 06.12.95 Arnim. Angelegt.                                                                                  */
/* - 10.12.95 Arnim. NFBildung mit Bevorzugung der Gleichungsmenge hinzugefuegt.                                */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* - In diesem Modul werden Regeln und Gleichungen 'gleichberechtigt' bei der NFBildung verwendet. Dies bedeu-  */
/* - tet, dass jede Stelle des zu durchlaufenden Terms nacheinander auf Reduzierbarkeit mit der Regel- und der  */
/* - Gleichungsmenge geprueft wird. Ob nun zunaechst die Regelmenge oder die Gleichungsmenge angewendet wird,   */
/* - ist aus dem Praefix der NFBildungs-Routine ersichtlich (NFORE bzw. NFOER).                                 */
/* -------------------------------------------------------------------------------------------------------------*/


#include "general.h"
#include "NFBasis.h"
#include "TermOperationen.h"

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Initialisieren -----------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

void NFORE_Init(void);
/* Die Initialisierung geschieht in Abhaengigkeit von der eingestellten NF-Strategie. Deshalb wird sie auch erst beim
   Initialisieren der NFBildung aufgerufen, und nicht direkt aus IN_AllesAprioriInit.*/


/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Gleichungen ------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den uebergebenen Term bzgl. des aktuellen Glgs.systems auf Normalform. Dabei wird aber */
/* jeweils eine andere Variante der Leftmost-Outermost-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste   */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFORE_NormalformZuGleichungenAbDirekt (TermT Term, ObjektT objekt);   
/* nach Reduktion direkt Abstieg von Wurzel */

BOOLEAN NFORE_NormalformZuGleichungenAbNochmal (TermT Term, ObjektT objekt);  
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */

BOOLEAN NFORE_NormalformZuGleichungenAufDirekt (TermT Term, ObjektT objekt);  
/*  nach Reduktion direkt Aufstieg */

BOOLEAN NFORE_NormalformZuGleichungenAufNochmal (TermT Term, ObjektT objekt);
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln -----------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den uebergebenen Term bzgl. des aktuellen Regelsystems auf Normalform. Dabei wird aber */
/* jeweils eine andere Variante der Leftmost-Outermost-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste   */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFORE_NormalformZuRegelnAufNochmal(TermT Term, ObjektT objekt);
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */

BOOLEAN NFORE_NormalformZuRegelnAufDirekt(TermT Term, ObjektT objekt);  
/*  nach Reduktion direkt Aufstieg */

BOOLEAN NFORE_NormalformZuRegelnAbNochmal(TermT Term, ObjektT objekt);  
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */

BOOLEAN NFORE_NormalformZuRegelnAbDirekt(TermT Term, ObjektT objekt);   
/* nach Reduktion direkt Abstieg von Wurzel */


/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln und Gleichungen -------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren sorgen dafuer, dass nach einer Reduktion die der reduzierten Stelle uebergeordneten Stellen    */
/* auf Reduzierbarkeit geprueft werden. Dazu werden alle Stellen bis zu der Ebene, bei der Matches wegen direkter Clashes  */
/* gescheitert sind, probiert sowie alle darueberliegenden Stellen, an denen indirekte Clashes Reduktionen verhinderten.   */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFORE_NormalformAufNochmal(TermT Term, ObjektT objekt);
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */

BOOLEAN NFORE_NormalformAufDirekt(TermT Term, ObjektT objekt);  
/*  nach Reduktion direkt Aufstieg */

BOOLEAN NFORE_NormalformAbNochmal(TermT Term, ObjektT objekt);  
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */

BOOLEAN NFORE_NormalformAbDirekt(TermT Term, ObjektT objekt);   
/* nach Reduktion direkt Abstieg von Wurzel */

BOOLEAN NFOER_NormalformAufNochmal(TermT Term, ObjektT objekt);
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */

BOOLEAN NFOER_NormalformAufDirekt(TermT Term, ObjektT objekt);  
/*  nach Reduktion direkt Aufstieg */

BOOLEAN NFOER_NormalformAbNochmal(TermT Term, ObjektT objekt);  
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */

BOOLEAN NFOER_NormalformAbDirekt(TermT Term, ObjektT objekt);   
/* nach Reduktion direkt Abstieg von Wurzel */


#endif

#ifdef __cplusplus
}
#endif
