#ifdef __cplusplus
extern "C" {
#endif

#ifndef NFOUTERMOST_SPECIAL2_H
#define NFOUTERMOST_SPECIAL2_H
/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFOutermostSpecial2.h                                                                                      */
/* - *********************                                                                                      */
/* -                                                                                                            */
/* - Enthaelt die Operationen zur Normalform-Bildung (Leftmost-Outermost)                                       */
/* -                                                                                                            */
/* - 15.04.94 Arnim. Angelegt.                                                                                  */
/* - 26.10.95 Arnim. Aus NFOutermost erzeugt.                                                                   */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#include "general.h"
#include "NFBasis.h"
#include "TermOperationen.h"


/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Initialisieren -----------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

void NFOS2_Init(void);
/* Die Initialisierung geschieht in Abhaengigkeit von der eingestellten NF-Strategie. Deshalb wird sie auch erst beim
   Initialisieren der NFBildung aufgerufen, und nicht direkt aus IN_AllesAprioriInit.*/


/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Gleichungen ------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den uebergebenen Term bzgl. des aktuellen Glgs.systems auf Normalform. Dabei wird aber */
/* jeweils eine andere Variante der Leftmost-Outermost-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste   */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFOS2_NormalformZuGleichungenAbDirekt (TermT Term, ObjektT objekt);   
/* nach Reduktion direkt Abstieg von Wurzel */

BOOLEAN NFOS2_NormalformZuGleichungenAbNochmal (TermT Term, ObjektT objekt);  
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */

BOOLEAN NFOS2_NormalformZuGleichungenAufDirekt (TermT Term, ObjektT objekt);  
/*  nach Reduktion direkt Aufstieg */

BOOLEAN NFOS2_NormalformZuGleichungenAufNochmal (TermT Term, ObjektT objekt);
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln -----------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den uebergebenen Term bzgl. des aktuellen Regelsystems auf Normalform. Dabei wird aber */
/* jeweils eine andere Variante der Leftmost-Outermost-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste   */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFOS2_NormalformZuRegelnAufNochmal(TermT Term, ObjektT objekt);
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */

BOOLEAN NFOS2_NormalformZuRegelnAufDirekt(TermT Term, ObjektT objekt);  
/*  nach Reduktion direkt Aufstieg */

BOOLEAN NFOS2_NormalformZuRegelnAbNochmal(TermT Term, ObjektT objekt);  
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */

BOOLEAN NFOS2_NormalformZuRegelnAbDirekt(TermT Term, ObjektT objekt);   
/* nach Reduktion direkt Abstieg von Wurzel */


/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln und Gleichungen -------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den  Term bzgl. des akt. Regel- und Gleichungssystems auf Normalform. Dabei wird aber  */
/* jeweils eine andere Variante der Leftmost-Outermost-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste   */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/* Es wird erst dann versucht, Gleichungen anzuwenden, wenn es keine anwendbaren Regeln mehr gibt.                         */
/* In diesem Fall wird versucht, eine anwendbare Gleichung zu finden. Gelingt dies, so wird die Gleichung angewendet, und  */
/* anschliessend wird an dieser Stelle mit Regeln fortgefahren. Dies wiederholt sich, bis es weder anwendbare Regeln noch  */
/* Gleichungen mehr gibt.                                                                                                  */
/*-------------------------------------------------------------------------------------------------------------------------*/


BOOLEAN NFOS2_NormalformAufNochmal(TermT Term, ObjektT objekt);
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */

BOOLEAN NFOS2_NormalformAufDirekt(TermT Term, ObjektT objekt);  
/*  nach Reduktion direkt Aufstieg */

BOOLEAN NFOS2_NormalformAbNochmal(TermT Term, ObjektT objekt);  
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */

BOOLEAN NFOS2_NormalformAbDirekt(TermT Term, ObjektT objekt);   
/* nach Reduktion direkt Abstieg von Wurzel */


#endif

#ifdef __cplusplus
}
#endif
