/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - PriorityQueue.h                                                                                            */
/* - ***************                                                                                            */
/* -                                                                                                            */
/* - Schnittstelle fuer das Set of unselected equations.                                                        */
/* -                                                                                                            */
/* - 14.03.96 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#ifndef PQ_H
#define PQ_H


#include "TermOperationen.h"
#include "general.h"
#include "RUndEVerwaltung.h"
#include "InfoStrings.h"
#include "Classification.h"
#include "TermpairRepresentation.h"
#include "SUEBasis.h"

#ifdef __cplusplus
extern "C" {
#endif

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Initialize ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void PQ_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR);

BOOLEAN PQ_INITIALIZE(InfoStringT DEF, InfoStringT USR, int WeightEntryType);
/*
   WeightEntryType: 0 -> 1 long int
                    1 -> 2 long ints
                    2 -> 2 long ints, 1 addref; Anwendung der comparisonProcedure aus SUEBasis.h
   immortalCPsDynamic = TRUE  : CPs zu SUE-Demigod sind dynamische Objekte (extra freizugeben)
                        FALSE : ... koennen ohne Extra-Aufruf freigegeben werden

*/

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Insert ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
extern void    (* PQ_InsertProc)           (TermpairT , WeightEntryT , RegelOderGleichungsT , RegelOderGleichungsT);
extern void    (* PQ_InsertImmortalProc)   (TermpairT , WeightEntryT , RegelOderGleichungsT , RegelOderGleichungsT);
extern BOOLEAN (* PQ_PostponeProc)         (TermpairT , WeightEntryT , RegelOderGleichungsT , RegelOderGleichungsT);

#define PQ_INSERT(tp, we, actualParent, otherParent) ( (* PQ_InsertProc) (tp, we, actualParent, otherParent))
/* Das aus den uebergebenen Daten bestehende KP in die PQ einfuegen. */

#define PQ_INSERTImmortal(tp, we, actualParent, otherParent) \
( (* PQ_InsertImmortalProc) (tp, we, actualParent, otherParent))
/* Das aus den uebergebenen Daten bestehende KP in die PQ einfuegen. */

#define PQ_POSTPONE(tp, we, ActualParent, OtherParent) ( (* PQ_PostponeProc) (tp, we, ActualParent, OtherParent))
/* */

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Select ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern BOOLEAN (* PQ_SelectInitialProc)      (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *, void **); 
/* void* fuer addRef */

#define PQ_SELECTINITIAL(tpPtr, ActualParentPtr, OtherParentPtr, AddRefPtr)\
( (* PQ_SelectInitialProc) (tpPtr, ActualParentPtr, OtherParentPtr, AddRefPtr) )
/* Rueckgabe der Daten des minimalen KPs. Wurde AddRef nicht verwendet, so ist AddRefPtr==NULL. */


extern BOOLEAN (* PQ_SelectProc)      (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *, void **); 
/* void* fuer addRef */

#define PQ_SELECT(tpPtr, ActualParentPtr, OtherParentPtr, AddRefPtr)\
( (* PQ_SelectProc) (tpPtr, ActualParentPtr, OtherParentPtr, AddRefPtr) )
/* Rueckgabe der Daten des minimalen KPs. Wurde AddRef nicht verwendet, so ist AddRefPtr==NULL. */


#if COMP_WMTW

extern BOOLEAN (*PQ_OriginalSelectProc)(TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *, void **); 

#define /*void*/ PQ_SELECT2INITIAL(/*void*/) /* Select-Funktion temporaer umbiegen, so dass nur initiale KPs */\
  PQ_OriginalSelectProc = PQ_SelectProc;     /* ausgewaehlt werden */\
  PQ_SelectProc = PQ_SelectInitialProc;

#define /*void*/ PQ_SELECT2NORMAL(/*void*/) /* Umbiegen rueckgaengig machen */\
  PQ_SelectProc = PQ_OriginalSelectProc;

#endif


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Parents ---------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
extern void    (* PQ_NewParentProc)   (RegelOderGleichungsT);
extern void    (* PQ_KillParentProc)  (RegelOderGleichungsT);
extern void    (* PQ_RedirectParentProc) (void);
extern void    (* PQ_FinishParentProc) (void);

#define PQ_NEWPARENT(re) ( (* PQ_NewParentProc) (re))
/* Referenz (bei Regeln), Refernzen (bei Gleichungen) in re (bzw re und Gegenrichtung eintragen) */

#define PQ_KILLPARENT(re) ( (* PQ_KillParentProc) (re))
/* Referenzen loeschen. */

#define PQ_RedirectParent() ( (* PQ_RedirectParentProc) ())
/* Gegenrichtung verwenden. */

#define PQ_FinishParent() ( (* PQ_FinishParentProc) ())
/* KPBildung zum aktuellen Elternteil abgeschlossen. */

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Walkthrough -----------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
extern void    (* PQ_WalkthroughProc) (PQ_WalkthroughProcedureT, BOOLEAN);
extern void    (* PQ_WatchingWalkthroughProc) (PQ_WatchingWalkthroughProcedureT);
extern void    (* PQ_TermVerfolgungsProc) (void (*) (TermT, TermT));
extern unsigned long int (* PQ_AnzKPsToGoProc) (void);

#define PQ_WALKTHROUGH(ApplicationProcedure, NeedsTerms) ( (* PQ_WalkthroughProc) (ApplicationProcedure, NeedsTerms))
/* ApplicationProcedure auf alle gueltigen KPs anwenden und je nach Rueckgabewert weiterbehandeln.
   Je nach NeedsTerms werden die Termpaare entpackt.
*/

#define PQ_AnzKPsToGo() ( (* PQ_AnzKPsToGoProc) ())

#define PQ_WatchingWalkthrough(WP) ( (* PQ_WatchingWalkthroughProc) (WP))

#define PQ_TermeVerfolgen(OperationAufTermen) ( (*PQ_TermVerfolgungsProc) (OperationAufTermen))

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Statistics ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
extern void    (* PQ_StatisticsProc)  (InfoStringT);
extern void   (* PQ_ActSizeProc) (unsigned long int *, unsigned long int *);

void PQ_AktuelleMengenGroesse(unsigned long int *AktKPZahl, unsigned long int *AktGueltigeKPZahl);

#define PQ_AktuelleMengenGroesse(/*unsigned long int **/ AktKPZahlPtr,          \
                                 /*unsigned long int**/ AktGueltigeKPZahlPtr)   \
((* PQ_ActSizeProc) (AktKPZahlPtr, AktGueltigeKPZahlPtr))

#define PQ_Statistics(Degree) ( (* PQ_StatisticsProc) (Degree))
/* Mitgerechnete Statistik ausgeben auf IO_Strom. */


/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ CLEANING WOMAN ---------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* PQ_CleanupProc) (void);
extern void (* PQ_SoftCleanupProc) (void);

#define PQ_Cleanup() ( (* PQ_CleanupProc)())
#define PQ_SoftCleanup() ( (* PQ_SoftCleanupProc)())


#ifdef __cplusplus
}
#endif

#endif /* PQ_H */
