/*=================================================================================================================
  =                                                                                                               =
  =  RUndEVerwaltung.h                                                                                            =
  =                                                                                                               =
  =  Verwaltung der Regel- und der Gleichungsmenge                                                                =
  =                                                                                                               =
  =                                                                                                               =
  =  22.03.94  Thomas.                                                                                            =
  =  29.06.94  Arnim. Operationen III. eingebaut und teilweise bestehende Operationen geaendert:                  =
  =            RE_???EinfuegenMitRueckgabe eingebaut, RE_???Entfernen geaendert, RE_Loesche???menge geaendert     =
  =            RE_IstRegel, RE_DatenHolenObjektLoeschen eingebaut.                                                =
  =                                                                                                               =
  =                                                                                                               =
  =================================================================================================================*/

#ifndef RUNDEVERWALTUNG_H
#define RUNDEVERWALTUNG_H


#include "DSBaumKnoten.h"
#include "general.h"
#include "InfoStrings.h"
#include "TermOperationen.h"
#include "Termpaare.h"

#ifdef __cplusplus
extern "C" {
#endif

/*=================================================================================================================*/
/*= I. Testfunktion und bedingter Testbetrieb                                                                     =*/
/*=================================================================================================================*/

void RE_RUndETest(void);                                              /* Interne Konsistenzpruefung beider Mengen. */


/*=================================================================================================================*/
/*= II. Regelmenge                                                                                                =*/
/*=================================================================================================================*/

/* 0. Typen und Objekte ------------------------------------------------------------------------------------------ */

#define RegelOderGleichungsT GNTermpaarT

#define RegelT GNTermpaarT

extern DSBaumT RE_Regelbaum;


/* 1. Erzeugen von Regeln aus Termpaaren ------------------------------------------------------------------------- */

RegelT RE_ErzeugteRegel(TermT links, TermT rechts, unsigned long Nummer, PCL_ObjektNummerT *PclID);


/* 2. Praedikate und nebenwirkungsfreie Funktionen --------------------------------------------------------------- */


#define /*BOOLEAN*/ RE_RegelmengeLeer(/*void*/)                                                                     \
  (!RE_Regelbaum.Wurzel)

#define /*unsigned int*/ RE_maxTiefeRegel(/*void*/)                                                                 \
  (RE_Regelbaum.maxTiefeLinks)



/* 3. Einfuegeoperation ------------------------------------------------------------------------------------------ */

void RE_ZuletztErzeugteRegelEinfuegen(void);


/* 4. Entfernungs- und Loeschoperationen ------------------------------------------------------------------------- */
                                                     /* Das Loeschen der beteiligten Terme muss explizit erfolgen! */

void RE_RegelEntfernen(RegelT Regel); /* aus Baum entfernen, weder wegwerfen noch als geloescht markieren! (->KPV) */

void RE_RegelLoeschen(RegelT Regel);                   /* Regel als Objekt loeschen sowie Meldung an KP-Verwaltung
                                                          Terme nicht loeschen*/
void RE_LoescheRegelmenge(void);


/* 5. Ausgabeoperation ------------------------------------------------------------------------------------------- */

void RE_AusgabeRegelmenge(int Grad);
/* Je nach Einstellung auch graphische Ausgabe:
   1 - Textuell
   2 - Baum
   3 - Baum mit Grafik
   4 - Textuell ausfuehrlich (hier wie 1)
   5 - Baumdump
   6 - Baumdump mit Grafik */


/* 6. Schleifen-Konstrukte --------------------------------------------------------------------------------------- */

#define RE_forRegelnRobust(/*RegelT*/ RegelIndex, /*Block*/ Action)                                                 \
   BK_forRegelnRobust(RegelIndex, RE_Regelbaum, Action)

  
/* 7. Umbenennungen ---------------------------------------------------------------------------------------------- */

void RE_RegelExternalisieren(RegelT Regel);
/* Ist Regel ein Element des Regelbaumes, so nimmt die Funktion eine Umbenennung der Variablen vor, 
   so dass die resultierenden Variablen nicht im Baum vorkommen. 
   Kann auch fuer einzelne Richtungen von Gleichungen verwendet werden. */

void RE_RegelReinternalisieren(RegelT Regel);
/* Stellt nach dem Aufruf von RE_RegelExternalisieren den status quo ante wieder her, allerdings nur
   solange, wie keine Regel oder Gleichung mit neuer maximaler Variablenanzahl eingefuegt worden ist. */



/*=================================================================================================================*/
/*= III. Gleichungsmenge                                                                                          =*/
/*=================================================================================================================*/

/* 0. Typen und Objekte ------------------------------------------------------------------------------------------ */

extern DSBaumT RE_Gleichungsbaum;
#define GleichungsT GNTermpaarT


/* 1. Erzeugen von Gleichungen aus Termpaaren -------------------------------------------------------------------- */

GleichungsT RE_ErzeugteGleichung(TermT links, TermT rechts, unsigned long Nummer, PCL_ObjektNummerT *PclID);


/* 2. Praedikate und nebenwirkungsfreie Funktionen --------------------------------------------------------------- */

#define /*BOOLEAN*/ RE_GleichungsmengeLeer(/*void*/)                                                                \
  (!RE_Gleichungsbaum.Wurzel)

#define /*BOOLEAN*/ RE_GleichungsmengeNichtLeer(/*void*/)                                                           \
  (RE_Gleichungsbaum.Wurzel)

#define /*unsigned int*/ RE_maxTiefeGleichung(/*void*/)                                                             \
  (RE_Gleichungsbaum.maxTiefeLinks)

unsigned int RE_GradVonPermPaar(TermT lhs, TermT rhs);
/* Ist lhs = rhs eine permutative Gleichung?
   Grad = 0 gdw Gleichung ist keine Permgleichung. */


/* 3. Einfuegeoperation ------------------------------------------------------------------------------------------ */

void RE_ZuletztErzeugteHinrichtungEinfuegen(KopfZellenT *Referenz);

void RE_ZuletztErzeugteHerrichtungEinfuegen(void);


/* 4. Entfernungs- und Loeschoperationen ------------------------------------------------------------------------- */
                                                     /* Das Loeschen der beteiligten Terme muss explizit erfolgen! */

void RE_GleichungEntfernen(GleichungsT Gleichung);
/* Nur eine Richtung der Gleichung aus dem Baum entfernen. 
   ACHTUNG: Gegenrichtung muss entfernt sein, bevor erneut mit Gleichungsbaum reduziert wird. */

void RE_GleichungLoeschen(GleichungsT Gleichung);  /* Gleichung als Objekt loeschen sowie Meldung an KP-Verwaltung.
       Subgleichungen loeschen & Meldung an NFBildung, falls es sich um ausgezeichnete Seite der Gleichung handelt.*/

void RE_LoescheGleichungsmenge(void);


/* 5. Ausgabeoperationen ----------------------------------------------------------------------------------------- */

void RE_AusgabeGleichungsmenge(int Grad);
/* Je nach Einstellung auch graphische Ausgabe:
   1 - Textuell
   2 - Baum
   3 - Baum mit Grafik
   4 - Textuell ausfuehrlich
   5 - Baumdump
   6 - Baumdump mit Grafik
*/

/* 6. Schleifen-Konstrukte --------------------------------------------------------------------------------------- */

#define RE_forGleichungenRobust(/*GleichungsT*/ GleichungsIndex, /*Block*/ Action)                                  \
   BK_forRegelnRobust(GleichungsIndex, RE_Gleichungsbaum, Action)

#define RE_forGMReferenzen(/*Testproc*/ testen, /*LoeschProc*/ loeschen)\
   BK_ReferenzDurchlauf(RE_Gleichungsbaum, testen, loeschen)
/* Auf jede Referenz-Gleichung jedes Blattes im E-Baum wird die Test-Prozedur testen angewendet.
   Liefert diese TRUE, so wird loeschen auf alle Gleichungen des Blattes angewendet.
*/



/*=================================================================================================================*/
/*= IV. Ergebnisausgabe R-, E- und KP-Menge                                                                       =*/
/*=================================================================================================================*/

void RE_SonstigeErgebnisse(BOOLEAN flag);
/* flag dient zur Detailsteuerung in den verschiedenen Modi:
   convergence: TRUE  = im TP-Puffer befinden sich divergente KPs
                FALSE = im TP-Puffer befinden sich unrichtbare Ausgangsgleichungen
*/

#if COMP_PCL
void RE_DruckeEndsystemFuerPCL(void);
#else
#define RE_DruckeEndsystemFuerPCL()                                                                     /* nix tun */
#endif



/*=================================================================================================================*/
/*= V. Regeln und Gleichungen als Eltern von KPs                                                                  =*/
/*=================================================================================================================*/

/* Prozeduren, die zum Loesen, Weiterbehandeln und Loeschen von Regeln und Gleichungen waehrend der Interreduktion
   benoetigt werden, und zwar unabhaengig davon, ob es sich um Regeln oder Gleichungen handelt. Deshalb einheit-
   liche Sichtweise auf 'Objekte'.                                                                                 */

#define /*BOOLEAN*/ RE_IstRegel(/*RegelOderGleichungsT*/ objekt) /* Liefert TRUE, wenn das angegebene Objekt eine */\
  TP_TermpaarIstRegel(objekt)                                                                         /* Regel ist */



/*=================================================================================================================*/
/*= VI. Initialisierungen                                                                                         =*/
/*=================================================================================================================*/

void RE_ReInitApriori(void);
void RE_InitApriori(void);

/*=================================================================================================================*/
/*= VII. Statistik                                                                                                =*/
/*=================================================================================================================*/

extern unsigned long int RE_AnzRegelnGesamt, RE_AnzGleichungenGesamt;

#define RE_AnzRegelnGesamt()              RE_AnzRegelnGesamt 
#define RE_AnzGleichungenGesamt()         RE_AnzGleichungenGesamt

#define RE_IncAnzRegelnGesamt() (RE_AnzRegelnGesamt++) 
#define RE_IncAnzGleichungenGesamt() (RE_AnzGleichungenGesamt++)  

void RE_Statistik(InfoStringT IS);

void RE_AktuelleGroessen(unsigned long int *Regelmenge, unsigned long int *Gleichungsmenge);


#ifdef __cplusplus
}
#endif

#endif
