/**********************************************************************
* Filename : StandardHeuristiken.c
*
**********************************************************************/

#ifndef STANDARDHEURISTIKEN_H
#define STANDARDHEURISTIKEN_H

#ifdef __cplusplus
extern "C" {
#endif
   
#include "TermOperationen.h"
#include "SymbolOperationen.h"
#include "general.h"
#include "SUEBasis.h"
#include "InfoStrings.h"
#include "RUndEVerwaltung.h"


/* 
   Ein uebergeordnetes Modul wird in einer Tabelle die fuer die drei KP-Typen und fuer Killer-KPs 
   noetigen Funktionen aufrufen. 
   
   Bei Reclassify ist nur dann eine Unterscheidung noetig, falls es sich um KillerKps handelt.
   An dieser Stelle ist auch ein Abschalten der dynamischen Neubewertung moeglich.
   
   Evtl. sollte man die die Funktionen dieses Moduls in ihrer Parametrisierung soweit erweitern,
   dass sie das vorgegebene Schema fuer Classify-Routinen erfuellen und direkt ueber Funktionspointer
   nach aussen gereicht werden koennen sollten.

*/


/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Initialize ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

BOOLEAN STD_INITIALIZE(InfoStringT DEF, InfoStringT USR, int *WeightEntryType);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Classify --------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


extern void (* STD_0)(WeightEntryT *, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT, UTermT);

extern void (* STD_50) (WeightEntryT *, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT, UTermT);

extern void (* STD_100) (WeightEntryT *, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT, UTermT);

extern void (* STD_200) (WeightEntryT *, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT, UTermT);

extern void (* STD_pen)(WeightEntryT *, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT, UTermT);

extern void (* STD_Prepare) (void);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- ReClassify ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

int STD_0_Re(WeightEntryT *, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT);

extern int (* STD_50_Re) (WeightEntryT *, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT);

extern int (* STD_100_Re) (WeightEntryT *, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT);

extern int (* STD_200_Re) (WeightEntryT *, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT);

int STD_pen_Re(WeightEntryT *, TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Statistics ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void STD_Statistics(InfoStringT StatisticsIS);


void STD_WeightStatisticsStart(InfoStringT StatisticsIS);

extern void (* STD_WeightStatProc) (WeightEntryT wt, TermpairT tp, RegelOderGleichungsT actualP, 
                             RegelOderGleichungsT otherP);

void STD_WeightStatisticsEnd(void);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Output ----------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* STD_PrintProc) (WeightEntryT *, RegelOderGleichungsT, RegelOderGleichungsT);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ CLEANING WOMAN ---------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void STD_CleanUp(void);

#ifdef __cplusplus
}
#endif

#endif /* STANDARDHEURISTIKEN_H */
