#ifdef __cplusplus
extern "C" {
#endif

#ifndef SUBGLEICHUNGEN_H
#define SUBGLEICHUNGEN_H
/*=================================================================================================================
  =                                                                                                               =
  = Subgleichungen.h                                                                                              =
  =                                                                                                               =
  = Realisiereung einer Menge von Subgleichungen                                                                  =
  =                                                                                                               =
  = 28.07.94 Angelegt. Arnim.                                                                                     =
  =                                                                                                               =
  =                                                                                                               =
  =                                                                                                               =
  =                                                                                                               =
  =================================================================================================================*/

#include "TermOperationen.h"
#include "general.h"

typedef struct KopfZellenTS KopfZellenT;

#define SGM_ReferenzT KopfZellenT *
#define SGM_BetroffenesObjektT void *

/* ----------------- Initialisieren ------------------------*/

void SGM_InitAposteriori(void);
/* Nur aufrufen, wenn der externe Modus proof oder completion ist.*/

/* ----------------- SG's einfuegen und loeschen -----------*/

#define SGM_InitRef() NULL
/* Referenz zum Initialisieren. Wird die Loeschroutine mit dieser Referenz aufgerufen, so wird keine
   Taetigkeit ausgeloest, es wird auch kein Fehler produziert.!*/

void SGM_LoescheSubgleichungen(SGM_ReferenzT Referenz);
/* Loescht die zu einem Objekt eingetragenen Subgleichungen. Diese Prozedur muss immer dann aufgerufen werden,
   wenn eine solches Objekt geloescht wird, und zwar unabhaengig davon, ob die Referenz==NULL ist oder nicht.
   Die Subgleichungen zu einem Objekt werden in IR_GMSubsummiertGleichung erzeugt und mit
   SGM_SubgleichungenZuObjektAufnehmen in die Subgleichungsmenge aufgenommen. Dabei ist zu beachten, dass die
   hierbei zurueckgegebene Referenz diejenige ist, die beim Loeschen mit SGM_LoescheSubgleichungen angegeben werden
   muss.*/

SGM_ReferenzT SGM_SubgleichungenZuObjektAufnehmen(SGM_BetroffenesObjektT objekt);
/* Die im Puffer vorhandenen Gleichungen werdenn nun mit Verweis auf objekt (Gleichung) in die SGM aufgenommen.
   Die zurueckgegebene Referenz ist zur Gleichung zu speichern und beim Loeschen der Subgleichungen anzugeben.*/

/* ----------------- Durchlauf der SGM ---------------------*/

void SGM_StarteDurchlauf(void);
/* Durchlauf initialisieren */

void SGM_NaechsteSubgleichung(UTermT *links, UTermT *rechts, SGM_ReferenzT *objekt);
/* Naechste zu behandelnde Subgleichung holen; links und rechts enthalten nach dem Aufruf die linke und
   rechte Seite dieser Subgleichung.
   Vorbed.: Es gibt noch zu behandelnde SGs.*/

void SGM_NaechsteGleichung(void);
/* Schaltet beim Durchlauf auf die Subgleichungen der naechsten Gleichung, falls sie existieren.
   Wird verwendet, wenn die aktuelle Subgleichung subsummiert werden konnte und deshalb die uebrigen Subgleichungen
   der aktuellen Gleichung nicht mehr betrachtet werden muessen.*/

BOOLEAN SGM_NochSubgleichungen(void);
/* Liefert TRUE, wenn es in diesem Durchlauf noch unbehandelte SGs gibt.*/

/* Bemerkung: es ist nicht moeglich, die Subgleichungen zur aktuellen SG sofort zu loeschen, da dies ohnehin
   beim Loeschen des betroffenen Objekts passiert. Allerdings muss dies der SGM mit der folgenden Prozedur
   gemeldet werden, um eine geordnete Fortsetzung des Durchlaufs zu ermoeglichen.*/

void SGM_DruckeSGM(void);
/* Zum Testen */


#endif

#ifdef __cplusplus
}
#endif
