/*=================================================================================================================
  =                                                                                                               =
  =  SymbolOperationen.h                                                                                          =
  =                                                                                                               =
  =  In diesem Modul sind die Operationen auf Symbolen, d.h. auf Variablen-,                                      =
  =  Konstanten- und Funktions-Symbolen, realisiert.                                                              =
  =                                                                                                               =
  =  23.02.94 Angelegt.                                                                                           =
  =  02.03.94 Weitergemacht. Getestet.                                                                            =
  =  02.03.94 Mit einigen Beispielen getestet, SO_TestAusgaben eingebaut, LAEUFT soweit.                          =
  =  16.03.94 Neue Schleifen-Makros eingebaut.                                                                    =
  =                                                                                                               =
  =================================================================================================================*/

#ifndef SYMBOLOPERATIONEN_H
#define SYMBOLOPERATIONEN_H


#include "general.h"
#include "InfoStrings.h"
#include "parser.h"

#ifdef __cplusplus
extern "C" {
#endif


/*=================================================================================================================*/
/*= I. Typdefinitionen                                                                                            =*/
/*=================================================================================================================*/

typedef signed   short int SymbolT;

typedef unsigned short int StelligkeitsT;



/*=================================================================================================================*/
/*= II. Globale Variablen mit Zugriffsmakros                                                                      =*/
/*=================================================================================================================*/

/* Signaturinformationen */
extern SymbolT        SO_GroesstesESymb,
                      SO_GroesstesSpezSymb,
                      SO_GroesstesFktSymb;
extern unsigned int   SO_AnzahlFunktionen;
extern StelligkeitsT  *SO_Stelligkeit;
extern unsigned int   SO_MaximaleStelligkeit;

/* Ordnungsinformationen */
extern BOOLEAN        SO_PraezedenzTotal;
extern VergleichsT    **SO_Praezedenz;
extern SymbolT        *SO_TotaleKette;                 /* fallend angeordnet; Indizierung 0..SO_AnzahlFunktionen-1 */
extern weight         *SO_FunktionsGewichte;

/* Semantische Informationen */
extern BOOLEAN        *SO_IstACSymbol;
extern BOOLEAN        SO_Paramodulation;
extern SymbolT        SO_EQSymbol,
                      SO_TRUESymbol,
                      SO_FALSESymbol;

#define SO_KleinstesFktSymb()    0                                           /* Liefert das erste Funktions-Symbol */
#define SO_GroesstesFktSymb()    SO_GroesstesFktSymb/* Liefert das groesste Funktions-Symbol ggf. mit Paramod-Fkt. */
#define SO_ErstesVarSymb()       -1                                                  /* Liefert die erste Variable */
#define SO_VorErstemVarSymb()    0                           /* Ergebnis von SO_VorigesVarSymb(SO_ErstesVarSymb()) */
#define SO_AnzahlFunktionen()    SO_AnzahlFunktionen
#define SO_Stelligkeit(symbol)   (SO_Stelligkeit[symbol])                              /* SymbolT -> StelligkeitsT */
#define SO_MaximaleStelligkeit() SO_MaximaleStelligkeit                               /* 0, falls keine Operatoren */

#define SO_PraezedenzTotal()   SO_PraezedenzTotal
#define SO_SymbGewicht(symbol) (SO_FunktionsGewichte[symbol])                              /* SymbolT -> GewichtsT */

#define SO_IstACSymbol(Symbol) (SO_IstACSymbol[Symbol])                                      /* SymbolT -> BOOLEAN */
#define SO_Paramodulation()    SO_Paramodulation
#define SO_EQSymbol()          SO_EQSymbol
#define SO_TRUESymbol()        SO_TRUESymbol
#define SO_FALSESymbol()       SO_FALSESymbol


/* Arrangement der Funktionssymbole nach Normierung:
   SO_KleinstesFktSymb() <= SO_GroesstesESymb <= SO_GroesstesSpezSymb <= SO_GroesstesFktSymb() */


/*=================================================================================================================*/
/*= III. Weitere elementare Praedikate und Funktionen; Iterationsmakros                                           =*/
/*=================================================================================================================*/

#define SO_SymbIstFkt(symbol)  ((symbol) >= 0)                                               /* SymbolT -> BOOLEAN */
#define SO_SymbIstVar(symbol) ((symbol) < 0)                                                 /* SymbolT -> BOOLEAN */

#define SO_NaechstesVarSymb(var) (var-1)                                          /* Liefert die naechste Variable */
#define SO_VorigesVarSymb(var) (var+1)                                              /* Liefert die vorige Variable */
#define SO_NaechstesVarSymbD(var) (--var)                              /* Liefert destruktiv die naechste Variable */
#define SO_VorigesVarSymbD(var) (++var)                                  /* Liefert destruktiv die vorige Variable */

#define SO_SymbVergleich(symbol1, symbol2) SO_Praezedenz[symbol1][symbol2]       /* SymbolT SymbolT -> VergleichsT */
#define SO_SymbGleich(symbol1, symbol2) ((symbol1)==(symbol2))                       /* SymbolT SymbolT -> BOOLEAN */
#define SO_SymbUngleich(symbol1, symbol2) ((symbol1)!=(symbol2))                     /* SymbolT SymbolT -> BOOLEAN */
#define SO_SymbGroesser(symbol1, symbol2) (SO_Praezedenz[symbol1][symbol2]==Groesser)/* SymbolT SymbolT -> BOOLEAN */
#define SO_VariableIstNeuer(x1,x2) ((x1)<(x2))
#define SO_VariableIstNichtNeuer(x1,x2) ((x1)>=(x2))

#define SO_VarNummer(var)    ((unsigned)(-(var)))           /* SymbolT -> Nat; liefert die Nummer einer Variablen. */
#define SO_NummerVar(nummer) (-((signed)(nummer)))              /* Nat -> SymbolT; liefert soundsovielte Variable. */

#define SO_forSymboleAufwaerts(Index,MinSymb,MaxSymb) \
  for(Index=(MinSymb);Index<=(MaxSymb);Index++)

#define SO_forSymboleAbwaerts(Index,MaxSymb,MinSymb) \
  for(Index=(MaxSymb);Index>=(MinSymb);Index--)

#define SO_forFktSymbole(Index) \
  SO_forSymboleAufwaerts(Index,SO_KleinstesFktSymb(),SO_GroesstesFktSymb())                     /* alle Operatoren */

#define SO_forSpezSymbole(Index) \
  SO_forSymboleAufwaerts(Index,SO_KleinstesFktSymb(),SO_GroesstesSpezSymb)/* alle Operatoren ausser Paramodulation */

#define SO_forEquationSymbole(Index) \
  SO_forSymboleAufwaerts(Index,SO_KleinstesFktSymb(),SO_GroesstesESymb)           /* alle Operatoren aus EQUATIONS */

#define SO_forConclusionSymbole(Index) \
  SO_forSymboleAufwaerts(Index,SO_GroesstesESymb+1,SO_GroesstesSpezSymb)         /* alle Operatoren aus CONCLUSION */
                                                                                               /* sowie unbenutzte */
#define SO_forParaSymbole(Index) \
  SO_forSymboleAufwaerts(Index,SO_GroesstesSpezSymb+1,SO_GroesstesFktSymb())                     /* falls benutzt, */
                                                                                 /* alle Paramodulationsoperatoren */
#define SO_forKonstSymbole(Index) \
  SO_forFktSymbole(Index) \
    if (!SO_Stelligkeit(Index))                                          /* von allen Operatoren die nullstelligen */

#define SO_forUnkonstSymbole(Index) \
  SO_forFktSymbole(Index) \
    if (SO_Stelligkeit(Index))                                     /* von allen Operatoren die nicht nullstelligen */



/*=================================================================================================================*/
/*= IV. Symbolindizierte Felder                                                                                   =*/
/*=================================================================================================================*/

SymbolT SO_AusAltMachNeu(func_symb fkt);
/* Liefert das dem FunktionsSymbol fkt intern zugeordnete Symbol, so moeglich; sonst our_assertion_failed. */

void SO_VariablenfeldAnpassen(SymbolT NeuesteBaumvariable);                  /* Anpassen des Feldes SO_Stelligkeit */



/*=================================================================================================================*/
/*= VI. Symbolinformationen und initiale Praezedenz holen                                                         =*/
/*=================================================================================================================*/

void SO_InitAposteriori(void);
/* Besetzt die noetigen Strukturen wie Praezedenz, Stelligkeit, Gewichte etc. mit den Daten aus der Spezifikation */



/*=================================================================================================================*/
/*= VII. Praezedenz aus Kommandozeile lesen und auswerten                                                         =*/
/*=================================================================================================================*/

BOOLEAN SO_PraezedenzKorrekt(char *ps);

BOOLEAN SO_PraezedenzErneuert(char *ps);



/*=================================================================================================================*/
/*= VIII. Testausgaben, Cleanup und Statistik                                                                     =*/
/*=================================================================================================================*/

void SO_TestAusgaben(void);
/* Gibt die eingelesenen Daten am Bildschirm aus. */

void SO_CleanUp(void);

void SO_Statistik(InfoStringT IS);



#ifdef __cplusplus
}
#endif

#endif
