/**********************************************************************
* Filename : Zielverwaltung.h
*
* Kurzbeschreibung : "Mittelschicht" Zielverwaltung,
*                    also gemeinsame Schnittstelle von NormaleZiele und
*                    MNF_Ziele nach oben
*
* Autoren : Andreas Jaeger
*
* 10.07.96: Erstellung aus der urspruenglichen Zielverwaltung heraus
**********************************************************************/

#ifndef ZIELVERWALTUNG_H
#define ZIELVERWALTUNG_H


#include "TermOperationen.h"
#include "InfoStrings.h"
#include "RUndEVerwaltung.h"

#ifdef __cplusplus
extern "C" {
#endif


typedef void (*Z_BeobachtungsDurchlaufProcT) (TermT, TermT, unsigned long);

typedef void (*Z_BeobachtungsProzT) (TermT, TermT);

typedef enum {
   Z_NichtDefiniert, Z_NormaleZiele, Z_MNF_Ziele, Z_Paramodulation
} Z_ZielbehandlungsModulT;

extern Z_ZielbehandlungsModulT Z_ZielModul;

void Z_ZieleAusSpezifikationHolen(void);

void Z_ZielEinfuegen(unsigned long int ZielIndex, TermT lhs, TermT rhs);

BOOLEAN Z_ZielmengeLeer(void);

BOOLEAN Z_ZieleBewiesen(void);

/* Zielmengen-Durchlauf */
/* -------------------- */

void Z_ZielmengenBeobachtungsDurchlauf(Z_BeobachtungsDurchlaufProcT AP);
/* Wendet AP auf jedes Ziel an, wobei AP die Ziele nicht veraendert.*/

void Z_ZielBeobachtungsProzAnmelden(Z_BeobachtungsProzT proz);
/* Anmeldung einer Prozedur proz, die aufgerufen wird, sobald sich das 1. Ziel aendert.
   Wurde das erste Ziel bewiesen, so wird proz mit dem naechsten Ziel aufgerufen.
   proz wird insbesondere auch dann aufgerufen, sobald das erste Ziel eingefuegt wird.
   
   Aenderungen werden nur innerhalb der Prozedur Z_ZielmengenDurchlauf mitgeteilt,
   die uebrigen Durchlaufmoeglichkeiten rufen proz nicht auf!
*/

/* Zielmengen-Ausgabe */       
/* ------------------ */       

typedef enum  {Z_AusgabeAlles, Z_AusgabeInitial, Z_AusgabeSpez} Z_AusgabeArtT;

void Z_AusgabeZielmenge (Z_AusgabeArtT Z_AusgabeArt);


/*=================================================================================================================*/
/*=  Zielbehandlung                                                                                               =*/
/*=================================================================================================================*/

void Z_Statistik(InfoStringT str);

BOOLEAN Z_ZieleErledigtMitNeuerRegel(RegelT Regel);

BOOLEAN Z_ZieleErledigtMitNeuerGleichung(GleichungsT Gleichung);

BOOLEAN Z_FinalerZielDurchlauf(void);

/* Rueckgabe der letzten drei Funktionen ist TRUE, wenn gemaess Parametrisierung
   eines oder alle Ziele bewiesen wurden. */

void Z_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR);

BOOLEAN  Z_InitAposteriori(InfoStringT DEF, InfoStringT USR);
   
void Z_Cleanup (void);
   

#ifdef __cplusplus
}
#endif

#endif /* ZIELVERWALTUNG_H */
