/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - pcl.h                                                                                                      */
/* - *****                                                                                                      */
/* -                                                                                                            */
/* - Enthaelt alle fuer den PCL-Modus notwendigen Routinen, Zaehler etc..                                       */
/* -                                                                                                            */
/* - ACHTUNG: Das Abklemmen des PCL-Modus (ueber die Makerules) bewirkt, dass - aehnlich wie im Modul           */
/* -          Ausgaben.h die Makros - die Funktionen auf 'Nix tun'-Makros  umgestellt werden. Es ist also       */
/* -          keine weitere Aenderung erforderlich. Ferner finden auch alle Abfragen innerhalb dieser Funktionen*/
/* -          statt, so dass an den betreffenden Stellen im Code nur der jeweilige Makro-Aufruf erscheint.      */
/* -                                                                                                            */
/* - 30.11.94 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - Diese Datei enthaelt die verschiedenen Ausgaberoutinen, die zum Erstellen eines PCL-Protokolls noetig sind.*/
/* - Wird beim Aufruf von WaldmeisterII bei der Option -pcl ein Dateiname angegeben, so wird diese Datei in     */
/* - PCL_InitAposteriori geoeffnet. Am Programmende wird diese Datei dann mit PCL_Beenden wieder geschlossen.   */
/* - Aufrufe der einzelnen Ausgaberotinen sind im Code von WMII an den entsprechenden Stellen zu finden. Da aber*/
/* - bei jedem Aufruf der Ausgaberoutinen eine Abfrage (PCL-Modus eingestellt?) faellig wird, gibt es die Moeg- */
/* - lichkeit, ueber Makerules mit -DCOMP_PCL=0 (automatisch immer ausser bei gmake pcl) die Funktionsdeklara-  */
/* - tionen in diesem Modul durch 'nichtstuende' Makros zu ersetzen (s.u.). Dies fuehrt dazu, dass WMII ohne Ge-*/
/* - schwindigkeitsverlust laufen kann. In diesem Falle werden auch die PCL-IDs aus den verschiedenen Daten-    */
/* - strukturen gestrichen. Da die Makro-Aufrufe aber vom Praeprozessor ersetzt werden, stoeren die fuer den    */
/* - Compiler dann ja unbekannten Komponenten 'PCL-ID' nicht weiter. Deshalb muessen die PCL-IDs in den ver-    */
/* - schiedenen Datenstrukturen durch #ifdef eingeklammert werden.                                              */
/* - Beim Aufruf der Ausgaberoutinen ist jeweils ein VAR-Parameter nr anzugeben. Dieser wird dann auf die als   */
/* - naechstes zu vergebende PCL-ID gesetzt, ist also nicht weiter anzupassen.                                  */
/* - Zu Beachten ist noch, dass Zuweisungen von PCL-IDs (zB von einer Datenstruktur in die naechste) nur ueber  */
/* - das zu diesem Zweck vorgesehene Makro ablaufen, da sonst im 'rauskompilierten Falle' der Compiler die      */
/* - Komponenten gar nicht kennt, dieses Makro aber ebenso durch 'nix tun' ersetzt wird und so keine Probleme   */
/* - bereiten duerfte.                                                                                          */
/* - Der Einbau der PCL-IDs in die einzelnen Datenstrukturen wurde wie folgt realisiert:                        */
/* - - es wurde eine Komponente PclNr eingefuegt                                                                */
/* - - diese Komponente kann mit bedingter Compilierung (COMP_PCL=0) wieder entfernt werden                     */
/* - - ferner wird eine Zugriffsfunktion angeboten, die die Adresse (!) der Komponente PclNr liefert, um sie    */
/* -   in pcl.c manipulieren zu koennen. Diese Zugriffsfunktion wird beim rauuscompilieren durch ein 0-Makro    */
/* -   ersetzt.                                                                                                 */
/* -                                                                                                            */
/* - Einige Erweiterungen fuer integrated PCL                                                                   */
/* - Bernd, 15.05.97                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#ifndef PCL_H
#define PCL_H

#if 0
#if COMP_PCL == 1
#undef COMP_PCL
#define COMP_PCL 2
#endif
#endif

#include <limits.h>  /* fuer USHRT_MAX */

#include "TermOperationen.h"
#include "general.h"
#include "RUndEVerwaltung.h"
#include "InfoStrings.h"

#ifdef __cplusplus
extern "C" {
#endif


/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Typ-Definitionen ----------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

/*typedef unsigned StellenT;
#define MaxStelle UINT_MAX*/

                                /* oder: */

typedef unsigned short StellenT;
#define MaxStelle USHRT_MAX

/* Falls ein Ueberlauf von StellenT auftritt:
typedef unsigned long StellenT;
#define MaxStelle ULONG_MAX
*/

#if COMP_PCL >= 1

/* ----------------------------------------------------------------------------------------------------------------------- */
/* -- iPCL --------------------------------------------------------------------------------------------------------------- */
/* ----------------------------------------------------------------------------------------------------------------------- */

typedef enum {
  iPCL_init, iPCL_hypo, iPCL_red, iPCL_cp, iPCL_orient, iPCL_inst, iPCL_proved, iPCL_final
} iPCL_EntryType;

#define iPCL_EntryTypeSize 8

typedef enum {
  iPCL_eqn, iPCL_rule, iPCL_goal, iPCL_finalr
} iPCL_ResultType;

typedef unsigned long misctp;

/* todo: bitfield Ansatz:
typedef struct {
   unsigned int pos   :23;
   unsigned int pside : 1;
   unsigned int side  : 1;
   unsigned int used  : 1;
   unsigned int etp   : 4;
} misctp;
*/

typedef union {
    struct {
      misctp misc;
      PCL_ObjektNummerT id1, id2;
    } id;
    struct {
      misctp misc;
      TermT p1, p2;
    } p;
} iPCL_Rec;

/* The contents of the fields in the different cases:
 * misc.et | id1                | id2                     | misc.pos          | misc.side
 * --------+--------------------+-------------------------+-------------------+----------------------------
 * init:   | pointer to lhs     | pointer to rhs          | n.a.              | n.a.
 * hypo:   | pointer to lhs     | pointer to rhs          | n.a.              | n.a.
 * red:    | id of reduced eqn  | id of reducing eqn      | where reduced     | L/R of reducing eqn matches
 * cp:     | id of father       | id of mother (toplevel) | where overlapped  | L/R of mother for overlap
 * orient: | id of oriented eqn | n.a.                    | n.a.              | l->r or r->l orientation
 * inst:   | id of subsumed eqn | id of subsuming eqn     | (where subsumed)  | (LR or RL subsumption)
 * proved: | id of proved eqn   | n.a.                    | n.a.              | n.a.
 * final:  | id of final eqn    | n.a.                    | n.a.              | n.a.
 *
 * (..) == n.a. at the moment
 *
 * misc is a packed record wich contains
 * - an EntryType   --> T  
 * - a position     --> P  for 3 ETs (postprocessing: the new PCL-ID)
 * - a side         --> S  for 4 ETs
 * - a used bit     --> U  set by by post-processing and the ETs: iPCL_inst, iPCL_proved, and iPCL_final
 * - some free bits --> F  for further extension
 * 
 * assuming 32 Bit ulongs, the structure of misc is:
 *  3         2         1
 * 10987654321098765432109876543210 Bit
 * PPPPPPPPPPPPPPPPPPPPPPPPFFUSTTTT
 *                        ^-- L/R of position
 *
 * in the postprocessing phase, id1 & id2 contain pointer to lhs/rhs
 * the misc field is (I == new Id, R == result type):
 *  3         2         1
 * 10987654321098765432109876543210 Bit
 * IIIIIIIIIIIIIIIIIIIIIIIIFFFFFFRR
 */

#define iPCL_GetEntryType(ipr) 		   	 ((ipr).id.misc & 0xF)
#define iPCL_GetSide(ipr)      		   	 (((ipr).id.misc >> 4) & 1)
#define iPCL_GetPos(ipr)       		   	 ((ipr).id.misc >> 9)
#define iPCL_GetPSide(ipr)     		   	 (((ipr).id.misc >> 8) & 1)
#define iPCL_ComposeMisc(et,pos,pside,side)      ((pos) << 9 | ((pside &1) << 8) | (side & 1) << 4 | (et & 0xF))
#define iPCL_IsUsed(ipr)       		   	 ((ipr).id.misc & (1 << 5))
#define iPCL_SetUsed(ipr)      		   	 ((ipr).id.misc |= (1 << 5))
#define iPCL_StoreNewID(ipr,id)            	 ((ipr).p.misc = (((ipr).p.misc & 0xFF) | (id << 8)))
#define iPCL_GetNewId(ipr)                 	 ((ipr).p.misc >> 8)
#define iPCL_SetResultType(ipr,tp)         	 ((ipr).p.misc = (((ipr).p.misc & 0xFFFFFF00) | (tp & 3)))
#define iPCL_GetResultType(ipr)                  ((ipr).p.misc & 3)

/*
#define iPCL_RecordPerformed(tp) iPCL_Performed[tp]++
#define iPCL_RecordRecorded(tp)  iPCL_Recorded[tp]++
#define iPCL_RecordUsed(tp)      iPCL_Used[tp]++
*/
#define iPCL_RecordPerformed(tp) 
#define iPCL_RecordRecorded(tp)  
#define iPCL_RecordUsed(tp)      

#endif

#if COMP_PCL == 2
extern iPCL_Rec *iPCL_Memo;
extern int iPCL_doIt, iPCL_doItReally;
extern long iPCL_lastId, iPCL_AktPclID;
extern PCL_ObjektNummerT *iPCL_AktuelleID;
extern PCL_SeitenT iPCL_AktuelleSeite;

extern unsigned long iPCL_Performed[iPCL_EntryTypeSize], iPCL_Recorded[iPCL_EntryTypeSize], iPCL_Used[iPCL_EntryTypeSize];
#endif

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Die PCL-Ausgabe-Prozeduren /-Makros ---------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#if COMP_PCL == 1

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Initialisierung/Abschluss des Protokolls ----------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void PCL_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR);

void pcltest(TermT term);

void PCL_InitAposteriori(InfoStringT DEF, InfoStringT USR);
/* PCL-Protokoll-Datei oeffnen, Funktionspointer besetzen.
   Parametrisierung:
   PCL=Dateiname oder uPCL=Dateiname oder uPCLQuelle=Datei1:uPCLZiel=Datei1
*/

void PCL_Beenden(void);
/* PCL-Protokoll-Datei wieder schliessen.*/

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Gleichung/Ziel aus der Spezifikation --------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* PCLProc_GleichungAusDerSpezifikation) (TermT, TermT, PCL_ObjektNummerT *);
#define PCL_GleichungAusDerSpezifikation(links, rechts, nrPtr)\
((* PCLProc_GleichungAusDerSpezifikation)(links, rechts, nrPtr))

/* Die Gleichungen aus der Spezifiaktion werden beim Einfuegen in die KP-Menge mit Gewicht 0 als 'initial' ins PCL-Protokoll (PP)
   geschrieben. Die Nummer wird mit einer neuen PCL-ID besetzt.
<pcl-id> : tes-eqn : links = rechts : initial
*/

extern void (* PCLProc_ZielAusDerSpezifikation) (TermT, TermT, PCL_ObjektNummerT *);
#define PCL_ZielAusDerSpezifikation(links, rechts, nrPtr)\
      ((* PCLProc_ZielAusDerSpezifikation)(links, rechts, nrPtr))
/* Einlesen der Ziele aus der Spezifikation:
   Die Ziele werden bei der Aufnahme in die interne Datenstruktur als tes-goal gemeldet.
   Die Nummer wird mit einer neuen PCL-ID besetzt.
<pcl-id> : tes-goal : links = rechts : hypothesis
*/

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------------- REDUKTION -----------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* PCLProc_Reduktionsbeginn) (PCL_ObjektT, PCL_ObjektNummerT *, PCL_SeitenT, TermT);
#define PCL_Reduktionsbeginn(objektArt, nrPtr, seite, andereSeite)\
      ((* PCLProc_Reduktionsbeginn)(objektArt, nrPtr, seite, andereSeite))

/* Zu Beginn der Normalformbildung wird das zu behandelnde Objekt angemeldet.
   Die nr wird bei jedem Reduktionsschrittauf die neue PCL-ID gesetzt.*/

extern void (* PCLProc_Reduktion) (TermT, UTermT, PCL_ObjektNummerT *, PCL_SeitenT);
#define PCL_Reduktion(ergebnis, betroffenerTeilterm,AngewandtesFaktIDPtr, angewandteSeite)\
      ((* PCLProc_Reduktion) (ergebnis, betroffenerTeilterm,AngewandtesFaktIDPtr, angewandteSeite))
/* Reduktion:
   Jeder Reduktionsschritt wird mit betroffenem <fact>, angewandter <rule>/<eqn>, betroffener Seite+Stelle sowie angewandter Seite
   ins PPP geschrieben. 
   Aus dem betroffenen Teilterm werden Tiefe sowie Stelle berechnet (spart deren Berechnung bei per Option abgeschaltetem
   PCL-Protokoll.
<pcl-id> : <step-type> : links = rechts : tes-red( <pcl-expr>, <place>, <pcl-expr>, <side>)
  ^Nummer   ^Art des entst.  ^Ergebnis              ^Nr. des red. ^Pos.     ^Nr der  ^Richtung der Anwendung(l->r bzw. r->l)
             Faktums          der Red.               Faktums       der Red.  angewandten Regel/Gleichung
*/

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Unifikation ---------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* PCLProc_Unifikation) (PCL_ObjektNummerT *, TermT, TermT, RegelOderGleichungsT,
                     TermT/* ggf umbenannt!!! */, UTermT, RegelOderGleichungsT);
#define PCL_Unifikation(nrPtr, links, rechts, Vater, betroffenerVaterTerm, betroffenerVaterTeilterm, Mutter)\
      ((* PCLProc_Unifikation) (nrPtr, links, rechts, Vater, betroffenerVaterTerm, betroffenerVaterTeilterm, Mutter))

/* Unifikation:
   Jede gelungene Unifikation wird als <cp-expr> ins PP geschrieben: beide Elternteile (<rule> und/oder <eqn>) sowie genaue Beschreibung
   der Position. Die nr wird mit einer neuen PCL-ID besetzt.
   Dabei beschreibt Tiefe die Anzahl von Eintraegen im Stellen-Array
   (dh. Tiefe = 0 bedeutet Stelle lambda).
<pcl-id> : tes-eqn : links=rechts : cp(Vater, <place>, Mutter, <side>)
*/

extern void (* PCLProc_ReUnifikation) (PCL_ObjektNummerT *, TermT, TermT, RegelOderGleichungsT, int,
                                RegelOderGleichungsT);
#define PCL_ReUnifikation(nrPtr, links, rechts, Vater, NummerBetroffeneVaterStelle, Mutter)\
      ((* PCLProc_ReUnifikation)(nrPtr, links, rechts, Vater, NummerBetroffeneVaterStelle, Mutter))

/* NummerBetroffeneVaterStelle ist die Nummer desjenigen Symbols in der linken Vater-Seite,
   das an der Stelle steht, wo die Ueberlappung stattfand.*/

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Aufnahme von KPs als Regeln oder Gleichungen ------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* PCLProc_KPZuRegel) (TermT, TermT, PCL_RichtungsRichtungT, PCL_ObjektNummerT *);
#define PCL_KPZuRegel(links, rechts, RichtungsRichtung, RegelIDPtr) \
      ((* PCLProc_KPZuRegel) (links, rechts, RichtungsRichtung, RegelIDPtr))

/* Orientierung:
   Die Orientierung von kritischen Paaren, also <eqn>'s, wird nur im Erfolgsfall ins PP gemeldet:
   <pcl-id> : tes-rule: links=rechts : orient(<pcl-id>,x|u)
   Die Pcl-ID der Regel enthaelt bereits diejenige des KPs und wird aktualisiert.
*/


/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Ziele subsummieren/zusammenfuehren ----------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* PCLProc_ZielSubsummiert) (PCL_ObjektNummerT *,
                         TermT, TermT,   /* linke und rechte Seite des subsummierten Zieles */
                         PCL_ObjektNummerT *);
#define PCL_ZielSubsummiert(subsummiertesZielIDPtr, links, rechts,subsummierendeGleichungIDPtr)\
      ((* PCLProc_ZielSubsummiert) (subsummiertesZielIDPtr, links, rechts,subsummierendeGleichungIDPtr))

/* Subsumption:
   Wurde ein Ziel erfolgreich subsummiert, so ist dies als 
   <pcl-id> : tes-final : links = rechts : instance(<pcl-expr>, <pcl-expr>)
                                           das subs. Ziel^           ^die subsummierende Gleichung
   *subsummiertesZiel enthaelt beim Aufruf die PCL-ID des subsummierten Zieles und wird dann mit der neuen PCL-ID besetzt.
*/

extern void (* PCLProc_ZielBewiesen) (PCL_ObjektNummerT *, 
                               TermT, TermT);   /* linke und rechte Seite des bewiesenen Zieles */
#define PCL_ZielBewiesen(bewiesenesZielIDPtr, links,rechts)\
      ((* PCLProc_ZielBewiesen) (bewiesenesZielIDPtr, links,rechts))

/* Beweis von Zielen:
   Konnte ein Ziel bewiesen, i.e. zusammengefuehrt, werden, so wird dies mit <pcl-id> : tes-final: links=rechts:<quote-expr> gemeldet,
   wobei die <quote-expr> gerade die vorhergehende PCL-ID des Ziels bezeichnet.
   <pcl-id> : tes-final: links=rechts: <quote-expr>
   *bewiesenesZiel enthaelt die PCL-ID des bewiesenen Zieles und wird dann auf die neue ID gesetzt.
*/

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Abschluss der Vervollstaendigung: alle Fakten ausgeben --------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* PCLProc_VVAbschlussFaktDrucken) (PCL_ObjektNummerT *, TermT, TermT, PCL_ObjektT);

#define PCL_VVAbschlussFaktDrucken(nrPtr, links, rechts, art)\
      ((* PCLProc_VVAbschlussFaktDrucken) (nrPtr, links, rechts, art))

/* Abschluss der Vervollstaendigung: Wurde ein (grund?, siehe PCL, S. 20) konvergentes Regelsystem erzielt, so
   sollen alle erzeugten Regeln und Gleichungen nochmal mit tes-final gemeldet werden.
   Dies soll hier geschehen. Mit jeder Regel und Gleichung ist diese Prozedur aufzurufen.
   Die neue PCL-ID wird an nr weitergegeben.*/

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- Zuweisung -----------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define PCL_IDZuweisung(/* *PCL_ObjektNummerT*/ ziel, /* PCL_ObjektNummerT */ quelle)\
 (*ziel = quelle)

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- CHAINING ------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* PCLProc_StartChain) (void);
#define PCL_StartChain() ((* PCLProc_StartChain) ())
/* Alle folgenden Schritte nur vorlaeufig protokollieren. */

extern void (* PCLProc_StartFeyChain) (void);
#define PCL_StartFeyChain() ((* PCLProc_StartFeyChain) ()) 
/* Alle folgenden Schritte gar nicht protokollieren. */

extern void (* PCLProc_SaveChain) (void);
#define PCL_SaveChain() ((* PCLProc_SaveChain) ())
/* Vorlaeufig protokollierte Schritte speichern. */

extern void (* PCLProc_DeleteChain) (void);
#define PCL_DeleteChain() ((* PCLProc_DeleteChain) ())
/* Vorlaeufig protokollierte Schritte loeschen. */


/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- CORRESPONDANCE ------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

extern void (* PCLProcNeueRegel) (PCL_ObjektNummerT *, long int);
#define PCL_NeueRegel(pclIDPtr, RegelNummer) ( (* PCLProcNeueRegel) (pclIDPtr, RegelNummer))
/* PclID und Regel-Nummer im Korrespondenz-Protokoll vermerken */

extern void (* PCLProcNeueGleichung) (PCL_ObjektNummerT *, long int);
#define PCL_NeueGleichung(pclIDPtr, GleichungsNummer) ( (* PCLProcNeueGleichung) (pclIDPtr, GleichungsNummer))
/* PclID und Gleichungs-Nummer im Korrespondenz-Protokoll vermerken */

extern void (* PCLProcRegelRechtsReduziert) (PCL_ObjektNummerT *, long int);
#define PCL_RegelRechtsReduziert(pclIDPtr, RegelNummer) ( (* PCLProcRegelRechtsReduziert)(pclIDPtr, RegelNummer))
/* Neue PclID und (alte) Regel-Nummer im Korrespondenz-Protokoll vermerken */


/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- INLINED iPCL --------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#elif COMP_PCL == 2    /* iPCL als inlined ! */

void PCL_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR);

void pcltest(TermT term);

void PCL_InitAposteriori(InfoStringT DEF, InfoStringT USR);
/* PCL-Protokoll-Datei oeffnen, Funktionspointer besetzen.
   Parametrisierung:
   PCL=Dateiname oder uPCL=Dateiname oder uPCLQuelle=Datei1:uPCLZiel=Datei1
*/

void PCL_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR);

void PCL_Beenden(void);
/* PCL-Protokoll-Datei wieder schliessen.*/

#define /*void*/ PCL_GleichungAusDerSpezifikation(/*TermT*/ links, /*TermT*/ rechts, /*PCL_ObjektNummerT**/ nr)	\
{														\
  if (iPCL_doIt){												\
    iPCL_Memo[iPCL_lastId].p.p1 = TO_Termkopie(links);								\
    iPCL_Memo[iPCL_lastId].p.p2 = TO_Termkopie(rechts);								\
    iPCL_Memo[iPCL_lastId].p.misc = iPCL_ComposeMisc(iPCL_init,0,0,0);						\
    *(nr) = iPCL_lastId;											\
    iPCL_lastId++;												\
    iPCL_RecordPerformed(iPCL_init);										\
  }														\
}

#define /*void*/ PCL_ZielAusDerSpezifikation(/*TermT*/ links, /*TermT*/ rechts, /*PCL_ObjektNummerT**/ nr)	\
{														\
  if (iPCL_doIt){												\
    iPCL_Memo[iPCL_lastId].p.p1 = TO_Termkopie(links);								\
    iPCL_Memo[iPCL_lastId].p.p2 = TO_Termkopie(rechts);								\
    iPCL_Memo[iPCL_lastId].p.misc = iPCL_ComposeMisc(iPCL_hypo,0,0,0);						\
    *(nr) = iPCL_lastId;											\
    iPCL_lastId++;												\
    iPCL_RecordPerformed(iPCL_hypo);										\
  }														\
}

#define /*void*/ PCL_Reduktionsbeginn(/*PCL_ObjekT*/ objektArt,	\
                          /*PCL_ObjektNummerT **/ nr,		\
                          /*PCL_SeitenT*/ seite,		\
                          /*TermT*/ andereSeite)		\
{								\
  if (iPCL_doIt){						\
    iPCL_AktuelleID = nr;					\
    iPCL_AktuelleSeite = seite;					\
  }								\
}

#define /*void*/ PCL_Reduktion(/*TermT*/ ergebnis,							\
                   /*UTermT*/ betroffenerTeilterm,							\
                   /*PCL_ObjektNummerT*/ AngewandtesFakt,						\
                   /*PCL_SeitenT*/ angewandteSeite) 							\
{													\
  if (iPCL_doIt){											\
    iPCL_Memo[iPCL_lastId].id.id1 = *iPCL_AktuelleID;							\
    iPCL_Memo[iPCL_lastId].id.id2 = *AngewandtesFakt;							\
    iPCL_Memo[iPCL_lastId].id.misc = iPCL_ComposeMisc(iPCL_red,						\
						      TO_StelleInTerm(betroffenerTeilterm,ergebnis),	\
						      iPCL_AktuelleSeite,				\
						      angewandteSeite);					\
    *iPCL_AktuelleID = iPCL_lastId;									\
    iPCL_lastId++;											\
    iPCL_RecordPerformed(iPCL_red);									\
  }													\
}

#define /*void*/ PCL_Unifikation(/*PCL_ObjektNummerT**/ nr,								\
                     /*TermT*/ links, /*TermT*/ rechts,									\
                     /*RegelOderGleichungsT*/ Vater,									\
                     /*TermT*/ betroffenerVaterTerm,									\
                     /*UTermT*/ betroffenerVaterTeilterm,								\
                     /*RegelOderGleichungsT*/ Mutter)									\
{															\
  if (iPCL_doIt){													\
    iPCL_Memo[iPCL_lastId].id.id1 = *(TP_PclID(Vater));									\
    iPCL_Memo[iPCL_lastId].id.id2 = *(TP_PclID(Mutter));								\
    iPCL_Memo[iPCL_lastId].id.misc = 											\
      iPCL_ComposeMisc(iPCL_cp,												\
		       TO_StelleInTerm(betroffenerVaterTeilterm,betroffenerVaterTerm),					\
		       RE_IstRegel(Vater) ? linkeSeite : TP_RichtungAusgezeichnet(Vater) ? linkeSeite : rechteSeite,	\
		       RE_IstRegel(Mutter)? linkeSeite : TP_RichtungAusgezeichnet(Mutter)? linkeSeite : rechteSeite);	\
    *nr = iPCL_lastId;													\
    iPCL_lastId++;													\
    iPCL_RecordPerformed(iPCL_cp);											\
  }															\
}

#define /*void*/ PCL_ReUnifikation(/*PCL_ObjektNummerT**/ nr,								\
                     /*TermT*/ links, /*TermT*/ rechts,									\
                     /*RegelOderGleichungsT*/ Vater,									\
                     /*int*/ NummerBetroffeneVaterStelle,								\
                     /*RegelOderGleichungsT*/ Mutter)									\
{															\
  if (iPCL_doIt){													\
    iPCL_Memo[iPCL_lastId].id.id1 = *(TP_PclID(Vater));									\
    iPCL_Memo[iPCL_lastId].id.id2 = *(TP_PclID(Mutter));								\
    iPCL_Memo[iPCL_lastId].id.misc = 											\
      iPCL_ComposeMisc(iPCL_cp,												\
		       NummerBetroffeneVaterStelle,									\
		       RE_IstRegel(Vater) ? linkeSeite : TP_RichtungAusgezeichnet(Vater) ? linkeSeite : rechteSeite,	\
		       RE_IstRegel(Mutter)? linkeSeite : TP_RichtungAusgezeichnet(Mutter)? linkeSeite : rechteSeite);	\
    *nr = iPCL_lastId;													\
    iPCL_lastId++;													\
    iPCL_RecordPerformed(iPCL_cp);											\
  }															\
}

#define /*void*/ PCL_KPZuRegel(/*TermT*/ links, /*TermT*/ rechts,				\
                               /*PCL_RichtungsRichtungT*/ RichtungsRichtung,			\
                               /*PCL_ObjektNummerT**/ RegelID) 					\
{												\
  if (iPCL_doIt){										\
    iPCL_Memo[iPCL_lastId].id.id1 = *RegelID;							\
    iPCL_Memo[iPCL_lastId].id.id2 = 0;								\
    iPCL_Memo[iPCL_lastId].id.misc = iPCL_ComposeMisc(iPCL_orient,0,0,RichtungsRichtung);	\
    *RegelID = iPCL_lastId;									\
    iPCL_lastId++;										\
    iPCL_RecordPerformed(iPCL_orient);								\
  }												\
}

#define /*void*/ PCL_ZielSubsummiert(/*PCL_ObjektNummerT**/ subsummiertesZiel,						\
                         /*TermT*/ links, /*TermT*/ rechts,   /* linke und rechte Seite des subsummierten Zieles */	\
                         /*PCL_ObjektNummerT**/ subsummierendeGleichung)						\
{															\
  if (iPCL_doIt){													\
    iPCL_Memo[iPCL_lastId].id.id1 = *subsummiertesZiel;									\
    iPCL_Memo[iPCL_lastId].id.id2 = *subsummierendeGleichung;								\
    iPCL_Memo[iPCL_lastId].id.misc = iPCL_ComposeMisc(iPCL_inst,0,0,0);							\
    iPCL_SetUsed(iPCL_Memo[iPCL_lastId]);										\
    *subsummiertesZiel = iPCL_lastId;											\
    iPCL_lastId++;													\
    iPCL_RecordPerformed(iPCL_inst);											\
  }															\
}

#define /*void*/ PCL_ZielBewiesen(/*PCL_ObjektNummerT**/ bewiesenesZiel,					\
                      /*TermT*/ links, /*TermT*/ rechts)   /* linke und rechte Seite des bewiesenen Zieles */	\
{														\
  if (iPCL_doIt){												\
    iPCL_Memo[iPCL_lastId].id.id1 = *bewiesenesZiel;								\
    iPCL_Memo[iPCL_lastId].id.id2 = 0;										\
    iPCL_Memo[iPCL_lastId].id.misc = iPCL_ComposeMisc(iPCL_proved,0,0,0);					\
    iPCL_SetUsed(iPCL_Memo[iPCL_lastId]);									\
    *bewiesenesZiel = iPCL_lastId;										\
    iPCL_lastId++;												\
    iPCL_RecordPerformed(iPCL_proved);										\
  }														\
}

#define /*void*/ PCL_VVAbschlussFaktDrucken(/*PCL_ObjektNummerT**/ nr, /*TermT*/ links, /*TermT*/ rechts, /*PCL_ObjektT*/ art)	\
{																\
  if (iPCL_doIt){														\
    iPCL_Memo[iPCL_lastId].id.id1 = *nr;											\
    iPCL_Memo[iPCL_lastId].id.id2 = 0;												\
    iPCL_Memo[iPCL_lastId].id.misc = iPCL_ComposeMisc(iPCL_final,0,0,0);							\
    iPCL_SetUsed(iPCL_Memo[iPCL_lastId]);											\
    *nr = iPCL_lastId;														\
    iPCL_lastId++;														\
    iPCL_RecordPerformed(iPCL_final);												\
  }																\
}


/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- CHAINING --------------------------------------------- --------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define PCL_StartChain()					\
{								\
  if (iPCL_doIt){						\
   iPCL_AktPclID = iPCL_lastId; /* Aktuelle Pcl-ID merken */	\
  }								\
}

#define PCL_SaveChain()						\
{								\
  if (iPCL_doItReally && !iPCL_doIt){ /* end of fey chain ?? */	\
   printf(">>>>> ERROR: SaveFeyChain\n");			\
  }								\
}

#define PCL_DeleteChain()				\
{							\
  if (iPCL_doIt){					\
    iPCL_lastId = iPCL_AktPclID;			\
  }							\
  else if (iPCL_doItReally){ /* end of fey chain ?? */	\
    iPCL_doIt = 1;					\
  }							\
}

#define PCL_StartFeyChain() 			\
{						\
   iPCL_doIt = 0;				\
}

#define PCL_IDZuweisung(/* *PCL_ObjektNummerT*/ ziel, /* PCL_ObjektNummerT */ quelle)\
 (*ziel = quelle)

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- CORRESPONDANCE ------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define PCL_NeueRegel(pclIDPtr, RegelNummer)
#define PCL_NeueGleichung(pclIDPtr, GleichungsNummer)
#define PCL_RegelRechtsReduziert(pclIDPtr, RegelNummer)

/*---------------------------------*/
/* Und jetzt die Nixtuenden Makros */
/*---------------------------------*/

#else    /* Der PCL-Protokoll-Modus ist komplett abgeklemmt. */

void PCL_InitAposteriori(InfoStringT DEF, InfoStringT USR);
/* Nix tun */
void PCL_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR);

#define /*void*/ PCL_Beenden(/*void*/) /* Nix tun */

#define /*void*/ PCL_GleichungAusDerSpezifikation(/*TermT*/ links, /*TermT*/ rechts, /*PCL_ObjektNummerT**/ nr) /* Nix tun */

#define /*void*/ PCL_ZielAusDerSpezifikation(/*TermT*/ links, /*TermT*/ rechts, /*PCL_ObjektNummerT**/ nr) /* Nix tun */

#define /*void*/ PCL_Reduktionsbeginn(/*PCL_ObjekT*/ objektArt, \
                          /*PCL_ObjektNummerT **/ nr, \
                          /*PCL_SeitenT*/ seite, \
                          /*TermT*/ andereSeite) /* Nix tun */

#define /*void*/ PCL_Reduktion(/*TermT*/ ergebnis, \
                   /*UTermT*/ betroffenerTeilterm, \
                   /*PCL_ObjektNummerT*/ AngewandtesFakt,  \
                   /*PCL_SeitenT*/ angewandteSeite) /* Nix tun */

#define /*void*/ PCL_Unifikation(/*PCL_ObjektNummerT**/ nr, \
                     /*TermT*/ links, /*TermT*/ rechts, \
                     /*RegelOderGleichungsT*/ Vater, \
                     /*TermT*/ betroffenerVaterTerm, \
                     /*UTermT*/ betroffenerVaterTeilterm, \
                     /*RegelOderGleichungsT*/ Mutter) /* Nix tun */

#define /*void*/ PCL_ReUnifikation(/*PCL_ObjektNummerT**/ nr, \
                     /*TermT*/ links, /*TermT*/ rechts, \
                     /*RegelOderGleichungsT*/ Vater, \
                     /*int*/ NummerBetroffeneVaterStelle, \
                     /*RegelOderGleichungsT*/ Mutter) /* Nix tun */

#define /*void*/ PCL_KPZuRegel(/*TermT*/ links, /*TermT*/ rechts, \
                               /*PCL_RichtungsRichtungT*/ RichtungsRichtung,\
                               /*PCL_ObjektNummerT**/ RegelID) /* Nix tun */

#define /*void*/ PCL_KPZuGleichung(/*PCL_ObjektNummerT**/ KpID, \
                                   /*PCL_ObjektNummerT**/ GleichungsID) /* Nix tun */

#define /*void*/ PCL_ZielSubsummiert(/*PCL_ObjektNummerT**/ subsummiertesZiel, \
                         /*TermT*/ links, /*TermT*/ rechts,   /* linke und rechte Seite des subsummierten Zieles */ \
                         /*PCL_ObjektNummerT**/ subsummierendeGleichung) /* Nix tun */

#define /*void*/ PCL_ZielBewiesen(/*PCL_ObjektNummerT**/ bewiesenesZiel, \
                      /*TermT*/ links, /*TermT*/ rechts) /* Nix tun */   /* linke und rechte Seite des bewiesenen Zieles */

#define /*void*/ PCL_VVAbschlussFaktDrucken(/*PCL_ObjektNummerT**/ nr, /*TermT*/ links, /*TermT*/ rechts, /*PCL_ObjektT*/ art) /* Nix tun */

#define /*PCL_ID*/ PCL_IDZuweisung(/*PCL_ObjektNummerT**/ ziel, /* PCL_ObjektNummerT* */ quelle) /* Nix tun */



/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- CHAINING --------------------------------------------- --------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define PCL_StartChain()
#define PCL_StartFeyChain() 
#define PCL_SaveChain()
#define PCL_DeleteChain()

/* -------------------------------------------------------------------------------------------------------------*/
/* --------------------------- CORRESPONDANCE ------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define PCL_NeueRegel(pclIDPtr, RegelNummer)
#define PCL_NeueGleichung(pclIDPtr, GleichungsNummer)
#define PCL_RegelRechtsReduziert(pclIDPtr, RegelNummer)

#endif  /* if COMP_PCL */




#ifdef __cplusplus
}
#endif

#endif  /* ifdef PCL_H */
