/*=================================================================================================================
  =                                                                                                               =
  =  DomaenenMemorium.c                                                                                           =
  =                                                                                                               =
  =  Kapselung mmap-Speicherzugriff                                                                               =
  =                                                                                                               =
  =  19.02.97 Thomas: aus StrahlMemorium.c ausgelagert.                                                           =
  =                                                                                                               =
  =================================================================================================================*/

/* 02.09.96: AJ: Signalbehandlung fuer SUNOS5 (=Solaris2) und Linux */

#ifdef LINUX
#define _GNU_SOURCE 1
#endif

#include "antiwarnings.h"
#include "Ausgaben.h"
#include "DomaenenMemorium.h"
#include <fcntl.h>
#include "FehlerBehandlung.h"
#include <signal.h>
#include "SpeicherVerwaltung.h"
#include <sys/mman.h>
#include <sys/types.h>
#include <unistd.h>

#if defined (LINUX)
# if !defined __GLIBC__
#  include <asm/sigcontext.h>
#  include <signal.h>
#  define sigcontext sigcontext_struct
   typedef __sighandler_t sighandler_t;
# endif
#elif defined (SUNOS5)
#include <sys/siginfo.h>
#elif defined (SUNOS4)
#else
#error "Signalbehandlung nur fuer SunOS4, SunOS5 und Linux definiert"
#endif


/* Bemerkung zur Implementierung:
   - AlterSignalbehandler kann den Wert SIG_DFL haben, in diesem Fall
     ist AlterSignalbehandler kein gueltiger Funktionsaufruf
   
   LINUX:  Um an die Adresse, die den Speicherfehler versucht hat, zu kommen, wurden die Kernel-Strukturen untersucht
           und der zusaetzliche Parameter sigcontext_struct lokalisiert. Dies funktioniert mit Kernel 2.0.17, aber
           muss nicht mit aktuelleren Kernels funktionieren.  Linux unterstuetzt bei mmap auch sogenanntes anonymous
           Mapping. Dazu muss kein File geoeffnet sein, aber das Flag MAP_ANON (manchmal auch MAP_ANONYMOUS)
           uebergegeben werden.

   SUNOS5: Solaris definiert bei sigaction zwei moegliche Funktionen: sa_handler (POSIX konform) und sa_sigaction
           (Solaris-spezifisch). sa_sigaction liefert die Informationen, die wir benoetigen.  */


/* -------------------------------------------------------*/
/* --------------- Globals -------------------------------*/
/* -------------------------------------------------------*/

size_t DM_Seitengroesse;

/*=================================================================================================================*/
/*= II. Initialisierung; Allgemeines                                                                              =*/
/*=================================================================================================================*/

#define EinViertelGByte     0x10000000
#define CharToLong(a,b,c,d) 256*(256*(256*a+b)+c)+d                            /* fuer Markierung unserer Domaenen */
#define WMIdentifikation    CharToLong('W','M','I','I')                 /* im Ggs. zu von anderen gemappten Seiten */

static volatile DomaenenT Domaenenliste;

#ifndef MAP_ANON
  static int Orkus;
#endif
#if   defined (SUNOS4)
 typedef void (*WM_SignalHandlerT)(int Signal, int Code, struct sigcontext *Kontext, char *Adresse);
 static void Kollisionsbehandlung(int Signal, int Code, struct sigcontext *Kontext, char *Adresse);
#elif defined (SUNOS5)
  static void Kollisionsbehandlung(int Signal, siginfo_t *Kontext, void *Adresse);
#elif defined (LINUX)
  typedef void (*WM_SignalHandlerT) (int Signal, struct sigcontext Kontext);
  static void Kollisionsbehandlung(int Signal, struct sigcontext Kontext);
#endif
static struct sigaction AlterSignalbehandler;


void DM_InitApriori(void)
{
  struct sigaction sigact;
  DM_Seitengroesse = getpagesize();
#if !defined (MAP_ANON)
  if ((Orkus = open("/dev/zero",O_RDWR))==-1)
    our_error("/dev/zero kann nicht geoeffnet werden.");
#endif
#if defined (SUNOS4)
  sigact.sa_handler = &Kollisionsbehandlung;
  sigact.sa_flags =  0;
#elif defined (SUNOS5)
  sigact.sa_handler = NULL;
  sigact.sa_sigaction  = &Kollisionsbehandlung;
  sigact.sa_flags =  SA_SIGINFO|SA_RESTART;
#elif defined (LINUX)
  sigact.sa_handler = (sighandler_t)&Kollisionsbehandlung;
  sigact.sa_flags =  SA_RESTART;
#endif                                   
  sigemptyset(&sigact.sa_mask);
  if (sigaction(SIGSEGV, &sigact, &AlterSignalbehandler))
    our_error("Signalbehandler fuer SIGSEGV kann nicht gesetzt werden.");
  Domaenenliste = NULL;
}


void DM_NieMehrDomaenen_SagNiemalsNie(void)
{
#ifndef MAP_ANON   
  if (close(Orkus))
    our_error("/dev/zero kann nicht geschlossen werden.");
#endif
}


/*=================================================================================================================*/
/*= III. Mapping, Unmapping, Seitenschutz                                                                         =*/
/*=================================================================================================================*/

void DM_mmap(void *Start, void *Ende)              /* Seiten [Start,Ende) holen. Kein Test, ob Ausdehnung erlaubt. */
{                                                                           /* Start und Ende sind Seitenadressen. */
  /* printf("mmap[%p,%p[\n",Start,Ende); */
  if (mmap(Start,(char*)Ende-(char*)Start,PROT_READ|PROT_WRITE,MAP_PRIVATE|MAP_FIXED          
#ifdef MAP_ANON
                                                                         |MAP_ANON,-1
#else
                                                                                  ,Orkus
#endif
                                                                                        ,0)==(caddr_t)-1)
    our_error("mmap kann nicht ausgefuehrt werden.");
}


void DM_munmap(void *Start, void *Ende)                                           /* Seiten [Start,Ende) weggeben. */
{                                                                           /* Start und Ende sind Seitenadressen. */
  if (munmap(Start,(size_t)((char*)Ende-(char*)Start)))
    our_error("munmap fehlgeschlagen");
}


typedef enum { 
  Gesperrt = PROT_NONE,
  Ungesperrt = PROT_READ|PROT_WRITE
} SchutzT;

static BOOLEAN our_mprotect(void *Ende, SchutzT Schutz, BOOLEAN Abbruch)            /* Fuer die Seite vor Ende Schutzbits */
{                                                                               /* setzen. Ende ist Seitenadresse. */
  if (mprotect((char*)Ende-DM_Seitengroesse(),DM_Seitengroesse(),Schutz)==-1) {                      /* Fehlerfall */
    if (Abbruch)
      our_error("mprotect kann nicht ausgefuehrt werden.");
    return FALSE;
  }
  else
    return TRUE;
}



/*=================================================================================================================*/
/*= IV. Domaenen                                                                                                  =*/
/*=================================================================================================================*/

void DM_AmEndeGesperrtAllokieren(DomaenenT Domaene, unsigned long int Delta, void *Hakendatum)
             /* Seiten [Ende,Ende+Delta) holen; zuvor jedoch testen, ob erlaubte Ausdehnung ueberschritten wuerde; */
{                       /* Domaenenadresse, Identifikation und Hakendatum in letzte Seite schreiben; diese sperren */
  void *NeuesEnde = (char*)Domaene->Ende+Delta;
  if ((unsigned long int)NeuesEnde > (unsigned long int)Domaene->Obergrenze)
    our_error("Domaene waechst ueber vorgesehenen Adressraum.");
  DM_mmap(Domaene->Ende,NeuesEnde);
  *((DomaenenT*)NeuesEnde-1) = Domaene;
  *((unsigned long int*)NeuesEnde-2) = WMIdentifikation;
  *((unsigned long int**)NeuesEnde-3) = Hakendatum;
  our_mprotect(NeuesEnde,Gesperrt,TRUE);
  Domaene->Ende = NeuesEnde;
}


static void LastRessortSignalHandler(void)  /* Wenn kein Signalbehandler gesetzt ist, hilft nur eins: raus! Nichts */
{       /* mehr ausgeben, da es sonst zu einer Endlosschleife kommen kann (auf Linux). Prozedur wurde ausgelagert, */
#if defined SUNOS4 || defined SUNOS5                               /* um einen Breakpoint darauf setzen zu koennen */
  fputs("Unrecoverable Segmentation Fault\n",stderr);                              /* mal austesten, ob das klappt */
#endif
  abort ();                                                                        /* weniger Aufwand als exit(-1) */
}

#define TestAlterSignalbehandler(Signalbehandler)         /* pruefen, ob ueberhaupt eigener Behandler gesetzt war */\
  if ((void *)Signalbehandler==(void *)SIG_DFL) {  /* Diese Abfrage wird vom Apogee bemeckert mit Warnung invalid */\
    LastRessortSignalHandler(); /* type conversion. Es ist aber die einzige Moeglichkeit, den GCC ruhigzustellen. */\
  }

#if defined (SUNOS4)
#define SonstigerFehler() (Code!=SEGV_PROT)
#define AlteBehandlung()                                                                                            \
{ TestAlterSignalbehandler(AlterSignalbehandler.sa_handler);                                                        \
  (*AlterSignalbehandler.sa_handler)(Signal,Code,Kontext,Adresse); }
#elif defined (SUNOS5)
#define SonstigerFehler() (Info->si_code!=SEGV_ACCERR)
#define AlteBehandlung()                                                                                            \
{ if ((AlterSignalbehandler.sa_flags) && SA_SIGINFO) {                                                              \
    TestAlterSignalbehandler(AlterSignalbehandler.sa_sigaction);                                                    \
    (*(AlterSignalbehandler.sa_sigaction))(Signal, Info, Kontext);                                                  \
  }                                                                                                                 \
  else {                                                                                                            \
    TestAlterSignalbehandler(AlterSignalbehandler.sa_handler);                                                      \
    (*(AlterSignalbehandler.sa_handler))(Signal);                                                                   \
  } }
#elif defined (LINUX)
#define SonstigerFehler() (Signal!=SIGSEGV)
#define AlteBehandlung()                                                                                            \
{ TestAlterSignalbehandler (AlterSignalbehandler.sa_handler);                                                       \
  (*(WM_SignalHandlerT)AlterSignalbehandler.sa_handler)(Signal,Kontext); }
#endif

static void Kollisionsbehandlung(int Signal, 
#if   defined (SUNOS4)
                                             int Code, struct sigcontext *Kontext, char *Adresse)
#elif defined (SUNOS5)    
                                             siginfo_t *Info, void *Kontext)
#elif defined (LINUX)    
                                             struct sigcontext Kontext)
#endif
{
  if (SonstigerFehler())
    AlteBehandlung()
  else {
#if   defined (SUNOS5)
    unsigned long Adresse = (unsigned long) Info->si_addr;
#elif defined (LINUX)
    unsigned long Adresse = Kontext.cr2;
#endif
    char *Kollision = (char*)((unsigned long int)Adresse & ~(DM_Seitengroesse()-1)) + DM_Seitengroesse();
    DomaenenT Domaene;
    if (!our_mprotect(Kollision,Ungesperrt,FALSE) ||   /* Kollisionsseite zum Lesen freigeben, sofern das moeglich */
      *((unsigned long int*)Kollision-2)!=WMIdentifikation)                                /* ist, sonst Ausstieg. */
      AlteBehandlung();                                          /* Dito, falls Seite keine WM-Identifikation hat. */
    Domaene = *((DomaenenT*)Kollision-1);
    (*Domaene->Haken)(Kollision);                                                        /* Domaenenhaken aufrufen */
  }
}


DomaenenT DM_NeueDomaene(DomaenenartenT Domaenenart, unsigned int Start, unsigned int Obergrenze,  
  DomaenenhakenT Haken, const char *Name)
{
  DomaenenT Domaene = our_alloc(sizeof(struct DomaenenTS));
  volatile DomaenenT Index = Domaenenliste;
  Domaene->Name = Name;
  if (Start>Obergrenze || Obergrenze>0xF)
    our_error("Ungueltige Adressraumangabe in DM_NeueDomaene.");
  Domaene->Start = Domaene->Ende = Domaene->Frei = (void*)(Start*EinViertelGByte);
  Domaene->Obergrenze =(void*)((Obergrenze+1)*EinViertelGByte);/* umsteigen auf Angabe als rechtsoffenes Intervall */
  while (Index) {
    if ((unsigned long int)Domaene->Start < (unsigned long int)Index->Obergrenze 
      && (unsigned long int)Domaene->Obergrenze > (unsigned long int)Index->Start)
      our_error("Der Adressraum einer neuer Domaene ueberschneidet sich mit dem einer alten.");
    Index = Index->Nachf;
  }
  if ((Domaene->Domaenenart = Domaenenart)!=Strahldomaene && Domaenenart!=Managerdomaene)
    our_error("Unbekannte Domaenenart.");
  Domaene->Haken = Haken;
  Domaene->Vorg = NULL;
  Domaene->Nachf = Domaenenliste;
  Domaenenliste = Domaene;
  if_sst(
    Domaene->AnzahlDomaenenausdehnungen = 
    Domaene->AnzahlGefroreneStrahlen = 
    Domaene->AnzahlGeloeschteStrahlen = 
    Domaene->AnzahlGeunmapptePages = 
    Domaene->AnzahlGemapptePages = 
    Domaene->AnzahlMaxGemapptePages = 0);
  return Domaene;
}


void DM_DomaeneLeeren(volatile DomaenenT Domaene)
{
  DM_munmap(Domaene->Start,Domaene->Ende);  /* Adressbereich freigeben. Nach munmap-Dokumentation (natuerliche In- */
  Domaene->Ende = Domaene->Frei =Domaene->Start;/* terpretation) muss nicht mehr der gesamte Bereich gemappt sein. */
  if_sst(
    Domaene->AnzahlDomaenenausdehnungen = 
    Domaene->AnzahlGefroreneStrahlen = 
    Domaene->AnzahlGeloeschteStrahlen = 
    Domaene->AnzahlGeunmapptePages =
    Domaene->AnzahlGemapptePages = 
    Domaene->AnzahlMaxGemapptePages = 0);
}


void DM_DomaeneLoeschen(volatile DomaenenT Domaene)
{
  DM_DomaeneLeeren(Domaene);                                                            /* Adressbereich freigeben */
  if (Domaene->Vorg)
    Domaene->Vorg->Nachf = Domaene->Nachf;
  else                                                                /* erstes Element der Domaenenliste entfernt */
    Domaenenliste = Domaene->Nachf;
  if (Domaene->Nachf)
    Domaene->Nachf->Vorg = Domaene->Vorg;
  our_dealloc(Domaene);
}



/*=================================================================================================================*/
/*= V. Statistik                                                                                                  =*/
/*=================================================================================================================*/

void DM_Statistik(InfoStringT IS)
{
  if (!(IS_IsIn("all", IS) || IS_IsIn("DM", IS)))
    return; 
  if (IS_IsIn("standard", IS)){
    volatile DomaenenT Domaene = Domaenenliste;
    while (Domaene) {
      IO_DruckeKurzUeberschrift(Domaene->Name,"+");
      IO_DruckeStatistikZeile(Domaene->Domaenenart==Strahldomaene ? "Mapped memory in bytes (beams)"
       : "Mapped memory in bytes (managers)",(unsigned long int)((char*)Domaene->Ende-(char*)Domaene->Start));
      if_sst(
        IO_DruckeStatistikZeile("Domain expansions",Domaene->AnzahlDomaenenausdehnungen);
        IO_DruckeStatistikLeerZeile();
        IO_DruckeStatistikZeile("Mapped pages", Domaene->AnzahlGemapptePages);
        IO_DruckeStatistikZeile("Unmapped pages", Domaene->AnzahlGeunmapptePages);
        IO_DruckeStatistikZeile("Max. mapped pages", Domaene->AnzahlMaxGemapptePages);
        IO_DruckeStatistikLeerZeile();
        IO_DruckeStatistikZeile("Mapped memory", Domaene->AnzahlGemapptePages * DM_Seitengroesse());
        IO_DruckeStatistikZeile("Unmapped memory", Domaene->AnzahlGeunmapptePages * DM_Seitengroesse());
        IO_DruckeStatistikZeile("Max. mapped memory", Domaene->AnzahlMaxGemapptePages * DM_Seitengroesse());
        switch (Domaene->Domaenenart) {
        case Strahldomaene:
          IO_DruckeStatistikZeile("Frozen beams", Domaene->AnzahlGefroreneStrahlen);
          IO_DruckeStatistikZeile("Removed beams", Domaene->AnzahlGeloeschteStrahlen);
          break;
        case Managerdomaene:
        default:
          break;
        }
      );
      Domaene = Domaene->Nachf;
    }
  }
  else if (IS_IsIn("latex-head", IS)){
    volatile DomaenenT Domaene = Domaenenliste;
    if (Domaene)
      fprintf(IO_Strom, "Domains: ");
    while (Domaene) {
      fprintf(IO_Strom, "%s : mapped mem (bytes) & ", Domaene->Name);
      if_sst(fprintf(IO_Strom, "expand & frozen & deleted & "));
      Domaene = Domaene->Nachf;
    }
  }
}
