
/* Datei: FehlerBehandlung.c
 * Zweck: Stellt verschiedene Behandlungsm"oglichkeiten f"ur Fehler zur Verf"ugung
 * Autor: Arnim Buch
 * Stand: 28.02.94
 * Ersterstellung: 28.02.94
 * Aenderungen:
 * 
 * 
 * 
 * 
 * 
 * 
 */

/* ACHTUNG !!!!!!!!!!!!!!

Um nach einem Fehler auf die Oberfl"ache zur"uckspringen zu k"onnen, ist vor(!) dem Aufruf von
create_base der aktuelle Zustand mit der Befehls-Sequenz

  if (setjmp(Environment_before_Computation_of_GB) == 0){   %R"ucksprung f"ur Fehler setzen. 
    create_base(...);
    }
  else{
    <Fehler ist aufgetreten>;  %hierhin nach Auftreten eines Fehlers gesprungen
  };

zu speichern. In der Fehler-Routine ist dann folgendes aufzurufen, um den Zustand vor dem Aufruf
 wiederherszustellen:

  longjmp (Environment_before_Computation_of_GB,1); % nach Fehler zur"uck auf Top-Level 

Genaueres findet Ihr im Anhang (ich  glaube, in Abschnitt B) vom Kerningham-Ritchie.
Diese Geschichten habe ich im Polynom-Parser benutzt, und dort funktionieren sie auch.
(Siehe parser.c ).

Ach ja, die Typen etc.:
#include <setjmp.h>   % ist die zu importierende Datei.

jmp_buf ist der Typ der Environment_before_... -Variablen.

So, damit mue/3te es eigentlich schon klappen.


Arnim, 21.02.94
*/


#include "FehlerBehandlung.h"
#include "general.h"
#include <stdio.h>
#include "antiwarnings.h"
#include "Ausgaben.h"
#include "Statistiken.h"           /* nach our_error wird die Statistik ausgegeben */
#include "SpeicherVerwaltung.h"
#include "RUndEVerwaltung.h"
#include "Zielverwaltung.h"
#include "ZeitMessung.h"
#include "SUEM.h"

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe Fehlerbehandlung.h)-------------------*/
/* ------------------------------------------------------------------------*/
unsigned long FB_Z[10];
BOOLEAN volatile FB_IRFlag;
BOOLEAN volatile FB_Usr1IRFlag;

/* ================================================================================= */

void our_assertion_failed(const char *hoppala)
{
  fprintf (stderr, "Annahme fehlgeschlagen :");
  fprintf (stderr, hoppala);
  exit(6);
}

/**************************************************************************************/

void our_warning(const char *autsch)
{
  fprintf (stderr, "Warnung :");
  fprintf (stderr, autsch);
}

/**************************************************************************************/

void our_error(const char *aua)
{
   ZM_StoppGesamtZeit();
   ST_ErrorStatistik(aua, FALSE);
   exit(4);
}

/**************************************************************************************/

void not_yet_implemented(const char *functionName)
{
  fprintf (stderr, "Noch nicht implementiert: %s", functionName);
  exit(5);
}

/**************************************************************************************/

void our_fatal_error(const char *aechhh)
{
  fprintf (stderr, "Fataler Fehler: ");
  fprintf (stderr, aechhh);
  fflush(stderr);
  exit(3);
}

/* -------------------------------------------------------------------------------*/
/* --------------- Fehlersuche und -beseitigung ----------------------------------*/
/* -------------------------------------------------------------------------------*/

/* Es werden drei Arten von Fehlertest bereitgestellt:
   1. Unbedingter Fehlertest.
   2. Von bis zu vier Zaehlern abhaengiger Fehlertest, der bei Erreichen/Ueberschreiten der zuvor entweder
      per Aufruf-Option oder durch expliziten Prozeduraufruf eingestellten Marken aktiviert wird.
      Die Zaehler koennen mit den vier angegebenen INC-Makros erhoeht werden. Der Fehlertest selbst
      ist NICHT Interrupt-Gesteuert, muss also an die entsprechenden Code-Stellen eingebaut werden.
      Werden die AutoInc-Zaehl-Makros verwendet, so erfolgt der Fehlertest sofort nach der Erhoehung
      des Zaehlers.
      Werden die Aufruf-Optionen zur Besetzung der Marken herangezogen, so wird dabei die zusaetzliche
      Option NoDebug herangezogen, bei deren Angabe kein Abbruch bei Erreichen der Marken erfolgt.
      Dies ermoeglicht bei eincompilierten Abbruch-Aufrufen normale Laeufe.
   3. Von Signal abhaengiger Fehlertest, dessen Aufruf zwar vom Setzen des Signals abhaengig,
      zeitlich aber davon entkoppelt ist.
   Alle Fehlertests loeschen saemtliche zugaengliche Systemstrukturen und aktivieren den Speichertest.
   Ferner wird die Anzahl der noch auf dem Zeitstack liegenden Zeiten angegeben.
   Nach Ablauf des Tests wird das Programm verlassen.
*/


unsigned long FB_M[10]; /* Stopp-Marker */

void FB_Fehlertest(void)
/* Nicht von Bedingungen abhaengiger Fehlertest */
{
   int i;
   fprintf(IO_Strom, "\n\n************** FEHLERTEST ************\n\n");
   fprintf(IO_Strom, "Aktuelle Zaehler: ");
   for(i=0; i<10; i++)
      fprintf(IO_Strom, "%ld, ", FB_Z[i]);
   fprintf(IO_Strom, "\nGesetzte Marke: ");
   for(i=0; i<10; i++)
      fprintf(IO_Strom, "%ld, ", FB_M[i]);
   fprintf(IO_Strom, "\n");
   IO_AusgabenAus();                                                   /* Verbliebene Strukturen loeschen */
   SUE_CLEANUP();                                                      /* muss vor dem Loeschen der R- bzw. E-Menge passieren */
   RE_LoescheGleichungsmenge();
   RE_LoescheRegelmenge();                                                      /* E- und R-Menge loeschen */
   TO_FeldVariablentermeLeeren();                                             /* ... */
   Z_Cleanup();                                                       /* Ziele loeschen */
   IO_AusgabenEin();
   
   SV_Statistik(FALSE, FALSE, "complete");                                                /* FALSE: auch die Termzellen werden getestet. */
   
   ZM_ZeitstackOK();                                                   /* Zeitstack pruefen */
   exit(0);
}

void FB_MarkeSetzen(unsigned long m0, unsigned long m1,unsigned long m2,unsigned long m3,unsigned long m4,
                    unsigned long m5,unsigned long m6,unsigned long m7,unsigned long m8,
                    unsigned long m9)
/* Loescht bei gesetztem Stopp-Signal alle zugreifbaren System-Strukturen,
   so dass im fehlerfreien Falle keine Diskrepanzen in der Speicherstatistik auftreten
   duerften.*/
{
   FB_M[1] = m1;
   FB_M[2] = m2;
   FB_M[3] = m3;
   FB_M[4] = m4;
   FB_M[5] = m5;
   FB_M[6] = m6;
   FB_M[7] = m7;
   FB_M[8] = m8;
   FB_M[9] = m9;
   FB_M[0] = m0;
}

void FB_DebugOptionAlsMarke(void)
/* Setzt die Stopp-Signale auf die Werte der Debug-Option 1,..4 */
{
   if (!PA_NoDebug()){
      if (PA_DebugOption1Angegeben()) FB_M[1] = atoi(PA_DebugOption1());
      if (PA_DebugOption2Angegeben()) FB_M[2] = atoi(PA_DebugOption2());
      if (PA_DebugOption3Angegeben()) FB_M[3] = atoi(PA_DebugOption3());
      if (PA_DebugOption4Angegeben()) FB_M[4] = atoi(PA_DebugOption4());
      if (PA_DebugOption5Angegeben()) FB_M[5] = atoi(PA_DebugOption5());
      if (PA_DebugOption6Angegeben()) FB_M[6] = atoi(PA_DebugOption6());
      if (PA_DebugOption7Angegeben()) FB_M[7] = atoi(PA_DebugOption7());
      if (PA_DebugOption8Angegeben()) FB_M[8] = atoi(PA_DebugOption8());
      if (PA_DebugOption9Angegeben()) FB_M[9] = atoi(PA_DebugOption9());
      if (PA_DebugOption0Angegeben()) FB_M[0] = atoi(PA_DebugOption0());
   }
}

void FB_FehlertestBeiErreichterMarke(void)
{
   int i;
   BOOLEAN erreicht = TRUE;
   for(i=0;i<10;i++)
      if (FB_Z[i]<FB_M[i]) erreicht = FALSE;
   if (erreicht && !PA_NoDebug())
      FB_Fehlertest();
}

BOOLEAN FB_MarkeErreicht(void)
/* Liefert TRUE, wenn die eingestellte Marke erreicht wurde und nicht NoDebug gewaehlt wurde. */
{
   int i;
   BOOLEAN erreicht = TRUE;
   for(i=0;i<10;i++)
      if (FB_Z[i]<FB_M[i]) erreicht = FALSE;
   return (erreicht && !PA_NoDebug());
}

BOOLEAN Signal = FALSE;

void FB_DruckeAktuelleZaehler(void)
/* Gibt die aktuellen Zaehlerstaende aus */
{
   int i;
   fprintf(IO_Strom, "Fehler-Zaehler: ");
   for(i=0; i<10; i++)
      fprintf(IO_Strom, "%ld, ", FB_Z[i]);
   fprintf(IO_Strom, "\n");
}

void FB_FehlerSignalLoeschen(void)
/* Initialisieren der Fehlermarke. */
{
   Signal = FALSE;
}

void FB_FehlerSignalSetzen(void)
/* Setzt ein Fehlerflag,... */
{
   Signal = TRUE;
}

void FB_FehlertestBeiSignal(void)
/* auf das diese Prozedur mit dem Abraeumen aller Strukturen und einem Speichertest reagiert.*/
{
   if (Signal)
      FB_Fehlertest();
}



