/***********************************************************************************************************************
 * Autor:      Juergen Sellentin, 13.02.94
 * Angepasst : Arnim, 04.03.94
 * Stackverwaltung zugunsten eines Arrays 'rausgeworfen. 11.04.94, Arnim.
 ***********************************************************************************************************************
 * 
 * Bemerkungen zur Zeitmessung:
 * ----------------------------
 *
 * Die Zeitmessungen werden mit der Systemfunktion "getrusage" realisiert. Diese protokolliert die seit dem Beginn des
 * Prozesses verbrauchte Zeit (als user- und system- Anteil). 
 * Die Genauigkeit dieser Messung haengt sehr stark von der update-Frequenz von getrusage ab !!!
 * Bei sehr kurzen Prozeduraufrufen werden die getrusage-Daten evtl. NICHT veraendert, so dass die Laufzeit 0
 * als Ergebnis herauskommt. Andererseits kann eine zu hohe Laufzeit gemessen werden, wenn das Aktualisieren der Daten
 * (von getrusage) gerade in einer sehr kurzen Prozedur durchgefuehrt wird.
 * Im statistischen Mittel gleichen sich diese Fehler hoffentlich aus!
 *
 ***********************************************************************************************************************/

#include "antiwarnings.h"
#include <sys/time.h>
#include <sys/resource.h>
#include <stdio.h>
#include <string.h>

#include "Ausgaben.h"
#include "FehlerBehandlung.h"
#include "InfoStrings.h"
#include "ZeitMessung.h"
#include "SpeicherVerwaltung.h"

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe ZeitMessung.h) -----------------------*/
/* ------------------------------------------------------------------------*/
#if ZM_TEST
struct timeval ZeitDummy;
#endif

struct timeval       ZM_GesamtZeit_ut;               /* insgesamt verbrauchte Zeit exklusive Parsen und Plattklopfen. */
struct timeval       ZM_GesamtZeit_st;

/* ... fuer : Reduktion ......................................................................................................*/
struct timeval       ZM_NFBildung_ut;                /* enthaelt alle Unter-Zeiten */
struct timeval       ZM_NFBildung_st;

struct timeval       ZM_MatchenRm_ut;                /* Matchen mit Regel- bzw. GleichungsMENGE (feiner wird nicht gemessen!) */
struct timeval       ZM_MatchenRm_st;
struct timeval       ZM_MatchenGm_ut;
struct timeval       ZM_MatchenGm_st;
struct timeval       ZM_TERedGm_ut;                  /* Ueberfluessige(!) Termerzeugung bei RE-Schritt */
struct timeval       ZM_TERedGm_st;
struct timeval       ZM_Rewrite_ut;
struct timeval       ZM_Rewrite_st;
/* ... fuer : Unifikation ......................................................................................................*/
struct timeval       ZM_KPBildungZuRegel_ut;         /* enthaelt alle Unter-Zeiten */
struct timeval       ZM_KPBildungZuRegel_st;
struct timeval       ZM_KPBildungZuGleichung_ut;
struct timeval       ZM_KPBildungZuGleichung_st;
struct timeval       ZM_MGUZuR_ut;                   /* Zeit fuer MGU-Suche mit neuer Regel */
struct timeval       ZM_MGUZuR_st;
struct timeval       ZM_MGUZuG_ut;                   /* Zeit fuer MGU-Suche mit neuer Gleichung */
struct timeval       ZM_MGUZuG_st;
struct timeval       ZM_Kapurtest_ut;                /* Ueberpruefung des Kapur-Kriteriums bei KP-Bildung */
struct timeval       ZM_Kapurtest_st;
/* ... fuer : Subsumption ......................................................................................................*/
struct timeval       ZM_Subsumption_ut;              /* Zeit fuer Subsumption */
struct timeval       ZM_Subsumption_st;
/* ... fuer : Interreduktion ...................................................................................................*/
struct timeval       ZM_Interred_ut;                 /* Zeit fuer Interreduktion der verschiedenen Mengen */
struct timeval       ZM_Interred_st;
/* ....fuer : Ordnungen........................................................................................................*/
struct timeval       ZM_OrdnungEntscheidung_ut;      /* Zeit, die in der Ordnung (Entsch) verbraucht wird.   */
struct timeval       ZM_OrdnungEntscheidung_st;
struct timeval       ZM_OrdnungRedTest_ut;           /* Zeit, die in der Ordnung (>-Test bei Red.) verbraucht wird.   */
struct timeval       ZM_OrdnungRedTest_st;
struct timeval       ZM_OrdnungUnifTest_ut;          /* Zeit, die in der Ordnung (>-Test bei Unif.) verbraucht wird.   */
struct timeval       ZM_OrdnungUnifTest_st;
/* ....fuer : Parsen und Plattklopfen .........................................................................................*/
struct timeval       ZM_PP_ut;                       /* user- / system- Zeit, die zum Parsen und Plattklopfen verbraucht wird. */
struct timeval       ZM_PP_st;
/* ... fuer : Regelmengenverwaltung ...........................................................................................*/
struct timeval       ZM_RegelEinf_ut;                /* user- / system- Zeit, die zum Einfuegen von Regeln verbraucht wird. */
struct timeval       ZM_RegelEinf_st;
struct timeval       ZM_RegelLoeschen_ut;            /* user- / system- Zeit, die zum Loeschen von Regeln verbraucht wird. */
struct timeval       ZM_RegelLoeschen_st;
/* ... fuer : Gleichungsmengenverwaltung ......................................................................................*/
struct timeval       ZM_GleichgEinf_ut;              /* user- / system- Zeit, die zum Einfuegen von Gleichungen verbraucht wird. */
struct timeval       ZM_GleichgEinf_st;
struct timeval       ZM_GleichgLoeschen_ut;          /* user- / system- Zeit, die zum Loeschen von Gleichungen verbraucht wird. */
struct timeval       ZM_GleichgLoeschen_st;
/* ... fuer : Kritische-Paar-Mengen-Verwaltung ................................................................................*/
struct timeval       ZM_KPEinf_ut;                   /* user- / system- Zeit, die zum Einfuegen von KP's verbraucht wird. */
struct timeval       ZM_KPEinf_st;
struct timeval       ZM_KPLoeschen_ut;               /* user- / system- Zeit, die zum Loeschen von KP's verbraucht wird. */
struct timeval       ZM_KPLoeschen_st;
struct timeval       ZM_KPEinpacken_ut;              /* user- / system- Zeit, die zum Einpacken von KP's verbraucht wird. */
struct timeval       ZM_KPEinpacken_st;
struct timeval       ZM_KPAuspacken_ut;              /* user- / system- Zeit, die zum Auspacken von gepackten KP's verbraucht wird. */
struct timeval       ZM_KPAuspacken_st;
struct timeval       ZM_GepackteKPsLoeschen_ut;      /* user- / system- Zeit, die zum Loeschen von gepackten KP's verbraucht wird. */
struct timeval       ZM_GepackteKPsLoeschen_st;
/* ... fuer : Speicherverwaltung  ................................................................................*/
struct timeval       ZM_NeuerSpeicherBlock_ut;       /* user- / system- Zeit zum Anlegen und Initialisierung neuer Speicherbloecke. */
struct timeval       ZM_NeuerSpeicherBlock_st;

/***********************************************************************************************************************
 * globale Prozeduren
 ***********************************************************************************************************************/

#define ZM_StackSize 10

unsigned long int AnzahlZeitFehler = 0;

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++++++++++++++++++++++++ Korrektheitstest ++++++++++++++++++++++++++++++++ */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#if ZM_TEST

static BOOLEAN ANGEHALTEN = FALSE;
static unsigned long angeh_Stackpos;


typedef struct TestStackTS{
  struct timeval *AdresseDerLaufendenUserZeit;
  char *Aufruf;
}TestStackZellenT;

static TestStackZellenT *TestStack;
static unsigned TestStackzeiger = 0;      /* Naechster freier Eintrag. */

static char *blubb = "|/-\\|/-\\";
static int modulo = 8;
static int zaehler = 0;

#define Blubb() printf("%c\r", blubb[zaehler = (zaehler+1)%modulo])

static void TestPush(struct timeval *user, char *Aufruf)
{
  TestStack[TestStackzeiger].AdresseDerLaufendenUserZeit = user;
  TestStack[TestStackzeiger++].Aufruf = Aufruf;

  if ((user == &ZM_KapurtestZuRegel_ut)&& (TestStack[TestStackzeiger-2].AdresseDerLaufendenUserZeit != &ZM_KPBildungZuRegel_ut)){
    printf("ZM_KapurtestZuRegel gestartet, Oberzeit NICHT ZM_KPBildungZuRegel, sondern %s\n", TestStack[TestStackzeiger-2].Aufruf);
    AnzahlZeitFehler++;
  }
  else if ((user == &ZM_KapurtestZuGleichung_ut) && (TestStack[TestStackzeiger-2].AdresseDerLaufendenUserZeit != &ZM_KPBildungZuGleichung_ut)){
    printf("ZM_KapurtestZuGleichung gestartet, Oberzeit NICHT ZM_KPBildungZuGleichung, sondern %s\n", TestStack[TestStackzeiger-2].Aufruf);
    AnzahlZeitFehler++;
  }
  if ((user == &ZM_MGUZuR_ut)&& (TestStack[TestStackzeiger-2].AdresseDerLaufendenUserZeit != &ZM_KPBildungZuRegel_ut)){
    printf("ZM_MGUZuR gestartet, Oberzeit NICHT ZM_KPBildungZuRegel, sondern %s\n", TestStack[TestStackzeiger-2].Aufruf);
    AnzahlZeitFehler++;
  }
  else if ((user == &ZM_MGUZuG_ut) && (TestStack[TestStackzeiger-2].AdresseDerLaufendenUserZeit != &ZM_KPBildungZuGleichung_ut)){
    printf("ZM_MGUZuG gestartet, Oberzeit NICHT ZM_KPBildungZuGleichung, sondern %s\n", TestStack[TestStackzeiger-2].Aufruf);
    AnzahlZeitFehler++;
  }
  else if ((user == &ZM_NFBildung_ut) && (TestStack[TestStackzeiger-2].AdresseDerLaufendenUserZeit == &ZM_KPBildungZuRegel_ut) && !ANGEHALTEN){
    printf("ZM_NFBildung gestartet, Oberzeit  ZM_KPBildungZuRegel\n");
    AnzahlZeitFehler++;
  }
  else if ((user == &ZM_NFBildung_ut) && (TestStack[TestStackzeiger-2].AdresseDerLaufendenUserZeit == &ZM_KPBildungZuGleichung_ut)&& !ANGEHALTEN){
    printf("ZM_NFBildung gestartet, Oberzeit  ZM_KPBildungZuGleichung\n");
    AnzahlZeitFehler++;
  }
}

static void TestPop(struct timeval *user, char *Aufruf)
{
  if (TestStack[TestStackzeiger-1].AdresseDerLaufendenUserZeit != user){ /* Fehler! */
    printf("\nFehler in der Zeitmessung bei TestPop!\n"
           "%s, %s\n\n", TestStack[TestStackzeiger-1].Aufruf, Aufruf);
    AnzahlZeitFehler++;
 }
  --TestStackzeiger;
}



void ZM_ZeitStartTest(struct timeval *user, char *Aufruf)
/* Die Adresse der angestossenen User-Zeit-Variablen sowie der Name des aufgerufenen Makros
   werden auf einem Stack gespeichert.
   */
{
   Blubb();
  TestPush(user, Aufruf);
}

void ZM_ZeitSwitchTest(struct timeval *Stoppuser, char *SwitchAufruf, struct timeval *Startuser)
/* Die Adresse der gestoppten User-Time wird mit derjenigen auf dem Test-Stack verglichen.
   Bei Unterschied werden der gespeicherte Start-Aufruf sowie der Stopp-Aufruf ausgegeben.
   Ferner wird ein Fehler-Zaehler erhoeht, der spaeter bei ZM_ZeitstackOK ueberprueft wird.
   
   Anschliessend wird der alte, oberste Stackeintrag durch Startuser und StartAufruf ersetzt.
   */
{
   Blubb();
  TestPop(Stoppuser, SwitchAufruf);
  TestPush(Startuser, SwitchAufruf);
}


void ZM_ZeitStoppTest(struct timeval *user, char *Aufruf)
/* Die Adresse der gestoppten User-Time wird mit derjenigen auf dem Test-Stack verglichen.
   Bei Unterschied werden der gespeicherte Start-Aufruf sowie der Stopp-Aufruf ausgegeben.
   Ferner wird ein Fehler-Zaehler erhoeht, der spaeter bei ZM_ZeitstackOK ueberprueft wird.
   */
{
   Blubb();
  TestPop(user, Aufruf);
}

int ZM_ZeitStartTestF(struct timeval *user, char *Aufruf)
{
   Blubb();
  TestPush(user, Aufruf);
  return 0;
}

int ZM_ZeitSwitchTestF(struct timeval *Stoppuser, char *SwitchAufruf, struct timeval *Startuser)
{
   Blubb();
  TestPop(Stoppuser, SwitchAufruf);
  TestPush(Startuser, SwitchAufruf);
  return 0;
}


int ZM_ZeitStoppTestF(struct timeval *user, char *Aufruf)
{
   Blubb();
  TestPop(user, Aufruf);
  return 0;
}


#endif
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++++++++++++++++++++++++ Zeitmessung +++++++++++++++++++++++++++++++++++++ */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

#if ZM_STATISTIK
#define if_ZM_ZAEHLEN(Action) Action
static unsigned long int gettimeCalls = 0;
static unsigned long int stackMax = 0;
#else
#define if_ZM_ZAEHLEN(Action) /* Nix tun */
#endif

#define AnzMikrosekProSek 1000000

static struct timezone timezonedummy; /* global wegen Initialisierung */

#define /*void*/ ZeitdifferenzAddieren(/*struct timeval*/ Summe, /*struct timeval*/ Start, /*struct timeval*/ Ende) \
  (Summe).tv_sec += (Ende).tv_sec - (Start).tv_sec;              \
   if (((Summe).tv_usec += (Ende).tv_usec - (Start).tv_usec)<0)  \
   {  (Summe).tv_usec += AnzMikrosekProSek;                      \
      (Summe).tv_sec--;                                          \
   }                                                             \
   else if ((Summe).tv_usec >= AnzMikrosekProSek)                \
   {  (Summe).tv_usec -= AnzMikrosekProSek;                      \
      (Summe).tv_sec++;                                          \
   }

#define /*void*/ ZeitdifferenzBestimmen(/*struct timeval*/ StartZeit, /*struct timeval*/ EndZeit, /*unsigned long*/ d_sec, /*unsigned long*/ d_usec)\
{\
    d_sec = EndZeit.tv_sec - StartZeit.tv_sec;\
    if (StartZeit.tv_usec > EndZeit.tv_usec){\
        d_usec = AnzMikrosekProSek + EndZeit.tv_usec - StartZeit.tv_usec;\
        d_sec--;\
   }\
   else\
      d_usec = EndZeit.tv_usec - StartZeit.tv_usec;\
}

void ZM_ZeitenAddieren(struct timeval *alteUT, struct timeval *alteST, struct timeval *summandUT, struct timeval *summandST)
/* Addiert die summandxT zu altexT, die SummandxT wird anschliessend auf Null gesetzt.*/
{
   alteUT->tv_sec += summandUT->tv_sec;
   alteUT->tv_usec += summandUT->tv_usec;
   if (alteUT->tv_usec > AnzMikrosekProSek){
      (alteUT->tv_sec)++;
      alteUT->tv_usec -= AnzMikrosekProSek;
   }
   summandUT->tv_usec = summandUT->tv_sec = 0;

   alteST->tv_sec += summandST->tv_sec;
   alteST->tv_usec += summandST->tv_usec;
   if (alteST->tv_usec > AnzMikrosekProSek){
      (alteST->tv_sec)++;
      alteST->tv_usec -= AnzMikrosekProSek;
   }
   summandST->tv_usec = summandST->tv_sec = 0;
}


#ifdef ZM_RUSAGE

/********** Realisierung des Zeitstacks ***************/

typedef struct StackZellenTS{
  struct timeval userZeit;
  struct timeval systemZeit;
}StackZellenT;

static StackZellenT *stack;
static unsigned Stackzeiger = 0;      /* Naechster freier Eintrag. */

static void push(struct timeval ut, struct timeval st)
{
  if_ZM_ZAEHLEN(MaximierenLong(&stackMax,Stackzeiger);)
  stack[Stackzeiger].userZeit = ut;
  stack[Stackzeiger++].systemZeit = st;
}

static void pop(struct timeval *ut, struct timeval *st)
/* Test auf leeren Stack entfaellt, da zu zeitintensiv! */
{
  *ut = stack[--Stackzeiger].userZeit;
  *st = stack[Stackzeiger].systemZeit;
}

void ZM_ZeitStart(void)

  /* Die Prozedur speichert den aktuellen Stand der bisher verbrauchten system- und user- Zeit
   * auf den Stack.
   */
{
   struct rusage rusageBuf;

   if_ZM_ZAEHLEN(gettimeCalls++;)
   if (getrusage (RUSAGE_SELF, &rusageBuf) == -1)
     our_assertion_failed ("Zeitmessung: getrusage fehlgeschlagen");
   push(rusageBuf.ru_utime, rusageBuf.ru_stime);
   /* DruckeZeitenAufStack();*/
}


void ZM_ZeitStopp(struct timeval *usertime,struct timeval *systemtime)
   /* Die Prozedur bildet die Differenz zwischen dem derzeitigen Stand der System-Zeit und
    *  der zuletzt gespeicherten Zeit, die oben auf dem Zeit-Stack liegt.
    */
{  
   struct rusage rusageBuf;
   struct timeval user_start_time;
   struct timeval system_start_time;
   
   if_ZM_ZAEHLEN(gettimeCalls++;)
   if (getrusage (RUSAGE_SELF, &rusageBuf) == -1)
      our_assertion_failed ("ZeitStopp: getrusage fehlgeschlagen");
   pop(&user_start_time, &system_start_time);
   /*DruckeZeitenAufStack();*/

   ZeitdifferenzAddieren(*usertime,user_start_time,rusageBuf.ru_utime);   /* user-time update */
   ZeitdifferenzAddieren(*systemtime,system_start_time,rusageBuf.ru_stime);   /* system-time update */

}

void ZM_ZeitSwitch(struct timeval *alteUT, struct timeval *alteST)
/* Diese Prozedur wechselt vom aktuellen Zeitmessungs-Kontext in einen anderen. Dadurch werden - verglichen mit einer
   Stopp-Start-Kombination - ein Aufruf der Zeitmessung sowie die entsprechenden Stack-Operationen eingespart.
   Bsp.: Waehrend der Unifikation wird die Normalformbildung aufgerufen => Switch vom Unifizieren zum Reduzieren.
   Die Zeit, in deren Kontext gewechselt wurde, kann wieder mit einem ZeitSwitch oder aber mit ZM_Stoppxxx gestoppt
   werden.
*/
{
   struct rusage rusageBuf;
   unsigned long StackPos = Stackzeiger-1;
   
   if_ZM_ZAEHLEN(gettimeCalls++;)
   if (getrusage (RUSAGE_SELF, &rusageBuf) == -1)
      our_assertion_failed ("ZeitStopp: getrusage fehlgeschlagen");
   
   ZeitdifferenzAddieren(*alteUT,stack[StackPos].userZeit, rusageBuf.ru_utime);   /* user-time update */
   ZeitdifferenzAddieren(*alteST,stack[StackPos].systemZeit,rusageBuf.ru_stime);   /* system-time update */
   
   stack[StackPos].userZeit = rusageBuf.ru_utime;
   stack[StackPos].systemZeit = rusageBuf.ru_stime;
}

static struct timeval angeh_st, angeh_ut;

void ZM_AlleZeitenAnhalten()
/* Haelt alle auf dem Stack liegenden Zeiten 'an'. Weitere Zeitmessungen koennen gestartet werden.
   Beachte: Aufrufe dieser Funktion sind _NICHT_ geschachtelt moeglich.*/
{
   struct rusage rusageBuf;

#if ZM_TEST
   if (ANGEHALTEN) our_error("\nZeitmessung: geschachtelte Aufrufe von AlleZeitenAnhalten.");
   ANGEHALTEN = TRUE;
   angeh_Stackpos = Stackzeiger;
#endif

   if_ZM_ZAEHLEN(gettimeCalls++;)
   if (getrusage (RUSAGE_SELF, &rusageBuf) == -1)
      our_assertion_failed ("AlleZeitenAnhalten: getrusage fehlgeschlagen");
   angeh_st = rusageBuf.ru_stime;
   angeh_ut = rusageBuf.ru_utime;
}   
  
void ZM_AlleZeitenWeiter()
/* Laesst die angehaltenen Zeiten weiterlaufen.
   Vor.: Alle in der Zwischenzeit auf den Stack gelegten Zeiten sind wieder entfernt worden.
*/
{
   struct rusage rusageBuf;
   unsigned long sec_s, usec_s, sec_u, usec_u;     /* Differenz zwischen akt. Zeitpunkt und Anhaltezeitpunkt */
   unsigned i;

#if ZM_TEST
   if (!ANGEHALTEN)
      our_error("\nZeitmessung: AlleZeitenWeiter ohne vorher angehalten zu halten\n");
   if (angeh_Stackpos != Stackzeiger)
      our_error("\nZeitmessung: AlleZeitenWeiter an falscher Position aufgerufen");
   ANGEHALTEN = FALSE;
#endif

   if_ZM_ZAEHLEN(gettimeCalls++;)
   if (getrusage (RUSAGE_SELF, &rusageBuf) == -1)
      our_assertion_failed ("AlleZeitenWeiter: getrusage fehlgeschlagen");

   /* Differenz zwischen Anhalten und jetzt berechnen */
   ZeitdifferenzBestimmen(angeh_ut, rusageBuf.ru_utime, sec_u, usec_u);
   ZeitdifferenzBestimmen(angeh_st, rusageBuf.ru_stime, sec_s, usec_s);

   for (i=0; i<Stackzeiger; i++){
      stack[i].userZeit.tv_sec += sec_u;
      stack[i].userZeit.tv_usec += usec_u;
      if (stack[i].userZeit.tv_usec > AnzMikrosekProSek){
         stack[i].userZeit.tv_sec++;
         stack[i].userZeit.tv_usec -= AnzMikrosekProSek;
      }
      stack[i].systemZeit.tv_sec += sec_s;
      stack[i].systemZeit.tv_usec += usec_s;
      if (stack[i].systemZeit.tv_usec > AnzMikrosekProSek){
         stack[i].systemZeit.tv_sec++;
         stack[i].systemZeit.tv_usec -= AnzMikrosekProSek;
      }
   }
}

/* ++++++++++++++++++++++++ Zeitmessung ohne Zeitstack ++++++++++++++++++++++++++*/
/* Wird fuer Speicherverwaltung benoetigt, um zyklische Aufrufe zu vermeiden. */

static struct timeval user_start_time;
static struct timeval system_start_time;


void ZM_ZeitStartOhneStack(void)

  /* Die Prozedur speichert den aktuellen Stand der bisher verbrauchten system- und user- Zeit
   * auf den Stack.
   */
{
   struct rusage rusageBuf;

   if (getrusage (RUSAGE_SELF, &rusageBuf) == -1)
      our_assertion_failed ("write_start_log_intereduction: getrusage fehlgeschlagen");
   user_start_time = rusageBuf.ru_utime, system_start_time = rusageBuf.ru_stime;
   /* DruckeZeitenAufStack();*/
}

void ZM_ZeitStoppOhneStack(struct timeval *usertime,struct timeval *systemtime)
   /* Die Prozedur bildet die Differenz zwischen dem derzeitigen Stand der System-Zeit und
    *  der zuletzt gespeicherten Zeit, die oben auf dem Zeit-Stack liegt.
    */
{  
   struct rusage rusageBuf;

   if (getrusage (RUSAGE_SELF, &rusageBuf) == -1)
      our_assertion_failed ("ZeitStopp: getrusage fehlgeschlagen");

   ZeitdifferenzAddieren(*usertime,user_start_time,rusageBuf.ru_utime);   /* user-time update */
   ZeitdifferenzAddieren(*systemtime,system_start_time,rusageBuf.ru_stime);   /* system-time update */

}

#else  /* Zeitmessung mit gettimeofday */

/********** Realisierung des Zeitstacks ***************/

typedef struct StackZellenTS{
  struct timeval realZeit;
}StackZellenT;

static StackZellenT *stack;
static unsigned Stackzeiger = 0;      /* Naechster freier Eintrag. */

static void push(struct timeval ut)
{
  if_ZM_ZAEHLEN(MaximierenLong(&stackMax,Stackzeiger);)
  stack[Stackzeiger++].realZeit = ut;
}

static void pop(struct timeval *ut)
/* Test auf leeren Stack entfaellt, da zu zeitintensiv */
{
  *ut = stack[--Stackzeiger].realZeit;
}

void ZM_ZeitStart(void)

  /* Die Prozedur speichert den aktuellen Stand der bisher verbrauchten system- und user- Zeit
   * auf den Stack.
   */
{
   struct timeval tp;

   if_ZM_ZAEHLEN(gettimeCalls++;)
   if (gettimeofday(&tp, &timezonedummy) == -1)
     our_assertion_failed ("Zeitmessung: gettimeofday fehlgeschlagen");
   push(tp);
   /* DruckeZeitenAufStack();*/
}


void ZM_ZeitStopp(struct timeval *usertime,struct timeval *systemtime)
   /* Die Prozedur bildet die Differenz zwischen dem derzeitigen Stand der System-Zeit und
    *  der zuletzt gespeicherten Zeit, die oben auf dem Zeit-Stack liegt.
    */
{  
   struct timeval tp;
   struct timeval user_start_time;

   if (gettimeofday(&tp, &timezonedummy) == -1)
      our_assertion_failed ("ZeitStopp: gettimeofday fehlgeschlagen");
   pop(&user_start_time);
   /*DruckeZeitenAufStack();*/

   ZeitdifferenzAddieren(*usertime,user_start_time,tp);   /* user-time update */
   if_ZM_ZAEHLEN(gettimeCalls++;systemtime->tv_sec++;)        /* Zaehler, der die Anzahl der Zeitmessungsaufrufe protokolliert */
}

void ZM_ZeitSwitch(struct timeval *alteUT, struct timeval *alteST)
/* Diese Prozedur wechselt vom aktuellen Zeitmessungs-Kontext in einen anderen. Dadurch werden - verglichen mit einer
   Stopp-Start-Kombination - ein Aufruf der Zeitmessung sowie die entsprechenden Stack-Operationen eingespart.
   Bsp.: Waehrend der Unifikation wird die Normalformbildung aufgerufen => Switch vom Unifizieren zum Reduzieren.
   Die Zeit, in deren Kontext gewechselt wurde, kann wieder mit einem ZeitSwitch oder aber mit ZM_Stoppxxx gestoppt
   werden.
*/
{
   struct timeval tp;
   unsigned long StackPos = Stackzeiger-1;
   
   if (gettimeofday(&tp, &timezonedummy) == -1)
      our_assertion_failed ("ZeitStopp: gettimeofday fehlgeschlagen");
   
   ZeitdifferenzAddieren(*alteUT,stack[StackPos].realZeit, tp);   /* user-time update */
   
   stack[StackPos].realZeit = tp;
   if_ZM_ZAEHLEN(gettimeCalls++;alteST->tv_sec++;)        /* Zaehler, der die Anzahl der Zeitmessungsaufrufe protokolliert */
}

static struct timeval angeh_t;
#if ZM_TEST
static unsigned long angeh_Stackpos;
#endif

void ZM_AlleZeitenAnhalten()
/* Haelt alle auf dem Stack liegenden Zeiten 'an'. Weitere Zeitmessungen koennen gestartet werden.
   Beachte: Aufrufe dieser Funktion sind _NICHT_ geschachtelt moeglich.*/
{
#if ZM_TEST
   if (ANGEHALTEN) our_error("\nZeitmessung: geschachtelte Aufrufe von AlleZeitenAnhalten.");
   ANGEHALTEN = TRUE;
   angeh_Stackpos = Stackzeiger;
#endif

   if_ZM_ZAEHLEN(gettimeCalls++;)
   if (gettimeofday(&angeh_t, &timezonedummy) == -1)
      our_assertion_failed ("ZeitStopp: gettimeofday fehlgeschlagen");
}   
  
void ZM_AlleZeitenWeiter()
/* Laesst die angehaltenen Zeiten weiterlaufen.
   Vor.: Alle in der Zwischenzeit auf den Stack gelegten Zeiten sind wieder entfernt worden.
*/
{
   struct timeval tp;
   unsigned long sec, usec;     /* Differenz zwischen akt. Zeitpunkt und Anhaltezeitpunkt */
   unsigned int i;

#if ZM_TEST
   if (!ANGEHALTEN)
      our_error("\nZeitmessung: AlleZeitenWeiter ohne vorher angehalten zu halten\n");
   ANGEHALTEN = FALSE;
   if (angeh_Stackpos != Stackzeiger)
      our_error("\nZeitmessung: AlleZeitenWeiter an falscher Position aufgerufen");
#endif

   if_ZM_ZAEHLEN(gettimeCalls++;)
   if (gettimeofday(&tp, &timezonedummy) == -1)
      our_assertion_failed ("ZeitStopp: gettimeofday fehlgeschlagen");

   /* Differenz zwischen Anhalten und jetzt berechnen */
   ZeitdifferenzBestimmen(angeh_t, tp, sec, usec);

   for (i=0; i<Stackzeiger; i++){
      stack[i].realZeit.tv_sec += sec;
      stack[i].realZeit.tv_usec += usec;
      if (stack[i].realZeit.tv_usec > AnzMikrosekProSek){
         stack[i].realZeit.tv_sec++;
         stack[i].realZeit.tv_usec -= AnzMikrosekProSek;
      }
   }
}
      
/* ++++++++++++++++++++++++ Zeitmessung ohne Zeitstack ++++++++++++++++++++++++++*/
/* Wird fuer Speicherverwaltung benoetigt, um zyklische Aufrufe zu vermeiden. */

static struct timeval user_start_time;

void ZM_ZeitStartOhneStack(void)

  /* Die Prozedur speichert den aktuellen Stand der bisher verbrauchten system- und user- Zeit
   * auf den Stack.
   */
{
   struct timeval tp;

   if_ZM_ZAEHLEN(gettimeCalls++;)
   if (gettimeofday(&tp, &timezonedummy) == -1)
      our_assertion_failed ("ZeitStopp: gettimeofday fehlgeschlagen");
   user_start_time = tp;
   /* DruckeZeitenAufStack();*/
}

void ZM_ZeitStoppOhneStack(struct timeval *usertime,struct timeval *systemtime)
   /* Die Prozedur bildet die Differenz zwischen dem derzeitigen Stand der System-Zeit und
    *  der zuletzt gespeicherten Zeit, die oben auf dem Zeit-Stack liegt.
    */
{  
   struct timeval tp;

   if_ZM_ZAEHLEN(gettimeCalls++;)
   if (gettimeofday(&tp, &timezonedummy) == -1)
      our_assertion_failed ("ZeitStopp: gettimeofday fehlgeschlagen");

   ZeitdifferenzAddieren(*usertime,user_start_time,tp);   /* user-time update */
}



#endif

int ZM_ZeitStartF(void)
{
   ZM_ZeitStart();
   return 0;
}

int ZM_ZeitStoppF(struct timeval *user,struct timeval *system)
{
   ZM_ZeitStopp(user,system);
   return 0;
}

int ZM_ZeitSwitchF(struct timeval *alteUT, struct timeval *alteST)
{
   ZM_ZeitSwitch(alteUT, alteST);
   return 0;
}





static struct timeval GesamtzeitStart_user_time;
static struct timeval GesamtzeitStart_system_time;
static struct timeval GesamtzeitStart_wallclock_time;
static struct timeval GesamtzeitWallClock;

void ZM_StartGesamtZeit(void)
{
   struct rusage rusageBuf;

   if (getrusage (RUSAGE_SELF, &rusageBuf) == -1)
      our_assertion_failed ("Start time: getrusage didn't work");
   GesamtzeitStart_user_time = rusageBuf.ru_utime, GesamtzeitStart_system_time = rusageBuf.ru_stime;
   /* DruckeZeitenAufStack();*/
   if (gettimeofday(&GesamtzeitStart_wallclock_time, &timezonedummy) == -1)
     our_assertion_failed ("Start time: gettimeofday didn't work");
}

void ZM_StoppGesamtZeit(void)
{  
   struct timeval GesamtzeitEnde_wallclock_time;
   struct rusage rusageBuf;

   if (getrusage (RUSAGE_SELF, &rusageBuf) == -1)
      our_assertion_failed ("Stop time: getrusage didn't work");
   if (gettimeofday(&GesamtzeitEnde_wallclock_time, &timezonedummy) == -1)
      our_assertion_failed ("Stop time: gettimeofday didn't work");

   ZeitdifferenzAddieren(ZM_GesamtZeit_ut,GesamtzeitStart_user_time,rusageBuf.ru_utime);   /* user-time update */
   ZeitdifferenzAddieren(ZM_GesamtZeit_st,GesamtzeitStart_system_time,rusageBuf.ru_stime);   /* system-time update */
   ZeitdifferenzAddieren(GesamtzeitWallClock, GesamtzeitStart_wallclock_time, GesamtzeitEnde_wallclock_time);
}

void ZM_Times(unsigned long int *cpu, unsigned long int *wct)
/* Liefert cputime und wallclocktime in msec zurueck  */
{
  
  *cpu = (ZM_GesamtZeit_ut.tv_sec + ZM_GesamtZeit_st.tv_sec)*1000
    + (ZM_GesamtZeit_ut.tv_usec + ZM_GesamtZeit_st.tv_usec)/1000
    + 1;
  if (*cpu==0) *cpu=1;

  *wct = GesamtzeitWallClock.tv_sec*1000 + GesamtzeitWallClock.tv_usec/1000 ;
  if (*wct == 0) *wct=1;
}


BOOLEAN ZM_ZeitstackOK(void)
{
  BOOLEAN Fehler = FALSE;

  fprintf(IO_Strom, "\nZeitmessungstest ...\n");
  if (Stackzeiger > 0){
    fprintf(IO_Strom, "Auf dem Zeitstack liegen noch %d Zeiten.\n",Stackzeiger-1);
    Fehler = TRUE;
  }
  if_ZM_ZAEHLEN(fprintf(IO_Strom, "%ld Aufrufe der Zeitmessung\n", gettimeCalls);)
  if_ZM_ZAEHLEN(fprintf(IO_Strom, "Maximale Ausdehnung des Zeitstacks ist %ld\n", stackMax);)
  if (AnzahlZeitFehler>0){
    fprintf(IO_Strom, "%ld Fehler in der Zeitmessung.\n", AnzahlZeitFehler);
    Fehler = TRUE;
  }
  if (!Fehler)
    fprintf(IO_Strom, "                 ... ok\n");
  return Fehler;
}

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */
/* +++++++++++++++++++++++++ Initialisierung +++++++++++++++++++++++++++++++++ */
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++ */

void ZM_ReInitApriori(void)
{
  return;
   timezonedummy.tz_minuteswest = 0;
   timezonedummy.tz_dsttime = 0;

   ZM_NFBildung_ut.tv_sec=0; 
   ZM_NFBildung_st.tv_sec=0; 
   ZM_MatchenRm_ut.tv_sec=0; 
   ZM_MatchenRm_st.tv_sec=0; 
   ZM_MatchenGm_ut.tv_sec=0; 
   ZM_MatchenGm_st.tv_sec=0; 
   ZM_TERedGm_ut.tv_sec=0;   
   ZM_TERedGm_st.tv_sec=0;
   ZM_Rewrite_ut.tv_sec=0;
   ZM_Rewrite_st.tv_sec=0;

   ZM_KPBildungZuRegel_ut.tv_sec=0;      
   ZM_KPBildungZuRegel_st.tv_sec=0;      
   ZM_KPBildungZuGleichung_ut.tv_sec=0; 
   ZM_KPBildungZuGleichung_st.tv_sec=0; 
   ZM_MGUZuR_ut.tv_sec=0;                    
   ZM_MGUZuR_st.tv_sec=0;                    
   ZM_MGUZuG_ut.tv_sec=0;                    
   ZM_MGUZuG_st.tv_sec=0;                    
   ZM_Kapurtest_ut.tv_sec = 0;
   ZM_Kapurtest_st.tv_sec = 0;

   ZM_Subsumption_ut.tv_sec=0;              
   ZM_Subsumption_st.tv_sec=0;              
   ZM_Interred_ut.tv_sec=0;                 
   ZM_Interred_st.tv_sec=0;                 
   ZM_OrdnungEntscheidung_ut.tv_sec=0;      
   ZM_OrdnungEntscheidung_st.tv_sec=0;      
   ZM_OrdnungRedTest_ut.tv_sec=0;           
   ZM_OrdnungRedTest_st.tv_sec=0;           
   ZM_OrdnungUnifTest_ut.tv_sec=0;          
   ZM_OrdnungUnifTest_st.tv_sec=0;          
   ZM_PP_ut.tv_sec=0;                       
   ZM_PP_st.tv_sec=0;
   ZM_RegelEinf_ut.tv_sec=0;                
   ZM_RegelEinf_st.tv_sec=0;                
   ZM_RegelLoeschen_ut.tv_sec=0;            
   ZM_RegelLoeschen_st.tv_sec=0;            
   ZM_GleichgEinf_ut.tv_sec=0;              
   ZM_GleichgEinf_st.tv_sec=0;              
   ZM_GleichgLoeschen_ut.tv_sec=0;          
   ZM_GleichgLoeschen_st.tv_sec=0;          
   ZM_KPEinf_ut.tv_sec=0;                   
   ZM_KPEinf_st.tv_sec=0;                   
   ZM_KPLoeschen_ut.tv_sec=0;               
   ZM_KPLoeschen_st.tv_sec=0;               
   ZM_KPEinpacken_ut.tv_sec=0;              
   ZM_KPEinpacken_st.tv_sec=0;              
   ZM_KPAuspacken_ut.tv_sec=0;              
   ZM_KPAuspacken_st.tv_sec=0;              
   ZM_GepackteKPsLoeschen_ut.tv_sec=0;      
   ZM_GepackteKPsLoeschen_st.tv_sec=0;      
   ZM_NeuerSpeicherBlock_ut.tv_sec=0;       
   ZM_NeuerSpeicherBlock_st.tv_sec=0;

   ZM_GesamtZeit_ut.tv_usec=0;
   ZM_GesamtZeit_st.tv_usec=0;
   ZM_NFBildung_ut.tv_usec=0; 
   ZM_NFBildung_st.tv_usec=0; 
   ZM_MatchenRm_ut.tv_usec=0; 
   ZM_MatchenRm_st.tv_usec=0; 
   ZM_MatchenGm_ut.tv_usec=0; 
   ZM_MatchenGm_st.tv_usec=0; 
   ZM_TERedGm_ut.tv_usec=0;   
   ZM_TERedGm_st.tv_usec=0;
   ZM_Rewrite_ut.tv_usec=0;
   ZM_Rewrite_st.tv_usec=0;
   ZM_KPBildungZuRegel_ut.tv_usec=0;      
   ZM_KPBildungZuRegel_st.tv_usec=0;      
   ZM_KPBildungZuGleichung_ut.tv_usec=0; 
   ZM_KPBildungZuGleichung_st.tv_usec=0; 
   ZM_MGUZuR_ut.tv_usec=0;                    
   ZM_MGUZuR_st.tv_usec=0;                    
   ZM_MGUZuG_ut.tv_usec=0;
   ZM_MGUZuG_st.tv_usec=0;
   ZM_Kapurtest_ut.tv_usec = 0;
   ZM_Kapurtest_st.tv_usec = 0;
   ZM_Subsumption_ut.tv_usec=0;              
   ZM_Subsumption_st.tv_usec=0;              
   ZM_Interred_ut.tv_usec=0;                 
   ZM_Interred_st.tv_usec=0;                 
   ZM_OrdnungEntscheidung_ut.tv_usec=0;      
   ZM_OrdnungEntscheidung_st.tv_usec=0;      
   ZM_OrdnungRedTest_ut.tv_usec=0;           
   ZM_OrdnungRedTest_st.tv_usec=0;           
   ZM_OrdnungUnifTest_ut.tv_usec=0;          
   ZM_OrdnungUnifTest_st.tv_usec=0;          
   ZM_PP_ut.tv_usec=0;                       
   ZM_PP_st.tv_usec=0;
   ZM_RegelEinf_ut.tv_usec=0;                
   ZM_RegelEinf_st.tv_usec=0;                
   ZM_RegelLoeschen_ut.tv_usec=0;            
   ZM_RegelLoeschen_st.tv_usec=0;            
   ZM_GleichgEinf_ut.tv_usec=0;              
   ZM_GleichgEinf_st.tv_usec=0;              
   ZM_GleichgLoeschen_ut.tv_usec=0;          
   ZM_GleichgLoeschen_st.tv_usec=0;          
   ZM_KPEinf_ut.tv_usec=0;                   
   ZM_KPEinf_st.tv_usec=0;                   
   ZM_KPLoeschen_ut.tv_usec=0;               
   ZM_KPLoeschen_st.tv_usec=0;               
   ZM_KPEinpacken_ut.tv_usec=0;              
   ZM_KPEinpacken_st.tv_usec=0;              
   ZM_KPAuspacken_ut.tv_usec=0;              
   ZM_KPAuspacken_st.tv_usec=0;              
   ZM_GepackteKPsLoeschen_ut.tv_usec=0;      
   ZM_GepackteKPsLoeschen_st.tv_usec=0;      
   ZM_NeuerSpeicherBlock_ut.tv_usec=0;       
   ZM_NeuerSpeicherBlock_st.tv_usec=0;

   ZM_GesamtZeit_ut.tv_sec=0;
   ZM_GesamtZeit_st.tv_sec=0;

   GesamtzeitWallClock.tv_sec = 0;
   GesamtzeitWallClock.tv_usec = 0;

}

void ZM_InitApriori(void)
{
   stack = our_alloc((ZM_StackSize)*sizeof(StackZellenT));

#if ZM_TEST
   TestStack = our_alloc((ZM_StackSize)*sizeof(TestStackZellenT));
#endif 
   ZM_ReInitApriori();
}


#if (ZM_MESSLEVEL !=0)
static BOOLEAN ZeitenDa(int anz, ...)
/* Gibt TRUE zurueck, wenn einer der uebergebenen Werte > 0 ist */
{
   va_list ap;
   int i;
   struct timeval stval;
   BOOLEAN ok=FALSE;

   va_start(ap, anz);
   for (i=0; i<anz; i++){
      stval = va_arg(ap, struct timeval);
      if ((stval.tv_sec>0) || (stval.tv_usec>0)){
         ok=TRUE;
         break;
      }
   }
   va_end(ap);
   return ok;
}
#endif
 
static void DruckeZeitSolo(FILE *fp, struct timeval time)
{
  int             i;
  char            buf1 [256];
  char            buf2 [256];
  const int       digits = 3; /* strlen gibt size_t zurueck */
  const int       maxDigits = 6;
  int             cutAfterDigit;
  
  sprintf (buf1, "%li", time.tv_sec);
  sprintf (buf2, "%li", time.tv_usec);
  if (! ( (signed)strlen (buf2) <= maxDigits ))
    our_assertion_failed("Fehler in IO_Drucke-Zeit.");
  cutAfterDigit = (signed)strlen (buf2) - digits;
  if ( cutAfterDigit >=0 )
    buf2 [cutAfterDigit] = '\000';                                                                  /* abschneiden */
  else
    buf2 [0] = '\000';                                                                        /* komplett loeschen */
  for ( i = 10 - (signed) strlen (buf1); i > 0; i-- )
    fprintf(fp, " ");
  fprintf(fp, "%s,", buf1);
  if (! ( (signed)strlen (buf2) <= digits ))
    our_assertion_failed("Fehler in IO_Drucke-Zeit.");
  for ( i = digits - (signed)strlen (buf2); i > 0; i-- )
    fprintf(fp,  "0");
  fprintf(fp, "%s", buf2);
}

#if (ZM_MESSLEVEL !=0)
static void DruckeZeiten (char *str,struct timeval utime,struct timeval stime)
/* Gibt die beiden Zeiten und dahinter den String aus.*/
{
   if (utime.tv_sec || utime.tv_usec || stime.tv_sec || stime.tv_usec){
      DruckeZeitSolo(IO_Strom, utime);
      fprintf(IO_Strom, "s ut");
      if (ZM_GettimeofdayEingestellt())
         if (ZM_AufrufeZaehlen())
            fprintf(IO_Strom, "%14ld calls", stime.tv_sec);
         else
            fprintf(IO_Strom, "              ");
      else{
         DruckeZeitSolo(IO_Strom, stime);
         fprintf(IO_Strom, "s st");
       }
      fprintf(IO_Strom, "   %s",str);
      IO_LF();
   }
}
#endif

void ZM_Statistik(InfoStringT IS)
{
   if (!(IS_IsIn("ZM", IS) || IS_IsIn("all", IS)))
     return;

   if (IS_IsIn("standard", IS)){
      IO_DruckeUeberschrift("TIMES (user- and system-time in seconds)");
#if (ZM_MESSLEVEL ==0)
      fprintf(IO_Strom, "... no detailed timing enabled\n\n");
#else
      fprintf(IO_Strom, "\nVerwendete Zeitmessung : %s\n\n", ZM_VerwendeteZeitmessung());
      if (ZM_GetrusageEingestellt())
         fprintf(IO_Strom, "\nZeiten jeweils user- und system-Zeit des Prozesses.\n");
      if (ZM_GettimeofdayEingestellt()){
         fprintf(IO_Strom, "\nEinzelzeiten sind Realzeiten, System-Zeiten wurden nicht gemessen.\n");
         if (ZM_AufrufeZaehlen())
            fprintf(IO_Strom, "Die zweite Zahl gibt die Anzahl der gemessenen Teilzeiten an.\n");
      }
      if (ZeitenDa(2,ZM_PP_ut,ZM_PP_st)){
         IO_DruckeKurzUeberschrift("Initialisierungen","+");
         DruckeZeiten("Parsen und Plattklopfen",ZM_PP_ut,ZM_PP_st);
      }
      if (ZeitenDa(2,ZM_NFBildung_ut, ZM_NFBildung_st)){
         IO_DruckeKurzUeberschrift("Normalformbildung","+");
         if (ZeitenDa(10, ZM_MatchenRm_ut,ZM_MatchenRm_st, ZM_MatchenGm_ut,ZM_MatchenGm_st,
                      ZM_TERedGm_ut,ZM_TERedGm_st, ZM_Rewrite_ut, ZM_Rewrite_st, ZM_OrdnungRedTest_ut,ZM_OrdnungRedTest_st)){
            DruckeZeiten("Normalformbildung insgesamt, davon", ZM_NFBildung_ut, ZM_NFBildung_st);
            DruckeZeiten("... Matchen mit Regelmenge",ZM_MatchenRm_ut,ZM_MatchenRm_st);
            DruckeZeiten("... Matchen mit Gleichungsmenge",ZM_MatchenGm_ut,ZM_MatchenGm_st);
            DruckeZeiten("... Ueberfluessige Termerzeugung bei E-Schritt",ZM_TERedGm_ut,ZM_TERedGm_st);
            DruckeZeiten("... Rewrite insgesamt", ZM_Rewrite_ut, ZM_Rewrite_st);
            DruckeZeiten("... Ordnungen-Test in Reduktion",ZM_OrdnungRedTest_ut,ZM_OrdnungRedTest_st);
         }
         else
            DruckeZeiten("Normalformbildung insgesamt", ZM_NFBildung_ut, ZM_NFBildung_st);
      }
      
      if(ZeitenDa(4, ZM_KPBildungZuRegel_ut, ZM_KPBildungZuRegel_st,ZM_KPBildungZuGleichung_ut, ZM_KPBildungZuGleichung_st)){
         IO_DruckeKurzUeberschrift("Unifikation","+");
         
         DruckeZeiten("KP-Bildung zu neuen Regeln", ZM_KPBildungZuRegel_ut, ZM_KPBildungZuRegel_st);
         DruckeZeiten("... davon MGU-Suche",ZM_MGUZuRegel_ut, ZM_MGUZuRegel_st);
         DruckeZeiten("... davon Kapur-Test",ZM_KapurtestZuRegel_ut,ZM_KapurtestZuRegel_st);

         if (ZeitenDa(2, ZM_KPBildungZuGleichung_ut,ZM_KPBildungZuGleichung_st)){
            fprintf(IO_Strom, "\n");
            DruckeZeiten("KP-Bildung zu neuen Gleichungen", ZM_KPBildungZuGleichung_ut, ZM_KPBildungZuGleichung_st);
            DruckeZeiten("... davon MGU-Suche",ZM_MGUZuGleichung_ut, ZM_MGUZuGleichung_st);
            DruckeZeiten("... davon Kapur-Test", ZM_KapurtestZuGleichung_ut, ZM_KapurtestZuGleichung_st);
            DruckeZeiten("... davon Ordnungen-Test in Unifikation",ZM_OrdnungUnifTest_ut,ZM_OrdnungUnifTest_st);
         }
      }
      
      if (ZeitenDa(2,ZM_Subsumption_ut,ZM_Subsumption_st)){
         IO_DruckeKurzUeberschrift("Subsumption","+");
         DruckeZeiten("Subsumption insgesamt",ZM_Subsumption_ut,ZM_Subsumption_st);
      }
      if (ZeitenDa(2,ZM_Interred_ut,ZM_Interred_st)){
         IO_DruckeKurzUeberschrift("Interreduktion","+");
         DruckeZeiten("Interreduktion insgesamt", ZM_Interred_ut,ZM_Interred_st);
      }
      if (ZeitenDa(2,ZM_OrdnungEntscheidung_ut,ZM_OrdnungEntscheidung_st)){
         IO_DruckeKurzUeberschrift("Ordnungen","+");
         DruckeZeiten("Entscheidungsfuktion im Vervollstaendiger",ZM_OrdnungEntscheidung_ut,ZM_OrdnungEntscheidung_st);
      }
      if (ZeitenDa(4,ZM_RegelEinf_ut,ZM_RegelEinf_st,ZM_RegelLoeschen_ut,ZM_RegelLoeschen_st)){
         IO_DruckeKurzUeberschrift("Regelmengenverwaltung","+");
         DruckeZeiten("Einfuegen",ZM_RegelEinf_ut,ZM_RegelEinf_st);
         DruckeZeiten("Loeschen",ZM_RegelLoeschen_ut,ZM_RegelLoeschen_st);
      }
      if (ZeitenDa(4,ZM_GleichgEinf_ut,ZM_GleichgEinf_st,ZM_GleichgLoeschen_ut,ZM_GleichgLoeschen_st)){
         IO_DruckeKurzUeberschrift("Gleichungsmengenverwaltung","+");
         DruckeZeiten("Einfuegen",ZM_GleichgEinf_ut,ZM_GleichgEinf_st);
         DruckeZeiten("Loeschen",ZM_GleichgLoeschen_ut,ZM_GleichgLoeschen_st);
      }
      if (ZeitenDa(4,ZM_KPEinf_ut,ZM_KPEinf_st,ZM_KPLoeschen_ut,ZM_KPLoeschen_st)){
         IO_DruckeKurzUeberschrift("Kritische-Paar-Mengen-Verwaltung","+");
         DruckeZeiten("Einfuegen",ZM_KPEinf_ut,ZM_KPEinf_st);
         DruckeZeiten("Loeschen",ZM_KPLoeschen_ut,ZM_KPLoeschen_st);
      }
      if (PA_KPMengePacken()){
         if (ZeitenDa(6,ZM_KPEinpacken_ut,ZM_KPEinpacken_st,ZM_KPAuspacken_ut,ZM_KPAuspacken_st,ZM_GepackteKPsLoeschen_ut,ZM_GepackteKPsLoeschen_st)){
            IO_DruckeKurzUeberschrift("Gepackte KPs", "+");
            DruckeZeiten("Einpacken",ZM_KPEinpacken_ut,ZM_KPEinpacken_st);
            DruckeZeiten("Auspacken",ZM_KPAuspacken_ut,ZM_KPAuspacken_st);
            DruckeZeiten("Loeschen",ZM_GepackteKPsLoeschen_ut,ZM_GepackteKPsLoeschen_st);
         }
      }
      if (ZM_NeuerSpeicherBlock_ut.tv_sec || ZM_NeuerSpeicherBlock_ut.tv_usec || 
          ZM_NeuerSpeicherBlock_st.tv_sec || ZM_NeuerSpeicherBlock_st.tv_usec){
         IO_DruckeKurzUeberschrift("Zeitbedarf der Speicherverwaltung", "+");
         DruckeZeiten("Anlegen und Initialisieren neuer Bloecke",ZM_NeuerSpeicherBlock_ut,
                      ZM_NeuerSpeicherBlock_st);
         fprintf(IO_Strom, "\n\n");
      }
      fprintf(IO_Strom, "\n");
      IO_DruckeZeile("=");
      fprintf(IO_Strom, "\n");
#endif
      DruckeZeitSolo(IO_Strom, ZM_GesamtZeit_ut);
      fprintf(IO_Strom, "s ut");
      DruckeZeitSolo(IO_Strom, ZM_GesamtZeit_st);
      fprintf(IO_Strom, "s st (");
      DruckeZeitSolo(IO_Strom, GesamtzeitWallClock);
      fprintf(IO_Strom, " wct)");
      fprintf(IO_Strom, "   total time (without initialisation)\n");
   }
   else if (IS_IsIn("latex", IS)){
      DruckeZeitSolo(IO_Strom, ZM_GesamtZeit_ut);
      fprintf(IO_Strom, "s ut");
      DruckeZeitSolo(IO_Strom, ZM_GesamtZeit_st);
      fprintf(IO_Strom, "s st (");
      DruckeZeitSolo(IO_Strom, GesamtzeitWallClock);
      fprintf(IO_Strom, " wct)");
   }
   else if (IS_IsIn("intern", IS)){
#if ZM_STATISTIK
      IO_DruckeStatistikZeile("timer queries", gettimeCalls);
#else
      ;
#endif
   }
   else if (IS_IsIn("config", IS)){
      fprintf(IO_Strom, "ZM: Level %d, ", ZM_MESSLEVEL);
#ifdef ZM_TIMEOFDAY
      fprintf(IO_Strom, "gettimeofday, ");
#elif defined ZM_RUSAGE
      fprintf(IO_Strom, "getrusage, ");
#endif
#if ZM_STATISTIK
      fprintf(IO_Strom, "count queries, ");
#else
      fprintf(IO_Strom, "don't count queries, ");
#endif
#if ZM_TEST
      fprintf(IO_Strom, "Timertest  ON.\n");
#else
      fprintf(IO_Strom, "Timertest OFF.\n");
#endif
   }
}
