/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - Classification.c                                                                                           */
/* - ****************                                                                                           */
/* -                                                                                                            */
/* - Schnittstelle der Klassifikationsroutinen fuer SUEM.                                                       */
/* -                                                                                                            */
/* - 14.03.96 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#include <limits.h>
#include <string.h>

#include "Ausgaben.h"
#include "Classification.h"
#include "Hauptkomponenten.h"
#include "InfoStrings.h"
#include "Interreduktion.h"
#include "RUndEVerwaltung.h"
#include "TermOperationen.h"
#include "general.h"
#include "SH2.h"
#include "CStatistics.h"
#include "CGHeuristiken.h"
#include "Ordnungen.h"

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe Classification.h) --------------------*/
/* ------------------------------------------------------------------------*/

ClassifyProcT   C_ClassifyProcAR [4], C_CG_ClassifyProc;
ReClassifyProcT C_ReClassifyProc;
PrepareProcT    C_PrepareProc;
void (* C_PrintProc) (WeightEntryT *, RegelOderGleichungsT, RegelOderGleichungsT);
void (* C_WeightStatisticsStartProc) (InfoStringT);
void (* C_WeightStatisticsEndProc) (void);
PQ_WatchingWalkthroughProcedureT C_WeightStatProc;
int (* C_AdditionalComparisonProc) (long int, long int, void *, TermpairT, RegelOderGleichungsT, RegelOderGleichungsT,
                                    long int, long int, void *, TermpairT, RegelOderGleichungsT, RegelOderGleichungsT);
void (* C_KillAdditionalReferenceProc) (void *);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ CLEANING WOMAN ---------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
static void (* C_CleanUpProc) (void);

void C_CleanUp(void){
   CS_CleanUp();
   (* C_CleanUpProc)();
}

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ Standard classification ------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


static void StandardKillAddRefProc(void *addref)
{
   ;
}

static void StandardWeightStatProc(WeightEntryT wt, TermpairT tp, RegelOderGleichungsT actualP, 
                            RegelOderGleichungsT otherP)
{
   ;
}

static void StandardCleanUpProc(void){;}


ClassifyProcT CLAS_0, CLAS_50, CLAS_100, CLAS_200, CLAS_pen, CLAS_DEFAULT;
ReClassifyProcT RECLAS_0, RECLAS_50, RECLAS_100, RECLAS_200, RECLAS_pen, RECLAS_DEFAULT, RECLAS_Dummy;

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ Criteria ---------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

typedef BOOLEAN (* CritPraedikatsT) (TermT, TermT, RegelOderGleichungsT, RegelOderGleichungsT);


#define AnzahlDefinierterKriterien 8

typedef struct{
   const char *Name;
   CritPraedikatsT PraedikatsProc;
} CritDefStructT;

CritDefStructT CritDefAR[AnzahlDefinierterKriterien];



typedef struct{
   const char *Name;
   CritPraedikatsT PraedikatsProc;
   ClassifyProcT ClassifyProc;
   ReClassifyProcT ReclassifyProc;
   unsigned long int Fulfilled;
} CritAnwStructT;

typedef struct{
  char *DEF, *USR;
  CritAnwStructT CritAnwAR[AnzahlDefinierterKriterien];
  int AnzahlAnzuwendendeKriterien;
  BOOLEAN PermGradBerechnen;
} CriteriaT;


CriteriaT crit[2];              /* CLAS, RECLAS */


static BOOLEAN KillerPraedikatR(TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, 
                         RegelOderGleichungsT otherParent)
{
   switch (ORD_VergleicheTerme(lhs, rhs)) {
   case Kleiner:
      if (IR_TPRReduziertRE(rhs, lhs)) {
         fprintf(IO_Strom,"Killer-KP!\n");
         return TRUE;
      }
      else
         return FALSE;
      break;
   case Groesser:
      if (IR_TPRReduziertRE(lhs, rhs)) {
         fprintf(IO_Strom,"Killer-KP!\n");
         return TRUE;
      }
      else
         return FALSE;
      break;
   default:
      return FALSE;
      break;
   }
}

BOOLEAN KillerPraedikatE(TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, 
                         RegelOderGleichungsT otherParent)
{
   if ((ORD_VergleicheTerme(lhs, rhs) == Unvergleichbar) &&
       (IR_TPEReduziertRE(lhs, rhs))){
      fprintf(IO_Strom,"Killer-KP!\n");
      return TRUE;
   }
   else
      return FALSE;
}

static BOOLEAN KillerPraedikatRE(TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, 
                         RegelOderGleichungsT otherParent)
{
   switch (ORD_VergleicheTerme(lhs, rhs)) {
   case Kleiner:
      if (IR_TPRReduziertRE(rhs, lhs)) {
         fprintf(IO_Strom,"Killer-KP!\n");
         return TRUE;
      }
      else
         return FALSE;
      break;
   case Groesser:
      if (IR_TPRReduziertRE(lhs, rhs)) {
         fprintf(IO_Strom,"Killer-KP!\n");
         return TRUE;
      }
      else
         return FALSE;
      break;
   case Unvergleichbar:
      if (IR_TPEReduziertRE(lhs, rhs)){
         fprintf(IO_Strom,"Killer-KP!\n");
         return TRUE;
      }
      else
         return FALSE;
   default:
      return FALSE;
      break;
   }
}

static unsigned long int PermGrad;

static BOOLEAN PermutationsPraedikat3(TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, 
                                      RegelOderGleichungsT otherParent)
{
   return (PermGrad == 3);
}

static BOOLEAN PermutationsPraedikat4(TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, 
                                      RegelOderGleichungsT otherParent)
{
   return (PermGrad == 4);
}

static BOOLEAN PermutationsPraedikat5(TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, 
                                      RegelOderGleichungsT otherParent)
{
   return (PermGrad == 5);
}

static BOOLEAN PermutationsPraedikat6(TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, 
                                      RegelOderGleichungsT otherParent)
{
   return (PermGrad >= 6);
}

static BOOLEAN EChildPraedikat(TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, 
                               RegelOderGleichungsT otherParent)
{
   return (!RE_IstRegel(actualParent) || !RE_IstRegel(otherParent));
}





static void ClassifyWithCriteria(WeightEntryT *wt, TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, 
                                 RegelOderGleichungsT otherParent, UTermT actualPosition)
{
   int i;

   if (crit[0].PermGradBerechnen)
      PermGrad = RE_GradVonPermPaar(lhs, rhs);

   for (i=0; i<crit[0].AnzahlAnzuwendendeKriterien; i++){
      if ( (* (crit[0].CritAnwAR[i].PraedikatsProc) ) (lhs,  rhs, actualParent, otherParent)){
         crit[0].CritAnwAR[i].Fulfilled++;
         (* (crit[0].CritAnwAR[i].ClassifyProc)) (wt, lhs, rhs, actualParent, otherParent, actualPosition);
         return;
      }
   }
   (* CLAS_100) (wt, lhs, rhs, actualParent, otherParent, actualPosition); 
   /* kein Kriterium anwendbar: normale Gewichtung */
}


static int ReClassifyWithCriteria(WeightEntryT *wt, TermT lhs, TermT rhs, RegelOderGleichungsT actualParent, 
                                  RegelOderGleichungsT otherParent)
{
   int i;
   
   if (crit[1].PermGradBerechnen)
      PermGrad = RE_GradVonPermPaar(lhs, rhs);

   for (i=0; i<crit[1].AnzahlAnzuwendendeKriterien; i++){
      if ( (* (crit[1].CritAnwAR[i].PraedikatsProc) ) (lhs, rhs, actualParent, otherParent)){
         crit[1].CritAnwAR[i].Fulfilled++;
         return (* (crit[1].CritAnwAR[i].ReclassifyProc)) (wt, lhs, rhs, actualParent, otherParent);
      }
   }
   return (* RECLAS_100) (wt, lhs, rhs, actualParent, otherParent); 
   /* kein Kriterium anwendbar: normale Gewichtung */
}



/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ Statistics -------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void (* C_StatisticsProc) (InfoStringT); /* Die Statistik-Prozedur des Submoduls */

void C_Statistics(InfoStringT IS)
{
   if (!(IS_IsIn("C", IS) || IS_IsIn("all", IS)))
      return;
   if (IS_IsIn("standard", IS)){
      if (crit[0].AnzahlAnzuwendendeKriterien
          + crit[1].AnzahlAnzuwendendeKriterien > 0){
         int i, j;
         char str[100];
         const char *bla[2] = {" during CLASSIFY", " during RECLASSIFY"};

         IO_DruckeKurzUeberschrift("Classification", "+");
         for (j=0; j<2; j++){
           for (i=0; i<crit[j].AnzahlAnzuwendendeKriterien; i++){
             strcpy (str, crit[j].CritAnwAR[i].Name);
             strcat (str, " ");
             strcat (str, bla[j]);
             IO_DruckeStatistikZeile(str, crit[j].CritAnwAR[i].Fulfilled);
           }
         }
       }
      ((* C_StatisticsProc) (IS));
   }
   else if (IS_IsIn("latex-head", IS)){
      if (crit[0].AnzahlAnzuwendendeKriterien
          + crit[1].AnzahlAnzuwendendeKriterien > 0){
         int i, j;
         const char *bla[2] = {"CLAS: ", "RECLAS: "};
         for (j=0; j<2; j++){
            fprintf(IO_Strom, bla[j]);
            for (i=0; i<crit[j].AnzahlAnzuwendendeKriterien; i++){
               fprintf(IO_Strom, " %s & ", crit[j].CritAnwAR[i].Name);
            }
         }
         ((* C_StatisticsProc) (IS));
      }
   }
   else if (IS_IsIn("config", IS)){
      ;
   }
   else if (IS_IsIn("intern", IS)){
      ;
   }
   else if (IS_IsIn("time", IS))
      ((* C_StatisticsProc) (IS));
}      

/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ Initialize -------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

#define KriteriumEinrichten(DEF, USR, Name, ArrayEintragCLAS, ArrayEintragRECLAS)                               \
{                                                                                                               \
   if (IS_FindValuedOption(DEF, USR, 1, Name) == 0){                                                            \
      if (strcmp(IS_ValueOfOption(DEF, USR, Name), "half") == 0){                                               \
         fprintf(IO_Strom, "Prefer %s half way, ", Name);                                                       \
         ArrayEintragCLAS = CLAS_50;                                                                            \
         ArrayEintragRECLAS =  RECLAS_50;                                                                       \
      }                                                                                                         \
      else if (strcmp(IS_ValueOfOption(DEF, USR, Name), "ultimate") == 0){                                      \
         fprintf(IO_Strom, "Prefer %s ultimate, ", Name);                                                       \
         ArrayEintragCLAS = CLAS_0;                                                                             \
         ArrayEintragRECLAS =  RECLAS_0;                                                                        \
      }                                                                                                         \
      else if (strcmp(IS_ValueOfOption(DEF, USR, Name), "double") == 0){                                        \
         fprintf(IO_Strom, "Defer %s (double), ", Name);                                                        \
         ArrayEintragCLAS = CLAS_200;                                                                           \
         ArrayEintragRECLAS =  RECLAS_200;                                                                      \
      }                                                                                                         \
      else if (strcmp(IS_ValueOfOption(DEF, USR, Name), "penalty") == 0){                                       \
         fprintf(IO_Strom, "Defer %s (penalty), ", Name);                                                       \
         ArrayEintragCLAS = CLAS_pen;                                                                           \
         ArrayEintragRECLAS =  RECLAS_pen;                                                                      \
      }                                                                                                         \
   }                                                                                                            \
}

void C_InitApriori(void)
{
     CritDefAR[0].Name = "killerR";
   CritDefAR[0].PraedikatsProc = KillerPraedikatR;
   CritDefAR[1].Name = "killerE";
   CritDefAR[1].PraedikatsProc = KillerPraedikatE;
   CritDefAR[2].Name = "killerRE";
   CritDefAR[2].PraedikatsProc = KillerPraedikatRE;
   CritDefAR[3].Name = "echild";
   CritDefAR[3].PraedikatsProc = EChildPraedikat;
   CritDefAR[4].Name = "perm3";
   CritDefAR[4].PraedikatsProc = PermutationsPraedikat3;
   CritDefAR[5].Name = "perm4";
   CritDefAR[5].PraedikatsProc = PermutationsPraedikat4;
   CritDefAR[6].Name = "perm5";
   CritDefAR[6].PraedikatsProc = PermutationsPraedikat5;
   CritDefAR[7].Name = "perm6";
   CritDefAR[7].PraedikatsProc = PermutationsPraedikat6;
}
  
BOOLEAN C_INITIALIZE(InfoStringT clasDEF, InfoStringT clasUSR, 
                     InfoStringT reclasDEF, InfoStringT reclasUSR, 
                     InfoStringT cgDEF, InfoStringT cgUSR, 
                     int *WeightEntryType)
{
   int i, j;
   
   crit[0].DEF = clasDEF;
   crit[0].USR = clasUSR;
   crit[1].DEF = reclasDEF;
   crit[1].USR = reclasUSR;

   fprintf(IO_Strom, "\n* CLASSIFICATION\n\n");

   switch(IS_FindOption(clasDEF, clasUSR, 1, "STD")){
   case 0:
   default:

      SH2_INITIALIZE(clasDEF, clasUSR, WeightEntryType);
      CLAS_0 = SH2_0;
      CLAS_50 = SH2_50;
      CLAS_100 = SH2_100;
      CLAS_200 = SH2_200;
      CLAS_pen = SH2_pen;
      RECLAS_0 = SH2_0_Re;
      RECLAS_50 = SH2_50_Re;
      RECLAS_100 = SH2_100_Re;
      RECLAS_200 = SH2_200_Re;
      RECLAS_pen = SH2_pen_Re;
      
      CGH_INITIALIZE(cgDEF, cgUSR, *WeightEntryType);
      
      C_CG_ClassifyProc = CGH_100;
      
      C_StatisticsProc = SH2_Statistics;
      
      C_CleanUpProc = SH2_CleanUp;

      C_KillAdditionalReferenceProc = StandardKillAddRefProc;

      C_PrintProc = SH2_PrintProc;

      break;
   }

   C_WeightStatProc = CS_WeightStatProc;
   C_WeightStatisticsStartProc = CS_WeightStatisticsStart;
   C_WeightStatisticsEndProc = CS_WeightStatisticsEnd;
   
   CS_INITIALIZE(clasDEF, clasUSR, *WeightEntryType);
      
   CLAS_DEFAULT = CLAS_100;
   RECLAS_DEFAULT = RECLAS_100;

   
   /* KP-bedingte Kriterien etablieren */
   fprintf(IO_Strom, "  Criteria:\n");
   for (j=0;  j<2; j++){
     const char *bla[2] = {"    Classification: ", "    Re-Classification: "};
     fprintf(IO_Strom, bla[j]);
     crit[j].AnzahlAnzuwendendeKriterien = 0;
     crit[j].PermGradBerechnen = FALSE;
     for (i=0; i<AnzahlDefinierterKriterien; i++){
       if (IS_ValueOfOption(crit[j].DEF, crit[j].USR, CritDefAR[i].Name) &&
           ((strcmp(IS_ValueOfOption(crit[j].DEF, crit[j].USR, CritDefAR[i].Name), "half") == 0) ||
            (strcmp(IS_ValueOfOption(crit[j].DEF, crit[j].USR, CritDefAR[i].Name), "ultimate") == 0) ||
            (strcmp(IS_ValueOfOption(crit[j].DEF, crit[j].USR, CritDefAR[i].Name), "double") == 0) ||
            (strcmp(IS_ValueOfOption(crit[j].DEF, crit[j].USR, CritDefAR[i].Name), "penalty") == 0))){
         
         if (j==0) CLAS_DEFAULT = ClassifyWithCriteria;
         if (j==1) RECLAS_DEFAULT = ReClassifyWithCriteria;
         
         crit[j].CritAnwAR[crit[j].AnzahlAnzuwendendeKriterien].PraedikatsProc =  CritDefAR[i].PraedikatsProc;
         crit[j].CritAnwAR[crit[j].AnzahlAnzuwendendeKriterien].Name =  CritDefAR[i].Name;
         
         KriteriumEinrichten(crit[j].DEF, crit[j].USR,
                             CritDefAR[i].Name, crit[j].CritAnwAR[crit[j].AnzahlAnzuwendendeKriterien].ClassifyProc,
                             crit[j].CritAnwAR[crit[j].AnzahlAnzuwendendeKriterien].ReclassifyProc);
         
         crit[j].CritAnwAR[crit[j].AnzahlAnzuwendendeKriterien].Fulfilled = 0;
         
         if ((i>1) && (i<6))
            crit[j].PermGradBerechnen = TRUE;
         crit[j].AnzahlAnzuwendendeKriterien++;
       }
     }
     fprintf(IO_Strom, "\n");
   }

   /* Historien-bedingte Kriterien etablieren */

   fprintf(IO_Strom, "    General: ");

   C_ClassifyProcAR[Initial] = CLAS_DEFAULT;

   KriteriumEinrichten(crit[0].DEF, crit[0].USR, "initial", C_ClassifyProcAR[Initial], RECLAS_Dummy);

   C_ClassifyProcAR[KilledRE] = CLAS_DEFAULT;

   KriteriumEinrichten(crit[0].DEF, crit[0].USR, "killedRE", C_ClassifyProcAR[KilledRE], RECLAS_Dummy);
   
   C_ClassifyProcAR[CPWithFather] = 
      C_ClassifyProcAR[CPWithMother] = CLAS_DEFAULT;
   
   C_ReClassifyProc = RECLAS_DEFAULT;

   fprintf(IO_Strom, "\n");

   return TRUE;
}
/* -------------------------------------------------------------------------------------------------------------*/
/* ------------ ParamInfo --------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

CriteriaT localCrit[2];

#define ParamInfoKriterium(DEF, USR, Name)                               \
{                                                                                                               \
   if (IS_FindValuedOption(DEF, USR, 1, Name) == 0){                                                            \
      if (strcmp(IS_ValueOfOption(DEF, USR, Name), "half") == 0){                                               \
         fprintf(IO_Strom, "Prefer %s half way, ", Name);                                                       \
      }                                                                                                         \
      else if (strcmp(IS_ValueOfOption(DEF, USR, Name), "ultimate") == 0){                                      \
         fprintf(IO_Strom, "Prefer %s ultimate, ", Name);                                                       \
      }                                                                                                         \
      else if (strcmp(IS_ValueOfOption(DEF, USR, Name), "double") == 0){                                        \
         fprintf(IO_Strom, "Defer %s (double), ", Name);                                                        \
      }                                                                                                         \
      else if (strcmp(IS_ValueOfOption(DEF, USR, Name), "penalty") == 0){                                       \
         fprintf(IO_Strom, "Defer %s (penalty), ", Name);                                                       \
      }                                                                                                         \
   }                                                                                                            \
}

void C_ParamInfo(ParamInfoArtT what, int whichModule, 
                 InfoStringT clasDEF, InfoStringT clasUSR, 
                 InfoStringT reclasDEF, InfoStringT reclasUSR, 
                 InfoStringT cgDEF, InfoStringT cgUSR)
{
   int i;
   
   if (what == ParamEvaluate){
      char *localDEF, *localUSR;

      switch(whichModule){                                               /* Ueberschrift */
      case 2:
        localDEF = reclasDEF; 
        localUSR = reclasUSR; 
        fprintf(IO_Strom, "\n* RE-CLASSIFICATION\n\n"); break;
      case 3:   
        localDEF = cgDEF; 
        localUSR = cgUSR; 
        fprintf(IO_Strom, "\n* CG-CLASSIFICATION\n\n"); break;
      case 99:
      case 1:
      default:  
        localDEF = clasDEF; 
        localUSR = clasUSR;
        fprintf(IO_Strom, "\n* CLASSIFICATION\n\n"); break;
      }

      if (whichModule == 99){
        SH2_ParamInfo(ParamEvaluate, clasDEF, clasUSR);
        CGH_ParamInfo(ParamEvaluate, cgDEF, cgUSR);
        CS_ParamInfo(ParamEvaluate, clasDEF, clasUSR);
        /* KP-bedingte Kriterien */
         
        fprintf(IO_Strom, "  Criteria:\n");
        for (i=0; i<AnzahlDefinierterKriterien; i++)
          ParamInfoKriterium(localDEF, localUSR, CritDefAR[i].Name);
        fprintf(IO_Strom, "\n");
        
        /* Historien-bedingte Kriterien */
        
        fprintf(IO_Strom, "    -General: ");
        
        ParamInfoKriterium(localDEF, localUSR, "initial");
        
        ParamInfoKriterium(localDEF, localUSR, "killedRE");
      }
      else{
        if (whichModule==1){
          localDEF = clasDEF; 
          localUSR = clasUSR;
          SH2_ParamInfo(ParamEvaluate, clasDEF, clasUSR);
          CS_ParamInfo(ParamEvaluate, clasDEF, clasUSR);
        }
        else if (whichModule==2){
          localDEF = reclasDEF; 
          localUSR = reclasUSR; 
        }
        else if (whichModule==3){
          localDEF = cgDEF; 
          localUSR = cgUSR; 
          CGH_ParamInfo(ParamEvaluate, cgDEF, cgUSR);
        }

        if ((whichModule < 3) || (whichModule == 99)){     /* clas, reclas, all */
          
          /* KP-bedingte Kriterien */
          
          fprintf(IO_Strom, "  Criteria:\n");
          for (i=0; i<AnzahlDefinierterKriterien; i++)
            ParamInfoKriterium(localDEF, localUSR, CritDefAR[i].Name);
          fprintf(IO_Strom, "\n");
        }
        
        /* Historien-bedingte Kriterien */
        
        fprintf(IO_Strom, "    -General: ");
        
        ParamInfoKriterium(localDEF, localUSR, "initial");
        
        ParamInfoKriterium(localDEF, localUSR, "killedRE");
      }
        fprintf(IO_Strom, "\n\n");
   }
   else{
      if (whichModule == 1){
         fprintf(IO_Strom, "Classification\n");
         fprintf(IO_Strom, "**************\n");
         fprintf(IO_Strom, "This parameter guides the assessment of critical pairs,\n");
         fprintf(IO_Strom, "i.e. the computation of one or more integers (w1/w2) that make up an ordering\n");
         fprintf(IO_Strom, "of the stored cps. This computation of basic weights can be influenced\n");
         fprintf(IO_Strom, "by assigning different weights to single symbols, or by some criteria.\n\n");
            
      }
      else if (whichModule == 2){
         fprintf(IO_Strom, "Re-Classification\n");
         fprintf(IO_Strom, "*****************\n");
         fprintf(IO_Strom, "This parameter guides the criteria to apply on critical pairs during the\n");
         fprintf(IO_Strom, "intermediate reprocessing of the CP-Set. The basic classifying weight is\n");
         fprintf(IO_Strom, "the same as given by the -clas option.\n\n");
      }
      else if (whichModule == 3){
         fprintf(IO_Strom, "CG-Classification\n");
         fprintf(IO_Strom, "*****************\n");
         fprintf(IO_Strom, "This parameter guides the classification of critical goals, and is similar to\n");
         fprintf(IO_Strom, "the -clas option. For critical goals, no criteria are available.\n\n");
         CGH_ParamInfo(ParamExplain, clasDEF, clasUSR);
      }         
      if (whichModule < 3){
         fprintf(IO_Strom, "Criteria that can be applied to cps detect certain properties. If such a property\n");
         fprintf(IO_Strom, "has been detected, the cp will be prefered resp. defered as stated in <treatment>.\n");
         fprintf(IO_Strom, "\n");
         fprintf(IO_Strom, "<criterion>=<treatment>\n");
         fprintf(IO_Strom, "\n");
         fprintf(IO_Strom, "<criterion>=killerR|killerE|killerRE|echild|perm3|perm4|perm5|perm6|initial|killedRE\n");
         fprintf(IO_Strom, "<treatment>=ultimate|half|double|penalty\n");
         fprintf(IO_Strom, "\n");
         fprintf(IO_Strom, "killerR/E/RE  detects those cps that would remove a rule/equation via interreduction\n");
         fprintf(IO_Strom, "echild        detects cps a parent of which is an equation\n");
         fprintf(IO_Strom, "perm4/5/6     detects permutative equations of degree 4/5/6 or higher\n");
         fprintf(IO_Strom, "initial       detects cps from the initial specification\n");
         fprintf(IO_Strom, "killedRE      detects cps that arised from interreduction victims\n\n");
      }
      if (whichModule == 1){
         SH2_ParamInfo(ParamExplain, clasDEF, clasUSR);
         CS_ParamInfo(ParamExplain, clasDEF, clasUSR);
      }
   }
}
