#include "OccNest.h"
#include "TermOperationen.h"
#include "general.h"
#include "Ausgaben.h"
#include "SymbolOperationen.h"
#include "Zielverwaltung.h"

/* interne Konsistenztests, Alternativen 0,1 */
#define ON_TEST 0

/***********************************************************************************/
/********** Globale Typen und Variablen ********************************************/
/***********************************************************************************/

typedef struct GlobalePfadInfoTS{
   int unbehNachf;
   long int *cur;                    /* Ruecksetzinfo fuer alle Symbole */
} GlobalePfadInfoT;

typedef struct SymbolEintragTS{ /* fuer jedes Symbol wird ein solcher Eintrag angelegt  */
   long int cur;                     /* Laenge der bisher angetroffenen Symbolkette */
   long int nest;                    /* Laenge der insgesamt laengsten bisher angetroffenen Symbolkette */
   long int occ;                     /* # Vorkommen des Symbols im Term */
} SymbolEintragT;
   
typedef struct SeitenInfoTS{
   long int *Occ, *Nest;
} SeitenInfoT;



SymbolEintragT *SymbolTabelle;  /* fuer jedes Symbol anzulegen */

GlobalePfadInfoT *GlobalerPfad; /* Array ueber die Tiefe */

static unsigned int OccTiefe = 1;

SeitenInfoT OccInfoLinks, OccInfoZiel; /* zum Sichern von Daten */

/***********************************************************************************/
/********** Basis-Funktionen *******************************************************/
/***********************************************************************************/

#define SymbolInOccnestBeruecksichtigen(/*SymbolT*/ symb) TRUE

#define PSI(/*long int*/ x) ((x<0)? 1:(x+1))

#define PSI2(/*long int*/ x, /*long int*/ y) ((x>y)?(x-y+1):1)

/* Es gilt: PSI(x-y) = PSI2(x,y) */

/***********************************************************************************/
/********** Ausgabe-Funktionen *****************************************************/
/***********************************************************************************/
#if ON_TEST
static void DruckeGlobaleInfo(char *Text, TermT T)
{
   SymbolT symb;
   IO_DruckeFlex("Global. %s, %t\n\n", Text,T);
   fprintf(IO_Strom, "Symbol  Occ   Nest\n-----------------\n");
   SO_forFktSymbole(symb)
      fprintf(IO_Strom, "%6s%5ld%5ld\n", IO_FktSymbName(symb), SymbolTabelle[symb].occ, SymbolTabelle[symb].nest);
}
   
static void DruckeSeitenInfo(char *Text, SeitenInfoT SI)
{
   SymbolT symb;
   fprintf(IO_Strom, "SeitenInfo. %s\n", Text);
   fprintf(IO_Strom, "Symbol  Occ   Nest\n-----------------\n");
   SO_forFktSymbole(symb)
      fprintf(IO_Strom, "%6s%5ld%5ld\n", IO_FktSymbName(symb), SI.Occ[symb], SI.Nest[symb]);
   fprintf(IO_Strom, "\n\n");
}
   
static void Mom(void)
{
   SymbolT symb;
   unsigned long i;

   DruckeSeitenInfo("Ziel", OccInfoZiel);
   DruckeSeitenInfo("Links", OccInfoLinks);
   IO_DruckeFlex("Global.\n");
   fprintf(IO_Strom, "Symbol  Occ Nest  Cur\n---------------------\n");
   SO_forFktSymbole(symb)
      fprintf(IO_Strom, "%6s%5ld%5ld%5ld\n", IO_FktSymbName(symb), 
             SymbolTabelle[symb].occ, SymbolTabelle[symb].nest,
             SymbolTabelle[symb].cur);
   fprintf(IO_Strom, "Globaler Pfad\n");
   fprintf(IO_Strom, "Tiefe | UnBHN Cur:");
   SO_forFktSymbole(symb)
      fprintf(IO_Strom, "%6s", IO_FktSymbName(symb));
   fprintf(IO_Strom, "\n");
   fprintf(IO_Strom, "------------------");
   SO_forFktSymbole(symb)
      fprintf(IO_Strom, "------");
   fprintf(IO_Strom, "\n");
   for (i=0;i<15;i++){
      fprintf(IO_Strom, "%5ld | %5d     ", i, GlobalerPfad[i].unbehNachf);
      SO_forFktSymbole(symb)
         fprintf(IO_Strom, " %5ld", GlobalerPfad[i].cur[symb]);
      fprintf(IO_Strom, "\n");
   }
   fprintf(IO_Strom, "\nOccTiefe = %d\n\n", OccTiefe);
}

#endif

/***********************************************************************************/
/********** Zielanmeldung **********************************************************/
/***********************************************************************************/
static void OccNestSchleife(TermT term);

static void ZielIst(TermT term1, TermT term2)
{
   SymbolT Index;

   OccNestSchleife(term1);

   SO_forUnkonstSymbole(Index){
      OccInfoLinks.Occ[Index] = SymbolTabelle[Index].occ;                                      /* Occ sichern und loeschen */
      SymbolTabelle[Index].occ = 0;
      OccInfoLinks.Nest[Index] = SymbolTabelle[Index].nest;                                   /* Nest sichern und loeschen */
      SymbolTabelle[Index].nest=0;
   }
   SO_forKonstSymbole(Index){
      OccInfoLinks.Occ[Index] = SymbolTabelle[Index].occ;                                      /* Occ sichern und loeschen */
      SymbolTabelle[Index].occ = 0;
   }

   if (OccTiefe!=1) fprintf(IO_Strom, "OccTiefe = %d\n",OccTiefe);

   OccNestSchleife(term2);

   SO_forFktSymbole(Index){
      OccInfoZiel.Occ[Index] = max(SymbolTabelle[Index].occ, OccInfoLinks.Occ[Index]);
      SymbolTabelle[Index].occ = 0;
      OccInfoZiel.Nest[Index] = max(SymbolTabelle[Index].nest, OccInfoLinks.Nest[Index]);
      SymbolTabelle[Index].nest= 0;
      SymbolTabelle[Index].cur = 0;
   }
   /*   fprintf(IO_Strom, "*************************Occ-Nest-ZIEL****************************\n");
   IO_DruckeFlex("Terme: %t %t\n\n", term1, term2);
   if (OccTiefe != 1) fprintf(IO_Strom, "OccTiefe != 1\n");*/
   OccTiefe = 1;
}


/***********************************************************************************/
/********** Occ- und Nest-Berechnung auf Termen ************************************/
/***********************************************************************************/

/* Iterative Funktion zur Berechnung von nest und occ(auf allen Symbolen).
   Nest bestimmt die groesste Zahl von hintereinanderfolgenden Vorkommen von f in Aesten vonT, wobei
   T als Baum gesehen wird. 
*/

static void OccNestSchleife(TermT term)
{
   SymbolT symb, Index;

   TO_doSymbole(term, symb,{

      /* aktuelles Symbol behandeln */

      if (SO_SymbIstVar(symb))
         SO_forUnkonstSymbole(Index)
            SymbolTabelle[Index].cur = 0;
      else{
         SymbolTabelle[symb].occ++;

         SO_forUnkonstSymbole(Index)
            if (SO_SymbUngleich(Index,symb))
              SymbolTabelle[Index].cur = 0;

         if (SO_Stelligkeit(symb)){ /*d. h. Symbol ist keine Konstante */
            SymbolTabelle[symb].cur++;
            SymbolTabelle[symb].nest = max(SymbolTabelle[symb].cur,SymbolTabelle[symb].nest);
         }

      }
      /* Pfad-Position des folgenden Symbols bestimmen: */

      GlobalerPfad[OccTiefe].unbehNachf = SO_Stelligkeit(symb);

      if (GlobalerPfad[OccTiefe].unbehNachf > 0){ /* es gibt noch unbeh. Nachf => naechstes Symbol ist Nachfolger von symb */
         GlobalerPfad[OccTiefe].unbehNachf--;     /* ein Nachf wird dann behandelt */
         SO_forUnkonstSymbole(Index)
            GlobalerPfad[OccTiefe].cur[Index] = SymbolTabelle[Index].cur; /* aktuelle Werte zum Ruecksetzen merken */
         OccTiefe++;                      /* Abstieg */
      }
      else{       /* symb war nullstellig oder alle Teilterme abgearbeitet => naechstes Symbol ist Vorgaenger davon */
         do{                    /* Also Aufstieg, bis unbeh. Geschwister oder Wurzel erreicht */
            OccTiefe--;
         }while((OccTiefe>0) && (GlobalerPfad[OccTiefe].unbehNachf==0));
         GlobalerPfad[OccTiefe].unbehNachf--; /* einer der bisher unbeh. Nachf. wird jetzt beh. */
         SO_forUnkonstSymbole(Index)
            SymbolTabelle[Index].cur = GlobalerPfad[OccTiefe].cur[Index];    /* akt. Werte zuruecksetzen */
         OccTiefe++;                  /* und absteigen */
      }
   });
}


static long int OccNest(TermT term1, TermT term2, long int Phi)
{
   SymbolT Index;

   OccNestSchleife(term1);

   /* Werte sichern: */
   /*    DruckeGlobaleInfo("Linke Seite", term1); */

   SO_forUnkonstSymbole(Index){
      OccInfoLinks.Occ[Index] = SymbolTabelle[Index].occ;                                      /* Occ sichern und loeschen */
      SymbolTabelle[Index].occ = 0;
      OccInfoLinks.Nest[Index] = SymbolTabelle[Index].nest;                                   /* Nest sichern und loeschen */
      SymbolTabelle[Index].nest=0;
   }
   SO_forKonstSymbole(Index){
      OccInfoLinks.Occ[Index] = SymbolTabelle[Index].occ;                                      /* Occ sichern und loeschen */
      SymbolTabelle[Index].occ = 0;
   }

   if (OccTiefe!=1) fprintf(IO_Strom, "OccTiefe = %d\n",OccTiefe);

   /*    DruckeSeitenInfo("Gerettet als", OccInfoLinks); */

   OccNestSchleife(term2);

   SO_forUnkonstSymbole(Index){
      if(SymbolInOccnestBeruecksichtigen(Index)){
         OccInfoLinks.Occ[Index] = max(OccInfoLinks.Occ[Index], SymbolTabelle[Index].occ);
         OccInfoLinks.Nest[Index] = max(OccInfoLinks.Nest[Index], SymbolTabelle[Index].nest);
            Phi *= PSI2(OccInfoLinks.Occ[Index], OccInfoZiel.Occ[Index])
            * PSI2(OccInfoLinks.Nest[Index], OccInfoZiel.Nest[Index]);
      }
      SymbolTabelle[Index].occ = 0;
      SymbolTabelle[Index].nest= 0;
   }
   SO_forKonstSymbole(Index){
      if(SymbolInOccnestBeruecksichtigen(Index)){
         OccInfoLinks.Occ[Index] = max(OccInfoLinks.Occ[Index], SymbolTabelle[Index].occ);
         Phi *= PSI2(OccInfoLinks.Occ[Index], OccInfoZiel.Occ[Index]);
      }
      SymbolTabelle[Index].occ = 0;
   }
/*   if (OccTiefe!=1) fprintf(IO_Strom, "OccTiefe = %d\n",OccTiefe);*/
   return Phi;
}

TermT aktuellesKPlinks, aktuellesKPrechts; /* Terme des zu bewertenden KPs */
long int aktuellesPhi,          /* kleinstes bisher erzieltes Phi */
  aktuellesBasisPhi;            /* Basis-Wert des KPs */

static void OccNestEinZiel(TermT term1, TermT term2, unsigned long number)
{
  long int localPhi;
  ZielIst(term1, term2);
  localPhi = OccNest(aktuellesKPlinks, aktuellesKPrechts, aktuellesBasisPhi);
  if (localPhi < aktuellesPhi) aktuellesPhi = localPhi;
}

long int ON_OccNest(TermT term1, TermT term2, long int Phi)
     /* Berechnet das Minimum aller OccNest-Werte des KPs mit allen Zielen. */
{
  aktuellesKPlinks = term1, aktuellesKPrechts = term2;
  aktuellesPhi = LONG_MAX;
  aktuellesBasisPhi = Phi;

  Z_ZielmengenBeobachtungsDurchlauf(OccNestEinZiel);
  
  return aktuellesPhi;
}

void ON_Init(void)
{
   SymbolT Index;
   unsigned long i;
   
   OccInfoLinks.Occ = our_alloc(SO_AnzahlFunktionen()*sizeof(long int));
   OccInfoLinks.Nest = our_alloc(SO_AnzahlFunktionen()*sizeof(long int));
   OccInfoZiel.Occ = our_alloc(SO_AnzahlFunktionen()*sizeof(long int));
   OccInfoZiel.Nest = our_alloc(SO_AnzahlFunktionen()*sizeof(long int));
   SymbolTabelle = our_alloc(SO_AnzahlFunktionen()*sizeof(SymbolEintragT));
   GlobalerPfad = our_alloc(100*sizeof(GlobalePfadInfoT));
   for(i=0; i<100; i++){
      GlobalerPfad[i].cur = our_alloc(SO_AnzahlFunktionen()*sizeof(long int));
      SO_forFktSymbole(Index)
         GlobalerPfad[i].cur[Index] = 0;
      GlobalerPfad[i].unbehNachf = 0;
   }
   SO_forFktSymbole(Index){
      SymbolTabelle[Index].cur = 0;
      SymbolTabelle[Index].nest = 0;
      SymbolTabelle[Index].occ = 0;
      OccInfoLinks.Occ[Index]=0;
      OccInfoLinks.Nest[Index]=0;
      OccInfoZiel.Nest[Index]=0;
      OccInfoZiel.Occ[Index]=0;
   }
   OccTiefe = 1;
}
