#include    <stdio.h>
#include    "NFBildung.h"
#include    "ZeitMessung.h"
#include    "Referee.h"
#include    "RUndEVerwaltung.h"
#include    "Ausgaben.h"
#include    "Termpaare.h"
#include    "SUEM.h"

/*
//------------------------------------------------------------------------
//      globale Variablen
//------------------------------------------------------------------------  
*/

report   REF_referee_report, *referee_report;

unsigned long int altG=0, altCG=0, altCPGueltig=0, altCPAktuell=0, altCPGesamt=0,
  altRGesamt=0, altEGesamt=0, altRAktuell=0, altEAktuell=0,
  altAnzRed = 0;

/*
//-----------------------------------------------------------------------------
//      Nach RE zu verschiebende RE-Operationen
//-----------------------------------------------------------------------------
*/

#define State(RE) 0
#define INITIAL 0



typedef enum{StatRef, NoRef} RefereeT;
typedef enum{LastSel, StatSel, NoSel} SelectionT;

RefereeT REF_Referee;
SelectionT REF_Selection;


#define LiesParameter(par) if (IS_ValueOfOption(DEF,USR,#par)) par=atoi(IS_ValueOfOption(DEF,USR,#par))

/*
//-----------------------------------------------------------------------------
//      lokale Funktiondefinitionen
//-----------------------------------------------------------------------------
*/

/* --------------- Basisfunktionen zur Aufnahme gewichteter Regeln/Gleichungen in Bestenliste */
static void     InsertRefRule ( RegelOderGleichungsT rule ); /* Einfuegen einer Regel in die Bestenliste */
static void     InsertRefEqu  ( RegelOderGleichungsT rule ); /* Einfuegen einer Gleichung in die Bestenliste */


/*
//=============================================================================
//      Regeln und Gleichungen aufsammeln
//=============================================================================
*/

static RegelOderGleichungsT *rulesAR, *equsAR;
static int maxR, maxE, anzR, anzE;

#define InsertObject(ROG, AR, ANZ, MAX)                                         \
{                                                                               \
    short   i;                                                                  \
    if (ANZ < MAX) AR[(ANZ)++] = ROG;                                           \
                                                                                \
    for (i = (ANZ)-2; i >=0; i--) /* einsortieren zwischen bisherige ROG */     \
      if (TP_Weight(AR[i]) > TP_Weight(ROG)){                                   \
        AR[i+1] = ROG;                                                          \
        return;                                                                 \
      }                                                                         \
      else AR[i+1] = AR[i];                                                     \
    AR[0] = ROG;                                                                \
}

static void   InsertRefRule ( RegelOderGleichungsT rule )
{
  InsertObject(rule, rulesAR, anzR, maxR);
}


static void   InsertRefEqu ( RegelOderGleichungsT equ )
{
  InsertObject(equ, equsAR, anzE, maxE);
}


static void OutputRulesEqus(char *description)
{
  int i;
  fprintf(IO_Strom,  "The %s %d rules.\n", description, anzR );
  for (i = 0; i < anzR; i++)
    IO_DruckeFlex("%d %t = %t\n", TP_Weight(rulesAR[i]),
                  TP_LinkeSeite(rulesAR[i]),
                  TP_RechteSeite(rulesAR[i]));

  fprintf(IO_Strom,  "\nThe %s %d equations.\n", description, anzE );
  for (i = 0; i < anzE; i++)
    IO_DruckeFlex("%d %t = %t\n", TP_Weight(equsAR[i]),
                  TP_LinkeSeite(equsAR[i]),
                  TP_RechteSeite(equsAR[i]));

  fprintf(IO_Strom,  "\n-------------------------------------------------------------------\n" );
}

/*
//-----------------------------------------------------------------------------
//  REFEREES:
//-----------------------------------------------------------------------------
*/

#define ZeileDrucken(str,anz,fac)                                                               \
{                                                                                               \
  fprintf(IO_Strom,  "    Number of %31s%9ld * %6ld = %15ld\n", str, anz, fac, anz*fac);       \
  result+= anz*fac;                                                                             \
}

long int REF_StatReferee(long int sizeR, long int sizeE, long int sizeCP, long int sizeG,
                         long int sizeCG, long int newR, long int redR, long int remR,
                         long int newE, long int redE, long int remE, long int newCP,
                         long int newG, long int redG, long int newCG, long int red)
{
    long int result = 0;
    unsigned long int AnzRegelnAktuell=0, AnzGleichungenAktuell=0, SUE_AnzKPsAktuell=0,
      SUE_AnzKPsGueltig=0, SUE_AnzKPsGesamt=0, AnzZieleAktuell=0, AnzkritZieleAktuell=0, 
      AnzRedRegeln=0, AnzRemRegeln=0, AnzRedGleichungen=0,
      AnzRemGleichungen=0, AnzRedZiele=0,
      RedCount=0;
    long int AnzNeueRegeln, AnzNeueGleichungen, AnzNeueKPs=0, AnzNeueZiele, AnzNeueCG;

    RE_AktuelleGroessen(&AnzRegelnAktuell, &AnzGleichungenAktuell);
    SUE_AktuelleGroesse(&SUE_AnzKPsGesamt, &SUE_AnzKPsAktuell, &SUE_AnzKPsGueltig);
    NF_AnzReduktionenGesamt(&RedCount);

    AnzNeueRegeln      = RE_AnzRegelnGesamt - altRGesamt;                    /* Neu gebildete Regeln */
    AnzNeueGleichungen = RE_AnzGleichungenGesamt - altEGesamt;               /* Neu gebildete Gleichungen */

    AnzRemRegeln = altRAktuell + AnzNeueRegeln - AnzRegelnAktuell;           /* Entfernte Regeln */
    AnzRemGleichungen = altEAktuell + AnzNeueGleichungen - AnzGleichungenAktuell; /* Entfernte Gleichungen */

    AnzNeueKPs= SUE_AnzKPsGesamt - altCPGesamt;
    AnzNeueZiele=0;
    AnzNeueCG=0;

    IO_DruckeUeberschrift("Statistical Referee");

    fprintf(IO_Strom,  "    Assessment of current system\n" );
    fprintf(IO_Strom,  "    ----------------------------\n\n" );

    ZeileDrucken("Rules:",                         AnzRegelnAktuell,      sizeR);
    ZeileDrucken("-New rules:",                   AnzNeueRegeln,         newR);
    ZeileDrucken("-Reduced rules (---):",             AnzRedRegeln,          redR);
    ZeileDrucken("-Removed rules:",              AnzRemRegeln,          remR);
    ZeileDrucken("Equations:",                    AnzGleichungenAktuell, sizeE);
    ZeileDrucken("-New equations:",              AnzNeueGleichungen,    newE);
    ZeileDrucken("-Reduced Equations (---):",        AnzRedGleichungen,     redE);
    ZeileDrucken("-Removed Equations:",         AnzRemGleichungen,     remE);
    ZeileDrucken("Critical pairs:",               SUE_AnzKPsAktuell,     sizeCP);
    ZeileDrucken("-New critical pairs:",         AnzNeueKPs,            newCP);
    ZeileDrucken("Goals (---):",                    AnzZieleAktuell,       sizeG);
    ZeileDrucken("-New Goals (---):",                    AnzNeueZiele,          newG);
    ZeileDrucken("-Reduced Goals (---):",              AnzRedZiele,           redG);
    ZeileDrucken("Critical goals (---):",              AnzkritZieleAktuell,   sizeCG);
    ZeileDrucken("-New critical Goals (---):",              AnzNeueCG,             newCG);
    ZeileDrucken("Reductions:",                    (RedCount-altAnzRed),              red);
    fprintf(IO_Strom,  "\n" );
    return result;
}

long int REF_GenericReferee(InfoStringT DEF, InfoStringT USR)
{
  unsigned long int    runtime, cputime;
  long int result = 0;

  ZM_Times(&cputime, &runtime);

  switch(IS_FindOption(DEF,USR,1, "refstat")){
  case 0: {
    long int sizeR=0, sizeE=0,sizeCP=0,sizeG=0,sizeCG=0,newR=0,redR=0,remR=0,
      newE=0,redE=0,remE=0,newCP=0,newG=0,redG=0,newCG=0,red=0;

    LiesParameter(sizeR); LiesParameter(sizeE); LiesParameter(sizeCP);
    LiesParameter(sizeG); LiesParameter(sizeCG); LiesParameter(newR);
    LiesParameter(redR); LiesParameter(remR); LiesParameter(newE);
    LiesParameter(redE); LiesParameter(remE); LiesParameter(newCP);
    LiesParameter(newG); LiesParameter(redG); LiesParameter(newCG);
    LiesParameter(red);

    result= REF_StatReferee(sizeR, sizeE, sizeCP, sizeG, sizeCG, newR, redR, remR,
                         newE,redE, remE, newCP, newG, redG, newCG, red);
    break;
  }
  case -1:
  default:
    IO_DruckeUeberschrift("Empty Referee");
    result = -1000000;
  }
  fprintf(IO_Strom,  "    Bewertung:                                      %10ld\n\n", result );
  fprintf(IO_Strom,  "    CPU-Auslastung:   %3ld %%\n\n",  cputime*100/runtime);
  fprintf(IO_Strom,  "\n-------------------------------------------------------------------\n" );
  return result;
}

/*
//-----------------------------------------------------------------------------
//  AUSWAHL:
//-----------------------------------------------------------------------------
*/

#define SelectionInit()                         \
  rulesAR = rules;                              \
  equsAR = equs;                                \
  maxR = AnzRegeln;                             \
  maxE = AnzGleichungen;                        \
  anzR = 0;                                     \
  anzE = 0;


/*
//-----------------------------------------------------------------------------
//  Funktion:       LastSelect
//-----------------------------------------------------------------------------
*/
void REF_LastSelect(unsigned int         AnzRegeln, 
                    unsigned int         AnzGleichungen,
                    unsigned int         *rule_count,
                    RegelOderGleichungsT *rules,
                    unsigned int         *equ_count,
                    RegelOderGleichungsT *equs)

/* Liefert die letzten hoechstens AnzRegeln Regeln und hoechstens AnzGleichungen Gleichungen in
   rules[0..rule_count] bzw. equs[0..equ_count].
   Die Arrays rules bzw. equs. muessen mit der Groesse AnzRegeln bzw. AnzGleichungen
   bereits angelegt sein.
*/
{
  RegelOderGleichungsT rog;

  SelectionInit();

  RE_forRegelnRobust(rog, {
    if (!((maxR==0) || (State (rog) & INITIAL))){
      TP_SetWeight(rog,rog->Nummer);
      if ((anzR < maxR)||(TP_Weight(rulesAR[maxR-1]) < TP_Weight(rog)))
        InsertRefRule ( rog );
    }
  });
  RE_forGleichungenRobust(rog, {
    if (!((maxE==0) || (State (rog) & INITIAL)) && (TP_RichtungAusgezeichnet(rog))){
      TP_SetWeight(rog,-(rog->Nummer));
      if ((anzE < maxE)||(TP_Weight(equsAR[maxE-1]) < TP_Weight(rog)))
        InsertRefEqu ( rog );
    }
  });

  *rule_count = anzR;
  *equ_count = anzE;
  OutputRulesEqus("last");
}


/*
//-----------------------------------------------------------------------------
//  Funktion:       StatisticSelect
//-----------------------------------------------------------------------------
*/

#define TestStat(){                                                                     \
  TP_SetWeight(rog,                                                                     \
               (anzRed       *(TP_RedCp(rog)+ TP_RedGoal (rog)+ TP_RedEqu(rog) +        \
                               TP_RedLeft (rog) + TP_RedRight(rog)))                    \
               + (anzRedrhsR * TP_RedRight(rog))                                        \
               + (anzRedlhsR * TP_RedLeft(rog))                                         \
               + (anzRedE    * TP_RedEqu(rog))                                          \
               + (anzRedG    * TP_RedGoal(rog))                                         \
               + (anzSub     * TP_Subsum(rog))                                          \
               + (anzKPs     * TP_CpCount(rog)));                                       \
}

void REF_StatSelect(unsigned int AnzRegeln, 
                    unsigned int AnzGleichungen,
                    unsigned int    *rule_count,
                    RegelOderGleichungsT *rules,
                    unsigned int    *equ_count,
                    RegelOderGleichungsT *equs,
                    long int anzRed,
                    long int anzRedrhsR,
                    long int anzRedlhsR,
                    long int anzRedE,
                    long int anzRedG,
                    long int anzSub,
                    long int anzKPs,
                    long int anzCGs,
                    long int schwellwert)
/* Liefert wie LastSelect hoechstens .....
   Dabei findet die Auswahl aber ueber speziellere Bewertungen statt (Diss von Joerg, S. 122),
   wobei schwellwert nicht unterschritten werden darf.
   */
{
  RegelOderGleichungsT rog;
  
  SelectionInit();
  
  RE_forRegelnRobust(rog, {
    if (!((maxR==0)||(State (rog) & INITIAL))){
      TestStat();
      if ((TP_Weight(rog) >= schwellwert)&&
          ((anzR < maxR) || (TP_Weight(rulesAR[maxR-1]) < TP_Weight(rog))))
        InsertRefRule ( rog );
    }
  } );
  RE_forGleichungenRobust(rog, {
    if (!((maxR==0)||(State (rog) & INITIAL))){
      TestStat();        
      if ((TP_Weight(rog) >= schwellwert)&&
          ((anzE < maxE) || (TP_Weight(equsAR[maxE-1]) < TP_Weight(rog))))
        InsertRefEqu ( rog );
    }
  } );
  *rule_count = anzR;
  *equ_count = anzE;
  OutputRulesEqus("best");
}


void REF_GenericSelect(unsigned int         AnzRegeln, 
                       unsigned int         AnzGleichungen,
                       unsigned int         *rule_count,
                       RegelOderGleichungsT *rules,
                       unsigned int         *equ_count,
                       RegelOderGleichungsT *equs,
                       InfoStringT DEF, 
                       InfoStringT USR)
/* Waehlt eine der vorstehenden Select-Funktionen aus und bestueckt die Parameter entsprechend.
   Syntax: [sellast|selstat]:anzRegeln=..:...:anzRed=...:schwellwert=....
   */
{
  switch(IS_FindOption(DEF,USR,2, "sellast", "selstat")){
  case 0:                       /* last */
    REF_LastSelect(AnzRegeln, AnzGleichungen, rule_count, rules, equ_count, equs);
    break;
  case 1:                       /* stat */
    {
      long int anzRed=0, anzRedrhsR=0, anzRedlhsR=0, anzRedE=0, 
        anzRedG=0, anzSub=0, anzKPs=0, anzCGs=0, schwellwert=0;

      LiesParameter(anzRed);
      LiesParameter(anzRedrhsR);
      LiesParameter(anzRedlhsR);
      LiesParameter(anzRedE);
      LiesParameter(anzRedG);
      LiesParameter(anzSub);
      LiesParameter(anzKPs);
      LiesParameter(anzCGs);
      LiesParameter(schwellwert);
      REF_StatSelect(AnzRegeln, AnzGleichungen, rule_count, rules, equ_count, equs,
                     anzRed, anzRedrhsR, anzRedlhsR, anzRedE, anzRedG, anzSub, anzKPs, 
                     anzCGs, schwellwert);
    }
    break;
  case -1:                      /* none o. ae. */
  default:
    *rule_count=anzR=0;
    *equ_count=anzE=0;
    OutputRulesEqus("none");
    break;
  }
}
  

void REF_SetzeStartPunkt(void)
{
  RegelOderGleichungsT rog;

  altRGesamt = RE_AnzRegelnGesamt;
  altEGesamt = RE_AnzGleichungenGesamt;

  NF_AnzReduktionenGesamt(&altAnzRed);
  
  RE_AktuelleGroessen(&altRAktuell, &altEAktuell);
  SUE_AktuelleGroesse(&altCPGesamt, &altCPAktuell, &altCPGueltig);
  altG  = (long int) 0;
  altCG = (long int) 0;

  RE_forRegelnRobust(rog, {TP_SetWeightZero(rog);});
  RE_forGleichungenRobust(rog, {TP_SetWeightZero(rog);});
}
