#include "glamc.h"
#include "codegen.h"

/* alle Parsefunktionen: */
void parsePrg(Tinp inp, Toutp outp);
void parseDef(Tinp inp, Toutp outp);
void parseExp(Tinp inp, Toutp outp);
void parseBor(Tinp inp, Toutp outp);
void parseBnd(Tinp inp, Toutp outp);
void parseRel(Tinp inp, Toutp outp);
void parseSum(Tinp inp, Toutp outp);
void parseFac(Tinp inp, Toutp outp);
void parseIdent(Tinp inp, Toutp outp);
void parseIf(Tinp inp, Toutp outp);
void parseWhile(Tinp inp, Toutp outp);
void parseLet(Tinp inp, Toutp outp);

void prepare(Tinp inp, Toutp outp){
  emit(outp, LDC, 0); /* die Programme werden weniger als 2^24-1 Worte lang sein... */
  emit(outp, JSR, 0);
  emit(outp, STP, 0);
}

void finish(Tinp inp, Toutp outp){
  patch(outp, 0, LDC, getFctAdr(inp, "main"));
}

void parsePrg(Tinp inp, Toutp outp){
  parseDef(inp,outp);
  while(getToken(inp)==Def){
    printf("---------------------\n");
    parseDef(inp,outp);
  }
  acceptToken(Eos,inp);
}

void parseDef(Tinp inp, Toutp outp){
  int noV = 0;

  acceptToken(Def, inp);

  if (getToken(inp)== Id){
    tokenIsFun(inp,pc2glam_f(getpc(outp)));
    eatToken(inp);
  }
  else {
    printf("Error: Function Identifier expected!\n");
    exit(9);
  }
  acceptToken(Lpar,inp);
   
  if(getToken(inp)== Id){
    tokenIsVar(inp);noV++;
    eatToken(inp);
    while(maybeToken(Com,inp)){
      if(getToken(inp)== Id){
        tokenIsVar(inp);noV++;
        eatToken(inp);
      }
      else {
        printf("Error: Parameter Identifier expected!\n");
        exit(9);
      }
    }
  }

  acceptToken(Rpar,inp);
  setArity(inp,noV);
  acceptToken(Eq,inp);

  parseExp(inp,outp);
  
  acceptToken(Fed,inp);
  emit(outp,RET,noV);
  popVars(inp,noV);
}

void parseExp(Tinp inp, Toutp outp){
  parseBor(inp,outp);

  while (getToken(inp) == Lor){
    eatToken(inp);
    parseBor(inp,outp);
    emit(outp,LOR,0);
  }
}

void parseBor(Tinp inp, Toutp outp){
  parseBnd(inp,outp);

  while (getToken(inp) == Land){
    eatToken(inp);
    parseBnd(inp,outp);
    emit(outp,LAND,0);
  }
}

void parseBnd(Tinp inp, Toutp outp){
  Op op;

  parseRel(inp,outp);
  
  switch (getToken(inp)){
  case Eqv: op = EQ;  break;
  case Neq: op = NEQ; break;
  case Lt:  op = LT;  break;
  case Leq: op = LEQ; break;
  case Gt:  op = GT;  break;
  case Geq: op = GEQ; break;
  default:  op = NOP; break;
  }
  if (op != NOP){
    eatToken(inp);
    parseRel(inp,outp);
    emit(outp,op,0);
  }
}

void parseRel(Tinp inp, Toutp outp){
  Token tk;

  if (maybeToken(Min,inp)){
    if (getToken(inp) == Num){
      emitConst(outp,-getTokenVal(inp));
      eatToken(inp);
    }
    else {
      parseSum(inp,outp);
      emit(outp,NEG,0);
    }
  }
  else {
    parseSum(inp,outp);
  }

  tk = getToken(inp);
  while ((tk == Pls) ||(tk == Mns)){
    eatToken(inp);
    parseSum(inp,outp);
    if (tk == Pls){
      emit(outp,ADD,0);
    }
    else {
      emit(outp,SUB,0);
    }
    tk=getToken(inp);
  }
}

void parseSum(Tinp inp, Toutp outp){
  Token tk;

  parseFac(inp,outp);

  tk = getToken(inp);
  while ((tk == Mul) || (tk == Div) || (tk == Mod)){
    eatToken(inp);
    parseFac(inp,outp);
    if (tk == Mul){
      emit(outp,MUL,0);
    }
    else if(tk == Div){
      emit(outp,DIV,0);
    }
    else {
      emit(outp,MOD,0);
    }
    tk=getToken(inp);
  }
}

void parseFac(Tinp inp, Toutp outp){
  glam_c code;

  switch (getToken(inp)){
  case Id: 
    parseIdent(inp,outp); 
    break;
  case Num:
    emitConst(outp,getTokenVal(inp));
    eatToken(inp);
    break;
  case If: 
    parseIf(inp,outp); 
    break;
/*  case While: 
    parseWhile(inp,outp); 
    break;
*/
  case Let: 
    parseLet(inp,outp); 
    break;
  case Lpar: 
    acceptToken(Lpar,inp);
    parseExp(inp,outp);
    acceptToken(Rpar,inp);     
    break;
/*  case Lbrc:  
    acceptToken(Lbrc,inp);
    parseExp(inp,outp);
    while(maybeToken(Com,inp)){
      parseExp(inp,outp);
    }
    acceptToken(Rbrc,inp);
    break;
*/
  case Min:
  case Max:
    code = (getToken(inp) == Min) ? MIN : MAX;
    eatToken(inp);
    acceptToken(Lpar,inp);
    parseExp(inp,outp);
    while(maybeToken(Com,inp)){
      parseExp(inp,outp);
      emit(outp,code,0);
    }
    acceptToken(Rpar,inp);
    break;
  default: 
    printf("Parse-Error\n");
    exit(9);
  }
}

void parseIdent(Tinp inp, Toutp outp){
  int no = 0;

  switch (getIdType(inp)){
  case Unknown:
    printf("Unkown Identifier >%s< used!\n", getTokenText(inp));
    exit(9);
    break;
  case Constant:
    emitConst(outp,getTokenVal(inp));
    eatToken(inp);
    return;
    break;
  case Variable:
    emit(outp,LDV,getVarAdr(inp));
    eatToken(inp);
    break;
  case Function:
  case Builtin:
    emitConst(outp,(glam_c)getAdr(inp));
    eatToken(inp);
    break;
  default:
    printf("Implementation Error!\n");
    exit(9);
    break;
  }
  /* jetzt kommt evtl noch der Applikationsteil (nur fuer Var, Fct, Bltin) */
  if (maybeToken(Lpar,inp)){
    if (getToken(inp) != Rpar){ 
      parseExp(inp,outp);no++;
      while (maybeToken(Com,inp)){
        parseExp(inp,outp);no++;
      }
    }
    acceptToken(Rpar,inp);
    emit(outp,JSR,no);
  }
}
                
void parseIf(Tinp inp, Toutp outp){
  int l0rs[100];  /* Ja, das ist eine Implementationskonstante! */
  int nl0rs = 0;
  int lir = 0;
  int i;

  acceptToken(If, inp);
  
  parseExp(inp,outp);
  
  acceptToken(Then, inp);  
  lir=getpc(outp);
  emit(outp,BRF,0); 

  parseExp(inp,outp);
  
  while(maybeToken(Elsif, inp)){
    l0rs[nl0rs++]=getpc(outp);
    emit(outp,JMP,0); 
    patch(outp,lir,BRF,getpc(outp)-lir);
    
    parseExp(inp,outp);
    
    acceptToken(Then, inp);  
    lir=getpc(outp);
    emit(outp,BRF,0); 

    parseExp(inp,outp);
  }
	
  acceptToken(Else, inp);
  l0rs[nl0rs++]=getpc(outp);
  emit(outp,JMP,0); 
  patch(outp,lir,BRF,getpc(outp)-lir);
    
  parseExp(inp,outp);
    
  acceptToken(Fi, inp);
  for(i=0; i<nl0rs; i++){
    patch(outp,l0rs[i],JMP,getpc(outp));
  }
}

void parseWhile(Tinp inp, Toutp outp){
  int l0r = 0;
  int l1 = 0;

  acceptToken(While, inp);
  l1 = getpc(outp);
  
  parseExp(inp,outp);
  
  acceptToken(Do, inp);  
  emit(outp,BRF,0); l0r=getpc(outp);

  parseExp(inp,outp);
  
  acceptToken(Od, inp);
  emit(outp,JMP,l1);
  patch(outp,l0r,BRF,getpc(outp)); 
}

void parseLet(Tinp inp, Toutp outp){
  int noV = 0;

  acceptToken(Let, inp);

  if(getToken(inp)== Id){
    tokenIsVar(inp);noV++;
    eatToken(inp);
  }
  else {
    printf("Error: Variable Identifier expected!\n");
    exit(9);
  }
  emit(outp,ALV,1);

  acceptToken(Eq, inp);
  
  parseExp(inp,outp);
  emit(outp,STV,1);

  while(maybeToken(Com, inp)){
    if(getToken(inp)== Id){
      tokenIsVar(inp);noV++;
      eatToken(inp);
    }
    else {
      printf("Error: Variable Identifier expected!\n");
      exit(9);
    }
    emit(outp,ALV,1);
    acceptToken(Eq, inp);
    
    parseExp(inp,outp);
    emit(outp,STV,1);
  }
 
  acceptToken(In,inp);

  parseExp(inp,outp);

  acceptToken(Tel,inp);
  emit(outp,DAV,noV);
  popVars(inp,noV);
}

