#include <ctype.h>
#include <string.h>
#include "scan.h"

#define TTEXT_INITAL_LEN 20
#define DELTA_TTEXTLEN   10

envEntry kws[] = {
  {NULL,      "def",   Keyword, Def,   0, NULL, 0},
  {&(kws[0]),  "fed",   Keyword, Fed,   0, NULL, 0},
  {&(kws[1]),  "if",    Keyword, If,    0, NULL, 0},
  {&(kws[2]),  "then",  Keyword, Then,  0, NULL, 0},
  {&(kws[3]),  "elsif", Keyword, Elsif, 0, NULL, 0},
  {&(kws[4]),  "else",  Keyword, Else,  0, NULL, 0},
  {&(kws[5]),  "fi",    Keyword, Fi,    0, NULL, 0},
  {&(kws[6]),  "let",   Keyword, Let,   0, NULL, 0},
  {&(kws[7]),  "in",    Keyword, In,    0, NULL, 0},
  {&(kws[8]),  "tel",   Keyword, Tel,   0, NULL, 0},
  {&(kws[9]),  "while", Keyword, While, 0, NULL, 0},
  {&(kws[10]), "do",    Keyword, Do,    0, NULL, 0},
  {&(kws[11]), "od",    Keyword, Od,    0, NULL, 0},
  {&(kws[12]), "min",   Keyword, Min,   0, NULL, 0},
  {&(kws[13]), "max",   Keyword, Max,   0, NULL, 0}
};

#define lastKW (&(kws[sizeof(kws)/sizeof(envEntry)-1]))

/* ----------------------------------------------------------------------*/

envEntry *lookupEnv (Env env, char* name){
  while ((env != NULL) && strcmp(name,env->name) != 0){
    env = env->next;
  }
  return env;
}

Env pushEnv(Env env, envEntry *e){
  e->next = env;
  return e;
}

Env popEnv(Env env){
  Env e;
  if(env == lastKW){
    printf("Error in deleting Environment Entry!\n");
    exit(9);
  }
  e = env->next;
  free(env);
  return e;
}

void deleteEnv(Env env){
  Env e;
  while(env != lastKW){
    e = env->next;
    free (env);
    env = e;
  }
}

int getVarNo(Env env, envEntry *e){
  int i = 0;

  while((env != NULL) && (env != e)){
    if (env->tp == Variable){
      i++;
    }
    env = env->next;
  }
  if (env == NULL){
    printf("Implementation error in getVarNo!\n");
    exit(9);
  }
  return i;
}

Env popVarEnv(Env env, int no){
  while ((no > 0) && (env->tp == Variable)){
    env = popEnv(env);
    no--;
  }
  if(no > 0){
    printf("Implementation error in popVarEnv!\n");
    exit(9);
  }
  return env;
}

envEntry *lookupLastFunction(Env env){
  while((env != NULL) && (env->tp != Function)){
    env = env->next;
  }
  return env;
}

void printEnvEntry (envEntry *e){
  printf("Next: %8x Name: >%-10s< Type: %-8s Token: %-6s Arity: %4d Adr: %8x Val: %8ld\n",
         e->next, e->name, idtypeStr[e->tp], TokenStr[e->tk], e->arity, e->adr, e->value);
}

void printEnv(Env env){
  while (env != NULL){
    printEnvEntry(env);
    env=env->next;
  }
}

/* ----------------------------------------------------------------------*/

Tinp newScanner(){
  Tinp inp;

  inp = (Tinp) malloc(sizeof(InpRec));

  inp->ttextlen = TTEXT_INITAL_LEN;
  inp->ttext = (char *)malloc(inp->ttextlen);

  inp->data = NULL;
  inp->env  = lastKW;
  inp->p    = NULL;
  inp->val  = 0;
  inp->tk   = Eos;
  inp->e    = NULL;
  return inp;
}

void deleteScanner(Tinp inp){
  free(inp->data);
  deleteEnv(inp->env);
  free(inp->ttext);
  free(inp);
}

void addBuiltin(Tinp inp, char *name, int arity, glam_f adr){
  envEntry *e = (envEntry *) malloc(sizeof(envEntry));
  e->name = name; /* evtl Kopieren ?? */
  e->tp = Builtin;
  e->arity = arity;
  e->adr = adr;
  e->tk = Id;
  inp->env = pushEnv(inp->env,e);
}

void addConstant(Tinp inp, char *name, glam_t val){
  envEntry *e = (envEntry *) malloc(sizeof(envEntry));
  e->name = name; /* evtl Kopieren ?? */
  e->tp = Constant;
  e->value = val;
  e->tk = Id;
  inp->env = pushEnv(inp->env,e);
}

void newInput(Tinp inp, char *data){
  if (inp->data != NULL){
    free(inp->data);
  }
  inp->data = data;
  inp->p = data;
  eatToken(inp);
}

/* ---------------------------------------------------------------------- */

void scanTokenText(Tinp inp, int testDigits){
  int i = 0;
  while(testDigits ? (isdigit(inp->p[i])) 
                   : (isalnum(inp->p[i]) || (inp->p[i] == '_'))){
    if (i >= inp->ttextlen){
      inp->ttextlen = inp->ttextlen + DELTA_TTEXTLEN;
      inp->ttext = (char *) realloc(inp->ttext, inp->ttextlen);
    }
    inp->ttext[i]=inp->p[i];
    i++;
  }
  inp->ttext[i] = '\0';
  inp->p = inp->p + i;
}

void scanIdent(Tinp inp){
  envEntry *e;

  scanTokenText(inp,0);
  
  inp->e = lookupEnv(inp->env,inp->ttext);
  if(inp->e == NULL){ /* allocate entry */
    e = (envEntry *) malloc (sizeof(envEntry));
    e->name = (char *) malloc(strlen(inp->ttext)+1);
    strcpy(e->name,inp->ttext);
    e->tp   = Unknown;
    e->tk   = Id;
    inp->env = pushEnv(inp->env,e);
    inp->e   = e;
  }
  inp->tk = inp->e->tk;
}

void scanNumber(Tinp inp){
  scanTokenText(inp,1);
  inp->val = atoi(inp->ttext);
  inp->tk = Num;
}

void eatToken(Tinp inp){
  /* scan for next token */
  inp->tk = Eos;
  while((inp->tk == Eos) && (inp->p[0] != '\0')){
    if (isalpha(inp->p[0])){
      scanIdent(inp);
    }
    else if (isdigit(inp->p[0])){
      scanNumber(inp);
    }
    else {
      switch (inp->p[0]) {
      case '(': inp->tk = Lpar; break;
      case ')': inp->tk = Rpar; break;
      case '{': inp->tk = Lbrc; break;
      case '}': inp->tk = Rbrc; break;
      case '-': /* comments or minus */
        if(inp->p[1] == '-'){do {inp->p++;} while ((inp->p[0] != '\0') && (inp->p[0] != '\n'));} 
        else {inp->tk = Mns;} 
        break;
      case '+': inp->tk = Pls; break;
      case '*': inp->tk = Mul; break;
      case '/': inp->tk = Div; break;
      case '%': inp->tk = Mod; break;
      case ',': inp->tk = Com; break;
      case ';': inp->tk = Semi; break;
      case '=': /* == or = */
        if(inp->p[1] == '='){inp->tk = Eqv; inp->p++;}
        else                {inp->tk = Eq;}             break;
      case '>': /* >= or > */
        if(inp->p[1] == '='){inp->tk = Geq; inp->p++;}
        else                {inp->tk = Gt;}             break;
      case '<': /* <= or < */
        if(inp->p[1] == '='){inp->tk = Leq; inp->p++;}
        else                {inp->tk = Lt;}             break;
      case '!': /* != or ! */
        if(inp->p[1] == '='){inp->tk = Neq; inp->p++;}
        else                {inp->tk = Lnot;}           break;
      case '&': /* && or & */
        if(inp->p[1] == '&'){inp->tk = Land; inp->p++;}
        else                {inp->tk = Band;}             break;
      case '^': /* ^^ or ^ */
        if(inp->p[1] == '^'){inp->tk = Lxor; inp->p++;}
        else                {inp->tk = Bxor;}             break;
      case '|': /* || or | */
        if(inp->p[1] == '|'){inp->tk = Lor; inp->p++;}
        else                {inp->tk = Bor;}             break;
      case '~': inp->tk = Bnot; break;
      case ' ':
      case '\n':
      case '\t': break;
      default:
        printf("Unexpected Char: >%c<\n", inp->p[0]);
        exit(9);
      }
      inp->p++;
    }
  }
}

void acceptToken(Token tk, Tinp inp){
  if (inp->tk == tk){
    eatToken(inp);
  }
  else {
    printf("Error: Unexpected Token %s, (%s expected)!\n", TokenStr[inp->tk],TokenStr[tk]);
    exit(9);
  }
}

int maybeToken(Token tk, Tinp inp){
  if (inp->tk == tk){
    eatToken(inp);
    return 1;
  }
  else {
    return 0;
  }
}

Token getToken(Tinp inp){
  return inp->tk;
}

long getTokenVal(Tinp inp){
  if (inp->tk == Num){
    return inp->val;
  }
  else if ((inp->tk == Id) && (inp->e->tp == Constant)){
    return inp->e->value;
  }
  else if ((inp->tk == Id) && (inp->e->tp == Variable)){
    return getVarNo(inp->env,inp->e);
  }
  else {
    printf("Error: no Token val!\n");
    exit(9);
  }
}

char *getTokenText(Tinp inp){
  if (inp->tk != Id){
    printf("Error in getTokenText!\n");
    exit(9);
  }
  return inp->ttext;
}

idtype getIdType(Tinp inp){
  if((inp->tk != Id)||(inp->e == NULL)){
    printf("Error in getIdType!\n");
    exit(9);
  }
  return inp->e->tp;
}

void tokenIsFun(Tinp inp, glam_f adr){
  if((inp->tk == Id)&&(inp->e->tp == Unknown)){
    inp->e->tp = Function;
    inp->e->arity = -1;
    inp->e->adr = adr;
  }
  else {
    printf("Error: tokenIsFun not applicable!\n");
    exit(9);
  }
}

void setArity(Tinp inp, int no){
  envEntry *e = lookupLastFunction(inp->env);
  if ((e != NULL) && (e->arity == -1)){
    e->arity = no;
  }
  else {
    printf("Error in setArity!\n");
    exit(9);
  }  
}

int getArity(Tinp inp){
  if((inp->tk != Id)||(inp->e == NULL)||!((inp->e->tp == Function)||(inp->e->tp == Builtin))){
    printf("Error in getIdType!\n");
    exit(9);
  }
  return inp->e->arity;
}

glam_f getAdr(Tinp inp){
  if((inp->tk != Id)||(inp->e == NULL)||!((inp->e->tp == Function)||(inp->e->tp == Builtin))){
    printf("Error in getAdr!\n");
    exit(9);
  }
  return inp->e->adr;
}

glam_f getFctAdr(Tinp inp, char *name){
  envEntry *e = lookupEnv(inp->env, name);
  if((e == NULL)||!((e->tp == Function)||(e->tp == Builtin))){
    printf("Error in getFctAdr!\n");
    exit(9);
  }
  return e->adr;
}

void tokenIsVar(Tinp inp){
  if((inp->tk == Id)&&(inp->e->tp == Unknown)){
    inp->e->tp = Variable;
  }
}

void popVars(Tinp inp, int no){
  inp->env = popVarEnv(inp->env, no);
}

int getVarAdr(Tinp inp){
  return getVarNo(inp->env, inp->e);
}

void printToken (Token tk){
  printf("%s\n", TokenStr[tk]);
}

void printScanner(Tinp inp){
  printf("ToDo: >>>%s<<<\n", inp->p);
  printf("Token: %s\n", TokenStr[inp->tk]);
  printf("Value: %ld\n", inp->val);
  printf("TText: >>>%s<<<\n", inp->ttext);
  printf("Environment:\n");
  printEnv(inp->env);
  printf("Entry:\n");
  printEnvEntry(inp->e);
}
