/**********************************************************************
* Filename : CritGoals.c
*
* Kurzbeschreibung : Verwalten der Kritischen Ziele
*
* Autoren : WEC
*
* 08.12.96: Erstellung aus SUEM.c heraus
* 22.02.97: Mehrfachaufrufclean. AJ.
* 02.07.97: Umstellung auf Heaps. AB.
*
*
* Benoetigt folgende Module:
*
* Benutzte Globale Variablen aus anderen Modulen:
*
**********************************************************************/

/* Die Datei wurde mal erst aus SUEM.c heraus ausgelagert. Die Schnittstelle
   ist aber noch nicht ganz sauber.
   */

#include "general.h"
#include "CritGoals.h"
#include "TermOperationen.h"
#include "RUndEVerwaltung.h"
#include "InfoStrings.h"
#include "Ausgaben.h"

/* Globale Variablen */
static SV_SpeicherManagerT CG_Speicher;
static BOOLEAN CG_Initialised = FALSE;

static CounterT CG_AnzCritGoalsGesamt, CG_AnzMaxCritGoals, CG_AnzCritGoalsActual;

/* --------------------------------------------------------------------------------------------------------*/
/* ------------ LISTE fuer Crit Goals ---------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/


/* Liste fuer neue CGs: */
typedef struct CGListTS {
  TermpairT            tp;
  long int             w1;
  long int             w2;
  RegelOderGleichungsT aP;
  RegelOderGleichungsT oP;
  struct CGListTS      *Nachf;
} CGListT, *CGListPtrT;

static CGListPtrT CGListNeu;

/* Heap fuer CG-Aufbewahrung */
typedef struct CGHeapTS {
  TermpairT            tp;
  long int             w1;
  long int             w2;
  BOOLEAN              LebtNoch;
  RegelOderGleichungsT aP;
  RegelOderGleichungsT oP;
} CGHeapT;

static CGHeapT *CGHeap;
static unsigned long int NaechstesFreies;
static unsigned long int CGHeapAktuelleGroesse;

#define TESTEN_AUS
#ifdef TESTEN

static void HeapTest(char *str)
{
   unsigned long int i;
   printf("************Heap-Test %s*************\n",str);
   for(i=0; i < NaechstesFreies; i++){
      if ((2*i+1< NaechstesFreies) &&
          ((CGHeap[2*i+1].w1<CGHeap[i].w1)||
           ((CGHeap[2*i+1].w1==CGHeap[i].w1)&&(CGHeap[2*i+1].w2<CGHeap[i].w2))))
         printf("CGHeap-Sortierung falsch (Vater (%ld,%ld), Sohn (%ld,%ld))\n",
                CGHeap[i].w1, CGHeap[i].w2, CGHeap[2*i+1].w1, CGHeap[2*i+1].w2);

      if ((2*i+2< NaechstesFreies) &&
          ((CGHeap[2*i+2].w1<CGHeap[i].w1)||
           ((CGHeap[2*i+2].w1==CGHeap[i].w1)&&(CGHeap[2*i+2].w2<CGHeap[i].w2))))
         printf("CGHeap-Sortierung falsch (Vater (%ld,%ld), Sohn (%ld,%ld))\n",
                CGHeap[i].w1, CGHeap[i].w2, CGHeap[2*i+2].w1, CGHeap[2*i+2].w2);
      /*      printf("%ld --- KP zu Eltern (%ld, %ld) \n", i, 
             CGHeap[i].LebtNoch?CGHeap[i].aP->Nummer:0,
             CGHeap[i].LebtNoch?CGHeap[i].oP->Nummer:0);*/
   }
}
#else
#define HeapTest(bla)
#endif


static BOOLEAN LasseEinsinken(unsigned long int position)
{
   BOOLEAN changed = FALSE,isHeap = FALSE;
   unsigned long j=position*2+1;
   CGHeapT merker=CGHeap[position];

   while ((!isHeap) && (j<NaechstesFreies)){
      if ((j+1<NaechstesFreies) &&                                      /* Position j+1 noch gueltig? */
          ((CGHeap[j].w1>CGHeap[j+1].w1) ||
           ((CGHeap[j].w1==CGHeap[j+1].w1)&&(CGHeap[j].w2>CGHeap[j+1].w2))))
        j++;                                                           /* => j hat jetzt den Wert des leichteren Nachfolgers */
      if ((merker.w1 > CGHeap[j].w1)||
          ((merker.w1 == CGHeap[j].w1)&&(merker.w2 > CGHeap[j].w2))){
         CGHeap[position]=CGHeap[j];
         changed = TRUE;
         position = j;
         j *=2;
         j++;
      }
      else
         isHeap = TRUE;
   }
   if (changed)
      CGHeap[position] = merker;
   return changed;
}

static void Insert(TermpairT tp, long int w1, long int w2, 
                   RegelOderGleichungsT actPar, RegelOderGleichungsT otherPar)
{
   unsigned long int i = NaechstesFreies;

   /* ggf. dynHeap ausdehnen */
   if (NaechstesFreies == CGHeapAktuelleGroesse){
      CGHeapAktuelleGroesse *= 2;
      CGHeap = our_realloc(CGHeap,
                           CGHeapAktuelleGroesse*sizeof(CGHeapT),
                           "CG: Expansion of Heap");
   }

   CGHeap[NaechstesFreies].tp = tp;
   CGHeap[NaechstesFreies].oP = otherPar;
   CGHeap[NaechstesFreies].aP = actPar;
   CGHeap[NaechstesFreies].w1 = w1;
   CGHeap[NaechstesFreies].w2 = w2;
   CGHeap[NaechstesFreies].LebtNoch = TRUE;

   NaechstesFreies++;

   do
      i = (i-1)/2;
   while(LasseEinsinken(i) && (i>0));
}

void CG_Collect(TermpairT tp,
                WeightEntryT wt,
                RegelOderGleichungsT aP,
                RegelOderGleichungsT oP)
     /* Einfuegen in die CGListe */
{
   CGListPtrT Neu;
   SV_Alloc(Neu,CG_Speicher,CGListPtrT);
   Neu->tp = tp;
   Neu->w1 = wt.w1;
   Neu->w2 = wt.w2;
   Neu->aP = aP;
   Neu->oP = oP;
   Neu->Nachf = CGListNeu;
   CGListNeu = Neu;
}


void CG_AddCollected(void)
     /* neue Liste in alte einsortieren */
{
   CGListPtrT Nachf;
   HeapTest("Anfang AddCollected");
   while (CGListNeu) {
      Nachf = CGListNeu->Nachf;
      /* TermT links,
         rechts,
         ZLinks;
      unsigned long int PclId;
      TPR_TERMPAIRtoPARALLEL_INTERMED(CGListNeu->tp,&links,&rechts,CGListNeu->aP,CGListNeu->oP,&PclId);

      ZLinks = TO_ErsterTeilterm(SO_SymbGleich(TO_TopSymbol(links),SO_FALSESymbol()) ? rechts : links);

      if (U1_Unifiziert(ZLinks,TO_NaechsterTeilterm(ZLinks)))
      CGListNeu->w1 = LONG_MIN;

      TPR_PARALLELtoTERMPAIR(CGListNeu->tp,links,rechts,CGListNeu->aP,CGListNeu->oP,FALSE,PclId);*/

      /*Einfuegen*/
      Insert(CGListNeu->tp, CGListNeu->w1, CGListNeu->w2, CGListNeu->aP, CGListNeu->oP);
#ifdef blubber
      IndexVorg = NULL;
      Index = CGList;
      while (Index &&
             (CGListNeu->w1 > Index->w1
              || (CGListNeu->w1 == Index->w1 && CGListNeu->w2 > Index->w2))) {
         IndexVorg = Index;
         Index = Index->Nachf;
      }
      CGListNeu->Nachf = Index;
      if (IndexVorg)                                                                          /* nicht am Listenanfang */
         IndexVorg->Nachf = CGListNeu;
      else
         CGList = CGListNeu;
#endif

      CG_AnzCritGoalsGesamt ++;
      CG_AnzCritGoalsActual ++;
      MaximierenLong(&CG_AnzMaxCritGoals, CG_AnzCritGoalsActual);
      CGListNeu = Nachf;
   }
   HeapTest("Ende AddCollected");
}


void CG_KillParent(RegelOderGleichungsT re)
     /* Ueberschreibt die Eltern-Eintraege der CGs im Heap mit NULL.
        Damit kann die Gueltigkeit von CGs einfach mit (aP==NULL?)
        getestet werden.*/
{
  unsigned long int i;
  HeapTest("Anfang KillParent");
  for (i=0; i<NaechstesFreies; i++)
    if ((CGHeap[i].aP->Nummer == re->Nummer) || (CGHeap[i].oP->Nummer == re->Nummer))
      CGHeap[i].LebtNoch = FALSE;
  HeapTest("Ende KillParent");
}
    
void CG_CleanUp (void)
{
  return;
}

void CG_SoftCleanUp (void)
{
  return;
}



BOOLEAN CG_InitAposteriori (void)
{

  if (!CG_Initialised)  SV_ManagerInitialisieren (&CG_Speicher, sizeof(CGListT), 100, "Critical Goals");

  CG_Initialised = TRUE;

  CGListNeu = NULL;

  CG_AnzCritGoalsGesamt = 0;
  CG_AnzMaxCritGoals = 0;
  CG_AnzCritGoalsActual = 0;
  
  NaechstesFreies=0;
  CGHeapAktuelleGroesse =2000;
  
  CGHeap = our_alloc(CGHeapAktuelleGroesse*sizeof(CGHeapT));
  
  return TRUE;
}


void CG_Statistics (InfoStringT IS)
{
   if (IS_IsIn("standard", IS)) {
      IO_DruckeKurzUeberschrift("Critical Goals","+");
      IO_DruckeStatistikZeile("total number of critical goals", CG_AnzCritGoalsGesamt);
      IO_DruckeStatistikZeile("current number of critical goals", CG_AnzCritGoalsActual);
      IO_DruckeStatistikZeile("max number of critical goals", CG_AnzMaxCritGoals);
   }
   else if (IS_IsIn("latex-head", IS)) { /* korrespondiert zu standard */
      fprintf(IO_Strom, "CG: total CGs & current CGs & Max CGs&");
   }
}



BOOLEAN CG_Select(TermpairT *tpPtr, RegelOderGleichungsT *aPPtr, RegelOderGleichungsT *oPPtr, void **addrefPtr)
{
  /* Tote an Top loeschen */
  HeapTest("Anfang Select");
  while ((NaechstesFreies>0) && (!CGHeap[0].LebtNoch)){
    TPR_DESTROYTERMPAIR(CGHeap[0].tp, CGHeap[0].aP, CGHeap[0].oP);
    CG_AnzCritGoalsActual--;
    if (--NaechstesFreies>0){
      CGHeap[0] = CGHeap[NaechstesFreies];
      LasseEinsinken(0);
    }
  }
  /* CGHeap jetzt leer? */
  if (NaechstesFreies==0) return FALSE;
  /* Sonst: Daten holen */
  CG_AnzCritGoalsActual--;
  *tpPtr = CGHeap[0].tp;
  *aPPtr = CGHeap[0].aP;
  *oPPtr = CGHeap[0].oP;
  *addrefPtr = NULL;
  if (--NaechstesFreies>0){
    CGHeap[0] = CGHeap[NaechstesFreies];
    LasseEinsinken(0);
  }
  HeapTest("Ende Select");
  return TRUE;
}
