/**********************************************************************
* Filename : NormaleZiele.c
*
* Kurzbeschreibung : Normale Zielverwaltung
*
* Autoren : Andreas Jaeger
*
* 10.10.96: Erstellung aus Zielverwaltung.h heraus
* 22.02.97: Mehrfachaufrufclean. AJ.
*
* Die Funktionen in diesem Modul werden nie direkt aufgerufen, sondern
* nur von der Zielverwaltung!
*
**********************************************************************/

#include <stdlib.h>
#include <string.h>

#include "compiler-extra.h"
#include "antiwarnings.h"
#include "general.h"
#include "pcl.h"

#include "Ausgaben.h"
#include "InfoStrings.h"
#include "MatchOperationen.h"
#include "NFBildung.h"
#include "NormaleZiele.h"
#include "RUndEVerwaltung.h"
#include "Subsumption.h"
#include "SymbolOperationen.h"
#include "TermOperationen.h"
#include "Termpaare.h"
#include "Zielverwaltung.h"
#include "Zielverwaltung_private.h"


/*********************/
/* Typdeklarationen  */
/*********************/

typedef struct privateNormaleZieleTS {
   TermT               links;
   TermT               rechts;
} privateNormaleZieleT, *privateNormaleZielePtrT;


void NZ_InitZielBeobachtungsProz (ZieleT  *ziel) {

  ziel->beobachtProz (ziel->privat.NZ->links, ziel->privat.NZ->rechts);
}


void NZ_ZielEinfuegen (ZieleT * ziel, TermT links, TermT rechts) 
{
   ziel->privat.NZ = (privateNormaleZieleT *) our_alloc (sizeof(privateNormaleZieleT));
   ziel->privat.NZ->links = TO_Termkopie (links);
   ziel->privat.NZ->rechts = TO_Termkopie (rechts);   
}


void NZ_LoeschePrivateData (ZieleT *ziel) 
{
   TO_TermLoeschen (ziel->privat.NZ->links);
   TO_TermLoeschen (ziel->privat.NZ->rechts);
   our_dealloc (ziel->privat.NZ);
}


void NZ_AusgabeZiel (ZieleT * ziel, Z_AusgabeArtT Z_AusgabeArt) 
{
   switch (Z_AusgabeArt) {
   case Z_AusgabeAlles:
   case Z_AusgabeInitial:
      IO_DruckeFlex(" current  %t ?= %t\n", ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      break;
   case Z_AusgabeSpez:
      IO_DruckeFlex("\t%t = %t\n", ziel->orginalLinks, ziel->orginalRechts);
      break;
   }
}

   
void NZ_ZielBeobachtung (ZieleT *ziel, Z_BeobachtungsDurchlaufProcT AP)  
{
   AP (ziel->privat.NZ->links, ziel->privat.NZ->rechts, ziel->nummer);
}


#define Test_SubsummiertVonGM(ZIEL)                                                                     \
       if (SS_TermpaarSubsummiertVonGM (ZIEL->privat.NZ->links, ZIEL->privat.NZ->rechts))             \
         {                                                                                              \
            /* falls Ziel subsummiert */                                                                \
            PCL_ZielSubsummiert (&ZIEL->pclID,                                                          \
                                 ZIEL->privat.NZ->links,                                               \
                                 ZIEL->privat.NZ->rechts,                                              \
                                 TP_PclID(MO_SubsummierendeGleichung));                                 \
            IO_DruckeMeldungZielSubsummiert(ZIEL->nummer, TP_ElternNr(MO_SubsummierendeGleichung),      \
                                            ZIEL->privat.NZ->links, ZIEL->privat.NZ->rechts,          \
                                            TP_LinkeSeite(MO_SubsummierendeGleichung),                  \
                                            TP_RechteSeite(MO_SubsummierendeGleichung));                \
            ZIEL->subSumNr = TP_ElternNr(MO_SubsummierendeGleichung);                                    \
            ZIEL->status = Z_Subsumed;                                                                  \
            return TRUE;                                                                                \
         }


BOOLEAN NZ_ZielErledigtMitNeuerRegel (ZieleT *ziel, RegelT Regel)
{
   BOOLEAN reduziert;
   
   PCL_Reduktionsbeginn (Ziel, &ziel->pclID, linkeSeite, ziel->privat.NZ->rechts);
   reduziert = NF_NFMitNeuerRegel(Regel, ziel->privat.NZ->links, Ziele);
   PCL_Reduktionsbeginn(Ziel, &ziel->pclID, rechteSeite, ziel->privat.NZ->links);
   reduziert |= NF_NFMitNeuerRegel(Regel, ziel->privat.NZ->rechts, Ziele);
   if (reduziert){
      if (ziel->beobachtProz)
         ziel->beobachtProz (ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      IO_DruckeMeldungZielReduziert(ziel->nummer, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      /* Ziel .. konnte reduziert werden: l =? r */
      if (TO_TermeGleich(ziel->privat.NZ->links, ziel->privat.NZ->rechts)){
         PCL_ZielBewiesen(&ziel->pclID, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
         IO_DruckeMeldungZielZusammengefuehrt(); /* ... und somit zusammengefuehrt.*/
         ziel->status = Z_Joined;
         return TRUE;
      }
      else
         Test_SubsummiertVonGM (ziel);
   }
   return FALSE;
}


BOOLEAN NZ_ZielErledigtMitNeuerGleichung (ZieleT *ziel, GleichungsT Gleichung)
{
   BOOLEAN reduziert;
   
   PCL_Reduktionsbeginn (Ziel, &ziel->pclID, linkeSeite, ziel->privat.NZ->rechts);
   reduziert = NF_NFMitNeuerGleichung(Gleichung, ziel->privat.NZ->links, Ziele);
   PCL_Reduktionsbeginn(Ziel, &ziel->pclID, rechteSeite, ziel->privat.NZ->links);
   reduziert |= NF_NFMitNeuerGleichung(Gleichung, ziel->privat.NZ->rechts, Ziele);
   if (reduziert){
      if (ziel->beobachtProz)
         ziel->beobachtProz (ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      IO_DruckeMeldungZielReduziert(ziel->nummer, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      /* Ziel .. konnte reduziert werden: l =? r */
      if (TO_TermeGleich(ziel->privat.NZ->links, ziel->privat.NZ->rechts)){
         PCL_ZielBewiesen(&ziel->pclID, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
         IO_DruckeMeldungZielZusammengefuehrt(); /* ... und somit zusammengefuehrt.*/
         ziel->status = Z_Joined;
         return TRUE;
      }
      else
         Test_SubsummiertVonGM(ziel);
   } else
      /* wenn nicht reduziert wurde muss nur auf Subsumption mit neuer Gleichung getestet
         werden */
      if (SS_GleichungSubsummiertTermpaar (Gleichung, ziel->privat.NZ->links,
                                           ziel->privat.NZ->rechts))
         {
            /* falls Ziel subsummiert */
            PCL_ZielSubsummiert (&ziel->pclID,
                                 ziel->privat.NZ->links,
                                 ziel->privat.NZ->rechts,
                                 TP_PclID(Gleichung));
            IO_DruckeMeldungZielSubsummiert(ziel->nummer, TP_ElternNr(Gleichung),
                                            ziel->privat.NZ->links,
                                            ziel->privat.NZ->rechts,
                                            TP_LinkeSeite(Gleichung),
                                            TP_RechteSeite(Gleichung));
            ziel->subSumNr = TP_ElternNr(Gleichung);
            ziel->status = Z_Subsumed;
            return TRUE;
         }
   return FALSE;
}


BOOLEAN NZ_ZielFinalerDurchlauf (ZieleT * ziel) 
{
   BOOLEAN reduziert;
   
   PCL_Reduktionsbeginn (Ziel, &ziel->pclID, linkeSeite, ziel->privat.NZ->rechts);
   reduziert = NF_NFBildung (ziel->privat.NZ->links, Ziele);
   PCL_Reduktionsbeginn(Ziel, &ziel->pclID, rechteSeite, ziel->privat.NZ->links);
   reduziert |= NF_NFBildung (ziel->privat.NZ->rechts, Ziele);
   if (reduziert){
      if (ziel->beobachtProz)
         ziel->beobachtProz (ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      IO_DruckeMeldungZielReduziert(ziel->nummer, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
      /* Ziel .. konnte reduziert werden: l =? r */
      if (TO_TermeGleich(ziel->privat.NZ->links, ziel->privat.NZ->rechts)){
         PCL_ZielBewiesen(&ziel->pclID, ziel->privat.NZ->links, ziel->privat.NZ->rechts);
         IO_DruckeMeldungZielZusammengefuehrt(); /* ... und somit zusammengefuehrt.*/
         ziel->status = Z_Joined;
         return TRUE;
      }
   }
   Test_SubsummiertVonGM (ziel);

   return FALSE;
}


/*=================================================================================================================
  =  Zielbehandlung                                                                                               =
  =================================================================================================================*/


void NZ_Statistik(InfoStringT str)
{
   if (!(IS_IsIn("all", str) || IS_IsIn("Z", str)))
      return;
}


static void NZ_ParseInfoStrings (InfoStringT DEF MAYBE_UNUSEDPARAM,
                                 InfoStringT USR MAYBE_UNUSEDPARAM) 
{
   ;
}


void NZ_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
   switch (what) {
   case ParamEvaluate:
      NZ_ParseInfoStrings (DEF, USR);
      break;
   case ParamExplain:
      fprintf (IO_Strom, "Classical Goal Treatment (without MNF)\n");
      fprintf (IO_Strom, "**************************************\n");
      fprintf (IO_Strom, "To select: snf\n");

      break;
   }
}


BOOLEAN NZ_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
   NZ_ParseInfoStrings (DEF, USR);
   return TRUE;
}

