/*=================================================================================================================
  =                                                                                                               =
  =  KPMengenInterreduktion.c                                                                                     =
  =                                                                                                               =
  =  Interreduktion/Subsumption der KP-Menge                                                                      =
  =                                                                                                               =
  =  27.03.96  Arnim. Aus der Interreduktion ausgegliedert.                                                       =
  =================================================================================================================*/

#include "Ausgaben.h"
#include "general.h"
#include "InfoStrings.h"
#include "KPMengenInterreduktion.h"
#include "NFBildung.h"
#include "Subsumption.h"
#include "SUEM.h"
#include "TermOperationen.h"

#include <string.h>

/* -------------------------------------------------------------------------------*/
/* --------------- Statistik -----------------------------------------------------*/
/* -------------------------------------------------------------------------------*/

static BOOLEAN StatistikAn;

static unsigned long        AnzKPIRBehandelt;                  /* Bei der KPM-IR insgesamt betrachtete KPs */
static unsigned long        AnzKPIRReduziert;                  /* Bei der KPM-IR reduzierte KPs */
static unsigned long        AnzKPIRZusammengefuehrt;           /* Bei der KPM-IR von der Anwendungsprozedur zusammengefuehrte KPs */
static unsigned long        AnzKPIRSubsummiert;                /* Bei der KPM-IR von der Anwendungsprozedur subsummierte KPs */
static unsigned long        AnzKPnurW;                         /* Beim Neugewichten angefasste KPs */

#define IncAnzKPIRBehandelt_Stat()        (AnzKPIRBehandelt++)
#define IncAnzKPIRReduziert_Stat()        (AnzKPIRReduziert++)
#define IncAnzKPIRZusammengefuehrt_Stat() (AnzKPIRZusammengefuehrt++)
#define IncAnzKPIRSubsummiert_Stat()      (AnzKPIRSubsummiert++)
#define IncAnzKPIR_nurW_Stat()            (AnzKPnurW++)

#define IncAnzKPIRBehandelt_NoStat()
#define IncAnzKPIRReduziert_NoStat()
#define IncAnzKPIRZusammengefuehrt_NoStat()
#define IncAnzKPIRSubsummiert_NoStat()
#define IncAnzKPIR_nurW_NoStat()

void KPIR_Statistik(InfoStringT IS)
{
   if (!(IS_IsIn("KPIR", IS) || IS_IsIn("all", IS)))
      return;
   if (IS_IsIn("standard", IS)){
      if (StatistikAn){
         int tiefe1 = IO_Platzverbrauch(Maximum(2, AnzKPIRReduziert, AnzKPIRSubsummiert));
         
         IO_DruckeKurzUeberschrift("Intermediate Reprocessing (IRP)", "+");
         IO_DruckeStatistikZeile("CPs touched (total)", AnzKPIRBehandelt);
         IO_DruckeStatistikZeileEingerueckt("of them subsumed", AnzKPIRSubsummiert, tiefe1);
         IO_DruckeStatistikZeileEingerueckt("of them simplified", AnzKPIRReduziert, tiefe1);
         IO_DruckeStatistikZeileEingerueckt("of them joined", AnzKPIRZusammengefuehrt, tiefe1+IO_Platzverbrauch(AnzKPIRZusammengefuehrt));
         IO_DruckeStatistikZeile("CPs touched only for reweighting", AnzKPnurW);
      }
   }
   else if (IS_IsIn("latex-head", IS)){
      if (StatistikAn)
         fprintf(IO_Strom, "KPIR: touched & subs & red & join & rew & ");
   }
   else if (IS_IsIn("config", IS)){
      fprintf(IO_Strom, "KPIR: Statistik %s\n", StatistikAn?" AN":"AUS");
   }
   else if (IS_IsIn("intern", IS)){
      ;
   }
}
      

/* -------------------------------------------------------------------------------*/
/* --------------- Strings in Strength umwandeln ---------------------------------*/
/* -------------------------------------------------------------------------------*/

static int String2Strength(char *str)
{
   int val = 0;

   if (strchr(str, 'r')) val |= 1;
   if (strchr(str, 'e')) val |= 2;
   if (strchr(str, 's')) val |= 4;
   if (strchr(str, 'w')) val |= 8;

   return val;
}
      
/* -------------------------------------------------------------------------------*/
/* --------------- Anwendungsprozeduren ------------------------------------------*/
/* -------------------------------------------------------------------------------*/

/* Anwendungsprozeduren, die dem KP-Mengendurchlauf uebergeben werden, um die KP-Mengen-Interreduktion
   durchzufuehren.
   Vor Aufruf jeder Anwendungsprozedur kann davon ausgegangen werden, dass das entsprechende KP als aktuelles (bzgl. IR) angemeldet
   und physikalisch noch intakt ist. Es sind also sowohl Zugriffe auf die Eintrage des Objekts kp als auch ueber die
   KPO_...AktKP()-Funktionen statthaft.*/


#define REDUKTIONS_BODY(/*{NFBildung, NFBildungZuRegeln, NFBildungZuGleichungen} */ Womit, /*{Stat, NoStat}*/ statistik)                \
{                                                                                                                                       \
   PCL_Reduktionsbeginn(KP, pclID, linkeSeite, rhs);                                                                                    \
   reduziert = NF_ ## Womit (lhs, KPsIntermed);                                                                                                   \
   PCL_Reduktionsbeginn(KP, pclID, rechteSeite, lhs);                                                                                   \
   reduziert |= NF_ ## Womit (rhs, KPsIntermed);                                                                                                  \
                                                                                                                                        \
   if (reduziert){                                                                                                                      \
      IncAnzKPIRReduziert_ ## statistik();                                                                                              \
      if (TO_TermeGleich(lhs, rhs)){                                                                                                    \
         IncAnzKPIRZusammengefuehrt_ ## statistik();                                                                                    \
         return 1;                                                                                                                      \
      }                                                                                                                                 \
   }                                                                                                                                    \
}   

#define SUBSUMPTIONS_BODY(/*{Stat, NoStat}*/ statistik)                                                                                 \
{                                                                                                                                       \
   if (SS_TermpaarSubsummiertVonGM(lhs, rhs)){                                                                                             \
      IncAnzKPIRSubsummiert_## statistik();                                                                                             \
      return 1;                                                                                                                         \
   }                                                                                                                                    \
}

#define COMBINE_RS_BODY(/*{NFBildung, NFBildungZuRegeln, NFBildungZuGleichungen} */ Womit, /*{Stat, NoStat}*/ statistik, ReturnExpr) \
{                                                                                                                                    \
   BOOLEAN reduziert;                                                                                                                \
                                                                                                                                     \
   IncAnzKPIRBehandelt_ ## statistik();                                                                                              \
   IO_DruckeMeldungKPInIRBehandelt();                                                                                                \
   REDUKTIONS_BODY(Womit, statistik);                                                                                                \
   SUBSUMPTIONS_BODY(statistik);                                                                                                     \
   return (ReturnExpr);                                                                                                              \
}

#define S_BODY(/*{Stat, NoStat}*/ statistik, ReturnExpr)                                                                                \
{                                                                                                                                       \
   IncAnzKPIRBehandelt_ ## statistik();                                                                                                 \
   IO_DruckeMeldungKPInIRBehandelt();                                                                                                \
   SUBSUMPTIONS_BODY(statistik);                                                                                                        \
   return (ReturnExpr);                                                                                                                 \
}

#define R_BODY(/*{NFBildung, NFBildungZuRegeln, NFBildungZuGleichungen} */ Womit, /*{Stat, NoStat}*/ statistik, ReturnExpr)             \
{                                                                                                                                       \
   BOOLEAN reduziert;                                                                                                                   \
                                                                                                                                        \
   IncAnzKPIRBehandelt_ ## statistik();                                                                                                 \
   IO_DruckeMeldungKPInIRBehandelt();                                                                                                \
   REDUKTIONS_BODY(Womit, statistik);                                                                                                   \
   return (ReturnExpr);                                                                                                                 \
}

#define W_BODY(/*{Stat, NoStat}*/ statistik)                                                                                            \
{                                                                                                                                       \
   IO_DruckeMeldungKPInIRBehandelt();                                                                                                \
   IncAnzKPIR_nurW_ ## statistik();                                                                                                     \
   return 2;                    /* immer neugewichten */                                                                                \
}


static int AP_15_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
COMBINE_RS_BODY(NFBildung, Stat, 2)
static int AP_15_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
COMBINE_RS_BODY(NFBildung, NoStat, 2)

static int AP_14_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
COMBINE_RS_BODY(NFBildungZuGleichungen, Stat, 2)
static int AP_14_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
COMBINE_RS_BODY(NFBildungZuGleichungen, NoStat, 2)

static int AP_13_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
COMBINE_RS_BODY(NFBildungZuRegeln, Stat, 2)
static int AP_13_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
COMBINE_RS_BODY(NFBildungZuRegeln, NoStat, 2)

static int AP_12_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
S_BODY(Stat,2)
static int AP_12_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
S_BODY(NoStat,2)

static int AP_11_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
R_BODY(NFBildung, Stat, 2)
static int AP_11_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
R_BODY(NFBildung, NoStat, 2)

static int AP_10_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
R_BODY(NFBildungZuGleichungen, Stat, 2)
static int AP_10_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
R_BODY(NFBildungZuGleichungen, NoStat, 2)

static int AP_9_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
R_BODY(NFBildungZuRegeln, Stat, 2)
static int AP_9_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
R_BODY(NFBildungZuRegeln, NoStat, 2)


static int AP_8_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
W_BODY(Stat)
static int AP_8_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
W_BODY(NoStat)

static int AP_7_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
COMBINE_RS_BODY(NFBildung, Stat, reduziert?2:0)
static int AP_7_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
COMBINE_RS_BODY(NFBildung, NoStat, reduziert?2:0)

static int AP_6_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
COMBINE_RS_BODY(NFBildungZuGleichungen, Stat, reduziert?2:0)
static int AP_6_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
COMBINE_RS_BODY(NFBildungZuGleichungen, NoStat, reduziert?2:0)

static int AP_5_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
COMBINE_RS_BODY(NFBildungZuRegeln, Stat, reduziert?2:0)
static int AP_5_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
COMBINE_RS_BODY(NFBildungZuRegeln, NoStat, reduziert?2:0)

static int AP_4_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
S_BODY(Stat,0)
static int AP_4_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
S_BODY(NoStat,0)

static int AP_3_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
R_BODY(NFBildung, Stat, reduziert?2:0)
static int AP_3_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
R_BODY(NFBildung, NoStat, reduziert?2:0)

static int AP_2_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
R_BODY(NFBildungZuGleichungen, Stat, reduziert?2:0)
static int AP_2_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
R_BODY(NFBildungZuGleichungen, NoStat, reduziert?2:0)

static int AP_1_Stat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
R_BODY(NFBildungZuRegeln, Stat, reduziert?2:0)
static int AP_1_NoStat(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
R_BODY(NFBildungZuRegeln, NoStat, reduziert?2:0)



static int AP_0(TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID)
{
   return 0;
}



const char *strengthArray[16]={"---", "r", "e", "re", "s", "rs", "es", "res", "w",
			       "rw", "ew", "rew", "sw", "rsw", "esw", "resw"};

typedef int (* WTProcT) (TermT lhs, TermT rhs, RegelOderGleichungsT Father, RegelOderGleichungsT Mother, PCL_ObjektNummerT *pclID);

static WTProcT AP_Stat[16] = {AP_0, AP_1_Stat, AP_2_Stat, AP_3_Stat, AP_4_Stat, AP_5_Stat, AP_6_Stat, AP_7_Stat, AP_8_Stat,
                              AP_9_Stat, AP_10_Stat, AP_11_Stat, AP_12_Stat, AP_13_Stat, AP_14_Stat, AP_15_Stat};

static WTProcT AP_NoStat[16] = {AP_0, AP_1_NoStat, AP_2_NoStat, AP_3_NoStat, AP_4_NoStat, AP_5_NoStat, AP_6_NoStat, AP_7_NoStat, AP_8_NoStat,
                                AP_9_NoStat, AP_10_NoStat, AP_11_NoStat, AP_12_NoStat, AP_13_NoStat, AP_14_NoStat, AP_15_NoStat};

static WTProcT *AP;


/* Staerke:
   0 -- nix (dh. nur Waisenmord)
   1 -- r
   2 -- e
   3 -- re
   4 -- s
   5 -- rs
   6 -- es
   7 -- res
   8 -- w        _W_eight
   9 -- rw
  10 -- ew
  11 -- rew
  12 -- sw
  13 -- rsw
  14 -- esw
  15 -- resw
*/

/* -------------------------------------------------------------------------------*/
/* --------------- KP-Mengen-Durchlaeufe -----------------------------------------*/
/* -------------------------------------------------------------------------------*/

typedef struct{
   unsigned long int sp;
   int strength;
}StichpunktT;

static StichpunktT *StichpunktListe;
static unsigned long int AnzahlStichpunkte;
static unsigned long int AktuellerStichpunkt;

static unsigned long int Period;
static int PeriodStrength;


static BOOLEAN IstStichpunkt(unsigned long int punkt, int *strength)
/* Prozedur wird immer aufgerufen (kein Stichpunkt wird durch periodische Interreduktion
   uebersprungen!
*/
{
   if (StichpunktListe && (AktuellerStichpunkt<AnzahlStichpunkte)){
      if (StichpunktListe[AktuellerStichpunkt].sp>punkt) return FALSE;
      else{                     /* nicht groesser => nach Vorbed. gleich */
         *strength= StichpunktListe[AktuellerStichpunkt].strength;
         AktuellerStichpunkt++;
         return TRUE;
      }
   }
   else
      return FALSE;
}

#define KPIR_BODY(strength)                                                                                     \
{                                                                                                               \
   IO_DruckeMeldungKPIRStart(strengthArray[strength], SUE_AnzKPsToGo());                                        \
   SUE_WALKTHROUGH(AP[strength], TRUE);                                                                         \
   IO_DruckeMeldungKPIREnde();                                                                                  \
}
   


void KPIR_KPMengeInterreduzieren(void)
/* Interreduktion der KP-Menge gemaess Aufruf-Einstellung.*/
{
   int strength;
   /* Bedingungsreihenfolge relevant: Stichpunkte werden immer geprueft. Damit
      wird kein Stichpukt durch paralleles periodisches Interreduzieren uebersprungen!
      */
   if (IstStichpunkt(RE_AnzRegelnGesamt()+RE_AnzGleichungenGesamt(), &strength))
      KPIR_BODY(strength)
   else if ((Period>0)&&((RE_AnzRegelnGesamt()+RE_AnzGleichungenGesamt())%Period == 0))
      KPIR_BODY(PeriodStrength)
}


void KPIR_Wahlfrei(char *howStrong)
/* Fuehrt die KP-Mengen-Interreduktion zu einem beliebigen Zeitpunkt aus.
   Funktionalitaet(Maechtigkeit und dyn. Neubewerten)  wie beim Programm-Start
   vorgesehen. */
{
   int strength = String2Strength(howStrong);

   KPIR_BODY(strength);
}


/* -------------------------------------------------------------------------------*/
/* --------------- Initialisieren ------------------------------------------------*/
/* -------------------------------------------------------------------------------*/

static int SPcompare(void const *i,const void *j)
{
   return(((StichpunktT *) i)->sp - ((StichpunktT *) j)->sp);
}

static int AnzEintraege(char* str)
{
   char *kopie, *token, *kopieDummy;
   int anzahl=0;
   
   if (!str) return 0;          /* String ist leer -> strtok schmiert ab */

   kopie = (char *) our_alloc(sizeof(char)*(strlen(str)+1));
   strcpy(kopie, str);
   kopieDummy = kopie;

   while((token = strtok(kopie, "."))){
      kopie = NULL;             /* gemaess Spezifikation von strtok */
      if (atoi(token)>0) anzahl++;
   }
   our_dealloc(kopieDummy);
   return anzahl;
}
   
static void StichpunkteEintragen(char *str, int strength)
{
   char *kopie, *token, *kopieDummy;
   
   if (!str) return;          /* String ist leer -> strtok schmiert ab */

   kopie = (char *) our_alloc(sizeof(char)*(strlen(str)+1));
   strcpy(kopie, str);
   kopieDummy = kopie;

   while((token = strtok(kopie, "."))){
      kopie = NULL;             /* gemaess Spezifikation von strtok */
      if (atoi(token)>0){
         StichpunktListe[AktuellerStichpunkt].sp = atoi(token);
         StichpunktListe[AktuellerStichpunkt++].strength = strength;
      }
   }
   our_dealloc(kopieDummy);
}

#define Auswertung(StatistikAnAction, StatistikAusAction)                                                                                      \
{                                                                                                                                              \
  char *Str_r, *Str_e, *Str_s, *Str_w,                                                                                                         \
    *Str_re, *Str_rs, *Str_es, *Str_rw, *Str_ew, *Str_sw,                                                                                      \
    *Str_rew, *Str_rsw, *Str_esw, *Str_res,                                                                                                    \
    *Str_resw, *Str_o;                                                                                                                         \
                                                                                                                                               \
  char *periodStr = IS_ValueOfOption(DEF, USR, "period");                                                                                      \
  unsigned long int i,j, NeueStichpunktZahl;                                                                                                   \
                                                                                                                                               \
  StatistikAn = (IS_ValueOfOption(DEF, USR, "statistics") && (strcmp(IS_ValueOfOption(DEF, USR, "statistics"), "on") == 0));                   \
                                                                                                                                               \
  if (StatistikAn){StatistikAnAction;}else{StatistikAusAction;}                                                                                \
                                                                                                                                               \
  if (periodStr){Period = atoi(periodStr);PeriodStrength = String2Strength(periodStr);}                                                        \
                                                                                                                                               \
  Str_o = IS_ValueOfOption(DEF, USR, "o");                                                                                                     \
                                                                                                                                               \
  Str_w = IS_ValueOfOption(DEF, USR, "w");                                                                                                     \
  Str_r = IS_ValueOfOption(DEF, USR, "r");                                                                                                     \
  Str_e = IS_ValueOfOption(DEF, USR, "e");                                                                                                     \
  Str_s = IS_ValueOfOption(DEF, USR, "s");                                                                                                     \
                                                                                                                                               \
  Str_rw = IS_ValueOfOption(DEF, USR, "rw");                                                                                                   \
  Str_ew = IS_ValueOfOption(DEF, USR, "ew");                                                                                                   \
  Str_sw = IS_ValueOfOption(DEF, USR, "sw");                                                                                                   \
  Str_re = IS_ValueOfOption(DEF, USR, "re");                                                                                                   \
  Str_rs = IS_ValueOfOption(DEF, USR, "rs");                                                                                                   \
  Str_es = IS_ValueOfOption(DEF, USR, "es");                                                                                                   \
                                                                                                                                               \
  Str_rew = IS_ValueOfOption(DEF, USR, "rew");                                                                                                 \
  Str_rsw = IS_ValueOfOption(DEF, USR, "rsw");                                                                                                 \
  Str_esw = IS_ValueOfOption(DEF, USR, "esw");                                                                                                 \
  Str_res = IS_ValueOfOption(DEF, USR, "res");                                                                                                 \
                                                                                                                                               \
  Str_resw = IS_ValueOfOption(DEF, USR, "resw");                                                                                               \
                                                                                                                                               \
  AnzahlStichpunkte =                                                                                                                          \
    AnzEintraege(Str_r) + AnzEintraege(Str_e) + AnzEintraege(Str_s)  + AnzEintraege(Str_w) +                                                   \
    AnzEintraege(Str_rw) + AnzEintraege(Str_ew) + AnzEintraege(Str_sw)  + AnzEintraege(Str_re) + AnzEintraege(Str_rs) + AnzEintraege(Str_es) + \
    AnzEintraege(Str_rew) + AnzEintraege(Str_rsw) + AnzEintraege(Str_esw) + AnzEintraege(Str_res) +                                            \
    AnzEintraege(Str_resw) + AnzEintraege(Str_o);                                                                                              \
                                                                                                                                               \
  if (AnzahlStichpunkte > 0){                                                                                                                  \
    StichpunktListe = our_alloc(AnzahlStichpunkte*sizeof(StichpunktT));                                                                        \
                                                                                                                                               \
    AktuellerStichpunkt = 0;                                                                                                                   \
                                                                                                                                               \
    StichpunkteEintragen(Str_o, 0);                                                                                                            \
                                                                                                                                               \
    StichpunkteEintragen(Str_r, 1);                                                                                                            \
    StichpunkteEintragen(Str_e, 2);                                                                                                            \
    StichpunkteEintragen(Str_s, 4);                                                                                                            \
    StichpunkteEintragen(Str_w, 8);                                                                                                            \
                                                                                                                                               \
    StichpunkteEintragen(Str_rw, 9);                                                                                                           \
    StichpunkteEintragen(Str_ew, 10);                                                                                                          \
    StichpunkteEintragen(Str_sw, 12);                                                                                                          \
    StichpunkteEintragen(Str_re, 3);                                                                                                           \
    StichpunkteEintragen(Str_rs, 5);                                                                                                           \
    StichpunkteEintragen(Str_es, 6);                                                                                                           \
                                                                                                                                               \
    StichpunkteEintragen(Str_res, 7);                                                                                                          \
    StichpunkteEintragen(Str_rew, 11);                                                                                                         \
    StichpunkteEintragen(Str_rsw, 13);                                                                                                         \
    StichpunkteEintragen(Str_esw, 14);                                                                                                         \
                                                                                                                                               \
    StichpunkteEintragen(Str_resw, 15);                                                                                                        \
                                                                                                                                               \
    AktuellerStichpunkt = 0;                                                                                                                   \
                                                                                                                                               \
    qsort(StichpunktListe,AnzahlStichpunkte,sizeof(StichpunktT),SPcompare);                                                                    \
                                                                                                                                               \
    /* Doppelte Stichpunkte eliminieren */                                                                                                     \
    /* (Liste ist bereits sortiert => doppelte folgen aufeinander) */                                                                          \
                                                                                                                                               \
    NeueStichpunktZahl = AnzahlStichpunkte;                                                                                                    \
                                                                                                                                               \
    for(i = 0; i<AnzahlStichpunkte; i++)                                                                                                       \
      for (j=i+1; j<AnzahlStichpunkte; j++)                                                                                                    \
        if ((StichpunktListe[i].sp!=0)&&(StichpunktListe[i].sp == StichpunktListe[j].sp)){                                                     \
          StichpunktListe[i].strength |= StichpunktListe[j].strength;                                                                          \
          StichpunktListe[j].sp = 0;                                                                                                           \
          NeueStichpunktZahl--;                                                                                                                \
        }                                                                                                                                      \
    j=0;                                                                                                                                       \
    for (i = 0; i<AnzahlStichpunkte; i++){                                                                                                     \
      if (StichpunktListe[i].sp!=0){                                                                                                           \
        StichpunktListe[j].sp=StichpunktListe[i].sp;                                                                                           \
        StichpunktListe[j].strength=StichpunktListe[i].strength;                                                                               \
        j++;                                                                                                                                   \
      }                                                                                                                                        \
    }                                                                                                                                          \
    AnzahlStichpunkte = NeueStichpunktZahl;                                                                                                    \
  }                                                                                                                                            \
  fprintf(IO_Strom, "\n* INTERMEDIATE CP-SET-REDUCTION\n\n  Period = %lu %s, %lu poke points, statistics %s\n",                                 \
          Period, strengthArray[PeriodStrength], AnzahlStichpunkte, StatistikAn?"on":"off");                                                   \
  if (AnzahlStichpunkte > 0){                                                                                                                  \
    fprintf(IO_Strom, "                ");                                                                                                     \
    for (i = 0; i<AnzahlStichpunkte; i++)                                                                                                      \
      fprintf(IO_Strom, "%ld %s, ", StichpunktListe[i].sp, strengthArray[StichpunktListe[i].strength]);                                        \
    fprintf(IO_Strom, "\n");                                                                                                                   \
  }                                                                                                                                            \
}
  

void KPIR_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
   if (what == ParamEvaluate){
     int AktuellerStichpunkt = 0; /* globale Variablen ueberschreiben */
     StichpunktT *localStichpunktListe = StichpunktListe; /* retten */
     long int Period=0, PeriodStrength=0;
     unsigned long int AnzahlStichpunkte;
     BOOLEAN StatistikAn;

     Auswertung({;},{;});

     our_dealloc(StichpunktListe);
     StichpunktListe = localStichpunktListe;
   }
   else{
      IO_DruckeKurzUeberschrift("INTERMEDIATE CP-SET-REDUCTION", "-");
      fprintf(IO_Strom, "1. [period=<n><strength>]\n");
      fprintf(IO_Strom, "2. [<strength>=<n1>.<n2>.<n3>...]\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "Explanation:\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "1. Tells Waldmeister to periodically reduce the SUE every <n> new rules/equations\n");
      fprintf(IO_Strom, "    with <strength>\n");
      fprintf(IO_Strom, "2. Tells Waldmeister to reduce the SUE after a certain number of new rules/equations\n");
      fprintf(IO_Strom, "    (<n1>, <n2>, <n3>) with <strength>.\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "<strength> is a string within which only the following letters will be recognized:\n");
      fprintf(IO_Strom, "r - rules           : use rules for normalization\n");
      fprintf(IO_Strom, "e - equations       : use equations for normalization\n");
      fprintf(IO_Strom, "s - subsumption     : subsume\n");
      fprintf(IO_Strom, "w - weight          : reweight every touched ue\n");
      fprintf(IO_Strom, "o - orphan-murder   : remove all the orphans.\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "If normalizing the cps intermediately, reweighting of the simplified ones\n");
      fprintf(IO_Strom, "will be done automatically. Those joined or subsumed will be removed from\n");
      fprintf(IO_Strom, "the set.\n");
      fprintf(IO_Strom, "It is not necessary to state an 'o' when intermediate reductions are invoked:\n");
      fprintf(IO_Strom, "full orphan murder will be executed on these ocasions automatically.\n");
   }
}
   

BOOLEAN KPIR_InitAposteriori(InfoStringT DEF, InfoStringT USR)
{
  AnzKPIRBehandelt = AnzKPIRReduziert = 
    AnzKPIRZusammengefuehrt = AnzKPIRSubsummiert = AnzKPnurW = 0;               

  Auswertung({AP = AP_Stat;},{AP = AP_NoStat;})
    
  return TRUE;
}
