/*=================================================================================================================
  =                                                                                                               =
  =  RUndEVerwaltung.c                                                                                            =
  =                                                                                                               =
  =  Verwaltung der Regel- und der Gleichungsmenge                                                                =
  =                                                                                                               =
  =                                                                                                               =
  =  22.03.94  Thomas.                                                                                            =
  =  25.03.94  Ziel-Verwaltung eingebaut. Arnim.                                                                  =
  =  27.03.94  Ausgaben eingebaut.                                                                                =
  =  11.06.96  Statistik eingebaut, Ziel-Verwaltung ausgelagert. Arnim.                                           =
  =================================================================================================================*/

#include "Ausgaben.h"
#include "DSBaumAusgaben.h"
#include "DSBaumOperationen.h"
#include "DSBaumTest.h"
#include "InfoStrings.h"
#include "NFBildung.h"
#include "Parameter.h"
#include "pcl.h"
#include "Puffer.h"
#include "RUndEVerwaltung.h"
#include "SUEM.h"
#include "Subgleichungen.h"
#include "SymbolOperationen.h"
#include "ZeitMessung.h"



/*=================================================================================================================*/
/*= 0. Globale Variablen                                                                                          =*/
/*=================================================================================================================*/

DSBaumT RE_Regelbaum;
DSBaumT RE_Gleichungsbaum;
unsigned long int RE_AnzRegelnGesamt, RE_AnzGleichungenGesamt;



/*=================================================================================================================*/
/*= I. Testfunktion und bedingter Testbetrieb                                                                     =*/
/*=================================================================================================================*/

static BOOLEAN RegelmengenTest(void);
static BOOLEAN GleichungsmengenTest(BOOLEAN GegenrichtungenSuchen);

void RE_RUndETest(void)
{
  fprintf(IO_Strom, "Regelbaum testen ... \n");
  if (!RegelmengenTest())
    our_error("Index inkonsistent");
  fprintf(IO_Strom, "                 ... ok\n\nGleichungsbaum testen ... \n");
  if (!GleichungsmengenTest(TRUE))
    our_error("Index inkonsistent");
  fprintf(IO_Strom, "                      ... ok\n");
}


#define TestMitSprungverweisen TRUE            /* Alternativen: TRUE und FALSE; sollte moeglichst auf TRUE stehen! */
/* Alternativen fuer RE_TEST: 0,1; keine Auswirkung auf RE_RUndETest */
#define RE_TEST 0
#define Testbeginn (unsigned long int) 1                                        /* legt fest, ab wievieltem Aufruf */
                                                              /* wirklich getestet bzw. getestet und gedruckt wird */

#if RE_TEST
static long unsigned int TestaufrufNr = 0;
#define RTest(/*char[]*/ Ankuendigungstext)                                                                         \
  TestaufrufNr++;                                                                                                   \
  if (TestaufrufNr>=Testbeginn) {                                                                                   \
    fprintf(IO_Strom, "\nRETest %ld: %s\n",TestaufrufNr,Ankuendigungstext);                                         \
    if (!RegelmengenTest())                                                                                         \
      our_fatal_error("Regelmenge inkonsistent.");                                                                  \
  }
#define ETest(/*BOOLEAN*/ GegenrichtungenSuchen, /*char[]*/ Ankuendigungstext)                                      \
  TestaufrufNr++;                                                                                                   \
  if (TestaufrufNr>=Testbeginn) {                                                                                   \
    fprintf(IO_Strom, "\nRETest %ld: %s\n",TestaufrufNr,Ankuendigungstext);                                         \
    if (!GleichungsmengenTest(GegenrichtungenSuchen))                                                               \
      our_fatal_error("Gleichungsmenge inkonsistent.");                                                             \
  }
#else
#define RTest(/*char[]*/ Ankuendigungstext)
#define ETest(/*BOOLEAN*/ GegenrichtungenSuchen, /*char[]*/ Ankuendigungstext)
#endif



/*=================================================================================================================*/
/*= II. Regelmenge                                                                                                =*/
/*=================================================================================================================*/

static unsigned long int AnzRegelnAktuell;/* Lokaler Zaehler. Enthaelt immer die aktuelle Regelzahl im System. */

/* 1. Erzeugen von Regeln aus Termpaaren ------------------------------------------------------------------------- */

static RegelT ZuletztErzeugteRegel;

static unsigned int TiefeLinks;

RegelT RE_ErzeugteRegel(TermT links, TermT rechts, unsigned long Nummer, PCL_ObjektNummerT *PclID)
{
  TP_GNTermpaarHolen(ZuletztErzeugteRegel);
  BO_TermpaarNormieren(links,rechts,&TiefeLinks);
  ZuletztErzeugteRegel->rechteSeite = rechts;
  ZuletztErzeugteRegel->linkeSeite = links;
  ZuletztErzeugteRegel->Nummer = Nummer;
  ZuletztErzeugteRegel->SUEInfo = NULL;
  TP_GegenrichtungBesetzen(ZuletztErzeugteRegel,NULL);
  PCL_IDZuweisung(TP_PclID(ZuletztErzeugteRegel),*PclID);
  return ZuletztErzeugteRegel;
}


/* 3. Einfuegeoperation ------------------------------------------------------------------------------------------ */

void RE_ZuletztErzeugteRegelEinfuegen(void)
{
  AnzRegelnAktuell++;
  ZM_StartRegelEinf();
  RTest("Regelbaum vor dem Einfuegen");
  BO_ObjektEinfuegen(&RE_Regelbaum,ZuletztErzeugteRegel,TiefeLinks);
  RTest("Regelbaum nach dem Einfuegen");
  ZM_StoppRegelEinf();
  NF_RegelAufgenommen();
}


/* 4. Loeschoperation -------------------------------------------------------------------------------------------- */

void RE_RegelEntfernen(RegelT Regel)
{
  AnzRegelnAktuell--;
  ZM_StartRegelLoeschen();
  RTest("Regelbaum vor dem Entfernen");
  BO_ObjektEntfernen(&RE_Regelbaum,Regel);
  RTest("Regelbaum nach dem Entfernen");
  ZM_StoppRegelLoeschen();
  SUE_KillParent(Regel);
  NF_RegelEntfernt();
}


void RE_RegelLoeschen(RegelT Regel)                    /* Regel als Objekt loeschen sowie Meldung an KP-Verwaltung */
{
   TP_GNTermpaarLoeschen(Regel);
}


void RE_LoescheRegelmenge(void)
{
  IO_AusgabenAus();
  while (RE_Regelbaum.Wurzel) {
#if TP_PETERGRAF
    RegelT Regel = (RegelT) RE_Regelbaum.ErstesBlatt;
#else
    RegelT Regel = BK_Regeln(RE_Regelbaum.ErstesBlatt);
#endif
    RE_RegelEntfernen(Regel);
    TP_LinksRechtsLoeschen(Regel);
    TP_GNTermpaarLoeschen(Regel);
  }
  NF_RegelEntfernt();
  IO_AusgabenEin();
}


/* 5. Ausgabeoperation ------------------------------------------------------------------------------------------- */

void RE_AusgabeRegelmenge(int Grad)                                  /* Abhaengig von der Option Ergebnis-Ausgabe. */
{
   if (Grad != 7)
      IO_DruckeUeberschrift("Regelmenge");
   switch (Grad) {
   case 1:
   case 4:
   {
      RegelT RegelIndex;
      BK_forRegeln(RegelIndex,RE_Regelbaum, {
         fprintf(IO_Strom, "Nr. %4ld     ",TP_ElternNr(RegelIndex));
         IO_DruckeFlex("%t -> %t\n",TP_LinkeSeite(RegelIndex),TP_RechteSeite(RegelIndex));
      });
      break;
   }
   case 2:
      BA_BaumMalenDefaultOhneG(RE_Regelbaum);
      break;
   case 3:
      BA_BaumMalenDefaultMitG(RE_Regelbaum);
   case 5:
      BA_DumpBaum(RE_Regelbaum, FALSE);
      break;
   case 6:
      BA_DumpBaum(RE_Regelbaum, TRUE);
      break;
   case 7: /* wird fuer Ausgabe als Spezifikation benutzt */
   {
      RegelT RegelIndex;
      BK_forRegeln(RegelIndex,RE_Regelbaum, {
         IO_DruckeFlex("\t%t = %t\n",TP_LinkeSeite(RegelIndex),TP_RechteSeite(RegelIndex));
      });
      break;
   }
   default:
      break;
   }
   if (Grad != 7)
      fprintf(IO_Strom, "\n\n");
}


/* 6. Testfunktion ----------------------------------------------------------------------------------------------- */

static BOOLEAN RegelPraedikat(RegelT Regel)
{
  if (TP_Nachf(Regel)) {
    fprintf(IO_Strom, "\nFehler: Regel mit eingetragener Nachfolgerregel zur selben linken Seite.\n");
    return FALSE;
  }
  if (TP_Gegenrichtung(Regel)) {
    fprintf(IO_Strom, "\nFehler: Regel mit eingetragener Gegenrichtung.\n");
    return FALSE;
  }
  if (TP_ElternNr(Regel)<=0) {
    fprintf(IO_Strom, "\nFehler: Regel mit nichtpositiver Nummer.\n");
    return FALSE;
  }
  if (!TO_TermKorrekt(TP_LinkeSeite(Regel))) {
    fprintf(IO_Strom, "\nFehler: Regel hat fehlerhafte linke Seite.\n");
    return FALSE;
  }
  if (!TO_TermKorrekt(TP_RechteSeite(Regel))) {
    fprintf(IO_Strom, "\nFehler: Regel hat fehlerhafte rechte Seite.\n");
    return FALSE;
  }
  return TRUE;
}


static BOOLEAN RegelmengenTest(void)
{
  if (BT_IndexTesten(RE_Regelbaum,RegelPraedikat,TestMitSprungverweisen))
    return TRUE;
  RE_AusgabeRegelmenge(1);
  return FALSE;
}


/* 7. Umbenennungen ---------------------------------------------------------------------------------------------- */

void RE_RegelExternalisieren(RegelT Regel)
{
  TO_VariablenUmbenennen(TP_LinkeSeite(Regel),BO_NeuesteBaumvariable());
  TO_VariablenUmbenennen(TP_RechteSeite(Regel),BO_NeuesteBaumvariable());
}


void RE_RegelReinternalisieren(RegelT Regel)
{
  TO_VariablenUmbenennen(TP_LinkeSeite(Regel),-BO_NeuesteBaumvariable());
  TO_VariablenUmbenennen(TP_RechteSeite(Regel),-BO_NeuesteBaumvariable());
}


  
/*=================================================================================================================*/
/*= III. Gleichungsmenge                                                                                          =*/
/*=================================================================================================================*/

static unsigned long int AnzGleichungenAktuell,                      /* lokale Zaehler, die die Ausdehnung */
                         AnzAktMonogleichungen,                      /* der Gleichungsmenge beschreiben */
                         AnzMonogleichungen;                         /* Insgesamt aufgetretene Monogleichungen */

/* 1. Erzeugen von Gleichungen aus Termpaaren -------------------------------------------------------------------- */

static GleichungsT ZuletztErzeugteGleichung;

static unsigned int TiefeRechts;

GleichungsT RE_ErzeugteGleichung(TermT links, TermT rechts, unsigned long Nummer, PCL_ObjektNummerT *PclID)
{
  GNTermpaarT ZuletztErzeugteAntigleichung;
  TP_GNTermpaarHolen(ZuletztErzeugteAntigleichung);
  TP_GNTermpaarHolen(ZuletztErzeugteGleichung);
  BO_TermpaarNormieren(links,rechts,&TiefeLinks);
  ZuletztErzeugteGleichung->rechteSeite = rechts;
  ZuletztErzeugteGleichung->linkeSeite = links;
  ZuletztErzeugteGleichung->SUEInfo = NULL;
  ZuletztErzeugteGleichung->Nummer = -Nummer;
  ZuletztErzeugteGleichung->RichtungAusgezeichnet = TRUE;
  TP_WeissMalen(ZuletztErzeugteGleichung);
  /* Dies ist die einzige Stelle im System, an der die Abbildung der Gleichungs-Nummern auf negative Zahlen stattfindet!*/
  TP_SubgleichungsReferenzSei(ZuletztErzeugteGleichung,SGM_InitRef());
  PCL_IDZuweisung(TP_PclID(ZuletztErzeugteGleichung),*PclID);

  BO_TermpaarNormieren(
    ZuletztErzeugteAntigleichung->linkeSeite = TO_Termkopie(rechts),
    ZuletztErzeugteAntigleichung->rechteSeite = TO_Termkopie(links),
    &TiefeRechts);
  if ((ZuletztErzeugteGleichung->IstMonogleichung = 
    ZuletztErzeugteAntigleichung->IstMonogleichung =
    TP_TermpaareGleich(ZuletztErzeugteGleichung,ZuletztErzeugteAntigleichung)))
    AnzMonogleichungen++;
  ZuletztErzeugteAntigleichung->RichtungAusgezeichnet = FALSE;
  ZuletztErzeugteAntigleichung->Nummer = -Nummer;
  ZuletztErzeugteAntigleichung->SUEInfo = NULL;
  TP_WeissMalen(ZuletztErzeugteAntigleichung);
  TP_SubgleichungsReferenzSei(ZuletztErzeugteAntigleichung,SGM_InitRef());
  PCL_IDZuweisung(TP_PclID(ZuletztErzeugteAntigleichung),*PclID);
  
  (ZuletztErzeugteGleichung->Gegenrichtung = ZuletztErzeugteAntigleichung)->Gegenrichtung = ZuletztErzeugteGleichung;

  return ZuletztErzeugteGleichung;
}


/* 2. Praedikate und nebenwirkungsfreie Funktionen --------------------------------------------------------------- */

static UTermT *PermArg;
static unsigned int PermArgIndex,
                    PermArgDimension;

#define /*void*/ PermInit(/*void*/)                                                                                 \
  PermArgIndex = 0

#define /*void*/ PermPush(/*UTermT*/ Term)                                                                          \
  if (PermArgIndex==PermArgDimension)                                                                               \
    PermArg = (UTermT *) our_realloc(PermArg+1,(PermArgDimension += RE_ErstePermArgDimension)*sizeof(UTermT),       \
      "Ausdehnung des PermArg-Stapels")-1;                                                                          \
  PermArg[++PermArgIndex] = Term;

static BOOLEAN PermPop(TermT Term)
{
  unsigned int i;
  for (i = 1; i<=PermArgIndex; i++)
    if (TO_TermeGleich(Term,PermArg[i])) {                                                           /* PermShrink */
      unsigned int j;
      for (j = i+1; j<=PermArgIndex; j++)
        PermArg[j-1] = PermArg[j];
      PermArgIndex--;
      return TRUE;
    }
  return FALSE;
}

#define /*BOOLEAN*/ PermEmpty(/*void*/)                                                                             \
  (!PermArgIndex)

unsigned int RE_GradVonPermPaar(TermT links, TermT rechts)                  /* sinnvoll nur, falls A-normalisiert! */
{                                                                /* Grad==0 gdw Gleichung ist keine Permgleichung. */
  SymbolT ACSymbol = TO_TopSymbol(links);
  if (SO_IstACSymbol(ACSymbol) && TO_TopSymboleGleich(links,rechts)) {
    unsigned int Grad;
    PermInit();
    do {
      PermPush(links = TO_Schwanz(links));/* Erster Teilterm beliebig; zweiter o(t1,t2) oder mit anderem Topsymbol */
      if (SO_SymbUngleich(TO_TopSymbol(links = TO_NaechsterTeilterm(links)),ACSymbol)) {      /* dann PermTerm-Ende erreicht */
        PermPush(links);
        break;
      }
    } while (TRUE);
    Grad = PermArgIndex;
    do {
      if (!PermPop(rechts = TO_Schwanz(rechts)))           /* Erster Teilterm muss jetzt irgendwo gefunden werden. */
        return 0;
      if (SO_SymbUngleich(TO_TopSymbol(rechts = TO_NaechsterTeilterm(rechts)),ACSymbol))
        if (!PermPop(rechts) || !PermEmpty())
          return 0;
        else
          return Grad;
    } while (TRUE);
  }
  else
    return 0;
}



/* 3. Einfuegeoperation ------------------------------------------------------------------------------------------ */

void RE_ZuletztErzeugteHinrichtungEinfuegen(KopfZellenT *Referenz)
{
  AnzGleichungenAktuell ++;
  ZM_StartGleichgEinf();
  ETest(TRUE,"Gleichungsbaum vor dem Einfuegen der letzten Hinrichtung");
  BO_ObjektEinfuegen(&RE_Gleichungsbaum,ZuletztErzeugteGleichung,TiefeLinks);
  ETest(FALSE,"Gleichungsbaum nach dem Einfuegen der letzten Hinrichtung");
  TP_SubgleichungsReferenzSei(ZuletztErzeugteGleichung,Referenz);
  ZM_StoppGleichgEinf();
}


void RE_ZuletztErzeugteHerrichtungEinfuegen(void)                    /* Nur wenn Gleichung keine Monogleichung ist.*/
{
  if (TP_IstKeineMonogleichung(ZuletztErzeugteGleichung)) {
    ZM_StartGleichgEinf();
    ETest(FALSE,"Gleichungsbaum vor dem Einfuegen der letzten Herrichtung");
    BO_ObjektEinfuegen(&RE_Gleichungsbaum,TP_Gegenrichtung(ZuletztErzeugteGleichung),TiefeRechts);
    ETest(TRUE,"Gleichungsbaum nach dem Einfuegen der letzten Herrichtung");
    ZM_StoppGleichgEinf();
  }
  else
    AnzAktMonogleichungen++;
  NF_GleichungAufgenommen();
}


/* 4. Loeschoperationen ------------------------------------------------------------------------------------------ */

void RE_GleichungEntfernen(GleichungsT Gleichung)            /* Nur, wenn nicht Gegenrichtung einer Monogleichung. */
{
  if (TP_RichtungAusgezeichnet(Gleichung) || TP_IstKeineMonogleichung(Gleichung)) {
    ZM_StartGleichgLoeschen();
    ETest(FALSE,"Gleichungsbaum vor dem Entfernen");
    BO_ObjektEntfernen(&RE_Gleichungsbaum,Gleichung);
    ETest(FALSE,"Gleichungsbaum nach dem Entfernen");
    ZM_StoppGleichgLoeschen();
  }
  if (TP_RichtungAusgezeichnet(Gleichung))
     SUE_KillParent(Gleichung);

  NF_GleichungEntfernt();
}


void RE_GleichungLoeschen(GleichungsT Gleichung)     /* Gleichung als Objekt loeschen, dabei Meldung an SUE und NF */
{                                                             /* sowie SGM loeschen, falls ausgezeichnete Richtung */
  if (TP_RichtungAusgezeichnet(Gleichung)){
    AnzGleichungenAktuell--;
    if (TP_IstMonogleichung(Gleichung))
      AnzAktMonogleichungen--;
    SGM_LoescheSubgleichungen(TP_SubgleichungsReferenz(Gleichung));
  }
  TP_GNTermpaarLoeschen(Gleichung);
}


void RE_LoescheGleichungsmenge(void)
{
  IO_AusgabenAus();
  while (RE_Gleichungsbaum.Wurzel) {
#if TP_PETERGRAF
    GleichungsT Gleichung = (GleichungsT) RE_Gleichungsbaum.ErstesBlatt;
#else
    GleichungsT Gleichung = BK_Regeln(RE_Gleichungsbaum.ErstesBlatt);
#endif
    RE_GleichungEntfernen(TP_Gegenrichtung(Gleichung));                 /* Test auf Monogl. erfolgt in 'Entfernen' */
    RE_GleichungEntfernen(Gleichung);
    TP_LinksRechtsLoeschen(TP_Gegenrichtung(Gleichung));
    RE_GleichungLoeschen(TP_Gegenrichtung(Gleichung));
    TP_LinksRechtsLoeschen(Gleichung);
    RE_GleichungLoeschen(Gleichung);
  }
  IO_AusgabenEin();
}


/* 5. Ausgabeoperationen ----------------------------------------------------------------------------------------- */

void RE_AusgabeGleichungsmenge(int Grad)                              /* Abhaengig von der Option Ergebnis-Ausgabe.*/
{
   if (Grad != 7)
      IO_DruckeUeberschrift("Gleichungsmenge");
   switch (Grad) {
   case 1: 
   {
      RegelT RegelIndex;
      BK_forRegeln(RegelIndex,RE_Gleichungsbaum, {
         if (TP_RichtungAusgezeichnet(RegelIndex)) {
            fprintf(IO_Strom, "Nr. %4ld     ",TP_ElternNr(RegelIndex));
            IO_DruckeFlex("%t = %t\n",TP_LinkeSeite(RegelIndex),TP_RechteSeite(RegelIndex));
         }
      });
      break;
   }
   case 2: 
      BA_BaumMalenDefaultOhneG(RE_Gleichungsbaum);
      break;
   case 3:
      BA_BaumMalenDefaultMitG(RE_Gleichungsbaum);
      break;
   case 4:
   {
      GleichungsT GleichungsIndex;
      BK_forRegeln(GleichungsIndex,RE_Gleichungsbaum, {
         fprintf(IO_Strom, "Nr. %4ld %sausgez., %so. ", TP_ElternNr(GleichungsIndex),
                 TP_RichtungAusgezeichnet(GleichungsIndex) ? "" : "un",
                 TP_IstMonogleichung(GleichungsIndex) ? "mon" : "stere");
         IO_DruckeFlex("%t = %t\n",TP_LinkeSeite(GleichungsIndex),TP_RechteSeite(GleichungsIndex));
      })
   }
   case 5:
      BA_DumpBaum(RE_Gleichungsbaum, FALSE);
      break;
   case 6:
      BA_DumpBaum(RE_Gleichungsbaum, TRUE);
      break;
   case 7: /* fuer Ausgabe als Spezifikation */
   {
      RegelT RegelIndex;
      BK_forRegeln(RegelIndex,RE_Gleichungsbaum, {
         if (TP_RichtungAusgezeichnet(RegelIndex)) {
            IO_DruckeFlex("\t%t = %t\n",TP_LinkeSeite(RegelIndex),TP_RechteSeite(RegelIndex));
         }
      });
      break;
   }
   default:
      break;
   }
   if (Grad != 7)
      fprintf(IO_Strom, "\n\n");
}


/* 6. Testfunktion ----------------------------------------------------------------------------------------------- */

static BOOLEAN ImGleichungsbaumVorhanden(GleichungsT Gleichung)
{
  KantenT Knoten = RE_Gleichungsbaum.Wurzel;
  UTermT TermIndex = TP_LinkeSeite(Gleichung),
         SuffixIndex;
  GleichungsT GleichungsIndex;
  if (!Knoten)
    return FALSE;
  do {
    Knoten = BK_Nachf(Knoten,TO_TopSymbol(TermIndex));
    if (!Knoten)
      return FALSE;
    TermIndex = TO_Schwanz(TermIndex);
  } while (!BK_IstBlatt(Knoten));
  SuffixIndex = BK_Rest(Knoten);
  while (SuffixIndex) {
    if (!TO_TopSymboleGleich(SuffixIndex,TermIndex))
      return FALSE;
    SuffixIndex = TO_Schwanz(SuffixIndex);
    TermIndex = TO_Schwanz(TermIndex);
  }
  GleichungsIndex = BK_Regeln(Knoten);
  while (GleichungsIndex) {
    if (GleichungsIndex==Gleichung)
      return TRUE;
    GleichungsIndex = TP_Nachf(GleichungsIndex);
  }
  return FALSE;
}


static BOOLEAN GleichungsPraedikatSchwach(GleichungsT Gleichung)
{
  GleichungsT Antigleichung = TP_Gegenrichtung(Gleichung);
  TermT Links = TP_LinkeSeite(Gleichung),
        Rechts = TP_RechteSeite(Gleichung),
        AntiLinks = TP_LinkeSeite(Antigleichung),
        AntiRechts = TP_RechteSeite(Antigleichung);
  unsigned int TiefeTerm1;

  if (!Antigleichung) {
    fprintf(IO_Strom, "\nFehler: Gleichung mit Antigleichung NULL.\n");
    return FALSE;
  }
  if (TP_RichtungAusgezeichnet(Gleichung)==TP_RichtungAusgezeichnet(Antigleichung)) {
    fprintf(IO_Strom, "\nFehler: Gleichung und Antigleichung beide ausgezeichnet oder beide unausgezeichnet.\n");
    return FALSE;
  }
  if (TP_ElternNr(Gleichung)>=0) {
    fprintf(IO_Strom, "\nFehler: Gleichung mit nichtnegativer Nummer.\n");
    return FALSE;
  }
  if (TP_ElternNr(Gleichung)!=TP_ElternNr(Antigleichung)) {
    fprintf(IO_Strom, "\nFehler: Gleichung und Antigleichung haben verschiedene Nummern.\n");
    return FALSE;
  }
  if (TP_Gegenrichtung(Antigleichung)!=Gleichung) {
    fprintf(IO_Strom, "\nFehler: Gleichung und Antigleichung sind nicht korrekt miteinander verzeigert.\n");
    return FALSE;
  }
  if (!TO_TermKorrekt(Links)) {
    fprintf(IO_Strom, "\nFehler: Gleichung hat fehlerhafte linke Seite.\n");
    return FALSE;
  }
  if (!TO_TermKorrekt(Rechts)) {
    fprintf(IO_Strom, "\nFehler: Gleichung hat fehlerhafte rechte Seite.\n");
    return FALSE;
  }
  BO_TermpaarNormieren(AntiRechts,AntiLinks,&TiefeTerm1);
  if (!TO_TermeGleich(Links,AntiRechts) || !TO_TermeGleich(Rechts,AntiLinks)) {
    fprintf(IO_Strom, "\nFehler: Gleichung und Antigleichung sind nicht Varianten voneinander.\n");
    return FALSE;
  }
  BO_TermpaarNormieren(AntiLinks,AntiRechts,&TiefeTerm1);
  if (TP_IstMonogleichung(Gleichung)!=TP_IstMonogleichung(Antigleichung)) {
    fprintf(IO_Strom, "\nFehler: Gleichung und Antigleichung haben unterschiedliche"
      " Monogleichungskennzeichnung.\n");
    return FALSE;
  }
  if (TP_IstMonogleichung(Gleichung)) {
    GleichungsT BaumExterne = TP_RichtungUnausgezeichnet(Gleichung) ? Gleichung : Antigleichung;
    if (!TO_TermKorrekt(TP_LinkeSeite(BaumExterne))) {
      fprintf(IO_Strom, "\nFehler: Unausgezeichnete Richtung von Monogleichung hat fehlerhafte linke Seite.\n");
      return FALSE;
    }
    if (!TO_TermKorrekt(TP_RechteSeite(BaumExterne))) {
      fprintf(IO_Strom, "\nFehler: Unausgezeichnete Richtung von Monogleichung hat fehlerhafte rechte Seite.\n");
      return FALSE;
    }
    if (ImGleichungsbaumVorhanden(BaumExterne)) {
      fprintf(IO_Strom, "\nFehler: Unausgezeichnete Richtung von Monogleichung steht im Gleichungsbaum.\n");
      return FALSE;
    }
  }
  return TRUE;
}


static BOOLEAN GleichungsPraedikatNormal(GleichungsT Gleichung)
{
  if (!GleichungsPraedikatSchwach(Gleichung))
    return FALSE;
  if (TP_IstKeineMonogleichung(Gleichung) && !ImGleichungsbaumVorhanden(TP_Gegenrichtung(Gleichung))) {
    fprintf(IO_Strom, "\nFehler: Gegenrichtung von Stereogleichung steht nicht im Gleichungsbaum.\n");
    return FALSE;
  }
  return TRUE;
}


static BOOLEAN GleichungsmengenTest(BOOLEAN GegenrichtungenSuchen)
{
  if (BT_IndexTesten(RE_Gleichungsbaum,GegenrichtungenSuchen ? GleichungsPraedikatNormal : 
    GleichungsPraedikatSchwach,TestMitSprungverweisen))
    return TRUE;
  RE_AusgabeGleichungsmenge(4);
  return FALSE;
}



/*=================================================================================================================*/
/*= IV. Ergebnisausgabe R-, E- und KP-Menge                                                                       =*/
/*=================================================================================================================*/

void RE_SonstigeErgebnisse(BOOLEAN flag) /* flag dient zur Detailsteuerung in den verschiedenen Modi:
                                   convergence: TRUE  = im TP-Puffer befinden sich divergente KPs */
{                               /*              FALSE = im TP-Puffer befinden sich unrichtbare Ausgangsgleichungen */
  if ((PA_ExtModus == convergence) || (PA_ExtModus == termination) || (PA_ExtModus == confluence)){
    TermT links, rechts;
    unsigned long int nr;
    if (PA_ExtModus == termination)
      IO_DruckeKurzUeberschrift("Unrichtbare Ausgangsgleichungen", "+");
    else if (PA_ExtModus == confluence)
      IO_DruckeKurzUeberschrift("Divergente kritische Paare", "+");
    else                        /* convergence */
      IO_DruckeKurzUeberschrift(flag ? "Divergente kritische Paare" : "Unrichtbare Ausgangsgleichungen", "+");
    while(PU_TPPufferNichtLeer()){
      PU_TPAuslesen(&links, &rechts, &nr);
      fprintf(IO_Strom, "%5lu  ", nr);
      IO_DruckeFlex("%t = %t\n", links, rechts);
    }
    fprintf(IO_Strom, "\n\n");
  }
}


#if COMP_PCL
void RE_DruckeEndsystemFuerPCL(void)
{
  if (PA_ExtModus == completion){
    GNTermpaarT Index;
    RE_forRegelnRobust(Index, {
      PCL_VVAbschlussFaktDrucken(TP_PclID(Index),TP_LinkeSeite(Index),TP_RechteSeite(Index),Regel);
    });
    RE_forGleichungenRobust(Index, {
      if (TP_RichtungAusgezeichnet(Index))
        PCL_VVAbschlussFaktDrucken(TP_PclID(Index),TP_LinkeSeite(Index),TP_RechteSeite(Index),Gleichung);
    });
  }
}
#endif



/*=================================================================================================================*/
/*= VI. Initialisierungen                                                                                         =*/
/*=================================================================================================================*/

void RE_ReInitApriori(void)
{
  BO_DSBaumInitialisieren(&RE_Regelbaum);
  BO_DSBaumInitialisieren(&RE_Gleichungsbaum);
  RE_AnzRegelnGesamt = RE_AnzGleichungenGesamt = 
    AnzRegelnAktuell = AnzGleichungenAktuell = AnzAktMonogleichungen = AnzMonogleichungen = 0;
}  

void RE_InitApriori()
{
  PermArg = (UTermT *)our_alloc((PermArgDimension = RE_ErstePermArgDimension)*sizeof(UTermT))-1;
  RE_ReInitApriori();
}


/*=================================================================================================================*/
/*= VII. Statistik                                                                                                =*/
/*=================================================================================================================*/


void RE_Statistik(InfoStringT IS)
{
   if (!(IS_IsIn("all", IS) || IS_IsIn("RE", IS)))
      return;
   if (IS_IsIn("standard", IS)){
      int tiefe1;
      /******** Regelmenge ****************************** */
      if (RE_AnzRegelnGesamt > 0){
         unsigned int KnotenImBaum, GesparteKnoten;
         BO_KnotenZaehlen(&KnotenImBaum, &GesparteKnoten, RE_Regelbaum);
         IO_DruckeKurzUeberschrift("Set of Rules","+");
         tiefe1 = IO_Einruecktiefe(0, AnzRegelnAktuell);
         IO_DruckeStatistikZeile("Total Number of Rules",RE_AnzRegelnGesamt);
         IO_DruckeStatistikZeileEingerueckt("of them still in R", AnzRegelnAktuell, tiefe1);
         IO_DruckeStatistikLeerZeile();
         IO_DruckeStatistikZeile("Nodes in R-tree", KnotenImBaum);
         IO_DruckeStatistikZeile("Nodes saved", GesparteKnoten);
         IO_DruckeStatistikZeileAnteil("Percentage of saved nodes (compared with SDT)", 
                                       GesparteKnoten, GesparteKnoten+KnotenImBaum);
      }
      /******** Gleichungsmenge ************************* */
      if (RE_AnzGleichungenGesamt > 0){
         unsigned int KnotenImBaum, GesparteKnoten;
         BO_KnotenZaehlen(&KnotenImBaum, &GesparteKnoten, RE_Gleichungsbaum);
         IO_DruckeKurzUeberschrift("Set of Equations","+");
         IO_DruckeStatistikZeile("Total Number of Equations",RE_AnzGleichungenGesamt);
         tiefe1 = IO_Einruecktiefe(0, Maximum(2, AnzGleichungenAktuell, AnzMonogleichungen));
         IO_DruckeStatistikZeileEingerueckt("of them still in E", AnzGleichungenAktuell, tiefe1);
         IO_DruckeStatistikZeileEingerueckt("of them Mono-Equations", AnzMonogleichungen, tiefe1);
         IO_DruckeStatistikZeileEingerueckt("of them still in E", AnzAktMonogleichungen, tiefe1+IO_Platzverbrauch(AnzAktMonogleichungen)+2);
         IO_DruckeStatistikLeerZeile();
         IO_DruckeStatistikZeile("Nodes in E-tree", KnotenImBaum);
         IO_DruckeStatistikZeile("Nodes saved", GesparteKnoten);
         IO_DruckeStatistikZeileAnteil("Percentage of saved nodes (compared with SDT)", 
                                       GesparteKnoten, GesparteKnoten+KnotenImBaum);
      }
   }
   else if (IS_IsIn("short", IS)){
      IO_DruckeKurzUeberschrift("Sets of Rules and Equations","+");
      IO_DruckeStatistikZeile("Total Number of Rules",RE_AnzRegelnGesamt);
      IO_DruckeStatistikZeile("Total Number of Equations",RE_AnzGleichungenGesamt);
   }
   else if (IS_IsIn("latex-head", IS)){
      if (RE_AnzRegelnGesamt > 0)
         fprintf(IO_Strom, "R: total & current & ");
      if (RE_AnzGleichungenGesamt > 0)
         fprintf(IO_Strom, "E: total & current & mono & curr. mono & ");
   }
   else if (IS_IsIn("slatex-head", IS)){
      fprintf(IO_Strom, "R & E & ");
   }
   else if (IS_IsIn("intern", IS)){
      IO_DruckeStatistikZeile("Max. Expansion of Leaf-depth-array in R-tree", RE_Regelbaum.TiefeBlaetterDimension);
      IO_DruckeStatistikZeile("Max. Expansion of Leaf-depth-array in E-tree", RE_Gleichungsbaum.TiefeBlaetterDimension);
      IO_DruckeStatistikZeile("Max. Expansion of PermArgStack", PermArgDimension);
   }
   else if (IS_IsIn("config", IS)){
#if RE_TEST
      fprintf(IO_Strom, "RE: Test  ON, Teststart %lud, %s Jumplist-Test.\n", Testbeginn, TestMitSprungverweisen?"with":"without");
#else
      fprintf(IO_Strom, "RE: Test OFF.\n");
#endif
   }
}

void RE_AktuelleGroessen(unsigned long int *Regelmenge, unsigned long int *Gleichungsmenge)
{
  *Regelmenge = AnzRegelnAktuell;
  *Gleichungsmenge = AnzGleichungenAktuell;
}

