/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFBasis.c                                                                                                  */
/* - *********                                                                                                  */
/* -                                                                                                            */
/* - Enthaelt die Basis-Operationen zur Normalform-Bildung (alle Verfahren) :                                   */
/* -  * Pfadverwaltung (ueber Makros etc.)                                                                      */
/* -  * RegelAngewendet                                                                                         */
/* -  * GleichungAngewendet                                                                                     */
/* -                                                                                                            */
/* - 15.04.94 Arnim. Angelegt.                                                                                  */
/* - 02.05.94 Arnim. Erweitert/Geaendert/Ergaenzt fuer Frame-Konzept fuer Top-Down-Strategie.                   */
/* - 10.10.95 Arnim. Modul bereinigt. Dazu Backtracking-Makros und Frame-Konzept in die Strategie-Module        */
/* -                 ausgelagert.                                                                               */
/* -------------------------------------------------------------------------------------------------------------*/



#include "general.h"
#include "SpeicherParameter.h"     /* wg. der initialen Pfad-Groesse */
#include "TermOperationen.h"
#include "MatchOperationen.h"
#include "FehlerBehandlung.h"
#include "RUndEVerwaltung.h"
#include "ZeitMessung.h"
#include "Ausgaben.h"
#include "DSBaumOperationen.h"
#include "NFBasis.h"
#include "Hauptkomponenten.h"
#include "pcl.h"
#include "NFBasis.h"

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe NFBasis.h) ---------------------------*/
/* ------------------------------------------------------------------------*/

unsigned long        AnzRedVersMitRegeln[AnzObjektTypen];   /* Reduktionsversuche mit der Regelmenge */
unsigned long        AnzRedErfRegeln[AnzObjektTypen];       /* Erfolgreiche Reduktionsversuche mit der Regelmenge */
unsigned long        AnzRedVersMitGleichgn[AnzObjektTypen]; /* Reduktionsversuche mit der Gleichungsmenge */
unsigned long        AnzRedErfGleichgn[AnzObjektTypen];     /* Erfolgreiche Reduktionsversuche mit der Gleichungsmenge */
unsigned long        AnzRedMisserfolgeICG;                  /* Matches mit Gleichungs-Menge wg. Indirekter Clashes gescheitert */
unsigned long        AnzRedMisserfolgeICR;                  /* Matches mit Regel-Menge wg. Indirekter Clashes gescheitert */

StellenT *PositionImAktuellenTeilterm;
StellenT        aktuelleStelle;
StellenT        aktuelleGroessePfad;
NFB_PfadZellenT *EigentlPfad;
ObjektT         objektArt;
unsigned long int ObjektNummer;

BOOLEAN NFB_Reduziert;
UTermT aktTeiltermAH, aktTermAH, EinfuegepunktAH, EndeAH;
void (*NFB_DerGrosseExekutor)(void);

/*-------------------------------------------------------------------------------------------------------------------------*/
/*------ Initialisieren ---------------------------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static void (* Ausdehnungsproz) (StellenT) = NULL;

static BOOLEAN NFB_SchonInitialisiert = FALSE;

void NFB_Init(void)
/* Die Variable Pfad einrichten (dh. das Array EigentlicherPfad. */
{
   int Zaehler;
   
   /* Statistik initialisieren */
   for (Zaehler = 0; Zaehler<AnzObjektTypen; Zaehler++){
      AnzRedVersMitRegeln[Zaehler] = 0;
      AnzRedErfRegeln[Zaehler] = 0;
      AnzRedVersMitGleichgn[Zaehler] = 0;
      AnzRedErfGleichgn[Zaehler] = 0;
   }
   AnzRedMisserfolgeICG = 0;
   AnzRedMisserfolgeICR = 0;       

   if (NFB_SchonInitialisiert)
      return;
   NFB_SchonInitialisiert = TRUE;
   
   EigentlPfad = (NFB_PfadZellenT *) our_alloc (NF_PfadArrayGroesse*sizeof(NFB_PfadZellenT));
   aktuelleGroessePfad = NF_PfadArrayGroesse -1;

}

void NFB_AusdehnungsprozedurAnmelden(void (* fkt) (StellenT))
/* Diese Prozedur wird aufgerufen, wenn der Reduktionspfad ausgedehnt werden muss (Erspart Abfragen).
   Nur eine Prozedur anmeldbar.
*/
{
   Ausdehnungsproz = fkt;
}


/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- PfadEinsWeiterSetzen -----------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/*                                                                                                                         */
/* Vor.: die aktuelle Stelle im Pfad zeigt auf einen Knoten im Term (Term als Baum interpretiert), dessen Nachfolger noch  */
/*       nicht alle abgehandelt sind. Dann wird ein neuer Eintrag im Pfad erzeugt, der den ersten noch nicht behandelten   */
/*       Teilterm des aktuellen Terms als Term hat; der aktuelle Term ist jetzt dieser Teilterm.                           */
/*                                                                                                                         */
/*-------------------------------------------------------------------------------------------------------------------------*/


void NFB_PfadEinsWeiterSetzen(void)
/* Der naechste noch zu behandelnde Teilterm an der aktuellen Stelle wird als neue aktuelle Stelle eingerichtet.
   Vor.: aktuelle Stelle zeigt auf einen Eintrag, der noch nicht abgearbeitete Teilterme besitzt. */
{
   if (aktuelleStelle == aktuelleGroessePfad){                 /* Im Pfad kein Platz mehr zum Tiefersteigen => realloc versuchen*/
      if (aktuelleGroessePfad >= MaxStelle - NF_PfadArrayGroesse)
         our_error ("Ueberlauf in StellenT, ggf. pcl.h anpassen");
      aktuelleGroessePfad += NF_PfadArrayGroesse;
      EigentlPfad = (NFB_PfadZellenT *) our_realloc(EigentlPfad, sizeof(NFB_PfadZellenT)*(aktuelleGroessePfad+1),
                                                    "Erweitern des Normalform-Pfades");
      if (Ausdehnungsproz)
         (* Ausdehnungsproz) (aktuelleGroessePfad+1);
   }
   EigentlPfad[aktStelle()].AnzNochZuBehandelnderTeilterme--; /* ein Teilterm wird jetzt behandelt */
   if (EigentlPfad[aktStelle()].EinfuegePunkt == NULL){      /* bisher noch kein Teilterm behandelt */
      EigentlPfad[aktStelle()].EinfuegePunkt =               /* => der behandelte Teilterm faengt direkt hinter Topsymbol an*/
         EigentlPfad[aktStelle()].Stelle;       
   }
   else{                                                     /* es wurde bereits ein Teilterm behandelt */
      EigentlPfad[aktStelle()].EinfuegePunkt =               /* => EinfuegePunkt Ende des zuletzt behandelten Teilterms setzen */
         TO_TermEnde( TO_Schwanz( EigentlPfad[aktStelle()].EinfuegePunkt)); /* (der direkt hinter altem EinfuegePunkt anfing.) */
   }
   EigentlPfad[aktStelle()+1].Stelle =                       /* der jetzt zu behandelnde Teilterm faengt direkt hinter dem */
      TO_Schwanz(EigentlPfad[aktStelle()].EinfuegePunkt);    /* neuen EinfuegePunkt an */
   PfadEinsAb();
   EigentlPfad[aktStelle()].AnzNochZuBehandelnderTeilterme = 
      TO_AnzahlTeilterme(EigentlPfad[aktStelle()].Stelle);
   EigentlPfad[aktStelle()].EinfuegePunkt = NULL;
}

BOOLEAN AdHocPfadWeitersetzen(void)
/* Setzt den AdHoc-Pfad eins weiter. Rueckgabe ist TRUE, wenn die letzte Stelle noch nicht ueberschritten wurde.*/
{
   EinfuegepunktAH = aktTeiltermAH;
   aktTeiltermAH = TO_Schwanz(aktTeiltermAH);
   return (TO_PhysischUngleich(aktTeiltermAH, EndeAH));
}




/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------- Termersetzungsschnittstele zu MatchOperationen----------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

/* Falls die Reduktion an der Wurzel des zu reduzierenden Terms stattfindet, wird die Termersetzung uniform von
   MO_SigmaRInEZ vorgenommen. Andernfalls ist MO_SigmaRInLZ zustaendig. Dabei sind jedoch noch einige je verschiedene
   Verwaltungsinformationen zu aktualisieren. Daher folgen kurze Makros und eine Funktion. */

#define /*void*/ TermersetzungInnenAnPfadstelleMitPfadUpdate(/*void*/)                                              \
  TO_NachfBesetzen(EigentlPfad[aktStelle()-1].EinfuegePunkt,EigentlPfad[aktStelle()].Stelle = MO_SigmaRInLZ())


#define /*void*/ TermersetzungInnenAnAdhocstelle(/*void*/)                                                          \
  TO_NachfBesetzen(EinfuegepunktAH,aktTeiltermAH = MO_SigmaRInLZ())

#define /*void*/ TermersetzungInnenAnAdhocstelleOhneUpdate(/*void*/)\
   TO_NachfBesetzen(EinfuegepunktAH, MO_SigmaRInLZ())

/* ---------- fuer Exekutor (und ggf. PCL) ---------------------*/

/* Der nachfolgende Teil ist umstaendlich, muss aber so sein, da fuer das PCL-Protokoll der entsprechende Reduktions-
   schritt erst ausgegeben werden kann, wenn er tatsaechlich mit dem Exekutor ausgefuehrt wird. Deshalb muss der
   entsprechende PCL-Aufruf auch im Exekutor erfolgen.*/

/* Falls PCL-Protokoll an: Angewendetes Objekt merken. In der Excekutor-Prozedur Meldung absetzen */
/* andernfalls           : Aufrufe wie gehabt. */

#if COMP_PCL

static RegelOderGleichungsT zuletztVersuchtesObjekt;
static PCL_SeitenT zuletztVersuchteSeite;


#define zuletztVersuchtesObjektFuerPclMerken(/*RegelOderGleichungsT*/ objekt, /*PCL_SeitenT*/ seite) \
  (zuletztVersuchtesObjekt = objekt, zuletztVersuchteSeite = seite)

static void InnenExekutor(void)
{
   TO_NachfBesetzen(EinfuegepunktAH, MO_SigmaRInLZ());
   PCL_Reduktion(aktTermAH, TO_Schwanz(EinfuegepunktAH), TP_PclID(zuletztVersuchtesObjekt), zuletztVersuchteSeite);
}

static void WurzelExekutor(void)
{
  MO_SigmaRInEZ();
  PCL_Reduktion(aktTermAH, aktTermAH, TP_PclID(zuletztVersuchtesObjekt), zuletztVersuchteSeite);
}

#else

#define zuletztVersuchtesObjektFuerPclMerken(/*RegelOderGleichungsT*/ objekt, /*PCL_SeitenT*/ seite) /* Nix tun */

static void InnenExekutor(void)
{
   TO_NachfBesetzen(EinfuegepunktAH, MO_SigmaRInLZ());
}

#define  WurzelExekutor MO_SigmaRInEZ

#endif

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Regel angewendet ? -------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die Prozedur Regel angewendet versucht destruktiv, einen Reduktionschritt an der aktuellen, im Pfad gespeicherten       
   Stelle des auf Normalform zu bringenden Terms auszufuehren. Zurueckgegeben wird, ob dies moeglich war.                  
   Dabei muss unterschieden werden, ob es sich bei der Stelle, an der reduziert werden soll, um die Wurzel des Terms       
   handelte oder nicht:                                                                                                    

   1. Stelle = Wurzel von Term                                                                                             
      (d.h., es zeigen keine Endezeiger von ausserhalb auf das Ende des Termes, sehr wohl verweisen aber moeglicherweise   
       mehrere Zeiger von ausserhalb auf den Anfang (dh. die erste Zelle des Termes). Deshalb ist hier erste Buerger-     
       pflicht, die bisherige erste Termzelle auch als erste Termzelle im reduzierten Term zu haben.                       
       Die Funktion MO_SigmaRInEZ() erzeugt die aussubstituierte rechte Seite der Regel so, dass sie mit der ersten
       Zelle des Terms beginnt, und zwar fuer alle moeglichen Laengen des Terms und der rechten Seite. Ausserdem
       werden die nicht mehr benoetigten Termzellen geloescht.
     
   2. Stelle ist nicht die Wurzel im Term                                                             
      (d.h. es zeigen moeglicherweise Ende-Zeiger von ausserhalb auf das Ende des aktuellen Teilterms. Aus diesem Grunde   
       ist es hier erforderlich, die letzte Zelle des aktuellen Teilterms auch nach der Reduktion noch in dieser Funktion  
       zu haben.
       Die Funktion MO_SigmaRInLZ() erzeugt die aussubstituierte rechte Seite der Regel so, dass sie mit der letzten
       Zelle des Terms endet, und zwar fuer alle moeglichen Laengen des Terms und der rechten Seite. Ausserdem
       werden die nicht mehr benoetigten Termzellen geloescht.

   Von jeder benoetigten Funktion existieren hier zwei Varianten, die den beiden verschiedenen Pfad-Formen Rechnung tragen.
  -------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFB_RegelAngewendet(void)
/* Versucht, eine Regel auf den aktuellen Teilterm im Pfad anzuwenden. Dies geschieht dann ggf. destruktiv.
   Das Ergebnis des Reduktionsversuches wird zurueckgegeben.
   Nach erfolgreicher Reduktion steht in der aktuellen Stelle des Pfades der frisch reduzierte Term, wobei Stelligkeit
   und Einfuegepunkt bereits wieder neu initialisiert sind.*/
{
  IncAnzRedVersMitRegeln(objektArt);
  if (TO_TermIstNichtVar(EigentlPfad[aktStelle()].Stelle))
     if (MO_RegelGefunden(EigentlPfad[aktStelle()].Stelle)) {
        /* Reduktionsversuch erfolgreich? */
        NFB_Reduziert = TRUE;
        if (Wurzel()){                                                                               /* Fall 1: Wurzel */
           ZM_StartRewrite();
           MO_SigmaRInEZ();                                                    /* dann bedarfsgerecht einpassen lassen */
           ZM_StoppRewrite();
        }
        else{
           ZM_StartRewrite();
           TermersetzungInnenAnPfadstelleMitPfadUpdate();
           ZM_StoppRewrite();
        }
        aktStelleVonVorne();
        IncAnzRedErfRegeln(objektArt, MO_AngewendetesObjekt);
        PCL_Reduktion(EigentlPfad[0].Stelle, EigentlPfad[aktStelle()].Stelle, 
                      TP_PclID(MO_AngewendetesObjekt), linkeSeite);
        return TRUE;
     }
     else{
        IncAnzRedMisserfolgeICR(MO_IndirekterClashImBaum());
        return FALSE;
     }
  else
     return FALSE;
}

BOOLEAN NFB_RegelAngewendetAH(void)
/* Versucht, eine Regel auf den aktuellen Teilterm im Adhoc-Pfad anzuwenden. Dies geschieht dann ggf. destruktiv.
   Das Ergebnis des Reduktionsversuches wird zurueckgegeben.
   Nach erfolgreicher Reduktion steht in der aktuellen Stelle des Pfades der frisch reduzierte Term.*/
{
  IncAnzRedVersMitRegeln(objektArt);
  if (TO_TermIstNichtVar(aktTeiltermAH))
     if (MO_RegelGefunden(aktTeiltermAH)){     /* Reduktionsversuch erfolgreich? */
        NFB_Reduziert = TRUE;
        if (EinfuegepunktAH){                                        /* Fall 1: nicht Wurzel (EinfuegepunktAH != NULL) */
           ZM_StartRewrite();
           TermersetzungInnenAnAdhocstelle();
           ZM_StoppRewrite();
        }
        else{                                                                                         /* Fall 2: Wurzel */
           ZM_StartRewrite();
           MO_SigmaRInEZ();                                                     /* dann bedarfsgerecht einpassen lassen */
           ZM_StoppRewrite();
        }
        IncAnzRedErfRegeln(objektArt, MO_AngewendetesObjekt);
        PCL_Reduktion(aktTermAH, aktTeiltermAH, TP_PclID(MO_AngewendetesObjekt), linkeSeite);
        return TRUE;
     }
     else{
        IncAnzRedMisserfolgeICR(MO_IndirekterClashImBaum());
        return FALSE;
     }
  else
     return FALSE; 
}

/* ACHTUNG: Bei den folgenden Prozeduren wird der Pfad nach erfolgter Reduktion _nicht_ angepasst. */
/* ----------------------------------------------------------------------------------------------- */

static BOOLEAN EineRegelAngewendetAH(RegelT Regel)
/* Versucht, die angegebene Regel auf den aktuellen Teilterm im Adhoc-Pfad anzuwenden. Dies geschieht dann ggf. destruktiv.
   Das Ergebnis des Reduktionsversuches wird zurueckgegeben.
   Die Anzahl der Reduktionsversuche sowie die Anzahl der Reduktionserfolge (bezogen auf den in PfadNeu angegebenen
   Objekttyp) wird hier gegebenenfalls erhoeht.*/
{
  IncAnzRedVersMitRegeln(objektArt);
  if (TO_TermIstNichtVar(aktTeiltermAH) && MO_RegelPasst(Regel,aktTeiltermAH)) {                    /* Reduktionsversuch erfolgreich? */
    if (EinfuegepunktAH){                                                   /* Fall 1: nicht Wurzel (EinfuegepunktAH != NULL) */
      ZM_StartRewrite();
      TermersetzungInnenAnAdhocstelleOhneUpdate();
      ZM_StoppRewrite();
    }
    else{
      ZM_StartRewrite();                                                              /* Fall 2: Wurzel */
      MO_SigmaRInEZ();                                                      /*         dann bedarfsgerecht einpassen lassen */
      ZM_StoppRewrite();
    }
    PCL_Reduktion(aktTermAH, (EinfuegepunktAH)?TO_Schwanz(EinfuegepunktAH):aktTermAH, TP_PclID(Regel), linkeSeite);
    IncAnzRedErfRegeln(objektArt, Regel);
    return TRUE;
  }
  else
    return FALSE;
}


static BOOLEAN EineRegelAnwendbarAH(RegelT Regel)
{
  IncAnzRedVersMitRegeln(objektArt);
  if (TO_TermIstNichtVar(aktTeiltermAH) && MO_RegelPasst(Regel, aktTeiltermAH)) {
    zuletztVersuchtesObjektFuerPclMerken(Regel, linkeSeite);
    NFB_DerGrosseExekutor = (EinfuegepunktAH) ? InnenExekutor : WurzelExekutor;
    IncAnzRedErfRegeln(objektArt, Regel);
    return TRUE;
  }
  else
    return FALSE;
}



/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- GleichungAngewendet ------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Diese Prozedur versucht destruktiv, einen Reduktionschritt mit einer Gleichung an der aktuell im Pfad gespeicherten     */
/* Stelle auszufuehren.                                                                                                    */
/*                                                                                                                         */
/* Dabei wird ein moeglicher Reduktionnsschritt nur dann ausgefuehrt, wenn er vertraeglich mit der Ordnung ist.            */
/* Bei den Prozeduren, die die Anwendbarkeit einer vorgegebenen Gleichung pruefen, wird die angegebene Gleichung in        */
/* beiden Richtungen geprueft (natuerlich incl. Ordnungstest).                                                             */
/*-------------------------------------------------------------------------------------------------------------------------*/


BOOLEAN NFB_GleichungAngewendet(void)
{
  IncAnzRedVersMitGleichgn(objektArt);
  if (TO_TermIstNichtVar(EigentlPfad[aktStelle()].Stelle))
     if (MO_GleichungGefunden(EigentlPfad[aktStelle()].Stelle)) {
        /* Reduktionsversuch erfolgreich? */
        NFB_Reduziert = TRUE;
        if (Wurzel()){                                                                                /* Fall 1: Wurzel */
           ZM_StartRewrite();
           MO_SigmaRInEZ();                                                     /* dann bedarfsgerecht einpassen lassen */
           ZM_StoppRewrite();
        }
        else{
           ZM_StartRewrite();
           TermersetzungInnenAnPfadstelleMitPfadUpdate();
           ZM_StoppRewrite();
        }
        aktStelleVonVorne();
        IncAnzRedErfGleichgn(objektArt, MO_AngewendetesObjekt);
        PCL_Reduktion(EigentlPfad[0].Stelle, EigentlPfad[aktStelle()].Stelle, 
                      TP_PclID(MO_AngewendetesObjekt),
                      TP_RichtungAusgezeichnet(MO_AngewendetesObjekt)?linkeSeite:rechteSeite);
        return TRUE;
     }
     else{
        IncAnzRedMisserfolgeICG(MO_IndirekterClashImBaum());
        return FALSE;
     }
  else
     return FALSE;
}

BOOLEAN NFB_GleichungAngewendetAH(void)
{
  IncAnzRedVersMitGleichgn(objektArt);
  if (TO_TermIstNichtVar(aktTeiltermAH))
     if (MO_GleichungGefunden(aktTeiltermAH)) { /* Reduktionsversuch erfolgreich? */
        NFB_Reduziert = TRUE;
        if (EinfuegepunktAH){                                          /* Fall 1: nicht Wurzel (EinfuegepunktAH != NULL) */
           ZM_StartRewrite();
           TermersetzungInnenAnAdhocstelle();
           ZM_StoppRewrite();
        }
        else{                                                          /* Fall 2: Wurzel */
           ZM_StartRewrite();
           MO_SigmaRInEZ();                                            /*         dann bedarfsgerecht einpassen lassen */
           ZM_StoppRewrite();
        }
        IncAnzRedErfGleichgn(objektArt, MO_AngewendetesObjekt);
        PCL_Reduktion(aktTermAH, aktTeiltermAH, 
                      TP_PclID(MO_AngewendetesObjekt),
                      TP_RichtungAusgezeichnet(MO_AngewendetesObjekt)?linkeSeite:rechteSeite);
        return TRUE;
     }
     else{
        IncAnzRedMisserfolgeICG(MO_IndirekterClashImBaum());
        return FALSE;
     }
  else
     return FALSE;
}

/* ACHTUNG: Bei den folgenden Prozeduren wird der Pfad nach erfolgter Reduktion _nicht_ angepasst. */

static BOOLEAN EineGleichungAngewendetAH(GleichungsT Gleichung)
/* Versucht, die angegebene Gleichung auf den aktuellen Teilterm im Adhoc-Pfad anzuwenden. Dies geschieht dann ggf. destruktiv.
   Das Ergebnis des Reduktionsversuches wird zurueckgegeben.
   Die Anzahl der Reduktionsversuche sowie die Anzahl der Reduktionserfolge (bezogen auf den in PfadNeu angegebenen
   Objekttyp) wird hier gegebenenfalls erhoeht.*/
{
  IncAnzRedVersMitGleichgn(objektArt);

  if (TO_TermIstNichtVar(aktTeiltermAH) && MO_GleichungPasst(Gleichung,aktTeiltermAH)){ /* Reduktionsversuch erfolgreich? */

    if (EinfuegepunktAH){                                                    /* Fall 1: nicht Wurzel (EinfuegepunktAH != NULL) */
      ZM_StartRewrite();
      TermersetzungInnenAnAdhocstelleOhneUpdate();
      ZM_StoppRewrite();
    }
    else{                                                                    /* Fall 2: Wurzel */
      ZM_StartRewrite();
      MO_SigmaRInEZ();                                                      /*         dann bedarfsgerecht einpassen lassen */             
      ZM_StoppRewrite();
    }

    PCL_Reduktion(aktTermAH, (EinfuegepunktAH)?TO_Schwanz(EinfuegepunktAH):aktTermAH, TP_PclID(Gleichung),
                  TP_RichtungAusgezeichnet(MO_AngewendetesObjekt)?linkeSeite:rechteSeite);
    IncAnzRedErfGleichgn(objektArt, Gleichung);
    return TRUE;
  }
  else
    return FALSE;
}


static BOOLEAN GleichungAnwendbarAH(GleichungsT Gleichung)
/* Wenn die Gleichung auf den aktuellen Teilterm im Adhoc-Pfad anwendbar ist, so wird der grosse Exekutor vorbereitet und
   TRUE zurueckgegeben. */
{
  IncAnzRedVersMitGleichgn(objektArt);
  if (TO_TermIstNichtVar(aktTeiltermAH) && MO_GleichungPasst(Gleichung, aktTeiltermAH)) {
    zuletztVersuchtesObjektFuerPclMerken(Gleichung, TP_RichtungAusgezeichnet(MO_AngewendetesObjekt)?linkeSeite:rechteSeite);
    NFB_DerGrosseExekutor = (EinfuegepunktAH) ? InnenExekutor : WurzelExekutor;
    IncAnzRedErfGleichgn(objektArt, Gleichung);
    return TRUE;
  }
  else
    return FALSE;
}


/*-------------------------------------------------------------------------------------------------------------------------*/
/*------ Anwendbarkeit eines bestimmten Objektes irgendwo auf einen Term --------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Folgende Prozeduren verfolgen die Vorordnungs-Strategie, um die ihnen uebergebenen Terme auf Reduzierbarkeit mit dem    */
/* uebergebenen Objekt zu pruefen. Die Prozeduren mit 'Anwendbar' im Namen bereiten nur die Reduktionsschritt vor, welcher */
/* dann mit dem grossen Exekutor ausgefuehrt werden kann. Die Prozeduren mit 'Angewendet' im Namen fuehren diesen Schritt  */
/* unmittelbar aus. Alle Prozeduren kehren danach aber sofort mit TRUE zurueck, ohne weiterzusuchen.                       */
/*-------------------------------------------------------------------------------------------------------------------------*/
   
BOOLEAN NFB_EineRegelAnwendbar(RegelT regel, TermT Term, ObjektT objekt)
{
   AdHocPfadNeu(Term, objekt);
   do{
      if (EineRegelAnwendbarAH(regel))
         return TRUE;
   }while(AdHocPfadWeitersetzen());
   return FALSE;
}


BOOLEAN NFB_EineRegelAngewendet(RegelT regel, TermT Term, ObjektT objekt)
{
   AdHocPfadNeu(Term, objekt);
   do{
      if (EineRegelAngewendetAH(regel))
         return TRUE;
   }while(AdHocPfadWeitersetzen());
   return FALSE;
}
   
BOOLEAN NFB_EineGleichungAnwendbar(GleichungsT gleichung, TermT Term, ObjektT objekt)
{
   AdHocPfadNeu(Term, objekt);
   do{
      if (GleichungAnwendbarAH(gleichung))
         return TRUE;
   }while(AdHocPfadWeitersetzen());
   return FALSE;
}


BOOLEAN NFB_EineGleichungAngewendet(GleichungsT gleichung, TermT Term, ObjektT objekt)
{
   AdHocPfadNeu(Term, objekt);
   do{
      if (EineGleichungAngewendetAH(gleichung))
         return TRUE;
   }while(AdHocPfadWeitersetzen());
   return FALSE;
}

BOOLEAN NFB_TermInnenReduzibelE(UTermT Term)
{
  UTermT NULLTerm = TO_NULLTerm(Term);

  while (TO_PhysischUngleich(Term = TO_Schwanz(Term),NULLTerm))
    if (TO_TermIstNichtVar(Term) && (MO_GleichungGefunden(Term)))
      return TRUE;
  
  return FALSE;
}

BOOLEAN NFB_TermInnenReduzibelR(UTermT Term)
{
  UTermT NULLTerm = TO_NULLTerm(Term);

  while (TO_PhysischUngleich(Term = TO_Schwanz(Term),NULLTerm))
    if (TO_TermIstNichtVar(Term) && (MO_RegelGefunden(Term)))
      return TRUE;

  return FALSE;
}

BOOLEAN NFB_TermInnenReduzibelRE(UTermT Term)
{
  UTermT NULLTerm = TO_NULLTerm(Term);
  TermT TermIndex = TO_Schwanz(Term);

  do
    if (TO_TermIstNichtVar(TermIndex) && (MO_RegelGefunden(TermIndex)))
      return TRUE;
  while (TO_PhysischUngleich(TermIndex = TO_Schwanz(TermIndex),NULLTerm));

  while (TO_PhysischUngleich(Term = TO_Schwanz(Term),NULLTerm))
    if (TO_TermIstNichtVar(Term) && (MO_GleichungGefunden(Term)))
      return TRUE;

  return FALSE;
}

int NFB_AnzTermReduzibelR(TermT Term)
{
   TermT TermIndex;
   int anzahl = 0;

   TO_doStellen(Term, TermIndex, {if (TO_TermIstNichtVar(Term) && (MO_RegelGefunden(Term))) anzahl ++;});
   return anzahl;
}

int NFB_AnzTermReduzibelE(TermT Term)
{
   TermT TermIndex;
   int anzahl = 0;
   
   TO_doStellen(Term, TermIndex, {if (TO_TermIstNichtVar(Term) && (MO_GleichungGefunden(Term))) anzahl ++;});
   return anzahl;
}

int NFB_AnzTermReduzibelRE(TermT Term)
{
   TermT TermIndex;
   int anzahl = 0;
   
   TO_doStellen(Term, TermIndex, {if (TO_TermIstNichtVar(Term) && (MO_RegelGefunden(Term))) anzahl ++;});
   TO_doStellen(Term, TermIndex, {if (TO_TermIstNichtVar(Term) && (MO_GleichungGefunden(Term))) anzahl ++;});
   
   return anzahl;
}

