/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFBildung.c                                                                                                */
/* - ***********                                                                                                */
/* -                                                                                                            */
/* - Enthaelt die Operationen zur Normalform-Bildung (normales Verfahren)                                       */
/* -                                                                                                            */
/* - 23.03.94 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#include "Ausgaben.h"

#include "general.h"		   /* wg. Typ BOOLEAN */
#include "TermOperationen.h"	   /* wg. Typ TermT */
#include "FehlerBehandlung.h"
#include "RUndEVerwaltung.h"
#include "NFInnermost.h"
#include "NFOutermost.h"
#include "NFOutermostSpecial1.h"
#include "NFOutermostSpecial2.h"
#include "NFOutermostRE.h"
#include "NFTopDown.h"
#include "NFAdhoc.h"
#include "NFBildung.h"
#include "NFBasis.h"              /* wg. Exekutor und Statistik*/

#include <string.h>

/* ------------------------------------------------------------------------*/
/* ---------- globale Variablen (siehe NFBildung.h) -----------------------*/
/* ------------------------------------------------------------------------*/

NFProzedurT           NFBildungProz, NFZuRegelnProz, NFZuGleichungenProz, NFBildungStratProz, NFZuRegelnStratProz, NFZuGleichungenStratProz;
TermpaarIrgendwoT     GleichungAlsTermpaarAnwendbarProz, RegelAlsTermpaarAnwendbarProz;
NFInnenReduzibelT     NFInnenReduzibelProzE, NFInnenReduzibelProzRE, NFInnenReduzibelProzR;
NFRedStellenT         NFAnzRedStellenProzR, NFAnzRedStellenProzE, NFAnzRedStellenProzRE;
IrrTestProcT IrrNFBildungProz, IrrNFZuRegelnProz, IrrNFZuGleichungenProz;
BOOLEAN _RMleer, _GMleer;
#if NF_TEST
/***************** Irreduzibilitaets-Test *********************************/
BOOLEAN ___booly___;
#endif

/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Dummy ---------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFInnenReduzibelDummyProz(UTermT Term)
{
  return FALSE;
}

BOOLEAN NFDummyProz(TermT Term, ObjektT objekt)
{
   return FALSE;
}

int NFRedStellenDummyProz(UTermT Term)
{
  return 0;
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Irreduzibilitaets-Test ----------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
#include "MatchOperationen.h"

long int NFEEges = 0;
long int NFRRges = 0;
long int NFRERges = 0, NFREEges = 0;

static const char *blubb = "|/-\\|/-\\";
static int modulo = 8;
static int zaehler = 0;

#define Blubb() printf("%c\r", blubb[zaehler = (zaehler+1)%modulo])

void NF_IrrTestR(TermT Term)
{
   UTermT TIndex = Term;
   int count = 0;
   Blubb();
   while (TIndex){
      if (TO_TermIstNichtVar(TIndex) && MO_RegelGefunden(TIndex)){
         IO_DruckeFlex("%t mit R reduzierbar an Stelle %t.\n", Term, TIndex);
         count++;
      }
      TIndex = TO_Schwanz(TIndex);
   }
   NFRRges += count;
   if (count>0)fprintf(IO_Strom, "R: Rges = %ld\n", NFRRges);
}

void NF_IrrTestE(TermT Term)
{
   UTermT TIndex = Term;
   int count = 0;
   Blubb();
   while (TIndex){
      if (TO_TermIstNichtVar(TIndex) && MO_GleichungGefunden(TIndex)){
         IO_DruckeFlex("%t mit E reduzierbar an Stelle %t.\n", Term, TIndex);
         count++;
      }
      TIndex = TO_Schwanz(TIndex);
   }
   NFEEges += count;
   if (count>0)fprintf(IO_Strom, "E: Eges = %ld\n", NFEEges);
}

void NF_IrrTestRE(TermT Term)
{
   UTermT TIndex = Term;
   int Rcount = 0, Ecount = 0;
   Blubb();
   while (TIndex){
      if (TO_TermIstNichtVar(TIndex)){
         if (MO_RegelGefunden(TIndex)){
            IO_DruckeFlex("%t mit R reduzierbar an Stelle %t.\n", Term, TIndex);
            Rcount++;
         }
         if (MO_GleichungGefunden(TIndex)){
            IO_DruckeFlex("%t mit E reduzierbar an Stelle %t.\n", Term, TIndex);
            Ecount++;
         }
      }
      TIndex = TO_Schwanz(TIndex);
   }
   NFRERges += Rcount; NFREEges += Ecount;
   if (Rcount+Ecount>0){ 
      fprintf(IO_Strom, "RE: Rges = %ld, Eges = %ld\n", NFRERges, NFREEges);
   }
}

void NF_IrrTestDummy(TermT T)
{
   ;
}

BOOLEAN NF_IrrtestFehler(void)
{
   return (BOOLEAN) (NFEEges+NFRRges+NFRERges+NFREEges>0);
}
 
/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Extra-Test-Prozedur zum Vergleich verschiedener Strategien ----------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/
#define VariablenDeklarieren()                                                                                  \
   unsigned long        AnzRedVersMitRegelnAlt[AnzObjektTypen];                                                 \
   unsigned long        AnzRedErfRegelnAlt[AnzObjektTypen];                                                     \
   unsigned long        AnzRedMisserfolgeICRAlt;                                                                \
   unsigned long        AnzRedVersMitGleichgnAlt[AnzObjektTypen];                                               \
   unsigned long        AnzRedErfGleichgnAlt[AnzObjektTypen];                                                   \
   unsigned long        AnzRedMisserfolgeICGAlt;                                                                \
                                                                                                                \
   unsigned long        AnzRedVersMitRegelnNeu[AnzObjektTypen];                                                 \
   unsigned long        AnzRedErfRegelnNeu[AnzObjektTypen];                                                     \
   unsigned long        AnzRedMisserfolgeICRNeu;                                                                \
   unsigned long        AnzRedVersMitGleichgnNeu[AnzObjektTypen];                                               \
   unsigned long        AnzRedErfGleichgnNeu[AnzObjektTypen];                                                   \
   unsigned long        AnzRedMisserfolgeICGNeu;                                                                \
                                                                                                                \
   TermT kopie;                                                                                                 \
   BOOLEAN result1, result2, ZaehlerWeichenAb = FALSE;                                                          \
   int Zaehler;

#define AlteZaehlerstaendeMerken()                                                                              \
{                                                                                                               \
   for (Zaehler = 0; Zaehler < AnzObjektTypen; Zaehler++){                                                      \
      AnzRedVersMitRegelnAlt[Zaehler] = AnzRedVersMitRegeln[Zaehler];                                           \
      AnzRedErfRegelnAlt[Zaehler]  =             AnzRedErfRegeln[Zaehler];                                      \
      AnzRedVersMitGleichgnAlt[Zaehler]  =       AnzRedVersMitGleichgn[Zaehler];                                \
      AnzRedErfGleichgnAlt[Zaehler]  =           AnzRedErfGleichgn[Zaehler];                                    \
   }                                                                                                            \
   AnzRedMisserfolgeICGAlt  =        AnzRedMisserfolgeICG;                                                      \
   AnzRedMisserfolgeICRAlt  =        AnzRedMisserfolgeICR;                                                      \
}


#define NeueZaehlerstaendeMerken()                                                                              \
{                                                                                                               \
   for (Zaehler = 0; Zaehler < AnzObjektTypen; Zaehler++){                                                      \
      AnzRedVersMitRegelnNeu[Zaehler]=      AnzRedVersMitRegeln[Zaehler];                                       \
      AnzRedErfRegelnNeu[Zaehler]=          AnzRedErfRegeln[Zaehler];                                           \
      AnzRedVersMitGleichgnNeu[Zaehler]=    AnzRedVersMitGleichgn[Zaehler];                                     \
      AnzRedErfGleichgnNeu[Zaehler]=        AnzRedErfGleichgn[Zaehler];                                         \
   }                                                                                                            \
   AnzRedMisserfolgeICRNeu=     AnzRedMisserfolgeICR;                                                           \
   AnzRedMisserfolgeICGNeu=     AnzRedMisserfolgeICG;                                                           \
}

#define ZaehlerstaendeRestaurieren()                                                                            \
{                                                                                                               \
   for (Zaehler = 0; Zaehler < AnzObjektTypen; Zaehler++){                                                      \
      AnzRedVersMitRegeln[Zaehler]=     AnzRedVersMitRegelnAlt[Zaehler];                                        \
      AnzRedErfRegeln[Zaehler]=         AnzRedErfRegelnAlt [Zaehler];                                           \
      AnzRedVersMitGleichgn[Zaehler]=   AnzRedVersMitGleichgnAlt[Zaehler];                                      \
      AnzRedErfGleichgn[Zaehler]=       AnzRedErfGleichgnAlt [Zaehler];                                         \
   }                                                                                                            \
   AnzRedMisserfolgeICR=    AnzRedMisserfolgeICRAlt ;                                                           \
   AnzRedMisserfolgeICG=    AnzRedMisserfolgeICGAlt ;                                                           \
}

#define ZaehlerstaendeTesten()                                                                                  \
{                                                                                                               \
   for (Zaehler = 0; Zaehler < AnzObjektTypen; Zaehler++)                                                       \
      if ((AnzRedErfRegelnNeu[Zaehler]!=          AnzRedErfRegeln[Zaehler]) ||                                  \
          (AnzRedErfGleichgnNeu[Zaehler]!=        AnzRedErfGleichgn[Zaehler]))                                  \
         ZaehlerWeichenAb = TRUE;                                                                               \
                                                                                                                \
   if ( (result1 != result2)  ||  ZaehlerWeichenAb ||                                                           \
       (AnzRedMisserfolgeICRNeu!=     AnzRedMisserfolgeICR)||                                                   \
       (AnzRedMisserfolgeICGNeu!=     AnzRedMisserfolgeICG)){                                                   \
      printf("\nDISKREPANZ bei FB_Z[1] == %ld\n\n", FB_Z[1]);                                                   \
      for (Zaehler = 0; Zaehler < AnzObjektTypen; Zaehler++){                                                   \
         printf(" %10ld <---> %10ld\n", AnzRedErfRegelnNeu[Zaehler], AnzRedErfRegeln[Zaehler]);                 \
         printf(" %10ld <---> %10ld\n", AnzRedErfGleichgnNeu[Zaehler], AnzRedErfGleichgn[Zaehler]);             \
      }                                                                                                         \
      printf(" %10ld <---> %10ld\n", AnzRedMisserfolgeICRNeu   , AnzRedMisserfolgeICR  );                       \
      printf(" %10ld <---> %10ld\n", AnzRedMisserfolgeICGNeu   , AnzRedMisserfolgeICG);                         \
      printf(" %s  <---> %s\n", result1?"T":"F",  result2?"T":"F");                                             \
      IO_DruckeFlex("Term nach Strategie LO: %t <---> Term nach Strategie LOS2: %t \n", Term, kopie);           \
      our_error("bla");                                                                                         \
   }                                                                                                            \
}

static BOOLEAN ExtraTest(TermT Term, ObjektT objekt)
/* Vergleicht die Zaehlerstaende der Reduktionsstatistik bzgl. Diskrepanzen in den 
   Strategien Leftmost-Outermost und Leftmost-Outermost-Special-2
*/
{
#if NFB_STATISTIK

   VariablenDeklarieren()

   /* 0. Fehlerzaehler FB_Z[1] erhoehen */

   FB_IncZ(1);

   /* 1. Term kopieren */

   kopie = TO_Termkopie(Term);

   /* 2. Zaehlerstaende merken */

   AlteZaehlerstaendeMerken();
   
   /* 3. Term nach Strategie 1 reduzieren */

   result1 = NFO_NormalformAufNochmal(Term, objekt);

   /* 4. Zaehlerstaende merken */

   NeueZaehlerstaendeMerken();

   /* 5. Zaehlerstaende restaurieren nach Werten aus 2*/
   
   ZaehlerstaendeRestaurieren();

   /* 6. Termkopie nach Strategie 2 reduzieren */

   result2 = NFOS2_NormalformAufNochmal(kopie, objekt);

   /* 7. Zaehlerstaende mit den in 4 geretteten vergleichen, */
   /*    Bei Diskrepanzen Ausgabe des Fehler-Zaehlers FB_Z[1] */
   
   ZaehlerstaendeTesten();

   /* 9. Termkopie loeschen */
   TO_TermLoeschen(kopie);

   return result1;
#else
 return TRUE;
#endif
}
   
static BOOLEAN ExtraTestR(TermT Term, ObjektT objekt)
/* Vergleicht die Zaehlerstaende der Reduktionsstatistik bzgl. Diskrepanzen in den 
   Strategien Leftmost-Outermost und Leftmost-Outermost-Special-2
*/
{
#if NFB_STATISTIK

   VariablenDeklarieren()

   /* 0. Fehlerzaehler FB_Z[2] erhoehen */

   FB_IncZ(2);

   /* 1. Term kopieren */

   kopie = TO_Termkopie(Term);

   /* 2. Zaehlerstaende merken */

   AlteZaehlerstaendeMerken();  
   
   /* 3. Term nach Strategie 1 reduzieren */
   
   result1 = NFO_NormalformZuRegelnAufNochmal(Term, objekt);

   /* 4. Zaehlerstaende merken */

   NeueZaehlerstaendeMerken();

   /* 5. Zaehlerstaende restaurieren nach Werten aus 2*/
   
   ZaehlerstaendeRestaurieren();

   /* 6. Termkopie nach Strategie 2 reduzieren */

   result2 = NFOS2_NormalformZuRegelnAufNochmal(kopie, objekt);

   /* 7. Zaehlerstaende mit den in 4 geretteten vergleichen, */
   /*    Bei Diskrepanzen Ausgabe des Fehler-Zaehlers FB_Z[1] */
   
   ZaehlerstaendeTesten();

   /* 9. Termkopie loeschen */
   TO_TermLoeschen(kopie);

   return result1;
#else
   return TRUE;
#endif
}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Initialisierung -----------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void NF_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
/* what == ParamEvaluate: Auswerten und Ausgeben, aber nix einstellen.
   what == ParamExplain: Syntax ausgeben.
*/
{
   if (what == ParamEvaluate){
      int Strategie = IS_FindOption(DEF, USR, 8, "outermost", "innermost", "topdown", "adhoc", "outermost1", 
                                    "outermost2", "outermostRE", "outermostER");
      BOOLEAN Wiederholen = (IS_FindOption(DEF, USR, 2, "repeat", "continue") == 0);
      BOOLEAN Aufsteigen  = (IS_FindOption(DEF, USR, 2, "up", "down") == 0);
      fprintf(IO_Strom, "\n* NORMALIZATION:    ");
      
      switch (Strategie)
      {
      case 1: fprintf(IO_Strom, "innermost"); break;
      case 2: fprintf(IO_Strom, "topdown"); break;
      case 3: fprintf(IO_Strom, "adhoc"); break;
      case 4: fprintf(IO_Strom, "outermost1"); break;
      case 5: fprintf(IO_Strom, "outermost2"); break;
      case 6: fprintf(IO_Strom, "outermostRE"); break;
      case 7: fprintf(IO_Strom, "outermostER"); break;
      case 0:
      default: fprintf(IO_Strom, "outermost"); break;
      }
      
      fprintf(IO_Strom, ", %s, %s\n", Wiederholen?"repeat":"continue", Aufsteigen?"up":"down");
   }
   else{
      IO_DruckeKurzUeberschrift("Normalization", "-");
      fprintf(IO_Strom, "1. [outermost|innermost|topdown|adhoc|outermost1|outermost2|outermostRE|outermostER]\n");
      fprintf(IO_Strom, "2. [up|down]\n");
      fprintf(IO_Strom, "3. [repeat|continue]\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "1. Strategy\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "outermost   = leftmost-outermost\n");
      fprintf(IO_Strom, "innermost   = leftmost-innermost\n");
      fprintf(IO_Strom, "topdown     = breadth-first search, level after level\n");
      fprintf(IO_Strom, "adhoc       = normalization without overhead\n");
      fprintf(IO_Strom, "outermost1  = outermost with marking indirect clashes and pointers\n");
      fprintf(IO_Strom, "outermost2  = outermost without marking indirect clashes\n");
      fprintf(IO_Strom, "outermostRE = try each position with equations immediately after rules have failed\n");
      fprintf(IO_Strom, "outermostER = like outermostRE, but try equations first\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "2. Backtracking\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "up   = climb up from the current position\n");
      fprintf(IO_Strom, "down = jump to the top-most position and go down\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "3. Behaviour after reduction\n");
      fprintf(IO_Strom, "\n");
      fprintf(IO_Strom, "repeat   = try current position again\n");
      fprintf(IO_Strom, "continue = backtrack immediately\n");
      fprintf(IO_Strom, "\n");
   }      
}   

void NF_InitAposteriori(InfoStringT DEF, InfoStringT USR)

/* Initialisieren der fuer die eingestellte NF-Strategie und -Variante benoetigten Module. 
Strategie: 0 = leftmost-outermost
           1 = leftmost-innermost
           2 = top-down
           3 = adhoc
           4 = leftmost-outermost-special-1 (Fassung mit Faedelung)
           5 = leftmost-outermost-special-2 (Fassung ohne IC-Verwaltung)
*/
{
   int Strategie = IS_FindOption(DEF, USR, 8, "outermost", "innermost", "topdown", "adhoc", "outermost1", 
                                 "outermost2", "outermostRE", "outermostER");
   BOOLEAN Wiederholen = (IS_FindOption(DEF, USR, 2, "repeat", "continue") == 0);
   BOOLEAN Aufsteigen  = (IS_FindOption(DEF, USR, 2, "up", "down") == 0);

   _RMleer = _GMleer = TRUE;


   fprintf(IO_Strom, "\n* NORMALIZATION:    ");

   switch (Strategie){

   case 0:                      /* Leftmost-Outermost */
      fprintf(IO_Strom, "Leftmost-Outermost, ");
      NFO_Init();
      if (Wiederholen){
        if (Aufsteigen){
          NFZuRegelnStratProz= NFO_NormalformZuRegelnAufNochmal; /* LeftmostOutermost, nach Reduktion aktuelle Stelle nochmal, Aufstieg */
          NFZuGleichungenStratProz = NFO_NormalformZuGleichungenAufNochmal;
          NFBildungStratProz = NFO_NormalformAufNochmal;
        }
        else{
          NFZuRegelnStratProz= NFO_NormalformZuRegelnAbNochmal;  /* LeftmostOutermost, nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFO_NormalformZuGleichungenAbNochmal; 
          NFBildungStratProz = NFO_NormalformAbNochmal;
        }
      }
      else{
        if (Aufsteigen){
          NFZuRegelnStratProz= NFO_NormalformZuRegelnAufDirekt;  /* LeftmostOutermost, nach Reduktion direkt Aufstieg */
          NFZuGleichungenStratProz = NFO_NormalformZuGleichungenAufDirekt; 
          NFBildungStratProz = NFO_NormalformAufDirekt;
        }
        else{
          NFZuRegelnStratProz= NFO_NormalformZuRegelnAbDirekt;   /* LeftmostOutermost, nach Reduktion direkt Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFO_NormalformZuGleichungenAbDirekt;  
          NFBildungStratProz = NFO_NormalformAbDirekt;
        }
      }
      break;

   case 1:                      /* Leftmost-Innermost */
      fprintf(IO_Strom, "Leftmost-Innermost, ");
      NFI_Init();
      if(Wiederholen){
        NFZuRegelnStratProz= NFI_NormalformZuRegelnNochmal;
        NFZuGleichungenStratProz = NFI_NormalformZuGleichungenNochmal;
        NFBildungStratProz = NFI_NormalformBildenNochmal;
      }
      else{
        NFZuRegelnStratProz= NFI_NormalformZuRegelnDirekt;
        NFZuGleichungenStratProz = NFI_NormalformZuGleichungenDirekt;
        NFBildungStratProz = NFI_NormalformBildenDirekt;
      }
      break;

   case 2:                      /* Top-Down */
      fprintf(IO_Strom, "Top-Down, ");
      NFT_Init();
      if (Aufsteigen){
        if (Wiederholen){
          NFZuRegelnStratProz= NFT_NormalformZuRegelnAufNochmal; /* Top-Down, nach Reduktion aktuelle Stelle nochmal, Aufstieg */
          NFZuGleichungenStratProz = NFT_NormalformZuGleichungenAufNochmal;
          NFBildungStratProz = NFT_NormalformAufNochmal;
        }
        else{
          NFZuRegelnStratProz= NFT_NormalformZuRegelnAufDirekt;  /* Top-Down, nach Reduktion direkt Aufstieg */
          NFZuGleichungenStratProz = NFT_NormalformZuGleichungenAufDirekt; 
          NFBildungStratProz = NFT_NormalformAufDirekt;
        }
      }
      else{
        if (Wiederholen){
          NFZuRegelnStratProz= NFT_NormalformZuRegelnAbNochmal; /* Top-Down, nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFT_NormalformZuGleichungenAbNochmal; 
          NFBildungStratProz = NFT_NormalformAbNochmal;
        }
        else{
          NFZuRegelnStratProz= NFT_NormalformZuRegelnAbDirekt;   /* Top-Down, nach Reduktion direkt Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFT_NormalformZuGleichungenAbDirekt;  
          NFBildungStratProz = NFT_NormalformAbDirekt;
        }
      }
      break;

   case 3:                      /* Adhoc */
      fprintf(IO_Strom, "Adhoc, ");
      NFA_Init();
      if (Wiederholen){
        NFZuRegelnStratProz= NFA_NormalformZuRegelnNochmal; /* Adhoc, nach Reduktion Stelle fertigreduzieren */
        NFZuGleichungenStratProz = NFA_NormalformZuGleichungenNochmal; 
        NFBildungStratProz = NFA_NormalformBildenNochmal;
      }
      else{
        NFZuRegelnStratProz= NFA_NormalformZuRegelnDirekt; /* Adhoc, nach Reduktion sofort weitergehen */
        NFZuGleichungenStratProz = NFA_NormalformZuGleichungenDirekt;  
        NFBildungStratProz = NFA_NormalformBildenDirekt;
       }
      break;

   case 4:                      /* Leftmost-Outermost-Special-1 */
      fprintf(IO_Strom, "Leftmost-Outermost-Special-1, ");
      NFOS1_Init();
      if (Wiederholen){
        if (Aufsteigen){
          NFZuRegelnStratProz= NFOS1_NormalformZuRegelnAufNochmal; /* LeftmostOutermost, nach Reduktion aktuelle Stelle nochmal, Aufstieg */
          NFZuGleichungenStratProz = NFOS1_NormalformZuGleichungenAufNochmal;
          NFBildungStratProz = NFOS1_NormalformAufNochmal;
        }
        else{
          NFZuRegelnStratProz= NFOS1_NormalformZuRegelnAbNochmal;  /* LeftmostOutermost, nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFOS1_NormalformZuGleichungenAbNochmal; 
          NFBildungStratProz = NFOS1_NormalformAbNochmal;
        }
      }
      else{
        if (Aufsteigen){
          NFZuRegelnStratProz= NFOS1_NormalformZuRegelnAufDirekt;  /* LeftmostOutermost, nach Reduktion direkt Aufstieg */
          NFZuGleichungenStratProz = NFOS1_NormalformZuGleichungenAufDirekt; 
          NFBildungStratProz = NFOS1_NormalformAufDirekt;
        }
        else{
          NFZuRegelnStratProz= NFOS1_NormalformZuRegelnAbDirekt;   /* LeftmostOutermost, nach Reduktion direkt Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFOS1_NormalformZuGleichungenAbDirekt;  
          NFBildungStratProz = NFOS1_NormalformAbDirekt;
        }
      }
      break;

   case 5:                      /* Leftmost-Outermost-Special-2 */
      fprintf(IO_Strom, "Leftmost-Outermost-Special-2, ");
      NFOS2_Init();
      if (Wiederholen){
        if (Aufsteigen){
          NFZuRegelnStratProz= NFOS2_NormalformZuRegelnAufNochmal; /* LeftmostOutermost, nach Reduktion aktuelle Stelle nochmal, Aufstieg */
          NFZuGleichungenStratProz = NFOS2_NormalformZuGleichungenAufNochmal;
          NFBildungStratProz = NFOS2_NormalformAufNochmal;
        }
        else{
          NFZuRegelnStratProz= NFOS2_NormalformZuRegelnAbNochmal;  /* LeftmostOutermost, nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFOS2_NormalformZuGleichungenAbNochmal; 
          NFBildungStratProz = NFOS2_NormalformAbNochmal;
        }
      }
      else{
        if (Aufsteigen){
          NFZuRegelnStratProz= NFOS2_NormalformZuRegelnAufDirekt;  /* LeftmostOutermost, nach Reduktion direkt Aufstieg */
          NFZuGleichungenStratProz = NFOS2_NormalformZuGleichungenAufDirekt; 
          NFBildungStratProz = NFOS2_NormalformAufDirekt;
        }
        else{
          NFZuRegelnStratProz= NFOS2_NormalformZuRegelnAbDirekt;   /* LeftmostOutermost, nach Reduktion direkt Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFOS2_NormalformZuGleichungenAbDirekt;  
          NFBildungStratProz = NFOS2_NormalformAbDirekt;
        }
      }
      break;
   case 6:                      /* Leftmost-Outermost mit Gleichberechtigung (R vor E) */
      fprintf(IO_Strom, "Leftmost-Outermost RE, ");
      NFORE_Init();
      if (Wiederholen){
        if (Aufsteigen){
          NFZuRegelnStratProz= NFORE_NormalformZuRegelnAufNochmal; /* LeftmostOutermost, nach Reduktion aktuelle Stelle nochmal, Aufstieg */
          NFZuGleichungenStratProz = NFORE_NormalformZuGleichungenAufNochmal;
          NFBildungStratProz = NFORE_NormalformAufNochmal;
        }
        else{
          NFZuRegelnStratProz= NFORE_NormalformZuRegelnAbNochmal;  /* LeftmostOutermost, nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFORE_NormalformZuGleichungenAbNochmal; 
          NFBildungStratProz = NFORE_NormalformAbNochmal;
        }
      }
      else{
        if (Aufsteigen){
          NFZuRegelnStratProz= NFORE_NormalformZuRegelnAufDirekt;  /* LeftmostOutermost, nach Reduktion direkt Aufstieg */
          NFZuGleichungenStratProz = NFORE_NormalformZuGleichungenAufDirekt; 
          NFBildungStratProz = NFORE_NormalformAufDirekt;
        }
        else{
          NFZuRegelnStratProz= NFORE_NormalformZuRegelnAbDirekt;   /* LeftmostOutermost, nach Reduktion direkt Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFORE_NormalformZuGleichungenAbDirekt;  
          NFBildungStratProz = NFORE_NormalformAbDirekt;
        }
      }
      break;
   case 7:                      /* Leftmost-Outermost mit Gleichberechtigung (E vor R)*/
      fprintf(IO_Strom, "Leftmost-Outermost ER, ");
      NFORE_Init();
      if (Wiederholen){
        if (Aufsteigen){
          NFZuRegelnStratProz= NFORE_NormalformZuRegelnAufNochmal; /* LeftmostOutermost, nach Reduktion aktuelle Stelle nochmal, Aufstieg */
          NFZuGleichungenStratProz = NFORE_NormalformZuGleichungenAufNochmal;
          NFBildungStratProz = NFOER_NormalformAufNochmal;
        }
        else{
          NFZuRegelnStratProz= NFORE_NormalformZuRegelnAbNochmal;  /* LeftmostOutermost, nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFORE_NormalformZuGleichungenAbNochmal; 
          NFBildungStratProz = NFOER_NormalformAbNochmal;
        }
      }
      else{
        if (Aufsteigen){
          NFZuRegelnStratProz= NFORE_NormalformZuRegelnAufDirekt;  /* LeftmostOutermost, nach Reduktion direkt Aufstieg */
          NFZuGleichungenStratProz = NFORE_NormalformZuGleichungenAufDirekt; 
          NFBildungStratProz = NFOER_NormalformAufDirekt;
        }
        else{
          NFZuRegelnStratProz= NFORE_NormalformZuRegelnAbDirekt;   /* LeftmostOutermost, nach Reduktion direkt Abstieg von Wurzel */
          NFZuGleichungenStratProz = NFORE_NormalformZuGleichungenAbDirekt;  
          NFBildungStratProz = NFOER_NormalformAbDirekt;
        }
      }
      break;
    default:
      our_error("Nicht implementierte NF-Variante!\n");
      break;
   }

   NFBildungProz = NFZuRegelnProz = NFZuGleichungenProz = NFDummyProz;
   IrrNFBildungProz = IrrNFZuRegelnProz = IrrNFZuGleichungenProz = NF_IrrTestDummy;
   NFInnenReduzibelProzR = NFInnenReduzibelProzE = NFInnenReduzibelProzRE = NFInnenReduzibelDummyProz;
   NFAnzRedStellenProzR = NFAnzRedStellenProzE = NFAnzRedStellenProzRE = NFRedStellenDummyProz;

   if (0){
      fprintf(IO_Strom, "IrrTest on, ");
      NFZuRegelnStratProz = ExtraTestR;
      NFZuGleichungenStratProz = ExtraTest;
   }
   else
      fprintf(IO_Strom, "IrrTest off, ");

   fprintf(IO_Strom, "%s, %s\n", Wiederholen?"repeat":"continue", Aufsteigen?"up":"down");

}

/* -------------------------------------------------------------------------------------------------------------*/
/* --------- Statistik -----------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/

void NF_Statistik(InfoStringT IS)
{
   unsigned long int 
      AnzRedVersMitRegelnGes = 0, AnzRedVersMitGleichgnGes = 0, 
      AnzRedErfMitRegelnGes = 0, AnzRedErfMitGleichgnGes = 0,
      maxAnz = 0;
   int Zaehler;

   if (!(IS_IsIn("all", IS) || IS_IsIn("NF", IS)))
      return;

#if NFB_STATISTIK
   for (Zaehler = 0; Zaehler<AnzObjektTypen; Zaehler++){
      AnzRedVersMitRegelnGes += AnzRedVersMitRegeln[Zaehler];
      AnzRedVersMitGleichgnGes += AnzRedVersMitGleichgn[Zaehler];
      AnzRedErfMitRegelnGes += AnzRedErfRegeln[Zaehler];
      AnzRedErfMitGleichgnGes += AnzRedErfGleichgn[Zaehler];
      if (AnzRedVersMitRegeln[Zaehler] > maxAnz) maxAnz = AnzRedVersMitRegeln[Zaehler];
      if (AnzRedVersMitGleichgn[Zaehler] > maxAnz) maxAnz = AnzRedVersMitGleichgn[Zaehler];
   }
#endif   
   if (IS_IsIn("standard", IS)){
#if NFB_STATISTIK
      int tiefe1;
      const char *RedObjekteIO[AnzObjektTypen] = {"ues after generation", "ues during IRP", 
                                               "ues after selection", "rhs of rules",
                                               "lhs of rules", "equations", "goals"};
      
      if (WerteDa(2, AnzRedVersMitRegelnGes, AnzRedVersMitGleichgnGes)){
         IO_DruckeKurzUeberschrift("rewriting","+");

         tiefe1 = IO_Einruecktiefe(0, Maximum(3, maxAnz, AnzRedMisserfolgeICR, AnzRedMisserfolgeICG));

         IO_DruckeStatistikZeile("match queries R",AnzRedVersMitRegelnGes);
         for (Zaehler = 0; Zaehler<AnzObjektTypen; Zaehler++){
            char str[100];
            strcpy(str, "of them concerning ");
            strcat(str, RedObjekteIO[Zaehler]);
            IO_DruckeStatistikZeileEingerueckt(str, AnzRedVersMitRegeln[Zaehler], tiefe1);
         }
         IO_DruckeStatistikZeileEingerueckt("of them failed due to indir. clashes",AnzRedMisserfolgeICR, 
                                            tiefe1);         
         IO_DruckeStatistikLeerZeile();
         IO_DruckeStatistikZeile("Reductions with R",AnzRedErfMitRegelnGes);
         for (Zaehler = 0; Zaehler<AnzObjektTypen; Zaehler++){
            char str[100];
            strcpy(str, "of them concerning ");
            strcat(str, RedObjekteIO[Zaehler]);
            IO_DruckeStatistikZeileEingerueckt(str, AnzRedErfRegeln[Zaehler], tiefe1);
         }

         IO_DruckeStatistikLeerZeile();
         IO_DruckeStatistikZeile("match queries E",AnzRedVersMitGleichgnGes);
         for (Zaehler = 0; Zaehler<AnzObjektTypen; Zaehler++){
            char str[100];
            strcpy(str, "of them concerning ");
            strcat(str, RedObjekteIO[Zaehler]);
            IO_DruckeStatistikZeileEingerueckt(str, AnzRedVersMitGleichgn[Zaehler], tiefe1);
         }
         IO_DruckeStatistikZeileEingerueckt("of them failed due to indir. clashes",AnzRedMisserfolgeICG, tiefe1);
         IO_DruckeStatistikLeerZeile();
         IO_DruckeStatistikZeile("Reductions with E",AnzRedErfMitGleichgnGes);
         for (Zaehler = 0; Zaehler<AnzObjektTypen; Zaehler++){
            char str[100];
            strcpy(str, "of them concerning ");
            strcat(str, RedObjekteIO[Zaehler]);
            IO_DruckeStatistikZeileEingerueckt(str, AnzRedErfGleichgn[Zaehler], tiefe1);
         }
         IO_DruckeStatistikLeerZeile();
         IO_DruckeStatistikZeile("total match queries (CoM)", AnzRedVersMitRegelnGes+AnzRedVersMitGleichgnGes);
         IO_DruckeStatistikZeile("total reductions (Red)", AnzRedErfMitRegelnGes+AnzRedErfMitGleichgnGes);
      }
#else
      IO_DruckeStatistikZeile("rewriting-statistics OFF", 0);
#endif
   }
   else if (IS_IsIn("short", IS)){
#if NFB_STATISTIK
      IO_DruckeKurzUeberschrift("rewriting","+");
      IO_DruckeStatistikZeile("Match Queries", AnzRedVersMitRegelnGes + AnzRedVersMitGleichgnGes);
      IO_DruckeStatistikZeile("Reductions", AnzRedErfMitRegelnGes+ AnzRedErfMitGleichgnGes);
#else
      ;
#endif
   }
   else if (IS_IsIn("latex-head", IS)){
#if NFB_STATISTIK
      fprintf(IO_Strom, "Red R: CoM & Red & IC--Fail & Red E: CoM & Red & IC--Fail & Rewrite--Steps: total & CPs & R--rhs & R--lhs & E & Goals & ");
#else
      ;
#endif
   }
#if NFB_STATISTIK
   else if (IS_IsIn("slatex-head", IS)){
      fprintf(IO_Strom, "Match Queries &Reductions & ");
      ;
   }
#endif
   else if (IS_IsIn("intern", IS)){
      IO_DruckeStatistikZeile("max. expansion of rewriting path",NFB_AusdehnungPfad());
   }
   else if (IS_IsIn("config", IS)){
#if NFB_STATISTIK
      fprintf(IO_Strom, "NF: statistics  ON, ");
#else
      fprintf(IO_Strom, "NF: statistics OFF,");
#endif
#if NF_TEST
      fprintf(IO_Strom, "Irred-test  ON, ");
#else
      fprintf(IO_Strom, "Irred-test OFF, ");
#endif
   }
}

void NF_AnzReduktionenGesamt(unsigned long int *AnzRedGes)
{
  int Zaehler;
  *AnzRedGes = 0;
#if NFB_STATISTIK
   for (Zaehler = 0; Zaehler<AnzObjektTypen; Zaehler++)
      *AnzRedGes += AnzRedErfRegeln[Zaehler] + AnzRedErfGleichgn[Zaehler];
#endif
}
