/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - NFOutermostRE.c                                                                                            */
/* - ***************                                                                                            */
/* -                                                                                                            */
/* - Normalform-Bildung mit gleichberechtigter Beruecksichtigung von Regeln und Gleichungen.                    */
/* -                                                                                                            */
/* - 06.12.95 Arnim. Angelegt durch Anpassung aus NFOutermost.c (Version mit einfacher IC-Verwaltung)           */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/



#include "general.h"
#include "TermOperationen.h"
#include "FehlerBehandlung.h"
#include "RUndEVerwaltung.h"
#include "ZeitMessung.h"
#include "Ausgaben.h"
#include "NFBasis.h"
#include "pcl.h"
#include "NFOutermostRE.h"

typedef struct ICStructRE{
   BOOLEAN IC;
} ICZellenRET;

static ICZellenRET       *IC = NULL;

static void NFORE_Ausdehnungsproz(StellenT neueGroesse)
{
   IC = our_realloc(IC, neueGroesse*sizeof(ICZellenRET), "Ausdehnung des IC1-Pfades");
}

static BOOLEAN NFORE_SchonInitialisiert = FALSE;

void NFORE_Init(void)
{
   NFB_Init();
   NFB_AusdehnungsprozedurAnmelden(NFORE_Ausdehnungsproz);
   if (!NFORE_SchonInitialisiert)
     IC = our_alloc(NF_PfadArrayGroesse*sizeof(ICZellenRET));
   NFORE_SchonInitialisiert = TRUE;
}

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Lokale Makros ------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static StellenT        letzteRed,         /* Verweis auf die letzte reduzierte Stelle   */
                       hoechsteRed;       /* Verweis auf die hoechste reduzierte Stelle */

static unsigned long int MaxTiefeRegeln, MaxTiefeGleichungen, MaxTiefeRundE; 
/* wegen Tiefe("f(a)") == 2 wird hier in PfadNeu MaxTiefe??? auf */
/* RE_MaxTiefe??? ()-1 gesetzt. Sonst Aufstieg eins zu weit!       */

#define PfadInit(/*TermT*/ Term, /*ObjektT*/ objekt)\
{\
    NFB_PfadInit(Term, objekt);\
    letzteRed =0;\
    hoechsteRed = 0;\
    MaxTiefeRegeln = RE_maxTiefeRegel()-1;\
    MaxTiefeGleichungen = RE_maxTiefeGleichung()-1;\
    MaxTiefeRundE = (MaxTiefeRegeln>MaxTiefeGleichungen)?MaxTiefeRegeln:MaxTiefeGleichungen;\
    IC[0].IC=FALSE;\
}

#define hoechsteRedAnpassen(Kandidat) (Kandidat<hoechsteRed?(hoechsteRed = Kandidat):0)

#define GipfelZuR(/*StellenT*/ stelle)            ((stelle > MaxTiefeRegeln)? (stelle - MaxTiefeRegeln):0)
#define GipfelZuE(/*StellenT*/ stelle)            ((stelle > MaxTiefeGleichungen)? (stelle - MaxTiefeGleichungen):0)
#define GipfelZuRE(/*StellenT*/ stelle)            ((stelle > MaxTiefeRundE)? (stelle - MaxTiefeRundE):0)


/* Unter Gipfel versteht man hier diejenige Stelle im Pfad, bis zu der sich eine Reduktion an der aktuellen Stelle maximal
   auswirken kann (bzgl. Regeln). Hierzu wird die Tiefe der tiefsten Regel nach oben gestiegen, da oberhalb dieser Stelle
   keine Regel mehr anwendbar geworden sein kann, die vorher nicht anwendbar war.*/

   
/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- SetzePfadAufNaechstenInVorOrdnung ----------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/*                                                                                                                         */
/* Nachdem die aktuelle Stelle im Pfad nicht mehr reduzierbar ist, wird der naechste Teilterm in Vorordnung gesucht und    */
/* zum aktuellen Term gemacht. Dabei wird auch das noetige Backtracking durchgefuehrt                                      */
/*                                                                                                                         */
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN WeiterInVO(void)
{
   while (NichtWurzel() && EsGibtKeineZuBehNachf()) /* Aufstieg bis Wurzel oder Knoten mit noch nicht abgearbeiteten Teiltermen */
      PfadEinsAuf();
   if (EsGibtKeineZuBehNachf())                     /* Keinen Knoten mit noch nicht abgearbeiteten Teiltermen gefunden */
      return FALSE;                                     /* => Pfad ist am Ende, dh. Term ist fertig durchlaufen.*/
   else{
      NFB_PfadEinsWeiterSetzen();                           /* sonst: Weiter im Durchlauf */
      IC[aktuelleStelle].IC = FALSE;
      return TRUE;
   }
}



/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Backtracking zu Gleichungen ----------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren sorgen dafuer, dass nach einer Reduktion die der reduzierten Stelle uebergeordneten Stellen    */
/* auf Reduzierbarkeit geprueft werden. Dazu werden alle Stellen bis zu der Ebene, bei der Matches wegen direkter Clashes  */
/* gescheitert sind, probiert sowie alle darueberliegenden Stellen, an denen indirekte Clashes Reduktionen verhinderten.   */
/*-------------------------------------------------------------------------------------------------------------------------*/

/*---------------------------------- Aufstieg -----------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN GAlleAufsteigenBisReduziertOderStop(StellenT start, StellenT stop)
/*
- alle Stellen (start, stop] probieren (Ist start = stop, wird stop nicht probiert).
- FALSE: aktuelleStelle = stop
- TRUE : aktuelleStelle = reduzierte Stelle = letzteRed
*/
{
   aktuelleStelle = start;

   while(aktuelleStelle>stop){
      PfadEinsAuf();
      if (NFB_GleichungAngewendet())
         return TRUE;
      else                                                                      /* ICUpdateBeiAufstiegZuGipfel(); */
         IC[aktuelleStelle].IC = NFB_IC();
   }
   return FALSE;
}

static BOOLEAN GICsAufsteigenBisReduziertOderStop(StellenT start)
/*
- alle ICs excl. start aufsteigen und probieren
- FALSE: aktuelleStelle = 0
- TRUE : aktuelleStelle = letzter reduzierter IC = letzteRed
*/
{
   aktuelleStelle = start;
   while(aktuelleStelle>0){
      PfadEinsAuf();
      if (IC[aktuelleStelle].IC && NFB_GleichungAngewendet())
         return TRUE;
   }
   return FALSE;
}
            
/*---------------------------------- Abstieg ------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN GICsAbsteigenBisReduziertOderStop(StellenT stop)
/*
- alle ICs von Wurzel ausgehend absteigend probieren, excl. stop
- FALSE: aktuelleStelle = letzterIC oberhalb von stop bzw. MaxStelle, falls keine ICs ex.
- TRUE : aktuelleStelle = reduzierter IC = letzteRed
*/
{
   aktuelleStelle = 0;

   while (aktuelleStelle < stop){
      if (IC[aktuelleStelle].IC && NFB_GleichungAngewendet())
         return TRUE;
      PfadEinsAb();
   }
   return FALSE;
}

static BOOLEAN GAlleAbsteigenBisReduziertOderStop(StellenT start, StellenT stop)
/*
- alle Stellen [start, stop) probieren (Ist start = stop, wird start nicht probiert)
- FALSE: aktuelleStelle = '1 ueber stop' oder start, falls start=stop
- TRUE : aktuelleStelle = letzteRed
*/
{
   aktuelleStelle = start;
   while (aktuelleStelle < stop){
      if (NFB_GleichungAngewendet())
         return TRUE;
      else
         IC[aktuelleStelle].IC = NFB_IC();
      PfadEinsAb();
   }
   return FALSE;
}



/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Gleichungen ------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den uebergebenen Term bzgl. des aktuellen Glgs.systems auf Normalform. Dabei wird aber */
/* jeweils eine andere Variante der Leftmost-Outermost-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste   */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/* ++++ Kommentare sind analog zu denjenigen in den entsprechenden Prozeduren NormalformZuRegeln***  ++++                  */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFORE_NormalformZuGleichungenAufDirekt(TermT Term, ObjektT objekt)
/*  nach Reduktion direkt Aufstieg. Unkommentierte Stellen siehe NFORE_NormalformZuRegelnAufNochmal. */
{
   BOOLEAN weiter;

   ZM_StartNFBildung();
   PfadInit(Term, objekt);
   while (NFB_GleichungAngewendet());
   IC[aktuelleStelle].IC = NFB_IC();
   weiter =  WeiterInVO();

   while(weiter)
      if (NFB_GleichungAngewendet()){
	 letzteRed = aktuelleStelle;
         while(GAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuE(letzteRed)))
            /* TRUE : aktuelleStelle = reduzierte Stelle */
            /* FALSE: aktuelleStelle = GipfelZuE(letzteRed) */
            letzteRed = aktuelleStelle;
         while(GICsAufsteigenBisReduziertOderStop(GipfelZuE(letzteRed))){
            letzteRed = aktuelleStelle;
            /* TRUE : aktuelleStelle = letzter reduzierter IC */
            /* FALSE: aktuelleStelle = hoechsterIC */
            while(GAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuE(letzteRed)))
               /* TRUE : aktuelleStelle = reduzierte Stelle */
               /* FALSE: aktuelleStelle = GipfelZuE(letzteRed) */
               letzteRed = aktuelleStelle;
         }
         aktuelleStelle = letzteRed;				    /* ... und am Ende an der zuletzt red. Stelle fortfahren */
      }
      else{
         IC[aktuelleStelle].IC = NFB_IC();
	 weiter =  WeiterInVO();
      }
   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
}


BOOLEAN NFORE_NormalformZuGleichungenAufNochmal(TermT Term, ObjektT objekt)
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */
{
   ZM_StartNFBildung();
   PfadInit(Term, objekt);

   while (NFB_GleichungAngewendet());			    /* Es gibt keine leeren Terme, also mind. ein Red.versuch */
   IC[aktuelleStelle].IC = NFB_IC();

   while(WeiterInVO()){
      if (NFB_GleichungAngewendet()){			    /* Reduktionserfolg => Stelle merken */
	 while(NFB_GleichungAngewendet());			    /*           Stelle fertigreduzieren */
         IC[aktuelleStelle].IC = NFB_IC();
	 letzteRed = aktuelleStelle;	                            /* (Abfrage, ob Wurzel, wuerde zwar an der Wurzel OPs einsparen, */
							    /*  dies passiert aber zu selten (s.o.))*/
         while (GAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuE(letzteRed))){
            while(NFB_GleichungAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
         }
         while (GICsAufsteigenBisReduziertOderStop(GipfelZuE(letzteRed))){
            while(NFB_GleichungAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
            while (GAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuE(letzteRed))){
               while(NFB_GleichungAngewendet());
               IC[aktuelleStelle].IC = NFB_IC();
               letzteRed = aktuelleStelle;
            }
         }
	 aktuelleStelle = letzteRed;				    /* nach Aufstieg-Ende an der zuletzt red. Stelle mit dem Durchlauf */
      }	                                                    /* fortfahren, da diese Stelle schon erledigt ist  */
      else
         IC[aktuelleStelle].IC = NFB_IC();
   }

   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
}

BOOLEAN NFORE_NormalformZuGleichungenAbNochmal(TermT Term, ObjektT objekt)
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel.*/

{
   ZM_StartNFBildung();
   PfadInit(Term, objekt);
         
   do{
      if (NFB_GleichungAngewendet()){ 
	 while(NFB_GleichungAngewendet());			    /* Stelle fertigreduzieren */
         IC[aktuelleStelle].IC = NFB_IC();
	 letzteRed = aktuelleStelle;				    /* sonst: red. Stelle merken und */
         while(GICsAbsteigenBisReduziertOderStop(GipfelZuE(aktuelleStelle))){
            while(NFB_GleichungAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
         }
         while (GAlleAbsteigenBisReduziertOderStop(GipfelZuE(letzteRed), letzteRed)){
            while(NFB_GleichungAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
            while(GICsAbsteigenBisReduziertOderStop(GipfelZuE(letzteRed))){
               while(NFB_GleichungAngewendet());
               IC[aktuelleStelle].IC = NFB_IC();
               letzteRed = aktuelleStelle;
            }
         }
         aktuelleStelle = letzteRed;
      }
      else
         IC[aktuelleStelle].IC = NFB_IC();
   }while(WeiterInVO());
   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
   
}

BOOLEAN NFORE_NormalformZuGleichungenAbDirekt(TermT Term, ObjektT objekt)
/* nach Reduktion direkt Abstieg von Wurzel */
{
   BOOLEAN nochmal;

   ZM_StartNFBildung();
   PfadInit(Term, objekt);

   do{
      if (NFB_GleichungAngewendet()){
	 letzteRed = aktuelleStelle;
         nochmal = TRUE;
         while(GICsAbsteigenBisReduziertOderStop(GipfelZuE(letzteRed)))
            letzteRed = aktuelleStelle;
         while (GAlleAbsteigenBisReduziertOderStop(GipfelZuE(letzteRed), letzteRed)){
            letzteRed = aktuelleStelle;
            while(GICsAbsteigenBisReduziertOderStop(GipfelZuE(letzteRed)))
               letzteRed = aktuelleStelle;
         }
         aktuelleStelle =letzteRed;
      }
      else{
         IC[aktuelleStelle].IC = NFB_IC();
         nochmal = WeiterInVO();
      }
   }while(nochmal);

   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
}



/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Backtracking zu Regeln ---------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren sorgen dafuer, dass nach einer Reduktion die der reduzierten Stelle uebergeordneten Stellen    */
/* auf Reduzierbarkeit geprueft werden. Dazu werden alle Stellen bis zu der Ebene, bei der Matches wegen direkter Clashes  */
/* gescheitert sind, probiert sowie alle darueberliegenden Stellen, an denen indirekte Clashes Reduktionen verhinderten.   */
/*-------------------------------------------------------------------------------------------------------------------------*/

/*---------------------------------- Aufstieg -----------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN RAlleAufsteigenBisReduziertOderStop(StellenT start, StellenT stop)
/*
- alle Stellen (start, stop] probieren (Ist start = stop, wird stop nicht probiert).
- FALSE: aktuelleStelle = stop
- TRUE : aktuelleStelle = reduzierte Stelle = letzteRed
*/
{
   aktuelleStelle = start;

   while(aktuelleStelle>stop){
      PfadEinsAuf();
      if (NFB_RegelAngewendet())
         return TRUE;
      else
         IC[aktuelleStelle].IC = NFB_IC();
   }
   return FALSE;
}

static BOOLEAN RICsAufsteigenBisReduziertOderStop(StellenT start)
/*
- alle ICs excl. start aufsteigen und probieren
- FALSE: aktuelleStelle = hoechsterIC
- TRUE : aktuelleStelle = letzter reduzierter IC = letzteRed
*/
{
   aktuelleStelle = start;
   while(aktuelleStelle > 0){
      PfadEinsAuf();
      if (IC[aktuelleStelle].IC && NFB_RegelAngewendet())
         return TRUE;
   }
   return FALSE;
}
            
/*---------------------------------- Abstieg ------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN RICsAbsteigenBisReduziertOderStop(StellenT stop)
/*
- alle ICs von Wurzel ausgehend absteigend probieren, excl. stop
- FALSE: aktuelleStelle = letzterIC oberhalb von stop bzw. MaxStelle, falls keine ICs ex.
- TRUE : aktuelleStelle = reduzierter IC = letzteRed
- ICs bleiben ICs, auch nach Aenedrungen weiter unten im Term
*/
{
   aktuelleStelle = 0;

   while (aktuelleStelle < stop){
      if (IC[aktuelleStelle].IC && NFB_RegelAngewendet())
         return TRUE;
      PfadEinsAb();
   }
   return FALSE;
}

static BOOLEAN RAlleAbsteigenBisReduziertOderStop(StellenT start, StellenT stop)
/*
- alle Stellen [start, stop) probieren (Ist start = stop, wird start nicht probiert)
- FALSE: aktuelleStelle = '1 ueber stop'
- TRUE : aktuelleStelle = letzteRed
*/
{
   aktuelleStelle = start;
   while (aktuelleStelle < stop){
      if (NFB_RegelAngewendet())
         return TRUE;
      else
         IC[aktuelleStelle].IC = NFB_IC();
      PfadEinsAb();
   }
   return FALSE;
}

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln -----------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren bringen den uebergebenen Term bzgl. des aktuellen Regelsystems auf Normalform. Dabei wird aber */
/* jeweils eine andere Variante der Leftmost-Outermost-Strategie verfolgt. Die Prozeduren arbeiten destruktiv, die erste   */
/* Termzelle bleibt aber jeweils erhalten, so dass keine Probleme mit verschiedenen Verweisen auf diesen Term entstehen.   */
/* Die Varianten unterscheiden sich nur im Verhalten nach einem Reduktionserfolg (s. Kommentare unten).                    */
/* Der Rueckgabewert gibt an, ob ueberhaupt eine Reduktion stattgefunden hat.                                              */
/* Allen Prozeduren gemeinsam ist, da\3 der Term zunaechst auf Toplevel reduziert wird, bis dort keine Reduktionen mehr    */
/* moeglich sind. Anschlie\3end wird der Durchlauf in Vorordnung durchgefuehrt. Die vorgezogene NF-Bildung auf Toplevel hat*/
/* den Vorteil, da\3 hierbei keinerlei Backtracking-Overhead erzeugt werden mu\3, wie das innerhalb der 'normalen' Reduk-  */
/* tionsschleife der Fall ist. Zur Begruendung dient auch folgende Tabelle (Resultate aus ca. 20 Laeufen je Variante+Bsp.)

                 vorher while             innen auch                weder noch
                                                                                                
-------------------------------------------------------------------------------------------------------------------------  
                                                                                               
ackermannViel
                 121.68   0.11913        123.268  0.0995833          121.509   0.105833                                     
                 121.595  0.106842       122.725  0.103              121.314   0.097                                                   
                 118.940 0.080           120.230 0.070               118.580 0.060                                                     
                                                                                                
                                                                                                
                                                                                                
braid4
                20.2867  0.40375         20.4683   0.374167          20.2325   0.39375                                                     
                20.2535  0.383           20.4265   0.371             20.209    0.3795                                               
                19.760 0.530             19.940 0.420                19.830 0.420                                                     
                                                                                                
                                                                                                
                                                                                                
z22             46.6633   0.310417       47.2667  0.224167          46.96     0.242083                                                        
                46.46     0.3105         46.976   0.2255            46.6255   0.242
                46.070 0.250             46.540 0.180               46.120 0.260                                                    
                                                                                                
                                                                                                
-------------------------------------------------------------------------------------------------------------------------  
vorherwhile: Zusaetzliche while-Schleife zur Reduktion der Wurzel vor der eigentlichen Reduktionsschleife
innen auch : Auch in der eigentlichen Reduktionsschleife wird nach jeder Reduktion geprueft, ob man sich an der Wurzel
             befindet, und falls ja, diese zuerst in einer weiteren Schleife kleinreduziert.
weder noch : 'normales' Verfahren, dh. ohne zusaetzliche while-Schleifen.
-------------------------------------------------------------------------------------------------------------------------  

1.Zeile  : jeweils Durchschnitte komplett
2. Zeile : ohne die 2 groessten und die zwei kleinsten Werte
3. Zeile : bester erreichter Wert

ERGEBNIS : 

- Durchschnittswerte : 2xwedernoch, 1xvorherwhile(z22)
- DurchschnittswerteII: 2xwedernoch, 1xvorherwhile(z22)
- Spitzenwerte : 2xvorherwhile, 1xwedernoch(braid4)


Weitere Ueberlegung:
--------------------
Innerhalb der Reduktionsschleifen lassen sich mit der Abfrage, ob man sich an der Wurzel befindet, im Erfolgsfall Ops einsparen.
Dies saehe folgendermassen aus:

if(NichtWurzel()){                        <--- hier wird mit Abfrage abgebrochen
   letzteRed = aktuelleStelle;
   aktStelleSei(neuerGipfel());
   while(aktIstNichtLetzteRed()){         <--- hier ohne
      if(NFB_RegelAngewendet()){
	  letzteRed = aktuelleStelle;
	  while(NFB_RegelAngewendet());
	  aktStelleSei(neuerGipfel());
      }
      else{
	  PfadEinsAb();
      }
   }
}

Daraus ergibt sich folgendes:

- mit Abfrage:

Stelle = Wurzel: 1 Op (Abfrage)                           = 1 Op
Stelle # Wurzel: 1 Op (Abfrage) + 2 Zuweisungen + Abfrage = 4 Ops

- ohne Abfrage:

immer 2 Zuweisungen + 1 Abfrage = 3 Ops

Soll die Abfrage Vorteile bringen, ergibt sich (sei w = P(Stelle = Wurzel)):

    w * (1 Abfrage) + (1-w)*(2 Abfrage + 2 Zuweisungen) < 2 Zuweisungen + 1 Abfrage
<=> w*A + 2A + 2Z - w*2A - w*2Z                         < 2Z + 1A
<=> A - w*A - 2w*Z                                      < 0
<=> A                                                   < w*(A + 2Z)

Unter der Annahme, dass eine Zuweisung doppelten Aufwand zu einer Abfrage erfordert, ergibt sich:
w > 20 %.

==> Es muesste in mehr als 20 % der Faelle die behandelte Stelle tatsaechlich die Wurzel sein, was m.E. utopisch ist.
---------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFORE_NormalformZuRegelnAufDirekt(TermT Term, ObjektT objekt)
/*  nach Reduktion direkt Aufstieg. Unkommentierte Stellen siehe NFORE_NormalformZuRegelnAufNochmal. */
{
   BOOLEAN weiter;

   ZM_StartNFBildung();
   PfadInit(Term, objekt);
   while (NFB_RegelAngewendet());
   IC[aktuelleStelle].IC = NFB_IC();
   weiter =  WeiterInVO();

   while(weiter)
      if (NFB_RegelAngewendet()){
	 letzteRed = aktuelleStelle;
         while(RAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuR(letzteRed)))
            /* TRUE : aktuelleStelle = reduzierte Stelle */
            /* FALSE: aktuelleStelle = GipfelZuR(letzteRed) */
            letzteRed = aktuelleStelle;
         while(RICsAufsteigenBisReduziertOderStop(GipfelZuR(letzteRed))){
            /* TRUE : aktuelleStelle = letzter reduzierter IC */
            /* FALSE: aktuelleStelle = hoechsterIC */
            letzteRed = aktuelleStelle;
            while(RAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuR(letzteRed)))
               /* TRUE : aktuelleStelle = reduzierte Stelle */
               /* FALSE: aktuelleStelle = GipfelZuR(letzteRed) */
               letzteRed = aktuelleStelle;
         }
         aktuelleStelle = letzteRed;				    /* ... und am Ende an der zuletzt red. Stelle fortfahren */
      }
      else{
         IC[aktuelleStelle].IC = NFB_IC();
	 weiter =  WeiterInVO();
      }
   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
}


BOOLEAN NFORE_NormalformZuRegelnAufNochmal(TermT Term, ObjektT objekt)
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */
{
   ZM_StartNFBildung();
   PfadInit(Term, objekt);

   while (NFB_RegelAngewendet());			    /* Es gibt keine leeren Terme, also mind. ein Red.versuch */
   IC[aktuelleStelle].IC = NFB_IC();

   while(WeiterInVO()){
      if (NFB_RegelAngewendet()){			    /* Reduktionserfolg => Stelle merken */
	 while(NFB_RegelAngewendet());			    /*           Stelle fertigreduzieren */
         IC[aktuelleStelle].IC = NFB_IC();
	 letzteRed = aktuelleStelle;	                            /* (Abfrage, ob Wurzel, wuerde zwar an der Wurzel OPs einsparen, */
							    /*  dies passiert aber zu selten (s.o.))*/
         while (RAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuR(letzteRed))){
            while(NFB_RegelAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
         }
         while (RICsAufsteigenBisReduziertOderStop(GipfelZuR(letzteRed))){
            while(NFB_RegelAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
            while (RAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuR(letzteRed))){
               while(NFB_RegelAngewendet());
               IC[aktuelleStelle].IC = NFB_IC();
               letzteRed = aktuelleStelle;
            }
         }
	 aktuelleStelle = letzteRed;				    /* nach Aufstieg-Ende an der zuletzt red. Stelle mit dem Durchlauf */
      }	                                                    /* fortfahren, da diese Stelle schon erledigt ist  */
      else
         IC[aktuelleStelle].IC = NFB_IC();
   }

   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
}

BOOLEAN NFORE_NormalformZuRegelnAbNochmal(TermT Term, ObjektT objekt)
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel.
   Das Flag 'Aufstieg' bedeutet in dieser Prozedur eigentlich Abstieg!*/
{
   ZM_StartNFBildung();
   PfadInit(Term, objekt);
         
   do{
      if (NFB_RegelAngewendet()){ 
	 while(NFB_RegelAngewendet());			    /* Stelle fertigreduzieren */
         IC[aktuelleStelle].IC = NFB_IC();
	 letzteRed = aktuelleStelle;				    /* sonst: red. Stelle merken und */
         while(RICsAbsteigenBisReduziertOderStop(GipfelZuR(aktuelleStelle))){
            while(NFB_RegelAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
         }
         while (RAlleAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed), letzteRed)){
            while(NFB_RegelAngewendet());
            IC[aktuelleStelle].IC = NFB_IC();
            letzteRed = aktuelleStelle;
            while(RICsAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed))){
               while(NFB_RegelAngewendet());
               IC[aktuelleStelle].IC = NFB_IC();
               letzteRed = aktuelleStelle;
            }
         }
         aktuelleStelle = letzteRed;
      }
      else
         IC[aktuelleStelle].IC = NFB_IC();
   }while(WeiterInVO());
   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
   
}

BOOLEAN NFORE_NormalformZuRegelnAbDirekt(TermT Term, ObjektT objekt)
/* nach Reduktion direkt Abstieg von Wurzel */
{
   BOOLEAN nochmal;

   ZM_StartNFBildung();
   PfadInit(Term, objekt);

   do{
      if (NFB_RegelAngewendet()){
	 letzteRed = aktuelleStelle;
         nochmal = TRUE;
         while(RICsAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed)))
            letzteRed = aktuelleStelle;
         while (RAlleAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed), letzteRed)){
            letzteRed = aktuelleStelle;
            while(RICsAbsteigenBisReduziertOderStop(GipfelZuR(letzteRed)))
               letzteRed = aktuelleStelle;
         }
         aktuelleStelle =letzteRed;
      }
      else{
         IC[aktuelleStelle].IC = NFB_IC();
         nochmal = WeiterInVO();
      }
   }while(nochmal);

   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
}

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Backtracking zu Regeln und Gleichungen -----------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/* Die folgenden Prozeduren sorgen dafuer, dass nach einer Reduktion die der reduzierten Stelle uebergeordneten Stellen    */
/* auf Reduzierbarkeit geprueft werden. Dazu werden alle Stellen bis zu der Ebene, bei der Matches wegen direkter Clashes  */
/* gescheitert sind, probiert sowie alle darueberliegenden Stellen, an denen indirekte Clashes Reduktionen verhinderten.   */
/*                                                                                                                         */
/* Die Normlformbildung zu Regeln UND Gleichungen wird hier in zwei verschiedenen Varianten realisiert. Zwar wird an jeder */
/* Stelle die des Terms die Reduktion mit Regeln und Gleichungen versucht, aber die Reihenfolge der Anwendung unterscheidet*/
/* sich: in den mit 'RE' gekennzeichneten Varianten wird zunaechst der Reduktionsversuch mit der Regelmenge unternommen,   */
/* in den mit 'ER' gekennzeichneten wird zunaechst die Gleichungsmenge auf Anwendbarkeit hin ueberprueft.                  */
/* Aus diesem Grunde existieren alle folgenden Prozeduren in zwei Varianten, die sich im nur in diesem Detail unter-       */
/* scheiden.                                                                                                               */
/*                                                                                                                         */
/*-------------------------------------------------------------------------------------------------------------------------*/

/*---------------------------------- Reduktionsversuch --------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

#define REAnwendung(/*BOOLEAN*/ ICFlag) \
                      (NFB_RegelAngewendet()?TRUE:((ICFlag = NFB_IC(), NFB_GleichungAngewendet())?TRUE:(ICFlag|=NFB_IC(), FALSE)))

#define ERAnwendung(/*BOOLEAN*/ ICFlag) \
                      (NFB_GleichungAngewendet()?TRUE:((ICFlag = NFB_IC(), NFB_RegelAngewendet())?TRUE:(ICFlag|=NFB_IC(), FALSE)))

/*---------------------------------- Aufstieg -----------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN REAlleAufsteigenBisReduziertOderStop(StellenT start, StellenT stop)
/*
- alle Stellen (start, stop] probieren (Ist start = stop, wird stop nicht probiert).
- FALSE: aktuelleStelle = stop
- TRUE : aktuelleStelle = reduzierte Stelle = letzteRed
*/
{
   aktuelleStelle = start;

   while(aktuelleStelle>stop){
      PfadEinsAuf();
      if (REAnwendung(IC[aktuelleStelle].IC))
         return TRUE;
   }
   return FALSE;
}

static BOOLEAN REICsAufsteigenBisReduziertOderStop(StellenT start)
/*
- alle ICs excl. start aufsteigen und probieren
- FALSE: aktuelleStelle = hoechsterIC
- TRUE : aktuelleStelle = letzter reduzierter IC = letzteRed
*/
{
   aktuelleStelle = start;
   while(aktuelleStelle > 0){
      PfadEinsAuf();
      if (IC[aktuelleStelle].IC && (NFB_RegelAngewendet() || NFB_GleichungAngewendet()))
         return TRUE;
   }
   return FALSE;
}

static BOOLEAN ERAlleAufsteigenBisReduziertOderStop(StellenT start, StellenT stop)
{
   aktuelleStelle = start;

   while(aktuelleStelle>stop){
      PfadEinsAuf();
      if (ERAnwendung(IC[aktuelleStelle].IC))
         return TRUE;
   }
   return FALSE;
}

static BOOLEAN ERICsAufsteigenBisReduziertOderStop(StellenT start)
{
   aktuelleStelle = start;
   while(aktuelleStelle > 0){
      PfadEinsAuf();
      if (IC[aktuelleStelle].IC && (NFB_GleichungAngewendet() || NFB_RegelAngewendet()))
         return TRUE;
   }
   return FALSE;
}
            
/*---------------------------------- Abstieg ------------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/

static BOOLEAN REICsAbsteigenBisReduziertOderStop(StellenT stop)
/*
- alle ICs von Wurzel ausgehend absteigend probieren, excl. stop
- FALSE: aktuelleStelle = letzterIC oberhalb von stop bzw. MaxStelle, falls keine ICs ex.
- TRUE : aktuelleStelle = reduzierter IC = letzteRed
- ICs bleiben ICs, auch nach Aenedrungen weiter unten im Term
*/
{
   aktuelleStelle = 0;

   while (aktuelleStelle < stop){
      if (IC[aktuelleStelle].IC && (NFB_RegelAngewendet() || NFB_GleichungAngewendet()))
         return TRUE;
      PfadEinsAb();
   }
   return FALSE;
}

static BOOLEAN REAlleAbsteigenBisReduziertOderStop(StellenT start, StellenT stop)
/*
- alle Stellen [start, stop) probieren (Ist start = stop, wird start nicht probiert)
- FALSE: aktuelleStelle = '1 ueber stop'
- TRUE : aktuelleStelle = letzteRed
*/
{
   aktuelleStelle = start;
   while (aktuelleStelle < stop){
      if (REAnwendung(IC[aktuelleStelle].IC))
         return TRUE;
      PfadEinsAb();
   }
   return FALSE;
}

static BOOLEAN ERICsAbsteigenBisReduziertOderStop(StellenT stop)
{
   aktuelleStelle = 0;

   while (aktuelleStelle < stop){
      if (IC[aktuelleStelle].IC && (NFB_GleichungAngewendet() || NFB_RegelAngewendet()))
         return TRUE;
      PfadEinsAb();
   }
   return FALSE;
}

static BOOLEAN ERAlleAbsteigenBisReduziertOderStop(StellenT start, StellenT stop)
{
   aktuelleStelle = start;
   while (aktuelleStelle < stop){
      if (ERAnwendung(IC[aktuelleStelle].IC))
         return TRUE;
      PfadEinsAb();
   }
   return FALSE;
}

/*-------------------------------------------------------------------------------------------------------------------------*/
/*---------------------------------- Normalform zu Regeln und Gleichungen -------------------------------------------------*/
/*-------------------------------------------------------------------------------------------------------------------------*/
/*                                                                                                                         */
/* Bringt den uebergebenen Term bzgl. des aktuellen Regel- und Gleichungssystems auf Normalform. Dabei werden Regeln und   */
/* Gleichungen beinahe gleichberechtigt eingesetzt: An jeder zu probierenden Stelle wird zunaechst versucht, eine Regel    */
/* anzuwenden, und bei Misserfolg sofort dasselbe mit der Gleichungsmenge versucht. Das Backtracking erfolgt daher eben-   */
/* falls nicht mehr getrennt nach Regeln und Gleichungen, sondern genauso wie eben beschrieben.                            */
/*                                                                                                                         */
/*-------------------------------------------------------------------------------------------------------------------------*/

BOOLEAN NFORE_NormalformAufDirekt(TermT Term, ObjektT objekt)
/*  nach Reduktion direkt Aufstieg. Unkommentierte Stellen siehe NFORE_NormalformZuRegelnAufNochmal. */
{
   BOOLEAN weiter;

   ZM_StartNFBildung();
   PfadInit(Term, objekt);

   while (REAnwendung(IC[aktuelleStelle].IC));

   weiter =  WeiterInVO();

   while(weiter)
      if (REAnwendung(IC[aktuelleStelle].IC)){
	 letzteRed = aktuelleStelle;
         while(REAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuRE(letzteRed)))
            /* TRUE : aktuelleStelle = reduzierte Stelle */
            /* FALSE: aktuelleStelle = GipfelZuRE(letzteRed) */
            letzteRed = aktuelleStelle;
         while(REICsAufsteigenBisReduziertOderStop(GipfelZuRE(letzteRed))){
            /* TRUE : aktuelleStelle = letzter reduzierter IC */
            /* FALSE: aktuelleStelle = hoechsterIC */
            letzteRed = aktuelleStelle;
            while(REAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuRE(letzteRed)))
               /* TRUE : aktuelleStelle = reduzierte Stelle */
               /* FALSE: aktuelleStelle = GipfelZuRE(letzteRed) */
               letzteRed = aktuelleStelle;
         }
         aktuelleStelle = letzteRed;				    /* ... und am Ende an der zuletzt red. Stelle fortfahren */
      }
      else
	 weiter =  WeiterInVO();
   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
}


BOOLEAN NFORE_NormalformAufNochmal(TermT Term, ObjektT objekt)
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */
{
   ZM_StartNFBildung();
   PfadInit(Term, objekt);

   while (REAnwendung(IC[aktuelleStelle].IC));			    /* Es gibt keine leeren Terme, also mind. ein Red.versuch */

   while(WeiterInVO()){
      if (REAnwendung(IC[aktuelleStelle].IC)){		    /* Reduktionserfolg => Stelle merken */
	 while(REAnwendung(IC[aktuelleStelle].IC));			    /*           Stelle fertigreduzieren */
	 letzteRed = aktuelleStelle;	                            /* (Abfrage, ob Wurzel, wuerde zwar an der Wurzel OPs einsparen, */
							    /*  dies passiert aber zu selten (s.o.))*/
         while (REAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuRE(letzteRed))){
            while(REAnwendung(IC[aktuelleStelle].IC));
            letzteRed = aktuelleStelle;
         }
         while (REICsAufsteigenBisReduziertOderStop(GipfelZuRE(letzteRed))){
            while(REAnwendung(IC[aktuelleStelle].IC));
            letzteRed = aktuelleStelle;
            while (REAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuRE(letzteRed))){
               while(REAnwendung(IC[aktuelleStelle].IC));
               letzteRed = aktuelleStelle;
            }
         }
	 aktuelleStelle = letzteRed;				    /* nach Aufstieg-Ende an der zuletzt red. Stelle mit dem Durchlauf */
      }	                                                    /* fortfahren, da diese Stelle schon erledigt ist  */
   }

   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
}

BOOLEAN NFORE_NormalformAbNochmal(TermT Term, ObjektT objekt)
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel.
   Das Flag 'Aufstieg' bedeutet in dieser Prozedur eigentlich Abstieg!*/
{
   ZM_StartNFBildung();
   PfadInit(Term, objekt);
         
   do{
      if (REAnwendung(IC[aktuelleStelle].IC)){ 
	 while(REAnwendung(IC[aktuelleStelle].IC));			    /* Stelle fertigreduzieren */
	 letzteRed = aktuelleStelle;				    /* sonst: red. Stelle merken und */
         while(REICsAbsteigenBisReduziertOderStop(GipfelZuRE(aktuelleStelle))){
            while(REAnwendung(IC[aktuelleStelle].IC));
            letzteRed = aktuelleStelle;
         }
         while (REAlleAbsteigenBisReduziertOderStop(GipfelZuRE(letzteRed), letzteRed)){
            while(REAnwendung(IC[aktuelleStelle].IC));
            letzteRed = aktuelleStelle;
            while(REICsAbsteigenBisReduziertOderStop(GipfelZuRE(letzteRed))){
               while(REAnwendung(IC[aktuelleStelle].IC));
               letzteRed = aktuelleStelle;
            }
         }
         aktuelleStelle = letzteRed;
      }
   }while(WeiterInVO());
   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
   
}

BOOLEAN NFORE_NormalformAbDirekt(TermT Term, ObjektT objekt)
/* nach Reduktion direkt Abstieg von Wurzel */
{
   BOOLEAN nochmal;

   ZM_StartNFBildung();
   PfadInit(Term, objekt);

   do{
      if (REAnwendung(IC[aktuelleStelle].IC)){
	 letzteRed = aktuelleStelle;
         nochmal = TRUE;
         while(REICsAbsteigenBisReduziertOderStop(GipfelZuRE(letzteRed)))
            letzteRed = aktuelleStelle;
         while (REAlleAbsteigenBisReduziertOderStop(GipfelZuRE(letzteRed), letzteRed)){
            letzteRed = aktuelleStelle;
            while(REICsAbsteigenBisReduziertOderStop(GipfelZuRE(letzteRed)))
               letzteRed = aktuelleStelle;
         }
         aktuelleStelle =letzteRed;
      }
      else
         nochmal = WeiterInVO();
   }while(nochmal);

   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
}

BOOLEAN NFOER_NormalformAufDirekt(TermT Term, ObjektT objekt)
/*  nach Reduktion direkt Aufstieg. Unkommentierte Stellen siehe NFOER_NormalformZuRegelnAufNochmal. */
{
   BOOLEAN weiter;

   ZM_StartNFBildung();
   PfadInit(Term, objekt);

   while (ERAnwendung(IC[aktuelleStelle].IC));

   weiter =  WeiterInVO();

   while(weiter)
      if (ERAnwendung(IC[aktuelleStelle].IC)){
	 letzteRed = aktuelleStelle;
         while(ERAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuRE(letzteRed)))
            /* TRUE : aktuelleStelle = reduzierte Stelle */
            /* FALSE: aktuelleStelle = GipfelZuRE(letzteRed) */
            letzteRed = aktuelleStelle;
         while(ERICsAufsteigenBisReduziertOderStop(GipfelZuRE(letzteRed))){
            /* TRUE : aktuelleStelle = letzter reduzierter IC */
            /* FALSE: aktuelleStelle = hoechsterIC */
            letzteRed = aktuelleStelle;
            while(ERAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuRE(letzteRed)))
               /* TRUE : aktuelleStelle = reduzierte Stelle */
               /* FALSE: aktuelleStelle = GipfelZuRE(letzteRed) */
               letzteRed = aktuelleStelle;
         }
         aktuelleStelle = letzteRed;				    /* ... und am Ende an der zuletzt red. Stelle fortfahren */
      }
      else
	 weiter =  WeiterInVO();
   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
}


BOOLEAN NFOER_NormalformAufNochmal(TermT Term, ObjektT objekt)
 /*  nach Reduktion aktuelle Stelle nochmal, Aufstieg */
{
   ZM_StartNFBildung();
   PfadInit(Term, objekt);

   while (ERAnwendung(IC[aktuelleStelle].IC));			    /* Es gibt keine leeren Terme, also mind. ein Red.versuch */

   while(WeiterInVO()){
      if (ERAnwendung(IC[aktuelleStelle].IC)){		    /* Reduktionserfolg => Stelle merken */
	 while(ERAnwendung(IC[aktuelleStelle].IC));			    /*           Stelle fertigreduzieren */
	 letzteRed = aktuelleStelle;	                            /* (Abfrage, ob Wurzel, wuerde zwar an der Wurzel OPs einsparen, */
							    /*  dies passiert aber zu selten (s.o.))*/
         while (ERAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuRE(letzteRed))){
            while(ERAnwendung(IC[aktuelleStelle].IC));
            letzteRed = aktuelleStelle;
         }
         while (ERICsAufsteigenBisReduziertOderStop(GipfelZuRE(letzteRed))){
            while(ERAnwendung(IC[aktuelleStelle].IC));
            letzteRed = aktuelleStelle;
            while (ERAlleAufsteigenBisReduziertOderStop(letzteRed, GipfelZuRE(letzteRed))){
               while(ERAnwendung(IC[aktuelleStelle].IC));
               letzteRed = aktuelleStelle;
            }
         }
	 aktuelleStelle = letzteRed;				    /* nach Aufstieg-Ende an der zuletzt red. Stelle mit dem Durchlauf */
      }	                                                    /* fortfahren, da diese Stelle schon erledigt ist  */
   }

   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
}

BOOLEAN NFOER_NormalformAbNochmal(TermT Term, ObjektT objekt)
/* nach Reduktion aktuelle Stelle nochmal, Abstieg von Wurzel.
   Das Flag 'Aufstieg' bedeutet in dieser Prozedur eigentlich Abstieg!*/
{
   ZM_StartNFBildung();
   PfadInit(Term, objekt);
         
   do{
      if (ERAnwendung(IC[aktuelleStelle].IC)){ 
	 while(ERAnwendung(IC[aktuelleStelle].IC));			    /* Stelle fertigreduzieren */
	 letzteRed = aktuelleStelle;				    /* sonst: red. Stelle merken und */
         while(ERICsAbsteigenBisReduziertOderStop(GipfelZuRE(aktuelleStelle))){
            while(ERAnwendung(IC[aktuelleStelle].IC));
            letzteRed = aktuelleStelle;
         }
         while (ERAlleAbsteigenBisReduziertOderStop(GipfelZuRE(letzteRed), letzteRed)){
            while(ERAnwendung(IC[aktuelleStelle].IC));
            letzteRed = aktuelleStelle;
            while(ERICsAbsteigenBisReduziertOderStop(GipfelZuRE(letzteRed))){
               while(ERAnwendung(IC[aktuelleStelle].IC));
               letzteRed = aktuelleStelle;
            }
         }
         aktuelleStelle = letzteRed;
      }
   }while(WeiterInVO());
   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
   
}

BOOLEAN NFOER_NormalformAbDirekt(TermT Term, ObjektT objekt)
/* nach Reduktion direkt Abstieg von Wurzel */
{
   BOOLEAN nochmal;

   ZM_StartNFBildung();
   PfadInit(Term, objekt);

   do{
      if (ERAnwendung(IC[aktuelleStelle].IC)){
	 letzteRed = aktuelleStelle;
         nochmal = TRUE;
         while(ERICsAbsteigenBisReduziertOderStop(GipfelZuRE(letzteRed)))
            letzteRed = aktuelleStelle;
         while (ERAlleAbsteigenBisReduziertOderStop(GipfelZuRE(letzteRed), letzteRed)){
            letzteRed = aktuelleStelle;
            while(ERICsAbsteigenBisReduziertOderStop(GipfelZuRE(letzteRed)))
               letzteRed = aktuelleStelle;
         }
         aktuelleStelle =letzteRed;
      }
      else
         nochmal = WeiterInVO();
   }while(nochmal);

   ZM_StoppNFBildung();
   return PfadWurdeReduziert();
}

