/**********************************************************************
* Filename : Ordnungen.c
*
* Kurzbeschreibung : Ordnungen
*
* Autoren : Andreas Jaeger
*
* 14.12.96: Erstellung
* 25.02.97: Vergleichs- und Groessertest-Funktionen aus Hauptkomponenten
*           hierher verlagert. Arnim.
*
*
* Benoetigt folgende Module:
*
* Benutzte Globale Variablen aus anderen Modulen:
*
**********************************************************************/

#include "Ausgaben.h"
#include "general.h"
#include "parser.h"
#include "FehlerBehandlung.h"
#include "InfoStrings.h"
#include "KBO.h"
#include "LPO.h"
#include "LPOVortests.h"
#include "Ordnungen.h"
#include "Ordnungsgenerator.h"
#include <string.h>
#include "Statistiken.h"
#include "SymbolOperationen.h"
#include "ZeitMessung.h"

/* -------------------------------------------------------------------------------*/
/* --------------- Term-Vergleich (ORD-1) ----------------------------------------*/
/* -------------------------------------------------------------------------------*/

typedef VergleichsT (*ORDVergleichsT)  (UTermT, UTermT);
typedef BOOLEAN     (*ORDTestT)  (UTermT, UTermT);

static ORDVergleichsT ORD_Vergleichsfkt;
static ORDTestT       ORD_UnifTestfkt, ORD_RedTestfkt;


/* ORD-1 */
/*-------*/

VergleichsT ORD_VergleicheTerme(UTermT term1, UTermT term2)
/* Vergleicht die beiden angegebenen Terme nach der eingestellten Ordnung. */
{
  VergleichsT result;

  ZM_StartOrdnungEntsch();
  result = ((*ORD_Vergleichsfkt) (term1, term2));
  ZM_StoppOrdnungEntsch();
  return result;
}

/* ORD-1a*/
/*-------*/
BOOLEAN ORD_TermGroesserRed(UTermT term1, UTermT term2)
/* Testet, ob term1 groesser als term2 ist (bzgl. der eingestellten Ordnung). */
{
  ZM_TERedGm2RedTest();
  return  ((*ORD_RedTestfkt) (term1, term2));
  /* Ende der Zeitmessung erfolgt in MatchOPs:GleichungGefunden */
}

BOOLEAN ORD_TermGroesserUnif(UTermT term1, UTermT term2)
/* Testet, ob term1 groesser als term2 ist (bzgl. der eingestellten Ordnung). */
{
  ZM_StartOrdnungUnifTest();
  if ((*ORD_UnifTestfkt) (term1, term2)) {
    ZM_StoppOrdnungUnifTest();
    return TRUE;
  }
  else {
    ZM_StoppOrdnungUnifTest();
    return FALSE;
  }
}

static void NewKbo (void) 
{
   SymbolT index;
   /* Alle Gewichte auf 1, Praezedenz nicht veraendert */
   SO_forFktSymbole(index){
      SO_FunktionsGewichte[index] = 1;
   }
}

#define Auswertung(AUTO, ORD, KBOACTION, LPOACTION,                                     \
  PREC, PRECTEST, PREC_NOK_ACTION, PERM_OK_ACTION, PERM_NOK_ACTION)                     \
{                                                                                       \
  char *newperm;                                                                        \
  AUTO = FALSE;                                                                         \
  fprintf(IO_Strom, "\n* ORDERING: ");                                                  \
                                                                                        \
  if (IS_FindOption (DEF, USR, 1, "kbo") == 0)                                          \
    ORD = kbo;                                                                          \
  else if (IS_FindOption (DEF, USR, 1, "lpo") == 0)                                     \
    ORD = lpo;                                                                          \
  switch (ORD){                                                                         \
  case kbo: KBOACTION;fprintf(IO_Strom, "KBO");break;                                   \
  case lpo: LPOACTION;fprintf(IO_Strom, "LPO");break;                                   \
  default:  ST_ErrorStatistik("Falsche Ordnung",FALSE);break;                           \
  }                                                                                     \
  if (IS_FindOption (DEF, USR, 1, "auto") == 0){                                        \
    fprintf(IO_Strom, ", Auto-Mode");                                                   \
    AUTO = TRUE;                                                                        \
  }                                                                                     \
  if (!AUTO){                                                                           \
    PREC=IS_ValueOfOption (DEF,USR,"prec");                                             \
    newperm=IS_ValueOfOption (DEF,USR,"perm");                                          \
                                                                                        \
    if (PREC){                                                                          \
      fprintf(IO_Strom, ", (new) precedence %s", PREC);                                 \
      if (!PRECTEST)                                                                    \
        PREC_NOK_ACTION;                                                                \
    }                                                                                   \
    else if (newperm){                                                                  \
      char *punkt;                                                                      \
      int permno;                                                                       \
      int permdegree=SO_AnzahlFunktionen();                                             \
      punkt=strchr(newperm,'.');                                                        \
      if (punkt){                                                                       \
        *punkt='\0';                                                                    \
        permdegree = atoi(punkt+1);                                                     \
      }                                                                                 \
      permno = atoi(newperm);                                                           \
      if ((PREC=OG_Permutator(permno, permdegree))){                                    \
        fprintf(IO_Strom, ", permutation %u is %s with degree %d", permno,              \
                PREC, permdegree);                                                      \
        PERM_OK_ACTION;                                                                 \
      }                                                                                 \
      else{                                                                             \
        fprintf(IO_Strom,                                                               \
                ", permutation %u with degree %d invalid or precedence not total",      \
                permdegree, permno);                                                    \
        PERM_NOK_ACTION;                                                                \
      }                                                                                 \
    }                                                                                   \
  }                                                                                     \
  fprintf(IO_Strom, "\n            current precedence is ");                            \
  if (SO_PraezedenzTotal()) {                                                           \
    unsigned int i;                                                                     \
    for(i=0; i<SO_AnzahlFunktionen(); i++){                                             \
      IO_DruckeSymbol(SO_TotaleKette[i]);                                               \
      if (i<SO_AnzahlFunktionen-1) fprintf(IO_Strom, " > ");                            \
    }                                                                                   \
  }                                                                                     \
  else                                                                                  \
    fprintf(IO_Strom, "not total");                                                     \
                                                                                        \
  fprintf(IO_Strom, "\n");                                                              \
}  

void ORD_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
  ordering ord = PA_Ordering;
  char *newprec;
  BOOLEAN autoOrd;

  if (what == ParamEvaluate){
    Auswertung(autoOrd,
               ord, 
               {;}, 
               {;},
               newprec, 
               (SO_PraezedenzKorrekt(newprec)),
               {fprintf(IO_Strom, " invalid, leaving unchanged");},
               {;},
               {;}
               );
  }
  else{
    fprintf(IO_Strom, "-ord [kbo|lpo]:prec=a.b.c..d.e.a  redefine ordering and precedence\n"
            "KBO only with weights=1; precedence above would be a>b>c;d>e>a\n"
            "-ord auto causes WM to generate an ordering\n");
  }
}
  
#if COMP_WMTW
static BOOLEAN ORD_SchonInitialisiert = FALSE;
#endif


BOOLEAN ORD_InitAposteriori (InfoStringT DEF, InfoStringT USR) 
{
  ordering oldOrd = PA_Ordering;
  char *newprec;
  BOOLEAN autoOrd;

#if COMP_WMTW
  if (ORD_SchonInitialisiert) return TRUE;
  else ORD_SchonInitialisiert=TRUE;
#endif
  Auswertung(autoOrd,
             PA_Ordering, 
             {
               ORD_Vergleichsfkt = KB_KBOVergleich;
               ORD_RedTestfkt = ORD_UnifTestfkt = KB_KBOGroesser;
               if (oldOrd != PA_Ordering) {
                 NewKbo();
                 fprintf(IO_Strom, "changed to ");
               } 
               KB_InitAposteriori();   /* Phi auf Gueltigkeit testen.*/
             }, 
             {
               if (oldOrd != PA_Ordering) fprintf(IO_Strom, "changed to ");
               ORD_Vergleichsfkt = LP_LPO;
               ORD_RedTestfkt =  ORD_UnifTestfkt = LP_LPOGroesser;
               LV_InitAposteriori();
             },
             newprec, 
             (SO_PraezedenzErneuert(newprec)),
             {ST_ErrorStatistik("Invalid Precedence", FALSE);},
             {if(!SO_PraezedenzErneuert(newprec)) ST_ErrorStatistik("Permutator generated invalid precedence\n",FALSE);},
             {ST_ErrorStatistik("Invalid permutation", FALSE);}
             );
  OG_InitAposteriori(autoOrd);
  return TRUE;
}

void ORD_OrdnungUmstellen(ordering Ordnung)
{
  switch (Ordnung){
  case lpo:
    ORD_Vergleichsfkt = LP_LPO;
    ORD_RedTestfkt =  ORD_UnifTestfkt = LP_LPOGroesser;
    LV_InitAposteriori();
    break;
  case kbo:
    ORD_Vergleichsfkt = KB_KBOVergleich;
    ORD_RedTestfkt = ORD_UnifTestfkt = KB_KBOGroesser;
    if (Ordnung != PA_Ordering) NewKbo();
    KB_InitAposteriori();   /* Phi auf Gueltigkeit testen.*/
    break;
  default:
    ST_ErrorStatistik("Invalid Ordering in ORD_OrdnungUmstellen", FALSE);
    break;
  }
  PA_Ordering = Ordnung;
}
