/*=================================================================================================================
  =                                                                                                               =
  =  Ordnungsgenerator.c                                                                                          =
  =                                                                                                               =
  =                                                                                                               =
  =  26.07.97 Angelegt. Thomas.                                                                                   =
  =                                                                                                               =
  =================================================================================================================*/



#include "Ausgaben.h"
#include "compiler-extra.h"
#include "DSBaumOperationen.h"
#include "MatchOperationen.h"
#include "Ordnungen.h"
#include "Ordnungsgenerator.h"
#include <stdarg.h>
#include <string.h>
#include "SpeicherVerwaltung.h"
#include "SymbolOperationen.h"
#include "Unifikation1.h"

/* Achtung: Beim Re-Init Liste "OG_Gleichungen" wieder auf NULL setzen. 
   Garbage Collection, ordentlicher Zellenfluss oder Manager komplett leeren.
*/

/*=================================================================================================================*/
/*= I. Lokale Cons-Zellen; Listen-Funktionen                                                                      =*/
/*=================================================================================================================*/


static SV_SpeicherManagerT OG_FreispeicherManager;

static ListenT cons(void *car, void *cdr)
{
  ListenT Zelle;
  SV_Alloc(Zelle,OG_FreispeicherManager,ListenT);
  Zelle->car = car;
  Zelle->cdr = cdr;
  return Zelle;
}

static void Cons(void *car, ListenT *cdr)
{
  *cdr = cons(car,*cdr);
}

#define /*void**/ car(/*ListenT*/ Zelle)\
  ((Zelle)->car)

#define /*void**/ cdr(/*ListenT*/ Zelle)\
  ((Zelle)->cdr)

static ListenT swap(ListenT Zelle)
{
  return cons(cdr(Zelle),car(Zelle));
}

static BOOLEAN member(void *Element, ListenT Liste)
{
  while (Liste) {
    if (Element==car(Liste))
      return TRUE;
    Liste = cdr(Liste);
  }
  return FALSE;
}

static ListenT insert(void *Element, ListenT Liste)
{
  return member(Element,Liste) ? Liste : cons(Element,Liste);
}

static ListenT append(ListenT, ListenT) MAYBE_UNUSEDFUNC;
static ListenT append(ListenT Liste1, ListenT Liste2)
{
  if (Liste1) {
    ListenT Index = Liste1;
    while (cdr(Index))
      Index = cdr(Index);
    cdr(Index) = Liste2;
    return Liste1;
  }
  else
    return Liste2;
}

static void mapc(void (*Funktion)(void*), ListenT Liste)
{
  while (Liste) {
    (*Funktion)(car(Liste));
    Liste = cdr(Liste);
  }
}

static void filter(BOOLEAN (*)(ListenT), ListenT, ListenT*, ListenT*) MAYBE_UNUSEDFUNC;
static void filter(BOOLEAN (*Funktion)(ListenT), ListenT Liste, ListenT *Sieger, ListenT *Verlierer)
{
  *Sieger = *Verlierer = NULL;
  while (Liste) {
    ListenT Element = car(Liste);
    if ((*Funktion)(Element))
      *Sieger = cons(Element, *Sieger);
    else
      *Verlierer = cons(Element, *Verlierer);
    Liste = cdr(Liste);
  }
}

static void maparrange3(VergleichsT (*)(ListenT), ListenT, ListenT*, ListenT*) MAYBE_UNUSEDFUNC;
static void maparrange3(VergleichsT (*Funktion)(ListenT), ListenT Liste,
  ListenT *Vergleichbare, ListenT *Unvergleichbare)
{
  *Vergleichbare = *Unvergleichbare = NULL;
  while (Liste) {
    ListenT Element = car(Liste);
    switch ((*Funktion)(Element)) {
    case Groesser:
      *Vergleichbare = cons(Element, *Vergleichbare);
      break;
    case Kleiner:
      *Vergleichbare = cons(swap(Element), *Vergleichbare);
      break;
    default:
      *Unvergleichbare = cons(Element, *Unvergleichbare);
    }
    Liste = cdr(Liste);
  }
}

static void maparrange4(VergleichsT (*)(ListenT), ListenT, ListenT*, ListenT*, ListenT*) MAYBE_UNUSEDFUNC;
static void maparrange4(VergleichsT (*Funktion)(ListenT), ListenT Liste,
    ListenT *EchtVergleichbare, ListenT *Gleiche, ListenT *Unvergleichbare)
  {
    *EchtVergleichbare = *Gleiche = *Unvergleichbare = NULL;
    while (Liste) {
      ListenT Element = car(Liste);
      switch ((*Funktion)(Element)) {
      case Groesser:
        *EchtVergleichbare = cons(Element, *EchtVergleichbare);
        break;
      case Kleiner:
        *EchtVergleichbare = cons(swap(Element), *EchtVergleichbare);
        break;
      case Gleich:
        *Gleiche = cons(Element, *Gleiche);
        break;
      default:
        *Unvergleichbare = cons(Element, *Unvergleichbare);
      }
      Liste = cdr(Liste);
    }
  }

static ListenT list(unsigned int AnzElemente, ...)
{
  va_list Argumente;
  ListenT Liste = NULL;
  va_start(Argumente,AnzElemente);
  if (AnzElemente) {
    void *Element = va_arg(Argumente,void*);
    ListenT Index = cons(Element,NULL);
    Liste = Index;
    while (--AnzElemente) {
      Element = va_arg(Argumente,void*);
      Index = cdr(Index) = cons(Element,NULL);
    }
  }
  va_end(Argumente);
  return Liste;
}


#define /*static void ArraysFktSymb(unsigned int AnzFelder, ...)*/ TypeArrays(/*Typ*/ Type, /*Expression*/ Initial) \
static void Type ## Arrays(unsigned int AnzFelder, ...)                                                             \
{                                                                                                                   \
  va_list Argumente;                                                                                                \
  va_start(Argumente,AnzFelder);                                                                                    \
  while (AnzFelder--) {                                                                                             \
    Type **Feld = va_arg(Argumente,Type**);                                                                         \
    SymbolT f;                                                                                                      \
    *Feld = array_alloc(SO_AnzahlFunktionen(),Type);                                                                \
    SO_forFktSymbole(f)                                                                                             \
      (*Feld)[f] = Initial;                                                                                         \
  }                                                                                                                 \
  va_end(Argumente);                                                                                                \
}

/*static void BOOLEANArrays(unsigned int AnzFelder, ...) */
TypeArrays(BOOLEAN, FALSE)

/*static void ListenTArrays(unsigned int AnzFelder, ...) */
TypeArrays(ListenT, NULL)



/*=================================================================================================================*/
/*= II. Aufsammeln des Gleichungssystems                                                                          =*/
/*=================================================================================================================*/

ListenT OG_Gleichungen;

static BOOLEAN GeneratorAktiviert;

void OG_GleichungMelden(TermT links, TermT rechts)
{
  if (GeneratorAktiviert) {
    ListenT Eintrag = cons(TO_Termkopie(links),TO_Termkopie(rechts));
    OG_Gleichungen = cons(Eintrag, OG_Gleichungen);
  }
}



/*=================================================================================================================*/
/*= IV. Generierung von Praezedenzen                                                                              =*/
/*=================================================================================================================*/

unsigned int OG_Schranke1;
char *OG_Strategie2;

static BOOLEAN *IstASymbol,
               *IstCSymbol,
               *IstIdempotent;
static ListenT *IstLinksEinsZu,          *IstRechtsEinsZu,
               *DistribuiertLinksUeber,  *DistribuiertRechtsUeber,
               *IstLinksInverseZu,       *IstRechtsInverseZu,
               *IstAssoziatorZu,         *IstKommutatorZu;
static BOOLEAN MaybeLattice,
               MaybeNassring;

static char *Praezedenzfeld(void)
{
  char *Praezedenzfeld = array_alloc((MaximaleBezeichnerLaenge+1)*SO_AnzahlFunktionen(),char);
  Praezedenzfeld[0] = '\0';
  return Praezedenzfeld;
}

static void SymbolAnhaengen(char *Bisher, SymbolT f)
{
  if (Bisher[0])
    strcat(Bisher,".");
  strcat(Bisher,IO_FktSymbName(f));
}

static ListenT DefinierteSuchen(ListenT Liste)
{
  if (Liste) {
    TermT l = car((ListenT)car(Liste)),
          r = cdr((ListenT)car(Liste));
    SymbolT f = TO_TopSymbol(l),
            g = TO_TopSymbol(r);
    ListenT Rest = DefinierteSuchen(cdr(Liste));
    return 
      TO_DefinierendeGleichung(l,r) ? insert((void*)(int)f,Rest) :
        (TO_DefinierendeGleichung(r,l) ? insert((void*)(int)g,Rest) :
          Rest);
  }
  else
    return NULL;
}


static char *FuchsPraezedenz(void)
{
  /* Erzeugte Praezedenz:
     (1) mit Paramod:  definierte EQUATION-Symbole > sonstige EQUATION- und      CONCLUSION-Symbole > Paramodulations-Symbole
     (2) ohne Paramod: definierte EQUATION-Symbole > sonstige EQUATION-Symbole > CONCLUSION-Symbole

     Definierte Symbole werden nur fuer die LPO verwendet, und zwar aus den EQUATION-Symbolen herausgenommen.

     innerhalb der definierten Symbole entsprechend Symbolzuordnung
       (Dies ist natuerlich fragwuerdig, betrachte etwa a=b, b=c, c=a, jedoch praktisch ausreichend.)
     innerhalb der EQUATION-Symbole 1-stellig > 2-stellig > ... > n-stellig > Konstanten
       bei gleicher Stelligkeit entsprechend Symbolzuordnung
     ebenso innerhalb von EQUATION- und CONCLUSION-Symbolen
     innerhalb der CONCLUSION-Symbole 0-stellig > 1-stellig > 2-stellig > ... > n-stellig
       bei gleicher Stelligkeit entsprechend Symbolzuordnung
     innerhalb der Paramodulations-Symbole EQ > TRUE > FALSE
  */
  char *Praezedenz = Praezedenzfeld();
  unsigned int i,
               N = SO_MaximaleStelligkeit();
  SymbolT f;
  ListenT DefinierteSymbole;

  /* definierte Symbole */
  DefinierteSymbole = PA_Ordering==lpo ? DefinierteSuchen(OG_Gleichungen) : NULL;
  SO_forEquationSymbole(f)
    if (member((void*)(int)f,DefinierteSymbole))
      SymbolAnhaengen(Praezedenz,f);

  /* (1) mit Paramodulation */

  if (SO_Paramodulation()) {
    /* sonstige EQUATION- und CONCLUSION-Symbole */
    for(i=1; i<=N+1; i++)
      SO_forSpezSymbole(f)
        if (i%(N+1)==SO_Stelligkeit(f) && !member((void*)(int)f,DefinierteSymbole))
          SymbolAnhaengen(Praezedenz,f);
  
    /* Paramodulations-Symbole */
    SymbolAnhaengen(Praezedenz,SO_EQSymbol());
    SymbolAnhaengen(Praezedenz,SO_TRUESymbol());
    SymbolAnhaengen(Praezedenz,SO_FALSESymbol());
  }

  /* (2) ohne Paramodulation */

  else {
    /* EQUATION-Symbole */
    for(i=1; i<=N+1; i++)
      SO_forEquationSymbole(f)
        if (i%(N+1)==SO_Stelligkeit(f) && !member((void*)(int)f,DefinierteSymbole))
          SymbolAnhaengen(Praezedenz,f);
  
    /* CONCLUSION-Symbole */
    for(i=0; i<=N; i++)
      SO_forConclusionSymbole(f)
        if (i==SO_Stelligkeit(f))
          SymbolAnhaengen(Praezedenz,f);
  }
  return Praezedenz;
}


static char *LatticePraezedenz(void)
{
  BOOLEAN IsThat = TRUE;
  char *Praezedenz = Praezedenzfeld();
  ListenT Benutzt = NULL;
  SymbolT f;

  #define /*void*/ OperatorFischen(/*char**/ Text, /*BOOLEAN*/ Bedingung, /*BOOLEAN*/ MussSein)                 \
    SO_forFktSymbole(f)                                                                                         \
      if (Bedingung && !member((void*)(int)f,Benutzt)) {                                                        \
        SymbolAnhaengen(Praezedenz,f);                                                                          \
        Cons((void*)(int)f,&Benutzt);                                                                           \
        break;                                                                                                  \
      }                                                                                                         \
    if (f>SO_GroesstesFktSymb() && MussSein) {                                                                  \
      IsThat = FALSE;                                                                                           \
    }

  #define /*void*/ OperatorenNehmen(/*char**/ Text, /*BOOLEAN*/ Bedingung)                                      \
    SO_forFktSymbole(f)                                                                                         \
      if (Bedingung && !member((void*)(int)f,Benutzt)) {                                                        \
        SymbolAnhaengen(Praezedenz,f);                                                                          \
        Cons((void*)(int)f,&Benutzt);                                                                           \
      }

  OperatorFischen("inverse",  IstRechtsInverseZu[f], TRUE);
  OperatorFischen("multiply", IstASymbol[f] && DistribuiertLinksUeber[f] && DistribuiertRechtsUeber[f], TRUE);
  OperatorFischen("lub/glb",  IstASymbol[f] && IstCSymbol[f] && IstIdempotent[f], TRUE);
  OperatorFischen("lub/glb",  IstASymbol[f] && IstCSymbol[f] && IstIdempotent[f], TRUE);
  OperatorFischen("neg/pos",  SO_Stelligkeit(f)==1, FALSE);
  OperatorFischen("neg/pos",  SO_Stelligkeit(f)==1, FALSE);
  OperatorenNehmen("restE",   !IstRechtsEinsZu[f] && f<=SO_GroesstesESymb);
  OperatorFischen("identity", IstRechtsEinsZu[f], TRUE);
  OperatorenNehmen("restG",   TRUE);
  return IsThat ? Praezedenz : NULL;
}


static char *NassringPraezedenz(void)
{
  BOOLEAN IsThat = TRUE;
  char *Praezedenz = Praezedenzfeld();
  ListenT Benutzt = NULL;
  SymbolT f;
  OperatorFischen("commut", IstKommutatorZu[f], TRUE);
  OperatorFischen("assoc",  IstAssoziatorZu[f], TRUE);
  OperatorFischen("minus",  IstRechtsInverseZu[f], TRUE);
  OperatorFischen("mal",    !IstASymbol[f] && !IstCSymbol[f] && DistribuiertRechtsUeber[f], TRUE);
  OperatorFischen("plus",   IstASymbol[f] && IstCSymbol[f], TRUE);
  OperatorFischen("null",   IstRechtsEinsZu[f], TRUE);
  OperatorenNehmen("restE", f<=SO_GroesstesESymb);
  OperatorenNehmen("restG", TRUE);
  if (IsThat) {
    char *Praezedenz1 = Praezedenz;
    Praezedenz = Praezedenzfeld();
    Benutzt = NULL;
    OperatorFischen("assoc",  IstAssoziatorZu[f], TRUE);
    OperatorFischen("commut", IstKommutatorZu[f], TRUE);
    OperatorFischen("plus",   IstASymbol[f] && IstCSymbol[f], TRUE);
    OperatorFischen("mal",    !IstASymbol[f] && !IstCSymbol[f] && DistribuiertRechtsUeber[f], TRUE);
    OperatorFischen("minus",  IstRechtsInverseZu[f], TRUE);
    OperatorFischen("null",   IstRechtsEinsZu[f], TRUE);
    OperatorenNehmen("restE", f<=SO_GroesstesESymb);
    OperatorenNehmen("restG", TRUE);
    OG_Schranke1 = 1000;
    OG_Strategie2 = Praezedenz;
    return Praezedenz1;
  }
  else
    return NULL;
}



/*=================================================================================================================*/
/*= V. Analyse                                                                                                    =*/
/*=================================================================================================================*/

/*--- 1. Analyse einer Gleichung ----------------------------------------------------------------------------------*/

static unsigned int Ccount, lEcount, rEcount, lDcount, rDcount, Acount, 
                    lInvcount, rInvcount, Idemcount, AssCount, CommCount;

static void GleichungAnalysieren(void *Gleichung)
{
  TermT s = car((ListenT)Gleichung),
        t = cdr((ListenT)Gleichung);
  BOOLEAN links;
  SymbolT was, worueber, wozu, ergibt;
  SymbolT ass, comm, mal, plus, minus;
  if (TO_IstAssoziativitaet(s,t)) {
    IstASymbol[TO_TopSymbol(s)] = TRUE;
    Acount++;
  }
  else if (TO_IstKommutativitaet(s,t)) {
    IstCSymbol[TO_TopSymbol(s)] = TRUE;
    Ccount++;
  }
  else if (TO_IstDistributivitaet(s,t,&was,&worueber,&links)) {
    SymbolT *Eintrag = array_alloc(1,SymbolT);
    Eintrag[0] = worueber;
    if (links) {
      Cons(Eintrag,&DistribuiertLinksUeber[was]);
      lDcount++;
    }
    else {
      Cons(Eintrag,&DistribuiertRechtsUeber[was]);
      rDcount++;
    }          
  }
  else if (TO_IstEinheit(s, t, &was, &wozu, &links)) {
    SymbolT *Eintrag = array_alloc(1,SymbolT);
    Eintrag[0] = wozu;
    if (links) {
      Cons(Eintrag,&IstLinksEinsZu[was]);
      lEcount++;
    }
    else {
      Cons(Eintrag,&IstRechtsEinsZu[was]);
      rEcount++;
    }         
  }
  else if (TO_IstInverse(s, t, &was, &wozu, &ergibt, &links)) {
    SymbolT *Eintrag = array_alloc(2,SymbolT);
    Eintrag[0] = wozu;
    Eintrag[1] = ergibt;
    if (links) {
      Cons(Eintrag,&IstLinksInverseZu[was]);
      lInvcount++;
    }
    else {
      Cons(Eintrag,&IstRechtsInverseZu[was]);
      rInvcount++;
    }          
  }
  else if (TO_IstIdempotenz(s, t, &was)) {
    IstIdempotent[was] = TRUE;
    Idemcount++;
  }
  else if (TO_IstAssoziator(s, t, &ass, &mal, &plus, &minus)) {
    SymbolT *Eintrag = array_alloc(3,SymbolT);
    Eintrag[0] = mal;
    Eintrag[1] = plus;
    Eintrag[2] = minus;
    Cons(Eintrag,&IstAssoziatorZu[ass]);
    AssCount++;
  }
  else if (TO_IstKommutator(s, t, &comm, &mal, &plus, &minus)) {
    SymbolT *Eintrag = array_alloc(3,SymbolT);
    Eintrag[0] = mal;
    Eintrag[1] = plus;
    Eintrag[2] = minus;
    Cons(Eintrag,&IstKommutatorZu[comm]);
    CommCount++;
  }
}


/*--- 2. Ordnungsgenerierung --------------------------------------------------------------------------------------*/

void OG_OrdnungGenerieren(void)
{
  /* Fuer die Gleichungsanalyse wird vorausgesetzt, dass die Gleichungen car-cdr-variablennormiert sind. */

  if (GeneratorAktiviert) {
    char *Praezedenz;
    SymbolT f;
    unsigned long int stell = 1;
    OG_Schranke1 = 0;
    OG_Strategie2 = NULL;

    BOOLEANArrays(3, &IstASymbol, &IstCSymbol, &IstIdempotent);
    ListenTArrays(8, &IstLinksEinsZu, &IstRechtsEinsZu, &DistribuiertLinksUeber, &DistribuiertRechtsUeber, 
      &IstLinksInverseZu, &IstRechtsInverseZu, &IstAssoziatorZu, &IstKommutatorZu);
    MaybeLattice = MaybeNassring = FALSE;
    Ccount = lEcount = rEcount = lDcount = rDcount = Acount = 
      lInvcount = rInvcount = Idemcount = AssCount =  CommCount = 0;

    /* Strukturen analysieren */

    mapc(GleichungAnalysieren, OG_Gleichungen);
    SO_forFktSymbole(f)
      stell *= PrimzahlNr(SO_Stelligkeit(f));

    /* Ordnung generieren */

    ORD_OrdnungUmstellen(
       ((Acount==1 && Ccount==1 && rEcount==1) || 
         Acount==2 ||
        (Acount==3 && Ccount==2)) ? lpo : kbo);
    MaybeLattice = Acount==3 && Ccount==2;
    MaybeNassring = Acount==1 && Ccount==1 && rEcount==1;

    /* Praezedenz besorgen */

    if (MaybeLattice && !SO_Paramodulation()) {
      char *LPraezedenz = LatticePraezedenz();
      Praezedenz = LPraezedenz ? LPraezedenz : FuchsPraezedenz();
    }
    else if (MaybeNassring && !SO_Paramodulation()) {
      char *NPraezedenz = NassringPraezedenz();
      Praezedenz = NPraezedenz ? NPraezedenz : FuchsPraezedenz();
    }
    else
      Praezedenz = FuchsPraezedenz();

    SO_PraezedenzErneuert(Praezedenz);
  }
}



/*=================================================================================================================*/
/*= VI. Grosser Permutator                                                                                        =*/
/*=================================================================================================================*/

static BOOLEAN NaechsterWert(unsigned int *Ist, unsigned int N)
{
  unsigned int i;
  for(i=0; i<N; i++)
    if (SO_TotaleKette[i]==(signed)*Ist)
      break;
  if (i==N-1) {
    *Ist = SO_AnzahlFunktionen();
    return FALSE;
  }
  else {
    *Ist = SO_TotaleKette[i+1];
    return TRUE;
  }
}

static BOOLEAN Benutzt(unsigned int n, unsigned int *Feld, unsigned int Limit) 
{
  unsigned int i;
  for(i=0; i<Limit; i++)
    if (Feld[i]==n)
      return TRUE;
  return FALSE;
}

static unsigned int *Permutation(unsigned int Nummer, unsigned int N)
{
  unsigned int i, j,
               *Feld = array_alloc(N,unsigned int);
  for(i=0; i<N; i++)
    Feld[i] = SO_TotaleKette[i];

  while (Nummer--) {                                                        /* Vorwaerts zur naechsten Permutation */
    unsigned int Level = N;
    do {
      while (NaechsterWert(&Feld[Level-1],N) && Benutzt(Feld[Level-1],Feld,Level-1));
      if (Feld[Level-1]<SO_AnzahlFunktionen())
        break;
    } while (--Level);
  /*if (!Level)                                                         Nummer<fac(N) wird vor dem Aufruf getestet   
      our_assertion_failed("Permutator enters next cycle\n");*/ 
    for(i=Level; i<N; i++)
      for(j=0; j<N; j++)
        if (!Benutzt(SO_TotaleKette[j],Feld,i)) {
          Feld[i] = SO_TotaleKette[j];
          break;
        }
  }
  return Feld;
}

char *OG_Permutator(unsigned int Nummer, unsigned int Grad)
/* Eingabe: natuerliche Zahlen Nummer, Grad
   Ausgabe: Fuer Grad<=SO_AnzahlFunktionen, Nummer<fac(Grad) 
              Zeiger auf eine Zeichenkette mit den Grad groessten Funktionssymbolen der initialen Praezedenz
              in entsprechender Verdrehung; 0 = keine Verdrehung. Der Permutator verdreht von hinten nach vorne.
            Ansonsten NULL. */
{
  if (Grad>SO_AnzahlFunktionen() || Nummer>=fac(Grad) || !SO_PraezedenzTotal())
    return NULL;
  else {
    unsigned int *Feld = Permutation(Nummer,Grad);
    char *Praezedenz = Praezedenzfeld();
    unsigned int i;
    for(i=0; i<Grad; i++)
      SymbolAnhaengen(Praezedenz,(signed)Feld[i]);
    our_dealloc(Feld);
    for(i=Grad; i<SO_AnzahlFunktionen(); i++)
      SymbolAnhaengen(Praezedenz,SO_TotaleKette[i]);
    return Praezedenz;
  }
}

/*=================================================================================================================*/
/*= VII. Initialisierung                                                                                          =*/
/*=================================================================================================================*/

void OG_InitAposteriori(BOOLEAN GeneratorAktivieren)
{
  if ((GeneratorAktiviert = GeneratorAktivieren)) {
    SV_ManagerInitialisieren(&OG_FreispeicherManager, sizeof(struct ListenTS), OG_Blockgroesse, 
      "Conszellen fuer Ordnungsgenerator");
    OG_Gleichungen = NULL;
  }
}



