/* 
   protExtr.c:
   
   Korrespondez-File-Extraktion gemaess vorgegebenem PCL-Extrakt-File:
   
   mit Hilfe des extract-Files (aus einem PCL-Protokoll) werden aus dem zu diesem
   Lauf gehoerenden Korrespondenz-File diejenigen Zeilen extrahiert, die
   'relevante' Regeln bzw. Gleichungen beschreiben.
   
   Im Anschluss wird aus dem extrahierten Korrespondenz-File und dem
   Waldmeister-Protokoll (Level 2) ein extrahiertes Laufprotokoll
   generiert, welches nur noch diejenigen Zeilen enthaelt, die fuer den
   Beweislauf noetig waren.

 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "antiwarnings.h"


#define MARKER "***Extr-Fact***   "
#define KPIRPattern "KP-Mengen-IR"

int main(int argc, char *argv[])
{
   char *corrFilename, *extrFilename, tmpFilename[L_tmpnam], *protFilename, 
        extrStr[10000], corrStr[10000], protStr[10000], tmpStr[10000], *SuchStr,
        *targetFilename;

   FILE *corrFile, *extrFile, *tmpFile, *protFile, *targetFile;

   char *SuchPattern[2] = {"Neue Regel:", "Neue Gleichung:"};
   int was;
   
   unsigned long extrNr, corrNr, protNr;
   unsigned long int SuchNr, corrFacts = 0, protFacts = 0;

   if (argc == 1){
      printf("\n\nprotExtr    (C) 1996 by A.Buch & Th. Hillenbrand\n\n"
             "Protokoll-Extraktion aus vorgegebenem PCL-Extrakt-File:\n"
             "*******************************************************\n\n"
             "Mit Hilfe des extract-Files (aus einem PCL-Protokoll) werden aus dem zu diesem\n"
             "Lauf gehoerenden Korrespondenz-File diejenigen Zeilen extrahiert, die\n"
             "beweisrelevante Regeln bzw. Gleichungen beschreiben.\n\n"
             "Im Anschluss wird aus dem extrahierten Korrespondenz-File und dem\n"
             "Waldmeister-Protokoll (Level 2) ein extrahiertes Laufprotokoll\n"
             "generiert, welches nur noch diejenigen Zeilen enthaelt, die fuer den\n"
             "Beweislauf noetig waren.\n\n"
             "Usage: protExtr <extr-File> <corr-File> <prot-File> <target-file>\n\n");
      exit(-1);
   }
      
   if (argc != 5){
      printf("Usage: protExtr <extr-File> <corr-File> <prot-File> <target-file>\n\n");
      exit(-1);
   }
   
   tmpnam(tmpFilename);
   printf("%s\n", tmpFilename);
   extrFilename = argv[1];
   corrFilename = argv[2];
   protFilename = argv[3];
   targetFilename = argv[4];

   if (!(extrFile = fopen(extrFilename, "r"))){
      printf("Error: Couldn't open %s.\n\n", extrFilename);
      exit(-1);
   }

   if (!(corrFile = fopen(corrFilename, "r"))){
      printf("Error: Couldn't open %s.\n\n", corrFilename);
      fclose(corrFile);
      exit(-1);
   }

   if (!(protFile = fopen(protFilename, "r"))){
      printf("Error: Couldn't open %s.\n\n", protFilename);
      fclose(extrFile);
      fclose(corrFile);
      exit(-1);
   }

   if (!(tmpFile = fopen(tmpFilename, "w"))){
      printf("Error: Couldn't open temporary file.\n\n");
      fclose(extrFile);
      fclose(corrFile);
      fclose(protFile);
      exit(-1);
   }

   if (!(targetFile = fopen(targetFilename, "w"))){
      printf("Error: Couldn't open target file %s.\n\n", targetFilename);
      fclose(extrFile);
      fclose(corrFile);
      fclose(protFile);
      fclose(tmpFile);
      exit(-1);
   }
   

   extrNr = 0;

   printf("********** Corr-Extraction started\n");

   while(fgets(corrStr, 10000, corrFile)){
      corrNr = atoi(corrStr);

      /* Solange aus extr-file lesen, bis extrNr groesser oder gleich corrNr ist */
      /* Lesen aber nur dann beginnen, wenn extrNr auch kleiner ist als corrNr.*/
      /* Andernfalls wird nur im corrFile weiter gelesen. */

      while((extrNr<corrNr) && fgets(extrStr, 10000, extrFile) && 
            ((extrNr = atoi(extrStr)) < corrNr));

      if (extrNr == corrNr){
         fprintf(tmpFile, "%s", corrStr);
         printf(corrStr);
         corrFacts ++;
      }
   }
   
   printf("********** Corr-Extraction finished (%ld facts extracted)\n", corrFacts);
   fclose(tmpFile);

   if (!(tmpFile = fopen(tmpFilename, "r"))){
      printf("Error: Couldn't open temporary file.\n\n");
      exit(-1);
   }
   
   printf("********** Prot-Extraction started\n");

   while(fgets(tmpStr, 10000, tmpFile)){ /* extrahiertes corr-File durchgehen */

      if ((SuchStr = strstr(tmpStr, "R")))          /* Regel oder */
         was = 0;
      else if ((SuchStr = strstr(tmpStr, "E")))     /* ... Gleichung ? */
         was = 1;
      else
         was = 2;
      
      SuchNr = abs(atoi(SuchStr+1)); /* Zu findende Nummer einer Regel bzw. Gleichung */
/*      printf("Suche nach %s %ld\n", SuchPattern[was], SuchNr);*/
      while (fgets(protStr, 10000, protFile)){ /* naechste Zeile mit Regel oder Gleichung im Prot. suchen */
         int flag = 0;
         if ((was!=2) && (SuchStr = strstr(protStr, SuchPattern[was]))){
            /* das gefundene Suchpattern entspricht dem gesuchten Typ (R/E) */
            
            /* Nummer bestimmen: */
            SuchStr = strstr(protStr, ":");
            SuchStr = SuchStr++;
            protNr = abs(atoi(SuchStr));
/*            printf("Gefunden : %ld\n", protNr);*/
            
            if (protNr == SuchNr){
               fprintf(targetFile, MARKER);
               protFacts ++;
               flag = 1;
            }
         }
         if (strstr(protStr, KPIRPattern))
            fprintf(targetFile, MARKER);
         fprintf(targetFile, protStr);
         if (flag){
            flag = 0;
            break;
         }
      }
      
   }

   printf("********** Prot-Extraction finished(%ld facts extracted)\n", protFacts);
      
   fclose(extrFile);
   fclose(corrFile);
   fclose(protFile);
   fclose(targetFile);
   fclose(tmpFile);
   remove(tmpFilename);
   exit(0);
}

