#include "antiwarnings.h"
#include <stdio.h>
#include <string.h>
#include "Ausgaben.h"
#include "Classification.h"
#include "general.h"
#include "HeapOfHeaps.h"
#include "InfoStrings.h"
#include "PQStatistik.h"
#include "RUndEVerwaltung.h"
#include "SUEBasis.h"
#include "SpeicherVerwaltung.h"
#include "StrahlMemorium.h"
#include "Termpaare.h"
#include "TermpairRepresentation.h"

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe HeapOfHeaps.h) -----------------------*/
/* ------------------------------------------------------------------------*/
void (* HOH_INSERT)   (TermpairT, WeightEntryT, RegelOderGleichungsT, RegelOderGleichungsT);
void (* HOH_INSERTImmortal)   (TermpairT, WeightEntryT, RegelOderGleichungsT, RegelOderGleichungsT);
BOOLEAN (* HOH_SELECT) (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *, void **);
void (* HOH_NEWPARENT) (RegelOderGleichungsT re);
void (* HOH_KILLPARENT) (RegelOderGleichungsT re);
void (* HOH_REDIRECTPARENT) (void);
void (* HOH_FINISHPARENT) (void);
void (* HOH_WALKTHROUGH) (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms);

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Data-Structures --------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

typedef struct SHTS_0 SubHeapZellenT_0;
typedef struct SHTS_1 SubHeapZellenT_1;
typedef struct SHTS_2 SubHeapZellenT_2;

#define INFOTS_BODY(StrukturTyp)                                                                                \
   long int ElternNummer;                                                                                       \
   unsigned long AnzNochNichtBehandelterKPs;                                                                    \
   RegelOderGleichungsT re;                                                                                     \
   BOOLEAN LebtNoch;                                                                                            \
   SubHeapZellenT_ ## StrukturTyp *subheap;                                                                     \
   unsigned long NaechstesFreies;                                                                               \
   unsigned long AnzInitialerKPs;                                                                               \
   BOOLEAN InTopHeap;


typedef struct InfoTS_0{        /* Info-Block mit Statistik */
   INFOTS_BODY(0)
} InfoT_0;

typedef struct InfoTS_1{        /* Info-Block mit Statistik */
   INFOTS_BODY(1)
} InfoT_1;

typedef struct InfoTS_2{        /* Info-Block mit Statistik */
   INFOTS_BODY(2)
} InfoT_2;

typedef struct{
   InfoT_0 *info;
   long int w1;                 /* Die Gewichte des Top-Elements, um eine Indirektion zu sparen */
} TopEntryT_0;

typedef struct{
   InfoT_1 *info;
   long int w1;                 /* Die Gewichte des Top-Elements, um eine Indirektion zu sparen */
   long int w2;
} TopEntryT_1;

typedef struct{
   InfoT_2 *info;
   long int w1;                 /* Die Gewichte des Top-Elements, um eine Indirektion zu sparen */
   long int w2;
   void *addRef;
   TermpairT tp;
   InfoT_2 *OtherParent;
} TopEntryT_2;



struct SHTS_0{                  /* Eintrag im SubHeap: 1 long als Gewicht */
   InfoT_0 *OtherParent;
   TermpairT tp;
   long int w1;
} ;

struct SHTS_1{                  /* Eintrag im SubHeap: 2 long als Gewicht */
   InfoT_1 *OtherParent;
   TermpairT tp;
   long int w1, w2;
} ;

struct SHTS_2{                  /* Eintrag im SubHeap: 2 long + 1 pointer als Gewicht */
   InfoT_2 *OtherParent;
   TermpairT tp;
   long int w1, w2;
   void *addRef;
} ;


/* --------------------------------------------------------------------------------------------------------*/
/* ------- Globale Variablen ------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

static unsigned long  TopHeapAktuelleGroesse;
static unsigned long  DynHeapAktuelleGroesse;

static unsigned long TopHeapNaechstesFreies;

static TopEntryT_0 *TopHeap_0;
static TopEntryT_1 *TopHeap_1;
static TopEntryT_2 *TopHeap_2;

static InfoT_0 *DynInfo_0;
static InfoT_1 *DynInfo_1;
static InfoT_2 *DynInfo_2;

static SV_SpeicherManagerT InfoManager;

static int WeightEntryStrukturTyp; 
/* 0 -> 1 long 
   1 -> 2 long
   2 -> 2 long + 1 pointer
*/


static DomaenenT Domaene;
static unsigned long int DomainGran;

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Statistics 1 -----------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

static unsigned long int AnzRegelnMitInitial0KPs = 0, AnzGleichungenMitInitial0KPs = 0,
                         AnzRegelnMitInitialenKPs = 0, AnzGleichungenMitInitialenKPs = 0,
                         SummeInitialerKPsZuRegeln = 0, SummeInitialerKPsZuGleichungen = 0;

#define IncAnzI0KPR_Stat() (AnzRegelnMitInitial0KPs++)
#define IncAnzI0KPG_Stat() (AnzGleichungenMitInitial0KPs++)
#define IncSIKPR_Stat(anz) (AnzRegelnMitInitialenKPs++, SummeInitialerKPsZuRegeln += anz)
#define IncSIKPG_Stat(anz) (AnzGleichungenMitInitialenKPs++, SummeInitialerKPsZuGleichungen += anz)

#define IncAnzI0KPR_NoStat() 
#define IncAnzI0KPG_NoStat() 
#define IncSIKPR_NoStat(anz)
#define IncSIKPG_NoStat(anz)

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Heap-Behandlung fuer allgemeine Heaps ----------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

/* Hier finden sich Prozeduren, die zur Verwaltung allgemeiner Heaps benutzt werden koennen.
   Dabei ist der Top-Heap moeglichst so zu gestalten, dass er sich wie die Sub-Heaps verwalten laesst (insb.
   mit denselben Prozeduren). Diese erhalten dieselbe Aufgliederung wie beim regular heap, dh. mit
   den Varianten e1...e4 und random. Ferner wird unterschieden, ob die Statistik ein- oder ausgeschaltet ist
   und natuerlich, welche Groesse die Weight-Entries haben.
   
   Die Info-Bloecke enthalten im wesentlichen einen Zeiger auf den anhaengigen Subheap und Informationen zu
   dessen Verwaltung. Ferner die ueblichen Informationen bzgl. Status (lebend/tot) und noch anhaengenden KPs.
   
   Ein KP-Eintrag enthaelt:
   - einen Zeiger auf das Termpaar
   - einen Gewichtseintrag
   - einen Zeiger auf den anderen Elternteil

   Jedes Element des Top-Heaps erhaelt das Gewicht des anhaengenden Subheaps fest eingeschrieben.
   Dies bedingt, dass jede INSERT- bzw. SELECT-Ausfuehrung auf einem Subheap die Uebertragung des
   Gewichts des aktuellen Top-Elementes nach sich zieht.
 
   KPs zu Gleichungen werden in einem einzigen Array abgespeichert, welches aber aus zwei Subheaps besteht.
   Dies garantiert die noetige Trennung nach den beiden Richtungen als Eltern.

   Dynamischer Extra-Subheap fuer unsterbliche KPs ist vorzusehen.

   Extra-Funktion, die das Ende der KP-Bildung zu einem Elternteil angibt.

   Operationen auf Top-Heap:
   
   - Einfuegen eines Elementes
   - Entfernen eines Elementes (?)
     InfoBloecke zu toten RE fruehestens dann loeschen, wenn zugehoeriges TopHeap-Element an Top-Position angekommen ist.
     Loeschen aber erst dann, wenn es keine spaeteren KPs dazu mehr gibt.
     ==> Loeschen? == Aus TopHeapEntfernt && KeineKPsMehr 
   - Neu-Einordnen eines Elementes (LasseEinsinken nach oben sowie unten fortlaufend)
   - Durchlauf

   ==> LasseEinsinken fuer TopHeap mit der ueblichen Funktionalitaet, bis auf addref-Vergleiche, die auf
       die den zu vergleichenden Elementen anhaengenden Subheaps durchgreifen. BoundCheck ausfuehren.

   Operationen auf SubHeaps und dem dyn.Heap:
   
   - Einfuegen
   - Top entfernen
   - Neusortieren nach Durchlauf
   
   KPZuElternGeloescht darf die zugehoerigen InfoBloecke noch nicht entfernen. Diese werden erst geloescht, wenn sie
   im TopHeap auf der Top-Position angekommen sind.

   Leere SubHeaps werden NICHT angelegt. Es gibt eine zusaetzliche Funktion, die anzeigt, wenn die KP-Bildung
   zur aktuellen RE abgeschlossen ist. Diese friert den zugehoerigen Strahl ein und sorgt dafuer, dass ggf.
   eine NULL als Subheap eingetragen wird. In diesem Falle wird der Info-Block natuerlich nicht in den TopHeap
   eingetragen.

*/

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Operationen auf Info-Bloecken ------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

#define GGfStrahlLoeschen(/*{0,1,2}*/ WeightEntryType, INFO)                                                                    \
/* Voraussetzung: info->NaechstesFreies == 0 */                                                                                 \
{                                                                                                                               \
   if (RE_IstRegel(INFO->re)){                                                                                                  \
      if (INFO->subheap){                                                                                                       \
         SM_StrahlLoeschen(INFO->subheap);                                                                                      \
         INFO->subheap = NULL;                                                                                                  \
      }                                                                                                                         \
   }                                                                                                                            \
   else                                                                                                                         \
      if (((InfoT_ ## WeightEntryType *) TP_SUEInfo(TP_Gegenrichtung(INFO->re))) -> NaechstesFreies == 0){                      \
         if (TP_RichtungAusgezeichnet(INFO->re) && (INFO->subheap))                                                             \
            SM_StrahlLoeschen(INFO->subheap);                                                                                   \
         else if (TP_RichtungUnausgezeichnet(INFO->re) &&                                                                       \
                  (((InfoT_ ## WeightEntryType *) TP_SUEInfo(TP_Gegenrichtung(INFO->re))) -> subheap))                          \
            SM_StrahlLoeschen(((InfoT_ ## WeightEntryType *) TP_SUEInfo(TP_Gegenrichtung(INFO->re))) -> subheap);               \
         INFO->subheap = NULL;                                                                                                  \
         ((InfoT_ ## WeightEntryType *) TP_SUEInfo(TP_Gegenrichtung(INFO->re))) -> subheap = NULL;                              \
      }                                                                                                                         \
}
         

#define InfoBlockInitialisieren(info)                                                                                           \
      info -> AnzNochNichtBehandelterKPs = 0;                                                                                   \
      info -> AnzInitialerKPs = 0;                                                                                              \
      info -> NaechstesFreies = 0;                                                                                              \
      info -> InTopHeap = FALSE;                                                                                                \
      info -> LebtNoch = TRUE;                                                                                                  \


#define GGfInfoBlockLoeschen(INFO)                                                                                              \
{                                                                                                                               \
   if ( (INFO->AnzNochNichtBehandelterKPs==0) && (!INFO->InTopHeap) && (!INFO->LebtNoch) && (!INFO->subheap))                   \
      SV_Dealloc(InfoManager, INFO);                                                                                            \
}

  

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Operationen auf Subheap-Elementen --------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

#define KPVonElternGeloescht(actPar, oPar)                                                                                      \
{                                                                                                                               \
   actPar->AnzNochNichtBehandelterKPs--;                                                                                        \
   oPar->AnzNochNichtBehandelterKPs--;                                                                                          \
   GGfInfoBlockLoeschen(oPar);                                                                                                  \
}


#define TermeAuspacken_terms(tp, lhsPtr, rhsPtr, actualP, otherP, pclIDPtr) TPR_TERMPAIRtoPARALLEL_INTERMED(tp, lhsPtr, rhsPtr, actualP, otherP, pclIDPtr)
#define TermeAuspacken_noterms(tp, lhsPtr, rhsPtr, actualParent, otherParent, pclIDPtr) /* nix tun */

#define TermeWiederEinpacken_terms(tp, lhs, rhs, actualP, otherP, changed, pclID) TPR_PARALLELtoTERMPAIR(tp, lhs, rhs, actualP, otherP, changed, pclID)
#define TermeWiederEinpacken_noterms(tp, lhs, rhs, actualParent, otherParent, changed, pclID)  tp

#define TermeZerstoeren_terms(tp, lhs, rhs, actualP, otherP) TPR_DESTROYPARALLEL(tp, lhs, rhs, actualP, otherP)
#define TermeZerstoeren_noterms(tp, lhs, rhs, actualParent, otherParent) /* nix tun */


#define GewichteUebertragen_0(ziel, quelle) (ziel.w1=quelle.w1)
#define GewichteUebertragen_1(ziel, quelle) {ziel.w1=quelle.w1; ziel.w2=quelle.w2;}
#define GewichteUebertragen_2(ziel, quelle) {ziel.w1=quelle.w1; ziel.w2=quelle.w2; ziel.addRef=quelle.addRef;}

#define GewichteHochziehen_0(ziel, quelle) (ziel.w1 = quelle.w1)
#define GewichteHochziehen_1(ziel, quelle) {ziel.w1 = quelle.w1; ziel.w2 = quelle.w2;}
#define GewichteHochziehen_2(ziel, quelle) {ziel.w1 = quelle.w1; ziel.w2 = quelle.w2; ziel.addRef = quelle.addRef;              \
                                            ziel.OtherParent = quelle.OtherParent; ziel.tp = quelle.tp;}

#define LebtNoch(elt) (elt.OtherParent->LebtNoch)



#define C_Groesser_0(info1, info2, e1, e2) (e1.w1 > e2.w1)
 
#define C_Groesser_1(info1, info2, e1, e2) ((e1.w1 > e2.w1) || ((e1.w1 == e2.w1) && (e1.w2 > e2.w2)))

#define C_Groesser_2(info1, info2, e1, e2)                                                                                      \
((e1.w1 > e2.w1) ||                                                                                                             \
 ((e1.w1 == e2.w1) && (e1.w2 > e1.w2)) ||                                                                                       \
 ((e1.w1 == e2.w1) && (e1.w2 == e2.w2) &&                                                                                       \
  info1->LebtNoch && info2->LebtNoch && e1.OtherParent->LebtNoch && e2.OtherParent->LebtNoch &&                                 \
  C_AdditionalComparison (e1.w1, e1.w2, e1.addRef, info1->re, e1.OtherParent->re, e1.tp,                                        \
                          e2.w1, e2.w2, e2.addRef, info2->re, e2.OtherParent->re, e2.tp)))
/* Die additional comparison darf nur mit lebenden KPs aufgerufen werden! */


#define DESTROYADDREF_0(addref)
#define DESTROYADDREF_1(addref)
#define DESTROYADDREF_2(addref) (C_KillAdditionalReference(addref))

#define ADDREF_BESETZEN_0(addrefPtr, addref) 
#define ADDREF_BESETZEN_1(addrefPtr, addref) 
#define ADDREF_BESETZEN_2(addrefPtr, addref) (*addrefPtr = addref) 

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Operationen auf Subheaps -----------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

#define LasseEinsinkenBody(infoPtr, position, /*{0,1,2}*/ WeightEntryType)                                                      \
{                                                                                                                               \
   BOOLEAN changed = FALSE,isHeap = FALSE;                                                                                      \
   unsigned long j=position*2+1;                                                                                                \
   SubHeapZellenT_ ## WeightEntryType *ElementArray = infoPtr->subheap;                                                         \
   SubHeapZellenT_ ## WeightEntryType merker= ElementArray[position];                                                           \
                                                                                                                                \
   while ((!isHeap) && (j<infoPtr->NaechstesFreies)){                                                                           \
      if ((j+1<infoPtr->NaechstesFreies) &&                                      /* Position j+1 noch gueltig? */               \
          (C_Groesser_ ## WeightEntryType (infoPtr, infoPtr, ElementArray [j], ElementArray[j+1])))                             \
         j++;                                                  /* => j hat jetzt den Wert des leichteren Nachfolgers */         \
      if (C_Groesser_ ## WeightEntryType (infoPtr, infoPtr, merker, ElementArray[j])){                                          \
         ElementArray[position]=ElementArray[j];                                                                                \
         changed = TRUE;                                                                                                        \
         position = j;                                                                                                          \
         j *=2;                                                                                                                 \
         j++;                                                                                                                   \
      }                                                                                                                         \
      else                                                                                                                      \
         isHeap = TRUE;                                                                                                         \
   }                                                                                                                            \
   if (changed)                                                                                                                 \
      ElementArray[position] = merker;                                                                                          \
   return changed;                                                                                                              \
}

static BOOLEAN SHLasseEinsinken_0(InfoT_0 *info, unsigned long position)
LasseEinsinkenBody(info, position, 0)

static BOOLEAN SHLasseEinsinken_1(InfoT_1 *info, unsigned long position)
LasseEinsinkenBody(info, position, 1)

static BOOLEAN SHLasseEinsinken_2(InfoT_2 *info, unsigned long position)
LasseEinsinkenBody(info, position, 2)



#define SHTopElementLoeschenUndReorganisieren(/*{0,1,2}*/ WeightEntryType, info)                                                \
{                                                                                                                               \
   if (--(info->NaechstesFreies)>0){                                                                                            \
      info->subheap[0] = info->subheap[info->NaechstesFreies];                                                                  \
      SHLasseEinsinken_ ## WeightEntryType(info, 0);                                                                            \
   }                                                                                                                            \
}

#define SHKomplettReorganisieren(/*{0,1,2}*/ WeightEntryType, info)                                                             \
{                                                                                                                               \
   if (info->NaechstesFreies > 0){                                                                                              \
      unsigned long position = (info->NaechstesFreies)/2;                                                                       \
      while (position-->0)                                                                                                      \
         SHLasseEinsinken_ ## WeightEntryType(info, position);                                                                  \
   }                                                                                                                            \
}


#define SHVonTotenAufTopBefreien(/*{0,1,2}*/ WeightEntryType, info, changed, removed, /*{Stat, NoStat}*/ statisticsMode)        \
{ /* changed und removed enthalten am Ende des Makros gueltige Werte, werde aber zu Beginn nicht initialisiert. */              \
   SubHeapZellenT_ ## WeightEntryType *sh = info->subheap;                                                                      \
                                                                                                                                \
   while ((info->NaechstesFreies>0) && (!LebtNoch(sh[0]))){                                                                     \
      changed = TRUE;                                                                                                           \
      TPR_DESTROYTERMPAIR(sh[0].tp, info->re, sh[0].OtherParent->re);                                                           \
      DESTROYADDREF_ ## WeightEntryType (sh[0].addRef);                                                                         \
      IncAnzKPSpaeterWaisenmord_ ## statisticsMode();                                                                           \
      IncAnzKPEntfernt_ ## statisticsMode();                                                                                    \
      KPVonElternGeloescht(info, sh[0].OtherParent);                                                                            \
      SHTopElementLoeschenUndReorganisieren(WeightEntryType, info);                                                             \
   }                                                                                                                            \
   if (info->NaechstesFreies==0){                                                                                               \
      removed = TRUE;                                                                                                           \
      changed = FALSE;                                                                                                          \
      GGfStrahlLoeschen(WeightEntryType, info);                                                                                 \
   }                                                                                                                            \
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Operationen auf dem Top-Heap -------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

#define THC_Groesser_0(e1, e2) (e1.w1 > e2.w1)

#define THC_Groesser_1(e1, e2) ((e1.w1 > e2.w1) || ( (e1.w1 == e2.w1) && (e1.w2 > e2.w2)))

#define THC_Groesser_2(e1, e2) ((e1.w1 > e2.w1) || ( (e1.w1 == e2.w1) && (e1.w2 > e2.w2)) ||                                    \
    ((e1.w1 == e2.w1) && (e1.w2 == e2.w2) &&                                                                                    \
     (e1.info->LebtNoch && e2.info->LebtNoch && e1.OtherParent->LebtNoch && e2.OtherParent->LebtNoch &&                         \
      C_AdditionalComparison (e1.w1, e1.w2, e1.addRef, e1.info->re, e1.OtherParent->re, e1.tp,                                  \
                              e2.w1, e2.w2, e2.addRef, e2.info->re, e2.OtherParent->re, e2.tp))))
/* C_AdditionalComparison darf nur mit lebenden Eltern aufgerufen werden. */

#define THLasseEinsinkenBody(position, /*{0,1,2}*/ WeightEntryType)                                                             \
/* Einsinken, bis ein nicht-kleineres Gewicht angetroffen wird */                                                               \
{                                                                                                                               \
   BOOLEAN changed = FALSE, isHeap = FALSE;                                                                                     \
   unsigned long j=position*2+1;                                                                                                \
   TopEntryT_ ## WeightEntryType merker= TopHeap_ ## WeightEntryType[position];                                                 \
                                                                                                                                \
   while ((!isHeap) && (j<TopHeapNaechstesFreies)){                                                                             \
      if ((j+1<TopHeapNaechstesFreies) &&                                      /* Position j+1 noch gueltig? */                 \
          (THC_Groesser_ ## WeightEntryType (TopHeap_ ## WeightEntryType [j], TopHeap_ ## WeightEntryType[j+1])))               \
         j++;                                                 /* => j hat jetzt den Wert des leichteren Nachfolgers */          \
      if (THC_Groesser_ ## WeightEntryType (merker, TopHeap_ ## WeightEntryType[j])){                                           \
         TopHeap_ ## WeightEntryType[position]=TopHeap_ ## WeightEntryType[j];                                                  \
         changed = TRUE;                                                                                                        \
         position = j;                                                                                                          \
         j *=2;                                                                                                                 \
         j++;                                                                                                                   \
      }                                                                                                                         \
      else                                                                                                                      \
         isHeap = TRUE;                                                                                                         \
   }                                                                                                                            \
   if (changed)                                                                                                                 \
      TopHeap_ ## WeightEntryType[position] = merker;                                                                           \
   return changed;                                                                                                              \
}

static BOOLEAN THLasseEinsinken_0(unsigned long int position)
THLasseEinsinkenBody(position, 0)

static BOOLEAN THLasseEinsinken_1(unsigned long int position)
THLasseEinsinkenBody(position, 1)

static BOOLEAN THLasseEinsinken_2(unsigned long int position)
THLasseEinsinkenBody(position, 2)


#define THNeuenSubheapEinhaengenUndReorganisieren(/*{0,1,2}*/ WeightEntryType, INFO)                                            \
/* info ist der dem neuen subheap zugehoerige Info-Block */                                                                     \
{                                                                                                                               \
   unsigned long int position = TopHeapNaechstesFreies;                                                                         \
                                                                                                                                \
   if (TopHeapNaechstesFreies==TopHeapAktuelleGroesse){                                                                         \
      TopHeapAktuelleGroesse += 1000;                                                                                           \
      TopHeap_ ## WeightEntryType = our_realloc(TopHeap_ ## WeightEntryType,                                                    \
                                                TopHeapAktuelleGroesse*sizeof(TopEntryT_ ## WeightEntryType),                   \
                                                "HoH: Ausdehnung Top-Heap");                                                    \
   }                                                                                                                            \
   TopHeap_ ## WeightEntryType[TopHeapNaechstesFreies].info = INFO;                                                             \
   GewichteHochziehen_ ## WeightEntryType(TopHeap_ ## WeightEntryType[TopHeapNaechstesFreies], (INFO->subheap)[0]);             \
   TopHeapNaechstesFreies++;                                                                                                    \
                                                                                                                                \
   do                                                                                                                           \
      position = (position-1)/2;                                                                                                \
   while (THLasseEinsinken_ ## WeightEntryType(position) && (position>0));                                                      \
}

#define THTopElementLoeschenUndReorganisieren(/*{0,1,2}*/ WeightEntryType)                                                      \
{                                                                                                                               \
   if (--TopHeapNaechstesFreies>0){                                                                                             \
      TopHeap_ ## WeightEntryType[0] = TopHeap_ ## WeightEntryType[TopHeapNaechstesFreies];                                     \
                                                                                                                                \
      THLasseEinsinken_ ## WeightEntryType(0);                                                                                  \
   }                                                                                                                            \
}

#define THVonTotenAufTopBefreien(/*{0,1,2}*/ WeightEntryType)                                                                   \
{                                                                                                                               \
   while( (TopHeapNaechstesFreies>0) && (!TopHeap_ ## WeightEntryType[0].info->LebtNoch)){                                      \
      TopHeap_ ## WeightEntryType[0].info->InTopHeap = FALSE;                                                                   \
      GGfInfoBlockLoeschen(TopHeap_ ## WeightEntryType[0].info);                                                                \
      THTopElementLoeschenUndReorganisieren(WeightEntryType);                                                                   \
   }                                                                                                                            \
}

#define THKomplettReorganisieren(/*{0,1,2}*/ WeightEntryType)                                                                   \
{                                                                                                                               \
   unsigned long i = TopHeapNaechstesFreies/2;                                                                                  \
                                                                                                                                \
   while(i>0)                                                                                                                   \
      THLasseEinsinken_ ## WeightEntryType(--i);                                                                                \
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Testing ----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/
void HOHAusgeben(void);

#define HOH_TEST 0
/* Alternativen: 0 fuer aus, 1 fuer an */
#if HOH_TEST

static unsigned long int testNr = 0;
static BOOLEAN ok = TRUE;

#define HeapFehler(str, nr1, nr2) (ok = FALSE, printf("%ld: %s (%ld, %ld)\n", i, str, nr1, nr2))


#define TestKleiner_0(e1, e2) (e1.w1 < e2.w1)
#define TestKleiner_1(e1, e2) ((e1.w1 < e2.w1) || ( (e1.w1 == e2.w1) && (e1.w2 < e2.w2)))
#define TestKleiner_2(e1, e2) ((e1.w1 < e2.w1) || ( (e1.w1 == e2.w1) && (e1.w2 < e2.w2)))

#define GewichteIdentisch_0(e1, e2) (e1.w1 == e2.w1)
#define GewichteIdentisch_1(e1, e2) ((e1.w1 == e2.w1) && (e1.w2 == e2.w2))
#define GewichteIdentisch_2(e1, e2) ((e1.w1 == e2.w1) && (e1.w2 == e2.w2) && (e1.addRef == e2.addRef))

#define SHTest(/*{0,1,2}*/ WeightEntryType, info)                                                                               \
/* Vor.: Subheap existiert */                                                                                                   \
{                                                                                                                               \
   unsigned long int i;                                                                                                         \
   SubHeapZellenT_ ##  WeightEntryType *sh = info->subheap;                                                                     \
                                                                                                                                \
   for(i=0; i < info->NaechstesFreies; i++){                                                                                    \
      if ((2*i+1< info->NaechstesFreies) &&                                                                                     \
         TestKleiner_ ##  WeightEntryType (sh[2*i+1], sh [i]))                                                                  \
         HeapFehler("SH-Sortierung falsch (Vater, Sohn)",sh[i].w1, sh[2*i+1].w1);                                               \
                                                                                                                                \
      if ((2*i+2< info->NaechstesFreies) &&                                                                                     \
           TestKleiner_ ##  WeightEntryType (sh [2*i+2], sh [i]))                                                               \
         HeapFehler("SH-Sortierung falsch (Vater, Sohn)", sh[i].w1, sh[2*i+2].w1);                                              \
   }                                                                                                                            \
}



#define THTest(/*{0,1,2}*/ WeightEntryType)                                                                                     \
{                                                                                                                               \
   unsigned long i;                                                                                                             \
   for (i=0; i<TopHeapNaechstesFreies; i++){                                                                                    \
      TopEntryT_ ## WeightEntryType elt = TopHeap_ ## WeightEntryType[i];                                                       \
                                                                                                                                \
      if (elt.info->LebtNoch){                                                                                                  \
         if (elt.info->re){                                                                                                     \
            if (TP_SUEInfo(elt.info->re) != (void *) elt.info)                                                                  \
               HeapFehler("Info-Block !=  InfoBlock->RE->InfoBlock (Nr)", elt.info->ElternNummer, 0L);                          \
            if (RE_IstRegel(elt.info->re)){                                                                                     \
               if (elt.info->ElternNummer > RE_AnzRegelnGesamt())                                                               \
                  HeapFehler("Regel mit falscher Nummer(reNr, maxNr)", elt.info->ElternNummer, RE_AnzRegelnGesamt());           \
                                                                                                                                \
               if (elt.info->AnzNochNichtBehandelterKPs > 2*(AnzKPEingefuegt-AnzKPGeloescht))                                   \
                  HeapFehler("Regel mit zuvielen noch unbeh. KPs (nnbhKPs, max)",elt.info->AnzNochNichtBehandelterKPs,          \
                              2*(AnzKPEingefuegt-AnzKPGeloescht));                                                              \
            }                                                                                                                   \
            else{                                                                                                               \
               if (elt.info->ElternNummer < -RE_AnzGleichungenGesamt())                                                         \
                  HeapFehler("Gleichung mit falscher Nummer (Nr, maxNr)",elt.info->ElternNummer,  -RE_AnzGleichungenGesamt());  \
                                                                                                                                \
               if (elt.info->AnzNochNichtBehandelterKPs > 2*(AnzKPEingefuegt-AnzKPGeloescht))                                   \
                  HeapFehler("Gleichung mit zuvielen noch unbeh. KPs (nnbhKPs, max)",elt.info->AnzNochNichtBehandelterKPs,      \
                              2*(AnzKPEingefuegt-AnzKPGeloescht));                                                              \
                                                                                                                                \
            }                                                                                                                   \
         }                                                                                                                      \
         else                                                                                                                   \
            HeapFehler("Keine RE an lebendem Info-Block (Nr)", elt.info->ElternNummer, 0L);                                     \
         if (elt.info->subheap){                                                                                                \
            if (!GewichteIdentisch_ ## WeightEntryType(elt, elt.info->subheap[0]))                                              \
               HeapFehler("Gewichte nicht richtig hochgezogen (RE-Nr)", elt.info->ElternNummer, 0L);                            \
            SHTest(WeightEntryType, elt.info);                                                                                  \
         }                                                                                                                      \
      }                                                                                                                         \
      else{                                                                                                                     \
         if (elt.info->re)                                                                                                      \
            HeapFehler("RE an totem Info-Block (Nr)", elt.info->ElternNummer, 0L);                                              \
         if (elt.info->subheap)                                                                                                 \
            HeapFehler("Nicht genullter Strahl an totem Info-Block (Nr)", elt.info->ElternNummer, 0L);                          \
      }                                                                                                                         \
      if ((2*i+1< TopHeapNaechstesFreies) &&                                                                                    \
         TestKleiner_ ##  WeightEntryType (TopHeap_ ## WeightEntryType [2*i+1], TopHeap_ ## WeightEntryType [i]))               \
         HeapFehler("TH-Sortierung falsch (Vater,Sohn)",i,  2*i+1);                                                             \
                                                                                                                                \
      if ((2*i+2< TopHeapNaechstesFreies) &&                                                                                    \
           TestKleiner_ ##  WeightEntryType (TopHeap_ ## WeightEntryType [2*i+2], TopHeap_ ## WeightEntryType [i]))             \
         HeapFehler("TH-Sortierung falsch (Vater,Sohn)", i, 2*i+2);                                                             \
   }                                                                                                                            \
   SHTest(WeightEntryType, DynInfo_ ##  WeightEntryType);                                              \
   if (!ok) HOHAusgeben();                                                                                                      \
}




void HeapTest(char *str)
{
   printf("%s HEAPOfHEAPS-TEST %ld %s\n", ok?"++":"--", ++testNr, str);
   
   if (testNr == 50)
      printf("bla\n");
   switch (WeightEntryStrukturTyp){
   case 0:
      THTest(0);
      break;
   case 1:
      THTest(1);
      break;
   case 2:
      THTest(2);
      break;
   default:
      break;
   }

}

#else
#define HeapTest(str)
#endif

#define ElementGewichtAusgeben_0(ELT) printf(" w1 = %ld\n", (ELT).w1)
#define ElementGewichtAusgeben_1(ELT) printf(" w1 = %ld, w2 = %ld\n", (ELT).w1, (ELT).w2)
#define ElementGewichtAusgeben_2(ELT) printf(" w1 = %ld, w2 = %ld\n", (ELT).w1, (ELT).w2)


#define SHAusgeben(/*{0,1,2}*/ WeightEntryType, info)                                                                           \
/* Vor.: Subheap existiert */                                                                                                   \
{                                                                                                                               \
   unsigned long int i;                                                                                                         \
   SubHeapZellenT_ ##  WeightEntryType *sh = info->subheap;                                                                     \
                                                                                                                                \
   for(i=0; i < info->NaechstesFreies; i++){                                                                                    \
      printf("   OtherPar = %ld, %s, ", sh[i].OtherParent->ElternNummer, (sh[i].OtherParent->LebtNoch)?"*":"+");                \
      ElementGewichtAusgeben_ ## WeightEntryType(sh[i]);                                                                        \
   }                                                                                                                            \
}



#define THAusgeben(/*{0,1,2}*/ WeightEntryType)                                                                                 \
{                                                                                                                               \
   unsigned long i;                                                                                                             \
   for (i=0; i<TopHeapNaechstesFreies; i++){                                                                                    \
      TopEntryT_ ## WeightEntryType elt = TopHeap_ ## WeightEntryType[i];                                                       \
      if (elt.info->LebtNoch){                                                                                                  \
         if (RE_IstRegel(elt.info->re))                                                                                         \
            printf("TH[%3ld]: Regel Nr. %ld, %s, %lu init., %lu unbeh. KPs. NF = %lu.", i,                                      \
                   elt.info->ElternNummer, (elt.info->LebtNoch)?"*":"+",                                                        \
                   elt.info->AnzInitialerKPs,                                                                                   \
                   elt.info->AnzNochNichtBehandelterKPs,                                                                        \
                   elt.info->NaechstesFreies);                                                                                  \
         else                                                                                                                   \
            printf("TH[%3ld]: Gleichung Nr. %ld (%s), %s, %lu init., %lu unbeh. KPs. NF = %lu.",  i,                            \
                   elt.info->ElternNummer,                                                                                      \
                   TP_RichtungAusgezeichnet(elt.info->re)?"ausgez":"unausgez",                                                  \
                   (elt.info->LebtNoch)?"*":"+",                                                                                \
                   elt.info->AnzInitialerKPs,                                                                                   \
                   elt.info->AnzNochNichtBehandelterKPs,                                                                        \
                   elt.info->NaechstesFreies);                                                                                  \
         ElementGewichtAusgeben_ ## WeightEntryType(TopHeap_ ## WeightEntryType[i]);                                            \
         SHAusgeben(/*{0,1,2}*/ WeightEntryType, elt.info);                                                                     \
      }                                                                                                                         \
      else{                                                                                                                     \
         printf("TH[%3ld]: RE Nr. %ld, +, %lu init., %lu unbeh. KPs. NF = %lu.", i,elt.info->ElternNummer,                      \
                elt.info->AnzInitialerKPs, elt.info->AnzNochNichtBehandelterKPs,                                                \
                   elt.info->NaechstesFreies);                                                                                  \
         ElementGewichtAusgeben_ ## WeightEntryType(TopHeap_ ## WeightEntryType[i]);                                            \
      }                                                                                                                         \
   }                                                                                                                            \
   printf("TH[---]: RE Nr. 0, *, %lu init., %lu unbeh. KPs. NF = %lu. \n",                                                      \
          DynInfo_ ##  WeightEntryType -> AnzInitialerKPs,                                                                      \
          DynInfo_ ##  WeightEntryType -> AnzNochNichtBehandelterKPs,                                                           \
          DynInfo_ ##  WeightEntryType -> NaechstesFreies);                                                                     \
   SHAusgeben(WeightEntryType, DynInfo_ ##  WeightEntryType);                                                                   \
}
         
void HOHAusgeben(void)
{
   switch (WeightEntryStrukturTyp){
   case 0:
      THAusgeben(0);
      break;
   case 1:
      THAusgeben(1);
      break;
   case 2:
      THAusgeben(2);
      break;
   default:
      break;
   }
}

      
/* --------------------------------------------------------------------------------------------------------*/
/* ------- Insert -----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

#define INSERT_BODY(tp, wt, father,  mother, /*{0,1,2}*/ WeightEntries, /*{Stat, NoStat}*/ statisticsMode)                   \
{                                                                                                                            \
   SubHeapZellenT_ ## WeightEntries *sh;                                                                                     \
   InfoT_ ## WeightEntries *info = (InfoT_ ## WeightEntries *) TP_SUEInfo(father);                                           \
                                                                                                                             \
   HeapTest("INSERT Anfang");                                                                                                \
   /* ggf. dynHeap ausdehnen */                                                                                              \
   sh = info->subheap;                                                                                                       \
                                                                                                                             \
   sh[info->NaechstesFreies].tp = tp;                                                                                        \
   sh[info->NaechstesFreies].OtherParent = TP_SUEInfo(mother);                                                               \
   GewichteUebertragen_ ## WeightEntries (sh[info->NaechstesFreies], wt);                                                    \
                                                                                                                             \
   info->NaechstesFreies++;                                                                                                  \
                                                                                                                             \
   info->AnzInitialerKPs++;                                                                                                  \
   IncAnzKPEingefuegt_ ## statisticsMode();                                                                                  \
   info->AnzNochNichtBehandelterKPs++;                                                                                       \
   (((InfoT_ ## WeightEntries *) TP_SUEInfo(mother))->AnzNochNichtBehandelterKPs)++;                                         \
   HeapTest("INSERT Ende");                                                                                                  \
}

#define INSERTImmortal_BODY(tp, wt, father,  mother, /*{0,1,2}*/ WeightEntries, /*{Stat, NoStat}*/ statisticsMode)           \
{                                                                                                                            \
   SubHeapZellenT_ ## WeightEntries *sh;                                                                                     \
   InfoT_ ## WeightEntries *info = DynInfo_ ## WeightEntries;                                                                \
   unsigned int i = info->NaechstesFreies;                                                                                   \
                                                                                                                             \
   HeapTest("INSERTImmortal Anfang");                                                                                        \
   /* ggf. dynHeap ausdehnen */                                                                                              \
   if (info->NaechstesFreies == DynHeapAktuelleGroesse){                                                                     \
      DynHeapAktuelleGroesse *= 2;                                                                                           \
      info->subheap = our_realloc(info->subheap,                                                                             \
                                  DynHeapAktuelleGroesse*sizeof(SubHeapZellenT_ ## WeightEntries),                           \
                                  "HoH: Ausdehnung dyn. Heap");                                                              \
   }                                                                                                                         \
                                                                                                                             \
   sh = info->subheap;                                                                                                       \
                                                                                                                             \
   sh[info->NaechstesFreies].tp = tp;                                                                                        \
   sh[info->NaechstesFreies].OtherParent = TP_SUEInfo(mother);                                                               \
   GewichteUebertragen_ ## WeightEntries (sh[info->NaechstesFreies], wt);                                                    \
                                                                                                                             \
   info->NaechstesFreies++;                                                                                                  \
                                                                                                                             \
   info->AnzInitialerKPs++;                                                                                                  \
   IncAnzKPEingefuegt_ ## statisticsMode();                                                                                  \
                                                                                                                             \
   do                                                                                                                        \
      i = (i-1)/2;                                                                                                           \
   while(SHLasseEinsinken_ ## WeightEntries(info, i) && (i>0));                                                              \
                                                                                                                             \
   info->AnzNochNichtBehandelterKPs += 2;                                                                                    \
                                                                                                                             \
   HeapTest("INSERTImmortal Ende");                                                                                          \
}




static void INSERT_0_STATISTICS(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP, RegelOderGleichungsT oP)
INSERT_BODY(tp, wt, aP, oP, 0, Stat)

static void INSERT_0_NOSTAT(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP, RegelOderGleichungsT oP)
INSERT_BODY(tp, wt, aP, oP, 0, NoStat)


static void INSERT_1_STATISTICS(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP, RegelOderGleichungsT oP)
INSERT_BODY(tp, wt, aP, oP, 1, Stat)

static void INSERT_1_NOSTAT(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP, RegelOderGleichungsT oP)
INSERT_BODY(tp, wt, aP, oP, 1, NoStat)


static void INSERT_2_STATISTICS(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP, RegelOderGleichungsT oP)
INSERT_BODY(tp, wt, aP, oP, 2, Stat)

static void INSERT_2_NOSTAT(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP, RegelOderGleichungsT oP)
INSERT_BODY(tp, wt, aP, oP, 2, NoStat)


static void INSERTImmortal_0_STATISTICS(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP, RegelOderGleichungsT oP)
INSERTImmortal_BODY(tp, wt, aP, oP, 0, Stat)

static void INSERTImmortal_0_NOSTAT(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP, RegelOderGleichungsT oP)
INSERTImmortal_BODY(tp, wt, aP, oP, 0, NoStat)


static void INSERTImmortal_1_STATISTICS(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP, RegelOderGleichungsT oP)
INSERTImmortal_BODY(tp, wt, aP, oP, 1, Stat)

static void INSERTImmortal_1_NOSTAT(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP, RegelOderGleichungsT oP)
INSERTImmortal_BODY(tp, wt, aP, oP, 1, NoStat)


static void INSERTImmortal_2_STATISTICS(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP, RegelOderGleichungsT oP)
INSERTImmortal_BODY(tp, wt, aP, oP, 2, Stat)

static void INSERTImmortal_2_NOSTAT(TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP, RegelOderGleichungsT oP)
INSERTImmortal_BODY(tp, wt, aP, oP, 2, NoStat)



BOOLEAN HOH_POSTPONE (TermpairT tp, WeightEntryT wt, RegelOderGleichungsT aP, RegelOderGleichungsT oP)
{
   return FALSE;
}


/* --------------------------------------------------------------------------------------------------------*/
/* ------- Select -----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/


#define SELECT_BODY(/*{0,1,2}*/ WeightEntryType, tprPtr, actParPtr, oParPtr, addRefPtr,                                       \
                    /*{Stat, NoStat}*/ statisticsMode)                                                                                  \
/* Ggf. Subheap fuer zuletzt eingefuegte KPs anlegen und diese dort hineinkopieren. (Von hinten!)                                       \
   Dabei jedes Element Einsinken lassen.                                                                                                \
   Leichtesten Heap aus Top-Heap auswaehlen, dort Top-Element entnehmen. GGf. Top-Heap reorganisieren.                                  \
                                                                                                                                        \
   while (progress && THNaechstesFreies>0)                                                                                              \
     TH von toten auf Top befreien                                                                                                      \
     Top-SH von toten auf top befreien                                                                                                  \
     if(Top-SH removed)                                                                                                                 \
        THTopElementLoeschenUndReorganisieren                                                                                           \
     else if changed                                                                                                                    \
        THLasseEinsinken(0)                                                                                                             \
     else                                                                                                                               \
        progress = FALSE                                                                                                                \
 */                                                                                                                                     \
{                                                                                                                                       \
   BOOLEAN removed, changed, progress = TRUE;                                                                                           \
   SubHeapZellenT_ ## WeightEntryType top;                                                                                              \
                                                                                                                                        \
   HeapTest("SELECT Anfang");                                                                                                           \
                                                                                                                                        \
                                /* zuerst den Top-Heap von toten Sub-Heaps auf seinem Gipfel befreien */                                \
                                /* und den obersten Subheap von toten KPs auf seinem Gipfel befreien. */                                \
                                /* Dabei ggf. den Top-Heap reorganisieren und das Procedere wiederholen. */                             \
   while(progress && (TopHeapNaechstesFreies > 0)){                                                                                     \
                                                                                                                                        \
      THVonTotenAufTopBefreien(WeightEntryType);                                                                                        \
      if (TopHeapNaechstesFreies == 0)                                                                                                  \
         break;                                                                                                                         \
      removed=FALSE, changed=FALSE;                                                                                                     \
      SHVonTotenAufTopBefreien(WeightEntryType, TopHeap_ ## WeightEntryType[0].info, changed, removed, statisticsMode);                 \
      if (removed){                                                                                                                     \
         TopHeap_ ## WeightEntryType[0].info -> InTopHeap = FALSE;                                                                      \
         GGfInfoBlockLoeschen(TopHeap_ ## WeightEntryType[0].info);                                                                     \
         THTopElementLoeschenUndReorganisieren(WeightEntryType);                                                                        \
      }else if (changed){                                                                                                               \
         GewichteHochziehen_ ## WeightEntryType(TopHeap_ ## WeightEntryType [0],                                                        \
            (TopHeap_ ## WeightEntryType [0].info->subheap)[0]);                                                                        \
         THLasseEinsinken_ ## WeightEntryType(0);                                                                                       \
      }else                                                                                                                             \
         progress = FALSE;                                                                                                              \
   }                                                                                                                                    \
                                                                                                                                        \
   if ((TopHeapNaechstesFreies > 0) &&                                                                                                  \
       ((DynInfo_ ##  WeightEntryType->NaechstesFreies == 0) ||                                                                         \
        (C_Groesser_ ## WeightEntryType(DynInfo_ ## WeightEntryType,                                                                    \
                                        TopHeap_ ## WeightEntryType [0].info,                                                           \
                                        DynInfo_ ## WeightEntryType->subheap[0],                                                        \
                                        (TopHeap_ ## WeightEntryType [0].info->subheap)[0])))){ /* TopHeap auslesen */                  \
                                                                                                                                        \
      top = TopHeap_ ## WeightEntryType [0].info->subheap[0];                                                                           \
      *tprPtr = top.tp;                                                                                                                 \
      *actParPtr = (TopHeap_ ## WeightEntryType [0].info)->re;                                                                          \
      *oParPtr = top.OtherParent->re;                                                                                                   \
      ADDREF_BESETZEN_ ## WeightEntryType(addRefPtr, top.addRef);                                                                       \
                                                                                                                                        \
      KPVonElternGeloescht(TopHeap_ ## WeightEntryType [0].info,                                                                        \
                           TopHeap_ ## WeightEntryType [0].info->subheap[0].OtherParent);                                               \
                                                                                                                                        \
      SHTopElementLoeschenUndReorganisieren(WeightEntryType, TopHeap_ ## WeightEntryType [0].info);                                     \
                                                                                                                                        \
      if (TopHeap_ ## WeightEntryType [0].info->NaechstesFreies == 0){                                                                  \
         GGfStrahlLoeschen(WeightEntryType, TopHeap_ ## WeightEntryType [0].info);                                                      \
      }                                                                                                                                 \
      else{                                                                                                                             \
         GewichteHochziehen_ ## WeightEntryType(TopHeap_ ## WeightEntryType [0],                                                        \
                                                (TopHeap_ ## WeightEntryType [0].info->subheap)[0]);                                    \
         THLasseEinsinken_ ## WeightEntryType(0);                                                                                       \
      }                                                                                                                                 \
   }                                                                                                                                    \
   else if (DynInfo_ ##  WeightEntryType->NaechstesFreies > 0){                                                                         \
      top = DynInfo_ ##  WeightEntryType->subheap[0];                                                                                   \
      DynInfo_ ##  WeightEntryType->AnzNochNichtBehandelterKPs -=2;                                                                     \
      *tprPtr = top.tp;                                                                                                                 \
      *oParPtr = *actParPtr = &SUE_Demigod;                                                                                             \
      /* ? addref */                                                                                                                    \
      SHTopElementLoeschenUndReorganisieren( WeightEntryType, DynInfo_ ##  WeightEntryType);                                            \
   }                                                                                                                                    \
   else                                                                                                                                 \
      return FALSE;                                                                                                                     \
                                                                                                                                        \
   IncAnzKPEntfernt_ ## statisticsMode();                                                                                               \
                                                                                                                                        \
   HeapTest("SELECT Ende");                                                                                                             \
                                                                                                                                        \
   return TRUE;                                                                                                                         \
}   

static BOOLEAN SELECT_0_STATISTICS(TermpairT *tpPtr, RegelOderGleichungsT *actParPtr, RegelOderGleichungsT *oParPtr, 
                                   void **addrefPtr)
SELECT_BODY(0, tpPtr, actParPtr, oParPtr, addrefPtr, Stat)

static BOOLEAN SELECT_0_NOSTAT(TermpairT *tpPtr, RegelOderGleichungsT *actParPtr, RegelOderGleichungsT *oParPtr, 
                               void **addrefPtr)
SELECT_BODY(0, tpPtr, actParPtr, oParPtr, addrefPtr, NoStat)


static BOOLEAN SELECT_1_STATISTICS(TermpairT *tpPtr, RegelOderGleichungsT *actParPtr, RegelOderGleichungsT *oParPtr, 
                                   void **addrefPtr)
SELECT_BODY(1, tpPtr, actParPtr, oParPtr, addrefPtr, Stat)

static BOOLEAN SELECT_1_NOSTAT(TermpairT *tpPtr, RegelOderGleichungsT *actParPtr, RegelOderGleichungsT *oParPtr, 
                               void **addrefPtr)
SELECT_BODY(1, tpPtr, actParPtr, oParPtr, addrefPtr, NoStat)


static BOOLEAN SELECT_2_STATISTICS(TermpairT *tpPtr, RegelOderGleichungsT *actParPtr, RegelOderGleichungsT *oParPtr, 
                                   void **addrefPtr)
SELECT_BODY(2, tpPtr, actParPtr, oParPtr, addrefPtr, Stat)

static BOOLEAN SELECT_2_NOSTAT(TermpairT *tpPtr, RegelOderGleichungsT *actParPtr, RegelOderGleichungsT *oParPtr, 
                               void **addrefPtr)
SELECT_BODY(2, tpPtr, actParPtr, oParPtr, addrefPtr, NoStat)



/* --------------------------------------------------------------------------------------------------------*/
/* ------- Parents ----------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

#define NEWPARENT_BODY(/*{0,1,2}*/ WeightEntryType)                                                                             \
/* Info-Block anlegen, zugehoerigen Sub-Heap mit NULL initialisieren.                                                           \
   Bei Gleichungen zweiten Info-Block anlegen, aber ohne Sub-Heap!*/                                                            \
{                                                                                                                               \
                                                                                                                                \
   if (IsDemigod(re)){                                                                                                          \
      InfoT_ ## WeightEntryType *info = DynInfo_ ## WeightEntryType;                                                            \
      TP_SUEInfo(re) = (void *) DynInfo_ ## WeightEntryType;                                                                    \
      InfoBlockInitialisieren(info);                                                                                            \
      info -> ElternNummer = TP_ElternNr(re);                                                                                   \
      info -> re = re;                                                                                                          \
   }                                                                                                                            \
   else{                                                                                                                        \
      InfoT_ ## WeightEntryType *info;                                                                                          \
                                                                                                                                \
      SV_Alloc(TP_SUEInfo(re),InfoManager,void *);                                                                              \
      info = (InfoT_ ## WeightEntryType *) TP_SUEInfo(re);                                                                      \
                                                                                                                                \
      InfoBlockInitialisieren(info);                                                                                            \
      info -> ElternNummer = TP_ElternNr(re);                                                                                   \
      info -> re = re;                                                                                                          \
      info -> subheap = SM_NeuerStrahl(Domaene);                                                                                \
                                                                                                                                \
      if (!RE_IstRegel(re)){        /* also Gleichung */                                                                        \
         SV_Alloc(TP_SUEInfo(TP_Gegenrichtung(re)),InfoManager,void *);                                                         \
         info = (InfoT_ ## WeightEntryType *) TP_SUEInfo(TP_Gegenrichtung(re));                                                 \
         InfoBlockInitialisieren(info);                                                                                         \
         info -> ElternNummer = TP_ElternNr(re);                                                                                \
         info -> re = TP_Gegenrichtung(re);                                                                                     \
         info -> subheap = NULL;                                                                                                \
      }                                                                                                                         \
   }                                                                                                                            \
}

static void NEWPARENT_0(RegelOderGleichungsT re)
NEWPARENT_BODY(0)

static void NEWPARENT_1(RegelOderGleichungsT re)
NEWPARENT_BODY(1)

static void NEWPARENT_2(RegelOderGleichungsT re)
NEWPARENT_BODY(2)



#define SubheapLeeren(/*{0,1,2}*/ WeightEntryType, actualInfo, /*{Stat, NoStat}*/ statisticsMode)                               \
/* Loescht alle Eintraege im actualInfo anhaengenden Subheap von Index 0 bis NaechstesFreies-1. */                              \
{                                                                                                                               \
   unsigned long i;                                                                                                             \
   SubHeapZellenT_ ## WeightEntryType *sh = actualInfo->subheap;                                                                \
                                                                                                                                \
   for (i=0; i < actualInfo->NaechstesFreies; i++){                                                                             \
      IncAnzKPFrueherWaisenmord_ ## statisticsMode();                                                                           \
      IncAnzKPEntfernt_ ## statisticsMode();                                                                                    \
/*      IO_DruckeMeldungKPGeloeschtWegenElternTot(0L, actualInfo->ElternNummer, sh[i].OtherParent->ElternNummer); */                \
      DESTROYADDREF_ ## WeightEntryType(sh[i].addRef);                                                                          \
      TPR_DESTROYTERMPAIR(sh[i].tp, actualInfo->re, sh[i].OtherParent->re);                                                     \
      KPVonElternGeloescht(actualInfo, sh[i].OtherParent);                                                                      \
   }                                                                                                                            \
}   
   
#define KILLPARENT_BODY(/*{0,1,2}*/ WeightEntryType, re, /*{Stat, NoStat}*/ statisticsMode)                                     \
/* Gesamten Sub-Heap durchlaufen und jedes Element loeschen, Sub-Heap freigeben. */                                             \
{                                                                                                                               \
   InfoT_ ## WeightEntryType *info = (InfoT_ ## WeightEntryType *) TP_SUEInfo(re);                                              \
                                                                                                                                \
   HeapTest("KILLPARENT Anfang");                                                                                               \
   info -> LebtNoch = FALSE;                                                                                                    \
                                                                                                                                \
   IO_DruckeMeldungWaisenmordStart();                                                                                           \
   SubheapLeeren(WeightEntryType, info, statisticsMode);                                                                        \
   if (RE_IstRegel(re)){                                                                                                        \
      if (IsDemigod(re))                                                                                                   \
         our_dealloc(info->subheap);                                                                                            \
      else if (info->subheap)                                                                                                   \
         SM_StrahlLoeschen(info->subheap);                                                                                      \
      info -> subheap=NULL;                                                                                                     \
      info -> NaechstesFreies = 0;                                                                                              \
      info -> re = NULL;                                                                                                        \
      GGfInfoBlockLoeschen(info);                                                                                               \
   }                                                                                                                            \
   else{                        /* Gleichung */                                                                                 \
      InfoT_ ## WeightEntryType *infoGegenrichtung = (InfoT_ ## WeightEntryType *) TP_SUEInfo(TP_Gegenrichtung(re));            \
                                                                                                                                \
      infoGegenrichtung  -> LebtNoch = FALSE;                                                                                   \
      SubheapLeeren(WeightEntryType, infoGegenrichtung, statisticsMode);                                                        \
      if (info->subheap)                                                                                                        \
         SM_StrahlLoeschen(info->subheap);                                                                                      \
      info->subheap=NULL;                                                                                                       \
      info->NaechstesFreies = 0;                                                                                                \
      infoGegenrichtung->subheap=NULL;                                                                                          \
      infoGegenrichtung->NaechstesFreies = 0;                                                                                   \
      info -> re = NULL;                                                                                                        \
      infoGegenrichtung-> re = NULL;                                                                                            \
      GGfInfoBlockLoeschen(info);                                                                                               \
      GGfInfoBlockLoeschen(infoGegenrichtung);                                                                                  \
   }                                                                                                                            \
   IO_DruckeMeldungWaisenmordEnde();                                                                                            \
   HeapTest("KILLPARENT Ende");                                                                                                 \
}

static void KILLPARENT_0_STATISTICS(RegelOderGleichungsT re)
KILLPARENT_BODY(0, re, Stat)

static void KILLPARENT_0_NOSTAT(RegelOderGleichungsT re)
KILLPARENT_BODY(0, re, NoStat)

static void KILLPARENT_1_STATISTICS(RegelOderGleichungsT re)
KILLPARENT_BODY(1, re, Stat)

static void KILLPARENT_1_NOSTAT(RegelOderGleichungsT re)
KILLPARENT_BODY(1, re, NoStat)

static void KILLPARENT_2_STATISTICS(RegelOderGleichungsT re)
KILLPARENT_BODY(2, re, Stat)

static void KILLPARENT_2_NOSTAT(RegelOderGleichungsT re)
KILLPARENT_BODY(2, re, NoStat)


#define REDIRECTPARENT_BODY(WeightEntryType)                                                                                    \
/* Neuen Sub-Heap zur Gegenrichtung anlegen und fuellen. Es wird ein gemeinsames Array verwendet, das                           \
   dann auch gemeinsam freigegeben werden kann.                                                                                 \
   Die Gleichungsrichtung in SUE_ActualParent ist noch gueltig, wird aber sofort im Anschluss umgedreht.                        \
   Jede Gleichung wird 'umgedreht', bevor FinishParent aufgerufen wird!                                                         \
 */                                                                                                                             \
{                                                                                                                               \
   InfoT_ ## WeightEntryType *info = (InfoT_ ## WeightEntryType *) TP_SUEInfo(SUE_ActualParent);                                \
   InfoT_ ## WeightEntryType *infoGegenrichtung = (InfoT_ ## WeightEntryType *) TP_SUEInfo(TP_Gegenrichtung(SUE_ActualParent)); \
                                                                                                                                \
   infoGegenrichtung->subheap = &(info->subheap[info->NaechstesFreies]);                                                        \
}

static void REDIRECTPARENT_0(void)
REDIRECTPARENT_BODY(0)

static void REDIRECTPARENT_1(void)
REDIRECTPARENT_BODY(1)

static void REDIRECTPARENT_2(void)
REDIRECTPARENT_BODY(2)

#define FINISHPARENT_BODY(WeightEntryType, /*{Stat, NoStat}*/ statisticsMode)                                                   \
/* Zeigt an, wenn die KP-Bildung zum zuletzt angemeldeten Elternteil abgeschlossen ist.                                         \
   Wurden keine KPs zu diesem Elternteil eingefuegt, so muss der anhaengende Strahl sofort                                      \
   freigegeben werden. Info-Bloecke koennen hier noch nicht geloescht werden, weil die                                          \
   Elternregel auf jeden Fall noch lebt.                                                                                        \
*/                                                                                                                              \
{                                                                                                                               \
   HeapTest("FINISHPARENT Anfang");                                                                                             \
   if (RE_IstRegel(SUE_ActualParent)){                                                                                          \
      InfoT_ ## WeightEntryType *info = (InfoT_ ## WeightEntryType *) TP_SUEInfo(SUE_ActualParent);                             \
      if (info->NaechstesFreies>0){                                                                                             \
         IncSIKPR_ ## statisticsMode(info->NaechstesFreies);                                                                    \
         SM_StrahlGefrieren(info->subheap, &(info->subheap[info->NaechstesFreies]));                                            \
         info->InTopHeap = TRUE;                                                                                                \
         SHKomplettReorganisieren(WeightEntryType, info);                                                                       \
         THNeuenSubheapEinhaengenUndReorganisieren(WeightEntryType, info);                                                      \
      }                                                                                                                         \
      else{                                                                                                                     \
         IncAnzI0KPR_## statisticsMode();                                                                                       \
         SM_NichtGefrorenenStrahlLoeschen(info->subheap);                                                                       \
         info->subheap = NULL;                                                                                                  \
      }                                                                                                                         \
   }                                                                                                                            \
   else{  /* SUE_ActualParent enthaelt nach REDIRECT die Gegenrichtung der aktuellen Gleichung */                               \
      InfoT_ ## WeightEntryType *infoGegenrichtung = (InfoT_ ## WeightEntryType *) TP_SUEInfo(SUE_ActualParent),                \
      *info = (InfoT_ ## WeightEntryType *) TP_SUEInfo(TP_Gegenrichtung(SUE_ActualParent));                                     \
                                                                                                                                \
      if ((info->NaechstesFreies>0) || (infoGegenrichtung->NaechstesFreies>0)){                                                 \
                                                                                                                                \
         IncSIKPG_ ## statisticsMode(info->NaechstesFreies + infoGegenrichtung->NaechstesFreies);                               \
         SM_StrahlGefrieren(info->subheap, &(infoGegenrichtung->subheap[infoGegenrichtung->NaechstesFreies]));                  \
                                                                                                                                \
         if  (infoGegenrichtung->NaechstesFreies>0){                                                                            \
            infoGegenrichtung->InTopHeap = TRUE;                                                                                \
            SHKomplettReorganisieren(WeightEntryType, infoGegenrichtung);                                                       \
            THNeuenSubheapEinhaengenUndReorganisieren(WeightEntryType, infoGegenrichtung);                                      \
         }                                                                                                                      \
         if (info->NaechstesFreies>0){                                                                                          \
            info->InTopHeap = TRUE;                                                                                             \
            SHKomplettReorganisieren(WeightEntryType, info);                                                                    \
            THNeuenSubheapEinhaengenUndReorganisieren(WeightEntryType, info);                                                   \
         }                                                                                                                      \
      }                                                                                                                         \
      else{                                                                                                                     \
         IncAnzI0KPG_## statisticsMode();                                                                                       \
         SM_NichtGefrorenenStrahlLoeschen(info->subheap);                                                                       \
         info->subheap = NULL;                                                                                                  \
         infoGegenrichtung->subheap = NULL;                                                                                     \
      }                                                                                                                         \
   }                                                                                                                            \
   HeapTest("FINISHPARENT Ende");                                                                                               \
}

static void FINISHPARENT_0_STATISTICS(void)
FINISHPARENT_BODY(0, Stat)

static void FINISHPARENT_1_STATISTICS(void)
FINISHPARENT_BODY(1, Stat)

static void FINISHPARENT_2_STATISTICS(void)
FINISHPARENT_BODY(2, Stat)

static void FINISHPARENT_0_NOSTAT(void)
FINISHPARENT_BODY(0, NoStat)

static void FINISHPARENT_1_NOSTAT(void)
FINISHPARENT_BODY(1, NoStat)

static void FINISHPARENT_2_NOSTAT(void)
FINISHPARENT_BODY(2, NoStat)


/* --------------------------------------------------------------------------------------------------------*/
/* ------- Walkthrough ------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

unsigned long int HOH_AnzKPsToGo(void)
{
   return PQStatisticsAn?(AnzKPEingefuegt - AnzKPGeloescht) : 0;
}


#define SHWALKTHROUGH(WeightEntryType, statisticsMode, /*{terms, noterms} */ MitTermen, RESULT, INFO)                           \
/* AP auf jedes Element des Subheaps anwenden. Subheap am Ende ggf. reorganisieren,                                             \
   ggf. auch Strahl loeschen.                                                                                                   \
   RESULT ist:                                                                                                                  \
   0 - noop                                                                                                                     \
   1 - delete                                                                                                                   \
   2 - reweight                                                                                                                 \
*/                                                                                                                              \
{                                                                                                                               \
   unsigned long int i = 0;                                                                                                     \
   SubHeapZellenT_ ## WeightEntryType *sh = INFO->subheap;                                                                      \
   int CPresult;                                                                                                                \
   BOOLEAN SHchanged = FALSE;                                                                                                   \
                                                                                                                                \
   if (INFO->LebtNoch){                                                                                                         \
      while(i<INFO->NaechstesFreies){                                                                                           \
         IncAnzKPIRBetrachtet_ ## statisticsMode();                                                                             \
         if (LebtNoch(sh[i])){                                                                                                  \
            TermT lhs=NULL, rhs=NULL;                                                                                           \
            PCL_ObjektNummerT pclID;                                                                                            \
            /* Daten an Globale Variablen zuweisen */                                                                           \
                                                                                                                                \
            TermeAuspacken_ ## MitTermen(sh[i].tp, &lhs, &rhs, INFO->re, sh[i].OtherParent->re, &pclID);                        \
                                                                                                                                \
            if ((CPresult = (* AP) (lhs, rhs, INFO->re, sh[i].OtherParent->re,                                                  \
                                    &pclID)) == 1){                                               /* loeschen */                \
               TermeZerstoeren_ ## MitTermen(sh[i].tp, lhs, rhs, INFO->re,sh[i].OtherParent->re );                              \
               DESTROYADDREF_ ## WeightEntryType(sh[i].addRef);                                                                 \
               KPVonElternGeloescht(INFO, sh[i].OtherParent);                                                                   \
               sh[i] = sh[--(INFO->NaechstesFreies)];                                                                           \
               SHchanged = TRUE;                                                                                                \
               IncAnzKPEntfernt_ ## statisticsMode();                                                                           \
               IncAnzKPIRGeloescht_ ## statisticsMode();                                                                        \
            }                                                                                                                   \
            else if (CPresult == 2){                                                             /* neugewichten */             \
               WeightEntryT wt;                                                                                                 \
                                                                                                                                \
               IncAnzKPIRVeraendert_ ## statisticsMode();                                                                       \
               SHchanged = TRUE;                                                                                                \
               GewichteUebertragen_ ## WeightEntryType(wt, sh[i]);                                                              \
               C_RECLASSIFY(&wt, lhs, rhs, INFO->re,sh[i].OtherParent->re);                                                     \
               sh[i].tp = TermeWiederEinpacken_ ## MitTermen(sh[i].tp, lhs, rhs, INFO->re, sh[i].OtherParent->re, TRUE, pclID); \
               GewichteUebertragen_ ## WeightEntryType(sh[i], wt);                                                              \
               i++;                                                                                                             \
            }                                                                                                                   \
            else{                                                                                                               \
               sh[i].tp = TermeWiederEinpacken_ ## MitTermen(sh[i].tp, lhs, rhs, INFO->re, sh[i].OtherParent->re, FALSE, pclID);\
               i++;                                                                                                             \
            }                                                                                                                   \
         }                                                                                                                      \
         else{                                                                             /* KP ist bereits tot */             \
            TPR_DESTROYTERMPAIR(sh[i].tp, INFO->re, sh[i].OtherParent->re);                                                     \
            DESTROYADDREF_ ## WeightEntryType(sh[i].addRef);                                                                    \
                                                                                                                                \
            KPVonElternGeloescht(INFO, sh[i].OtherParent);                                                                      \
                                                                                                                                \
            IncAnzKPIRWaisenmord_ ## statisticsMode();                                                                          \
            IncAnzKPEntfernt_ ## statisticsMode();                                                                              \
                                                                                                                                \
            sh[i] = sh[--(INFO->NaechstesFreies)];                                                                              \
            SHchanged = TRUE;                                                                                                   \
         }                                                                                                                      \
      }                                                                                                                         \
      if ((INFO->NaechstesFreies == 0) && (!IsDemigod(INFO->re))){                                                      \
         GGfStrahlLoeschen(WeightEntryType, INFO);   /* Der Info-Block kann noch nicht gel. werden, da er noch im TH steht */   \
         RESULT = 1;            /* loeschen */                                                                                  \
      }                                                                                                                         \
      else if (SHchanged){                                                                                                      \
         SHKomplettReorganisieren(WeightEntryType, INFO);                                                                       \
         RESULT = 2;            /* neubewerten */                                                                               \
      }                                                                                                                         \
      else                                                                                                                      \
         RESULT = 0;            /* nix tun */                                                                                   \
   }                                                                                                                            \
   else                                                                                                                         \
      RESULT = 1;               /* schon tot: loeschen */                                                                       \
}
         

#define THWALKTHROUGHBODY(WeightEntryType, statisticsMode, /*{terms, noterms} */ MitTermen)                                     \
/* Doppelte Schleife ueber Top-Level-Heap und alle Sub-Heaps, dabei vorzeitigen Waisenmord ausfuehren. */                       \
{                                                                                                                               \
   unsigned long int thindex = 0;                                                                                               \
   int result;                                                                                                                  \
   BOOLEAN changed = FALSE;                                                                                                     \
                                                                                                                                \
   while(thindex<TopHeapNaechstesFreies){                                                                                       \
      if (TopHeap_ ## WeightEntryType[thindex].info->LebtNoch){                   /* tot oder nach Durchlauf leer */            \
         SHWALKTHROUGH(WeightEntryType, statisticsMode, MitTermen, result, TopHeap_ ## WeightEntryType[thindex].info);          \
      }                                                                                                                         \
      else                                                                                                                      \
         result = 1;            /* tot => loeschen */                                                                           \
                                                                                                                                \
      if (result == 1){                                                                                                         \
         TopHeap_ ## WeightEntryType[thindex].info->InTopHeap = FALSE;                                                          \
         GGfInfoBlockLoeschen(TopHeap_ ## WeightEntryType[thindex].info);                                                       \
         TopHeap_ ## WeightEntryType[thindex] = TopHeap_ ## WeightEntryType[--TopHeapNaechstesFreies];                          \
         changed = TRUE;                                                                                                        \
      }                                                                                                                         \
      else if (result == 2){                                                                                                    \
         GewichteHochziehen_ ## WeightEntryType(TopHeap_ ## WeightEntryType[thindex],                                           \
                                                TopHeap_ ## WeightEntryType[thindex].info->subheap[0]);                         \
         changed = TRUE;                                                                                                        \
         thindex++;                                                                                                             \
      }                                                                                                                         \
      else                                                                                                                      \
         thindex++;                                                                                                             \
   }                                                                                                                            \
   if (changed){                                                                                                                \
      THKomplettReorganisieren(WeightEntryType);                                                                                \
   }                                                                                                                            \
   /* SH-WT auf DynHeap!, anschliessend diesen ggf. Reorganisieren. */                                                          \
   SHWALKTHROUGH(WeightEntryType, statisticsMode, MitTermen, result, DynInfo_ ## WeightEntryType);                              \
}


#define WALKTHROUGHBODY(a,b)                                                                                                    \
{                                                                                                                               \
   if (NeedsTerms)                                                                                                              \
      THWALKTHROUGHBODY(a,b,terms)                                                                                              \
   else                                                                                                                         \
      THWALKTHROUGHBODY(a,b,noterms)                                                                                            \
}



static void WALKTHROUGH_0_STATISTICS (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
WALKTHROUGHBODY(0, Stat)

static void WALKTHROUGH_0_NOSTAT (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
WALKTHROUGHBODY(0, NoStat)

static void WALKTHROUGH_1_STATISTICS (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
WALKTHROUGHBODY(1, Stat)

static void WALKTHROUGH_1_NOSTAT (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
WALKTHROUGHBODY(1, NoStat)

static void WALKTHROUGH_2_STATISTICS (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
WALKTHROUGHBODY(2, Stat)

static void WALKTHROUGH_2_NOSTAT (PQ_WalkthroughProcedureT AP, BOOLEAN NeedsTerms)
WALKTHROUGHBODY(2, NoStat)





#define WW_CONDITION(element) (element.OtherParent->LebtNoch)



#define SH_WWBODY(/*{0,1,2}*/ WeightEntryType, INFO)                                                                         \
/* Vor.: Subheap existiert */                                                                                                \
{                                                                                                                            \
   unsigned long int j;                                                                                                      \
                                                                                                                             \
   for(j=0; j < INFO->NaechstesFreies; j++){                                                                                 \
      if (WW_CONDITION(INFO->subheap[j])){                                                                                   \
        WeightEntryT wt;                                                                                                     \
        GewichteUebertragen_ ## WeightEntryType(wt, INFO->subheap[j]);                                                       \
        (* AP)(wt, INFO->subheap[j].tp, INFO->re,                                                                            \
                   INFO->subheap[j].OtherParent->re);                                                                        \
      }                                                                                                                      \
   }                                                                                                                         \
}


#define TH_WWBODY(/*{0,1,2}*/ WeightEntryType)                                                                               \
{                                                                                                                            \
   unsigned long i;                                                                                                          \
   for (i=0; i<TopHeapNaechstesFreies; i++){                                                                                 \
      if (TopHeap_ ## WeightEntryType[i].info->LebtNoch){                                                                    \
         SH_WWBODY(/*{0,1,2}*/ WeightEntryType, TopHeap_ ## WeightEntryType[i].info);                                        \
      }                                                                                                                      \
   }                                                                                                                         \
   if (DynInfo_ ##  WeightEntryType -> NaechstesFreies>0)                                                                    \
      SH_WWBODY(WeightEntryType, DynInfo_ ##  WeightEntryType);                                                              \
}


void HOH_WATCHINGWALKTHROUGH(PQ_WatchingWalkthroughProcedureT AP)
/* s.o., aber ohne Waisenmord. */
{
   switch(WeightEntryStrukturTyp){
   case 1:
      TH_WWBODY(1);
      break;     
   case 2:       
      TH_WWBODY(2);
      break;     
   case 0:       
   default:      
      TH_WWBODY(0);
      break;
   }
}

#define SH_TVBODY(/*{0,1,2}*/ WeightEntryType, INFO)                                                                         \
/* Vor.: Subheap existiert */                                                                                                \
{                                                                                                                            \
   unsigned long int j;                                                                                                      \
                                                                                                                             \
   for(j=0; j < INFO->NaechstesFreies; j++)                                                                                  \
      TPR_OperationWaehrendTermverfolgung(OperationAufTermen, INFO->subheap[j].tp,                                           \
                                            INFO->re, INFO->subheap[j].OtherParent->re);                                     \
}


#define TH_TVBODY(/*{0,1,2}*/ WeightEntryType)                                                                               \
{                                                                                                                            \
   unsigned long i;                                                                                                          \
   for (i=0; i<TopHeapNaechstesFreies; i++){                                                                                 \
      if (TopHeap_ ## WeightEntryType[i].info->LebtNoch){                                                                    \
         SH_TVBODY(/*{0,1,2}*/ WeightEntryType, TopHeap_ ## WeightEntryType[i].info);                                        \
      }                                                                                                                      \
   }                                                                                                                         \
   if (DynInfo_ ##  WeightEntryType -> NaechstesFreies>0)                                                                    \
      SH_TVBODY(WeightEntryType, DynInfo_ ##  WeightEntryType);                                                              \
}


void HOH_TermVerfolgung(void (* OperationAufTermen) (TermT, TermT))
/* s.o., aber ohne Waisenmord. */
{
   switch(WeightEntryStrukturTyp){
   case 1:
      TH_TVBODY(1);
      break;
   case 2:
      TH_TVBODY(2);
      break;
   case 0:
   default:
      TH_TVBODY(0);
      break;
   }
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Statistics 2 -----------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

static unsigned long int Lebende;

static void LebendsKPsZaehlen(WeightEntryT wt, TermpairT tp, RegelOderGleichungsT actualP, RegelOderGleichungsT otherP)
{
   Lebende++;
}

void HOH_STATISTICS(InfoStringT is)
{
   if (!(IS_IsIn("all", is) || (IS_IsIn("PQ", is))))
      return;                   /* Modul nicht angesprochen */
   if (IS_IsIn("standard", is)){
      IO_DruckeKurzUeberschrift("Priority-Queue: Heap of heaps", "+");
      if (PQStatisticsAn){
         Lebende = 0;
         HOH_WATCHINGWALKTHROUGH(LebendsKPsZaehlen);
         PQStatistik(is, Lebende);
         fprintf(IO_Strom, "\n");
         IO_DruckeStatistikZeile("Regeln ohne initiale KPs", AnzRegelnMitInitial0KPs);
         IO_DruckeStatistikZeile("Regeln mit initialen KPs", AnzRegelnMitInitialenKPs);
         if (AnzRegelnMitInitialenKPs>0)
            IO_DruckeStatistikZeile("Durchschnittliche initiale KPs zu Regeln", SummeInitialerKPsZuRegeln/AnzRegelnMitInitialenKPs);
         IO_DruckeStatistikZeile("Gleichungen ohne initiale KPs", AnzGleichungenMitInitial0KPs);
         IO_DruckeStatistikZeile("Gleichungen mit initialen KPs", AnzGleichungenMitInitialenKPs);
         if (AnzGleichungenMitInitialenKPs>0)
            IO_DruckeStatistikZeile("Durchschnittliche initiale KPs zu Gleichungen", SummeInitialerKPsZuGleichungen/AnzGleichungenMitInitialenKPs);
      }
   }
   else if(IS_IsIn("latex-head", is)){
      if (PQStatisticsAn)
         PQStatistik(is, 0);
      fprintf(IO_Strom, "R: mit init & ohne init & ");
      if (AnzRegelnMitInitialenKPs>0)
         fprintf(IO_Strom, "avg init & ");
      fprintf(IO_Strom, "E: mit init & ohne init & ");
      if (AnzGleichungenMitInitialenKPs>0)
         fprintf(IO_Strom, "avg init & ");
   }
   else if(IS_IsIn("intern", is))
      IO_DruckeStatistikZeile("domain granularity", DomainGran);
   else if(IS_IsIn("config", is)){
      fprintf(IO_Strom, "HoH: Statistik %s, ", PQStatisticsAn?" AN":"AUS");
#if HOH_TEST
      fprintf(IO_Strom, "Heap-Test-Modus  ANn");
#else
      fprintf(IO_Strom, "Heap-Test-Modus AUS\n");
#endif
   }      
}

void HOH_ActSize(unsigned long int *AktKPZahl, unsigned long int *AktGueltigeKPZahl)
{
  Lebende = 0;
  HOH_WATCHINGWALKTHROUGH(LebendsKPsZaehlen);
  
  *AktGueltigeKPZahl = Lebende;
  *AktKPZahl = AnzKPEingefuegt - AnzKPGeloescht;
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Clean up ---------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/

#define CleanUpBody(WeightEntryType)                                                                                            \
{                                                                                                                               \
   unsigned long int i;                                                                                                         \
   for (i=0; i<TopHeapNaechstesFreies; i++){                                                                                    \
      if (TopHeap_ ## WeightEntryType[i].info->subheap){                                                                        \
         unsigned long j;                                                                                                       \
         SubHeapZellenT_ ## WeightEntryType *sh = TopHeap_ ## WeightEntryType[i].info->subheap;                                 \
                                                                                                                                \
         for (j=0; j < TopHeap_ ## WeightEntryType[i].info->NaechstesFreies; j++){                                              \
            DESTROYADDREF_ ## WeightEntryType(sh[j].addRef);                                                                    \
            TPR_DESTROYTERMPAIR(sh[j].tp, TopHeap_ ## WeightEntryType[i].info->re, sh[j].OtherParent->re);                      \
            KPVonElternGeloescht(TopHeap_ ## WeightEntryType[i].info, sh[j].OtherParent);                                       \
         }                                                                                                                      \
         TopHeap_ ## WeightEntryType[i].info->NaechstesFreies = 0;                                                              \
         GGfStrahlLoeschen(WeightEntryType, TopHeap_ ## WeightEntryType[i].info);                                               \
      }                                                                                                                         \
      TopHeap_ ## WeightEntryType[i].info->InTopHeap = FALSE;                                                                   \
      GGfInfoBlockLoeschen(TopHeap_ ## WeightEntryType[i].info);                                                                \
   }                                                                                                                            \
}

void HOH_Abraeumen(void)
/* Alles loeschen*/
{
   switch (WeightEntryStrukturTyp){
   case 0: CleanUpBody(0); break;
   case 1: CleanUpBody(1); break;
   case 2: CleanUpBody(2); break;
   default:
      break;
   }
   SM_StrahldomaeneLoeschen(Domaene);
}

void HOH_SoftAbraeumen(void)
     /* Nur die Nicht-SpeicherManager-Objekte freigeben, wiederaufsetzen-kompatibel. */
{
   switch (WeightEntryStrukturTyp){
   case 0: our_dealloc(DynInfo_0->subheap); break;
   case 1: our_dealloc(DynInfo_1->subheap); break;
   case 2: our_dealloc(DynInfo_2->subheap); break;
   default:
      break;
   }
  SM_StrahldomaeneLoeschen(Domaene);
}

/* --------------------------------------------------------------------------------------------------------*/
/* ------- Initialize -------------------------------------------------------------------------------------*/
/* --------------------------------------------------------------------------------------------------------*/


static BOOLEAN HOH_SchonInitialisiert_0 = FALSE;
static BOOLEAN HOH_SchonInitialisiert_1 = FALSE;
static BOOLEAN HOH_SchonInitialisiert_2 = FALSE;

#define HOH_SchonInitialisiert (HOH_SchonInitialisiert_0 || HOH_SchonInitialisiert_1 || HOH_SchonInitialisiert_2)

#define INIT_BODY(WeightEntryType, statisticsMode)                                                                            \
{                                                                                                                             \
   TopHeap_ ## WeightEntryType = our_alloc(TopHeapAktuelleGroesse * sizeof(TopEntryT_ ## WeightEntryType));                   \
   if (!HOH_SchonInitialisiert_##WeightEntryType){                                                                            \
     SV_ManagerInitialisieren(&InfoManager, sizeof(InfoT_ ## WeightEntryType),                                                \
                            PQ_AnzahlInfoBloeckeJeBlock, "KP-Infobloecke (HoH)");                                             \
   }                                                                                                                          \
   SV_Alloc(DynInfo_ ## WeightEntryType,InfoManager,void *);                                                                  \
   DynInfo_ ## WeightEntryType -> subheap = our_alloc(DynHeapAktuelleGroesse * sizeof(SubHeapZellenT_ ## WeightEntryType));   \
                                                                                                                              \
   HOH_INSERT = INSERT_ ## WeightEntryType ## _ ## statisticsMode;                                                            \
   HOH_INSERTImmortal = INSERTImmortal_ ## WeightEntryType ## _ ## statisticsMode;                                            \
   HOH_SELECT = SELECT_ ## WeightEntryType ## _ ## statisticsMode;                                                            \
   HOH_NEWPARENT = NEWPARENT_## WeightEntryType;                                                                              \
   HOH_KILLPARENT = KILLPARENT_## WeightEntryType ## _ ## statisticsMode;                                                     \
   HOH_REDIRECTPARENT = REDIRECTPARENT_ ## WeightEntryType;                                                                   \
   HOH_FINISHPARENT = FINISHPARENT_  ## WeightEntryType ## _ ## statisticsMode;                                               \
   HOH_WALKTHROUGH = WALKTHROUGH_  ## WeightEntryType ## _ ## statisticsMode;                                                 \
   HOH_SchonInitialisiert_##WeightEntryType = TRUE;                                                                           \
}

BOOLEAN HOH_INITIALIZE(InfoStringT DEF, InfoStringT USR, int WeightEntryType)
/* Info-BlockManager init.,  Funktionspointer in .h-File besetzen, Top-Level-Heap allokieren */
{
   char *statOption;

   if ((statOption = IS_ValueOfOption(DEF, USR, "statistics")))
      PQStatisticsAn = (strcmp(statOption, "on") == 0);
   else
      PQStatisticsAn = FALSE;
   
   if (PQStatisticsAn)
      PQStatistikInitialisieren();
   
   if (!HOH_SchonInitialisiert){
     if (IS_ValueOfOption(DEF, USR, "domgran"))
       DomainGran = atoi(IS_ValueOfOption(DEF, USR, "domgran"));
     else
       DomainGran = 10;
     
     /* dyn. Heap initialisieren */
     DynHeapAktuelleGroesse = 100;

     TopHeapAktuelleGroesse = PQ_AnzahlInfoBloeckeJeBlock;
   }

   Domaene = SM_NeueStrahldomaene(HOH_Domaenenstart, HOH_Domaenenende, DomainGran, "Domain for heap of heaps");

   WeightEntryStrukturTyp = WeightEntryType;

   /* Top-Heap initialisieren */
   TopHeapNaechstesFreies = 0;
   
   
   fprintf(IO_Strom, "Heap of Heaps. Statistics = %s. domgran = %ld.\n",
           PQStatisticsAn?"on":"off", DomainGran);

   switch (WeightEntryType){
   case 0:
      if (PQStatisticsAn)
         INIT_BODY(0, STATISTICS)
      else
         INIT_BODY(0, NOSTAT)
      break;
   case 1:
      if (PQStatisticsAn)
         INIT_BODY(1, STATISTICS)
      else
         INIT_BODY(1, NOSTAT)
      break;
   case 2:
      if (PQStatisticsAn)
         INIT_BODY(2, STATISTICS)
      else
         INIT_BODY(2, NOSTAT)
      break;
   default:
      return FALSE;
      break;
   }
   
   return TRUE;
}

