/* -------------------------------------------------------------------------------------------------------------*/
/* -                                                                                                            */
/* - PriorityQueue.c                                                                                            */
/* - ***************                                                                                            */
/* -                                                                                                            */
/* - Schnittstelle fuer die Priority Queue.                                                                     */
/* -                                                                                                            */
/* - 14.03.96 Arnim. Angelegt.                                                                                  */
/* -                                                                                                            */
/* -                                                                                                            */
/* -------------------------------------------------------------------------------------------------------------*/

#include "Ausgaben.h"
#include "TermOperationen.h"
#include "general.h"
#include "RUndEVerwaltung.h"
#include "InfoStrings.h"
#include "FehlerBehandlung.h"

#include "Classification.h"
#include "TermpairRepresentation.h"
#include "PriorityQueue.h"
#include "PQStatistik.h"

#include "KPHeap.h"
#include "KPPairing.h"
#include "HeapOfHeaps.h"

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe PriorityQueue.h) ---------------------*/
/* ------------------------------------------------------------------------*/

void    (* PQ_InsertProc)           (TermpairT , WeightEntryT , RegelOderGleichungsT , RegelOderGleichungsT);
void    (* PQ_InsertImmortalProc)   (TermpairT , WeightEntryT , RegelOderGleichungsT , RegelOderGleichungsT);
BOOLEAN (* PQ_PostponeProc)         (TermpairT , WeightEntryT , RegelOderGleichungsT , RegelOderGleichungsT);

void    (* PQ_WalkthroughProc) (PQ_WalkthroughProcedureT, BOOLEAN);
void    (* PQ_WatchingWalkthroughProc) (PQ_WatchingWalkthroughProcedureT);
void    (* PQ_TermVerfolgungsProc) (void (*) (TermT, TermT));
unsigned long int (* PQ_AnzKPsToGoProc) (void);

void    (* PQ_StatisticsProc)  (InfoStringT);
void   (* PQ_ActSizeProc) (unsigned long int *, unsigned long int *);

void (* PQ_CleanupProc) (void);
void (* PQ_SoftCleanupProc) (void);

BOOLEAN (* PQ_SelectInitialProc)      (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *, void **); 
BOOLEAN (* PQ_SelectProc)      (TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *, void **);
#if COMP_WMTW
BOOLEAN (*PQ_OriginalSelectProc)(TermpairT *, RegelOderGleichungsT *, RegelOderGleichungsT *, void **); 
#endif
void    (* PQ_NewParentProc)   (RegelOderGleichungsT);
void    (* PQ_KillParentProc)  (RegelOderGleichungsT);
void    (* PQ_RedirectParentProc) (void);
void    (* PQ_FinishParentProc) (void);

/* -------------------------------------------------------------------------------------------------------------*/
/* ------- Initialize ------------------------------------------------------------------------------------------*/
/* -------------------------------------------------------------------------------------------------------------*/


static void VoidDummy(void)
{
   ;
}

static BOOLEAN SelectInitialDummy(TermpairT *tp, RegelOderGleichungsT *v, RegelOderGleichungsT *m, void **a)
{
   our_error("Falsche Einstellung in einem Nicht-VV-Modus\n");
   return FALSE;
}

void PQ_ParamInfo(ParamInfoArtT what, InfoStringT DEF, InfoStringT USR)
{
   fprintf(IO_Strom, "\n* PRIORITY-QUEUE\n\n  ");
   fprintf(IO_Strom, "PQ_ParamInfo not yet implemented\n\n");
   
/*   switch (IS_FindOption(DEF, USR, 4, "heap", "pairing", "hoh", "HeapOfHeaps", "list")){
   case 1:
      KPP_ParamInfo(what, DEF, USR);
      break;
   case 2:
   case 3:
      HOH_ParamInfo(what, DEF, USR);
      break;
   case 0:
   default:
      KPH_ParamInfo(what, DEF, USR);
      break;
   }*/
   
}

BOOLEAN PQ_INITIALIZE(InfoStringT DEF, InfoStringT USR, int WeightEntryType)
{
   fprintf(IO_Strom, "\n* PRIORITY-QUEUE\n\n  ");
   switch (IS_FindOption(DEF, USR, 4, "heap", "pairing", "hoh", "HeapOfHeaps", "list")){
   case 1:
      KPP_INITIALIZE(DEF, USR, WeightEntryType);
      PQ_InsertProc         = KPP_INSERT;
      PQ_InsertImmortalProc = KPP_INSERTImmortal;
      PQ_PostponeProc       = KPP_POSTPONE;
      PQ_SelectProc         = KPP_SELECT;
      PQ_NewParentProc      = KPP_NEWPARENT;
      PQ_KillParentProc     = KPP_KILLPARENT;
      PQ_RedirectParentProc = KPP_REDIRECTPARENT;
      PQ_WalkthroughProc    = KPP_WALKTHROUGH;
      PQ_WatchingWalkthroughProc = KPP_WATCHINGWALKTHROUGH;
      PQ_StatisticsProc     = KPP_STATISTICS;
      PQ_CleanupProc        = KPP_Abraeumen;
      PQ_SoftCleanupProc    = KPP_SoftAbraeumen;
      PQ_FinishParentProc   = VoidDummy;
      PQ_TermVerfolgungsProc = KPP_TermVerfolgung;
      PQ_AnzKPsToGoProc     = KPP_AnzKPsToGo;
      PQ_SelectInitialProc  = KPP_SELECTINITIAL;
      PQ_ActSizeProc        = KPP_ActSize;
      return TRUE;
      break;
   case 2:
   case 3:
      HOH_INITIALIZE(DEF, USR, WeightEntryType);
      PQ_InsertProc         = HOH_INSERT;
      PQ_InsertImmortalProc = HOH_INSERTImmortal;
      PQ_PostponeProc       = HOH_POSTPONE;
      PQ_SelectProc         = HOH_SELECT;
      PQ_NewParentProc      = HOH_NEWPARENT;
      PQ_KillParentProc     = HOH_KILLPARENT;
      PQ_RedirectParentProc = HOH_REDIRECTPARENT;
      PQ_WalkthroughProc    = HOH_WALKTHROUGH;
      PQ_WatchingWalkthroughProc = HOH_WATCHINGWALKTHROUGH;
      PQ_StatisticsProc     = HOH_STATISTICS;
      PQ_CleanupProc        = HOH_Abraeumen;
      PQ_SoftCleanupProc    = HOH_SoftAbraeumen;
      PQ_FinishParentProc   = HOH_FINISHPARENT;
      PQ_TermVerfolgungsProc = HOH_TermVerfolgung;
      PQ_AnzKPsToGoProc     = HOH_AnzKPsToGo;
      PQ_SelectInitialProc  = SelectInitialDummy;
      PQ_ActSizeProc        = HOH_ActSize;
      return TRUE;
      break;
   case 0:
   default:
      /* Initialize PQ-Module with IS, NumIntsInWeightEntry, AdditionalComparisonProcedure */
      KPH_INITIALIZE(DEF, USR, WeightEntryType);
      PQ_InsertProc         = KPH_INSERT;
      PQ_InsertImmortalProc = KPH_INSERTImmortal;
      PQ_PostponeProc       = KPH_POSTPONE;
      PQ_SelectProc         = KPH_SELECT;
      PQ_NewParentProc      = KPH_NEWPARENT;
      PQ_KillParentProc     = KPH_KILLPARENT;
      PQ_RedirectParentProc = KPH_REDIRECTPARENT;
      PQ_WalkthroughProc    = KPH_WALKTHROUGH;
      PQ_WatchingWalkthroughProc = KPH_WATCHINGWALKTHROUGH;
      PQ_StatisticsProc     = KPH_STATISTICS;
      PQ_CleanupProc        = KPH_Abraeumen;
      PQ_SoftCleanupProc    = KPH_SoftAbraeumen;
      PQ_FinishParentProc   = VoidDummy;
      PQ_TermVerfolgungsProc = KPH_TermVerfolgung;
      PQ_AnzKPsToGoProc     = KPH_AnzKPsToGo;
      PQ_SelectInitialProc  = KPH_SELECTINITIAL;
      PQ_ActSizeProc        = KPH_ActSize;
      return TRUE;
      break;
   }
}
         
