#include "antiwarnings.h"
#include "Ausgaben.h"
#include "general.h"
#include "SpeicherVerwaltung.h"
#include "FileMenue.h"
#include <ctype.h>
#include <dirent.h>
#include <pwd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


/* ----------------------------------------------------------------------------------------------------- */
/* ------------------------ Typen und Variablen -------------------------------------------------------- */
/* ----------------------------------------------------------------------------------------------------- */
#define TStartSize 50
#define TInc 25
#define LStartSize 500
#define LInc 100

typedef struct ListenTypStruct{
   char *Name;
   unsigned int Nr;
   char *Pfad;
   BOOLEAN IsDir;
} ListenTyp;


static ListenTyp *L;                   /* Array vom Listentyp */
static char **T;                       /* Aufgedroeselter Filename */
static char **RootListe;               /* Liste der Such-Verzeichnisse: Home, ENVIRONMENT-VARIABLE etc. */
static char *aktMenuePfad;             /* Pfad, der zum aktuellen Menue-Verzeichnis fuehrt */
static unsigned int Lsize,             /* akt. Ausdehnung von L */
                    RootListSize,      /* Ausdehnung der Root-Liste */
                    Lpos,              /* letzter gueltiger Eintrag in L */
                    aktMenueTiefe;     /* aktuelle Absteigtiefe im Menue (0 bedeutet Root-Ebene) */
static unsigned int Tsize;

static char  *Hilfedatei, *Musterdatei; /* Namen (incl. Pfad) der beiden Dateien. */
static BOOLEAN  HilfedateiExistiert, MusterdateiExistiert;

static const char *FestInstallierteVerzeichnisseUser = NULL; /*"indexing";*/
static char *FestInstallierteVerzeichnisse = NULL; /*"Specifications";*/

static char           HomeDir[FILENAME_MAX]; /* HomeDir hat Slash am Ende */
      

/* Syntax: Pfad1:Pfad2:Pfad3:..:Pfadn oder NULL*/

/* ----------------------------------------------------------------------------------------------------- */
/* ------------------------ Hilfsprozeduren ------------------------------------------------------------ */
/* ----------------------------------------------------------------------------------------------------- */

static void RingBell(void)
{
   if (PA_InteraktionAn())
      fprintf(IO_Strom, "\a");
}


static BOOLEAN Bestaetigt(void)
/* Fragt auf j/n ab. */
{
  char key[INPUT_STRLEN];

  rewind(stdin);
  do{
     fprintf(IO_Strom, "OK? (y/n)\n>>> ");
     fgets(key, INPUT_STRLEN, stdin);
     fprintf(IO_Strom, "\n");
  }while(!((key[0] == 'n') || (key[0] == 'j') || (key[0] == 'y')));
  return ((key[0] == 'j') || (key[0] == 'y'));
}


static BOOLEAN DateiExistiert(char *fn)
/* Prueft, ob die angegebene Datei auf der Platte existiert.*/
{
  FILE *test;
  BOOLEAN existiert;

  test = fopen(fn, "r");
  existiert = (test!=NULL);
  if (existiert)
     fclose(test); /* darf nur aufgerufen werden, wenn test != NULL ist */
  return existiert;
}

static BOOLEAN passt(char *Kandidat, char *Name)
/* Makrofizieren */
{
   return (BOOLEAN) (strstr(Name, Kandidat) != NULL);
}

/* ----------------------------------------------------------------------------------------------------- */
/* ------------------------ Besetzen der Root-Liste ---------------------------------------------------- */
/* ----------------------------------------------------------------------------------------------------- */

static void RootListeBesetzen(void)
/* Baut eine Liste aller Verzeichnisse auf, in denen nach Beispielen zu suchen ist.
   Dazu gehoeren: 1) Ein hart eincodiertes Verzeichnis
                  2) Das lokale Verzeichnis (dh. ohne Praefix)
                  3) Die Verzeichnisse aus der ENVIRONMENT-Variablen WMPATHS
*/
{
   char *name = our_alloc(200*sizeof(char)), *env, *env2;
   int  i;
   RootListSize = 0;
   
   /* Anzahl der Pfade in FestInstallierteVerzeichnisse zaehlen */
   if (FestInstallierteVerzeichnisse != NULL){
      const char *Index = FestInstallierteVerzeichnisse;

      RootListSize=1;
      while((Index = strchr(Index, ':'))){
         ++RootListSize;
         ++Index;
      }
   }

   env = getenv("WM_PATHS");
   if (env && (env[0] == '\0')){
      fprintf(IO_Strom, "\n\n****Note: Search-Paths for Waldmeister can be set via\n");
      fprintf(IO_Strom, "          setenv WM_PATHS path1:path2:path3...\n\n");
      env = NULL;
   }

   /* Anzahl der Pfade in env zaehlen */
   if (env != NULL){
      char *Index = env;
      
      RootListSize++;
      while((Index = strchr(Index, ':'))){
         ++RootListSize;
         ++Index;
      }
   }

   /* Groesse der Root-Liste um ENVIRONMENT-unabhaengige Eintraege erhoehen */

   RootListSize += 1;

   /* Root-Liste allokieren */

   RootListe = our_alloc(RootListSize * sizeof(char *));

   /* ENVIRONMENT-unabhaengige Eintraege in RootListe schreiben, Offset vornehmen */
   
   env2 = getenv("PWD");
   if (env2 && (env2[0] == '\0'))
     env2 = NULL;

   i = 0;                       /* Index des ersten ENVIRONMENT-abhaengigen Eintrages */


   /* Eintraege aus FestInstallierteVerzeichnisse in RootListe schreiben */
   
   if (FestInstallierteVerzeichnisse != NULL){
      char *NaechsteStelle = NULL;
      char *FestInstallierteVerzeichnisseKopie = strdup(FestInstallierteVerzeichnisse);
      char *Index, *myHomeDir;
      DIR *dir;
      Index = FestInstallierteVerzeichnisseKopie;
      
      myHomeDir = our_alloc(sizeof(char) * (strlen(HomeDir)+strlen(FestInstallierteVerzeichnisseKopie)));

      do{                       /* es gibt min. einen Eintrag in FestInstallierteVerzeichnisse, da FestInstallierteVerzeichnisse!= NULL */
         Index = strchr(Index, ':'); /* naechsten Trenner suchen */
         if (Index){
            NaechsteStelle = Index+1; /* erstes Zeichen des naechsten Eintrages merken */
            *Index = '\0';           /* String-Ende simulieren */
            Index = NaechsteStelle;
         }
         strcpy (myHomeDir, HomeDir); /* HomeDir hat Slash am Ende */
         strcat (myHomeDir, FestInstallierteVerzeichnisseKopie);

         if ((dir = opendir(FestInstallierteVerzeichnisseKopie))){
            RootListe[i++] = strdup(FestInstallierteVerzeichnisseKopie);
            closedir(dir);
         }
	 else if ((dir = opendir(myHomeDir))) {
	    RootListe[i++] = strdup(myHomeDir);
	    closedir(dir);
	 }
         else{
            fprintf(IO_Strom, "%s is not a directory!\n", FestInstallierteVerzeichnisseKopie);
            RootListSize--;
         }
         FestInstallierteVerzeichnisseKopie = NaechsteStelle;
      }while(Index);
   }

   /* ENVIRONMENT-abhaengige Eintraege in RootListe schreiben */
   
   if (env != NULL){
      char *NaechsteStelle = NULL, *envKopie = strdup(env), *Index = envKopie;
      DIR *dir;

      do{                       /* es gibt min. einen Eintrag in env, da env!= NULL */
         Index = strchr(Index, ':'); /* naechsten Trenner suchen */
         if (Index){
            NaechsteStelle = Index+1; /* erstes Zeichen des naechsten Eintrages merken */
            *Index = '\0';           /* String-Ende simulieren */
            Index = NaechsteStelle;
         }
         if ((dir = opendir(envKopie))){
            RootListe[i++] = strdup(envKopie);
            closedir(dir);
         }
         else{
            fprintf(IO_Strom, "%s is not a directory!\n", envKopie);
            RootListSize--;
         }
         envKopie = NaechsteStelle;
      }while(Index);
   }
   RootListe[i] = (env2!=NULL) ? strdup(env2) : ".";
   our_dealloc(name);
}

/* ----------------------------------------------------------------------------------------------------- */
/* ------------------------ Filenamen-Expansion -------------------------------------------------------- */
/* ----------------------------------------------------------------------------------------------------- */

static void InhaltVonDirAnfuegen(unsigned int i, unsigned int Nr)
/* Inhalt des in L[i] bezeichneten Verzeichnisses ans Ende von L anhaengen. */
{
   char *neuerPfad = our_alloc(strlen(L[i].Pfad) + strlen(L[i].Name) + 2);
   DIR *dir;
   struct dirent *eintrag;
   
   strcpy(neuerPfad, L[i].Pfad);
   /* Kosmetik: Wenn Pfad bereits mit "/" aufhoert, dann wird nicht noch eins drangehaengt */
   if (neuerPfad == '\0' || neuerPfad[strlen(neuerPfad)-1] != '/')
      strcat(neuerPfad, "/");
   strcat(neuerPfad, L[i].Name);
   neuerPfad[strlen(L[i].Pfad) + strlen(L[i].Name) + 1] = '\0';
   if ((dir = opendir(neuerPfad))){
      while((eintrag = readdir(dir))){
         char *letzterPunkt = strrchr(eintrag->d_name, '.');
         
         if (letzterPunkt && ((strcmp(letzterPunkt, ".pr")==0) || (strcmp(letzterPunkt, ".rd")==0) ||
                              (strcmp(letzterPunkt, ".cp")==0) || (strcmp(letzterPunkt, ".cf")==0) ||
                              (strcmp(letzterPunkt, ".cv")==0) || (strcmp(letzterPunkt, ".tm")==0) ||
                              (strcmp(letzterPunkt, ".WMdir") == 0))){
            if (Lpos == Lsize){
               Lsize += LInc;
               L = realloc(L, Lsize * sizeof(ListenTyp));
            }
            L[Lpos].Nr = Nr;
            L[Lpos].IsDir = (strcmp(letzterPunkt, ".WMdir") == 0);
            L[Lpos].Name = strdup(eintrag->d_name);
            L[Lpos].Pfad = strdup(neuerPfad);
            Lpos++;
         }
      }
      closedir(dir);
   }
   our_dealloc(neuerPfad);
}


static char *ExpandiereFilenamen(char *fn)
/* Es wird versucht, fn in L oder den dort bezeichneten Verzeichnissen zu finden.
   L hat nach Beendigung dieselbe Gestalt wie vor, ggf. ist Lsize groesser.
   Freizugeben ist hernach nur Speicher bis incl. Lpos.
*/
{
   unsigned int i, j, Tpos = 1, AltLpos;
   char *ptr;
   char *Name = NULL;

   T[0] = fn;
   ptr = fn;

   /* Aufdroeseln des Filenamens */

   while ((ptr = strchr(ptr, '/'))){
      if (Tpos == Tsize){
         Tsize += TInc;
         T = realloc (T, Tsize * sizeof(char *));
      }
      *ptr = '\0';
      T[Tpos++] = ++ptr;
   }
   Tpos--;
   
   /* Suchen des Filenamens ab aktueller Menue-Ebene */
   fprintf(IO_Strom, "\nfile is `%s'\n", fn);
   i = 0;
   AltLpos = Lpos;
   while(i < Lpos){
      if (L[i].IsDir){
         if ( (L[i].Nr < Tpos) && passt(T[L[i].Nr], L[i].Name)) /* noch nicht beim Dateinamen angekommen */
            InhaltVonDirAnfuegen(i, L[i].Nr + 1);
         else 
            InhaltVonDirAnfuegen(i, 0);
         if ( (L[i].Nr == Tpos) && passt(T[0], L[i].Name))
            InhaltVonDirAnfuegen(i, 1);
      }
      else if ((L[i].Nr == Tpos) && passt(T[Tpos], L[i].Name)){
         fprintf(IO_Strom, "\nentering directory `%s'\n", L[i].Pfad);
         Name = our_alloc(strlen(L[i].Pfad) + strlen(L[i].Name) + 2); /* +2 wegen "/" und '/0' */
         strcpy(Name, L[i].Pfad);
         strcat(Name, "/");
         strcat(Name, L[i].Name);
         Name[strlen(L[i].Pfad) + strlen(L[i].Name) + 1] = '\0';
         break;
      }
      i++;
   }
   /* Freigeben des zum Suchen verwendeten Schwanzes von L. */
   for(j=AltLpos;j<Lpos;j++){
      /*
	fprintf(IO_Strom, "%s (%s) Nr %d IsDir %d\n", L[j].Name, L[j].Pfad, L[j].Nr, L[j].IsDir);
	fprintf(IO_Strom, "%lu %lu\n", (unsigned long int) L[j].Name, L[j].Pfad);
	*/
      our_dealloc(L[j].Name);
      our_dealloc(L[j].Pfad);
   }
   Lpos = AltLpos;
   return Name;
}
   

/* ----------------------------------------------------------------------------------------------------- */
/* ------------------------ Behandlung von L ----------------------------------------------------------- */
/* ----------------------------------------------------------------------------------------------------- */

static void EintraegeAusLFreigeben(void)
/* Loescht die Eintraege in L bis Lpos - 1.
   Der Eintrag Pfad darf nicht geloescht werden, da er nur ein Verweis auf die Variable
   aktMenuePfad ist und keine eigene Instanz des Pfadnamens.
*/
{
   while(Lpos > 0)
      our_dealloc(L[--Lpos].Name);
}

static void SortiereL(void)
/* Sortiert die Eintraege in L nach Directories und Dateien. */
{
   unsigned int vorne, hinten;
   
   if (Lpos == 0) return;
   vorne = 0; hinten = Lpos-1;
   while(vorne < hinten){
      while ((vorne < Lpos) && L[vorne].IsDir)
         vorne++;
      while ((hinten >0) && !L[hinten].IsDir)
         hinten--;
      if (vorne < hinten){
         char *name, *pfad;
         name = L[vorne].Name;
         pfad = L[vorne].Pfad;

         L[vorne].IsDir = TRUE;
         L[vorne].Name = L[hinten].Name;
         L[vorne].Pfad = L[hinten].Pfad;
         L[hinten].IsDir = FALSE;
         L[hinten].Name = name;
         L[hinten].Pfad = pfad;
      }
      vorne++;
      if (hinten>0) hinten--;
   }
}

static void EintraegeInLAuflisten(void)
/* Mit um 1 erhoehter Nummer! */
{
   unsigned int i, len;
   BOOLEAN VerzeichnisseFertig = FALSE;
   char *neuerPfad;

   if (((aktMenueTiefe > 0) && (RootListSize > 1)) || ((RootListSize == 1) && (aktMenueTiefe > 1))){
      fprintf(IO_Strom, "\nDirectory '%s':\n", aktMenuePfad);
      fprintf(IO_Strom, "*************");
      for(i=0;i<strlen(aktMenuePfad);i++)
         fprintf(IO_Strom, "*");
      fprintf(IO_Strom, "\n");
      len = strlen(aktMenuePfad) + 2 + strlen(".WMII.Kommentar");
      neuerPfad = our_alloc(len*sizeof(char));
      strcpy(neuerPfad, aktMenuePfad);
      strcat(neuerPfad, "/");
      strcat(neuerPfad, ".WMII.Kommentar");
      neuerPfad[len-1] = '\0';
      if (DateiExistiert(neuerPfad)){
         IO_DateiAusgabe(neuerPfad, FALSE);
         fprintf(IO_Strom, "------------------------------\n");
      }
   }
   else if ( (aktMenueTiefe == 1) && (RootListSize == 1)){
      fprintf(IO_Strom, "\nDirectory '%s':\n", aktMenuePfad);
      fprintf(IO_Strom, "*************");
      for(i=0;i<strlen(aktMenuePfad);i++)
         fprintf(IO_Strom, "*");
      fprintf(IO_Strom, "\n");
      len = strlen(aktMenuePfad) + 2 + strlen(".WMII.Kommentar");
      neuerPfad = our_alloc(len*sizeof(char));
      strcpy(neuerPfad, aktMenuePfad);
      strcat(neuerPfad, "/");
      strcat(neuerPfad, ".WMII.Kommentar");
      neuerPfad[len-1] = '\0';
      if (DateiExistiert(neuerPfad)){
         IO_DateiAusgabe(neuerPfad, FALSE);
         fprintf(IO_Strom, "------------------------------\n");
      }
   }     
   else{
      fprintf(IO_Strom, "\nDirectories");
      fprintf(IO_Strom, "\n***********\n");
   }
   if (Lpos>0)
      for(i=0; i<Lpos; i++){
         if (!VerzeichnisseFertig && !L[i].IsDir){ /* Leerzeile nach Ausgabe der Directories */
            fprintf(IO_Strom, "\n");
            VerzeichnisseFertig = TRUE;
         }
         if(strcmp(L[i].Name, "") == 0)
            fprintf(IO_Strom, "%5d. %s\n", i+1, L[i].Pfad);
         else
            fprintf(IO_Strom, "%5d. %s\n", i+1, L[i].Name);
      }
}

static void EintraegeAusRootListeInLSchreiben(void)
/* In Lpos steht dann die letzte besetzte Position + 1, also die naechste freie Position.
   VonAnzahlAbhaengig == TRUE: nur ein Eintrag -> dessen Inhalt wird uebertragen
                         FALSE oder mehrere Eintraege -> alles wird uebertragen*/
{
   if (RootListSize == 1){ /* Falls nur ein Verzeichnis angegeben -> Inhalt uebernehmen */
      DIR *dir;
      struct dirent *eintrag;

      aktMenueTiefe = 1;
      aktMenuePfad = strdup(RootListe[0]);  
      Lpos = 0;
      dir = opendir(RootListe[0]);
      while((eintrag = readdir(dir))){
         char *letzterPunkt = strrchr(eintrag->d_name, '.');
         
         if (letzterPunkt && ((strcmp(letzterPunkt, ".pr")==0) || (strcmp(letzterPunkt, ".rd")==0) ||
                              (strcmp(letzterPunkt, ".cp")==0) || (strcmp(letzterPunkt, ".cf")==0) ||
                              (strcmp(letzterPunkt, ".cv")==0) || (strcmp(letzterPunkt, ".tm")==0) ||
                              (strcmp(letzterPunkt, ".WMdir") == 0))){
            if (Lpos == Lsize){
               (Lsize) += LInc;
               L = realloc(L, Lsize * sizeof(ListenTyp));
            }
            L[Lpos].IsDir = (strcmp(letzterPunkt, ".WMdir") == 0);
            L[Lpos].Name = strdup(eintrag->d_name);
            L[Lpos].Pfad = strdup(RootListe[0]);
            Lpos++;
         }
      }
      SortiereL();
   }
   else{                       /* ... andernfalls die Verzeichnisse eintragen */
      aktMenueTiefe = 0;
      aktMenuePfad = strdup("");
      for (Lpos = 0; Lpos<RootListSize; Lpos++){
         if (Lpos == Lsize){
            Lsize += LInc;
            L = realloc(L, Lsize * sizeof(ListenTyp));
         }
         L[Lpos].IsDir = TRUE;
         L[Lpos].Name = strdup("");
         L[Lpos].Pfad = strdup(RootListe[Lpos]);
      }
   }
}


static void EintraegeAusMenuePfadNachLSchreiben(void)
/* Vor.: Eintraege in L sind geloescht.
   Schreibt den Inhalt des in AktMenuePfad bezeichneten Verzeichnisses in L.
*/
{
   DIR *dir;
   struct dirent *eintrag;

   if((dir = opendir(aktMenuePfad))){
      while((eintrag = readdir(dir))){
         char *letzterPunkt = strrchr(eintrag->d_name, '.');
         
         if (letzterPunkt && ((strcmp(letzterPunkt, ".pr")==0) || (strcmp(letzterPunkt, ".rd")==0) ||
                              (strcmp(letzterPunkt, ".cp")==0) || (strcmp(letzterPunkt, ".cf")==0) ||
                              (strcmp(letzterPunkt, ".cv")==0) || (strcmp(letzterPunkt, ".tm")==0) ||
                              (strcmp(letzterPunkt, ".WMdir") == 0))){
            if (Lpos == Lsize){
               Lsize += LInc;
               L = realloc(L, Lsize * sizeof(ListenTyp));
            }
            L[Lpos].Nr = 0;
            L[Lpos].IsDir = (strcmp(letzterPunkt, ".WMdir") == 0);
            L[Lpos].Name = strdup(eintrag->d_name);
            L[Lpos].Pfad = aktMenuePfad;
            Lpos++;
         }
      }
      closedir(dir);
   }
   SortiereL();
   /*      IO_WarteAufReturnMitText("Weiter");*/
}


/* ----------------------------------------------------------------------------------------------------- */
/* ------------------------ Menue-Prozeduren ----------------------------------------------------------- */
/* ----------------------------------------------------------------------------------------------------- */

static char *FM_FilenamenAuswaehlen(void)
/* Wurde beim Aufruf kein oder ein fehlerhafter Dateiname angegeben, so wird diese Prozedur angewendet.
   Dazu werden die Verzeichnisse der RootListe ausgegeben bzw. - falls nur eines existiert - dessen Inhalt.
   Befinden sich in einem solchen Verzeichnis wieder Verzeichnisse der Gestalt 'bla.WMdir', so koennen auch diese wieder
   angewaehlt und aufgelistet werden. Spezifikationen mit Endung .pr, .cp, .rd, .cf, .tm bzw. .cv  koennen angewaehlt und dann
   direkt gestartet werden.
   Das Anwaehlen erfolgt ueber den Namen bzw. einen Teilstring davon oder ueber die Nummer in der Auflistung.
   Ruecksprung auf Toplevel der Menue-Struktur erfolgt durch Eingabe von '..'.
   Ist die Root-Liste leer, so wird lediglich der Name einer existierenden Datei akzeptiert.
   
   Der Aufrufparameter verhindert ggf. (wenn FALSE), dass die angewaehlte Datei direkt gestartet werden kann.

1. Besetzten der Root-Liste
  a. Durchzaehlen der zu besetzenden Felder
  b. Liste allokieren
  c. Liste besetzen
2. if (anzVerzeichnisse == 1) then Inhalt des einzigen Verzeichnisses auflisten
   else Verzeichnisse auflisten
   (auflisten == am Bildschirm und in weiteres Array)
3. Abfrage nach Nummer bzw. Name bzw. '..'
4. if '..' then Aufstieg im Datei-Baum, Auflisten, goto 3
5. if Verzeichnis angefragt then dieses ausgeben
6. if Spezifikationsname gewaehlt, then diesen ansehen, starten, neuWaehlen (wie gehabt)
7. goto 3

*/
{
  char *eingname = our_alloc(INPUT_STRLEN * sizeof(char));
  unsigned int eingabe;
  BOOLEAN Gefunden = FALSE;
  char *Name = NULL;
  
  IO_DruckeUeberschrift("Select a specification");
  
  if (RootListSize == 0){
     do{
        fprintf(IO_Strom, "\nName of specification.\n>>> ");
        fgets(eingname, INPUT_STRLEN, stdin);
	chop(eingname);
        fprintf(IO_Strom, "\n");
     }while(!DateiExistiert(eingname));
     return eingname;
  }
  
  EintraegeAusRootListeInLSchreiben();
  EintraegeInLAuflisten();

  do{
     if (Lpos > 0){
        fprintf(IO_Strom, "\nNumber of desired specification/subdirectory (abbrev's accepted)");
        if ((aktMenueTiefe>1) || ((aktMenueTiefe ==1) && (RootListSize > 1)))
           fprintf(IO_Strom, " or\n'..' for higher level");
        fprintf(IO_Strom, "\n>>> ");
        fgets(eingname, INPUT_STRLEN, stdin);
	chop (eingname);
        fprintf(IO_Strom, "\n");
     }
     else{
        fprintf(IO_Strom, "\n... no WALDMEISTER-specifications/directories found.\n\n");
        IO_WarteAufReturnMitText("RETURN");
        if ((aktMenueTiefe == 1) && (RootListSize == 1)){
          fprintf(IO_Strom, "\nNo files found!\n\n");
          exit(10);
        }
        eingname = strcpy(eingname, "..");
     }
     if (strcmp(eingname, "..") == 0){
        if ((aktMenueTiefe>1) || ((aktMenueTiefe ==1) && (RootListSize > 1))){          
           /* eins hochsteigen, aber nur, wenn die akt. Tiefe > 1 ist oder sie gleich 1 ist und mehr als ein Verzeichnis
              in der Root-Liste steht.*/
           char *neuerPfad, *backslashpos;
           
           aktMenueTiefe--;
           if (aktMenueTiefe > 0){
              backslashpos = strrchr(aktMenuePfad, '/');
              if (backslashpos)
                 *backslashpos = '\0';
              neuerPfad = strdup (aktMenuePfad);
              if (backslashpos)
                 *backslashpos = '/';
              our_dealloc(aktMenuePfad);
              aktMenuePfad = neuerPfad;
              EintraegeAusLFreigeben();
              EintraegeAusMenuePfadNachLSchreiben();
           }
           else{
              EintraegeAusLFreigeben();
              EintraegeAusRootListeInLSchreiben();
           }
        }
        EintraegeInLAuflisten();
     }
     else{
        eingabe = atoi(eingname);
        /* beginnt die eingegebene Zeichenkette mit einem Buchstaben, so liefert atoi '0' zurueck.
           In diesem Falle wird die eingegebene Zeichenkette als Dateiname interpretiert.*/
        if((eingabe>0) && (eingabe <=Lpos)){
           eingabe--;
           Gefunden = TRUE;
        }
        else{                   /* keine gueltige Nummer => Eingabe als Abkuerzung interpretieren */
           unsigned int i;
           for(i=0; i<Lpos; i++)
              if (((strcmp(L[i].Name, "") == 0) && strstr(L[i].Pfad, eingname))
                  || (strstr(L[i].Name, eingname))){
                 eingabe = i;
                 Gefunden = TRUE;
                 break;
              }
        }
        if (Gefunden){
           if (L[eingabe].IsDir){          /* Directory gewaehlt */
              if ((aktMenueTiefe == 0) && (RootListSize > 1)){
                 our_dealloc(aktMenuePfad);
                 aktMenuePfad = strdup(L[eingabe].Pfad);
              }
              else{
                 char *neuerPfad;
                 neuerPfad = our_alloc(strlen(aktMenuePfad) + strlen(L[eingabe].Name) + 2);
                 strcpy(neuerPfad, aktMenuePfad);
                 strcat(neuerPfad, "/");
                 strcat(neuerPfad, L[eingabe].Name);
                 neuerPfad[strlen(aktMenuePfad) + strlen(L[eingabe].Name) + 1] = '\0';
                 our_dealloc(aktMenuePfad);
                 aktMenuePfad = neuerPfad;
              }
              aktMenueTiefe++;
              EintraegeAusLFreigeben();
              EintraegeAusMenuePfadNachLSchreiben();
              EintraegeInLAuflisten();
              Gefunden = FALSE;
           }
           else{
              Name = our_alloc(strlen(L[eingabe].Name) + strlen(aktMenuePfad)/*strlen(L[eingabe].Pfad)*/ + 2);
              strcpy(Name, /*L[eingabe].Pfad*/ aktMenuePfad);
              strcat(Name, "/");
              strcat(Name, L[eingabe].Name);
              Name[strlen(L[eingabe].Name) + /*strlen(L[eingabe].Pfad)*/ strlen(aktMenuePfad) + 1] = '\0';
              /* Behandlung des Namens */
           }
        }
        else{
           Name = ExpandiereFilenamen(eingname);
           Gefunden = (Name != NULL);
        }
     }
     if (Gefunden){
        char c[INPUT_STRLEN];
        fprintf(IO_Strom, "Your selection is '%s'.\n", Name);
        fprintf(IO_Strom, "_v_iew, _s_tart or new  _c_hoice ?\n>>> ");
        do{
           fgets(c, INPUT_STRLEN, stdin);
        }while(!((c[0]=='v') || (c[0]=='c') || (c[0]=='s')));
        if (c[0]=='v'){
           IO_DateiAusgabe(Name, TRUE);
           fprintf(IO_Strom, "\n\nIs this your desired specification?\n");
           if (!(Gefunden= Bestaetigt())){
              IO_DruckeUeberschrift("Select a specification");
              our_dealloc(Name);
              EintraegeInLAuflisten();
           }
        }
        else if (c[0]=='c'){
           Gefunden = FALSE;
           IO_DruckeUeberschrift("Select a specification");
           EintraegeInLAuflisten();
        }
     }
  }while(!Gefunden);
  
  our_dealloc(eingname);
  return Name;
}

/* ----------------------------------------------------------------------------------------------------- */
/* ------------------------ Exportierte Prozeduren ----------------------------------------------------- */
/* ----------------------------------------------------------------------------------------------------- */

void FM_InitApriori(char *aufruf)
{
   int PfadLaenge;
   char *BackslashPosition;
   struct passwd *HomePasswdPtr;

   Tsize = TStartSize;
   Lsize = LStartSize; Lpos = 0;
   T = our_alloc(Tsize * sizeof(char *));
   L = our_alloc(Lsize * sizeof(ListenTyp));
   /* RootListe wird lokal allokiert, wenn Groesse bekannt ist. */
   
#if COMP_STATIC
   strcpy (HomeDir, "/home/indexing/");
#else
   HomePasswdPtr= FestInstallierteVerzeichnisseUser ? getpwnam(FestInstallierteVerzeichnisseUser) : NULL;
   if (HomePasswdPtr != NULL) {
     strcpy (HomeDir, HomePasswdPtr->pw_dir);
     strcat (HomeDir, "/");
   }
#endif

   RootListeBesetzen();

   /* Hilfedatei suchen */

   BackslashPosition = strrchr(aufruf, '/');
   
   if (BackslashPosition){
      PfadLaenge = (BackslashPosition - aufruf);
      Hilfedatei = our_alloc((PfadLaenge + strlen("/WMII.Hilfe")+1)*sizeof(char));
      strncpy (Hilfedatei, aufruf, PfadLaenge);
      Hilfedatei[PfadLaenge] = '\0';
      strcat(Hilfedatei, "/WMII.Hilfe");
      Musterdatei = our_alloc((PfadLaenge + strlen("/WMII.Muster")+1)*sizeof(char));
      strncpy (Musterdatei, aufruf, PfadLaenge);
      Musterdatei[PfadLaenge] = '\0';
      strcat(Musterdatei, "/WMII.Muster");
   }
   else{
     Hilfedatei = our_alloc((strlen("WMII.Hilfe")+1)*sizeof(char));
     strcpy (Hilfedatei, "WMII.Hilfe");
     Musterdatei = our_alloc((strlen("WMII.Muster")+1)*sizeof(char));
     strcpy (Musterdatei, "WMII.Muster");
   }
   if (DateiExistiert(Hilfedatei))
     HilfedateiExistiert = TRUE;
   else{
     unsigned int i;
     our_dealloc(Hilfedatei);
     for (i=0; i<RootListSize; i++){
       Hilfedatei = our_alloc((strlen(RootListe[i]) + strlen("/WMII.Hilfe")+1)*sizeof(char));
       strncpy (Hilfedatei,RootListe[i], strlen(RootListe[i]));
       Hilfedatei[strlen(RootListe[i])] = '\0';
       strcat(Hilfedatei,  "/WMII.Hilfe");
       if (DateiExistiert(Hilfedatei)){
         HilfedateiExistiert = TRUE;
         break;
       }
       else
         our_dealloc(Hilfedatei);
     }
   }
   if (DateiExistiert(Musterdatei))
     MusterdateiExistiert = TRUE;
   else{
     unsigned int i;
     our_dealloc(Musterdatei);
     for (i=0; i<RootListSize; i++){
       Musterdatei = our_alloc((strlen(RootListe[i]) + strlen("/WMII.Muster")+1)*sizeof(char));
       strncpy (Musterdatei,RootListe[i], strlen(RootListe[i]));
       Musterdatei[strlen(RootListe[i])] = '\0';
       strcat(Musterdatei,  "/WMII.Muster");
       if (DateiExistiert(Musterdatei)){
         MusterdateiExistiert = TRUE;
         break;
       }
       else
         our_dealloc(Musterdatei);
     }
   }
}

char *FM_HilfeDatei(void)
/* Gibt den absoluten Dateinamen der Hilfedatei zurueck, falls diese existiert, sonst NULL. */
{
   if (HilfedateiExistiert)
      return Hilfedatei;
   else
      return NULL;
}

char *FM_MusterDatei(void)
/* Gibt den absoluten Dateinamen der Musterdatei zurueck, falls diese existiert, sonst NULL. */
{
   if (MusterdateiExistiert)
      return Musterdatei;
   else
      return NULL;
}

char *FM_SpezifikationErkannt(int argc, char *argv[], BOOLEAN MenueErlaubt)
/* String fn zu komplettem Dateinamen machen.
   1. Als Datei suchen
   2. In Beispielverzeichnissen suchen
   3. Durch Menue eingeben lassen, falls MenueErlaubt.
   Rueckgabewert ist der komplette Pfad oder NULL, falls Suche erfolglos war und Menue unerwunescht.
*/
{
   char *ExpFN, *fn = NULL;
   int i;

   for (i=1; i<argc; i++){
      char *letzterPunkt;

      if (argv[i] == NULL)
         continue;              /* Argument bereits verbraucht */

      letzterPunkt = strrchr(argv[i], '.');
      
      if (letzterPunkt && ((strcmp(letzterPunkt, ".pr")==0) || (strcmp(letzterPunkt, ".rd")==0) ||
                           (strcmp(letzterPunkt, ".cp")==0) || (strcmp(letzterPunkt, ".cf")==0) ||
                           (strcmp(letzterPunkt, ".cv")==0) || (strcmp(letzterPunkt, ".tm")==0))){
         fn = argv[i];
         argv[i] = NULL;
         break;
      }
      
   }

   if (fn != NULL){
      if (DateiExistiert(fn)) return fn;
      else{
         EintraegeAusRootListeInLSchreiben();
         if ((ExpFN = ExpandiereFilenamen(fn))) return ExpFN;
         else if (MenueErlaubt){
            RingBell();
            fprintf(IO_Strom, "\nFile %s not found.\n", fn);
            return FM_FilenamenAuswaehlen();
         }
         else return NULL;
      }
   }
   else if (MenueErlaubt) return FM_FilenamenAuswaehlen();
   else return NULL;
}

/* ----------------------------------------------------------------------------------------------------- */
/* ------------------------ Optionsdateien Parsen ------------------------------------------------------ */
/* ----------------------------------------------------------------------------------------------------- */

typedef struct OptlistTS{
  char *string;
  struct OptlistTS *next;
} OptlistT;

#define MAX_OPTFILES 10

static char *OptfileSuchen(char *optfilename)
     /* Aus optfilename einen gueltigen Dateinamen incl. Pfad machen, falls existiert, oder
        NULL zurueckgeben, falls Datei nicht gefunden.*/
{
  if (DateiExistiert(optfilename))
    return optfilename;
  else{
    char *withpath;
    withpath = our_alloc(sizeof(char)*(strlen(HomeDir)+strlen("wm_options/")+strlen(optfilename)));
    strcpy (withpath, HomeDir);
    strcat (withpath, "wm_options/");
    strcat (withpath, optfilename);
    if (DateiExistiert(withpath))
      return withpath;
    else
      our_dealloc(withpath);
  }
  return NULL;
}


static BOOLEAN OptlisteAusDateiLesen(OptlistT *Anfang)
     /*     1) Optfile suchen
               a) im aktuellen Verzeichnis
               b) im Verzeichnis ~/WM_OptionFiles
            2) Nicht gefunden: FALSE zurueckgeben
            3) Sonst: Eintraege im Optfile durchgehen und in Listenzellen kopieren.
            3) Anfang und Ende besetzen, TRUE zurueckgeben
        Ende zeigt nach dieser Prozedur auf eine leere Listen-Zelle.
      */
            
{
  FILE *OPT;
  char c, *optfileInclPfad;
  int  pos=0;
  BOOLEAN wasspace=TRUE;
  
  fprintf(IO_Strom, "%s", Anfang->string);

  if (!(optfileInclPfad = OptfileSuchen(Anfang->string))){
    fprintf(IO_Strom, " (not found!),");
    return FALSE;
  }
  
  if (!(OPT = fopen(optfileInclPfad, "r"))){
    fprintf(IO_Strom, " (found, but couldn't open),");
    return FALSE;
  }
  
  fprintf(IO_Strom, ", ");
  while ((c = fgetc(OPT))!= EOF){
    if (wasspace && !isspace(c)){
      long position;
      int anzZeichen = 1;
      OptlistT *haken;
      
      ungetc(c, OPT);           /* Zeichen zaehlen */
      position = ftell(OPT);
      while (((c = fgetc(OPT)) !=EOF) && (! isspace(c)))
        anzZeichen++;

      haken = our_alloc(sizeof(OptlistT));
      haken -> next = Anfang->next;
      Anfang -> next = haken;
      Anfang = haken;

      Anfang->string = our_alloc((anzZeichen+2)*sizeof(char));
      fseek(OPT, position, SEEK_SET); /* zuruecksetzen und Zeichen holen */
      fgets(Anfang->string, anzZeichen, OPT);

      pos++;
      wasspace = FALSE;
    }
    else if (!wasspace  && isspace(c)){
      wasspace = TRUE;
    }
  }

  fclose(OPT);

  return TRUE;
}
  

  
char ** FM_NeueParameterliste(int *altargc, char **altargv)
/* Ergaenzt die Parameterliste in altargv durch den Inhalt von <optfilename>.
   Ferner werden die bereits verbrauchten Positionen in atlargv aufgefuellt.
   Die Suche nach der Optionsdatei erfolgt
   1) Im aktuellen Verzeichnis
   2) Im einem fest eincodierten Verzeichnis
*/
{
  /* Ablauf:
     1) Argumente in verkettete Liste stecken
     2) Dabei "-optfile" suchen und
     3) jedes Vorkommen von -optfile "Optfile" durch die dort spezifizierten Argumente ersetzen
     4) Maximal so oft wie vorgegebene Konstante
     5) Neues argv nach Laenge der Liste anlegen, argc und argv besetzen und zurueckgeben
 */
  char **argv;
  OptlistT *Optionen = our_alloc(sizeof(OptlistT)), *runny, *haken; 
  int neuargc=0,                /* neuargc=0, da immer eine leere, aber angelegte Listenzelle am Ende steht */
    i, parsedOptfiles = 0;
  BOOLEAN IGNORE_PRINTED = FALSE;
  
  runny = Optionen;
  
  for (i=0; i<*altargc; i++){
    runny->string = altargv[i];
    runny->next = our_alloc(sizeof(OptlistT));
    runny = runny->next;
    runny->next = NULL;
  }
  
  for (runny=Optionen; ((runny->next) && (runny->next->next)); runny=runny->next)
    if ((strcmp(runny->string,"-optfile")==0)){ /* OptFile als Argument */
      if (parsedOptfiles ==0)
        fprintf(IO_Strom, "\nparsing additional options from ");
      runny = runny->next;
      if (parsedOptfiles > MAX_OPTFILES){ /* zu viele Optfiles */
        if (!IGNORE_PRINTED){
          fprintf(IO_Strom, "\n  ... too many Optfiles. Ignoring further ones.");
          IGNORE_PRINTED = TRUE;
        }
      }
      else{
        if (OptlisteAusDateiLesen(runny)){ /* runny steht auf dem File-Namen */
          parsedOptfiles++;
        }
      }
      runny -> string = "-optfile";
    }
  
  for (runny = Optionen; runny->next; runny=runny->next)
    if (strcmp(runny->string, "-optfile")!=0)
      neuargc++;
  
  argv = our_alloc(neuargc * sizeof(char *));
  
  for (runny = Optionen,i=0; runny->next; runny=runny->next,i)
    if (strcmp(runny->string, "-optfile")!=0)
      argv[i++] = runny->string;
  
  for (runny = haken = Optionen; haken; runny=haken){
    haken = runny->next;
    our_dealloc(runny);
  }
  
  if (parsedOptfiles>0){
    fprintf(IO_Strom, "\n  extended command line is: ");
    for (i=0; i<neuargc; i++)
      fprintf(IO_Strom, "%s ", argv[i]);
    fprintf(IO_Strom, "\n...done\n\n");
  }
    
  *altargc = neuargc;
  
  return argv;
}
