/*
Dieses Modul dient dem Initialisieren aller Strukturen nach dem Einlesen einer Spezifikation.

28.02.94 Angelegt.
03.03.94 Auf TermOperationen erweitert.

*/

/* -------------------------------------------------------*/
/* ------------- Includes --------------------------------*/
/* -------------------------------------------------------*/

#include "Ausgaben.h"
#include "DomaenenMemorium.h"
#include "DSBaumKnoten.h"
#include "DSBaumOperationen.h"
#include "DSBaumTest.h"
#include "FehlerBehandlung.h"
#include "Hauptkomponenten.h"
#include "Interreduktion.h"
#include "KBO.h"
#include "KPBildung.h"
#include "KPMengenInterreduktion.h"
#include "LPOVortests.h"
#include "MatchOperationen.h"
#include "MNF.h"
#include "NFBildung.h"
#include "Ordnungsgenerator.h"
#include "Parameter.h"
#include "Puffer.h"
#include "RUndEVerwaltung.h"
#include "Subgleichungen.h"
#include "SUEM.h"
#include "SymbolOperationen.h"
#include "TermOperationen.h"
#include "TermVerfolgung.h"
#include "Termpaare.h"
#include "Unifikation1.h"
#include "VariablenMengen.h"
#include "ZeitMessung.h"
#include "Zielverwaltung.h"
#include "general.h"
#include "parser.h"
#include "pcl.h"
#include "Initialisierungen.h"

#include <stdlib.h>

/* -------------------------------------------------------*/
/* ------------- Funktionen ------------------------------*/
/* -------------------------------------------------------*/

void IN_SpezifikationEinlesen(char *filename)
/* Nur Spezifikation einlesen, damit extModus bekannt wird. */
{
#if !COMP_ACIDWM
   fprintf(IO_Strom, "\nparsing %s", filename);
#endif
   parse_spec(filename);
   fprintf(IO_Strom, "... done\n");
}
   

void IN_AllesAprioriInitialisieren(void)
/* Alles, was VOR Einlesen der Spezifikation passieren mu\3 oder kann. Die Aufrufparameter sind aber noch nicht eingestellt! */
{
   printf("\napriori-initializing");
   IO_InitApriori();  /* IO_Strom initialisieren */
   PA_InitApriori();  /* Parameter definieren */
   DM_InitApriori();  /* Signalbehandlung umbiegen fuer Domaenenverwaltung */
   SV_InitApriori();  /* Falls SV-Debug-Betrieb: Aufruf malloc_debug(2) */
   ZM_InitApriori();  /* Zeit-Messung initialisieren: Freispeicherverwaltung. */
   BK_InitApriori();  /* Speichermanager fuer innere und Blattknoten und Baumverwaltungseintraege */
   BO_InitApriori();  /* Variablenbank; Speichermanager fuer gerichtet normierte Termpaare */
   RE_InitApriori();  /* Variablenumbenennungsbank dynamisch erzeugen; NaechsteBaumvariable auf x1 setzen */
   MO_InitApriori();  /* Variablenbank fuer Matching dynamisch erzeugen */
   VM_InitApriori();  /* Speichermanager fuer Variablenmengen */
   U1_InitApriori();  /* Bindungsstack, Substitutions-Variablenbank, TermIteration */
   TP_InitApriori();  /* Termpaar-Manager initialisieren. */
   TV_InitApriori();  /* Termiteration initialisieren. */
   BT_InitApriori();  /* Stapel in DSBaumTest initialisieren. */
   SUE_InitApriori();
   fprintf(IO_Strom, "... done\n");
}

void IN_ReAllesAprioriInitialisieren(void)
/* Alles, was VOR Einlesen der Spezifikation passieren mu\3 oder kann. Die Aufrufparameter sind aber noch nicht eingestellt! */
{
   printf("\nre-apriori-initializing");
   PA_ReInitApriori();  /* Parameter definieren */
   /*SV_InitApriori();  nicht noetig */
   ZM_ReInitApriori();  /* Zeiten wieder nullen, falls Neustart */
   /* entfaellt: BK_ReInitApriori(); nur Speichermanager*/
   BO_ReInitApriori();  /* Variablen reinitialisieren */
   RE_ReInitApriori();  /* NaechsteBaumvariable auf x1 setzen */
   /* MO_ReInitApriori(); nicht noetig */
   /* VM_ReInitApriori(); nicht noetig, da nur Speichermanager fuer Variablenmengen */
   U1_ReInitApriori();  /* Substitutions-Variablenbank neu-initialisieren */
   /* TP_InitApriori();  nur Termpaar-Manager initialisieren. */
   /* TV_InitApriori();  nur Termiteration initialisieren. */
   BT_ReInitApriori();  /* Stapel in DSBaumTest initialisieren. */
   /* SUE_InitApriori(); setzt nur Kriterien ein */
   MNF_ReInitApriori(); /* Statistik fuer MNF */
   fprintf(IO_Strom, "... done\n");
}



void IN_SpezifikationGelesen(void)
/* Eingelesene Strukturen freigeben. */
{
   clear_spec();
}
