#include <string.h>
#include <stdlib.h>
#include "Ausgaben.h"
#include "antiwarnings.h"
#include "FehlerBehandlung.h"
#include "FileMenue.h"
#include "Parameter.h"

#include "KBO.h"
#include "LPOVortests.h"
#include "NFBildung.h"
#include "Unifikation1.h"
#include "KPBildung.h"
#include "KPMengenInterreduktion.h"
#include "Zielverwaltung.h"
#include "Interreduktion.h"
#include "Puffer.h"
#include "Hauptkomponenten.h"
#include "Ordnungen.h"
#include "SUEM.h"
#include "Statistiken.h"
#include "pcl.h"


/* ------------------------------------------------------------------------*/
/* -------------------- globale Variablen (siehe Parameter.h) -------------*/
/* ------------------------------------------------------------------------*/
ordering        PA_Ordering;
BOOLEAN PA_unfairness;
mode PA_ExtModus;
BOOLEAN PA_InteraktionAn;
BOOLEAN PA_FailingCompletion;
BOOLEAN PA_PostponingAn;
AufrufOptionT AufrufOptionen[OptNrLAST];
/* ------------------------------------------------------------------------*/
/* -------------------- Initialisierung -----------------------------------*/
/* ------------------------------------------------------------------------*/

const char *GruppenNamen[8] = { "Arbeitsmodi", "KPStrategien", "Normalform-Strategien", "Ziel-Behandlung", 
                            "Interreduktion", "Technische Details", "Zusatz-Optionen", "Debug-Optionen"};

static char *Zeichenkette, *ProtokolldateiName = "stdout";
static BOOLEAN XPRTModus, XPRTModusSichtbar, HilfeVerlangt, Vervollstaendigung, VervollstaendigungZugaenglich;

BOOLEAN VerwendungVonGleichungen;                                         /* Ist das Auftreten von Gleichungen ueberhaupt moeglich? */


#define XPRTModusWechseln() (XPRTModus = !XPRTModus)

#define OptionZugaenglich(option) (AufrufOptionen[option].aktiviert && (!AufrufOptionen[option].IsXPRT || XPRTModus) &&\
                                   (!AufrufOptionen[option].NurVervollstaendigung || Vervollstaendigung))

#define HoechsterSichtbarerIndex(option) (XPRTModus?AufrufOptionen[option].ParameterTyp.zweistellig.AnzahlVerschiedenerWerte:\
                                          AufrufOptionen[option].ParameterTyp.zweistellig.ErsterXPRTWert)


BOOLEAN PA_XPRTModus(void)
{
  return XPRTModus;
}

BOOLEAN PA_XPRTModusSichtbar(void)
{
  return XPRTModusSichtbar;
}

#define INIT_VARS_1                                             \
   PA_InteraktionAn = FALSE;                                     \
   XPRTModus = TRUE;                                            \
   XPRTModusSichtbar = TRUE;                                    \
   VervollstaendigungZugaenglich = TRUE;                        \
   Vervollstaendigung = VervollstaendigungZugaenglich;          \
   PA_unfairness = FALSE;                                       \
      /* Fuer Sichtbarkeit der Modi z.B. im Hilfe-Menue */

#define INIT_VARS_2                                                     \
{                                                                       \
  int i;                                                                \
  for (i=0; i<OptNrLAST; i++)                                           \
    switch (AufrufOptionen[i].ParameterArt){                            \
    case 0:                     /* frei */                              \
      AufrufOptionen[i].ParameterTyp.frei.eingestellterWert =           \
        AufrufOptionen[i].ParameterTyp.frei.defaultWert;                \
      AufrufOptionen[i].ParameterTyp.frei.wurdeAngegeben = FALSE;       \
      break;                                                            \
    case 1:                     /* einstellig */                        \
      AufrufOptionen[i].ParameterTyp.einstellig.eingestellterWert =     \
        AufrufOptionen[i].ParameterTyp.einstellig.defaultWert;          \
      break;                                                            \
    case 2:                     /* zweistellig */                       \
      AufrufOptionen[i].ParameterTyp.zweistellig.eingestellterWert =    \
        AufrufOptionen[i].ParameterTyp.zweistellig.defaultWert;         \
      break;                                                            \
    case 3:                     /* InfoString */                        \
      AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString = "";    \
      break;                                                            \
    default: break;                                                     \
    }                                                                   \
}   

void PA_ReInitApriori(void)
{
  INIT_VARS_1;
  INIT_VARS_2;
}

void PA_InitApriori(void)
/* Aufruf-Optionen besetzen.*/
{
  INIT_VARS_1;

/*------ Arbeitsmodi-------------------------------------*/

   AufrufOptionen[OptNrFailing].Kuerzel           = "-f";
   AufrufOptionen[OptNrFailing].Name              = "'f'ailing completion";
   AufrufOptionen[OptNrFailing].Bedeutung         = "Abbruch beim Auftreten unrichtbarer kritischer Paare";
   AufrufOptionen[OptNrFailing].ParameterTyp.einstellig.BedeutungT        = "Failing Completion";
   AufrufOptionen[OptNrFailing].ParameterTyp.einstellig.BedeutungF        = "Unfailing Completion";
   AufrufOptionen[OptNrFailing].Gruppe            = Arbeitsmodi;
   AufrufOptionen[OptNrFailing].ParameterArt      = 1;
   AufrufOptionen[OptNrFailing].aktiviert = TRUE;
   AufrufOptionen[OptNrFailing].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrFailing].IsXPRT = FALSE;
   AufrufOptionen[OptNrFailing].ParameterTyp.einstellig.defaultWert = FALSE;

/*------ SUE-Statistics ---------------------------------*/

   AufrufOptionen[OptNrSUE].Kuerzel           = "-sue";
   AufrufOptionen[OptNrSUE].Name              = "'SUE'";
   AufrufOptionen[OptNrSUE].Bedeutung         = "SUEM. Syntax:\nstatistics=[on|off]";
   AufrufOptionen[OptNrSUE].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrSUE].ParameterArt      = 3;
   AufrufOptionen[OptNrSUE].aktiviert = TRUE;
   AufrufOptionen[OptNrSUE].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrSUE].IsXPRT = FALSE;
   AufrufOptionen[OptNrSUE].ParameterTyp.InfoString.DefaultInfoString = "statistics=on:paraint=3";
      
/*------ Classification ---------------------------------*/

   AufrufOptionen[OptNrClassification].Kuerzel           = "-clas";
   AufrufOptionen[OptNrClassification].Name              = "'Clas'sification";
   AufrufOptionen[OptNrClassification].Bedeutung         = "Classification of unselected equations. Syntax:\n"
      "w[1|2]=[[lifo|fifo]|comb.weight|occnest.comb.weight|\n"
      "        entry.comb1.weight1.comb2.weight2]             --- primary heuristic\n"
      "DEF=n1:VAR=n2:f=n3:g=n4...:                            --- symbol weights\n"
      "statistics=[on|off]:                                   --- statistics\n" 
      "killer[R|E|RE]=[regular|half|ultimate|double|penalty]: --- killer criterion\n"
      "echild=[regular|half|ultimate|double|penalty]:         --- children of equations\n"
      "perm[3|4|5|6]=[regular|half|ultimate|double|penalty]:  --- permutative equations of degree ...\n"
      "initial=[regular|half|ultimate|double|penalty]:        --- initial equations\n"
      "killedRE=[regular|half|ultimate|double|penalty]:       --- killed rules/equations\n"
      "cut1=n:                                                --- throw away CPs with weight w1 > n\n"
      "statint=n                                              --- intervall for time distribution of weights\n";
   AufrufOptionen[OptNrClassification].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrClassification].ParameterArt      = 3;
   AufrufOptionen[OptNrClassification].aktiviert = TRUE;
   AufrufOptionen[OptNrClassification].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrClassification].IsXPRT = FALSE;
   AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.DefaultInfoString = "def=1:var=1:"
      "initial=ultimate:statistics=off:statint=1:w1=prodsum.gt.length.plus.length:w2=fifo";
      
   AufrufOptionen[OptNrReClassification].Kuerzel           = "-reclas";
   AufrufOptionen[OptNrReClassification].Name              = "'Re-Clas'sification";
   AufrufOptionen[OptNrReClassification].Bedeutung         = "Reclassification of unselected equations during SUE-. Syntax:\n"
     "killer[R|E|RE]=[regular|half|ultimate|double|penalty]:\n"
     "echild=[regular|half|ultimate|double|penalty]:\n"
     "perm[3|4|5|6]=[regular|half|ultimate|double|penalty]\n"; 
   AufrufOptionen[OptNrReClassification].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrReClassification].ParameterArt      = 3;
   AufrufOptionen[OptNrReClassification].aktiviert = TRUE;
   AufrufOptionen[OptNrReClassification].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrReClassification].IsXPRT = FALSE;
   AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.DefaultInfoString = "";

   AufrufOptionen[OptNrCGClassification].Kuerzel           = "-cgclas";
   AufrufOptionen[OptNrCGClassification].Name              = "'CG-Clas'sification";
   AufrufOptionen[OptNrCGClassification].Bedeutung         = "classification of critical goals. Syntax like clas\n"; 
   AufrufOptionen[OptNrCGClassification].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrCGClassification].ParameterArt      = 3;
   AufrufOptionen[OptNrCGClassification].aktiviert = TRUE;
   AufrufOptionen[OptNrCGClassification].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrCGClassification].IsXPRT = FALSE;
   AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.DefaultInfoString = "g1=prod.plus.length.plus.unif:g2=fifo:def=2";
      
/*------ Priority Queue ---------------------------------*/

   AufrufOptionen[OptNrPriorityQueue].Kuerzel           = "-pq";
   AufrufOptionen[OptNrPriorityQueue].Name              = "'P'riority-'Q'ueue";
   AufrufOptionen[OptNrPriorityQueue].Bedeutung         = "'P'riority-'Q'ueue. Syntax:\n"
      "[regular|pairing|hoh]:"
      "[e1|e2|e3|e4|random|unambiguous]:   --- unambiguous nicht bei 'hoh' \n"
      "[left|bin|right]:                   --- nur bei pairing\n"
      "statistics=[on|off]:\n"
      "domain=Granularity";
   AufrufOptionen[OptNrPriorityQueue].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrPriorityQueue].ParameterArt      = 3;
   AufrufOptionen[OptNrPriorityQueue].aktiviert = TRUE;
   AufrufOptionen[OptNrPriorityQueue].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrPriorityQueue].IsXPRT = FALSE;
   AufrufOptionen[OptNrPriorityQueue].ParameterTyp.InfoString.DefaultInfoString = "heap:e1:domain=30:statistics=on";

/*------ Termpair-Representation ------------------------*/

   AufrufOptionen[OptNrTermpairRepresentation].Kuerzel           = "-tpr";
   AufrufOptionen[OptNrTermpairRepresentation].Name              = "'T'erm'p'air-'r'epresentation";
   AufrufOptionen[OptNrTermpairRepresentation].Bedeutung         = "Syntax:\n"
      "[regular|packed|implicit]:statistics=[on|off]";
   AufrufOptionen[OptNrTermpairRepresentation].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrTermpairRepresentation].ParameterArt      = 3;
   AufrufOptionen[OptNrTermpairRepresentation].aktiviert = TRUE;
   AufrufOptionen[OptNrTermpairRepresentation].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrTermpairRepresentation].IsXPRT = FALSE;
   AufrufOptionen[OptNrTermpairRepresentation].ParameterTyp.InfoString.DefaultInfoString = "packed::statistics=on:managers=50";

/*------ CP-Criteria ------------------------------------*/

   AufrufOptionen[OptNrKapur].Kuerzel           = "-subcon";
   AufrufOptionen[OptNrKapur].Name              = "'Subcon'nectedness-criterion";
   AufrufOptionen[OptNrKapur].Bedeutung         = "Vermeidung unnoetiger KPs. Argument = Staerke des Tests\n"
      "['r'|'e']" ;
   AufrufOptionen[OptNrKapur].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrKapur].ParameterArt      = 0;
   AufrufOptionen[OptNrKapur].aktiviert = TRUE;
   AufrufOptionen[OptNrKapur].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrKapur].IsXPRT = FALSE;
   AufrufOptionen[OptNrKapur].ParameterTyp.frei.wurdeAngegeben = FALSE;
   AufrufOptionen[OptNrKapur].ParameterTyp.frei.defaultWert = "";

/*------ Treatment of CPs -------------------------------*/

   AufrufOptionen[OptNrKPInterreduktion].Kuerzel           = "-ki";
   AufrufOptionen[OptNrKPInterreduktion].Name              = "'K'P-Mengen-'I'nterreduktions";
   AufrufOptionen[OptNrKPInterreduktion].Bedeutung         = "Durchreduzieren von SUE. Syntax:\n"
      "period=<periode>[res]:[res]=<stichpunkt>.<stichpunkt>:[res]=<stichpunkt>.<stichpunkt>...[:statistis=[on|off]]";
   AufrufOptionen[OptNrKPInterreduktion].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrKPInterreduktion].ParameterArt      = 3;
   AufrufOptionen[OptNrKPInterreduktion].aktiviert = TRUE;
   AufrufOptionen[OptNrKPInterreduktion].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrKPInterreduktion].IsXPRT = FALSE;
   AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.DefaultInfoString = "statistics=on";

   AufrufOptionen[OptNrKPgen].Kuerzel           = "-kg";
   AufrufOptionen[OptNrKPgen].Name              = "'K'P-Behandlung 'g'eneration";
   AufrufOptionen[OptNrKPgen].Bedeutung         = "Behandlung der KPs after generation.Syntax\n"
      "r:e:s:permsub=[n]";
   AufrufOptionen[OptNrKPgen].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrKPgen].ParameterArt      = 3;
   AufrufOptionen[OptNrKPgen].aktiviert = TRUE;
   AufrufOptionen[OptNrKPgen].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrKPgen].IsXPRT = TRUE;
   AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.DefaultInfoString = "r";

   AufrufOptionen[OptNrKPsel].Kuerzel           = "-ks";
   AufrufOptionen[OptNrKPsel].Name              = "'K'P-Behandlung 's'election";
   AufrufOptionen[OptNrKPsel].Bedeutung         = "Behandlung der KPs after selection.Syntax\n"
      "r:e:s:permsub=[n]";
   AufrufOptionen[OptNrKPsel].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrKPsel].ParameterArt      = 3;
   AufrufOptionen[OptNrKPsel].aktiviert = TRUE;
   AufrufOptionen[OptNrKPsel].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrKPsel].IsXPRT = TRUE;
   AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.DefaultInfoString = "r:e:s:permsubE=4";

/*------ Diverses ---------------------------------------*/

   AufrufOptionen[OptNrPostponing].Kuerzel           = "-pp";
   AufrufOptionen[OptNrPostponing].Name              = "'P'ost-'P'oning";
   AufrufOptionen[OptNrPostponing].Bedeutung         = "Zurueckstellung unrichtbarer KPs";
   AufrufOptionen[OptNrPostponing].ParameterTyp.einstellig.BedeutungT        = "An";
   AufrufOptionen[OptNrPostponing].ParameterTyp.einstellig.BedeutungF        = "Aus";
   AufrufOptionen[OptNrPostponing].Gruppe            = KPStrategien;
   AufrufOptionen[OptNrPostponing].ParameterArt      = 1;
   AufrufOptionen[OptNrPostponing].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrPostponing].aktiviert = FALSE;
   AufrufOptionen[OptNrPostponing].IsXPRT = FALSE;
   AufrufOptionen[OptNrPostponing].ParameterTyp.einstellig.defaultWert = FALSE;

/*------ NormalformStrategien ---------------------------*/

   AufrufOptionen[OptNrNFBildung].Kuerzel                   = "-nf";
   AufrufOptionen[OptNrNFBildung].Name              = "'N'ormal'f'ormbildung";
   AufrufOptionen[OptNrNFBildung].Bedeutung = "Normalization-Strategy and Backtracking-Behaviour.";
   AufrufOptionen[OptNrNFBildung].Gruppe = NormalformStrategien;
   AufrufOptionen[OptNrNFBildung].ParameterArt = 3;
   AufrufOptionen[OptNrNFBildung].aktiviert = TRUE;
   AufrufOptionen[OptNrNFBildung].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrNFBildung].IsXPRT = FALSE;
   AufrufOptionen[OptNrNFBildung].ParameterTyp.InfoString.DefaultInfoString = "outermost:up:repeat";

/*------ Zielbehandlung ---------------------------------*/

   AufrufOptionen[OptNrZielbehandlung].Kuerzel           = "-zb";
   AufrufOptionen[OptNrZielbehandlung].Name              = "'Z'iel-'B'ehandlung";
   AufrufOptionen[OptNrZielbehandlung].Bedeutung         = "Proof-Modus: Behandlung der Ziele";
   AufrufOptionen[OptNrZielbehandlung].Gruppe            = Zielbehandlung ;
   AufrufOptionen[OptNrZielbehandlung].ParameterArt      = 3;
   AufrufOptionen[OptNrZielbehandlung].aktiviert = TRUE;
   AufrufOptionen[OptNrZielbehandlung].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrZielbehandlung].IsXPRT = TRUE;
   AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.DefaultInfoString = "period=1:continue"
      ":mnf=re.0:s:mnfanalysis=100.200.4:mnfantiWOVar:mnfantir=5:mnfup=5";

 /*------ Interreduktion ---------------------------------*/

   AufrufOptionen[OptNrRMrechtsIR].Kuerzel           = "-irr";
   AufrufOptionen[OptNrRMrechtsIR].Name              = "'I'R 'r'ight hand sides of 'r'ules";
   AufrufOptionen[OptNrRMrechtsIR].Bedeutung         = "Interval between subsequent interreductions of rhs of rules";
   AufrufOptionen[OptNrRMrechtsIR].Gruppe            = Interreduktion;
   AufrufOptionen[OptNrRMrechtsIR].ParameterArt      = 3;
   AufrufOptionen[OptNrRMrechtsIR].aktiviert = TRUE;
   AufrufOptionen[OptNrRMrechtsIR].IsXPRT = TRUE;
   AufrufOptionen[OptNrRMrechtsIR].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.DefaultInfoString="period=1:strength=re";

/*------ Ordnungen ---------------------------------*/

   AufrufOptionen[OptNrOrdnungen].Kuerzel           = "-ord";
   AufrufOptionen[OptNrOrdnungen].Name              = "Orderings";
   AufrufOptionen[OptNrOrdnungen].Bedeutung         = "Override ordering from specification ";
   AufrufOptionen[OptNrOrdnungen].Gruppe            = ZusatzOptionen;
   AufrufOptionen[OptNrOrdnungen].ParameterArt      = 3;
   AufrufOptionen[OptNrOrdnungen].aktiviert = TRUE;
   AufrufOptionen[OptNrOrdnungen].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrOrdnungen].IsXPRT = TRUE;
   AufrufOptionen[OptNrOrdnungen].ParameterTyp.InfoString.DefaultInfoString = "auto";
   
/*------ TechnischeDetails ------------------------------*/

   AufrufOptionen[OptNrAusgabeLevel].Kuerzel                   = "-a";
   AufrufOptionen[OptNrAusgabeLevel].Name              = "'A'usgabe-Level";
   AufrufOptionen[OptNrAusgabeLevel].Bedeutung = "Einstellen des Ausgabe-Levels (Wert ist proportional zur Geschwaetzigkeit)";
   AufrufOptionen[OptNrAusgabeLevel].Gruppe    = TechnischeDetails;
   AufrufOptionen[OptNrAusgabeLevel].ParameterArt = 2;
   AufrufOptionen[OptNrAusgabeLevel].aktiviert = TRUE;
   AufrufOptionen[OptNrAusgabeLevel].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrAusgabeLevel].IsXPRT = FALSE;
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.defaultWert = 0;
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.AnzahlVerschiedenerWerte = 5;
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.ErsterXPRTWert = 5;
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.WerteImKlartext[0] = "0";
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.WerteImKlartext[1] = "1";
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.WerteImKlartext[2] = "2";
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.WerteImKlartext[3] = "3";
   AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.WerteImKlartext[4] = "4";

   AufrufOptionen[OptNrPCLProt].Kuerzel           = "-pcl";
   AufrufOptionen[OptNrPCLProt].Name              = "'PCL'-Protokoll-Datei";
   AufrufOptionen[OptNrPCLProt].Bedeutung         = "PCL-Protokoll fuer Beweisauswertung. Syntax:\n"
      "PCL=Zieldatei                             --- normales PCL-Protokoll \n"
      "uPCL=Zieldatei                            --- Mikro-Protokoll\n"
      "uPCLQuelle=Quelldatei:uPCLZiel=Zieldatei  --- normales aus (extrahiertem) Mikro-Protokoll generieren:\n"
      "corr=<Corrdatei>                          --- Korrespondenz zw. PCL-ID und WM-IDs";
   AufrufOptionen[OptNrPCLProt].Gruppe            = TechnischeDetails;
   AufrufOptionen[OptNrPCLProt].ParameterArt      = 3;
#if COMP_PCL
   AufrufOptionen[OptNrPCLProt].aktiviert = TRUE;
#else
   AufrufOptionen[OptNrPCLProt].aktiviert = FALSE;
#endif
   AufrufOptionen[OptNrPCLProt].NurVervollstaendigung = TRUE;
   AufrufOptionen[OptNrPCLProt].IsXPRT = FALSE;
   AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.DefaultInfoString = "iPCL";
   
   AufrufOptionen[OptNrZeitlimit].Kuerzel           = "-tl";
   AufrufOptionen[OptNrZeitlimit].Name              = "set 't'ime-'l'imit";
   AufrufOptionen[OptNrZeitlimit].Bedeutung         = "Waldmeister terminates after <sec> process time";
   AufrufOptionen[OptNrZeitlimit].Gruppe            = TechnischeDetails;
   AufrufOptionen[OptNrZeitlimit].ParameterArt      = 0;
   AufrufOptionen[OptNrZeitlimit].aktiviert = TRUE;
   AufrufOptionen[OptNrZeitlimit].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrZeitlimit].IsXPRT = TRUE;
   AufrufOptionen[OptNrZeitlimit].ParameterTyp.frei.defaultWert = "0";
   
   AufrufOptionen[OptNrRealZeitlimit].Kuerzel           = "-rtl";
   AufrufOptionen[OptNrRealZeitlimit].Name              = "set 'r'eal-'t'ime-'l'imit";
   AufrufOptionen[OptNrRealZeitlimit].Bedeutung         = "Waldmeister terminates after <sec> real time";
   AufrufOptionen[OptNrRealZeitlimit].Gruppe            = TechnischeDetails;
   AufrufOptionen[OptNrRealZeitlimit].ParameterArt      = 0;
   AufrufOptionen[OptNrRealZeitlimit].aktiviert = TRUE;
   AufrufOptionen[OptNrRealZeitlimit].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrRealZeitlimit].IsXPRT = TRUE;
   AufrufOptionen[OptNrRealZeitlimit].ParameterTyp.frei.defaultWert = "0";
   AufrufOptionen[OptNrRealZeitlimit].ParameterTyp.frei.wurdeAngegeben = FALSE;
   
   AufrufOptionen[OptNrErgebnis].Kuerzel           = "-res";
   AufrufOptionen[OptNrErgebnis].Name              = "'Res'ult if interaction is off";
   AufrufOptionen[OptNrErgebnis].Bedeutung         = "Syntax:\n h    - Head\n"
      "r    - Regelmenge\n"
      "e    - Gleichungsmenge\n"
      "s    - Statistik (komplett)\n"
      "k    - Konfiguration\n"
      "i    - Intern\n"
      "p    - Prozess\n"
      "t    - Times\n"
      "pa   - Aufruf-Parameter\n"
      "sh   - Short Statistik\n"
      "tex  - LaTeX-Statistik\n"
      "stex - short LaTeX-Statistik\n";
   AufrufOptionen[OptNrErgebnis].Gruppe            = TechnischeDetails;
   AufrufOptionen[OptNrErgebnis].ParameterArt      = 3;
   AufrufOptionen[OptNrErgebnis].aktiviert = TRUE;
   AufrufOptionen[OptNrErgebnis].IsXPRT = TRUE;
   AufrufOptionen[OptNrErgebnis].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrErgebnis].ParameterTyp.InfoString.DefaultInfoString="-:stex";
   
/*------ DebugOptionen ----------------------------------*/

   AufrufOptionen[OptNrDebugOption1].Kuerzel = "-debug1";
   AufrufOptionen[OptNrDebugOption1].Name= "Debug-Parameter 1";
   AufrufOptionen[OptNrDebugOption1].Bedeutung = "freier Parameter fuer Debugging-Zwecke, ansonsten ungebunden";
   AufrufOptionen[OptNrDebugOption1].Gruppe= DebugOptionen;
   AufrufOptionen[OptNrDebugOption1].aktiviert=TRUE;
   AufrufOptionen[OptNrDebugOption1].IsXPRT=TRUE;
   AufrufOptionen[OptNrDebugOption1].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrDebugOption1].ParameterArt = 0;
   AufrufOptionen[OptNrDebugOption1].ParameterTyp.frei.defaultWert = NULL;
   AufrufOptionen[OptNrDebugOption1].ParameterTyp.frei.wurdeAngegeben = FALSE;
      
   AufrufOptionen[OptNrDebugOption2].Kuerzel = "-debug2";
   AufrufOptionen[OptNrDebugOption2].Name= "Debug-Parameter 2";
   AufrufOptionen[OptNrDebugOption2].Bedeutung = "freier Parameter fuer Debugging-Zwecke, ansonsten ungebunden";
   AufrufOptionen[OptNrDebugOption2].Gruppe= DebugOptionen;
   AufrufOptionen[OptNrDebugOption2].ParameterArt = 0;
   AufrufOptionen[OptNrDebugOption2].aktiviert=TRUE;
   AufrufOptionen[OptNrDebugOption2].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrDebugOption2].IsXPRT=TRUE;
   AufrufOptionen[OptNrDebugOption2].ParameterTyp.frei.defaultWert = NULL;
   AufrufOptionen[OptNrDebugOption2].ParameterTyp.frei.wurdeAngegeben = FALSE;
      
   AufrufOptionen[OptNrDebugOption3].Kuerzel = "-debug3";
   AufrufOptionen[OptNrDebugOption3].Name= "Debug-Parameter 3";
   AufrufOptionen[OptNrDebugOption3].Bedeutung = "freier Parameter fuer Debugging-Zwecke, ansonsten ungebunden";
   AufrufOptionen[OptNrDebugOption3].Gruppe= DebugOptionen;
   AufrufOptionen[OptNrDebugOption3].aktiviert=TRUE;
   AufrufOptionen[OptNrDebugOption3].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrDebugOption3].IsXPRT=TRUE;
   AufrufOptionen[OptNrDebugOption3].ParameterArt = 0;
   AufrufOptionen[OptNrDebugOption3].ParameterTyp.frei.defaultWert = NULL;
   AufrufOptionen[OptNrDebugOption3].ParameterTyp.frei.wurdeAngegeben = FALSE;
      
   AufrufOptionen[OptNrDebugOption4].Kuerzel = "-debug4";
   AufrufOptionen[OptNrDebugOption4].Name= "Debug-Parameter 4";
   AufrufOptionen[OptNrDebugOption4].Bedeutung = "freier Parameter fuer Debugging-Zwecke, ansonsten ungebunden";
   AufrufOptionen[OptNrDebugOption4].Gruppe= DebugOptionen;
   AufrufOptionen[OptNrDebugOption4].aktiviert=TRUE;
   AufrufOptionen[OptNrDebugOption4].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrDebugOption4].IsXPRT=TRUE;
   AufrufOptionen[OptNrDebugOption4].ParameterArt = 0;
   AufrufOptionen[OptNrDebugOption4].ParameterTyp.frei.defaultWert = NULL;
   AufrufOptionen[OptNrDebugOption4].ParameterTyp.frei.wurdeAngegeben = FALSE;
      
   AufrufOptionen[OptNrDebugOption5].Kuerzel = "-debug5";
   AufrufOptionen[OptNrDebugOption5].Name= "Debug-Parameter 5";
   AufrufOptionen[OptNrDebugOption5].Bedeutung = "freier Parameter fuer Debugging-Zwecke, ansonsten ungebunden";
   AufrufOptionen[OptNrDebugOption5].Gruppe= DebugOptionen;
   AufrufOptionen[OptNrDebugOption5].aktiviert=TRUE;
   AufrufOptionen[OptNrDebugOption5].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrDebugOption5].IsXPRT=TRUE;
   AufrufOptionen[OptNrDebugOption5].ParameterArt = 0;
   AufrufOptionen[OptNrDebugOption5].ParameterTyp.frei.defaultWert = NULL;
   AufrufOptionen[OptNrDebugOption5].ParameterTyp.frei.wurdeAngegeben = FALSE;
      
   AufrufOptionen[OptNrDebugOption6].Kuerzel = "-debug6";
   AufrufOptionen[OptNrDebugOption6].Name= "Debug-Parameter 6";
   AufrufOptionen[OptNrDebugOption6].Bedeutung = "freier Parameter fuer Debugging-Zwecke, ansonsten ungebunden";
   AufrufOptionen[OptNrDebugOption6].Gruppe= DebugOptionen;
   AufrufOptionen[OptNrDebugOption6].ParameterArt = 0;
   AufrufOptionen[OptNrDebugOption6].aktiviert=TRUE;
   AufrufOptionen[OptNrDebugOption6].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrDebugOption6].IsXPRT=TRUE;
   AufrufOptionen[OptNrDebugOption6].ParameterTyp.frei.defaultWert = NULL;
   AufrufOptionen[OptNrDebugOption6].ParameterTyp.frei.wurdeAngegeben = FALSE;
      
   AufrufOptionen[OptNrDebugOption7].Kuerzel = "-debug7";
   AufrufOptionen[OptNrDebugOption7].Name= "Debug-Parameter 7";
   AufrufOptionen[OptNrDebugOption7].Bedeutung = "freier Parameter fuer Debugging-Zwecke, ansonsten ungebunden";
   AufrufOptionen[OptNrDebugOption7].Gruppe= DebugOptionen;
   AufrufOptionen[OptNrDebugOption7].aktiviert=TRUE;
   AufrufOptionen[OptNrDebugOption7].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrDebugOption7].IsXPRT=TRUE;
   AufrufOptionen[OptNrDebugOption7].ParameterArt = 0;
   AufrufOptionen[OptNrDebugOption7].ParameterTyp.frei.defaultWert = NULL;
   AufrufOptionen[OptNrDebugOption7].ParameterTyp.frei.wurdeAngegeben = FALSE;
      
   AufrufOptionen[OptNrDebugOption8].Kuerzel = "-debug8";
   AufrufOptionen[OptNrDebugOption8].Name= "Debug-Parameter 8";
   AufrufOptionen[OptNrDebugOption8].Bedeutung = "freier Parameter fuer Debugging-Zwecke, ansonsten ungebunden";
   AufrufOptionen[OptNrDebugOption8].Gruppe= DebugOptionen;
   AufrufOptionen[OptNrDebugOption8].aktiviert=TRUE;
   AufrufOptionen[OptNrDebugOption8].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrDebugOption8].IsXPRT=TRUE;
   AufrufOptionen[OptNrDebugOption8].ParameterArt = 0;
   AufrufOptionen[OptNrDebugOption8].ParameterTyp.frei.defaultWert = NULL;
   AufrufOptionen[OptNrDebugOption8].ParameterTyp.frei.wurdeAngegeben = FALSE;
      
   AufrufOptionen[OptNrDebugOption9].Kuerzel = "-debug9";
   AufrufOptionen[OptNrDebugOption9].Name= "Debug-Parameter 9";
   AufrufOptionen[OptNrDebugOption9].Bedeutung = "freier Parameter fuer Debugging-Zwecke, ansonsten ungebunden";
   AufrufOptionen[OptNrDebugOption9].Gruppe= DebugOptionen;
   AufrufOptionen[OptNrDebugOption9].aktiviert=TRUE;
   AufrufOptionen[OptNrDebugOption9].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrDebugOption9].IsXPRT=TRUE;
   AufrufOptionen[OptNrDebugOption9].ParameterArt = 0;
   AufrufOptionen[OptNrDebugOption9].ParameterTyp.frei.defaultWert = NULL;
   AufrufOptionen[OptNrDebugOption9].ParameterTyp.frei.wurdeAngegeben = FALSE;
      
   AufrufOptionen[OptNrDebugOption0].Kuerzel = "-debug0";
   AufrufOptionen[OptNrDebugOption0].Name= "Debug-Parameter 0";
   AufrufOptionen[OptNrDebugOption0].Bedeutung = "freier Parameter fuer Debugging-Zwecke, ansonsten ungebunden";
   AufrufOptionen[OptNrDebugOption0].Gruppe= DebugOptionen;
   AufrufOptionen[OptNrDebugOption0].ParameterArt = 0;
   AufrufOptionen[OptNrDebugOption0].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrDebugOption0].aktiviert=TRUE;
   AufrufOptionen[OptNrDebugOption0].IsXPRT=TRUE;
   AufrufOptionen[OptNrDebugOption0].ParameterTyp.frei.defaultWert = NULL;
   AufrufOptionen[OptNrDebugOption0].ParameterTyp.frei.wurdeAngegeben = FALSE;
      
   AufrufOptionen[OptNrNoDebug].Kuerzel = "-noDebug";
   AufrufOptionen[OptNrNoDebug].Name = "Debugmarken beruecksichtigen";
   AufrufOptionen[OptNrNoDebug].Bedeutung = "Debugmarken beruecksichtigen";
   AufrufOptionen[OptNrNoDebug].ParameterTyp.einstellig.BedeutungT = "Aus";
   AufrufOptionen[OptNrNoDebug].ParameterTyp.einstellig.BedeutungF = "An";
   AufrufOptionen[OptNrNoDebug].Gruppe= DebugOptionen;
   AufrufOptionen[OptNrNoDebug].ParameterArt = 1;
   AufrufOptionen[OptNrNoDebug].aktiviert=TRUE;
   AufrufOptionen[OptNrNoDebug].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrNoDebug].IsXPRT=TRUE;
   AufrufOptionen[OptNrNoDebug].ParameterTyp.einstellig.eingestellterWert = FALSE;
   
   AufrufOptionen[OptNrTestOption].Kuerzel = "-test";
   AufrufOptionen[OptNrTestOption].Name = "Schalter fuer Testmodus";
   AufrufOptionen[OptNrTestOption].Bedeutung = "Verzweigung in Testroutine";
   AufrufOptionen[OptNrTestOption].ParameterTyp.einstellig.BedeutungT = "An";
   AufrufOptionen[OptNrTestOption].ParameterTyp.einstellig.BedeutungF = "Aus";
   AufrufOptionen[OptNrTestOption].Gruppe= DebugOptionen;
   AufrufOptionen[OptNrTestOption].ParameterArt = 1;
   AufrufOptionen[OptNrTestOption].aktiviert=TRUE;
   AufrufOptionen[OptNrTestOption].NurVervollstaendigung = FALSE;
   AufrufOptionen[OptNrTestOption].IsXPRT=TRUE;
   AufrufOptionen[OptNrTestOption].ParameterTyp.einstellig.eingestellterWert = FALSE;
   
   INIT_VARS_2;

   Zeichenkette = our_alloc(INPUT_STRLEN * sizeof(char));
}


void PA_InitAposteriori(void)
{
   int i;
   
   if ((PA_ExtModus == proof) || (PA_ExtModus == completion)){ /* bei Proof oder Completion */
     SGM_InitAposteriori(); /* Subgleichungen */
   }
   else{                                                       /* sonst */
      /* In den verschiedenen Modi ungueltige Parameter ueberschreiben  */

      AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.UserInfoString = "-";
      AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.UserInfoString = "-:r"; /* Nur Regeln */
      AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.UserInfoString = "-"; /* Kein PCL */
      AufrufOptionen[OptNrKapur].ParameterTyp.frei.eingestellterWert = "-";      /* Kein Kapur */
      AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.UserInfoString = "-:w1=fifo:intial=ultimate";
      AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.UserInfoString = "-";
      AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.UserInfoString = "-";
      AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.UserInfoString = "-";
      AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.UserInfoString = "";
      AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.UserInfoString = "";
   }

   for (i=0; i<OptNrLAST; i++)
      if (AufrufOptionen[i].ParameterArt == 3){ /* InfoString */
         if (IS_IsIn("-", AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString))
            AufrufOptionen[i].ParameterTyp.InfoString.DefaultInfoString = "";
      }

   IO_DruckeKopf();             /* Programm-Kopf ausgeben */

   IO_DruckeBeispiel();

   IO_DruckeKurzUeberschrift("INITIALIZATION", "+");

   IO_ParamInfo(ParamEvaluate, atoi(AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.
         WerteImKlartext[AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.eingestellterWert]),
         ProtokolldateiName);   /* nur ausgeben! */

   if (Vervollstaendigung){
     fprintf(IO_Strom, "\n* MODE:  %s completion\n", 
             AufrufOptionen[OptNrFailing].ParameterTyp.einstellig.eingestellterWert?
             "failing":"unfailing");
   }
   ORD_InitAposteriori (AufrufOptionen[OptNrOrdnungen].ParameterTyp.InfoString.DefaultInfoString,
			AufrufOptionen[OptNrOrdnungen].ParameterTyp.InfoString.UserInfoString);
   
   BK_InitAposteriori();

   /* An dieser Stelle muessten noch Ueberpruefungen der Gleichungen und Ziele stattfinden:
      - verwendete Argumentsorten;
      - Verwendung von Funktionen nur mit den richtigen Stelligkeiten.*/
   /* Skolemisierung der Ziele: Passiert aus dem Hauptprogramm heraus.*/
   /* clear_spec(); Erst einfuegen, wenn NACH Ablauf dieser Funktion nicht mehr auf 
		    die eingelesenen Daten zugegriffen werden muss.
		    ! Kann hier noch nicht aufgerufen werden, 
		    da die Ziele erst spaeter eingebaut werden. */

   PU_InitAposteriori();

   NF_InitAposteriori(AufrufOptionen[OptNrNFBildung].ParameterTyp.InfoString.DefaultInfoString,
                      AufrufOptionen[OptNrNFBildung].ParameterTyp.InfoString.UserInfoString);

   U1_InitAposteriori(AufrufOptionen[OptNrKapur].ParameterTyp.frei.eingestellterWert);

   KPB_InitAposteriori(AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.DefaultInfoString,
                       AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.UserInfoString);

   ST_InitAposteriori(AufrufOptionen[OptNrErgebnis].ParameterTyp.InfoString.DefaultInfoString,
                      AufrufOptionen[OptNrErgebnis].ParameterTyp.InfoString.UserInfoString);

   HK_InitAposteriori(AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.DefaultInfoString,
                      AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.UserInfoString);

   IR_InitAposteriori(AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.DefaultInfoString,
                      AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.UserInfoString);
   SUE_INITIALIZE(AufrufOptionen[OptNrSUE].ParameterTyp.InfoString.DefaultInfoString,
                  AufrufOptionen[OptNrSUE].ParameterTyp.InfoString.UserInfoString,
                  AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.DefaultInfoString,
                  AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.UserInfoString,
                  AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.DefaultInfoString,
                  AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.UserInfoString,
                  AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.DefaultInfoString,
                  AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.UserInfoString,
                  AufrufOptionen[OptNrPriorityQueue].ParameterTyp.InfoString.DefaultInfoString,
                  AufrufOptionen[OptNrPriorityQueue].ParameterTyp.InfoString.UserInfoString,
                  AufrufOptionen[OptNrTermpairRepresentation].ParameterTyp.InfoString.DefaultInfoString,
                  AufrufOptionen[OptNrTermpairRepresentation].ParameterTyp.InfoString.UserInfoString);

   KPIR_InitAposteriori(AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.DefaultInfoString,
                        AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.UserInfoString);

   PCL_InitAposteriori(AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.DefaultInfoString,
                       AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.UserInfoString);
   Z_InitAposteriori(AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.DefaultInfoString,
                     AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.UserInfoString);
}

/* ------------------------------------------------------------------------------------*/
/* -------------------- Ausgabe der Einstellungen im Statistik-Kopf -------------------*/
/* ------------------------------------------------------------------------------------*/

static void SchreibeStringOhneHochkommataNachZeichenkette(const char *string)
/* s.Name: nach Ende dieser Prozedur steht string bis auf Hochkommata in Zeichenkette.*/
{
   const char *p;
   char       *ziel;
  for (p = string, ziel = Zeichenkette; *p; p++){ /* Hochkommas aus String eliminieren */
    if (*p!='\''){
      *ziel = *p;
      ziel++;
    }
  }
  *ziel = '\0';
}

void PA_DruckeModi(void)
/* Gibt die eingestellten Modi auf den IO_Strom aus.
   Verwendung bei Ausgabe der Statistik.*/
{
  if (Vervollstaendigung){
    fprintf(IO_Strom, "\n* MODE:  %s completion\n", 
            AufrufOptionen[OptNrFailing].ParameterTyp.einstellig.eingestellterWert?
            "failing":"unfailing");
  }
  
  IO_ParamInfo(ParamEvaluate, atoi(AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.
         WerteImKlartext[AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.eingestellterWert]),
         ProtokolldateiName);

  ORD_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrOrdnungen].ParameterTyp.InfoString.DefaultInfoString,
			AufrufOptionen[OptNrOrdnungen].ParameterTyp.InfoString.UserInfoString);
  
  NF_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrNFBildung].ParameterTyp.InfoString.UserInfoString,
               AufrufOptionen[OptNrNFBildung].ParameterTyp.InfoString.DefaultInfoString);

  U1_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrKapur].ParameterTyp.frei.eingestellterWert);

  KPB_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.UserInfoString);
  HK_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.UserInfoString);
  if (Vervollstaendigung){
    IR_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.UserInfoString,
                 AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.DefaultInfoString);
  }

  SUE_ParamInfo(ParamEvaluate, 99, 
                AufrufOptionen[OptNrSUE].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrSUE].ParameterTyp.InfoString.UserInfoString,
                AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.UserInfoString,
                AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.UserInfoString,
                AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.UserInfoString,
                AufrufOptionen[OptNrPriorityQueue].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrPriorityQueue].ParameterTyp.InfoString.UserInfoString,
                AufrufOptionen[OptNrTermpairRepresentation].ParameterTyp.InfoString.DefaultInfoString,
                AufrufOptionen[OptNrTermpairRepresentation].ParameterTyp.InfoString.UserInfoString);
  
  KPIR_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.DefaultInfoString,
                 AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.UserInfoString);

  if (PA_ExtModus==proof){
     Z_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.DefaultInfoString,
                 AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.UserInfoString);
  }
  PCL_ParamInfo(ParamEvaluate, AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.DefaultInfoString,
                 AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.UserInfoString);

  fprintf(IO_Strom, "\n\n");
}  

static void DruckeOptionDefault(FILE *fp, AufrufOptionenTyp i)
/* Gibt eine einzelne Option mit Kuerzel und Namen auf Strom fp aus.*/
{
   if (OptionZugaenglich(i)){
      switch (AufrufOptionen[i].ParameterArt){
      case 0:
         if (AufrufOptionen[i].ParameterTyp.frei.eingestellterWert!=NULL)
            fprintf(fp, "%s %s ",AufrufOptionen[i].Kuerzel, AufrufOptionen[i].ParameterTyp.frei.eingestellterWert);
         break;
      case 1:
         fprintf(fp, "%s %s ",AufrufOptionen[i].Kuerzel,(AufrufOptionen[i].ParameterTyp.einstellig.eingestellterWert)?" AN":"AUS");
         break;
      case 2:
         fprintf(fp, "%s %30s ",AufrufOptionen[i].Kuerzel,
                 AufrufOptionen[i].ParameterTyp.zweistellig.WerteImKlartext[AufrufOptionen[i].ParameterTyp.zweistellig.eingestellterWert]);
      case 3:
         fprintf(fp, "%s %s%s%s ",
                 AufrufOptionen[i].Kuerzel,
                 AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString,
                 AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString?":":"",
                 AufrufOptionen[i].ParameterTyp.InfoString.DefaultInfoString);
         break;
      default:
         break;
      }
   }
}

void PA_DefaultsAusgeben(char *Filename)
{
   FILE *fp;
   int i;
   /*   ParameterGruppenT gruppe;*/
   unsigned int gruppe;
   
   if (strcmp (Filename, "stdout") ==0)
      fp = stdout;
   else
      fp = fopen(Filename, "a");
   
   for (gruppe = (int)Arbeitsmodi; gruppe < (int)TechnischeDetails; gruppe++)
      for (i=0; i<OptNrLAST; i++)
         if ((AufrufOptionen[i].Gruppe == gruppe)&&OptionZugaenglich(i))
            DruckeOptionDefault(fp,i);
   
   fprintf(fp, "\n");
   if (strcmp (Filename, "stdout") != 0)
      fclose(fp);
}

/* ------------------------------------------------------------------------*/
/* -------------------- Sonstige Ausgaben bzgl. Optionen ------------------*/
/* ------------------------------------------------------------------------*/

static void DruckeOption(AufrufOptionenTyp i)
/* Gibt eine einzelne Option ausfuehrlich am Bildschirm aus.*/
{
   int j, sueModule=0;
   
   SchreibeStringOhneHochkommataNachZeichenkette(AufrufOptionen[i].Name);

   switch(i){
   case OptNrNFBildung: 
      NF_ParamInfo(ParamExplain, AufrufOptionen[OptNrNFBildung].ParameterTyp.InfoString.UserInfoString,
                   AufrufOptionen[OptNrNFBildung].ParameterTyp.InfoString.DefaultInfoString);
      break;
   case OptNrRMrechtsIR:
      IR_ParamInfo(ParamExplain, AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.UserInfoString,
                   AufrufOptionen[OptNrRMrechtsIR].ParameterTyp.InfoString.DefaultInfoString);
      break;
   case OptNrKapur:
      U1_ParamInfo(ParamExplain, AufrufOptionen[OptNrKapur].ParameterTyp.frei.eingestellterWert);
      break;
   case OptNrZielbehandlung:
      Z_ParamInfo(ParamExplain, AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.DefaultInfoString,
                  AufrufOptionen[OptNrZielbehandlung].ParameterTyp.InfoString.UserInfoString);
      break;
   case OptNrKPgen:
      KPB_ParamInfo(ParamExplain, AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.DefaultInfoString,
                    AufrufOptionen[OptNrKPgen].ParameterTyp.InfoString.UserInfoString);
      break;
   case OptNrKPsel:
      HK_ParamInfo(ParamExplain, AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.DefaultInfoString,
                    AufrufOptionen[OptNrKPsel].ParameterTyp.InfoString.UserInfoString);
      break;
   case OptNrKPInterreduktion:
      KPIR_ParamInfo(ParamExplain, AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.DefaultInfoString,
                           AufrufOptionen[OptNrKPInterreduktion].ParameterTyp.InfoString.UserInfoString);
      break;
   case OptNrAusgabeLevel:
     IO_ParamInfo(ParamExplain, atoi(AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.
                          WerteImKlartext[AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.eingestellterWert]),
                  ProtokolldateiName);
       break;
   case OptNrOrdnungen:
     ORD_ParamInfo(ParamExplain, AufrufOptionen[OptNrOrdnungen].ParameterTyp.InfoString.DefaultInfoString,
                   AufrufOptionen[OptNrOrdnungen].ParameterTyp.InfoString.UserInfoString);
     break;
   case OptNrTermpairRepresentation: sueModule++;
   case OptNrPriorityQueue:sueModule++;
   case OptNrCGClassification:sueModule++;
   case OptNrReClassification:sueModule++;
   case OptNrClassification:sueModule++;
   case OptNrSUE:
     SUE_ParamInfo(ParamExplain, sueModule, 
                   AufrufOptionen[OptNrSUE].ParameterTyp.InfoString.DefaultInfoString,
                   AufrufOptionen[OptNrSUE].ParameterTyp.InfoString.UserInfoString,
                   AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.DefaultInfoString,
                   AufrufOptionen[OptNrClassification].ParameterTyp.InfoString.UserInfoString,
                   AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.DefaultInfoString,
                   AufrufOptionen[OptNrReClassification].ParameterTyp.InfoString.UserInfoString,
                   AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.DefaultInfoString,
                   AufrufOptionen[OptNrCGClassification].ParameterTyp.InfoString.UserInfoString,
                   AufrufOptionen[OptNrPriorityQueue].ParameterTyp.InfoString.DefaultInfoString,
                   AufrufOptionen[OptNrPriorityQueue].ParameterTyp.InfoString.UserInfoString,
                   AufrufOptionen[OptNrTermpairRepresentation].ParameterTyp.InfoString.DefaultInfoString,
                   AufrufOptionen[OptNrTermpairRepresentation].ParameterTyp.InfoString.UserInfoString);
     break;
   case OptNrPCLProt:
     PCL_ParamInfo(ParamExplain, AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.DefaultInfoString,
                   AufrufOptionen[OptNrPCLProt].ParameterTyp.InfoString.UserInfoString);
     break;
   default: IO_DruckeKurzUeberschrift(Zeichenkette, "-");
      break;
   }

   fprintf(IO_Strom, "%s >>> %s.\n", AufrufOptionen[i].Kuerzel, AufrufOptionen[i].Name);
   switch (AufrufOptionen[i].ParameterArt){
   case 0:
     fprintf(IO_Strom, "%s\n", AufrufOptionen[i].Bedeutung);
     fprintf(IO_Strom, "This option requires an argument. ");
     if (AufrufOptionen[i].ParameterTyp.frei.wurdeAngegeben)
       fprintf(IO_Strom, "Argument given is %s.", AufrufOptionen[i].ParameterTyp.frei.eingestellterWert);
     else
       fprintf(IO_Strom, "Argument set to %s per default.", 
              AufrufOptionen[i].ParameterTyp.frei.eingestellterWert?
              AufrufOptionen[i].ParameterTyp.frei.eingestellterWert:"(nothing)");
     break;
   case 1:
     fprintf(IO_Strom, "This Option is a switch: ");
     fprintf(IO_Strom, "%s.\n Position: %s\n Alternative: %s\n", 
            AufrufOptionen[i].Bedeutung,
            AufrufOptionen[i].ParameterTyp.einstellig.eingestellterWert?
            AufrufOptionen[i].ParameterTyp.einstellig.BedeutungT:AufrufOptionen[i].ParameterTyp.einstellig.BedeutungF,
            AufrufOptionen[i].ParameterTyp.einstellig.eingestellterWert?
            AufrufOptionen[i].ParameterTyp.einstellig.BedeutungF:AufrufOptionen[i].ParameterTyp.einstellig.BedeutungT);
     break;
   case 2:
     fprintf(IO_Strom, "%s\n", AufrufOptionen[i].Bedeutung);
     fprintf(IO_Strom, "This option requires an argument from ");
     for (j=0; j<HoechsterSichtbarerIndex(i); j++){
        fprintf(IO_Strom, " %s",AufrufOptionen[i].ParameterTyp.zweistellig.WerteImKlartext[j]);
        if (j==AufrufOptionen[i].ParameterTyp.zweistellig.eingestellterWert)
           fprintf(IO_Strom, "*");
        if (j+1<HoechsterSichtbarerIndex(i))
           fprintf(IO_Strom, ", ");
     }
     fprintf(IO_Strom, ".\n");
     break;
  case 3:
     fprintf(IO_Strom, "%s\n", AufrufOptionen[i].Bedeutung);
     fprintf(IO_Strom, "This option requires an Info-String (opt1:opt2:opt3=val:...)\n");
     fprintf(IO_Strom, "Default-Info-String is %s,\n", AufrufOptionen[i].ParameterTyp.InfoString.DefaultInfoString);
     fprintf(IO_Strom, "Your input was %s.\n", 
             AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString?"nothing":AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString);
     break;
  default:
     break;
  }
   fprintf(IO_Strom, "Option is %s and has influence %s.\n\n\n",
           AufrufOptionen[i].IsXPRT?"available only in expert-mode":"always available",
           AufrufOptionen[i].NurVervollstaendigung?"only on the modes proof and completion":"on all modes");
}


static void DruckeOptionKurz(AufrufOptionenTyp i)
/* Gibt eine einzelne Option mit Kuerzel und Namen am Bildschirm aus.*/
{
   int j;
   
   fprintf(IO_Strom, "%6s  >>>  %s",AufrufOptionen[i].Kuerzel+1, AufrufOptionen[i].Name);
   switch (AufrufOptionen[i].ParameterArt){
   case 0:
      if (AufrufOptionen[i].ParameterTyp.frei.wurdeAngegeben)
         fprintf(IO_Strom, ". Value is %s.\n", AufrufOptionen[i].ParameterTyp.frei.eingestellterWert);
      else
         fprintf(IO_Strom, ". Default is %s.\n", 
                 AufrufOptionen[i].ParameterTyp.frei.eingestellterWert?
                 AufrufOptionen[i].ParameterTyp.frei.eingestellterWert:"(nichts)");
      break;
   case 1:
      fprintf(IO_Strom, ". Schalter-Stellung: %s.\n",(AufrufOptionen[i].ParameterTyp.einstellig.eingestellterWert)?
              AufrufOptionen[i].ParameterTyp.einstellig.BedeutungT:AufrufOptionen[i].ParameterTyp.einstellig.BedeutungF);
      break;
   case 2:
      fprintf(IO_Strom, " (Values: ");
      for (j=0; j<HoechsterSichtbarerIndex(i); j++){
         fprintf(IO_Strom, "%s",AufrufOptionen[i].ParameterTyp.zweistellig.WerteImKlartext[j]);
         if (j==AufrufOptionen[i].ParameterTyp.zweistellig.eingestellterWert)
            fprintf(IO_Strom, "*");
         if (j+1<HoechsterSichtbarerIndex(i))
            fprintf(IO_Strom, ", ");
      }
      fprintf(IO_Strom, ").\n");
      break;
   case 3:
      fprintf(IO_Strom, ". Setting: %s%s%s.\n", 
              AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString,
              AufrufOptionen[i].ParameterTyp.InfoString.UserInfoString?":":"",
              AufrufOptionen[i].ParameterTyp.InfoString.DefaultInfoString);
      break;
   default:
      break;
   }
}

static int OptionMitKuerzelWaehlen(BOOLEAN NurErklaeren);

static void Hilfe(void)
{
   int eingabe;
   char text[INPUT_STRLEN];
   BOOLEAN zurueck = FALSE;
   
   do{
      int nr=1;
     IO_DruckeUeberschrift("Waldmeister II - HILFE");
     if (FM_HilfeDatei())
       fprintf(IO_Strom, "%d  General Information\n", nr++);
     fprintf(IO_Strom, "%d  Options\n", nr++);
     fprintf(IO_Strom, "%d  Specifications\n", nr++);
     fprintf(IO_Strom, "q  Quit\n");
     fprintf(IO_Strom, "c  Continue\n");
     fprintf(IO_Strom, "\n\n>>> ");
     fgets(text, INPUT_STRLEN, stdin);
     chop (text);
     if (strcmp(text,"q") == 0)
       exit(0);
     else if (strcmp(text, "c") == 0)
       zurueck = TRUE;
     else{
        eingabe = atoi(text);
        if (!FM_HilfeDatei())
           eingabe++;
        switch (eingabe){
        case 1:
           if (FM_HilfeDatei())
              IO_DateiAusgabe(FM_HilfeDatei(), TRUE);
           break;
        case 2:
        {
           int i, eingabe;
           unsigned int gruppe;
           
           do{
              IO_DruckeUeberschrift("Options");
              for (gruppe = (int)Arbeitsmodi; gruppe <= (int)ZusatzOptionen; gruppe++){
                 BOOLEAN EineZugaenglicheOptionGefunden = FALSE;
                 for (i=0; i<OptNrLAST; i++)
                    if ((AufrufOptionen[i].Gruppe == gruppe)&&OptionZugaenglich(i)){
                       EineZugaenglicheOptionGefunden = TRUE;
                       break;
                    }
                 if (EineZugaenglicheOptionGefunden){
                    IO_DruckeKurzUeberschrift(GruppenNamen[gruppe],"=");
                    for (i=0; i<OptNrLAST; i++)
                       if ((AufrufOptionen[i].Gruppe == gruppe)&&OptionZugaenglich(i))
                           DruckeOptionKurz(i);
                 }
              }
              IO_DruckeKurzUeberschrift("Miscellaneous", "=");
              fprintf(IO_Strom, "    def >>> program-call with current parameter-setting\n");
              if (XPRTModusSichtbar)
                 fprintf(IO_Strom, "      x >>> turn expert-mode %s\n", XPRTModus?"off":"on");
              fprintf(IO_Strom, "     ia >>> turn interactivity %s\n", PA_InteraktionAn?"off":"on");
              fprintf(IO_Strom, "optfile >>> parse more options from <file>\n");
              fprintf(IO_Strom, "\nFor explanation,");
              eingabe = OptionMitKuerzelWaehlen(TRUE);
              if (eingabe >= 0){
                 DruckeOption(eingabe);
                 IO_WarteAufReturnMitText("WEITER");
              }
           }while(eingabe != -2);
        }
        break;
        case 3: 
        {
           fprintf(IO_Strom, "\nIt is possible to view specifications during the selection.\n"
                   "From every menu you can 'view' the currently selected specification.\n"
                   "You can tell Waldmeister to look for specifications in certain directories\n"
                   "by setting the environment-variable WM_PATHS:\n"
                   "setenv WM_PATHS path1:path2:path3\n");
           IO_WarteAufReturnMitText("RETURN");
           if (FM_MusterDatei())
              IO_DateiAusgabe(FM_MusterDatei(), TRUE);
           break;
        }
        default :break;
        }
     }
  }while(!zurueck);
}


/* ------------------------------------------------------------------------*/
/* -------------------- Option per Kuerzel waehlen ------------------------*/
/* ------------------------------------------------------------------------*/

static int OptionMitKuerzelWaehlen(BOOLEAN NurErklaeren)
/* Eingabe einer Option, die durch das Kuerzel fn angegeben wurde. Beruecksichtigt
   werden nur aktive Optionen, die bzgl. XPRT-Modus zugaenglich sind.
   Bei Eingabe einer Pseudo-Option, deren Behandlung bereits hier erledigt wurde, wird -1 zurueckgegeben.
   Bei Eingabe von c oder q wird abgebrochen, in diesem Falle wird -2 zurueckgeliefert.
*/
{
   int option = 0;
   BOOLEAN Gefunden;

   do{
      fprintf(IO_Strom, "\nChoose option, continue with 'c'\n>>> ");
      fgets(Zeichenkette, INPUT_STRLEN, stdin);
      chop (Zeichenkette);
      fprintf(IO_Strom, "\n");
      if (strcmp("c", Zeichenkette)==0){
         Gefunden = TRUE;
         option = -2;
      }
      else if (strcmp("def", Zeichenkette)==0){
         Gefunden = TRUE;
         option = -1;
         if (NurErklaeren)
            fprintf(IO_Strom, "Current parameter setting as comand line.\n");
         else
            PA_DefaultsAusgeben("stdout");
         IO_WarteAufReturnMitText("RETURN");
      }
      else if (XPRTModusSichtbar && (strcmp("x", Zeichenkette)==0)){
         Gefunden = TRUE;
         option = -1;
         if (NurErklaeren){
            fprintf(IO_Strom, "Turn expert-mode on resp. off.\n");
            IO_WarteAufReturnMitText("RETURN");
         }
         else{
            XPRTModusWechseln();
            fprintf(IO_Strom, "EXPERT-MODE  : %s\n", XPRTModus?"An":"Aus");
         }
      }
      else if (strcmp("exp", Zeichenkette)==0){
         Gefunden = FALSE;
         if (NurErklaeren)
            fprintf(IO_Strom, "Unknown option.\n");
         else{
            do{
               fprintf(IO_Strom, "\nEnter option to be explained.\n>>> ");
               fgets(Zeichenkette, INPUT_STRLEN, stdin);
	       chop (Zeichenkette);
               fprintf(IO_Strom, "\n");
               option=0; 
               do
                  if ((strcmp((AufrufOptionen[option].Kuerzel+1), Zeichenkette) == 0) && OptionZugaenglich(option))
                     Gefunden = TRUE;
               while((!Gefunden)&&(++option<OptNrLAST));
               if (Gefunden) DruckeOption(option);
               else fprintf(IO_Strom, "Unknown option.\n");
            }while(!Gefunden);
            option = -1;
            Gefunden = TRUE;
            IO_WarteAufReturnMitText("RETURN");
         }
      }
      else if (strcmp("view", Zeichenkette)==0){
         if (NurErklaeren){
            fprintf(IO_Strom, "View currently selected specification.\n");
            IO_WarteAufReturnMitText("RETURN");
         }
         else
            IO_DateiAusgabe(IO_Spezifikationsname(), TRUE);

         option = -1;
         Gefunden = TRUE;
      }
      else if (strcmp("setup", Zeichenkette)==0){
         if (NurErklaeren){
            fprintf(IO_Strom, "Show current setup.\n");
            IO_WarteAufReturnMitText("RETURN");
         }
         else{
           IO_DruckeUeberschrift("SETUP");
           PA_DruckeModi();
           IO_WarteAufReturnMitText("RETURN");
         }
         option = -1;
         Gefunden = TRUE;
      }
      else if (strcmp("ia", Zeichenkette)==0){
         if (NurErklaeren){
            fprintf(IO_Strom, "Turn on resp. off interactivity (File-, Options-, Interrupt- and result-menus).\n");
            IO_WarteAufReturnMitText("RETURN");
         }
         else{
            PA_InteraktionAn = !PA_InteraktionAn;
            fprintf(IO_Strom, "OPTION    -ia: %s\n", PA_InteraktionAn?"An":"Aus");
         }
         Gefunden = TRUE;
         option = -1;
      }
      else if (strcmp("optfile", Zeichenkette)==0){
         if (NurErklaeren){
           fprintf(IO_Strom, "Parse more options from file.\n");
           IO_WarteAufReturnMitText("RETURN");
         }
         else{
           fprintf(IO_Strom, "Sorry, no longer available at this stage of computation.\n");
         }
         Gefunden = TRUE;
         option = -1;
      }
        
      else if (strcmp("quit", Zeichenkette)==0){
         if (NurErklaeren){
           fprintf(IO_Strom, "Quits the run. No further actions take place.\n");
           IO_WarteAufReturnMitText("RETURN");
         }
         else{
           exit(0);
         }
         Gefunden = TRUE;
         option = -1;
      }
        
      else{
         option=0;
         Gefunden = FALSE;
         do
            if ((strcmp((AufrufOptionen[option].Kuerzel+1), Zeichenkette) == 0) && OptionZugaenglich(option))
               Gefunden = TRUE;
         while((!Gefunden)&&(++option<OptNrLAST));
      }
      if (!Gefunden)
         fprintf(IO_Strom, "Unknown option.\n");
   }while(!Gefunden);
   return option;
}


/* ------------------------------------------------------------------------*/
/* -------------------- Einstellen von Optionen ---------------------------*/
/* ------------------------------------------------------------------------*/

static BOOLEAN NeuerWertVonToggle(BOOLEAN AlterWert, char *Argument, BOOLEAN *ArgumentVerbraucht,
                                  BOOLEAN EndeGelaende)
/* Liefert den neuen Wert des Toggles in Abhaengigkeit von Argument. */
{
   *ArgumentVerbraucht = FALSE;

   if (EndeGelaende) return !AlterWert;

   if (Argument!=NULL){
      if ((strcmp(Argument,"AN")==0)||(strcmp(Argument,"an")==0)||
          (strcmp(Argument,"ON")==0)||(strcmp(Argument,"on")==0)){
         AlterWert = TRUE;
         *ArgumentVerbraucht = TRUE;
      }
      else if ((strcmp(Argument,"AUS")==0)||(strcmp(Argument,"aus")==0)||
               (strcmp(Argument,"OFF")==0)||(strcmp(Argument,"off")==0)){
         AlterWert = FALSE;
         *ArgumentVerbraucht = TRUE;
      }
      else
         AlterWert = !AlterWert;
   }
   else 
      AlterWert = !AlterWert;      
   return AlterWert;
}

static BOOLEAN OptionEingestellt(AufrufOptionenTyp OptNr, BOOLEAN ArgumentAlsIntErlaubt, BOOLEAN AnAus, 
                                 char *Argument, BOOLEAN *ArgumentVerbraucht, BOOLEAN EndeGelaende)
{
   BOOLEAN Gefunden = FALSE;

   *ArgumentVerbraucht = TRUE;

   fprintf(IO_Strom, "  OPTION %8s: ", AufrufOptionen[OptNr].Kuerzel);

   switch (AufrufOptionen[OptNr].ParameterArt){
   case 0:
      if (Argument){
         AufrufOptionen[OptNr].ParameterTyp.frei.wurdeAngegeben = TRUE;
         AufrufOptionen[OptNr].ParameterTyp.frei.eingestellterWert = (char *) our_alloc((1+strlen(Argument))*sizeof(char));
         strcpy(AufrufOptionen[OptNr].ParameterTyp.frei.eingestellterWert, Argument);
         fprintf(IO_Strom, "%s ", Argument);
         Gefunden = TRUE;
      }
      else{
         fprintf(IO_Strom, "*** Missing argument ");
         *ArgumentVerbraucht = FALSE;
      }
      if (AufrufOptionen[OptNr].NurVervollstaendigung && !Vervollstaendigung)
         fprintf(IO_Strom, "*** will be ignored in this mode.\n");
      else
         fprintf(IO_Strom, "\n");
      break;
   case 1:
      AufrufOptionen[OptNr].ParameterTyp.einstellig.eingestellterWert = 
         NeuerWertVonToggle(AufrufOptionen[OptNr].ParameterTyp.einstellig.eingestellterWert, 
                            AnAus?Argument:NULL, ArgumentVerbraucht, EndeGelaende);
      fprintf(IO_Strom, "%s ",(AufrufOptionen[OptNr].ParameterTyp.einstellig.eingestellterWert)?
             AufrufOptionen[OptNr].ParameterTyp.einstellig.BedeutungT:AufrufOptionen[OptNr].ParameterTyp.einstellig.BedeutungF);
      Gefunden = TRUE;
      if (AufrufOptionen[OptNr].NurVervollstaendigung && !Vervollstaendigung)
         fprintf(IO_Strom, "*** will be ignored in this mode.\n");
      else
         fprintf(IO_Strom, "\n");
      break;
   case 2:
   {
      int wert , i;
      if (ArgumentAlsIntErlaubt && ((wert= atoi(Argument))>0) &&
          (wert<=HoechsterSichtbarerIndex(OptNr))){
         AufrufOptionen[OptNr].ParameterTyp.zweistellig.eingestellterWert = wert-1;
         Gefunden = TRUE;
      }
      else
         if (Argument)
            for (i=0; i<HoechsterSichtbarerIndex(OptNr); i++)
               if (strstr(AufrufOptionen[OptNr].ParameterTyp.zweistellig.WerteImKlartext[i], Argument)){
                  AufrufOptionen[OptNr].ParameterTyp.zweistellig.eingestellterWert = i;
                  Gefunden = TRUE;
                  break;
               }
      if (Gefunden)
         fprintf(IO_Strom, "%s ", AufrufOptionen[OptNr].
                ParameterTyp.zweistellig.WerteImKlartext[AufrufOptionen[OptNr].
                                                         ParameterTyp.zweistellig.eingestellterWert]);
      else if (Argument){
         *ArgumentVerbraucht = FALSE;
         fprintf(IO_Strom, "%s *** unknown value ", Argument);
      }
      else{
         *ArgumentVerbraucht = FALSE;
         fprintf(IO_Strom, "*** missing argument ");
      }

      if (OptNr == OptNrAusgabeLevel)
        IO_AusgabeLevelBesetzen(atoi(AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.
           WerteImKlartext[AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.eingestellterWert]));

      if (AufrufOptionen[OptNr].NurVervollstaendigung && !Vervollstaendigung)
         fprintf(IO_Strom, "*** will be ignored in this mode.\n");
      else
         fprintf(IO_Strom, "\n");
      break;
   }
   case 3:
      if (Argument){
         AufrufOptionen[OptNr].ParameterTyp.InfoString.UserInfoString = (char *) our_alloc((1+strlen(Argument))*sizeof(char));
         strcpy(AufrufOptionen[OptNr].ParameterTyp.InfoString.UserInfoString, Argument);
         fprintf(IO_Strom, "%s ", Argument);
         Gefunden = TRUE;
      }
      else{
         fprintf(IO_Strom, "*** missing argument ");
         *ArgumentVerbraucht = FALSE;
      }
      if (AufrufOptionen[OptNr].NurVervollstaendigung && !Vervollstaendigung)
         fprintf(IO_Strom, "*** will be ignored in this mode.\n");
      else
         fprintf(IO_Strom, "\n");
      break;
      
   default:
      *ArgumentVerbraucht = FALSE;
      our_error("\nError in SetOption.\n\n");
      break;
   }
   return Gefunden;
}
   

/* ------------------------------------------------------------------------*/
/* -------------------- Konsistenztest fuer Einstellung -------------------*/
/* ------------------------------------------------------------------------*/

static BOOLEAN ParameterKonsistent(void)
/* Testet die innere Konsistenz bzw. Gueltigkeit der Aufruf-Parameter.
   Ferner werden hier voneinander abhaengige Parameter gesetzt.
*/
{
   if (Vervollstaendigung){

/*      PA_KapurR = (BOOLEAN) strstr(AufrufOptionen[OptNrKapur].ParameterTyp.frei.eingestellterWert, "r");
      PA_KapurG = (BOOLEAN) strstr(AufrufOptionen[OptNrKapur].ParameterTyp.frei.eingestellterWert, "g");
      PA_Kapur = PA_KapurR | PA_KapurG;*/
   
      VerwendungVonGleichungen = !PA_FailingCompletion;
   }
   else {                       /* Einfache Modi: keine KP-Behandlung bei ihrer Entstehung, keine Interreduktion */

/*      PA_RMrechtsB  = PA_RMrechtsBR = PA_RMrechtsBG = FALSE;*/
      
      VerwendungVonGleichungen = FALSE;
   }

   return TRUE;
}

/* ------------------------------------------------------------------------*/
/* -------------------- Parameter-Einstellung -----------------------------*/
/* ------------------------------------------------------------------------*/

static void PA_ParameterAuswaehlen(void)
/* Manuelles Einstellen der Parameter (menuegesteuert!!!!)*/
{
   int i, eingabe;
   unsigned int gruppe;
   BOOLEAN Booly;

   do{
      IO_DruckeUeberschrift("Waldmeister Interactive Parameter Menue");
      for (gruppe = (int)Arbeitsmodi; gruppe <= (int)ZusatzOptionen; gruppe++){
         BOOLEAN ZugaenglicheOptionGefunden = FALSE;
         for (i=0; i<OptNrLAST; i++)
            if ((AufrufOptionen[i].Gruppe == gruppe)&&OptionZugaenglich(i)){
               ZugaenglicheOptionGefunden = TRUE;
               break;
            }
         if (ZugaenglicheOptionGefunden){
            IO_DruckeKurzUeberschrift(GruppenNamen[gruppe], "=");
            for (i=0; i<OptNrLAST; i++)
               if ((AufrufOptionen[i].Gruppe == gruppe)&&OptionZugaenglich(i))
                  DruckeOptionKurz(i);
         }
      }
      IO_DruckeKurzUeberschrift("Sonstiges", "=");
      fprintf(IO_Strom, "    def >>> current parameter-setting as command-line\n");
      if (XPRTModusSichtbar)
         fprintf(IO_Strom, "      x >>> turn expert-mode %s\n", XPRTModus?"off":"on");
      fprintf(IO_Strom, "     ia >>> turn interactivity %s\n", PA_InteraktionAn?"off":"on");
      fprintf(IO_Strom, "    exp >>> explain option\n");
      fprintf(IO_Strom, "   view >>> view selected spezification\n");
      fprintf(IO_Strom, "  setup >>> Show current parameter-setup\n");
      fprintf(IO_Strom, "   quit >>> Quit Waldmeister.\n");
      eingabe = OptionMitKuerzelWaehlen(FALSE);
      if (eingabe >= 0){
         if (!AufrufOptionen[eingabe].NurVervollstaendigung || Vervollstaendigung){
            switch (AufrufOptionen[eingabe].ParameterArt){
            case 0:
            case 3:
               SchreibeStringOhneHochkommataNachZeichenkette(AufrufOptionen[eingabe].Name);
               fprintf(IO_Strom, "\nEnter argument for option '%s': ", Zeichenkette);
               fgets(Zeichenkette, INPUT_STRLEN, stdin);
	       chop (Zeichenkette);
            case 1:
               OptionEingestellt(eingabe, FALSE, FALSE, Zeichenkette, &Booly, FALSE);
               break;
            case 2:
               do{
                  SchreibeStringOhneHochkommataNachZeichenkette(AufrufOptionen[eingabe].Name);
                  fprintf(IO_Strom, "\n%s (currently %s): \n",
                         Zeichenkette, 
                         AufrufOptionen[eingabe].ParameterTyp.zweistellig.WerteImKlartext[AufrufOptionen[eingabe].ParameterTyp.zweistellig.eingestellterWert]);
                  for (i=1; i<=HoechsterSichtbarerIndex(eingabe); i++)
                     fprintf(IO_Strom, "%2d. %s\n", i, AufrufOptionen[eingabe].ParameterTyp.zweistellig.WerteImKlartext[i-1]);
                  fprintf(IO_Strom, ">>> ");
		  fgets(Zeichenkette, INPUT_STRLEN, stdin);
		  chop (Zeichenkette);
                  fprintf(IO_Strom, "\n");
               }while(!OptionEingestellt(eingabe, TRUE, FALSE, Zeichenkette, &Booly,FALSE));
               break;
            default:
               our_error("\nFauler Fehler in Optionen eingeben.\n\n");
               break;
            }
         }
         else{
            fprintf(IO_Strom, "Option will be ignored in this mode.\n");
            IO_WarteAufReturnMitText("RETURN");
         }
      }
      if (!ParameterKonsistent()){
         IO_WarteAufReturnMitText("Please correct.");
         eingabe = 1;           /* Damit nicht abgebrochen wird */
      }
   }while  (eingabe >= -1);
}   

BOOLEAN PA_Vorbehandeln(int argc, char *argv[])
/* - Reaktion auf Hilfe-Anforderung
   - Einstellung der Pseudo-Parameter, soweit moeglich
   - Einlesen von Optionen aus Datei
   Rueckgabe ist FALSE, wenn das Vorbehandeln bereits einen Abbruchgrund lieferte.
   Die vernutzten Argumente werden auf NULL gesetzt.
*/
{
   BOOLEAN Fehler = FALSE, protdateiAngegeben = FALSE;
   int pos=1;

   HilfeVerlangt = FALSE;

   /* argv[0] == Programmname,  deshalb ist argc >= 2 (sonst wird diese Prozedur nicht aufgerufen aus main.c) */
   fprintf(IO_Strom, "\npreparsing options");

   while (pos<argc){
      if (strcmp(argv[pos], "-def") ==0){
         if (pos+1<argc)
            PA_DefaultsAusgeben(argv[pos+1]);
         else
            PA_DefaultsAusgeben("stdout");
         exit(0);
      }
      else if ((strcmp(argv[pos], "-help") == 0) ||                        /* Hilfe verlangt? */
          (strcmp(argv[pos], "-h") == 0) ||
          (strcmp(argv[pos], "help") == 0)){
         HilfeVerlangt = TRUE;
         fprintf(IO_Strom, "  HELP \n");
         argv[pos] = NULL;
         pos++;
      }
      else if ((strcmp(argv[pos], "-ia") == 0)){                        /* Hilfe verlangt? */
         BOOLEAN Verbraucht;
         PA_InteraktionAn = NeuerWertVonToggle(PA_InteraktionAn, argv[pos+1], &Verbraucht, (pos>=argc));
         argv[pos] = NULL;
         fprintf(IO_Strom, "  INTERACTION  : %s\n", PA_InteraktionAn?"An":"Aus");
         if (Verbraucht){
            pos++;
            argv[pos] = NULL;
         }
         pos++;
      }
      else if ((strcmp(argv[pos], "-x") == 0) && XPRTModusSichtbar){
         BOOLEAN ArgVerbraucht;
         XPRTModus = NeuerWertVonToggle(XPRTModus, (pos<argc-1)?argv[pos+1]:NULL, &ArgVerbraucht, (pos>=argc));
         argv[pos]=NULL;
         if (ArgVerbraucht){
            pos++;
            argv[pos]=NULL;
         }
         fprintf(IO_Strom, "  EXPERT-MODE  : %s\n", XPRTModus?"An":"Aus");
         pos++;
      }
      else if ((strcmp(argv[pos], "-prot") == 0)){
         argv[pos]=NULL;
         if (pos+1<argc){
            ProtokolldateiName = argv[pos+1];
            pos++;
            argv[pos]=NULL;
            protdateiAngegeben = TRUE;
            fprintf(IO_Strom, "  PROTOCOL     : %s\n", ProtokolldateiName);
         }
         pos++;
      }
      else                      /* sonst: Weiterlaufen */
         pos++;
   }
   
   if (protdateiAngegeben) PA_InteraktionAn = FALSE;

   if (!isatty(STDIN_FILENO))/* 960825 AJ: Interaktion macht nur Sinn, wenn Benutzer tippen kann */
     PA_InteraktionAn = FALSE;
   
   if ((Fehler || HilfeVerlangt) && PA_InteraktionAn()) Hilfe();

   fprintf(IO_Strom, "...done\n");

   return Fehler;
}


BOOLEAN PA_AufrufParameterEinstellen(int argc, char *argv[], mode Modus)
/* Einstellen der beim Aufruf mitgegebenen Parameter. argc und argv sind die Parameter von main,
   es wird also davon ausgegangen, dass argv[0] den Programmnamen enthaelt, und das argv[argc] der
   Name der 21zu bearbeitenden Spezifikation ist. Die Angaben zwischen argv[1] und argv[argc-1] werden
   ausgewertet. Dabei werden die in Parameter.c spezifizierten Aufruf-Optionen herangezogen.
   Inkorrekte Parameter sowie der Filename werden ueberlesen.
   Auf Anforderung von Hilfe sowie Menue wird reagiert.
   Rueckgabe ist, ob - ggf. mit Menue-Hilfe - eine korrekte Einstellung vorgenommen werden konnte.
   Vor.: Die Spezifikation befindet sich bereits im Speicher.
*/
{
   BOOLEAN Gefunden = FALSE, PseudoOption, ArgumentVerbraucht, MenueVerlangt = FALSE;
   int pos=1, option=0, aktuelleOption = 0, AnzahlFehler = 0;
   
   /* argv[0] == Programmname,  deshalb ist argc >= 2 (sonst wird diese Prozedur nicht aufgerufen aus main.c) */

   PA_ExtModus = Modus;

   Vervollstaendigung = (PA_ExtModus == proof) || (PA_ExtModus == completion);

   if (Vervollstaendigung && !VervollstaendigungZugaenglich){
      fprintf(IO_Strom, "Completion not available!\n\n");
      return FALSE;
   }

   PA_Ordering = ord_name(get_ordering());

   SO_InitAposteriori();

   TO_InitAposteriori();

   IO_InitAposteriori(atoi(AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.
         WerteImKlartext[AufrufOptionen[OptNrAusgabeLevel].ParameterTyp.zweistellig.eingestellterWert]),
         ProtokolldateiName);

   fprintf(IO_Strom, "\nparsing options\n");

   while (pos<argc){

      /* Pseudo-Optionen finden */
      /* ---------------------- */

      PseudoOption = Gefunden = TRUE;
      if (argv[pos] == NULL){
         pos++;
         continue;
      }
      if ((strcmp(argv[pos], "-menue") == 0) ||                            /* Menue verlangt? */
          (strcmp(argv[pos], "menue") == 0) ||
          (strcmp(argv[pos], "-m") == 0)){
         fprintf(IO_Strom, "  MENUE\n");
         MenueVerlangt = TRUE;
      }
      else{                                                                /* sonst: Option suchen */

         /* Richtige Option finden */
         /* ---------------------- */

         PseudoOption = Gefunden = FALSE;
         for (Gefunden = FALSE, option=0; (!Gefunden)&&(option<OptNrLAST); option++)
            if ((strcmp(AufrufOptionen[option].Kuerzel, argv[pos]) == 0) && (OptionZugaenglich(option))){
               Gefunden = TRUE;
               aktuelleOption = option;
            }
      }

      pos++;                                       /* pos steht jetzt auf dem naechsten Element der Kommando-Zeile */

      /* unbekannte Option */
      /* ----------------- */

      if (!Gefunden){
         fprintf(IO_Strom, "Unknown option or invalid filename: %s\n", argv[pos-1]);
         AnzahlFehler++;
      }
      else if (!PseudoOption)

         /* Richtige Option einstellen */
         /* -------------------------- */
         
         if (!OptionEingestellt(aktuelleOption, FALSE, TRUE, argv[pos], &ArgumentVerbraucht, (pos>=argc)))
            AnzahlFehler++;  /* Bei Fehler wird kein Argument vernutzt! */
         else if (ArgumentVerbraucht) 
            pos++;

   }
   
   fprintf(IO_Strom, "... done\n");

   PA_FailingCompletion = AufrufOptionen[OptNrFailing].ParameterTyp.einstellig.eingestellterWert;
   PA_PostponingAn = AufrufOptionen[OptNrPostponing].ParameterTyp.einstellig.eingestellterWert;


   if ((!ParameterKonsistent()) || (AnzahlFehler>0) || MenueVerlangt || HilfeVerlangt)
      if (PA_InteraktionAn){
        IO_DruckeKopf();             /* Programm-Kopf ausgeben */
        IO_DruckeBeispiel();
        PA_ParameterAuswaehlen();
        return TRUE;
      }
      else
         return FALSE;
   else
      return TRUE;
   
}


