/*
01.03.94 Angelegt.
02.03.94 Weitergemacht. Optionen eingebaut. Arnim.
04.03.94 Aufruf mit symbolischen Dateinamen eingebaut. Arnim.
07.03.94 KPHeuristik als Aufrufoption eingebaut.
*/

/* ------------------------------------------------*/
/* ------------- Includes -------------------------*/
/* ------------------------------------------------*/
#include "antiwarnings.h"
#include <string.h>
#include <errno.h>
#include <stdarg.h>

#include <stdlib.h>
#include <string.h>
#include <dirent.h>
#include <signal.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>

#include "compiler-extra.h"
#include "general.h"
#include "Ausgaben.h"
#include "DomaenenMemorium.h"
#include "FehlerBehandlung.h"
#include "FileMenue.h"
#include "Hauptkomponenten.h"
#include "Initialisierungen.h"
#include "MNF.h"
#include "Ordnungsgenerator.h"
#include "Parameter.h"
#include "pcl.h"
#include "Puffer.h"             /* wg. Puffer leeren */
#include "Statistiken.h"
#include "SUEM.h"
#include "WaldmeisterII.h"
#include "ZeitMessung.h"
#include "Zielverwaltung.h"

#define CASC14

/* ------------------------------------------------*/
/* -------- lokale Prozeduren ---------------------*/
/* ------------------------------------------------*/

/* zum Suchen in den angegebenen Beispielverzeichnissen.*/

static const char  *LaufErgebnis;         /* Ergebnis des Laufes im Klartext */
BOOLEAN LaufBeendet = FALSE, Resultat;

/* Signalhandler */
/* ============= */
#if defined (SUNOS4)
/* SUNOS4 kennt SA_RESTART nicht, da dies standardmaessig gesetzt
   ist, stattdessen gibt's SA_INTERRPUT mit der negierten Bedeutung */
# define SIGNAL_DEFINITION(name)    static void name (void)
# define SA_RESTART 0
#else
# define SIGNAL_DEFINITION(name)    static void name (int signal MAYBE_UNUSEDPARAM)
#endif

SIGNAL_DEFINITION (ontimeout)
{
   if (!LaufBeendet){           /* Falls gerade eben Beweis gefunden ==> Lauf regulaer beenden */
      ZM_StoppGesamtZeit();
      ST_TimeoutStatistik("Timeout", FALSE);
   }
   else
      return;
   /* exit-Werte:
      0 - success;
      1 - failure;
      2 - timeout;
      3 - error;
      */
   exit(2);
}

SIGNAL_DEFINITION (onquit)
{
   FB_InterruptflagSetzen();
}

/* Ausgeben der AutoStatistik */
SIGNAL_DEFINITION (onusr1)
{
   FB_InterruptflagSetzen();
   FB_Usr1InterruptflagSetzen();
}



/* Signalbehandlung:
   Es werden SIGUSR1, SIGINT, SIGVTALRM, SIGALRM behandelt.
   SIGSEGV wird in Strahlmemorium behandelt.
   Reaktion auf SIGUSR1 ist Ausgabe von einigen Infos zum Lauf - ohne Anzuhalten.
   Reaktion auf SIGINT ist Ausgabe des Menues.
   
   Reaktion auf Interrupt erfolgt nur, wenn 
   - Interaktion gewuenscht ist,
   - nicht in Datei geschrieben wird und
   - es sich um einen Vervollstaendigungs- oder Proof-Lauf handelt.
*/


static void SignalsSetzen (void) 
{
   struct itimerval timer;
   struct sigaction new_action;

   sigemptyset (&new_action.sa_mask);

   if (PA_ZeitlimitGesetzt()){
      timer.it_interval.tv_sec  =  0;
      timer.it_interval.tv_usec =  0;
      
      timer.it_value.tv_sec  =  PA_Zeitlimit();
      timer.it_value.tv_usec =  0;

      new_action.sa_handler = ontimeout;
      new_action.sa_flags =  SA_RESETHAND|SA_RESTART; /* Signalhandler wird nur einmal aufgerufen */
      sigaction (SIGVTALRM, &new_action, NULL);
      setitimer (ITIMER_VIRTUAL, &timer, NULL);
   }
   if (PA_RealZeitlimitGesetzt()){
      /* Problem: Wenn Zeitlimit nicht erreicht wird, z.B. wegen
	 Swapping oder anderer Systemaktivitaeten, dann laeuft der Prozess
	 ein mehrfaches der gewuenschten Zeit. Deswegen wird noch ein
	 zweiter Timer mit einem Timelimit in realer Zeit gesetzt,
	 der dafuer sorgt, dass der Prozess wirklich terminiert.
	 Achtung: ITIMER_REAL wird auch von alarm gebraucht - also
	 auf keinen Fall alarm benutzen!
	 */

      timer.it_interval.tv_sec  =  0;
      timer.it_interval.tv_usec =  0;

      timer.it_value.tv_sec  =  PA_RealZeitlimit();
      timer.it_value.tv_usec =  0;

      new_action.sa_handler = ontimeout;
      new_action.sa_flags =  SA_RESETHAND|SA_RESTART; /* Signalhandler wird nur einmal aufgerufen */

      sigaction (SIGALRM, &new_action, NULL);
      setitimer (ITIMER_REAL, &timer, NULL);
   }
   
   FB_InterruptflagLoeschen();  /* Mit FALSE initialisieren */
   if (PA_InteraktionAn() && ((PA_ExtModus == completion) || PA_ExtModus == proof)) {
      new_action.sa_handler = onquit;
      new_action.sa_flags =  SA_RESTART; /* interrupted system calls werden neu gestartet */
      sigaction (SIGINT, &new_action, NULL);
   }
   FB_Usr1InterruptflagLoeschen();  /* Mit FALSE initialisieren */
   new_action.sa_handler = onusr1;
   new_action.sa_flags =  SA_RESTART; /* interrupted system calls werden neu gestartet */
   sigaction (SIGUSR1, &new_action, NULL);
}

static void SignalsDeaktivieren (void) 
{
   struct itimerval timer;
   struct sigaction new_action;

   timer.it_interval.tv_sec  =  0;
   timer.it_interval.tv_usec =  0;
   
   timer.it_value.tv_sec  =  0;
   timer.it_value.tv_usec =  0;

   sigemptyset (&new_action.sa_mask);
   new_action.sa_handler = SIG_DFL;
   new_action.sa_flags =  0;

   if (PA_ZeitlimitGesetzt()){
      setitimer (ITIMER_VIRTUAL, &timer, NULL);
      sigaction (SIGVTALRM, &new_action, NULL);
   }

   if (PA_RealZeitlimitGesetzt()){
      setitimer (ITIMER_REAL, &timer, NULL);
      sigaction (SIGALRM, &new_action, NULL);
   }
   
   FB_InterruptflagLoeschen();  /* Mit FALSE initialisieren */
   sigaction (SIGINT, &new_action, NULL);
   
   sigaction (SIGUSR1, &new_action, NULL);
}


/* ------------------------------------------------*/
/* ---------- DAS HAUPTPROGRAMM -------------------*/
/* ------------------------------------------------*/

int    
#if COMP_WMTW
    WM_main
#else
       main
#endif
            (int argc, char *argv[])
{
   int i;
   char *Spezifikation;
   BOOLEAN AbbruchWegenGarNichtErstGestartet, SVStatistikZeigen, quitDesired;
   char newargv[INPUT_STRLEN];
   char again[INPUT_STRLEN];
   int CycleCount = 0;

   do{
     /*************************
      Variablen-Initialisierung
      *************************/

     Spezifikation = NULL;
     AbbruchWegenGarNichtErstGestartet = quitDesired = FALSE;
     LaufErgebnis = "Fehler! Wert des Strings 'LaufErgebnis' nicht ueberschrieben!";
     again[0]='n';
     again[1]='\0';
     
     Resultat = FALSE;
     
     /************************
       Vorab-Initialisierung
      *************************/
     if (CycleCount==0){
       IN_AllesAprioriInitialisieren(); /* Alle Module initialisieren. */
#ifndef CASC14
       FM_InitApriori(argv[0]);         /* File-Menues mit AufrufNamen initialisieren */
#endif
     }
     else
       IN_ReAllesAprioriInitialisieren();

     /************************************************************
       Aufruf abarbeiten: Spezifikation und Parameter einstellen
      *************************************************************/

#ifndef CASC14
     fprintf(IO_Strom, "\nReproducing command-line: ");  /* Aufruf reproduzieren (Kontrolle fuer User) */
     for(i=0; i<argc; i++)
       fprintf(IO_Strom, "%s ", argv[i]);
     fprintf(IO_Strom, "\n...done\n");

     argv = FM_NeueParameterliste(&argc, argv);
     /* argv ggf. um die Parameter aus Optfiles ergaenzen. */
#endif

     PA_Vorbehandeln(argc, argv); 
     /* beim ersten Mal: help, interaction, protocol etc. */
     /* bei weiteren Malen ggf. weniger */

#ifdef CASC14
     if (argc!=2) {             /* Fuer Geoff: Auch ohne '.pr' */
       printf("ERROR: Please use only one argument!\n\n");
       exit(-1);
     }
     else{
       Spezifikation = argv[1];
       argv[1]=NULL;
     }
#else
     Spezifikation = FM_SpezifikationErkannt(argc, argv, PA_InteraktionAn());
#endif

     if (Spezifikation == NULL){
       fprintf(IO_Strom, "\n\nTerminated: no specification!\n\n");
       exit(1);
     }

     /*********************************
       Gewaehlte Spezifikation einlesen
      ********************************/

     IO_SpezifikationsnamenSetzen(Spezifikation);

     ZM_StartPP();                /* Zeit-Start fuer Parsen und Plattklopfen */
     IN_SpezifikationEinlesen(IO_Spezifikationsname());
     ZM_StoppPP();

     /****************************************
       Aufruf abarbeiten: Parameter einstellen
      ***************************************/

     /* Betroffene Routinen merken sich ihre Initialisiertheit! */

     if (!PA_AufrufParameterEinstellen(argc,argv, get_mode())){
       fprintf(IO_Strom, "\nProgrammende wegen inkonsistenter Aufrufparameter!\n\n");
       exit(1);
     }
       
     /**************************************************
       Gewaehlte Spezifikation in Systemstrukturen laden
      *************************************************/

     ZM_StartPP();
        
     PA_InitAposteriori(); /* System gemaess Parameter-Einstellung initialisieren. */

#ifndef CASC14
     IO_DruckeZeile("+");

     if (PA_InteraktionAn()){
       char eing[INPUT_STRLEN];
       do{
         printf("\nConfiguration ok (y/n)?\n>>> ");
         fgets(eing, INPUT_STRLEN, stdin);
       }while ((eing[0] != 'j') && (eing[0] != 'n') && (eing[0] != 'y') && (eing[0]!= 'c'));
       if (eing[0] == 'n') {
         printf("\n\n... try it again, Sam.\n\n");
         exit(0);
       }
     }
#endif

     HK_GleichungenUndZieleHolen(); /* Gleichungen und Ziele in Systemstrukturen uebernehmen */
     /* (plattklopfen) */

     OG_OrdnungGenerieren();

     IN_SpezifikationGelesen();

     ZM_StoppPP();

     /*************************************************************************
       Eigentlicher Programm-Lauf nach Art der Spezifikationsdatei
      *************************************************************************/

     SignalsSetzen ();
   
     ZM_StartGesamtZeit();        /* Gesamtzeit starten */

     if (SO_PraezedenzTotal()){
       if (Z_ZielmengeLeer()){
         LaufErgebnis = "proof mode: no goals.";
         AbbruchWegenGarNichtErstGestartet = TRUE;
       }
       else {
         HK_Vervollstaendigung();
         LaufErgebnis =  Z_ZieleBewiesen() ? "GOAL PROVED" : "NO PROOF FOUND";
       }
     }
     else{
       LaufErgebnis = "precedence not total";
       AbbruchWegenGarNichtErstGestartet = TRUE;
     }
     
     LaufBeendet = TRUE;
     
     ZM_StoppGesamtZeit();
     
     SignalsDeaktivieren ();
     

     /**************************************************
       Ausgabe der Ergebnisse
      ***************************************************/

     RE_DruckeEndsystemFuerPCL(); /* Fuer PCL-Modus ggf. Abschlussfakten drucken */

     PCL_Beenden();               /* PCL-Protokoll abschliessen */

     /**************************************************
       Statistiken
      ***************************************************/

     if (AbbruchWegenGarNichtErstGestartet)
       ST_ErrorStatistik(LaufErgebnis, Resultat);
     else
       ST_SuccessStatistik(LaufErgebnis, Resultat);

   
     /**************************************************
       Speicher aufraeumen
      ***************************************************/
     
#ifndef CASC14
     IO_AusgabenAus();     /* Verbliebene Strukturen loeschen */
     RE_RUndETest();
     SVStatistikZeigen=FALSE;
     if (PA_InteraktionAn() && PA_XPRTModusSichtbar()){
       
       printf("\n\nTo proceed, enter new command line (memtest for SV_Statistik)\n"
              "To exit with memtest enter `q'.\n>>> ");
       fgets(newargv, INPUT_STRLEN, stdin);
       printf("\n\n");
     }
     else{
       newargv[0]='q';
       newargv[1]='\0';
     }
     
     quitDesired = (strspn(newargv,"q")>0);
     SVStatistikZeigen= quitDesired || (strstr(newargv, "memtest"));


     if (quitDesired || SVStatistikZeigen){
       char *where = strstr(newargv, "memtest"); /* pattern memtest wieder aus newargv loeschen */
       if (where) strncpy(where, "          ", 7);

       Z_Cleanup();    /* Ziele loeschen */
       MNF_Cleanup();
       SUE_CLEANUP();         /* muss vor dem Loeschen der R- bzw. E-Menge passieren */
       
       RE_LoescheGleichungsmenge();
       RE_LoescheRegelmenge();   /* E- und R-Menge loeschen */
       
       PU_TPPufferLoeschen();
       TO_FeldVariablentermeLeeren(); /* ... */
     }
     else{
       SUE_SOFTCLEANUP();
       SV_AlleManagerLeeren();
     }

     IO_AusgabenEin();
     ZM_ZeitstackOK();         /* Zeitstack pruefen */
     SO_CleanUp();
#endif
     
       
     if (SVStatistikZeigen)      /* bei stop und 'm' */
       SV_Statistik(FALSE, TRUE, "complete");      /* FALSE: kein System-Fehler aufgetreten */
     
#ifndef CASC14
     
     fprintf(IO_Strom, "Insgesamt %lu Abfragen auf angemessene Stapelausdehnung bei TIx_Pushxx\n",maekmaek());
     IO_CleanUp(); /* falls Ausgaben umgelenkt waren, die Protokoll-Datei schliessen */
     
     /**************************************************
       Ausgaben beenden
       ***************************************************/
      
     IO_DruckeKurzUeberschrift(LaufErgebnis, "+");

     if (!quitDesired){ /* nicht anhalten */
       char *cptr, *dummy = newargv;
       int newargc = 0;
       if ((cptr=strstr(newargv, "\n"))) *cptr='\0'; /* newline ueberschreiben */
       while ((cptr=strtok(dummy," "))){
         dummy=NULL;
         argv[++newargc] = cptr;
         printf("%s ", argv[newargc-1]);
       }
       argc = newargc + 1;
       clear_spec();
     }      
#else 
     quitDesired = TRUE;
#endif

     CycleCount++;
   }while(!quitDesired);
   
   exit(0);                    /* Programm-Ende */
}

