/*
Dieses Modul enthaelt die Ausgabe-Funktionen sowie alle 'Bezeichner' der eingelesenen Spezifikation.

28.02.94 Angelegt.
02.03.94 DruckeKopf und DruckeStatistik eingebaut.

*/
/*
  ToDo:
  Anpassung an englische Geflogenheiten: Decimalpunkt, Tausenderkomma?
  */
/* -------------------------------------------------------*/
/* ------------- Includes --------------------------------*/
/* -------------------------------------------------------*/
#include "antiwarnings.h"
#include <signal.h> 
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#if defined (LINUX) || defined (SUNOS5)
#include <sys/wait.h>
#endif

#include "Ausgaben.h"
#include "FehlerBehandlung.h"
#include "general.h"
#include "Parameter.h"
#include "parser.h"
#include "SymbolOperationen.h"
#include "TermOperationen.h"
#include "TermIteration.h"
#include "version.h"

TI_IterationDeklarieren(IO)

unsigned long int IO_AusdehnungAusgabestapel(void)
{
   return IO_AusdehnungKopierstapel();
}

/* ------------------------------------------------------------------------*/
/* -------------------- globale Variablen (siehe Ausgaben.h) --------------*/
/* ------------------------------------------------------------------------*/
BOOLEAN IO_level0, IO_level1, IO_level2, IO_level3, IO_level4;
BOOLEAN IO_GUeberschrift, IO_TWUeberschrift;
FILE *IO_Strom;
long IO_Nr;
unsigned long int IO_AnzAng;

/* -------------------------------------------------------*/
/* ------------- Globals ---------------------------------*/
/* -------------------------------------------------------*/

static char *Spezifikationsname = NULL;
static char **IO_SymbolNamen;
static char IO_SpezName[MaximaleBezeichnerLaenge];
static char IO_Modus[MaximaleBezeichnerLaenge];
static int Ausgabelevel;

static BOOLEAN LatexAusgabeAktiviert = FALSE;

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* +++++++++++++ AusgabeStrom steuern ++++++++++++++++++++*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

static FILE *AlterStrom = NULL;

void IO_AusgabeStromUmlenken(const char *Filename)
{
   if (AlterStrom != NULL)
      our_fatal_error("Ausgabestrom zweimal umgelenkt!");
   else{
      AlterStrom = IO_Strom;
      if (!(IO_Strom = fopen(Filename, "a"))){
         IO_Strom = AlterStrom;
         AlterStrom = NULL;
         fprintf(IO_Strom, "\nAusgabe konnte nicht umgelenkt werden!\n\n");
      }
   }
}

void IO_AusgabeStromZuruecksetzen(void)
{
   if (AlterStrom == NULL)
      our_fatal_error("Ausgabestrom nicht umgelenkt -> kein Zuruecksetzen moeglich!\n");
   else{
      fclose(IO_Strom);
      IO_Strom = AlterStrom;
      AlterStrom = NULL;
   }
}

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* +++++++++++++ Initialisierung +++++++++++++++++++++++++*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void IO_InitApriori(void)
/* Vor der ersten Ausgabe auf IO_Strom aufzurufen!!!! */
{
   IO_Strom = stdout;
}

BOOLEAN IO_SymbolErkannt(char *str, SymbolT *Symbol)
/* Liefert TRUE, wenn str dem Ausgabe-String eines Symbols aus der Spezifikation oder aus der Paramodulation entspricht. */
{
   SymbolT i;

   SO_forFktSymbole(i)
      if (strcmp(str, IO_SymbolNamen[i]) == 0){
         *Symbol = i;
         return TRUE;
      }
   return FALSE;
}


static void ParamodSymbolnamenEtablieren(SymbolT ParamodSymbol, const char *Name)
{
  unsigned int n = 0;
  IO_SymbolNamen[ParamodSymbol] = our_alloc(MaximaleBezeichnerLaenge*sizeof(char *));
  while (TRUE) {
    SymbolT i;
    BOOLEAN ok = TRUE;
    if (n)
      sprintf(IO_SymbolNamen[ParamodSymbol],"%s%d",Name,n);
    else
      sprintf(IO_SymbolNamen[ParamodSymbol],"%s",Name);
    SO_forSpezSymbole(i)
      if (strcmp(IO_SymbolNamen[i],IO_SymbolNamen[ParamodSymbol])==0) 
        ok = FALSE;
    if (ok)
      return;
    n++;
  }
}

void IO_CleanUp(void)
{
  SymbolT i;

  SO_forFktSymbole(i)
    our_dealloc(IO_SymbolNamen[i]);
  our_dealloc(IO_SymbolNamen);

  if (IO_Strom != stdout){
     fclose(IO_Strom);
     IO_Strom = stdout;
  }
}

static BOOLEAN IO_SchonInitialisiert = FALSE;

void IO_ParamInfo(ParamInfoArtT what, int level, char* protfile)
{
  if (what == ParamEvaluate){
    fprintf(IO_Strom, "\n* OUTPUT:  Level %d to %s%s\n", 
            level, 
            (protfile||(strcmp(protfile,"stdout")!=0))?"file ":"",
            (protfile||(strcmp(protfile,"stdout")!=0))?protfile:"stdout");
  }
  else{
    fprintf(IO_Strom, "-a (0..5) -prot <protfile>\n");
  }
}

void IO_InitAposteriori(int level, char *Protokolldatei)
/* Daten aus der Spezifikation holen. */
{
   func_symb_list p;
   SymbolT i;
   
#if COMP_WMTW
   if (IO_SchonInitialisiert) return;
#endif
   if (!IO_SchonInitialisiert){TI_IterationInitialisieren();}
   IO_SchonInitialisiert = TRUE;
   
   IO_SymbolNamen = our_alloc(SO_AnzahlFunktionen * sizeof(char *));
   p=get_sig();
   SO_forSpezSymbole(i) {
     IO_SymbolNamen[SO_AusAltMachNeu(sig_first(p))] = (char *) our_alloc(MaximaleBezeichnerLaenge*sizeof(char));
     strcpy(IO_SymbolNamen[SO_AusAltMachNeu(sig_first(p))],map_func(sig_first(p)));
     p = sig_rest(p);
   }

   if (SO_Paramodulation()) {
     ParamodSymbolnamenEtablieren(SO_EQSymbol()   ,"eq");
     ParamodSymbolnamenEtablieren(SO_TRUESymbol() ,"true");
     ParamodSymbolnamenEtablieren(SO_FALSESymbol(),"false");
   }
   sprintf(IO_SpezName,get_spec_name());
   switch (PA_ExtModus){
      case completion  :  sprintf(IO_Modus,"completion");
                          break;
      case reduction   :  sprintf(IO_Modus,"reduction");
                          break;
      case confluence  :  sprintf(IO_Modus,"confluence");
                          break;
      case proof       :  sprintf(IO_Modus,"proof");
                          break;
      case termination :  sprintf(IO_Modus,"termination");
                          break;
      case convergence :  sprintf(IO_Modus,"convergence");
                          break;
      default : 
         our_error("Fehlerhafter ArbeitsModus. IO_Initialisieren.");
         break;
      }
   Ausgabelevel = level;
   IO_level0 = (level==0);
   IO_level1 = (level==1);
   IO_level2 = (level==2);
   IO_level3 = (level==3);
   IO_level4 = (level==4);
#if IO_AUSGABEN
   IO_GUeberschrift = IO_TWUeberschrift = FALSE;
#endif
   if (strcmp(Protokolldatei, "stdout") != 0){
     fprintf(IO_Strom, "\n\n>>>> working! (outputs in  %s)\n\n", Protokolldatei);
     if (!(IO_Strom = fopen(Protokolldatei, "a"))){
        IO_Strom = stdout;
        fprintf(IO_Strom, "\n\nCouldn't open protocol file %s!\n\n", Protokolldatei);
     }
     else
        PA_InteraktionAusschalten();
  }
   LatexAusgabeAktiviert = FALSE;
}  


int IO_AusgabeLevel(void)
{
   return Ausgabelevel;
}

void IO_AusgabeLevelBesetzen(int level)
{
   IO_level0 = (level ==0);
   IO_level1 = (level ==1);
   IO_level2 = (level ==2);
   IO_level3 = (level ==3);
   IO_level4 = (level ==4);
   Ausgabelevel = level;
}

void IO_AusgabenAus(void)
/* Schaltet unabhaengig vom aktuellen Ausgabelevel die Ausgaben aus.*/
{
   IO_level0 = TRUE;
   IO_level1 = IO_level2 = IO_level3 = IO_level4 = FALSE;
}

void IO_AusgabenEin(void)
/* Schaltet die mit IO_AusgabenAus zuvor ausgeschalteten Ausgaben gemaess Einstellung wieder ein.*/
{
   IO_level0 = (Ausgabelevel==0);
   IO_level1 = (Ausgabelevel==1);
   IO_level2 = (Ausgabelevel==2);
   IO_level3 = (Ausgabelevel==3);
   IO_level4 = (Ausgabelevel==4);
}

void IO_RingBell(void)
{
   if (PA_InteraktionAn())
      fprintf(IO_Strom, "\a");
}

void IO_SpezifikationsnamenSetzen(char *Filename)
/* Setzt den Namen der bearbeiteten Spezifikation (incl. Pfad) ein.*/
{
   Spezifikationsname = strdup(Filename);
}

char *IO_Spezifikationsname(void)
/* Gibt den Namen (und ggf. Pfad) der zu verwendeten Spez.datei zurueck.*/
{
   return Spezifikationsname;
}

/* -------------------------------------------------------*/
/* -------------- Kopf und Spezifikation -----------------*/
/* -------------------------------------------------------*/

void IO_DruckeKopf(void)
{
  fprintf(IO_Strom, "\n\n******************************************************************************\n");
  fprintf(IO_Strom, "**                            Waldmeister  II                               **\n");
  fprintf(IO_Strom, "**                                                                          **\n");
  fprintf(IO_Strom, "**              (C) 1994/95/96 by A.Buch and Th.Hillenbrand                 **\n");
  fprintf(IO_Strom, "**              (C)  1996/97   by A.Buch, Th.Hillenbrand, A. Jaeger         **\n");
  fprintf(IO_Strom, "**                <indexing@gentzen.informatik.uni-kl.de>                   **\n");
  fprintf(IO_Strom, "******************************************************************************\n\n");
}

void IO_DruckeBeispiel(void)
{
   if (LatexAusgabeAktiviert){
      fprintf(IO_Strom, "%s & ", Spezifikationsname);
   }
   else{
      IO_DruckeKopfZeileText(TRUE, "problem           ", IO_SpezName);
      IO_LF();
      IO_DruckeKopfZeileText(TRUE, "... from file     ", Spezifikationsname);
      IO_LF();
      IO_DruckeKopfZeileText(TRUE, "Computation Mode", IO_Modus);
   }
}  

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* +++++++++++++ Elementare Druck-Funktionen +++++++++++++*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void IO_DruckeGanzzahl(int n)
{
  fprintf(IO_Strom, "%d",n);
}

void IO_DruckeAdresse(void *Adresse)
{
  fprintf(IO_Strom, "%p",Adresse);
}

void IO_DruckeFlex(const char *fmt, ...)
/* Neufassung vom 1.4.95, Arnim. */
{
   va_list ap;
   int ival;
   double dval;
   unsigned long uval;
   long int lval;
   TermT tval;

   char *ArgPosition;
   
   va_start(ap, fmt);
   do{
     ArgPosition = strchr(fmt, '%');
     if (ArgPosition != NULL){
       fwrite(fmt, sizeof(char), ArgPosition-fmt, IO_Strom); /* (ArgPosition-fmt) Objekte aus fmt auf Ausgabe-Strom schreiben */
       switch (*++ArgPosition){
       case 'd':
         ival = va_arg(ap, int);
         fprintf(IO_Strom, "%d", ival);
         break;
       case 'f':
         dval = va_arg(ap, double);
         fprintf(IO_Strom, "%f", dval);
         break;
       case 's':
         fprintf(IO_Strom, va_arg(ap, char *));
         break;
       case 'l':
         lval = va_arg(ap, long);
         fprintf(IO_Strom, "%ld",lval);
         break;
       case 'u':
         uval = va_arg(ap, unsigned long);
         fprintf(IO_Strom, "%lu", uval);
         break;
       case 't':
         tval = va_arg(ap, TermT);
         IO_DruckeTerm(tval);
         break;
       default:
         putc(*ArgPosition, IO_Strom);
         break;
       }
       fmt = ArgPosition + 1;
     }
     else
       fprintf(IO_Strom, fmt);
   }while(ArgPosition);
   va_end(ap);
   return;
}

void IO_DruckeFlexStrom(FILE *Strom, const char *fmt, ...)
{
   va_list ap;
   int ival;
   double dval;
   unsigned long uval;
   long int lval;
   TermT tval;

   char *ArgPosition;
   
   va_start(ap, fmt);
   do{
     ArgPosition = strchr(fmt, '%');
     if (ArgPosition != NULL){
       fwrite(fmt, sizeof(char), ArgPosition-fmt, Strom); /* (ArgPosition-fmt) Objekte aus fmt auf Ausgabe-Strom schreiben */
       switch (*++ArgPosition){
       case 'd':
         ival = va_arg(ap, int);
         fprintf(Strom, "%d", ival);
         break;
       case 'f':
         dval = va_arg(ap, double);
         fprintf(Strom, "%f", dval);
         break;
       case 's':
         fprintf(Strom, va_arg(ap, char *));
         break;
       case 'l':
         lval = va_arg(ap, long);
         fprintf(Strom, "%ld",lval);
         break;
       case 'u':
         uval = va_arg(ap, unsigned long);
         fprintf(Strom, "%lu", uval);
         break;
       case 't':
         tval = va_arg(ap, TermT);
         IO_DruckeTermStrom(Strom,tval);
         break;
       default:
         putc(*ArgPosition, Strom);
         break;
       }
       fmt = ArgPosition + 1;
     }
     else
       fprintf(Strom, fmt);
   }while(ArgPosition);
   va_end(ap);
   return;
}


/* Diese beiden Funktionen geben eine Zeile des Programm-Kopfes aus.
   Dabei wird die Ausrichtung automatisch vorgenommen.
   Die Ausgabe erfolgt nur, wenn Bedingung erfuellt ist.*/

void IO_DruckeKopfZeileText(BOOLEAN Bedingung, const char *Titel, const char *Text)
{
  if(Bedingung){
    fprintf(IO_Strom, "   %-40s: ", Titel);
    fprintf(IO_Strom, "%s\n", Text);
  }
}

void IO_DruckeKopfZeileInfoStrings(BOOLEAN Bedingung, const char *Titel, InfoStringT is1, InfoStringT is2)
{
  if(Bedingung){
     fprintf(IO_Strom, "   %-40s: ", Titel);
     if ((strlen(is1)>0) && (strlen(is2)>0))
        fprintf(IO_Strom, "%s:%s \n", is1, is2);
     else if (strlen(is1)>0)
        fprintf(IO_Strom, "%s \n", is1);
     else if (strlen(is2)>0)
        fprintf(IO_Strom, "%s \n", is2);
     else fprintf(IO_Strom, "\n");
  }
}


void IO_DruckeKopfZeileZahl(BOOLEAN Bedingung, const char *Titel, int Zahl)
{
  if(Bedingung){
    fprintf(IO_Strom, "   %-40s: ", Titel);
    fprintf(IO_Strom, "%d\n", Zahl);
  }
}



void IO_DruckeZeile(const char *zeichen)
/* Druckt eine ganze Zeile mit dem Zeichen zeichen. */
{
  short i;

  for (i=0; i<_IO_Ueberschriftenbreite; i++){
    fprintf(IO_Strom, zeichen);
  }
  IO_LF();
}

void IO_DruckeUeberschrift(const char *text)
/* Ueberschrift drucken */
{
  short i, rahmen;

  IO_LF();
  IO_DruckeZeile("*");
  rahmen = ((_IO_Ueberschriftenbreite - strlen(text) - 2) / 2);
  for (i=0; i<rahmen; i++){
    fprintf(IO_Strom, "*");
  }
  fprintf(IO_Strom, " ");
  fprintf(IO_Strom, text);
  fprintf(IO_Strom, " ");
  for (i=rahmen+strlen(text)+2; i<_IO_Ueberschriftenbreite; i++){
    fprintf(IO_Strom, "*");
  }
  IO_LF();
  IO_DruckeZeile("*");
  IO_LF();
}  

void IO_WarteAufReturnMitText(const char *text)
/* Gibt text aus und Wartet auf Druck der Return-Taste.*/
{
  char bla[INPUT_STRLEN];

  fprintf(IO_Strom, "\n>> %s <<\n", text);
  rewind(stdin);
  fgets(bla, INPUT_STRLEN, stdin);
}


/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* +++++++++++++ Druck-Funktionen fuer System-Strukturen +*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

void IO_DruckeVergleich(VergleichsT Vergleich)
{
  switch (Vergleich) {
  case Kleiner:        fprintf(IO_Strom, " < "); break;
  case Gleich:         fprintf(IO_Strom, " = "); break;
  case Groesser:       fprintf(IO_Strom, " > "); break;
  case Unvergleichbar: fprintf(IO_Strom, " # "); break;
  }
}


void IO_DruckeSymbol(SymbolT symbol)
{
  if (SO_SymbIstFkt(symbol))
    fprintf(IO_Strom, IO_SymbolNamen[symbol]);
  else
    fprintf(IO_Strom, "x%d",SO_VarNummer(symbol));
}

char *IO_FktSymbName(SymbolT FktSymb)
/* Zeiger auf Zeichenketten-Darstellung des Druckenamens des Funktionssymbols */
{
  return IO_SymbolNamen[FktSymb];
}

void IO_DruckeTerm(UTermT Term)
{
  if (TO_Nullstellig(Term))
    IO_DruckeSymbol(TO_TopSymbol(Term));
  else
    do
      switch (TO_AnzahlTeilterme(Term)) {
      case 0: {
        IO_DruckeSymbol(TO_TopSymbol(Term));
        while (!(--TI_TopZaehler)) {
          fprintf(IO_Strom, ")");
          if (TI_PopUndLeer())
            return;
        }
        fprintf(IO_Strom, ",");
        Term = TO_Schwanz(Term);
        break;
      }
      case 1: {
        SymbolT Symbol = TO_TopSymbol(Term);
        UTermT Teilterm = TO_ErsterTeilterm(Term);
        unsigned int n = 1;
        while (SO_SymbGleich(Symbol,TO_TopSymbol(Teilterm))) {
          n++;
          Teilterm = TO_ErsterTeilterm(Teilterm);
        }
        if (n>=StartzahlVerkettungskuerzung) {
          IO_DruckeSymbol(Symbol);
          fprintf(IO_Strom, "^%d(",n);
          TI_PushOT(Term);
        }
        else
          while (n--) {
            IO_DruckeSymbol(Symbol);
            fprintf(IO_Strom, "(");
            TI_PushOT(Term);
          }
        Term = Teilterm;
        break;
      }
      default: {
        IO_DruckeSymbol(TO_TopSymbol(Term));
        fprintf(IO_Strom, "(");
        TI_PushOT(Term);
        Term = TO_Schwanz(Term);
        break;
      }
      }
    while (TRUE);
}

void IO_DruckeSymbolStrom(FILE *Strom, SymbolT symbol)
/* Ausgabe eines Symbols auf dem angegebenen Ausgabestrom.*/
{
  if (SO_SymbIstFkt(symbol))
    fprintf(Strom,IO_SymbolNamen[symbol]);
  else
    fprintf(Strom,"x%d",SO_VarNummer(symbol));
}


void IO_DruckeTermStrom(FILE *Strom, UTermT Term)
/* Druckt den Term Term auf den Ausgabestrom Strom aus. Dabei werden Folgen gleicher, einstelliger
   Funktionssymbole nicht gekuerzt und Konstanten mit nachfolgendem Klammernpaar ausgegeben.*/
{
  if (TO_Nullstellig(Term)){
    IO_DruckeSymbolStrom(Strom, TO_TopSymbol(Term));
    if (TO_TermIstNichtVar(Term)) /* bei Konstanten Klammernpaar ausgeben */
      fprintf(Strom, "()");
  }
  else
    do
      switch (TO_AnzahlTeilterme(Term)) {
      case 0: {
        IO_DruckeSymbolStrom(Strom, TO_TopSymbol(Term));
        if (TO_TermIstNichtVar(Term)) /* bei Konstanten Klammernpaar ausgeben */
          fprintf(Strom, "()");
        while (!(--TI_TopZaehler)) {
          fprintf(Strom, ")");
          if (TI_PopUndLeer())
            return;
        }
        fprintf(Strom, ",");
        Term = TO_Schwanz(Term);
        break;
      }
      case 1: {
        IO_DruckeSymbolStrom(Strom, TO_TopSymbol(Term));
        fprintf(Strom, "(");
        TI_PushOT(Term);
        Term = TO_ErsterTeilterm(Term);
        break;
      }
      default: {
        IO_DruckeSymbolStrom(Strom, TO_TopSymbol(Term));
        fprintf(Strom, "(");
        TI_PushOT(Term);
        Term = TO_Schwanz(Term);
        break;
      }
      }
    while (TRUE);
}

#if 0
#define IO_DruckeSymbolString(string, symb)     \
{                                               \
  if (SO_SymbIstFkt(symb))                      \
    strcat(string,IO_SymbolNamen[symb]);        \
  else{                                         \
    strcat(string,"x");                         \
    strcat(string,itoa(SO_VarNummer(symb)));    \
  }                                             \
}

void IO_DruckeTermString(char* str, UTermT Term)
/* Druckt den Term Term auf den Ausgabestring str aus. Dabei werden Folgen gleicher, einstelliger
   Funktionssymbole nicht gekuerzt und Konstanten mit nachfolgendem Klammernpaar ausgegeben.
   In str muss genuegend Platz vorhanden sein.*/
{
  if (TO_Nullstellig(Term)){
    IO_DruckeSymbolString(str, TO_TopSymbol(Term));
    if (TO_TermIstNichtVar(Term)) /* bei Konstanten Klammernpaar ausgeben */
      strcat(str, "()");
  }
  else
    do
      switch (TO_AnzahlTeilterme(Term)) {
      case 0: {
        IO_DruckeSymbolString(str, TO_TopSymbol(Term));
        if (TO_TermIstNichtVar(Term)) /* bei Konstanten Klammernpaar ausgeben */
          strcat(str, "()");
        while (!(--TI_TopZaehler)) {
          strcat(str, ")");
          if (TI_PopUndLeer())
            return;
        }
        strcat(str, ",");
        Term = TO_Schwanz(Term);
        break;
      }
      case 1: {
        IO_DruckeSymbolString(str, TO_TopSymbol(Term));
        strcat(str, "(");
        TI_PushOT(Term);
        Term = TO_ErsterTeilterm(Term);
        break;
      }
      default: {
        IO_DruckeSymbolString(str, TO_TopSymbol(Term));
        strcat(str, "(");
        TI_PushOT(Term);
        Term = TO_Schwanz(Term);
        break;
      }
      }
    while (TRUE);
}

unsigned int IO_PCLAusgabelaenge(UTermT Term)
{
  UTermT NULLTerm = TO_NULLTerm(Term);
  unsigned int Laenge = 0;
  do {
    SymbolT Symbol = TO_TopSymbol(Term),
            Stelligkeit = SO_Stelligkeit(Symbol);
    Laenge += SO_SymbIstVar(Symbol) ? 
      1+LaengeNatZahl(SO_VarNummer(Symbol))
      : 
      strlen(IO_SymbolNamen[Symbol])+2+(Stelligkeit>=2 ? Stelligkeit-1 : 0);
  } while (TO_PhysischUngleich(NULLTerm,Term = TO_Schwanz(Term)));
  return Laenge;
}
#endif
void IO_DruckeTermPaar(UTermT term1,const char *trenner,UTermT term2)
{
   IO_DruckeFlex("%t %s %t", term1, trenner, term2);
}
/* -------------------------------------------------------*/
/* -------------------------------------------------------*/
/* -------------- Hilfe-Text -----------------------------*/
/* -------------------------------------------------------*/
/* -------------------------------------------------------*/
extern char **environ;

static void DateiAusgabeOhnePager(const char *name)
{
   FILE *test;
   if (!(test = fopen(name, "r"))){
      fprintf(IO_Strom, "File '%s' not found.", name);
      fclose(test);
   }
   else{
      int c;
      
      while( (c = getc(test)) != EOF)
         putc(c, IO_Strom);
   }
}

void IO_DateiAusgabe(const char* name, BOOLEAN pager)
/* Ausgabe eines Hilfe-Textes (System-Beschreibung).*/
{
  if (pager){
     int PID, status, ww;
     
     PID = fork();                 /* fork erforderlich, weil execve sonst den WM-Prozess terminiert! */

     if (PID==0){
	const char *argv[3];
	
        
        argv[0] = getenv("PAGER");
        argv[1] = name;
        argv[2] = NULL;
        
        if (argv[0] == NULL)
           argv[0] = "more";
        
        if (execve(argv[0],  (char *const *) argv,environ) == -1){
           printf("ENVIRONMENT-VARIABLE PAGER not specified and 'more' not available.\n");
           perror(NULL);
           DateiAusgabeOhnePager(name);
           IO_WarteAufReturnMitText("RETURN");
           exit(-1);            /* Kindprozess terminieren */
        }
     }
     /* Jetzt Vaterprozess */
     ww=wait(&status);
  }
  else
     DateiAusgabeOhnePager(name);
}

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* ++++++++++++++ Statistik ++++++++++++++++++++++++++++++*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

static int Stelligkeit(unsigned long int zahl)
{
   int stellenzahl = 0;
   while (zahl > 0){
      zahl = zahl / 10;
      stellenzahl++;
   }
   return stellenzahl;
}

int IO_Platzverbrauch(unsigned long int zahl)
{
   int st = Stelligkeit(zahl);

   return st + ((st-1)/3);
}

int IO_Einruecktiefe(int bisherigeTiefe, unsigned long int maxEintrag)
/* Berechnet die Einruecktiefe, die fuer die folgenden Eintraege noetig ist.
   Dabei ist maxEintrag der groesste der folgenden Eintraege.
*/
{
   return (bisherigeTiefe + IO_Platzverbrauch(maxEintrag) +2);
}

static void DZFR(unsigned long int zahl, int *stellen, int stellenzahl)
{
  if (zahl == 0){
    *stellen = 0;
    return;
  }
  else if ((zahl / 10) == 0){
    fprintf(IO_Strom, "%lu", zahl);
    *stellen = 1;
  }
  else{
    DZFR(zahl/10, stellen, stellenzahl);
    /*
    if ( (((stellenzahl - *stellen)/3)*3 == stellenzahl-*stellen) && ((stellenzahl-*stellen) > 0))
    */
    if (((stellenzahl - *stellen) % 3 == 0) && ((stellenzahl-*stellen) > 0))
      fprintf(IO_Strom, ".");
    /*
      fprintf(IO_Strom, "%lu", zahl - (zahl/10)*10);
      */
    fprintf(IO_Strom, "%lu", zahl % 10);
    (*stellen)++;
  }
}
      

static void DruckeZahlFormatiert(unsigned long int zahl, int platz)
/* Gibt die Zahl zahl aus, setzt nach jeder dritten Ziffer (von hinten) einen Punkt,
   und fuellt die Ausgabe auf insgesamt platz Zeichen von links mit Leerzeichen auf.*/
{
  int stellenzahl = Stelligkeit(zahl), Leerzeichen, i;
  Leerzeichen = platz - IO_Platzverbrauch(zahl);
  for (i = 0; i < Leerzeichen; i++)
    fprintf(IO_Strom, " ");
  i = 0;
  DZFR(zahl,&i, stellenzahl);
}

static void DruckeStringFormatiert(const char *str, int platz)
/* Gibt die Zeichenkette str aus, nachdem die Ausgabe auf insgesamt platz Zeichen von links 
   mit Leerzeichen aufgefuellt wurde. */
{
  unsigned int i = platz-strlen(str);
  while (i--)
    fprintf(IO_Strom," ");
  fprintf(IO_Strom,"%s",str);
}


const int RechterRandDatum = 14;


/* -------------------------------------------------------*/
/* ------------- Statistik-Ausgaben ----------------------*/
/* -------------------------------------------------------*/

void IO_StatistikModusAufLatexSetzen(void)
{
   LatexAusgabeAktiviert = TRUE;
}

void IO_StatistikModusAufNormalSetzen(void)
{
   LatexAusgabeAktiviert = FALSE;
}

void IO_DruckeKurzUeberschrift(const char *text, const char *zeichen)
{
   if (!LatexAusgabeAktiviert){
      short i, rahmen;
      
      IO_LF();
      rahmen = 5; /*((_IO_Ueberschriftenbreite - strlen(text) - 2) / 2);*/
      for (i=0; i<rahmen; i++){
         fprintf(IO_Strom, zeichen);
      }
      fprintf(IO_Strom, " ");
      fprintf(IO_Strom, text);
      fprintf(IO_Strom, " ");
      for (i=rahmen+strlen(text)+2; i<_IO_Ueberschriftenbreite; i++){
         fprintf(IO_Strom, zeichen);
      }
      IO_LF();
      IO_LF();
   }
}  


   
void IO_DruckeStatistikZeile (const char *str,unsigned long Datum)
/* Gibt den String und dahinter das Datum aus, falls Datum nicht null ist.*/
{
   if (LatexAusgabeAktiviert){
      if (Datum == 0)
         fprintf(IO_Strom, " 0 & ");
      else{
         int stellenzahl = Stelligkeit(Datum), i;
         DZFR(Datum, &i, stellenzahl);
         fprintf(IO_Strom, " & ");
      }
   }
   else if (Datum){
      DruckeZahlFormatiert(Datum, RechterRandDatum);
      fprintf(IO_Strom, " %s\n", str);
  }
}

void IO_DruckeStatistikZeileString(const char *str, const char *Datum)
/* Gibt den String und dahinter das Datum aus, falls Datum nicht null ist.*/
{
   if (LatexAusgabeAktiviert)
      fprintf(IO_Strom, " %s & ", Datum);
   else if (Datum){
      DruckeStringFormatiert(Datum, RechterRandDatum);
      fprintf(IO_Strom, " %s\n", str);
  }
}

void IO_DruckeStatistikZeileEingerueckt(const char *str, unsigned long Datum, int Einruecktiefe)
/* Druckt die Statistik-Zeile um Einruecktiefe 'Stellen' eingerueckt aus.
   Einruecktiefe gibt die Position des rechten Zeichens des Datums an.
*/
{
   if (LatexAusgabeAktiviert){
      if (Datum == 0)
         fprintf(IO_Strom, " 0 & ");
      else{
         int stellenzahl = Stelligkeit(Datum), i;
         DZFR(Datum, &i, stellenzahl);
         fprintf(IO_Strom, " & ");
      }
   }
   else if (Datum){
      DruckeZahlFormatiert(Datum, RechterRandDatum + Einruecktiefe);
      fprintf(IO_Strom, " %s\n", str);
   }
      
}

void IO_DruckeStatistikLeerZeile(void)
{
   if (!LatexAusgabeAktiviert)  fprintf(IO_Strom, "\n");
}

void IO_DruckeStatistikZeileAnteil(const char *str,unsigned long int Zaehler, unsigned long int Nenner)
/* Gibt den String und dahinter das Datum aus, falls der Nenner nicht null ist.*/
{
  if (LatexAusgabeAktiviert){
    if ((Zaehler == 0) || (Nenner == 0))
      fprintf(IO_Strom, " 0 & ");
    else{
      double Quotient = 100.0*(double)Zaehler/(double)Nenner;
      fprintf (IO_Strom, "%3.1f & \n", Quotient);
    }
  }
  else if (Nenner) {
    double Quotient = 100.0*(double)Zaehler/(double)Nenner;
    unsigned int Vorkomma = (Quotient = ((double)(unsigned int)(10.0*Quotient+0.5))/10.0),
                 Nachkomma = (unsigned int)(10.0*(Quotient-(double)Vorkomma)+0.5),
                 i = RechterRandDatum-LaengeNatZahl(Vorkomma)-LaengeNatZahl(Nachkomma)-1;
    while (i--)
      fprintf(IO_Strom," ");
    fprintf(IO_Strom,"%d,%d %s\n",Vorkomma,Nachkomma,str);
  }
}

/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/
/* +++++++++++++++++++++ Sonstiges +++++++++++++++++++++++*/
/* +++++++++++++++++++++++++++++++++++++++++++++++++++++++*/

unsigned int IO_SymbAusgabelaenge(SymbolT Symbol)
{
  return SO_SymbIstFkt(Symbol) ? strlen(IO_SymbolNamen[Symbol]) : (1+LaengeNatZahl(SO_VarNummer(Symbol)));
}
