/*****************************************************************************************/
/* Normieren von Spezifikationen                                                         */
/*****************************************************************************************/

#include "general.h"
#include "parser.h"
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "SpeicherVerwaltung.h"
#include "SpezNormierung.h"

/*****************************************************************************************/
/* 1) Variablen                                                                          */
/*****************************************************************************************/

#define Norm_Output_Off
/* Achtung: Im Parser sind Funktionssymbole mit 1 beginnend aufsteigend nummeriert! */

typedef unsigned long int kodierung_typ;
const double Primzahlen [] ={2,3,5,7,11,13,17,19,23,29,31,37,41,43,47};
const int AnzPrimzahlen = sizeof(Primzahlen) / sizeof(double);


/* TODO: Nachfolgende Symbole, sofern moeglich, als static definieren */
static kodierung_typ *LFeld, *RFeld;

static kodierung_typ *Stelligkeit;

static kodierung_typ *AnzVorkGleichungen;
static kodierung_typ *PositionenGleichungen;

static kodierung_typ *AnzVorkZiele;
static kodierung_typ *PositionenZiele;


static kodierung_typ *KombiKontextGleichungenEinzeln;
static kodierung_typ *KombiKontextGleichungenEinzelnGegenueber;
static kodierung_typ *KombiKontextGleichungen;

static kodierung_typ *KombiKontextZieleEinzeln;
static kodierung_typ *KombiKontextZieleEinzelnGegenueber;
static kodierung_typ *KombiKontextZiele;

static kodierung_typ *FinaleBewertung;

static int AnzahlSymbole;

/* exportiert: */
int SN_AnzahlGleichungen = 0;

/*****************************************************************************************/
/* 2) Daten sammeln                                                                      */
/*****************************************************************************************/

static void VorkommensDatenAusTermSammeln(term Term, kodierung_typ *WerteFeld)
     /* Zaehlt die Vorkommen der Symbole im angegebenen Term und erhoeht die entsprechenden
        Slots im WerteFeld.*/
{
  int i;
  if (!variablep(Term)){
    WerteFeld[term_func(Term)-1]++;

    i = Stelligkeit[term_func(Term)-1];
    while(i>0)
      VorkommensDatenAusTermSammeln(term_arg(Term,--i),WerteFeld);
  }
}
    
static void PositionsDatenAusTermSammeln(term Term, kodierung_typ *WerteFeld, kodierung_typ Position, int Tiefe)
     /* Berechnet die Positions-Vorkommen der Symbole in Term in eindeutiger Weise und addiert sie
        auf die entsprechenden Slots in WerteFeld.*/
{
  int i;
  /*    if (Tiefe==0) printf("\n\n"); printf("PDAT(%lu,%d)\n", Position, Tiefe);*/
  if (Tiefe >=  AnzPrimzahlen) return;       /* Keine Primzahlen mehr! */
  if (!variablep(Term)){
    i = sig_arity(term_func(Term));
    while(i>0){
      PositionsDatenAusTermSammeln(term_arg(Term,i-1),
                                   WerteFeld, 
                                   Position*((kodierung_typ)pow(Primzahlen[Tiefe],(double)i)),
                                   Tiefe+1);
      i--;
    }
    WerteFeld[term_func(Term)-1] += Position;
    /*    printf("Erhoehe Wert(%ld, %s) um %lu auf %lu\n", term_func(Term)-1, map_func(term_func(Term)),
           Position, WerteFeld[term_func(Term)-1]);*/
  }
}

static kodierung_typ TermKontext(term Term, kodierung_typ context)
     /* Verwendet Daten aus Zielen und Gleichungen! Darf daher erst nach der Berechnung
        der AnzVork und Positionen aufgerufen werden.*/
{
  if (variablep(Term)) context++;
  else{
    int i = sig_arity(term_func(Term));
    context+= 
      AnzVorkGleichungen[term_func(Term)-1]
      +AnzVorkZiele[term_func(Term)-1]
      +PositionenGleichungen[term_func(Term)-1]
      +PositionenZiele[term_func(Term)-1];
    
    while(i>0)
      context += TermKontext(term_arg(Term,--i), context);
  }
  return context;
}

static kodierung_typ TermKontext2(term Term, kodierung_typ context)
{
  if (variablep(Term)) context++;
  else{
    int i = sig_arity(term_func(Term));
    context+=
      KombiKontextGleichungen[term_func(Term)-1]
      +KombiKontextZiele[term_func(Term)-1];
    
    while(i>0)
      context += TermKontext2(term_arg(Term,--i), context);
  }
  return context;
}

static void FinaleSymbolBewertung(term Term, kodierung_typ VaterBewertung)
     /* Jedes Symbol erhaelt die Bewertung seines direkten Vater-Terms
        addiert.*/
{
  if (variablep(Term)) return;
  else{
    int i = sig_arity(term_func(Term));

    FinaleBewertung[term_func(Term)-1] += VaterBewertung;

    VaterBewertung = TermKontext2(Term,0);
    
    while(i>0)
      FinaleSymbolBewertung(term_arg(Term,--i), VaterBewertung);
  }
}

#define KontextBewertungAusfuehren(getWhat,Was)                                  \
   for (eql=get_##getWhat(); eql; eql=eq_rest(eql)){                             \
     kodierung_typ GesKonL = TermKontext(eq_first_lhs(eql),0),                          \
         GesKonR = TermKontext(eq_first_rhs(eql),0);                             \
     VorkommensDatenAusTermSammeln(eq_first_lhs(eql), LFeld);                    \
     VorkommensDatenAusTermSammeln(eq_first_rhs(eql), RFeld);                    \
     for (p=get_sig(); p ; p = sig_rest(p)){                                     \
       if (RFeld[sig_first(p)-1] > 0){                                           \
         KombiKontext##Was[sig_first(p)-1] += GesKonR+GesKonL;                   \
         KombiKontext##Was##Einzeln[sig_first(p)-1] += GesKonR;                  \
       }                                                                         \
       if (LFeld[sig_first(p)-1] > 0){                                           \
         KombiKontext##Was[sig_first(p)-1] += GesKonR+GesKonL;                   \
         KombiKontext##Was##Einzeln[sig_first(p)-1] += GesKonL;                  \
       }                                                                         \
                                                                                 \
       if ((RFeld[sig_first(p)-1] > 0) && (LFeld[sig_first(p)-1] == 0))          \
         KombiKontext##Was##EinzelnGegenueber[sig_first(p)-1] += GesKonL;        \
       if ((LFeld[sig_first(p)-1] > 0) && (RFeld[sig_first(p)-1] == 0))          \
         KombiKontext##Was##EinzelnGegenueber[sig_first(p)-1] += GesKonR;        \
                                                                                 \
       if ((RFeld[sig_first(p)-1] > 0) || (LFeld[sig_first(p)-1] > 0))           \
       LFeld[sig_first(p)-1] = RFeld[sig_first(p)-1] = 0;                        \
     }                                                                           \
   }


#define FinaleBewertungAusfuehren(getWhat)                                             \
   for (eql=get_##getWhat(); eql; eql=eq_rest(eql)){                                   \
     FinaleSymbolBewertung(eq_first_lhs(eql), 0);                                      \
     FinaleSymbolBewertung(eq_first_rhs(eql), 0);                                      \
   }
    
static void SymboldatenSammeln(void)
{
   func_symb_list p;
   unsigned long int anz;
   eq_list eql;

   /* 1) Funktionssymbole zaehlen */
   for (p=get_sig(), anz=0; p ; p = sig_rest(p),anz++); 

   AnzahlSymbole = anz;
   
   /* 2) Felder allokieren */
   LFeld = our_alloc(anz*sizeof(kodierung_typ));
   RFeld = our_alloc(anz*sizeof(kodierung_typ));

   Stelligkeit = our_alloc(anz*sizeof(kodierung_typ));

   AnzVorkGleichungen = our_alloc(anz*sizeof(kodierung_typ));
   PositionenGleichungen = our_alloc(anz*sizeof(kodierung_typ));      

   AnzVorkZiele = our_alloc(anz*sizeof(kodierung_typ));
   PositionenZiele = our_alloc(anz*sizeof(kodierung_typ));
 
   KombiKontextGleichungenEinzeln = our_alloc(anz*sizeof(kodierung_typ));
   KombiKontextGleichungenEinzelnGegenueber = our_alloc(anz*sizeof(kodierung_typ));
   KombiKontextGleichungen= our_alloc(anz*sizeof(kodierung_typ));
   KombiKontextZieleEinzeln = our_alloc(anz*sizeof(kodierung_typ));
   KombiKontextZieleEinzelnGegenueber = our_alloc(anz*sizeof(kodierung_typ));
   KombiKontextZiele= our_alloc(anz*sizeof(kodierung_typ));

   FinaleBewertung = our_alloc(anz*sizeof(kodierung_typ));

   /* 3) Felder initialisieren */
   for (p=get_sig(), anz=0; p ; p = sig_rest(p),anz++){
     /*     prunsigned long intf("%d) Code=%d\n", anz, sig_first(p));*/
     Stelligkeit[sig_first(p)-1] = sig_arity(sig_first(p));
     AnzVorkGleichungen[sig_first(p)-1] 
       = LFeld[sig_first(p)-1] 
       = RFeld[sig_first(p)-1] 
       = AnzVorkZiele[sig_first(p)-1] 
       = PositionenGleichungen[sig_first(p)-1] 
       = PositionenZiele[sig_first(p)-1] 
       = KombiKontextGleichungenEinzeln[sig_first(p)-1] 
       = KombiKontextGleichungenEinzelnGegenueber[sig_first(p)-1] 
       = KombiKontextGleichungen[sig_first(p)-1] 
       = KombiKontextZieleEinzeln[sig_first(p)-1] 
       = KombiKontextZieleEinzelnGegenueber[sig_first(p)-1] 
       = KombiKontextZiele[sig_first(p)-1] 
       = FinaleBewertung[sig_first(p)-1] 
       = 0;
   }
   /* 4) Erster Durchlauf durch Ziele und Gleichungen:
      Vorkommens- und positionsabhaengige Daten berechnen
      */
   
   for (eql=get_conclusions(); eql; eql=eq_rest(eql)){
     VorkommensDatenAusTermSammeln(eq_first_lhs(eql), LFeld);
     VorkommensDatenAusTermSammeln(eq_first_rhs(eql), RFeld);
     for (p=get_sig(); p ; p = sig_rest(p)){
       AnzVorkZiele[sig_first(p)-1] += RFeld[sig_first(p)-1] +LFeld[sig_first(p)-1];
       LFeld[sig_first(p)-1] = RFeld[sig_first(p)-1] = 0;
     }
     PositionsDatenAusTermSammeln(eq_first_lhs(eql), LFeld,1,0);
     PositionsDatenAusTermSammeln(eq_first_rhs(eql), RFeld,1,0);
     for (p=get_sig(); p ; p = sig_rest(p)){
       PositionenZiele[sig_first(p)-1] += RFeld[sig_first(p)-1] +LFeld[sig_first(p)-1];
       LFeld[sig_first(p)-1] = RFeld[sig_first(p)-1] = 0;
     }
   }
   for (eql=get_equations(); eql; eql=eq_rest(eql)){
     SN_AnzahlGleichungen++;
     VorkommensDatenAusTermSammeln(eq_first_lhs(eql), LFeld);
     VorkommensDatenAusTermSammeln(eq_first_rhs(eql), RFeld);
     for (p=get_sig(); p ; p = sig_rest(p)){
       AnzVorkGleichungen[sig_first(p)-1] += RFeld[sig_first(p)-1] +LFeld[sig_first(p)-1];
       LFeld[sig_first(p)-1] = RFeld[sig_first(p)-1] = 0;
     }
     
     PositionsDatenAusTermSammeln(eq_first_lhs(eql), LFeld,1,0);
     PositionsDatenAusTermSammeln(eq_first_rhs(eql), RFeld,1,0);
     for (p=get_sig(); p ; p = sig_rest(p)){
       PositionenGleichungen[sig_first(p)-1] += RFeld[sig_first(p)-1] +LFeld[sig_first(p)-1];
       LFeld[sig_first(p)-1] = RFeld[sig_first(p)-1] = 0;
     }
   }

   
   /* 5) Zweiter Durchlauf durch Ziele und Gleichungen:
         Kontextabhaengige Daten auf Basis von Vorkommen und Positionen berechnen
      */
   
   KontextBewertungAusfuehren(equations,Gleichungen)
   KontextBewertungAusfuehren(conclusions,Ziele)

   /* 6) Dritter Durchlauf durch Ziele und Gleichungen:
         Kontextdaten aus 2. Durchlauf verwenden
      */

   FinaleBewertungAusfuehren(equations)
   FinaleBewertungAusfuehren(conclusions)
     
}
 
/*****************************************************************************************/
/* 3) Daten auswerten                                                                    */
/*****************************************************************************************/

#define DatenVergleichMin(datum)                      \
  if (resultat==Gleich)                               \
    resultat = (datum[f-1]>datum[g-1])?               \
      Groesser:(datum[f-1]<datum[g-1])?               \
      Kleiner:Gleich;                                 \
  else return resultat;

#define DatenVergleichMax(datum)                      \
  if (resultat==Gleich)                               \
    resultat = (datum[f-1]<datum[g-1])?               \
      Groesser:(datum[f-1]>datum[g-1])?               \
      Kleiner:Gleich;                                 \
  else return resultat;

static VergleichsT SymbolVergleich(func_symb f, func_symb g)
{
  VergleichsT resultat = Gleich;

  DatenVergleichMax(AnzVorkGleichungen);
  DatenVergleichMin(Stelligkeit);
  DatenVergleichMin(PositionenGleichungen);
  DatenVergleichMin(PositionenZiele);
  DatenVergleichMin(KombiKontextGleichungenEinzeln);
  DatenVergleichMin(KombiKontextGleichungenEinzelnGegenueber);
  DatenVergleichMin(KombiKontextGleichungen);
  DatenVergleichMin(KombiKontextZieleEinzeln);
  DatenVergleichMin(KombiKontextZieleEinzelnGegenueber);
  DatenVergleichMin(KombiKontextZiele);
  DatenVergleichMin(FinaleBewertung);
  return resultat;
}

static void   SymboldatenAusgeben(void)
{
  func_symb_list p1,p2;
  BOOLEAN IstEindeutig = TRUE, SkolemFall = FALSE, NichtSkolemFall = FALSE;
#ifdef Norm_Output
  printf("\n\nAnzahl Symbole: %d\n", AnzahlSymbole);
  printf("\n        |Gleichungen                           |Ziele                         |Finale");
  printf("\nSymb|tau| #G| POS    |   KOKOE|   KOKOZ|#Z |    POS |   KOKOE|   KOKOZ|Bew.");
  printf("\n----+---+---+--------+--------+--------+---+--------+--------+--------+---------\n");
  for (p=get_sig(); p ; p = sig_rest(p)){
    char *nom = map_func(sig_first(p));
    /*               tau     #G   POS    KOK   KOK  #Z    POS   KOK   KOK   FBew */
    printf("%c%c%c%c|%8lu|%8lu|%8lu|%8lu|%8lu|%8lu|%8lu|%8lu|%8lu|%8lu\n", 
           nom[0],
           strlen(nom)>1?nom[1]:' ',
           strlen(nom)>2?nom[2]:' ',
           strlen(nom)>3?nom[3]:' ',
           Stelligkeit[sig_first(p)-1],
           AnzVorkGleichungen[sig_first(p)-1],
           PositionenGleichungen[sig_first(p)-1],
           KombiKontextGleichungenEinzeln[sig_first(p)-1],
           KombiKontextGleichungen[sig_first(p)-1],
           AnzVorkZiele[sig_first(p)-1],
           PositionenZiele[sig_first(p)-1],
           KombiKontextZieleEinzeln[sig_first(p)-1],
           KombiKontextZiele[sig_first(p)-1],
           FinaleBewertung[sig_first(p)-1]);
  }

  printf("\n  ");
  for (p1=get_sig(); p1 ; p1 = sig_rest(p1))
    printf("%c ", *map_func(sig_first(p1)));
  printf("\n");
#endif
  for (p1=get_sig(); p1 ; p1 = sig_rest(p1)){
#ifdef Norm_Output
    printf("%c ", *map_func(sig_first(p1)));
#endif
    for (p2=get_sig(); p2 ; p2 = sig_rest(p2)){
      if (sig_first(p1)==sig_first(p2)){
#ifdef Norm_Output
        printf("= ");
#endif
        continue;
      }
      switch (SymbolVergleich(sig_first(p1), sig_first(p2))){
      case Gleich: 
#ifdef Norm_Output
        printf("= ");
#endif
        IstEindeutig = FALSE;
        if ((Stelligkeit[sig_first(p1)-1] == 0)
            && (Stelligkeit[sig_first(p2)-1] == 0)
            && (AnzVorkGleichungen[sig_first(p1)-1] == 0)
            && (AnzVorkGleichungen[sig_first(p2)-1] == 0))
          SkolemFall = TRUE;
        else
          NichtSkolemFall = TRUE;
        break;
#ifdef Norm_Output
      case Kleiner: printf("< ");break;
      case Groesser: printf("> ");break;
      case Unvergleichbar: printf("# ");break;
#endif
      default: break;
      }
    }
#ifdef Norm_Output
    printf("\n");
#endif
  }
}

/* 
   0) Reihenfolge der Symbole festlegen
      Zugriff: PositionVon(Symbol) und SymbolNr(Position)
      
   1) Links-Rechts-Sortierung von Gleichungen
      a) Variablen mit Positionsmethode verarbeiten
      b) Lexikographischer Term-Vergleich mit Symbol- und Variablen-Nummern

   2) Variablen-Normierung der Gleichungen

   2) Reihenfolge der Gleichungen
      Lexikographischer Vergleich von Gleichungen
 */
/*****************************************************************************************/
/* 4) Reihenfolge der Symbole festlegen                                                  */
/*****************************************************************************************/


static int *SymbNr;
/* exportiert */
int *SN_PosVon, SN_GroesstesESymbol;
#define PositionVon(symbol) SN_PosVon[symbol-1]
#define SymbolNr(position) SymbNr[position]

/* exportiert */
term *SN_LHS, *SN_RHS;

static void SymbolReihenfolgeFestlegen(void)
{
  func_symb_list p;
  int i,j,k;
  SN_PosVon = our_alloc(AnzahlSymbole*sizeof(int));
  SymbNr  = our_alloc(AnzahlSymbole*sizeof(int));

  for (i=0,p=get_sig(); p ; p = sig_rest(p),i++)
    SymbolNr(i) = sig_first(p);
  
  for (i=1;i<AnzahlSymbole; i++){
    for(j=0; j<AnzahlSymbole-i;j++)
      if (SymbolVergleich(SymbolNr(j),SymbolNr(j+1)) == Groesser){
        func_symb temp = SymbolNr(j);
        SymbolNr(j) = SymbolNr(j+1);
        SymbolNr(j+1) = temp;
      }
  }
  printf("Symbols: ");
  SN_GroesstesESymbol=-1;
  for (k = 0; k < AnzahlSymbole; k++){
    printf("%s  ", map_func(SymbolNr(k)));
    if (AnzVorkGleichungen[SymbolNr(k)-1])
      SN_GroesstesESymbol = k;
    PositionVon(SymbolNr(k)) = k;
  }
  printf("\n");
#ifdef Norm_Output
  for (p=get_sig(); p ; p = sig_rest(p))
    printf("%s auf Pos. %d\n", map_func(sig_first(p)), PositionVon(sig_first(p)));
#endif
  /* Alt:
     if ((sig_arity(SymbolNr(k))==0) && (AnzVorkGleichungen[SymbolNr(k)-1]==0) && (SN_GroesstesESymbol==0))
       SN_GroesstesESymbol=k-1;
  */
}

/*****************************************************************************************/
/* 5) Gleichungen anordnen                                                               */
/*****************************************************************************************/

static VergleichsT LRSortierenRek(term lhs, term rhs)
     /* Zunaechst Variablen als Joker behandeln */
{
  VergleichsT resultat;
  if (variablep(lhs) && variablep(rhs)) return Gleich;
  if (variablep(lhs)) return Kleiner;
  if (variablep(rhs)) return Groesser;
  if ((resultat=SymbolVergleich(term_func(lhs), term_func(rhs)))==Gleich){
    int i = 0;
    if (sig_arity(term_func(lhs))>0){
      do{
        resultat=LRSortierenRek(term_arg(lhs,i),term_arg(rhs,i));
        i++;
      }while((resultat==Gleich) && (i<sig_arity(term_func(lhs))));
    }
  }
  return resultat;
}    
      
static void LRSortieren(void)
     /* Alle Gleichungen LR-Sortieren */
{
  BOOLEAN ok=TRUE;
  int i=0;
  eq_list eql;
  SN_LHS=our_alloc(SN_AnzahlGleichungen*sizeof(term));
  SN_RHS=our_alloc(SN_AnzahlGleichungen*sizeof(term));
  for(eql=get_equations();eql;eql=eq_rest(eql)){
    switch(LRSortierenRek(eq_first_lhs(eql), eq_first_rhs(eql))){
    case Groesser:              /* l>r --> tauschen */
      SN_LHS[i] = eq_first_rhs(eql); SN_RHS[i] = eq_first_lhs(eql);
      /*print_pair(eq_first_rhs(eql), " X ", eq_first_lhs(eql));*/
      break;
    case Kleiner:               /* l<r --> nicht tauschen */
      SN_LHS[i] = eq_first_lhs(eql); SN_RHS[i] = eq_first_rhs(eql);
      /*print_pair(eq_first_lhs(eql), " || ", eq_first_rhs(eql));*/
      break;
    case Gleich: 
#ifdef Norm_output
      if (!((sig_arity(term_func(eq_first_lhs(eql))) ==2) /* kein AC-Symbol?? */
            && (term_func(eq_first_lhs(eql))==term_func(eq_first_rhs(eql)))
            && (variablep(term_arg(eq_first_lhs(eql),0)))
            && (variablep(term_arg(eq_first_lhs(eql),1)))
            && (variablep(term_arg(eq_first_rhs(eql),0)))
            && (variablep(term_arg(eq_first_rhs(eql),1)))
            && (term_func(term_arg(eq_first_lhs(eql),0)) == term_func(term_arg(eq_first_rhs(eql),1)))
            && (term_func(term_arg(eq_first_lhs(eql),1)) == term_func(term_arg(eq_first_rhs(eql),0))))){
        printf("Nicht eindeutig: ");
        print_pair(eq_first_lhs(eql), " = ", eq_first_rhs(eql));
      }
      else{
        printf("Nicht eindeutig(AC): ");
        print_pair(eq_first_lhs(eql), " = ", eq_first_rhs(eql));
      }
#endif
      SN_LHS[i] = eq_first_lhs(eql); SN_RHS[i] = eq_first_rhs(eql);
      ok=FALSE;break;
    default: 
      SN_LHS[i] = eq_first_lhs(eql); SN_RHS[i] = eq_first_rhs(eql);
      printf("error: Nicht vergleichbare Gleichung\n");ok=FALSE;break;
    }
    i++;
  }
}

static int *varlist;
static int AktVarNummer;

static void TermVariablenNormieren(term Term)
{
  if(variablep(Term)){
    if (varlist[-term_func(Term)] == 1){ /* erstes Vorkommen dieser Variablen */
      varlist[-term_func(Term)] = -AktVarNummer;
      AktVarNummer++;
    }
    Term->code = varlist[-term_func(Term)];
  }
  else{
    int j=0;
    while(j<sig_arity(term_func(Term)))
      TermVariablenNormieren(term_arg(Term,j++));
  }
}
  
static void GleichungenVariablenNormieren(void)
     /* Alle Gleichungen Variablen-Normieren */
{
  eq_list eql;
  int j=0;
  varlist = our_alloc(128*sizeof(func_symb));
  for(eql=get_equations();eql;eql=eq_rest(eql)){
    int i;
    for (i=0; i<128;i++) varlist[i] = 1; /* Feld initialisieren */
    AktVarNummer=1;
    TermVariablenNormieren(SN_LHS[j]);
    TermVariablenNormieren(SN_RHS[j]);
    /*print_pair(SN_LHS[j],"=",SN_RHS[j]);*/
    j++;
  }
  printf("\n\n");
  our_dealloc(varlist);
}

static VergleichsT LRSymbSortierenRek(term SN_LHS, term SN_RHS)
     /* 2 variablen-normierte Terme vergleichen (Variablen _nicht_ als Joker) */
{
  int i=0;
  VergleichsT resultat=Gleich;
  if (variablep(SN_LHS) && variablep(SN_RHS)){
     if (SN_LHS->code < SN_RHS->code) return Kleiner;
     else if (SN_LHS->code > SN_RHS->code) return Groesser;
     else return Gleich;
  }
  if (variablep(SN_LHS)) return Kleiner;
  if (variablep(SN_RHS)) return Groesser;
  
  if (PositionVon(term_func(SN_LHS)) < PositionVon(term_func(SN_RHS))) return Kleiner;
  if (PositionVon(term_func(SN_LHS)) > PositionVon(term_func(SN_RHS))) return Groesser;
  
  while ((i<sig_arity(term_func(SN_LHS)))&&(resultat==Gleich)){
    resultat = LRSymbSortierenRek(term_arg(SN_LHS,i),term_arg(SN_RHS,i));
    i++;
  }
  return resultat;
}
  
 
static void GleichungenSortieren(void)
     /* Alle Gleichungen in pseudo-eindeutiger Art sortieren.
        Am Ende stehen sind die Gleichungen in (SN_LHS[i],SN_RHS[i]) zu finden.*/
{
  int i,j;
  for (i=1;i<SN_AnzahlGleichungen; i++){
    for(j=0; j<SN_AnzahlGleichungen-i;j++)
      switch (LRSymbSortierenRek(SN_LHS[j],SN_LHS[j+1])){
      case Groesser:
        {
          term temp = SN_LHS[j];
          SN_LHS[j]=SN_LHS[j+1];
          SN_LHS[j+1]=temp;
          temp = SN_RHS[j];
          SN_RHS[j]=SN_RHS[j+1];
          SN_RHS[j+1]=temp;
        }
      break;
      case Gleich: 
        if(LRSymbSortierenRek(SN_RHS[j],SN_RHS[j+1])==Groesser){
          term temp = SN_LHS[j];
          SN_LHS[j]=SN_LHS[j+1];
          SN_LHS[j+1]=temp;
          temp = SN_RHS[j];
          SN_RHS[j]=SN_RHS[j+1];
          SN_RHS[j+1]=temp;
        }
        break;
      default:
        break;
      }
  }
#ifdef Norm_Output
  for(j=0; j<SN_AnzahlGleichungen;j++)    print_pair(SN_LHS[j],"=",SN_RHS[j]);
#endif
}

/*****************************************************************************************/
/* 6) Exportierte Funktion                                                               */
/*****************************************************************************************/

void SN_SpezifikationAuswerten(short int *GroesstesESymbol)
{
  
  SymboldatenSammeln();
  SymboldatenAusgeben();
  SymbolReihenfolgeFestlegen();
  LRSortieren();
  GleichungenVariablenNormieren();
  GleichungenSortieren();
  *GroesstesESymbol = SN_GroesstesESymbol;
}
