/*=================================================================================================================
  =                                                                                                               =
  =  SymbolOperationen.c                                                                                          =
  =                                                                                                               =
  =  In diesem Modul sind die Operationen auf Symbolen, d.h. auf Variablen-,                                      =
  =  Konstanten- und Funktions-Symbolen, realisiert.                                                              =
  =                                                                                                               =
  =  24.02.94 Angelegt. Arnim.                                                                                    =
  =  25.02.94 1 Fehler beseitigt, IntegerVergleich. Thomas.                                                       =
  =  25.02.94 Weitergemacht. Arnim.                                                                               =
  =  04.03.94 Kommentiert, etc.                                                                                   =
  =                                                                                                               =
  =================================================================================================================*/

#include "Ausgaben.h"
#include "FehlerBehandlung.h"
#include "general.h"
#include "InfoStrings.h"
#include "parser.h"
#include <stdlib.h>
#include <string.h>
#include "SpezNormierung.h"
#include "SymbolOperationen.h"


/*=================================================================================================================*/
/*= II. Globale Variablen                                                                                         =*/
/*=================================================================================================================*/

/* Signaturinformationen */

SymbolT             SO_GroesstesESymb,
                    SO_GroesstesSpezSymb,
                    SO_GroesstesFktSymb;
static unsigned int AnzahlSpezFunktionen;
unsigned int        SO_AnzahlFunktionen;
StelligkeitsT       *SO_Stelligkeit;
static unsigned int GroesseStelligkeit;
 unsigned int       SO_MaximaleStelligkeit;
static func_symb    *DISCOUNTvonWM;

/* Ordnungsinformationen */

BOOLEAN     SO_PraezedenzTotal;
VergleichsT **SO_Praezedenz;
SymbolT     *SO_TotaleKette;
weight      *SO_FunktionsGewichte, 
            Variablengewicht;

/* Semantische Informationen */

BOOLEAN *SO_IstACSymbol,
        SO_Paramodulation;
SymbolT SO_EQSymbol,
        SO_TRUESymbol,
        SO_FALSESymbol;



/*=================================================================================================================*/
/*= IV. Symbolindizierte Felder                                                                                   =*/
/*=================================================================================================================*/

static VergleichsT **Praezedenzfeld(void)
{
  SymbolT i,j;
  VergleichsT **Praezedenz = array_alloc(SO_AnzahlFunktionen, VergleichsT*);
  SO_forFktSymbole(i) {
    Praezedenz[i] = array_alloc(SO_AnzahlFunktionen, VergleichsT);     /* Zeilen der Praezedenz-Matrix anlegen. */
    SO_forFktSymbole(j)
      Praezedenz[i][j] = Unvergleichbar;
  }
  return Praezedenz;
}

static void PraezedenzfeldLoeschen(VergleichsT **Praezedenz)
{
  SymbolT i;
  SO_forFktSymbole(i)
    our_dealloc(Praezedenz[i]);
  our_dealloc(Praezedenz);
}

static void ArraysMitFunktionsInformationenAnlegen(void)
{   /* legt die globalen Variablen SO_Stelligkeit, SO_Praezedenz etc., mit der angegebenen Anzahl von Plaetzen an. */
  SymbolT i;
  GroesseStelligkeit = SO_AnzahlFunktionen+SO_AnzErsterVariablen;
  SO_Stelligkeit =       negarray_alloc(GroesseStelligkeit, StelligkeitsT, SO_AnzErsterVariablen);
  SO_FunktionsGewichte = negarray_alloc(GroesseStelligkeit, weight,        SO_AnzErsterVariablen);
  SO_IstACSymbol =       negarray_alloc(GroesseStelligkeit, BOOLEAN,       SO_AnzErsterVariablen);
  SO_forSymboleAbwaerts(i,SO_ErstesVarSymb(),SO_NummerVar(SO_AnzErsterVariablen)) {
    SO_Stelligkeit[i] = 0;
    SO_IstACSymbol[i] = FALSE;
  }
  DISCOUNTvonWM = array_alloc(SO_AnzahlFunktionen,func_symb);
  SO_Praezedenz = Praezedenzfeld();
  SO_TotaleKette = array_alloc(SO_AnzahlFunktionen,SymbolT);
}


SymbolT SO_AusAltMachNeu(func_symb fkt)
/* Liefert das dem FunktionsSymbol fkt intern zugeordnete Symbol, so moeglich; sonst our_assertion_failed. */
{
  SymbolT i;

  SO_forFktSymbole(i)
    if (DISCOUNTvonWM[i]==fkt)
      return i;
  our_assertion_failed("FunktionsSymbol nicht in der Spezifikation!");
  return SO_KleinstesFktSymb();
}


void SO_VariablenfeldAnpassen(SymbolT NeuesteBaumvariable)                   /* Anpassen des Feldes SO_Stelligkeit */
{                                                                            /* und des Feldes SO_FunktionsGewicht */
  if (SO_VariableIstNeuer(NeuesteBaumvariable,SO_NummerVar(GroesseStelligkeit-SO_AnzahlFunktionen()))) {
    unsigned int AnzVariablenNeu = SO_AnzErsterVariablen*((-NeuesteBaumvariable)/SO_AnzErsterVariablen+1),
                 GroesseStelligkeitNeu = AnzVariablenNeu+SO_AnzahlFunktionen();
    StelligkeitsT *StelligkeitNeu = negarray_alloc(GroesseStelligkeitNeu, StelligkeitsT, AnzVariablenNeu);
    weight *FunktionsGewichteNeu =  negarray_alloc(GroesseStelligkeitNeu, weight,        AnzVariablenNeu);
    BOOLEAN *IstACSymbolNeu =       negarray_alloc(GroesseStelligkeitNeu, BOOLEAN,       AnzVariablenNeu);
    SymbolT SymbolIndex;

    SO_forFktSymbole(SymbolIndex) {
      StelligkeitNeu[SymbolIndex] = SO_Stelligkeit[SymbolIndex];
      FunktionsGewichteNeu[SymbolIndex] = SO_FunktionsGewichte[SymbolIndex];
      IstACSymbolNeu[SymbolIndex] = SO_IstACSymbol[SymbolIndex];
    }

    SO_forSymboleAbwaerts(SymbolIndex,SO_ErstesVarSymb(),SO_NummerVar(AnzVariablenNeu)) {
      StelligkeitNeu[SymbolIndex] = 0;
      FunktionsGewichteNeu[SymbolIndex] = Variablengewicht;
      IstACSymbolNeu[SymbolIndex] = FALSE;
    }

    negarray_dealloc(SO_Stelligkeit,       GroesseStelligkeit-SO_AnzahlFunktionen());
    negarray_dealloc(SO_FunktionsGewichte, GroesseStelligkeit-SO_AnzahlFunktionen());
    negarray_dealloc(SO_IstACSymbol,       GroesseStelligkeit-SO_AnzahlFunktionen());

    GroesseStelligkeit = GroesseStelligkeitNeu;
    SO_Stelligkeit = StelligkeitNeu;
    SO_FunktionsGewichte = FunktionsGewichteNeu;
    SO_IstACSymbol = IstACSymbolNeu;
  }
}



/*=================================================================================================================*/
/*= V. Von der Praezedenz zur Totalordnung                                                                        =*/
/*=================================================================================================================*/

static void NeueFunktionenEinbauen(VergleichsT **Praezedenz)        /* ggf. Eintraege f>p und EQ>TRUE>FALSE setzen */
{
  if (SO_Paramodulation()) {
    SymbolT f,p;
    SO_forSpezSymbole(f)                                                 /* Jeder Operator f aus der Spezifikation */
      SO_forParaSymbole(p)                                     /* ist groesser als jeder Paramodulationsoperator p */
      Praezedenz[f][p] = Groesser;
    Praezedenz[SO_EQSymbol]  [SO_TRUESymbol]  = Groesser;
    Praezedenz[SO_TRUESymbol][SO_FALSESymbol] = Groesser;
  }
}

static void TransitivenAbschlussBilden(VergleichsT **Praezedenz)
{ 
  BOOLEAN geaendert;
  do {
    SymbolT f,g,h;
    geaendert = FALSE;
    SO_forFktSymbole(f)                                                                /* Zeilenweise durchlaufen. */
      SO_forFktSymbole(g)
        if (Praezedenz[f][g]==Groesser)                                                                     /* f>g */
          SO_forFktSymbole(h)
            if (Praezedenz[g][h]==Groesser && Praezedenz[f][h]==Unvergleichbar) {                           /* g>h */
              Praezedenz[f][h] = Groesser;                                                               /* => f>h */
              geaendert = TRUE;
            }
  } while (geaendert);
}


VergleichsT **qcmp_Praezedenz;

static int qcmp(const void *x, const void *y)
{
  SymbolT f = *(SymbolT *)x,
          g = *(SymbolT *)y;
  switch (qcmp_Praezedenz[f][g]) {
  case Groesser:
    return -1;
  case Kleiner:
    return +1;
  default:
    return 0;
  }
}

static BOOLEAN Komplettiert(VergleichsT **Praezedenz, BOOLEAN *PraezedenzTotal, SymbolT *TotaleKette)
/* Fuegt die '=' auf der Diagonalen und die zu den '>' korrespondierenden '<' ein. Dabei wird auf Zyklen getestet
   und festgestellt, ob die Praezedenz total ist.
   Wurde ein Zyklus entdeckt, so wird TRUE zurueckgeliefert.*/
{
  SymbolT i,j;
  *PraezedenzTotal = TRUE;
  SO_forFktSymbole(i)                                                /* Praezedenz-Matrix zeilenweise durchlaufen. */
    SO_forFktSymbole(j) {
      if (i==j)                                                                                      /* Diagonale. */
        if (Praezedenz[i][j]==Unvergleichbar)                                         /* Noch nicht besetzt -> '=' */
          Praezedenz[i][j] = Gleich;
        else                                                                            /* Schon besetzt -> Fehler */
          return FALSE;
      else if (Praezedenz[i][j]==Groesser)                                                                  /* i>j */
        if (Praezedenz[j][i]==Unvergleichbar || Praezedenz[j][i]==Kleiner)                               /* => j<i */
          Praezedenz[j][i] = Kleiner;
        else
          return FALSE;                                                                /* falls auch j>i => Fehler */
      if (j<i && Praezedenz[i][j] == Unvergleichbar)     /* i und j bleiben nach dem Komplettieren unvergleichbar. */
        *PraezedenzTotal = FALSE;                                                /* => Praezedenz ist nicht total. */
    }
  if (*PraezedenzTotal) {
    SO_forFktSymbole(i)
      TotaleKette[i] = i;
    qcmp_Praezedenz = Praezedenz;
    qsort(TotaleKette, SO_AnzahlFunktionen, sizeof(SymbolT), qcmp);
    /*printf("\nTotalpraezedenz:");
      SO_forFktSymbole(i) {
      func_symb fkt = DISCOUNTvonWM[TotaleKette[i]];
      printf(" %s",map_func(fkt));
      }
      printf("\n");
    */
  }
  return TRUE;                                                                         /* Kein Fehler aufgetreten. */
}



/*=================================================================================================================*/
/*= VI. Symbolinformationen und initiale Praezedenz holen                                                         =*/
/*=================================================================================================================*/

static void PraezedenzEinlesen(void)
/* Liest die in der Spezifikation angegebene Praezedenz ein und besetzt das Ordnungs-Array damit. */
{
  ord_chain ketten = ord_chain_list(get_ordering());                           /* Alle OrdnungsKetten durchlaufen. */
  while (ketten) {
    ord_chain aktKette = ord_first(ketten);                                  /* Aktuelle OrdnungsKette durchlaufen */
    while (aktKette) {
      ord_chain rest = ord_chain_rest(aktKette);                   /* Rest der aktuellen Ordnungskette durchlaufen */
      SymbolT Symbol1 = SO_AusAltMachNeu(ord_chain_first(aktKette)); 
      while (rest) {                         /* und den Vergleich mit dem ersten Element der akt. Kette eintragen. */
        SymbolT SymbolN = SO_AusAltMachNeu(ord_chain_first(rest)); 
        SO_Praezedenz[Symbol1][SymbolN] = Groesser;
        rest = ord_chain_rest(rest);
      }
      aktKette = ord_chain_rest(aktKette);
    }
    ketten = ord_rest(ketten);
  }
}


static BOOLEAN NichtGrundterm(term tt)                               /* Zum Erkennen existenzquantifizierter Ziele */
{
  if variablep(tt)
    return TRUE;
  else {
    int runny;
    for (runny=0; runny<sig_arity(term_func(tt)); runny++)
      if (NichtGrundterm(term_arg(tt,runny)))
        return TRUE;
    return FALSE;
  }                                                                            /* 'else': Konstante => Ueberlesen. */
}

static BOOLEAN ParamodulationNoetig(void)                            /* Zum Erkennen existenzquantifizierter Ziele */
{
  eq_list ziele = get_conclusions();
  SO_Paramodulation = FALSE;
  while (ziele) {
    if (NichtGrundterm(eq_first_lhs(ziele)) || NichtGrundterm(eq_first_rhs(ziele)))
      return (SO_Paramodulation = TRUE);
    ziele = eq_rest(ziele);
  }
  return FALSE;
}


#if COMP_WMTW
static BOOLEAN SO_SchonInitialisiert = FALSE;
#endif

void SO_InitAposteriori(void)
{
  func_symb_list fktsliste = get_sig(),
                 p;
  SymbolT index;
  weight Konstantengewicht = 0;

#if COMP_WMTW
  if (SO_SchonInitialisiert) return;
  else SO_SchonInitialisiert=TRUE;
#endif
  SN_SpezifikationAuswerten(&SO_GroesstesESymb);
  AnzahlSpezFunktionen = 0;
  for (p=fktsliste; p; p = sig_rest(p))                                                /* Funktionssymbole zaehlen */
    AnzahlSpezFunktionen++;
  SO_Paramodulation = ParamodulationNoetig();
  SO_AnzahlFunktionen = AnzahlSpezFunktionen + (SO_Paramodulation() ? 3 : 0);
  SO_GroesstesFktSymb = SO_AnzahlFunktionen() - 1;
  SO_GroesstesSpezSymb = AnzahlSpezFunktionen - 1;
  ArraysMitFunktionsInformationenAnlegen();                 /* allokiert und initialisiert die noetigen Strukturen */
  SO_MaximaleStelligkeit = 0;
  for (p=fktsliste; p; p = sig_rest(p)) {
    func_symb    fkt         = sig_first(p);
    SymbolT      WMIndex     =  SN_WMVonDiscountSymb(fkt);
    unsigned int Stelligkeit = sig_arity(fkt);
    weight       Gewicht     = ord_func_weight(fkt);
    DISCOUNTvonWM[WMIndex] = fkt;
    SO_Stelligkeit[WMIndex] = Stelligkeit;
    Maximieren(&SO_MaximaleStelligkeit, Stelligkeit);
    SO_FunktionsGewichte[WMIndex] = Gewicht;
    if (!Stelligkeit && (!Konstantengewicht || Gewicht<Konstantengewicht))                      /* erste Konstante */
      Konstantengewicht = ord_func_weight(fkt);                            /* oder Konstante mit kleinerem Gewicht */
  }

  Variablengewicht = Konstantengewicht>0 ? Konstantengewicht : 1;
  SO_forSymboleAbwaerts(index,SO_ErstesVarSymb(),SO_NummerVar(SO_AnzErsterVariablen))
    SO_FunktionsGewichte[index] = Variablengewicht;

  if (SO_Paramodulation()) {
    SO_EQSymbol    = SO_GroesstesSpezSymb+1;
    SO_TRUESymbol  = SO_GroesstesSpezSymb+2;
    SO_FALSESymbol = SO_GroesstesSpezSymb+3;
    SO_Stelligkeit[SO_EQSymbol]    = 2;
    SO_Stelligkeit[SO_TRUESymbol]  = 0;
    SO_Stelligkeit[SO_FALSESymbol] = 0;
    SO_forParaSymbole(index)
      SO_FunktionsGewichte[index] = Variablengewicht;
  }

  PraezedenzEinlesen();                                                             /* Praezedenz-Matrix besetzen. */
  NeueFunktionenEinbauen(SO_Praezedenz);               /* Die Spalten und Zeilen mit den neuen Konstanten besetzen */
  TransitivenAbschlussBilden(SO_Praezedenz);
  if (!Komplettiert(SO_Praezedenz, &SO_PraezedenzTotal, SO_TotaleKette))         /* die entsprechenden '<' und '=' */
    our_error("Praezedenz ist zyklisch!");                       /* einbauen und auf Zykel sowie Totalitaet testen */
}



/*=================================================================================================================*/
/*= VII. Praezedenz aus Kommandozeile lesen und auswerten                                                         =*/
/*=================================================================================================================*/

static BOOLEAN PraezedenzAnalyse(char *ps, BOOLEAN PraezedenzSetzen)
{
  VergleichsT **neuePraezedenz;
  BOOLEAN neuePraezedenzTotal;
  SymbolT *neueTotaleKette;
  
  /* 1. Neues Praezedenz-Array anlegen + initialisieren (eq,false,true beachten) */
  neuePraezedenz = Praezedenzfeld();
  /* Der Einbau der Paramodulationsfunktionen erfolgt nur in der initialen Praezedenz.
     Gibt man in der Kommandozeile eine Praezedenz an oder laesst eine solche durch Permutation oder 
     Ordnungsgenerator erzeugen, so muessen die Paramodulationssymbole explizit angegeben werden. */

  /* 2. Neue Praezedenz dort eintragen (Fehler -> freigeben, return FALSE) */
  do {
    SymbolT GroesseresSymbol = SO_ErstesVarSymb();
    do {                                                                      /* Start einer neuen Praezedenzkette */
      char *NaechsterPunkt = strchr(ps,'.');
      SymbolT Symbol;
      if (NaechsterPunkt)
        *NaechsterPunkt = '\0';
      SO_forFktSymbole(Symbol)
        if (!strcmp(ps,IO_FktSymbName(Symbol)))
          break;
      if (Symbol>SO_GroesstesFktSymb()){
        PraezedenzfeldLoeschen(neuePraezedenz);
        return FALSE;
      }
      if (SO_SymbUngleich(GroesseresSymbol,SO_ErstesVarSymb()))
        neuePraezedenz[GroesseresSymbol][Symbol] = Groesser;
      GroesseresSymbol = Symbol;
      if (NaechsterPunkt) {
        *NaechsterPunkt = '.';
        ps = NaechsterPunkt+1;
      }
      else
        ps += strlen(ps);
    } while (*ps && *ps!='.');
    ps++;
  } while (*(ps-1));

  /* 3. Trans. Abschl. bilden */
  TransitivenAbschlussBilden(neuePraezedenz);

  /* 4. Komplettieren (Fehler -> freigeben, return FALSE) */
  neueTotaleKette = array_alloc(SO_AnzahlFunktionen,SymbolT);
  if (!Komplettiert(neuePraezedenz, &neuePraezedenzTotal, neueTotaleKette)){
    PraezedenzfeldLoeschen(neuePraezedenz);
    our_dealloc(neueTotaleKette);
    return FALSE;
  }

  /* 5. Ggf. neue Praezedenz setzen */
  fprintf(IO_Strom, "\n");
  if (PraezedenzSetzen) {
    PraezedenzfeldLoeschen(SO_Praezedenz);
    SO_PraezedenzTotal = neuePraezedenzTotal;
    SO_Praezedenz = neuePraezedenz;
    our_dealloc(SO_TotaleKette);
    SO_TotaleKette = neueTotaleKette;
  }
  else {
    PraezedenzfeldLoeschen(neuePraezedenz);
    our_dealloc(neueTotaleKette);
  }
  return TRUE;
}  


BOOLEAN SO_PraezedenzKorrekt(char *ps)
{
  return PraezedenzAnalyse(ps,FALSE);
}

BOOLEAN SO_PraezedenzErneuert(char *ps)
{
  return PraezedenzAnalyse(ps,TRUE);
}

    
/*=================================================================================================================*/
/*= VIII. Testausgaben, Cleanup und Statistik                                                                     =*/
/*=================================================================================================================*/

static void PraezedenzDrucken(VergleichsT **praez)
{
  char AusgabeString[80];
  SymbolT i,j;
  SO_forFktSymbole(i){
    fprintf(IO_Strom, "  ");
    IO_DruckeSymbol(i);
  }
  IO_LF();
  SO_forFktSymbole(i){
    IO_DruckeSymbol(i);
    SO_forFktSymbole(j){
      switch (praez[i][j]){
      case Kleiner:fprintf(IO_Strom, " < ");break;
      case Groesser:fprintf(IO_Strom, " > ");break;
      case Gleich:fprintf(IO_Strom, " = ");break;
      default:fprintf(IO_Strom, " # ");
      }
    }
    sprintf(AusgabeString,"    Stelligkeit: %d   Gewicht: %ld\n",SO_Stelligkeit(i),SO_SymbGewicht(i));
    fprintf(IO_Strom, AusgabeString);
  }
}

void SO_TestAusgaben(void)
{
  fprintf(IO_Strom, "\nPraezedenz :\n");
  PraezedenzDrucken(SO_Praezedenz);
  fprintf(IO_Strom, "\nFunktionen aus Spezifikation: %d\nParamodulationsfunktionen: %d\nPraezedenz ist %s total.\n",
    AnzahlSpezFunktionen,SO_AnzahlFunktionen-AnzahlSpezFunktionen, (SO_PraezedenzTotal())? "":"nicht");
}

void SO_CleanUp(void)
{
  negarray_dealloc(SO_Stelligkeit,       GroesseStelligkeit-SO_AnzahlFunktionen());
  negarray_dealloc(SO_FunktionsGewichte, GroesseStelligkeit-SO_AnzahlFunktionen());
  negarray_dealloc(SO_IstACSymbol,       GroesseStelligkeit-SO_AnzahlFunktionen());
  our_dealloc(DISCOUNTvonWM);
  PraezedenzfeldLoeschen(SO_Praezedenz);
}


void SO_Statistik(InfoStringT IS)
{
  if (!(IS_IsIn("all", IS) || IS_IsIn("SO", IS)))
    return;
  if (IS_IsIn("intern", IS)) {
    IO_DruckeStatistikZeile("Max. Expansion of SO_Stelligkeit", GroesseStelligkeit);
    IO_DruckeStatistikZeile("Max. Expansion of SO_FunktionsGewichte", GroesseStelligkeit);
    IO_DruckeStatistikZeile("Max. Expansion of SO_IstACSymbol", GroesseStelligkeit);
  }
}
