/*=================================================================================================================
  =                                                                                                               =
  =  TermOperationen.c                                                                                            =
  =                                                                                                               =
  =  Definition des Datentyps Term und Realisation elementarer Operationen                                        =
  =                                                                                                               =
  =  02.03.94 Thomas.                                                                                             =
  =  01.02.95 Thomas. Rekursionen eliminiert.                                                                     =
  =                                                                                                               =
  =================================================================================================================*/

#include "Ausgaben.h"
#include "DSBaumOperationen.h"
#include "FehlerBehandlung.h"
#include "general.h"
#include "SpeicherVerwaltung.h"
#include "TermOperationen.h"
#include "TermIteration.h"
#include <stdarg.h>
#include <string.h>

TI_IterationDeklarieren(TO)

/* ------------------------------------------------------------------------*/
/* -------- globale Variablen (siehe TermOperationen.h) -------------------*/
/* ------------------------------------------------------------------------*/
SV_NestedSpeicherManagerT TO_FreispeicherManager;
TermT *TO_Variablenterm;


/*=================================================================================================================*/
/*= O. Initialisierung                                                                                            =*/
/*=================================================================================================================*/

static int VTermeDimension;

typedef struct VLZellenTS *VLZellenZeigerT;
typedef struct VLZellenTS {
  SymbolT VIntern;
  term VExtern;
  VLZellenZeigerT Vorg;
} VLZellenT;
#define GroesseVLZellenT sizeof(VLZellenT)
static SV_SpeicherManagerT VLZellenSpeicherManager;


static TermT Variablenterm(SymbolT Variablensymbol);


static BOOLEAN TO_SchonInitialisiert = FALSE;

void TO_InitAposteriori(void)
{
  SymbolT VarIndex;
  if (!TO_SchonInitialisiert){
    TI_IterationInitialisieren();
    SV_NestedManagerInitialisieren(TO_FreispeicherManager, TermzellenT, TO_AnzahlTermZellenJeBlock, "Termzellen");
    SV_ManagerInitialisieren(&VLZellenSpeicherManager, GroesseVLZellenT, TO_AnzVLZellenJeBlock, "TO_VLZellen");
    TO_Variablenterm = (TermT *) our_alloc(TO_AnzErsterVariablenterme*sizeof(TermT)) + TO_AnzErsterVariablenterme;
    VTermeDimension = -TO_AnzErsterVariablenterme;
  }
  TO_SchonInitialisiert = TRUE;
  SO_forSymboleAbwaerts(VarIndex,SO_ErstesVarSymb(),VTermeDimension)
    TO_Variablenterm[VarIndex] = Variablenterm(VarIndex);
}


void TO_FeldVariablentermeLeeren(void)
{
  SymbolT VarIndex;

  SO_forSymboleAbwaerts(VarIndex,SO_ErstesVarSymb(),VTermeDimension)
    TO_TermLoeschen(TO_Variablenterm[VarIndex]);
}



/*=================================================================================================================*/
/*= II. Erzeugen,Besetzen und Loeschen von Termzellen                                                             =*/
/*=================================================================================================================*/


TermT TO_NullstelligerTerm(SymbolT Symbol)
{
  TermT Term;
  TO_TermzelleHolen(Term);
  TO_SymbolBesetzen(Term,Symbol);
  TO_NachfBesetzen(Term,NULL);
  TO_EndeBesetzen(Term,Term);
  return Term;
}



/*=================================================================================================================*/
/*= IV. Einfache Praedikate auf Termen                                                                            =*/
/*=================================================================================================================*/

BOOLEAN TO_NachfolgendeTeiltermeGleich(UTermT Term1, UTermT Teilterm1, UTermT Teilterm2)
/* Prueft, ob die auf Teilterm1 folgenden Teilterme von Term1 mit den korrespondierenden Teiltermen, die Teilterm2
   nachfolgen, uebereinstimmen. Der hier nicht anzugebende term2, dessen Teilterm Teilterm2 ist, hat das gleiche
   Topsymbol wie Term1, und Teilterm1 wie Teilterm2 sind der jeweils i-te Teilterm von Term1 bzw. term2.*/
{
  TermzellenZeigerT Ende = TO_NULLTerm(Term1);
  Teilterm1 = TO_NaechsterTeilterm(Teilterm1);                                /* auf nachfolgenden Teilterm setzen */
  Teilterm2 = TO_NaechsterTeilterm(Teilterm2);
  while (TO_PhysischUngleich(Teilterm1,Ende)) {
    if (TO_TopSymboleGleich(Teilterm1,Teilterm2)) {
      Teilterm1 = TO_Schwanz(Teilterm1);                                               /* Symbole gleich -> weiter */
      Teilterm2 = TO_Schwanz(Teilterm2);
    }
    else
      return FALSE;                                                  /* unterschiedliches Symbol gefunden -> FALSE */
  }
  return TRUE;                                                                  /* erfolgreich durchgelaufen -> ok.*/
}


BOOLEAN TO_TermeGleich(UTermT Term1, UTermT Term2)
/* Prueft, ob UTerme Term1, Term2 denselben Praefix haben. Der Vergleich erfolgt von links nach rechts Symbol fuer
   Symbol. Abbruch, falls unterschiedliche Symbole oder Term1 abgearbeitet. Dass Term2 abgearbeitet ist und die-
   selben Symbole hat wie Term1 und Term1 noch nicht abgearbeitet ist, kommt wegen der Termstruktur nicht vor.     */
{
  TermzellenZeigerT Ende1;

  Ende1 = TO_NULLTerm(Term1);
  while (SO_SymbGleich(TO_TopSymbol(Term1),TO_TopSymbol(Term2))) {
    if (TO_PhysischGleich(Term1 = TO_Schwanz(Term1),Ende1))
      return TRUE;
    Term2 = TO_Schwanz(Term2);
  }
  return FALSE;
}


static BOOLEAN TopsymbolKorrekt(UTermT Term)
{
  SymbolT Symbol = TO_TopSymbol(Term);
  if (!(2*BO_NeuesteBaumvariable()<=Symbol && Symbol<=SO_GroesstesFktSymb())) {
    fprintf(IO_Strom, "\nFehler: Unzulaessiges Symbol.\n");
    return FALSE;
  }
  return TRUE;
}


BOOLEAN TO_TermKorrekt(TermT Term)                                       /* Prueft Termdarstellung auf Konsistenz. */
{                                                     /* Nur fuer Terme aus Regeln, Gleichungen, Kritischen Paaren */
  unsigned int TI_KSIndexAlt = TI_KSIndex;

  if (TO_NULLTerm(Term)) {
    fprintf(IO_Strom, "\nFehler: Beschnittener Term hat am Ende Nachfolger.\n");
    return FALSE;
  }
  if (!TopsymbolKorrekt(Term)) {
    fprintf(IO_Strom, "\nFehler: Term hat unzulaessiges Topsymbol.\n");
    return FALSE;
  }
  if (TO_Nullstellig(Term)) {
    if (TO_TermEnde(Term)!=Term) {
      fprintf(IO_Strom, "\nFehler: Ende-Zeiger von Konstanter weist nicht auf Konstante.\n");
      return FALSE;
    }
    return TRUE;
  }
  TI_Push(Term);
  do {
    if (!(Term = TO_Schwanz(Term))) {
      fprintf(IO_Strom, "\nFehler: Term endet vorzeitig.\n");
      TI_KSIndex = TI_KSIndexAlt;
      return FALSE;
    }
    if (!TopsymbolKorrekt(Term)) {
      fprintf(IO_Strom, "\nFehler: Term hat im Inneren unzulaessiges Symbol.\n");
      TI_KSIndex = TI_KSIndexAlt;
      return FALSE;
    }
    if (TO_Nullstellig(Term)) {
      if (TO_TermEnde(Term)!=Term) {
        fprintf(IO_Strom, "\nFehler: Ende-Zeiger von nullstelligem Teilterm weist nicht auf diesen Teilterm.\n");
        TI_KSIndex = TI_KSIndexAlt;
        return FALSE;
      }
      while (!(--TI_TopZaehler)) {
        if (Term!=TO_TermEnde(TI_TopTeilterm)) {
          fprintf(IO_Strom, "\nFehler: Ende-Zeiger weist nicht auf tatsaechliches (Teilterm-)Ende.\n");
          TI_KSIndex = TI_KSIndexAlt;
          return FALSE;
        }
        if (TI_PopUndLeer())
          return TRUE;
      }
    }
    else
      TI_Push(Term);
  } while (TRUE);
}


static unsigned int AbgeflachtGrad(UTermT Term)
{
  unsigned int Grad = 2;                                                    /* Voraussetzung: top(t) ist AC-Symbol */
  SymbolT ACSymbol = TO_TopSymbol(Term);
  while (SO_SymbGleich(ACSymbol,TO_TopSymbol(Term = TO_NaechsterTeilterm(TO_ErsterTeilterm(Term)))))
    Grad++;
  return Grad;
}

static void TermzelleMarkieren(TermzellenZeigerT Zelle)
{
  SymbolT Symbol = TO_TopSymbol(Zelle);
  Symbol = -Symbol+2*SO_GroesstesFktSymb()+1;
  TO_SymbolBesetzen(Zelle,Symbol);
}

#define /*void*/ TermzelleEntmarkieren(/*TermzellenZeigerT*/ Zelle)                                                 \
  TermzelleMarkieren(Zelle)

#define /*BOOLEAN*/ TermzelleIstMarkiert(/*TermzellenZeigerT*/ Zelle)                                               \
  (TO_TopSymbol(Zelle)>SO_GroesstesFktSymb())

#define ACnVerschieden 0

static unsigned int ACnGleich(UTermT s, UTermT t)
{
  SymbolT Symbol = TO_TopSymbol(s);
  if (SO_SymbUngleich(Symbol,TO_TopSymbol(t)))
    return ACnVerschieden;
  else if (!SO_Stelligkeit(Symbol))
    return 1;
  else if (SO_IstACSymbol(Symbol)) {
    unsigned int m = AbgeflachtGrad(s);
    if (m!=AbgeflachtGrad(t))
      return ACnVerschieden;
    else {
      #define /*BOOLEAN*/ Weiter(/*Bezeichner*/i)                                                                   \
      (i##Letzter ? FALSE :                                                                                         \
       SO_SymbGleich(Symbol,TO_TopSymbol(i = TO_NaechsterTeilterm(i))) ?                                            \
       (i = TO_ErsterTeilterm(i), TRUE) :                                                                           \
       (i##Letzter = TRUE, TRUE))
      BOOLEAN sLetzter = FALSE;
      #define /*BOOLEAN*/ sWeiter(/*void*/) Weiter(s)
      UTermT t1;
      BOOLEAN t1Letzter;
      #define /*void*/ t1Start(/*void*/)                                                                            \
        t1 = TO_ErsterTeilterm(t); t1Letzter = FALSE
      #define /*BOOLEAN*/ t1Weiter(/*void*/) Weiter(t1)
      #define /*void*/ t1Austragen(/*void*/)                                                                        \
        TermzelleMarkieren(t1)
      #define /*void*/ tRestaurieren(/*void*/)                                                                      \
        t1Start();                                                                                                  \
        do {                                                                                                        \
          if (TermzelleIstMarkiert(t1))                                                                             \
            TermzelleEntmarkieren(t1);                                                                              \
        } while (t1Weiter());

      s = TO_ErsterTeilterm(s);
      do {
        t1Start();
        do {
          unsigned int m1 = ACnGleich(s,t1);
          if (m1!=ACnVerschieden) {
            if (m1>m)
              m = m1;
            t1Austragen();
            goto Weiter;
          }
        } while (t1Weiter());
        tRestaurieren();
        return ACnVerschieden;
        Weiter:;
      } while (sWeiter());
      tRestaurieren();
      return m;
    }
  }
  else {                                                                             /* Nun fuer Non-AC(n)-Symbol. */
    unsigned int m = 1;
    UTermT sNULL = TO_NULLTerm(s);
    s = TO_ErsterTeilterm(s);
    t = TO_ErsterTeilterm(t);
    do {
      unsigned int m1 = ACnGleich(s,t);
      if (m1==ACnVerschieden)
        return ACnVerschieden; 
      if (m1>m)
        m = m1;
      s = TO_NaechsterTeilterm(s);
      t = TO_NaechsterTeilterm(t);
    } while (TO_PhysischUngleich(s,sNULL));
    return m;
  }
}


BOOLEAN TO_ACnGleich(unsigned int n, UTermT s, UTermT t)
{
  unsigned int ACn = ACnGleich(s,t);
  /*IO_DruckeFlex("AC%d %t %t %d\n",n,s,t,ACn);*/
  return ACn!=ACnVerschieden && ACn>=n;
}

/* Hier wurde implementiert |-|(n) := ={f(x_)=f(pi(x_)): x_ aus V^n, pi aus PI(n)} auf Term(F*,V)
   (repraesentiert durch Term(F,V)/=A mit rechtsgeklammerten Repraesentanten in Term(F,V))
   wobei in F* das Funktionssymbol f variable Stelligkeit hat, die f-Terme abgeflacht sein
   muessen und auf einen k-stelligen f-Term nur die Permutationen vom Grad k anwendbar sind,
   und damit s=(n)t := s =perm t  und nicht  s (|-|1 U ... U |-|n-1)+ t. */


BOOLEAN TO_HomoeomorphDrin(UTermT s, UTermT t)
{  /* Ist s homoeomorph in t eingebettet? Das ganze inklusive Aequivalenzanteil (==syntaktischer Gleichheit). */
  if (TO_TermIstVar(s)) {
    SymbolT x = TO_TopSymbol(s);
    UTermT NULLt = TO_NULLTerm(t);
    do {
      if (SO_SymbGleich(x,TO_TopSymbol(t)))
        return TRUE;
    } while (TO_PhysischUngleich(t = TO_Schwanz(t),NULLt));
    return FALSE;
  }
  else if (TO_TermIstVar(t))
    return FALSE;
  else {
      UTermT NULLt = TO_NULLTerm(t),
             j = TO_ErsterTeilterm(t);

    /* alpha */
    if TO_TopSymboleGleich(s,t) {
      UTermT i = TO_ErsterTeilterm(s);
      while (TO_PhysischUngleich(j,NULLt)) {
        if (!TO_HomoeomorphDrin(i,j))
          break;
        i = TO_NaechsterTeilterm(i);
        j = TO_NaechsterTeilterm(j);
      }
      if (TO_PhysischGleich(j,NULLt))
        return TRUE;
      j = TO_ErsterTeilterm(t);
    }

    /* beta */
    while (TO_PhysischUngleich(j,NULLt)) {
      if (TO_HomoeomorphDrin(s,j))
        return TRUE;
      j  = TO_NaechsterTeilterm(j);
    }

    return FALSE;
  }
}


BOOLEAN TO_SymbolEnthalten(SymbolT f, UTermT t)
{ /* Ist die f-Laenge von t positiv? */
  SymbolT g;
  TO_doSymbole(t, g, {
    if (SO_SymbGleich(f,g))
      return TRUE;
  });
  return FALSE;
}

BOOLEAN TO_VarGroesserGleich(UTermT s, UTermT t)
{ /* Ist Var(s) Obermenge von Var(t)? -- Effizientere Basisoperationen dafuer in VariablenMengen.h! */
  SymbolT x;
  TO_doSymbole(t, x, {
    if (SO_SymbIstVar(x) && !TO_SymbolEnthalten(x,s))
      return FALSE;
  });
  return TRUE;
}


BOOLEAN TO_DefinierendeGleichung(UTermT s, UTermT t)
{ /* Ist s=t eine definierende Gleichung?
     Gdw. s==f(x(i1), ..., x(in)); |t|(f)=0; Var(s)>=Var(t)
     Achtung: s und t werden von links nach rechts normiert. */
  if (TO_TermIstVar(s))
    return FALSE;
  else {
    UTermT si;
    SymbolT f = TO_TopSymbol(s);
    unsigned int i=1;
    BO_TermpaarNormierenAlsKP(s,t);
    TO_forTeilterme(s, si, {
      SymbolT fi = TO_TopSymbol(si);
      SymbolT xi = SO_NummerVar(i++);
      if (SO_SymbUngleich(fi,xi))
        return FALSE;
    });
    return !TO_SymbolEnthalten(f,t) && TO_VarGroesserGleich(s,t);
  }
}



/*=================================================================================================================*/
/*= VI. Funktionen zum Kopieren                                                                                   =*/
/*=================================================================================================================*/

/* ------- 1. Allgemeine Kopierfunktion -------------------------------------------------------------------------- */

TermT TO_Termkopie(UTermT TermAlt)
{                               /* Erzeugt Kopie des Praefixes von TermAlt. Nachf- und Ende-Zeiger werden besetzt. */
  if (TO_Nullstellig(TermAlt)) {
    TermT TermNeu;
    TO_TermzelleHolen(TermNeu);
    TO_SymbolBesetzen(TermNeu,TO_TopSymbol(TermAlt));
    TO_NachfBesetzen(TermNeu,NULL);
    TO_EndeBesetzen(TermNeu,TermNeu);
    return TermNeu;
  }
  else {
    TermT TermNeu = TO_HoleErsteNeueZelle(),
          IndexNeu = TermNeu;
    TO_SymbolBesetzen(TermNeu,TO_TopSymbol(TermAlt));
    TI_Push(TermNeu);
    do {
      TO_AufNaechsteNeueZelle(IndexNeu);
      TO_SymbolBesetzen(IndexNeu,TO_TopSymbol(TermAlt = TO_Schwanz(TermAlt)));
      if (TO_Nullstellig(TermAlt)) {
        TO_EndeBesetzen(IndexNeu,IndexNeu);
        while (!(--TI_TopZaehler)) {
          TO_EndeBesetzen(TI_TopTeilterm,IndexNeu);
          if (TI_PopUndLeer()) {
            TO_LetzteNeueZelleMelden(IndexNeu);
            TO_NachfBesetzen(IndexNeu,NULL);
            return TermNeu;
          }
        }
      }
      else
        TI_Push(IndexNeu);
    } while (TRUE);
  }
}


/* ------- 2. Besondere Funktionen fuer MatchOperatioen ---------------------------------------------------------- */

TermT TO_Termkopie0(UTermT TermAlt)                                 /* wie vor, aber Ende-Nachf-Zeiger undefiniert */
{
  if (TO_Nullstellig(TermAlt)) {
    TermT TermNeu;
    TO_TermzelleHolen(TermNeu);
    TO_SymbolBesetzen(TermNeu,TO_TopSymbol(TermAlt));
    TO_EndeBesetzen(TermNeu,TermNeu);
    return TermNeu;
  }
  else {
    TermT TermNeu = TO_HoleErsteNeueZelle(),
          IndexNeu = TermNeu;
    TO_SymbolBesetzen(TermNeu,TO_TopSymbol(TermAlt));
    TI_Push(TermNeu);
    do {
      TO_AufNaechsteNeueZelle(IndexNeu);
      TO_SymbolBesetzen(IndexNeu,TO_TopSymbol(TermAlt = TO_Schwanz(TermAlt)));
      if (TO_Nullstellig(TermAlt)) {
        TO_EndeBesetzen(IndexNeu,IndexNeu);
        while (!(--TI_TopZaehler)) {
          TO_EndeBesetzen(TI_TopTeilterm,IndexNeu);
          if (TI_PopUndLeer()) {
            TO_LetzteNeueZelleMelden(IndexNeu);
            return TermNeu;
          }
        }
      }
      else
        TI_Push(IndexNeu);
    } while (TRUE);
  }
}


TermT TO_TermkopieEZ(UTermT TermAlt, TermzellenZeigerT ersteTermzelle)     /* Erzeugt eine Kopie des Praefixes von */
{                /* TermAlt. Nachf- und Ende-Zeiger werden besetzt. Erste Termzelle der Kopie wird ersteTermzelle. */
  UTermT IndexNeu = TO_HoleErsteNeueZelle();                                           /* TermAlt ist mehrstellig. */
  TO_SymbolBesetzen(ersteTermzelle,TO_TopSymbol(TermAlt));
  TI_Push(ersteTermzelle);
  {              /* gesonderter erster Schleifendurchlauf, da einmal Aufruf von TO_HoleErsteNeueZelle erforderlich */
    TO_NachfBesetzen(ersteTermzelle,IndexNeu);
    TO_SymbolBesetzen(IndexNeu,TO_TopSymbol(TermAlt = TO_Schwanz(TermAlt)));
    if (TO_Nullstellig(TermAlt)) {
      TO_EndeBesetzen(IndexNeu,IndexNeu);
      while (!(--TI_TopZaehler)) {
        TO_EndeBesetzen(TI_TopTeilterm,IndexNeu);
        if (TI_PopUndLeer()) {
          TO_LetzteNeueZelleMelden(IndexNeu);
          TO_NachfBesetzen(IndexNeu,NULL);
          return ersteTermzelle;
        }
      }
    }
    else
      TI_Push(IndexNeu);
  }
  do {
    TO_AufNaechsteNeueZelle(IndexNeu);
    TO_SymbolBesetzen(IndexNeu,TO_TopSymbol(TermAlt = TO_Schwanz(TermAlt)));
    if (TO_Nullstellig(TermAlt)) {
      TO_EndeBesetzen(IndexNeu,IndexNeu);
      while (!(--TI_TopZaehler)) {
        TO_EndeBesetzen(TI_TopTeilterm,IndexNeu);
        if (TI_PopUndLeer()) {
          TO_LetzteNeueZelleMelden(IndexNeu);
          TO_NachfBesetzen(IndexNeu,NULL);
          return ersteTermzelle;
        }
      }
    }
    else
      TI_Push(IndexNeu);
  } while (TRUE);
}


TermT TO_TermkopieEZx1Subst(UTermT TermAlt, TermzellenZeigerT ersteTermzelle)            /* Erzeugt eine Kopie des */
{  /* Praefixes von TermAlt. Nachf- und Ende-Zeiger werden besetzt. Erste Termzelle der Kopie wird ersteTermzelle. */
                         /* TermAlt ist mehrstellig. Ausserdem wird x1 durch TO_Topsymbol(ersteTermzelle) ersetzt. */
  SymbolT Substitut = TO_TopSymbol(ersteTermzelle);
  UTermT IndexNeu = TO_HoleErsteNeueZelle();
  TO_SymbolBesetzen(ersteTermzelle,TO_TopSymbol(TermAlt));
  TI_Push(ersteTermzelle);
  {
    TO_NachfBesetzen(ersteTermzelle,IndexNeu);
    TermAlt = TO_Schwanz(TermAlt);
    if (TO_Nullstellig(TermAlt)) {
      SymbolT Topsymbol = TO_TopSymbol(TermAlt);
      TO_SymbolBesetzen(IndexNeu,SO_SymbGleich(Topsymbol,SO_ErstesVarSymb()) ? Substitut : Topsymbol);
      TO_EndeBesetzen(IndexNeu,IndexNeu);
      while (!(--TI_TopZaehler)) {
        TO_EndeBesetzen(TI_TopTeilterm,IndexNeu);
        if (TI_PopUndLeer()) {
          TO_LetzteNeueZelleMelden(IndexNeu);
          TO_NachfBesetzen(IndexNeu,NULL);
          return ersteTermzelle;
        }
      }
    }
    else {
      TO_SymbolBesetzen(IndexNeu,TO_TopSymbol(TermAlt));
      TI_Push(IndexNeu);
    }
  }
  do {
    TO_AufNaechsteNeueZelle(IndexNeu);                                                                              \
    TermAlt = TO_Schwanz(TermAlt);
    if (TO_Nullstellig(TermAlt)) {
      SymbolT Topsymbol = TO_TopSymbol(TermAlt);
      TO_SymbolBesetzen(IndexNeu,SO_SymbGleich(Topsymbol,SO_ErstesVarSymb()) ? Substitut : Topsymbol);
      TO_EndeBesetzen(IndexNeu,IndexNeu);
      while (!(--TI_TopZaehler)) {
        TO_EndeBesetzen(TI_TopTeilterm,IndexNeu);
        if (TI_PopUndLeer()) {
          TO_LetzteNeueZelleMelden(IndexNeu);
          TO_NachfBesetzen(IndexNeu,NULL);
          return ersteTermzelle;
        }
      }
    }
    else {
      TO_SymbolBesetzen(IndexNeu,TO_TopSymbol(TermAlt));
      TI_Push(IndexNeu);
    }
  } while (TRUE);
}


TermT TO_TermkopieLZm(UTermT TermAlt, TermzellenZeigerT letzteTermzelle)
/* Erzeugt eine Kopie des Praefixes von TermAlt. Letzte Termzelle der Kopie wird letzteTermzelle. Deren Nachfol-
   ger-Zeiger wird nicht neu besetzt. Ebensowenig deren Ende-Zeiger; dieser muss also schon auf letzteTermzelle    */
          /* weisen. Aufruf erfolgt nur in MatchOperationen; Zusicherung gilt dort. TermAlt muss mehrstellig sein. */
#define RumpfTO_TermkopieLZm                                                                                        \
{                                                                                                                   \
  TermT TermNeu = TO_HoleErsteNeueZelle(),                                                                          \
        TermAltEnde = TO_TermEnde(TermAlt);                                                                         \
  UTermT IndexNeu = TermNeu;                                                                                        \
                                                                                                                    \
  TO_SymbolBesetzen(TermNeu,TO_TopSymbol(TermAlt));                                                                 \
  TI_Push(TermNeu);                                                                                                 \
  while (TO_PhysischUngleich(TermAlt = TO_Schwanz(TermAlt),TermAltEnde)) {                                          \
    TO_AufNaechsteNeueZelle(IndexNeu);                                                                              \
    TO_SymbolBesetzen(IndexNeu,TO_TopSymbol(TermAlt));                                                              \
    if (TO_Nullstellig(TermAlt)) {                                                                                  \
      TO_EndeBesetzen(IndexNeu,IndexNeu);                                                                           \
      while (!(--TI_TopZaehler)) {                                                                                  \
        TO_EndeBesetzen(TI_TopTeilterm,IndexNeu);                                                                   \
        TI_Pop();                                                                                                   \
      }                                                                                                             \
    }                                                                                                               \
    else                                                                                                            \
      TI_Push(IndexNeu);                                                                                            \
  }                                                                                                                 \
                                                                                                                    \
  TO_LetzteNeueZelleMelden(IndexNeu);                                                                               \
  TO_NachfBesetzen(IndexNeu,letzteTermzelle);                                                                       \
  TO_SymbolBesetzen(letzteTermzelle,TO_TopSymbol(TermAltEnde));                                                     \
  do                                                                                                                \
    TO_EndeBesetzen(TI_TopTeilterm,letzteTermzelle);                                                                \
  while (TI_PopUndNichtLeer());                                                                                     \
                                                                                                                    \
  return TermNeu;                                                                                                   \
}
RumpfTO_TermkopieLZm


TermT TO_TermkopieLZ(UTermT TermAlt, TermzellenZeigerT letzteTermzelle)
/* Erzeugt eine Kopie des Praefixes von TermAlt. Letzte Termzelle der Kopie wird letzteTermzelle. Deren Nachfol-
   ger-Zeiger wird nicht neu besetzt. Ebensowenig deren Ende-Zeiger; dieser muss also schon auf letzteTermzelle    */
{                                        /* weisen. Aufruf erfolgt nur in MatchOperationen; Zusicherung gilt dort. */
  if (TO_Nullstellig(TermAlt)) {
    TO_SymbolBesetzen(letzteTermzelle,TO_TopSymbol(TermAlt));
    return letzteTermzelle;
  }
  else
    RumpfTO_TermkopieLZm;
}     


TermT TO_TermkopieLZx1Subst(UTermT TermAlt, TermzellenZeigerT letzteTermzelle)
{                     /* Wie TO_TermkopieLZm, aber zusaetzlich wird x1 durch TO_Topsymbol(ersteTermzelle) ersetzt. */
  SymbolT Substitut = TO_TopSymbol(letzteTermzelle);
  TermT TermNeu = TO_HoleErsteNeueZelle(),
        TermAltEnde = TO_TermEnde(TermAlt);
  UTermT IndexNeu = TermNeu;
  
  TO_SymbolBesetzen(TermNeu,TO_TopSymbol(TermAlt));
  TI_Push(TermNeu);
  while (TO_PhysischUngleich(TermAlt = TO_Schwanz(TermAlt),TermAltEnde)) {
    TO_AufNaechsteNeueZelle(IndexNeu);
    if (TO_Nullstellig(TermAlt)) {
      SymbolT Topsymbol = TO_TopSymbol(TermAlt);
      TO_SymbolBesetzen(IndexNeu,SO_SymbGleich(Topsymbol,SO_ErstesVarSymb()) ? Substitut : Topsymbol);
      TO_EndeBesetzen(IndexNeu,IndexNeu);
      while (!(--TI_TopZaehler)) {
        TO_EndeBesetzen(TI_TopTeilterm,IndexNeu);
        TI_Pop();
      }
    }
    else {
     TO_SymbolBesetzen(IndexNeu,TO_TopSymbol(TermAlt));
     TI_Push(IndexNeu);
    }
  }
  
  TO_LetzteNeueZelleMelden(IndexNeu);
  TO_NachfBesetzen(IndexNeu,letzteTermzelle);
  TO_SymbolBesetzen(letzteTermzelle,TO_TopSymbol(TermAltEnde));
  do
    TO_EndeBesetzen(TI_TopTeilterm,letzteTermzelle);
  while (TI_PopUndNichtLeer());
  
  return TermNeu;
}     


/* ------- 3. Besondere Funktion fuer Unifikation ---------------------------------------------------------------- */

TermT TO_TermkopieVarUmbenannt(UTermT TermAlt, SymbolT NeuesteVar)
{                          /* Erzeugt eine Kopie des Praefixes von TermAlt. Nachf- und Ende-Zeiger werden besetzt. */
  if (TO_Nullstellig(TermAlt)) {
    TermT TermNeu;
    SymbolT Topsymbol = TO_TopSymbol(TermAlt);
    TO_TermzelleHolen(TermNeu);
    TO_SymbolBesetzen(TermNeu,SO_SymbIstVar(Topsymbol) ? Topsymbol+NeuesteVar : Topsymbol);
    TO_NachfBesetzen(TermNeu,NULL);
    TO_EndeBesetzen(TermNeu,TermNeu);
    return TermNeu;
  }
  else {
    TermT TermNeu = TO_HoleErsteNeueZelle(),
          IndexNeu = TermNeu;
    TO_SymbolBesetzen(TermNeu,TO_TopSymbol(TermAlt));
    TI_Push(TermNeu);
    do {
      TO_AufNaechsteNeueZelle(IndexNeu);
      TermAlt = TO_Schwanz(TermAlt);
      if (TO_Nullstellig(TermAlt)) {
        SymbolT Topsymbol = TO_TopSymbol(TermAlt);
        TO_SymbolBesetzen(IndexNeu,SO_SymbIstVar(Topsymbol) ? Topsymbol+NeuesteVar : Topsymbol);
        TO_EndeBesetzen(IndexNeu,IndexNeu);
        while (!(--TI_TopZaehler)) {
          TO_EndeBesetzen(TI_TopTeilterm,IndexNeu);
          if (TI_PopUndLeer()) {
            TO_LetzteNeueZelleMelden(IndexNeu);
            TO_NachfBesetzen(IndexNeu,NULL);
            return TermNeu;
          }
        }
      }
      else {
        TO_SymbolBesetzen(IndexNeu,TO_TopSymbol(TermAlt));
        TI_Push(IndexNeu);
      }
    } while (TRUE);
  }
}



/*=================================================================================================================*/
/*= VIII. Funktionen zur Umwandlung von externer in interne Termdarstellung                                       =*/
/*=================================================================================================================*/

#define VListenT VLZellenZeigerT
static VListenT Variablenliste;                  /* Erlaeuterung der Struktur siehe unter Funktion PlattklopfenRek */


static VLZellenZeigerT VLZelleHolen(SymbolT VIntern, term VExtern, VLZellenZeigerT Vorg)
{
  VLZellenZeigerT Index;
  SV_Alloc(Index,VLZellenSpeicherManager,VLZellenZeigerT);
  Index->VIntern = VIntern;
  Index->VExtern = VExtern;
  Index->Vorg = Vorg;
  return Index;
}


static void VariablenlisteLoeschen(void)
{
  VLZellenZeigerT Vorg;
  do {
    Vorg = Variablenliste->Vorg;
    SV_Dealloc(VLZellenSpeicherManager,Variablenliste);
  } while ((Variablenliste=Vorg));
}


static TermT Variablenterm(SymbolT Variablensymbol)
{
  TermT Term;
  TO_TermzelleHolen(Term);
  TO_SymbolBesetzen(Term,Variablensymbol);
  TO_NachfBesetzen(Term,NULL);
  TO_EndeBesetzen(Term,Term);
  return Term;
}


static TermT PlattklopfenRek(term Term)
/* Interne Darstellung von Term erzeugen und zurueckgeben. Variablen werden mit Hilfe der dateiglobalen Variable
   Variablenliste negativen ganzen Zahlen zugeordnet; die Zuordnung der Funktionssymbole wird aus SymbolOperationen
   uebernommen. Variablenliste ist ein Stapel der Gestalt
   ((n. Variable extern | Nummer der n. Variable intern) ...(1. Variable extern | Nummer der 1. Variable intern)),
   so dass die interne Nummer beim obersten Stapelelement vermindert um eins die Nummer der jeweils naechsten
   neuen Variable ist. Beim ersten Aufruf von PlattklopfenRek muss Variablenliste daher initialisiert sein auf
   ((neue externe Variable, die nirgends sonst vorkommt | 0)).                                                     */
{
  if (variablep(Term)) {                                                                        /* fuer Variablen: */
    VLZellenZeigerT Index;
    Index = Variablenliste;
    while (Index && !term_eq(Index->VExtern,Term))                 /* nachsehen, ob Variable schon aufgetreten ist */
      Index = Index->Vorg;
    if (Index)                                         /* falls ja: Termzelle mit zugeordneter Nummer zurueckgeben */
      return Variablenterm(Index->VIntern);
    else {                                       /* sonst neue Nummer vergeben und entsprechende Termzelle zurueck */
      SymbolT neueNummer;
      neueNummer = Variablenliste->VIntern-1;
      Variablenliste = VLZelleHolen(neueNummer,Term,Variablenliste);         /* ausserdem Variablenliste erweitern */
      return Variablenterm(neueNummer);
    } 
 }
  else {                                                                                       /* fuer Konstanten: */
    short int i;
    UTermT Flunder=NULL, Teilterm=NULL, Ende=NULL;        /* alle Teilterme in umgekehrter Reihenfolge durchlaufen */
    for(i=sig_arity(term_func(Term))-1; i>=0; i--) {   /* Invariante: Flunder zeigt auf zuletzt erzeugten Teilterm */
      TO_NachfBesetzen(TO_TermEnde(Teilterm = PlattklopfenRek(term_arg(Term,i))),Flunder);     /* Nachf-Verkettung */
      if (!Ende)                       /* vermerkt Endzeiger des letzten Teilterms; wird Endzeiger des Gesamtterms */
        Ende = TO_TermEnde(Teilterm);
      Flunder = Teilterm;
    }
    TO_TermzelleHolen(Flunder);                                                 /* Termzelle fuer Topsymbol bilden */
    TO_NachfBesetzen(Flunder,Teilterm);
    TO_EndeBesetzen(Flunder,Ende ? Ende : Flunder);   /* Ende-Zeiger mit Ende aus Schleife besetzen bzw. mit Term- */
    TO_SymbolBesetzen(Flunder,SO_AusAltMachNeu(term_func(Term)));          /* zelle selbst, falls Konstantensymbol */
    return Flunder;
  }
}


TermT TO_Plattklopfen(term Term)

{
  TermT Flunder;

  Variablenliste = VLZelleHolen(0,create_var(sorts_first(get_sorts())),NULL);
  Flunder = PlattklopfenRek(Term);
  SO_VariablenfeldAnpassen(Variablenliste->VIntern);
  TO_FeldVariablentermeAnpassen(Variablenliste->VIntern);
  VariablenlisteLoeschen();
  return Flunder;
}


void TO_PlattklopfenLR(term Term1, term Term2, TermT *Flunder1, TermT *Flunder2)
/* Aufruf von PlattklopfenRek mit Initialisierung der Variablenliste; danach Loeschen derselbigen.
   Fuer Term2 mit bereits in Term1 erklaerten Variablenersetzungen. */
{
  Variablenliste = VLZelleHolen(0,create_var(sorts_first(get_sorts())),NULL);
  *Flunder1 = PlattklopfenRek(Term1);
  *Flunder2 = PlattklopfenRek(Term2);
  SO_VariablenfeldAnpassen(Variablenliste->VIntern);
  TO_FeldVariablentermeAnpassen(Variablenliste->VIntern);
  VariablenlisteLoeschen();
}



/*=================================================================================================================*/
/*= IX. Sonstiges                                                                                                 =*/
/*=================================================================================================================*/

unsigned long int TO_StelleInTerm(UTermT Stelle, TermT Term)                 /* Stelle des Teiltermzeigers als int */
{
   UTermT Index;
   unsigned long int zaehlen = 0;

   TO_doStellen(Term, Index,{
      if (TO_PhysischGleich(Index, Stelle))
         return zaehlen;
      zaehlen++;
   });
   IO_DruckeFlex("\nFEHLER: Stelle nicht im Term!\nStelle %t, Term %t\n", Stelle, Term);
   return 0;
}

UTermT TO_TermAnStelle(TermT Term, unsigned long int Stellennr)                          /* Teiltermzeigers von int */
{
  UTermT Stelle = Term;
  while (Stellennr--)
    Stelle = TO_Schwanz(Stelle);
  return Stelle;
}

unsigned long int TO_Termlaenge(UTermT Term)             /* Laenge (= Anzahl von Symbolen) eines Termes bestimmen. */
{
  unsigned long Laenge = 1;
  UTermT NULLTerm = TO_NULLTerm(Term);
  while (TO_PhysischUngleich(Term = TO_Schwanz(Term),NULLTerm))
    Laenge++;
  return Laenge;
}


unsigned int TO_Termtiefe(UTermT Term)
{
  if (TO_Nullstellig(Term))
    return 1;
  else {
    unsigned int maximaleTiefeMinus1 = 1;
    TI_PushOT(Term);        /* TI_Stapelhoehe()+1 ist aktuelle Termtiefe fuer den Fall, dass Term gerade auf einen */
    do {                            /* nullstelligen Teilterm zeigt. Da nur dort Maxima angenommen werden koennen, */
      Term = TO_Schwanz(Term);
      if (TO_Nullstellig(Term)) {                              /* muss maximaleTiefe nur dort aktualisiert werden. */
        if (TI_Stapelhoehe()>maximaleTiefeMinus1)
          maximaleTiefeMinus1 = TI_Stapelhoehe();
        while (!(--TI_TopZaehler))
          if (TI_PopUndLeer())
            return maximaleTiefeMinus1+1;
      }
      else
        TI_PushOT(Term)
    } while (TRUE);
  }
}


SymbolT TO_SymbolNeuesterVariablenterm(void)
{
  return VTermeDimension;
}


void TO_FeldVariablentermeAnpassen(SymbolT NeuesteBaumvariable)
{
  if (SO_VariableIstNeuer(NeuesteBaumvariable,VTermeDimension)) {
    int VTermeDimensionNeu = -TO_AnzErsterVariablenterme*((-NeuesteBaumvariable)/TO_AnzErsterVariablenterme+1);
    UTermT *VariablentermeNeu = (UTermT *) our_alloc(-VTermeDimensionNeu*sizeof(TermT)) - VTermeDimensionNeu;
    SymbolT VarIndex;                             /* Addition von -VTermeDimensionNeu wegen negativer Adressierung */
    SO_forSymboleAbwaerts(VarIndex,SO_ErstesVarSymb(),VTermeDimension)
      VariablentermeNeu[VarIndex] = TO_Variablenterm[VarIndex];
    SO_forSymboleAbwaerts(VarIndex,VTermeDimension-1,VTermeDimensionNeu)
      VariablentermeNeu[VarIndex] = Variablenterm(VarIndex);
    our_dealloc(TO_Variablenterm+VTermeDimension);
    VTermeDimension = VTermeDimensionNeu;
    TO_Variablenterm = VariablentermeNeu;
  }
 }


UTermT TO_VorigeTermzelle(UTermT Teilterm, UTermT Term)
/* Zeiger auf die der Termzelle Teilterm im Term Term vorangehende Termzelle */
{
  if (TO_PhysischGleich(Term,Teilterm))
    return NULL;
  else {
    UTermT TermVorg;
    while (TO_PhysischUngleich(Term = TO_Schwanz(TermVorg = Term),Teilterm));
    return TermVorg;
  }
}


UTermT TO_VorletzteZelle(UTermT Term)    /* Zeiger auf vorletzte Zelle von Term. Term darf nicht nullstellig sein. */
{
  do
    if (TO_MehrstelligerTermEinstellig(Term))
      if (TO_Nullstellig(TO_ErsterTeilterm(Term)))
        return Term;
      else
        Term = TO_ErsterTeilterm(Term);
    else {
      UTermT vorletzterTeilterm = TO_ErsterTeilterm(Term),
             letzterTeilterm = TO_NaechsterTeilterm(vorletzterTeilterm);
      while (TO_PhysischUngleich(TO_TermEnde(letzterTeilterm),TO_TermEnde(Term)))
        letzterTeilterm = TO_NaechsterTeilterm(vorletzterTeilterm = letzterTeilterm);
      if (TO_Nullstellig(letzterTeilterm))
        return TO_TermEnde(vorletzterTeilterm);
      else
        Term = letzterTeilterm;
    }
  while (TRUE);
}


void TO_VariablenUmbenennen(TermT Term, SymbolT Delta)
{                                               /* Sei Delta = x(q). Jede Variable x(i) wird ersetzt durch x(i+q). */
  do
    if (TO_TermIstVar(Term))
      TO_SymbolBesetzen(Term,TO_TopSymbol(Term)+Delta);
  while ((Term = TO_Schwanz(Term)));
}


VergleichsT TO_EntscheidungSymbLex(TermT Term1, TermT Term2)        /* stellt Verhaeltnis von Term1 und Term2 beim */
{                 /* Vergleich als Listen ueber ganzen Zahlen mittels lexikographischer Erweiterung von >(Z) fest. */
  do {
    SymbolT Symbol1 = TO_TopSymbol(Term1),
            Symbol2 = TO_TopSymbol(Term2);
    if (Symbol1>Symbol2)
      return Groesser;
    else if (Symbol1<Symbol2)
      return Kleiner;
    Term1 = TO_Schwanz(Term1);
    Term2 = TO_Schwanz(Term2);
  } while (Term1);
  return Gleich;
}


BOOLEAN TO_TestSymbLex(TermT Term1, TermT Term2)                                     /* Test, ob Term1 >(Z)* Term2 */
{
  do {
    SymbolT Symbol1 = TO_TopSymbol(Term1),
            Symbol2 = TO_TopSymbol(Term2);
    if (Symbol1>Symbol2)
      return TRUE;
    else if (Symbol1<Symbol2)
      return FALSE;
    Term1 = TO_Schwanz(Term1);
    Term2 = TO_Schwanz(Term2);
  } while (Term1);
  return FALSE;
}


VergleichsT TO_TPEntscheidungSymbLex(TermT TP1Links, TermT TP1Rechts, TermT TP2Links, TermT TP2Rechts)
{                 /* Vergleich als Listen ueber ganzen Zahlen mittels lexikographischer Erweiterung von >(Z) fest. */
  do {
    SymbolT Symbol1 = TO_TopSymbol(TP1Links),
            Symbol2 = TO_TopSymbol(TP2Links);
    if (Symbol1>Symbol2)
      return Groesser;
    else if (Symbol1<Symbol2)
      return Kleiner;
    TP1Links = TO_Schwanz(TP1Links);
    TP2Links = TO_Schwanz(TP2Links);
  } while (TP1Links);
  do {
    SymbolT Symbol1 = TO_TopSymbol(TP1Rechts),
            Symbol2 = TO_TopSymbol(TP2Rechts);
    if (Symbol1>Symbol2)
      return Groesser;
    else if (Symbol1<Symbol2)
      return Kleiner;
    TP1Rechts = TO_Schwanz(TP1Rechts);
    TP2Rechts = TO_Schwanz(TP2Rechts);
  } while (TP2Links);
  return Gleich;
}


BOOLEAN TO_TPTestSymbLex(TermT TP1Links, TermT TP1Rechts, TermT TP2Links, TermT TP2Rechts)               /* analog */
{
  do {
    SymbolT Symbol1 = TO_TopSymbol(TP1Links),
            Symbol2 = TO_TopSymbol(TP2Links);
    if (Symbol1>Symbol2)
      return TRUE;
    else if (Symbol1<Symbol2)
      return FALSE;
    TP1Links = TO_Schwanz(TP1Links);
    TP2Links = TO_Schwanz(TP2Links);
  } while (TP1Links);
  do {
    SymbolT Symbol1 = TO_TopSymbol(TP1Rechts),
            Symbol2 = TO_TopSymbol(TP2Rechts);
    if (Symbol1>Symbol2)
      return TRUE;
    else if (Symbol1<Symbol2)
      return FALSE;
    TP1Rechts = TO_Schwanz(TP1Rechts);
    TP2Rechts = TO_Schwanz(TP2Rechts);
  } while (TP1Rechts);
  return FALSE;
}


static BOOLEAN TermEntspricht(TermT Term, unsigned int AnzSymbole, ...)
{                                                                  /* Die Folge muss einen korrekten Term ergeben. */
  va_list Argumente;
  unsigned int i;
  SymbolT SymbolLinks, SymbolRechts;
  va_start(Argumente,AnzSymbole);
  for(i=1; i<=AnzSymbole; i++) {
            SymbolLinks = TO_TopSymbol(Term);
            SymbolRechts = (SymbolT)va_arg(Argumente,int);
    if (SO_SymbUngleich(SymbolLinks,SymbolRechts)) {
      va_end(Argumente);
      return FALSE;
    }
    Term = TO_Schwanz(Term);
  }
  va_end(Argumente);
  return TRUE;
}


BOOLEAN TO_IstAssoziativitaet(TermT links, TermT rechts)
/* Sei links=rechts variablennormiert von links nach rechts mit Status l.
   sei f=TO_TopSymbol(links); f aus F U V. Test, ob links=rechts aus
   { A1: f(f(x1,x2),x3)=f(x1,f(x2,x3));  A2: f(x1,f(x2,x3))=f(f(x1,x2),x3) } */
{
  SymbolT f = TO_TopSymbol(links);
  return 
    SO_Stelligkeit(f)==2 && (
      (TermEntspricht(links,5,f,f,-1,-2,-3) && TermEntspricht(rechts,5,f,-1,f,-2,-3)) ||
      (TermEntspricht(links,5,f,-1,f,-2,-3) && TermEntspricht(rechts,5,f,f,-1,-2,-3))
    );
}


BOOLEAN TO_IstKommutativitaet(TermT links, TermT rechts)
/* Sei links=rechts variablennormiert von links nach rechts mit Status l.
   sei f=TO_TopSymbol(links); f aus F U V. Test, ob links=rechts aus
   { C: f(x1,x2)=f(x2,x1) } */
{
  SymbolT f = TO_TopSymbol(links);
  return 
    SO_Stelligkeit(f)==2 &&
    TermEntspricht(links,3,f,-1,-2) && 
    TermEntspricht(rechts,3,f,-2,-1);
}

BOOLEAN TO_IstDistributivitaet(TermT links, TermT rechts, SymbolT *was, SymbolT *worueber, BOOLEAN *linke)
{
  SymbolT f = TO_TopSymbol(links), g = TO_TopSymbol(rechts);
  BOOLEAN resultat=(SO_Stelligkeit(f)==2)&&(SO_Stelligkeit(g) ==2);
  if (!resultat) return FALSE;

  if (TermEntspricht(links, 5, f, -1, g, -2, -3) 
      && TermEntspricht(rechts, 7, g, f, -1, -2, f, -1, -3)){
    *linke = TRUE;
    *was = f;
    *worueber= g;
    return TRUE;
  }
  else if (TermEntspricht(links, 7, f, g, -1, -2, g, -1, -3) 
           && TermEntspricht(rechts, 5, g, -1, f, -2, -3)){
    *linke = TRUE;
    *was = g;
    *worueber= f;
    return TRUE;
  }
  else if (TermEntspricht(links, 5, f, g, -1, -2, -3) 
      && TermEntspricht(rechts, 7, g, f, -1, -3, f, -2, -3)){
    *linke = FALSE;
    *was = f;
    *worueber= g;
    return TRUE;
  }
  else if (TermEntspricht(links, 7, f, g, -1, -2, g, -3, -2) 
           && TermEntspricht(rechts, 5, g, f, -1, -3, -2)){
    *linke = FALSE;
    *was = g;
    *worueber= f;
    return TRUE;
  }
  return FALSE;
}

BOOLEAN TO_IstEinheit(TermT links, TermT rechts, SymbolT *was, SymbolT *wozu, BOOLEAN *linke)
/* Sei links=rechts variablennormiert von links nach rechts mit Status l.
   sei f=TO_TopSymbol(links); f aus F U V. Test, ob links=rechts aus
   { E1: f(x1,const)=x1, E2: x1=f(x1,const)} fuer links=TRUE
   { E1: f(const,x1)=x1, E2: x1=f(const,x1)} fuer links=FALSE
   Dann ist was=const, wozu=f.
   */
{
  SymbolT f = TO_TopSymbol(links), g = TO_TopSymbol(rechts);
  TermT gross, klein;

  if (SO_SymbIstVar(f) && !SO_SymbIstVar(g))      {klein = links; gross = rechts;}
  else if (!SO_SymbIstVar(f) && SO_SymbIstVar(g)) {klein = rechts; gross = links;}
  else return FALSE;
  
  if  (SO_Stelligkeit(TO_TopSymbol(gross)) != 2) return FALSE;
  
  if (TO_TermIstNichtVar(TO_ErsterTeilterm(gross)) && 
      (SO_Stelligkeit(TO_TopSymbol(TO_NaechsterTeilterm(TO_ErsterTeilterm(gross))))==0)&&
      (TO_TermIstNichtVar(TO_NaechsterTeilterm(TO_ErsterTeilterm(gross))))){
    *linke = TRUE;
    *was   = TO_TopSymbol(TO_NaechsterTeilterm(TO_ErsterTeilterm(gross)));
    *wozu  = TO_TopSymbol(gross);
    return TRUE;
  }
  else if (TO_TermIstVar(TO_NaechsterTeilterm(TO_ErsterTeilterm(gross))) &&
           (SO_Stelligkeit(TO_TopSymbol(TO_ErsterTeilterm(gross)))==0)&&
           (TO_TermIstNichtVar(TO_ErsterTeilterm(gross)))){
    *linke = FALSE;
    *was   = TO_TopSymbol(TO_ErsterTeilterm(gross));
    *wozu  = TO_TopSymbol(gross);
    return TRUE;
  }
  return FALSE;
}

BOOLEAN TO_IstInverse(TermT links, TermT rechts, SymbolT *was, SymbolT *wozu, SymbolT *ergibt, BOOLEAN *linke)
/* Sei links=rechts variablennormiert von links nach rechts mit Status l.
   sei f=TO_TopSymbol(links); f aus F U V. Test, ob links=rechts aus
   { E1: f(g(x1),x1)=const, E2: const=f(g(x1),x1)} fuer linksseitig=TRUE
   { E1: f(x1,g(x1))=const, E2: const=f(x1,g(x1))} fuer linksseitig=FALSE
   Dann ist was=g, wozu=f, ergibt=const
   */
{
  TermT gross, klein;
  SymbolT f,g,x;
  
  if (TO_Nullstellig(links)&&TO_TermIstNichtVar(links)&&SO_Stelligkeit(TO_TopSymbol(rechts))==2){
    gross = rechts;klein=links;
  }
  else if (TO_Nullstellig(rechts)&&TO_TermIstNichtVar(rechts)&&SO_Stelligkeit(TO_TopSymbol(links))==2){
    gross = links; klein=rechts;
  }
  else return FALSE;
  
  f = TO_TopSymbol(gross);
  
  if (TO_TermIstVar(TO_ErsterTeilterm(gross))){
    g = TO_TopSymbol(TO_NaechsterTeilterm(TO_ErsterTeilterm(gross)));
    x = TO_TopSymbol(TO_ErsterTeilterm(gross));
    if ((SO_Stelligkeit(g)==1) && (x==TO_TopSymbol(TO_ErsterTeilterm(TO_NaechsterTeilterm(TO_ErsterTeilterm(gross)))))){
      /* gross = f(x,g(x)), klein = const */
      *ergibt = TO_TopSymbol(klein);
      *was = g;
      *wozu = f;
      *linke=TRUE;
      return TRUE;
    }
  }
  else if (TO_TermIstVar(TO_NaechsterTeilterm(TO_ErsterTeilterm(gross)))){
    g = TO_TopSymbol(TO_ErsterTeilterm(gross));
    x = TO_TopSymbol(TO_NaechsterTeilterm(TO_ErsterTeilterm(gross)));
    if ((SO_Stelligkeit(g)==1) && (x==TO_TopSymbol(TO_ErsterTeilterm(TO_ErsterTeilterm(gross))))){
      /* gross = f(g(x),x), klein = const */
      *ergibt = TO_TopSymbol(klein);
      *was = g;
      *wozu = f;
      *linke=FALSE;
      return TRUE;
    }      
  }
  return FALSE;
}


BOOLEAN TO_IstIdempotenz(TermT links, TermT rechts, SymbolT *was)
/* Sei links=rechts variablennormiert von links nach rechts mit Status l.
   sei f=TO_TopSymbol(links/rechts); f aus F U V. Test, ob links=rechts aus
   { E1: f(x1,x1)=x1, E2: x1=f(x1,x1)}
   */
{
  if (TO_TermIstVar(links)){
    *was=TO_TopSymbol(rechts);
    if ((SO_Stelligkeit(*was)==2) && TermEntspricht(rechts, 3, *was, -1, -1))
      return TRUE;
  }
  else if (TO_TermIstVar(rechts)){
    *was=TO_TopSymbol(links);
    if ((SO_Stelligkeit(*was)==2) && TermEntspricht(links, 3, *was, -1, -1))
      return TRUE;
  }
  return FALSE;
}

static BOOLEAN ItesSymbolKstelligeFkt(TermT t, unsigned int i, unsigned int k, SymbolT *f)
{
  SymbolT g;
  while (--i)
    t = TO_Schwanz(t);
  g = TO_TopSymbol(t);
  if (SO_SymbIstFkt(g) && SO_Stelligkeit(g)==k) {
    *f = g;
    return TRUE;
  }
  else
    return FALSE;
}

BOOLEAN TO_IstAssoziator(TermT links, TermT rechts, 
  SymbolT *ass, SymbolT *mal, SymbolT *plus, SymbolT *minus)
/* Sei links=rechts variablennormiert von links nach rechts mit Status l. Test, ob links=rechts aus
   { E1: ass(x1,x2,x3) = plus(mal(mal(x1,x2),x3),minus(mal(x1,mal(x2,x3)))),
     E2: plus(mal(mal(x1,x2),x3),minus(mal(x1,mal(x2,x3)))) = ass(x1,x2,x3) }
   ACHTUNG: Nach dem Aufruf ist links=rechts normiert.
 */
{
  if (TO_DefinierendeGleichung(links,rechts))
    return
      ItesSymbolKstelligeFkt(links, 1,3,ass) &&
      TO_Termlaenge(rechts)==12 &&
      ItesSymbolKstelligeFkt(rechts,1,2,plus) &&
      ItesSymbolKstelligeFkt(rechts,2,2,mal) &&
      ItesSymbolKstelligeFkt(rechts,7,1,minus) &&
      TermEntspricht(rechts, 12, *plus, *mal, *mal, -1, -2, -3, *minus, *mal, -1, *mal, -2, -3);
  else if (TO_DefinierendeGleichung(rechts,links)) {
    BOOLEAN Ergebnis = 
      ItesSymbolKstelligeFkt(rechts, 1,3,ass) &&
      TO_Termlaenge(links)==12 &&
      ItesSymbolKstelligeFkt(links,1,2,plus) &&
      ItesSymbolKstelligeFkt(links,2,2,mal) &&
      ItesSymbolKstelligeFkt(links,7,1,minus) &&
      TermEntspricht(links, 12, *plus, *mal, *mal, -1, -2, -3, *minus, *mal, -1, *mal, -2, -3);
    BO_TermpaarNormierenAlsKP(links,rechts);
    return Ergebnis;
  }
  else
    return FALSE;
}

BOOLEAN TO_IstKommutator(TermT links, TermT rechts, 
  SymbolT *comm, SymbolT *mal, SymbolT *plus, SymbolT *minus)
/* Sei links=rechts variablennormiert von links nach rechts mit Status l. Test, ob links=rechts aus
   { E1: comm(x1,x2) = plus(mal(x2,x1),minus(mal(x1,x2))),
     E2: plus(mal(x1,x2),minus(mal(x2,x1))) = comm(x2,x1) }
   Kreuzweise ergaenzen
 */
{
  if (TO_DefinierendeGleichung(links,rechts))
    return
      ItesSymbolKstelligeFkt(links, 1,2,comm) &&
      TO_Termlaenge(rechts)==8 &&
      ItesSymbolKstelligeFkt(rechts,1,2,plus) &&
      ItesSymbolKstelligeFkt(rechts,2,2,mal) &&
      ItesSymbolKstelligeFkt(rechts,5,1,minus) &&
      TermEntspricht(rechts, 8, *plus, *mal, -2, -1, *minus, *mal, -1, -2);
  else
    return
      TO_Termlaenge(rechts)==3 &&
      ItesSymbolKstelligeFkt(rechts, 1,2,comm) &&
      TermEntspricht(rechts, 3, *comm, -2, -1) &&
      TO_Termlaenge(links)==8 &&
      ItesSymbolKstelligeFkt(links,1,2,plus) &&
      ItesSymbolKstelligeFkt(links,2,2,mal) &&
      ItesSymbolKstelligeFkt(links,5,1,minus) &&
      TermEntspricht(links, 8, *plus, *mal, -1, -2, *minus, *mal, -2, -1);
}
