/*-------------------------------------------------------------*/
/*      const.h            constants used by TGTP
/*
/*      Copyright Newborn Software 1996
/*
/*---------------------------------------------------------------*/
 
/*----------------------------------------------*/
/* MAX_VARS (in theorem) must be a power of 2   */
/*----------------------------------------------*/
 
#define MAX_VARS           32L   /* max # distinct vars in theorem */
#define MAX_VARS_M1        31L   /* MAX_VARS_M1 = MAX_VARS - 1     */
#define MAX_VARS_X2        64L   /* MAX_VARS_X2 = MAX_VARS * 2     */

/*----------------------------------------------*/
/* A theorem can have at most 150 base clauses  */
/* and at most 20 literals in any base clause   */
/* or any inferred clause.                      */
/*----------------------------------------------*/
 
#define MAX_BASE_CLAUSES   500L           /* max # base clauses     */
#define MAX_LIT_IN_CLAUSE  20L            /* max # literals in cls  */

/*------------------------------------------------------------------*/
/* MAX_PF = max number of predicates,functions permitted in theorem */
/* Of course, each pred or fxn could appear many times.             */
/* PF_SPACE_SIZE = size of the array "pfspace".                     */
/*------------------------------------------------------------------*/

#define MAX_PF             400L /* MAX_PF=MAX_PRED_IDS+MAX_FXNS_IDS */ 
#define PF_SPACE_SIZE      6*MAX_PF /* space for 100 6 char pred/fxn*/ 
 
#define MAX_PRED           4000L  /* max # of predicate appearances */
#define MAX_CHAR           16000L /* MAX_CHAR = MAX_PRED * 4        */
#define PRED_PTR_SIZE      3*MAX_PRED /* size of pred_ptr_list      */
 
#define MAX_PRED_IDS       200L    /* max number of different preds */
#define MAX_FXNS_IDS       200L    /* max number of different fxns  */

#define MAX_ITERATIONS     12L

/*------------------------------------------------------------------*/
/* MAX_INPUT_THM_SIZE is the maximum number of characters that      */
/* can appear in the input clauses.                                 */
/* MAX_INPUT_THM_SIZE = size of array "kode".                       */
/*                                                                  */
/* MAX_COMPILED_THM_SIZE is the maximum number of long words that   */
/* that the compiled theorem can occupy.                            */
/* MAX_COMPILED_THM_SIZE = size of array "compiled_code_buffer.     */
/*------------------------------------------------------------------*/
 
#define MAX_INPUT_THM_SIZE     40000  
#define MAX_COMPILED_THM_SIZE  40000  
 
/*------------------------------------------------------------------*/
/* ELIM_ARRAY_SIZE is the maximum number of resolvents that can     */
/* be generated during the eliminate phase (Phase 0)                */
/* ELIM_ARRAY_SIZE = size of array "elim_inferences".               */
/*------------------------------------------------------------------*/
 
#define ELIM_ARRAY_SIZE        40000            
 
/*------------------------------------------------------------------*/
/* MAX_PROOF_LENGTH is maximum number of inferences in a proof.     */
/* PROOF_ARRAY_SIZE must be twice the value of MAX_PROOF_LENGTH.    */
/* DEPTH_PTR_SPACE = 4*MAX_PROOF_LENGTH = size of array "depth".    */
/* PROOF_ARRAY_SIZE = size of the array "cl" (used in pass2.c)      */
/*------------------------------------------------------------------*/
 
#define MAX_PROOF_LENGTH       400L  /* size of depth array */
#define PROOF_ARRAY_SIZE       800L     /* size of cl_array */
#define PROOF_ARRAY_SIZE_M1    799L     /* size of cl_array */

#define DEPTH_PTR_SPACE        4*MAX_PROOF_LENGTH
 
/*----------------------------------------------------------*/
/* CSPACE_CLAUSES is the maximum number of clauses that can */
/* be kept on "clist".  The size of cspace is CSPACE_SIZE   */
/*----------------------------------------------------------*/
 
#define CSPACE_CLAUSES     8000
#define CSPACE_SIZE      800000
 
/*----------------------------------------------------------*/
/* constants used by hash.c                                 */
/*                                                          */
/* HT_SIZE is the number of entries that can be placed      */
/* in the clause_hash_table.  It must be a power of 2!!!!   */
/*----------------------------------------------------------*/
 
#define HT_SIZE     1024L*256L
#define HT_MASK     HT_SIZE-1L
 
#define PLY_MASK    0x000000ffL /* mask off ply; used in hash.c */

#define RAND_SIZE   40000
 
/*--------------------------------------------------------------*/
/* constants used by pass2.c to decode infer's in array "cl".   */
/*--------------------------------------------------------------*/
 
#define PARENT_MASK  0x00000fffL              /*   parent mask  */
#define INDEX_MASK   0x0000000fL
#define CIDMASK      0x03ff0000L
#define CIDMASK2     0x000003ffL

/*--------------------------------------------------*/
/* message tags                                     */
/*--------------------------------------------------*/

#define CCS	3	/* message for "compiled code size */
#define CCB     4       /* message for compiled code buffr */
#define YN      5       /* message for yes/no to continue  */
#define YNR     6       /* message for yes/no reply        */
#define CCSR    7       /* message for reply on buf size   */
#define CNAME   8       /* message for name of computer    */
#define MPROOF  9       /* message master found proof      */
#define SPROOF  10      /* message slave found proof       */

