/*-------------------------------------------------------------*/
/*      elim.c
/*
/*      Copyright Newborn Software 1996
/*
/*      functions:
/*          bres_ply_1()            .................... #1
/*          res_ply_1()             .................... #2
/*          uncomple_lit()          .................... #3
/*          base_subsume()          .................... #4
/*          successor_subsume()     .................... #5
/*          single_resolvent()      .................... #6
/*          elim_link_lits()        .................... #7
/*          elim_long_cls()         .................... #8
/*          det_min_link_lit()      .................... #9
/*          add_link_resolvents()   .................... #10
/*          delete_link_clauses()   .................... #11
/*          simple_subsume()        .................... #12
/*          delete_clause()         .................... #13
/*          add_clause()            .................... #14
/*          bres_replace()          .................... #15
/*          bres_link()             .................... #16
/*          eliminate()              ....................#17
/*-------------------------------------------------------------
/*      function linkage
/*
/*      eliminate()
/*          res_ply_1()
/*              bres_ply_1()
/*          uncomple_lit()
/*              delete_clause()
/*          base_subsume()
/*              simple_subsume()
/*                  delete_clause()
/*          successor_subsume()
/*              add_clause()
/*                  bres_replace()
                delete_clause() 
/*          single_resolvent()
/*              add_clause()
/*                  bres_replace()
                delete_clause() 
/*          elim_link_lits()
/*              det_min_link_lit()
/*              add_link_resolvents()
/*                  bres_link()
/*              delete_link_clauses()
/*          elim_long_cls()
/*--------------------------------------------------------------*/
 
#include <stdio.h>
 
#include "clause.h"
#include "const.h"
 
/*------------------------------------------------------------*/
/* these masks are to allow a resolution to be compacted into */
/* one long word                                              */
/*------------------------------------------------------------*/
 
#define CLPARMASK       0xffc00000L /* mask for bits 23 - 32  */
#define CRPARMASK       0x003ff000L /* mask for bits 13 - 22  */
#define CNLITMASK       0x00000f00L /* mask for bits  9 - 12  */ 
#define CINDEX1         0x000000f0L /* mask field for intx1   */
#define CINDEX2         0x0000000fL /* mask field for intx2   */
#define HT_HIT_MASK     0xfffff0ffL /* mask field for c_nlit  */
#define P1023           0x000003ffL /* 1023 mask              */

/*------------------------------------------------------------*/
/* externally defined functions                               */
/*------------------------------------------------------------*/
	
extern print0();                                /* tdump.c    */
extern fatal();                                 /* tgtp.c     */
extern long time_now();                         /* tgtp.c     */
extern clear_var_substitution_tble();           /* infer.c    */
extern standardize();                           /* infer.c    */
extern unify();                                 /* unify.c    */
extern install_clause();                        /* search.c   */
extern clause *generate_clause();               /* gen.c      */
extern int test_equality();                     /* gen.c      */
extern is_clause_in_ht();                       /* hash.c     */
extern long add_clause_to_ht();                 /* hash.c     */

/*------------------------------------------------------------*/
/* externally defined constants                               */
/*------------------------------------------------------------*/
 
extern FILE *outfile;                           /* tgtp.c     */
 
extern long elim_inferences[];/* where cls_code-ed cl's are   */
extern clause *clist[];  /* search.c:  pointers to clauses    */
extern dnode depth[]; /* search.c:  pointers to levels of cls */
 
extern long base, org_base, nextc, ply, iter_depth, nc;
extern long org_base;     /* tgtp.c:  orig # of base clauses  */
extern long clauses_alive;/* count of the # of ACTIVE clauses */
extern long equal_pred;   /* = 1 if equal pred is in theorem  */
extern int o_elimprint;  /* tgtp.c: o_elimprint=1 => prt info */
extern int o_maxlit, o_maxvar, o_maxterm;
extern int o_merge, o_nc, o_hash, o_extend;
extern int o_statprint;
extern int o_noprint;
extern int number_preds;    /* tgtp.c:  # of preds in theorem */
extern int in_elim;
extern int simple_sol;                        /* tgtp.c       */
extern long eliminate_time;
extern long elim_starting_time;


/*------------------------------------------------------------*/
/* internally defined functions                               */
/*------------------------------------------------------------*/

void bres_ply_1();
long res_ply_1();
int uncomple_lit();
int base_subsume();
int successor_subsume();
int single_resolvent();
int unit_resolvent();
int elim_link_lits();
void elim_long_cls();
long det_min_link_lit();
long add_link_resolvents();
long delete_link_clauses();
int simple_subsume();
void delete_clause();
int add_clause();
int bres_replace();
long bres_link();
void eliminate();

/*------------------------------------------------------------*/
/* internally defined constants                               */
/*------------------------------------------------------------*/
 
long subsumes_parent[50];/* encoding of subsuming cls's info */
long subsumes_parent_id[50];
long sp;
long infoptr;
long units, nc_units;

long clause_code;

long reinstate[1000];
 
/*------------------------------------------------------------------*/ 
/*      bres_ply_1(c1,c2,intx1,intx2)         .................. #1
/*
/*      generate a single resolvent: on the intx1 lit of c1 and the
/*      intx2 lit of c2.  Extract information from the clause and
/*      store it in clause_code.  The format for clause_code is:
/*
/*      bits:  32 - 23    22 - 13   12 - 9    8 - 5     4 - 1
/*             l_par      r_par     c_nlit    intx1     intx2
/*
/*      called by:      res_ply_1()                        elim.c
/*
/*      calls:          standardize()                      infer.c
/*                      clear_var_substitution_tble()      infer.c
/*                      unify()                            unify.c
/*                      generate_clause()                  gen.c
/*                      is_clause_in_ht()                  hash.c
/*                      simple_subsume()                   elim.c
/*                      print_clause()                     tdump.c
/*
/*      returns:        nil
/*------------------------------------------------------------------*/ 
 
void bres_ply_1(c1,c2,int1x,int2x)
clause *c1, *c2;
int int1x, int2x;
{
register clause *c;
register literal *l1, *l2;
long id1, id2;
int ix1, ix2, j;
	
clause_code = 0L;
		
id1 = c1->c_bits; id2 = c2->c_bits;     
if(((id1 & id2) & ACTIVE) == 0L) {
   return;}
	
standardize(c1,c2);                /* standardize vars apart */
 
ix1 = int1x;
ix2 = int2x;
 
l1 = c1->c_lit;
for( j = 1 ; j < int1x; j++) {
   l1 = l1->l_next;}
l2 = c2->c_lit;
for( j = 1 ; j < int2x; j++) {
   l2 = l2->l_next;}
	
id1 = l1->l_id & IDMASK;               /* pred id's  */
id2 = l2->l_id & IDMASK;
	
/*----------------------------------------------------------*/
/* if IDs are the same then we are in a predicate group.    */
/* we attempt unification if literals are complementary.    */
/*----------------------------------------------------------*/
 
if(id1 == id2) {
   if((l1->l_id ^ l2->l_id) == NEGMASK) {
	
/*------------------------------------------------------*/
/* if id's are the same and complementary, try unify    */
/*------------------------------------------------------*/
 
      clear_var_substitution_tble(c1->c_bits,c1->c_nv);
      clear_var_substitution_tble(c2->c_bits,c2->c_nv);
      if(unify(&l1->l_id,&l2->l_id)) { 

	     if((c = generate_clause(c1,c2,l1,l2,ix1,ix2)) != 0L) {

/*----------------------------------------------------------------*/
/* pack into clause_code a description of the                     */
/* binary resolution as follows:                                  */
/*  bits 23-32     bits 13-22    bits 9-12   bits 5-8   bits 1-4  */
/*  c_lpar->c_id   c_rpar->c_id  c->c_nlit   ix2        ix1       */
/*----------------------------------------------------------------*/

	         clause_code  = c->c_lpar->c_id * 4194304L;
	         clause_code += c->c_rpar->c_id * 4096L;
	         clause_code += (long) c->c_nlit * 256L;
	         clause_code += (long) ix1 * 16L + (long) ix2;
 
           if(c->c_nlit == 1) {
              units++;
              if(c->c_lpar->c_id > org_base - nc + 1L) {
                 if(c->c_lpar->c_id <= org_base + 1L) {
                    nc_units++;}}}

/*--------------------------------------------------------------*/
/* if clause simple-subsumes either its left or right parent,   */
/* then set subsumes_parent to that encoded clause              */
/*--------------------------------------------------------------*/
 
	         if(simple_subsume(c,c1) == 2) {    /* record clause */
                     if(sp <49L) sp++;
	             subsumes_parent[sp] = clause_code;/*subs parent*/
	             subsumes_parent_id[sp] = c1->c_id;}
	         if(simple_subsume(c,c2) == 2) {
                     if(sp < 49L) sp++; 
	             subsumes_parent[sp] = clause_code;
	             subsumes_parent_id[sp] = c2->c_id;}}}}}      

if(o_elimprint > 2 && clause_code != 0L) {
   print0("\n");
   print_clause(c);
   if(o_elimprint > 3) {getchar();}}
return;
}
 
 
/*------------------------------------------------------------------*/
/*      res_ply_1( )                    .............. #2
/*
/*------------------------------------------------------------------*/
/*------------------------------Step 0------------------------------*/
/*------------------------------------------------------------------*/
/*
/*      Generate all first-level resolvents one at a time.  They are
/*      placed in "clause_code" and then added to elim_inferences 
/*      list.  At most ELIM_ARRAY_SIZE resolvents can be formed.
/*
/*      called by:          eliminate()           elim.c
/*
/*      calls:              bres_ply_1()          elim.c
/*
/*      returns:            the # of resolvents of the current
/*                          base clauses
/*------------------------------------------------------------------*/
 
long res_ply_1()
{
dnode *d;
register long i, j;
register int int1, int2;

depth[0L].d_tc = base;  /* pters to beginning, end of base clauses  */
depth[0L].d_bc = 0L;
 
d = &depth[1L];
d->d_tc = nextc;
d->d_bc = nextc;
sp = -1L;
 
units = nc_units = 0L;

infoptr = 0L;
ply = 1L;
 
if(o_elimprint == 2) {
   print0("\n");}
 
/*----------------------------------------------------------*/
/* Loop through base clauses generating resolvents          */
/*----------------------------------------------------------*/
 
for(j = base; j >= 0L; j--) {
   if(j > 0L) {
      for(i = j - 1L; i >= 0L; i--) {
	     for(int1 = 1 ; int1 <= clist[j]->c_nlit ; int1++) {
	        for(int2 = 1 ; int2 <= clist[i]->c_nlit ; int2++) {
	           if(infoptr == ELIM_ARRAY_SIZE) {
		          print0("\nToo many clauses. Skip Elim.\n");
		          return(-1L);}
	           bres_ply_1(clist[j],clist[i],int1,int2);
	           if((clause_code&CNLITMASK)==0L&&clause_code!= 0L) {
	              return(-1L);}     /* return on null clause */
	           if(clause_code != 0L) {
		          elim_inferences[infoptr++]=clause_code;}}}}}}

return(infoptr);                /*      return to eliminate      */
}
 
/*------------------------------------------------------------------*/
/*      uncomple_lit(numb_res)   ............................... #3 
/* 
/*------------------------------------------------------------------*/
/*------------------------------Step 1------------------------------*/
/*------------------------------------------------------------------*/
/* 
/*      Search for literal which does not appear complemented in 
/*      another clause.  Delete clause.
/* 
/*      called by:        eliminate()             elim.c 
/* 
/*      calls:            delete_clause()         elim.c
/*                        print_clause()          tdump.c
/* 
/*      returns:          1 if there is an uncomplemented lit in
/*                        some clause 0 otherwise 
/*------------------------------------------------------------------*/
 
int uncomple_lit(numb_res)
long numb_res;
{
register int l, lit_max, lit_count;
int deleted;
long i, j;
		
/*--------------------------------------------------------------*/
/* delete all clauses containing uncomplemented literals        */
/*--------------------------------------------------------------*/
			
deleted = 0;
for(i = base; i >= 0L; i--) {
   if((clist[i]->c_bits & ACTIVE) != 0L) {
      lit_max = clist[i]->c_nlit;       
      for(l = 1;l <= lit_max;l++) { /* search for lit "l" in cl "i" */
	     lit_count = 0;
	     for(j = 0L; j < numb_res ;  j++) {   
	        if(((elim_inferences[j]&CLPARMASK)>>22L)==clist[i]->c_id) { 
	           if(((elim_inferences[j] & CINDEX1) >> 4L)==(long) l) { 
		          lit_count++; break;}}                
																		
	        if(((elim_inferences[j] & CRPARMASK)>>12L)==clist[i]->c_id){ 
	           if((elim_inferences[j] & CINDEX2)  == (long) l) {       
		          lit_count++; break;}}}                        
																				
	     if(lit_count == 0)  {   /* lit "l" in clause "i" not found */  
	        if(o_elimprint != 0) {      
	           printf("\n\nUncomp_litrl(%d):  delete ",l);        
	           fprintf(outfile,"\n\nUncomp_litrl(%d):  delete ",l);
	           print_clause(clist[i]);}
	        delete_clause(i); 
          deleted = 1;
          break;}}}}
			
return(deleted);
}
 
/*------------------------------------------------------------------*/
/*      base_subsume()                  ..................... #4
/*
/*      if base clause j is subsumed by some other base clause,
/*      eliminate cls j
/*
/*------------------------------------------------------------------*/
/*------------------------------Step 2------------------------------*/
/*------------------------------------------------------------------*/
/*      called by:      eliminate()         elim.c
/*
/*      calls:          simple_subsume()    elim.c
/*                      delete_clause()     elim.c
/*                      print_clause()      tdump.c
/*
/*      returns:        1 if some clause is subsumed
/*                      0 otherwise
/*------------------------------------------------------------------*/
 
int base_subsume()
{
register long i, j;
int deleted;

deleted = 0;	
for(i = 0L; i <= base ; i++) {
   if((clist[i]->c_bits & ACTIVE) == 0L) {
      continue;}
   for(j = 0L ; j <= base ; j++) {
      if((clist[j]->c_bits & ACTIVE) == 0L) {
         continue;}
      if(i != j) {
	     if(simple_subsume(clist[i],clist[j]) != 0) {
	        if(o_elimprint != 0) {
	           print0("\n\nBase_subsume: delete ");
	           print_clause(clist[j]);
	           print0("\n        Subsumed by: ");
	           print_clause(clist[i]);}
	        delete_clause(j);
	        deleted = 1;}}}}
	
return(deleted);
}
 
/*------------------------------------------------------------------*/
/*      successor_subsume(c)       ........................ #5
/*
/*------------------------------------------------------------------*/
/*------------------------------Step 3------------------------------*/
/*------------------------------------------------------------------*/
/*
/*      if some clause with id sub_clause_id is subsumed by its child
/*      subsumes_parent, replace clause.
/*
/*      called by:          eliminate()        elim.c
/*
/*      calls:              add_clause()       elim.c
/*                          print_clause()     tdump.c
/*
/*      returns:            1 if a successor is subsumed; 0 otherwise
/*------------------------------------------------------------------*/
 
int successor_subsume()
{
long i;
int ssp;
clause *c;
int added;

ssp = 0L;

if(sp == -1L) {
   return(0);}
for(i = 0L ; i <= sp ; i++) { 
   c = clist[subsumes_parent_id[i] - 1L];
   if((c->c_bits & ACTIVE) == 0L) {continue;}
   ssp = 1;
   if(o_elimprint != 0) {
      print0("\n\nSuccessr_Sub: replace ");
      print_clause(clist[subsumes_parent_id[i] - 1L]);}
   added = add_clause(subsumes_parent_id[i] - 1L,subsumes_parent[i]);
   delete_clause(subsumes_parent_id[i] - 1L);}
return(ssp);
}
 
/*------------------------------------------------------------------*/
/*      single_resolvent(numb_res)    ............................ #6
/*
/*------------------------------------------------------------------*/
/*------------------------------Step 4------------------------------*/
/*------------------------------------------------------------------*/
/*
/*      If literal in clause has only one resolvent, replace
/*
/*      called by:  eliminate()                 elim.c
/*
/*      calls:      add_clause()                elim.c
/*                  delete_clause()             elim.c
/*                  print_clause()              tdump.c
/*
/*      returns:    1 if clauses replaced ; 0 otherwise
/*----------------------------------------------------------------*/
 
int single_resolvent(numb_res)
long numb_res;
{
literal *lit;
clause *c;
register long i, j, k, l, num, lit_max, min_lit;
int i1, repl_base[MAX_BASE_CLAUSES], repl_resol[MAX_BASE_CLAUSES];
int litl_resol[MAX_BASE_CLAUSES], done, lit_count;
int added;
		
min_lit = 1000L;
num = 0L;
added = 0;
	
for(i = 0L ; i < MAX_BASE_CLAUSES ; i++) {
   repl_resol[i] = 0;}
	
/*----------------------------------------------------------*/
/* replace all single resolvent clauses with its resolvent  */
/*----------------------------------------------------------*/
	
for(i = base ; i >= 0L && min_lit != 1L; i--) {         
   if((clist[i]->c_bits & ACTIVE) == 0L) {
      continue;}
   lit_max = clist[i]->c_nlit;
   for(l = 1L ; l <= lit_max ; l++) {/*for each ACTIVE base clause */
      lit_count = 0;
      for(j = 0L ; j < numb_res ; j++ ) {
	     if(((elim_inferences[j] & CLPARMASK) >> 22L) == 
                                         clist[i]->c_id) {
	        if(((elim_inferences[j] & CINDEX1) >> 4L) == l) {
	           lit_count++;     
	           k = j;   
	           if(lit_count > 1) break;
	           lit = clist[i]->c_lit;   
	           for(i1 = 1; i1 < (int) l ; i1++) {
		          lit = lit->l_next;}}}
	     if(((elim_inferences[j]&CRPARMASK) >> 12L) == clist[i]->c_id) {
	        if((elim_inferences[j] & CINDEX2) == l) {
	           lit_count++;
	           k = j;
	           if(lit_count > 1) break;
	           lit = clist[i]->c_lit;
	           for(i1 = 1; i1 < (int) l ; i1++) {
		          lit = lit->l_next;}}}}
			
/*----------------------------------------------------------*/
/* cannout use NULL clause to replace another clause        */
/* only can replace a longer clause                         */
/*----------------------------------------------------------*/
		
      if(lit_count == 1 && 
          (((elim_inferences[k] & CNLITMASK) >> 8L) != 0L)) {
	       done = 1; 
			
/*----------------------------------------------------------*/
/* DON'T LOOP (See s18.dbabhp.thm                           */
/* The next several lines were modified Dec 13, 1994        */
/*----------------------------------------------------------*/
 
         if(nextc - 1L < org_base) {
            done = 0;}

         if((((elim_inferences[k] & CLPARMASK) >> 22L) == nextc) ||
             (((elim_inferences[k] & CRPARMASK) >> 12L) == nextc)) {
             if(clist[nextc]->c_rpar->c_id == nextc ||
                       clist[nextc]->c_lpar->c_id == nextc) {
              if((clist[(elim_inferences[k] & CLPARMASK) >> 22L]->c_id
           +clist[(elim_inferences[k] & CRPARMASK) >> 12L]->c_id - 1L)
                        == (clist[nextc - 1]->c_lpar->c_id +
                        clist[nextc - 1]->c_rpar->c_id)) {
                    done = 0;}}}
 	     if(done == 1) { 
				
	         if((elim_inferences[k]&CNLITMASK) != 0L) {/*if not in HT:*/
	             repl_base[num] = (int) i;
	             repl_resol[num] = (int) k;
	             litl_resol[num++] = 
                       (int)((elim_inferences[k] & CNLITMASK) >> 8L);
	             if(((elim_inferences[k] & CNLITMASK) >> 8L) < min_lit) {
		             min_lit = 
                       ((elim_inferences[k] & CNLITMASK) >> 8L);}}}}}} 

if(done == 0) return(0);

for(j = 0L ; j < num; j++) {                            
   i = (long) repl_base[j];     
   k = (long) repl_resol[j];    
   if(reinstate[i] == 0L) {
      if(o_elimprint != 0) {
	        print0("\n\nSingle_resol: replace ");
	        print_clause(clist[i]);
	        print0("\n                 with ");}
	
/*-------------------------------------------*/
/* add clause with condensed info in         */
/* elim_inferences[k].                       */
/*-------------------------------------------*/
	
      added = added + add_clause(i,elim_inferences[k]);
      reinstate[i] = 1L;}} 

if(added != 0) return(1);
return(0);
}
 
/*------------------------------------------------------------------*/
/*------------------------------------------------------------------*/
/*      unit_resolvent()    ........................ #6a
/*
/*------------------------------------------------------------------*/
/*------------------------------Step x -----------------------------*/
/*------------------------------------------------------------------*/
/*
/*      add unit clauses
/*
/*      called by:  eliminate()                 elim.c
/*
/*      calls:      add_clause()                elim.c
/*                  print_clause()              tdump.c
/*
/*      returns:    1 if clauses added ; 0 otherwise
/*----------------------------------------------------------------*/
 
int unit_resolvent()
{
literal *lit;
clause *c, *rp;
register long i, j, k, num;
int repl_base[MAX_BASE_CLAUSES], repl_resol[MAX_BASE_CLAUSES];
int added;
		
num = 0L;
added = 0;
	
for(i = 0L ; i < MAX_BASE_CLAUSES ; i++) {
   repl_resol[i] = 0;}
	
/*----------------------------------------------------------*/
/* replace all single resolvent clauses with its resolvent  */
/*----------------------------------------------------------*/
	
for(j = 0L ; j < infoptr ; j++ ) {
   if(num >= MAX_BASE_CLAUSES) {break;}
   if(((elim_inferences[j] & CNLITMASK) >> 8L) == 1L ) {
	     i = (elim_inferences[j] & CLPARMASK) >> 22L; 
             repl_base[num] = (int) i;
             repl_resol[num++] = (int) j;}}


for(j = 0L ; j < num; j++) { 
   i = (long) repl_base[j];     
   k = (long) repl_resol[j];
   c = clist[i - 1L];
   if(reinstate[i] == 0L) { 
      if(o_elimprint > 0) {
	        print0("\n\nUnit_resol: add     ");
	        print_clause(clist[i - 1L]);
	        print0("\n");} 

/*-------------------------------------------*/
/* add clause with condensed info in         */
/* elim_inferences[k].                       */
/*-------------------------------------------*/ 

      added = added + add_clause(i,elim_inferences[k]);
      reinstate[i] = 1L;}}

if(added != 0) return(1);
return(0);
}
 
/*------------------------------------------------------------------*/
/*      elim_link_lits()              .................... #7
/*
/*------------------------------------------------------------------*/
/*------------------------------Step 5------------------------------*/
/*------------------------------------------------------------------*/
/*
/*      determine the simplest link literal and then add and
/*      delete clauses
/*
/*      called by:  eliminate()                elim.c
/*
/*      calls:      det_min_link_lit()         elim.c
/*                  add_link_resolvents()      elim.c
/*                  delete_link_clauses()      elim.c
/*
/*      returns:    1 if link lit is found and clauses changed
/*                  0 otherwise
/*------------------------------------------------------------------*/
 
int elim_link_lits()
{
long base0, min_lit, number_resolvents, deleted_link_cls;
	
if((min_lit = det_min_link_lit()) == -1L)  {
   return(0);}
base0 = base;
number_resolvents = add_link_resolvents(min_lit);
base = base + number_resolvents;
clauses_alive = clauses_alive + number_resolvents;
deleted_link_cls = delete_link_clauses(min_lit,base0); 
if(deleted_link_cls == 0L) {
   return(0);}	
return(1);
}
 
/*------------------------------------------------------------------*/
/*      elim_long_cls(maxlitrl)          .................... #8
/*
/*------------------------------------------------------------------*/
/*------------------------------Step 6------------------------------*/
/*------------------------------------------------------------------*/
/*
/*      sets the ACTIVE flag to "off" on all "long" clauses that have
/*      been added by the elim. procedure.  Clauses are considered
/*      to be long if they have more than maxlit + 1 literals.
/*
/*      called by:     eliminate()          elim.c
/*
/*      calls:          nobody
/*
/*      returns:        Nil
/*------------------------------------------------------------------*/
 
void elim_long_cls(maxlitrl)
int maxlitrl;
{
long i;
 
for(i = org_base + 1L ; i <= base ; i++) {
   if((clist[i]->c_bits & ACTIVE) != 0L) {
      if(clist[i]->c_nlit > maxlitrl + 1) {
	     clauses_alive--;
	     clist[i]->c_bits &= ~ACTIVE;}}}
return;
}
 
/*------------------------------------------------------------------*/
/*      det_min_link_lit()                 .................... #9
/*
/*      determine the literal that is to be called the minimum link
/*      literal
/*
/*      called by:      elim_link_lits()            elim.c
/*
/*      calls:          Nobody
/*
/*      returns:        the number of the min_link_literal
/*------------------------------------------------------------------*/
 
long det_min_link_lit()
{
register literal *litrl;
register long lit,j;
long min_link_lit;      /* returned by subroutine as pred to elim   */
long min_link_lit_cls;  /* number of cls in which min_link_lit is   */
long lit_total_count;   /* number of appearances of lit under study */
long lit_clause_count;  /* number of times lit appears in a clause  */
long dead_clause;
	
lit_clause_count = 0L;
min_link_lit = -1L;
min_link_lit_cls = 1000L;
	
for(lit = 1L ; lit <= (long) number_preds ; lit++) {
   dead_clause = 0L;
   lit_total_count = 0L;
   for(j = 0L ;j <= base ;j++) {
      lit_clause_count = 0L;
      if((clist[j]->c_bits & ACTIVE) != 0L) {
	       lit_clause_count = 0L;
	       for(litrl=clist[j]->c_lit;litrl!=0L;litrl=litrl->l_next) {
	          if(((long) ((litrl->l_id&IDMASK)>>ARGSHIFT))==lit) {
	             lit_clause_count++;
	             lit_total_count++;}}}
      if(lit_clause_count == 1L && reinstate[j] != 0L) {dead_clause++;}
      if(lit_clause_count > 1L) {
         break;}}
   if(lit_clause_count <= 1L && dead_clause == 0L) {
      if(lit_total_count < min_link_lit_cls && lit_total_count>0L) {
	        min_link_lit = lit;
	        min_link_lit_cls = lit_total_count;}}}

return(min_link_lit);
}
 
/*------------------------------------------------------------------*/
/*      add_link_resolvents(min_lit)      .................... #10
/*
/*      add clauses that are the resolvents of link literals
/*
/*      called by:      elim_link_lits()            elim.c
/*
/*      calls:          bres_link()                 elim.c
/*                      add_clause_to_ht()          hash.c
/*                      print_clause()              tdump.c
/*
/*      returns:        the number of new resolvents added
/*------------------------------------------------------------------*/
 
long add_link_resolvents(min_lit)
long min_lit;
{
dnode *d;
register long i, j, cr_bits, cl_bits;
long num_resol, resol;     /*   number of resolvents           */
clause *c;
	
depth[0L].d_tc = base;      /* initial generating clause       */
depth[0L].d_bc = 0L;
d = &depth[1L];
	
ply = 1L;
d->d_tc = nextc;
d->d_bc = nextc;
	
num_resol = 0L;
	
if(o_elimprint != 0) {
   print0(
	 "\n\nAdd clauses produced by eliminating link literal:\n");}
	
for(j = base; j >= 0L; j--) {
   if(reinstate[j] == 1L) {continue;}
   if(j > 0L) {
      for(i = j - 1L; i >= 0L; i--) {
       if(reinstate[i] == 1L) {continue;}
	     resol = bres_link(clist[i],clist[j],0L,min_lit);
 
	     num_resol = num_resol + resol;
	     if(resol != 0L) {
	        clist[nextc-1L]->c_id = nextc;
	        clist[nextc-1L]->c_bits |= MERGE | BASE;
	        cr_bits = clist[nextc - 1L]->c_rpar->c_bits;
	        cl_bits = clist[nextc - 1L]->c_lpar->c_bits;
	        if((cr_bits & NC) != 0L || (cl_bits & NC) != 0L) {
	           clist[nextc - 1L]->c_bits |= NC;} 
          c = clist[nextc - 1L];
	        add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);
	        if(o_elimprint != 0) {
	           print0("                      ");
	           print_clause(clist[nextc - 1L]);}}}}}
return(num_resol);           /*   return to elim_link_lits()  */
 
}
 
/*------------------------------------------------------------------*/
/*      delete_link_clauses(min_lit,base0)     ............... #11
/*
/*      delete clauses containing the link literal
/*
/*      called by:          elim_link_lits()        elim.c
/*
/*      calls:              print_clause()          tdump.c
/*
/*      returns:            nil
/*------------------------------------------------------------------*/
 
long delete_link_clauses(min_lit,base0)
long min_lit, base0;
{
literal *litrl;
long j, del_cls;
	
if(o_elimprint != 0) {
   print0("\nDelete clauses with link literal:\n");}
	
del_cls = 0L;
for(j = 0L ; j <= base0 ; j++) {
   if(((clist[j]->c_bits & ACTIVE)) != 0L ) {
      for(litrl=clist[j]->c_lit ;litrl!=0L ;litrl=litrl->l_next) {
	     if(((litrl->l_id & IDMASK) >> ARGSHIFT) == min_lit) {
          if(reinstate[j] == 0L) {
             del_cls = 1L;}
          reinstate[j] = 1L;
	        if(o_elimprint != 0) {
	           print0("                      ");
	           print_clause(clist[j]);}
	        break;}}}}
 
if(o_elimprint != 0) {
   print0("\n");}
return(del_cls);
}
 
/*------------------------------------------------------------------*/
/*      simple_subsume(c1,c2)           .................... #12
/*
/*      simple_subsume examines the hash codes of the literals.
/*      Returns "1" if c1 = c2.  Returns 2 if c1 s-subsumes c2.
/*      That is, if every literal in c1 appears in exactly the
/*      same way in c2.
/*
/*      called by:  bres_ply_1()                   elim.c
/*                  b_subsume()                    elim.c
/*
/*      calls:      Nobody
/*
/*      returns:    1 if c1 = c2, 2 if c1 s-subsumes c2; 0 otherwise
/*------------------------------------------------------------------*/
 
int simple_subsume(c1,c2)
clause *c1, *c2;
{
register literal *l1, *l2;
int found;
	
if(c1->c_nlit > c2->c_nlit) {
   return(0);}
		
for(l1 = c1->c_lit ; l1 != 0L ; l1 = l1->l_next) {
   found = 0;
   for( l2 = c2->c_lit ; l2 != 0L ; l2 = l2->l_next) {
      if(l1->l_X == l2->l_X && l1->l_Y == l2->l_Y) {
	     found = 1;
	     break;}}
   if(found == 0) {
      return(0);}}

if(c1->c_nlit == c2->c_nlit) {
   return(1);}
return(2);
}                       
	
/*------------------------------------------------------------------*/
/*      delete_clause(i)                 .................... #13
/*
/*      delete the clause on clist whose index is i
/*
/*      called by:              uncomple_lit()          elim.c
/*                              base_subsume()          elim.c
/*
/*      calls:                  nobody
/*
/*      returns:                Nil
/*------------------------------------------------------------------*/
 
void delete_clause(i)
long i;
{       
clist[i]->c_bits &= ~ACTIVE;    
clauses_alive--;
return;             /*  return to uncomple_lit    */
}
 
/*------------------------------------------------------------------*/
/*      add_clause(i,k)                 .................... #14
/*
/*      replace clause i with clause k
/*
/*      called by:          successor_subsume()         elim.c
/*                          single_resolvent()          elim.c
/*
/*      calls:              bres_replace()              elim.c
/*                          add_to_clause_ht()          hash.c
/*                          print_clause()              tdump.c
/*
/*      returns:            nil
/*------------------------------------------------------------------*/
 
int add_clause(i,k)
long i, k;
{
register clause *l_p, *r_p, *c;
register int l_p_lit, r_p_lit;
long cl_bits, cr_bits;
int resolvent;

l_p      = clist[(((k & CLPARMASK) >> 22L) & P1023) - 1L]; 
l_p_lit  = (int) (k & CINDEX1)>>4;
r_p      = clist[((k & CRPARMASK) >> 12L) - 1L];
r_p_lit  = (int) (k & CINDEX2);
cr_bits  = clist[((k & CRPARMASK) >> 12L) - 1L]->c_bits;
cl_bits  = clist[(((k & CLPARMASK) >> 22L) & P1023) - 1L]->c_bits;

resolvent = bres_replace(l_p,r_p,l_p_lit,r_p_lit);

c = clist[nextc - 1L];
if(c->c_lit->l_id == equal_pred) {
   if(test_equality(c->c_lit) == 1) {
      resolvent = 0;}}

if(resolvent == 1) {
   base++;
   clauses_alive++;
   clist[nextc - 1L]->c_id        = nextc;
   clist[nextc - 1L]->c_lpar      = ( i != k ) ? l_p : r_p;
   clist[nextc - 1L]->c_rpar      = ( i != k ) ? r_p : l_p;
   clist[nextc - 1L]->c_indices   = ( i != k ) ? l_p_lit : r_p_lit;
   clist[nextc - 1L]->c_indices   = 
                  clist[nextc - 1L]->c_indices << LEFT_INDEX_SHIFT;
   clist[nextc - 1L]->c_indices  += ( i != k ) ? r_p_lit : l_p_lit;
   clist[nextc - 1L]->c_bits     |= MERGE | BASE; 

   if(((cr_bits & NC) != 0L) || ((cl_bits & NC) != 0L)) {
      clist[nextc - 1L]->c_bits |= NC;} 

   c = clist[nextc - 1L];
   add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);
   if(o_elimprint != 0) {
      print_clause(clist[nextc - 1L]);}}
if(resolvent == 0) {
   if(o_elimprint != 0) {
      print0("..NO! SIMPLY DELETE CLAUSE\n");}}
 
return(resolvent);            /*  return to single resolvent  */
}
 
/*------------------------------------------------------------------*/
/*      bres_replace(c1,c2,l1_int,l2_int)    ................. #15
/*
/*      resolve the l1_int lit of c1 with the l2_int literal of c2
/*
/*      called by:      add_clause()                       elim.c
/*
/*      calls:          standardize()                      infer.c
/*                      clear_var_substitution_tble()      unify.c
/*                      unify()                            unify.c
/*                      generate_clause()                  gen.c
/*                      install_clause()                   search.c
/*                      print_clause()                     tdump.c
/*
/*      returns:        1 if clause is installed; 0 otherwise
/*------------------------------------------------------------------*/
 
int bres_replace(c1,c2,l1_int,l2_int)
clause *c1, *c2;
int l1_int, l2_int;
{
clause *c;
register literal *l1, *l2;
register int ix1, ix2, i, j, new;
	
ix1 = l1_int;  ix2 = l2_int;
	
standardize(c1,c2);              /* standardize variables apart  */
	
l1 = c1->c_lit;   for ( j = 1 ; j < l1_int; j++) l1 = l1->l_next;
l2 = c2->c_lit;   for ( j = 1 ; j < l2_int; j++) l2 = l2->l_next;
		
/*------------------------------------------------------*/
/* attempt unification if literals are complementary    */
/*------------------------------------------------------*/
		
clear_var_substitution_tble(c1->c_bits,c1->c_nv);
clear_var_substitution_tble(c2->c_bits,c2->c_nv);
in_elim = 1;                   /* necessary to elim tautologys */
if(unify(&l1->l_id,&l2->l_id)) {
   if((c = generate_clause(c1,c2,l1,l2,ix1,ix2)) != 0L) {
      in_elim = 0;
 
/*------------------------------------------------------------------*/
/*      see if clause is already on clist by checking hashcodes
/*------------------------------------------------------------------*/
 
      new = 0;
      for(i = 0L; i <= base ; i++) {
         if(c->c_X == clist[i]->c_X && c->c_Y == clist[i]->c_Y) {
            new = 1;break;}}
 
      if(new == 1) {
	     if(o_elimprint != 0L) {
	        print_clause(c);
          print0("\n..... Wait!...\n ");
	        print0(
	            "This clause is already in the clause hash table.\n");}
	     return(-1);}
     if(install_clause(c)) {
	     if(c->c_nlit == 0) simple_sol = 1; 
	     return(1);} 
      return(0);}}
in_elim = 0;
return(0);
}
 
/*------------------------------------------------------------------*/ 
/*      bres_link(c1,c2,m,min_lit)           ................. #16
/*
/*      generate all resolvents of c1 and c2 and put them on clist
/*
/*      called by:      calc_resolvents()                  elim.c
/*
/*      calls:          standardize()                      bres.c
/*                      clear_var_substitution_tble()      infer.c
/*                      unify()                            unify.c
/*                      generate_clause()                  gen.c
/*                      install_clause()                   search.c
/*
/*      returns:        nr, the number of resolvents
/*------------------------------------------------------------------*/ 
 
long bres_link(c1,c2,m,min_lit)
clause *c1, *c2;
long min_lit;
{
clause *c;
register literal *l1, *l2;
register int ix1, ix2;  
long nr, id1, id2, cid1, cid2;
 
/*----------------------------------------------*/
/* check to see if both c1 and c2 are ACTIVE    */
/*----------------------------------------------*/
 
cid1 = c1->c_bits; 
cid2 = c2->c_bits;
if(((cid1 & cid2) & ACTIVE) == 0L) {
   return(0L);}
	
/*--------------------------*/
/* standardize vars apart   */
/*--------------------------*/
 
standardize(c1,c2);                  
	
/*------------------------------------------*/
/* set counter of number of resolvents to 0 */
/*------------------------------------------*/
 
nr = 0L;
					     
ix1 = 1;
 
/*----------------------*/
/* scan down literals   */
/*----------------------*/
 
for(l1 = c1->c_lit ; l1 != 0L ; l1 = l1->l_next) { 
   ix2 = 1;
   for(l2 = c2->c_lit ; l2 != 0L ; l2 = l2->l_next) {
      id1 = l1->l_id & IDMASK;             /* predicate IDs  */
      id2 = l2->l_id & IDMASK;
 
/*----------------------------------------------------------*/
/* if IDs are the same then we are in a predicate group.    */
/* we attempt unification if literals are complementary.    */
/*----------------------------------------------------------*/
 
      if(id1 == id2 && (id1 >> ARGSHIFT) == min_lit) {
	     if((l1->l_id ^ l2->l_id) == NEGMASK) {
	        clear_var_substitution_tble(c1->c_bits,c1->c_nv);
	        clear_var_substitution_tble(c2->c_bits,c2->c_nv);
	        in_elim = 1;
	        if(unify(&l1->l_id,&l2->l_id)) {
	           if((c=generate_clause(c1,c2,l1,l2,ix1,ix2))!=0L){
		          in_elim = 0;
		          if(install_clause(c)) {
		             c->c_bits |= BASE;   
		             if(c->c_nlit == 0) {
		                simple_sol = 1;}
		             nr++; }}}
	        in_elim = 0;}}
  
      ix2++;}
   ix1++;}      
return (nr);
}
 
/*------------------------------------------------------------------*/
/*      eliminate()                 ..........................#17
/*
/*      Jan 26, 1991 (added to mac version,mn)
/*
/*      called by:      main()                  tgtp.c
/*
/*      calls:          time_now()              tgtp.c
/*                      res_ply_1()             elim.c
/*                      print_clause()          tdump.c
/*                      uncomple_lit()          elim.c
/*                      base_subsume()          elim.c
/*                      successor_subsume()     elim.c
/*                      single_resolvent()      elim.c
/*                      elim_link_lits()        elim.c
/*                      elim_long_cls()         elim.c
/*                      fatal()                 tgtp.c
/*
/*      returns:        nil 
/*------------------------------------------------------------------*/
 
void eliminate()
{ 
int modify_clause;
int o_nc_temp,o_hash_temp,o_extend_temp;
int o_maxlit_temp,o_merge_temp;
long resol_count,cl_ptr,ii;
int input_char, input_char2;
clause *c;
	
/*----------------------------------------------------------*/
/* Allocate space for 6000 resolvents of the current base   */
/* clauses.  If more memory space is available to the user, */
/* the size of this array should be increased.              */
/* The array is not big enough to handle the "lions &       */
/* unicorn puzzle                                           */
/*----------------------------------------------------------*/
 
iter_depth = 1L;
 
elim_starting_time = time_now(); /* det starting time of elim proc */
	
o_nc_temp     = o_nc;        /* save search opts and restore     */
o_hash_temp   = o_hash;     /* when leaving subroutine          */
o_extend_temp = o_extend;
o_maxlit_temp = o_maxlit;
o_merge_temp  = o_merge;
	
o_nc = o_hash = o_extend = o_merge = 0;
o_maxlit = 1000;
 
clauses_alive = base + 1L;
modify_clause = 1;

for(ii = 0L; ii <= 1000L; ii++) {
   reinstate[ii] = 0L;}
	
/*----------------------------------------------------------*/
/* Top of main loop of Steps 0 - 5                          */
/* Loop back to this point until no clause is modified      */
/* or the number of clauses_alive is less than or equal     */
/* to one or a simple_sol has been found                    */
/*----------------------------------------------------------*/
 
while(modify_clause != 0 && clauses_alive > 1L && simple_sol == 0) {
   modify_clause = 0;
							
/*----------------------------------------------------------*/
/* Step 0                                                   */
/* Form a list of all binary resolvents of the base         */
/* clauses. The resolvents are placed on the list           */
/* elim_inferences.  Each resolvent takes one word.         */
/* the resolution is performed by bres_ply1                 */
/* on return, resol_count = the number of resolvents        */
/*----------------------------------------------------------*/
 
   resol_count = res_ply_1();            /* form all resol  */
   if(resol_count == -1L)  {
         goto exit_elim;}   /* too many clses or NULL found */
	
   if(o_elimprint != 0) {
      printf("\nThere are %ld base clauses;",clauses_alive);
      printf(" they have %ld resolvents.",resol_count);
      fprintf(outfile,"\nThere are %ld base clauses;",clauses_alive);
      fprintf(outfile," they have %ld resolvents.",resol_count);
      print0(
      "\n?(procede:'CR', list clauses:'c', list active clauses:'a')");
      if(o_elimprint == 2) {input_char = ' ';}
      else {input_char = getchar();}
      if(input_char == 'c' || input_char == 'a') {
	     for(ii = 0L ; ii <= base ; ii++) {
		    if(input_char == 'c')  {
		       print_clause(clist[ii]);}
		    else if((clist[ii]->c_bits & ACTIVE) == ACTIVE) {
		       print_clause(clist[ii]);}}}}
 
/*----------------------------------------------------------*/
/* Step 1                                                   */
/* See if there is a literal that is unresolable            */
/* in some clause.  If so, eliminate the clause             */
/*----------------------------------------------------------*/
	
   if(resol_count != -1L) {
      if(o_elimprint != 0) {
         print0("u");}
      modify_clause = uncomple_lit(resol_count);}
 
/*----------------------------------------------------------*/
/* Step 2                                                   */
/* See if there is a base clause that s-subsumes another    */
/* If so, eliminate base clause that is subsumed            */
/*----------------------------------------------------------*/
 
   if(/*modify_clause == 0 && */resol_count != -1L) {
      if(o_elimprint != 0) {
         fprintf(outfile,"b%ld ",base);
         printf("b%ld ",base);}
      modify_clause += base_subsume();}
		
/*----------------------------------------------------------*/
/* Step 3                                                   */
/* Successor subsumption:       if the successor of some    */
/* base clause s-subsumes a base clause, replace base       */
/* clause with successor                                    */
/*----------------------------------------------------------*/

      if(o_elimprint != 0) {
         fprintf(outfile,"s%ld ",base);
         printf("s%ld ",base);}
      modify_clause += successor_subsume(); /*}*/

/*----------------------------------------------------------*/
/* Step 4                                                   */
/* Single resolvent: if a literal in a clause resolves with */
/* only one other literal in one other clause, replace      */
/* the clause with its resolvent                            */
/*----------------------------------------------------------*/
 
   if(modify_clause == 0) {
      if(o_elimprint != 0) {
         fprintf(outfile,"r%ld ",base);
         printf("r%ld ",base);}
      modify_clause += single_resolvent(resol_count);} 

/*----------------------------------------------------------*/
/* Step 5                                                   */
/* determine link literal in fewest clauses                 */
/* then replace clauses having link literal with their      */
/* resolvents                                               */
/*----------------------------------------------------------*/

   if(modify_clause == 0) {
      if(o_elimprint != 0) {
         print0("l");}
      modify_clause += elim_link_lits();}}

/*----------------------------------------------------------*/
/* Step 6                                                   */
/* Eliminate long clauses (longer than o_maxlit_temp)       */
/*----------------------------------------------------------*/
 
if(o_elimprint != 0) {
   print0("x");}
elim_long_cls(o_maxlit_temp);
 
 
/*----------------------------------------------------------*/
/* Exit eliminate: return variables to original values      */
/* then free space taken by elim_inferences                 */
/*----------------------------------------------------------*/
 
exit_elim:              
unit_resolvent();


o_nc        = o_nc_temp;
o_hash      = o_hash_temp;
o_extend    = o_extend_temp;
o_maxlit    = o_maxlit_temp;
o_merge     = o_merge_temp;
ply         = 0L;
 
/*----------------------------------------------------------*/
/* User interface:                                          */
/* when e-option is on, user can now quit (q), print        */
/* the clauses (c), or print only the "ACTIVE" clauses, ie, */
/* the clauses that will be used in the search for a        */
/* contradiction, or proceed (p).                           */
/*----------------------------------------------------------*/
 
if(o_elimprint != 0) {
   print0("\n\nExit eliminate:\n");
decide:
   printf("proceed:'CR'; quit:'q'; list clauses: 'c';");
   printf(" list active clauses: 'a'\n?");
get_character:
   if(o_elimprint == 2) {input_char = '\n';}
   else {input_char = getchar();}
   if(input_char != '\n') {
      while((input_char2 = getchar()) != '\n') { };}
   if(input_char == 'c' || input_char == 'a') { /* print clauses   */
      print0("\n");
      for(cl_ptr = 0L ; cl_ptr <= base ; cl_ptr++ ) {
	       if(input_char=='a'&&(clist[cl_ptr]->c_bits & ACTIVE) == 0L) {
            continue;}
	       print_clause(clist[cl_ptr]);}
      print0("\n");
      goto decide;}
 
   else if(input_char == 'q') simple_sol = 1; /* quit; no search    */
   else if(input_char == '\n') {
      print0("\n");}                          /* proceed to search  */
   else {
      printf("\n? ");
      goto decide;}}                        /* re-enter a character */
 
/*-------------------------------------------------------*/
/* print time taken by eliminate(), followed by          */
/* the original number of base clauses and the revised   */
/* number after eliminate()                              */
/*-------------------------------------------------------*/
 
eliminate_time = time_now() - elim_starting_time;
printf("   %ld <BC: %ld NC: %ld AC: %ld U: %ld NCU: %ld",
   eliminate_time,org_base + 1L,nc,clauses_alive,units,nc_units);
if(o_noprint == 0) {
fprintf(outfile,
     "   %ld <BC: %ld NC: %ld AC: %ld U: %ld NCU: %ld",
   eliminate_time,org_base + 1L,nc,clauses_alive,units,nc_units);}
printf(" L%d V%d X%d Z%d>\n",o_maxlit,o_maxvar,o_maxterm,o_nc);
if(o_noprint == 0) {
fprintf(outfile,
     " L%d V%d X%d Z%d>\n",o_maxlit,o_maxvar,o_maxterm,o_nc);}
if(o_statprint == 1) {print0("\n");}
else print0(" : ");

return;
}
