/*------------------------------------------------------------------*/
/*      tdump.c
/*
/*      Copyright Newborn Software 1995
/*
/*      functions:
/*
/*      the following three functions actually print the clause
/*
/*      print_arguments()       .......................#1
/*      print_clause()          .......................#2
/*
/*      information on the options available are printed out by this
/*      routine
/*
/*      help_halted_search()       ....................#3
/*
/*      user can request information or change options or
/*      guide the search by choosing various options.
/*
/*      halted_search_options() .......................#4
/*      print_continuation()    .......................#5
/*      dump_calloc_info()      .......................#6
/*
/*      prints a string to the screen and to the disk
/*
/*      print0()                .......................#7
/*-----------------------------------------------------------------*/
 
#include "const.h"
#include "clause.h"
 
#include <stdio.h>
#include <termio.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>

/*------------------------------------------------------------------*/
/* externally defined functions                                     */
/*------------------------------------------------------------------*/
 
extern print_cl_array(), print_hash_code_list();  /* in pass2.c     */
extern long time_now();                           /* in tgtp.c      */
extern fatal();
extern int grab_number();

/*------------------------------------------------------------------*/
/* externally defined constants                                     */
/*------------------------------------------------------------------*/
 
extern char input_buffer[];
extern clause *clist[];              /* pointers to clauses         */
extern dnode depth[];            /* pointers to pointers of clauses */
 
extern int phase, big_bug;
extern int o_step, o_print, o_cr;
extern long base, ply, iter_depth, s_nod;
extern long s_ltentry;
extern int x_stop, o_debug, o_maxlit;
extern int o_maxterm;
extern char *pidtbl[], *fidtbl[];
extern long search_start_time;
extern long proof_length;
extern long tab_amount;      /* number of tabs before printing cls  */
 
extern FILE *outfile;


long *print_arguments();
void print_clause();
void help_halted_search();
int halted_search_options();
void print_continuation();
void dump_calloc_info();
void print0();

/*------------------------------------------------------------------*/
/*      print_arguments(p)          .......................#1
/*
/*      print the arguments of the predicate
/*
/*      called by:    print_clause()           tdump.c
/*
/*      calls:        nobody
/*
/*      returns:      Nil
/*------------------------------------------------------------------*/
 
long *print_arguments(p)
long *p;
{
long a, p_prime;
static char var_[] =
   "xyzuvwtsrqponmlkKLMNOPQRSTWVUZYXxyzuvwtsrqponmlkKLMNOPQRSTWVUZYX";
 
if((*p & FCNMASK) == 0L) {
	
/*------------------*/
/* print a variable */  
/*------------------*/
	
   p_prime = *p;        
   if(p_prime >= MAX_VARS) {
      p_prime = p_prime - MAX_VARS;}            /* standardize */
   printf("%c",var_[p_prime]); 
   fprintf(outfile,"%c",var_[p_prime]); 
   return(++p);}
    
/*------------------*/
/* print a function */  
/*------------------*/
 
printf("%s",fidtbl[(*p&IDMASK)>>ARGSHIFT]);
fprintf(outfile,"%s",fidtbl[(*p&IDMASK)>>ARGSHIFT]);
	
/*------------------------------------------------*/
/* recursively print the remaining arguments of p */
/*------------------------------------------------*/
 
for(a = *p++ & ARGMASK; a > 0L; a--)  {
   p = print_arguments(p);}
return(p);
}
 
/*-----------------------------------------------------------------*/
/*      print_clause(c)               .......................#2
/*
/*      called by:      write_out_results()       tgtp.c
/*                      search_tree()             search.c
/*                      advance_to_successor()    search.c
/*
/*      calls:          print_arguments()         tdump.c
/*
/*      returns:        nil
/*-----------------------------------------------------------------*/
 
void print_clause(c)
clause *c;
{
long i, j, *a;
literal *l, *litrl;
static char AB[] = ".abcdefghijklmnopqrstuvwxyz";
 
/*--------------------------------------------------------*/
/* decide how far to tab before beginning to print clause */
/*--------------------------------------------------------*/
 
if((c->c_bits & BASE) == 0L) {
   if(c->c_lpar->c_id != c->c_id - 1L && c->c_id > base + 1L) {
      tab_amount = c->c_id - base - 2L;}}
else {
   tab_amount = 0L;}
 
if((c->c_bits & BASE) == 0L) {
   for(i = c->c_id - base - 2L - tab_amount; i > 0L; i--) {
	    print0(" ");}}       
				
if((c->c_bits & ACTIVE) != 0L ) {
   if((c->c_bits & NC) != 0L) {
				
/*--------------------------------------*/
/* active clause in NC; denote by S     */
/*--------------------------------------*/
 
	     if((c->c_bits & BASE) == BASE) {
	        if((c->c_bits & LIST) == LIST) {
	           printf ("%2ldS>",c->c_id);
	           fprintf (outfile,"%2ldS>",c->c_id);}
	        else {
	           printf ("%2ldS ",c->c_id);
	           fprintf (outfile,"%2ldS ",c->c_id);}}
	     else {
	        printf ("%2ldS>",c->c_id);
	        fprintf (outfile,"%2ldS>",c->c_id);}}
   else {               
 
/*--------------------------------------*/
/* active clause not in NC; leave " "   */
/*--------------------------------------*/
 
	     if((c->c_bits & BASE) == BASE) {
	        if((c->c_bits & LIST) == LIST) {
	           printf ("%2ld >",c->c_id);
	           fprintf (outfile,"%2ld >",c->c_id);}
	        else {
	           printf("%2ld: ",c->c_id);
	           fprintf(outfile,"%2ld: ",c->c_id);}}
	     else {
	        printf("%2ld: ",c->c_id);
	        fprintf(outfile,"%2ld: ",c->c_id);}}}
 
else {
 
/*--------------------------------------*/
/* clause is not active; denote by #    */
/*--------------------------------------*/
 
   if((c->c_bits & LIST) == LIST) {
      printf ("%2ld#>",c->c_id);
      fprintf (outfile,"%2ld#>",c->c_id);}
   else {
      printf("%2ld# ",c->c_id);
      fprintf(outfile,"%2ld# ",c->c_id);}}
		
if(c->c_lpar != 0L) {
   if(c->c_rpar == 0L) {
 
	 /*--------------------------*/
	 /* clause is a factor       */
	 /*--------------------------*/
 
	     printf("(%ld%c%c)    ",c->c_lpar->c_id,
		     AB[c->c_indices >> LEFT_INDEX_SHIFT], 
		     AB[c->c_indices & RIGHT_INDEX_MASK]);
	     fprintf(outfile,"(%ld%c%c)    ",c->c_lpar->c_id,
		     AB[c->c_indices >> LEFT_INDEX_SHIFT], 
		     AB[c->c_indices & RIGHT_INDEX_MASK]);
	     if(c->c_lpar->c_id<10L) {
	        print0(" ");}}
   else {
 
	 /*----------------------------------*/
	 /* clause is a binary resolvent     */
	 /*----------------------------------*/ 
if(c->c_bits & MERGE) {	
	     printf("(%ld%c*%ld%c",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);       
	    fprintf(outfile,"(%ld%c*%ld%c",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}
      

else { 

	     printf("(%ld%c,%ld%c",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);       
	    fprintf(outfile,"(%ld%c,%ld%c",c->c_lpar->c_id, 
	       AB[c->c_indices >> LEFT_INDEX_SHIFT],c->c_rpar->c_id,
	       AB[c->c_indices & RIGHT_INDEX_MASK]);}

	    print0(") ");       
	 		
	    if(c->c_lpar->c_id < 10L) print0(" ");
	    if(c->c_rpar->c_id < 10L) print0(" ");}}
			 
/*----------------------------------------------*/
/* if null clause, print []                     */
/*----------------------------------------------*/
 
if(c->c_lit == 0L) {
   print0("[]\n");}
		 
/*----------------------------------------------*/
/* else print clause                            */
/*----------------------------------------------*/
 
else {
   for(l = c->c_lit; l != 0L; l = l->l_next) {
      if((l->l_id & HHITMASK) != 0L) {
	       print0("< ");}
      
      if((l->l_id & NEGMASK) != 0L)  {
	       print0("~");}     
				 
	    printf("%s",pidtbl[(l->l_id & IDMASK) >> ARGSHIFT]); 
	    fprintf(outfile,"%s",pidtbl[(l->l_id & IDMASK) >> ARGSHIFT]);
 
      a = &l->l_id;
      a++;
      for(i = l->l_id & ARGMASK; i > 0L; i--) {
	       a = print_arguments(a);}
 
/*----------------------------------------------*/
/* see if literal is represented by hash code   */
/*----------------------------------------------*/              
	
      if((l->l_id & HHITMASK) != 0L) {/*
	       printf(" %ld/%ld >",l->l_X,l->l_Y);     
	       fprintf(outfile," %ld/%ld >",l->l_X,l->l_Y);*/}    
 
      if((l->l_id & HHITMASK) != 0L) {
	       print0("...>");}    
		
/*----------------------------------------------*/
/* print a space after each literal             */
/*----------------------------------------------*/
 
      if(l->l_next != 0L) {
         print0(" ");}}}
 
if(o_debug > 0 || big_bug != 0) {
   printf("\nH:%ld %ld\n ",c->c_X,c->c_Y);
   fprintf(outfile,"\nH:%ld %ld\n ",c->c_X,c->c_Y);

   j = 0L;	
   for(litrl = c->c_lit; litrl != 0L ; litrl = litrl->l_next) {
      fprintf(outfile,"L%ld %ld ",litrl->l_X,litrl->l_Y);
      printf("L%ld %ld ",litrl->l_X,litrl->l_Y);
      if(++j % 2L == 0L) print0("\n");}               
 
   printf("<%ld> ",s_nod);
   fprintf(outfile,"<%ld> ",s_nod);
				
   if(ply <= (long) o_step) {
      print0("...");}}
		
if(o_step <= 0 || (o_step > 0 && ply < (long) o_step)) { 
   if(o_cr == 0) {
   print0("\n");}}
if(o_cr == 1 && ply > o_step) {
   print0("\n");}

return; 
}
 
 
/*---------------------------------------------------------------*/
/*      help_halted_search()           .....................#3
/*
/*      print information to help human during search
/*
/*      called by:        halted_search_options()     tdump.c
/*
/*      calls:            nobody
/*
/*      returns:          Nil
/*---------------------------------------------------------------*/
 
void help_halted_search()
{
printf("\n\n  - - - - - - - - - - - - - - - - - - - - - - - - - - -");
printf("\n |  The following keys control the search:");
printf("\n |");
printf("\n | 'q'  => quit prove, try another thoerem");
printf("\n | 'k'  => quit iteration, begin the next one");
printf("\n | 'r'  => quit iteration, repeat iteration");
printf("\n | 'x'  => terminate search on this branch");
printf("\n | 't'  => print total search time");
printf("\n | 'n'  => print total search nodes");
printf("\n | 'a'  => print partially formed proof in cl_array");
printf("\n | 'h'  => print hash_code_list");
printf("\n | 'c'  => print current continuation and stats");
printf("\n | 'y'  => print tree structure");
printf("\n | 'vx' => print clause x on clist");
printf("\n | 'V'  => stop printing clauses from clist");
printf("\n | '+'  => increment step and print depth");
printf("\n | '-'  => decrement step and print depth");
printf("\n | 'p'  => print step and print depth");
printf("\n | 'd'  => turn on debug option");
printf("\n | '?'  => print this list!!");
printf("\n | 'CR' => type a 'return' to continue with the search");
printf("\n |");
printf("\n  - - - - - - - - - - - - - - - - - - - - - - - - - - -\n");
}
 
/*---------------------------------------------------------------*/
/*      halted_search_options()         .....................#4
/*
/*      called by:    search_tree()              search.c
/*
/*      calls:        time_now()                 tgtp.c
/*                    print_continuation()       tdump.c
/*                    print_cl_array()           pass2.c
/*                    print_hash_code_list()     pass2.c
/*                    grab_number()              tgtp.c
/*                    print_clause()             tdump.c
/*                    help_interrupted_search()  tdump.c
/*
/*      returns:      Nil
/*---------------------------------------------------------------*/
 
int halted_search_options()
{
int zz, i, o_debugt;
long kk, clause_to_print, current_time;
char in_char;
dnode *dd;
clause *c;
 
print0(" ");
 
clause_to_print = -1L;
 
search_options:

for(i = 0; i < 80 ; i++) {
   input_buffer[i] = getchar();
   if(input_buffer[i] == '\n') {break;}}

zz = input_buffer[0];
fprintf(outfile,"%c",(char) zz); 
 
if(clause_to_print >= 0L && zz == '\n') {
   goto print_this_clause;}
   
clause_to_print = -1L;
 
/*-----------------------------------------*/
/* if 'q', quit search                     */
/*-----------------------------------------*/
 
if(zz == 'q') {
   print0("\n\nDo you wish to quit(Y or N)? ");
   zz = getchar();
   printf("%c ",zz); 
   fprintf(outfile,"%c ",zz);
   if(zz == 'y' || zz == 'Y') {
      return(1);}
   else goto search_options;}
 
/*-----------------------------------------*/
/* if 'k', kill iteration, start next      */
/*-----------------------------------------*/
 
if(zz == 'k')  {
   print0("\nQuitting this iteration, beginning the next.\n");
   x_stop = 2;
   return(0);}
 
/*-----------------------------------------*/
/* if 'r', repeat iteration                */
/*-----------------------------------------*/
 
if(zz == 'r')  {
   x_stop = 2; 
   iter_depth--;
   print0("\nQuitting this iteration, but repeating it.\n");
   return(0);}
 
/*-----------------------------------------*/
/* if 'x', terminate search at this node   */
/*-----------------------------------------*/
 
if(zz == 'x')  {
   x_stop = 1;
   return(0);}
 
/*-----------------------------------------*/
/* if 't', print search time in seconds    */
/* if 'n', print nodes searched            */
/*-----------------------------------------*/
 
if(zz == 't') {
   current_time = time_now() - search_start_time;
   printf("TIME: %ld sec.",current_time); 
   fprintf(outfile,"TIME: %ld sec.",current_time);
   print0("\n");
   goto search_options;}
 
if(zz == 'n') {
   printf("NODES: %ld",s_nod);
   fprintf(outfile,"NODES: %ld",s_nod);
   print0("\n");
   goto search_options;}
 
/*----------------------------------------*/
/* if 'a', print proof array              */
/*----------------------------------------*/
 
if(zz == 'a') {
   if(proof_length == 0L) {
      print0("NO PROOF FOUND YET!\n");
      goto search_options;}
   else {
      print0("PROOF AS FOUND THUS FAR:\n");
      printf("\nProof length = %ld ",proof_length);
      fprintf(outfile,"\nProof length = %ld ",proof_length);
      print_cl_array(PROOF_ARRAY_SIZE - proof_length);
      goto search_options;}}
 
/*----------------------------------------*/
/* if 'h', print hash_code_list           */
/*----------------------------------------*/
 
if(zz == 'h') {
   if(proof_length == 0L) {
      print0("NO PROOF FOUND YET AND THUS NO HASH CODES TO FIND\n");
      goto search_options;}
   o_debugt = o_debug;
   o_debug = 1;
   print_hash_code_list();
   print0("\n");
   o_debug = o_debugt;
   goto search_options;}
		
/*------------------------------------------*/
/* if 'c', print current continuation       */
/*------------------------------------------*/
 
if(zz == 'c') {
   print0("CURRENT CONTINUATION UNDER SEARCH:");
   printf("\n (NODES: %ld) <UNITS: %ld>",s_nod,s_ltentry);
   fprintf(outfile,"\n (NODES: %ld) <UNITS: %ld>",s_nod,s_ltentry);
   current_time = time_now() - search_start_time;
   printf(" (TIME: %ld) (ITER: %ld) ",current_time,iter_depth);
   fprintf(outfile," (TIME: %ld) (ITER: %ld) ",
                                      current_time,iter_depth);
   printf("(l%d) (x%d) ",o_maxlit,o_maxterm);
   fprintf(outfile,"(l%d) (x%d) ",o_maxlit,o_maxterm);
   printf("<LEN: %ld> ",proof_length);
   fprintf(outfile,"<LEN: %ld> ",proof_length);
   print_continuation();
   goto search_options;}
 
/*------------------------------------------*/
/* if 'y', print current tree info          */
/*------------------------------------------*/
 
if(zz == 'y') {
   print0("CURRENT SEARCH INFORMATION:\n");
   dd = &depth[0];
   printf("\nBase clauses:    %3ld %3ld %3ld %8ld",
		 dd->d_bc + 1L, dd->d_tc + 1L, dd->d_ec + 1L, dd->d_rc);
   printf("\n--------------------------------");
 
   fprintf(outfile,"\nBase clauses:    %3ld %3ld %3ld %8ld",
		 dd->d_bc + 1L, dd->d_tc + 1L, dd->d_ec + 1L, dd->d_rc);
   fprintf(outfile,"\n--------------------------------");
 
   for(kk = 1L; kk <= ply; kk++) {
      dd = &depth[kk];
      printf("\nPly  %2ld clauses: %3ld %3ld %3ld %8ld",
	   kk, dd->d_bc + 1L, dd->d_tc + 1L, dd->d_ec + 1L, dd->d_rc);}
   print0("\n");
   goto search_options;}
 
/*----------------------------------------*/
/* if 'vx', print clause x on clist       */
/*----------------------------------------*/
 
if(zz == 'v') {
   clause_to_print = (long) grab_number(1) - 1L;
 
print_this_clause:
   dd = &depth[ply];
   if(clause_to_print > dd->d_ec || clause_to_print < 0L) {
      print0("\nInvalid clause index.\n");
      clause_to_print = -1L;
      goto search_options;}
   c = clist[clause_to_print];
   print_clause(c);
   print0("\n");
   clause_to_print++;
   goto search_options;}

if(zz == 'V') {
   clause_to_print = -1L;
   goto search_options;}
		
/*------------------------------------------*/
/* if '+', increment step,print depth       */
/* if '-', decrement step,print depth       */
/* if 'p', print out step,print depth       */
/*------------------------------------------*/
 
if(zz == '+') {
   o_print++; 
   o_step++;
   printf(" P%d S%d\n",o_print,o_step);
   fprintf(outfile," P%d S%d\n",o_print,o_step);
   goto search_options;}
 
if(zz == '-' ) {
   if(o_step > 0) {
      o_print--;
      o_step--;         
      printf(" P%d S%d\n",o_print,o_step);
      fprintf(outfile," P%d S%d\n",o_print,o_step);
      goto search_options;}
   else {
      printf(" P%d S%d\n",o_print,o_step);
      fprintf(outfile," P%d S%d\n",o_print,o_step);
      goto search_options;}}
			
if(zz == 'p') {
   printf(" P%d S%d\n",o_print,o_step);
   fprintf(outfile," P%d S%d\n",o_print,o_step);
   goto search_options;}        
 
/*-------------------------------------------*/
/* if 'd', then if debug option is on, turn  */
/* it off, else if off, turn it on           */
/*-------------------------------------------*/
 
if(zz == 'd')  {
   if(o_debug == 0) {
      o_debug = 1;
      goto search_options;}
   else {
      o_debug = 0;      
      goto search_options;}}   
 
/*-------------------------------------------*/
/* if '?', print halted_search_options       */
/*-------------------------------------------*/
 
if(zz == '?' ) {
   help_halted_search();
   goto search_options;}
 
if(zz == 10) {      /* if zz is a carriage return, return */
   return(0);}
 
else {
   print0("\n?? ");
   goto search_options;}
 
}
 
 
/*---------------------------------------------------------------*/
/*      print_continuation()            .......................#5
/*
/*      prints the current continuation under search
/*
/*      called by:   halted_search_options()    tdump.c
/*                   interrupt_options()        tdump.c
/*                   install_clause()           search.c
/*
/*      calls:       print_clause(c)            tdump.c
/*
/*      returns:     nil
/*---------------------------------------------------------------*/
 
void print_continuation()
{
long i;
clause *c;
	
print0("\n\n");
for(i = 1L ; i <= ply ; i++) {
   c = clist[depth[i].d_tc];
   print_clause(c);
   print0("\n");}
return;
}
 
/*---------------------------------------------------------------*/
/*      dump_calloc_info()         ....................#6
/*
/*      called by:    main()                     tgtp.c
/*                    init_clause_hash_table()   hash.c
/*
/*      calls:        fatal()                    tgtp.c
/*
/*      returns:      Nil
/*---------------------------------------------------------------*/
 
void dump_calloc_info(name,nameptr,index,data_type,size)
char *name;
long nameptr;
long index;
char *data_type;
long size;
{
if(o_debug > 1) {
   if(index == -1L) {
      printf("\n%s \t@%ld  #%ld  %s",name,nameptr,size,data_type);
      fprintf(outfile,
	     "\n%s \t@%ld  #%ld  %s",name,nameptr,size,data_type);}
 
   else {
      if(index % 4L == 0L) {
	     printf("\n%s[%ld] \t@%ld  #%ld  %s",
			  name,index,nameptr,size,data_type);
	     fprintf(outfile,"\n%s[%ld] \t@%ld  #%ld  %s",
			 name,index,nameptr,size,data_type);}
      else {
	     printf("  @%ld",nameptr);
	     fprintf(outfile,"  @%ld",nameptr);}}
   getchar();}
 
if(nameptr == 0L) {
   if(index == -1) {
      printf("Insufficient memory (%ld) for %s",size,name);
      fprintf(outfile,"Insufficient memory (%ld) for %s",size,name);}
   else {
      printf("Insufficient memory for %s[%ld]",name,index);
      fprintf(outfile,"Insufficient memory for %s[%ld]",name,index);}
   fatal("\nQuitting PROVE.\n");}      
return;
}       
	
/*---------------------------------------------------------------*/
/*      print0()         .......................#7
/*
/*      prints a string to the screen and to the disk
/*
/*      returns:      Nil
/*---------------------------------------------------------------*/
	
void print0(msg)
char *msg;
{
printf("%s",msg);
fprintf(outfile,"%s",msg);	
}
