/*------------------------------------------------------------*/
/*      tgtp.c
/*
/*      Copyright Newborn Software 1995
/*
/*      functions:
/*      time_now()..................... #1
/*      init_stats()................... #2
/*      init_options()................. #3
/*      grab_number(s)................. #4
/*      fatal(msg)..................... #5
/*      enter_theorem_name()........... #6
/*      init_search_space()............ #7
/*      init_clause_hashtable()........ #8
/*      init_rand_hash_array()......... #9             
/*      free_space_close_files()....... #10
/*      read_buffer_word()............. #11
/*      interchange_fxn_const()........ #12
/*      is_fxn_a_constant()............ #13
/*      sort_constants_fxns().......... #14
/*      format_clause()................ #15
/*      count_lit_terms().............. #16
/*      calc_max_term_in_any_lit()..... #17
/*      get_pred_fxn_symbols()......... #18
/*      prepare_clauses().............. #19
/*      hash_all_base_clauses()........ #20
/*      prove_by_iterative_deepening(). #21
/*      get_clause_stats()............. #22
/*      write_out_results()............ #23
/*      main()........>>>>>>>>>>>>>>>>> #24
/*-------------------------------------------------------------*/
/*
/*      LINKAGE OF FUNCTIONS
/*
/*main()--------------<<<<<<<<<<<<<<<<<<<<<<<<<.........tgtp.c
/*
/* begin_proof:
/* 
/*       enter_theorem_name()...........................tgtp.c
/*       compile_clauses()..............................affc.c
/*       init_cl_array()................................pass2.c
/*       init_clause_hash_table().......................tgtp.c
/*       init_stats()...................................tgtp.c
/*       init_search_space()............................tgtp.c
/*       prepare_clauses()..............................tgtp.c
/*             get_pred_fxn_symbols()...................tgtp.c
/*             read_buffer_word().......................tgtp.c
/*             format_clause()..........................tgtp.c
/*                   read_buffer_word().................tgtp.c
/*             sort_constants_fxns()....................tgtp.c
/*                   is_fxn_a_constant()................tgtp.c
/*                   interchange_fxn_const()............tgtp.c
/*             calc_max_term_in_any_lit()...............tgtp.c
/*                   count_lit_terms()..................tgtp.c
/*       init_random_hash_array().......................tgtp.c
/*       hash_all_base_clauses()........................tgtp.c
/*             calc_lit_hc()............................hash.c
/*             calc_clause_hashcode()...................hash.c
/*             add_clause_to_ht().......................hash.c
/*             calc_alt_hashcodes().....................hash.c
/* 
/*----------------------------------------------------*/
/* PHASE 0                                            */
/* -------                                            */
/* simply input clauses if possible.                  */
/*----------------------------------------------------*/
/* 
/*        eliminate()...................................elim.c
/* 
/*----------------------------------------------------*/
/* PHASE 1                                            */
/* -------                                            */
/* begin searching for a proof.                       */
/*----------------------------------------------------*/
/* 
/*        prove_by_iterative_deepening()................tgtp.c
/*              init_clause_hash_table()................tgtp.c
/*              add_clause_to_ht()......................hash.c
/*              calc_alt_hashcodes()....................hash.c
/*              search_tree()...........................search.c
/*              save_line().............................pass2.c
/*              time_now()..............................tgtp.c
/* 
/*        stack_hash_codes()............................pass2.c
/*      
/*----------------------------------------------------*/
/* PHASE 2                                            */
/* -------                                            */
/* search for hash codes on hash_code_list.           */
/*----------------------------------------------------*/
/* 
/*       prove_by_iterative_deepening().................tgtp.c
/*       stack_hash_codes().............................pass2.c
/*      
/*----------------------------------------------------*/
/* PHASE 3                                            */
/* -------                                            */
/* set phase to 3 check proof, then print results     */
/*----------------------------------------------------*/
/* 
/*       rearrange_cl_array()...........................pass2.c
/*       search_proof_line()............................pass2.c
/*       write_out_results()............................tgtp.c
/*             print_clause()...........................tdump.c
/*             get_clause_stats().......................tgtp.c
/*
/*----------------------------------------------*/
/* free allocated space and close output file   */
/*----------------------------------------------*/
/* 
/*       free_space_close_files().......................tgtp.c
/*
/*------------------------------------------------------------------*/
			 
#include "const.h"
#include "clause.h"

#include <pvm3.h>
#include <strings.h>
#include <stdio.h>
#include <malloc.h>
#include <time.h>
#include <math.h>
 
/*----------------------------------------------------------------*/
/* external functions called from this file                       */
/*----------------------------------------------------------------*/
 
extern get_clauses(); 
extern search_tree();                   /* functions in search.c  */
 
extern calc_lit_hc();                    /* functions in hash.c   */
extern long calc_alt_hashcodes();
extern calc_clause_hashcode();
extern long add_clause_to_ht();
extern normalize_literal();
 
extern eliminate();                      /* function in elim.c    */
 
extern init_cl_array();                  /* functions in pass2.c  */
extern save_line(); 
extern stack_hash_codes();
extern rearrange_cl_array();
extern search_proof_line();
 
extern print_clause();                   /* function in tdump.c   */
extern dump_calloc_info();
extern print_continuation();

/*----------------------------------------------------------------*/
/* hash_code_list_active is a bit mapping of the hash codes on    */
/* the hash code list that remain to be found.  bit i is a 1      */
/* if hash code i has yet to be found.                            */
/*----------------------------------------------------------------*/
 
extern long hash_code_list_active;

/*----------------------------------------------------------------*/
/* the hash code of the clause whose complement is found in the   */
/* clause_hash_table implying a contradiction; it is thus magic!  */
/*----------------------------------------------------------------*/

extern long magic_hashx, magic_hashy;
 
/*----------------------------------------------------------------*/
/* internally defined functions                                   */
/*----------------------------------------------------------------*/


long time_now();
void init_stats();
int init_options();
int grab_number();
void fatal();
void enter_theorem_name();
void init_search_space();
void init_clause_hashtable();
void init_rand_hash_array();
void free_space_close_files();
void read_buffer_word();
void interchange_fxn_const();
int is_fxn_a_constant();
void sort_constants_fxns();
void format_clause();
int count_lit_terms();
void calc_max_term_in_any_lit();
void equal_predicate();
void esa_for_fxns();
void esa_for_lits();
int equality_sub_axiom();
int test_lits();
int get_pred_fxn_symbols();
void prepare_clauses();
void hash_all_base_clauses();
int prove_by_iterative_deepening();
void get_clause_stats();
void write_out_results();
void main();
 
/*----------------------------------------------------------------*/
/* arrays used for reading in input file and converting           */
/* to format compatible with the search.                          */
/*----------------------------------------------------------------*/
 
FILE *outfile;

char compiled_code_buffer[MAX_COMPILED_THM_SIZE];
char *comp_code_buffer_index;
long compiled_code_size[11];
 
pfnode pfspace[PF_SPACE_SIZE];
long kode[MAX_INPUT_THM_SIZE];
char name_buff[MAX_CHAR];
cnode pred_ptr_list[PRED_PTR_SIZE];

/*----------------------------------------------------------------*/
/* space for clause list generated in elim.c for simplifying      */
/*----------------------------------------------------------------*/

long elim_inferences[ELIM_ARRAY_SIZE]; /* clause list for elim.c  */

/*----------------------------------------------------------------*/
/* clist contains pointers to the clauses (base clauses and       */ 
/* clauses generated during the search.)                          */
/*----------------------------------------------------------------*/

clause *clist[CSPACE_CLAUSES];    /* first few are base clauses   */

/*----------------------------------------------------------------*/
/* each entry in depth is the top, bottom, and current clause     */
/* under search as well as a pointer to the current clause.       */
/* each entry is thus four long words.                            */
/*----------------------------------------------------------------*/

dnode  depth[DEPTH_PTR_SPACE];    /* depth ptr info               */

/*----------------------------------------------------------------*/
/* the following two arrays occupy a large amount of memory       */
/* the space is dynamically allocated and is currently set to:    */
/* clause space: 256K words     clause_hash_table: 256K entries   */
/*----------------------------------------------------------------*/  

long  *clause_space;
hnode *clause_hash_table;           /* ptrs to clause hash tables */
unsigned int array_size;
long array_size_m1;

/*----------------------------------------------------------------*/
/* rand_array is dynamically allocated. its size depends on the   */
/* particular theorem.                                            */
/*----------------------------------------------------------------*/
 
rnode rand_array[RAND_SIZE];
 
/*----------------------------------------------------------------*/
/*  arrays for storing name of input and outpout theorem          */
/*----------------------------------------------------------------*/

char *theorem_ptr, *theorem_out_ptr, *theorem_out_ptr2;
char theorem_name[256];  /* name of theorem read in                */
char theorem_out_name[256];/* name of file where output is placed  */
char theorem_out_name2[256];/* if sol found, new name of theorem   */
long theorem_out_name2end; /* ptr to last char in output file name*/

/*----------------------------------------------------------------*/
/* variables that are concerned with recording time               */
/*----------------------------------------------------------------*/
 
long elim_starting_time; 
long eliminate_time;
long search_start_time;
long search_time;
long phase1_time;                     /* time for phase 1         */
 
long base;                          /* ID base last base clause   */
long org_base;                /* original number of base clauses  */
long clauses_alive;      /* count of the number of ACTIVE clauses */
long nc;                         /* # clauses in negated concl.   */
long iter_depth;                   /* current iteration depth     */
long ply;                          /* current search depth        */
long nextc;                    /* location where next clause goes */
long lit_in_hash;
 
/*----------------------------------------------------------*/
/* number_rows, numb_cols, numb_entries in the random array */
/*----------------------------------------------------------*/
 
int numb_rows, numb_cols, numb_entries, rows_less_nfxns;
 
/*----------------------------------------------------------------*/
/* variables concerned with the state of the search               */
/*----------------------------------------------------------------*/
 
int in_elim;                    /* in_elim = 0 when in eliminate  */
int simple_sol;              /* simple_sol = 1 if elim finds sol  */
int extra_pass;     /* denotes whether more than 1 pass was done  */
int phase; /* search phase: elim = 0, 1st pass = 1, 2nd pass = 2  */
int big_bug;      /* set to 1 when a hash error misleads prover   */
 
char input_buffer[256];  /* buffer for input file name and options */ 
 
long ww;  /* temp loc for reading chars from compiled_code_buffer */
 
long *var_substitution_tbl[MAX_VARS*2]; /* var substitution table */
long pstack[500],*psp;              /* used in unify.c             */
 
long equal_pred;     /* =1 if "equal" or "Equal" pred is in thm   */
long equal_identity; /* the clause number (id) of the eq identity */
long max_var;
int number_preds;  /* number of different preds in input clauses  */
int number_fxns;   /* number of different fxns in input clauses   */
int number_constants; /* # of different constants in input cls    */
int number_const_nc;  /* number of constants in negated conclus   */
int number_const_hash; /* number of constants for hashing         */
int max_lit;
int max_term_in_lit;                /* maximum terms in a literal */
int max_term_in_nc;                 /* fattest literal in nc      */
long tab_amount;     /* used to keep track of the number of tabs  */

long maxl, maxv, maxx; /* max lits, vars, trms in clause in proof */
 
char *pidtbl[MAX_PRED_IDS];   /* ptrs to strings for predicates   */
char *fidtbl[MAX_FXNS_IDS];   /* ptrs to strings for functions    */
cnode *ctbl[MAX_BASE_CLAUSES];
 
/*----------------------------------------------------------------*/ 
/* The following are vars that record statistics on the search.   */
/*----------------------------------------------------------------*/
 
long s_nod;                       /* # nodes on all iterations    */
long s_nod0;                      /* # nodes at beginning of pass */
long s_res;              /* # resolvents from binary resolution   */
long s_fac;              /* # factors inferred                    */
long s_subsume;          /* # of s-subsumed clauses (see 6.14)    */
 
long s_htentry; /* # of entries in clause hash table (see 6.10.3) */
long s_hthit;   /* # of clause hash table hits                    */
long s_htelim;  /* # of clauses elim by ht hits (see 6.10.4)      */
long s_htfail;  /* # of failures to insert clause in hash table   */
 
long s_htentry_phase1;    /* # of entries in cl ht on phase 1     */
long s_hthit_phase1;      /* # of ht hits in phase 1              */
long s_htelim_phase1;     /* # of clauses elim by ht hits in ph 1 */
long s_htfail_phase1;     /* # of ht failures in phase 1          */

long s_ltentry_phase1;    /* # of entries in lit ht on phase 1    */
long s_lthit_phase1;      /* # of unit ht resolutions in phase 1  */
long s_ltfail_phase1;     /* # of unit clause failures in phase 1 */
long s_unit_res_phase1;   /* # of unit ht resolutions in phase 1  */
long s_unit_sub_phase1;   /* # of unit clause subsump in phase 1  */
long s_subsume_phase1;    /* # of s-subsumed clauses in phase 1   */

long s_ltentry;           /* # of single literal clauses          */
long s_lthit;     /* # unit hash table resolutions (see 6.11.3)   */
long s_ltfail;    /* # unit clause failures on entering ht        */
long s_unit_res;  /* # of hash table resolutions (see 6.11.2)     */
long s_unit_sub;  /* # of unit clause subsumptions (see 6.11.3)   */
long s_default;   /* equals zero if default parameters are used   */
 
/*----------------------------------------------------------------*/
/* the following are variables that (in most cases)               */
/* the user can set as desired.  They control the search.         */
/*----------------------------------------------------------------*/
 
int o_elim;       /* o_elim = 1 => do elim, else do not do elim   */
int o_merge;      /* MERGE = ancestors must be merge clauses      */
int o_nc;         /* 1 = nc-strategy                              */
int o_hash;       /* 1 = clause hash table used in proof          */
int o_extend;     /* UNIT = allow unit extension past maxd        */
int o_subsume;    /* 1 = simple-subsumbtion used (0 => not used)  */
int o_maxlit;     /* n = max size of clauses (0 = no limit)       */
int o_deltalit;
int o_maxvar;     /* o_maxvar = max # vars in a clause            */
int o_maxterm;    /* max terms in literal                         */
int o_deltaterm;
int o_maxtermM1;  /* max terms in literal - 1                     */
int o_maxsterm;   /* max term in literal in search phase          */
int o_maxtime;    /* max time to search for a proof               */
int o_constants;  /* determines how many const to hash*/
 
int o_step;       /* step depth                                   */
int o_print;      /* print depth                                  */
int o_cr;         /* print carriage return if 0, otherwise no cr  */
int o_debug;      /* user is supposed to be unaware of this optn  */
int o_statprint;  /* =1 => print statistics after each iteration  */
int o_relevant;   /* = 1 => print only clauses involved in proof  */
int o_elimprint;  /* o_elimprint = 1 => print elim info           */
int o_cade;
int o_short;      /* o_short = 1 => no phase 2   */
int o_noprint;    /* o_noprint = 1 => no printing */

int x_maxterm;    /* x_maxterm = 0 => calc o_maxterm              */
int x_stop;       /* can be set to 1 by user in interactive mode  */
 
 
int mytid;
int myparent;
int cc;
char buf[20000];
int SlaveFoundProof;
long Proofresults[2000],prp;
int slave_quit;
int theorem_number;
int rcvd_theorem_number;
int spawned;      /* number of computers in network  */
int this_computer; /* number assigned to this computer */


/*----------------------------------------------------------------*/
/*      time_now()                    ...................... #1
/*
/*      added to consolidate all system calls to tgtp 
/*
/*      called by:  main()                         tgtp.c
/*                  prove_by_iterative_deepening   tgtp.c
/*                  halted_search_options()        tdump.c
/*
/*      calls:      nobody
/*
/*      returns:    time of day as a long integer
/*---------------------------------------------------------------*/
 
long time_now()
{
return((long) time(0));
}
 
/*--------------------------------------------------------------*/
/*      init_stats()                   ..................... #2
/*
/*      init statistical parameters gathered during the search
/*
/*      called from:    main()       tgtp.c
/*
/*      calls:          nobody
/*
/*      returns:        nil
/*---------------------------------------------------------------*/
 
void init_stats()
{
s_nod  = 0L; 
s_fac  = 0L;             
s_res  = 0L;
s_nod  = 0L;             
s_nod0 = 0L;
 
/*---------------------------------------------------------*/
/* stats for clause hash table and lit clause hash table   */
/*---------------------------------------------------------*/
	
s_htentry = 0L;       
s_hthit   = 0L;      
s_htfail  = 0L;
s_ltentry = 0L;       
s_lthit   = 0L;      
s_ltfail  = 0L; 
s_htelim  = 0L;
s_unit_res = 0L;
s_unit_sub = 0L;
s_subsume = 0L;
 
tab_amount = 0L;        
	
}
 
/*----------------------------------------------------------------*/
/*      init_options()                          .............. #3
/*
/*      reads the search options that the user requests in the
/*      command line.
/*
/*      called from:  enter_theorem_name()    tgtp.c
/*
/*      calls:        grab_number()           tgtp.c
/*
/*      returns:      the number of chars in the name of the theorem
/*                    or a -1 if help was requested
/*----------------------------------------------------------------*/
 
int init_options()
{
int i,j;
unsigned int k;
 
/*------------------------------------------*/
/* initialize all options to default values */
/*------------------------------------------*/
    
o_elim      = 1;  /* Option 1... elim phase: 1/0/2 => do/skip/only */
o_merge     = 1;  /* Option 2... merge strat: 1/0 => off/on        */
o_nc        = 0;  /* Option 3... linear-nc: 0/1 => off/on          */
o_hash      = 1;  /* Option 4... clause ht on: 1/0=> cht off/on    */
o_extend    = 1;  /* Option 6... extended search: 1/0=> on/off     */
o_subsume   = 1;  /* Option 7... s-subsumption: 1/0 => use/don't   */
o_deltalit  = 0;
o_maxterm   = 16; /* Option 9... maximum # terms in a literal      */
o_deltaterm = 0;
o_maxsterm  = 0;
o_constants = 0; 
o_maxvar    = 0;   /* Option 10.. maximum # diff vars in a clause  */
 
o_maxtime   = 6000;/* Option 11.. maximum time for proof..minutes  */
o_debug     = 0;   /* Option 12.. debug option: 0/1 => off/on      */
 
o_step      = 0;   /* Option 13.. step through tree to ply o_step  */ 
o_print     = 0;   /* Option 14.. print tree to ply o_step         */
o_cr        = 0;
 
o_elimprint = 0;   /* Option 15.. 0/1 do/ do not give phase 0 info */
o_statprint = 1;   /* Option 16.. 0/1 do/do not print statistics   */
o_relevant  = 0;   /* Option 17.. 0/1 do/do not prt all proof cls  */

s_default = 0L;
 
big_bug     = 0;       /* set to 1 if a hash error found inphase 3 */
/*
array_size  = 2097152;
array_size_m1 = 2097151L; */ 

array_size  = 1048576;
array_size_m1 = 1048575L;/*
array_size  = 1048576/8;
array_size_m1 = (1048575L + 1L)/8L - 1L; */

/*------------------------------------------*/
/* scan the arguments passed to the program */
/* if a "?" appears anywhere on the line    */
/* print out a "help" message               */
/*------------------------------------------*/
 
for(i = 0; i < 77 ; i++) {
   if(input_buffer[i] == '?') {
	
/*--------------------------------------------------------------*/
/* a help message, triggered by a "?" anywhere on input line    */
/*--------------------------------------------------------------*/

printf("\n--------------------------------------------------------");
printf("\n| Enter name of the theorem and then options as in the example below:");
      printf(
      "\n| ");
      printf(
      "\n| photon:ths:stark103 p3 s2 z1 m0 t0 n0 l4 x6 k0 e1 g1");
      printf(
      "\n| ");
      printf(
      "\n| where:");
      printf(
      "\n|   k0 -- skip elim; k1 carry out elim; k2 carry out only elim");
      printf(
      "\n|   m0 -- turn off linear-merge strategy");
      printf(
      "\n|   z1 -- turn on linear-nc search strategy");
      printf(
      "\n|   h0/u0 -- turn off hash table for clauses (for unit clauses)");
      printf(
      "\n|   n0 -- turn off extended search strategy");
      printf(
      "\n|   b0 -- turn off s-subsumption strategy");
      printf(
      "\n|   l4 -- set maximum number of literals in a clause to four");
      printf(
      "\n|   x6 -- set maximum number of terms in a literal to six");
      printf(
      "\n|   v8 -- set maximum number of variables in a literal to eight");
      printf(
      "\n|   t7 -- set maximum search time to 7 minutes");
      printf(
      "\n|   e1 -- print info on the elim fxn; e2 denotes include resolutions");
      printf(
      "\n|   p3 -- print three levels of search tree");
      printf(
      "\n|   s2 -- step two levels of search tree");
      printf(
      "\n|   d1 -- turn on debugger to obtain iformation on search");
      printf(
      "\n|   g1 -- prints statistics after each iteration");
      printf(
      "\n|   r1 -- when proof is found, list only relevant clauses");
      printf(
      "\n| by default, k=1, m=1, z=0, h=u=1, n=1, b=1, t=60, e=0, p=s=0, d=0, g=0, r=0");
      printf(
      "\n| default values for l,x,v depend on the theorem");
      printf(
      "\n------------------------------------------------------");
      printf("\n");
		
      for(i = 0; i < 256; i++) {
         input_buffer[i] = ' ';}           /* clr input buffer  */
      return(-1);}                       /* -1 => re-enter      */
			
   if(input_buffer[i] == ' ') {
			
/*------------------------------------------------------*/
/* k0 => skip elim phase, k1 => carry out elim phase    */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'k') {
         s_default = 1L;
	       o_elim = grab_number(i + 2);
	       if(o_elim > 2 || o_elim < 0) {
	          printf("\nbad argument\n");return(-1);}}
	
/*------------------------------------------------------*/
/* m0 => no merge strategy; m1 => uses merge startegy   */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'm') {
         s_default = 1L;
	       o_merge = grab_number(i + 2);
	       if(o_merge > 1 || o_merge < 0) {
	          printf("\nbad argument\n");return(-1);}}
 
/*------------------------------------------------------*/
/* z0 => no nc strategy; z1 => uses nc strategy         */
/*------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'z') {
         s_default = 1L;
	       o_nc = grab_number(i + 2);
	       if(o_nc > 1 || o_nc < 0) {
	          printf("\nbad argument\n");return(-1);}}
 
/*------------------------------------------------------------------*/
/* h0 => turn off clause_hash_table h1 => turn on clause_hash_table */
/*------------------------------------------------------------------*/
 
      if(input_buffer[i + 1] == 'h') {
         s_default = 1L;
	       o_hash = grab_number(i + 2);
	       if(o_hash > 1 || o_hash < 0) {
	          printf("\nbad argument\n");return(-1);}}
 
/*--------------------------------------------------------------*/
/* n0 => no ext'd search str'gy; n1 => uses ext'd search str'gy */
/*--------------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'n') {
         s_default = 1L;
	       o_extend = grab_number(i + 2);
	       if(o_extend > 1 || o_extend < 0) {
	          printf("\nbad argument\n");return(-1);}}
	
/*--------------------------------------------------------------*/
/* b0 => no simple subsumptions; b1 => yes..simple subsumptions */
/*--------------------------------------------------------------*/
		
      if(input_buffer[i + 1] == 'b') {
         s_default = 1L;
	       o_subsume = grab_number(i + 2);
	       if(o_subsume  > 1 || o_subsume  < 0) {
	          printf("\nbad argument\n");return(-1);}}
	    
/*--------------------------------------------------------------*/
/* if lx is found, set max number of literals in a clause to x  */
/* x is restricted to fall between 2 and 20                     */
/*--------------------------------------------------------------*/
						
      if(input_buffer[i + 1] == 'l') {
         s_default = 1L;
	       o_maxlit = grab_number(i + 2);
	       if(o_maxlit < 2 || o_maxlit > 20) {
	          printf("\nbad argument\n");return(-1);}}
			
/*------------------------------------------------------------------*/
/* if xX is found, set max number of terms in a literal to X        */
/*------------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 'x') {
         s_default = 1L;
         o_maxsterm = grab_number(i + 2);
         x_maxterm = 1;
         if(o_maxsterm < 2 || o_maxsterm > 200) {
            printf("\nbad argument\n");return(-1);}}
			
/*------------------------------------------------------------------*/
/* if vX is found, set max number of variables in a clause to X     */
/*------------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 'v') {
         s_default = 1L;
         o_maxvar = grab_number(i + 2);
         if(o_maxvar < 2 || o_maxvar > 32) {
            printf("\nbad argument\n");return(-1);}}
 
/*------------------------------------------------------------------*/
/* if tX is found, set max search time to X                         */
/*------------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 't') {
         o_maxtime = grab_number(i + 2);
         if(o_maxtime < 1 || o_maxtime > 200000) {
            printf("\nbad argument\n");return(-1);}}
			
/*----------------------------------------------*/
/* if e0 => turn elim info off,                 */
/* if e1 => set eliminate info on               */
/* if e2 => set elim info on, list resolutions  */
/*----------------------------------------------*/
			
      if(input_buffer[i + 1] == 'e') {
         o_elimprint = grab_number(i + 2);
         if(o_elimprint < 0 || o_elimprint > 20) {
            printf("\nbad argument\n");return(-1);}}
 
/*----------------------------------------------*/
/* if px is found, set printing depth to x      */
/*----------------------------------------------*/
			
      if(input_buffer[i + 1] == 'p') {
         o_print = grab_number(i + 2);
         if(o_print < o_step) {
            o_print = o_step;}
         if(o_print < 0 || o_print > 40) {
            printf("\nbad argument\n");return(-1);}}
			
/*--------------------------------------------------------------*/
/* if sx is found, set depth of stepping during printing to x   */
/*--------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 's') {
         o_step = grab_number(i + 2);
         if(o_print < o_step) {
            o_print = o_step;}
         if(o_step < 0 || o_step > 40) {
            printf("\nbad argument\n");return(-1);}}
			
/*----------------------------------*/
/* if d is found, set debugger on?  */
/*----------------------------------*/
			
      if(input_buffer[i + 1] == 'd') {
         o_debug = grab_number(i + 2);
         if(o_debug  > 2 || o_debug  < 0) {
            printf("\nbad argument\n");return(-1);}}
				
/*------------------------------------------------------------------*/
/* if g is found , set statistics to be printed on each iteration   */
/*------------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 'g') {
         o_statprint = grab_number(i + 2);
         if(o_statprint > 1 || o_statprint < 0) {
            printf("\nbad argument\n");return(-1);}}
 
/*------------------------------------------------------------------*/
/* if r1 is found , set o_relevant=1 to print only relevant clauses */
/*------------------------------------------------------------------*/
			
      if(input_buffer[i + 1] == 'r') {
         o_relevant = grab_number(i + 2);
         if(o_relevant > 1 || o_relevant < 0) {
            printf("\nbad argument\n");return(-1);}}
		
/*------------------------------------------------------------------*/
/* axxxxx determines the size of clause_space and the size of       */
/* clause_hash_table.  It must be a power of 2!!!                   */
/*------------------------------------------------------------------*/

      if(input_buffer[i + 1] == 'a') {
         s_default = 1L;
         array_size = (unsigned int) grab_number(i + 2);
         array_size_m1 = (long) (array_size - 1);
         for(k = 16384; k != array_size ; k = k * 2) {
            if(k > array_size) {
               printf("\nbad argument;");
               if(k == 16384) {printf(" not enough memory space.\n");}
               else {printf(" memory space must be a power of 2.\n");}
               return(-1);}}}}}	
         

/*---------------------------------------------------*/
/* end of option list:                               */ 
/* blank out input_buffer except for name of theorem */
/*---------------------------------------------------*/
	
for(i = 0; i < 255; i++) {
   if(input_buffer[i] == ' ' || input_buffer[i] == '\n') {
      j = i;
      for(j = i; j < 256 ; j++) input_buffer[j] = ' ';
      break;}}

/*--------------------------------------------------*/ 
/* return the number of characters in the name      */
/*--------------------------------------------------*/

return(i);
}
 
/*-----------------------------------------------------------------*/
/*      grab_number(s)                       .............. #4
/*
/*      reads a string of characters and converts them to a
/*      integer equivalent
/*
/*      called from:  init_options()        tgtp.c
/*                    print_clause()        tdump.c
/*
/*      calls:        nobody
/*
/*      returns:      a long integer correspond to the string s
/*-----------------------------------------------------------------*/
 
int grab_number(s)
int s;
{
int n;
	
for(n = 0; ; s++) {
   if(input_buffer[s] >= '0' && input_buffer[s] <= '9')
      n = n * 10 + ( (int) (input_buffer[s] - '0') );
   else
      return(n);}
}                                                       
 
/*-----------------------------------------------------------------*/ 
/*      fatal(msg)                          ................. #5
/*
/*      fatal will print an error message on the screen and quit
/*
/*      called by:    main()              tgtp.c
/*                    prepare_clauses()   tgtp.c
/*                    format_clause()     tgtp.c
/*                    eliminate()         elim.c
/*
/*      calls:        nobody
/*
/*      returns:      to the operating system after reporting
/*                    error
/*-----------------------------------------------------------------*/
 
void fatal(msg)
char *msg;
{
   if(o_noprint == 0) {printf("%s",msg);}
   pvm_exit();
   exit(0);
}
 
 
/*-------------------------------------------------------------*/
/*      enter_theorem_name()           ................#6
/*
/*      read in the name of the theorem to be proved
/*
/*      called by:   enter_theorem_name()      tgtp.c
/*
/*      calls:       nobody
/*
/*      returns:     nil
/*--------------------------------------------------------------*/
 
void enter_theorem_name()
{
int number_chars;      /* number of characters in name of theorem  */ 
long bad_name;       /* bad_name = 0 until theorem found in file   */
int j, k, m;
long ask_for_help;    /* normally 0, set to 1 if "?" on input line */
char zz;

for(j = 0 ; j < 256 ; j++) {
   theorem_name[j] = theorem_out_name[j] = theorem_out_name2[j] = 0;
   input_buffer[j] = ' ';}
   
init_options();  

theorem_ptr = &theorem_name[0];
theorem_out_ptr = &theorem_out_name[0];
theorem_out_ptr2 = &theorem_out_name2[0];
 
}
 
/*-------------------------------------------------------------------*/
/*      init_search_space()            ......................#7
/*
/*      allocates space for clauses.
/*      allocates space for clause hash table.
/*      allocates space for pointers at each level in tree
/*
/*      called from:   main()       tgtp.c
/*
/*      calls:         nobody
/*
/*      returns:       nil
/*-------------------------------------------------------------------*/
 
void init_search_space()
{       

clause_space = (long *) calloc(array_size,4);   
   dump_calloc_info(
	 "clause_space",(long) clause_space,-1L,
		 "longs",(long) array_size);
 
clause_hash_table =  (hnode *) calloc(4 * array_size,4);
dump_calloc_info(
      "clause_hash_table",
      (long) clause_hash_table,-1L,"hnodes",(long) (array_size));

nextc = 0L;
clist[nextc] = (clause *) clause_space;
}
 
/*-----------------------------------------------------------------*/
/*      init_clause_hash_table()          .................#8
/*
/*      initialize the clause_hash_table to all 0's
/*
/*      called by:      main()                             tgtp.c
/*                      prove_by_iterative_deepening()     tgtp.c
/*
/*      calls:          nobody
/*
/*      returns:        nil
/*-----------------------------------------------------------------*/
 
void init_clause_hash_table()
{
 
/*----------------------------------------------*/
/* allocate 4 * HT_SIZE long words for the hash */
/* table.  initially set all words to 0.        */
/*----------------------------------------------*/
 
clause_hash_table =  (hnode *) calloc(4 * array_size,4);
dump_calloc_info(
      "clause_hash_table",
      (long) clause_hash_table,-1L,"hnodes",(long) (array_size));
return;
}
 

/*----------------------------------------------------------------*/
/*      init_random_hash_array(numb_terms,numb_fxns)    ........#9
/*
/*      initialize the rand_array to random numbers 
/*
/*      called from:   main()         tgtp.c 
/*
/*      calls:         nobody
/*
/*      returns:       nil
/*-----------------------------------------------------------------*/
 
void init_random_hash_array(numb_terms, numb_fxns)
int numb_terms, numb_fxns;
{
rnode *r1;
register int i;
unsigned int num_ent;
	
/*--------------------------------------------------------------*/
/* the number of rows in rand_array = MAX_VARS + numb_fxns      */
/* the number of columns in rand_array = numb_chars             */
/*--------------------------------------------------------------*/
 
numb_rows = MAX_VARS + number_fxns;
numb_cols = numb_terms;
rows_less_nfxns = numb_rows - number_fxns - 1;
 
/*--------------------------------------------------------------*/
/* the number of entries in the array rand_array is the         */
/* product of the numb_rows * numb_cols                         */ 
/*--------------------------------------------------------------*/
 
numb_entries = numb_rows * numb_cols;
num_ent = (unsigned int) (numb_entries * 2 + 1); 
if(num_ent < 50) {
   numb_entries = 25;
   num_ent = 51;}


/*--------------------------------------------------------------*/
/* make 32-bit random number for each element in rand_array     */
/*--------------------------------------------------------------*/
 
srand(1);
 
for(i = 0 ; i < RAND_SIZE ; i++) {
   r1 = &rand_array[i];
   r1->r_hash1 = (long) rand();
   r1->r_hash2 = (long) rand();}
}

/*-------------------------------------------------------------*/
/*      free_space_close_files()  ...................#10
/*
/*      frees space and closes files used in phases 1 and 2
/*
/*      called by:   main()                 tgtp.c
/*
/*      calls:       nobody
/*
/*      returns:     nil
/*--------------------------------------------------------------*/
 
void free_space_close_files()
{
free(clause_hash_table);
free(clause_space);      
}
 
/*--------------------------------------------------------------*/
/*      read_buffer_word()               .............. #11
/*
/*      read a word (integer) from the input file
/*
/*      called by:    prepare_clauses()     tgtp.c
/*                    format_clause()       tgtp.c
/*
/*      calls:        nobody
/*
/*      returns:      nil
/*----------------------------------------------------------------*/
 
void read_buffer_word()
{
union u_tag {
   long w; 
   char ttext[4];} 
   temp;
			
int j;
	
for(j = 0; j < 4; j++) {
   temp.ttext[j] = *comp_code_buffer_index++;}
ww = temp.w;
}
 
 
/*--------------------------------------------------------------*/
/*      interchange_fxn_const()          ............. #12
/*
/*       called by:      sort_constants_fxns()      tgtp.c
/*
/*       calls:          nobody
/*
/*       returns:         nil
/*--------------------------------------------------------------*/
 
void interchange_fxn_const(i,j)
long i,j;
{
long k, *p, a;
clause *c;
literal *l;
char *fide;
	
for(k = 0L; k <= base ; k++) {
   c = clist[k];
   for(l = c->c_lit ; l != 0L ; l = l->l_next) {
      p = &l->l_id;
      for(a = (*p++ & ARGMASK) ; a > 0L ; a--) {
	     if((*p & FCNMASK) != 0L) {
	        if(((*p & IDMASK) >> ARGSHIFT) == i) {
	           *p = (*p & NIDMASK) + (j << ARGSHIFT);}
	        else if(((*p & IDMASK) >> ARGSHIFT) == j) {
	           *p = (*p & NIDMASK) + (i << ARGSHIFT);}
	        a += *p++ &ARGMASK;}
	     else {
	        p++;}}}}
			
fide = fidtbl[i];
fidtbl[i] = fidtbl[j];
fidtbl[j] = fide;
}
 
/*--------------------------------------------------------------*/
/*      is_fxn_a_constant()                .............. #13
/*
/*      called by:      sort_constants_fxns()       tgtp.c
/*
/*      calls:          nobody
/*
/*      returns:        1 if fxn is a constant, 0 otherwise
/*-----------------------------------------------------------*/
 
int is_fxn_a_constant(i,top_clause)
long i,top_clause;
{
long k, a, *p;
clause *c;
literal *l;
	
for(k = top_clause; k <= base ; k++) { 
   c = clist[k];
   for(l = c->c_lit ; l != 0L ; l = l->l_next) {  /* scan lits  */
      p = &l->l_id;
      for(a = (*p++ & ARGMASK) ; a > 0L ; a--) {
	     if((*p & FCNMASK) != 0L) {
	        if(((*p & IDMASK) >> ARGSHIFT) == i) {
	           if((*p & ARGMASK) == 0L) return(1);
	           else return(0);}
	        a += *p++ & ARGMASK;}
	     else {
	        p++;}}}}
return(0);
}
 
/*--------------------------------------------------------------*/
/*      sort_constants_fxns()           .............. #14
/*
/*      sort the constants and functions so that the constants
/*      are placed at the top of the list of functions and constants
/*      once that is done, place the constants in the nc at the
/*      head of the constants list.
/*
/*      called by:   prepare_clauses()          tgtp.c
/*
/*      calls:       is_fxn_a_constant()        tgtp.c
/*                   interchange_fxn_const()    tgtp.c
/*
/*      returns:     nil
/*---------------------------------------------------------------*/
 
void sort_constants_fxns()
{
long i, k;
 
/*-----------------------------------------------*/
/*    put the constants at the head of the list  */
/*-----------------------------------------------*/
 
number_constants = 0;
for(i = 1L; i <= (long) number_fxns ; i++) {
   if(is_fxn_a_constant(i,0L) != 0) {
      number_constants++;       
      if(i >= (long) number_fxns) return;}
   else {
      for(k = (long) number_fxns; k > i; k--) {
	     if(is_fxn_a_constant(k,0L) != 0) {
	        interchange_fxn_const(i,k);
	        number_constants++;
	        break;}}}}  

/*-----------------------------------------------*/
/*   now resort the constants in the nc to the   */
/*   top of the list.                            */
/*-----------------------------------------------*/

number_const_nc = 0;
for(i = 1L; i <= (long) number_constants; i++) {
   if(is_fxn_a_constant(i,base - nc + 1L) != 0) { 
      number_const_nc++; 
      continue;}
   else {
      for(k =  i + 1L; k <= (long) number_constants; k++) {
	       if(is_fxn_a_constant(k,base - nc + 1L) != 0) {
            number_const_nc++;
	          interchange_fxn_const(i,k);
	          break;}}}} 

if(number_constants <= 5) {
   number_const_hash = number_constants;}
else {
   number_const_hash = number_const_nc;
   if(number_const_hash < 3) {
      number_const_hash = 3;}}
if(o_constants == 1) {number_const_hash = number_constants;}

}

/*--------------------------------------------------------------*
/*      equal_predicate()      ...................... #15
/*
/*      see if "Equal" predicate is implicitly one of the     
/*      predicates.  Look for p(x,x), the ~p(x,y) | p(y,x), etc.
/*
/*--------------------------------------------------------------*/

void equal_predicate()
{
literal *l, *l1, *l2, *l3; 
clause *c;
long i, j, k, v1, v2, v3, v4, v5, v6, *p, *q, *r;
long pargs, qargs;
long l1id, l2id, l3id;
long ee, plus_lit;

for(i = 0L; i <= base ; i++) { 
   c = clist[i]; 
   if(c->c_nlit != 1) continue;
   if(c->c_nv != 1L) continue;
   if((c->c_lit->l_id & ARGMASK) != 2L) continue;
   if((c->c_lit->l_id & NEGMASK) != 0L) continue;
   p = &c->c_lit->l_id;
   p++; 
   v1 = *p++;
   v2 = *p;
   if(v1 != v2) continue;
   if((v1 & FCNMASK) != 0L) continue;
   ee = c->c_lit->l_id & IDMASK;
   c->c_bits |= EQCLS;

   if(c->c_lit->l_id == equal_pred) {
      equal_identity = i;}

   for(j = 0L; j <= base; j++) {
      c = clist[j];
      if(c->c_nlit != 2) continue;
      if(c->c_nv != 2) continue;
      l1 = c->c_lit;
      if((l1->l_id & IDMASK) != ee) continue;
      l2 = l1->l_next;
      if((l1->l_id & NEGMASK) == (l2->l_id & NEGMASK)) continue;
      if((l1->l_id & IDMASK) != (l2->l_id & IDMASK)) continue;
      p = &l1->l_id;
      q = &l2->l_id;
      p++; q++;
      v1 = *p++;
      v2 = *p;
      v3 = *q++;
      v4 = *q; 
      if(v1 != v4) continue;
      if(v2 != v3) continue;
      if((v1 & FCNMASK) != 0L) continue;
      if((v2 & FCNMASK) != 0L) continue;

      c->c_bits |= EQCLS; 

      for(k = 0L; k <= base; k++) {
         c = clist[k]; 
         if(c->c_nlit != 3) continue;
         if(c->c_nv != 3) continue;
         l1 = c->c_lit;
         if((l1->l_id & IDMASK) != ee) continue;
         l2 = l1->l_next;
         l3 = l2->l_next;
         l1id = l1->l_id; l2id = l2->l_id; l3id = l3->l_id;
         if((l1id & NEGMASK) == (l2id & NEGMASK) &&
              (l1id & NEGMASK) == (l3id & NEGMASK)) continue;

         if((l1->l_id & IDMASK) != (l2->l_id & IDMASK)) continue;
         if((l1->l_id & IDMASK) != (l3->l_id & IDMASK)) continue;

         if(((l1->l_id & NEGMASK) == 0L) &&
            ((l2->l_id & NEGMASK) != 0L) &&
            ((l3->l_id & NEGMASK) != 0L))  {
            plus_lit = 1L;}
         if(((l1->l_id & NEGMASK) != 0L) &&
            ((l2->l_id & NEGMASK) == 0L) &&
            ((l3->l_id & NEGMASK) != 0L))  {
            plus_lit = 2L;}
         if(((l1->l_id & NEGMASK) != 0L) &&
            ((l2->l_id & NEGMASK) != 0L) &&
            ((l3->l_id & NEGMASK) == 0L))  {
            plus_lit = 3L;}
         p = &l1->l_id;
         q = &l2->l_id;
         r = &l3->l_id;

         p++; q++; r++;
         v1 = *p++;
         v2 = *p;
         v3 = *q++;
         v4 = *q; 
         v5 = *r++;
         v6 = *r;
         if(v1 == v3 && v2 == v6 && v4 == v5) goto equal3;
         if(v1 == v3 && v2 == v5 && v4 == v6) goto equal3;
         if(v1 == v4 && v2 == v6 && v3 == v5) goto equal3;
         if(v1 == v4 && v2 == v5 && v3 == v6) goto equal3;
         if(v1 == v5 && v2 == v4 && v3 == v6) goto equal3;
         if(v1 == v5 && v2 == v3 && v4 == v6) goto equal3;
         if(v1 == v6 && v2 == v4 && v3 == v5) goto equal3;
         if(v1 == v6 && v2 == v3 && v4 == v5) goto equal3;
         continue;
         equal3:
         if((v1 & FCNMASK) != 0L) continue;
         if((v2 & FCNMASK) != 0L) continue;
         if((v3 & FCNMASK) != 0L) continue;
         if((v4 & FCNMASK) != 0L) continue;
         if((v5 & FCNMASK) != 0L) continue;
         if((v6 & FCNMASK) != 0L) continue;

         c->c_bits |= EQCLS;
         break;}}}

return;
}

/*-----------------------------------------------------*/
/*      esa_for_fxns()    ................#16          */
/*                                                     */
/*      determine whether clause c is an equality      */
/*      substitution axiom, e. g. ~equal(x,y) |        */
/*      equal(f(x),f(y)                                */
/*                                                     */
/*      called by:   prepare clauses()     tgtp.c      */
/*                                                     */
/*      calls:       equality_sub_fxn()    tgtp.c      */
/*                                                     */
/*      returns:     1 is clause is an esa             */
/*                   0 if it is not.                   */
/*-----------------------------------------------------*/
 
void esa_for_fxns()
{
clause *c;
literal *l1, *l2;
long i, v1, v2, *p;

for(i = 0L; i <= base; i++) { 
   c = clist[i]; 
   if(c->c_nlit != 2) {continue;}
 
   l1 = c->c_lit;
   l2 = l1->l_next;
 
   if((l1->l_id & ~NEGMASK) != equal_pred) {continue;}
   if((l2->l_id & ~NEGMASK) != equal_pred) {continue;}
   if((l1->l_id ^ l2->l_id ^ NEGMASK) != 0L) {continue;} 
 
   if((l1->l_id & NEGMASK) == 0L) {
      l1 = l2;
      l2 = c->c_lit;}
/*-------------------------------------------------------*/
/*   clause with two equal lits.  l1 is the "~" one.     */
/*   now see if l1 is of form equal(x,y)                 */
/*-------------------------------------------------------*/

   p = &l1->l_id;
   p++;

   v1 = *p++;
   v2 = *p;
   if(((v1 & FCNMASK) == 0L) && ((v2 & FCNMASK) == 0L && v1 != v2)) {
 
/*--------------------------------------------------------*/
/*   thus far, we have found that "~equal(x,y) is present */
/*--------------------------------------------------------*/

      if(test_equality_fxn(l2,v1,v2)) {
         c->c_bits |= EQCLS;}}}
return;
}
 
/*-----------------------------------------------------------------*/
/*   test_equality_fxn()           ................#17             */
/*                                                                 */
/*   now see if the equality literal is a equality substitution    */
/*   for example, equal(f(x,y),f(z,y)) would pass.                 */ 
/*                                                                 */
/*   called by:  esa_for_fxns()       tgtp.c                       */
/*                                                                 */
/*   calls:      nobody                                            */
/*                                                                 */
/*   returns:    0 if l_equal is not an equality substitution fxn  */
/*               1 if it is                                        */ 
/*-----------------------------------------------------------------*/
 
int test_equality_fxn(l_equal,v1,v2)
literal *l_equal;
long v1, v2;
{
long *p, *p1, a;
int esv;                    /* equality substitution variable  */
 
esv = 0;
 
p1 = p = &l_equal->l_id;
 
for(a = (*p++ & ARGMASK) ; a > 1L ; a--) {
   if(*p & FCNMASK) {
      a += *p++ & ARGMASK;}
   else {
      p++;}}
/*--------------------------------------*/
/* if first symbol is a variable, then  */
/* return, since clause shows symmetry  */
/* in the equality predicate.           */
/*--------------------------------------*/

if((*p & FCNMASK) == 0L) {return(0);} 

/*--------------------------------------*/
/* p now points to beginning of 2nd     */
/* half of the equality predicate.      */
/* p1 now points to the beginning.      */
/* now a scan is made comparing each    */
/* symbol in the two halfs              */
/*--------------------------------------*/

for(a = (*p1++ & ARGMASK) ; a > 1L ; a--) {
   if(*p != *p1) {
      if((*p == v1 && *p1 == v2) || (*p == v2 && *p1 == v1)) {
         esv++;
         if(esv > 1) {
            return(0);}}}
   if(*p1 & FCNMASK) {
      a +=*p1++ & ARGMASK;
      p++;}
   else {
      p++; p1++;}} 
 

return(1);
}
 
 
/*-----------------------------------------------------*/
/*      esa_for_lits()    ................#18          */
/*                                                     */
/*      determine whether clause c is an equality      */
/*      substitution axiom  for literals.              */
/*      for example: ~equal(x,y) | ~P(x) | P(y)        */
/*                                                     */
/*      called by:   prepare clauses()     tgtp.c      */
/*                                                     */
/*      calls:       test_lits()           tgtp.c      */
/*                                                     */
/*      returns:     1 is clause is an esa             */
/*                   0 if it is not.                   */
/*-----------------------------------------------------*/
 
void esa_for_lits()
{
clause *c;
literal *l1, *e_lit, *p_lit, *n_lit;
long i, v1, v2, *p;

for(i = 0L; i <= base; i++) { 
   c = clist[i]; 
   if(c->c_nlit != 3) {continue;}

   e_lit = p_lit = n_lit = 0L;

   for(l1 = c->c_lit; l1 != 0L; l1 = l1->l_next) {
      if((l1->l_id & ~NEGMASK) == equal_pred) {e_lit = l1;}
      else if((l1->l_id & ~NEGMASK) != equal_pred) {
         if((l1->l_id & NEGMASK) == 0L) {p_lit = l1;}
         else {
            n_lit = l1;}}}
   if(e_lit == 0L || p_lit == 0L || n_lit == 0L) {continue;}
   if((p_lit->l_id & ~NEGMASK) != (n_lit->l_id & ~NEGMASK)) {continue;}

   p = &e_lit->l_id;
   p++;

   v1 = *p++;
   v2 = *p;
   if(((v1 & FCNMASK) == 0L) && ((v2 & FCNMASK) == 0L && v1 != v2)) {
 
/*--------------------------------------------------------*/
/*   thus far, we have found that "~equal(x,y) is present */
/*   now test the two literals.                           */
/*--------------------------------------------------------*/

      if(test_lits(p_lit,n_lit,v1,v2)) { 
         c->c_bits |= EQCLS;}}}

return;
}
  
/*-----------------------------------------------------------------*/
/*   test_lits()                   ................#19             */
/*                                                                 */
/*   now see if other two literals differ by only one var.         */
/*   for example, p(x,y) and p(x,z)    would pass.                 */ 
/*                                                                 */
/*   called by:  esa_for_lits()       tgtp.c                       */
/*                                                                 */
/*   calls:      nobody                                            */
/*                                                                 */
/*   returns:    0 if l_equal is not an equality substitution fxn  */
/*               1 if it is                                        */ 
/*-----------------------------------------------------------------*/
 
int test_lits(l1,l2,v1,v2)
literal *l1, *l2;
long v1, v2;
{
long *p1, *p2, a;
int esv;                    /* equality substitution variable  */
 
esv = 0;
 
p1 = &l1->l_id;
p2 = &l2->l_id;
 
for(a = (*p1++ & ARGMASK) ; a > 0L ; a--) {
   if(*p2 != *p1) {
      if((*p2 == v1 && *p1 == v2) || (*p2 == v2 && *p1 == v1)) {
         esv++;
         if(esv > 1) {
            return(0);}}}
   if(*p1 & FCNMASK) {
      a +=*p1++ & ARGMASK;
      p2++;}
   else {
      p2++; p1++;}} 
return(1);
}
/*--------------------------------------------------------------*/
/*      format_clause()                   .............. #20
/*
/*      format_clause will read clause from the input and stuff
/*      it into the clist
/*
/*      called from:    prepare_clauses()   tgtp.c
/*
/*      calls:          read_buffer_word()  tgtp.c
/*                      fatal()             tgtp.c
/* 
/*      returns:        nil
/*--------------------------------------------------------------*/
 
void format_clause()
{
clause *c;
literal *previous_literal;
long *p,a;
	
ply = 0L;       
read_buffer_word();read_buffer_word();/* these are leading 0's */
 
/*------------------------------------------------------*/
/* initialize clause header (this is not part of the    */
/* input file, we add it as we read in the clause)      */
/*------------------------------------------------------*/
 
c = clist[nextc++];                  /* the new clause         */
c->c_id = nextc;
c->c_bits = ACTIVE | MERGE | BASE; /* base clauses are merges  */
if(nextc > base - nc +1L) {
    c->c_bits |= NC;}                      /* clause in NC     */
c->c_rpar = c->c_lpar = 0L;      /* no parents (base clause)   */
c->c_lit = 0L;                       /* ^ first literal        */
c->c_nlit = 0;                             /* # literals       */
c->c_nv   = 0;                       /* # distinct variables   */
	
p = (long *) (c + 1);           /* where predicate code goes   */
previous_literal = 0L;                /* literal chain         */
 
/*----------------------------------------*/
/* read in each literal, end clause at -1 */
/*----------------------------------------*/    
 
read_buffer_word(); 
while(ww != -1L) {
   if(previous_literal == 0L)            /* link last literal  */
      c->c_lit = (literal *) p;
   else {
      previous_literal->l_next = (literal *) p;}
      
   previous_literal = (literal *) p;
   previous_literal->l_id = ww;
   previous_literal->l_next = 0L;
 
   p = (long *) (previous_literal + 1);
 
   for(a = ww & ARGMASK; a > 0L; a--) {
      read_buffer_word();
      if((ww & FCNMASK) != 0L) {               /* function? */
	     *p++ = ww;                            /* yes       */
	     a += ww & ARGMASK;}                   /* end if    */ 
      else {
	     ww >>= ARGSHIFT;                      /* var ID    */
	     *p++ = ww;
	     if(ww > c->c_nv) {
	        c->c_nv = ww;}}}                     /* end for */
   
/*----------------------------------------------------------------*/
/* transform equality predicates to standard form                 */
/*----------------------------------------------------------------*/
 
   c->c_nlit++;
		
   read_buffer_word();}                       /* end while  */
 
/*-----------------------------------------------------------------*/
/* set up a ptr to the next clause, pointed to by clist[nextc]     */
/*-----------------------------------------------------------------*/
 
c->c_end = (clause *) p;
clist[nextc] = c->c_end;
	
/*--------------------------------------------------*/
/* if clause has only one literal, set UNIT flag on */
/*--------------------------------------------------*/
 
if(c->c_nlit == 1) {
   c->c_bits |= UNIT;}
 
if(++c->c_nv > MAX_VARS) {
   fatal("Too many variables\n");}
 
if(c->c_nlit == 0) {
   fatal("The NULL clause??\n");}
 
}
 
/*---------------------------------------------------------------*/
/*      count_lit_terms()                     .............. #21
/*
/*      count_lit_terms counts terms in lit l
/*
/*      called from:  calc_max_term_in_any_lit()        tgtp.c
/*
/*      calls:        nobody
/*
/*      returns:      number of terms in literal l
/*----------------------------------------------------------------*/
 
int count_lit_terms(l)
literal *l;                    /* l is a literal in clause c      */
{
register long a,*p;
register int literal_terms; /* count of the number of terms in lit */
	
p = &l->l_id;        
literal_terms = 0;     
	
for(a = *p++ & ARGMASK; a > 0L; a--) {
   if((*p & FCNMASK) != 0L) { 
      literal_terms++;  /* function; inc count; det. # arguments   */
      a += *p++ & ARGMASK;}
   else {               
      literal_terms++;
      p++;}}              /* variable; inc count                   */
 
return(literal_terms);
}
 
/*--------------------------------------------------------------*/
/*      calc_max_term_in_any_lit()           .............. #22
/*
/*      calculates the maximum number of terms in any literal
/*      o_maxsterm records this maximum number
/*
/*      called from:    prepare_clauses()         tgtp.c
/*
/*      calls:          count_lit_terms()         tgtp.c
/*
/*      returns:        nil
/*--------------------------------------------------------------*/
 
void calc_max_term_in_any_lit()
{
clause *c;
literal *lit;
long clause_index, ncc;
int lit_term_cnt;

max_term_in_nc = 0;
max_term_in_lit = 0;

if(nc == 0L) ncc = 1L; else ncc = nc; 

for(clause_index = 0L; clause_index<= base; clause_index++) {
   c = clist[clause_index];
   for(lit = c->c_lit ; lit != 0L ; lit = lit->l_next) {
      lit_term_cnt = count_lit_terms(lit);
      if(lit_term_cnt > max_term_in_nc && 
                      clause_index >= base - ncc + 1L &&
                      clause_index <= base) {
         max_term_in_nc = lit_term_cnt;}
      if(lit_term_cnt > max_term_in_lit) {
	     max_term_in_lit = lit_term_cnt;}}}

return;
}
			
/*--------------------------------------------------------------*/ 
/*      get_pred_fxn_symbols()        .............. #23
/*
/*      get_pred_fxn_symbols will read in a symbol table and write
/*      it to the output file (and to the log file if there is one
/*
/*      called from:   prepare_clauses()          tgtp.c
/*
/*      calls:         nobody
/*
/*      returns:       the number of symbols (pred/fxns)
/*---------------------------------------------------------------*/
 
int get_pred_fxn_symbols(qtbl)
char *qtbl[];
{
int a, k;
char *ii, *jj;
static char Eq[] = "Equal";
static char eq[] = "equal";
		
for(a = 1L; ;a++) {  

/*----------------------------------------*/
/* check first if at least one pred, fxn  */           
/*----------------------------------------*/

   if(*comp_code_buffer_index == 0xa) {
      comp_code_buffer_index++; 
      return(0);}

/*----------------------------------------*/
/* set qtbl[a] to point to beginning of   */
/* the a-th predicate or function         */
/*----------------------------------------*/

   qtbl[a] = ii = comp_code_buffer_index;

   while(*comp_code_buffer_index != 0xa) { 
      comp_code_buffer_index++;}
 
/*----------------------------------------*/
/* replace terminating '0xa' with a 0     */
/*----------------------------------------*/
  
   *comp_code_buffer_index++ = 0;
 
/*--------------------------------------------------*/
/* print string of characters representing pred     */    
/*--------------------------------------------------*/
    
/*--------------------------------------------------*/
/* test for equality predicate                      */
/*--------------------------------------------------*/
   jj = ii;
	
   for(k = 0; k < 5 ; k++) {
      if(*jj++ != Eq[k]) {k = 0; break;}}
   if(k != 0 && *jj == 0) { 
      equal_pred = (long) (a << ARGSHIFT) + 2L;}

   jj = ii;
   for(k = 0; k < 5 ; k++) {
      if(*jj++ != eq[k]) {k = 0; break;}}
   if(k != 0 && *jj == 0) { 
      equal_pred = (long) (a << ARGSHIFT) + 2L;} 
    
   if(*comp_code_buffer_index == 0xa) {
      comp_code_buffer_index++;
      return(a);}}                      
}                                                       
 
/*--------------------------------------------------------------*/
/*      prepare_clauses()                .............. #24
/*
/*      prepare_clauses will read the clauses currently stored
/*      in the input_buffer and format them, placing the resulting
/*      clauses in cspace with pointers to each in the clist
/*
/*      NOTES:
/*      - we assume that all variables are numbered from 0..n so
/*        that n+1 is the number of distinct variables
/*
/*      - we pass the predicate and function symbol tables on to
/*        the output directly since we have no need for them
/*
/*      called from:    main()                          tgtp.c
/*
/*      calls:          get_pred_fxn_symbols()          tgtp.c
/*                      read_buffer_word()              tgtp.c
/*                      format_clause()                 tgtp.c
/*                      calc_max_term_in_any_litar()    tgtp.c
/*                      sort_constants_fxns()           tgtp.c
/* 
/*      returns:        nil
/* -------------------------------------------------------------*/
 
void prepare_clauses()
{
static long characters[] =
	 {0L,3L,5L,7L,9L,12L,15L,18L,21L,24L,26L,28L,30L};
long n;
int max_char_difference, char_difference;
 
/*--------------------------------------------------*/
/* set pointer to beginning of compiled_code_buffer */
/*--------------------------------------------------*/
 
comp_code_buffer_index = compiled_code_buffer;
 
/*--------------------------------------------------------------*/
/* We first pass the symbol table on to the output file.        */
/* This symbol table is two sets of non-empty lines followed    */
/* by an empty line (the predicate and function symbol tables   */
/*--------------------------------------------------------------*/
    
number_preds = get_pred_fxn_symbols(pidtbl); /* pred sym table   */

number_fxns = get_pred_fxn_symbols(fidtbl);/* function sym table */
 
read_buffer_word(); 
 
n = ww; 
base = ww - 1L;              /* clist[base] is last base clause  */
org_base = base;              /* original number of base clauses */
 
read_buffer_word();nc = ww;   /* # clauses in negated conclusion */
	
if(n < 2L) {
   fatal("Too few clauses!?\n");}
 
/*--------------------------------------------*/
/* read the clauses and put them in the clist */
/*--------------------------------------------*/
  
while(n-- > 0L) { 
   format_clause();}
 
sort_constants_fxns();
 
calc_max_term_in_any_lit();
 
/*-----------------------------------------------*/
/* given the maximum number of terms in          */
/* any literal in any base clause,               */
/* determine o_maxsterm, the maximum number of   */
/* terms in any literal kept during the          */
/* search.  Using the default value:             */
/*                                               */
/*       max_term_in_lit    o_maxsterm           */
/*           0                  0                */
/*           1                  3                */
/*           2                  4                */
/*           3                  5                */
/*           4                  6                */
/*           5                  7                */
/*           6                  8                */
/*           7                  9                */
/*           8                 11                */
/*           9                 13                */
/*          10                 16                */
/*          11                 19                */
/*          12                 22                */
/*          13+       3 * max_term_in_lit - 14   */
/*                                               */
/* the user can set o_maxsterm using the xi      */
/* option.                                       */
/*-----------------------------------------------*/
 
max_char_difference = 2*max_term_in_lit;
char_difference = max_char_difference/spawned + 1;
if(char_difference < 2) char_difference = 2;

if(x_maxterm == 0) {
   if(max_term_in_lit <= 12) o_maxsterm = characters[max_term_in_lit];
   else o_maxsterm = 2 * max_term_in_lit + 6;
   o_maxsterm = o_maxsterm + 7;
   if(o_deltalit == 0) {
      o_maxterm = max_term_in_nc+this_computer*char_difference;}
   else {
      o_maxterm = o_maxsterm - 13;}
   if(o_maxterm < 3) o_maxterm = 3;
   o_maxtermM1 = o_maxterm - 1;}

if(x_maxterm != 0) {
   if(o_deltaterm == 0 && o_deltalit == 0) {
      o_deltaterm = char_difference;}
   o_maxterm = x_maxterm + this_computer*o_deltaterm;
   o_maxtermM1 = o_maxterm - 1;
   if(o_maxsterm < max_term_in_lit) {
      o_maxsterm = 2*max_term_in_lit+6;}}
 
/*--------------------------------------------------*/
/* determine the max number of literals in any      */
/* base clause and the maximum number of            */
/* different variable names in any base literal     */
/*--------------------------------------------------*/
 
max_lit = 1;
max_var = 0L;
	
for(n = 0L ; n <= base ; n++) {
   if(clist[n]->c_nv > max_var)  {
      max_var = clist[n]->c_nv;}}
if(o_maxvar == 0) o_maxvar = 2 * (int) max_var;
      
for(n = 0L ; n <= base ; n++) {
   if(clist[n]->c_nlit > max_lit) {
      max_lit = clist[n]->c_nlit;}}

if(o_maxlit == 0) {
   o_maxlit = max_lit - 1;
   if(o_maxlit < 5) {o_maxlit = 5;}}
if(max_lit < o_maxlit) {max_lit = o_maxlit;}

/*--------------------------------------*/
/* search for equal predicate           */
/*--------------------------------------*/

equal_predicate();

/*--------------------------------------*/
/* search for equality substit. axioms  */
/*--------------------------------------*/

esa_for_fxns();
esa_for_lits();
}
 
 
/*---------------------------------------------------------------*/
/*      hash_all_base_clauses()            .............. #25
/*
/*      calculates a hash code for each literal and a hash code
/*      for each base clause.  It adds the clause to the literal
/*      hash table if there is only one literal and to the
/*      the clause hash table in any case
/*
/*      called from: main()                      tgtp.c
/*
/*      calls:       calc_lit_hc()               hash.c
/*                   calc_clause_hashcode()      hash.c
/*                   add_clause_to_ht()          hash.c
/*                   calc_alt_hashcodes()        hash.c
/*
/*      returns:     nil
/*--------------------------------------------------------------*/

void hash_all_base_clauses()
{
clause *c;
literal *l;
long i;
int l_cnt, o_maxterm_tmp;
	
o_maxterm_tmp = o_maxterm;
o_maxterm = 200;
 
for(i = 0L ; i <= base ; i++) {
   c = clist[i];
   for(l = c->c_lit ; l != 0L ; l = l->l_next) {
      l_cnt = calc_lit_hc(l,c);}
   calc_clause_hashcode(c);
   add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);
   if(c->c_nlit == 1) {
      normalize_literal(c->c_lit,c);
	    add_clause_to_ht(c->c_lit->l_X,c->c_lit->l_Y,1);
      if(c->c_nv !=  0L) {  
	       calc_alt_hashcodes(c->c_lit,c);}}}
 
o_maxterm = o_maxterm_tmp;
return;
}       
 
					
/*----------------------------------------------------------------*/
/*      prove_by_iterative_deepening()      .............. #26
/*
/*      prove will actually perform the iteratively deepening
/*      depth first search.  We return the depth at which we stop
/*
/*      called by:      main()                     tgtp.c
/*
/*      calls:          free_hash_tables()         tgtp.c
/*                      init_clause_hash_table()   tgtp.c
/*                      add_clause_to_ht()         hash.c
/*                      calc_alt_hashcodes()       hash.c
/*                      search_tree()              search.c
/*                      time_now()                 tgtp.c
/*
/*      returns:        integer regarding the result of search
/*---------------------------------------------------------------*/
 
int prove_by_iterative_deepening()
{
int s;              /* normally s = 0; if proof found, s = 1     */
long i; 
clause *c;
 
if(o_print != 0) {
   printf("\n\nSearch dump: \n");}
 
/*---------------------------------------------------------------*/
/*                                                               */
/* Main iterative deepening loop begins here                     */
/*                                                               */
/*---------------------------------------------------------------*/
	
iter_depth = 0L;
s_nod0 = s_nod;
 
/*----------------------------------------------------------------*/
/* before beginning search, clear out hash tables                 */
/*----------------------------------------------------------------*/ 
 
free(clause_hash_table);
init_clause_hash_table();       
 
/*------------------------------------------*/
/* add clauses to clause_hash_table         */
/* and unit clauses to literal_hash_table   */
/*------------------------------------------*/
 
for(i = 0L ; i <= base ; i++) {                         
   c = clist[i];
   if((c->c_bits & ACTIVE) == ACTIVE) { 
      add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);}
   if(c->c_nlit == 1) {
      normalize_literal(c->c_lit,c);
	    add_clause_to_ht(c->c_lit->l_X,c->c_lit->l_Y,1L);
      if(c->c_nv != 0L) {
	       calc_alt_hashcodes(c->c_lit,c);}}}
 
while(iter_depth < MAX_ITERATIONS) {                    
 
/*--------------------------*/
/* inc iteration depth      */    
/*--------------------------*/

   if(iter_depth > 10L) {
      o_maxlit++;o_maxterm++;o_maxvar++;}
   iter_depth++;
		
/*--------------------------*/
/* Begin search!!!!         */
/*--------------------------*/
 
   s = search_tree();   
    
/*----------------------------------------*/
/* if proof found (s != 0), and in        */
/* phase 1, then save line                */
/*----------------------------------------*/
 
   if(s == 1 && phase == 1) {
      save_line(0L);}
 
/*----------------------------------------*/
/* save the current time                  */
/*----------------------------------------*/

   search_time = time_now();
		
/*----------------------------------------------------------*/
/* there are two printing routines during the search:       */ 
/* one is long, one is short                                */
/*                                                          */
/* extended printing routine: on each iteration print:      */
/*          1. time so far                                  */
/*          2. nodes searched on this pass                  */
/*          3. clause hash table entries                    */
/*          4. clause hash table failures                   */
/*          5. unit clause hash table entries               */
/*          6. unit clause hash table failures              */
/*                                                          */      
/* extended printing routine,  controlled by o_statprint    */
/* use g1 in line of theorem to use "long" print, g0 short  */
/*----------------------------------------------------------*/

if(o_noprint == 0) {
   printf("%ld {T%ld N%ld R%ld F%ld C%ld H%ld h%ld U%ld r%ld}\n",
	    iter_depth,(search_time - search_start_time),
	    s_nod - s_nod0,s_res,s_fac,s_htentry,s_hthit,s_htfail,
      s_ltentry,s_unit_res);
   fflush(stdout);}

/*------------------------------*/
/* if proof found, print *, ret */
/*------------------------------*/

   if(s == 1 && o_statprint == 1)  {
      if(o_noprint == 0) {
         printf("*\n"); 
         fflush(stdout);}
      return(1);}
   if(s == 1 && o_statprint == 0)  {
      if(o_noprint == 0) {
         printf("*");
         fflush(stdout);}
      return(1);}
   if(s != 1 && o_statprint == 1)  {
      if(o_noprint == 0) {fflush(stdout);}}
   
/*------------------------------*/
/* too much time,...quit        */
/*------------------------------*/
 
   if(s == 2)  {/*
      printf("\n\nTime has expired; no proof found\n");
      printf("Last continuation searched was:");
      print_continuation(); */
      ply = 0L;
      fflush(stdout);
      return(2);}
      
/*------------------------------*/
/* giving up search,....quit    */
/*------------------------------*/
 
   if(s == 3)  { /*
      printf("\n\nGiving up the search; no proof found\n"); */
      ply = 0L;
      fflush(stdout);
      return(2);}
   
   if(s_nod == 0L) { 
      if(o_noprint == 0) { 
         printf("\n\nNo resolvents found!\n"); 
         fflush(stdout);}
      o_maxlit++;o_maxterm++;}}
	       
/*----------------------*/              
/* return with no proof */
/*----------------------*/

if(o_noprint == 0) {fflush(stdout);}
return(4);
}

/*-------------------------------------------------------------*/
/*      get_clause_stats()         .............. #27
/*
/*      determine the number of chars, vars, and lits
/*      in clause c
/*
/*      called by:     write_out_results    tgtp.c
/*
/*      calls:         nobody
/*
/*      returns:       nil
/*
/*--------------------------------------------------------------*/

void get_clause_stats(c)
clause *c;
{
literal *l;
long lit_length;

if(c->c_nlit > maxl) maxl = c->c_nlit;
if(c->c_nv   > maxv) maxv = c->c_nv;
for(l = c->c_lit ; l != 0L ; l = l->l_next) {
   if(l->l_next != 0L) {
      lit_length = ((long) l->l_next - (long) l)/4L - 4L;}
   else {
      lit_length = ((long) c->c_end - (long) l)/4L - 4L;}
   if(lit_length > maxx) {
      maxx = lit_length;}}

}

 
/*------------------------------------------------------------*/
/*      write_out_results()         .............. #28
/*
/*      write_out_results writes out the cl's that form the proof
/*      (if a proof found) and write statistics out. it writes
/*      the results on the screen and to a file "in-theorem.u".
/*
/*      called by:      main()          tgtp.c
/*
/*      calls:          print_clause()  tdump.c
/*
/*      returns:        nil
/*---------------------------------------------------------------*/
 
void write_out_results(proof_found)
int proof_found;
{
clause *c, *ci;
long i, k;         
long total_search_time;     /* total search time                 */
long good_inferred_cls;     /* # inferred clauses used in proof  */
long yes_inferred_cls;
long t_temp, t1_temp, t_power;
char t1_char;
long leading_zeros;
 
/*--------------------------------------------------*/
/* set screen printing to print to a depth of 1000. */
/*--------------------------------------------------*/

o_print = 1000; o_step = -1; o_debug = 0; 
maxx = maxl = maxv = 0L;
	
/*----------------------------------------------------------------*/
/* determine which of the inferred clauses were used in the proof.*/  
/* fix c->c_bits by adding LIST bit for inferred clauses that     */ 
/* should be printed. for each clause in the proof found in phases*/
/* 1,2, check LIST bit of parents                                 */
/*----------------------------------------------------------------*/
	
for(i = 1L ; i <= ply ; i++) {
   c = clist[depth[i].d_tc];
   ci = c->c_lpar;
   ci->c_bits |= LIST;  /* check left par of clist[depth[i].d_tc  */
   if(c->c_rpar != 0L) {           /* check if c is a factor      */
      ci = c->c_rpar;
      ci->c_bits |= LIST;}}/* check rt par of clist[depth[i].d_tc */
			
/*-------------------------------------------------------------*/
/* now check LIST bit of other inferred cl's to det parents    */
/*-------------------------------------------------------------*/

yes_inferred_cls = 0L; 
for(i = base ; i > org_base ; i--)  {
   c = clist[i];
   if(c->c_nlit == 0) c->c_bits |= LIST; /* null cls is inc */
   if((c->c_bits & LIST) == LIST) {
      yes_inferred_cls++;
      ci = c->c_lpar;
      ci->c_bits |= LIST;
      ci = c->c_rpar;
      ci->c_bits |= LIST;}}
			 
good_inferred_cls = 0L;       /* count # of inferred cls used   */

/*--------------------------------------------------------------*/
/* if there are axioms, print "axioms". otherwise print         */
/* "negated conclusion".                                        */
/*--------------------------------------------------------------*/
 
if(org_base + 1L - nc != 0L ) {
   /*printf("\n\nAxioms:\n");*/}
else {
   /*printf("\nNegated conclusion:\n");*/}

/*--------------------------------------------------------------*/
/* write out the base clauses and the clauses used from phase 0 */
/*--------------------------------------------------------------*/

tab_amount = 0L;
prp = 32L;

for(i = 0L; i <= base; i++) {
   c = clist[i];
   if(i <= org_base) {
      if(o_relevant == 0) {
         if((c->c_bits & LIST) == LIST) {
            get_clause_stats(c);}
         print_clause(c);}
      else if((c->c_bits & LIST) == LIST) {
         get_clause_stats(c);
         print_clause(c);}}             
   else {

         get_clause_stats(c);
         print_clause(c); 
	 Proofresults[prp++] = c->c_lpar->c_id;
	 if(c->c_rpar != 0L) {
	    Proofresults[prp++] = c->c_rpar->c_id;}
         else {Proofresults[prp++] = 0L;}
	 Proofresults[prp++] = c->c_indices;
	good_inferred_cls++;}
	
   if((i == org_base) && (org_base != base)) {
      if(yes_inferred_cls != 0L) {
         /*printf("\nClauses from phase 0 used in proof:\n");*/}}
   if(i == (org_base - nc))  { 
      /*printf("\nNegated conclusion:\n");*/}} 
	
/*----------------------------------*/
/* write out the clauses from       */
/* phases 1 and 2 used in proof     */
/* if proof not found in phase 0    */
/*----------------------------------*/
	
if(clist[base]->c_nlit != 0 && ply != 0L) {
   /*printf("\nClauses from phases 1 and 2 used in proof:\n");*/
   for(i = 1L;i <= ply ; i++)  {
      c = clist[depth[i].d_tc];
      if(c->c_lpar->c_id != c->c_id - 1L && i != 1L) {
         /*printf("\n");*/}
      get_clause_stats(c);
      Proofresults[prp++] = c->c_lpar->c_id;
      if(c->c_rpar != 0L) {
         Proofresults[prp++] = c->c_rpar->c_id;}
      else {Proofresults[prp++] = 0L;}
      Proofresults[prp++] = c->c_indices;
      print_clause(c);}} 

Proofresults[prp] = 0L;
total_search_time = search_time - search_start_time; 

if(c->c_nlit == 0L) {
   Proofresults[0] = good_inferred_cls +  ply;
   Proofresults[1] = eliminate_time;
	
   if(simple_sol == 0) {
      Proofresults[2] = phase1_time;}
   else {
      Proofresults[2] = -1L;}
	
   if(extra_pass == 1) { 
      Proofresults[3] = total_search_time - phase1_time;}
   else {
      Proofresults[3] = -1L;}

   if(simple_sol == 0) {
      Proofresults[4] = eliminate_time + total_search_time;}
   else {
      Proofresults[4] = -1L;}

   Proofresults[5] = s_nod;  
   Proofresults[6] = s_res;
   Proofresults[7] = s_fac;       
   Proofresults[8] = o_maxterm;
   Proofresults[9] = o_maxvar;
   Proofresults[10] = o_maxlit;        
   Proofresults[11] = s_htentry_phase1;  
   Proofresults[12] = s_hthit_phase1;      
   Proofresults[13] = s_htfail_phase1;      
   Proofresults[14] = s_ltentry_phase1;      
   Proofresults[15] = s_lthit_phase1;     
   Proofresults[16] = s_ltfail_phase1;      
   Proofresults[17] = maxx;
   Proofresults[18] = maxv;
   Proofresults[19] = maxl;
   Proofresults[20] = base;
   Proofresults[21] = clauses_alive;
   Proofresults[22] = good_inferred_cls;
   Proofresults[23] = i;
   Proofresults[24] = o_elim;
   Proofresults[25] = o_merge;
   Proofresults[26] = o_nc;
   Proofresults[27] = o_hash;
   Proofresults[28] = o_extend;
   Proofresults[29] = o_subsume;
   Proofresults[30] = o_maxtime;
   Proofresults[31] = theorem_number;

   pvm_initsend(PvmDataDefault);
   pvm_pklong(Proofresults,2000,1);
   pvm_send(myparent, MPROOF);}

}
 
/*-----------------------------------------------------------------*/
/*      main()                             .............. #29
/*
/*      The theorem prover enters here. We set the options specified
/*      by the  user, read in the clause system, prove the theorem,
/*      and write the proof out.
/*
/*      calls:   fatal()                        tgtp.c
/*               init_options()                 tgtp.c
/*               compile_clauses()              affc.c
/*               init_cl_array()                pass2.c
/*               init_clause_hash_table()       tgtp.c
/*               init_clause_space()            tgtp.c
/*               init_stats()                   tgtp.c
/*               prepare_clauses()              tgtp.c
/*               init_random_hash_array()       tgtp.c
/*               hash_all_base_clauses()        tgtp.c
/*          (PHASE 0)
/*               eliminate()                    elim.c
/*          (PHASE 1)
/*               init_stats()                   tgtp.c
/*               prove_by_iterative_deepening() tgtp.c
/*               stack_hash_codes()             pass2.c
/*          (PHASE 2)
/*               prove_by_iterative_deepening() tgtp.c
/*               stack_hash_codes()             pass2.c
/*               rearrange_cl_array()           pass2.c
/*               search_proof_line()            pass2.c
/*          (PHASE 3)
/*               write_out_results()            tgtp.c
/*-----------------------------------------------------------------*/
 
void main()
{
int zz;
int proof_found, p;
char decision_char;   /* char to decide whether to do another thm  */
long i;
int ii;

/*------------------------------------*/
/* constants associated with pvm      */
/* functions, etc.                    */
/*------------------------------------*/

theorem_number = 0;
mytid = pvm_mytid();

myparent = pvm_parent();

gethostname(buf, 64);
pvm_initsend(PvmDataDefault);
pvm_pkstr(buf);
pvm_send(myparent, CNAME);


o_short = 0;
o_noprint = 0;
o_cade = 0;

begin_proof:

proof_found = 0;
magic_hashx = 0L;
magic_hashy = 0L;
extra_pass  = 0;
phase       = 0;
iter_depth  = 0L;
equal_pred  = 0L;
equal_identity = 0L;
slave_quit = 0;

for(p = 0L; p < 2000L ; p++) {Proofresults[p] = 0L;}

/*------------------------------------------------------*/
/* attempt to determine theorem's name                  */
/*------------------------------------------------------*/
 
enter_theorem_name(); 
 
/*--------------------------------------------------*/
/* allocate space for data structures in affc.c     */
/*                                                  */
/* read in clauses and format for theorem prover    */
/* This is done in the function compile_clauses()   */
/* which is found in affc.c.                        */
/*                                                  */
/* free space used by compile; close input file     */
/*                                                  */
/*--------------------------------------------------*/
 
get_clauses();
 
/*--------------------------------------------------*/
/* init arrays for phase 0, then fix up clauses     */
/*--------------------------------------------------*/
 
init_cl_array();          /* init cl list in pass2.c         */
init_stats();             /* zero our statistics variables   */
init_search_space();      /* init search space               */

prepare_clauses();
 
/*---------------------------------------------------*/
/* initialize the random_hash_array and then hash    */
/* all base clauses                                  */
/*---------------------------------------------------*/
 
init_random_hash_array(o_maxsterm,number_fxns);
hash_all_base_clauses();
 
/*----------------------------------------------------*/
/* PHASE 0                                            */
/* -------                                            */
/* simplify input clauses if possible.                */
/* the array elim_inferences stores clauses generated */
/* in elim.                                           */
/*----------------------------------------------------*/
 
in_elim = 0;
if(o_elim != 0) {
   simple_sol = 0;
   eliminate();}
 
in_elim = 1;
		
/*----------------------------------------*/
/* if proof found, don't carry out search */
/*----------------------------------------*/
 
if(o_elim == 2) simple_sol = 1;
	
/*----------------------------------------------------*/
/* PHASE 1                                            */
/* -------                                            */
/* begin searching for a proof.                       */
/*----------------------------------------------------*/

phase = 1;      
ply = 0L;
lit_in_hash = 0L;
	
search_start_time = time_now(); /* start time for search  */ 
 
init_stats();           /* zero our statistics variables  */
	
if(simple_sol == 0) {
   proof_found = prove_by_iterative_deepening();
   if(proof_found == 4) {
      o_maxlit++;
      o_maxterm = o_maxterm * 2;
      if(o_maxterm > 50) {o_maxterm = 50;}
      proof_found = prove_by_iterative_deepening();}
   if(proof_found == 1) {
      stack_hash_codes();}
   phase1_time = search_time - search_start_time;
   s_ltentry_phase1  = s_ltentry;
   s_htentry_phase1  = s_htentry;
   s_hthit_phase1    = s_hthit;
   s_lthit_phase1    = s_lthit;
   s_htfail_phase1   = s_htfail;
   s_ltfail_phase1   = s_ltfail;
   s_htelim_phase1   = s_htelim;
   s_subsume_phase1  = s_subsume;
   s_unit_res_phase1 = s_unit_res;
   s_unit_sub_phase1 = s_unit_sub;
   s_ltentry = s_htentry = s_hthit = s_lthit = 0L;
   s_htfail = s_ltfail = s_subsume = s_unit_res = s_unit_sub = 0L;
   s_htelim = 0L;
	
/*----------------------------------------------------*/
/* PHASE 2                                            */
/* -------                                            */
/* search for hash codes on hash_code_list.           */
/*----------------------------------------------------*/
 
   phase = 2;
	
   while(hash_code_list_active != 0L && proof_found < 2) {
      ply = 0L; 
      extra_pass = 1;

      Proofresults[31] = theorem_number;
      Proofresults[0] = -1L;
      pvm_initsend(PvmDataDefault);
      pvm_pklong(Proofresults,2000,1);
      pvm_send(myparent, MPROOF);

      proof_found = prove_by_iterative_deepening();
      stack_hash_codes();}}

/*----------------------------------------------------*/
/* PHASE 3                                            */
/* -------                                            */
/* check proof.  if proof is  ok, then print results  */
/*----------------------------------------------------*/

phase = 3;
if(simple_sol == 0 && proof_found < 2 && slave_quit == 0) {
   rearrange_cl_array();
   search_proof_line();}

/*-----------------------------------------------------*/
/* if no proof found, don't proceed further until      */
/* a message is receive from the master                */
/*-----------------------------------------------------*/

if(slave_quit == 0) {
write_out_results(proof_found);}
	
/*----------------------------------------------*/
/* free allocated space and close output file   */
/*----------------------------------------------*/
 
free_space_close_files();
 
/*----------------------------------------------*/
/* prove another theorem?                       */
/*----------------------------------------------*/
 
another_theorem:
ii = 1;
while(ii < 2) {ii = ii;}
}
