/*---------------------------------------------------------------*/
/*      unify.c
/*
/*      Copyright Newborn Software 1996
/*
/*      functions:
/*         occurs()         ....................... #1
/*         unify()          ....................... #2
/*         pop_pstack()     ....................... #3
/*
/*      function linkage:
/*         unify()
/*            occurs()
/*            pop_stack()
/*            
/*---------------------------------------------------------------*/
 
 
#include "const.h"
#include "clause.h"
 
extern long *var_substitution_tbl[];
extern long pstack[], *psp;

/*----------------------------------------------*/
/* internally defined functions                 */
/*----------------------------------------------*/

long *occurs();
int unify();
int pop_stack();
 
/*----------------------------------------------*/
/* internally defined constants                 */
/*----------------------------------------------*/

int fxn_found;
int loop;
 
long occ_stack[200]; 
int occ_bottom;
 
/*---------------------------------------------------------------*/ 
/*
/*      *occurs(varbl,ffxn)             ....................... #1
/*
/*      occurs will see if variable varbl occurs in a given
/*      function ffxn. we return a pointer past the given function.
/*      We assume that a subst'tion has been made thus 
/*      var_substitution_tbl[varbl]!=NULL.
/*      Therefore varbl should not appear ANYWHERE in ffxn.
/*      We return NULL if the variable occurs
/*
/*      called by:  unify()           unify.c
/*
/*      calls:      nobody
/*
/*      returns     ptr to fxn or 0
/*
/*---------------------------------------------------------------*/
 
long *occurs(varbl,ffxn)
long varbl, *ffxn;
{
register long i, sf, *sv;
int occ_ptr;
 
occ_stack[occ_bottom++] = *ffxn;
 
for(i = 0L; i >= 0L; i--) {
   if(((sf = *ffxn) & FCNMASK) == 0L) {
 
/*------------------------------------------------------------------*/
/*	two variables are under consideration.                          */
/*------------------------------------------------------------------*/
 
      if(sf == varbl) {         /* if variables are the same, then: */
         loop++;                /* record fact by incrementing loop */
         return(0L);}               /* and then return to unify     */
			
      if(occ_bottom > 2) {
         for(occ_ptr = 1; occ_ptr < occ_bottom - 2; occ_ptr++) {
            if(sf == occ_stack[occ_ptr]) { 
               loop++;  
               return(0L);}}}
					
      if((sv=var_substitution_tbl[sf]) != 0L && occurs(varbl,sv)==0L) {
         return(0L);}}
   else {
     
/*-------------------------------------------------------------------*/
/* *f is a function. the number of arguments are determined and added*/
/* to the parameter i. fxn_found is incre'd to record the fact that a*/
/* function was encountered.                                         */
/*-------------------------------------------------------------------*/
      
      i += (sf & ARGMASK);   /* *f points to a function; det # args  */
      fxn_found++;}           /* fxn_found++ => function found       */
   ffxn++;}
return(ffxn);
}
 
/*-----------------------------------------------------------------*/
/*
/*      unify(l1,l2)                 ....................... #2
/*
/*      attempt to unify l1 and l2
/*
/*      called by:  binary_factors()        infer.c
/*                  binary_resolutions()    infer.c
/*                  bres_ply_1()            elim.c
/*                  bres_link()             elim.c
/*                  bres_replace()          elim.c
/*                  bres_proof_line()       pass2.c
/*                  factor_proof_line()     pass2.c
/*
/*      calls:      pop_pstack()            unify.c
/*                  occurs()                unify.c
/*
/*      returns:    0 if l1,l2 are not unifiable
/*                  1 if l1,l2 are unifiable
/*
/*-----------------------------------------------------------------*/
 
int unify(l1,l2)                       /* find mgu of l1 and l2  */
long *l1,*l2;
{
long i;
register long sl1, sl2, *l3;
	
psp = pstack;            /* psp is ptr to top position on stack  */
     	
i = *l1++ & ARGMASK; l2++;    /* set  i = # of arguments in pred */
	
/*------------------------------------------------------*/
/* push first arg of literal l1, ptr to it, on stack    */
/* push first arg of literal l2, ptr to it, on stack    */
/* push # of arguments onto stack, ptr to it            */
/* save a "3" to indicate 3 parameters stored on stack  */
/*------------------------------------------------------*/
	
*psp++ = (long) l1; *psp++ = (long) &l1;
*psp++ = (long) l2; *psp++ = (long) &l2;
*psp++ = (long) i;  *psp++ = (long) &i;
*psp++ = 3L;
 
while(pop_pstack()) {           /* until stack is not empty        */
   while(i > 0L) {           /* while the number of arguments > 0  */
      sl1 = *l1; sl2 = *l2;           /* ptrs to first arguments   */
      loop = 0;
      occ_bottom = 1; 
      if((sl1 & FCNMASK) != 0L) {
         if((sl2 & FCNMASK) != 0L) {                   /* FCN FCN  */
            if(sl1 != sl2) {
               return(0);}                /* can't unify 2 != fxns */
            i += sl1 & ARGMASK;               /* i = i+#args(sl1)  */
            l1++; l2++; i--;            /* inc ptrs to next arg    */
            continue;}		
  	
         if(var_substitution_tbl[sl2] == 0L) {/* FCN VAR (virgin)  */
            var_substitution_tbl[sl2] = l1;     /* replace var     */
            l1 = occurs(sl2,l1);
            if(l1 == 0L) {
               return(0);}
            l2++;i--;
            continue;}                                         
											
         i--; l2++;              			
											
/*------------------------------------------------------*/
/* push first arg of  l2, ptr to it, on stack           */
/* push # of arguments onto stack, ptr to it            */
/* save a "2" to indicate 2 parameters stored on stack  */
/*------------------------------------------------------*/
 
         *psp++ = (long) l2; *psp++ = (long) &l2;
         *psp++ = (long) i;  *psp++ = (long) &i;
         *psp++ = 2L;
         l2 = var_substitution_tbl[sl2]; i = 1L;         		
         continue;}
 
      if((sl2 & FCNMASK) != 0L) {
         if(var_substitution_tbl[sl1] == 0L){/* VAR (virgin) FCN */
            var_substitution_tbl[sl1] = l2;   /* replace var     */
            l2 = occurs(sl1,l2);
            if(l2 == 0L) {
               return(0);}
            l1++;i--;
            continue;}
                                            /* VAR (subst) FCN   */
         i--; l1++;               	      	/* see the subst.    */
 
/*------------------------------------------------------*/
/* push first arg of  l1, ptr to it, on stack           */
/* push # of arguments onto stack, ptr to it            */
/* save a "2" to indicate 2 parameters stored on stack  */
/*------------------------------------------------------*/
 
         *psp++ = (long) l1; *psp++ = (long) &l1;
         *psp++ = (long) i;  *psp++ = (long) &i;
         *psp++ = 2L;
         l1 = var_substitution_tbl[sl1]; i = 1L;
         continue;}
 
      if(sl1 == sl2) {               /* Two identical variables? */
         l1++;l2++;i--;                    /* yes: get next args */
         continue;}
        	
      if(var_substitution_tbl[sl1] == 0L) {  /* VAR (virgin) VAR */
         var_substitution_tbl[sl1] = l2;       /* replace var    */
         fxn_found = 0;
         l3 = occurs(sl1,l2);
         if(l3 == 0L && fxn_found != 0)  {
            return(0);}
         if(l3 == 0L && loop != 0) {
            var_substitution_tbl[sl1] = 0L;}
         l1++;l2++;i--;
         continue;}
          	
      if(var_substitution_tbl[sl2] == 0L) {  /* VAR VAR (virgin) */
         var_substitution_tbl[sl2] = l1;
         fxn_found = 0;
         l3 = occurs(sl2,l1);
         if(l3 == 0L && fxn_found != 0)  {
            return(0);}
         if(l3 == 0L && loop != 0) {
            var_substitution_tbl[sl2] = 0L;}
         l1++;l2++;i--;
         continue;}
           	
      l1++; l2++; i--;               /* see both subst.s   */
 
/*------------------------------------------------------*/
/* push first arg of  l1, ptr to it, on stack           */
/* push first arg of  l2, ptr to it, on stack           */
/* push # of arguments onto stack, ptr to it            */
/* save a "3" to indicate 3 parameters stored on stack  */
/*------------------------------------------------------*/
 
      *psp++ = (long) l1; *psp++ = (long) &l1;
      *psp++ = (long) l2; *psp++ = (long) &l2;
      *psp++ = (long) i;  *psp++ = (long) &i;
      *psp++ = 3L;
      l1 = var_substitution_tbl[sl1]; 
      l2 = var_substitution_tbl[sl2]; 
      i = 1L;}}
 
return(1);
}
 
/*------------------------------------------------------------------*/
/*
/*      pop_pstack()                ....................... #3
/*
/*      parameters:   l1,l2 : pointers to literals
/*
/*      called by:    unify()        unify.c
/*
/*      calls:        nobody
/*
/*      returns:      TRUE if unifier found(var_subs_tbl contains mgu)
/*
/*------------------------------------------------------------------*/
 
int pop_pstack()
{
register long pop_number;      /* pop_number = number of par to pop  */
register long *a;              /* pointer to element to pop          */
 
if(psp == pstack) {return(0);}                   /* nothing to pop   */
      
for(pop_number = *(--psp); pop_number > 0L; pop_number--) {
   a  = (long *) *(--psp);
   *a = *(--psp);}
return(1);
}
 
 
