(proclaim '(special *otter-results*))

(setf *time-limit* 1.5)
(setf *reductio-interest* .23)
(setf *reductio-discount* .23)
(setf *deductive-only* t)

(def-backwards-reason contra-conditionalization
    :conclusions "(p -> q)"
    :condition (or (and (not (occur 'equal p))
                                   (not (occur 'equal q))
                                   (not (u-genp p))
                                   (not (conjunctionp (quantifiers-matrix p)))
                                   (e-genp q))
                           (and (literal p)
                                    (literal q)
                                    (or (negationp p) (negationp q))))
    :backwards-premises "~p"
    :discharge "~q"
    :variables  p q)

(def-backwards-reason conditionalization
    :conclusions "(p -> q)"
    :condition (or (occur 'equal p)
                           (occur 'equal q)
                           (u-genp p)
                           (and (literal p) (not (e-genp q)))
                           (conjunctionp (quantifiers-matrix p))
                           (not (e-genp q)))
    :backwards-premises "q"
    :discharge "p"
    :variables  p q)

(setf *backwards-logical-reasons*
         (list adjunction neg-intro i-neg-disj i-neg-condit i-neg-bicondit
                 bicondit-intro disj-cond i-DM conditionalization
                 i-neg-ug i-neg-eg EG UG
                 disj-cond-2
                 contra-conditionalization
                ; vacuous-condition
                 ))

(defunction test (&rest rest)
    (terpri) (princ "(") (princ "                                 ") (princ *version*) (princ "          ")
    (let ((time (multiple-value-list (get-decoded-time))))
       (princ (mem5 time)) (princ "/") (princ (mem4 time)) (princ "/")
       (princ (mem6 time)) (princ "          ") (princ (mem3 time))
       (princ ":") (if (< (mem2 time) 10) (princ "0")) (princ (mem2 time))
       (princ ":") (if (< (mem2 time) 10) (princ "0")) (princ (mem1 time))
       (terpri) (terpri))
    (princ "*reductio-discount* = ") (princ *reductio-discount*) (terpri)
    (princ "*reductio-interest* = ") (princ *reductio-interest*) (terpri)
    (princ "*skolem-multiplier* = ") (princ *skolem-multiplier*) (terpri)
    (princ "*quantifier-discount* = ") (princ *quantifier-discount*) (terpri)
    (setf *test-log* nil)
    (cond ((null rest)
                 ;; (test)
                 (dolist (P *problems*)
                     (run-reasoning-problem P)
                     (when *pause*
                          (terpri) (princ "Type any key to continue.") (terpri) (terpri)
                          (read-char))))
                ((equal (mem1 rest) :skip)
                  ;; (test :skip i j k)
                 (dolist (P *problems*)
                     (when (not (mem (car P) (cdr rest)))
                          (run-reasoning-problem P)
                          (when *pause*
                               (terpri) (princ "Type any key to continue.") (terpri) (terpri)
                               (read-char)))))
                (t (let ((start (mem1 rest)))
                      (cond
                        ((null (cdr rest))
                          ;; (test n)
                          (run-reasoning-problem (assoc start *problems*)))
                        ((eq (mem2 rest) t)
                          ;; (test n t)
                          (dolist (P (mem (assoc start *problems*) *problems*))
                              (run-reasoning-problem P)
                              (when *pause*
                                   (terpri) (princ "Type any key to continue.") (terpri) (terpri)
                                   (read-char))))
                        ((equal (mem2 rest) :skip)
                          ;; (test n :skip i j k)
                          (dolist (P (mem (assoc start *problems*) *problems*))
                              (when (not (mem (car P) (cdr rest)))
                                   (run-reasoning-problem P)
                                   (when *pause*
                                        (terpri) (princ "Type any key to continue.") (terpri) (terpri)
                                        (read-char)))))
                        ((numberp (mem2 rest))
                          ;; (test n m) and (test n m :skip i j k)
                          (dolist (P (mem (assoc start *problems*) *problems*))
                              (when (and (not (mem (car P) (cdddr rest))) (not (> (car P) (mem2 rest))))
                                   (run-reasoning-problem P)
                                   (when (and *pause* (not (= (car P) (mem2 rest))))
                                        (terpri) (princ "Type any key to continue.") (terpri) (terpri)
                                        (read-char)))))))))
    (setf *test-log*
          (list *version* *reductio-discount* *reductio-interest* *skolem-multiplier* *quantifier-discount*
                *test-log*))
    (when (not *display?*)
       (display-test-log)
       (display-test-log :not-equal t)
       (display-test-log :equal t))
    (terpri)
    (princ *test-log*)
    (terpri) (princ ")")
    (terpri))

(defunction run-reasoning-problem (P)
    (setf *problem-number* (car P))
    (setf *pretty-list* nil *string-symbols* nil)
    (terpri)
    (princ
      "******************************************************************************************")
    (terpri)
    (princ
      "******************************************************************************************")
    (terpri)
    (display-problem P)
    (gc)
    (COGITATE)
    (terpri))

(defunction display-test-log (&key equal not-equal)
    (progn
        (cond (equal
                    (princ "============== TEST RESULTS FOR PROBLEMS WITH EQUALITY =============="))
                  (not-equal
                    (princ "============== TEST RESULTS FOR PROBLEMS WITHOUT EQUALITY =============="))
                  (t
                    (princ "=========================== TEST RESULTS ===========================")))
        (terpri) (terpri)
        (princ "                                                                                ") (princ (mem1 *test-log*))
        (princ "                        Otter                   ratio of run times")
        (terpri)
        (princ "          *reductio-discount*:                                 ") (princ (mem2 *test-log*)) (terpri)
        (princ "          *reductio-interest*:                                   ") (princ (mem3 *test-log*)) (terpri)
        (princ "          *skolem-multiplier*:                                   ") (princ (mem4 *test-log*)) (terpri)
        (princ "          *quantifier-discount*:                                   ") (princ (mem5 *test-log*)) (terpri) (terpri))
    (let ((ratios nil)
            (oscar-fail nil)
            (otter-fail nil))
      (dolist (L (reverse (mem6 *test-log*)))
          (let* ((n (mem1 L))
                    (p-assoc (assoc n *problems*))
                    (L* (assoc (read-from-string (mem8 p-assoc)) *otter-results* :test 'equal)))
            (when
                 (or (and equal (occur 'equal p-assoc))
                       (and not-equal (not (occur 'equal p-assoc)))
                       (and (not equal) (not not-equal)))
                 (princ "#") (princ n) (princ "   ") (princ (mem8 p-assoc))
                 (terpri)
                 (princ "                                                                                ")
                 (cond ((zerop (mem3 L)) (princ "no proof") (push (cons n (mem8 p-assoc)) oscar-fail))
                           (t (display-run-time-in-seconds (mem2 L))))
                 (when L*
                      (cond ((equal (mem2 L*) 'mp) (princ "                            ") (princ (mem3 L*)) (princ "  sec"))
                                (t (princ "                            no proof") (push (cons n (mem8 p-assoc)) otter-fail)))
                      (cond ((and (not (zerop (mem3 L))) (equal (mem2 L*) 'mp))
                                  (let ((ratio (/ (* internal-time-units-per-second (mem3 L*)) (mem2 L))))
                                    (push ratio ratios)
                                    (princ "                            ")
                                    (princ
                                      (if (< (abs (- (mem2 L) (* internal-time-units-per-second (mem3 L*)))) 15) 1.0
                                           (float (/ (truncate (* 100 ratio)) 100))))))))
                 (terpri)
                 (princ "           cumulative  argument length:                 ") (princ (mem3 L))
                 (terpri)
                 (princ "           total number of inferences:                     ") (princ (mem4 L))
                 (terpri) (terpri))
            ; (terpri)
            ))
      (let* ((both-fail (intersection oscar-fail otter-fail :test 'equal))
                (only-oscar-got (setdifference otter-fail both-fail))
                (only-otter-got (setdifference oscar-fail both-fail)))
        (princ "There were ") (princ (length both-fail)) (Princ " problems that neither OSCAR nor Otter solved:")
        (terpri)
        (dolist (x both-fail) (princ " #") (princ (car x)) (princ " (") (princ (cdr x)) (princ ") ") (terpri)) (terpri)
        (princ "There were ") (princ (length only-oscar-got))
        (Princ " problems that OSCAR solved but Otter failed to solve:") (terpri)
        (dolist (x only-oscar-got) (princ " #") (princ (car x)) (princ " (") (princ (cdr x)) (princ ") ") (terpri)) (terpri)
        (princ "There were ") (princ (length only-otter-got))
        (Princ " problems that Otter solved but OSCAR failed to solve:") (terpri)
        (dolist (x only-otter-got) (princ " #") (princ (car x)) (princ " (") (princ (cdr x)) (princ ") ") (terpri))) (terpri)
      (when ratios
           (princ "Of the problems solved by both OSCAR and Otter, OSCAR was on the average (geometric average) ")
           (let ((average (expt (apply '* ratios) (/ 1 (length ratios)))))
             (princ (float (/ (truncate (* 10000 average)) 10000))))
           (princ " times faster than Otter.") (terpri))) (terpri) (terpri))


(setf *otter-results*
'(
(com003+1		sos			4.77)
(com003+2		sos			1.7)
(com003+3		sos			4.12)

(grp194+1		mp			1.5)

(lcl181+1		mp			.58)
(lcl230+1		mp			.62)

(mgt001+1		mp			1.25)
(mgt002+1		mp			.93)
(mgt003+1		mp			.33)
(mgt004+1		mp			.43)
(mgt005+1		max-time		0)
(mgt005+2		mp			17.05)
(mgt006+1		mp			1.00)
(mgt007+1		mp			1.23)
(mgt008+1		mp			1.02)
(mgt009+1		mp			1.02)
(mgt010+1		mp			1.17)
(mgt011+1		mp			1.18)
(mgt012+1		mp			1.18)
(mgt013+1		mp			1.40)
(mgt014+1		mp			1.4)
(mgt015+1		mp			1.02)
(mgt016+1		mp			1.02)
(mgt017+1		mp			1.05)
(mgt018+1		mp			1.07)
(mgt019+2		max-time		0)
(mgt020+1		mp			1.72)
(mgt021+1		mp			1.07)
(mgt022+1		mp			.92)
(mgt022+2		mp			.90)
(mgt023+1		mp			3.18)
(mgt023+2		max-time		0)
(mgt024+1		mp			1.93)
(mgt025+1		max-time		0	)
(mgt026+1		mp			1.83)
(mgt027+1		max-time		0)
(mgt028+1		mp			1.27)
(mgt029+1		max-time		0)
(mgt030+1		mp			1.85)
(mgt031+1		max-time		0)
(mgt032+2		mp			.8)
(mgt033+1		max-time		0)
(mgt033+2		max-time		0)
(mgt034+1		max-time		0)
(mgt034+2		max-time		0)
(mgt035+1		max-time		0)
(mgt035+2		max-time		0)
(mgt036+1		mp			1.02)
(mgt036+2		mp			.97)
(mgt036+3		sos			2.3)
(mgt037+1		max-time		0)
(mgt037+2		max-time		0)
(mgt038+1		sos			.78)
(mgt038+2		max-time		0)
(mgt039+1		mp			6.25)
(mgt039+2		max-time		0)
(mgt040+1		max-time		0)
(mgt040+2		max-time		0)
(mgt041+2		mp			.78)

(msc009+1		sos			.68)

(puz001+1		max-time		0)
(puz031+1		mp			1.52)

(set043+1		sos			2.73)
(set044+1		mp			.67)
(set045+1		sos			1.98)
(set046+1		sos			4.43)
(set047+1		mp			1.18)

(syn001+1		sos			.4)
(syn007+1.005			mp			1.7)
(syn036+1			max-time		0)
(syn036+1			max-time		0)
(syn040+1			mp			.57)
(syn041+1			mp			.5)
(syn044+1			mp			.7)
(syn045+1			mp			.58)
(syn046+1			mp			.55)
(syn047+1			mp			.67)
(syn048+1			mp			.5)
(syn049+1			mp			.55)
(syn050+1			mp			.62)
(syn051+1			mp			.63)
(syn052+1			mp			.62)
(syn053+1			mp			.58)
(syn054+1			mp			.77)
(syn055+1			mp			.68)
(syn056+1			mp			.9)
(syn057+1			mp			.72)
(syn058+1			mp			.7)
(syn059+1			mp			1.27)
(syn060+1			mp			.7)
(syn061+1			mp			.65)
(syn062+1			mp			.7)
(syn063+1			mp			.6)
(syn064+1			mp			.53)
(syn065+1			mp			.62)
(syn066+1			mp			.68)
(syn067+1			max-time		0)
(syn068+1			mp			.77)
(syn069+1			mp			.93)
(syn070+1			mp			.83)
(syn071+1			mp			.78)
(syn072+1			mp			.97)
(syn073+1			mp			.55)
(syn074+1			mp			1.40)
(syn075+1			mp			1.35)
(syn076+1			max-time		0)
(syn077+1			max-time		0)
(syn078+1			mp			.93)
(syn079+1			mp			.62)
(syn080+1			mp			.67)
(syn081+1			mp			.6)
(syn082+1			mp			.7)
(syn083+1			mp			.65)
(syn084+1			mp			1.05)
(syn315+1			mp			.62)
(syn316+1			max term depth	82.30)
(syn317+1			mp			.62)
(syn318+1			mp			.6)
(syn319+1			mp			.67)
(syn320+1			sos			.43)
(syn321+1			mp			.63)
(syn322+1			sos			.37)
(syn323+1			mp			.63)
(syn324+1			max-time		0)
(syn325+1			mp			.63)
(syn326+1			mp			.68)
(syn327+1			mp			.60)
(syn328+1			mp			1.3)
(syn329+1			sos			.42)
(syn330+1			max-time		0)
(syn331+1			mp			.67)
(syn332+1			mp			1.17)
(syn333+1			mp			.62)
(syn334+1			mp			1.28)
(syn335+1			max-time		0)
(syn336+1			mp			.57)
(syn337+1			sos			.45)
(syn338+1			mp			.55)
(syn339+1			mp			.50)
(syn340+1			mp			.52)
(syn341+1			mp			.53)
(syn342+1			sos			.38)
(syn343+1			mp			.55)
(syn344+1			sos			.48)
(syn345+1			mp			.75)
(syn346+1			mp			.62)
(syn347+1			mp			7.70)
(syn348+1			sos			10.07)
(syn349+1			mp			1.05)
(syn350+1			mp			.72)
(syn351+1			mp			.62)
(syn352+1			mp			.78)
(syn353+1			max-time		0)
(syn354+1			mp			.7)
(syn355+1			mp			.63)
(syn356+1			mp			.68)
(syn357+1			mp			.53)
(syn358+1			mp			.62)
(syn359+1			mp			.63)
(syn360+1			mp			.58)
(syn361+1			mp			.65)
(syn362+1			mp			.58)
(syn363+1			mp			.58)
(syn364+1			mp			.70)
(syn365+1			mp			1.15)
(syn366+1			mp			.65)
(syn367+1			mp			.63)
(syn368+1			mp			.55)
(syn369+1			mp			.58)
(syn370+1			mp			.52)
(syn371+1			mp			.58)
(syn372+1			mp			.95)
(syn373+1			mp			.63)
(syn374+1			mp			.78)
(syn375+1			mp			.73)
(syn376+1			mp			.92)   ;; rerun this
(syn377+1			mp			.72)
(syn378+1			mp			.55)
(syn379+1			mp			.58)
(syn380+1			mp			.63)
(syn381+1			mp			.67)
(syn382+1			mp			.58)
(syn383+1			mp			.57)
(syn384+1			mp			.88)
(syn385+1			mp			.53)
(syn386+1			mp			.65)
(syn387+1			mp			.52)
(syn388+1			mp			.83)
(syn389+1			mp			.47)
(syn390+1			mp			.52)
(syn391+1			mp			.63)
(syn392+1			mp			.60)
(syn393+1			mp			.75)
(syn394+1			mp			.60)
(syn395+1			mp			.6)
(syn396+1			mp			.57)
(syn397+1			mp			.53)
(syn398+1			mp			.58)
(syn399+1			mp			.65)
(syn400+1			mp			.50)
(syn401+1			mp			.52)
(syn402+1			mp			.52)
(syn403+1			mp			.63)
(syn404+1			mp			.55)
(syn405+1			mp			.58)
(syn406+1			mp			.60)
(syn407+1			mp			.62)
(syn408+1			mp			.53)
(syn409+1			mp			.57)
(syn410+1			mp			.55)
(syn411+1			mp			.60)
(syn412+1			sos			2.95)
(syn413+1			mp			.58)
(syn414+1			sos			.73)
(syn415+1			mp			25.03)
(syn416+1			mp			.52)
(syn417+1			max-time		0)))


(setf *test-log* '(OSCAR_3.22 0.25 0.25 10 0.95
 ((217 543 58 208) (216 6 5 7) (215 116 40 60) (214 60092 0 803) (213 36 19 20) (212 13 8 8) (211 40 22 22) (210 8 6 6) (209 23 13 15) (208 11 8 8)
  (207 26 17 17) (206 17 11 11) (205 12 8 8) (204 5 4 4) (203 14 9 9) (202 6 5 5) (201 11 4 4) (200 10 7 7) (199 21 14 14) (198 25 15 17)
  (197 13 7 10) (196 14 7 10) (195 15 10 10) (194 12 8 8) (193 6 0 3) (192 36 0 18) (191 17 12 12) (190 3 0 1) (189 5 4 5) (188 4 3 3) (187 5 3 3)
  (186 346 40 73) (185 33 9 19) (184 33 10 15) (183 72 10 36) (182 24 10 15) (181 52 21 26) (180 38 16 18) (179 26 11 14) (178 22 9 12)
  (177 46 24 31) (176 105385 0 5176) (175 63 30 45) (174 67 41 43) (173 37 19 23) (172 13 9 9) (171 17 9 11) (170 30 12 14) (169 25 13 14)
  (168 23 8 12) (167 32 16 20) (166 46 18 19) (165 262 24 116) (164 363 45 75) (163 19 11 11) (162 17 10 10) (161 31 18 18) (160 20 10 12)
  (159 36 16 16) (158 26 15 17) (157 6 4 4) (156 36 18 18) (155 19 12 14) (154 541 38 105) (153 60023 0 1021) (152 95712 0 1328) (151 490 25 75)
  (150 901 27 125) (149 60266 0 1144) (148 60790 0 1687) (147 1275 20 134) (146 68 12 22) (145 748 37 128) (144 60432 0 928) (143 27 9 12)
  (142 21 0 12) (141 35 12 15) (140 40 12 15) (139 30 10 13) (138 39 11 18) (137 43 0 19) (136 58 14 23) (135 107963 0 770) (134 2171 41 227)
  (133 71 21 23) (132 413052 0 1885) (131 77354 0 4159) (130 60810 0 1120) (129 32 0 15) (128 61433 0 1608) (127 65 18 24) (126 161 20 55)
  (125 33305 18 1370) (124 46487 40 1066) (123 96 16 30) (122 659 0 69) (121 29 14 14) (120 59 0 24) (119 279 28 68) (118 72 24 34) (117 37 20 22)
  (116 83726 0 578) (115 278 28 71) (114 125 40 53) (113 557 17 136) (112 70 23 37) (111 20 12 13) (110 39 9 29) (109 14 9 10) (108 86 34 52)
  (107 61151 0 2177) (106 60052 0 2038) (105 69883 0 1131) (104 61111 0 1596) (103 18 11 11) (102 17085 67 1468) (101 297 52 119) (100 130 38 61)
  (99 57 30 38) (98 29 17 22) (97 270 40 74) (96 64 28 34) (95 27 13 20) (94 26 12 14) (63 85 31 45) (92 26 19 20) (91 20 16 17) (90 27 13 17)
  (89 86 38 46) (88 32 22 23) (87 40 26 27) (86 66 38 44) (85 15 12 13) (84 56 32 35) (83 38 19 27) (82 5 0 3) (81 13 0 10) (80 30 13 16)
  (79 51 13 24) (78 23 8 12) (77 83 40 53) (76 17 7 9) (75 46 27 36) (74 28 0 19) (73 5 4 6) (72 17 11 14) (71 62046 0 2912) (70 3018 201 735)
  (69 2 0 1) (68 4 3 3) (67 73 40 46) (66 74 27 35) (65 27 18 19) (64 32 14 14) (63 13 8 8) (62 712 131 280) (61 61159 0 3779) (59 57 37 41)
  (58 60257 0 1272) (57 60338 0 1311) (56 60723 0 5198) (55 819 110 387) (54 60263 0 6227) (53 60012 0 2743) (52 60511 0 3824) (51 60283 0 3824)
  (50 60702 0 1207) (49 989 50 84) (48 153 48 79) (47 60269 0 4812) (46 62732 0 2156) (45 60571 0 4308) (44 60734 0 4260) (43 60114 0 4135)
  (42 60162 0 3854) (41 65 28 34) (40 60096 0 4222) (39 61 12 21) (38 63629 0 4346) (37 124 32 47) (36 61060 0 2848) (35 64435 0 1890)
  (34 60426 0 4084) (33 437 52 229) (32 342 37 178) (31 325 29 180) (30 60 26 31) (29 60 26 31) (28 313 38 185) (27 62240 0 1367) (26 61218 0 987)
  (25 1195 100 264) (24 1618 100 338) (23 1905 100 264) (22 1621 100 338) (21 1321 117 498) (20 1325 115 502) (19 2829 111 874) (18 2832 111 874)
  (17 1864 95 260) (16 1024 93 247) (15 1498 95 363) (14 972 94 264) (13 491 75 142) (12 56324 46 2120) (11 3591 46 724) (10 1754 94 411)
  (9 606 74 195) (8 458 74 157) (7 1892 129 330) (6 14 11 12) (5 19 11 15) (4 60118 0 3179) (3 1412 68 395) (2 222 44 139) (1 596 62 167))))
