/* launcher.c */

/* Exitcodes: >0  Nummer des Erfolgreichen Wettbewerbers
               0  Terminierung durch total failure oder limit
              -1  pvm-Fehler (keine pvm oder kein spawn)
              -2  kein conf-File
              -3  Signal
              -4  sonst. Fehler
*/

#include <pvm3.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <setjmp.h>
#include <signal.h>

#define FOUND_TAG 101
#define EXIT_TAG 102
#define ANZ_TASKS 50
#define MAXLINE 512
#define OPTLENGTH 30
#define MAXOPTS 50       /* im Text Array aendern! */
#define COMMANDLENGTH 512

static int tid_array[ANZ_TASKS];
static char line[MAXLINE];
static int suctask=0;
static int errnumber=-3;
static int killflag=1;
static int compet_tasks=4;
static int tasks_launched=0;

char* t_argv[MAXOPTS+1];
char t_host[OPTLENGTH];
char t_command[COMMANDLENGTH];

/******************************************************************************
*  sig_handler()
*  Zweck:   Signalbehandlungsroutine, schickt kill-Messages an die einzelnen
*	    Tasks (bei allen Signalen). 
*  Eingabe: Signalnummer
*  Ausgabe: -
******************************************************************************/
void sig_handler(sig)
int sig;
{
  int i;

  for (i=0; i<ANZ_TASKS; i++) {
    if (tid_array[i]) {
      pvm_kill(tid_array[i]);
      fprintf(stderr,"launcher: task %d killed.\n",i+1);
      }
    }
  pvm_exit();
  exit(errnumber);
}

/******************************************************************************
*  launch_next_task()
*  Zweck:   liest aus dem config-File die naechste Task und startet sie 
*  Eingabe: filename des config-Files, Listennummer der Task
*  Ausgabe: 1: task gestartet, 0: spawn-Fehler, -1: keine weitere task da
*           (tasks_launched geaendert)
******************************************************************************/
int launch_next_task(file,number)
FILE *file;
int number;
{
  int tid, info, j;
  char line[MAXLINE];
  char **arg;
  int spawnmode;
  
  if (NULL == fgets(line,MAXLINE,file)) {
    return(-1);
    }
  for (j=0; j<MAXOPTS; j++) {
    if (NULL==(t_argv[j]=(char*)malloc((OPTLENGTH+1)*sizeof(char)))) {
      fprintf(stderr, "launcher: Not enough memory.\n");
      pvm_exit();
      exit(-4);
      }
    sprintf(t_argv[j],"");
    }
  t_argv[MAXOPTS]=NULL;

  /* nicht schoen, MAXOPTS-1 Feldelemente von t_argv scannen */
  sscanf(line, "%s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s %s", t_host, t_command, \
           t_argv[0], t_argv[1], t_argv[2], t_argv[3], t_argv[4], \
           t_argv[5], t_argv[6], t_argv[7], t_argv[8], t_argv[9], \
           t_argv[10], t_argv[11], t_argv[12], t_argv[13], t_argv[14], \
           t_argv[15], t_argv[16], t_argv[17], t_argv[18], t_argv[19], \
           t_argv[20], t_argv[21], t_argv[22], t_argv[23], t_argv[24], \
           t_argv[25], t_argv[26], t_argv[27], t_argv[28], t_argv[29], \
           t_argv[30], t_argv[31], t_argv[32], t_argv[33], t_argv[34], \
           t_argv[35], t_argv[36], t_argv[37], t_argv[38], t_argv[39], \
           t_argv[40], t_argv[41], t_argv[42], t_argv[43], t_argv[44], \
           t_argv[45], t_argv[46], t_argv[47], t_argv[48], t_argv[49] );
/*  fprintf(stderr, "launcher: task %d command %s ", i+1, t_command); */
  if (strcmp(t_host,"auto")==0) {
/*    fprintf(stderr, "host automatic_selection options "); */
    spawnmode=PvmTaskDefault;
    }
  else {
/*    fprintf(stderr, "host %s options", t_host); */
    spawnmode=PvmTaskHost;
    }
/*  for (arg=t_argv; *arg; arg++) fprintf(stderr, " %s", *arg);
  fprintf(stderr, "\n"); */
  info = pvm_spawn(t_command, t_argv, spawnmode, t_host, 1, &tid);
  if (info < 1) {
    fprintf(stderr,"launcher: error %d spawning task %d, retrying...\n", info, number+1); 
    sleep(1);
    info = pvm_spawn(t_command, t_argv, spawnmode, t_host, 1, &tid);
    if (info < 1) {
      fprintf(stderr,"launcher: error %d spawning task %d, giving up.\n", info, number+1); 
      tid_array[number] = 0;
      }
    }
  else {
    tid_array[number] = tid;
    tasks_launched++;
    fprintf(stderr,"launcher: task %d spawned, TID %x.\n", number+1, tid_array[number]); 
    pvm_notify(PvmTaskExit, EXIT_TAG, 1, tid_array+number);
    }
  for (j=0; j<MAXOPTS; j++) {
    free(t_argv[j]);
      }
  if (tid_array[number] == 0) {
    return(0);
    }
  else {
    return(1);
    }
}

/******************************************************************************
*  receive_message()
*  Zweck:   blocking warten auf ein Ergebnis einer task
*  Eingabe: -
*  Ausgabe: - (tasks_launched geaendert, FOUND_TAG mit killflag raeumt auf
*              und terminiert)
******************************************************************************/
void receive_message()
{
  int tid, bufid, tag, info, i, j;
  extern void sig_handler();

if (tasks_launched) {
  if (0 > (bufid = pvm_recv(-1, -1))) { 
    fprintf(stderr,"launcher: receive msg error %d.\n", bufid); 
    } 
  else {
    pvm_bufinfo(bufid, NULL, &tag, NULL);
    switch (tag) {
      case EXIT_TAG:
	tasks_launched--;
        if (0 > (info=pvm_upkint(&tid, 1, 1))) {
          fprintf(stderr, "launcher: error %d unpacking exit message.\n", info);
          break;
          } 
        for (i=0; i<ANZ_TASKS; i++) {
	  if (tid_array[i]==tid) {
            tid_array[i]=0;  
            fprintf(stderr, "launcher: task %d (tid %x) finished.\n", i+1, tid); 
            break;
	    }
	  }
	break;
      case FOUND_TAG:
        if (killflag == 0) {
          tasks_launched--;
          if (0 > (info=pvm_upkint(&tid, 1, 1))) {
            fprintf(stderr, "launcher: error %d unpacking exit message.\n", info);
            break;
            } 
          for (i=0; i<ANZ_TASKS; i++) {
	    if (tid_array[i]==tid) {
              tid_array[i]=0;  
              fprintf(stderr, "launcher: task %d (tid %x) successful (nokill).\n", i+1, tid); 
              break;
	      }
	    }
	  break;
	  }
        if (killflag != 0) {
          if (0 > (info=pvm_upkint(&tid, 1, 1))) {
            fprintf(stderr, "launcher: error %d unpacking exit message.\n", info);
            break;
            } 
          for (i=0; i<ANZ_TASKS; i++) {
            if (tid_array[i]==tid) {
	      tid_array[i]=0;  
              fprintf(stderr, "launcher: task %d (tid %x) successful.\n", i+1, tid);
              suctask=i+1; 
	      break;
	      }
	    }
          errnumber=suctask;
          sig_handler(2); /* Beweis gefunden, alles wird abgeschossen, dann exit */
	  break;
	  }
      default:
	fprintf(stderr, "launcher: Unexpected message tag %d.\n", tag);
      }
    pvm_freebuf(bufid);
    }
  }  
}

/******************************************************************************
*  main()
*  Zweck:   Programmrahmen, paralleler competitiver SETHEO-Aufruf mit pvm
*  Eingabe: -
*  Ausgabe: Fehlercode:  n: Nr. der erfolgreichen task +1
*                       -1 allg. Fehler 
*                       -2 keine pvm da     
******************************************************************************/

int main(argc, argv)
  int argc;
  char** argv;
{
  int mytid, i;
  extern void sig_handler();
  FILE *file;

  mytid = pvm_mytid(); 	/* PVM eingefuehrt */
  if (mytid < 0){  	/* PVM laeuft nicht */
    fprintf(stderr,"launcher: There is no PVM.\n");
    exit(-1);
  }
  fprintf(stderr,"launcher: PVM enrolled, TID: %x.\n", mytid); 

  signal(SIGTERM, sig_handler);
  signal(SIGINT, sig_handler);
  signal(SIGHUP, sig_handler);

  if (argc > 2) {
    sscanf(argv[2], "%d", &i);
    fprintf(stderr, "launcher: maximal %d tasks at the same time.\n", i); 
    if (i > 0) {
      compet_tasks = i;
      }
    }
  if (argc > 3) { /* hinter dem Problemnamen und der taskzahl steht "nokill" */
    killflag = 0;
    }
/*  file = fopen("psam.conf","r"); */
  file = fopen(argv[1],"r"); 
  if (file == NULL) {
    fprintf(stderr,"launcher: cannot open configuration file.\n");
    pvm_exit();
    exit(-2);
    }

  /* jetzt ist alles initialisiert, es muessen nun die Konfigurationen
     entsprechend den Vorgaben gestartet werden */
  i=0;
  while (i<ANZ_TASKS && launch_next_task(file,i) >= 0) {
    i++;
    if (tasks_launched >= compet_tasks) { /* jetzt haben wir die maximale 
                                             Zahl paralleler tasks, weiter
                                             gehts erst, wenn das weniger 
                                             werden */
      receive_message();
      }
    }
  fclose(file);
  if (0==tasks_launched) {
    pvm_exit();
    exit(-1);
    }

/* alles gestartet, nun warten wir (blocking) auf die restlichen Ergebnisse */

  while (tasks_launched) {
   receive_message();
   }

/* alle ohne Ergebnis terminiert */

  pvm_exit();
  exit(0);
}



