/*########################################################################
# File:			extern.h
# Author:		J. Schumann
# Date:			2.7.93
# Context:		inwasm (SETHEO)
# INWASMSCCS:		%W  07 Apr 1995
# SCCS        : @(#)extern.h	10.1 02 Apr 1996
# 
# Contents:		external definitions of global variables
#
# Modifications:	
#	(when)		(who)		(what)
#	12.7.93		jsc		grouping of global variables
#	14.12.93	jsc		purity, fold
#			hamdi		include ficonst.h
#	11.9.95		jsc		predicate_nr
#	3/12/96		jsc	moved to ~/include
#	3/12/96		jsc		wasm-less inwasm
#	3/26/96		jsc		does not include ficonst.h
#
# Bugs:
#
# --- Forschungsgruppe KI, Inst. fuer Informatik, TU Muenchen ---
########################################################################*/

#ifndef EXTERN_H
#define EXTERN_H

#include <stdio.h>
#include "symb.h"


/***************************************************************************/
/* local macros
/***************************************************************************/
	/* handle the initialization */
#ifndef extern
#	define INIT(X)
#else
#	define INIT(X) = X
#endif

/****************************************************************************
/* files
/***************************************************************************/
extern FILE *del_file;		/* deleted connection file    w */
extern FILE *code_file;		/* code output file    w */
extern FILE *lopout_file;	/* lop output file    w */
extern FILE *yyin;		/* lop input file      r */
extern FILE *constrout_file;    /* constr output file   w  */

/*****************************************************************************
/* matrix: intern representation of formula
/***************************************************************************/
extern modtype *matrix;		/* global, internal Structure	*/

extern t_fcb  fcb;


/***************************************************************************
/* globals for code generation
/***************************************************************************/
extern claustype *startclause INIT(NULL);

extern symbollist *redlist INIT(NULL);	/* list for reduction facts */
extern symbollist *cpylist INIT(NULL);	/* list for copy bounds	    */
extern int nrvars;				/* number of vars+strvars
						   in clause */ 
extern int nrstrvars;				/* number of strvars
						   in clause */ 
extern int failure_clause_needed INIT(0);

/***************************************************************************/
/* stuff for extended constraints
/***************************************************************************/
extern int constrnum;     /* cgen.c */
extern int nr_del_conn;   /* connopt.c */
extern int nr_links_tested;   /* connopt.c */

/***************************************************************************/
/* inwasm flags
/***************************************************************************/
extern inwasm_cmdline_flags	inwasm_flags;
extern float			sgreord_at; /* PJ opt params for sg reordering */ 
extern float			sgreord_av; /*    weight formula */
extern float			sgreord_ac;

/***************************************************************************/
/* Symbol table
/***************************************************************************/
extern syel *symbtab;		/*  pointer to symbol table 		*/
extern int nump	   INIT(0);	/* number of symbols in symbol table 	*/
extern int nrbuilt INIT(0);	/* number of built-ins in symbol table 	*/
extern int predicate_nr INIT(1);/* number of predicates in symbol table	*/
extern int max_arity_fsymb INIT(0);/* max. arity of function symbols    */

/***************************************************************************/
/* Counters
/***************************************************************************/
extern int maxunif	INIT(0);	/* maximum choicepoint	  */
extern int maxlit	INIT(0);	/* longest clause	  */
extern int orbcal	INIT(0);	/* counter for call or... */
extern int detcal	INIT(0);	/* counter for call l[np]... */
extern int nr_globals  INIT(0);		/* number of global variables */
extern int nr_constants INIT(0);	/* number of constants in formula */
extern int nr_functions INIT(0);	/* number of functions in formula */
extern int nr_numbers  INIT(0);		/* number of numbers in formula */
extern int nr_predicates  INIT(0);	/* number of real predicate symb's */
extern int nr_constraints  INIT(0);	/* number of constraints in formula */
extern int nr_clauses	INIT(0);	/* number of clauses in matrix */
extern int parsed_nr_clauses;           /* number of clauses in formula */

extern int nr_negated_clauses INIT(0);	  /* number of negative clauses */
extern int nr_definite_clauses INIT(0);	  /* number of definite clauses */
extern int nr_indefinite_clauses INIT(0); /* number of indefinite clauses */
extern int nr_facts  INIT(0);		  /* number of facts */
extern int nr_literals  INIT(0);	  /* number of literals */
extern int nr_connections  INIT(0);	  /* number of connections */
extern int maxlits  INIT(0);		  /* maximal clause length */
extern int minlits  INIT(0);		  /* minimal clause length */
extern float complexity INIT(0); 	  /* literal complexity */

extern int subs_unif_counter INIT(0); 
extern int subs_max_counter INIT(0);  /* in order to break subsumption */
/***************************************************************************/
/* etc
/***************************************************************************/
extern int ishorn	INIT(1);	/* is a horn formula		*/
extern int	opt_incount;		/* for hex_out-optimizer	*/
extern int	opt_outcount;		/* for hex_out-optimizer	*/

/***************************************************************************/
/* weak unification trail
/***************************************************************************/
extern trailelt * trail INIT(NULL);
extern int trailcnt;

/***************************************************************************/
/* errors
/***************************************************************************/
extern int errors	INIT(0);
/***************************************************************************/
/* scanner and parser
/***************************************************************************/
extern int yylineno;
extern int nofanning INIT(0);	/* there is at least one nofanning clause */
extern int loc_cpy INIT(0);	/* there is a local copybound */

/***************************************************************************/
/* subsumption and symmetry constraints
/***************************************************************************/
extern int nr_gen_constr INIT(0) ;

extern int no_subsumption;
extern int subsumption;
extern int is_subsumption;

extern int abbruch;
extern int var_counter INIT(0);          /* Zaehler zur Umbenennung von Var. */

extern sortarray *ssort;		/* sort.c 	*/

/***************************************************************************/
/* constraint subsumption
/***************************************************************************/

extern int Deleted; 
extern int Olddeleted; 
extern int Generated;

/***************************************************************************/
/* functions
/***************************************************************************/
extern symbollist *new_symboll();
extern char *get_name();
extern void enter_occ();
extern int genpurity();		        /* preproc/purity.c	*/

extern void genfanning();	        /* preproc/fanning.c	*/
extern void genreorder();	        /* preproc/reordering.c	*/
extern void gencode();		        /* codegen/cgen.c	*/
extern void enter_built_ins();
extern void init_trail();		/*   to initialize trail	  */
extern int unify();			/*   to unify uninstantiated terms  */
extern void rearrange();		/*   to clear the trail after unify  */

extern void enter_interface();	        /* make connection in submodules */
extern claustype *gen_s_constr();
extern claustype *gen_overl_constr();
extern claustype *gen_taut_constr();

/* extern functions for subsumption and symmetry constraint generation */
extern termtype *copy_terms();          /* hlpfkt.c 	*/
extern predtype *ErzeugeConstr();       /* hlpfkt.c 	*/
extern termtype *deref();               /* unify.c    	*/
extern int enter_name();                /* symb.c  	*/
extern int enter_token();               /* symb.c 	*/
extern int delete_p();                  /* purity.c 	*/
extern int del_occlist();               /* purity.c 	*/

extern int nr_bindings();	        /* constrsu.c   */
extern littype *make_cart_list();       /* substest.c   */
extern int clause_s_constr();           /* sconstr.c    */
extern void my_rename();                /* oconstr.c    */
extern symbollist *app_sym_to_list();   /* sconstr.c    */
extern void make_s_list();              /* sconstr.c    */
extern void make_sym_s_list();          /* symconstr.c  */

/* connopt.c */

extern predtype *get_red_pred();        /* oconstr.c */
extern termtype *replace_in_term2();    /* oconstr.c */
extern int FindeTautologie();           /* tconstr.c */

/***************************************************************************/
/* global counter for testing & newfunctions
/***************************************************************************/
extern void newrearrange();
extern termtype *newderef();
extern int newunify();
extern int newmakeref();
extern void high_rename();

extern void print_lop_term();

extern int globalcnt INIT(0);

#endif
