/****************************************************
;;; MODULE NAME : main.c
;;;
;;; PARENT      :
;;;
;;; PROJECT     : MPLOP
;;;
;;; SCCS        : @(#)main.c	15.2 01 Apr 1996
;;;
;;; AUTHOR      : Janos Breiteneicher
;;;
;;; DATE        : 15.5.91
;;;
;;; DESCRIPTION : main program
;;;
;;; REMARKS     :
;;;		22.2.93		jsc	new version
;;;		3.3.93		jsc	LOP output (fast hack)
;;;					-nocode generates file.lopout
;;;		4.3.93		jsc	mod'd functionality of
;;;					print_lop_formel,print_wunif
;;;					LOP output file named F.o.lop
;;;		8.3.93		jsc	output of execution times
;;;		11.3.93		jsc	usage
;;;		26.3.93		jsc	bug: _pp file
;;;						also to .._pp.lop
;;;		04.05.93	jsc	debugging of: ..
;;;		2.7.93		jsc	extern.h
;;;		21.7.93		jsc	clean-up
;;;		14.12.93	jsc	purity, fold
;;;		28.2.94		jsc	CLK_TCK
;;;		24.3.94		jsc	V3.2
;;;		28.7.94		jsc	-searchtree, -interactive
;;;		5.1.94		hamdi	constraints
;;;		9.2.94		hamdi	sg reordering
;;;		25.02.94	hamdi	linksubs,rsubs...
;;;		02.03.94	hamdi	rlinksubs
;;;             16.06.95	PJ	-sgreord, .lop suffix
;;;		11.9.95		jsc	fcb-mod'd
;;;		19.1.96		jsc	hex-code & clean-up
;;;
****************************************************/

#include <sys/types.h>
#include <sys/times.h>
#include <time.h>
#ifndef CLK_TCK
# define CLK_TCK 60
#endif
#include "types.h"
#include "extern.h"

/***************************************************/
/* XXXXXX
/***************************************************/
#define PF_NO		0  
#define PF_PARS		1 
#define PF_TCONSTR 	2 
#define PF_SCONSTR 	3 
#define PF_FIRSTDEL 	4 
#define PF_PUR 		5 
#define PF_FAN 		6 
#define PF_SYMCONSTR 	7 
#define PF_SECONDDEL 	8 
#define PF_OCONSTR 	9 
#define PF_THIRDDEL 	10 
#define PF_SGREORD 	11 
#define PF_CLREORD 	12 
#define PF_ALWAYS 	99 

/***************************************************/
/* statics
/***************************************************/

#ident "inwasm: @(#)main.c	15.2 01 Apr 1996"

static char logo[] = 
  "inwasm V3.3.1 [wasm-less] Copyright TU Munich (March 12, 1996) ";

static int debug_mode	    = 0;
static int overlap_mode     = 2;    /* maximal overlap-complexity */
static int deletion_mode    = 6;    /* maximal link-deletion-complexity */
				     /* test up to 5000 connections */


/*****************************************************************************/
/* main
/*****************************************************************************/
main(argc,argv)
int argc;
char *argv[];
{
/*
char filename_cdel[100];
*/
char filename_base[100];
char s[100];
struct tms      buffer;
int ti;
int rv;
int i;

/***************************************************/
/* start with logo 
/***************************************************/
fprintf(stderr,"%s\n",logo);

/***************************************************/
/* check arguments/options */
/***************************************************/
if (argc < 2) {
	usage(argv[0]);
 	exit(0);
	}
	
if ((argc == 2) && (!strcmp(argv[1],"-verbose"))) {
	verbose_usage(argv[0]);
  	exit(0);
  	}

if (*argv[argc-1] == '-') {
  fprintf(stderr,"please specify an input file.\n");
  exit(0);
  }
	
/***************************************************/
/* echo the command line */
/***************************************************/
fprintf(stderr,"command line: ");
for (rv=0; rv< argc;rv++){
	fprintf(stderr,"%s ",argv[rv]);
	}
fprintf(stderr,"\n");

/***************************************************/
/* get the beginning time	*/
/***************************************************/
times (&buffer);
ti = buffer.tms_utime;

/***************************************************/
/* set default flags
/***************************************************/
set_inwasm_defaults();

/***************************************************/
/* process arguments/options */
/***************************************************/
if (process_argv(argv,argc)){
	exit(1);
	}

/***************************************************/
/* check consistency of the arguments
/***************************************************/
if (check_argument_consistency()){
	exit(3);
	}

/***************************************************/
/* open input file for parsing */
/***************************************************/
if ((yyin = fopen(argv[argc-1],"r")) == NULL) {
	/* now, let's try if the file extension is
	/* missing */
	sprintf(s,"%s.lop",argv[argc-1]);
	if ((yyin = fopen(s,"r")) == NULL) {
        	printf("Could not open %s or %s.lop\n",
			argv[argc-1],argv[argc-1]);
		exit(2);
		}
	else {
		strcpy(filename_base,argv[argc-1]);
		}
	}
else {
	/* we have a filename: x.lop */
	/* strip off the .lop */
	strcpy(filename_base,argv[argc-1]);
	*(filename_base+strlen(argv[argc-1])-4)= '\0';
	}

/***************************************************/
/* open output file for code generation */
/***************************************************/
sprintf(s,"%s.%s",filename_base,
		  (inwasm_flags.hex_out)?"hex":"s");
if (!inwasm_flags.nocode && (code_file = fopen(s,"w")) == NULL) {
	printf("Could not open %s\n",s);
	exit(2);
	}

/***************************************************/
/* open del file for storing deleted connections */
/***************************************************/
#ifdef CDEL_FILE
sprintf(filename_cdel,"%s.cdel",filename_base);
if (inwasm_flags.linksubs && (del_file = fopen(filename_cdel,"w")) == NULL) {
	printf("Could not open %s\n",filename_cdel);
	exit(2);
	}
#endif

/***************************************************/
/* open _pp.lop file
/***************************************************/
/* if nocode is set, generate the lop-output */
if (inwasm_flags.nocode){
    sprintf(s,"%s_pp.lop",argv[argc-1]);
    if ((lopout_file = fopen(s,"w")) == NULL) {
	printf("Could not open %s.lopout\n",argv[argc-1]);
	exit(2);
	}
}

/***************************************************/
/* parse input, generate symboltable and matrix */
/***************************************************/
printf("Parsing \n");
if (yyparse()) {
	fprintf(stderr,"Too many errors. Cannot continue \n");
	exit(1);
	}

ishorn = ishorn && negative_goals(matrix->c_list);
parsed_nr_clauses = nr_clauses;

/***************************************************/
/* non horn formula => by default: reductionsteps */
if (!ishorn && !inwasm_flags.noreduct) {
	inwasm_flags.reduct = 1;
	inwasm_flags.noreduct = 0;
	}

/*******************************************/
/* non horn formula => by default: fanning */
if (!ishorn && !inwasm_flags.nofan) {
	inwasm_flags.fan = 1;
	inwasm_flags.nofan = 0;
	}

/******************************/
/* horn formula => no fanning */
if (ishorn && inwasm_flags.fan) {
	inwasm_flags.fan = 0;
	inwasm_flags.nofan = 1;
	}

/********************************************************************/
/* horn formula => no reduction steps exept for folding up and down */
if (ishorn && inwasm_flags.reduct) 
	if (!inwasm_flags.foldup && !inwasm_flags.foldupx &&
	    !inwasm_flags.folddown && !inwasm_flags.folddownx) {
	inwasm_flags.reduct = 0;
	inwasm_flags.noreduct = 1;
	}
if (ishorn && inwasm_flags.noreduct) 
	if (inwasm_flags.foldup || inwasm_flags.foldupx ||
	    inwasm_flags.folddown || inwasm_flags.folddownx) {
	inwasm_flags.reduct = 1;
	inwasm_flags.noreduct = 0;
	}


/*******************************/
/* some warnings after parsing */
	
if (!ishorn && inwasm_flags.noreduct)
	warning("","A NON-HORN-formula is processed with -noreduct option");
if (!ishorn && inwasm_flags.nofan)
	warning("","A NON-HORN-formula is processed with -nofan option");

if (nump > MAXPREDSYMB-1) {
	warning("","Too many predicate symbols");
	}

/*****************************/
/* generate weak unification */
genwunif(); 


/*********************************************/
/* print formula in lop syntax after parsing */
if (inwasm_flags.verbose == PF_PARS || 
    inwasm_flags.verbose == PF_ALWAYS) 
 print_lop_formel(matrix,stdout,1,0);

/*******************/
/* generate purity */
if (inwasm_flags.purity)  {  
	printf("Preprocessing: purity");
	rv = genpurity(matrix);
		/* rv <> 0 iff purity removed query */
	printf("\n");
	if (nr_clauses == 0 || rv) {
	   fprintf(stderr,"Purity removed all clauses from matrix\n");
	   fprintf(code_file,"fail\n");
	   exit(4);
	   }
        if (inwasm_flags.verbose == PF_PUR || 
	    inwasm_flags.verbose == PF_ALWAYS) 
         print_lop_formel(matrix,stdout,1,0);
	}


if (inwasm_flags.clausreord || inwasm_flags.linksubs || inwasm_flags.subs) 
   make_sortarray(matrix->c_list->next);

/*************************/
/*  orbranch reordering  */
if (inwasm_flags.clausreord) {  
	printf("Preprocessing: orbranch reordering ");
	orbranch_reorder(matrix->c_list);
	printf("\n");
        if (inwasm_flags.verbose == PF_CLREORD || 
	    inwasm_flags.verbose == PF_ALWAYS) 
	 print_wunif(matrix,stdout,0);
	}


/*=======================*/
/* tautology constraints */
if (inwasm_flags.tauto) {
        printf("Preprocessing: inserting tautology constraints\n");
        matrix->c_list = gen_taut_constr(matrix->c_list);
        if (inwasm_flags.verbose == PF_TCONSTR || 
	    inwasm_flags.verbose == PF_ALWAYS) 
         print_lop_formel(matrix,stdout,1,0);
        }

/*=========================*/
/* subsumption constraints */
if (inwasm_flags.subs)
 {
   printf("Preprocessing: inserting subsumption constraints\n");
   matrix->c_list = gen_s_constr(matrix->c_list,inwasm_flags.subs);
   if (inwasm_flags.verbose == PF_SCONSTR || 
       inwasm_flags.verbose == PF_ALWAYS) 
    print_lop_formel(matrix,stdout,1,0);
 }


/*======================================*/
/* constr. subsumption after tauto,subs */
if (inwasm_flags.tauto || inwasm_flags.subs) {
    printf("Preprocessing: removing redundant constraints\n");
    gen_constr_subs(matrix->c_list);
    if (inwasm_flags.verbose == PF_FIRSTDEL || 
	inwasm_flags.verbose == PF_ALWAYS) 
     print_lop_formel(matrix,stdout,1,0);
} 

/*======================*/
/* optimize connections */
if (inwasm_flags.linksubs) 
 {
    printf("Preprocessing: testing connections \n");
    opt_connections(matrix->c_list,inwasm_flags.linksubs);
    printf("   %d connections tested", nr_links_tested);
    printf(" and %d deleted\n",nr_del_conn);

    /*==========================================*/
    /* purtity after having deleted connections */
    if (nr_del_conn > 0) {
        printf("Preprocessing: purity again");
    	rv = genpurity(matrix);
		/* rv <> 0 iff purity removed query */
    	printf("\n");
	if (nr_clauses == 0 || rv) {
	   fprintf(stderr,"Purity removed all clauses from matrix\n");
	   fprintf(code_file,"fail\n");
	   exit(4);
	   }
    	if (inwasm_flags.verbose == PF_PUR || 
    	   inwasm_flags.verbose == PF_ALWAYS) 
         print_lop_formel(matrix,stdout,1,0);
    }
 } 

nr_clauses = parsed_nr_clauses;

/********************/
/* set up statistics */
set_fcb();

/********************/
/* generate fanning */
if (!inwasm_flags.nofan && !ishorn) {
	printf("Preprocessing: fanning   ");
	genfanning(matrix); 
	printf("\n");
        if (inwasm_flags.verbose == PF_FAN || 
	    inwasm_flags.verbose == PF_ALWAYS) 
         print_lop_formel(matrix,stdout,1,0);
	}

/*******************/
/* subgoal weights */
if (inwasm_flags.sgreord || inwasm_flags.overlap)
	gen_subgoal_weights(matrix->c_list);

/*******************************************/
/* tell that reduct instructions are added */
if (inwasm_flags.reduct) {
	printf("Preprocessing: inserting reduction steps\n");
	if (inwasm_flags.nocode && lopout_file){
		fprintf(lopout_file,"\n# inserting reduction steps\n");
		}
}

/*======================*/
/* symmetry constraints */
if (!ishorn && !inwasm_flags.nofan && inwasm_flags.subs) {
        printf("Preprocessing: inserting symmetry constraints\n");
        gen_sym_constr(matrix->c_list);
        if (inwasm_flags.verbose == PF_SYMCONSTR || 
	    inwasm_flags.verbose == PF_ALWAYS) 
         print_lop_formel(matrix,stdout,1,0);
        }


/*===========================*/
/* constr. subsumption again */
if (!ishorn && inwasm_flags.subs) {
           printf("Preprocessing: removing redundant constraints\n");
           gen_constr_subs(matrix->c_list);
           if (inwasm_flags.verbose == PF_SECONDDEL || 
	       inwasm_flags.verbose == PF_ALWAYS) 
            print_lop_formel(matrix,stdout,1,0);
}


/*=====================*/
/* overlap constraints */
if (inwasm_flags.overlap) {
        printf("Preprocessing: inserting overlap constraints\n");
        matrix->c_list = gen_overl_constr(matrix->c_list,overlap_mode,
                         (inwasm_flags.verbose == PF_OCONSTR || 
	                  inwasm_flags.verbose == PF_ALWAYS)? 1 : 0); 
        if (inwasm_flags.verbose == PF_OCONSTR || 
	    inwasm_flags.verbose == PF_ALWAYS) 
         {
	    printf("  Output after removing DELETED-clauses:\n");
            print_lop_formel(matrix,stdout,1,0);
	 }
        }


/*===========================*/
/* constr. subsumption again */
if (inwasm_flags.overlap) {
           printf("Preprocessing: removing redundant constraints\n");
           gen_constr_subs(matrix->c_list);
           if (inwasm_flags.verbose == PF_THIRDDEL || 
	       inwasm_flags.verbose == PF_ALWAYS) 
            print_lop_formel(matrix,stdout,1,0);
}

/*========================*/
/* generate sg reordering */
if (inwasm_flags.sgreord) {
          printf("Preprocessing: subgoal reordering\n");
          sg_reordering(matrix->c_list);
          if (inwasm_flags.verbose == PF_SGREORD || 
	      inwasm_flags.verbose == PF_ALWAYS) 
           print_lop_formel(matrix,stdout,1,0);
          }

/*****************/
/* generate code */
if (!inwasm_flags.nocode && !errors) {
	count_codenr();
	gencode(matrix);
	printf("Codegeneration.\n");
	}

/***************************************/
if (inwasm_flags.nocode && !errors){
	lop_output(matrix,lopout_file);
	}

/***************************************/
/* print information about compilation */
/*
if (debug_mode) {
	printf("debug = %d\n",debug_mode);
	printf("\tsymboltable : %d\n",1);
	printf("\tmatrix      : %d\n",2);
	printf("\toccurrences : %d\n",4);
	printf("\tweak unif   : %d\n",8);
	printf("\tsymboltable : %d\n",16);
	printf("\tlop-formel  : %d\n",32);
	printf("\tstatistic   : %d\n",64));
	}
if (DEB(1)) printsytab(nrbuilt); 
if (DEB(2)) print_matrix(matrix,stdout);
if (DEB(3)) print_occ(stdout);
if (DEB(4)) print_wunif(matrix,stdout,0);
if (DEB(5)) printsytab(0); 
if (DEB(6)) print_statistics(stdout,filename_lop);
if (DEB(7)) print_lop_formel(matrix,stdout,1,1);
*/

if (errors) {
   	fprintf(stderr,"\n %d errors found, no code generated\n",errors);
   	exit(1);
   	}


times (&buffer);
fprintf(stderr,
	"%s%s generated in %5.2f seconds\n",
	filename_base,
	(inwasm_flags.nocode)?"_pp.lop":
		((inwasm_flags.hex_out)?".hex":".s"),
	((float)((buffer.tms_utime - ti) * 1000 /
		(float)CLK_TCK))/1000.0);

exit(0);
}

/*######################################################################*/

/* Johann ... */

#define SET_MAX(W,V) if ((V)>(W)) W=(V)
/***************************************************************************/
/* set_fcb
/***************************************************************************/
set_fcb()
{
claustype *cl;

fcb.nr_clauses = parsed_nr_clauses;

fcb.nr_const =nr_constants;
fcb.nr_fsymb =nr_functions;
fcb.nr_predsymb=nr_predicates;
fcb.max_arity_fsymb=max_arity_fsymb;

for (cl=matrix->c_list;cl;cl=cl->next){
	fcb.nr_connections += cl->connections; 

	if ((cl->p_list->symb == 0) || (cl->parsed_rule_type == PQUERY)){
		/* we have a query */
		if (cl->nr_lits == 1) {
			/* this is the artificial query, */
			/* we don't count this		 */
			continue;
			}
		fcb.nr_queries++;

		SET_MAX(fcb.max_lit_queries,cl->nr_lits-1);
		fcb.sum_lit_queries +=cl->nr_lits-1;
		fcb.nr_literals += cl->nr_lits-1;
		SET_MAX(fcb.max_vars_queries,cl->nrvars);
		SET_MAX(fcb.max_termdepth_queries,cl->max_termdepth);
		SET_MAX(fcb.max_termsize_queries,cl->max_termsize);
		}
	else {
	    fcb.nr_literals += cl->nr_lits;
	    if (cl->nr_lits == 1){
		if (cl->poslits > 0){
			fcb.nr_punits++;
			SET_MAX(fcb.max_vars_punits,cl->nrvars);
			SET_MAX(fcb.max_termdepth_punits,cl->max_termdepth);
			SET_MAX(fcb.max_termsize_punits,cl->max_termsize);
			}
		else {
			fcb.nr_nunits++;
			SET_MAX(fcb.max_vars_nunits,cl->nrvars);
			SET_MAX(fcb.max_termdepth_nunits,cl->max_termdepth);
			SET_MAX(fcb.max_termsize_nunits,cl->max_termsize);
			}
		}
	else  {
		if (cl->poslits > 1 ) {
			/* indefinite clause */
		    fcb.nr_indefclauses++;
		    fcb.sum_lit_indefclauses += cl->nr_lits;
		    fcb.sum_plit_indefclauses += cl->poslits;
		    SET_MAX(fcb.max_lit_indefclauses,cl->nr_lits);
		    SET_MAX(fcb.max_vars_indefclauses,cl->nrvars);
		    SET_MAX(fcb.max_termdepth_indefclauses,cl->max_termdepth);
		    SET_MAX(fcb.max_termsize_indefclauses,cl->max_termsize);
		    SET_MAX(fcb.max_plit_indefclauses,cl->poslits);
			}
		else {
			/* definite clause */
		    fcb.nr_defclauses++;
		    fcb.sum_lit_defclauses += cl->nr_lits;
		    SET_MAX(fcb.max_lit_defclauses,cl->nr_lits);
		    SET_MAX(fcb.max_vars_defclauses,cl->nrvars);
		    SET_MAX(fcb.max_termdepth_defclauses,cl->max_termdepth);
		    SET_MAX(fcb.max_termsize_defclauses,cl->max_termsize);
			}
		}
	   }
	}

}
	

/************************************************/
/* usage
/************************************************/
usage(name)
char *name;
{
 fprintf(stderr,"\nusage: %s\n",name);
 fprintf(stderr,"\t[ -[no]purity]    # pure clauses are [not] removed\n");
 fprintf(stderr,"\t[ -[no]fan]       # axioms and goals are [not] fanned\n");
 fprintf(stderr,"\t[ -[no]reduct]    # reduction steps are [not] inserted\n");
 fprintf(stderr,"\t[ -[no]sgreord [at av ac]]   # subgoal ordering is [not] changed\n");
 fprintf(stderr,"\t[ -[no]clreord]   # clause ordering is [not] changed\n");
 fprintf(stderr,"\t[ -randreord]     # clause ordering is randomized\n");
 fprintf(stderr,"\t[ -all]           # enumerate all proofs\n");
 fprintf(stderr,"constraint choices:\n");
 fprintf(stderr,"\t[ -eqpred]        # enable equal predecessor fails\n");
 fprintf(stderr,"\t[ -reg]           # generate regularity constraints \n");
 fprintf(stderr,"\t[ -taut]          # generate tautology constraints\n");
 fprintf(stderr,"\t[ -subs]          # generate subsumption constraints\n");
 fprintf(stderr,"\t[ -rsubs]         # subsumtion constraints without facts\n");
 fprintf(stderr,"\t[ -overlap [n]]   # overlap constraints [n]\n");
 fprintf(stderr,"\t[ -overlapx [n]]  # extensive overlap constraints [n]\n");
 fprintf(stderr,"\t[ -cons]          # -reg + -taut + -rsubs\n");
 fprintf(stderr,"\t[ -ocons [n]]     # -reg + -taut + -rsubs -overlap [n]\n");
 fprintf(stderr,"\t[ -oconsx [n]]    # -reg + -taut + -rsubs -overlapx [n]\n");
 fprintf(stderr,"link deletion:\n");
 fprintf(stderr,"\t[ -linksubs [n]]  # remove redundant connections \n");
 fprintf(stderr,"\t[ -rlinksubs [n]] # link subsumption between rules \n");
 fprintf(stderr,"additional inference rules:\n");
 fprintf(stderr,"\t[ -foldup]        # insert folding-up operations\n");
 fprintf(stderr,"\t[ -foldupx]       # folding-up with extended regularity\n");
 fprintf(stderr,"\t[ -folddown]      # insert folding-down operations\n");
 fprintf(stderr,"\t[ -folddownx]     # folding-down with extended regularity\n");
 fprintf(stderr,"verbostity options:\n");
 fprintf(stderr,"\t[ -verbose [n]]   # verbosity of preproc phase [n]\n");
 fprintf(stderr,"\t[ -lop]           # generate lop-file file_pp.lop \n");
 fprintf(stderr,"\t[ -scode]         # generate s-file file.s \n");
 fprintf(stderr,"\t[ -noopt]         # do not optimize .hex file\n");
 fprintf(stderr,"\t[ -notree]        # the proof tree is not generated\n");
 fprintf(stderr,"\t[ -searchtree]    # enable displaying the search tree\n");
 fprintf(stderr,"\t[ -interactive]   # enable interactive proof control\n");
 fprintf(stderr,"input formula:\n");
 fprintf(stderr,"\tfile              # compile input formula from file.lop\n");
}

/************************************************/
/* usage for verbose mode
/************************************************/
verbose_usage(name)
char *name;
{

  fprintf(stderr,"usage of verbose mode in %s: %s -verbose [nu] ...\n\n",
	  name,name);
  fprintf(stderr,"print formula in lop syntax on stdout.\n");
  fprintf(stderr,"'nu' is one of the following (default = 0):\n\n");
  fprintf(stderr,"	%d  : no printing at all\n",PF_NO);
  fprintf(stderr,"	%d  : after parsing (before preprocessing)\n",PF_PARS);
  fprintf(stderr,"	%d  : after tautology constraints\n",PF_TCONSTR);
  fprintf(stderr,"	%d  : after subsumption constraints\n",PF_SCONSTR);
  fprintf(stderr,"	%d  : after constraint deletion\n",PF_FIRSTDEL);
  fprintf(stderr,"	%d  : after purity\n",PF_PUR);
  fprintf(stderr,"	%d  : after orbranch reordering\n",PF_CLREORD);
  fprintf(stderr,"	%d  : after fanning\n",PF_FAN);
  fprintf(stderr,"	%d  : after symmetry constraints\n",PF_SYMCONSTR);
  fprintf(stderr,"	%d  : after overlap constraints\n",PF_OCONSTR);
  fprintf(stderr,"	%d  : after constraint deletion\n",PF_SECONDDEL);
  fprintf(stderr,"	%d  : after subgoal reordering\n",PF_SGREORD);
  fprintf(stderr,"	%d  : always\n",PF_ALWAYS);
}

/************************************************/
/* set defaults for flags
/************************************************/
set_inwasm_defaults()
{
  inwasm_flags.clausreord = 1;
  inwasm_flags.sgreord = 1;
  inwasm_flags.ptree = 1;
  inwasm_flags.all = 0;
  inwasm_flags.verbose = 0;
  inwasm_flags.reduct = 1;
  inwasm_flags.noreduct = 0;
  inwasm_flags.fan = 1;
  inwasm_flags.nofan = 0;
  inwasm_flags.predfail = 0;
  inwasm_flags.ptree = 1;
  inwasm_flags.purity = 1;
  inwasm_flags.randreord = 0;
  inwasm_flags.genreg = 0;
  inwasm_flags.set_subgoal_list = 1;
  inwasm_flags.cpybd = 0;
  inwasm_flags.cpylimit = 0;
  inwasm_flags.foldup = 0;
  inwasm_flags.foldupx = 0;
  inwasm_flags.folddown = 0;
  inwasm_flags.folddownx = 0;
  inwasm_flags.interactive = 0;
  inwasm_flags.searchtree = 0;
  inwasm_flags.overlap = 0;
  inwasm_flags.linksubs = 0;
  inwasm_flags.hex_out = 1;
  inwasm_flags.hex_out_opt = 1;


sgreord_at=2.0; /* PJ defaults for sgreordering */
sgreord_av=1.0;
sgreord_ac=1.0;

}
  
/************************************************/
/* process_argv
/************************************************/
process_argv(argv,ac)
char **argv;
int ac;
{
while ((ac > 1) && (*argv[1] == '-')) {
	if (!strcmp(argv[1],"-all")) {
		inwasm_flags.all 	 = 1;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-noclreord")) {
		inwasm_flags.clausreord = 0;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-reg")) {
		inwasm_flags.genreg    = 1;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-purity")) {
		inwasm_flags.purity   = 1;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-nopurity")) {
		inwasm_flags.purity   = 0;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-interactive")) {
		inwasm_flags.interactive = 1;
		inwasm_flags.set_subgoal_list = 1;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-searchtree")) {
		inwasm_flags.searchtree = 1;
		inwasm_flags.interactive = 1;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-lop")) {
		inwasm_flags.nocode = 1;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-scode")) {
		inwasm_flags.hex_out = 0;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-noopt")) {
		inwasm_flags.hex_out_opt = 0;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-fan")) {
		inwasm_flags.nofan = 0;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-nofan")) {
		inwasm_flags.nofan = 1;
		argv++;
		ac--;
		continue;
		}
        if (!strcmp(argv[1],"-sgreord")) {
		inwasm_flags.sgreord = 2;
                if (isdigit(*argv[2])) {
                   sscanf(argv[2], "%f", &sgreord_at);
                   argv++;
                   ac--;
                   }
                if (isdigit(*argv[2])) {
                   sscanf(argv[2], "%f", &sgreord_av);
                   argv++;
                   ac--;
                   }
                if (isdigit(*argv[2])) {
                   sscanf(argv[2], "%f", &sgreord_ac);
                   argv++;
                   ac--;
                   }
                argv++;
                ac--;
		/* printf("Using sgreord parameters: %f %f %f\n",
		   sgreord_at,sgreord_av,sgreord_ac); */
                continue;
                }
	if (!strcmp(argv[1],"-nosgreord")) {
		inwasm_flags.sgreord = 0;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-eqpred")) {
		inwasm_flags.predfail = 1;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-reduct")) {
		inwasm_flags.reduct 	 = 1;
		inwasm_flags.noreduct 	 = 0;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-noreduct")) {
		inwasm_flags.reduct 	 = 0;
		inwasm_flags.noreduct 	 = 1;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-notree")) {
                inwasm_flags.ptree    = 0;
		inwasm_flags.set_subgoal_list = 1; /* is a must now */
                argv++;
                ac--;
                continue;
                }
	if (!strcmp(argv[1],"-subs")) {
		inwasm_flags.subs = 1;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-rsubs")) {
		inwasm_flags.subs = 2;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-taut")) {
		inwasm_flags.tauto = 1;
		argv++;
		ac--;
		continue;
		}
        if (!strcmp(argv[1],"-overlap")) {
                if (isdigit(*argv[2])) {
                   sscanf(argv[2], "%d", &overlap_mode);
                   argv++;
                   ac--;
                   }
                inwasm_flags.overlap = 1;
                argv++;
                ac--;
                continue;
                }
        if (!strcmp(argv[1],"-overlapx")) {
                if (isdigit(*argv[2])) {
                   sscanf(argv[2], "%d", &overlap_mode);
                   argv++;
                   ac--;
                   }
                inwasm_flags.overlap = 2;
                argv++;
                ac--;
                continue;
                }
        if (!strcmp(argv[1],"-linksubs")) {
                if (isdigit(*argv[2])) {
                   sscanf(argv[2], "%d", &deletion_mode);
                   argv++;
                   ac--;
                   }
                inwasm_flags.linksubs = 1;
                argv++;
                ac--;
                continue;
                }
        if (!strcmp(argv[1],"-rlinksubs")) {
                if (isdigit(*argv[2])) {
                   sscanf(argv[2], "%d", &deletion_mode);
                   argv++;
                   ac--;
                   }
                inwasm_flags.linksubs = 2;
                argv++;
                ac--;
                continue;
                }
	if (!strcmp(argv[1],"-cons")) {
		inwasm_flags.tauto   = 1;
		inwasm_flags.subs    = 2;
		inwasm_flags.genreg  = 1;
                inwasm_flags.overlap = 0;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-ocons")) {
		inwasm_flags.tauto   = 1;
		inwasm_flags.subs    = 2;
		inwasm_flags.genreg  = 1;
                if (isdigit(*argv[2])) {
                   sscanf(argv[2], "%d", &overlap_mode);
                   argv++;
                   ac--;
                   }
                inwasm_flags.overlap = 1;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-oconsx")) {
		inwasm_flags.tauto   = 1;
		inwasm_flags.subs    = 2;
		inwasm_flags.genreg  = 1;
                if (isdigit(*argv[2])) {
                   sscanf(argv[2], "%d", &overlap_mode);
                   argv++;
                   ac--;
                   }
                inwasm_flags.overlap = 2;
		argv++;
		ac--;
		continue;
		}
/*
	if (!strcmp(argv[1],"-debug_mode")) {
		if (isdigit(*argv[2])) {
		   sscanf(argv[2], "%d", &debug_mode);
		   argv++;
		   ac--;
		   }
		else debug_mode = 128;
		argv++;
		ac--;
		continue;
		}
*/
	if (!strcmp(argv[1],"-verbose")) {
		if (isdigit(*argv[2])) {
		   sscanf(argv[2], "%d", &inwasm_flags.verbose);
		   argv++;
		   ac--;
		   }
		else inwasm_flags.verbose = 0;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-randreord")) {
		inwasm_flags.randreord  = 1;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-foldup")) {
		inwasm_flags.foldup  = 1;
		inwasm_flags.reduct = 1;
		inwasm_flags.noreduct = 0;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-foldupx")) {
		inwasm_flags.foldupx  = 1;
		inwasm_flags.genreg = 1;
		inwasm_flags.set_subgoal_list = 1;
		inwasm_flags.reduct = 1;
		inwasm_flags.noreduct = 0;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-folddown")) {
		inwasm_flags.folddown  = 1;
		inwasm_flags.set_subgoal_list = 1;
		inwasm_flags.reduct = 1;
		inwasm_flags.noreduct = 0;
		argv++;
		ac--;
		continue;
		}
	if (!strcmp(argv[1],"-folddownx")) {
		inwasm_flags.folddownx  = 1;
		inwasm_flags.genreg = 1;
		inwasm_flags.set_subgoal_list = 1;
		inwasm_flags.reduct = 1;
		inwasm_flags.noreduct = 0;
		argv++;
		ac--;
		continue;
		}
	warning(argv[1]," unknown switch");
	return 1;
	}
return 0;
}


/***********************************************************/
/* check consistency of the arguments
/*	return:	0 can continue
/***********************************************************/
check_argument_consistency()
{
if (inwasm_flags.foldup && inwasm_flags.foldupx){
	warning("","-foldupx overrides -foldup");
	inwasm_flags.foldup = 0;
	inwasm_flags.set_subgoal_list = 1;
	}
if (inwasm_flags.folddown && inwasm_flags.folddownx){
	warning("","-folddownx overrides -folddown");
	inwasm_flags.folddown = 0;
	inwasm_flags.set_subgoal_list = 1;
	}
if (inwasm_flags.interactive && inwasm_flags.randreord){
	warning("",
	 "-interactive is not compatible with -randreord: -norandreord set");
	inwasm_flags.randreord = 0;
	}
if (inwasm_flags.interactive){
	inwasm_flags.set_subgoal_list = 1;
	}

return 0;
}
	
