/**************************************************************************
;;; MODULE NAME : constrsu.c
;;;
;;; PARENT      : main.c
;;;
;;; PROJECT     : MPLOP
;;;
;;; SCCS        : @(#)constrsu.c	13.2 07 Apr 1995
;;;
;;; AUTHOR      : Mohamed Salah, Hamdi
;;;
;;; DATE        : 15.11.92
;;;
;;; DESCRIPTION : generate constraint subsumption 
;;;
;;; REMARKS     :
;;;
***************************************************************************/

 
/*********************** INCLUDE ******************************************/

#include "extern.h"  /* symb.h: types.h, defines.h, symbtab ... */

/*********************** DEFINE *******************************************/

#define TRUE 1
#define FALSE 0
#define DELETED (-9999)
#define SV_TO_XV 6        /* mode fuer my_rename */
#define TO_OLD 4        /* mode fuer my_rename */

/*********************** EXTERNE DATEN ************************************/


/*********************** EXTERNE FUNKTIONEN *******************************/

extern void my_rename();                /* sconstr.c */
extern termtype *copy_terms();          /* hlpfkt.c */

/*********************** GLOBALE DATEN ************************************/

static int Deleted = 0; 
static int Olddeleted = 0; 
static int Generated = 0;


/*********************** PROTOTYPEN ***************************************/

predtype *constr_subsume();
void free_term();

/*********************** FUNKTIONEN ***************************************/


/*---------------------- gen_constr_subs ---------------------------------*/

   /* Loescht subsumierbare Constr. der Klauselliste cll.
     Hauptprogramm fuer Constr.-Subsumption.
   */

/*---------------------- gen_constr_subs ---------------------------------*/

gen_constr_subs(cll)

 claustype *cll;
{
  claustype *fann;

   Generated = 0;
   Olddeleted = Deleted;
   Deleted = 0;
   
  while (cll)
   {
     /* optimiere Constr. der original Klausel */
     cll->constr = constr_subsume(cll->constr);

     if (cll->fann) /* cll hat fann-Kompn. */
      {
	/* optimiere Constr. der fann-Komp. */
	fann = cll->fann;
	while (fann)
	 {
	   fann->constr = constr_subsume(fann->constr);

	   fann = fann->next;
	 }
      }

     cll = cll->next;
   }
   printf("  Deleted constraints: %d\n",Deleted);
}


/*---------------------- constr_subsusme ---------------------------------*/

   /* Entfernt subsumierbare Constr. aus cl.
      Jedes Constr. mit all seine Nachfolger.

      Return-Wert: optimierte Liste von Constr.
   */

/*---------------------- constr_subsusme ---------------------------------*/

predtype *constr_subsume(cl)

 predtype *cl;
{
   predtype *hp1,*hp2;

   /* Subsumierbare Constr. markieren: ->sign = DELETED */

   hp1 = cl;

   while (hp1)
    {
      /* zaehle vorhandene Constraints */
      if (hp1->litnr > Generated)
         	Generated = hp1->litnr;

      if (hp1->sign != DELETED)
       {
	 /* vergleiche hp1 mit allen hinteren */
	 hp2 = hp1->next;
	 while (hp2 && hp1->sign != DELETED)
	  {
	    if (hp2->sign != DELETED)
	     {
	       switch( subsumes_c(hp1,hp2) )
		{
		  case 1: /* hp1 wird von hp2 subsumiert: loesche hp1 */
			  hp1->sign = DELETED;
			  Deleted += 1;
			  break;
		  case 2: /* hp2 wird von hp1 subsumiert: loesche hp2 */
			  hp2->sign = DELETED;
			  Deleted += 1;
			  break;
		  default: /* -1: Keine Subsumption: tue nichts */
			   break;
		}
	     }

	    hp2 = hp2->next;
	  }
       }

      hp1 = hp1->next;
    }


   /* DELETED-Constr. loeschen */

   hp1 = cl;
   hp2 = NULL;
   while (hp1)
    {

      if (hp1->sign == DELETED)
       {
	 if (hp1 == cl)
	  hp1 = cl = hp1->next;
	 else
	  {
	    hp2->next = hp1->next;
	    hp1 = hp1->next;
	  }
       }
      else
       {
	 hp2 = hp1;
	 hp1 = hp1->next;
       }
    }

   return(cl);
}


/*---------------------- subsumes_c --------------------------------------*/

   /* Untersucht c1 und c2 auf Subsumption.
      Return-Wert:
		  1: c1 wird von c2 subsumiert.
		  2: c2 wird von c1 subsumiert.
		 -1: keine Subsumption zwischen c1 und c2.
   */

/*---------------------- subsumes_c --------------------------------------*/

int subsumes_c(c1,c2)

 predtype *c1,*c2;
{
  switch( c_subs(c1,c2) )
   {
     case 1: /* c1 wird von c2 subsumiert */
	     return(1);
     case 0: /* Moeglichkeit der Subs. von c2 durch c1 */
	     if ( c_subs(c2,c1) == 1 )
	      return(2);
	     else
              return(-1);
     case -1: return(-1);
   }
}


/*---------------------- c_subs ------------------------------------------*/

   /* Prueft ob das Constr. c1 vom Constr. c2 subsumiert werden kann.
      Return-Wert:
		  1: c1 wird von c2 subsumiert.
		  0: c1 kann NICHT von c2 subsumiert werden aber
		     es ist moeglich, dass c2 von c1 subsumiert wird.
		 -1: Keine Subsumption zwischen c1 und c2 moeglich.
		     (in beiden Richtungen!)
   */
/*---------------------- c_subs ------------------------------------------*/

int c_subs(c1,c2)

 predtype *c1,*c2;
{
  termtype *l1,*r1,*l2,*r2;
  int rv1,rv2,n,m;
  int erg = -1; /* Ergebnis */

  /* Es  werden nur nicht leere Constr. betrachtet */
  if (!c2)
      return(1);
  if (!c1)
      return(-1);

  if (c2->sign == TAUT_CONSTR || c2->sign == SUB_CONSTR ||
      c1->sign == SYMM_CONSTR || c1->sign == OVER_CONSTR ||
      (c1->sign == TAUT_CONSTR && !c1->claus->fann) ||
      (c1->sign == SUB_CONSTR && !c1->claus->fann)) {
	
  l1 = c1->t_list->t_list;
  r1 = c1->t_list->next->t_list;

  /* #V... ---> V...: damit werden Bind. von SV von c1 gezaehlt */
  my_rename(r1,0,SV_TO_XV);

  l2 = c2->t_list->t_list;
  r2 = c2->t_list->next->t_list;

  rearrange();
  rv1 = unify(l1,r1);
  n = nr_bindings();   /* Anz. der geb. Var. (SV nicht betrachten) */
  rv2 = unify(l2,r2);
  m = nr_bindings();   /* Anz. der geb. Var. (SV nicht betrachten) */

  my_rename(r1,0,TO_OLD);
  if (rv1 == SUCCESS && rv2 == SUCCESS)
   {
     if (m != n) /* c1 NICHT subsumierbar aber andere Richtg moeglich */
      erg = 0;
     else        /* m == n: c1 wird von c2 subsumiert */
      erg = 1;
   }
  else           /* Keine Subs. in beiden Richtgen */
   erg = -1;
 }
 else 
   erg = 0;


  return(erg);
}


/*---------------------- nr_bindings -------------------------------------*/

   /* liefert die Anzahl der Bindungen aus dem Trail.
      Bindungen fuer SV werden nicht mitgezaehlt.
   */

/*---------------------- nr_bindings -------------------------------------*/

int nr_bindings()
{
  int i;
  int erg = 0; /* Ergebnis */
  trailelt * trp;
  int varsymb;
  termtype *right;

  for (i = 0; i < trailcnt; i++)
   {
        trp = (trailelt*)symbtab[trail[i].symb].loclist;
        varsymb = trp->symb;
        right   = trp->term;
 
     if ( symbtab[varsymb].name[0] != '#' )
      /* there is a bound stucture variable */
      erg += 1;
   }

  return(erg);
}

/*---------------------- test_subs -------------------------------------*/

int testsubs(oconstr,clist)

predtype *oconstr, *clist;
{
 while (clist) 
  {
   if (clist->sign != DELETED && c_subs(oconstr,clist) == 1)
      return(1);
   clist = clist->next;
  }
 return (0);
}



/*########################################################################*/
