/**************************************************************************
;;; MODULE NAME : symconstr.c 
;;;
;;; PARENT      : main.c 
;;;
;;; PROJECT     : MPLOP
;;;
;;; SCCS        : @(#)symconstr.c	13.2 07 Apr 1995
;;;
;;; AUTHOR      : Mohamed Salah, Hamdi 
;;;
;;; DATE        : 15.11.92 
;;;
;;; DESCRIPTION : generate symmetry constraints 
;;;
;;; REMARKS     :
;;;		30.12.1993	Hamdi	extern.h
;;;		25.02.94	Hamdi	SUBS,RSUBS ..
;;;
;;;
***************************************************************************/



/*********************** INCLUDE ******************************************/

#include "extern.h"  /* symb.h: types.h, defines.h, symbtab ... */
#include "stypes.h"
#include "defines.h" /* PJ */
/*********************** DEFINE *******************************************/

#define NORMALC 0   	/* Subsumption zwischen normalen Klauseln */
#define CONTRAC 1   	/* Subsumption zwischen Kontrapositiven einer
                           Klausel. */
#define DELETED (-9999)
#define SUBS 1   	/* modes fuer den Test auf Subsumption */
#define RSUBS 2  	/* SUBS: alle Klauseln vergleichen     */
		 	/* RSUBS: nur Regeln vergleichen       */

/*********************** EXTERNE DATEN ************************************/


/*********************** EXTERNE FUNKTIONEN *******************************/

extern termtype *copy_terms();          /* hlpfkt.c */
extern unify();                         /* unify.c    */
extern littype *make_cart_list();       /* sconstr.c  */
extern clause_s_constr();               /* sconstr.c  */
extern void my_rename();                /* sconstr.c  */
extern void rearrange();                /* unify.c    */
extern delete_p();                      /* purity.c */
extern symbollist *app_sym_to_list();   /* sconstr.c */

/*********************** GLOBALE DATEN ************************************/

extern int no_subsumption;

/*********************** PROTOTYPEN ***************************************/

littype *sym_make_cart_list();

/*********************** FUNKTIONEN ***************************************/

/*---------------------- gen_sym_constr ----------------------------------*/

   /* Generiert Symmetrie Constraints */

/*---------------------- gen_sym_constr ----------------------------------*/

gen_sym_constr(Gesamtstruktur)

 claustype *Gesamtstruktur;
{
   claustype *hc; 

   nr_gen_constr = 0;
   hc = Gesamtstruktur;

   if (!hc) 
    return -1;

   /*******************************/
   /* Default query "uberspringen */
   /*******************************/
   if (!hc->p_list->symb) 
    hc = hc->next;

   while (hc)
    {
      if (hc->fann)    
       {
         if (inwasm_flags.linksubs) 
	  { 	
	    if (pos_lits(hc)<=1) 
	      /* Fann-Komponenten vergleichen */
              sym_constr(hc);
          }
	 else 
	  {
	    /* Fann-Komponenten vergleichen */
            sym_constr(hc);
          }
       }

      hc = hc->next;
    }
    printf("  Generated constraints: %d\n", nr_gen_constr);


}


/*---------------------- sym_constr --------------------------------------*/

   /* Bearbeitet die Kontrapositiven von clause und generiert
      gegebenfalls Constraints.
   */

/*---------------------- sym_constr --------------------------------------*/

sym_constr(clause)

 claustype *clause;
{
  claustype *hp,*hp2;
  int rv;
  symbollist *dellist = NULL; 
 
  if (! clause)
   return 0;

  if (clause->delpos == PRULE) 
  /********************************/
  /* Kontrapositiv ist keine Query */
  /********************************/
   {
      /*******************************************/
      /* Erstes Kontrapositiv mit allen hinteren */
      /*******************************************/
      hp = clause->fann;
      while (hp) 
       {
         rv = clause_s_constr(clause,hp,CONTRAC,SUBS); 

         /*
            result_msg(clause->clnr,hp->clnr,rv);
         */
    
         if (rv == 1)  
	    /****************************************/
	    /* Subsumption zwischen Kontrapositiven */
	    /****************************************/
          {
	    /************************/
            /* Aktualisiere dellist */
	    /************************/
            dellist = app_sym_to_list(dellist,(- hp->clnr));

            hp->del = DELETED;
          }

         hp = hp->next;
       }  
    }
  /***********************************************/ 
  /* Restliche Kontrapositive mit allen hinteren */
  /***********************************************/ 
  hp = clause->fann;
  while (hp)
   {
     if (hp->del != DELETED)
      {
        hp2 = hp->next;
        while (hp2)
         {
           if (hp2->del != DELETED)
            {
              rv = clause_s_constr(hp,hp2,CONTRAC,SUBS);

              /*
              	result_msg(hp->clnr,hp2->clnr,rv);
              */
         
              if (rv == 1)  
	      /****************************************/
	      /* Subsumption zwischen Kontrapositiven */
	      /****************************************/
               {
	         /*************************/
                 /* Aktualisiere dellist */
	         /*************************/
                 dellist = app_sym_to_list(dellist,(- hp2->clnr));

                 hp2->del = DELETED;
               }

            }

           hp2 = hp2->next;
         }
      }

     hp = hp->next;
   }
 
  /**************************************/
  /* Ausgabe der DELETED-Kontrapositive */
  /**************************************/
  if (dellist)
   print_del_contra(clause->clnr,dellist);

  return 0;
}


/*---------------------- sym_make_cart_list ------------------------------*/

   /* Erstellt Kopie von kpp 
   	- mit Umbenennung der Variablen
      	- und Berechnung von s_list fuer jedes Literal.

      Return-Wert: Erstellte Kopie.
      Nebeneffekt: Falls eine s_list leer ist, dann wird die Variable
		   no_subsumption = 1 und das bereits gebildete Teil
		   der Kopie zurueckgegeben.

      Falls Head1 (kpp) und Head2 (lpp) keine Partner-Literale sind,
         - dann ist das Ergebnis NULL.

      Die Funktion wird von sconstr.c aus aufgerufen.
   */

/*---------------------- sym_make_cart_list ------------------------------*/

littype *sym_make_cart_list(kpp,clnr,lpp)

 predtype *kpp;
 int clnr;
 predtype *lpp;
{
   littype *newlit = NULL;
   termtype *htermlist;
   plittype *newerg;


   if (kpp && lpp && kpp->symb >= nrbuilt)
     {
        htermlist = copy_terms(kpp->t_list,ALL_TERM);
        my_rename(htermlist,clnr,ALL_TO_SV);
  
        rearrange();
        if ( (kpp->symb == lpp->symb) && (kpp->sign == lpp->sign)
             && (unify(lpp->t_list,htermlist) == SUCCESS) )
        /*****************************/
	/* Kopfliterale unifizierbar */
        /*****************************/
         {
	   newlit = (littype *) malloc(sizeof(littype));
	   newlit->symb = kpp->symb;
 	   newlit->sign = kpp->sign;
	   newlit->t_list = htermlist; 

           /*********************/
	   /* Initialisierungen */
           /*********************/
	   newerg = (plittype *) malloc(sizeof(plittype));
	   newerg->pr = lpp;
	   newerg->notunif = 0;
	   newerg->next = NULL;
	   newlit->s_list = newerg;

           /*************************/
	   /* Berechnung von s_list */
           /*************************/
	   newlit->next = make_cart_list(kpp->next,clnr,lpp->next);
         }
        else 
        /***********************************/
	/* Kopfliterale nicht unifizierbar */
        /***********************************/
	     {
	       no_subsumption = 1;
	       return(NULL);
             }
     }

    return(newlit);

}




/*---------------------- print_del_contra --------------------------------*/


print_del_contra(clnr,list)
int clnr;
symbollist  *list;
{
  printf("  Deleted contrapositives: ");
  while  (list)
   {
     printf("%d.%d ",clnr,list->symbol);
     list = list->next;
   }
  printf("\n");
}



/*########################################################################*/
