/*########################################################################
# Include File:		tags.h
# SCCS-Info:		8/26/92, @(#)tags.h	1.8
# Author:		J. Schumann
# Date:			88
# Context:		wasm (SETHEO)
# 
# Contents:		???
#
# Modifications:	
#	(when)		(who)		(what)
#		 			tags mod'd to fit any address-start
#	13.4.88 			CREF
#       13.4. 				atari V.
#	14.7. 				signals
#	5.10. 				return values
#	7.11. 				global variables
#	30.11. 				hypercube
#	27.12. 				transputer
#	31.1.89 			Makefile
#	14.2.89				IDTAG
#	19.6.89 			ABORT
#	3.10.89 			reordering of return codes
#	10.10.90 			new return codes
#       24.6.91 	Goller		changes to ISVAR
#	23.10.91	Max Moser	`restyled'
#       11.02.92        Markus Bschorer #ifdef CONSTRAINTS
#	   09.03.92		Markus Bschorer New Tag T_STRVAR
#	18.02.93	jsc		T_STRVAR always active
#	25.2.93		Goller		T_CSTRVAR added, T_STRVAR -> 9
#	11.3.93	Mueller,Goller neue Constraints, Statistics
#	29.6.93		jsc		include-defines
#	1.6.94		jsc		new macros
#
# Bugs:
#
# --- Forschungsgruppe KI, Inst. fuer Informatik, TU Muenchen ---
########################################################################*/
#ifndef SAM_TAGS_H
#define SAM_TAGS_H


/*------------------------------------------------------------------------
- !!!!!MACHINE DEPENDENT !!!!!!!!!!!!!!
------------------------------------------------------------------------*/


#ifdef sun4
#define  PTR_MASK	0		/* mask for pointer     	*/
typedef  unsigned int   WORD;           /* 32-bit wordlength 		*/
#endif

#ifdef sun3
#define  PTR_MASK	0		/* mask for pointer     	*/
typedef  unsigned int   WORD;           /* 32-bit wordlength 		*/
#endif

#ifdef hpux
#define  PTR_MASK	4		/* mask for pointer     	*/
typedef  int   WORD;                    /* 32-bit wordlength 		*/
#endif

#ifndef PTR_MASK
#define  PTR_MASK	0		/* mask for pointer     	*/
typedef  unsigned int   WORD;           /* 32-bit wordlength 		*/
#endif


#ifdef hpux

/*------------------------------------------------------------------------
               WORD-format:

    4 bits 4 bits      24 bits
   +------+------+---------------------------+
   | RES  | TAG  |  VALUE                    |
   +------+------+---------------------------+
    31  28 27  24 23                        0

------------------------------------------------------------------------*/

#define TAGMASK         0x0f000000      /* 4-bit tags 			*/
#define TAGSTART  	24
#define VALMASK 	16777215	/* 2**24-1   			*/
#define PTRMASK		4043309055	/* 0xf0ffffff			*/
#define IDMASK          1048575         /* ID-nr mask 2**20 -1 		*/
#define IDTAG		0x100000	/* if on, is a constant		*/
#define NUMMASK		0x0000ffff      /* only short-arithmetic        */
#define MAXNUMBER       32767           /* must correspond to NUMMASK   */
#define MINNUMBER       -32767          /* must correspond to NUMMASK   */
#define WORDLENG        32              /* 32 bit wordlength 		*/
#define T_ISVAR         8               /* bitpos of VAR-types 		*/
#define MAXPREDSYMB     10000           /* must be equal to mplop       */

#else  /* if not hpux */

/*------------------------------------------------------------------------

                          WORD-format:

                4 bits     28 bits 
               +------+----------------------------+
               | TAG  |  VALUE                     |
               +------+----------------------------+
                31  28 27                         0

------------------------------------------------------------------------*/

#define TAGMASK         0xf0000000      /* 4-bit tags 			*/
#define TAGSTART  	28
#define VALMASK 	0x0fffffff
#define PTRMASK		0x0fffffff  
#define IDMASK          0x000fffff      /* ID-nr mask 2**20 - 1         */
#define IDTAG		0x00100000      /* if on, is a constant		*/
#define NUMMASK		0x0000ffff      /* only short-arithmetic        */
#define MAXNUMBER       32767           /* must correspond to NUMMASK   */
#define MINNUMBER       -32767          /* must correspond to NUMMASK   */
#define WORDLENG        32              /* 32 bit wordlength 		*/
#define T_ISVAR         8               /* bitpos of VAR-types 		*/
#define MAXPREDSYMB     10000           /* must be equal to mplop       */

#endif  /* if not hpux */



/*----------------------------------------------------------------------*/

#define GETTAG(obj) \
        (((obj) & TAGMASK) >> (TAGSTART))

#define GETVAL(obj) \
        ((obj) & VALMASK)

#define GETPTR(obj)\
        ((WORD*) ((obj) & PTRMASK))

#define GETNUMBER(OBJECT) \
        ((short) ((OBJECT) & NUMMASK))

#define GETSYMBOL(OBJECT) \
        ((OBJECT) & IDMASK)

#define GETSYMBOLINDEX(OBJECT) \
        (GETSYMBOL(OBJECT) \
         - (ISNEGATED(OBJECT)? MAXPREDSYMB : 0))

#define GETARITY(OBJECT)  \
        symbtab[GETSYMBOLINDEX(OBJECT)].arity

#define GETPRINTNAME(OBJECT)  \
        symbtab[GETSYMBOLINDEX(OBJECT)].name

#define GETSIGN(OBJECT)                 \
        ((ISNEGATED(OBJECT))?-1:1)


#define CHANGE_SIGN(OBJECT) \
        ((ISNEGATED(OBJECT)) ? (OBJECT) - MAXPREDSYMB : \
                               (OBJECT)+ MAXPREDSYMB)

#define ISCOMPLEX(OBJECT) \
        ((GETTAG(OBJECT) == T_CRTERM) || 			\
         (GETTAG(OBJECT) == T_GTERM)  || 			\
         (GETTAG(OBJECT) == T_NGTERM))

#define ISNUMBER(OBJECT) \
        ((GETTAG(OBJECT) == T_CONST) && \
        !((OBJECT) & IDTAG))

#define ISCONSTANT(OBJECT) \
        ((GETTAG(OBJECT) == T_CONST) && \
        ((OBJECT) & IDTAG))

#define ISSYMBOL(OBJECT) \
        (ISCONSTANT(OBJECT) || ISCOMPLEX(OBJECT) )

#define ISVAR(OBJECT) \
        (GETTAG(OBJECT) & T_ISVAR)

#define ISLITERAL(OBJECT) \
        (   (ISCONSTANT(OBJECT) || ISCOMPLEX(OBJECT)) \
         && symbtab[GETSYMBOLINDEX(OBJECT)].type == SYMB_P)

#define ISNEGATED(OBJECT) \
        (GETSYMBOL(OBJECT) >= MAXPREDSYMB)

#define CONSTCELLISNUMBER(OBJECT) \
        !((OBJECT) & IDTAG)

#define CONSTCELLISSYMBOL(OBJECT) \
        ((OBJECT) & IDTAG)

#define INDEXTOSYMBOL(OBJECT) \
        (OBJECT | IDTAG)


/* Generate arbitrary cells: */
#define GENOBJ(where,val,tag) \
        ((where)  =  ((val) | ((tag) << (TAGSTART))))

/* Generate cells for numeric values (remember: only short-arithmetic !): */
#define GENNUMBER(where,val) \
        GENOBJ(where, (((WORD) (val)) & NUMMASK), T_CONST)

/* Generate cells for symbols (predicate, function, constant symbols): */
#define GENSYMBOL(where,val,tag) \
        GENOBJ(where, ((((WORD) (val)) & IDMASK) | IDTAG), tag)

/* Generate cells for pointers: */
#define GENPTR(where,val) \
        GENOBJ(where, (WORD) (val), T_BVAR)


/* for compatibility */
#define GETID(X) GETSYMBOL(X)



/*------------------------------------------------------------------------
- Tags
------------------------------------------------------------------------*/

#define T_EMPTY         0               /* empty cell or nil    	*/
#define T_CONST         1               /* Constant             	*/
#define T_GTERM         2               /* Ground term, no copy 	*/
#define T_NGTERM        3               /* non-ground, code-area	*/
#define T_CRTERM        4               /* term in heap         	*/
#define T_EOSTR         5               /* end of structure     	*/
#define T_CREF		6		/* reference in code ar.	*/

#define T_VAR           11              /* free variable        	*/
#define T_BVAR          12              /* bound var == REF     	*/
#define T_FVAR		13		/* variable cell, no inf	*/

#define T_GLOBL		10		/* Global variable	        */
					/* T_GLOBAL is a tag similar to 
					   T_FVAR, not an offset as
					   defined in Johann's Diss.
					   Goller                       */

#define T_HREF		14		/* reference to heap	        */
					/* used for globals	        */

#define T_CSTRVAR	8
#define T_STRVAR        9               /* tag for structure-variables  */
                                        /* neccessary for handling sub- */
                                        /* sumption- and antilemma-     */
                                        /* constraints.                 */
                                        /* added by bschorer            */

#define T_MARK          T_EMPTY         /* Tag for marking cells (used 
					   for hornanl, nonhornanl, 
					   tc_bound, dynsgreord)        */
#define T_MARKED_FVAR   7               /* Tag for marking cells (used 
					   for delay of subgoals)       */

/*------------------------------------------------------------------------
- Signal handling
------------------------------------------------------------------------*/

#define SIGNALS		/**/

/*------------------------------------------------------------------------
- general purpose
------------------------------------------------------------------------*/

#ifndef TRUE
#define TRUE		'\1'
#define FALSE		'\0'
#endif

#define NVAL		(environ *)((-1) & PTRMASK)

#define LISTFUNCT	1	/* index of list functor in symbtab	*/
#define EMPTYLIST	2	/* index of empty list in symbtab	*/


/*------------------------------------------------------------------------
- instr() return values
------------------------------------------------------------------------*/
typedef enum {
	success =0,
	failure,
	sam_abort,
	error,
	mem_overflow,
	totfail,
        boundfail,
	stop_cycle		= 4096
	} instr_result;

#define REMODE		(-7)
#define NORMAL		(-8)


/*------------------------------------------------------------------------
- tags for Regularity-Constraints
------------------------------------------------------------------------*/

#ifdef CONSTRAINTS

#define REGULARITY      1     /* Tag for regulartity-constr.  */
#define TAUTOLOGIE      2     /* Tag for tautologic-constr.   */
#define SUBSUMPTION     3     /* Tag for subsupmtion-constr.  */
#define ANTILEMMA       4     /* Tag for antilemma-constr.    */

#endif



/*------------------------------------------------------------------------
- inference types used as tags for environ->tp
------------------------------------------------------------------------*/

typedef enum {
	EXTENSION = 0,
	REDUCTION,
	FACTORIZATION_PESS,
	FACTORIZATION_OPT,
	LEMMA,
	NOT_SOLVED
	} inf_types;



/*------------------------------------------------------------------------
- types of alternatives at a choicepoint
------------------------------------------------------------------------*/

typedef enum {
    declarative,     /* declarative rule */
    proc_fact,       /* procedural fact */
    proc_query,      /* procedural query */
    proc_rule,       /* procedural rule */
    reduct           /* reduction step */
} alt_types;


/*------------------------------------------------------------------------
- Semantics of environ->flags  
-----------------------------------------------------------------------------*/

/* In order to preserve compatibility with Antilemmata (everything on the
trail has to be tagged) flags are tagged with T_EMPTY */

#ifdef hpux

/*----------------------------------------------------------------------------
           folded_up (bit 28)
           |
-          |                                             0 
-  +--------+---------+-----------------------------------+
-  |        | T_EMPTY |        inference-number           |
-  +--------+---------+-----------------------------------+
-

The inference-number is set in i_ptree 
-----------------------------------------------------------------------------*/

#ifdef FOLD_UP
#define NOT_FOLDED_UP       0xE0FFFFFF
#define FOLDED_UP           0x10000000
#endif

#else  /* if not hpux */

/*----------------------------------------------------------------------------
              folded_up (bit 27)
              |
-             |                                          0 
-  +---------+--------------------------------------------+
-  | T_EMPTY |                 inference-number           |
-  +---------+--------------------------------------------+
-

The inference-number is set in i_ptree 
-----------------------------------------------------------------------------*/

#ifdef FOLD_UP
#define NOT_FOLDED_UP       0x07FFFFFF
#define FOLDED_UP           0x08000000
#endif

#endif  /* if not hpux */


#endif

/*######################################################################*/
