/****************************************************
;;; MODULE NAME : output.c
;;;
;;; PARENT      :
;;;
;;; PROJECT     : MPLOP
;;;
;;; SCCS        : @(#)output.c	15.2  10 May 1996
;;;
;;; AUTHOR      : Janos Breiteneicher
;;;
;;; DATE        : 15.5.91
;;;
;;; DESCRIPTION : print matrix, print_occ, error messages
;;;
;;; REMARKS     :
;;;		2.7.93	jsc	extern.h
;;;		12.7.93	jsc	silent removed
;;;		30.3.94 hamdi	static functions decl.
;;;		2.2.96	jsc	fatal/2	
;;;
****************************************************/


#include "symb.h"
#include "extern.h"

/***************************************************/
/* Local Macros
/***************************************************/
#define IND i,' '	/* indent */

/***************************************************/
/* statics
/***************************************************/
static int i = 0;
static FILE *fout = stdout; 


/*******************************************/
/* print_occ()
/*******************************************/
/* print occurrence list on fout
/* in readable form
/*******************************************/
print_occ(fout)
FILE * fout;
{
int j;
predlist * lp;

fprintf(fout,"\nOccurencies of predicates and Built-ins : (clnr,litnr)\n");
for (j=0; j< nump; j++ ) {
	if (symbtab[j].type == PREDSYM || symbtab[j].type == BUILTIN ) {
		lp = symbtab[j].loclist;
		if(!lp) continue;
		/* print name and list of occurrences */
		fprintf(fout,"Symbol %10s /%d  occurrs: ",
			NAM(j),symbtab[j].arity);
		while (lp) {
			fprintf(fout,"%c(%d,%d) ",lp->pr->sign,
				lp->pr->claus->clnr,lp->pr->litnr);
			lp=lp->next;
			}
		fprintf(fout,"\n");
		}
	}
}

/*******************************************/
/* print_statistic()
/*******************************************/
/* print statistic after compilation	   */
print_statistics(fout,fnam)
FILE *fout;
char *fnam;
{
fprintf(fout,"\n/*  Information %s :\n",fnam);
fprintf(fout,"%d clauses \t%d constants  %d functions  %d numbers in formula\n",
	nr_clauses, nr_constants, nr_functions, nr_numbers);
fprintf(fout,"%d entries in symboltable  (%d predefined) %d global variables\n",
	nump,nrbuilt,nr_globals);
fprintf(fout,"longest clause %d literals\n",maxlit-1);
fprintf(fout,"%d deterministic calls  %d indet. calls with maximum %d labels\n",
	detcal,orbcal,maxunif);
fprintf(fout,"*/ \n");
}


/*******************************************/
/* warning
/*******************************************/
/* print warning on stdout
/*******************************************/
warning(t,s)
char *t,*s;
{
   fprintf(stderr,"Warning : %s %s\n",t,s);
}

/*******************************************/
/* nferror
/*******************************************/
/* print non fatal error on stdout
/*******************************************/
nferror(s,lineno)
char *s;
int lineno;
{
fprintf(stderr,"Non-fatal error : %s  in line %d\n",s,lineno);
errors++;
}

/*******************************************/
/* fatal
/* - report a fatal error and exit
/*******************************************/
fatal (s,exitval)

    char           *s;		/* error message			 */
    int		   exitval;

{
    fprintf (stderr,"inwasm: fatal error: %s\n",s);
    exit(exitval);
}
