
/*----------------------------------------------------------------------------*/
/* Programm: FindConstraints                                                  */
/* Datei:    Subsumptions-Funktionen                                          */
/* Autor:    Martin Ziegenaus                                                 */
/* Datum:    02.10.91                                                         */
/* SCCS:     @(#)cl_reordering.c	15.1 01 Apr 1996                                                          */
/* Modif:    12.11.92 sort.c M. Salah Hamdi                                   */
/* Modif:    28.10.94 K. Mayr: weigt of consts and vars exchanged,            */
/*                    clause complexity is the sum of lit complexity,         */
/*                    less complex clauses get smaller indices.               */
/*                    Intialisierung von erg korrigiert: meist 1.0            */
/*                    Neue Sortierfunktion hinzugefuegt                       */
/* Modif:    16.06.95 PJ: weight function                                     */
/*----------------------------------------------------------------------------*/


#include "extern.h"
#define DELETED (-9999)

extern int pos_lits();
extern symbollist *get_all_var();


float BerechneSortierkriterium();
void SortiereNachKriterium();
int nr_var_in_lit();
void count_connections();

sortarray *ssort = NULL;
sortarray *hsort = NULL;



/*-------------------------- inspect_clauses --------------------------------*/

void inspect_clauses(cll)

claustype *cll;
{
 claustype *cl;
 int plits;

 minlits = 10000;

 for (cl = cll; cl; cl = cl->next) {
     plits = pos_lits(cl);
     if (plits == 0)
        nr_negated_clauses++;
     if (plits == 1)
        nr_definite_clauses++;
     if (plits > 1)
        nr_indefinite_clauses++;
     if (cl->nr_lits = 1)
        nr_facts++;
     if (cl->nr_lits > maxlits) 
        maxlits = cl->nr_lits; 
     if (cl->nr_lits < minlits) 
        minlits = cl->nr_lits; 
     nr_literals += cl->nr_lits; 
     nr_connections += cl->connections; 
     complexity += cl->complexity; 
 }
}

/*----------------------------------------------------------------------------*/
/* KriteriumEinesTerms 						 	      */
/*----------------------------------------------------------------------------*/
float KriteriumEinesTerms( term )
	termtype *term;

{
	int argnr;	/* zum Feststellen der Aritaet bei Funktionen */
	termtype *hpt;	/* zum Durchlaufen der Argumentenliste bei Funktionen */
	float erg;

erg = 0.0;

if ( term->type == VARIABLE )
	return( 0.0 );

if ( term->type == GVARIABLE )
	return( 0.0 );

if ( term->type == CONSTANT || term->type == NUMCONST || 
     term->type == STRINGCONST )
	{
	if ( term->t_list == NULL )
                /* Konstante haben Gewicht 1 */
		return( 1.0 );
	else
		{
		/* Funktionen haben Gewicht                 */ 
                /* 1 / Summe der Argumentkomplexitaeten + 1 */
		argnr = 1;
		hpt = term->t_list;
		while ( hpt != NULL )
			{
			argnr++;
			hpt = hpt->next;
			}
		
	/* durchlaufe Unterstruktur und rufe diese Funktion rekursiv auf */
		hpt = term->t_list;
		erg = 1.0;
		while ( hpt != NULL )
			{
			erg += KriteriumEinesTerms( hpt );

			hpt = hpt->next;
			}

		return( erg / (float) (argnr));
		}
	}
else
	return(-1.0);
}			

/*----------------------------------------------------------------------------*/
/* KriteriumEinesPraedikates						      */
/*----------------------------------------------------------------------------*/
float KriteriumEinesPraedikates( praedikat )
	predtype *praedikat;
	
{
	termtype *term;
	float erg;

erg = 1.0;

/* durchlaufe Termliste des Praedikates */
term = praedikat->t_list;

while ( term != NULL )
	{
	erg += KriteriumEinesTerms( term ) / (float)(praedikat->arity + 1);

	term = term->next;
	}


praedikat->complexity = erg ;

return( erg );
}

/*----------------------------------------------------------------------------*/
/* BerechneSortierkriterium						      */
/*----------------------------------------------------------------------------*/
float BerechneSortierkriterium( Clausel )
	claustype *Clausel;
	/* Berechnet ein float-Kriterium, nach dem die Klauseln sortiert werden
		koennen um bei der Subsumption moeglichst starke Constraints
		zu erhalten. */
{
	predtype *praedikat;
	float erg;



erg = 0.0;

/* durchlaufe Praedikatenliste der Klausel */
praedikat = Clausel->p_list;

while ( praedikat != NULL )
	{
	erg += KriteriumEinesPraedikates( praedikat );

	praedikat = praedikat->next;
	}
Clausel->complexity = erg;

return( erg );
}



/*----------------------------------------------------------------------------*/
/* tausche()								      */
/*----------------------------------------------------------------------------*/
static void tausche(sa,i,j)
sortarray *sa;
int i,j;
{
sortarray tausch;

tausch.kriterium = sa[i].kriterium;
tausch.varanz = sa[i].varanz;
tausch.Clause = sa[i].Clause;

sa[i].kriterium = sa[j].kriterium;
sa[i].varanz = sa[j].varanz;
sa[i].Clause = sa[j].Clause;

sa[j].kriterium = tausch.kriterium;
sa[j].varanz = tausch.varanz;
sa[j].Clause = tausch.Clause;

	/* Ordnung in der Klausel sichern */
sa[j].Clause->order = j;
sa[i].Clause->order = i;
}

	
/*----------------------------------------------------------------------------*/
/* SortiereNachKriterium						      */
/*----------------------------------------------------------------------------*/
void SortiereNachKriterium(ssort,Clausanz)
sortarray *ssort;
int Clausanz;
{
    int i,j,length_i,length_j;

    /* strukturelle Klauseln weiter nach hinten! */
    for (i = 0; i < Clausanz; i++) {
        if ((float) (ssort[i].Clause->nr_lits) == ssort[i].kriterium)
	  ssort[i].kriterium += 1;
	if (    ssort[i].Clause->parsed_rule_type == PQUERY
	    ||  ssort[i].Clause->parsed_rule_type == PRULE
	    ||  ssort[i].Clause->parsed_rule_type == PFACT)
	  ssort[i].kriterium -= 1000;
    }

    /* lexicographisch Ordnung der Klauseln nach Laenge und Komplexitaet */
    for (i = 0; i < Clausanz; i++) {
        for (j = i; j < Clausanz; j++) {
	    length_i = ssort[i].Clause->nr_lits - ssort[i].Clause->nr_built;
	    length_j = ssort[j].Clause->nr_lits - ssort[j].Clause->nr_built;
	    if (length_j < length_i)
	        tausche(ssort,i,j);
	    else {
	        if (    length_j == length_i
		    &&  ssort[j].kriterium < ssort[i].kriterium)
		  tausche(ssort,i,j);
	    }
	}
    }
}


/****************************************************************************/
/* make_sortarray()
/****************************************************************************/
/* allocate and fill an array with pointer in every clause		    */
/*  and add two static criteria: ->kriterium ->varanz			    */
int make_sortarray(Gesamtstruktur)
claustype *Gesamtstruktur;
{
int i;
claustype *hpClause;


if (matrix->m_list) {
	nferror("no sortarray for modules",0);
	return (1);
	}
if (!ssort) {
	/* Speicher fuer Sortierarray allocieren */
	if((ssort = (sortarray *)malloc(sizeof(sortarray) * nr_clauses))==NULL){
		nferror("no memory for sortarray",0);
		return (1);
		}
	}
else
	/*      array already built up 				  */
	/* may be second call by main.c before orbranch_reorder() */
	return (0);

count_connections(Gesamtstruktur);

hsort = ssort;
hpClause = Gesamtstruktur;
	
while ( hpClause )
	{
	hsort->Clause = hpClause;
	hsort->kriterium = BerechneSortierkriterium( hpClause );

/*
	printf("length:%d\n",hpClause->nr_lits);
	printf("complexity:%f\n",hsort->kriterium);
	printf("vars:%d\n\n",hsort->varanz);
*/

	hpClause = hpClause->next;	
	hsort++;
	}

/* Sortieren des Feldes */

 SortiereNachKriterium( ssort, nr_clauses );	


/* die Reihenfolge im lop File wird nach der hier berechneten */
/* Ordnung umsortiert JB				      */
if (inwasm_flags.clausreord) {

if (VERB(9)) {
	   printf("\n  i clnr   nrlits  kriterium  ordnung\n");
	   for ( i = 0; i < nr_clauses; i++) 
		printf("%3d %3d    %3d     %f	%d\n",i,ssort[i].Clause->clnr,
				ssort[i].Clause->nr_lits,ssort[i].kriterium,
				ssort[i].Clause->order);
	   }

	for ( i = 0; i < nr_clauses-1; i++) 
		ssort[i].Clause->next = ssort[i+1].Clause;
	if (nr_clauses > 0)
		ssort[nr_clauses-1].Clause->next = NULL;
	matrix->c_list->next = ssort[0].Clause; 
	}

return (0);
}

/****************************************************************************/
/* orbranch_reorder()
/****************************************************************************/
/* reorder the connections of every predicate				    */
/*  use clause->order							    */
orbranch_reorder(clause)
claustype * clause;
{
predtype * aktpred;
predtype * help;
predlist * connections;
predlist * hconnect;

/* go through all clauses, all literals, all connections lists */
while (clause) {
	aktpred = clause->p_list;

	while (aktpred) {
		connections = aktpred->wu;

		while (connections) {
			hconnect = connections->next;

			while (hconnect) {
				/* now compare */
				if (connections->pr->claus->order >
				    hconnect->pr->claus->order) {
					  /* change order */
				      help = connections->pr;
				      connections->pr = hconnect->pr;
				      hconnect->pr = help;
				      }

				hconnect = hconnect->next;
				}

			connections = connections->next;
			}

		aktpred = aktpred->next;
		}

	clause = clause->next;
	}
}


/*----------------------------------------------------------------------------*/
/* KriteriumEinesKontrapositivs					              */
/*----------------------------------------------------------------------------*/

float KriteriumEinesKontrapositivs( cp )
	claustype *cp;

{
	predtype *head;
	predtype *pred;
	symbollist *vlist;
	float erg;
        float old;

	head = cp->p_list;
	erg = 0.0;
	
	pred = head->next;
	head->var_dependency=0;
	while ( pred != NULL )
	  {
	    vlist = get_all_var(pred->t_list);
	    pred->var_dependency  = nr_var_in_lit(head->t_list,vlist);	
	    head->var_dependency += pred->var_dependency;
	    pred = pred->next;
	  }

	pred=head->next;
	while ( pred != NULL )
	  {
	    old=pred->complexity;
	    pred->complexity = 
	      sgreord_at * pred->complexity
	      + sgreord_av * (float) pred->var_dependency / (float) (head->var_dependency + 1)
	      + sgreord_ac * ( 1.0 - (float) pred->nr_connections / (float) (cp->connections + 1))
	      ;
	    erg+=pred->complexity;
	    
	    pred = pred->next;
	  }
	cp->complexity = erg;
	
	return ( erg );
}

/*----------------------------------------------------------------------------*/
/* gen_subgoal_weights()                                                      */
/*----------------------------------------------------------------------------*/
gen_subgoal_weights(cl)
 claustype *cl;
{
 claustype *hcl;
 claustype *hcp;
 float erg;

 hcl = cl;
 erg = 0.0;

 while (hcl) 
 	{
	erg += KriteriumEinesKontrapositivs(hcl);
	hcp = hcl->fann;
	while(hcp)
		{
		erg += KriteriumEinesKontrapositivs(hcp);
		hcp = hcp->next;
		}
        hcl = hcl->next;
	}
}
/*######################################################################*/

/*---------------------- count_connections ----------------------------*/

void count_connections (cll)

claustype *cll;
{
   predtype *pp;
   predlist *wu;

/* 12.6.97
   if (cll)
   	cll->connections = 0;
	*/
   while (cll)
   {
      cll->connections = 0;    /* 12.6.97 */
      for (pp = cll->p_list; pp; pp = pp->next) 
        {
          pp->nr_connections = 0;
          for (wu = pp->wu; wu; wu = wu->next)
                 pp->nr_connections += 1;
          cll->connections += pp->nr_connections;
      /* printf("\n\tpp%d",pp->nr_connections); */
        }
      /* printf("\n\tclause%d",cll->connections); */
      cll = cll->next;
   }
}

/*---------------------- nr_var_in_lit -----------------------------------*/

   /*
    Liefert f"ur tlist, die Anzahl von Vorkommen von Variablen aus vlist.
   */

nr_var_in_lit(tlist,vlist)
termtype *tlist;
symbollist *vlist;
{
   symbollist *hlpvlist;

   if (! tlist || ! vlist) return(0);
   else if (tlist->type == VARIABLE || tlist->type == SVARIABLE)
         {
           hlpvlist = vlist;
           while (hlpvlist)
            {
              if (hlpvlist->symbol == tlist->symb)
               return(1 + nr_var_in_lit(tlist->next,vlist));

              hlpvlist = hlpvlist->next;
            }
            return(nr_var_in_lit(tlist->next,vlist));
         }
   else  return(nr_var_in_lit(tlist->next,vlist) +
                nr_var_in_lit(tlist->t_list,vlist));
}

