/**************************************************************************
;;; MODULE NAME : connopt.c
;;;
;;; PARENT      : main.c
;;;
;;; PROJECT     : MPLOP
;;;
;;; SCCS        : @(#)connopt.c	15.1 01 Apr 1996
;;;
;;; AUTHOR      : Mohamed Salah, Hamdi
;;;
;;; DATE        : 10.2.93 
;;;
;;; DESCRIPTION : delete connections that produce taut. or subs. resolvents 
;;;
;;; REMARKS     :
;;;             30.12.93	Hamdi	extern.h
;;;		02.03.94	Hamdi	DELCONN,RDELCONN
;;;             14.06.95        PJ      do_res: reject some subsumptions
;;;
***************************************************************************/

/* Notes:

  ssort: 0..n - lexical index to clauses

*/


/*********************** INCLUDE ******************************************/

#include "extern.h" /* symb.h: types.h, defines.h, symbtab ... */

/*********************** DEFINE *******************************************/

#define TRUE 1
#define FALSE 0
#define ALL 1         /* modes fuer rename */
#define ALL_TO_SV 2   /* siehe my_rename() in  sconstr.c */
#define ONLY_SV 3
#define TO_OLD 4
#define SV_TO_V 5
#define NORMALC 0        /* vgl. sconstr.c */ 
#define NOOPT (-100)     /* zur Markierung eines Literals */ 
#define DELCONN 1   /* modes fuer Konnektionsoptimierung      */
#define RDELCONN 2  /* DELCONN: alle Klauseln vergleichen     */
		    /* RDELCONN: nur Regeln vergleichen       */

/*********************** EXTERNE DATEN ************************************/

extern sortarray *ssort;                /* sort.c   */

/*********************** EXTERNE FUNKTIONEN *******************************/

extern termtype *copy_terms();          /* hlpfkt.c */
extern predtype *ErzeugeConstr();       /* hlpfkt.c */
extern void my_rename();                /* sconstr.c  */
extern predtype *get_red_pred();        /* oconstr.c */
extern termtype *replace_in_term();     /* oconstr.c */
extern FindeTautologie();               /* tconstr.c */
extern FindeSubs();               	/* sconstr.c */
extern delete_p();                      /* purity.c */

/*********************** GLOBALE DATEN ************************************/

int nr_del_conn = 0; /* Anzahl der durch diese Optimierung 
                        gel"oschten Konnektionen.  
                     */

int nr_resolvents;    /* >= nr_clauses! */

claustype *rll = NULL;
/*********************** PROTOTYPEN ***************************************/

predtype *ren_get_red_pred();
claustype *do_resolution();
predtype *get_other_conn_lit();

/*********************** FUNKTIONEN ***************************************/


/*---------------------- opt_connections ---------------------------------*/

   /* Hauptprogramm zur Elimination von Konnectionen, die tautologische
      und subsumierbare Resolventen erzeugen.
      Return-Wert: gesamte Anzahl der durchgef"uhrten Resolutionen. 
      Nebeneffekt: die globale Var. `nr_del_conn' enth"alt die Anzahl
                   der durch diese Optimierung gel"oschten Konnektionen.
   */

void opt_connections(cll,compare_mode)
claustype *cll;
int compare_mode;
{
  claustype *resolvente,*hp1,*hp2;
  int erg = 0;
  int maxlength, reslength, i,j;

  subs_unif_counter = 0;
  subs_max_counter = 10000;

  printf("  Deleted: ");
  /*******************/
  /* Initialisierung */
  /*******************/
  nr_del_conn = 0;
  nr_resolvents = nr_clauses;

  /*******************************/
  /* Default query "uberspringen */
  /*******************************/
  hp1 = cll->next;
     
  maxlength = 0;
  while(hp1)
    {
     if (hp1->nr_lits > maxlength)
        maxlength = hp1->nr_lits;
	hp1 = hp1->next;
    }

  for (reslength = 0; reslength <= maxlength * 2 - 2; reslength++) {
      /*********************************************************/
      /* resolviere alle Tail-Literale mit allen Headliteralen */
      /*********************************************************/
      for (i = 0; i < nr_clauses &&
	             subs_unif_counter <= subs_max_counter; i++) {
      	  hp1 = ssort[i].Clause;
          for (j = 0; j < nr_clauses && 
	                 subs_unif_counter <= subs_max_counter; j++) { 
	      hp2 = ssort[j].Clause;
              if (hp1->nr_lits + hp2->nr_lits - 2 == reslength)
                 erg += opt_clause_conn(hp1,hp2,cll,compare_mode);
          }
      }
  }
  printf("\n");
  nr_links_tested = erg;
}



/*---------------------- pos_lits -----------------------------------------*/


int pos_lits(clause)

claustype *clause;
{
 predtype *pp;
 int i;

 for (i = 0,pp = clause->p_list; pp; pp = pp->next) {
     if (pp->sign == '+')
        i += 1; 
 }
 return(i);
}




/*---------------------- opt_clause_conn ---------------------------------*/
                               
int opt_clause_conn(clause1,clause2,cll,compare_mode)

 claustype *clause1,*clause2;
 claustype *cll; 	 
 int compare_mode; 
		   
{
   predtype *pp,*conn_pp,*hpp;
   claustype *resolvente;
   int erg = 0;
   /***********************************************************/
   /* RDELCONN: nur Konnektionen zwischen Regeln "uberpr"ufen */ 
   /***********************************************************/
   if (compare_mode == RDELCONN &&
       (clause1->nr_lits == 1 || clause2->nr_lits == 1))
      return(erg);

   if (pos_lits(clause1) > 1 || pos_lits(clause2) > 1)
      return(erg);

   /***********************************************/
   /* Optimiere Konnektionen clause1 <--> clause2 */
   /***********************************************/
   pp = clause1->p_list;
   while (pp)
    {
      conn_pp = clause2->p_list;
      while (conn_pp) {

         /*********************/
	 /* Resolvente bilden */
         /*********************/
	 if (conn_pp->sign == '+' && pp->sign == '-' 
	     && pp->symb == conn_pp->symb
	     && subs_unif_counter < subs_max_counter)
               resolvente = do_resolution(clause1,clause2,pp,conn_pp);
            else 
	       resolvente = NULL; 

         if (resolvente)
          {

            erg++;
/*
            printf("\n############## (%d.%d -> %d.%d) #############\n",
                 clause1->clnr,pp->litnr,clause2->clnr,conn_pp->litnr); 
            print_lop_clause(clause1,clause1->p_list,
	                     clause1->delpos,0,0);
            print_lop_clause(clause2,clause2->p_list,
	                     clause2->delpos,0,0);
            printf("\n---------------------------------------------\n");
            print_lop_clause(resolvente,resolvente->p_list,
	                     resolvente->delpos,0,0);
	    printf("\n ");
*/



	    /*********************/
            /* Resolvente testen */
	    /*********************/
            if  (FindeTautologie(resolvente) == 1 ||
                 is_subsumable(cll,resolvente) || 
                 is_subsumable(rll,resolvente))
             {
	        /***************************************/
                /* Konnektion (pp -- conn_pp) l"oschen */
	        /***************************************/
                delete_p(conn_pp,pp);
                delete_p(pp,conn_pp);
		pp->missing_link = 1;
		conn_pp->missing_link = 1;
                nr_del_conn++ ;
                printf("(%d.%d %d.%d) ", 
		        clause1->clnr,pp->litnr,clause2->clnr,conn_pp->litnr);
/*
                fprintf(del_file,"%d %d %d %d\n", 
		        clause1->clnr,pp->litnr,clause2->clnr,conn_pp->litnr);
                fprintf(del_file,"%d %d %d %d\n", 
			clause2->clnr,conn_pp->litnr,clause1->clnr,pp->litnr);
*/

/*
                printf("\n================================================\n ");
*/
                if (nr_del_conn % 5 == 0) 
		   printf("\n\t");
             }
	     else 
	     {
	       /************************/
               /* Resolvente einfuegen */
	       /************************/
	       resolvente->next = rll;
	       rll = resolvente;
	     }
          }

	 conn_pp = conn_pp->next;
        
       } 

      pp = pp->next;
    }
   return(erg);
}


/*---------------------- is_subsumable -----------------------------------*/

   /* pr"uft ob clause von einer Klausel der Liste cll subsumiert wird 
      Return-Wert:  JA  : TRUE
                    NEIN: FALSE
   */

int is_subsumable(cll,clause)
claustype *cll, *clause;
{
   int rv;


   while (cll)
    {
     if (cll->clnr && 
         cll->parsed_rule_type != PRULE && 
	 cll->parsed_rule_type != PQUERY && 
         clause->parsed_rule_type != PRULE && 
	 clause->parsed_rule_type != PQUERY)
       {
        rv = FindeSubs(cll,clause);
        if (rv == 1) 
          {
/*
           printf("\n is subsumed by\n ");
           print_lop_clause(cll,cll->p_list, cll->delpos,0,0);
	   printf("\n ");
*/



           /* don't return TRUE if the subsuming resolvent and
	      the subsumed resolvent have the same parent clauses and 
	      if the resolvent is not a self resolvent */
           if (!(
	         cll->clnr      >= nr_clauses     && /* cll is resolvent */
		 clause->clnr   >= nr_clauses     && /* clause is resolvent */ 
		 clause->rescl1 != clause->rescl2 && /* different parents */
		 cll->rescl1    == clause->rescl1 && 
		 cll->rescl2    == clause->rescl2 )) 
	      return(TRUE);
          }
       }   

     cll = cll->next;
    }

   return(FALSE); 
}



/*---------------------- do_resolution -----------------------------------*/

   /* f"uhrt die Resolution von cl1 und cl2 "uber die Literale 
      lit1 (aus cl1) und lit2 (aus cl2) durch.

      Return-Wert: - Resolvente.
                   - NULL: falls Resolution nicht m"oglich. 
		   
   */
/*---------------------- do_resolution -----------------------------------*/

claustype *do_resolution(cl1,cl2,lit1,lit2)

 claustype *cl1,*cl2;
 predtype *lit1,*lit2;
{

   predtype *orglit1, *orglit2, *hpp, *newlit; 

   termtype *term1,*term2;
   predtype *sigma; 		/* Mgu */
   claustype *erg = NULL; 	/* Ergebnis */
   predtype *litlist = NULL; 	/* Liste der Literale von erg */
   int i = 1; 			/* f"ur Numerierung der Literale in erg */ 

   /*******************/
   /* Initialisierung */
   /*******************/

   if (!lit1 || !lit2) 
    return(NULL);

   if ((lit1->symb != lit2->symb) || (lit1->sign == lit2->sign))
    return(NULL);

   /****************************/
   /* lit1 ist Literal von cl1 */
   /* lit2 ist Literal von cl2 */
   /****************************/
   if (lit1->claus != cl1 || lit2->claus != cl2) 
    return(NULL);

   /*********************************************/
   /* Argumentlisten von lit1 und lit2 kopieren */
   /*********************************************/
   term1 =  copy_terms(lit1->t_list,ALL_TERM);
   term2 =  copy_terms(lit2->t_list,ALL_TERM);
   my_rename(term2,cl2->clnr,ALL);

   /**************************************/
   /* Allgemeinsten Unifikator berechnen */
   /**************************************/
   rearrange();
   if (unify(term1,term2) != SUCCESS)
      return(NULL);
   else
      sigma = ErzeugeConstr(cl1);
/*
    print_lop_constr(sigma);
*/

   /********************************************************************/
   /* Resolvente konstruieren: sigma( cl1 - lit1) + sigma( cl2 - lit2) */ 
   /********************************************************************/

   erg = (claustype *) malloc(sizeof(claustype));

   /*****************************************************/

   /*********************************/
   /* sigma( cl1 - lit1 ) berechnen */
   /*********************************/
   orglit1 = cl1->p_list;
   while (orglit1)
    {
      /***************************************/
      /* lit1 und query nicht in Resolvente! */
      /***************************************/
      if (orglit1 != lit1 && orglit1->symb)
       {
	  newlit = get_red_pred(orglit1,sigma,erg);

          if (!litlist) litlist = newlit;
          else 
           {
             hpp = litlist;
             while (hpp->next)
              hpp = hpp->next;
             hpp->next = newlit;
           }
           
       }

      orglit1 = orglit1->next;
    }
   
   /*********************************/
   /* sigma( cl2 - lit2 ) berechnen */
   /*********************************/
   orglit2 = cl2->p_list;
   while (orglit2)
    {
      /***************************************/
      /* lit2 und query nicht in Resolvente! */
      /***************************************/
      if (orglit2 != lit2 && orglit2->symb)
       {
          newlit = ren_get_red_pred(orglit2,i++,sigma, erg,cl2->clnr);
          if (!litlist) litlist = newlit;
          else 
           {
             hpp = litlist;
             while (hpp->next)
              hpp = hpp->next;
             hpp->next = newlit;
           }
           
       }

      orglit2 = orglit2->next;
    }

   /**************************************************/
   /* falls cl1 oder cl2 QUERY, dann Resolvente auch */
   /**************************************************/
   if (cl1->p_list->symb == 0 || cl2->p_list->symb == 0) 
    {
       /*****************/
       /* query-Literal */
       /*****************/
       hpp = (predtype *) malloc(sizeof(predtype));
       hpp->symb  = 0;
       hpp->sign  = 0;
       hpp->arity = 0;
       hpp->litnr = 1;
       hpp->claus = erg;
       hpp->wu    = NULL;
       hpp->next  = litlist;

       litlist = hpp;
       /*************************************/
       /* litnr in litlist verschieben sich */
       /*************************************/
       hpp = litlist->next;
       while (hpp)
        {
          hpp->litnr++; 
          hpp = hpp->next;
        }
       
       i += 1;
    }
   /*******************/ 
   /* leere Resolvente */
   /*******************/ 
   if (!litlist) 
    {
     warning("","Preprocessing derived empty clause!"); 
     return(NULL);
    }

   /************/
   /* Ergebnis */          /* ATTN: Klaus - reduziere Size - eigene struct */
   /************/
   erg->clnr     = ++nr_resolvents; /* PJ - darf ich ? */
   erg->rescl1   = cl1->clnr;       /* PJ: speichere clnr der Clauses fuer */
   erg->rescl2   = cl2->clnr;       /*     diese Resolvente */
   erg->modul    = cl1->modul;
   erg->lineno   = 0;
   erg->constr   = NULL; 
   erg->fann     = NULL;
   erg->next     = NULL;
   erg->p_list   = litlist;
   erg->delpos   = PRULE;
   erg->ddepth   = 2;
   erg->dinf     = cl1->nr_lits  + cl2->nr_lits - 2;
   erg->nr_built = cl1->nr_built + cl2->nr_built;
   erg->nr_lits  = cl1->nr_lits  + cl2->nr_lits - 2;

   return(erg);
}


/*---------------------- ren_get_red_pred --------------------------------*/
 
   /* liefert eine Kopie von pr mit:
      		- Umbenennung der Variablen: V --> V<ren_nr> und
      		- Anwendung von sigma = constr auf die Argumentliste. 

      Wie get_red_pred() aber mit renaming!
   */ 

/*---------------------- ren_get_red_pred --------------------------------*/

predtype *ren_get_red_pred(pr,nr,constr,clause,ren_nr)

 predtype *pr;
 int nr;
 predtype *constr;
 claustype *clause;
 int ren_nr;
{
  predtype *erg;
  termtype *term,*left,*right;

  erg = (predtype *) malloc(sizeof(predtype));
  erg->symb  = pr->symb;
  erg->sign  = pr->sign;
  erg->arity = pr->arity;
  erg->litnr = nr;
  erg->claus = clause;
  erg->wu    = NULL;
  erg->next  = NULL;

  /***********************/
  /* Berechne erg->t_list */
  /***********************/

  term = copy_terms(pr->t_list,ALL_TERM);

  my_rename(term,ren_nr,ALL);

  if (constr)
   {
     /*******************************************/
     /* linke und rechte Seite der Substitution */
     /*******************************************/
     left = constr->t_list->t_list;
     right = constr->t_list->next->t_list;

     /*************/
     /* einsetzen */
     /*************/
     while (left && right)
      {
         term = replace_in_term(term,left,right);

         left  = left->next;
         right = right->next;
      }
   }

  erg->t_list = term;

  return(erg);
}

/*---------------------- get_other_conn_lit ------------------------------*/

   /* Liefert den einzigen Konnektionspartner von lit in clause.
      Falls lit mehrere Konnektionspartner hat dann NULL.
   */

/*---------------------- get_other_conn_lit ------------------------------*/

predtype *get_other_conn_lit(lit,clause)

 predtype *lit;
 claustype *clause;
{
  
   predlist *wul;
   int nr = 0;
   predtype *erg = NULL;

   wul = lit->wu;
   while (wul)
    {
       if (wul->pr->claus == clause)
        {
           nr++;
           erg = wul->pr;
        }

       wul = wul->next;
    }

   /* nur 1 Konnektionspartner ? */ 
   if (nr == 1) 
    return(erg);
 
   return(NULL);
}



/*########################################################################*/
