/*----------------------------------------------------------------------------*/
/* Programm: FindConstraints                                                  */
/* Datei:    Allgemein benutzte Funktionen                                    */
/* Autor:    Martin Ziegenaus                                                 */
/* Datum:    13.09.91                                                         */
/* SCCS:     @(#)hlpfkt.c	15.2 10 May 1996
/* Modif:    12.11.92 hlpfkt.c  hamdi                                         */
/*           12.01.94 extern.h  hamdi                                         */
/*----------------------------------------------------------------------------*/


#include "extern.h"

#define ONE_TERM 0
#define ALL_TERM 1
#define ALL_ARGS 2


/*----------------------------------------------------------------------------*/
/* externe Funktionen													      */
/*----------------------------------------------------------------------------*/
extern termtype *init_term();	/* init.c   Janos Breiteneicher   */
extern termtype *deref();	/* unify.c       - " - 		  */



/*----------------------------------------------------------------------------*/
/* Funktionen 								      */
/*----------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------*/
/* FuegeConstraintHinzu							      */
/*----------------------------------------------------------------------------*/
int FuegeConstraintHinzu( Constraint, Clausel)
	predtype *Constraint;
	claustype *Clausel;
	/* Der uebergebene Constraint wird hinten an die Constraints-Liste
		der uebergebenen Klausel angefuegt. */
{
	predtype *hpc; 	/* zum Durchlaufen */


if ( ! Clausel->constr )
	{
	Clausel->constr = Constraint;
	return(0);
	}

hpc = Clausel->constr;

while (hpc->next)
	hpc = hpc->next;

hpc->next = Constraint;

return(0);
}


/*----------------------------------------------------------------------------*/
/* KopiereTerm															      */
/*----------------------------------------------------------------------------*/
int KopiereTerm( ziel, quelle )
	termtype *ziel;
	termtype *quelle;
	/* Kopiert einen Term.
		Nur der Term selber, keine t_list- und next-Verweise, wird kopiert.
		Also nur bei Variablen verwenden !
		Eingabe: zwei Pointer auf Ziel und Quelle.
			Achtung: Ziel muss schon allociert sein.
		Ausgabe : 0 alles ok */
{
ziel->symb = quelle->symb;
ziel->type = quelle->type;
ziel->value = quelle->value;
ziel->grflag = quelle->grflag;
ziel->t_list = NULL;
ziel->next = NULL;

return( 0 );
}


/*************************************************************/
termtype *copy_terms(quelle,mode)
termtype * quelle;
int mode;
{
register termtype *tp;
termtype *htp;

if (!quelle) 
	return (NULL);

htp = quelle;

if (symbtab[quelle->symb].name[0] == '#' && symbtab[quelle->symb].loclist) 
	htp = deref(quelle);

tp	= (termtype *)malloc(sizeof(termtype));
KopiereTerm(tp, htp);

switch (mode) {
case ALL_TERM:
	/* copy all recursivly */
	tp->next = (quelle->next)?
		    copy_terms(quelle->next,mode) : NULL;
	tp->t_list = (htp->t_list)?
		    copy_terms(htp->t_list,ALL_TERM) : NULL;
	break;
case ONE_TERM:
	tp->next = NULL;
	tp->t_list = (htp->t_list)?
		    copy_terms(htp->t_list,ALL_ARGS) : NULL;
	break;
case ALL_ARGS:
	tp->next = (quelle->next)?
		    copy_terms(quelle->next,mode) : NULL;
	tp->t_list = (htp->t_list)?
		    copy_terms(htp->t_list,ALL_ARGS) : NULL;
	break;
default:
	nferror("wrong mode in copy_terms",0);
	}

if (symbtab[tp->symb].name[0] == '#') {
	symbtab[tp->symb].type = SVARIABLE;
	tp->type = SVARIABLE;
	}
return (tp);
}

/*************************************************************/
termtype *deref_copy_terms(quelle,mode)
termtype * quelle;
int mode;
{
register termtype *tp;
termtype *htp;

if (!quelle) 
	return (NULL);

htp = quelle;

	htp = deref(quelle);

tp	= (termtype *)malloc(sizeof(termtype));
KopiereTerm(tp, htp);

switch (mode) {
case ALL_TERM:
	/* copy all recursivly */
	tp->next = (quelle->next)?
		    deref_copy_terms(quelle->next,mode) : NULL;
	tp->t_list = (htp->t_list)?
		    deref_copy_terms(htp->t_list,ALL_TERM) : NULL;
	break;
case ONE_TERM:
	tp->next = NULL;
	tp->t_list = (htp->t_list)?
		    deref_copy_terms(htp->t_list,ALL_ARGS) : NULL;
	break;
case ALL_ARGS:
	tp->next = (quelle->next)?
		    deref_copy_terms(quelle->next,mode) : NULL;
	tp->t_list = (htp->t_list)?
		    deref_copy_terms(htp->t_list,ALL_ARGS) : NULL;
	break;
default:
	nferror("wrong mode in deref_copy_terms",0);
	}

if (symbtab[tp->symb].name[0] == '#') {
	symbtab[tp->symb].type = SVARIABLE;
	tp->type = SVARIABLE;
	}
return (tp);
}
/*************************************************************/
/* ErzeugeConstr()					     */
/*************************************************************/
predtype *ErzeugeConstr(claus)
claustype *claus;
{
trailelt * trp;
termtype * right;
termtype * htp;
predtype * constr = NULL;
termtype * newleft = NULL;
termtype * newright = NULL;
char functsymb[NAMELENGTH]; /* "const"arity_  */
int newarity = 0;
int newsymb;
int varsymb;
int i;

/* no bindings, no constraint */
/* printf("%d ",trailcnt); */
if (!trailcnt) 
 return (NULL);

for (i=trailcnt-1; i >=0; i--) {
	/* take binding from trail make subconstraintlist */
	trp = (trailelt*)symbtab[trail[i].symb].loclist;
	varsymb = trp->symb;
	right = trp->term;

	if (symbtab[varsymb].name[0] == '#') 
	   /* there is a bound stucture variable, no subconstraint */
	   continue;
        else if (symbtab[right->symb].name[0] == '#') 
           /* a variable is bound to a structure variable, no subconstr. */
	   continue;
	else {
	   newarity++;

	   /* enter variable in left side of the constraint */
	   /* only a variable can be bound to a term	    */
	   if (!newleft)
	      newleft  = init_term(NULL,varsymb,VARIABLE,0,NGROUND);
	   else {
	      htp = init_term(NULL,varsymb,VARIABLE,0,NGROUND);
	      htp->next = newleft;
	      newleft = htp;
	      }

	   /* enter term in right side of the constraint */
	   if (!newright)
	      newright = deref_copy_terms(right,ONE_TERM);
	   else {
	      htp = deref_copy_terms(right,ONE_TERM);
	      htp->next = newright;
	      newright = htp;
	      }
	   }
	}
	
if (!newarity) 
   /* no bindings of non-struct-vars, no constraint */
   {
   /* printf(" %d\n",newarity); */
   return (NULL);
   }

/* create a constraintsymbol  const newarity _    */
sprintf(functsymb,"const%d_",newarity);
newsymb = enter_token(functsymb);

/* generate constraint newleft =/= newright */
constr = (predtype *)malloc(sizeof(predtype));

constr->arity	= 2;
constr->symb	= NQAL_SYMB;
constr->claus	= claus;
constr->litnr	= ++nr_constraints;
constr->wu	= NULL;
constr->next 	= NULL;
constr->t_list	= init_term(newleft, newsymb, CONSTSYM,
		0, GROUND);
constr->t_list->next = init_term(newright, newsymb, CONSTSYM,
		0, GROUND);
		nr_gen_constr += 1;
return (constr);
}

/*######################################################################*/
