/**************************************************************************
;;; MODULE NAME : sconstr.c 
;;;
;;; PARENT      : main.c 
;;;
;;; PROJECT     : MPLOP
;;;
;;; SCCS        : @(#)sconstr.c	15.2 10 May 1996
;;;
;;; AUTHOR      : Mohamed Salah, Hamdi 
;;;
;;; DATE        : 15.11.92 
;;;
;;; DESCRIPTION : generate subsumption constraints 
;;;
;;; REMARKS     :
;;;		30.12.93	Hamdi	extern.h
;;;		25.02.94	Hamdi	SUBS,RSUBS 
;;;		02.03.94	Hamdi	my_rename: if nr < 0
;;;		10.03.94	Hamdi	my_makeref #V --> V
;;;             26.01.95        PJ      bvstack init removed
;;;             02.02.95        PJ      merged my_unify into unify
;;;             08.02.95        PJ      merged inwasm_rename into my_rename
;;;             07.05.95        PJ      map - attempt at bug fix
***************************************************************************/



/*********************** INCLUDE ******************************************/

#include "extern.h" /* ficonst.h,symb.h,types.h, defines.h, symbtab ... */ 
#include "stypes.h"
#include "defines.h"

/*********************** DEFINE *******************************************/

#define TRUE 1        
#define FALSE 0       
#define NORMALC 0
#define CONTRAC 1
#define DELETED (-9999)
#define SUBS 1   
#define RSUBS 2 
#define BVSTACKSIZE 100
/*********************** EXTERNE DATEN ************************************/

extern sortarray *ssort;		/* sort.c 	*/

/*********************** EXTERNE FUNKTIONEN *******************************/

extern termtype *copy_terms();          /* hlpfkt.c 	*/
extern predtype *ErzeugeConstr();       /* hlpfkt.c 	*/
extern termtype *deref();               /* unify.c    	*/
extern void init_trail();               /* unify.c    	*/
extern void rearrange();                /* unify.c    	*/
extern enter_name();                    /* symb.c  	*/
extern enter_token();                   /* symb.c 	*/
extern unify();                         /* unify.c    	*/
extern littype *sym_make_cart_list();   /* symconstr.c 	*/
extern delete_p();                      /* purity.c 	*/
extern del_occlist();                   /* purity.c 	*/
extern int nr_bindings();	        /* constrsu.c   */

/*********************** GLOBALE DATEN ************************************/

int no_subsumption = FALSE;
int subsumption = FALSE;

static int is_subsumption = FALSE;
static predtype *constr_list = NULL;
static symbollist *bvstack[BVSTACKSIZE];

static int bvsptr = 0;               /* Stackpointer */
static int abbruch = FALSE;
static int var_counter = 0;          /* Zaehler zur Umbenennung von Var. */

/*********************** PROTOTYPEN ***************************************/

littype *make_cart_list();
plittype *find_slist();
predtype *add_to_list();
void undo_unify();
void add_clist();
void ren_constr();
void my_rename();
symbollist *app_sym_to_list();
void map_all();
void map_all_and_genconstr();

/*********************** FUNKTIONEN ***************************************/



/*---------------------- gen_s_constr ------------------------------------*/

   /* 
   	Generiert Subsumptions Constraints 
   */
/*---------------------- gen_s_constr ------------------------------------*/

claustype *gen_s_constr(Gesamtstruktur,compare_mode)

claustype *Gesamtstruktur;
 int compare_mode; 
{
   int i,j;
   int rv;
   claustype *cll = Gesamtstruktur;
   symbollist *dellist = NULL;

   nr_gen_constr = 0;
   subs_unif_counter = 0;
   subs_max_counter = 10000;

   if (! cll)    
    /****************/
    /* Eingabe leer */ 
    /****************/
    return(cll);

   if (! cll->p_list->symb) 
      /***********************************/
      /* erste Klausel ist default query */
      /***********************************/
    cll = cll->next;


   /********************************************/
   /* jede Klausel mit allen hinteren Klauseln */
   /********************************************/
   for (i = 0; i < nr_clauses 
                 && subs_unif_counter < subs_max_counter; i++)
     {
       if (ssort[i].Clause->del != DELETED && 
	   ssort[i].Clause->parsed_rule_type != PRULE &&
	   ssort[i].Clause->parsed_rule_type != PQUERY) 
        {
           for (j = i+1; j < nr_clauses 
	   		   && subs_unif_counter < subs_max_counter; j++)
            {
              if (ssort[j].Clause->del != DELETED &&
	          ssort[j].Clause->parsed_rule_type != PRULE &&
	          ssort[j].Clause->parsed_rule_type != PQUERY)
               {
	         rv = clause_s_constr(ssort[i].Clause, 
                                      ssort[j].Clause,NORMALC,compare_mode);
                 /*
                 result_msg(ssort[i].Clause->clnr,ssort[j].Clause->clnr,rv);
                 */

                 if (rv == 1) 
		    /***************/
		    /* Subsumption */ 
		    /***************/
                  {
                    dellist = app_sym_to_list(dellist,ssort[j].Clause->clnr);
                    ssort[j].Clause->del = DELETED;  
                  }

               }
            }
        }
     }

/************************************/
/* print Liste der DELETED-Klauseln */ 
/************************************/
printf("  Generated constraints: %d\n", nr_gen_constr);
if (dellist)
 print_del_clauses(dellist);

return(Gesamtstruktur);

}


/*---------------------- clause_s_constr ---------------------------------*/

   /*
      langeClausel kommt (im sortierten Feld) nach kurzeClausel vor.
      Return-Wert:
         3 : illegal mode !
	 2 : kurzeClausel ist leer !
	 1 : langeClausel wird von kurzeClausel subsumiert !
	-1 : Die s_liste eines Literals von kurzeClausel ist leer.
	     Subsumption ist NICHT moeglich.
	     Also keine Constraints fuer langeClausel !
	-2 : Subsumption ist NICHT moeglich.
	     Also keine Constraints fuer langeClausel !
	-3 : Kein Vergleich noetig, da eine der Klauseln ein Fakt ist.
	     Gilt nur falls compare_mode = RSUBS !
	 0 : Es wurde eine Constraintliste berechnet und an langeClausel
	     geheangt.
   */

/*---------------------- clause_s_constr ---------------------------------*/

int clause_s_constr(kurzeClausel,langeClausel,mode,compare_mode)

 claustype *kurzeClausel, *langeClausel;
 int mode;
 int compare_mode; 
{
   predtype *kpp, *lpp;   /* zum Durchlaufen der Klauseln */
   littype *cart_list;    /* zur Literalzuordnung. */


   kpp = kurzeClausel->p_list;
   lpp = langeClausel->p_list;
   /********************************************/
   /* Wenn RSUBS, dann nur Regeln vergleichen  */
   /********************************************/
   if (   compare_mode == RSUBS 
       && (   kurzeClausel->nr_lits - kurzeClausel->nr_built == 1 
	   || (   kurzeClausel->nr_lits - kurzeClausel->nr_built == 2 
	       && !kurzeClausel->p_list->symb)))
    return(-3);
   /*******************************/ 
   /* default query ueberspringen */
   /*******************************/ 
   if ( ! kpp->symb )
      kpp = kpp->next;
   if ( ! lpp->symb )
      lpp = lpp->next;
   /**************************/
   /* Wenn kurzeClausel leer */
   /**************************/
   if (! kpp)
      return 2;
/*
   printf("kurze Klausel:%d.lange Klausel:%d\n",
	kurzeClausel->clnr,langeClausel->clnr);
*/

   /***********************************************************/
   /* Erstelle Kopie von kurzeClausel mit Umbenennung der     */
   /* Variablen und Berechnung von s_list fuer jedes Literal. */
   /***********************************************************/
   if (! trail)
      init_trail();
   no_subsumption = FALSE;
   if (mode == NORMALC)
      /*******************************************/
      /* verschiedene Klauseln werden verglichen */
      /*******************************************/
    cart_list = make_cart_list(kpp,kurzeClausel->clnr,lpp);
   else if (mode == CONTRAC)
           /*********************************************************/
           /* Kontrapositive der gleichen Klausel werden verglichen */
           /*********************************************************/
         cart_list = sym_make_cart_list(kpp,kurzeClausel->clnr,lpp);
   else  
     return 3; 

   /******************************************************************/
   /* Wenn ein Literal keine Partner-Literale hat: keine Subsumption */
   /******************************************************************/
   if (no_subsumption)
      {
	 return -1;
      }

   /*********************/
   /* Initialisierungen */
   /*********************/
   bvsptr = 0;
   bvstack[bvsptr]=NULL; 
   rearrange();

   constr_list = NULL;
   is_subsumption = FALSE;
   abbruch = FALSE;

   /********************************************/
   /* Zaehle alle Zuordnungsmoeglichkeiten auf */
   /********************************************/

   map_and_genconstr(cart_list);

   if (is_subsumption)       
      /****************************************/
      /* kurzeClausel subsumiert langeClausel */
      /****************************************/
      {
	 return 1;
      }
   else 
     if (! constr_list)   
        /************************/
        /* Constraintliste leer */
        /************************/
	   return -2;
     else {
          /************************************************/
          /* wenn nicht: ueberall Constraintart eintragen */
          /************************************************/
          if (mode == NORMALC)
             type_of_constr(constr_list,SUB_CONSTR);
          else  
             type_of_constr(constr_list,SYMM_CONSTR);

       add_clist(constr_list,langeClausel);
       return 0;
     }

}


/*---------------------- make_cart_list ----------------------------------*/

   /* 
      Erstellt Kopie von kpp mit Umbenennung der Variablen
      und berechnet dabei die s_list fuer jedes Literal.
      Return-Wert: Erstellte Kopie.
   */
/*---------------------- make_cart_list ----------------------------------*/


littype *make_cart_list(kpp,clnr,lpp)

 predtype *kpp;
 int clnr;
 predtype *lpp;
{
   littype *litlist,*newlit,*helplit;

   litlist = NULL;
   while (kpp)
     /*****************************************/
     /* f"ur alle Literale der kurzen Klausel */
     /*****************************************/
     {
	if (kpp->symb >= nrbuilt) {
	newlit = (littype *) malloc(sizeof(littype));

	newlit->symb = kpp->symb;
	newlit->sign = kpp->sign;

        /**********************/
	/* Termliste kopieren */
        /**********************/
	newlit->t_list = copy_terms(kpp->t_list,ALL_TERM);
        /***************************************************/
	/* Variablen der Kopie werden zu Strukturvariabeln */
        /***************************************************/
	my_rename(newlit->t_list,clnr,ALL_TO_SV);

        /*************************/
	/* Berechnung von s_list */
        /*************************/
	newlit->s_list = find_slist(newlit,lpp);
	if (! newlit->s_list)
	   {
	    /***********************************************/
	    /* Wenn ein Literal keine Partnerliterale hat. */
	    /***********************************************/
	     no_subsumption = TRUE;
	     return(litlist);
	   }

	newlit->next = NULL;
        /**************************************/
	/* newlit hinten an litlist anhaengen */
        /**************************************/
	if (! litlist)
	   litlist = newlit;
	else {
		helplit = litlist;
		while (helplit->next)
		      helplit = helplit->next;
		helplit->next = newlit;
	     }
	}
	kpp = kpp->next;
     }

   return(litlist);
}


/*---------------------- find_slist --------------------------------------*/

   /* 
       liefert die Liste von Partner-Literalen von lit in list.
   */
/*---------------------- find_slist --------------------------------------*/

plittype *find_slist(lit,list)

 littype *lit; 
 predtype *list;
{
   predtype *helplist;
   plittype *erglist,*newerg,*hptr;

   erglist = NULL;
   helplist = list;
   /*********************************************/
   /* fuer gegebenes Literal der kurzen Klausel */
   /*********************************************/
   
   while (helplist)
      /************************************************/
      /* suche Partner-Literale in der langen Klausel */
      /************************************************/
      {
	 rearrange();
         
	 if ((helplist->symb == lit->symb) && (helplist->sign == lit->sign)
	     && (unify(helplist->t_list,lit->t_list) == SUCCESS) )
	    {
	       newerg = (plittype *) malloc(sizeof(plittype));
	       newerg->pr = helplist;
	       newerg->notunif = FALSE;
	       newerg->next = NULL;
               /**************************************/
	       /* newerg hinten an erglist anhaengen */
               /**************************************/
		if (! erglist)
		   erglist = newerg;
		else {
			hptr = erglist;
			while (hptr->next)
			      hptr = hptr->next;
			hptr->next = newerg;
		      }
	    }

	 helplist = helplist->next;
      }

   return(erglist);
}




/*------------------------- map_and_genconstr -------------------------------*/

   /* Zaehlt alle Zuordnungsmoeglichkeiten baumartig auf.
	       - is_subsumption = TRUE : Subsumption liegt vor.
	       - Sonst: Constraintliste in constr_list.
   */
/*------------------------- map_and_genconstr -------------------------------*/

map_and_genconstr(l)      /* init flags (usage flag) */
littype *l;
{
   plittype *plist=l->s_list;
   while (plist) {
         plist->pr->flag=0;
         plist=plist->next;
   }
   is_subsumption = FALSE;
   abbruch = FALSE;
   map_all_and_genconstr(l);
}

void map_all_and_genconstr(l)

littype *l;      /* Literal der kurzen Klausel*/
{
   predtype *constr;
   plittype *plist=l->s_list; /* Liste deren Partner in der langen Klausel*/
   int flag;

   /*****************************************/
   /* Fuer alle Literale der langen Klausel */
   /*****************************************/
   while (plist)
      {
         flag = 0;
	 if (subs_unif_counter > subs_max_counter) {
	    is_subsumption = FALSE;
	    abbruch = TRUE;
	 }
	 if (abbruch) 
	    return;

	 if (l->next == NULL)
         /*************************************************************/
	 /* Wenn alle Literale der kurzen erfolgreich zugeordnet sind */
         /*************************************************************/
	  {
	   if ( plist->pr->flag == 0 &&
	       (flag = 1, new_unify(plist->pr->t_list,l->t_list) == SUCCESS))
             {
	       constr = ErzeugeConstr(plist->pr->claus);
	       if (constr == NULL)  
		{
		  is_subsumption = TRUE;
		  abbruch = TRUE;
		}
	       else 
	        {
		      constr_list = add_to_list(constr,constr_list);
	        }
	     }
	  }
         else if ( plist->pr->flag == 0 &&
	          (flag = 1, new_unify(plist->pr->t_list,l->t_list) == SUCCESS))
	      /******************************/
	      /* naechstes Literal zuordnen */
	      /******************************/
	       {
		 plist->pr->flag=1;
		 map_all_and_genconstr(l->next);
		 plist->pr->flag=0;
	       }

	 /*********************************************/ 
	 /* Wenn alle Zuordnungen erfolgreich waren,  */
	 /* oder die Zuordnung vorzeitig abgebrochen  */
	 /* wurde: Dann mache die Wirkung der letzten */
	 /* Unifikation im Trail rueckgaengig.        */
	 /*********************************************/ 
	 if (flag)
	    undo_unify(); /* backtracking */
	 plist = plist->next;
	 subs_unif_counter += 1;
      } /* while */

}


/*---------------------- new_unify ---------------------------------------*/

int new_unify(t1,t2)

 termtype *t1, *t2;
{
   int rv;
   /* PJ: flag special handling of unification */
   sconstr_makeref_flag=1;
   rv = unify(t1,t2);
   sconstr_makeref_flag=0;
   bvsptr++;              /* Stackpointer inkrementieren */
   return(rv);
}


/*---------------------- makeref_stackmanagement() ----------------------*/

void sconstr_makeref_stackmanagement(symb)
termtype *symb;
{
   symbollist *hp; /* Hilfsvar. zum Durchlaufen von bvstack[bvsptr] */
   symbollist *newsymb;
   /*****************************************************************/
   /* Stackverwaltung: Die Bindungen der letzten Zuordnungen merken */
   /*****************************************************************/


   newsymb = (symbollist *) malloc(sizeof(symbollist));
   newsymb->symbol = symb->symb;
   newsymb->next = NULL;

   /* newsymb hinten an bvstack[bvsptr] anhaengen */
   if (! bvstack[bvsptr])
      bvstack[bvsptr] = newsymb;
   else {
           hp = bvstack[bvsptr];
           while (hp->next)
                 hp = hp->next;
           hp->next = newsymb;
         }
   return;
}

/*---------------------- add_to_list -------------------------------------*/

   /* Fuegt ein Constraint zu einer Constraintliste hinten ein.
      Return-Wert: die neue Constraintliste.
   */
/*---------------------- add_to_list -------------------------------------*/

predtype *add_to_list(constr,list)

 predtype *constr,*list;
{
   predtype *hp;
   ren_constr(constr);

   if (!list)
      list = constr;
   else {
	   hp = list;
	   while (hp->next)
		 hp = hp->next;
	   hp->next = constr;
	}
   return(list);
}


/*---------------------- undo_unify --------------------------------------*/

   /* 
     Loescht Bindungen, die durch das letzte unify gemacht wurden. 
   */

/*---------------------- undo_unify --------------------------------------*/

void undo_unify()
{
   bvsptr--;
   while (bvstack[bvsptr])
      {
	 symbtab[bvstack[bvsptr]->symbol].loclist = NULL;
	 trailcnt--;


	 bvstack[bvsptr] = bvstack[bvsptr]->next;
      }
   return;
}


/*---------------------- add_clist ---------------------------------------*/

   /* Fuegt eine Constraintliste zu einer Klausel ein.
      Die Constraintliste wird vorne eingefuegt.
   */
/*---------------------- add_clist ---------------------------------------*/

void add_clist(clist,clause)

 predtype *clist;
 claustype *clause;
{
   predtype *hp;

   if (clist)
      {
	 hp = clist;
	 while (hp->next)
	       hp = hp->next;
	 hp->next = clause->constr;
	 clause->constr = clist;
       }
}


/*---------------------- ren_constr --------------------------------------*/

   /* Nennt Strukturvariablen auf der rechten Seite des Constraints um.
      #V  ---->  #V_counter
   */
/*---------------------- ren_constr --------------------------------------*/

void ren_constr(constr)

 predtype *constr;
{
   if (constr->t_list)
       {
	  my_rename(constr->t_list->next,var_counter,ONLY_SV);
	  var_counter++;
       }
}


/*---------------------- my_rename ---------------------------------------*/

   /* Benennt je nach mode die Variablen von tlist um.
      mode: ALL      :  V     ---->  V<nr>
	    ALL_TO_SV:  V     ---->  #V<nr>
	    ONLY_SV  :  #V... ---->  #V...<nr>
			In diesem Fall werden nur Strukturvar. umbenannt.
            SV_TO_SV :  #V<..>  ----> V<..>
			In diesem Fall werden nur Strukturvar. umbenannt.
            TO_OLD   :  #V    ---->  V
                        Arbeitet nur dann richtig, wenn davor Aufruf mit
                        ALL_TO_SV.
   */
/*---------------------- my_rename ---------------------------------------*/

void my_rename(tlist,nr,mode)

 termtype *tlist;
 int nr;
 int mode;
{
   char name[80];

   if (! tlist)
      return;
   /* if (tlist->symb == NIL_SYMB)
      return;
   */
   if (tlist->next)
      my_rename(tlist->next,nr,mode);
   if (tlist->t_list)
      my_rename(tlist->t_list,nr,mode);
   if (tlist->type == VARIABLE || tlist->type == SVARIABLE)
      {
	 switch(mode)
	  {
	     case ALL:
	      sprintf(name,"New_%s",symbtab[tlist->symb].name,nr);
              tlist->value = tlist->symb; /* memory */
	      tlist->symb = enter_name(enter_token(name),
                                symbtab[tlist->symb].type,0,0);
	      break;

	     case ALL_TO_SV:
	      sprintf(name,"#%s",symbtab[tlist->symb].name);
              tlist->value = tlist->symb; /* memory */
	      tlist->symb = enter_name(enter_token(name),SVARIABLE,0,0);
              /* tlist->type = SVARIABLE; */  
              tlist->type = symbtab[tlist->symb].type;
	      break;

	     case ONLY_SV:
	      if (symbtab[tlist->symb].name[0] == '#')  /* Strukturvar. */
	       {
		 sprintf(name,"%s_%d",symbtab[tlist->symb].name,nr);
                 tlist->value = tlist->symb; /* memory */
		 tlist->symb = enter_name(enter_token(name),SVARIABLE,0,0);
	       }
	      break;

	     case SV_TO_V:
	      if (symbtab[tlist->symb].name[0] == '#')  /* Strukturvar. */
	       {
		 sprintf(name,"NEW__%s",&(symbtab[tlist->symb].name[1]));
		 tlist->symb = enter_name(enter_token(name),VARIABLE,0,0);
                 tlist->type = symbtab[tlist->symb].type;
	       }
	      break;
	      
	     case SV_TO_XV:
	      if (symbtab[tlist->symb].name[0] == '#')  /* Strukturvar. */
	       {
		 sprintf(name,"*%s",&(symbtab[tlist->symb].name[1]));
                 tlist->value = tlist->symb; /* memory */
		 tlist->symb = enter_name(enter_token(name),VARIABLE,0,0);
                 tlist->type = symbtab[tlist->symb].type;
	       }
	      break;

             case TO_OLD:
	      if (symbtab[tlist->symb].name[0] == '#')  /* Strukturvar. */
               {
                 tlist->symb = tlist->value;
                 tlist->value = 0;
                 if (tlist->type != VARIABLE)
                  tlist->type = symbtab[tlist->symb].type;
               }
	      if (symbtab[tlist->symb].name[0] == '*')  /* Xvar. */
               {
                 tlist->symb = tlist->value;
                 tlist->value = 0;
                 if (tlist->type != SVARIABLE)
                  tlist->type = symbtab[tlist->symb].type;
               }
              break;

	     default: 
              nferror("wrong my_rename mode",0);
	  }
      }
}




/*---------------------- result_msg  -------------------------------------*/

result_msg(nr1,nr2,rv)
int nr1,nr2,rv;
{
  switch(rv)
   {
     case  3:
      printf("\nC%d - C%d : Illegal Mode !\n",nr1,nr2);
      break;
     case  2:
      printf("\nC%d - C%d : C%d is empty\n",nr1,nr2,nr1);
      break;
     case  1:
      printf("\nC%d - C%d : Subsumption\n",nr1,nr2);
      break;
     case  0:
      printf("\nC%d - C%d : Constraints\n",nr1,nr2);
      break;
     case -1:
      printf("\nC%d - C%d : No Constraints (one s_list empty)\n",nr1,nr2);
      break;
     case -2:
      printf("\nC%d - C%d : No Constraints (Subs. impossible)\n",nr1,nr2);
      break;
   }
}



/*---------------------- app_sym_to_list ---------------------------------*/

   /* fuegt symb an list hinten an.
      Return-Wert: die neue Liste.
   */
/*---------------------- app_sym_to_list ---------------------------------*/

symbollist *app_sym_to_list(list,sym)

 symbollist *list;
 int sym;
{
   symbollist *hp,*newsym;

   newsym = (symbollist *)  malloc(sizeof(symbollist));
   newsym->symbol = sym;
   newsym->next = NULL;
 
   if (!list)
    list = newsym;
   else {
          hp = list;
          while (hp->next)
           hp = hp->next;
          hp->next = newsym;
        }
   
   return(list);
}



/*---------------------- print_del_clauses -------------------------------*/

print_del_clauses(list)

 symbollist  *list;
{
  printf("  Deleted clauses: ");
  while  (list)
   {
     printf("%d ",list->symbol);
     list = list->next;
   }
  printf("\n");
}




/*---------------------- type_of_constr ----------------------------------*/


type_of_constr(constrl,type)

 predtype *constrl;
 int type;
{
  while (constrl)
   {
     constrl->sign = type;
     constrl = constrl->next;
   }
}



/*---------------------- Finde_subs ---------------------------------*/

int FindeSubs(kurzeClausel,langeClausel)

 claustype *kurzeClausel, *langeClausel;
{
   predtype *kpp, *lpp;   /* zum Durchlaufen der Klauseln */
   littype *cart_list;    /* zur Literalzuordnung. */

   kpp = kurzeClausel->p_list;
   lpp = langeClausel->p_list;
   /*********************************************/ 
   /* default query mit symb == 0 ueberspringen */
   /*********************************************/ 
   if ( ! kpp->symb )
      kpp = kpp->next;
   if ( ! lpp->symb )
      lpp = lpp->next;
   /**************************/
   /* Wenn kurzeClausel leer */
   /**************************/
   if (! kpp)
      return 1;

   if (kurzeClausel->nr_lits > langeClausel->nr_lits)
      return -1;

   /***********************************************************/
   /* Erstelle Kopie von kurzeClausel mit Umbenennung der     */
   /* Variablen und Berechnung von s_list fuer jedes Literal. */
   /***********************************************************/
   if (! trail)
      init_trail();

   abbruch = FALSE;
   no_subsumption = FALSE;
   is_subsumption = FALSE;

    cart_list = make_cart_list(kpp,kurzeClausel->clnr,lpp);

   /******************************************************************/
   /* Wenn ein Literal keine Partner-Literale hat: keine Subsumption */
   /******************************************************************/
   if (no_subsumption)
      {
	 return -1;
      }

   /*********************/
   /* Initialisierungen */
   /*********************/
   bvsptr = 0;
   bvstack[bvsptr]=NULL; 
   rearrange();

   /********************************************/
   /* Zaehle alle Zuordnungsmoeglichkeiten auf */
   /********************************************/
   map(cart_list/* PJ: ,cart_list->s_list*/);
   if (is_subsumption)
     {
      return 1;
     }
   else 
     {
      return -1;
     }
}


/*---------------------- map ------------------------------------*/

   /* Zaehlt alle injektiven Zuordnungsmoeglichkeiten baumartig auf.
	       - is_subsumption = TRUE : Subsumption liegt vor.
	       - Sonst: Constraintliste in constr_list.
   */
/*---------------------- map ------------------------------------*/

map(l)      /* init flags (usage flag) */
littype *l;
{
   plittype *plist=l->s_list;
   while (plist) {
      plist->pr->flag=0;
      plist=plist->next;
   }
   is_subsumption = FALSE;
   abbruch = FALSE;
   map_all(l);
}

void map_all(l)  /* Literal der kurzen Klausel */
littype *l;
{
   predtype *constr;
   plittype *plist=l->s_list; /* Liste deren Partner in der langen Klausel*/
   int flag;

   /*****************************************/
   /* Fuer alle Literale der langen Klausel */
   /*****************************************/
   while (plist)
      {
         flag = 0;
	 if (subs_unif_counter > subs_max_counter) {
	     is_subsumption = FALSE;
             abbruch = TRUE;
	 }
	 if (abbruch) 
	    return;

	 if (l->next == NULL) 
         /*************************************************************/
	 /* Wenn alle Literale der kurzen erfolgreich zugeordnet sind */
         /*************************************************************/
	  {
	    if ( plist->pr->flag == 0 && 
	        (flag = 1, new_unify(plist->pr->t_list,l->t_list) == SUCCESS) 
		&& nr_bindings() == 0)
             {
		  is_subsumption = TRUE;
		  abbruch = TRUE;
	     }
	  }
         else if ( plist->pr->flag == 0 && 
	          (flag = 1, new_unify(plist->pr->t_list,l->t_list) == SUCCESS) 
		  && nr_bindings() == 0) /* Rekursionsfall */
	      /******************************/
	      /* naechstes Literal zuordnen */
	      /******************************/
	       {
	         plist->pr->flag=1;
		 map_all(l->next);
		 plist->pr->flag=0;
	       }
	 /*********************************************/ 
	 /* Wenn alle Zuordnungen erfolgreich waren,  */
	 /* oder die Zuordnung vorzeitig abgebrochen  */
	 /* wurde: Dann mache die Wirkung der letzten */
	 /* Unifikation im Trail rueckgaengig.        */
	 /*********************************************/ 
        if (flag)  
	    undo_unify(); /* backtracking */
	 plist = plist->next;
	 subs_unif_counter += 1;
      } /* while */

}



/*########################################################################*/
