/****************************************************
;;; MODULE NAME : unify.c
;;;
;;; PARENT      : wunif.c
;;;
;;; PROJECT     : MPLOP
;;;
;;; SCCS        : @(#)unify.c	15.2 01 Apr 1996
;;;
;;; AUTHOR      : Janos Breiteneicher
;;;
;;; DATE        : 15.5.91
;;;
;;; DESCRIPTION : unify recursively 2 termlists
;;;
;;; REMARKS     : 
;;;		2.7.93	jsc	extern.h
;;;		10.3.94 Hamdi	makeref #V ---> V
;;;
****************************************************/


#include "symb.h"
#include "extern.h"

/******************************************************************************
/* init_trail
/*****************************************************************************/
/* initialise the trail, allocate memory for it
/**/
void init_trail()
{
if (!trail) 
	trail = (trailelt *)malloc(MAXTRAIL * sizeof(trailelt));
trailcnt=0;
}
		
/******************************************************************************
/* rearrange
/*****************************************************************************/
/* unwind the trail, i.e. reset the pointers to the trail in the
/* symbol table and make the trail empty (trailcnt=0)
/**/
void rearrange()
{
int i;

for (i=0; i<trailcnt; i++){
	symbtab[trail[i].symb].loclist = NULL;
	}
trailcnt = 0;
}
		
/******************************************************************************
/* deref
/*****************************************************************************/
/* dereference a variable cell
/* if it unbound (i.e. there is no trail-entry in the symbol table,
/* return the orig. term 
/* otherwise look at the term, the trail points to and follow up
/**/
termtype *deref(term)
termtype *term;
{ 
trailelt *tp;
termtype *term2;

/* dereference variables only */
while (term->type == VARIABLE || term->type == SVARIABLE) {
	if ( ! (tp = (trailelt *)symbtab[term->symb].loclist)){
		/* is a unbound variable */
		return term;
		}
	/* else : it has a trail cell and this direction is bound */
	else term2 = tp->term;

	/* follow up the reference */
	if (term != term2)
	   term = term2;
	else 
	   break;
	}
return term;
}


/******************************************************************************
/* unify
/*****************************************************************************/
/* unify 2 terms.
/* returns: SUCESS or FAIL 
/* needs an empty trail, does no rearrange
/**/
unify(t1,t2)
termtype *t1, *t2;
{
termtype *lt1, *lt2;
int rv;

/* move thru term list */
while (t1) {
	if (!t2) /* may not happen !! */
	   return FAIL;

	/* dereference the two terms */
	lt1 = deref(t1);
	lt2 = deref(t2);

	switch (lt1->type) {
	case STRINGCONST:
	case NUMCONST:
	case CONSTANT:
		switch (lt2->type) {
		case STRINGCONST:
		case NUMCONST:
		case CONSTANT:
			if (lt1->symb != lt2->symb) 
			   return FAIL; 

		   	/* there are function symbols */
		   	if (lt1->t_list) {
			      rv = unify(lt1->t_list,lt2->t_list);
			      if (rv != SUCCESS)
			         return rv;
			      }
			break;
		case GVARIABLE:
			/* there is a globvar array */
			if (lt2->t_list)
			   break;
		case SVARIABLE:
		case VARIABLE :
			 /* lt2 is a (unbound) variable */
			 /* bind to lt2 variable        */
			 if (makeref(lt1,lt2))
			    return FAIL;
			break;
		default :
			warning(NAM(lt1->symb)," weak-unif error ");
			break;
			}
		break;
	case GVARIABLE:
		/* there is a globvar array */
		if (lt1->t_list)
		   break;
	case SVARIABLE:
	case VARIABLE :
		/* lt1 is a (unbound) variable       */
		/* bind lt2 to lt1, whatever lt2 is  */
	     	if (makeref(lt1,lt2))
	      	   return FAIL;
		break;
	default :
		warning(NAM(lt1->symb)," weak-unif error ");
		break;
		}
	t1 = t1->next;
	t2 = t2->next;
	}

return SUCCESS;
}
		

/******************************************************************************
/* makeref
/*****************************************************************************/
/* generate a new entry on the trail and make an assignment
/* - pointer in the symbol-table for symb
/* - new trail entry with l/r-assignment
/* - there is at least one variable
/* - if necessary do the occur check
/* - returns: 0 if ok, -1 if occur check
/**/
makeref(term1,term2)
termtype *term1,*term2;
{
termtype *term;
termtype *symb;

if (term1->type==CONSTANT || term1->type==NUMCONST || term1->type==STRINGCONST) {
	/* bind term1 on symbol(term2) */
	term = term1;
	symb = term2;
	if (term1->t_list && occ_check(term2,term1->t_list)){
	   return (-1);
	   }
	}
else if (term2->type==CONSTANT || term2->type==NUMCONST || term2->type==STRINGCONST) {
	/* bind term2 on symbol(term1) */
	term = term2;
	symb = term1;
	if (term2->t_list && occ_check(term1,term2->t_list)){
	   return (-1);
	   }
	}
else {
	/* there are two variables      */
	/* bind lower to higher         */
	/* no occur check needed        */
	/* dont bind variable to itself */

	if (term1->symb == term2->symb )
	   /* no self reference !!  */
	   return (0);

	/* variable symb  is bound to variable term	*/
	/* use fixed order to avoid cycles ???     	*/
	/* structural variables ALWAYS bound to variables */
	if (symbtab[term1->symb].name[0] == '#' && /* term1 SV */
	    symbtab[term2->symb].name[0] != '#')   /* term2 V  */ {
	   term = term2;
	   symb = term1;
	    }
	else if (symbtab[term2->symb].name[0] == '#' && /* term2 SV */
	       symbtab[term1->symb].name[0] != '#')   /* term1 V  */ {
	   term = term1;
	   symb = term2;
	   }
	else if (term1 < term2) {
	   term = term1;
	   symb = term2;
	   }
	else {
	   term = term2;
	   symb = term1;
	   }
	}
if ((trailelt *)symbtab[symb->symb].loclist) {
	/* variable already bound */
	/* deref error ?!	  */
	/* may not happen	  */
	warning(symbtab[symb->symb].name,"makeref breaking");
	return (-1);
	}

/* unbound variable make new trail cell */
if (trailcnt == MAXTRAIL-1) {
	nferror("trail too small",0);
	return (0);
	}
trail[trailcnt].symb = symb->symb;
trail[trailcnt].term = term;
symbtab[symb->symb].loclist = (predlist *)(trail + trailcnt);
trailcnt++;

/* PJ: addon to merge hamdi's my_unify et al with unify; the sole difference
       being this routine. NOTE - currently I can only determine hamdi-ness
       by use of a global variable. This may break for more than one thread!

  This scrap is only used in file sconstr.c by new_unify frontend */

if (sconstr_makeref_flag) 
   sconstr_makeref_stackmanagement(symb);

return (0);
}

/******************************************************************************
/* occ_check
/*****************************************************************************/
/* do the occur check :
/*    a functor is bound to a variable
/*    test if the variable occurs in functor
/**/
int occ_check(varterm,functerm)
termtype *varterm, *functerm;
{
termtype *term2;

while (functerm) {
	term2 = deref(functerm);
	if ((term2->type == VARIABLE || term2->type == SVARIABLE) 
	     && term2->symb==varterm->symb) {
	      return (1);
	   }
	if (term2->t_list)
	   if (occ_check(varterm,term2->t_list))
	      return 1;
	functerm = functerm->next;
	}
return 0;
}

/******************************************************************************
/* print_binding
/*****************************************************************************/
/* write the bindings of actuell trail on stdout  */
void print_binding(kurz,lang)
int kurz,lang;
{
int i;
trailelt *trp;

printf("Versuch Klausel %d subsumiert Klausel %d ?\n",kurz,lang);
if (!trailcnt) return;
for (i=0; i<trailcnt; i++){
	trp = (trailelt*)symbtab[trail[i].symb].loclist;
	if (trp->term) {
	   printf("\tVariable %s  an ",
	   symbtab[trp->symb].name);
	   printf("%s",symbtab[trp->term->symb].name);
	   if (trp->term->t_list) {
		printf("(");
		print_lop_term(trp->term->t_list,stdout);
		printf(")");
		}
	   printf("\n");
	   }
	}
}


/*######################################################################*/
