/****************************************************
;;; MODULE NAME : wunif.c
;;;
;;; PARENT      : main.c
;;;
;;; PROJECT     : MPLOP
;;;
;;; SCCS        : @(#)wunif.c	15.1  01 Apr 1996
;;;
;;; AUTHOR      : Janos Breiteneicher
;;;
;;; DATE        : 15.5.91
;;;
;;; DESCRIPTION : make the weak unification ( graph )
;;;
;;; REMARKS     :
;;;		1.3.93	jsc	connections are always in both directions
;;;		3.3.93  jsc	print_wunif adapted to new format
;;;		4.3.93	jsc	start wunif-list with a # on each line
;;;		2.7.93	jsc	extern.h
;;;		21.7.93	jsc	debugging
;;;             8.2.95  PJ      redirected inwasm_rename to my_rename
;;;
****************************************************/

#include "symb.h"
#include "extern.h"
#include "defines.h"

#define DELETED (-9999)


/******************************************************************************
/* genwunif
/******************************************************************************
/* make weak unific. & create data structure
/* for each predicate (symmetric)
/* the unification is called only from '-' to '+' connections
/**/
void genwunif()
{
int i;

init_trail();
maxunif = 0;

	/* default query */
weakunif(0);

	/* make for every user defined predicate the connection list */
for (i=nrbuilt; i<nump; i++) 
	if (symbtab[i].type == PREDSYM) 
	   weakunif(i);
}

/******************************************************************************
/* weakunif
/******************************************************************************
/* enter connections of predicate with number symb
/**/
weakunif(symb)
int symb;
{
int nrunif;
predlist *p1, *p2;
termtype *t1, *t2;
predlist *winf;

/* move thru negative loclist and unify with positive list */
p1 = symbtab[symb].loclist;

while (p1) {
	nrunif = 0;
	/* get one predicate & its termlist t1 */
	if (p1->pr->sign == '+') {
	   p1 = p1->next;
	   continue;
	   }
	t1 = p1->pr->t_list;
	/* inner while loop */
	p2 = symbtab[symb].loclist;
	while (p2) {
	    if (p2->pr->sign == '-') {
		p2 = p2->next;
		continue;
		}
	    t2 = p2->pr->t_list;

	    my_rename(t2, p2->pr->claus->clnr, ALL_TO_SV);

	    switch (unify(t1,t2)) {
	    case SUCCESS:
		nrunif++;

		/* have to store this information */
		winf = (predlist *)malloc( 2 * sizeof(predlist));
		
		/* we always store the connection in both  */
		/* directions,  */
		/* predicate1 -> predicate2  :	*/
		winf->next = p1->pr->wu;
		winf->pr = p2->pr;
		p1->pr->wu = winf++;
		p1->pr->nr_connections += 1;
		/* predicate 2 -> predicate1 :	*/
		winf->next = p2->pr->wu;
		winf->pr = p1->pr;
		p2->pr->wu = winf;
		p2->pr->nr_connections += 1;
		break;
	    case FAIL:
		/* store nothing */
		break;
	    default:
		warning(symbtab[symb].name," weak unify: illegal switch");
		}

	    my_rename(t2, 0, TO_OLD);
	    rearrange();
	    p2 = p2->next;
	    } /* end of inner loop */

	if (nrunif > maxunif) maxunif = nrunif;
	p1 = p1->next;
	}
}


/******************************************************************************
/* print_wunif() 
/******************************************************************************
/* print weak unification  */
print_wunif(modulist,fout,skip)
modtype *modulist;
FILE *fout;
int skip;
{
claustype	*clauses;
predlist	*w;
predtype	*p;
int		cnt;

skip = 0;

while (modulist) {
	if (modulist->m_list) 
		print_wunif(modulist->m_list,fout,0);
	for (clauses=modulist->c_list; clauses; clauses=clauses->next) {
		/* move thru the literals of one clause */
		for (p = clauses->p_list;p;p=p->next) {
		   /* skip default query */
		   if (!p->symb)
		      continue;
		   /* skip head of prules and pfacts */
		   if (!clauses->fann && clauses->p_list == p)
		      continue;
	   	   fprintf(fout,"# %d.%d ", p->claus->clnr,p->litnr);
		   /* skip built-ins */
		   if (symbtab[p->symb].type == BUILTIN ) {
		      fprintf(fout," [built-in]\n");
		      continue;
		   }
		   fprintf(fout,"to ( ");
	           /* move thru wulist & print predics */
		   for (w = p->wu,cnt = 0; w; w=w->next,cnt++) 
		      if (w->pr->kp && w->pr->kp->del != DELETED)
		       fprintf(fout,"%d.%d ",
		           w->pr->claus->clnr,w->pr->litnr);
		   fprintf(fout,")\n");
		   }
		}
	modulist = modulist->next;
	}
}


/******************************************************************************
/* inwasm_rename()
/******************************************************************************
/* rename in termlist all variables back and forth  */
void inwasm_rename(term,clnr,mode)
termtype *term;
int clnr;
int mode;
{
char newname[NAMELENGTH+4];	/* # name clnr */
/* PJ - this is replaced by sconstr.c:my_rename */
fprintf(stderr,"wunif.c: obsolete inwasm_rename called\n.");
exit(42);


while (term) {
	if (term->type == VARIABLE) 
	   switch (mode) {
	   case ALL_TO_SV:
		/* renaming for weak unification: new */
		sprintf(newname,"#%s%d", symbtab[term->symb].name,clnr);

		term->value = term->symb;	/* memory */
		term->symb  = enter_name(enter_token(newname),SVARIABLE,0,0);
		break;
	   case TO_OLD:
		/* restore information before weak unification 	 */
		/* renaming for weak unification: back		 */
		term->symb  = term->value;
		term->value = 0;	/* variables have value 0 */
		break;
	    default:
		nferror("wrong rename mode",0);
	    }
	else if (term->type == CONSTANT && term->t_list)
	   /* recursion for functions  */
	   inwasm_rename(term->t_list,clnr,mode);
	
	term = term->next;
	}
}
	   
/*######################################################################*/
