#include "stack.h"
#include "stree.h"
#include "flags.h"
#include "clock.h"
#include "bar.h"

#if DEBUG
#define CheckFFree  ({ \
 If ffree > 40000 Then \
   printf("ffree overflow!\n\n"); exit(1); \
 EndIf })
#else
#define CheckFFree  ({})
#endif

#define bankQuery   (0)
#define bankDemod   (1)
#define bankSub     (2)
#define bankOld     (3)
#define bankTheorem (4)
#define bankSimplif (5)
#define iow(term)   HeapWriteTerm(term);
#define HeapIsNormalForm(t)    ( (int) ((t)->address) == nftimestamp)
#define HeapSetNormalForm(t)   ( (int) ((t)->address) =  nftimestamp)


/*-----------------------------------------Local procedures prototypes */

void         AddNewRule( _HeapAddr l, _HeapAddr r, int p1, int p2 );
void         AddNewRuleNO( _HeapAddr l, _HeapAddr r, int p1, int p2 );
void         AddOldRule( int n );
void         AddOldRuleNO( int n );
void         FreeRuleNode( int num );
void         FreeSharingSub(_HeapAddr head);
_HeapAddr    InsertSharingSub(_HeapAddr term, int num);
void         InsertSub(_HeapAddr term, int num);
void         PrintRules( void );
void         SizeInit( void );
int          GetOldestNewRuleNum();
int          GetSmallestNewRuleNum( void );
int          GetSmallestThmNewRuleNum( void );
void         AddNewRuleSize( int size, int rulenum );
void         AddNewThmRuleSize( int size, int rulenum );
int          ComputeRuleSize( _HeapAddr l, _HeapAddr r );
int          ComputeTermSize( _HeapAddr t );
void         CritPairs( int n);
void         SCritPairs( int n);
void         CritPairsOldWithAllSubterms( _HeapAddr l, _HeapAddr subt, 
                                       _HeapAddr r, _Bool or, int p1 );
void         ComputeNormalForm( _HeapAddr s, _Bool *touched );
void         TreatCP( _HeapAddr s, _HeapAddr t, int p1, int p2 );
void         BacktrackRewriteStack( void );
_Bool        IsGround(_HeapAddr th1);
_Bool        Subsumed(_HeapAddr ns, _HeapAddr nt);
void         PrintProof(int n);
void         RuleSimplification( int i);
_HeapAddr    NormalFormlhs( _HeapAddr s, _Bool *touched );
_HeapAddr    NormalFormrhs( _HeapAddr s, _Bool *touched );
_IntListNode *GetListDemodSteps();
void         RecursivelySimplify( int before );
void         RuleCopyToBank(short int bank, _HeapAddr   l, _HeapAddr   r, 
                                    _HeapAddr *nl, _HeapAddr *nr);


/*--------------------global variables this module-------------------------*/
Var
  _STree       demod, ndemod, sub, old, subold;
  _IntListNode *rulesizes[MAXRULESIZE];
  _IntListNode *thmrulesizes[MAXRULESIZE];
  int          smallestsize, thmsmallestsize, oldestnewrule=0;
  int          numSimplified, simplified[30000];
  int          numnewrules, numthmnewrules, numrules;
  _RuleNode    *rules[MAXNUMRULES];
  _HeapAddr    refTerm, refTermlhs, refTermrhs;
  _HeapAddr    terms[6000];
  _HeapAddr    teorema,equation, th1,th2,th1r,th2r;
  int          numterms, nftimestamp=1;
  _HeapCell    rewriteheap[10000];
  int          ffree=0, limit;
  _Stack       demodRefStack, refStack, rewritestack;
  _Stack       demodsteps;
  _HeapTag     theoremtop;
  _Bool        thmground;
EndVar


/*--------------------main program-----------------------------*/

void main()
Proc
Var
  int          inc,i, initialnumrules, cycles;
EndVar
  StackFlush(      refStack );
  StackFlush( demodRefStack );
  StackFlush( rewritestack  );
  Clock_Reset;
  Clock_Start;
  STInit();
  VBInit();
  LPOInit();  /* only if non-LPO ordering is used */
  STreeInitStacks();
  MemoryInit();
  SizeInit();
  HeapCreateRef(refTerm);    HeapSetSelfRef(refTerm);
  HeapCreateRef(refTermlhs); HeapSetSelfRef(refTermlhs);
  HeapCreateRef(refTermrhs); HeapSetSelfRef(refTermrhs);
  numnewrules = 0;
  numthmnewrules = 0;
  numterms = 0;
  numrules = 0;

/*----------------------------------------INPUT */
  yyparse();
  flag3(7,HeapIthArg(teorema,1),HeapIthArg(teorema,2));
  equation =  HeapCreateStruct( HeapTag(teorema) );
  theoremtop    = STCheckSymbol("thtop", 2);
  th1 =  HeapCreateStruct( theoremtop );
  th2 =  HeapCreateStruct( theoremtop );
  th1r = HeapCreateStruct( STCheckSymbol("thtrue",  0) );
  th2r = HeapCreateStruct( STCheckSymbol("thfalse", 0) );
  HeapSetAddr( HeapIthArg(th1,1), HeapIthArg(teorema,1) );
  HeapSetAddr( HeapIthArg(th1,2), HeapIthArg(teorema,2) );
  HeapSetAddr( HeapIthArg(th2,1), VBHeapAddr(bankQuery, 0) );
  HeapSetAddr( HeapIthArg(th2,2), VBHeapAddr(bankQuery, 0) );
  limit=10000;
  TreatCP(th1,th1r,0,0);
  If IsGround(th1) Then
    thmground = TRUE;
  Else
    thmground = FALSE;
  EndIf
  TreatCP(th2,th2r,0,0);
  For i=0;i<numterms;i++ Do
    TreatCP( HeapIthArg(terms[i],1), HeapIthArg(terms[i],2), 0,0);
  EndFor
  For i = 1; i <= numrules; i++ Do
    RuleCopyToBank( bankOld, rules[i]->demodlhs,  rules[i]->demodrhs, 
                            &rules[i]->oldlhs,   &rules[i]->oldrhs );
    InsertSharingSub(rules[i]->oldlhs, i);
    If HeapIsSymbol(rules[i]->oldrhs) Then 
      InsertSharingSub(rules[i]->oldrhs, i);
    EndIf
  EndFor



/*----------------------------------------COMPLETION PROCEDURE */
cycles = 0;
If thmground Then
  limit=27;
Else
  limit=30;
EndIf
inc=20;

While numrules<100000 Do
  While (numnewrules > 0) And (numrules<100*inc*limit) And (numrules < 100000)  Do
    cycles++;

    If thmground Then

      If (cycles Mod 2) == 0 And numthmnewrules > 0 Then
        i = GetSmallestThmNewRuleNum();
      ElseIf (cycles Mod 2) == 1 Then
        i = GetOldestNewRuleNum();
      Else
        i = GetSmallestNewRuleNum();
      EndIf

      flag4(2,(_HeapAddr)i, rules[i]->oldlhs, rules[i]->oldrhs);
      numnewrules--;
      If HeapTag(rules[i]->demodlhs)==theoremtop Then
        numthmnewrules--;
      EndIf
      rules[i]->isold = TRUE;
      initialnumrules = numrules;
      SCritPairs(i);
      RecursivelySimplify( initialnumrules );
    Else

      If (cycles Mod 3) == 0 And numthmnewrules > 0 Then
        i = GetSmallestThmNewRuleNum();
      ElseIf (cycles Mod 2) == 1 Then
        i = GetOldestNewRuleNum();
      Else
        i = GetSmallestNewRuleNum();
      EndIf

      flag4(2,(_HeapAddr)i, rules[i]->oldlhs, rules[i]->oldrhs);
      numnewrules--;
      If HeapTag(rules[i]->demodlhs)==theoremtop Then
        numthmnewrules--;
      EndIf
      rules[i]->isold = TRUE;
      initialnumrules = numrules;
      SCritPairs(i);
      RecursivelySimplify( initialnumrules );
    EndIf
  EndWhile

  inc++;
  limit = limit+5;
  oldestnewrule = 0;
  printf("Moving all old rules to NEW Again. Limit is now %d.\n",limit);
  For i=1; i<=numrules; i++ Do
    If Not DeletedRule(i) And rules[i]->isold Then
      numnewrules++;
      rules[i]->isold=FALSE;
      If HeapTag(rules[i]->demodlhs)==theoremtop Then
        numthmnewrules++;
        AddNewThmRuleSize( rules[i]->size, i );
      EndIf        
      AddNewRuleSize( rules[i]->size, i );
    EndIf
  EndFor

EndWhile
PrintRules();
EndProc

/*---------------------------------------- END MAIN  */


void RecursivelySimplify( int before )
Proc
Var
  int i, num1, num2;
EndVar
  num1 = before;
  While num1 != numrules Do
    For i = num1+1; i <= numrules; i++ Do
      RuleCopyToBank( bankOld, rules[i]->demodlhs,  rules[i]->demodrhs, 
                              &rules[i]->oldlhs,   &rules[i]->oldrhs );
      InsertSharingSub(rules[i]->oldlhs, i);
      If HeapIsSymbol(rules[i]->oldrhs) Then 
        InsertSharingSub(rules[i]->oldrhs, i);
      EndIf
    EndFor

    num2 = numrules;
    For i = num1+1; i <= num2; i++ Do
      If rules[i]->oriented And Not DeletedRule(i) Then
        RuleSimplification(i);
      EndIf
    EndFor
    num1 = num2;
  EndWhile
EndProc



void TreatCP( _HeapAddr s, _HeapAddr t, int p1, int p2 )
Proc
Var
  _Bool touched;
  _HeapAddr ns,nt,nns,nnt,ds,dt;
EndVar


  HeapDeref(s,ds);
  HeapDeref(t,dt);

  StackFlush( demodsteps );
  flag3(1,ds,dt);

  If (p2 != 0) And HeapTag(ds)!=theoremtop And 
                   HeapTag(dt)!=theoremtop And
     (ComputeTermSize(ds)+ComputeTermSize(dt) > limit*1.5) Then
    /* printf("       O         \n"); */
    BacktrackRewriteStack();
  ElseIf (p2 != 0) And 
     (ComputeTermSize(ds)+ComputeTermSize(dt) > limit*3.2) Then
    /* printf("       O         \n"); */
    BacktrackRewriteStack();
  Else
    nftimestamp++;
    ns = NormalFormlhs(ds, &touched);
    nt = NormalFormrhs(dt, &touched);
    If (p2 != 0) And HeapTag(ds)!=theoremtop And 
                     HeapTag(dt)!=theoremtop And
       (ComputeTermSize(ns)+ComputeTermSize(nt) > limit) Then
      /* printf("       X         \n"); */
      BacktrackRewriteStack();
    ElseIf HeapEqualTerms(ns,nt) Then
      flag1(1);
      BacktrackRewriteStack();
    ElseIf Subsumed(ns,nt) Then
      flag1(2);
      BacktrackRewriteStack();
    Else
      RuleCopyToBank(bankDemod,ns,nt,&nns,&nnt);
      BacktrackRewriteStack();
      AddNewRule( nns, nnt, p1, p2 ); 
    EndIf
  EndIf
EndProc



void AddNewRule( _HeapAddr s, _HeapAddr t, int p1, int p2 )
Proc
Var 
  int *info, size;
  _HeapAddr l, r;
EndVar
  numnewrules++;
  numrules++;

If Not (numrules Mod 1000) Then
  printf("%d\n",numrules);
EndIf
/*
If (numrules==47363) Then
  test = 1;
EndIf
*/


  rules[numrules] = (_RuleNode *) MemoryAlloc(sizeof(_RuleNode));
  rules[numrules]->parent1   = p1;
  rules[numrules]->parent2   = p2;
  rules[numrules]->demodlist = GetListDemodSteps();
  rules[numrules]->isold     = FALSE;
  rules[numrules]->isdeleted = FALSE;


  If LPOOrient(s,t,&l,&r) Then
    size = 2*ComputeTermSize(l)+ComputeTermSize(r);
    rules[numrules]->size      = size;
    rules[numrules]->oriented  = TRUE;
    flag4(1,(_HeapAddr)numrules,l,r);
    info = (int *) STreeInsert(&demod,l);
    *info = numrules;
    If HeapTag(l)==HeapTag(th1r) And HeapTag(r)==HeapTag(th2r) Then
      printf("Theorem Proved."); putchar(7);
      printf("\n");
      rules[numrules]->demodlhs  = l;
      rules[numrules]->demodrhs  = r;
      PrintProof(numrules);   
      printf("\n");
      Clock_Stop; Clock_Print;  MemoryStatistics();
      printf("\n");
      exit(1);
    EndIf
    If HeapTag(l) == theoremtop Then
      numthmnewrules++;
      AddNewThmRuleSize( size, numrules );
    EndIf
    AddNewRuleSize( size, numrules );
  Else
    size = 1.5 * (ComputeTermSize(l)+ ComputeTermSize(r));
    rules[numrules]->size      = size;
    rules[numrules]->oriented  = FALSE;
    flag4(5,(_HeapAddr)numrules,l,r);
    HeapSetAddr( HeapIthArg(equation,1),l);
    HeapSetAddr( HeapIthArg(equation,2),r);
    info = (int *) STreeInsert(&ndemod,equation);
    *info = numrules;
    If HeapIsRef(l) Or HeapIsRef(r) Then
      printf("\nTheorem Proved.\n\n"); putchar(7);  
      rules[numrules]->demodlhs  = l;
      rules[numrules]->demodrhs  = r;
      PrintProof(numrules); 
      printf("\nInconsistency!\n\n"); 
      Clock_Stop; Clock_Print;  MemoryStatistics();
      exit(1);
    EndIf
    AddNewRuleSize( size, numrules );
  EndIf
  rules[numrules]->demodlhs  = l;
  rules[numrules]->demodrhs  = r;
  rules[numrules]->oldlhs  = NULL;
  rules[numrules]->oldrhs  = NULL;
EndProc



void AddNumToSubTerms( _InfoNodeSub *node, int num )
Proc
Var
  _IntListNode *p;
  int arity;
  _HeapAddr ti,term;
EndVar

  If node->numslist->num != num Then
    p = (_IntListNode *) MemoryAlloc(sizeof(_IntListNode));
    p->num = num;
    p->next = node->numslist;
    node->numslist = p;
  EndIf
  (node->ctr)++;
  term = HeapAddr( (_HeapAddr)node);
  arity = HeapTagArity(term);
  While arity > 0 Do
    ti = HeapAddr( HeapIthArg(term,arity));
    If Not HeapIsSelfRef(ti) Then
      AddNumToSubTerms( (_InfoNodeSub*)ti, num);
    EndIf
    arity--;
  EndWhile
EndProc

_HeapAddr InsertSharingSub(_HeapAddr term, int num)
Proc
Var 
  _HeapAddr ti, newterm, refpos;
  _InfoNodeSub **info; 
  _IntListNode *p;
  int arity;
EndVar
  info = (_InfoNodeSub **) STreeFind(sub,term);
  If info Then
    AddNumToSubTerms(*info, num);
    Return( (_HeapAddr)*info );
  Else
    info = (_InfoNodeSub **) STreeInsert(&sub,term);
    *info = (_InfoNodeSub *) MemoryAlloc(sizeof(_InfoNodeSub));
    p = (_IntListNode *) MemoryAlloc(sizeof(_IntListNode));
    p->num = num;
    p->next = NULL;
    (*info)->numslist = p;
    (*info)->ctr = 1;
    newterm = HeapCreateStruct( HeapTag(term) );
    refpos = (_HeapAddr) *info;
    HeapSetRef( refpos );
    HeapSetAddr( refpos, newterm);
    arity = HeapTagArity(term);
    While arity > 0 Do
      ti = HeapAddr(HeapIthArg(term,arity));
      If HeapIsSymbol(ti) Then
        HeapSetAddr( HeapIthArg(newterm,arity), 
                 InsertSharingSub(ti,num) );
      Else
        HeapSetAddr( HeapIthArg(newterm,arity), ti );
      EndIf
      arity--;
    EndWhile
    Return( refpos );
  EndIf
EndProc

void RemoveNumToSubTerms(_InfoNodeSub *info, _HeapAddr term)
Proc
  Var 
    _HeapAddr ti, leafTerm;
    _InfoNodeSub *infoti; 
    _IntListNode *p;
    int arity;
  EndVar

  leafTerm = HeapAddr((_HeapAddr) info);
  arity = HeapTagArity(term);
  While arity > 0 Do
    ti = HeapAddr(HeapIthArg(term,arity));
    infoti = (_InfoNodeSub*) HeapAddr( HeapIthArg(leafTerm,arity) );
    If HeapIsSymbol(ti) Then
      RemoveNumToSubTerms( infoti, ti );
    EndIf
    arity--;
  EndWhile

  info->ctr--;
  If info->ctr == 0 Then
    STreeDelete(&sub, term);
    HeapFreeStruct(leafTerm);
    While info->numslist Do
      p = info->numslist;
      info->numslist = info->numslist->next;
      MemoryFree((_Ptr) p, sizeof(_IntListNode));
    EndWhile
    MemoryFree((_Ptr) info, sizeof(_InfoNodeSub));
  EndIf
EndProc

void RemoveSharingSub( _HeapAddr term )
Proc
Var 
  _InfoNodeSub **info; 
EndVar
  info = (_InfoNodeSub **) STreeFind(sub,term);
  RemoveNumToSubTerms(*info, term);
EndProc



void RuleSimplification(int i)
Proc
Var
  _IntListNode *p;
  _InfoNodeSub *infoSub;
  int hit,j;
EndVar

  If Not thmground And HeapTag( rules[i]->demodlhs )==theoremtop Then
    If TermUnify( HeapIthArg( rules[i]->oldlhs, 1), 
                  HeapIthArg( rules[i]->oldlhs, 2) ) Then
      TreatCP( th1r, th2r, i, 2 );
    EndIf
    RestoreRefsUntilEmpty( refStack );
  EndIf


  numSimplified = 0;
  flag2(7,(_HeapAddr)i);
  infoSub = STreeFirstReverseMatch( sub, rules[i]->demodlhs );
  While infoSub Do
    p = infoSub->numslist;
    While p Do
      hit = p->num;
      If hit != i And Not DeletedRule(hit) Then
        flag2(8,(_HeapAddr)hit);
        numSimplified++;
        simplified[numSimplified] = hit;
      EndIf
      p = p->next;
    EndWhile
    infoSub = STreeNextReverseMatch();
  EndWhile

  While numSimplified != 0 Do
    j = simplified[numSimplified--];
    If Not DeletedRule(j) Then
      If rules[j]->oriented Then
        STreeDelete( &demod, rules[j]->demodlhs );
      Else
        HeapSetAddr( HeapIthArg(equation,1), rules[j]->demodlhs );
        HeapSetAddr( HeapIthArg(equation,2), rules[j]->demodrhs );
        STreeDelete( &ndemod, equation );
      EndIf
      TreatCP( rules[j]->oldlhs, rules[j]->oldrhs, j, 0 );

      If Not rules[j]->isold Then
        numnewrules--;
        If HeapTag(rules[j]->demodlhs)==theoremtop Then
          numthmnewrules--;
        EndIf
      EndIf
      RemoveSharingSub(rules[j]->oldlhs);
      If HeapIsSymbol(rules[j]->oldrhs) Then 
        RemoveSharingSub(rules[j]->oldrhs);
      EndIf
      HeapFreeNormalTerm( rules[j]->oldlhs );
      HeapFreeNormalTerm( rules[j]->oldrhs );
      rules[j]->isdeleted = TRUE;
      rules[j]->oldlhs = NULL;
      rules[j]->oldrhs = NULL;
      flag2(6,(_HeapAddr)j);
    EndIf
  EndWhile
  flag2(10,(_HeapAddr)i);
EndProc



/*------------------------Rules Output----------------------------*/


void PrintRules()
Proc
Var
  int i;
EndVar

  For i=1;i<=numrules;i++ Do
    If Not DeletedRule(i) Then
      printf("%d.- ",i);
      HeapWriteTerm( rules[i]->demodlhs );
      printf(" --> ");
      HeapWriteTerm( rules[i]->demodrhs );
      If rules[i]->isold Then printf(" (old)"); EndIf
      printf("\n");
    EndIf
  EndFor
  printf("\n\n");

  For i=1;i<=numrules;i++ Do
    If Not DeletedRule(i) Then
      printf("%d.- ",i);
      HeapWriteTerm( rules[i]->oldlhs );
      printf(" --> ");
      HeapWriteTerm( rules[i]->oldrhs );
      If rules[i]->isold Then printf(" (old)"); EndIf
      printf("\n");
    EndIf
  EndFor
  printf("\n\n");
EndProc



void SuperposeTop(_HeapAddr l1, _HeapAddr r1, int p1, 
                  _HeapAddr l2, _HeapAddr r2, int p2)
Proc
  If p1 != p2 And TermUnify( l1, l2 ) Then
    TreatCP( r1, r2, p1, p2 );
  EndIf
  RestoreRefsUntilEmpty( refStack );
EndProc


void SuperposeSubt1(_HeapAddr s, _HeapAddr l1, _HeapAddr r1, int p1,  
                                 _HeapAddr l2, _HeapAddr r2, int p2)
Proc
Var
  _HeapAddr ti;
  int arity, i;
EndVar
  ti = HeapAddr( s );
  If HeapIsSymbol( ti ) Then
    If TermUnify( ti, l2 ) Then
      HeapSetAddr( s, r2 );
      TreatCP( l1, r1, p1, p2 );
      HeapSetAddr( s, ti );
    EndIf
    RestoreRefsUntilEmpty( refStack );
    arity = HeapTagArity(ti);
    For i=1;i<=arity;i++ Do
      SuperposeSubt1(HeapIthArg(ti,i), l1, r1, p1,   l2, r2, p2);
    EndFor
  EndIf
EndProc


void SuperposeSubt(_HeapAddr l1, _HeapAddr r1, int p1, 
                   _HeapAddr l2, _HeapAddr r2, int p2)
Proc
Var
  int arity, i;
EndVar
  If HeapIsSymbol( l1 ) Then
    arity = HeapTagArity(l1);
    For i=1;i<=arity;i++ Do
      SuperposeSubt1(HeapIthArg(l1,i), l1, r1, p1,    l2, r2, p2);
    EndFor
  EndIf
EndProc


void SCritPairs(int n)
Proc
Var
  _HeapAddr nl,nr;
  int i;
EndVar

  flag4(3,(_HeapAddr)n,rules[n]->demodlhs, rules[n]->demodrhs);
  RuleCopyToBank( bankQuery, rules[n]->demodlhs, rules[n]->demodrhs, &nl, &nr);
  StackFlush( refStack );

  For i=1;i<=numrules;i++ Do
    If Not DeletedRule(i) And rules[i]->isold Then
      SuperposeTop( nl, nr, n, rules[i]->oldlhs, rules[i]->oldrhs, i );
      SuperposeSubt( nl, nr, n, rules[i]->oldlhs, rules[i]->oldrhs, i );
      SuperposeSubt( rules[i]->oldlhs, rules[i]->oldrhs, i, nl, nr, n );
      If Not rules[i]->oriented Then
        SuperposeTop( nl, nr, n, rules[i]->oldrhs, rules[i]->oldlhs, i );
        SuperposeSubt( nl, nr, n, rules[i]->oldrhs, rules[i]->oldlhs, i );
        SuperposeSubt( rules[i]->oldrhs, rules[i]->oldlhs, i, nl, nr, n );
      EndIf
    EndIf
  EndFor
  If Not rules[n]->oriented Then
    For i=1;i<=numrules;i++ Do
      If Not DeletedRule(i) And rules[i]->isold Then
        SuperposeTop( nr, nl, n, rules[i]->oldlhs, rules[i]->oldrhs, i );
        SuperposeSubt( nr, nl, n, rules[i]->oldlhs, rules[i]->oldrhs, i );
        SuperposeSubt( rules[i]->oldlhs, rules[i]->oldrhs, i, nl, nr, n );
        If Not rules[i]->oriented Then
          SuperposeTop( nr, nl, n, rules[i]->oldrhs, rules[i]->oldlhs, i );
          SuperposeSubt( nr, nl, n, rules[i]->oldrhs, rules[i]->oldlhs, i );
          SuperposeSubt( rules[i]->oldrhs, rules[i]->oldlhs, i, nl, nr, n );
        EndIf
      EndIf
    EndFor
  EndIf
  flag2(5,(_HeapAddr)n);
  HeapFreeNormalTerm(nl);
  HeapFreeNormalTerm(nr);
EndProc

/*------------------Procedures for size maintenance--------------------*/
void SizeInit()
Proc
Var
  int i;
EndVar
  For i=0;i<MAXRULESIZE;i++ Do
    rulesizes[i]=NULL;
    thmrulesizes[i]=NULL;
  EndFor
  smallestsize=MAXRULESIZE;
  thmsmallestsize=MAXRULESIZE;
EndProc


int GetOldestNewRuleNum()
Proc
  oldestnewrule++;
  While rules[oldestnewrule]->isold Or DeletedRule(oldestnewrule) Do 
    oldestnewrule++;
  EndWhile
  Return(oldestnewrule);
EndProc


int GetSmallestNewRuleNum()
Proc
Var
  _IntListNode *p;
  int n;
EndVar
  n = rulesizes[smallestsize]->num;
  p = rulesizes[smallestsize];
  rulesizes[smallestsize] = (_IntListNode*)rulesizes[smallestsize]->next;
  MemoryFree((_Ptr) p, sizeof(_IntListNode));
  While numnewrules > 0 And rulesizes[smallestsize] == NULL Do
    smallestsize++;
  EndWhile
  If DeletedRule(n) Or rules[n]->isold Then
    n=GetSmallestNewRuleNum();
  EndIf
  Return n;
EndProc



int GetSmallestThmNewRuleNum()
Proc
Var
  _IntListNode *p;
  int n;
EndVar
  n = thmrulesizes[thmsmallestsize]->num;
  p = thmrulesizes[thmsmallestsize];
  thmrulesizes[thmsmallestsize] = (_IntListNode*)thmrulesizes[thmsmallestsize]->next;
  MemoryFree((_Ptr) p, sizeof(_IntListNode));
  While numthmnewrules > 0 And thmrulesizes[thmsmallestsize] == NULL Do
    thmsmallestsize++;
  EndWhile
  If DeletedRule(n) Or rules[n]->isold Then
    n=GetSmallestThmNewRuleNum();
  EndIf
  Return n;
EndProc


void AddNewRuleSize(int size, int rulenum)
Proc
Var
  _IntListNode *p;
EndVar
  p = (_IntListNode *) MemoryAlloc(sizeof(_IntListNode));
  p->next = rulesizes[size];
  p->num = rulenum;
  rulesizes[size] = p;
  If size < smallestsize Then
    smallestsize = size;
  EndIf
EndProc


void AddNewThmRuleSize(int size, int rulenum)
Proc
Var
  _IntListNode *p;
EndVar
  p = (_IntListNode *) MemoryAlloc(sizeof(_IntListNode));
  p->next = thmrulesizes[size];
  p->num = rulenum;
  thmrulesizes[size] = p;
  If size < thmsmallestsize Then
    thmsmallestsize = size;
  EndIf
EndProc


int ComputeRuleSize(_HeapAddr l, _HeapAddr r)
Proc
    Return ComputeTermSize(l)+ComputeTermSize(r);
EndProc

int ComputeTermSize(_HeapAddr t)
Proc
Var
  _HeapAddr auxt, auxs;
  int n,i,arity;
EndVar
  n = 1;
  HeapDeref(t, auxt);
  If HeapIsRef(auxt) Then
    Return(1);
  Else
    arity = HeapTagArity(auxt);
    For i = 1; i <= arity; i++ Do
      auxs = HeapIthArg(auxt,i);
      HeapDeref(auxs, auxs);
      n = n + 1 + ComputeTermSize( auxs );
    EndFor
  EndIf
  Return n;
EndProc

/*
int ComputeRuleSize(_HeapAddr l, _HeapAddr r)
Proc
    Return(ComputeTermSize(l)+ComputeTermSize(r));
EndProc


int ComputeTermSize(_HeapAddr t)
Proc
Var
  _HeapAddr auxt;
  int arity, n;
EndVar
  HeapDeref(t, auxt);
  If HeapIsRef(auxt) Then
    Return(1);
  Else
    n=1;
    arity = HeapTagArity(auxt);
    While arity>0 Do
      n += ComputeTermSize( HeapIthArg(auxt,arity));
    EndWhile
  EndIf
  Return n;
EndProc
*/



_Bool Subsumed(_HeapAddr s, _HeapAddr t)
Proc
Var
  int info;
EndVar
  HeapSetAddr( HeapIthArg(equation,1), s);
  HeapSetAddr( HeapIthArg(equation,2), t);

  info = (int) STreeFirstMatch( ndemod, equation );
  If info Then
    RestoreRefsUntilEmpty( demodRefStack );
    Return(TRUE);
  Else
    RestoreRefsUntilEmpty( demodRefStack );
    HeapSetAddr( HeapIthArg(equation,1), t);
    HeapSetAddr( HeapIthArg(equation,2), s);
    info = (int) STreeFirstMatch( ndemod, equation );
    If info Then
      RestoreRefsUntilEmpty( demodRefStack );
      Return(TRUE);
    Else
      RestoreRefsUntilEmpty( demodRefStack );
      Return(FALSE);
    EndIf
  EndIf
EndProc


void GetProofMembers( int n, _IntListNode **list )
Proc
Var
  _IntListNode *aux, *aux1;
EndVar
  If n !=0 And rules[n]->parent1 >= 0 Then
    aux = (_IntListNode*) MemoryAlloc( sizeof( _IntListNode ));
    aux->next = *list;
    aux->num = n;
    *list = aux;
    rules[n]->parent1 = -(rules[n]->parent1+1);
    GetProofMembers( -(rules[n]->parent1)-1, list );
    GetProofMembers( rules[n]->parent2, list );
    aux1 = rules[aux->num]->demodlist;
    While aux1 Do
      GetProofMembers( aux1->num, list );
      aux1 = aux1->next;
    EndWhile
  EndIf
EndProc


_IntListNode *GetListDemodSteps()
Proc
Var
  _IntListNode *aux, *aux1; 
EndVar
  aux = NULL;
  While Not StackIsEmpty( demodsteps ) Do
    aux1 = (_IntListNode*) MemoryAlloc( sizeof( _IntListNode ));
    aux1->num = (int) StackTopPop( demodsteps );
    aux1->next = aux;
    aux = aux1;
  EndWhile
  Return(aux);
EndProc

int intcompare(int *i, int *j)
Proc
  Return(*i - *j);
EndProc


void PrintProof(int n)
Proc
Var
  _IntListNode *aux, *aux1;
  int a[1000];
  int i,j;
  FILE *steps;
EndVar
  steps = fopen( "steps", "w" );
  aux = NULL;
  GetProofMembers( n, &aux );
  i=0;
  While aux Do
    a[i++]=aux->num;
    aux = aux->next;
  EndWhile
  qsort(a,i-1,sizeof(int), (void *) intcompare );
  For j=0;j<i;j++ Do
    If -(rules[a[j]]->parent1)-1 == 0 Then
      fprintf(steps, "[%d=axiom",a[j] );
    ElseIf rules[a[j]]->parent2 == 0 Then
      fprintf(steps, "[%d=demod(%d)",a[j], -(rules[a[j]]->parent1)-1 );
    Else
      fprintf(steps, "[%d=param(%d,%d)",a[j], 
                           -(rules[a[j]]->parent1)-1, rules[a[j]]->parent2);
    EndIf
    aux1 = rules[a[j]]->demodlist;
    If aux1 Then
      fprintf(steps, ",[");
      While aux1 Do
        fprintf(steps, "%d",aux1->num);
        aux1 = aux1->next;
        If aux1 Then
          fprintf(steps, ",");
        Else
          fprintf(steps, "],");
        EndIf
      EndWhile
    Else
      fprintf(steps, ",[],");
    EndIf
    fprintf(steps, "\n\t\t\t");
    HeapWriteTermFile( steps, rules[a[j]]->demodlhs );
    fprintf(steps, " \t= ");
    HeapWriteTermFile( steps, rules[a[j]]->demodrhs );
    fprintf(steps, "].\n");
  EndFor
EndProc

/*------------------- Rewriting -----------------------------*/

_HeapAddr NormalFormlhs(_HeapAddr s, _Bool *touched)
Proc
  HeapSetAddr( refTermlhs, s );
  ComputeNormalForm( refTermlhs, touched );
  Return( HeapAddr( refTermlhs ));
EndProc


_HeapAddr NormalFormrhs(_HeapAddr s, _Bool *touched)
Proc
  HeapSetAddr( refTermrhs, s );
  ComputeNormalForm( refTermrhs, touched );
  Return( HeapAddr( refTermrhs ));
EndProc


/* 
   Pre: RHS is built with all refs below a struct pointing
        to another struct or to a variable bank
*/
_HeapAddr CopyRHS(_HeapAddr rhs)
Proc
Var
  _HeapAddr aux;
  _HeapTag tag;
  int i,arity;
EndVar
  If HeapIsRef(rhs) Then  /* it is a BankRef */
    Return(HeapAddr(rhs));
  Else
    tag = HeapTag(rhs);
    arity = STArityOfSymbol(tag);
    aux = rewriteheap+ffree;
    ffree = ffree+arity+1;
    CheckFFree;
    HeapSetTag( aux, tag );
    HeapSetAddr( aux, (_HeapAddr) 0 );
    For i=1; i<=arity; i++ Do 
      HeapSetRef(  HeapIthArg(aux,i) ); 
      HeapSetAddr( HeapIthArg(aux,i),
                   CopyRHS( HeapAddr(HeapIthArg(rhs,i)) ) );
    EndFor
    Return(aux);
  EndIf
EndProc


/* PRE: s must be a ref to a term */
void ComputeNormalForm(_HeapAddr s, _Bool *touched)
Proc
Var
  _HeapAddr add,dref;
  int i,arity;
  _Bool intouched,toptouched;
  int info;
EndVar
  add = HeapAddr( s );
  HeapDeref( add, dref );
  *touched = FALSE;
  If HeapIsSymbol(dref) And Not HeapIsNormalForm(dref) Then
    arity = STArityOfSymbol( HeapTag(dref));
    For i=1; i<=arity; i++ Do
      ComputeNormalForm( HeapIthArg(dref,i), &intouched );
      *touched = (*touched Or intouched);
    EndFor
    toptouched = FALSE;
    info = (int) STreeFirstMatch( demod, dref );
    If info Then
      StackPush( demodsteps, (_HeapAddr) info );
      StackPush( rewritestack, add );
      StackPush( rewritestack, s );
      HeapSetAddr( s, CopyRHS( rules[info]->demodrhs ) );
      RestoreRefsUntilEmpty(demodRefStack);
      ComputeNormalForm( s, &intouched );
      *touched = TRUE;
      toptouched = TRUE;
    EndIf
    If Not toptouched Then
      HeapSetNormalForm(dref);
    EndIf
  EndIf
EndProc


void BacktrackRewriteStack()
Proc
  While Not StackIsEmpty( rewritestack ) Do
    HeapSetAddr( StackTopPop( rewritestack ), 
                 StackTopPop( rewritestack ) );
  EndWhile
  ffree=0;
EndProc



void RuleCopyToBank(short int bank, _HeapAddr   l, _HeapAddr   r, 
                                    _HeapAddr *nl, _HeapAddr *nr) 
Proc
Var
  int ctr,h;
EndVar
  *nl = HeapCopyTerm(l);
  *nr = HeapCopyTerm(r);
  ctr=0;
  h = StackHeight( refStack );
  *nl = HeapMoveToBankTerm(&ctr, *nl, bank);
  *nr = HeapMoveToBankTerm(&ctr, *nr, bank);
  RestoreRefsUntilHeight( refStack, h );
EndProc  


_Bool IsGround(_HeapAddr t)
Proc
Var
  int arity;
  _Bool ground;
  _HeapAddr term;
EndVar

  HeapDeref(t,term);
  If HeapIsRef(term) Then
    Return(FALSE);
  Else
    ground = TRUE;
    arity = HeapTagArity(term);
    While arity>0 And ground Do
      ground = IsGround( HeapAddr( HeapIthArg(term,arity) ));
      arity--;
    EndWhile
    Return(ground);
  EndIf
EndProc
