/* 
 * I-THOP  (A first-order theorem prover based on PTTP technology)
 * 
 *     version 1.01
 *     July 2, 1997
 *     Copyright (C) 1997, Koji Iwanuma (iwanuma@esi.yamanashi.ac.jp)
 *
 * This program is free software; you can redistribute it or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either versions 1, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful but
 * WITHOUT ANY WARRANTY.  See the GNU General Public License for more
 * details.
 */

#include <signal.h>
#include <sicstus/sicstus.h>

SP_pred_ref event_pred;

static int signal_event(signal_no)
     void *signal_no;
{
  SP_term_ref x=SP_new_term_ref();
   
  SP_put_integer(x, (int)signal_no);
  SP_query(event_pred,x) ;
  exit(0) ;
}

static void signal_handler(sig)
     int sig;
{
  SP_event(signal_event, (void *)sig);
  SP_reinstall_signal(sig,signal_handler); 
  SP_continue();
}

void signal_init()
{
  event_pred = SP_predicate("prolog_handler",1,"");

  SP_signal(SIGXCPU,signal_handler);
}

/*
 *  Each index i of carray[] is for counting the number of 
 *
 *  0:  
 *  1:   
 *  2: 
 *  3:
 *  4:
 *  5:  
 *  6:  
 *  7: 
 *  8:  
 *       .....
 *  Size_Array : 
 */

#define Size_CArray 128

static long carray[Size_CArray];

void reset_carray()
{	
	int i;

	for (i = 0 ; i < Size_CArray ; i++)
	   carray[i] = 0;
}

void inc_carray(ident)
long ident;
{
	(carray[ident])++;
}

long call_carray(ident)
long ident;
{
	return carray[ident];
}

/*************************************************/

#define Size_Array 32

/*  Each index i of number_array[] is for set and reset several flags:
 *
 *  0:  h_all_pruning_lazy
 *  1:  unit and/or lemma matching
 *  2:  
 *  3:  
 *  4:  
 *  5:  
 *  6:  
 *  7:  the number of lemma in lazy matching.
 *  8:  
 *  9:  
 *  10: the ID number of unit in replacement subusmption
 *  11: the ID number of lemma in replacement subusmption
 *
 *  12:   .....
 */

static long number_array[Size_Array];

void set_flag_number(index,number)
long index, number;
{	
         number_array[index] = number;
}

void reset_all_flag_number()
{	
	int i;
	for (i = 0 ; i < Size_Array ; i++)
	    number_array[i] = 0;
}

void reset_flag_number(index)
long index;
{
	number_array[index] = 0;
}

long call_flag_number(index)
long index;
{
	return number_array[index];
}

/***************************************************/

/*  Each index i of atom_array[] is for set and reset several flags:
 *
 *  1:  lemmaizing
 *  2:  DFlag in tail pruning
 *  3:  SFlag in tail pruning
 *
 *  4:  ?
 *  5:  SFlag in unit or/and lemma matching
 *  6:  Identifier of unit or/and lemma matching
 *
 *  7:  Flag for Lazy lemma matching
 *
 *  8:  Flag for unit subsumption in matching
 *  9:  anti_lemmaizing
 *
 * 10:  Flag for unit replacement subusmption
 * 11:  Flag for lemma replacement subusmption
 *
 * 12:  Flag for detecting "Finite Failures"
 *
 * 13:  Flag for unit consistenecy check
 * 14:  Flag for lemma consistenecy check
 */

static unsigned long atom_array[Size_Array];

void set_flag_atom(index,atom)
long index;
unsigned long atom;
{	
         atom_array[index] = atom;
}

void reset_all_flag_atom(atom)
unsigned long atom;
{	
	int i;
	for (i = 0 ; i < Size_Array ; i++)
	    atom_array[i] = atom;
}

unsigned long call_flag_atom(index)
long index;
{
	return atom_array[index];
}


/***************************************************/

#define Double_Size_Array 64

static long lemma_number_array[Double_Size_Array];
static long lemma_total_number;

void reset_all_lemma_number()
{	
	int i;
	for (i = 0 ; i < Double_Size_Array ; i++)
	    lemma_number_array[i] = 0;
	
	lemma_total_number = 0;
}

/*
 *void reset_lemma_number(index)
 *long index;
 *{
 *	lemma_number_array[index] = 0;
 *}
 */

void inc_lemma_number(index)
long index;
{	
         lemma_number_array[index]++;
	 lemma_total_number++;
}

long call_lemma_number(index)
long index;
{
	return lemma_number_array[index];
}

void inc_lemma_total_number()
{	
	 lemma_total_number++;
}

long call_lemma_total_number()
{
	return lemma_total_number;
}

/***************************************************/

static long tmp_number;

void reset_tmp_number()
{
	tmp_number = 0;
}

void inc_tmp_number()
{
        tmp_number++;
}

long call_tmp_number()
{
        return tmp_number;
}

/***************************************************/

static long anti_number;

void inc_anti()
{
        anti_number++;
      }

void dec_anti()
{
        anti_number--;
}

long call_anti_number()
{
        return anti_number;
}

void reset_anti()
{
       anti_number = 0;
}
