/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *               NAIVE CNF TRANSLATOR                     * */
/* *                                                        * */
/* *  $Module:   CNF                                        * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.63 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/02 14:26:51 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: cnf.c,v $ */

#include "cnf.h"
#include "rules-inf.h"

static void   cnf_ApplyDistributivity(TERM term);
static TERM   cnf_Cnf(TERM term, LIST* symblist);
static TERM   cnf_DistrQuantorNoVarSub(TERM Term);
static TERM   cnf_Flatten(TERM Term, SYMBOL Symbol);
static SYMBOL cnf_GetDualSymbol(SYMBOL symbol);
static int    cnf_GetFormulaPolarity(TERM term, TERM subterm);
static LIST   cnf_GetSymbolList(LIST varlist);
static BOOL   cnf_HaveProofOptSkolem(PROOFSEARCH, LIST, TERM);
static TERM   cnf_MakeOneAnd(TERM term);
static TERM   cnf_MakeOneAndPredicate(TERM term);
static TERM   cnf_MakeOneAndTerm(TERM term);
static TERM   cnf_MakeOneOr(TERM term);
static TERM   cnf_MakeOneOrPredicate(TERM term);
static TERM   cnf_MakeOneOrTerm(TERM term);
static TERM   cnf_MoveProvedTermToTopLevel(TERM, TERM, TERM, LIST, LIST);
static void   cnf_OptimizedSkolemFormula(PROOFSEARCH, TERM, TERM, LIST, BOOL);
static LIST   cnf_OptimizedSkolemization(PROOFSEARCH, TERM, LIST, BOOL, BOOL);
static void   cnf_PopAllQuantifier(TERM term);
static TERM   cnf_QuantifyAndNegate(TERM term, LIST VarList, LIST FreeList);
static TERM   cnf_QuantMakeOneVar(TERM term);
static TERM   cnf_MakeSkolemfunction(LIST varlist);
static TERM   cnf_RemoveTrivialOperators(TERM Term);
static TERM   cnf_RplacVarsymbFunction(TERM term, SYMBOL varsymb, TERM function);
static void   cnf_Skolemize(TERM Term, LIST FreeList);
static TERM   cnf_SimplifyQuantors(TERM Term);
static TERM   cnf_SkolemFunctionFormula(TERM, LIST, LIST, LIST*);
static LIST   cnf_SatUnit(PROOFSEARCH, LIST); 
static LIST   cnf_SubsumeClauseList(LIST clauselist);
static BOOL   cnf_TopIsAnd(LIST termlist);


static int cnf_GetFormulaPolarity(TERM term, TERM subterm)
     /**********************************************************
       INPUT:   Two terms term and subterm where subterm is a 
                subterm of term.
       RETURNS: The polarity of subterm in term.
       ********************************************************/
{  
  LIST   scan;
  TERM   term1;
  int    polterm1,bottom; 

  bottom = vec_ActMax();
  vec_Push((POINTER) 1);
  vec_Push(term);

  do {
    term1    = (TERM)vec_PopResult();
    polterm1 = (int)vec_PopResult();
    if (term1 == subterm){
      vec_SetMax(bottom);
       return polterm1;
    }else
      if (symbol_Equal(term_TopSymbol(term1),fol_Not())){
	vec_Push((POINTER) (- polterm1));
	vec_Push(list_Car(term_ArgumentList(term1)));
      }
    if (symbol_Equal(term_TopSymbol(term1),fol_Exist()) ||
	symbol_Equal(term_TopSymbol(term1),fol_All())){
      vec_Push((POINTER)polterm1);
      vec_Push(list_Second(term_ArgumentList(term1)));
    }
    else
      if (symbol_Equal(term_TopSymbol(term1),fol_Implies())){
	vec_Push((POINTER) (- polterm1));
	vec_Push(list_Car(term_ArgumentList(term1)));
	vec_Push((POINTER) polterm1);
	vec_Push(list_Second(term_ArgumentList(term1)));
      }
    else
      if (symbol_Equal(term_TopSymbol(term1),fol_Equiv())){
	vec_Push(0);
	vec_Push(list_Car(term_ArgumentList(term1)));
	vec_Push(0);
	vec_Push(list_Second(term_ArgumentList(term1)));
      }
    else
      if ((symbol_Equal(term_TopSymbol(term1),fol_And())) || 
	  (symbol_Equal(term_TopSymbol(term1),fol_Or()))){
	for (scan = term_ArgumentList(term1);
	     !list_Empty(scan);
	     scan = list_Cdr(scan)) {
	  vec_Push((POINTER) polterm1);
	  vec_Push(list_Car(scan));
	}
      }
  } while (bottom != vec_ActMax());
  vec_SetMax(bottom);
  printf("\nError in cnf_GetFormulaPolarity: Wrong arguments !! ");
  term_Print(term);printf("\n");
  term_Print(subterm);printf("\n");
  misc_DumpCore();
  return(-2);
}


TERM cnf_RemoveEquivImplFromFormula(TERM term)
     /**********************************************************
       INPUT:   A term term.
       RETURNS: The term  with replaced imiplications and 
                equivalences.
       CAUTION: The term is destructively changed.
       ********************************************************/
{
  TERM term1,termL,termR,termLneg,termRneg;
  LIST scan;
  int  bottom,pol;
  
  bottom = vec_ActMax();
  vec_Push(term);

  while (!(bottom == vec_ActMax())){
    term1 = (TERM)vec_PopResult();
    if (symbol_Equal(term_TopSymbol(term1),fol_Implies())){
      term_RplacTop(term1, fol_Or());
      list_Rplaca(term_ArgumentList(term1),
        term_Create(fol_Not(), list_List(list_Car(term_ArgumentList(term1)))));
    }else
      if (symbol_Equal(term_TopSymbol(term1),fol_Equiv())){
	pol      = cnf_GetFormulaPolarity(term,term1);
	termL    = (TERM)list_Car(term_ArgumentList(term1));
	termR    = (TERM)list_Second(term_ArgumentList(term1));
	termLneg = term_Create(fol_Not(),list_List(term_Copy(termL)));
	termRneg = term_Create(fol_Not(),list_List(term_Copy(termR)));
	if ((pol == 1)|| (pol == 0)){
	  term_RplacTop(term1, fol_And());
	  list_Rplaca(term_ArgumentList(term1), term_Create(fol_Or(),list_Cons(termLneg,list_List(termR))));
	  list_RplacSecond(term_ArgumentList(term1), term_Create(fol_Or(),list_Cons(termRneg,list_List(termL))));
	}else
	  if(pol == -1){
	    term_RplacTop(term1, fol_Or());
	    list_Rplaca(term_ArgumentList(term1), term_Create(fol_And(),list_Cons(termLneg,list_List(termRneg))));
	    list_RplacSecond(term_ArgumentList(term1), term_Create(fol_And(),list_Cons(termL,list_List(termR))));
	  }
      }  
    if (!list_Empty(term_ArgumentList(term1)))
      for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	vec_Push(list_Car(scan));
  }
  vec_SetMax(bottom);
  return term;
}


static SYMBOL cnf_GetDualSymbol(SYMBOL symbol)
     /**********************************************************
       INPUT:   A predefined symbol.
       RETURNS: The dual symbol.
       CAUTION: None.
       ********************************************************/
{
  SYMBOL dual;

  dual = symbol;
  if (symbol_Equal(symbol,fol_All()))
    dual = fol_Exist();
  else
    if (symbol_Equal(symbol,fol_Exist()))
      dual = fol_All();
    else
      if (symbol_Equal(symbol,fol_Or()))
	dual = fol_And();
      else
	if (symbol_Equal(symbol,fol_And()))
	  dual =  fol_Or();
  return dual;
}


TERM cnf_NegationNormalFormula(TERM term)
     /**********************************************************
       INPUT:   A term term.
       RETURNS: The negation normal form of the term.
       CAUTION: The term is destructively changed.
       ********************************************************/
{
  TERM   term1,subterm,termL;
  LIST   scan;
  SYMBOL symbol;
  int    bottom;
  
  bottom = vec_ActMax();
  vec_Push(term);

  while (!(bottom == vec_ActMax())){
    term1 = (TERM)vec_PopResult();
    if (symbol_Equal(term_TopSymbol(term1),fol_Not())){	    
      subterm = (TERM)list_Car(term_ArgumentList(term1));
      if (symbol_Equal(term_TopSymbol(subterm),fol_Not())){
	term_RplacTop(term1,term_TopSymbol(term_FirstArgument(subterm)));
	list_Delete(term_ArgumentList(term1));
	term_RplacArgumentList(term1,term_ArgumentList(term_FirstArgument(subterm)));
	term_Free(term_FirstArgument(subterm));
	list_Delete(term_ArgumentList(subterm));
	term_Free(subterm);
	vec_Push(term1);
      }else 
	if (fol_IsQuantifier(term_TopSymbol(subterm))){
	  symbol = (SYMBOL)cnf_GetDualSymbol(term_TopSymbol(subterm));
	  termL  = term_Create(fol_Not(),
			       list_List(term_SecondArgument(subterm)));
	  list_RplacSecond(term_ArgumentList(subterm), termL);
	  term_RplacTop(term1,symbol);
	  list_Delete(term_ArgumentList(term1));
	  term_RplacArgumentList(term1,list_CopyWithElement(
              term_ArgumentList(subterm),(POINTER (*)())term_Copy));
	  term_Delete(subterm);
	} else
	  if (symbol_Equal(term_TopSymbol(subterm),fol_Or()) || 
	      (symbol_Equal(term_TopSymbol(subterm),fol_And()))){
	    symbol = (SYMBOL)cnf_GetDualSymbol(term_TopSymbol(subterm));
	    for (scan = term_ArgumentList(subterm);
		 !list_Empty(scan);
		 scan = list_Cdr(scan)) {
	      termL = (TERM)list_Car(scan);
	      list_Rplaca(scan, term_Create(fol_Not(),list_List(termL)));
	    }
	    term_RplacTop(term1,symbol);
	    list_Delete(term_ArgumentList(term1));
	    term_RplacArgumentList(term1,list_CopyWithElement(
                term_ArgumentList(subterm), (POINTER (*)())term_Copy));
	    term_Delete(subterm); 
	  }
    }
    if (!list_Empty(term_ArgumentList(term1)))
      for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	vec_Push(list_Car(scan));
  }
  vec_SetMax(bottom);
  return term;
}


static TERM cnf_QuantMakeOneVar(TERM term)
/**************************************************************
  INPUT:   A term .
  RETURNS: The term where all quantifiers quantify over only one variable.
  CAUTION: The term in destrucively changed.
***************************************************************/
{
  TERM   term1,termL;
  SYMBOL quantor;
  LIST   scan,varlist;
  int    bottom;
  
  bottom = vec_ActMax();
  vec_Push(term);       

  while (bottom != vec_ActMax()) {
    term1 = (TERM)vec_PopResult();
    if (fol_IsQuantifier(term_TopSymbol(term1))){
      quantor = term_TopSymbol(term1);
      if (list_Length(term_ArgumentList(term_FirstArgument(term1))) > 1) {
	varlist =
	  list_Copy(list_Cdr(term_ArgumentList(term_FirstArgument(term1))));
	for (scan=varlist;!list_Empty(scan);scan=list_Cdr(scan)){
	  termL = term_SecondArgument(term1);
	  term_RplacSecondArgument(term1, fol_CreateQuantifier(quantor,list_List(list_Car(scan)),list_List(termL)));
	}
	for (scan=varlist;!list_Empty(scan);scan=list_Cdr(scan)) {
	  term_RplacArgumentList(term_FirstArgument(term1), 
				 list_PointerDeleteElement(term_ArgumentList(term_FirstArgument(term1)),list_Car(scan)));
	}
	list_Delete(varlist);
      }
    }
    if (!list_Empty(term_ArgumentList(term1)))
      for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	vec_Push(list_Car(scan));
  }
  vec_SetMax(bottom);
  return term;
}


static LIST cnf_GetSymbolList(LIST varlist)
/**************************************************************
  INPUT:   A list of variables
  RETURNS: The list of the sumbols of the variables.
***************************************************************/
{
  LIST scan,result;

  result = list_Nil();
  for (scan=varlist;!list_Empty(scan);scan=list_Cdr(scan))
    result = list_Cons((POINTER)term_TopSymbol((TERM)list_Car(scan)),result);

  return result;
}


static BOOL cnf_TopIsAnd(LIST termlist)
/**************************************************************
  INPUT:   A list of terms.
  RETURNS: True if one term in termlist is a conjunction.
  CAUTION: None.
***************************************************************/
{
  LIST scan;

  for (scan=termlist;!list_Empty(scan);scan=list_Cdr(scan))
    if (term_TopSymbol(list_Car(scan)) == fol_And())
      return TRUE;
  return FALSE;
}


static TERM cnf_MakeOneOr(TERM term)
/**************************************************************
  INPUT:   A term.
  RETURNS: Takes all arguments of an or together.
  CAUTION: The term in destrucively changed.
***************************************************************/
{

  LIST   scan;

  if (symbol_Equal(term_TopSymbol(term),fol_Or())){
    TERM   argterm;
    scan=term_ArgumentList(term);
    while (!list_Empty(scan)) {
      argterm = (TERM)list_Car(scan);
      cnf_MakeOneOr(argterm);
      if (symbol_Equal(term_TopSymbol(argterm),fol_Or())) {
	scan = list_Cdr(scan);
	term_RplacArgumentList(term,
	  list_Nconc(term_ArgumentList(argterm),list_PointerDeleteElement(term_ArgumentList(term),argterm)));
	term_Free(argterm);
      }
      else
	scan = list_Cdr(scan);
    }
  } else if (!symbol_IsPredicate(term_TopSymbol(term)))
    for (scan=term_ArgumentList(term);!list_Empty(scan);scan=list_Cdr(scan))
      cnf_MakeOneOr(list_Car(scan));
  
  return term;
}


static TERM cnf_MakeOneOrPredicate(TERM term)
/**************************************************************
  INPUT:   A term.
  RETURNS: Takes all predicates and negated predicates as arguments 
           of an or together.
  CAUTION: The term in destrucively changed.
***************************************************************/
{

  LIST   scan,scan1,predlist;

  if (cnf_TopIsAnd(term_ArgumentList(term))){
    
    for (scan1=term_ArgumentList(term);
	 !(list_Empty(scan1) || symbol_Equal(term_TopSymbol(list_Car(scan1)),fol_And()));
	 scan1=list_Cdr(scan1));

    if (!(list_Empty(scan1))){
      predlist = list_Nil();
      for (scan=scan1;!list_Empty(scan);scan=list_Cdr(scan)){
	if ((symbol_IsPredicate(term_TopSymbol(list_Car(scan)))) ||
	    (fol_IsNegativeLiteral(list_Car(scan)))){
	  predlist = list_Cons(list_Car(scan),predlist);
	}
      }
      for (scan=predlist;!list_Empty(scan);scan=list_Cdr(scan))
	term_RplacArgumentList(term,
	  list_PointerDeleteElement(term_ArgumentList(term),list_Car(scan)));
      
      if (!list_Empty(predlist))
	term_RplacArgumentList(term, list_Nconc(predlist,term_ArgumentList(term)));
    }
  }
  return term;
}


static TERM cnf_MakeOneOrTerm(TERM term)
/**************************************************************
  INPUT:   A term.
  RETURNS: Takes all predicates as arguments of an or together.
  CAUTION: The term in destrucively changed.
***************************************************************/
{
  return cnf_MakeOneOrPredicate(cnf_MakeOneOr(term));
}


static TERM cnf_MakeOneAnd(TERM term)
/**************************************************************
  INPUT:   A term.
  RETURNS: Takes all arguments of an and together.
  CAUTION: The term in destrucively changed.
***************************************************************/
{
  LIST   scan;

  if (symbol_Equal(term_TopSymbol(term),fol_And())) {
    TERM   argterm;
    scan=term_ArgumentList(term);
    while (!list_Empty(scan)) {
      argterm = (TERM)list_Car(scan);
      cnf_MakeOneAnd(argterm);
      if (symbol_Equal(term_TopSymbol(argterm),fol_And())) {
	scan = list_Cdr(scan);
	term_RplacArgumentList(term,
	  list_Nconc(term_ArgumentList(argterm),list_PointerDeleteElement(term_ArgumentList(term),argterm)));
	term_Free(argterm);
      }
      else
	scan = list_Cdr(scan);
    }
  } else if (!symbol_IsPredicate(term_TopSymbol(term)))
    for (scan=term_ArgumentList(term);!list_Empty(scan);scan=list_Cdr(scan))
      cnf_MakeOneAnd(list_Car(scan));

  return term;
}


static TERM cnf_MakeOneAndPredicate(TERM term)
/**************************************************************
  INPUT:   A term.
  RETURNS: Takes all predicates as arguments of one or together.
  CAUTION: The term in destrucively changed.
***************************************************************/
{
  LIST   scan,scan1,predlist;
  
  for (scan1=term_ArgumentList(term);
       !(list_Empty(scan1) || symbol_Equal(term_TopSymbol(list_Car(scan1)),fol_Or()));
       scan1=list_Cdr(scan1));
  
  if (!list_Empty(scan1)) {
    /* The car of scan1 points to a term with topsymbol 'or' */
    predlist = list_Nil();
    for (scan=scan1; !list_Empty(scan); scan=list_Cdr(scan)){
      if (symbol_IsPredicate(term_TopSymbol(list_Car(scan))) &&
	  fol_IsNegativeLiteral(list_Car(scan))) {
	predlist = list_Cons(list_Car(scan),predlist);
      }
    }
    for (scan=predlist; !list_Empty(scan); scan=list_Cdr(scan))
      term_RplacArgumentList(term, list_PointerDeleteElement(term_ArgumentList(term),list_Car(scan)));
    
    if (!list_Empty(predlist))
      term_RplacArgumentList(term, list_Nconc(predlist,term_ArgumentList(term)));
  }
  return term;
}


static TERM cnf_MakeOneAndTerm(TERM term)
/**************************************************************
  INPUT:   A term.
  RETURNS: Takes all predicates as arguments of an or together.
  CAUTION: The term in destrucively changed.
***************************************************************/
{
  return cnf_MakeOneAndPredicate(cnf_MakeOneAnd(term));
}


static void cnf_ApplyDistributivity(TERM term)
/**************************************************************
  INPUT:   A term which consists only of disjunctions and conjunctions.
  RETURNS: The term where the distributivity rule is exhaustively applied.
  CAUTION: The term in destrucively changed.
***************************************************************/
{
  if (symbol_Equal(term_TopSymbol(term),fol_Or())) {
    TERM andterm;
    LIST Scan;

    for (Scan=term_ArgumentList(term);!list_Empty(Scan);Scan=list_Cdr(Scan))
      cnf_ApplyDistributivity(list_Car(Scan));

    andterm = (TERM)NULL;

    for (Scan=term_ArgumentList(term);!(list_Empty(Scan) || andterm);Scan=list_Cdr(Scan))
      if (symbol_Equal(term_TopSymbol(list_Car(Scan)),fol_And()))
	andterm = (TERM)list_Car(Scan);

    if (andterm) {
      term_RplacArgumentList(term, list_PointerDeleteElement(term_ArgumentList(term),andterm));
      for (Scan=term_ArgumentList(andterm);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	list_Rplaca(Scan, 
	  term_Create(fol_Or(), 
	    list_Cons(list_Car(Scan),list_CopyWithElement(term_ArgumentList(term),
							    (POINTER (*)())term_Copy))));
      }
      term_RplacTop(term, fol_And());
      list_DeleteWithElement(term_ArgumentList(term),term_Delete);
      term_RplacArgumentList(term, term_ArgumentList(andterm));
      term_Free(andterm);
      cnf_ApplyDistributivity(term);
    }
  }
  else
    if (!symbol_Equal(term_TopSymbol(term),fol_Not()) &&
	!symbol_IsPredicate(term_TopSymbol(term))) {
      LIST Scan;
      for (Scan=term_ArgumentList(term);!list_Empty(Scan);Scan=list_Cdr(Scan))
	cnf_ApplyDistributivity(list_Car(Scan));
    }
}


TERM cnf_DistributivFormula(TERM term)
{
  cnf_ApplyDistributivity(term);
  cnf_MakeOneAnd(term);
  cnf_MakeOneOr(term);

  return term;
}


void cnf_FPrintClause(TERM term, FILE* file)
/**************************************************************
  INPUT:   A term and a file pointer.
  RETURNS: Print the term which contains only disjunctions to file.
           The disjunctions represent a clause.
  CAUTION: None.
***************************************************************/
{
    
 TERM   term1;
 LIST   scan;
 int    bottom;
  
  bottom = vec_ActMax();
  vec_Push(term);       

  while (!(bottom == vec_ActMax())){
    term1 = (TERM)vec_PopResult();
    if (symbol_Equal(term_TopSymbol(term1),fol_Or())){
      for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	vec_Push(list_Car(scan));
    }else
      term_FPrint(file,term1);
  }
 fprintf(file,".\n");
 vec_SetMax(bottom);
}


void cnf_FPrint(TERM term, FILE* file)
/**************************************************************
  INPUT:   A term and a file pointer.
  RETURNS: Print the term (in negation normal form)
           which contains only conjunctins of 
	   disjunctions to file. The conjunctions are interpreted 
	   to represent different clauses.
  CAUTION: None.
***************************************************************/
{
  TERM   term1;
  LIST   scan;
  int    bottom;
  
  bottom = vec_ActMax();
  vec_Push(term);       

  while (!(bottom == vec_ActMax())){
    term1 = (TERM)vec_PopResult();
    if (symbol_Equal(term_TopSymbol(term1),fol_And())){
      for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	vec_Push(list_Car(scan));
    }else
      if ((symbol_Equal(term_TopSymbol(term1),fol_Or())) ||
	  (symbol_IsPredicate(term_TopSymbol(term1)))  ||
	  (symbol_Equal(term_TopSymbol(term1),fol_Not())))
	cnf_FPrintClause(term1,file);
  }
  vec_SetMax(bottom);
}


void cnf_StdoutPrint(TERM term)
/**************************************************************
  INPUT:   A term and a file pointer.
  RETURNS: Print the term (in negation normal form)
           which contains only conjunctins of 
	   disjunctions to file. The conjunctions are interpreted 
	   to represent different clauses.
  CAUTION: None.
***************************************************************/
{
  LIST termlist,scan,scan1;

  for (scan=term_ArgumentList(term);!list_Empty(scan);scan=list_Cdr(scan)){
    termlist = list_Nil();
    if (!((symbol_IsPredicate(term_TopSymbol(list_Car(scan)))) ||
	  (fol_IsNegativeLiteral(list_Car(scan)))))
      termlist = term_ArgumentList(list_Car(scan));
    
    if (!list_Empty(termlist)){
      for (scan1=termlist;!list_Empty(scan1);scan1=list_Cdr(scan1))
	term_Print(list_Car(scan1));
      printf(".\n");
    }else{
      term_Print(list_Car(scan));
      printf(".\n");
    }
  }
  
}


void cnf_FilePrint(TERM term, FILE* file)
/**************************************************************
  INPUT:   A term and a file pointer.
  RETURNS: Print the term (in negation normal form)
           which contains only conjunctins of 
	   disjunctions to file. The conjunctions are interpreted 
	   to represent different clauses.
  CAUTION: None.
***************************************************************/
{
  LIST termlist,scan,scan1;

  for (scan=term_ArgumentList(term);!list_Empty(scan);scan=list_Cdr(scan)){
    termlist = list_Nil();
    if (!((symbol_IsPredicate(term_TopSymbol(list_Car(scan)))) ||
	  (fol_IsNegativeLiteral(list_Car(scan)))))
      termlist = term_ArgumentList(list_Car(scan));
    
    if (!list_Empty(termlist)){
      for (scan1=termlist;!list_Empty(scan1);scan1=list_Cdr(scan1))
	term_FPrint(file,list_Car(scan1));
      fprintf(file,".\n");
    }else{
      term_FPrint(file,list_Car(scan));
      fprintf(file,".\n");
    }
  }
  
}


void cnf_FilePrintPrefix(TERM term, FILE* file)
/**************************************************************
  INPUT:   A term and a file pointer.
  RETURNS: Prefix Print the term (in negation normal form)
           which contains only conjunctins of 
	   disjunctions to file. The conjunctions are interpreted 
	   to represent different clauses.
  CAUTION: None.
***************************************************************/
{
  LIST termlist,scan,scan1;

  for (scan=term_ArgumentList(term);!list_Empty(scan);scan=list_Cdr(scan)){
    termlist = list_Nil();
    if (!((symbol_IsPredicate(term_TopSymbol(list_Car(scan)))) ||
	  (fol_IsNegativeLiteral(list_Car(scan)))))
      termlist = term_ArgumentList(list_Car(scan));
    
    if (!list_Empty(termlist)){
      for (scan1=termlist;!list_Empty(scan1);scan1=list_Cdr(scan1)) {
	term_FPrintPrefix(file,list_Car(scan1));
	if (!list_Empty(list_Cdr(scan1)))
	  fprintf(file," | ");
      }
      fprintf(file,".\n");
    } else {
      term_FPrintPrefix(file,list_Car(scan));
      fprintf(file,".\n");
    }
  }
}



static LIST cnf_SubsumeClauseList(LIST clauselist)
/**********************************************************
  INPUT:   A list of clauses.
  RETURNS: The list of clauses without subsumed clauses.
  CAUTION: The list is destructively changed.
***********************************************************/
{
  LIST     scan,result;
  st_INDEX stindex;
  CLAUSE   actclause;

  stindex = st_IndexCreate();
  result  = list_Nil();

  for (scan=clauselist;!list_Empty(scan);scan=list_Cdr(scan))
    res_InsertClauseIndex(list_Car(scan),stindex);

  for (scan=clauselist;!list_Empty(scan);scan=list_Cdr(scan)){
    actclause     = list_Car(scan);
    res_DeleteClauseIndex(actclause,stindex);
    if (!(res_BackSubWithLength(actclause,stindex))){
      res_InsertClauseIndex(actclause,stindex);
      result = list_Cons(actclause,result);
    }
  }
  if (!(list_Length(result) == list_Length(clauselist))){
    for (scan=result;!list_Empty(scan);scan=list_Cdr(scan))
      clauselist = list_PointerDeleteElement(clauselist,list_Car(scan));
    if (!list_Empty(result))
      list_DeleteWithElement(clauselist,clause_Delete);
  }else
    list_Delete(clauselist);
  st_IndexDelete(stindex);
  return result;
}


LIST cnf_MakeClauseList(TERM term, BOOL Sorts, BOOL Conclause)
/**************************************************************
  INPUT:   A term in cnf and two boolean values indicating
           whether sorts should be generated and whether the
	   generated clauses are Conclauses.
  RETURNS: A list of clauses with respect to term. The terms 
           in the new clauses are the copied subterms from term.
  CAUTION: None.
***************************************************************/
{
  LIST   termlist,scan,clauselist,newclauselist,delclauselist,condlist;
  CLAUSE clause;
  int    j;

  termlist    = list_Nil();
  clauselist  = list_Nil();

  if (fol_IsTrue(term))
    return  clauselist;

  if ((fol_IsNegativeLiteral(term))||
      (symbol_IsPredicate(term_TopSymbol(term)))){
    termlist = list_List(term_Copy(term));
    clause = clause_CreateFromLiterals(termlist,Sorts,Conclause);
    clauselist = list_Nconc(clauselist,list_List(clause));
    term_StartMinRenaming();
    term_Rename(clause_GetLiteralTerm(clause,0));
    list_Delete(termlist);
    return clauselist;
  }

  delclauselist = list_Nil();
  term          = cnf_MakeOneAndTerm(term);
  if (symbol_Equal(term_TopSymbol(term), fol_And())){
    for (scan=term_ArgumentList(term);!list_Empty(scan);scan=list_Cdr(scan)) {
      list_Rplaca(scan, cnf_MakeOneOrTerm(list_Car(scan)));
      termlist = list_Nil();
      if (!(symbol_IsPredicate(term_TopSymbol(list_Car(scan))) ||
	    fol_IsNegativeLiteral(list_Car(scan)))) {
	termlist = list_CopyWithElement(term_ArgumentList(list_Car(scan)),
					(POINTER (*)())term_Copy);
	termlist = list_DeleteDuplicatesFree(termlist,term_Equal,term_Delete);
      }else
	termlist = list_List(term_Copy(list_Car(scan)));
  
      if (!list_Empty(termlist)){
	clause     = clause_CreateFromLiterals(termlist,Sorts,Conclause);
	term_StartMinRenaming();
	for (j=0;j<clause_Length(clause);j++)
	  term_Rename(clause_GetLiteralTerm(clause,j));
	clauselist = list_Cons(clause,clauselist);
	list_Delete(termlist);
      }
    }
  } else {
    /* Here the term is a disjunction, i.e. there is only one clause */
    term = cnf_MakeOneOrTerm(term);
    if (!((symbol_IsPredicate(term_TopSymbol(term))) ||
	  (fol_IsNegativeLiteral(term)))){
      termlist = list_CopyWithElement(term_ArgumentList(term),
				      (POINTER (*)())term_Copy);
      termlist = list_DeleteDuplicatesFree(termlist,term_Equal,term_Delete);
    } else
      termlist = list_List(term_Copy(term));
    
    if (!list_Empty(termlist)){
      clause     = clause_CreateFromLiterals(termlist,Sorts,Conclause);
      term_StartMinRenaming();
      for (j = 0; j < clause_Length(clause); j++)
	term_Rename(clause_GetLiteralTerm(clause,j));
      clauselist = list_Cons(clause,clauselist);
      list_Delete(termlist);
    }
  }

  for (scan=clauselist; !list_Empty(scan); scan=list_Cdr(scan)){
    condlist = cond_CondFast(list_Car(scan));
    if (!list_Empty(condlist))
      clause_DeleteLiterals(list_Car(scan), condlist);
    list_Delete(condlist);
  }
  clauselist = cnf_SubsumeClauseList(clauselist);
  newclauselist = list_Nil();
  while (!list_Empty(clauselist)){
    clause = res_SelectLightestClause(clauselist);
    newclauselist = list_Nconc(newclauselist,list_List(clause));
    clauselist = list_PointerDeleteElement(clauselist,clause);
  }
  list_Delete(clauselist);
  for (scan=newclauselist;!list_Empty(scan);scan=list_Cdr(scan)){
    if (res_HasTautology(list_Car(scan)))
      delclauselist = list_Cons(list_Car(scan),delclauselist);
  }
  for (scan=delclauselist;!list_Empty(scan);scan=list_Cdr(scan))
    newclauselist = list_PointerDeleteElement(newclauselist,list_Car(scan));
  list_DeleteWithElement(delclauselist,clause_Delete);
    
  return newclauselist;
}

static TERM cnf_Flatten(TERM Term, SYMBOL Symbol)
/**************************************************************
  INPUT:   A <Term> and <Symbol> that is assumed to be associative.
  RETURNS: If the top symol of <Term> is <Symbol> the <Term> is flattened
           as long as it contains further direct subterms starting with <Symbol>
  CAUTION: The term in destrucively changed.
***************************************************************/
{
  LIST   Scan1,Scan2;

  if (symbol_Equal(term_TopSymbol(Term), Symbol)) {
    TERM   Argterm;
    Scan1 =term_ArgumentList(Term);
    while (!list_Empty(Scan1)) {
      Argterm = (TERM)list_Car(Scan1);
      Scan2   = list_Cdr(Scan1);
      if (symbol_Equal(term_TopSymbol(Argterm),Symbol)) {
	cnf_Flatten(Argterm,Symbol);
	list_NInsert(Scan1,list_Cdr(term_ArgumentList(Argterm)));   /* Be efficent ... */
	list_Rplaca(Scan1,list_Car(term_ArgumentList(Argterm)));
	list_Free(term_ArgumentList(Argterm));
	term_Free(Argterm);
      }
      Scan1  = Scan2;
    }
  }
  return Term;
}

static TERM cnf_DistrQuantorNoVarSub(TERM Term)
/**************************************************************
  INPUT:   A formula in negation normal form starting with a universal 
           (existential) quantifier and a disjunction (conjunction) as argument.
  RETURNS: The term after application of anti-prenexing. Quantifiers
           are moved inside as long as possible.
  CAUTION: The term in destrucively changed.
***************************************************************/
{
  LIST   Variables,Subformulas,Scan1,Scan2,Rest;
  TERM   Subterm,Var,NewForm;
  SYMBOL Subtop,Top;

  /*printf("\nAN0:\t");term_Print(Term);*/

  Top       = term_TopSymbol(Term);
  Variables = list_Copy(fol_QuantifierVariables(Term));
  Subterm   = term_SecondArgument(Term);
  Subtop    = term_TopSymbol(Subterm);
  Subterm   = cnf_Flatten(Subterm,Subtop);

  /*printf("\nAN1:\t");term_Print(Subterm);*/

  for (Scan1=Variables;!list_Empty(Scan1);Scan1=list_Cdr(Scan1)) {
    Subformulas = list_Nil();
    Var         = (TERM)list_Car(Scan1);
    for(Scan2=term_ArgumentList(Subterm);!list_Empty(Scan2);Scan2=list_Cdr(Scan2))
      if (!fol_VarOccursFreely(Var,list_Car(Scan2)))
	Subformulas = list_Cons(list_Car(Scan2),Subformulas);
    if (!list_Empty(Subformulas)) {
      Rest     = list_NPointerDifference(term_ArgumentList(Subterm),Subformulas);
      if (list_Empty(list_Cdr(Rest)))/* One subformula */
	if (symbol_Equal(Top,term_TopSymbol(list_Car(Rest)))) { /* Optimization: quantifier still exist */
	  NewForm = (TERM)list_Car(Rest);
	  term_RplacArgumentList(term_FirstArgument(NewForm),
				 list_Cons((POINTER)Var,fol_QuantifierVariables(NewForm)));
	  list_Delete(Rest);
	}
	else
	  NewForm = fol_CreateQuantifier(Top,list_List((POINTER)Var),Rest);
      else
	NewForm = fol_CreateQuantifier(Top,list_List((POINTER)Var),list_List(term_Create(Subtop,Rest)));
      term_RplacArgumentList(Subterm,list_Cons(cnf_AntiPrenex(NewForm),Subformulas));
      term_RplacArgumentList(term_FirstArgument(Term),
			     list_PointerDeleteElement(fol_QuantifierVariables(Term),(POINTER)Var));
    } 
  }

  if (list_Empty(fol_QuantifierVariables(Term))) { /* All variables moved inside */
    term_Free(term_FirstArgument(Term));
    list_Delete(term_ArgumentList(Term));
    term_RplacTop(Term,Subtop);
    term_RplacArgumentList(Term,term_ArgumentList(Subterm));
    term_Free(Subterm);
  }
  list_Delete(Variables);
  
  return Term;
}

TERM cnf_AntiPrenex(TERM Term)
/**************************************************************
  INPUT:   A formula in negation normal form.
  RETURNS: The term after application of anti-prenexing. Quantifiers
           are moved inside as long as possible.
  CAUTION: The term in destrucively changed.
***************************************************************/
{
  LIST   Scan;
  SYMBOL Top;

  Top = term_TopSymbol(Term);

  if (fol_IsQuantifier(Top)) {
    TERM   Subterm,Actterm;
    SYMBOL DistrSymb,Subtop;          /* The junctor the respective quantifier distributes over */
    
    Subterm = term_SecondArgument(Term);
    Subtop  = term_TopSymbol(Subterm);

    if (!symbol_Equal(Subtop,fol_Not()) && !symbol_IsPredicate(Subtop)) { /* No Literals or Atoms */
      if (symbol_Equal(Top,fol_All()))
	DistrSymb = fol_And();
      else
	DistrSymb = fol_Or();
      if (fol_IsQuantifier(Subtop)) {
	cnf_AntiPrenex(Subterm);
	Subtop = term_TopSymbol(Subterm);
      }
      if (symbol_Equal(Subtop,DistrSymb)) {
	LIST Variables;	
	LIST NewVars;
	Variables = fol_QuantifierVariables(Term);
	Subterm   = cnf_Flatten(Subterm,DistrSymb);
	for (Scan=term_ArgumentList(Subterm);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	  Actterm = (TERM)list_Car(Scan);
	  NewVars = list_NIntersect(fol_FreeVariables(Actterm),Variables,term_Equal);
	  if (!list_Empty(NewVars))
	    if (symbol_Equal(Top,term_TopSymbol(Actterm)))           /* Quantor already there */
	      term_RplacArgumentList(term_FirstArgument(Actterm),
				     list_Nconc(fol_QuantifierVariables(Actterm),
						list_NMapCar(NewVars,(POINTER (*)())term_Copy)));
	    else
	      list_Rplaca(Scan,fol_CreateQuantifier(Top,list_NMapCar(NewVars,(POINTER (*)())term_Copy),
						    list_List(Actterm)));
	}
	term_Delete(term_FirstArgument(Term));   /* Delete old variable list */
	list_Delete(term_ArgumentList(Term)); 
	term_RplacTop(Term,DistrSymb);
	term_RplacArgumentList(Term,term_ArgumentList(Subterm));
	term_Free(Subterm);
	for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan)) 
	  list_Rplaca(Scan,cnf_AntiPrenex(list_Car(Scan)));
      }
      else 
	if (!fol_IsQuantifier(Subtop))
	  Term = cnf_DistrQuantorNoVarSub(Term);
    }
  }
  else
    if (!symbol_Equal(Top,fol_Not()) && !symbol_IsPredicate(Top))
      for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
	cnf_AntiPrenex(list_Car(Scan));

  return Term;
}

static TERM  cnf_RemoveTrivialOperators(TERM Term)
/**************************************************************
  INPUT:   A formula.
  RETURNS: The formula after
           removal of "or" and "and" with only one argument
  CAUTION: The term in destrucively changed.
***************************************************************/
{
  SYMBOL Top;
  LIST   Scan;

  Top = term_TopSymbol(Term);

  if (symbol_IsPredicate(Top))
    return Term;

  if ((symbol_Equal(Top,fol_And()) || symbol_Equal(Top,fol_Or())) &&
      list_Empty(list_Cdr(term_ArgumentList(Term)))) {
    TERM Result;
    Result = term_FirstArgument(Term);
    list_Delete(term_ArgumentList(Term));
    term_Free(Term);
    return cnf_RemoveTrivialOperators(Result);
  }

  for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
    list_Rplaca(Scan,cnf_RemoveTrivialOperators(list_Car(Scan)));
  
  return Term;
}

static TERM  cnf_SimplifyQuantors(TERM Term)
/**************************************************************
  INPUT:   A formula.
  RETURNS: The formula after
	   removal of bindings of variables that don't occur in the
	   respective subformula and possible mergings of quantors
  CAUTION: The term is destrucively changed.
***************************************************************/
{
   SYMBOL Top;
   LIST   Scan;

  Top = term_TopSymbol(Term);

  if (symbol_IsPredicate(Top) || symbol_Equal(Top,fol_Varlist()))
    return Term;

  if (fol_IsQuantifier(Top)) {
    LIST Vars;
    TERM Var,Subterm,Aux;
    Vars    = list_Nil();
    Subterm = term_SecondArgument(Term);

    while (symbol_Equal(term_TopSymbol(Subterm),Top)) {
      term_RplacArgumentList(term_FirstArgument(Term),
			     list_Nconc(fol_QuantifierVariables(Term),fol_QuantifierVariables(Subterm)));
      term_Free(term_FirstArgument(Subterm));
      Aux = term_SecondArgument(Subterm);
      list_Delete(term_ArgumentList(Subterm));
      term_Free(Subterm);
      list_Rplaca(list_Cdr(term_ArgumentList(Term)),Aux);
      Subterm = Aux;
    }
    
    for(Scan=fol_QuantifierVariables(Term);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Var = (TERM)list_Car(Scan);
      if (!fol_VarOccursFreely(Var,Subterm))
	Vars = list_Cons(Var,Vars);
    }
    if (!list_Empty(Vars)) {
      Subterm = term_FirstArgument(Term);
      term_RplacArgumentList(Subterm,list_NPointerDifference(term_ArgumentList(Subterm),Vars));
      list_DeleteWithElement(Vars,term_Delete);
      if (list_Empty(term_ArgumentList(Subterm))) {
	Subterm = term_SecondArgument(Term);
	term_Delete(term_FirstArgument(Term));
	list_Delete(term_ArgumentList(Term));
	term_Free(Term);
	return cnf_SimplifyQuantors(Subterm);
      }
    }
  }
  
  for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
    list_Rplaca(Scan,cnf_SimplifyQuantors(list_Car(Scan)));
  
  return Term; 
}

TERM  cnf_RemoveTrivialAtoms(TERM Term)
/**************************************************************
  INPUT:   A formula.
  RETURNS: The formula where occurrences of the atoms "true"
           and "false" are propagated and eventually removed.
  CAUTION: The term is destrucively changed.
***************************************************************/
{
  SYMBOL Top,Subtop;
  LIST   Scan;
  TERM   Result;
  BOOL   Update;

  if (!term_IsComplex(Term))
    return Term;
  
  Top    = term_TopSymbol(Term);
  Update = FALSE;

  if (symbol_Equal(Top,fol_And())) {
    Scan  = term_ArgumentList(Term);
    while (!list_Empty(Scan)) {
      Result = cnf_RemoveTrivialAtoms(list_Car(Scan));
      Subtop = term_TopSymbol(Result);
      if (symbol_Equal(Subtop,fol_True()))
	Update = TRUE;
      else
	if (symbol_Equal(Subtop,fol_False())) {
	  term_RplacTop(Term,fol_False());
	  list_DeleteWithElement(term_ArgumentList(Term),term_Delete);
	  term_RplacArgumentList(Term,list_Nil());
	  return Term;
	}
      Scan = list_Cdr(Scan);
    }
    if (Update) {
      term_RplacArgumentList(Term,list_DeleteElementIf(term_ArgumentList(Term),fol_IsTrue,term_Delete));
      if (list_Empty(term_ArgumentList(Term))) {
	term_RplacTop(Term,fol_True());
	return Term;
      }
    }
  }
  else if (symbol_Equal(Top,fol_Or())) {
    Scan  = term_ArgumentList(Term);
    while (!list_Empty(Scan)) {
      Result = cnf_RemoveTrivialAtoms(list_Car(Scan));
      Subtop = term_TopSymbol(Result);
      if (symbol_Equal(Subtop,fol_False()))
	Update = TRUE;
      else
	if (symbol_Equal(Subtop,fol_True())) {
	  term_RplacTop(Term,fol_True());
	  list_DeleteWithElement(term_ArgumentList(Term),term_Delete);
	  term_RplacArgumentList(Term,list_Nil());
	  return Term;
	}
      Scan = list_Cdr(Scan);
    }
    if (Update) {
      term_RplacArgumentList(Term,list_DeleteElementIf(term_ArgumentList(Term),fol_IsFalse,term_Delete));
      if (list_Empty(term_ArgumentList(Term))) {
	term_RplacTop(Term,fol_False());
	return Term;
      }
    }
  }
  else if (fol_IsQuantifier(Top) || symbol_Equal(Top,fol_Not())) {
    if (fol_IsQuantifier(Top))
      Result = cnf_RemoveTrivialAtoms(term_SecondArgument(Term));
    else
      Result = cnf_RemoveTrivialAtoms(term_FirstArgument(Term));
    Subtop = term_TopSymbol(Result);
    if ((symbol_Equal(Subtop,fol_False()) && symbol_Equal(Top,fol_Not())) ||
	(symbol_Equal(Subtop,fol_True()) && fol_IsQuantifier(Top))) {
      term_RplacTop(Term,fol_True());
      list_DeleteWithElement(term_ArgumentList(Term),term_Delete);
      term_RplacArgumentList(Term,list_Nil());
      return Term;
    }
    else
      if ((symbol_Equal(Subtop,fol_True()) && symbol_Equal(Top,fol_Not())) ||
	  (symbol_Equal(Subtop,fol_False()) && fol_IsQuantifier(Top))) {
	term_RplacTop(Term,fol_False());
	list_DeleteWithElement(term_ArgumentList(Term),term_Delete);
	term_RplacArgumentList(Term,list_Nil());
	return Term;
      }
  }
  else if (symbol_Equal(Top,fol_Implies())) {
    Result = cnf_RemoveTrivialAtoms(term_FirstArgument(Term));
    Subtop = term_TopSymbol(Result);
    if (symbol_Equal(Subtop,fol_False())) {
      term_RplacTop(Term,fol_True());
      list_DeleteWithElement(term_ArgumentList(Term),term_Delete);
      term_RplacArgumentList(Term,list_Nil());
      return Term;
    }
    else if (symbol_Equal(Subtop,fol_True())) {
      term_Delete(Result);
      Result = term_SecondArgument(Term);
      list_Delete(term_ArgumentList(Term));
      term_RplacTop(Term,term_TopSymbol(Result));
      term_RplacArgumentList(Term,term_ArgumentList(Result));
      term_Free(Result);
      return cnf_RemoveTrivialAtoms(Term);
    }
    Result = cnf_RemoveTrivialAtoms(term_SecondArgument(Term));
    Subtop = term_TopSymbol(Result);
    if (symbol_Equal(Subtop,fol_True())) {
      term_RplacTop(Term,fol_True());
      list_DeleteWithElement(term_ArgumentList(Term),term_Delete);
      term_RplacArgumentList(Term,list_Nil());
      return Term;
    }
    else if (symbol_Equal(Subtop,fol_False())) {
      term_RplacTop(Term,fol_Not());
      term_RplacArgumentList(Term,list_DeleteElementIf(term_ArgumentList(Term),fol_IsFalse,term_Delete));
    }
  }
  else if (symbol_Equal(Top,fol_Equiv())) {
    Result = cnf_RemoveTrivialAtoms(term_FirstArgument(Term));
    Subtop = term_TopSymbol(Result);
    if (symbol_Equal(Subtop,fol_False())) {
      term_RplacTop(Term,fol_Not());
      term_RplacArgumentList(Term,list_DeleteElementIf(term_ArgumentList(Term),fol_IsFalse,term_Delete));
      return cnf_RemoveTrivialAtoms(Term);
    }
    else if (symbol_Equal(Subtop,fol_True())) {
      term_Delete(Result);
      Result = term_SecondArgument(Term);
      list_Delete(term_ArgumentList(Term));
      term_RplacTop(Term,term_TopSymbol(Result));
      term_RplacArgumentList(Term,term_ArgumentList(Result));
      term_Free(Result);
      return cnf_RemoveTrivialAtoms(Term);
    }
    Result = cnf_RemoveTrivialAtoms(term_SecondArgument(Term));
    Subtop = term_TopSymbol(Result);
    if (symbol_Equal(Subtop,fol_False())) {
      term_RplacTop(Term,fol_Not());
      term_RplacArgumentList(Term,list_DeleteElementIf(term_ArgumentList(Term),fol_IsFalse,term_Delete));
    }
    else if (symbol_Equal(Subtop,fol_True())) {
      term_Delete(Result);
      Result = term_FirstArgument(Term);
      list_Delete(term_ArgumentList(Term));
      term_RplacTop(Term,term_TopSymbol(Result));
      term_RplacArgumentList(Term,term_ArgumentList(Result));
      term_Free(Result);
    }
  }

  return Term;
}

TERM  cnf_ObviousSimplifications(TERM Term)
/**************************************************************
  INPUT:   A formula.
  RETURNS: The formula after performing the following simplifications:
             - remove "or" and "and" with only one argument
	     - remove bindings of variables that don't occur in the
	       respective subformula
	     - merge quantors
  CAUTION: The term is destrucively changed.
***************************************************************/
{

  Term = cnf_RemoveTrivialAtoms(Term);
  Term = cnf_RemoveTrivialOperators(Term);
  Term = cnf_SimplifyQuantors(Term);

  return Term;
}

TERM cnf_SkolemFormula(TERM Term, LIST* Symblist)
/**************************************************************
  INPUT:   A term pointer term in negation normal form.
  RETURNS: The skolemized term and the list of introduced Skolem functions.
  CAUTION: The term in destrucively changed.
***************************************************************/
{
  SYMBOL Top,SkolemSymbol;
  TERM   Subterm,SkolemTerm;
  LIST   Varlist,Scan;
  NAT    Arity;

  Top = term_TopSymbol(Term);
    
  if (fol_IsQuantifier(term_TopSymbol(Term)))
    if (symbol_Equal(Top,fol_All())){
      term_Delete(term_FirstArgument(Term));
      Subterm = term_SecondArgument(Term);
      list_Delete(term_ArgumentList(Term));
      term_RplacTop(Term,term_TopSymbol(Subterm));
      term_RplacArgumentList(Term,term_ArgumentList(Subterm));
      term_Free(Subterm);
      return cnf_SkolemFormula(Term,Symblist);
    } else {
      Varlist = fol_FreeVariables(Term);
      Arity   = list_Length(Varlist);
      for (Scan=fol_QuantifierVariables(Term);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	SkolemSymbol = symbol_CreateSkolemFunction(Arity);
	*Symblist  = list_Cons((POINTER)SkolemSymbol,*Symblist);
	SkolemTerm = term_Create(SkolemSymbol,Varlist); /* Caution: Sharing of Varlist ! */
	term_ReplaceVariable(term_SecondArgument(Term),term_TopSymbol(list_Car(Scan)),SkolemTerm);
	term_Free(SkolemTerm);
      }
      list_Delete(Varlist);
      term_Delete(term_FirstArgument(Term));
      Subterm = term_SecondArgument(Term);
      list_Delete(term_ArgumentList(Term));
      term_RplacTop(Term,term_TopSymbol(Subterm));
      term_RplacArgumentList(Term,term_ArgumentList(Subterm));
      term_Free(Subterm);
      return cnf_SkolemFormula(Term,Symblist);
    } else
      if (symbol_Equal(Top,fol_And()) || symbol_Equal(Top,fol_Or()))
	for (Scan=term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan))
	  cnf_SkolemFormula(list_Car(Scan), Symblist);
  return Term;
}

static TERM cnf_Cnf(TERM term, LIST* symblist)
/**************************************************************
  INPUT:   A term and a list of symbols.
  RETURNS: The term is transformed to conjunctive normal form.
  CAUTION: The term is destructively changed and not normalized.
***************************************************************/
{
  term     = ren_Rename(term,symblist);
  term     = cnf_RemoveEquivImplFromFormula(term);
  term     = cnf_NegationNormalFormula(term);
  term     = cnf_SkolemFormula(cnf_AntiPrenex(term), symblist);
  term     = cnf_DistributivFormula(term);

  return term;
}

static BOOL cnf_HaveProofOptSkolem(PROOFSEARCH Search, LIST NonConClauses, TERM term2)
/**************************************************************
  INPUT:   Two terms.
  RETURNS: True if res_ResSubsIndex finds a proof for (impl term1 term2).
  CAUTION: The term in destrucively changed.
  ***************************************************************/
{
    LIST ConClauses, EmptyClauses;
    LIST scan,symblist;
    BOOL   found;
    LIST Usables;
    NAT  Depth;

    ConClauses = list_Nil();
    symblist   = list_Nil();
    found      = FALSE;
  
    /* List of clauses from term2 */
    term2      = cnf_Cnf(term2,&symblist);
    Usables    = cnf_MakeClauseList(term2,FALSE,FALSE);
    Depth      = clause_MaxTermDepthClauseList(Usables);
    term_Delete(term2);

    for (scan=Usables;!list_Empty(scan); scan = list_Cdr(scan))
      clause_SetFlag(list_Car(scan),CONCLAUSE);

    EmptyClauses = cnf_SatUnit(Search, Usables);
    if (!list_Empty(EmptyClauses)) {
      found = TRUE;
      list_DeleteWithElement(EmptyClauses, clause_Delete);
    }
    
    /* Removing ConClauses from UsablesIndex */
    ConClauses = list_Copy(prfs_UsableClauses(Search));
    for(scan = ConClauses; !list_Empty(scan); scan = list_Cdr(scan))
      prfs_DeleteUsable(Search, (CLAUSE) list_Car(scan));
    list_Delete(ConClauses);
	
    list_DeleteWithElement(symblist,symbol_Delete);
    return found;
}

static TERM cnf_RplacVarsymbFunction(TERM term, SYMBOL varsymb,
					TERM function)
/**********************************************************
  INPUT:   A terms,a variable symbol and a function.
  RETURNS: The variable with the symbol  varsymb in the term term 
           is replaced by the function function.
  CAUTION: The term is destructively changed.
***********************************************************/
{
    int  bottom;
  TERM term1;
  LIST scan;

  bottom = vec_ActMax();
  vec_Push(term);

  while (!(bottom == vec_ActMax())){
    term1 = (TERM)vec_PopResult();
    if (symbol_Equal(term_TopSymbol(term1),varsymb)){
      term_RplacTop(term1,term_TopSymbol(function));
      term_RplacArgumentList(term1,list_CopyWithElement(term_ArgumentList(function),( POINTER (*)())term_Copy));
    }else
      if (!list_Empty(term_ArgumentList(term1)))
	for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	  vec_Push(list_Car(scan));
  }
  vec_SetMax(bottom);
  return term;
}


static TERM cnf_MakeSkolemfunction(LIST varlist)
/**************************************************************
  INPUT:   A  list of variables.
  RETURNS: The new term oskf... (oskc...) which is a function 
           with varlist as arguments.
  CAUTION: None.
***************************************************************/
{
  TERM   term;
  SYMBOL skolem;
  int    n;
  
  n = list_Length(varlist);
  skolem    = symbol_CreateSkolemFunction(n);
  term      = term_Create(skolem,list_CopyWithElement(varlist,(POINTER (*)())term_Copy));
  return term;
}


static void cnf_PopAllQuantifier(TERM term) {
    /********************************************************
      INPUT:   A term whose top symbol is fol_all
      EFFECT:  Removes the quantifier
      RETURNS: A list of the quantifiers variables
      *******************************************************/
    TERM SubTerm;
    LIST VarList;
#ifdef DEBUG
    if (term_TopSymbol(term) != fol_All()) {
	printf("Top symbol of ");
	term_Print(term);
	printf(" is not fol_All !\n");
	misc_DumpCore();
    }
#endif
    VarList = term_ArgumentList(term_FirstArgument(term));
    list_DeleteWithElement(VarList, term_Delete);
    term_Free(term_FirstArgument(term));
    SubTerm = term_SecondArgument(term);
    list_Delete(term_ArgumentList(term));
    term_RplacTop(term,term_TopSymbol(SubTerm));
    term_RplacArgumentList(term,term_ArgumentList(SubTerm));
    term_Free(SubTerm);
}

static TERM cnf_QuantifyAndNegate(TERM term, LIST VarList, LIST FreeList) {
    /****************************************************************
      INPUT: A term, a list of variables to be exist-quantified,
      a list of variables to be all-quantified
      OUTPUT: not(forall[FreeList](exists[VarList](term)))
      MEMORY: The term, the lists and their arguments are copied.
      ***************************************************************/
    TERM Result;
    TERM TermCopy;
    LIST VarListCopy;
    LIST FreeListCopy;

    TermCopy = term_Copy(term);
    VarListCopy = list_CopyWithElement(VarList, (POINTER (*) ()) term_Copy);
    Result = fol_CreateQuantifier(fol_Exist(),VarListCopy,list_List(TermCopy));
    FreeListCopy = list_Nil();

    FreeList = fol_FreeVariables(Result);
    if (!list_Empty(FreeList)) {
      FreeListCopy = list_CopyWithElement(FreeList, 
					  (POINTER (*) ()) term_Copy);
      list_Delete(FreeList);
      Result = fol_CreateQuantifier(fol_All(), FreeListCopy, 
				    list_List(Result));
    }
    Result = term_Create(fol_Not(), list_List(Result));
    return Result;
}

static TERM cnf_MoveProvedTermToTopLevel(TERM Term, TERM Term1, TERM Proved,
				     LIST VarList, LIST FreeList) {
    /********************************************************************
      INPUT: A top-level term Term, which must be a conjunction, 
      a subterm Term1 of Term,
      a subterm Proved of Term1, a list of existence quantified variables
      VarList and a list of free variables FreeList.
      Term1 is of the form exists[...](t1 and t2 and ... and Proved and ..)
      EFFECT: Proved is removed from the arguments of Term1.
      *******************************************************************/

    TERM termR;
    TERM skolemfunction;
    SYMBOL varsymb;
    LIST scan;
    termR = term_SecondArgument(Term1); /* t1 and t2 and ... and Proved ... */
    term_RplacArgumentList(termR,
			   list_PointerDeleteElement(term_ArgumentList(termR),
						     Proved));
    if (list_Length(term_ArgumentList(termR)) < 2) {
	TERM termRL = term_FirstArgument(termR);     /* t1 */
	list_Delete(term_ArgumentList(termR));
	term_RplacTop(termR, term_TopSymbol(termRL));
	term_RplacArgumentList(termR,term_ArgumentList(termRL));
	term_Free(termRL);
    }

/*     varsymb = term_TopSymbol(list_Car(VarList)); */
/*     skolemfunction = cnf_MakeSkolemfunction(FreeList); */
/*     termR = cnf_RplacVarsymbFunction(termR,varsymb,skolemfunction); */
/*     Proved = cnf_RplacVarsymbFunction(Proved,varsymb,skolemfunction); */
/*     term_Delete(skolemfunction); */
    for(scan = VarList; scan != list_Nil(); scan = list_Cdr(scan)) {
	varsymb = term_TopSymbol(list_Car(scan));
	skolemfunction = cnf_MakeSkolemfunction(FreeList);
	termR = cnf_RplacVarsymbFunction(termR,varsymb,skolemfunction);
	Proved = cnf_RplacVarsymbFunction(Proved,varsymb,skolemfunction); 
	term_Delete(skolemfunction); 
    }

    if (!list_Empty(FreeList)) {
	Proved = 
	    fol_CreateQuantifier(fol_All(),
				 list_CopyWithElement(FreeList,
						      (POINTER(*)())term_Copy),
				 list_List(Proved));
	if (list_Length(FreeList) > 1)
	    Proved = cnf_QuantMakeOneVar(Proved);
    }
 
    term_Delete(term_FirstArgument(Term1)); /* Variables of "exists" */
    list_Delete(term_ArgumentList(Term1));
    term_RplacTop(Term1,term_TopSymbol(termR));
    term_RplacArgumentList(Term1,term_ArgumentList(termR));
    term_Free(termR);
    
    term_RplacArgumentList(Term, list_Cons(Proved, term_ArgumentList(Term)));
    return Proved;
}

static void cnf_Skolemize(TERM Term, LIST FreeList) {
    /********************************************************
      INPUT: A existence quantified term and the list of free variables
      EFFECT: The existence quantifier is removed by skolemization
      *******************************************************************/
    LIST exlist;
    TERM subterm;
    LIST symblist;
    symblist = list_Nil();
    exlist = cnf_GetSymbolList(term_ArgumentList(term_FirstArgument(Term)));
    term_Delete(term_FirstArgument(Term));
    subterm = term_SecondArgument(Term);
    list_Delete(term_ArgumentList(Term));
    term_RplacTop(Term,term_TopSymbol(subterm));
    term_RplacArgumentList(Term,term_ArgumentList(subterm));
    term_Free(subterm);
    Term = cnf_SkolemFunctionFormula(Term, FreeList, exlist, &symblist);
    list_Delete(exlist);
    list_Delete(symblist);
}

static void cnf_OptimizedSkolemFormula(PROOFSEARCH Search, 
				       TERM topterm, TERM term, 
				       LIST NonConClauses, BOOL result1)
/**************************************************************
  INPUT:   Two terms in negation normal form.
  RETURNS: The  skolemized term with the optimized skolemization
           due to Ohlbach and Weidenbach of <term> and further improvments.
	   <rest> is used as additional, conjunctively added information.
  CAUTION: The term in destrucively changed.
***************************************************************/
{
    TERM   termL2,provedterm;
    LIST   FreeVariables,scan,varlist;
    SYMBOL Top;
    BOOL   result2;
    BOOL   optimized;

    result2 = FALSE;
    FreeVariables = list_Nil();

    Top = term_TopSymbol(term);
    if (fol_IsQuantifier(Top)) {
	if (symbol_Equal(Top,fol_All())) {
	    cnf_PopAllQuantifier(term);
	    cnf_OptimizedSkolemFormula(Search,topterm, term, 
				       NonConClauses, result1);
	    return;
	}
	FreeVariables = fol_FreeVariables(term);
	optimized = FALSE;
	if (symbol_Equal(term_TopSymbol(term_SecondArgument(term)),
			 fol_And())) {
	    scan = term_ArgumentList(term_SecondArgument(term));
	    varlist = term_ArgumentList(term_FirstArgument(term));
	    while((!list_Empty(scan)) && (!optimized)){
		if (result1 == FALSE) {
		    termL2 = cnf_QuantifyAndNegate((TERM) list_Car(scan), 
						      varlist, FreeVariables);
		    result2 = cnf_HaveProofOptSkolem(Search, 
							NonConClauses,
							termL2);
		}
			
		if ((result1 == TRUE) || (result2 == TRUE)) {
		    optimized   = TRUE;
		    if(flag_GetFlagValue(flag_POPTSKOLEM)) {
		      printf("In term ");
		      term_Print(topterm);
		      printf("\n subterm ");
		      term_Print((TERM) list_Car(scan));
		      printf(" is moved to toplevel.\n");
		    }
		    provedterm 
			= cnf_MoveProvedTermToTopLevel(topterm, term, 
							  (TERM) list_Car(scan), 
							  varlist,
							  FreeVariables);
		    if(flag_GetFlagValue(flag_POPTSKOLEM)) {
		      printf("Result : ");
		      term_Print(topterm);
		      printf("\n");
		    }

		    cnf_OptimizedSkolemFormula(Search, topterm, 
					       provedterm,
					       NonConClauses, 
					       result1);
		}
		else 
		  scan = list_Cdr(scan);
	    }
	}
	if(!optimized)
	    cnf_Skolemize(term, FreeVariables);
	list_Delete(FreeVariables);
	cnf_OptimizedSkolemFormula(Search, topterm, 
				   term,
				   NonConClauses, 
				   result1);
	return;
    } 
    else {
	if ((symbol_Equal(Top,fol_And())) || (symbol_Equal(Top,fol_Or()))) { 
	    for (scan=term_ArgumentList(term);!list_Empty(scan);
		 scan=list_Cdr(scan))
		cnf_OptimizedSkolemFormula(Search, topterm, 
					   (TERM) list_Car(scan),
					   NonConClauses, 
					   result1);
	}
    }
    return;
}


static TERM cnf_SkolemFunctionFormula(TERM term, LIST allist, LIST exlist,
				  LIST* symblist)
/**************************************************************
  INPUT:   A term  term and a  list allist of variables and a list exlist
           of variable symbols.
  RETURNS: None.
  CAUTION: The term term is destrucively changed. All variable symbols
           in exlist which appear in term are replaced by skolem functions
	   with respect to allist which contains the universally quantified
	   variables.
***************************************************************/
{
  TERM   term1;
  LIST   scan,scan1;
  SYMBOL skolem;
  int    bottom,n;
  
  bottom = vec_ActMax();
  n = list_Length(allist);
  
  for (scan1=exlist;!list_Empty(scan1);scan1=list_Cdr(scan1)){
    vec_Push(term);
    skolem    = symbol_CreateSkolemFunction(n);
    *symblist = list_Cons((POINTER)skolem,*symblist);
    
    while (!(bottom == vec_ActMax())){
      term1 = (TERM)vec_PopResult();
      if (symbol_Equal(term_TopSymbol(term1),(SYMBOL)list_Car(scan1))){
	term_RplacTop(term1,skolem);
	list_Delete(term_ArgumentList(term1));
	term_RplacArgumentList(term1,list_CopyWithElement(allist,(POINTER (*)())term_Copy));
      }
      if (!list_Empty(term_ArgumentList(term1)))
	for (scan=term_ArgumentList(term1);!list_Empty(scan);scan=list_Cdr(scan))
	  vec_Push(list_Car(scan));
    }
  }
  vec_SetMax(bottom);
  return term;
}


LIST cnf_OptimizedSkolemization(PROOFSEARCH Search, TERM Term, LIST NonConClauses, BOOL result, BOOL Conjecture) {
/**************************************************************
  INPUT: A term, a shared index and a list of non-ConClauses.
  EFFECT: The term is skolemized using optomized skolemization wrt ShIndex.
  OUTPUT: The list of clauses derived from Term
**************************************************************/
    LIST Clauses;
    TERM FirstArg;
#ifdef DEBUG
    if (Term == NULL) {
	printf("Term is NULL in cnf_OptimizedSkolemization\n");
	misc_DumpCore();
    }
#endif
    FirstArg = Term;
    if (!fol_IsLiteral(Term)) {
      if (flag_GetFlagValue(flag_CNFOPTSKOLEM)) {
	Term = term_Create(fol_And(), list_List(Term)); /* CW hack */
	cnf_OptimizedSkolemFormula(Search, Term, FirstArg,  NonConClauses, result);
      }
      else {
	LIST Symbols;
	Symbols = list_Nil();
	Term    = cnf_SkolemFormula(Term,&Symbols);
	list_Delete(Symbols);
      }
    }
    Term    = cnf_DistributivFormula(Term);
    Clauses = cnf_MakeClauseList(Term,FALSE,Conjecture);
    term_Delete(Term);
#ifdef DEBUG
    /*    checkClauseListsConsistency(ShIndex, NonConClauses); */
#endif     
    return Clauses;
}


LIST cnf_GetSkolemFunctions(TERM Term, LIST ArgList, LIST* SkolToExVar) {
/**************************************************************
  INPUT: A term, the argumentlist of a skolem function, a mapping from
         a skolem function to a variable
  EFFECT: Computes information for renaming variables and replacing 
          skolem functions during de-skolemization.
  OUTPUT: The longest argumentlist of all skolem functions found so far.
**************************************************************/ 
    LIST Scan;
    SYMBOL Top;
    
    Top = term_TopSymbol(Term);

    if (fol_IsQuantifier(Top)) 
	return cnf_GetSkolemFunctions(term_SecondArgument(Term), ArgList, 
				      SkolToExVar);

    if (symbol_IsFunction(Top) && symbol_HasProperty(Top, SKOLEM)) {
      BOOL found;
      SYMBOL Var = 0;
      int Arity;
      found = FALSE;
      
      /* Keep longest argument list of all skolem functions in the clause for renaming */
      /* Delete all other argument lists */
      Arity = list_Length(term_ArgumentList(Term));
      if (Arity > list_Length(ArgList)) {
	list_DeleteWithElement(ArgList, term_Delete);
	ArgList = term_ArgumentList(Term);
      }
      else 
	list_DeleteWithElement(term_ArgumentList(Term), term_Delete);
      term_RplacArgumentList(Term, NULL);
      
      /* Replace skolem function by variable */
      if (list_Length(*SkolToExVar) > Arity) {
	NAT i;
	LIST SkolScan = *SkolToExVar;
	for(i = 0; i < Arity; i++)
	  SkolScan = list_Cdr(SkolScan);
	for(SkolScan = (LIST) list_Car(SkolScan); 
	    (SkolScan != list_Nil()) && (found == FALSE); 
	    SkolScan = list_Cdr(SkolScan)) {
	  SYMBOL Skol;
	  Skol = (SYMBOL) list_PairFirst((LIST) list_Car(SkolScan));
	  if (Skol == term_TopSymbol(Term)) {
	    Var = (SYMBOL) list_PairSecond((LIST) list_Car(SkolScan));
	    found = TRUE;
	  }
	}
      }
      if (found == FALSE) {
	LIST Pair;
	NAT i;
	LIST SkolScan;

	SkolScan = *SkolToExVar;
	for(i=0; i < Arity; i++) {
	  if (list_Cdr(SkolScan) == list_Nil())
	    list_Rplacd(SkolScan, list_List(NULL));
	  SkolScan = list_Cdr(SkolScan);
	}

	Var = symbol_CreateStandardVariable();
	Pair = list_PairCreate((POINTER) term_TopSymbol(Term), 
			       (POINTER) Var);
	if (list_Car(SkolScan) == list_Nil())
	  list_Rplaca(SkolScan, list_List(Pair));
	else
	  list_Rplaca(SkolScan, list_Nconc((LIST) list_Car(SkolScan), 
					   list_List(Pair)));
      }
      term_RplacTop(Term, Var);
    }
    else {
      for(Scan = term_ArgumentList(Term); Scan != list_Nil(); 
	  Scan = list_Cdr(Scan))
	ArgList = cnf_GetSkolemFunctions((TERM) list_Car(Scan), ArgList, 
					 SkolToExVar);
    }
    return ArgList;
}

void cnf_ReplaceVariable(TERM Term, SYMBOL Old, SYMBOL New) {
/**************************************************************
  INPUT: A term, two symbols that are variables
  EFFECT: In term every occurence of Old is replaced by New
**************************************************************/ 
  LIST Scan;

#if DEBUG
  if (!symbol_IsVariable(Old)) {
      printf("Illegal symbol\n");
      misc_DumpCore();
  }
#endif

  if (symbol_Equal(term_TopSymbol(Term), Old))
    term_RplacTop(Term, New);
  else 
      for (Scan = term_ArgumentList(Term); !list_Empty(Scan); 
	   Scan = list_Cdr(Scan))
	  cnf_ReplaceVariable(list_Car(Scan), Old, New);
}

LIST cnf_RemoveSkolemFunctions(CLAUSE Clause, LIST* SkolToExVar, LIST Vars) {
/**************************************************************
  INPUT: A clause, a list which is a mapping from skolem functions to variables
         and a list of variables for forall-quantification.
  EFFECT: Arguments of skolem functions are renamed consistently.
          Skolemfunctions are replaced by variables.
  OUTPUT: A list of terms derived from the clause using deskolemization
**************************************************************/
    LIST Scan;
    LIST TermScan, TermList, ArgList;
    TERM Term;
    int i;

    TermList = list_Nil();

    ArgList = list_Nil();
    for (i = 0; i < clause_Length(Clause); i++) {
	Term = term_Copy(clause_GetLiteralTerm(Clause, i));
	ArgList = cnf_GetSkolemFunctions(Term, ArgList, SkolToExVar);
	TermList = list_Cons(Term, TermList);
    }
    
    if (list_Length(ArgList) == 0)
	return TermList;
    
    /* Rename variables */
    for(Scan = ArgList; Scan != list_Nil(); Scan = list_Cdr(Scan)) {
	for (TermScan = TermList; TermScan != list_Nil(); 
	     TermScan = list_Cdr(TermScan)) {
	    Term = (TERM) list_Car(TermScan);
	    cnf_ReplaceVariable(Term,
				term_TopSymbol((TERM) list_Car(Scan)),
				(SYMBOL) list_Car(Vars));
	}
	if (list_Cdr(Vars) == list_Nil()) {
	    SYMBOL New = symbol_CreateStandardVariable();
	    Vars = list_Nconc(Vars, list_List((POINTER) New));
	}
	Vars = list_Cdr(Vars);
    }
    list_DeleteWithElement(ArgList, term_Delete);
    return TermList;
}


TERM   cnf_DeSkolemFormula(LIST Clauses) {
/**************************************************************
  INPUT: A list of clauses.
  EFFECT: All skolem functions are removed from the clauses.
  OUTPUT: A term built from the clauses
**************************************************************/
    LIST Scan;
    LIST SkolToExVar;
    LIST Vars;
    LIST FreeVars;
    LIST FreeVarsCopy;
    TERM VarListTerm;
    LIST VarScan;
    LIST TermList;
    TERM TopTerm;
    TERM Term;
    NAT i;
    BOOL First;

    SkolToExVar = list_List(NULL);
    Vars = list_List((POINTER) symbol_CreateStandardVariable());

    TopTerm = term_Create(fol_And(), NULL);

    for(Scan = Clauses; Scan != list_Nil(); Scan = list_Cdr(Scan)) {
	TermList =  cnf_RemoveSkolemFunctions((CLAUSE) list_Car(Scan),
					      &SkolToExVar, Vars);
	Term = term_Create(fol_Or(), TermList);
	FreeVars = fol_FreeVariables(Term);
	if (!list_Empty(FreeVars)) {
	  FreeVarsCopy = list_CopyWithElement(FreeVars, (POINTER (*) ()) term_Copy);
	  list_Delete(FreeVars);
	  Term = fol_CreateQuantifier(fol_All(), FreeVarsCopy,
				      list_List(Term));
	}
	term_RplacArgumentList(TopTerm, list_Cons(Term, 
						  term_ArgumentList(TopTerm)));
    }
    
    i = 0;
    VarScan = Vars;
    First = TRUE;

    for (Scan = SkolToExVar; Scan != list_Nil(); Scan = list_Cdr(Scan)) {
	if (list_Empty(list_Car(Scan))) {
	  if (term_TopSymbol(TopTerm) == fol_All())
	    term_RplacArgumentList(TopTerm, list_Cons(term_Create((SYMBOL) list_Car(VarScan), NULL),  
						      term_ArgumentList(TopTerm)));
	    if (First == FALSE)
	      TopTerm = fol_CreateQuantifier(fol_All(),
					     list_List(term_Create((SYMBOL) list_Car(VarScan), NULL)),
					     list_List(TopTerm)); 
	}
	else {
	    LIST ExVarScan;
	    LIST ExVars = list_Nil();
	    for(ExVarScan = list_Car(Scan); ExVarScan != list_Nil();
		ExVarScan = list_Cdr(ExVarScan)) {
		if (ExVars == list_Nil())
		    ExVars = list_List(term_Create((SYMBOL) list_PairSecond((LIST) list_Car(ExVarScan)), NULL));
		else
		    ExVars = list_Cons(term_Create((SYMBOL) list_PairSecond((LIST) list_Car(ExVarScan)), NULL), ExVars);
		list_PairFree((LIST) list_Car(ExVarScan));
	    }
	    list_Delete((LIST) list_Car(Scan));
	    list_Rplaca(Scan, NULL);

	    if (term_TopSymbol(TopTerm) == fol_Exist()) {
	      VarListTerm = (TERM) list_Car(term_ArgumentList(TopTerm));
	      term_RplacArgumentList(VarListTerm,
				     list_Nconc(term_ArgumentList(VarListTerm),
				     ExVars));
	    }
	    else
	      TopTerm = fol_CreateQuantifier(fol_Exist(), ExVars, list_List(TopTerm));
	    ExVars = list_Nil();

	    if (First == FALSE) 
		TopTerm = fol_CreateQuantifier(fol_All(), 
					       list_List(term_Create((SYMBOL) list_Car(VarScan), NULL)),
					       list_List(TopTerm));
	}
	if (First == FALSE) 
	    VarScan = list_Cdr(VarScan);
	else
	    First = FALSE;
	
    }
    list_Delete(SkolToExVar);
    list_Delete(Vars);

    return TopTerm;
}

#ifdef OPTCHECK
LIST checkOptimizedSkolemization(LIST* AxClauses, LIST* ConClauses,
				 TERM AxTerm, TERM ConTerm,
				 LIST NonConClauses, LIST* SkolemPredicates,
				 SHARED_INDEX ShIndex,
				 BOOL result) {
/**********************************************************
  EFFECT: Used to check the correctness of optimized skolemization
***********************************************************/

  TERM DeSkolemizedAxOpt, DeSkolemizedConOpt, DeSkolemizedAx, DeSkolemizedCon;
  TERM TopOpt, Top, ToProve;
  LIST SkolemFunctions2;

  if (*AxClauses != list_Nil()) {
    DeSkolemizedAxOpt = cnf_DeSkolemFormula(*AxClauses);
    if (*ConClauses != list_Nil()) {
      DeSkolemizedConOpt = cnf_DeSkolemFormula(*ConClauses);
      TopOpt = term_Create(fol_And(), 
			   list_Cons(DeSkolemizedAxOpt, 
				     list_List(DeSkolemizedConOpt)));
    }
    else 
      TopOpt = DeSkolemizedAxOpt;
  }
  else {
    DeSkolemizedConOpt = cnf_DeSkolemFormula(*ConClauses);     
    TopOpt = DeSkolemizedConOpt;
  }
  
  list_DeleteWithElement(*AxClauses,clause_Delete);
  list_DeleteWithElement(*ConClauses,clause_Delete); 
  *AxClauses = list_Nil();
  *ConClauses = list_Nil();
  
  flag_SetFlagValue(flag_CNFOptSkolem, FALSE);  
  if (AxTerm) {
    *AxClauses = cnf_OptimizedSkolemization(term_Copy(AxTerm), ShIndex, NonConClauses, result,FALSE);
  }
  if (ConTerm) {
    *ConClauses = cnf_OptimizedSkolemization(term_Copy(ConTerm), ShIndex, NonConClauses, result,TRUE);
  }
  
  if (*AxClauses != list_Nil()) {
    DeSkolemizedAx = cnf_DeSkolemFormula(*AxClauses);
    if (*ConClauses != list_Nil()) {
      DeSkolemizedCon = cnf_DeSkolemFormula(*ConClauses);
      Top = term_Create(fol_And(), 
			list_Cons(DeSkolemizedAx, 
				  list_List(DeSkolemizedCon)));
    }
    else 
      Top = DeSkolemizedAx;
  }
  else {
    DeSkolemizedCon = cnf_DeSkolemFormula(*ConClauses);    
    Top = DeSkolemizedCon;
  }
  
  list_DeleteWithElement(*AxClauses,clause_Delete);
  list_DeleteWithElement(*ConClauses,clause_Delete); 
  *AxClauses = list_Nil();
  *ConClauses = list_Nil();
  
  ToProve = term_Create(fol_Equiv(), list_Cons(TopOpt, list_List(Top)));
  ToProve = term_Create(fol_Not(), list_List(ToProve));
  fol_NormalizeVars(ToProve); 
  ToProve  = cnf_ObviousSimplifications(ToProve);
  ToProve  = ren_Rename(ToProve,SkolemPredicates);
  ToProve  = cnf_RemoveEquivImplFromFormula(ToProve);
  ToProve  = cnf_NegationNormalFormula(ToProve);
  ToProve  = cnf_AntiPrenex(ToProve);
  
  SkolemFunctions2 = list_Nil();
  ToProve = cnf_SkolemFormula(ToProve, &SkolemFunctions2);
  ToProve    = cnf_DistributivFormula(ToProve);
  *ConClauses = cnf_MakeClauseList(ToProve);
  if (ToProve)
    term_Delete(ToProve);
  *AxClauses = list_Nil();
  return SkolemFunctions2;
}
#endif

void cnf_Flotter(LIST AxiomList, LIST ConjectureList, LIST* AxClauses, LIST* ConClauses)
{
  LIST         Scan, NonConClauses, FormulaClauses,Usables,SkolemFunctions,SkolemPredicates, EmptyClauses;
  TERM         AxTerm,ConTerm,Formula;
  BOOL         Result;
  PROOFSEARCH Search;

  Search = prfs_Create(FALSE); /* No DocProof in Flotter */

  AxTerm           = ConTerm = (TERM)NULL;
  SkolemPredicates = Usables = list_Nil();
  Result           = FALSE;

  symbol_ReinitGenericNameCounters();

  if (!list_Empty(AxiomList)) {
    for (Scan=AxiomList;!list_Empty(Scan);Scan=list_Cdr(Scan))
      fol_NormalizeVars(list_Car(Scan));   
    if (list_Empty(list_Cdr(AxiomList))) {
      AxTerm    = (TERM)list_Car(AxiomList);
      list_Delete(AxiomList);
      AxiomList = list_Nil();
    }
    else
      AxTerm   = term_Create(fol_And(),AxiomList);
    AxTerm   = cnf_ObviousSimplifications(AxTerm);
    /*printf("\nAx:\t");term_Print(AxTerm);printf("\n");*/
    AxTerm   = ren_Rename(AxTerm,&SkolemPredicates);
    AxTerm   = cnf_RemoveEquivImplFromFormula(AxTerm);
    AxTerm   = cnf_NegationNormalFormula(AxTerm);
    /*printf("\nAn:\t");term_Print(AxTerm);printf("\n");*/
    AxTerm   = cnf_AntiPrenex(AxTerm);
  }
  
  if (!list_Empty(ConjectureList)) {
    if (list_Empty(list_Cdr(ConjectureList)))
      ConTerm  = term_Create(fol_Not(),ConjectureList);
    else
      ConTerm  = term_Create(fol_Not(),list_List(term_Create(fol_And(),ConjectureList)));
    fol_NormalizeVars(ConTerm); 
    ConTerm  = cnf_ObviousSimplifications(ConTerm);
    /*printf("\nCon:\t");term_Print(ConTerm);printf("\n");*/
    ConTerm   = ren_Rename(ConTerm,&SkolemPredicates);
    /*printf("\nRen:\t");term_Print(ConTerm);printf("\n");*/
    ConTerm  = cnf_RemoveEquivImplFromFormula(ConTerm);
    ConTerm  = cnf_NegationNormalFormula(ConTerm);
    /*printf("\nAn:\t");term_Print(ConTerm);printf("\n");*/
    ConTerm  = cnf_AntiPrenex(ConTerm);
    /*printf("\nPr:\t");term_Print(ConTerm);printf("\n");*/
  }

  SkolemFunctions = list_Nil();
  Formula = (TERM)NULL;
  FormulaClauses  = list_Nil();

  if (AxTerm)
    if (ConTerm)
      Formula = term_Create(fol_And(), list_Cons(term_Copy(AxTerm),list_List(term_Copy(ConTerm))));
    else
      Formula = term_Copy(AxTerm);
  else
    Formula = term_Copy(ConTerm);
  
  Formula         = cnf_SkolemFormula(Formula, &SkolemFunctions);
  Formula         = cnf_DistributivFormula(Formula);
  FormulaClauses  = cnf_MakeClauseList(Formula,FALSE,FALSE);
  term_Delete(Formula);
  
  /* printf("\nFormula : "); clause_ListPrint(FormulaClauses); */

  /* Trage nun Formula Clauses modulo Reduktion in einen Index ein */
  NonConClauses = list_Nil();
  /*    for(Scan = FormulaClauses; Scan != list_Nil(); Scan = list_Cdr(Scan))
	skol_Reduce(list_Car(Scan),ShIndex, &Result, NULL, &NonConClauses, &Usables);*/

  /* Number of derivables in red_SatUnit */
  /*flag_SetFlagValue(flag_SATUNIT, 1000);*/

  /* red_SatUnit works only on conclauses */
  for(Scan = FormulaClauses; !list_Empty(Scan); Scan = list_Cdr(Scan))
    clause_SetFlag((CLAUSE) list_Car(Scan), CONCLAUSE);
  /* For FormulaClauses a full saturation */
  /* List is deleted in red_SatUnit ! */
  EmptyClauses = red_SatUnit(Search, FormulaClauses, FALSE);
  if (!list_Empty(EmptyClauses)) {
    Result = TRUE;
    /*printf("\nPROOF in FormulaClauses\n");*/
    list_DeleteWithElement(EmptyClauses, clause_Delete);
  }

  /* Move all usables to workedoff */
  FormulaClauses = list_Copy(prfs_UsableClauses(Search));
  for(Scan = FormulaClauses; !list_Empty(Scan); Scan = list_Cdr(Scan))
    prfs_MoveUsableWorkedOff(Search, (CLAUSE) list_Car(Scan));
  list_Delete(FormulaClauses);
  FormulaClauses = list_Nil();


#ifdef DEBUG
  /*checkClauseListsConsistency(ShIndex, NonConClauses); */
#endif     
  
#ifdef OPTCHECK
  flag_SetFlagValue(flag_CNFOptSkolem, TRUE);
#endif

  if (AxTerm) {
    *AxClauses = cnf_OptimizedSkolemization(Search, term_Copy(AxTerm), NonConClauses, Result, FALSE);
  }
  if (ConTerm) {
    *ConClauses = cnf_OptimizedSkolemization(Search, term_Copy(ConTerm), NonConClauses, Result, TRUE);
  }

#ifdef OPTCHECK
  /*  SkolemFunctions2 = checkOptimizedSkolemization(AxClauses, ConClauses,
						 AxTerm, ConTerm, 
						 NonConClauses, 
						 &SkolemPredicates,
						 ShIndex, Result); */
#endif


  if (AxTerm)
    term_Delete(AxTerm);
  if (ConTerm)
    term_Delete(ConTerm);
  list_Delete(NonConClauses);
  /*  clause_DeleteAllIndexedClauses(ShIndex); */
  /*  sharing_IndexDelete(ShIndex); */
  list_DeleteWithElement(SkolemFunctions,symbol_Delete); 
  symbol_ResetSkolemIndex();

  list_Delete(SkolemPredicates);
#ifdef OPTCHECK
  list_DeleteWithElement(SkolemFunctions2, symbol_Delete);
#endif  
  prfs_Delete(Search);
}


#ifdef DEBUG
void checkClauseListsConsistency(SHARED_INDEX ShIndex, LIST NonConClauses) {
  /**************************************************************
  INPUT: A shared index and a list of non-ConClauses.
  EFFECT: When this function is called all clauses in the index must be
       non-ConClauses, which must also be members of the list.
  **************************************************************/
  LIST AllClauses, scan;

  AllClauses = clause_AllIndexedClauses(ShIndex);
  for(scan = AllClauses; scan != list_Nil(); scan = list_Cdr(scan)) {
    if (!list_PointerMember(NonConClauses, (CLAUSE) list_Car(scan))) {
      printf("Indexed clause \n");
      clause_Print((CLAUSE) list_Car(scan));
      printf("\nnot in NonConClauses.\n");
      misc_DumpCore();
    }
    if (clause_GetFlag((CLAUSE) list_Car(scan), CONCLAUSE)) {
      printf("Clause \n");
      clause_Print((CLAUSE) list_Car(scan));
      printf("\nis a CONCLAUSE.\n");
      misc_DumpCore();
    } 
    if (clause_GetFlag((CLAUSE) list_Car(scan), BLOCKED)) {
      printf("Clause \n");
      clause_Print((CLAUSE) list_Car(scan));
      printf("\nis BLOCKED.\n");
      misc_DumpCore();
    }
  }
  if (list_Length(AllClauses) != list_Length(NonConClauses)) {
    printf("There are clauses in NonConClauses that are not in the Index\n");
    misc_DumpCore();
  }  
  list_Delete(AllClauses);
}
#endif

LIST cnf_SatUnit(PROOFSEARCH Search, LIST ClauseList) 
/*********************************************************
  INPUT:   A list of unshared clauses, proof search object
  RETURNS: A possibly empty list of empty clauses.
**********************************************************/
{
  CLAUSE Given;
  LIST   Scan, Derivables, EmptyClauses, BackReduced;
  NAT    n,Derived;

  Derived      = flag_GetFlagValue(flag_CNFPROOFSTEPS);
  EmptyClauses = list_Nil();
  ClauseList   = clause_ListSortWeighed(ClauseList);
  
  while (!list_Empty(ClauseList) && list_Empty(EmptyClauses)) {
    Given = (CLAUSE)list_NCar(&ClauseList);
    Given = red_CompleteReductionOnDerivedClause(Search, Given, FALSE);
    if (Given)
      if (clause_IsEmptyClause(Given))
        EmptyClauses = list_List(Given);
      else {
        /*printf("\n\nGiven: ");clause_Print(Given);*/
	BackReduced = red_BackReduction(Search, Given, FALSE, red_USABLE);
	
        if (Derived != 0) {
	  Derivables = inf_BoundedDepthUnitResolution(Given, prfs_UsableSharingIndex(Search));
	  Derivables = list_Nconc(Derivables, 
			      inf_BoundedDepthUnitResolution(Given, prfs_WorkedOffSharingIndex(Search)));
          n          = list_Length(Derivables);
          if (n > Derived)
            Derived = 0;
          else
            Derived = Derived - n;
        }
        else 
          Derivables = list_Nil();
	
        Derivables  = list_Nconc(BackReduced,Derivables);
	/*printf("\nDerived: ");clause_ListPrint(Derivables); */
        Derivables  = split_ExtractEmptyClauses(Derivables, &EmptyClauses);
	
	prfs_InsertUsableClause(Search, Given);
	
	for(Scan = Derivables; !list_Empty(Scan); Scan = list_Cdr(Scan))
	  ClauseList = clause_InsertWeighed((CLAUSE) list_Car(Scan), ClauseList);
	list_Delete(Derivables);
      }
  }
  clause_DeleteClauseList(ClauseList);
  return EmptyClauses;
}

