/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                                                        * */
/* *                                                        * */
/* *  $Module:   CLAUSECONVERTER                            * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.59 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/06/20 09:12:34 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: cnf2dfg.c,v $ */

/*** MAINLOOP *************************************************/

#include "clauseinput.h"
#include "clause.h"


void main(int argc, char* argv[])
{
  LIST  Derivables,Scan;
  FILE  *Output;
  char  *OldString,*Filename;

  memory_Init(-1);
  symbol_Init(TRUE);
  stack_Init();
  flag_Init();
  term_Init();
  fol_Init(TRUE);
  in_Init();

  if (argc<2) {
    printf("\n\t          cnf2dfg Version 0.46");
    printf("\n\t       Usage: cnf2dfg <input-file>  <output-file> \n\n");
    return;
  }
  
  OldString = symbol_Name(fol_EQUALITY);  /* Hack because SPASS used "=" for DFG "equal" */
  symbol_SetName(fol_EQUALITY,"=");

  Derivables = in_CreateClauseSet(*++argv);

  symbol_SetName(fol_EQUALITY,OldString);

  Filename = *++argv;
  Output   = misc_OpenFile(Filename,"w");
  fprintf(Output, "begin_problem(Unknown).\n\n");
  fprintf(Output, "list_of_descriptions.\n");
  fprintf(Output, "name({*Unknown*}).\n");
  fprintf(Output, "author({*cnf2dfg*}).\n");
  fprintf(Output, "status(unknown).\n");
  fprintf(Output, "description({*File generated by cnf2dfg*}).\n");
  fprintf(Output, "end_of_list.\n\n");
  fprintf(Output, "list_of_symbols.\n");
  fol_PrintDFGSignature(Output);
  fprintf(Output, "end_of_list.\n\n");
  fprintf(Output, "list_of_clauses(axioms, cnf).\n");
  fprintf(Output, "end_of_list.\n\n");
  fprintf(Output, "list_of_clauses(conjectures, cnf).\n");
  
  for (Scan=Derivables;!list_Empty(Scan);Scan=list_Cdr(Scan))
    clause_FPrintDFG(Output,list_Car(Scan),FALSE);
 
  fprintf(Output, "end_of_list.\n\n");
  fprintf(Output, "end_problem.\n\n");
  misc_CloseFile(Output,Filename);
  printf("\nResult of Transformation is written to file: %s\n\n",*argv);

  list_DeleteWithElement(Derivables,clause_Delete);

  in_Free();
  fol_Free();
  symbol_FreeAllSymbols();
#ifdef DEBUG
  memory_Print();
#endif
}

