% I-THOP  (A first-order theorem prover based on PTTP technology)
% 
%      version 1.01
%      July 2, 1997
%      Copyright (C) 1997, Koji Iwanuma (iwanuma@esi.yamanashi.ac.jp)
% 
% This program is free software; you can redistribute it or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either versions 1, or (at your option)
% any later version.
%  
% This program is distributed in the hope that it will be useful but
% WITHOUT ANY WARRANTY.  See the GNU General Public License for more
% details.
% ---------------
%
%   comCOM.pl
%
% ---------------
%  This file contains a main part of the compilier for transformation 
%  from  clausal theories to prolog-codes


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%      Theorem compiler !!
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%% for SICStus ver 3%%%%%%%%%%%%%%%
:- use_module(library(system)),
	use_module(library(terms)).
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pttp(X) :-
	prolog_flag(unknown,Current,error),
	initialize_data_in_compiling,
	%-------------------------
	statistics(runtime,[X1,_]),
	retractall(start_time(preprocessing,_)),
	assert(start_time(preprocessing,X1)),
	%
	preprocessing(X),
	%-------------------------
	statistics(runtime,[X2,_]),
	%
	sorting,
	%-------------------------
	statistics(runtime,[X3,_]),
	%
        compiling_to_Prolog,
	%
	abolish_data_in_compiling,
	%-------------------------
        statistics(runtime,[X4,_]),
	%
	call(axiom_exec_file(EXEC)),
	compile(EXEC),
	%
	%-------------------------
	delete_tmp_files,
	%
        statistics(runtime,[X5,_]),
        %
	nl,
	Time12 is (X2 - X1) / 1000.0,
	write('Preprocessing = '),write(Time12),write(' sec'),nl,
	%
	Time23 is (X3 - X2) / 1000.0,
	write('Sorting = '),write(Time23),write(' sec'),nl,
	%
	Time34 is (X4 - X3) / 1000.0,
	write('Compiling_to_prolog = '),write(Time34),write(' sec'),nl,
	%
	Time45 is (X5 - X4) / 1000.0,
	write('Compiling_to_WAMcode = '),write(Time45),write(' sec'),nl,
	%
	Time15 is (X5 - X1) / 1000.0,
	write('Total time of translation = '),write(Time15),write(' sec'),nl,nl,
	%
	prolog_flag(unknown,_,Current),
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
initialize_data_in_compiling :- 
	abolish_data_in_compiling,
	set_initial_data_in_compiling.

abolish_data_in_compiling :-
        retractall(contrapositive(_,_,_,_)),
        retractall(pseudo_horn_clause(_,_,_,_)),
        retractall(sorted_pseudo_horn_clause(_,_,_,_)),
        retractall(predicate_name(_,_,_,_,_,_)),
        retractall(id_number(_,_)),
        retractall(axiom_att(_,_)),
        retractall(axiom_att_in_compiling(_,_)).

set_initial_data_in_compiling :-
        assert(id_number(unit,0)),
        assert(id_number(non_unit,0)),
        assert(id_number(goal,0)),
        assert(id_number(predicate,0)),
        assert(id_number(clause,1)),
        assert(axiom_att(max_term_depth,0)),
        assert(axiom_att(horn,horn)),
        assert(axiom_att_in_compiling(dumy,dumy)).

delete_tmp_files :-
	delete_temporal_files_mode(on) ->
	      prolog_flag(unknown,Current,fail),
	      (call(axiom_temp_data_file(DATA)),
	               delete_file(DATA,[ignore]) ;
	      %true ->
	               true
	      ),
	      (call(axiom_exec_file(EXEC))  ->
	               delete_file(EXEC,[ignore]) ;
	      %true ->
                       true 
	      ),
              prolog_flag(unknown,_,Current) ;
	delete_temporal_files_mode(off) ->
             true ;
	%true ->
             nl,write('Error 3 in deleting temporal files !!'),nl.

	
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

compiling_to_Prolog :-
	write('Compiling into Prolog program ...'),nl,

%%%%%%%%%  system depending predicate ??? %%%%%%%%%%%%%%%%%%
	prolog_flag(unknown,Current,fail),
	( call(axiom_exec_file(PRG1)) -> 
	     true 
	;
	    mktemp(theEXEXXXXXX,PRG1),      % for SICStus3
	    assert(axiom_exec_file(PRG1))
	),
	prolog_flag(unknown,_,Current),
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	tell(PRG1),
	write_flag,write_value,
	generate_int_clause,
	generate_ext_clause,
	generate_goal,
	told.

%%%%%%%%%%%%%%%%%%%%%%%%%%%
ancestor_matching(on) :-
	(ancestor_unit_lemma_matching_mode(on);ancestor_c_literal_matching_mode(on)),
	!.
ancestor_matching(off) :-
	\+ ancestor_matching(on).

lemmaizing(on) :-
	(unit_lemmaizing(on) ; non_unit_lemmaizing(on)),
	!.
lemmaizing(off) :-
	\+ lemmaizing(on).

unit_lemmaizing(on) :-
	(unit_lemma_matching_mode(on); ancestor_unit_lemma_matching_mode(on);
	    unit_lemma_resolution_mode(on); lemma_repl_subsumption_mode(on)),
	!.
unit_lemmaizing(off) :-
	\+ unit_lemmaizing(on).

non_unit_lemmaizing(on) :-
	(identical_c_reduction_mode(on); ancestor_c_literal_matching_mode(on);
	    full_c_reduction_mode(on); c_literal_repl_subsumption_mode(on)),
	!.
non_unit_lemmaizing(off) :-
	\+ non_unit_lemmaizing(on).

anti_lemmaizing(SW) :- anti_lemma_subsumption_mode(SW).

storing_lemma_proof(on) :- 
	(write_lemma_proof_mode(on) ; inconsistency_check_mode(on)),
	!.
storing_lemma_proof(off) :- 
	\+ storing_lemma_proof(on).

check_term_depth_bound(BB) :- 
	prolog_flag(unknown,Current,fail),
	(call(term_depth_bound_data(T0)) ->
	    BB = T0 ;
	call(axiom_att_in_compiling(max_term_depth,T1)) ->
	    BB is T1 + 1 ;    % This offset is a quite pragmatic value.
	%true ->
	    nl,write('Error in check_term_depth_bound !!!!'),nl
	),
	prolog_flag(unknown,_,Current).

%%%%%%%%%%%%%%%%%%%%%%%%%%%

write_flag :- 
	compute_and_write_runtime_flag,
	nl,   
	immediately_write_runtime_setting("_mode","_flag",[
           single_body_lemmaizing,
	   %
	   ancestor_unit_lemma_matching, ancestor_unit_lemma_matching_in_reduction,
	   unit_lemma_resolution, ancestor_c_literal_matching,
	   %
	   inconsistency_check, unit_repl_subsumption, 
	   lemma_repl_subsumption, c_literal_repl_subsumption,
	   last_goal_repl_subsumption,
	   %
	   cut_by_no_inst,
	   lemma_backward_subsumption, anti_lemma_backward_subsumption]),
	   %
	   nl.

write_value :- 
	immediately_write_runtime_setting("_data","_runtime_data",[
	   depth_bound_measure, head_pruning_position
	]),nl,
	check_term_depth_bound(BB),write(term_depth_bound_runtime_data(BB)),write('.'),nl,
        (axiom_att(horn,horn) ->
	       write('horn_or_non_horn_data(horn).'),nl ;
	 axiom_att(horn,non_horn_existed) ->
	       write('horn_or_non_horn_data(non_horn).'),nl ;
	 %true ->
	       nl,write('Error in axiom_att(horn,X) !!'),nl
	).

compute_and_write_runtime_flag :-
	%
	lemmaizing(SW0),write(lemmaizing_flag(SW0)),write('.'),nl,
	unit_lemmaizing(SW00),write(unit_lemmaizing_flag(SW00)),write('.'),nl,
	non_unit_lemmaizing(SW001),write(non_unit_lemmaizing_flag(SW001)),write('.'),nl,
	storing_lemma_proof(SW002),write(store_lemma_proof_flag(SW002)),write('.'),nl, 
	call(anti_lemma_subsumption_mode(SW1)),write(anti_lemmaizing_flag(SW1)),write('.'),nl,
	nl,
	%
        call(unit_subsumption_mode(SW2)),
	write(head_unit_subsump_flag(SW2)),write('.'),nl,
	call(anti_lemma_subsumption_mode(SW3)),
	write(head_anti_lemma_subsump_flag(SW3)),write('.'),nl,
	call(lemma_subsumption_mode(SW4)),
	write(head_lemma_subsump_flag(SW4)),write('.'),nl,
	call(ident_A_chk_mode(SW10)),
	write(head_ident_A_chk_flag(SW10)),write('.'),nl,
	call(compl_A_chk_mode(SW11)),
	write(head_compl_A_chk_flag(SW11)),write('.'),nl,
	%
	(tail_pruning_mode(on) ->
	    call(unit_subsumption_mode(SW5)),
	    write(tail_unit_subsump_flag(SW5)),write('.'),nl,
	    call(anti_lemma_subsumption_mode(SW6)),
	    write(tail_anti_lemma_subsump_flag(SW6)),write('.'),nl,
	    call(lemma_subsumption_mode(SW7)),
	    write(tail_lemma_subsump_flag(SW7)),write('.'),nl,
	    call(ident_A_chk_mode(SW8)),
	    write(tail_ident_A_chk_flag(SW8)),write('.'),nl,
	    call(compl_A_chk_mode(SW9)),
	    write(tail_compl_A_chk_flag(SW9)),write('.'),nl ;
	% tail_pruning_mode(off) ->
	    write(tail_unit_subsump_flag(off)),write('.'),nl,
	    write(tail_anti_lemma_subsump_flag(off)),write('.'),nl,
	    write(tail_lemma_subsump_flag(off)),write('.'),nl,
	    write(tail_ident_A_chk_flag(off)),write('.'),nl,
	    write(tail_compl_A_chk_flag(off)),write('.'),nl
	).

immediately_write_runtime_setting(_,_,[]).
immediately_write_runtime_setting(UsrSuf,RunSuf,[X|Y]) :-
	name(X,XStr),append(XStr,UsrSuf,UsrStr),name(UsrName,UsrStr),
	UsrPre =.. [UsrName,Data],
	name(X,XStr),append(XStr,RunSuf,RunStr),name(RunName,RunStr),
	RunPre =.. [RunName,Data],
	call(UsrPre),
	!,
	write(RunPre),write('.'),nl,
	immediately_write_runtime_setting(UsrSuf,RunSuf,Y).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%    generate_int_clause
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_int_clause :-
	call(id_number(unit,Num)),
	(Num > 0 ->
	     write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),nl,
             write('%%%  '),write(Num),write(' UNIT clauses exist.'),nl,nl,
	     generate_int_unit_clasuse(Num),
	     fail ;
        %true ->
	     write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),nl,
	     write('%%%  NO UNIT clauses exist'),nl,nl,
             fail
	).

generate_int_clause :-
	call(id_number(non_unit,Num)),
	(Num > 0 ->
	     write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),nl,
             write('%%%  '),write(Num),write(' Non-unit clauses exist.'),nl,nl,
	     generate_int_non_unit_clasuse(Num),
	     fail ;
        %true ->
	     write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),nl,
	     write('%%%  NO Non-unit clauses exist.'),nl,nl,
             fail
	).

generate_int_clause :-
	(retract(sorted_pseudo_horn_clause(unit,_,_,_)),fail ; true),
	(retract(sorted_pseudo_horn_clause(non_unit,_,_,_)),fail ; true).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%    generate_int_unit_clause      % for int clauses of unit clauses
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_int_unit_clasuse(0) :- !.

generate_int_unit_clasuse(Num) :-
	for(1,Num,CurNum),
	call(sorted_pseudo_horn_clause(unit,Unit,CurNum,[ClauseID])),
	gene_int_unit(Unit,CurNum,ClauseID), 
	fail.

generate_int_unit_clasuse(_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%    gene_int_unit      % for int clauses of unit clauses
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gene_int_unit(Unit,Id,ClauseID) :- 
	%
	check_and_collect_dis_equations(Unit,Id,EqList,SubFlag),
	(SubFlag = '@subsumed'(SubId) ->
	     write('%%% the Unit '),write(Unit),write(' [id: '),
	     write(Id),write('] is subsumed by id: '),write(SubId),write(' Unit !!'),nl ;
        %true ->
	     Unit =.. [PredName|Args],
	     name(PredName,PredStr),
	     (append("not_",_,PredStr) -> Pole = '@neg' ;  Pole = '@posi'),
     	     %--------------------------------------
	     %
	     append("unit_",PredStr,SubUnitStr),name(SubUnitName,SubUnitStr),
	     (Pole == '@neg' ->
	           NegClauseID is - ClauseID,
	           append(Args,[NegClauseID],SubUnitArgs) ;
	      Pole == '@posi' ->
	           append(Args,[ClauseID],SubUnitArgs)
	     ),
	     SubsumpUnitHead =.. [SubUnitName|SubUnitArgs],
	     write(SubsumpUnitHead),write('.'),nl,
	     %--------------------------------------
	     %
	     linearize(Unit,Head,[],_,true,true,SoundUnify,NoOcurUnify),
	     Head =..[_|LinearArgs],
	     %
	     append("int_unit_",PredStr,IntUnitStr),name(IntUnitName,IntUnitStr),

	     (Pole == '@neg' ->
	           NegClauseID is - ClauseID,
	           append(LinearArgs,['_SF',NegClauseID],IntUnitArgs) ;
	      Pole == '@posi' ->
	           append(LinearArgs,['_SF',ClauseID],IntUnitArgs)
	     ),

	     IntUnitHead =.. [IntUnitName|IntUnitArgs],
	     write(IntUnitHead),write(' :-  '),nl,
	     %--------------------------------------
	     %
	     ( SoundUnify == true -> 
        	    write('      inc_carray(42)') ;  
              %true ->
		    write('      (_SF==''@linear'' -> '),
		               write(NoOcurUnify),write(',inc_carray(42);'),nl,
		    write('                         '),
		               write(SoundUnify),write(',inc_carray(15))')
	     ),
	     (subsumption_in_units_mode(on) ->
		  write_dis_equations(EqList) ;
	      subsumption_in_units_mode(off) ->	     
		  true ;
	      %true ->
        	  nl,write('*** Error in subsumption_in_units_mode !! ***'),nl,nl
	     ),
	     write('.'),nl,nl
	),
	!.

%%%%%%%%%%%%%%%%%%%

check_and_collect_dis_equations(Unit,Id,EqList,SubFlag) :-
	functor(Unit,F,A),
	chk_and_collect_dis_eqs(1,Id,Unit,F,A,EqList,SubFlag).

chk_and_collect_dis_eqs(Num,IdNum,_,_,_,[],'@not_subsumed') :-
	Num == IdNum,!.
 
chk_and_collect_dis_eqs(Num,IdNum,Unit,F,A,EqList,SubFlag) :-
	functor(TmpUnit,F,A),
	call(sorted_pseudo_horn_clause(unit,TmpUnit,Num,_)),
	!,
	generate_dis_equations_for_subsump(TmpUnit,Unit,Eqs),
	(Eqs == false ->
	      NewNum is Num +1,
	      chk_and_collect_dis_eqs(NewNum,IdNum,Unit,F,A,EqList,SubFlag) ;
	 Eqs == true ->
	      EqList=[], SubFlag = '@subsumed'(Num);
	 %true ->
	      EqList = [Eqs|EqCdr],
	      NewNum is Num +1,
	      chk_and_collect_dis_eqs(NewNum,IdNum,Unit,F,A,EqCdr,SubFlag)
	).

chk_and_collect_dis_eqs(Num,IdNum,Unit,F,A,EqList,SubFlag) :-
	NewNum is Num + 1,
	chk_and_collect_dis_eqs(NewNum,IdNum,Unit,F,A,EqList,SubFlag).

%%%%%%%%%%%%%%%%%%%

write_dis_equations(EqList) :-
	EqList == [] ->
	     true ;
	%true ->
	     write(','),nl,
	     write('      '),
	     write_non_nil_dis_equation_list(EqList).

write_non_nil_dis_equation_list([Last]) :-
	!,write(\+ (Last,inc_carray(39))).

write_non_nil_dis_equation_list([X|Y]) :-
	write(\+ (X,inc_carray(39))),write(','),
	write_non_nil_dis_equation_list(Y).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%    gene_int_non_unit      % for int clauses of nonunit clauses
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


generate_int_non_unit_clasuse(0) :- !.
generate_int_non_unit_clasuse(Num) :-
	for(1,Num,CurNum),
	call(sorted_pseudo_horn_clause(non_unit,[Head|Body],CurNum,[ClauseID])),   
	gene_int_non_unit(Head,Body,ClauseID),
	fail.
generate_int_non_unit_clasuse(_).

gene_int_non_unit(Head,Body,ClauseID) :-  
	length(Body,NumGoals),
	gene_int_cl_head_operation(Head,NumGoals,Pole,ClauseID),
	gene_int_cl_body_lists(Body,NumGoals,Pole).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%    gene_int_cl_head_operation      % for head operation
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gene_int_cl_head_operation(P,NumGoals,Pole,ClauseID) :-
	%
	check_unit_subsumption(P,InitFlag),
	linearize(P,Head,[],_,true,true,SoundUni,NoOcurUni),
	%---------------------------------
	Head =..[HeadName|Args],name(HeadName,HeadStr),
	(append("not_",_,HeadStr) ->  Pole = '@neg' ;  Pole = '@posi'),
	%---------------------------------
	functor(P,HeadName,Arity),
	call(predicate_name(HeadName,Arity,_,UFlag,_,_)),
	(UFlag == '@unit_exist' ->
	        (NumGoals == 1 ->
		     Opt = 'UF,''@single_goal''' ; 
		 %true ->
		     Opt = 'UF,''@multi_goals'''
		);
	 %true ->
	        (NumGoals == 1 ->
		     Opt = '''@single_goal''' ; 
		 %true ->
		     Opt = '''@multi_goals'''
		)
	),
	%
	append("int_",HeadStr,IntStr),name(IntName,IntStr),
        (Pole=='@neg' ->
	        append(Args,['CS0,CA0,CAn,FLp-TrF,FLn,RL1,IF1,_SF',Opt,'Sol0,Soln'],IntArgs) ; 
	 Pole=='@posi' ->
	        append(Args,['CS0,CA0,CAn,FLp,FLn-TrF,RL1,IF1,_SF',Opt,'Sol0,Soln'],IntArgs) 
	),
	IntHead =.. [IntName|IntArgs],
	%---------------------------------
	%
	write(IntHead), write(' :- '),nl,
	%---------------------------------

	write('    check_depth_bound(CS0,CS1,CA0,CA1,'),write(NumGoals),write('),'),nl,
	(SoundUni == true -> 
	     write('    inc_carray(43),'),nl  ;  
         %true ->
	     write('    (_SF==''@linear''-> '),write(NoOcurUni),write(',inc_carray(43); '),nl,
	     write('                      '),write(SoundUni),write(',inc_carray(8)),'),nl
	),
	%---------------------------------
	write('    '),
	(UFlag == '@unit_exist' ->
	     (InitFlag == '''@unit_none''' ->
	           write('UF=''@unit_none'', ');
	      InitFlag == '''@unit_exist''' ->
	           write('set_U_flag(UF), ');
	      %true ->
	        nl,write('Some Errors in printing in UnitFlag in int-clauses'),nl,nl
	     ) ;
	 %true ->
	    true
	),
	%---------------------------------
	( Pole == '@posi' ->
	       write('dappend(Sol0,('),write(ClauseID),write(',''@non_unit'',FLn),Sol1),'),nl ;
	  Pole == '@neg' ->
	       NegClauseID is - ClauseID,
	       write('dappend(Sol0,('),write(NegClauseID),write(',''@non_unit'',FLp),Sol1),'),nl ;
	 %true ->
	     nl,write('Error in Pole of predicate !!'),nl,nl
	),
	!.    

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

check_unit_subsumption(Head,UnitFlag) :-
	prolog_flag(unknown,Current,fail),
	functor(Head,Name,Arity),functor(Template,Name,Arity),
	( call(sorted_pseudo_horn_clause(unit,Template,_,_)), \+ \+ unify(Template,Head) ->
	        UnitFlag = '''@unit_exist''';
	 %true ->
	        UnitFlag = '''@unit_none'''
	),
	prolog_flag(unknown,_,Current).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%     gene_int_cl_body_lists
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gene_int_cl_body_lists(Body,NumGoals,Pole) :-
	NumGoals == 1 ->
	      gene_single_int_body(Body,Pole) ;
        NumGoals >= 2 ->
	      gene_int_body(Body,1,Pole).


gene_single_int_body([Single],Pole) :-
	!,
	Single =..[ExtName|RealArgs],
	(Pole == '@neg' -> LastNum = -10000 ; LastNum = 10000),
	%
        append(RealArgs,['CS1,CA1,CAn,FLp,FLn,[]-RL1,[]-IF1',LastNum,'TrF,Sol1,Soln'],ExtArgs),
        GoalAtom =.. [ExtName|ExtArgs],
	%
	write('    '),write(GoalAtom),write('.'),nl,
	nl,nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  for last goal and tail_operation.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gene_int_body([Last],NumGoal,Pole) :- 
	!,
	make_var(NumGoal,"CA",Cn_1), make_var(NumGoal,"Sol",Sol_1),
	%
	(NumGoal==1 ->
	     RLold = [], IFold =[] ;
	%true ->
	     make_var(NumGoal,"RL",RLold),make_var(NumGoal,"IF",IFold)
	),
	%-----------------------------------
	Last =..[ExtName|RealArgs],
	(Pole == '@neg' -> LastNum = -1 ; LastNum = 1),
	%
        append(RealArgs,['CS1',Cn_1,'CAn,FLp,FLn',RLold-'RL1',IFold-'IF1',LastNum,'TrF',Sol_1,'Soln'],ExtArgs),
        GoalAtom =.. [ExtName|ExtArgs],
	%-----------------------------------
	%
	write('    '),write(GoalAtom),write('.'),nl,
	nl,nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  for intermediate goals
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gene_int_body([Q|R],NumGoal,Pole) :-
	%
	NewNum is NumGoal +1,
	make_var(NumGoal,"CA",C_1),make_var(NewNum,"CA",C_2),
	make_var(NumGoal,"Sol",Sol_1),make_var(NewNum,"Sol",Sol_2),
	%
	(NumGoal==1 ->
	     RLold = [], IFold = [] ;
	%true ->
	     make_var(NumGoal,"RL",RLold),make_var(NumGoal,"IF",IFold)
	),
	make_var(NewNum,"RL",RLnew),	make_var(NewNum,"IF",IFnew),
	%--------------------------------------
	Q=..[ExtName|RealArgs],
	(Pole == '@neg' ->  GoalID is -NewNum ; GoalID is NewNum),
	%
	append(RealArgs,['CS1',C_1,C_2,'FLp,FLn',RLold-RLnew,IFold-IFnew,GoalID,'TrF',Sol_1,Sol_2],ExtArgs),
	GoalAtom =.. [ExtName|ExtArgs],
	%--------------------------------------
	%
	write('    '),write(GoalAtom),write(','),nl,
	%--------------------------------------
	gene_int_body(R,NewNum,Pole).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%     generate_ext_clause    
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_ext_clause :-
	call(id_number(predicate,Num)),
	(Num > 0 ->
	     write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),nl,
             write('%%%  '),nl,
             write('%%%  '),write(Num),write(' predicate symbol exist.'),nl,
             write('%%%  '),nl,
	     write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),nl,nl,
	     generate_ext_clause(Num),
	     fail ;
        %true ->
	     nl,write('Error in generate_ext_clause !!!!'),nl,
             fail
	).
generate_ext_clause :-
	(retract(predicate_name(_,_,_,_,_,_)),fail :true).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_ext_clause(0) :-!.

generate_ext_clause(Num) :-
	for(1,Num,CurNum),
	call(predicate_name(Name,Arity,CurNum,UF,CompUF,NUF)),
	gene_ext_clause(Name,Arity,CurNum,UF,CompUF,NUF),
	fail.
generate_ext_clause(_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

gene_ext_clause(Name,Arity,IdNum,UFlag,_,NUFlag) :-
	UFlag=='@unit_none', NUFlag=='@non_unit_none',
	!,
	write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),nl,
	write('%%%  The predicate "'),write(Name),write('" does NOT have ANY clauses !!'),nl,
	write('%%%  The ID number is '),write(IdNum),write('.'),nl,
	write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),nl,nl,
	%
	%   Is the below setting really necessary?
	name(Name,NameST),
	append("lemma_",NameST,LemNameST),name(LemName,LemNameST),
        write('lemma_name_runtime_data('),write((LemName,Arity)),write(').'),nl,nl,
	%
	functor(ExtLite,Name,Arity),ExtLite =.. [_|Args],
	append(Args,['_CS,_CA0,_CAn,_FLp,_FLn,_RL,_IF,_GoNum,_TR_Flag,_Sol0,_Sol1'],ExtArgs),
	ExtHead =.. [Name|ExtArgs],
	%
	write(ExtHead), write(':-'),nl,
	write('       fail.'),nl,
	write('       %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% '),nl,
	write('       % This predicate is forced to immediately fail !!'),nl,
	write('       %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% '),nl,nl.


gene_ext_clause(Name,Arity,IdNum,UFlag,CompUFlag,NUFlag) :-
	name(Name,NameST),
	(append("not_",_,NameST) ->  Pole = '@neg' ;  Pole = '@posi'),

	functor(ExtLite,Name,Arity),add_string_to_literal(ExtLite,ExtLiteR,"not_"),
	ExtLite =.. [_|Args],

	(Pole == '@neg' -> Lite = ExtLiteR ; Lite = ExtLite),

	write('%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%'),nl,
	(UFlag == '@unit_exist' ->
	     write('%%%  The predicate "'),write(Name),
	     write('" has some UNIT clauses !!'),nl ;
	 UFlag == '@unit_none' ->
	     write('%%%  The predicate "'),write(Name),
	     write('" does NOT have any UNIT clauses !!'),nl
        ),	
	(NUFlag == '@non_unit_exist' ->
	     write('%%%  The predicate "'),write(Name),
	     write('" has some NON-UNIT clauses !!'),nl ;
	 NUFlag == '@non_unit_none' ->
	     write('%%%  The predicate "'),write(Name),
             write('" does NOT have any NON-UNIT clauses !!'),nl
        ),
	(CompUFlag == '@comp_unit_exist' ->
	     write('%%%  The complemented one of "'),write(Name),
	     write('" has some UNIT clauses !!'),nl ;
	 CompUFlag == '@comp_unit_none' ->
	     write('%%%  The complemented one of "'),write(Name),
	     write('" does NOT have any UNIT clauses !!'),nl
        ),	
	write('%%%  The ID number is '),write(IdNum),write('.'),nl,nl,

     %------------------------------------------------------------
     % Make the ext_head, the int_head and miscellaneous heads, 
     % and write varisous initial setting data 
     %------------------------------------------------------------

	append("int_",NameST,IntNameST),name(IntName,IntNameST),
	append("int_unit_",NameST,IntUnitNameST),name(IntUnitName,IntUnitNameST),
	%-------------------------------------------
	append(Args,['CS,CA0,CAn,FLp,FLn,RLin-RLout,IFin-IFout,_GoNum,TR_Flag,Sol0,Sol1'],ExtArgs),
	(UFlag=='@unit_exist' ->
	      append(Args,['CS,CA0,CAn,FL1p,FL1n,RL0,IF0,SF,UF,OF,Sol0,Sol1'],IntArgs) ;
	 UFlag=='@unit_none' ->
	      append(Args,['CS,CA0,CAn,FL1p,FL1n,RL0,IF0,SF,OF,Sol0,Sol1'],IntArgs) 
	),
	append(Args,['SF,ClauseId'],IntUnitArgs),
	ExtHead =.. [Name|ExtArgs],
	IntHead =.. [IntName|IntArgs], 
	IntUnitHead =.. [IntUnitName|IntUnitArgs],
	%-------------------------------------------
	(UFlag=='@unit_exist' ->
	     append("unit_",NameST,UnitNameST),name(UnitName,UnitNameST),
	     append(Args,['_UnitID'],UnitArgs),
	     UnitLite =.. [UnitName|UnitArgs] ;
	 UFlag=='@unit_none' ->
	     true ;
	 %true ->
	     nl,write('Error in UFlag in gene_ext_clause !'),nl,nl
	),
	%-------------------------------------------
	(CompUFlag=='@comp_unit_exist' ->
	     add_string_to_name(Name,CompName,"not_"),
	     add_string_to_name(CompName,CUnitName,"unit_"),
	     append(Args,['CUnitID'],CUnitArgs),
	     CompUnitLite =.. [CUnitName|CUnitArgs] ;
	 CompUFlag=='@comp_unit_none' ->
	     true ;
	 %true ->
	     nl,write('Error in CompUFlag in gene_ext_clause !'),nl,nl
	),
	%-------------------------------------------
        (unit_lemmaizing(on) ->
             append("lemma_",NameST,LemNameST),name(LemName,LemNameST),
	     add_string_to_name(Name,CompName,"not_"),
	     add_string_to_name(CompName,LemCompName,"lemma_"),
	     (unit_lemma_resolution_mode(on) ->
		       append(Args,['_GroFlag,_LemID'],LemArgs),
		       append(Args,['_GroFlag,_CLemID'],LemCompArgs) ;
	      unit_lemma_resolution_mode(off) ->
		       append(Args,['_LemID'],LemArgs),
		       append(Args,['_CLemID'],LemCompArgs) ;
	     %true ->
		  nl,write('Error in unit_lemma_resolution_mode !!'),nl,nl
	     ),
	     LemLite=.. [LemName|LemArgs],
	     LemCompLite=.. [LemCompName|LemCompArgs],
	     %-----------------------------
	     %
	     write('lemma_name_runtime_data('),write((LemName,Arity)),write(').'),nl ;
	     %-----------------------------	    
	 unit_lemmaizing(off) ->
	     true ;
	%true ->
	     nl,write('**** Anti_lemma_mode illegal !!  ****'),nl,nl
	),

	%-------------------------------------------
        (anti_lemmaizing(on) ->
	     append("anti_",NameST,AntiNameST),name(AntiName,AntiNameST),
	     append(['AntiNum'],Args,AntiArgs),
	     AntiLite =.. [AntiName|AntiArgs],
	     write('anti_lemma_name_runtime_data('),write((AntiName,Arity)),write(').'),nl,nl ;
	 anti_lemmaizing(off) ->	
	     true ;
	%true ->
	     nl,write('**** Anti_lemma_mode illegal !!  ****'),nl,nl
	),
%%%
%%%  Write the ext head and its content, e.g, various head pruning, unit and/or lemma matching,
%%%  expansion, reduction, and tail pruning routines. 
%%%
	write(ExtHead), write(':-'),nl,

%%%%%%%%  Pruning operations %%%%%%%%%%%%%%%

	write('  h_direct_pruning(TR_Flag,FLp,FLn,_GoNum),'),nl,

%%%%%%%%  The first initial setting -- 1 %%%%%%%%%%%%%%%

	write('  Lite = '),write(Lite),write(','),nl,
	%-------------------------------------------
	write('  (truncate(TR_Flag,LId) ->'),nl,
	(Pole == '@neg' ->
	      write('      inc_carray(13),dappend(Sol0,(trun_n(LId),Lite,FLp),Sol1)') ;
	 Pole == '@posi' ->
	      write('      inc_carray(13),dappend(Sol0,(trun(LId),Lite,FLp),Sol1)')
	),
	write(',CA0=CAn'),nl,
	write('  ;'), nl,
	%-------------------------------------------
	( Pole == '@neg' ->
            write('      nonidentical_member(Lite,FLp),') ;  
	  Pole == '@posi' ->
	    write('      nonidentical_member(Lite,FLn),')
	),
        nl,
	%-------------------------------------------
	(unit_lemmaizing(on) ->
	    write('      LemL= '),write(LemLite),write(','),nl,
	    write('      CLemL= '),write(LemCompLite),write(','),nl,
	    LemFL='LemL-_LemID',
	    CompLemFL='CLemL-_CLemID';
	 unit_lemmaizing(off) ->
	    LemFL='''$$LemL''-''$$LemID''',
	    CompLemFL='''$$CLemL''-''$$CLemID'''
	),
	%-------------------------------------------
	(UFlag=='@unit_exist' ->
	    write('      UnitL= '),write(UnitLite),write(','),nl,
	    UnitFL='UnitL',
	    UFFL='UF' ;
	 UFlag=='@unit_none' ->
	    UnitFL='''$$UnitL''',
	    UFFL='''@unit_none'''
	),
	%-------------------------------------------
	(CompUFlag=='@comp_unit_exist' ->
 	    write('      CUnitL= '),write(CompUnitLite),write(','),nl,
	    CompUnitFL='CUnitL',
	    CompUnitID='CUnitID' ;
	 CompUFlag=='@comp_unit_none' ->
	    CompUnitFL='''$$CUnitL''' ,
	    CompUnitID='''$$ID'''
	),

%%%%%%%% The first initial setting -- 2 %%%%%%%%%%%%

	(unit_lemma_resolution_mode(on) ->
	     write('      call_lemma_number('),write(IdNum),write(',LNum),'),nl ;
	 unit_lemma_resolution_mode(off) ->
 	     true
	),

	write('      ( ('),nl,

%%%%%%%% unit and/or lemma matching %%%%%%%%%%%%

	(UFlag=='@unit_exist' ->
	   (unit_matching_mode(on) ->
	       (unit_lemma_matching_mode(on) ->
	           write('          unit_lemma_match('),write(LemFL),
	           write(',UnitL-_UnitID,Lite,LU_flag,LU_lite),'),nl,
	           write('               !,RL=[],IF=[],dappend(Sol0,(LU_flag,LU_lite,FLp),Sol1)'),
	           write(',CA0=CAn ;'),nl ;
	        unit_lemma_matching_mode(off) ->
	           write('          unit_match(UnitL-_UnitID,Lite,ID),'),nl,
	           write('               !,RL=[],IF=[],dappend(Sol0,(u_match(ID),UnitL,FLp),Sol1)'),
	           write(',CA0=CAn ;'),nl 
	       ) ;
	    unit_matching_mode(off) ->
	       (unit_lemma_matching_mode(on) ->
	           write('          lemma_match('),write(LemFL),write(',Lite,ID),'),nl,
		   write('               !,RL=[],IF=[],dappend(Sol0,(l_match(ID),LemL,FLp),Sol1)'),
	           write(',CA0=CAn ;'),nl ;
	        unit_lemma_matching_mode(off) ->
	           true
	       ) 
	   ) ;
	UFlag=='@unit_none' ->
	   (unit_lemma_matching_mode(on) ->
	       write('          lemma_match('),write(LemFL),write(',Lite,ID),'),nl,
	       write('                   !,RL=[],IF=[],dappend(Sol0,(l_match(ID),LemL,FLp),Sol1)'), 
	       write(',CA0=CAn ;'),nl ;
	    unit_lemma_matching_mode(off) ->
	       true
	   )
	),

%%%%%%%% identical reduction %%%%%%%%%%%%
	
	(axiom_att_in_compiling(horn,horn) ->
	       write('          fail') ;
	 axiom_att_in_compiling(horn,non_horn_existed) ->
              (Pole == '@neg' ->
               write('          identical_reduction(Lite,FLn,Id),!,'),
		  write('RL=[Id],IF=[(Id,''@no_inst'')],dappend(Sol0,(id_red_n,Lite,FLp),Sol1)');  
	       Pole == '@posi' ->
               write('          identical_reduction(Lite,FLp,Id),!,'),
		  write('RL=[Id],IF=[(Id,''@no_inst'')],dappend(Sol0,(id_red,Lite,FLp),Sol1)')
	      ),
	      write(',CA0=CAn') ;
	 %true ->
	      nl,write('Error in axiom_att_in_compiling(horn,XXX) !!!!'),nl
	),

%%%%%%%% identical reduction with C-literals %%%%%%%%%%%%

	(identical_c_reduction_mode(on) ->
          (Pole == '@neg' ->
	   write(' ;'),nl,
	   write('          ident_c_reduction(Lite,FLn,RL,IF),!,'),
	      write('dappend(Sol0,(id_c_red_n,Lite,FLp),Sol1)');  
	   Pole == '@posi' ->
	   write(' ;'),nl,
	   write('          ident_c_reduction(Lite,FLp,RL,IF),!,'),
	      write('dappend(Sol0,(id_c_red,Lite,FLp),Sol1)')
	   ),
	   write(',CA0=CAn') ;
	 identical_c_reduction_mode(off) ->	
	   true
	),

%%%%%%%%%%%%%%%%%%%%

	nl,
	write('        )'),nl,
	write('      ;'),nl,

%%%%%%%%  SU_flag_setting %%%%%%%%%%%%

	(UFlag=='@unit_exist' ->
	   (unit_matching_mode(on) ->
	       (unit_lemma_matching_mode(on) ->
	           write('        set_SU_flag_LemUni(SF,UF),'),nl ;
	        unit_lemma_matching_mode(off) ->
	           write('        set_SU_flag_Uni(SF,UF),'),nl
	       ) ;
	    unit_matching_mode(off) ->
	       (unit_lemma_matching_mode(on) ->
	           write('        set_S_flag_Lem(SF),'),nl ;
	        unit_lemma_matching_mode(off) ->
	           write('        SF=''@not_examined'','),nl
	       ) 
	   ) ;
	UFlag=='@unit_none' ->
	   (unit_lemma_matching_mode(on) ->
	       write('        set_S_flag_Lem(SF),'),nl ;
	    unit_lemma_matching_mode(off) ->
	       write('        SF=''@not_examined'','),nl
	   )
	),

%%%%%%%  The second initial setting %%%%%%%%%%%%%%%%%%%%%

        (anti_lemmaizing(on) ->
	     write('        inc_anti,call_anti_number(AntiNum),'),nl,
	     write('        AntiL= '),write(AntiLite),write(','),nl,
	     AntiFL='AntiL' ;
	 anti_lemmaizing(off) ->	
	     AntiFL='Lite'
	),

        (tail_pruning_mode(on),cut_by_no_inst_mode(on) ->
	     write('        copy_term('),write(AntiFL),write(',_TmpL),'),nl ;
	 %true ->
	     true
	),
	write('        set_id_dcounter(CS,CA0,C),'),nl,

	(Pole == '@neg' ->
	     write('        FL1p = [(C,Lite,_CLp,TrF,'),
	           write((UFFL-UnitFL,LemFL,AntiFL)),write(')|FLp]-TrF,'),nl,
	     write('        FL1n = [(C,''$$dumL'',_CLn,TrF,'),
	           write((CompUnitFL-CompUnitID,CompLemFL,AntiFL)),write(')|FLn],'),nl,
	    CurFList='FL1n',
	    OppFList='FL1p' ; 
 	 Pole == '@posi' ->
	     write('        FL1n = [(C,Lite,_CLn,TrF,'),
	           write((UFFL-UnitFL,LemFL,AntiFL)),write(')|FLn]-TrF,'),nl,
	     write('        FL1p = [(C,''$$dumL'',_CLp,TrF,'),
	           write((CompUnitFL-CompUnitID,CompLemFL,AntiFL)),write(')|FLp],'),nl,
	    CurFList='FL1p',
	    OppFList='FL1n' ; 
	%true ->
	    nl,write('Error in Pole !!'),nl,nl
	),

	write('        ( ( '),nl,

%%%%%%% Expansion with untis %%%%%%%%%%%%%%%%%%%%%

	(UFlag == '@unit_exist' ->
	      write('            var(UF),'),nl,
	      write('            '),write(IntUnitHead),write(','),nl,
	      write('            RL=[],IF=[],dappend(Sol0,(ClauseId,Lite,FLp),Sol1)'),
	      write(',CA0=CAn'),
	     (tail_pruning_mode(off) ->
	            nl ;
	      pruning_after_unit_expan_mode(on),tail_pruning_mode(on) ->   
	            write(','),nl,
	            write('            t_pruning_after_unit_expansion('),
	            write(OppFList),write(')'),
                    nl ;
              %true ->
	            nl 
	     ),
	     write('           ;'),nl ;
	 UFlag == '@unit_none' ->
	     true
         ),

%%%%%%% Reductions  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	(axiom_att_in_compiling(horn,horn) ->
	     true  ;
	 axiom_att_in_compiling(horn,non_horn_existed) ->
	     (Pole == '@neg' ->
   	 	 write('            reduction(Lite,FLn,Id,GrF,RedF),'),nl,
		 write('            RL=[Id],IF=[(Id,GrF)],dappend(Sol0,(red_n,Lite,FLp),Sol1)') ;
   	      Pole == '@posi' ->
   	 	 write('            reduction(Lite,FLp,Id,GrF,RedF),'),nl,
		 write('            RL=[Id],IF=[(Id,GrF)],dappend(Sol0,(red,Lite,FLp),Sol1)');
             %true ->
   	 	 nl,write('Error in Pole !!'),nl,nl
	     ),
	     write(',CA0=CAn'),
	     (tail_pruning_mode(off) ->
	         nl ;
	      tail_pruning_mode(on) ->   
	         write(','),nl,
	         write('            t_pruning_after_reduction('),
		             write((OppFList,CurFList)),write(',Id,GrF,RedF)'),nl
	     ),
	     write('           ;'),nl ;
	%true ->
	     nl,write('Error in axiom_att_in_compiling(horn,YYY) !!!!'),nl
	),

%%%%%%% Expansion with non-untis  %%%%%%%%%%%%%%%%%%%%%
% 3/25 19:25
	(NUFlag == '@non_unit_exist' -> 
	    write('            '),write(IntHead),write(','),nl,
  
  	    write('            t_pruning_and_lemmaizing('),
            write((OppFList,CurFList)),write(',Sol0,RL0-RL,IF0-IF,_TmpL,_Cut2,OF,_GoNum)'),nl ;
         NUFlag == '@non_unit_none' -> 
            write('           fail'),nl
       ),

%%%%%%% Full C-Reductions %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	(full_c_reduction_mode(on) ->
	   write('           ;'),nl,
           (Pole == '@neg' ->
   		write('            c_reduction(Lite,FLn,RL,IF,GrF,RedF),'),nl,
		write('            dappend(Sol0,(c_red_n,Lite,FLp),Sol1)') ;
   	    Pole == '@posi' ->
   		write('            c_reduction(Lite,FLp,RL,IF,GrF,RedF),'),nl,
		write('            dappend(Sol0,(c_red,Lite,FLp),Sol1)');
           %true ->
   		nl,write('Error in Pole !!'),nl,nl
	   ),
	   write(',CA0=CAn'),

	   (tail_pruning_mode(off) ->
		nl ;
	    tail_pruning_mode(on) ->   
		write(','),nl,
		write('            t_pruning_after_C_reduction('),
	            write((OppFList,CurFList)),write(',RL,IF,GrF,RedF,_GoNum)'),nl
	   ) ;
	full_c_reduction_mode(off) ->
	  true 
	),

%%%%%%% Lemma resolutions  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

	(unit_lemma_resolution_mode(off) ->
	      true ;
	 unit_lemma_resolution_mode(on) ->
	     write('           ;'),nl,
	     (Pole == '@neg' ->
	        write('            unit_lemma_resolution('),write(LemFL),write(',LNum,SF,OF1,OF2),'),nl,
		write('            RL=[],IF=[],dappend(Sol0,(l_res_n(_LemID),Lite,FLp),Sol1)');
	      Pole == '@posi' ->
                write('            unit_lemma_resolution('),write(LemFL),write(',LNum,SF,OF1,OF2),'),nl,
		write('            RL=[],IF=[],dappend(Sol0,(l_res(_LemID),Lite,FLp),Sol1)');
	      %true ->
		nl,write('Error in Pole !!'),nl,nl
	     ),
	     write(',CA0=CAn'),

	     (tail_pruning_mode(off) ->
	          nl ;
	      tail_pruning_mode(on) ->
	          write(','),nl,
		  write('            t_pruning_after_lem_resol('),
		        write((OppFList,CurFList)),write(',_TmpL,_Cut2,OF1,OF2)'),nl
	     );
	%true ->
              nl,write('**** unit_lemma_resolution_mode illegal !! ****'),nl,nl
	),
	write('          )'),

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        (cut_by_no_inst_mode(on) ->
	     write(','),nl,
             write('          (_Cut2 == ''@cut'',!'),
	     (anti_lemmaizing(on) ->
	          write(',clear_anti_lemma(AntiL),dec_anti ; true)'),nl ;
	      anti_lemmaizing(off) ->
	          write(' ; true)'),nl ;
	      %true ->
	          nl,write('**** Anti_lemma_mode illegal !!  ****'),nl,nl
	     );
	 cut_by_no_inst_mode(off) ->
	     nl ;
        %true ->
	     nl,write('**** cut_by_no_inst_mode illegal !! ****'),nl,nl
	),

	write('        ;'),nl,

        (anti_lemmaizing(on) ->
	     write('          clear_anti_lemma(AntiL),dec_anti,!,fail'),nl ;
	 anti_lemmaizing(off) ->	
	     write('          fail'),nl ;
	%true ->
	     nl,write('**** Anti_lemma_mode illegal !!  ****'),nl,nl
	),

	write('        )'),nl,
	write('      )'),nl,
	write('  ),'),nl,
	write('  FLp = [(AncC,_)|_],'),nl,
	write('  merge_RedList(TR_Flag,RLin,RL,RLout,AncC),'),nl,
	write('  merge_InsFList(TR_Flag,IFin,IF,IFout,AncC).'),nl,nl,
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%      generate_goal
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_goal :-

	retract(pseudo_horn_clause(goal,GOALS,1,[ClauseID])),

	length(GOALS,NN),NN1 is NN+1,make_var(NN1,"Sol",SolL),
	%------------------------------------
	nl,nl,write('start(N,'),write(SolL),write(') :- '),nl,
	write('      reset_flag_atom(12),'),nl,
	write('      FLp=[(''@top'',''$$TopL_Pos'',_Cp,TrF,_Other_Pos)],'),nl,
	write('      FLn=[(''@top'',''$$TopL_Neg'',_Cn,TrF,_Other_Neg)],'),nl,
	%------------------------------------
	(NN == 1 ->
	       gene_single_subgoal(GOALS,ClauseID) ;
	%true ->
	       gene_subgoals(GOALS,1,ClauseID)
	),
	%------------------------------------
	write('      write_result(unsatisfiable).'),nl,nl,
	%------------------------------------
	%
        write('start(N,Sol) :- '),nl,
	write('      (call_flag_atom(12,FF),FF \\==''@Finitely_failed'',!,'),nl,
	write('                write_result(satisfiable) ; '),nl,
	write('       write_intermediate_counter(N), N1 is N+1, start(N1,Sol)).'),nl,nl,
	!.

gene_single_subgoal([GOAL],ClauseID) :-
	InitSol = [(ClauseID,'query','FLp')|'End']-'End',
	%
	GOAL =.. [Name|Args],
	%--------------
	%      Notice that the goal number is 0 !!
	%
	append(Args,['N,N,_C2,FLp,FLn, ''@top''-_RL,''@top''-_IF,0,TrF',InitSol,'Sol2'],ExtArgs),
	%
	ExtGoal =.. [Name|ExtArgs],
	write('      '),write(ExtGoal),write(','),nl,
	write('      nl,nl,'),nl.


gene_subgoals([Last],N,_) :-
	!,
	make_var(N,"_CA",C1),make_var(N,"Sol",Sol1),
	%
	NewN is N+1,
        make_var(NewN,"_CA",C2),make_var(NewN,"Sol",Sol2),
	%------------------------------------
	Last =.. [Name|Args],
	append(Args,['N',C1,C2,'FLp,FLn,''@top''-_RL,''@top''-_IF,1,TrF',Sol1,Sol2],ExtArgs),
	ExtGoal =.. [Name|ExtArgs],
	%------------------------------------
	write('      '),write(ExtGoal),write(','),nl,
	write('      nl,nl,'),nl.

gene_subgoals([Goal|CdrGOAL],N,ClauseID) :-
	%
	make_var(N,"_CA",C1),make_var(N,"Sol",Sol1), 
	%
	NewN is N+1,
        make_var(NewN,"_CA",C2),make_var(NewN,"Sol",Sol2),
	%------------------------------------
	InitSol = [(ClauseID,'query','FLp')|'End']-'End',
	%
	Goal =.. [Name|Args],
	(N==1 ->
  	       append(Args,['N,N',C2,'FLp,FLn,''@top''-_RL,''@top''-_IF',NewN,'TrF',InitSol,Sol2],ExtArgs);  
  	%true ->
  	       append(Args,['N',C1,C2,'FLp,FLn','''@top'''-'_RL','''@top'''-'_IF',NewN,'TrF',Sol1,Sol2],ExtArgs)
	),
	%------------------------------------
	ExtGoal =.. [Name|ExtArgs],
	write('      '),write(ExtGoal),write(','),nl,
	%------------------------------------
	gene_subgoals(CdrGOAL,NewN,_).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%  
%%%      utilities
%%%  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

append([],M,M) :- !.
append([F|R],M,[F|S]) :- append(R,M,S).

%%%%%%%%%%%%%

identical_member(X,[Y|_])  :-		
	X == Y,!.
identical_member(X,[_|L]) :-
	identical_member(X,L).

%%%%%%%%%%%%%

inc_id_number(ID) :-
	retract(id_number(ID,Num)),
	NewNum is Num + 1,
	assert(id_number(ID,NewNum)).

call_id_number(ID,Num) :-
	call(id_number(ID,Num)).

%%%%%%%%%%%%%

for(I, I, I) :- !.
for(I, _J, I).
for(I, J, K) :- I1 is I + 1, for(I1, J, K).

%%%%%%%%%%%%%

make_var(X,Y,Z):-
  name(X,Xst), append(Y,Xst,Zst), name(Z,Zst), !.

%%%%%%%%%%%%%

add_string_to_name(Name,AddedName,String) :-
	name(Name,NString),
	(append(String,AddedNString,NString) ->
	         true ;
	 %true ->
	         append(String,NString,AddedNString)
	),
	name(AddedName,AddedNString).


add_string_to_literal(Lite,AddedLite,String) :-
	Lite =.. [Name|Arg],
	add_string_to_name(Name,AddedName,String),
	AddedLite =.. [AddedName|Arg].

%%%%%%%%%%%%%

clause_to_list((CarL;CdrL),[CarL|List]) :-
	!,
	clause_to_list(CdrL,List).
clause_to_list(Unit,[Unit]).

%%%%%%%%%%%%%

conjunction_to_list((X,Z),[X|V]) :-
	!,
	conjunction_to_list(Z,V).
conjunction_to_list(Y,[Y]).

%%%%%%%%%%%%

list_to_conjunction([],true) :- !.
list_to_conjunction([X],X) :- !.
list_to_conjunction([X|Y],(X,Z)) :-
	list_to_conjunction(Y,Z).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%  
%%%%   A variant of LINEARIZATION developed by pttp,
%%%%   which is for linearing heads of clauses for sound unification
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

linearize(TermIn,TermOut,VarsIn,VarsOut,MIn1,MIn2,MOut1,MOut2) :-
	nonvar(TermIn) ->
		functor(TermIn,F,N),
		myfunctor(TermOut,F,N),
		linearize_args(TermIn,TermOut,VarsIn,VarsOut,
		               MIn1,MIn2,MOut1,MOut2,1,N);
	identical_member(TermIn,VarsIn) ->
		VarsOut = VarsIn,
		conjoin(MIn1,unify(TermIn,TermOut),MOut1),
		conjoin(MIn2,TermIn=TermOut,MOut2);
	%true ->
		TermOut = TermIn,
		VarsOut = [TermIn|VarsIn],
		MOut1 = MIn1,
		MOut2 = MIn2.

linearize_args(TermIn,TermOut,VarsIn,VarsOut,MIn1,MIn2,MOut1,MOut2,I,N) :-
	I > N ->
		VarsOut = VarsIn,
		MOut1 = MIn1,
		MOut2 = MIn2;
	%true ->
		arg(I,TermIn,ArgI),
		linearize(ArgI,NewArgI,VarsIn,VarsT,MIn1,MIn2,MT1,MT2),
		arg(I,TermOut,NewArgI),
		I1 is I + 1,
		linearize_args(TermIn,TermOut,VarsT,VarsOut,
		               MT1,MT2,MOut1,MOut2,I1,N).

myfunctor(Term,F,N) :-
	nonvar(F),
	atomic(F),
	N == 0,
	!,
	Term = F.
myfunctor(Term,F,N) :-
	nonvar(Term),
	atomic(Term),
	!,
	F = Term,
	N = 0.
myfunctor(Term,F,N) :-
	functor(Term,F,N).

conjoin(A,B,C) :-
	A == true ->
		C = B;
	B == true ->
		C = A;
	A == false ->
		C = false;
	B == false ->
		C = false;
	%true ->
		C = (A , B).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%   Generate Subsumption check routies with dis-equations
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_dis_equations_for_subsump(PreUnit,Unit,Equations) :-
	copy_term(Unit,TmpU),
	(unify(PreUnit,TmpU) ->
	     gene_subsump_frame(TmpU,Unit,AssocList,EqOut,SrOut,SdOut),
	     eliminate_local_variables(SrOut,AssocList,SrOut1),
	     var_list(Unit,VList),
	     make_subsump_chk_formula(EqOut,SrOut1,SdOut,VList,Equations) ;
	%true ->
	     Equations = false
	),
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%

gene_subsump_frame(Unified,Unit,AOut,EqOut,SrOut,SdOut) :-
	gene_frame(Unified,Unit,[],AOut,[],_,[],EqOut,[],SrOut,[],SdOut).

%%%%%%%%%%%

gene_frame(Unified,Unit,AIn,AIn,VIn,VIn,EqIn,EqIn,SrIn,SrIn,SdIn,SdIn) :-
	atomic(Unit),
	!,
	(atomic(Unified),Unified==Unit ->
	      true ;
	%true ->
	      nl,write('Error 1 in gene_subsump !!'),nl,nl,
	      fail
	).

gene_frame(Unified,Unit,AIn,AOut,VIn,VOut,EqIn,EqOut,SrIn,SrOut,SdIn,SdOut) :-
	var(Unit),
	!,
	(identical_member(Unit,VIn) ->
	      AOut=AIn, VOut=VIn, EqOut=EqIn, SrOut=SrIn, SdOut=SdIn ;
	%true ->
	      VOut=[Unit|VIn],
	      (var(Unified) ->
	 	   SrOut = SrIn, SdOut = SdIn,
	 	   (get_unified_variable(Unified,AIn,UniVar) ->
		         AOut = AIn, EqOut = [Unit==UniVar|EqIn] ;
	 	   %true ->
		         AOut = [(Unified,Unit)|AIn], EqOut = EqIn
	 	   ) ;
	       ground(Unified) ->
	           AOut = AIn, EqOut = [Unit==Unified|EqIn],
	 	   SrOut = SrIn, SdOut = SdIn ;
	      %true ->
	           AOut = AIn, EqOut=EqIn,
		   SrOut=[Unified|SrIn],SdOut=[Unit|SdIn]
	      )
	).
	
gene_frame(Unified,Unit,AIn,AOut,VIn,VOut,EqIn,EqOut,SrIn,SrOut,SdIn,SdOut) :-
	functor(Unit,F,A),
	(functor(Unified,F,A) ->
	        gene_frame_args(A,Unified,Unit,AIn,AOut,VIn,VOut,
	                                EqIn,EqOut,SrIn,SrOut,SdIn,SdOut) ;
	%true ->
	        nl,write('Error 2 in gene_subsump !!'),nl,nl
	).

%%%%%%%%%%%%%

gene_frame_args(0,_,_,AIn,AIn,VIn,VIn,EqIn,EqIn,SrIn,SrIn,SdIn,SdIn) :- !.

gene_frame_args(A,Unified,Unit,AIn,AOut,VIn,VOut,
	                         EqIn,EqOut,SrIn,SrOut,SdIn,SdOut) :-
	arg(A,Unit,UnitA),arg(A,Unified,UnifiedA),
	gene_frame(UnifiedA,UnitA,AIn,A1,VIn,V1,EqIn,Eq1,SrIn,Sr1,SdIn,Sd1),
	NewA is A-1,
	gene_frame_args(NewA,Unified,Unit,A1,AOut,V1,VOut,
	                 Eq1,EqOut,Sr1,SrOut,Sd1,SdOut).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% get_unified_variable(Unified,AIn,UniVar)

get_unified_variable(_,[],_) :- !,fail.

get_unified_variable(Unified,[(V1,V2)|_],V2) :-
	Unified == V1,!.

get_unified_variable(Unified,[_|R],UniVar) :-
	get_unified_variable(Unified,R,UniVar).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% eliminate_local_variables(SrOut,AssocList,SrOut1),

eliminate_local_variables([],_,[]) :- !.

eliminate_local_variables(SrOut,AssocList,SrOO) :-
	copy_term(SrOut,SrOO),
	search_and_replace(SrOut,SrOO,AssocList).

%%%%%%%%%%%%%%%%%

search_and_replace(Orig,Modified,_) :-
	atomic(Orig),!,
	(Orig == Modified ->
	     true ;
	 %true ->
	     nl,write('Error 1 in serach_and_replace !!'),nl,nl
	).

search_and_replace(Orig,Modified,AList) :-
	var(Orig),!,
	(var(Modified) ->
	      (get_unified_variable(Orig,AList,UniVar) ->
	           Modified = UniVar ;
	      %true ->
	           true
	      ) ;
        %true ->
	     nl,write('Error 2 in serach_and_replace !!'),nl,nl
	).

search_and_replace(Orig,Modified,AList) :-
	functor(Orig,F,A),
	(functor(Modified,F,A) ->
	      search_and_replace_args(A,Orig,Modified,AList) ;
	%true ->
     	      nl,write('Error 3 in serach_and_replace !!'),nl,nl
	).

%%%%%%%%%%

search_and_replace_args(0,_,_,_).
search_and_replace_args(A,Orig,Modified,AList) :-
	arg(A,Orig,Arg1),arg(A,Modified,Arg2),
	search_and_replace(Arg1,Arg2,AList),
	NewA is A -1,
	search_and_replace_args(NewA,Orig,Modified,AList).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% make_subsump_chk_formula(Eq,SrList,SdList,Formula)

make_subsump_chk_formula(Eq,[],SdList,_,Formula) :-
	!,
	(SdList == [] ->
	      list_to_conjunction(Eq,Formula) ;
	%true ->
  	     nl,write('Error 1 in make_subsump_chk_formula !!'),nl,nl
	).

make_subsump_chk_formula(Eq,SrList,SdList,VList,Formula) :-
	check_and_divide(SrList,SdList,VList,NewEq,SrOut,SdOut),
	(SrOut==[] ->
	     (SdOut == [] ->
	          append(Eq,NewEq,Equations),
		  list_to_conjunction(Equations,Formula) ;
              %true ->
  	         nl,write('Error 2 in make_subsump_chk_formula !!'),nl,nl
	     ) ;
	SrOut = [X] ->
	     (SdOut = [Y] ->
	          ChkTerm =.. [subsumes_chk,X,Y],
		  append(Eq, NewEq, EqTT),
		  append(EqTT,[ChkTerm],Equations),
		  list_to_conjunction(Equations,Formula) ;
              %true ->
  	         nl,write('Error 3 in make_subsump_chk_formula !!'),nl,nl
	     ) ;
	%true ->
	     Term1 =.. ['''@sub'''| SrOut],
	     Term2 =.. ['''@sub'''| SdOut],
	     ChkTerm =.. [subsumes_chk,Term1,Term2],
	     append(Eq, NewEq, EqTT),
	     append(EqTT,[ChkTerm],Equations),
	     list_to_conjunction(Equations,Formula)
	).

%%%%%%%%%%%%

check_and_divide([],SdList,_,[],[],[]) :-
	!,
	(SdList==[] ->
	      true ;
	%true ->
	    nl,write('Error 1 in check_and_divide !!'),nl,nl
	).

check_and_divide([Sr1|Sr2],[Sd1|Sd2],VList,NewEq,SrOut,SdOut) :-
	var_list(Sr1,VL1),
	(subset(VL1,VList) ->
	     NewEq = [Sr1==Sd1|EqCdr],
	     check_and_divide(Sr2,Sd2,VList,EqCdr,SrOut,SdOut);
	%true ->
	     SrOut = [Sr1|SrCdr],
	     SdOut = [Sd1|SdCdr],
	     check_and_divide(Sr2,Sd2,VList,NewEq,SrCdr,SdCdr)
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

var_list(Term,VList) :-
	var_list(Term,[],VList).

var_list(Term,VIn,VIn) :-
	atomic(Term),!.

var_list(Term,VIn,VOut) :-
	var(Term), !,
	(identical_member(Term,VIn) ->
	      VOut = VIn ;
	%true ->
	      VOut = [Term|VIn]
	).

var_list(Term,VIn,VOut) :-
	functor(Term,_,A),
	var_list_args(A,Term,VIn,VOut).

var_list_args(0,_,VIn,VIn) :- !.
var_list_args(A,Term,VIn,VOut) :- 
	arg(A,Term,Arg),var_list(Arg,VIn,VTT),
	A1 is A-1, var_list_args(A1,Term,VTT,VOut).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

subset([],_).
subset([X|Y],Z) :-
	identical_member(X,Z),
	subset(Y,Z).


