% I-THOP  (A first-order theorem prover based on PTTP technology)
% 
%      version 1.01
%      July 2, 1997
%      Copyright (C) 1997, Koji Iwanuma (iwanuma@esi.yamanashi.ac.jp)
% 
% This program is free software; you can redistribute it or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either versions 1, or (at your option)
% any later version.
%  
% This program is distributed in the hope that it will be useful but
% WITHOUT ANY WARRANTY.  See the GNU General Public License for more
% details.
% ---------------
%
%   comLIB.pl%
%
% ---------------
%  This file contains all of run-tim library.

:- use_module(library(timeout)).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%
%%%%%%   run-time predicate library
%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% Sound unification algorithm with occurs check that is called
%%% by code resulting from transforming clauses to eliminate
%%% repeated variables in the head.
%%% This should be coded in a lower-level language for efficiency.


unify(X,Y) :- unify_LHS(X,Y,_,_),!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

unify_check_inst_in_LHS(X,Y,GrFlag,RedFlag) :-
	unify_LHS(X,Y,TGrFlag,TRedFlag),!,
	(TGrFlag == '@done_inst' ->
	       GrFlag = '@done_inst' ;
	 var(TGrFlag) ->
	       GrFlag = '@no_inst' ;
	 %true ->
	       nl,write('Error 0 in unify_check_inst_in_LHS !'),nl,nl
	),
	(TRedFlag == '@done_inst' ->
	       RedFlag = '@done_inst' ;
	 var(TRedFlag) ->
	       RedFlag = '@no_inst' ;
	 %true ->
	       nl,write('Error 1 in unify_check_inst_in_LHS !'),nl,nl
	).

%%%%%%%%%%

unify_LHS(X,Y,GrFlag,RedFlag) :-
	var(X) ->
		(var(Y) ->
			X = Y,
		        GrFlag = '@done_inst';
		%true ->
			functor(Y,_,N),
			(N = 0 ->
				true;
			N = 1 ->
				arg(1,Y,Y1), not_occurs_in(X,Y1);
			%true ->
				not_occurs_in_args(X,Y,N)
			),
			X = Y,
			GrFlag = '@done_inst',
			RedFlag = '@done_inst'
		);
	var(Y) ->
		functor(X,_,N),
		(N = 0 ->
			true;
		N = 1 ->
			arg(1,X,X1), not_occurs_in(Y,X1);
		%true ->
			not_occurs_in_args(Y,X,N)
		),
		X = Y,
		(ground(X) ->
		        true ;
		 %true ->
			GrFlag = '@done_inst'
		);
	%true ->
		functor(X,F,N),
		functor(Y,F,N),
		(N = 0 ->
			true;
		N = 1 ->
			arg(1,X,X1), arg(1,Y,Y1), unify_LHS(X1,Y1,GrFlag,RedFlag);
		%true ->
			unify_LHS_args(X,Y,N,GrFlag,RedFlag)).

unify_LHS_args(X,Y,N,GrFlag,RedFlag) :-
	N = 2 ->
		arg(2,X,X2), arg(2,Y,Y2), unify_LHS(X2,Y2,GrFlag,RedFlag),
		arg(1,X,X1), arg(1,Y,Y1), unify_LHS(X1,Y1,GrFlag,RedFlag);
	%true ->
		arg(N,X,Xn), arg(N,Y,Yn), unify_LHS(Xn,Yn,GrFlag,RedFlag),
		N1 is N - 1, unify_LHS_args(X,Y,N1,GrFlag,RedFlag).

%%%%%%%%%%%%%%%%

not_occurs_in(Var,Term) :-
	Var == Term ->
		fail;
	var(Term) ->
		true;
	%true ->
		functor(Term,_,N),
		(N = 0 ->
			true;
		N = 1 ->
			arg(1,Term,Arg1), not_occurs_in(Var,Arg1);
		%true ->
			not_occurs_in_args(Var,Term,N)).

not_occurs_in_args(Var,Term,N) :-
	N = 2 ->
		arg(2,Term,Arg2), not_occurs_in(Var,Arg2),
		arg(1,Term,Arg1), not_occurs_in(Var,Arg1);
	%true ->
		arg(N,Term,ArgN), not_occurs_in(Var,ArgN),
		N1 is N - 1, not_occurs_in_args(Var,Term,N1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%  truncate
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

truncate(TR_Flag,ID) :-
	ground(TR_Flag),
	(TR_Flag='@truncation'(ID) ->
	    true ;
	%true ->
	    write('Error in truncate !!'),nl,fail
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%% unit and/or lemma matching 
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

unit_lemma_match(LemL-LemID,UnitL-UnitID,Lite,_,_):-
	reset_flag_atom(5),reset_flag_atom(6),reset_flag_atom(8),
	reset_flag_number(1),
	(\+ \+ (call(UnitL); set_flag_atom(8,'@unit_none'),call(LemL)) ->
	       numbervars_termdepth(Lite,100,_,SFlag),
	       set_flag_atom(5,SFlag),
	       (call(UnitL) ->
		       inc_carray(0),set_flag_atom(6,'@unit_match'),
		       set_flag_number(1,UnitID),
		       fail ;
		call(LemL) ->
		       inc_carray(5),set_flag_atom(6,'@lemma_match'),
		       set_flag_number(1,LemID),
		       fail ;
	       %true ->
		       !,fail
	       );
	%true ->
	       !,fail
	).

unit_lemma_match(LemL-_,UnitL-_,_,LU_flag,LU_lite):-
	call_flag_atom(6,X),
	call_flag_number(1,ID),
	(X == '@unit_match' ->
	        LU_flag = u_match(ID), LU_lite = UnitL;
	 X == '@lemma_match' ->
	        LU_flag = l_match(ID), LU_lite = LemL ;

	 %true ->
	        nl,write('Erreor 2 in matching resolution !!'),nl
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

unit_match(UnitL-UnitID,Lite,_):-
	reset_flag_atom(5),reset_flag_atom(8),reset_flag_number(1),
	(\+ \+ call(UnitL) ->
	       numbervars_termdepth(Lite,100,_,SFlag),
	       set_flag_atom(5,SFlag),
	       (call(UnitL) ->
		       inc_carray(0),set_flag_number(1,UnitID),
		       fail ;
	       %true ->
		       !,fail
	       );
	%true ->
	       set_flag_atom(8,'@unit_none'),!,fail
	).
unit_match(_,_,ID) :-
	call_flag_number(1,ID).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

lemma_match(LemL-LemID,Lite,_):-
	reset_flag_atom(5),reset_flag_number(1),
	(\+ \+ call(LemL) ->
	       numbervars_termdepth(Lite,100,_,SFlag),
	       set_flag_atom(5,SFlag),
	       (call(LemL) ->
		       inc_carray(5),set_flag_number(1,LemID),
		       fail ;
	       %true ->
		       !,fail
	       );
	%true ->
	       !,fail
	).
lemma_match(_,_,ID) :-
	call_flag_number(1,ID).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%%  reduction (with sound unification)
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%      

reduction(Lite,[(Id,ALite,_)|_],Id,GrF,RedF) :-
 	unify_check_inst_in_LHS(ALite,Lite,GrF,RedF),
 	(GrF == '@done_inst' ->
 	      inc_carray(1) ;
 	 GrF == '@no_inst' ->
 	      inc_carray(34) ;
         %true ->
 	      nl,write('Error in reduction !'),nl,nl
 	).
reduction(Lite,[_|R],Id,GrF,RedF) :- 
       reduction(Lite,R,Id,GrF,RedF).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%%  C-reduction (with sound unification)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

c_reduction(Lite,[(_,_,CList,_)|_],RL,IF,GrF,RedF) :-
	unify_C_lit(Lite,CList,RL,IF,GrF,RedF).
	
c_reduction(Lite,[_|Cdr],RL,IF,GrF,RedF) :-
	c_reduction(Lite,Cdr,RL,IF,GrF,RedF).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  This routine partially depends on 
%  the special built-in preicates of SICStus V3 !!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

unify_C_lit(_,CList,_,_,_,_) :-
	var(CList),!,fail.
unify_C_lit(Lite,[(CLite,SFlag,RL,Mutable)|_],RL,IF,GrF,RedF) :-
	SFlag == '@ground' ->
	      CLite=Lite,
	      inc_carray(53),
	      %--------------------
	      %  Notice !!!
	      %--------------------
	      get_mutable(IF,Mutable),
	      %------
	      GrF='@no_inst', RedF='@no_inst' ;
	unify_check_inst_in_LHS(CLite,Lite,GrF,RedF) ->
	      (GrF = '@no_inst' -> inc_carray(53) ; inc_carray(30)),
	      %--------------------
	      %  Notice !!!
	      %--------------------
	      get_mutable(AssList,Mutable),	       
	      convert_inst_flag(AssList,'@done_inst',IF),
	      update_mutable(IF,Mutable) ;
	      %------
        %true ->
	      fail.

unify_C_lit(Lite,[_|R],RL,IF,GrF,RedF) :-
	unify_C_lit(Lite,R,RL,IF,GrF,RedF).

convert_inst_flag([],_,[]) :- !.
convert_inst_flag([(Id,_)|Cdr],Flag,[(Id,Flag)|CCdr]) :-
	convert_inst_flag(Cdr,Flag,CCdr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%%  identical reduction 
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

identical_reduction(Lite,[(Id,ALite,_)|_],Id) :-
      Lite==ALite,!,inc_carray(2).

identical_reduction(Lite,[_|Cdr],Id) :-
      identical_reduction(Lite,Cdr,Id).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%%  identical_reduction with C-literals
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

ident_c_reduction(Lite,[(_,_,CList,_)|_],RL,IF) :-
	member_ident_CLit(Lite,CList,RL,IF),
	!,inc_carray(25).

ident_c_reduction(Lite,[_|Cdr],RL,IF) :-
	ident_c_reduction(Lite,Cdr,RL,IF).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  This routine partially depends on 
%  the special built-in preicates of SICStus V3 !!!!!!!!!
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

member_ident_CLit(_,CList,_,_) :-
	var(CList),!,fail.
member_ident_CLit(Lite,[(CLite,_,RL,Mutable)|_],RL,IF) :-
	Lite == CLite, !,
	%----------------------
	%  Notice !!
	%----------------------	
	get_mutable(IF,Mutable).

member_ident_CLit(Lite,[_|Cdr],RL,IF) :-
	member_ident_CLit(Lite,Cdr,RL,IF).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%% lemma resolution with full unification
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

unit_lemma_resolution(LemL-LemID,LNum,SF,OF1,OF2) :-
	functor(LemL,F,A),AA is A-1,
	(SF=='@linear' ->
	        arg(AA,LemL,TFlag),
	        call(LemL),
	        LemID \== '@initial_dummy',      % Ignore dummy lemmas !!
		inc_carray(32),
	        ( LemID =< LNum ->
	               OF1='@matching_done'; OF1='@not_matching' ),
		( TFlag=='@ground' ->
		       OF2='@ground_lem_resol' ; OF2='@not_ground_lem_resol') ;
	%true -> 
	        call(lemma_attrib(F,A,LemStat)), 
		((LemStat == '@ground' ; LemStat == '@linear') ->
			arg(AA,LemL,TFlag),
			call(LemL),
			LemID \== '@initial_dummy',     % Ignore dummy lemmas !!
			(TFlag == '@ground' ->
			       inc_carray(29);
			 TFlag == '@linear' ->
			       inc_carray(32);
			%true ->
			  nl,write('Error 2 occurs in lemma resolution !!'),nl,nl
			),
			( LemID =< LNum ->
			       OF1='@matching_done'; OF1='@not_matching' ),
			( TFlag=='@ground' ->
			       OF2='@ground_lem_resol'; OF2='@not_ground_lem_resol');
		 LemStat == '@non_linear' ->
			functor(Term,F,A),
			arg(A,Term,LemID),arg(AA,Term,TFlag),
			call(Term),
			LemID \== '@initial_dummy',      % Ignore dummy lemmas !!
			(TFlag=='@ground' ->
				LemL = Term,inc_carray(29); 
			 TFlag=='@linear' ->
				LemL = Term,inc_carray(32); 
			%true ->
				unify(LemL,Term),inc_carray(16)
			),
			( LemID =< LNum ->
			       OF1='@matching_done'; OF1='@not_matching' ),
			( TFlag=='@ground' ->
			       OF2='@ground_lem_resol'; OF2='@not_ground_lem_resol');
		%true ->
		      nl, write('Error 3 occurs in lemma resolution !!'),nl,nl
		)
	).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%  miscellaneous setting routines
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_SU_flag_LemUni(SF,UF) :-
	call_flag_atom(5,SS1),call_flag_atom(8,UU1),
	(UU1=='@unit_none' -> UF='@unit_none',inc_carray(47); true),
        (SS1=='@initial_data' ->
	      SF='@no_matched_Lem',inc_carray(49);
	 SS1 == '@ground' -> 
	      SF='@linear',inc_carray(62);
	 SS1 == '@linear' -> 
	      SF=SS1, inc_carray(31);
	 SS1 == '@non_linear' ->
	      SF=SS1, inc_carray(44);
        %true ->
	      nl,
	      write(SS1-UU1),
              nl,write('Error in set_SU_flag_LemUni !!'),nl,nl
	).

%%%%%%%%%%%%
set_SU_flag_Uni(SF,UF) :-
	call_flag_atom(5,SS1),call_flag_atom(8,UU1),
	(UU1=='@unit_none' -> UF='@unit_none',inc_carray(47); true),
        (SS1=='@initial_data' ->
	      SF='@not_examined',inc_carray(49);
	 SS1 == '@ground' -> 
	      SF='@linear',inc_carray(62);
	 SS1 == '@linear' -> 
	      SF=SS1, inc_carray(31);
	 SS1 == '@non_linear' ->
	      SF=SS1, inc_carray(44);
        %true ->
              nl,write('Error in set_SU_flag_Uni !!'),nl,nl
	).

%%%%%%%%%%%%
set_S_flag_Lem(SF) :-
	call_flag_atom(5,SS1),
        (SS1=='@initial_data' ->
	      SF='@no_matched_Lem',inc_carray(49);
	 SS1 == '@ground' -> 
	      SF='@linear',inc_carray(62);
	 SS1 == '@linear' -> 
	      SF=SS1, inc_carray(31);
	 SS1 == '@non_linear' ->
	      SF=SS1, inc_carray(44);
	%true ->
              nl,write('Error in set_S_flag_Lem !!'),nl,nl
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_id_dcounter(CS,CA0,C) :- 
        depth_bound_measure_runtime_data(sub) ->
	     C = CS ;
	%depth_bound_measure_runtime_data(all) ->
	     C = CA0.

%%%%%%%%%%%%%%%%%%

check_depth_bound(CS0,CS1,CA0,CA1,N) :-
	depth_bound_measure_runtime_data(sub) ->
	     (CS0 >= 1 ->
	           CS1 is CS0 -1 ;
	     %true ->
	           set_flag_atom(12,'@Finitely_failed'),
		   fail
	     ) ;
	%depth_bound_measure_runtime_data(all) ->
	     (CA0 > N  ->
	           CA1 is CA0 - N ;
	     %true ->
	           set_flag_atom(12,'@Finitely_failed'),
		   fail
	     ).

%%%%%%%%%%%%%%%%%%

set_U_flag(UF) :-
	var(UF) -> 
	     UF = '@unit_exist';
	UF == '@unit_none' ->
	     inc_carray(36);
	%true ->
	     nl,write('Error in set_U_flag !!'),nl,nl.

%%%%%%%%%%%%%%%%%%

dappend(L1-E1,Data,L2-E2) :-
	L2=L1, E1=[Data|E2].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%    merge_RedList
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

merge_RedList(_,RLin1,_,_,_) :-
	RLin1 == '@top',!.

merge_RedList(LaF,RLin1,RLin2,_,_) :-
	\+ truncate(LaF,_),\+ (ground(RLin1),ground(RLin2)),!,
	nl,write('Error 3 in merge_RedList !!'),nl,
	write(('LaF',LaF)),nl,
	write(('Rlin1',RLin1)),nl,
	write(('Rlin2',RLin2)),nl.

merge_RedList(LaF,RLin1,_,RLin1,_) :-
	truncate(LaF,_),!.

merge_RedList(LaF,RLin1,RLin2,RLout,C) :-
	(nonvar(LaF) -> 
	     write('Error 5 in merge_RedList !!'),nl ; true),
	%----------------------------------
	merge_RL(RLin1,RLin2,RLout,C),
	%----------------------------------
	(var(RLout) ->
	   write('Error 1 in merge_RedList !!!'),nl;
	%true ->
           true
	).

%%%%%%%%%%%%%%%%

merge_RL(Lin1,Lin2,Lout,C) :-
	C == '@top',!,
	((Lin1 \== [] ; Lin2 \== []) ->
	   nl,write('Error top level in merge_RL !!'),nl ;
	 %true ->
	       Lout = []
	).

merge_RL(Lin1,Lin2,Lin1,C) :-
	(var(Lin1) ; var(Lin2)) ->
    	       nl,write('Error 7 in merge_RL !!'),nl ;
	Lin1 = [Id|_], Id < C ->
    	       nl,write('Error 10 in merge_RL !!'),nl ;
        %----------------------------------
	Lin2==[] ->
	       ! ;
	%true ->
	       fail.

merge_RL([],Lin2,Lout,C) :-
	!,
	(
	 Lin2==[] ->
	       nl,write('Error 2 in merge_RL !!'),nl ;
	 Lin2=[Id|Cdr] ->
	      (
	       var(Id) ->
             	   nl,write('Error 8 in merge_RL !!'),nl ;
	       Id == '@top' ->
             	   nl,write('Error 3 in merge_RL !!'),nl ;
	       Id < C ->
		   nl,write('Error 1 in merge_RL !!'),nl,
		   write(('Lin2',Lin2)),nl,
		   write(('C',C)),nl ;
        %----------------------------------
	       Id == C ->
	           Lout=Cdr ;
	       Id > C ->    
	           Lout=Lin2 ;
	      %true ->
		   nl,write('Error 0 in merge_RL !!'),nl 
	      ) ;
        %----------------------------------
	%ture ->
	      write('Error 4 in merge_RL !!'),nl,
	      write(('Lin2',Lin2)),nl
	).

merge_RL([C1|R1],[C2|R2],Lout,C) :-
	(var(C1) ; var(C2)) ->
	     nl,write('Error 9 in merge_RL !!'),nl ;
	(C1 =< C ; C2 < C ) ->
	     nl,write('Error 5 in merge_RL !!'),nl ;
	(C1 == '@top' ; C2 == '@top') ->
	     nl,write('Error 6 in merge_RL !!'),nl ;
        %----------------------------------
	C2 == C ->
	     merge_RL([C1|R1],R2,Lout,C);
	C1 < C2 ->
	     Lout=[C1|Cdr],
	     merge_RL(R1,[C2|R2],Cdr,C) ;
	C1 == C2 ->	
	     Lout=[C1|Cdr],
	     merge_RL(R1,R2,Cdr,C) ;
	C1 > C2 ->
	   Lout=[C2|Cdr],
	   merge_RL([C1|R1],R2,Cdr,C).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%    merge_InsFList
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

merge_InsFList(_,IF1,_,_,_) :-
	IF1 == '@top',!.

merge_InsFList(LaF,IF1,IF2,_,_) :-
	\+ truncate(LaF,_),\+ (ground(IF1),ground(IF2)),!,
	nl,write('Error 3 in merge_InsFList !!'),nl,
	write(('LaF',LaF)),nl,
	write(('IF1',IF1)),nl,
	write(('IF2',IF2)),nl.

merge_InsFList(LaF,IF1,_,IF1,_) :-
	truncate(LaF,_),!.

merge_InsFList(LaF,IF1,IF2,IFout,C) :-
	(nonvar(LaF) -> 
	     write('Error 5 in merge_InsFList !!'),nl ; true),
	%----------------------------------
	merge_IF(IF1,IF2,IFout,C),
	%----------------------------------
	(var(IFout) ->
	   write('Error 1 in merge_InsFList !!!'),nl;
	%true ->
           true
	).

%%%%%%%%%%%%%%%%

merge_IF(Lin1,Lin2,Lout,C) :-
	C == '@top',!,
	((Lin1 \== [] ; Lin2 \== []) ->
	   nl,write('Error top lever in merge_IF !!'),nl ;
	 %true ->
	       Lout = []
	).

merge_IF(Lin1,Lin2,Lin1,C) :-
	(var(Lin1) ; var(Lin2)) ->
    	       nl,write('Error 7 in merge_IF !!'),nl ;
	Lin1 = [(Id,_)|_], Id < C ->
    	       nl,write('Error 10 in merge_IF !!'),nl ;
        %----------------------------------
	Lin2==[] ->
	       ! ;
	%true ->
	       fail.

merge_IF([],Lin2,Lout,C) :-
	!,
	(
         Lin2==[] ->
	       nl,write('Error 2 in merge_IF !!'),nl ;
	 Lin2=[(Id,_)|Cdr] ->
	      (
	       var(Id) ->
             	   nl,write('Error 8 in merge_IF !!'),nl ;
	       Id == '@top' ->
             	   nl,write('Error 3 in merge_IF !!'),nl ;
	       Id < C ->
		   nl,write('Error 1 in merge_IF !!'),nl ;
        %----------------------------------
	       Id == C ->
	           Lout=Cdr ;
	       Id > C ->    
	           Lout=Lin2 ;
	      %true ->
		   nl,write('Error 0 in merge_IF !!'),nl 
	      ) ;
        %----------------------------------
	%ture ->
	      write('Error 4 in merge_IF !!'),nl,
	      write(('Lin2',Lin2)),nl
	).

merge_IF([(C1,F1)|R1],[(C2,F2)|R2],Lout,C) :-
	(var(C1) ; var(C2)) ->
	     nl,write('Error 9 in merge_IF !!'),nl ;
	(C1 =< C ; C2 < C ) ->
	     nl,write('Error 5 in merge_IF !!'),nl ;
	(C1 == '@top' ; C2 == '@top') ->
	     nl,write('Error 6 in merge_IF !!'),nl ;
        %----------------------------------
	C2 == C ->
	     merge_IF([(C1,F1)|R1],R2,Lout,C);
	C1 < C2 ->
	     Lout=[(C1,F1)|Cdr],
	     merge_IF(R1,[(C2,F2)|R2],Cdr,C) ;
	C1 == C2 ->	
	     Lout=[(C1,FF)|Cdr],
	     merge_IF(R1,R2,Cdr,C),
	     ((F1=='@done_inst';F2=='@done_inst') ->
	            FF='@done_inst';
	      %true ->
	            FF='@no_inst'
	     );
	C1 > C2 ->
	   Lout=[(C2,F2)|Cdr],
	   merge_IF([(C1,F1)|R1],R2,Cdr,C).

%%%%%%%%%%%%%%%%%%%%%%%%%%%

collect_greater(_,X,_) :-
	var(X),!,
	nl,write('Error 0 in collect_greater !!'),nl.
collect_greater(C,In,Out)  :-
	collect(C,In,Out,_).

	
collect(_,[],[],_) :- !.
collect(C,[(Id,F)|Cdr],Out,Flag) :-
	 var(Flag), Id < C ->
	      collect(C,Cdr,Out,Flag) ;
	(\+ var(Flag)), Id < C ->
	      nl,write('Error 1 in collect_greater !!'),nl ;
	%true ->
	      Out = [(Id,F)|Out1],
	      collect(C,Cdr,Out1,'@greater').

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%  non-identical_member 
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

nonidentical_member(_,[]) :- !.
nonidentical_member(X,[(_,Y,_)|_]) :-
	X == Y, !, inc_carray(3),fail.
nonidentical_member(X,[_|L]) :-
	nonidentical_member(X,L).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%%     ancestor lemma matching and pruning for a direct ancestor
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

h_direct_pruning(TR_Flag,_,_,_) :- truncate(TR_Flag,_),!.
h_direct_pruning(_,[('@top',_)],[('@top',_)],_) :- !.
h_direct_pruning(_,FLp,FLn,GoNum) :-
	GoNum > 0 ->
	       h_pruning(FLn,FLp,GoNum) ;
	GoNum < 0 ->
	       h_pruning(FLp,FLn,GoNum) ;
	GoNum == 0 ->
	       true ;
	%true ->
               nl,write('Error in h_direct_pruning 1 !!'),nl.

h_pruning([(_,Lite,_)|_],_,_) :-
	Lite == '$$dumL', !, write('Error in h_pruning 2 !!'),nl.

h_pruning([(_,Lite,_,_,UF-UnitL,LemL-LemID,AntiL)|R1],[_|R2],GoNum) :-
	reset_flag_atom(7), reset_flag_number(7),
        \+ \+ (check_mode_head_pruning(GoNum),
	           (head_unit_subsump_flag(on),UF=='@unit_exist',call(UnitL);
		    head_anti_lemma_subsump_flag(on),call(AntiL);
		    head_lemma_subsump_flag(on),call(LemL)) ;
	        ancestor_unit_lemma_matching_flag(on),abs(GoNum) =:= 1,call(LemL)
	      ),
         f_numbervars(AntiL),
	 %-----------------------------------------
         (check_mode_head_pruning(GoNum) ->
	       (head_unit_subsump_flag(on),UF=='@unit_exist',call(UnitL) ->
		      !,inc_carray(4),fail ;
		head_anti_lemma_subsump_flag(on),call(AntiL) ->
		      !,inc_carray(48),fail ;
		head_ident_A_chk_flag(on),ident_chk(Lite,R1) ->
		      !,inc_carray(55),fail ;
		head_compl_A_chk_flag(on),ident_chk(Lite,R2) ->
		      !,inc_carray(56),fail ;
		head_lemma_subsump_flag(on),call(LemL) ->
		      !,inc_carray(17),fail ;
                %true ->		   
		      true 
	       ) ;
         %true ->
	       true
	 ),
	 %-----------------------------------------
	 (ancestor_unit_lemma_matching_flag(on),abs(GoNum)=:=1,
	  call(LemL) ->
	       set_flag_atom(7,'@truncation_flag_on'),
	       set_flag_number(7,LemID),
	       fail ;
	 %true ->
	       fail
	 ).
	 %-----------------------------------------


h_pruning([(_,Lite,_,TR_Flag,_)|_],[_|R2],GoNum) :-
	call_flag_atom(7,'@truncation_flag_on') ->
	      call_flag_number(7,LemID),
	      TR_Flag = '@truncation'(LemID),
	      inc_carray(12) ;
	ancestor_c_literal_matching_flag(on),abs(GoNum)=:=1,
	ident_C_Lite_chk(Lite,R2,Id,RL,IF) ->
	      TR_Flag = '@truncation'(cLit(Id,RL,IF)),
	      inc_carray(59) ;
	%true ->
	      true.

%%%%%%%%%%

check_mode_head_pruning(GoNum) :-
	head_pruning_position_runtime_data(X),
	AbG is abs(GoNum),
	(X==nowhere ->
	       fail ;
	 X==everywhere ->
	       true ;
	 X==only_first ->
	       ((AbG==2 ; AbG==10000) ->
		     true ;
	       %true ->
	             fail
	       ) ;
	 X==except_for_first ->
	       (AbG\==2, AbG\==10000 ->
		     true ;
	       %true ->
	             fail
	       ) ;
	 X==only_last ->
	       (AbG==1  ->
		     true ;
	       %true ->
 	             fail
	       ) ;
         %true ->
	       nl,write('Error in check_mode_head_pruning !!'),nl,
	       fail
	).

%%%%%%%%%%

ident_chk(Lite,[(_,ALite,_)|_]) :-
      Lite==ALite,!.
ident_chk(Lite,[_|R]) :-
      ident_chk(Lite,R).

%%%%%%%%%%

ident_C_Lite_chk(Lite,[(Id,_,CList,_)|_],Id,RL,IF) :-
	member_ident_CLit(Lite,CList,RL,IF),!.
ident_C_Lite_chk(Lite,[_|Cdr],Id,RL,IF) :-
	ident_C_Lite_chk(Lite,Cdr,Id,RL,IF).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%        tail pruning after unit expansions
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

t_pruning_after_unit_expansion(FLnew1-_LaF) :-
	FLnew1 = [(_,Lite,_)|FLnow1],
	( tail_ident_A_chk_flag(on),ident_chk(Lite,FLnow1) ->
	        !,inc_carray(82),fail ;
	 %true ->
		true
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%        tail pruning and lemmaizing for non-unit expansion
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

t_pruning_and_lemmaizing(_-LaF,_,_,RL-RL,IF-IF,_,_,_,_) :-
	truncate(LaF,Env),Env='@descend'(_),!.
	%%%%%%%%%%%%%%%%%%%%%%%%
	% nl,write((' LaF: ',LaF)),nl,write((' RL: ',RL)),nl.
	%%%%%%%%%%%%%%%%%%%%%%%%

t_pruning_and_lemmaizing(FLnew1-LaF,FLnew2,
	                 _Sol-End,RL0-RL1,IF0-IF1,TmpL,Cut,OF,GoNum) :-
        (var(RL0) ->
	    nl,write('Error 0 in t_pruning_and_lemmaizing !!'),nl ;
	%true ->
	    true
	),
	%---------------------------------------------
	FLnew1 = [(C,Lite,_,_,_,LemL-_,AntiL)|FLnow1],
	%---------------------------------------------
	(truncate(LaF,Env) ->
	       (Env = cLit(_,RList,IFList) ->
	             RL1 = RList,
	             merge_IF(IF0,IFList,IF1,C) ;
	        number(Env) ->
                     RL1 = [],
		     IF1 = IF0 ;
	        Env = '@reduc_match'(_) ->
                     RL1 = [],
		     IF1 = IF0 ;
	       (Env = '@unit_repl'(_,RList,IFList) ; 
		Env = '@lemma_repl'(_,RList,IFList) ;
		Env = '@c_lit_repl'(_,RList,IFList)) ->
	             merge_RL([],RList,RL1,C),
		     collect_greater(C,IFList,GreaterIF),
	             merge_IF(IF0,GreaterIF,IF1,C) ;
	       %true ->
	             nl,write('Error 0 in t_pruning !!'),nl,
	             write(('Env',Env)),nl
	       ) ;
	 %true ->
	       RL1 = RL0, IF1 = IF0
	),
	%---------------------------------------------
	%
	prune_and_chk_subsumption(FLnew1,FLnew2,RL1,OF,GoNum),
	%
	%---------------------------------------------
	call_flag_atom(3,SFlag),call_flag_atom(1,LemFlag),
	%
	(call_flag_atom(10,'@unit_repl') ->
	      call_flag_number(10,UID),
	      (RL1 == [] ->
       	             %----------
	             process_inconsistency(GoNum,End,'@unit_repl',UID) ;
		     %----------	             
	       RL1 = [TrID|_] ->
	             set_truncation_flag(FLnow1,TrID,'@unit_repl'(UID,RL1,IF1)) ;
	       %true ->
	             nl,write('Error 2 in t_pruning_and_lemmaizing !!'),nl,
	             write(('RL1',RL1)),nl
	      );
	 call_flag_atom(11,'@lemma_repl') ->
	      call_flag_number(11,LID),
	      (RL1 == [] ->
       	             %----------
	             process_inconsistency(GoNum,End,'@lemma_repl',LID) ;
		     %----------	             
	       RL1 = [TrID|_] ->
	             set_truncation_flag(FLnow1,TrID,'@lemma_repl'(LID,RL1,IF1)) ;
	       %true ->
	             nl,write('Error 3 in t_pruning_and_lemmaizing !!'),nl
	      );
	inconsistency_check_flag(on),
	RL1==[], SFlag\=='@ground', GoNum\==0 ->
	      %
	      % nl,write('Invoked: simple_check_inconsistency'),nl,
	      %
	      (simple_check_inconsistency(FLnew2,Kind,ID) ->
       	             %----------
       	             process_inconsistency(GoNum,End,Kind,ID) ;
       	             %----------
	      %true ->
	             %nl,write('Failed: simple_check_inconsistency'),nl,
		     %
		     true
	      ) ;
	 c_literal_repl_subsumption_flag(on),
	 ( (abs(GoNum)=\=1,abs(GoNum)=\=10000 ) ;     % for exception of the last goal in a clause
	           last_goal_repl_subsumption_flag(on)),
	 ident_C_Lite_chk(Lite,FLnow1,CID,CRL,CIF) ->
	      inc_carray(76),
	      merge_RL(RL1,CRL,CRLout,C),
	      merge_IF(IF1,CIF,CIFout,C),
	      (CRLout = [] ->
	             nl,write('Error 5 in t_pruning_and_lemmaizing !!'),nl ;
	       CRLout = [TrID|_] ->
	             set_truncation_flag(FLnow1,TrID,'@c_lit_repl'(CID,CRLout,CIFout)) ;
	       %true ->
	             nl,write('Error 4 in t_pruning_and_lemmaizing !!'),nl
	      );
	%true ->
	      true
	),
	%---------------------------------------------
	(LemFlag=='@lemmaizing' ->
	     (RL1==[] ->
	          (unit_lemma_resolution_flag(on) ->
		   %----------------------------
		   % Notice: the above automatically implies 
		   %         the unit_lemmaizing_flag(on).
		   %----------------------------
	                  lemma_assert(LemL,SFlag,End);
	           unit_lemmaizing_flag(on) ->
	                  simple_lemma_assert(LemL,SFlag,End);
		   non_unit_lemmaizing_flag(on) ->
		          generate_non_unit_lemma(Lite,'@top',SFlag,[],[],FLnew2);
		  %true ->
		          nl,write('Error in lemmaizing_flag !!'),nl
	          );
             non_unit_lemmaizing_flag(on),RL1=[Id|_] ->
	          generate_non_unit_lemma(Lite,Id,SFlag,RL1,IF1,FLnew2) ;
	     %true ->
	          write('Error 1 in t_pruning_and_lemmaizing !!'),nl
	     );
	%true ->
	     true 
	),
	%---------------------------------------------
	(flist_inst_chk(IF1) ->
	     (cut_by_no_inst_flag(on) ->
	          chk_cut_after_expan(AntiL,TmpL,Cut,SFlag);
	      cut_by_no_inst_flag(off) ->
	          Cut='@no_cut'
	     ),
	     (Cut=='@no_cut',anti_lemmaizing_flag(on) ->
		  (LemFlag=='@lemmaizing' ->
		      (unit_lemmaizing_flag(on),
		       tail_lemma_subsump_flag(on),RL1==[] ->
		           true ;
		      %true ->   
		           anti_lemma_assert(AntiL,SFlag)
		      ) ;
		  call_flag_atom(2,'@small') ->
		      anti_lemma_assert(AntiL,SFlag); 
		  %true ->
		      true 
		  );
	      %true ->
	          true 
	     );
	%true ->
	    Cut = '@no_cut'
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   prune_and_chk_subsumption
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

prune_and_chk_subsumption(FLnew1,FLnew2,RL,OF,GoNum) :-
	reset_flag_atom(1),reset_flag_atom(2),reset_flag_atom(3),
	reset_flag_atom(10),reset_flag_number(10),
	reset_flag_atom(11),reset_flag_number(11),
	FLnew1 = [(_,Lite,_,LaF,UF-UnitL,LemL-_,AntiL)|FLnow1],
	FLnew2 = [(_,_,_,_,CUnitL-CUnitID,CLemL-CLemID,_)|FLnow2],
	make_ground_and_check_term_at_tail(AntiL,DFlag,SFlag),
	%-----------------------------
	(tail_unit_subsump_flag(on),UF=='@unit_exist',call(UnitL) ->
	        !, inc_carray(10),fail ;
	 tail_anti_lemma_subsump_flag(on),call(AntiL) ->
	        !, inc_carray(20),fail ;
	 %%%%%
	 % horn_or_non_horn_data(horn),
	 %%%%%
	 tail_ident_A_chk_flag(on),ident_chk(Lite,FLnow1) ->
	        !,inc_carray(57),fail ;
	 tail_compl_A_chk_flag(on),ident_chk(Lite,FLnow2) ->
	        !,inc_carray(58),fail ;
	 tail_lemma_subsump_flag(on),call(LemL) ->
		!, inc_carray(18),fail ;
	%-----------------------------
	 unit_repl_subsumption_flag(on),
	 ( (abs(GoNum)=\=1,abs(GoNum)=\=10000 ) ; last_goal_repl_subsumption_flag(on)),
	 CUnitL \== '$$CUnitL',call(CUnitL) ->
	        inc_carray(72),
	        set_flag_atom(10,'@unit_repl'),
	        set_flag_number(10,CUnitID) ;
	 lemma_repl_subsumption_flag(on),
	 ( (abs(GoNum)=\=1,abs(GoNum)=\=10000 ) ; last_goal_repl_subsumption_flag(on)),
	 CLemL \== '$$CLemL',call(CLemL) ->
	        inc_carray(73),
	        set_flag_atom(11,'@lemma_repl'),
	        set_flag_number(11,CLemID) ;
	%-----------------------------
	%true ->
	        true
	),
	%-----------------------------
	set_flag_atom(3,SFlag),
	(lemmaizing_flag(on),
	 check_lemma_condition(LaF,OF,DFlag) ->
             (non_unit_lemmaizing_flag(on) ->
       	          (unit_lemmaizing_flag(on) ->
	               (tail_lemma_subsump_flag(on) ->
		  	      set_flag_atom(1,'@lemmaizing');
		  	forward_subsumption(LemL) ->
		  	      set_flag_atom(1,'@lemmaizing');
	          	%true ->
		  	      set_flag_atom(2,DFlag)
		  	) ;
		  %true ->
	                set_flag_atom(1,'@lemmaizing')
		  );
	      non_unit_lemmaizing_flag(off),
              RL==[],unit_lemmaizing_flag(on) ->
	          (tail_lemma_subsump_flag(on) ->
		       set_flag_atom(1,'@lemmaizing');
 		   forward_subsumption(LemL) ->
		       set_flag_atom(1,'@lemmaizing');
	           %true ->
		       set_flag_atom(2,DFlag)
		  );
	     %true ->
     		  set_flag_atom(2,DFlag)
	     );
         %true ->
              set_flag_atom(2,DFlag)
	),
	fail.

prune_and_chk_subsumption(_,_,_,_,_).


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   simple_check_inconsistency
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

simple_check_inconsistency(FLnew2,_Kind,_ID) :-
	reset_flag_atom(13),reset_flag_number(13),
	reset_flag_atom(14),reset_flag_number(14),
	%
	FLnew2 = [(_,_,_,_,CUnitL-CUnitID,CLemL-CLemID,_)|_],
	%
	%write('Disposed FLnew2 in simple_check_inconsistency !'),nl,
	(CUnitL\=='$$CUnitL' ->
	      functor(CUnitL,F1,A1),functor(TmpU,F1,A1),
	      (call(TmpU),unify(TmpU,CUnitL) ->
	             inc_carray(65),
	             set_flag_atom(13,'@unit_repl'),
		     set_flag_number(13,CUnitID),
		     fail ;
	       %true ->	  
                     true 
	      ) ;
	%true ->
	      true
	),
	(CLemL\=='$$CLemL' ->
	      functor(CLemL,F2,A2),functor(TmpL,F2,A2),
	      (call(TmpL),
	       unify(TmpL,CLemL),CLemID \=='@initial_dummy' ->
	             inc_carray(69),
	             set_flag_atom(14,'@lemma_repl'),
		     set_flag_number(14,CLemID),
		     fail ;
	       %true ->	  
                     !,fail
	      ) ;
         %true -> 
	      !,fail
	).

simple_check_inconsistency(_FLnew2,Kind,ID) :-
	call_flag_atom(13,'@unit_repl') ->
              Kind='@unit_repl',
	      call_flag_number(13,ID) ;
	call_flag_atom(14,'@lemma_repl') ->
	      Kind='@lemma_repl',
	      call_flag_number(14,ID) ;
	%true ->
	      nl,write('Error: simple_check_inconsistency !'),nl,
              call_flag_atom(13,X), call_flag_atom(14,Y),
	      write('flag_atoms 13 and 14 are '),write([X,Y]),nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   chk_cut_after_expan
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

chk_cut_after_expan(AntiL,TmpL,Cut,SFlag) :-
	SFlag=='@ground' -> 
	    (AntiL==TmpL ->
		  Cut = '@cut',inc_carray(45) ;
	     %true ->
		  Cut = '@no_cut'
	    );
	subsumes_chk(AntiL,TmpL) ->
	    Cut = '@cut',inc_carray(23);
	%true ->
	    Cut = '@no_cut'.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%
%%%%%%%    A pruning after reduction 
%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

t_pruning_after_reduction(FLnew1-_LaF,FLnew2,Id,GrF,RedF) :-
	%
	pruning_red(FLnew1,FLnew2,Id,GrF),
	%------------------------------
	FLnew1 = [(_,_,_,_,_,_,AntiL)|_],
	(RedF=='@no_inst' ->
	      (anti_lemmaizing_flag(on)  ->
	           call_flag_atom(3,SFlag),
		   anti_lemma_assert(AntiL,SFlag);
	       %true ->
		   true
	      ) ;
	 %true ->
	      true
	),
	%------------
	(call_flag_atom(7,'@truncation_in_red') ->
	      call_flag_number(7,LemID),
	      set_truncation_flag(FLnew2,Id,'@reduc_match'(LemID)) ;
	%true ->
	      true
	).

%%%%%%%%%%%%%%%%%%%%%%%

set_truncation_flag([],_,_) :-
	!,nl,write('Error in set_truncation_flag !!!'),nl.
set_truncation_flag([(C,_,_,LaF,_)|Cdr],Id,Flag) :-
	C \== Id ->
	     LaF = '@truncation'('@descend'(Flag)),
	     set_truncation_flag(Cdr,Id,Flag) ;
	C == Id ->
	     LaF = '@truncation'(Flag).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   pruning_red
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pruning_red(FLnew1,FLnew2,Id,GrF) :-
	reset_flag_atom(3),
	reset_flag_atom(7),reset_flag_number(7),
	FLnew1 = [(_,Lite,_,_,UF-UnitL,LemL-_,AntiL)|FLnow1],
	(GrF == '@no_inst' ->
	       SFlag = '@ground' ;
	 %true ->
	       make_ground_and_check_term_at_tail(AntiL,_,SFlag)
	),
	%--------------------------------------
	(tail_unit_subsump_flag(on), var(UF), %% This means UF=='@unit_exist' !!
	 call(UnitL) ->
	        !, inc_carray(38),fail ;
	 tail_anti_lemma_subsump_flag(on),call(AntiL) ->
	        !, inc_carray(19),fail ;
	 tail_lemma_subsump_flag(on), call(LemL) ->
		!, inc_carray(52),fail ;
	 tail_ident_A_chk_flag(on),ident_chk(Lite,FLnow1) ->
	        !,inc_carray(80),fail ;
	 %true ->
	        set_flag_atom(3,SFlag)
	),
	%--------------------------------------
        chk_lit_flist(FLnew2,Id,Result),
	(Result == '@failed' ->
	       ! ;
	Result == '@success' ->
	       true 
	),
	fail.

pruning_red(_,_,_,_).

%%%%%%%%%%%%%%%%%%%%%

chk_lit_flist([],_,_) :- 
	!,nl,write('Error in chk_lit_flist !!'),nl.
chk_lit_flist([(C,Other2)|Cdr2],Id,Result) :-
	C \== Id ->
	    chk_lit_flist(Cdr2,Id,Result) ;
	C == Id ->
	    reset_flag_atom(7),reset_flag_number(7),
	    Other2 = (_,_,_,UF-UnitL,LemL-LemID,AntiL),
	    %%%%%%%%%%%%%%%%%%%%
	    %f_numbervars(AntiL),
	    %-----------------------------------
	    % ?? The success of the below is extremely scare ????
	    %
	   (tail_unit_subsump_flag(on), var(UF), 
	                  %% This means UF=='@unit_exist' !!
	    call(UnitL) ->
	          inc_carray(66),Result='@failed' ;
	    %---
	    tail_anti_lemma_subsump_flag(on),call(AntiL) ->
	          inc_carray(70),Result='@failed' ;
	    tail_lemma_subsump_flag(on), call(LemL) ->
		  inc_carray(71),Result='@failed' ;
	    %-----------------------------------
	    ancestor_unit_lemma_matching_in_reduction_flag(on),
	    call(LemL) ->
		  inc_carray(63),
	          set_flag_atom(7,'@truncation_in_red'),
	          set_flag_number(7,LemID),
	          Result='@success';
	    %true ->
	          Result='@success'
	   ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%
%%%%%%%    A pruning after C-reduction 
%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

t_pruning_after_C_reduction(FLnew1-_LaF,FLnew2,RL,IF,GrF,RedF,GoNum) :-
	%
	pruning_C_red(FLnew1,FLnew2,GrF,GoNum),
	FLnew1 = [(C,Lite,_,_,_,_,AntiL)|FLnow1],
	%---------------------------------------------
	(call_flag_atom(10,'@unit_repl') ->
	      call_flag_number(10,UID),
	      (RL == [] ->
	             nl,write('Error 2 in t_pruning_after_C_reduction !!'),nl ;
	       RL = [TrID|_] ->
	             set_truncation_flag(FLnow1,TrID,'@unit_repl'(UID,RL,IF)) ;
	       %true ->
	             nl,write('Error 3 in t_pruning_and_lemmaizing !!'),nl
	      );
	 call_flag_atom(11,'@lemma_repl') ->
	      call_flag_number(11,LID),
	      (RL == [] ->
	             nl,write('Error 4 in t_pruning_after_C_reduction !!'),nl ;
	       RL = [TrID|_] ->
	             set_truncation_flag(FLnow1,TrID,'@lemma_repl'(LID,RL,IF)) ;
	       %true ->
	             nl,write('Error 5 in t_pruning_and_lemmaizing !!'),nl
	      );
	 c_literal_repl_subsumption_flag(on),
	 ( (abs(GoNum)=\=1,abs(GoNum)=\=10000) ; last_goal_repl_subsumption_flag(on)),
	 ident_C_Lite_chk(Lite,FLnow1,CID,CRL,CIF) ->
	      inc_carray(77),
	      merge_RL(RL,CRL,CRLout,C),
	      merge_IF(IF,CIF,CIFout,C),
	      (CRLout == [] ->
	             nl,write('Error 6 in t_pruning_after_C_reduction !!'),nl ;
	       CRLout = [TrID|_] ->
	             set_truncation_flag(FLnow1,TrID,'@c_lit_repl'(CID,CRLout,CIFout)) ;
	       %true ->
	             nl,write('Error 7 in t_pruning_and_lemmaizing !!'),nl
	      );
	%true ->
	      true
	),
	%---------------------------------------------
	(RedF=='@no_inst' ->
	      (anti_lemmaizing_flag(on) ->
	           call_flag_atom(3,SFlag),
		   anti_lemma_assert(AntiL,SFlag);
	       %true ->
		   true
	      ) ;
	 RedF=='@done_inst' ->
	      true ;
        %true ->
	      nl,write('Error 1 in t_pruning_after_C_reduction !'),nl,
	      write((val,RedF)),nl
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   pruning_red
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pruning_C_red(FLnew1,FLnew2,GrF,GoNum) :-
	reset_flag_atom(3),
	reset_flag_atom(10),reset_flag_number(10),
	reset_flag_atom(11),reset_flag_number(11),
	FLnew1 = [(_,Lite,_,_,UF-UnitL,LemL-_,AntiL)|FLnow1],
	FLnew2 = [(_,_,_,_,CUnitL-CUnitID,CLemL-CLemID,_)|FLnow2],
	(GrF == '@no_inst' ->
	        SFlag = '@ground' ;
	%true ->
	        make_ground_and_check_term_at_tail(AntiL,_,SFlag)
	),
	%--------------------------------------
	(tail_unit_subsump_flag(on), var(UF), %% This means UF=='@unit_exist' !!
	 call(UnitL) ->
	        !, inc_carray(26),fail ;
	 tail_anti_lemma_subsump_flag(on),call(AntiL) ->
	        !, inc_carray(60),fail ;
	 tail_lemma_subsump_flag(on), call(LemL) ->
		!, inc_carray(61),fail ;
         %%%
	 % horn_or_non_horn_data(horn),
         %%%
	 tail_ident_A_chk_flag(on),ident_chk(Lite,FLnow1) ->
	        !,inc_carray(78),fail ;
	 tail_compl_A_chk_flag(on),ident_chk(Lite,FLnow2) ->
	        !,inc_carray(79),fail ;
	%--------------------------------------
	 unit_repl_subsumption_flag(on),
	 ( (abs(GoNum)=\=1,abs(GoNum)=\=10000) ; last_goal_repl_subsumption_flag(on)),
	 CUnitL \== '$$CUnitL',call(CUnitL) ->
	        inc_carray(74),
	        set_flag_atom(10,'@unit_repl'),
	        set_flag_number(10,CUnitID),
		fail ;
	 lemma_repl_subsumption_flag(on),
	 ( (abs(GoNum)=\=1,abs(GoNum)=\=10000) ; last_goal_repl_subsumption_flag(on)),
	 CLemL \== '$$CLemL',call(CLemL) ->
	        inc_carray(75),
	        set_flag_atom(11,'@lemma_repl'),
	        set_flag_number(11,CLemID),
		fail ;
	%--------------------------------------
	 %true ->
	        set_flag_atom(3,SFlag),
		fail
	).
pruning_C_red(_,_,_,_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%
%%%%%%%    A pruning after unit_lemma_resolution 
%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

t_pruning_after_lem_resol(FLnew1-_LaF,FLnew2,TmpL,Cut,OF1,OF2) :-
	%
	pruning_lem_resol(FLnew1,FLnew2,OF2),
	%-----------------------------
	call_flag_atom(3,SFlag),
	FLnew1 = [(_,_,_,_,_,_,AntiL)|_],
	(cut_by_no_inst_flag(on) ->
	     chk_cut_after_lem_resol(AntiL,TmpL,Cut,OF1,SFlag) ;
	 cut_by_no_inst_flag(off) ->
	     Cut='@no_cut'
	),
	(Cut=='@no_cut',anti_lemmaizing_flag(on) ->
	     (SFlag=='@ground' ->
	         true ;
	      call_flag_atom(2,'@small') ->
	         anti_lemma_assert(AntiL,SFlag) ;
	     %true ->
	         true
	     );
	 %ture ->
	     true
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   pruning_lem_resol
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

pruning_lem_resol(FLnew1,FLnew2,OF2) :-
	reset_flag_atom(2),reset_flag_atom(3),
	FLnew1 = [(_,Lite,_,_,UF-UnitL,_,AntiL)|FLnow1],
	FLnew2 = [_|FLnow2],
	(OF2=='@ground_lem_resol' ->
	      DFlag='@small',SFlag='@ground';
	%true ->
	      make_ground_and_check_term_at_tail(AntiL,DFlag,SFlag)
	),
	(tail_unit_subsump_flag(on),  OF2 \== '@ground_lem_resol',
                                      %% Whether the above must be here or NOT ?
	 var(UF),  %% This means UF=='@unit_exist' !!
	 call(UnitL) ->
	        !, inc_carray(51),fail ;
	 tail_anti_lemma_subsump_flag(on),call(AntiL) ->
	        !, inc_carray(27),fail ;
	 tail_ident_A_chk_flag(on),ident_chk(Lite,FLnow1) ->
	        !,inc_carray(67),fail ;
	 tail_compl_A_chk_flag(on),ident_chk(Lite,FLnow2) ->
	        !,inc_carray(68),fail ;
	 %true ->
		set_flag_atom(2,DFlag),	set_flag_atom(3,SFlag),
		fail
	).
pruning_lem_resol(_,_,_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   chk_cut_after_lem_resol 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

chk_cut_after_lem_resol(AntiL,TmpL,Cut,OF1,SFlag) :-
	OF1=='@matching_done' ->
	     Cut='@no_cut' ;
	OF1=='@not_matching' ->	     
	     (SFlag=='@ground' ->
	          (AntiL==TmpL ->
	   	    Cut='@cut',inc_carray(45);
	          %true->
	   	    Cut='@no_cut' 
	          ) ;
             %true ->
	          (subsumes_chk(AntiL,TmpL) ->
	   	    Cut='@cut',inc_carray(23);
	          %true ->
	   	    Cut='@no_cut' 
	          )
	     );
        %true ->
            nl,write('Error 2 in chk_cut_after_lem_resol !'),nl,nl.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%    Subroutines for lemmaizing
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

check_lemma_condition(LaF,OF,DFlag) :-
	(truncate(LaF,_) ->
	       fail ;
	 single_body_lemmaizing_flag(off),OF=='@single_goal' ->
	       inc_carray(7),fail ;
	%true ->
	       inc_carray(6)
	),
	(DFlag == '@small' ->
	        inc_carray(9) ;
	%true ->
	        fail
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Notice, LemL is assumed to be ground.

forward_subsumption(LemL) :-
	call(LemL) ->
	       !,inc_carray(11),fail ;
        %true ->
               fail.
forward_subsumption(_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%

lemma_assert(LemL,SFlag,Dlist_Proof) :-
	%----------------------------------
        backward_subsumption(LemL,SFlag),
	%----------------------------------
	inc_lemma_total_number,call_lemma_total_number(CuNum),
	functor(LemL,F,A),arg(A,LemL,CuNum),
	AA is A-1,
	( SFlag == '@ground' ->
	     arg(AA,LemL,'@ground'),inc_carray(28) ;
	  SFlag == '@linear' ->
	     arg(AA,LemL,'@linear'),inc_carray(35),
	     call(lemma_attrib(F,A,LemStatus)),
	     (LemStatus \== '@ground' ->
	            !,true ; 
	      %true ->
	            retract(lemma_attrib(F,A,_)),
		    asserta(lemma_attrib(F,A,'@linear'))
	     ) ;
	  SFlag == '@non_linear' ->
	     arg(AA,LemL,'@non_linear'),
	     call(lemma_attrib(F,A,LemStatus)),
	     (LemStatus == '@non_linear' ->
	            !,true ; 
	      %true ->
	            retract(lemma_attrib(F,A,_)),
		    asserta(lemma_attrib(F,A,'@non_linear'))
	     ) ;
	% true ->    
	    nl,write('Error in lemma_assert !!'),nl,nl
	),
	asserta(LemL),
	(store_lemma_proof_flag(on) ->
	     assertz(lemma_proof(CuNum,Dlist_Proof)) ;
	 store_lemma_proof_flag(off) ->	    
	     true ;
	%true ->
	    nl,write('Error in store_lemma_proof_flag !!'),nl,nl
	),!.




%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

simple_lemma_assert(LemL,SFlag,Dlist_Proof) :-
	%----------------------------------
        backward_subsumption(LemL,SFlag),
	%----------------------------------
	inc_lemma_total_number,call_lemma_total_number(CuNum),
	functor(LemL,_,A),arg(A,LemL,CuNum),
	( SFlag == '@ground' ->
	     inc_carray(28) ;
	  SFlag == '@linear' ->
	     inc_carray(35) ; 
	  SFlag == '@non_linear' ->
	     true ;
	% true ->    
	    nl,write('Error in lemma_assert !!'),nl,nl
	),
	asserta(LemL),
	(store_lemma_proof_flag(on) ->
	     assertz(lemma_proof(CuNum,Dlist_Proof)) ;
	 store_lemma_proof_flag(off) ->	    
	     true ;
	%true ->
	    nl,write('Error in store_lemma_proof_flag !!'),nl,nl
	),
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%

backward_subsumption(LemL,SFlag) :-
	(SFlag == '@ground' -> 
	       true ;
	 lemma_backward_subsumption_flag(on) ->
               functor(LemL,F,A),functor(Term,F,A),
	       ( clause(Term,true,Ref),
		 subsumes_chk(LemL,Term),
		 erase(Ref),
		 inc_carray(14),
		 fail
	       ;
		 true
	       ) ;
	 %true ->
	       true
	),
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%  generate non_unit_lemma
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

generate_non_unit_lemma(_,_,_,_,_,[]) :-
	!,write('Error 1 in generate_non_unit_lemma !!'),nl.

generate_non_unit_lemma(Lite,Id,SFlag,_,_,FLnew2) :-
	Id == '@top',!,
	inc_carray(64),
	find_last(FLnew2,('@top',_,CList,_)),
	join_at_tail(Lite,SFlag,[],[],CList).

generate_non_unit_lemma(Lite,Id,SFlag,RL,IF,[(C,ALite,CList,_)|Cdr]) :-
	var(Id) ->
	    write('Error 2 in generate_non_unit_lemma !!'),nl ;	   
	Id < C ->
	    write('Error 3 in generate_non_unit_lemma !!'),nl ;
	%--------------------------------
	Id == C ->
	    (Lite == ALite ->
	          inc_carray(54) ;   %%% Non-sence ??????
	    %true ->
	          join_at_tail(Lite,SFlag,RL,IF,CList)
	    );
	Id > C ->
	    generate_non_unit_lemma(Lite,Id,SFlag,RL,IF,Cdr) ;
	%--------------------------------
	%true ->
	    write('Error 4 in generate_non_unit_lemma !!'),nl.


find_last([],[]) :- !, nl,write('Error: find_last !!'),nl.
find_last([X],X) :- !.
find_last([_F|R],X) :- find_last(R,X).

%%%%%%%%%%%%%%%%%%
%
%  This routine partially depends on 
%  the special built-in preicates of SICStus V3 !!!!!!!!!
%
%%%%%%%%%%%%%%%%%%%

join_at_tail(Lite,SFlag,RL,IF,CList) :-
	CList == [] ->
	    write('Error 1 in join_at_tail !!'),nl ;
	var(CList) ->
	   %----------------------------
	   %   Notice !!!
	   %----------------------------
            inc_carray(33),
	    (SFlag == '@ground' -> inc_carray(81); true),
	    create_mutable(IF,Mutable),
	    CList = [(Lite,SFlag,RL,Mutable)|_TT];
	   %----------------------------
	CList = [(LL,_)|Cdr] ->
	    (LL == Lite ->
	           inc_carray(37);
	    %true ->
	           join_at_tail(Lite,SFlag,RL,IF,Cdr)
	    );
	%true ->
            !,write('Error 2 in join_at_tail !!'),nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%

process_inconsistency(GoNum,End,Kind,ID) :-
	GoNum \== 0 ->
	      %
	      write_result(unsatisfiable),
      	      %----------------------------
	      statistics(runtime,[ETime,_]),
	      call(start_time(proving,STime)),
	      call(start_time(preprocessing,X0)),
	      Time is (ETime - STime) / 1000.0 ,
	      TotalTime is (ETime - X0) / 1000.0 ,
	      nl,
	      write('CPU Time = '),write(Time),write(' sec'),nl,
	      write('Total CPU Time = '),write(TotalTime),write(' sec'),nl,
	      process_proof(End,MainProof,UsedLemma1),
	      %--------
              nl,nl,
	      write('Inconsistency was found between'),nl,nl,
	      %--------
              MainProof =[[_,_,Head,_,_]|_],
	      write('     '),write(Head), write(' and '),
	      %--------
	      (Kind == '@unit_repl' ->
	           AbsIII is abs(ID),
	           write('Unit clause No.'),write(AbsIII),write(' !!'),
		   nl,nl ;
	       Kind == '@lemma_repl' ->
	           %%%%%%%%%%%
	           %
	           call(lemma_proof(ID,LemProof1)),
		   process_proof(LemProof1,LemmaProof,UsedLemma2),
		   LemmaProof = [[_,_,LemmaHead,_,_]|_],
		   %
	           %%%%%%%%%%%
		   write('Unit lemma No.'),write(ID),write(': '),
		   write(LemmaHead),
		   nl,nl ;
	       %true ->
	           nl,write('Error 0 in process_inconsistency !!!'),nl
	      ),
	      %----------------------------
              (write_proof_mode(on) ->
	           write_sub_proof('subgoal',Head,MainProof),
		   nl,
	           (write_lemma_proof_mode(on) ->
	                write_lemma_proof(UsedLemma1),nl,
			nl ;
	           %true ->
                        true 
		   ),
		   nl,
	           (Kind == '@unit_repl' ->
		        write('See the above unit axiom No.'),
			AbsID is abs(ID),
		        write(AbsID),write(' !!!'),nl,
			nl ;
		    Kind == '@lemma_repl' ->
	                write_sub_proof(ID,LemmaHead,LemmaProof),
			nl,
	                (write_lemma_proof_mode(on) ->
	                     write_lemma_proof(UsedLemma2),
			     nl ;
	                %true ->
                             true
			) ;
	            %true ->
	               nl,write('Error 1 in process_inconsistency !!!'),nl
		   ) ;
               write_proof_mode(off) ->
	           true ;
              %true ->
	           nl,write('Error 2 in process_inconsistency !!!'),nl
	      ),
	      %-----------------------
	      %   Notice !!
	      %-----------------------
	      raise_exception('@lemma_inconsistent') ;
	%true ->
              true.

write_sub_proof(_,_,Proof) :- 
	Proof=='@No_proof_found',
	!,
	nl,write('No sub proof was found in inconsistency !!'),nl,
	nl.
write_sub_proof(ID,Head,Proof) :-
	write('Sub Proof: '),write(ID),write(', '),write(Head),nl,
	write('Goal#        Wff#      Wff Instance'),nl,
	write('-----     --------   ------------------'),
	\+ \+ (lettervars(Proof),write_proof1(Proof)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%    Subroutines for "anti_lemmas"
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

anti_lemma_assert(AntiL,SFlag) :-
	inc_carray(24),
	anti_lemma_backward_subsumption(AntiL,SFlag),
	asserta(AntiL).

%%%%%%%%%%%%%%%%%%%%%

anti_lemma_backward_subsumption(AntiL,SFlag) :-
	(SFlag == '@ground' ->
	        inc_carray(50) ;
	 anti_lemma_backward_subsumption_flag(on) ->
                functor(AntiL,F,A),functor(Term,F,A),
		arg(1,AntiL,AntiNum),arg(1,Term,AntiNum),
		( clause(Term,true,Ref),
		  subsumes_chk(AntiL,Term),
		  erase(Ref),
		  inc_carray(22),
		  fail
		;
		  true
		) ;
	 %true ->
	        true
	),
	!.

%%%%%%%%%%%%%%%%%%%%%

clear_anti_lemma(AntiLite) :-
	retract(AntiLite),
	fail.
clear_anti_lemma(_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%   flist_instantiation_check  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

flist_inst_chk(X) :- 
	var(X),!,
	nl,write('Error in flist_inst_chk !!'),nl,fail.

flist_inst_chk([]) :- 
	!,
	inc_carray(21).

flist_inst_chk([(_,F1)|Cdr]) :- 
	F1 =='@no_inst' ->
	     flist_inst_chk(Cdr) ;
	F1=='@done_inst' ->
	     !,fail ;
	%true ->
	   write('Error 2 in flist_inst_chk !!'),nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%
%%%%   Subroutines for groundrizing and term check
%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

make_ground_and_check_term_at_tail(Lite,DFlag,SFlag) :-
	term_depth_bound_runtime_data(TDepth),
	numbervars_termdepth(Lite,TDepth,DFlag,SFlag),
	(SFlag == '@ground' ->    inc_carray(46);
	 SFlag == '@linear' ->    inc_carray(40);
	 SFlag == '@non_linear' -> inc_carray(41);
	 %true ->
	    nl,write('Error in make_ground_and_check_term_at_tail !!'),nl,nl
	).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%  numbervars_and term depth check
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

numbervars_termdepth(Term,Limit,DFlag,SFlag) :- 
	numbervars_tdepth(Term,0,Num,0,Limit,DFlag,TT),!,
	(var(DFlag) -> 
	         DFlag = '@small'; true
	),
	(Num == 0 -> 
	         SFlag = '@ground';
	 var(TT) ->
                 SFlag = '@linear';
	 TT == '@non_linear' ->
	         SFlag = '@non_linear'
	).

numbervars_tdepth('@VAR'(N),N,N1,_,_,_,_):-
	!,N1 is N+1.

numbervars_tdepth(X,N,N,_,_,_,TT) :-
 	( atom(X),! ; X='@VAR'(_),!,TT='@non_linear').

numbervars_tdepth(Term,N1,N2,Nest,Limit,DFlag,TT) :-
          (Limit > Nest -> true; DFlag = '@large'),
           functor(Term,_,S),
	   Nest1 is Nest + 1,
	   numbervars_tdepth_args(S,Term,N1,N2,Nest1,Limit,DFlag,TT).

numbervars_tdepth_args(0,_,N1,N1,_,_,_,_):- !.
numbervars_tdepth_args(S,Term,N1,N3,Nest,Limit,DFlag,TT) :-
	arg(S,Term,Arg),
        numbervars_tdepth(Arg,N1,N2,Nest,Limit,DFlag,TT),
	S1 is S-1, numbervars_tdepth_args(S1,Term,N2,N3,Nest,Limit,DFlag,TT).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%  f_numbervars
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

f_numbervars(T) :- f_numbervars(T,0,_),!.

f_numbervars('@VAR'(N),N,N1) :- 
	!, N1 is N+1.
f_numbervars(Term,N1,N2) :-
	functor(Term,_,S),f_numbervars(S,Term,N1,N2).

f_numbervars(0,_,N1,N1):- !.
f_numbervars(S,Term,N1,N3) :-
	arg(S,Term,Arg),f_numbervars(Arg,N1,N2),
	S1 is S-1, f_numbervars(S1,Term,N2,N3).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%
%%%                   start !!!
%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

:- dynamic user:prove_this/0.

prove_this_in_time_limit :-
	prolog_flag(unknown,Current,fail),
	(
	    call(usr_set_data(time_limit,LTime)),!, LimitTime is LTime*1000 ;
	    LimitTime = 3600000
	),
	prolog_flag(unknown,_,Current),
	%
	TTT is LimitTime // 1000.0,
        nl,write('Time limit '),write(TTT),write(' sec'),nl,nl,
	%
        time_out(user:prove_this,LimitTime,Result),
	%
	( Result == time_out -> 
	       T1 is LimitTime // 1000.0 ,
	       write_time_out,
	       write('Time limit '),write(T1),write(' sec has arrived just now ...'),nl,
	       nl,
	       write_counter,
	       delete_tmp_files ;
	  Result == success ->
	       true ;
	%true ->
	       nl,write('Error 1: Time_out routine !!'),nl
	).

prove(query) :- 
	prolog_flag(unknown,Current,error),
	%
	reset_carray, reset_all_flag_number,reset_all_flag_atom,
	reset_all_lemma_number,	reset_anti,
	initialize_lemma, initialize_anti_lemma,
	abolish_lemma_proof,
	%
	write('Proving started ...'),nl,
	on_exception('@lemma_inconsistent',
	             limiting_execution_time,exception_handler),
	%
	write_counter,listing_lemma,
	%
        abolish_lemma, abolish_anti_lemma,
	abolish_lemma_proof,
	%
	prolog_flag(unknown,_,Current).

exception_handler :- true.

limiting_execution_time :-
	batch_flag(on) ->	
              statistics(runtime,[X1,_]),
	      retractall(start_time(proving,_)),
	      assert(start_time(proving,X1)),
	      %
	      start(0,Sol),
	      %
	      statistics(runtime,[X2,_]),
	      call(start_time(preprocessing,X0)),
	      ExecTime is (X2 - X1) / 1000.0 ,
	      TotalTime is (X2 - X0) / 1000.0 ,
	      nl,
	      write('CPU Time for proving= '),write(ExecTime),write(' sec'),nl,
	      write('Total CPU Time = '),write(TotalTime),write(' sec'),nl,
	      %
              write_solution(Sol) ;
	      %
	batch_flag(off) ->
	      prolog_flag(unknown,Current,fail),
  	      (call(time_limit_data(LimitTime)),! ; LimitTime = 3600000),
	      prolog_flag(unknown,_,Current),
	      %
              statistics(runtime,[X1,_]),
	      retractall(start_time(proving,_)),
	      assert(start_time(proving,X1)),
	      %
              time_out(start(0,Sol),LimitTime,Result),
	      %
	      ( Result == time_out -> 
		      T1 is LimitTime // 1000.0 ,
		      write_time_out,
		      write('Time limit '),write(T1),write(' sec has arrived just now ...'),nl,
		      nl ;
		Result == success ->
		      statistics(runtime,[X2,_]),
		      call(start_time(preprocessing,X0)),
		      ExecTime is (X2 - X1) / 1000.0 ,
		      TotalTime is (X2 - X0) / 1000.0 ,
		      nl,
		      write('CPU Time for proving = '),write(ExecTime),write(' sec'),nl,
		      write('Total CPU Time = '),write(TotalTime),write(' sec'),nl,
		      %
		      write_solution(Sol) ;
	       %true ->
		      nl,write('Error 2 : Time_out routine  !!'),nl
	      ).

%%%%%%%%%%%%%%%%%%%%%

prolog_handler(X) :-
	(X==30 ; X==24 ) ->
	   write_time_out,
	   write('Signal '),write(X),write(' has arrived just now.'),nl,
	   nl,
	   delete_tmp_files ;
        %true ->
	   nl,write('Error: Signal '),write(X),write(' has reached !!'),nl.

%%%%%%%%%%%%%%%%%%%%%

write_time_out :-
	nl,nl,write('## No solution found ... time out !!'),nl.

write_result(X) :-
	X==unsatisfiable ->
	     nl,nl,write('## A proof exists !!'),nl,nl ;
	X==satisfiable ->
	     nl,nl,write('## No proof exists ... satisfiable !!'),nl,nl.


%%%%%%%%%%%%%%%%%%%%%

initialize_lemma :-
	abolish_lemma,
	make_initial_lemma.

abolish_lemma :-  
	retractall(lemma_attrib(_,_,_)),
	(unit_lemmaizing_flag(off) ->
	      !,true ;
	 unit_lemmaizing_flag(on) ->
	      lemma_name_runtime_data(X,_),
	      assert(X),  %%  a dummy for escaping an error at the next "abolish"
	      abolish(X),
	      fail
	).
abolish_lemma.

make_initial_lemma :-
	unit_lemmaizing_flag(off) ->
	    !,true ;
	unit_lemmaizing_flag(on) ->
	    lemma_name_runtime_data(Name,Arity),
	    (unit_lemma_resolution_flag(on) ->
	          Arity2 is Arity + 2 ;
	    %true ->
	          Arity2 is Arity + 1
	    ),
	    (unit_lemma_resolution_flag(on) ->
	         assertz(lemma_attrib(Name,Arity2,'@ground')); true),
	    functor(Dummy_lemma,Name,Arity2),
	    arg(Arity2,Dummy_lemma,'@initial_dummy'),
	    g_numbervars(Dummy_lemma),
	    assertz(Dummy_lemma),
	    fail.
make_initial_lemma.

abolish_lemma_proof :-
	retractall(lemma_proof(_,_)).

listing_lemma :-
	(unit_lemmaizing_flag(off); write_lemma_mode(off)) ->
	     !,nl ;
	unit_lemmaizing_flag(on),write_lemma_mode(on) ->
	     number_of_lemma_data(Num),
      	     (Num =< 0 ->
      		    !,true ;
      	      % true ->
      		     write(' ********* Lemma table ***********'),nl,
      		     lemma_name_runtime_data(Name,Arity),
      		     (unit_lemma_resolution_flag(on) ->
		           Arity2 is Arity + 2 ;
      		      %true ->
      			   Arity2 is Arity + 1
		     ),
      		     functor(LemTemp,Name,Arity2),
      		     reset_tmp_number,
      		     listing_lemma(LemTemp,Num,Arity2),
      		     fail
	     );
	%true ->
	     write('Error in listing_lemma_mode !!!!'),nl.
listing_lemma.

listing_lemma(LemTemp,N,Last) :- 
	arg(Last,LemTemp,LastArg),
	(\+ \+ (call(LemTemp), LastArg \== '@initial_dummy') ->
	     call(LemTemp), 
	     LastArg \== '@initial_dummy',
	     write_lettervars(LemTemp),nl,
	     inc_tmp_number,
	     call_tmp_number(Cur),
	     (Cur < N  -> fail ; !,nl) ;
        %true ->
	     !, true
	).
listing_lemma(_,_,_) :- nl.

%%%%%%%%%%%%%%%%%%%%%

initialize_anti_lemma :-
	abolish_anti_lemma,
	make_initial_anti_lemma.

abolish_anti_lemma :-  
	anti_lemmaizing_flag(off) ->
	    !,true ;
	anti_lemmaizing_flag(on) -> 
	    anti_lemma_name_runtime_data(X,_),
	    assert(X),
	    abolish(X),
	    fail.

abolish_anti_lemma.

make_initial_anti_lemma :-
	anti_lemmaizing_flag(off) ->
	    !,true ;
	anti_lemmaizing_flag(on) -> 
	    anti_lemma_name_runtime_data(Name,Arity),
	    Arity1 is Arity + 1,
	    functor(Dumy_anti_lemma,Name,Arity1),
	    arg(1,Dumy_anti_lemma,'@dummyNum'),
	    g_numbervars(Dumy_anti_lemma),
	    assertz(Dumy_anti_lemma),
	    fail.
make_initial_anti_lemma.

listing_anti_lemma :-
	anti_lemmaizing_flag(off) ->
	    !,true ;
	anti_lemmaizing_flag(on) -> 
	    anti_lemma_name_runtime_data(X,_),
	    listing(X),
	    fail.
listing_anti_lemma.

%%%%%%%%%%%%%%%%%%%%%

g_numbervars(T) :- g_numbervars(T,0,_),!.

g_numbervars('''@dummy_arg'''(N),N,N1) :- 
	!, N1 is N+1.
g_numbervars(Term,N1,N2) :-
	functor(Term,_,S),g_numbervars(S,Term,N1,N2).

g_numbervars(0,_,N1,N1):- !.
g_numbervars(S,Term,N1,N3) :-
	arg(S,Term,Arg),g_numbervars(Arg,N1,N2),
	S1 is S-1, g_numbervars(S1,Term,N2,N3).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%    write_solution for 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%% Proof Printing.
%%%
%%% Add extra arguments to each goal so that information
%%% on what inferences were made in the proof can be printed
%%% at the end.

write_solution(Proof) :-
	write_proof_mode(on) ->
	     Proof=Proof1-_End,
	     process_proof(Proof1,MainProof,UsedLemma),
	     nl,
	     write_main_proof(MainProof),
	     nl,
	     (write_lemma_proof_mode(on) ->
	         write_lemma_proof(UsedLemma),
		 nl ;
	     %true ->
                 true 
	     ) ;
	write_proof_mode(off) ->
	     true ;
	%treu ->
	     write('Error in write_proof_mode !!!'),nl.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

write_main_proof(Proof) :- 
	Proof=='@No_proof_found',
	!.
write_main_proof(Proof) :-
	nl,
	write('Main Proof:'),nl,
	write('Goal#        Wff#      Wff Instance'),nl,
	write('-----     --------   ------------------'),
	\+ \+ (lettervars(Proof),write_proof1(Proof)).

write_proof1([]).
write_proof1([[LineNum,X,Head,Depth,Subgoals]|Y]) :-
	nl,
	write_indent_for_number(LineNum),
	write('['),write(LineNum),write(']  '),
	write_indent_for_num_and_op(X),
	write(X),write('   '),
	write_proof_indent(Depth),
	write(Head),
	(Subgoals = [] ->
		true;
	%true ->
		write(' :- '),
		write_proof_subgoals(Subgoals)),
	write(.),
	write_proof1(Y).	

write_indent_for_number(N) :-
	((number(N) , N <  100) -> write(' ') ; true),
	((number(N) , N <   10) -> write(' ') ; true).

write_indent_for_num_and_op(X) :-
	number(X) ->
	        write('       '),   % X 4ʲȲ
		(X < 1000 -> write(' ') ; true),
		(X < 100  -> write(' ') ; true),
		(X < 10   -> write(' ') ; true);
	(X = u_mat(N) ; X = l_mat(N) ; 
	    X = l_res(N) ;  X = trunc(N) ; 
	    X = tr_AR(N) ;  X = tr_UR(N) ; X = tr_LR(N)) ->
		(N < 1000 -> write(' ') ; true),
		(N < 100  -> write(' ') ; true),
		(N < 10   -> write(' ') ; true);
	(X = tr_AC(N) ;  X = tr_CR(N)) ->
	        (N == '@top' ->
		       true ;
		 %true ->    
		    (N < 1000 -> write(' ') ; true),
		    (N < 100  -> write(' ') ; true),
		    (N < 10   -> write(' ') ; true)
		) ;
	%true ->
	        name(X,SList),length(SList,L),
		BN is 11-L,        % 11ǥ
		write_blank(BN).

write_blank(BN) :- 
	BN =:= 0 -> 
	        true ;
	%true ->
	   write(' '),BN1 is BN-1,
	   write_blank(BN1).
 
write_proof_subgoals([X,Y|Z]) :-
	write('['),
	write(X),
	write('] , '),
	write_proof_subgoals([Y|Z]).
write_proof_subgoals([X]) :-
	write('['),
	write(X),
	write(']').

write_proof_indent(N) :-
	N < 0 ->
	    nl,write('Error in Indent !!!!'),nl ;
	N == 0 ->
	    true ;
	N > 0  ->
	    write('   '),
	    N1 is N - 1,
	    write_proof_indent(N1).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

process_proof(X,MainProof,UsedLemma) :-
	var(X), !,
	MainProof='@No_proof_found', UsedLemma=[].

process_proof([],MainProof,UsedLemma) :-
	!,
	nl,write('Error:  NIL Proof found !!'),nl,
	MainProof=[], UsedLemma=[].

process_proof(Proof1,MainProof,UsedLemma) :-
	Proof1=[(_,TTHead,Ancestors)|_],
	(TTHead == query ->
	      Ind0 is 0 ;
	 TTHead == '@non_unit' ->
	      length(Ancestors,X),
	      Ind0 is X -1 ;
%	      Ind0 is X ;
	 %true ->
              nl,write('Error in process_proof 1 !!!'),nl
	),
	process_and_check_proof(Proof1,0,Ind0,MainProof,UsedLemma).

process_and_check_proof(X,_LineNum,_Indent,[],[]) :-
	var(X),!.

process_and_check_proof([(X,TTHead,TTAncestors)|Y],
	                LineNum,Indent,MainProof,UsedLemma) :-
	LineNum1 is LineNum + 1,
	process_and_check_proof(Y,LineNum1,Indent,MP,UL),
	(TTHead == '@non_unit' ->
	      TTAncestors=[(_,Head,_)|PosAncestors] ;
	%true ->
              Head = TTHead,
	      PosAncestors = TTAncestors
	),
%%%%%%%%%%%%%%%%%%%%%%%
% compute indentation to show level of goal nesting
%%%%%%%%%%%%%%%%%%%%%%%
	(Head == query ->
		Depth is 0 ;
	 %true ->                                
	        length(PosAncestors,LL),
		Depth is LL - Indent
	),                  		 
	Depth1 is Depth + 1,
	collect_proof_subgoals(Depth1,MP,Subgoals),
	(X = red_n ->
		X1 = red,
		add_string_to_literal(Head,Head1,"not_");
	 X = id_red_n ->
	        X1 = id_red,
		add_string_to_literal(Head,Head1,"not_");
	 X = c_red_n ->
		X1 = c_red,
		add_string_to_literal(Head,Head1,"not_");
	 X = id_c_red_n ->
		X1 = id_c_red,
		add_string_to_literal(Head,Head1,"not_");
	 %--------------------
	 X = trun(N) ->
	        classify(N,X1,LemID),
		Head1 = Head;
	 X = trun_n(N) ->
	        classify(N,X1,LemID),
		add_string_to_literal(Head,Head1,"not_");
	 %--------------------
	 X = u_match(ClauseID) ->
	        (ClauseID < 0 -> NewID is - ClauseID ; NewID = ClauseID),
		X1 = u_mat(NewID),
		add_string_to_literal(Head,Head2,"unit_"),
	        delete_last_arg(Head2,Head1);
	 X = l_match(N) ->
		LemID = N, X1 = l_mat(N), 
		add_string_to_literal(Head,Head2,"lemma_"),
	        delete_last_arg(Head2,Head1);
	 X = l_res(N) ->
		LemID = N, X1 = l_res(N),
		Head1 = Head ;
	 X = l_res_n(N) ->
		LemID = N, X1 = l_res(N),
		add_string_to_literal(Head,Head1,"not_") ;
	 (number(X), X < 0) ->
		X1 is - X,
		add_string_to_literal(Head,Head1,"not_");
	%true ->
		X1 = X,	Head1 = Head
	),
	MainProof = [[LineNum,X1,Head1,Depth,Subgoals]|MP],
	(integer(LemID) ->
	       UsedLemma = [LemID|UL] ;
	%true ->
	       UsedLemma = UL
	).

collect_proof_subgoals(_Depth1,[],Result) :-
	Result = [].
collect_proof_subgoals(Depth1,[[LineNum,_,_,Depth,_]|P],Result) :-
	Depth = Depth1,
	collect_proof_subgoals(Depth1,P,R),
	Result = [LineNum|R].
collect_proof_subgoals(Depth1,[[_,_,_,Depth,_]|P],Result) :-
	Depth > Depth1,
	collect_proof_subgoals(Depth1,P,Result).
collect_proof_subgoals(Depth1,[[_,_,_,Depth,_]|_],Result) :-
	Depth < Depth1,
	Result = [].

classify(N,X,LemID) :-
	number(N) -> 
	    X = trunc(N), LemID = N  ;
	N = cLit(Id,_RList,_IFList) ->
	    X = tr_AC(Id), LemID = '@not_lemma' ;
	N = '@reduc_match'(Id) -> 
	    X = tr_AR(Id), LemID = Id  ;
	%---------------------------
	N = '@descend'(Flag) ->
	    (Flag =  '@reduc_match'(TId) -> 
	         X = tr_AR(TId), LemID = TId ;
	     Flag = '@unit_repl'(TId,_,_) ->
	         (TId < 0 -> NewId is -TId ; NewId = TId),
	         X = tr_UR(NewId), LemID = '@not_lemma' ;
	     Flag = '@lemma_repl'(TId,_,_) ->
	         X = tr_LR(TId), LemID = TId ;
	     Flag = '@c_lit_repl'(TId,_,_) ->
	         X = tr_CR(TId), LemID = '@not_lemma';
	     %true ->
		 nl,write('Error 1 in classify !!'),nl
	    ) ;
	%---------------------------
	N = '@unit_repl'(Id,_RList,_IFList) ->
	    (Id < 0 -> NewId is -Id ; NewId = Id),
	    X = tr_UR(NewId), LemID = '@not_lemma' ;
	N = '@lemma_repl'(Id,_RList,_IFList) ->
	    X = tr_LR(Id), LemID = Id ;
	N = '@c_lit_repl'(Id,_RList,_IFList) ->
	    X = tr_CR(Id), LemID = '@not_lemma';
	%true ->
            nl,write('Error 0 in classify !!'),nl,
	    write((form_is,N)).

reverse(In,Out) :-qrev(In,[],Out).
qrev([],Out,Out) :-!.
qrev([X|Y],Imm,Out) :- qrev(Y,[X|Imm],Out).

delete_last_arg(Lite1,Lite2) :-
	Lite1 =.. [F|Args1],
	reverse(Args1,[_|RevCdr]),
	reverse(RevCdr,Args2),
	Lite2 =.. [F|Args2].

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

write_lemma_proof(UsedLemma) :-
	write_lemma_proof1(UsedLemma,[],_).

write_lemma_proof1([],X,X) :-!.
write_lemma_proof1([CarLem|CdrLem],P1,P2) :-
	(simple_member(CarLem,P1) ->
	     write_lemma_proof1(CdrLem,P1,P2) ;
	%true ->
	     call(lemma_proof(CarLem,Proof1)),
	     process_proof(Proof1,MainProof,UsedLemma),
	     nl,
	     nl,
	     write('Proof of Lemma '),write(CarLem),write(':'),nl,
	     write('-----     --------   ------------------'),
	     \+ \+ (lettervars(MainProof),write_proof1(MainProof)),
	     (UsedLemma == [] ->
	          P3 = [CarLem|P1] ;
	     %true ->
		  write_lemma_proof1(UsedLemma,[CarLem|P1],P3)
	     ),
	     write_lemma_proof1(CdrLem,P3,P2)
	).

simple_member(X,[Car|_]) :- X == Car,!.
simple_member(X,[_|Cdr]) :- simple_member(X,Cdr).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

write_intermediate_counter(N) :-
	call_carray(15,X15),call_carray(8,X8),call_carray(1,X1),call_carray(2,X2),
	call_carray(0,X0),call_carray(5,X5),call_carray(12,X12),call_carray(16,X16),
	call_carray(29,X29),call_carray(32,X32),call_carray(34,X34),
	call_carray(42,X42),call_carray(43,X43),call_carray(25,X25),
	call_carray(30,X30),call_carray(53,X53),call_carray(59,X59),
	Inference is X8+X1+X2+X15+X0+X5+X12+X16+X29+X32+X34+X42+X43+X25+X30+X53+X59,
	nl,
	write(N),write(' ...... '),write(Inference),write(' inferences').

write_counter :- nl,
  call_carray(0,X0),  call_carray(1,X1),  call_carray(2,X2),
  call_carray(3,X3),  call_carray(4,X4),  call_carray(5,X5),
  call_carray(6,X6),  call_carray(7,X7),  call_carray(8,X8),
  call_carray(9,X9),  call_carray(10,X10), call_carray(11,X11),  
  call_carray(12,X12), call_carray(13,X13), call_carray(14,X14),
  call_carray(15,X15), call_carray(16,X16), call_carray(17,X17),
  call_carray(18,X18), call_carray(19,X19), call_carray(20,X20),
  call_carray(21,X21), call_carray(22,X22), call_carray(23,X23),
  call_carray(24,X24), call_carray(25,X25), call_carray(26,X26), 
  call_carray(27,X27), call_carray(28,X28), call_carray(29,X29),
  call_carray(30,X30), call_carray(31,X31), call_carray(32,X32),
  call_carray(33,X33), call_carray(34,X34), call_carray(35,X35),
  call_carray(36,X36), call_carray(37,X37), call_carray(38,X38),
  call_carray(39,X39), call_carray(40,X40), call_carray(41,X41),
  call_carray(42,X42), call_carray(43,X43), call_carray(44,X44), 
  call_carray(45,X45), call_carray(46,X46), call_carray(47,X47),
  call_carray(48,X48), call_carray(49,X49), call_carray(50,X50),
  call_carray(51,X51), call_carray(52,X52), call_carray(53,X53),
  call_carray(54,X54), call_carray(55,X55), call_carray(56,X56), 
  call_carray(57,X57), call_carray(58,X58), call_carray(59,X59),
  call_carray(60,X60), call_carray(61,X61),call_carray(61,X62),
  call_carray(63,X63), call_carray(64,X64), call_carray(65,X65),
  call_carray(66,X66), call_carray(67,X67),  call_carray(68,X68),
  call_carray(69,X69), call_carray(70,X70), call_carray(71,X71),
  call_carray(72,X72), call_carray(73,X73), call_carray(74,X74), 
  call_carray(75,X75), call_carray(76,X76), call_carray(77,X77), 
  call_carray(78,X78), call_carray(79,X79), call_carray(80,X80),
  call_carray(81,X81), call_carray(82,X82),
  %
  %
  call_lemma_total_number(LemNum),
  %
  Inference is X8+X1+X2+X15+X0+X5+X12+X16+X29+X32+
               X34+X42+X43+X25+X30+X53+X59, 
  %
  %
  write(' ********* ME inference **********'),nl,
  write(' Total inferences = '),write(Inference),nl,
  RedNum is X1 + X34,

  write(' Reductions = '),write(RedNum),nl,
  write('  (With no instantiation = '),write(X34),write(')'),nl,
  write(' Identical reductions = '),write(X2),nl,
  Ext_unit is X15 + X42,
  write(' Expansions with units = '),write(Ext_unit),nl,
  write('  (With linear unification = '),write(X42),write(')'),nl,
  Ext_non_unit is X8 + X43,
  write(' Expansions with non-units = '),write(Ext_non_unit),nl,
  write('  (With linear unification = '),write(X43),write(')'),nl,

  write(' Matchings with units = '),write(X0),nl,
  write(' Matchings with unit-lemmas = '),write(X5),nl,
  Ancestor_match is X12 + X63, 
  write(' Ancestor unit-lemma matchings = '),write(Ancestor_match),nl,
  write('  (for A-lit. in reduction = '),write(X63),write(')'),nl,
  write(' Ancestor C-literal matchings = '),write(X59),nl,

  RS is X72 + X74 +X65,
  LS is X73 + X75 +X69,
  CS is X76 + X77, 
  write(' Unit replacement subsumption = '),write(RS),nl,
  write('  (in non-units expan. = '),write(X72),write(')'),nl,
  write('  (in C-reduction. = '),write(X74),write(')'),nl,
  write('  (using unification with unit lemmas = '),write(X65),write(')'),nl,
  write(' Lemma replacement subsumption = '),write(LS),nl,
  write('  (in non-units expan. = '),write(X73),write(')'),nl,
  write('  (in C-reduction = '),write(X75),write(')'),nl,
  write('  (using unification with unit lemmas = '),write(X69),write(')'),nl,
  write(' C-literal replacement subsumption = '),write(CS),nl,
  write('  (in non-units expan. = '),write(X76),write(')'),nl,
  write('  (in C-reduction. = '),write(X77),write(')'),nl,
  write(' (++ Truncated goals = '),write(X13),write(' + alpha .. ++)'),nl,

  Lem_Res_Num is X16 + X29 + X32,
  write(' Unit-lemma resolutions with unification = '),write(Lem_Res_Num),nl,
  write('  (Ground resolutions = '),write(X29),write(')'),nl,
  write('  (Linear resolutions = '),write(X32),write(')'),nl,

  write(' Identical reductions with C-literals = '),write(X25),nl,
  RedCLit is X30 + X53,
  write(' Reductions with C-literals = '),write(RedCLit),nl,
  write('  (With no instantiation = '),write(X53),write(')'),nl,

  write(' ********* Detection ***************'),nl,
  write(' At head:'),nl,
  write('   Ground goals = '),write(X62),nl,
  write('   Linear goals = '),write(X31),nl,
  write('   Non-linear goals = '),write(X44),nl,
  write('   Goals not examined rigidly = '),write(X49),nl,
  write('   Goals not unifiable with units= '),write(X47),nl,
  write(' At tail:'),nl,
  write('   Ground goals = '), write(X46),nl,
  write('   Linear goals = '), write(X40),nl,
  write('   Non-linear goals = '), write(X41),nl,
  write(' Non-subsumed goals pre-detected dynamically = '),write(X36),nl,
  write(' Checked local instantiations into goals = '),write(X21),nl,

  write(' ********* Pruning ***************'),nl,
  NumIdent is X3,
  write(' Failures by nonidentical members = '),write(NumIdent),nl,

  TS_Id is X57 + X67  + X78 + X80 + X82,
  write(' Failures by identical A-literals at head= '),write(X55),nl,
  write(' Failures by identical A-literals at tail= '),write(TS_Id),nl,
  write('    (# Failures of unit_expan.= '),write(X82),write(')'),nl,
  write('    (# Failures of non_unit_expan.= '),write(X57),write(')'),nl,
  write('    (# Failures of reduction= '),write(X80),write(')'),nl,
  write('    (# Failures of C-reduction= '),write(X78),write(')'),nl,
  write('    (# Failures of lemma resol.= '),write(X67),write(')'),nl,


  TS_Com is X58 + X68 +X79,
  write(' Failures by complemented A-literals at head = '),write(X56),nl,
  write(' Failures by complemented A-literals at tail = '),write(TS_Com),nl,
  write('    (# Failures of non_unit_expan.= '),write(X58),write(')'),nl,
  write('    (# Failures of C-reduction= '),write(X79),write(')'),nl,
  write('    (# Failures of lemma resol.= '),write(X68),write(')'),nl,

  HS_Unit is X4,
  TS_Unit is X10 + X38 + X26 + X51 + X66,
  Input_Unit is HS_Unit + TS_Unit + X39,
  write(' Input clauses subsumed by Units = '),write(Input_Unit),nl,
  write('  (Subsumptions in Units = '),write(X39),write(')'),nl,
  write('  (Head subsumptions = '),write(HS_Unit),write(')'),nl,
  write('  (Tail subsumptions = '),write(TS_Unit),write(')'),nl,
  write('   (# Subsumptions of non_unit_expan. = '),write(X10),write(')'),nl,
  write('   (# Subsumptions of reduc. at goals = '),write(X38),write(')'),nl,
  write('   (# Subsumptions of reduc. at A-lit. = '),write(X66),write(')'),nl,
  write('   (# Subsumptions of C-reduction = '),write(X26),write(')'),nl,
  write('   (# Subsumptions of lemma resol. = '),write(X51),write(')'),nl,

  HS_Lemma is X17,
  TS_Lemma is X18 + X61 + X52 + X71,
  Input_Lemma is HS_Lemma + TS_Lemma,
  write(' Input clauses subsumed by Lemmas = '),write(Input_Lemma),nl,
  write('  (Head subsumptions  = '),write(HS_Lemma),write(')'),nl,
  write('  (Tail subsumptions  = '),write(TS_Lemma),write(')'),nl,
  write('   (# Subsumptions of non_unit_expan. = '),write(X18),write(')'),nl,
  write('   (# Subsumptions of reduc. at goals = '),write(X52),write(')'),nl,
  write('   (# Subsumptions of reduc. at A-lit, = '),write(X71),write(')'),nl,
  write('   (# Subsumptions of C-reduc. = '),write(X61),write(')'),nl,

  NumCut is X23 + X45,
  write(' Cuts caused by no instantiation = '),write(NumCut),nl, 
  write('  (Cuts of ground goals = '),write(X45),write(')'),nl,  

  Head_Anti_Fail is X48,
  Tail_Anti_Fail is X19 + X20 + X60 + X27 + X70,
  Anti_Fail is Head_Anti_Fail + Tail_Anti_Fail,
  write(' Failures forced by anti-lemmas = '),write(Anti_Fail),nl,
  write('  (Failures at heads = '),write(Head_Anti_Fail),write(')'),nl,
  write('  (Failures at tails = '),write(Tail_Anti_Fail),write(')'),nl,
  write('   (# Failures of non_unit_expan. ='),write(X20),write(')'),nl,
  write('   (# Failures of reduc. at goals ='),write(X19),write(')'),nl,
  write('   (# Failures of reduc. at A-lit.='),write(X70),write(')'),nl,
  write('   (# Failures of C-reduc. ='),write(X60),write(')'),nl,
  write('   (# Failures of lemma resol. ='),write(X27),write(')'),nl,
  write(' ******** Anti Lemmas *************'),nl,

  AntiLemNum is X24,
  write(' Stored Anti-lemmas = '),write(AntiLemNum),nl,
  write('  (Ground anti-lemmas  = '),write(X50),write(')'),nl,
  write(' Anti-lemmas backward subsumed = '),write(X22),nl,
  write(' ********* Lemmas *****************'),nl,
  write(' Accepted Lemma Candidates = '),write(X6),nl,
  write('  (Rejected candidates in single bodies = '),write(X7),write(')'),nl,
  write(' Lemmas accepted by depth check = '),write(X9),nl,
  write(' Lemmas forwardly subsumed = '),write(X11),nl,
  write(' Stored unit lemmas = '),write(LemNum),nl,
  write('  (Ground unit lemmas = '),write(X28),write(')'),nl,
  write('  (Linear unit lemmas = '),write(X35),write(')'),nl,
  Non_line_lem is LemNum - X28 - X35,
  write('  (Non-linear lemmas = '),write(Non_line_lem),write(')'),nl,
  write('  (Unit lemmas backwardly subsumed = '),write(X14),write(')'),nl,
  write(' Stored C-lemmas = '),write(X33),nl,
  write('  (Ground C-lemmas = '),write(X81),write(')'),nl,
  write('  (Unit C-lemmas = '),write(X64),write(')'),nl,
  write('  (Rejected ones, identical to A-literals = '),write(X54),write(')'),nl,
  write('  (Rejected ones, identical to C-literals = '),write(X37),write(')'),nl.




