% I-THOP  (A first-order theorem prover based on PTTP technology)
% 
%      version 1.01
%      July 2, 1997
%      Copyright (C) 1997, Koji Iwanuma (iwanuma@esi.yamanashi.ac.jp)
% 
% This program is free software; you can redistribute it or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either versions 1, or (at your option)
% any later version.
%  
% This program is distributed in the hope that it will be useful but
% WITHOUT ANY WARRANTY.  See the GNU General Public License for more
% details.
% ---------------
%
%   comMODE.pl 
%
% ---------------

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%
%%%%%     Mode setting tools
%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

set_default_mode(S) :-
	%
	% Common setting
	%
        si_set(time_limit,3600),
        si_mode(delete_temporal_files,on),
        si_mode(write_proof,on),
	si_mode(write_lemma_proof,on),
        si_mode(write_lemma,off),
	si_set(number_of_lemma,20),
        si_mode(write_sorted_literal,off),
	%
        si_mode(lemma_backward_subsumption,off),
        si_mode(anti_lemma_backward_subsumption,off),
        %
        si_mode(unit_matching,on),
	si_mode(unit_lemma_matching,on),
	si_mode(unit_lemma_resolution,off),        
        %
        si_mode(pruning_after_unit_expan,on),
        si_mode(tail_pruning,on),
	%
        si_mode(unit_subsumption,on),
	si_mode(subsumption_in_units,on),
	si_mode(lemma_subsumption,off),
        %
        si_mode(anti_lemma_subsumption,on),
        %
        si_mode(sort_bliteral,auto),              %% (S8)
	%si_mode(sort_bliteral,[0,0,0,0,0,0]),%% off
        %si_mode(sort_bliteral,[0,1,4,0,0,1]).%% (S7) for Horn
	%si_mode(sort_bliteral,[0,1,4,2,1,0]),%% (S7) for Non-Horn
	%
	si_mode(sort_identical_value,on),
	si_mode(sort_clause,on),
	%
        (S == horn ->
              %
              si_set(depth_bound_measure,all), 
	      %
	      si_mode(single_body_lemmaizing,on),
	      %
	      si_mode(ancestor_unit_lemma_matching,on),
	      si_mode(ancestor_unit_lemma_matching_in_reduction,off), 
	      %
	      si_mode(unit_repl_subsumption,off),
	      si_mode(lemma_repl_subsumption,off),
	      si_mode(c_literal_repl_subsumption,off),
	      si_mode(last_goal_repl_subsumption,off),
	      si_mode(inconsistency_check,off),
	      %
	      si_mode(identical_c_reduction,off),  
	      si_mode(ancestor_c_literal_matching,off),
	      si_mode(full_c_reduction,off),        
	      %
	      si_set(head_pruning_position,nowhere),
	      %
	      % or si_set(head_pruning_position,only_last),
	      %
	      si_mode(cut_by_no_inst,off),
	      si_mode(ident_A_chk,on), %  Notice !!
	      si_mode(compl_A_chk,off)
	      ;
	%------------------------------
	 S == non_horn_existed ->
	      %
	      si_set(depth_bound_measure,sub), 
	      %
	      si_mode(single_body_lemmaizing,on),
	      %
	      si_mode(ancestor_unit_lemma_matching,on),  
	      si_mode(ancestor_unit_lemma_matching_in_reduction,on),
	      %----------------
	      si_mode(unit_repl_subsumption,on),
	      si_mode(lemma_repl_subsumption,on),
	      si_mode(c_literal_repl_subsumption,on),
	      si_mode(last_goal_repl_subsumption,off),
	      %----------------
	      %si_mode(inconsistency_check,off),
	      si_mode(inconsistency_check,on),
	      %
	      si_mode(identical_c_reduction,on),         % Notice !!!
	      si_mode(ancestor_c_literal_matching,off),
	      si_mode(full_c_reduction,on),       
	      %
	      si_set(head_pruning_position,everywhere),
	      %
	      % or si_set(head_pruning_position,except_for_first),
	      %
	      si_mode(anti_lemma_subsumption,on),
	      %
	      si_mode(cut_by_no_inst,on),
	      si_mode(ident_A_chk,on),
	      si_mode(compl_A_chk,on)
	).

si_mode(Name,F) :-
       (Name == unit_repl_subsumption ;
	Name == lemma_repl_subsumption ;
	Name == c_literal_repl_subsumption ;
	Name == last_goal_repl_subsumption ;
	%
	Name == single_body_lemmaizing ;
	Name == lemma_backward_subsumption ;
	Name == anti_lemma_backward_subsumption;
	%
        Name == unit_matching ;
	Name == unit_lemma_matching ;
	%
        Name == ancestor_unit_lemma_matching ;
        Name == ancestor_unit_lemma_matching_in_reduction ;
	%
        Name == unit_lemma_resolution ;
        %
	Name == identical_c_reduction ;
	Name == full_c_reduction ;
	Name == ancestor_c_literal_matching ;
	%
        Name == pruning_after_unit_expan ;
	Name == tail_pruning ;
	%
	Name == unit_subsumption ;
	Name == subsumption_in_units ;
	Name == lemma_subsumption ;
	%
	Name == ident_A_chk ;
	Name == compl_A_chk ;
	%
	%-------------------------
        Name == write_sorted_literal ;
        Name == sort_identical_value ;
	Name == sort_clause ;
	%-------------------------
	%
        Name == write_proof ;
	Name == write_lemma_proof ;
	Name == write_lemma ;
        Name == delete_temporal_files ;
	%
        fail  %%
    ) ->
        ((F == on ; F == off) ->
            int_mode(Name,F) ;
	 %true ->
	    nl,write('Irregular Data in 2nd. arg in mode !'),nl,
	    fail
	) ;
    %true ->
	fail.
%=======================

si_mode(Name,F):- 
        (Name==inconsistency_check ;
	 Name==anti_lemma_subsumption ;
	 Name==cut_by_no_inst
	) ->
	     (F == on ->
	 	   (Name==inconsistency_check  ->
		      si_mode(unit_repl_subsumption,on),
		      si_mode(lemma_repl_subsumption,on) ;
	            Name==anti_lemma_subsumption ->
                      si_mode(unit_subsumption,on) ;
	            Name==cut_by_no_inst ->
	              si_mode(unit_matching,on) ;
		   % true ->
            	      nl,write('Irregular Data in 2nd si_mode'),nl
		   ) ;
	      F == off ->
                   true ;
	      %true ->
	           nl,write('Irregular Data in 2nd. arg in mode !'),nl
	     ),
	     int_mode(Name,F) ;
	% true ->
             fail.

%-------------------------
si_mode(Name,F) :-
	 Name==sort_bliteral,F==auto ->
	     int_mode(Name, F) ;
         Name==sort_bliteral,F=[_X1,_X2,_X3,_X4,_X5,_X6] ->
             int_mode(Name,F) ;
	 Name==sort_bliteral,F==off ->
	     int_mode(Name, [0,0,0,0,0,0]) ;
	 % true ->
	    fail.

si_mode(Name,_) :-
	nl, write('Error in si_mode :  Mode Name is Irregular !! '),
	    write(Name),nl,fail.

%========================
si_set(Name,Data) :-
	(Name == depth_bound_measure,
	     (Data == all ; Data == sub) ;
	 %
         Name == head_pruning_position,
	      (Data == nowhere ;  Data == everywhere ; 
		  Data == only_first ;
		  Data == except_for_first ; Data == only_last) ;
	 %
	 Name == number_of_lemma ;
	 %
	 fail 
	) ->
	       int_set(Name,Data) ;
	%true ->
	       fail.
%------------------------

si_set(Name,X) :-
	Name==term_depth_bound ->
	      (X > 0 ->
	              int_set(term_depth_bound,X) ;
	       %true ->
	              nl,write('*** Term depth must be greater than 0 !! ***'),nl,
		      write('      set by the default value 3. !!'),nl,
		      int_set(term_depth_bound,3)
	      ) ;
	Name==time_limit ->
	      (integer(X) ->
	              Mil_sec is X * 1000,
		      int_set(time_limit,Mil_sec) ;
	      %true ->
	              write('****  The time-limit (sec) must be intger !!  ****'),nl,
		      write('     The  default time 3,600 sec has been set'),
		      int_set(time_limit,3600000)
	      ) ;
        %true ->
	      fail.

si_set(Name,_) :-
	nl, write('Error in si_set :  Data Name is Irregular !! '),
	    write(Name),nl,fail.

%========================
int_mode(X,F) :-
	add_string_to_name_at_tail(X,AName,"_mode"),
	Tmp =.. [AName,_],Atom =.. [AName,F],
	retractall(Tmp),assert(Atom).

int_set(X,F) :-
	add_string_to_name_at_tail(X,AName,"_data"),
	Tmp =.. [AName,_],Atom =.. [AName,F],
	retractall(Tmp),assert(Atom).

add_string_to_name_at_tail(Name,AddedName,String) :-
	name(Name,NString),
	append(NString,String,AddedNString),
	name(AddedName,AddedNString).

%------------------------------
listing_mode :-
	listing([
           single_body_lemmaizing_mode,
	   lemma_backward_subsumption_mode, 
	   anti_lemma_backward_subsumption_mode,
	   unit_matching_mode,
	   %
	   unit_lemma_matching_mode,
	   ancestor_unit_lemma_matching_mode,
	   ancestor_unit_lemma_matching_in_reduction_mode, 
	   unit_lemma_resolution_mode,
	   %
	   identical_c_reduction_mode, 
	   ancestor_c_literal_matching_mode, full_c_reduction_mode,
	   %
	   unit_repl_subsumption_mode, lemma_repl_subsumption_mode,
	   c_literal_repl_subsumption_mode,
	   last_goal_repl_subsumption_mode,
	   %
	   inconsistency_check_mode,
	   %
	   pruning_after_unit_expan_mode,
	   tail_pruning_mode,
           %
	   unit_subsumption_mode, subsumption_in_units_mode,
	   lemma_subsumption_mode, anti_lemma_subsumption_mode,
	   cut_by_no_inst_mode,
	   ident_A_chk_mode, compl_A_chk_mode,
	   %
	   %-------------------------------
	   sort_bliteral_mode,
	   write_sorted_literal_mode,
	   sort_identical_value_mode,
	   %
           sort_clause_mode,
	   %-------------------------------
	   %
	   write_proof_mode, write_lemma_proof,
	   write_lemma_mode
		]).

listing_set :-
	listing([
           depth_bound_measure_data, term_depth_bound_data,
	   head_pruning_position_data,
	   number_of_lemma_data, time_limit_data
		]).

%------------------------------
mode(X,Y) :- 
	retractall(usr_set_mode(X,_)),
	assert(usr_set_mode(X,Y)).
set(X,Y) :- 
	retractall(usr_set_data(X,_)),
	assert(usr_set_data(X,Y)).

set_user_mode_and_data :-
	%
	prolog_flag(unknown,Current,fail),
	%
	set_user_mode,
	set_user_data,
	%
	prolog_flag(unknown,_,Current).

set_user_mode :-
	call(usr_set_mode(X,Y)),
	si_mode_wrapper(X,Y),
	fail.
set_user_mode.
si_mode_wrapper(X,Y) :- si_mode(X,Y),!.

set_user_data :-
	call(usr_set_data(X,Y)),
	si_set_wrapper(X,Y),
	fail.
set_user_data.
si_set_wrapper(X,Y) :- si_set(X,Y),!.

listing_usr_mode :- 
	listing([usr_set_mode/2,usr_set_data/2]).