/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              DFG2TPTP                                  * */
/* *                                                        * */
/* *  $Module:    DFG2TPTP                                  * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.56 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/02 07:31:34 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: dfg2tptp.c,v $ */


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "dfgparser.h"

static void TPTPOutput(FILE* File, TERM Term);

void TPTPSymbolOutput(FILE* File, SYMBOL Symbol)
{
  if (symbol_Equal(Symbol,fol_All())) {
    fprintf(File,"\n!");
    return;
  }
  if (symbol_Equal(Symbol,fol_Exist())) {
    fprintf(File,"\n?");
    return;
  }
  if (symbol_Equal(Symbol,fol_Implies())) {
    fprintf(File,"=>\n");
    return;
  }
  if (symbol_Equal(Symbol,fol_Equiv())) {
    fprintf(File,"<=>\n");
    return;
  }
  if (symbol_Equal(Symbol,fol_Not())) {
    fprintf(File,"~");
    return;
  }
  if (symbol_Equal(Symbol,fol_And())) {
    fprintf(File,"&\n");
    return;
  }
  if (symbol_Equal(Symbol,fol_Or())) {
    fprintf(File,"|\n");
    return;
  }
  symbol_FPrint(File,Symbol);
  return;
}

void TPTPListOutput(FILE* File, LIST List)
/**************************************************************
  INPUT:   A list of terms.
  RETURNS: None.
  USAGE:   intern private.
***************************************************************/
{
  for(; !list_Empty(List); List=list_Cdr(List)) {
    TPTPOutput(File,list_Car(List));
    if (!list_Empty(list_Cdr(List)))
      fprintf(File,",");
  }
}


void TPTPOutput(FILE* File, TERM Term)
/**************************************************************
  INPUT:   A file and a term.
  RETURNS: none.
  SUMMARY: Prints the term in prefix notation to the file. 
***************************************************************/
{
  SYMBOL Top;
  Top = term_TopSymbol(Term);
  if (term_IsComplex(Term)) {
    if (fol_IsQuantifier(Top)) {
      TPTPSymbolOutput(File,Top);
      fprintf(File," [");
      TPTPListOutput(File,fol_QuantifierVariables(Term));
      fprintf(File,"] : ");
      TPTPOutput(File, term_SecondArgument(Term));
    }
    else
      if (symbol_Equal(Top,fol_And()) || symbol_Equal(Top,fol_Or()) ||
	  symbol_Equal(Top,fol_Implies()) || symbol_Equal(Top,fol_Equiv())) {
	LIST Scan;
	fprintf(File,"(");
	for (Scan = term_ArgumentList(Term);!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	  TPTPOutput(File,list_Car(Scan));
	  if (!list_Empty(list_Cdr(Scan))) {
	    fprintf(File," ");
	    TPTPSymbolOutput(File,Top);
	    fprintf(File," ");
	  }
	}
	fprintf(File,")");
      }
      else
	if (symbol_Equal(Top,fol_Not())) {
	  TPTPSymbolOutput(File,Top);
	  TPTPListOutput(File,term_ArgumentList(Term));
	}
	else {
	  TPTPSymbolOutput(File,Top);
	  fprintf(File,"(");
	  TPTPListOutput(File,term_ArgumentList(Term));
	  fprintf(File,")");
	}
  }
  else 
    TPTPSymbolOutput(File,Top);
}

/**************************************************************/
/* Main Function                                              */
/**************************************************************/

void main(int argc, char* argv[])
{
  LIST    Clauses,Axioms,Conjectures,Scan;
  FILE    *Output;
  int     i;
  CLAUSE  Clause;
  char    *Filename,*Outname;

  memory_Init(-1);
  symbol_Init(TRUE);
  stack_Init();
  term_Init();
  flag_Init();
  fol_Init(TRUE);
  clause_Init();

  if (argc < 3) {
    printf("\n\t          dfg2tptp Version 0.42");
    printf("\n\t       Usage: dfg2tptp <input-file> <output-fule>\n\n");
    return;
  }

  Axioms      = list_Nil();
  Conjectures = list_Nil();

  Filename = *++argv;
  Clauses  = dfg_DFGParser(Filename, &Axioms, &Conjectures);

  symbol_LowerSignature();

  Outname = *++argv;
  Output  = misc_OpenFile(Outname,"w");

  if (!list_Empty(Axioms) || !list_Empty(Conjectures)) {
    fprintf(Output,"%%--------------------------------------------------------------------------");
    fprintf(Output,"\n%% File     : %s",Filename);
    fprintf(Output,"\n%% Domain   : No information");
    fprintf(Output,"\n%% Problem  : No information");
    fprintf(Output,"\n%% Version  : No information");
    fprintf(Output,"\n%% English  : No information");
    fprintf(Output,"\n\n%% Refs     : No information");
    fprintf(Output,"\n%% Source   : No information");
    fprintf(Output,"\n%% Names    : No information");
    fprintf(Output,"\n\n%% Status   : %s",( dfg_ProblemStatus()== UNSATISFIABLE ? "theorem" :
					 ( dfg_ProblemStatus()== SATISFIABLE ? "satisfiable" : "unknown")));
    fprintf(Output,"\n%% Rating   : ?");
    fprintf(Output,"\n%% Syntax   : No information");
    fprintf(Output,"\n\n%% Comments : No information");
    fprintf(Output,"\n%%--------------------------------------------------------------------------");
    i=1;
    for (Scan = Axioms;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      fprintf(Output,"\ninput_formula(ax%d,axiom,(\n\t",i++);
      TPTPOutput(Output,list_Car(Scan));
      fprintf(Output,")).");
    }
    i=1;
    for (Scan = Conjectures;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      fprintf(Output,"\ninput_formula(co%d,conjecture,(\n\t",i++);
      TPTPOutput(Output,list_Car(Scan));
      fprintf(Output,")).\n");
    }
    fprintf(Output,"%%--------------------------------------------------------------------------");
  }
  else {
    fprintf(Output,"%%--------------------------------------------------------------------------");
    fprintf(Output,"\n%% File     : %s",Filename);
    fprintf(Output,"\n%% Domain   : No information");
    fprintf(Output,"\n%% Problem  : No information");
    fprintf(Output,"\n%% Version  : No information");
    fprintf(Output,"\n%% English  : No information");
    fprintf(Output,"\n\n%% Refs     : No information");
    fprintf(Output,"\n%% Source   : No information");
    fprintf(Output,"\n%% Names    : No information");
    fprintf(Output,"\n\n%% Status   : %s",( dfg_ProblemStatus()== UNSATISFIABLE ? "unsatisfiable" :
					 ( dfg_ProblemStatus()== SATISFIABLE ? "satisfiable" : "unknown")));
    fprintf(Output,"\n%% Rating   : ?");
    fprintf(Output,"\n%% Syntax   : No information");
    fprintf(Output,"\n\n%% Comments : No information");
    fprintf(Output,"\n%%--------------------------------------------------------------------------");
    for (Scan = Clauses;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      Clause = (CLAUSE)list_Car(Scan);
      fprintf(Output,"\ninput_clause(clause%d,%s,\n\t",clause_Number(Clause),
	      (clause_GetFlag(Clause,CONCLAUSE) ? "conjecture" : "axiom"));
      fprintf(Output,"[");
      for (i=clause_FirstLitIndex();i<clause_Length(Clause);i++) {
	if (i <= clause_LastAntecedentLitIndex(Clause))
	  fprintf(Output,"--");
	else
	  fprintf(Output,"++");
	TPTPOutput(Output,clause_GetLiteralAtom(Clause,i));
	if (i+1<clause_Length(Clause))
	  fprintf(Output,", ");
      }
      fprintf(Output,"]).\n");
    }
    fprintf(Output,"%%--------------------------------------------------------------------------");
  }

  misc_CloseFile(Output,Outname);


  list_DeleteWithElement(Clauses,clause_Delete);
  list_DeleteWithElement(Axioms,term_Delete);
  list_DeleteWithElement(Conjectures,term_Delete);
  /*symbol_Dump();*/
  fol_Free();
  symbol_FreeAllSymbols();
  dfg_Free();
#ifdef DEBUG
  memory_Print();
#endif
}


