/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *               INPUT OF SPASS CLAUSES                   * */
/* *                                                        * */
/* *  $Module:   DFGSCANNER                                 * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.56 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/05/26 12:32:51 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: dfgscanner.h,v $ */


#ifndef _DFGSCANNER_
#define _DFGSCANNER_


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "define.h"
#include "symbol.h"


/**************************************************************/
/* Type Definitions                                           */
/**************************************************************/

/* The tokens are the terminal symbols of this grammar               */
/* Don't change the order of tokens, because a range of tokens       */
/* might be tested by the parser !!!!!!                              */
/* If new tokens should be necessary add them at the end of the list */
typedef enum { DFG_CLAUSE = 256, DFG_FORMULA,       DFG_ENDLIST,
	       DFG_BEGPROBLEM,   DFG_ENDPROBLEM,    DFG_SYMBOLLIST,
	       DFG_CLAUSELIST,   DFG_FORMLIST,      DFG_PROOF,
	       DFG_FUNCTIONS,    DFG_OPERATORS,     DFG_PREDICATS,
	       DFG_QUANTIFS,     DFG_SORTS,         DFG_NUM,
	       DFG_ID,           DFG_FUNCTION,      DFG_PREDICATE,
	       DFG_JUNCTOR,      DFG_SETTINGS,      DFG_SPASS,
	       DFG_AXIOMS,       DFG_CONJECS,       DFG_RESOLUTION,
	       DFG_CNF,          DFG_DNF,           DFG_FORALL,
	       DFG_EXISTS,       DFG_AND,           DFG_OR,
	       DFG_NOT,          DFG_IMPLIES,       DFG_IMPLIED,
	       DFG_EQUIV,        DFG_EQUAL,         DFG_TRUE,
	       DFG_FALSE,        DFG_DESCLIST,      DFG_NAME,
	       DFG_AUTHOR,       DFG_VERSION,       DFG_LOGIC,
	       DFG_STATUS,       DFG_DESCRIPTION,   DFG_DATE,
	       DFG_SATISFIABLE,  DFG_UNSATISFIABLE, DFG_UNKNOWN,
	       DFG_DECLLIST,     DFG_SUBSORT,       DFG_SORTPREDICATE
} DFG_TOKEN;

typedef struct {
    long   number;
    char*  string;
    SYMBOL symbol;
} DFG_LVALTYPE;  

/**************************************************************/
/* Global Variables                                           */
/**************************************************************/

/* Used for information exchange of scanner and parser */
/* e.g. if the returned token of function dfg_Scanner  */
/* is DFG_ID dfg_LVAL.string contains the name of the  */
/* identifier.                                         */
extern DFG_LVALTYPE dfg_LVAL;

/* Used for handling of identifiers (not special keywords. */
/* If TRUE, the scanner searches the symbol table to find  */
/* a symbol with the same name. If a symbol <s> with same  */
/* name is found, dfg_LVAL.symbol is set to <s>.           */
/* If FALSE, dfg_LVAL.string is set to the name of the     */
/* identifier.                                             */
extern BOOL dfg_LOOKUPSYMBOLS;


/**************************************************************/
/* Prototypes                                                 */
/**************************************************************/

void  dfg_InitScanner(FILE*);
void  dfg_FreeScanner(void);

void  dfg_SetInputFile(FILE*);

NAT   dfg_ActLineNumber(void);
void  dfg_FPrintErrorLine(FILE*);

int   dfg_Scanner(void);

char* dfg_ReadText(BOOL);

char* dfg_StringCopy(const char*);
void  dfg_StringFree(char*);

void  dfg_FPrintToken(FILE*, int);

#endif
