/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                 PROOF DOCUMENTATION                    * */
/* *                                                        * */
/* *  $Module:   DOCPROOF                                   * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.50 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/06/20 09:15:18 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: doc-proof.c,v $ */

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "doc-proof.h"

/**************************************************************/
/* Global Variables                                           */
/**************************************************************/

int dp_DEPTH;


/**************************************************************/
/* Functions                                                  */
/**************************************************************/

void dp_Init(void)
{
  dp_DEPTH = 0;
}



static void dp_DocProof(PROOFSEARCH Search, CLAUSE Clause, int depth, LIST* ProofClausesPt)
/*********************************************************
  INPUT:   A proof search object, a clause, a depth parameter and a list (by reference)
           keeping all proof clauses detected so far.
  RETURNS: A list of clauses that participate in the proof of <Clause> if not
           already contained in <*ProofClausesPt>
  MEMORY:  Allocates memory for the resulting list.
**********************************************************/
{
  depth++;

  if (depth > dp_ProofDepth())
    dp_SetProofDepth(depth);

  /*  if (!list_PointerMember(*ProofClausesPt, (POINTER)Clause)) { */
  if (!clause_Exists(clause_GetNumberedCl(clause_Number(Clause),*ProofClausesPt))) {
    LIST   Scan, ParentClNums, PClauses;
    CLAUSE NextParentClause;
    int    NextParentNumber;

    *ProofClausesPt = list_Cons(Clause, *ProofClausesPt);
    ParentClNums    = list_PointerDeleteDuplicates( list_Copy(clause_ParentClauses(Clause)));
    PClauses        = list_Nil();

    for (Scan = ParentClNums; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
      NextParentNumber = (int)list_Car(Scan);
      /* printf("\nWe seek: %d",NextParentNumber);DBG */

      NextParentClause = clause_GetNumberedCl(NextParentNumber, prfs_WorkedOffClauses(Search));
      
      if (!clause_Exists(NextParentClause))
	NextParentClause = clause_GetNumberedCl(NextParentNumber, prfs_UsableClauses(Search));

      if (!clause_Exists(NextParentClause))
	NextParentClause = clause_GetNumberedCl(NextParentNumber, prfs_DocProofClauses(Search));

      if (!clause_Exists(NextParentClause))
	printf("\nError, no clause with number %d in any list!",
	       NextParentNumber);
      else
	PClauses = list_Cons(NextParentClause, PClauses);
    }
    list_Delete(ParentClNums);
    PClauses = list_NReverse(PClauses);


    for (Scan = PClauses; !list_Empty(Scan); ) { /* Recursive documentation! */
      Clause = list_Car(Scan);
      /* Free all memory, because list nodes are recursively needed! */
      Scan = list_Cdr(Scan);
      list_Free(PClauses);
      PClauses = Scan;

      dp_DocProof(Search, Clause, depth, ProofClausesPt);
    }
  }
}

void dp_FPrintDFGProof(LIST Clauses, char *argv)
/*********************************************************
  INPUT:   A list of clauses representing a proof and
	   a string indicating a file name.
  RETURNS: None.
  EFFECT:  Outputs the proof in DFG proof format to <*argv>.prf
  MEMORY:  No effects.
**********************************************************/
{
  FILE   *Output;
  CLAUSE Clause;
  LIST   AxClauses,ConClauses,ProofClauses,Scan;
  char   name[100];

  AxClauses = ConClauses = ProofClauses = list_Nil();

  sprintf(name,"%s.prf",argv);
  Output = misc_OpenFile(name,"w");
  fprintf(Output, "begin_problem(Unknown).\n\n");
  fprintf(Output, "list_of_descriptions.\n");
  fprintf(Output, "name({*%s*}).\n",argv);
  fprintf(Output, "author({*SPASS*}).\n");
  fprintf(Output, "status(unsatisfiable).\n");
  fprintf(Output, "description({*File generated by SPASS containing a proof.*}).\n");
  fprintf(Output, "end_of_list.\n\n");
  fprintf(Output, "list_of_symbols.\n");
  fol_PrintDFGSignature(Output);
  fprintf(Output, "end_of_list.\n\n");
  
  for(Scan=Clauses;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    Clause = (CLAUSE)list_Car(Scan);
    if (clause_IsFromInput(Clause))
      if (clause_GetFlag(Clause,CONCLAUSE))
	ConClauses = list_Cons(Clause,ConClauses);
      else
	AxClauses  = list_Cons(Clause,AxClauses);
    else
      ProofClauses  = list_Cons(Clause,ProofClauses);
  }
  ConClauses   = list_NReverse(ConClauses);
  AxClauses    = list_NReverse(AxClauses);
  ProofClauses = list_NReverse(ProofClauses);
  clause_FPrintCnfDFG(Output, AxClauses, ConClauses);
  fprintf(Output, "\nlist_of_proof(resolution).\n");
  for(Scan=ProofClauses;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    clause_FPrintDFG(Output,list_Car(Scan),TRUE);
  }
  fprintf(Output, "end_of_list.\n\n");
  fprintf(Output, "end_problem.\n\n");
  misc_CloseFile(Output,name);
  printf("\nDFG Proof printed to: %s\n",name);
  list_Delete(ConClauses);
  list_Delete(AxClauses);
  list_Delete(ProofClauses);
}


void dp_PrintProof(PROOFSEARCH Search, LIST Clauses, char *argv)
/*********************************************************
  INPUT:   A proofsearch object, a list of empty clauses and
           the name of an output file.
  RETURNS: None.
  MEMORY:  No effects.
**********************************************************/
{
  LIST   ProofClauses,Scan,EmptyClauses,AllClauses;
  CLAUSE Clause;

  EmptyClauses = list_Copy(Clauses);
  ProofClauses = list_Nil();
  AllClauses   = list_Nconc(list_Copy(prfs_DocProofClauses(Search)),
			    list_Nconc(list_Copy(prfs_UsableClauses(Search)),
				       list_Copy(prfs_WorkedOffClauses(Search))));

  for (Scan = AllClauses; !list_Empty(Scan); Scan = list_Cdr(Scan))
    if (clause_IsFromSplitting(list_Car(Scan)))
      EmptyClauses = list_Cons(list_Car(Scan),EmptyClauses);
  list_Delete(AllClauses);

  /* If we found an empty clause while reducing the input clause set, */
  /* we first have to delete <Clause> from <ProofClauses.             */

  for (Clauses = EmptyClauses;
       !list_Empty(Clauses);
       Clauses = list_Cdr(Clauses)) {

    Clause       = list_Car(Clauses);
    ProofClauses = clause_NumberDelete(ProofClauses, clause_Number(Clause));

    dp_DocProof(Search, Clause, -1, &ProofClauses);

    ProofClauses = clause_NumberSort(ProofClauses);
  }
  printf("\n\nHere is a proof with depth %d, length %d :\n",
	 dp_ProofDepth(), list_Length(ProofClauses));
  clause_ListPrint(ProofClauses);
  if (flag_GetFlagValue(flag_FPDFGPROOF))
    dp_FPrintDFGProof(ProofClauses,argv);
  list_Delete(ProofClauses);
  list_Delete(EmptyClauses);

  /* printf("\n\nHere is the dp_ClauseList():\n");
     dp_SetClauseList(clause_NumberSort(dp_ClauseList()));
     ProofClauses = dp_ClauseList();
     while (!list_Empty(ProofClauses)) {
     if (clause_Exists(clause_GetNumberedCl(clause_Number(
     list_Car(ProofClauses)), list_Cdr(ProofClauses)))) {
     
     printf("\nFirst double: ");
     clause_Print(list_Car(ProofClauses));
     ProofClauses = list_Nil();
     }
     else
     ProofClauses = list_Cdr(ProofClauses);
     }
     clause_ListPrint(dp_ClauseList());*/
}


