/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                 FLAGS OF SPASS                         * */
/* *                                                        * */
/* *  $Module:   FLAGS                                      * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.60 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/03 12:24:55 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: flags.h,v $ */

#ifndef _FLAGS_
#define _FLAGS_

#include "misc.h"

/**************************************************************/
/* Data Structures and Constants                              */
/**************************************************************/

typedef enum { flag_AUTO,

	       flag_CPARA,      flag_CORDPARA,   flag_CSUPE,
	       flag_CSUPP,

               flag_SPLITS,     flag_MEMORY,     flag_TIMELIMIT,
	       flag_DOCSST,     flag_DOCDST,     flag_DOCPROOF,
	       flag_DOCSPLIT,   flag_LOOPS,      flag_PSUB,      
	       flag_PREW,       flag_PCON,       flag_PTAUT,
	       flag_POBV,       flag_PSSI,       flag_PSST,
	       flag_PCLR,       flag_PDER,       flag_PGIVEN,
	       flag_PKEPT,      flag_PPROBLEM,   flag_PEMPTYCLAUSE,
	       flag_PSTATISTIC, flag_FPMODEL,    flag_FPDFGPROOF,
	       flag_PFLAGS,     flag_POPTSKOLEM,

	       flag_SELECT,     flag_RINPUT,     flag_SORTS,
	       flag_SATINPUT,   flag_WDRATIO,    flag_FULLRED,
	       flag_FUNCWEIGHT, flag_VARWEIGHT,  flag_PREFVAR,
	       
	       flag_CNFOPTSKOLEM, flag_CNFPROOFSTEPS, 
	       
	       flag_IEMS,       flag_ISOR,       flag_IEQR, 
	       flag_IEQF,       flag_IMPM,       flag_ISPR, 
	       flag_ISPL,       flag_IGER,       flag_IGEF,
	       flag_IUNR,       flag_IBUR,

	       flag_RFREW,      flag_RBREW, 
	       flag_RFCLR,      flag_RBCLR,
	       flag_ROBV,
	       flag_RTAUT,      flag_RSST,       flag_RSSI,
	       flag_RDST,       flag_RFSUB,      flag_RBSUB,      
	       flag_RCON,         
	       

	       flag_MAXFLAG } FLAG_FLAGS;     /* flag_MAXFLAG is a final Dummy */

extern int   flag_VALUES[flag_MAXFLAG];
extern char* flag_STRINGS[flag_MAXFLAG];

/**************************************************************/
/* Macros                                                     */
/**************************************************************/

static __inline__ void flag_SetFlagValue(FLAG_FLAGS FlagId, int Value)
{
  flag_VALUES[FlagId] = Value;
}

static __inline__ int flag_GetFlagValue(FLAG_FLAGS FlagId)
{
  return flag_VALUES[FlagId];
}

static __inline__ char* flag_Name(FLAG_FLAGS Flag)
{
#if DEBUG
  if (Flag <0 || Flag >= flag_MAXFLAG) {
    fputs("\n\nRange of flags exceeded in flag_Name\n", stderr);
    misc_DumpCore();
  }
#endif
 
  return(flag_STRINGS[Flag]);
}


static __inline__ void flag_ClearInferenceRules()
{
  flag_VALUES[flag_IEMS] = 0;
  flag_VALUES[flag_ISOR] = 0;
  flag_VALUES[flag_IEQR] = 0;
  flag_VALUES[flag_IEQF] = 0;
  flag_VALUES[flag_IMPM] = 0;
  flag_VALUES[flag_ISPR] = 0;
  flag_VALUES[flag_ISPL] = 0;
  flag_VALUES[flag_IGER] = 0;
  flag_VALUES[flag_IGEF] = 0;
  flag_VALUES[flag_IBUR] = 0;
}


static __inline__ void flag_ClearReductionRules()
{
  flag_VALUES[flag_RFREW] = 0;
  flag_VALUES[flag_RBREW] = 0;
  flag_VALUES[flag_RFCLR] = 0;
  flag_VALUES[flag_RBCLR] = 0;
  flag_VALUES[flag_ROBV]  = 0;
  flag_VALUES[flag_RTAUT] = 0;
  flag_VALUES[flag_RSST]  = 0;
  flag_VALUES[flag_RDST]  = 0;
  flag_VALUES[flag_RSSI]  = 0;
  flag_VALUES[flag_RFSUB] = 0;
  flag_VALUES[flag_RBSUB] = 0;
  flag_VALUES[flag_RCON]  = 0;
}

static __inline__ void flag_PrintReductionRules()
{
  int i;
  printf("\n Reductions: ");
  for(i=flag_RFREW;i<=flag_RCON;i++)
    if (flag_VALUES[i])
      printf("%s ",flag_STRINGS[i]);
} 

static __inline__ void flag_PrintInferenceRules()
{
  int i;
  printf("\n Inferences: ");
  for(i=flag_IEMS;i<=flag_IBUR;i++)
    if (flag_VALUES[i])
      printf("%s ",flag_STRINGS[i]);
} 



/**************************************************************/
/* Functions                                                  */
/**************************************************************/

void  flag_Init(void);
void  flag_Print(void);
void  flag_FPrint(FILE*);
char* flag_SPrint(char*);
BOOL  flag_Lookup(const char*);
int   flag_Id(const char*);

#endif






