/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *               TOP MODULE OF FLOTTER                    * */
/* *                                                        * */
/* *  $Module:   FLOTTER                                    * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.66 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/06/20 09:17:02 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: flotter.c,v $ */


/**************************************************************/
/*  DESCRIPTION:                                              */
/*                                                            */
/*  Main program for clause normal form tranformation with    */
/*  renaming and optimized skolemization. The input is the    */
/*  name of a file with a formel and an output file name.     */
/*  If no output file is designated the clauses are written   */
/*  to stdout. The clauses are tested for subsumption and     */
/*  for condensing.                                           */
/*  Additionally the programm tries to find a  a proof for    */
/*  the input formula. At the end the programm                */
/*  returns if it has found a proof and the time used for     */
/*  the search and the time used for the transformation       */
/*  to optimized clause normal form                           */
/*                                                            */
/**************************************************************/

#include "dfgparser.h"
#include "options.h"
#include "cnf.h"

void main(int argc, char* argv[])
{
  LIST Scan, AxiomList, ConjectureList, AxClauses, ConClauses;
  FILE *Output;
  char *Buffer,*BufferPos;
  NAT  BufferSize;

#ifdef OPTCHECK
  LIST SkolemFunctions2;
#endif

  memory_Init(-1);
  /*memory_Print();*/
  symbol_Init(TRUE);
  stack_Init();
  flag_Init(); 
  term_Init();
  fol_Init(TRUE);
  cont_Init();
  unify_Init();
  vec_Init();
  ren_Init();
  opts_Init();

  if (!opts_Read(argc, argv)) 
    return;

  if (argc != opts_Indicator()+2) {                 /* No input/output file given */
    printf("\n\t          FLOTTER Version 0.55");
    printf("\n\tUsage: FLOTTER <input-file> <output-file>\n\n");
    printf("Possible options:\n\n");
    opts_PrintNames(); 
    return;
  }

  AxiomList = ConjectureList = list_Nil();
  AxClauses = ConClauses = list_Nil();

  dfg_Parser(argv[opts_Indicator()],&AxiomList,&ConjectureList);

  opts_Transfer();
  opts_Free();

  cnf_Flotter(AxiomList,ConjectureList,&AxClauses,&ConClauses);

  
  BufferSize = clause_SizeOfDFGPrintedClauseSet(AxClauses)+
    clause_SizeOfDFGPrintedClauseSet(ConClauses) + dfg_DescriptionLength();
  Buffer     = (char *)memory_Malloc(BufferSize);
  BufferPos  = Buffer;

  /*printf("\n\n\t Size of printed problem: %d\n", BufferSize);*/
  
  BufferPos += sprintf(BufferPos, "begin_problem(Unknown).\n\n");
  BufferPos  = dfg_SPrintDescription(BufferPos);
  BufferPos += sprintf(BufferPos, "\n\n");  
  BufferPos += sprintf(BufferPos, "list_of_symbols.\n");
  BufferPos  = fol_SPrintDFGSignature(BufferPos);
  BufferPos += sprintf(BufferPos, "end_of_list.\n\n");
  BufferPos += sprintf(BufferPos, "list_of_clauses(axioms, cnf).\n");
  
  for (Scan=AxClauses;!list_Empty(Scan);Scan=list_Cdr(Scan))
    BufferPos = clause_SPrintDFG(BufferPos,list_Car(Scan));

  BufferPos += sprintf(BufferPos, "end_of_list.\n\n");
  BufferPos += sprintf(BufferPos, "list_of_clauses(conjectures, cnf).\n");
  
  for (Scan=ConClauses;!list_Empty(Scan);Scan=list_Cdr(Scan))
    BufferPos = clause_SPrintDFG(BufferPos,list_Car(Scan));
 
  BufferPos += sprintf(BufferPos, "end_of_list.\n\n");
  /*BufferPos  = flag_SPrint(BufferPos);*/
  BufferPos += sprintf(BufferPos, "\nend_problem.\n\n");

  Output = misc_OpenFile(argv[opts_Indicator()+1],"w");
  fwrite(Buffer,sizeof(char),strlen(Buffer),Output);
  misc_CloseFile(Output,argv[opts_Indicator()+1]);

  list_DeleteWithElement(AxClauses,clause_Delete);
  list_DeleteWithElement(ConClauses,clause_Delete);

  /*term_Delete(Formula);*/
  /*memory_Print();*/
  /*symbol_PrintAll();*/
  memory_Free(Buffer,BufferSize);
  unify_Free();
  cont_Free();
  fol_Free();
  /*Symbol_Dump();*/
  symbol_FreeAllSymbols();
  dfg_Free(); 
#ifdef DEBUG
  memory_Print();
#endif
}
