/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *         EXTENDED KNUTH BENDIX ORDERING                 * */
/* *                                                        * */
/* *  $Module:   KBO                                        * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.46 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/01 14:14:38 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: kbo.h,v $ */

#ifndef _KBO_
#define _KBO_

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "term.h"
#include "context.h"
#include "foldfg.h"

/**************************************************************/
/* Structures and constants.                                  */
/**************************************************************/

extern const int kbo_MINWEIGHT;     /* Gewicht fuer Variable  */

typedef enum {kbo_SMALLER_THAN,
	      kbo_EQUAL,
	      kbo_UNCOMPARABLE,
	      kbo_GREATER_THAN} kbo_RESULT;

/**************************************************************/
/*  MACROS                                                    */
/**************************************************************/

static __inline__ kbo_RESULT kbo_MinWeight(void)
{
  return kbo_MINWEIGHT;
}

static __inline__ kbo_RESULT kbo_Uncomparable(void)
{
  return kbo_UNCOMPARABLE;
}

static __inline__ kbo_RESULT kbo_Equal(void)
{
  return kbo_EQUAL;
}

static __inline__ kbo_RESULT kbo_GreaterThan(void)
{
  return kbo_GREATER_THAN;
}

static __inline__ kbo_RESULT kbo_SmallerThan(void)
{
  return kbo_SMALLER_THAN;
}

static __inline__ BOOL kbo_IsGreaterThan(kbo_RESULT Res)
{
  return kbo_GREATER_THAN == Res;
}

static __inline__ BOOL kbo_IsNotGreaterThan(kbo_RESULT Res)
{
  return kbo_GREATER_THAN != Res;
}

static __inline__ BOOL kbo_IsSmallerThan(kbo_RESULT Res)
{
  return kbo_SMALLER_THAN == Res;
}

static __inline__ BOOL kbo_IsNotSmallerThan(kbo_RESULT Res)
{
  return kbo_SMALLER_THAN != Res;
}

static __inline__ BOOL kbo_IsEqual(kbo_RESULT Res)
{
  return kbo_EQUAL == Res;
}

static __inline__ BOOL kbo_IsUncomparable(kbo_RESULT Res)
{
  return kbo_UNCOMPARABLE == Res;
}

static __inline__ int kbo_SymbolWeight(SYMBOL S)
{
  if (symbol_IsVariable(S))
    return kbo_MINWEIGHT;
  else
    return symbol_Weight(S);
}

            
/**************************************************************/
/*  FUNCTIONS                                                 */
/**************************************************************/

void       kbo_Print(kbo_RESULT);
int        kbo_TermWeight(TERM);
kbo_RESULT kbo_Compare(TERM, TERM);
kbo_RESULT kbo_ContCompare(CONTEXT, TERM, CONTEXT, TERM);
BOOL       kbo_ContGreater(CONTEXT, TERM, CONTEXT, TERM);
kbo_RESULT kbo_LiteralCompare(TERM,BOOL,TERM,BOOL,BOOL);

#endif
