/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *                     LISTS                              * */
/* *                                                        * */
/* *  $Module:   LIST                                       * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.49 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/01 14:06:44 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: list.h,v $ */

#ifndef _LIST_
#define _LIST_

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "memory.h"
#include "misc.h"

/**************************************************************/
/* Structures                                                 */
/**************************************************************/

typedef struct LIST_HELP {
  struct LIST_HELP *cdr;
  POINTER           car;
} LIST_NODE;

typedef LIST_NODE *LIST;

/**************************************************************/
/* Macros                                                     */
/**************************************************************/

static __inline__ void list_Free(LIST L)
{
  memory_Free((char*) L,sizeof(LIST_NODE));
}

static __inline__ LIST list_Nil(void)
{
  return (LIST)NULL;
}

static __inline__ BOOL list_Empty(LIST L)
{
  return L == (LIST) NULL;
}

static __inline__ BOOL list_Exist(LIST L)
{
  return L != (LIST) NULL;
}

static __inline__ POINTER list_Car(LIST L)
{
  return L->car;
}

static __inline__ POINTER list_NCar(LIST *L)
{
  POINTER Result;
  LIST    Help;

  Result = (*L)->car;
  Help   = (*L)->cdr;
  list_Free(*L);
  *L     = Help;
  return Result;
}

static __inline__ LIST list_Cdr(LIST L)
{
  return L->cdr;
}

static __inline__ POINTER list_First(LIST L)
{
  return list_Car(L);
}

static __inline__ POINTER list_Second(LIST L)
{
  return list_Car(list_Cdr(L));
}

static __inline__ POINTER list_Third(LIST L)
{
  return list_Car(list_Cdr(list_Cdr(L)));
}

static __inline__ void list_Rplacd(LIST L1, LIST L2)
{
  L1->cdr = L2;
}

static __inline__ void list_Rplaca(LIST L, POINTER P)
{
  L->car = P;
}

static __inline__ void list_RplacSecond(LIST L, POINTER P)
{
  list_Rplaca(list_Cdr(L), P);
}


/**************************************************************/
/* Functions                                                  */
/**************************************************************/

LIST    list_Copy(const LIST);
LIST    list_CopyWithElement(const LIST, POINTER (*)());
void    list_InsertNext(LIST, POINTER);

LIST    list_NMapCar(LIST, POINTER (*)());
void    list_Apply(void (*)(), LIST);

LIST    list_Reverse(const LIST);
LIST    list_NReverse(LIST);

LIST    list_PointerSort(LIST);
LIST    list_Sort(LIST, BOOL (*)());
LIST    list_NumberSort(LIST , NAT (*)());
LIST    list_GreaterNumberSort(LIST , NAT (*)());
LIST    list_NNumberMerge(LIST , LIST, NAT (*)());

POINTER list_DequeueNext(LIST);
POINTER list_NthElement(LIST, NAT);
void    list_DeleteWithElement(LIST, void (*)());
NAT     list_DeleteWithElementCount(LIST, void (*)());
LIST    list_DeleteElement(LIST, POINTER, BOOL (*)());
LIST    list_DeleteElementIf(LIST, BOOL (*)(), void (*)());
LIST    list_DeleteElementFree(LIST, POINTER, BOOL (*)(), void (*)());
LIST    list_PointerDeleteElement(LIST, POINTER );
LIST    list_PointerDeleteOneElement(LIST, POINTER);
BOOL    list_DeleteFromList(LIST*, POINTER);
BOOL    list_DeleteOneFromList(LIST*, POINTER);
LIST    list_DeleteDuplicates(LIST, BOOL (*)());
LIST    list_DeleteDuplicatesFree(LIST, BOOL (*)(), void (*)());
LIST    list_PointerDeleteDuplicates(LIST);

LIST    list_NPointerUnion(LIST, LIST);
LIST    list_NUnion(LIST, LIST, BOOL (*)());
LIST    list_NListTimes(LIST, LIST);
LIST    list_NIntersect(LIST, LIST, BOOL (*)());
void    list_NInsert(LIST, LIST);
LIST    list_NPointerIntersect(LIST, LIST);
BOOL    list_HasIntersection(LIST, LIST);
LIST    list_NPointerDifference(LIST, LIST);
LIST    list_NDifference(LIST, LIST, BOOL (*)());
BOOL    list_Member(LIST, POINTER, BOOL (*)());
BOOL    list_PointerMember(LIST, POINTER);
BOOL    list_PointerReplaceMember(LIST, POINTER, POINTER);

NAT     list_Length(LIST);
NAT     list_Bytes(LIST);

/**************************************************************/
/* Functional Inline Functions                                */
/**************************************************************/

static __inline__ LIST list_Cons(POINTER Ptr, const LIST List)
{ 
  LIST Cell;  

  Cell = (LIST)memory_Malloc(sizeof(LIST_NODE)); 
  Cell->car = Ptr;    
  Cell->cdr = List;            
  return Cell;
}


static __inline__ LIST list_Nconc(LIST List1, LIST List2)
{
  LIST Result;

  if (list_Empty(List1))  
    return List2;        

  if (list_Empty(List2)) 
    return List1;

  Result = List1;
  for (List1 = Result; !list_Empty(list_Cdr(List1));    
       List1 = list_Cdr(List1)) ;            
  List1->cdr = List2;
  return Result; 
}


static __inline__ LIST list_List(POINTER P)
{
  return list_Cons(P,list_Nil());
}


static __inline__ LIST list_Append(LIST List1, LIST List2)
{ 
  LIST Result;

  if (list_Empty(List1))  
    return List2; 
  if (list_Empty(List2))
    return list_Copy(List1); 

  Result = list_Copy(List1);  
  for (List1 = Result; !list_Empty(list_Cdr(List1));
       List1 = list_Cdr(List1)) ; 
  List1->cdr = List2;   
  return Result; 
}


static __inline__ void list_Delete(LIST L)
{
  LIST Current; 

  Current = L; 
  while (!list_Empty(Current)) { 
    L = list_Cdr(L);   
    list_Free(Current);
    Current = L; 
  }                
}


/**************************************************************/
/* Stack Macros                                               */
/**************************************************************/

static __inline__ LIST list_StackBottom(void)
{
  return list_Nil();
}


static __inline__ BOOL list_StackEmpty(LIST S)
{
  return list_Empty(S);
}


static __inline__ LIST list_Push(POINTER I, LIST L)
{
  return list_Cons(I, L);
}


static __inline__ POINTER list_Top(LIST L)
{
  return list_Car(L);
}


static __inline__ LIST list_Pop(LIST L)
{
  LIST Aux = L; 

  L = list_Cdr(L);   
  list_Free(Aux);
  return L;
}


static __inline__ void list_RplacTop(LIST L, POINTER P)
{
  list_Rplaca(L, P);
}


static __inline__ LIST list_StackFree(LIST L)
{
  while (!list_StackEmpty(L))
    L = list_Pop(L);
  return list_Nil();
}


/**************************************************************/
/* Pair Macros                                                */
/**************************************************************/

static __inline__ LIST list_PairNull(void)
{
  return list_Nil();
}


static __inline__ LIST list_PairCreate(POINTER P1, POINTER P2)
{
  return list_Cons(P1, P2);
}


static __inline__ void list_PairFree(LIST L)
{
  list_Free(L);
}


static __inline__ POINTER list_PairFirst(LIST L)
{
  return list_Car(L);
}


static __inline__ POINTER list_PairSecond(LIST L)
{
  return (POINTER)list_Cdr(L);
}

#endif
