#include "declarations.h"
#include "memory.h"

/* #define WORDSIZE            (2)    / * Bytes/Word For PC */
#define WORDSIZE            (4)    /* Bytes/Word For Unix                    */
#define PAGESIZE            (1024)   /* in words of 4 bytes                 */
#define MAXELEMSIZE         (256)   /* in words; has to be leq PAGESIZE/2  */

#define MemoryGetNewPage()  ((_Ptr)calloc(1,WORDSIZE*PAGESIZE))

struct {
  _Ptr freelist;         /* first elem of free list, if any                 */
  _Ptr nextnew;          /* next elem to be used in current page, if any    */
  _Ptr firstunusable;    /* first unusable position of this page, if any    */
   int usable;           /* size in words of usable part of this page, i.e. */
}marray[MAXELEMSIZE+1];  /* usable = max multiple of elemsize < PAGESIZE    */


int allocatedPages = 0;
int usedMemory   = 0;


void MemoryInit()
Proc
Var
  int i;
EndVar
  For i = 1; i <= MAXELEMSIZE; i++ Do
    marray[i].freelist     = NULL;
    marray[i].nextnew      = NULL;
    marray[i].usable       = PAGESIZE - (PAGESIZE Mod i);
  EndFor
EndProc


_Ptr MemoryAlloc(int size)
Proc
Var 
  _Ptr newMem;
  int i;
EndVar

  i = (size-1)/4+1;       /* i is the smallest integer such that 4i>=size */
  If marray[i].freelist != NULL Then
    newMem = marray[i].freelist;
    marray[i].freelist = (_Ptr)*marray[i].freelist;
  ElseIf marray[i].nextnew != NULL Then
    newMem = marray[i].nextnew;
    marray[i].nextnew = marray[i].nextnew + i;
    If marray[i].nextnew >= marray[i].firstunusable Then
      marray[i].nextnew = NULL;
    EndIf
  Else
    newMem = MemoryGetNewPage();
    If newMem == NULL Then
      printf(" error: out of system memory \n");
      exit(-1);
    EndIf
    allocatedPages++;
    marray[i].nextnew = newMem + i;
    marray[i].firstunusable = newMem + marray[i].usable;
    /* printf("new page for size %d\n", size);  */
  EndIf
  usedMemory = usedMemory + i;
  /* printf("allocated: %i bytes = %i words at pos %u\n",size,i,newMem); */
  Return newMem;
EndProc


void MemoryFree(_Ptr pos, int size)
Proc
Var
  int i;
  _Ptr aux;
EndVar
  i = (size-1)/4+1;
  /* printf("\n freeing: %ld words at pos %p\n",i,pos); */
  aux = marray[i].freelist;
  marray[i].freelist = pos;
  *(_Ptr*)marray[i].freelist = aux;
  usedMemory = usedMemory - i;
EndProc


void MemoryStatistics(void)
Proc
  printf("Space: %d bytes ", 4*usedMemory);
  printf("in %d pages of size %d bytes\n", allocatedPages, 4*PAGESIZE);
EndProc

/* ============== end memory.c =================================== */
