/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *        DYNAMIC MEMORY MANAGEMENT MODULE                * */
/* *                                                        * */
/* *  $Module:   MEMORY                                     * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.44 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/01/08 16:11:40 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: memory.h,v $ */


#ifndef _MEMORY_
#define _MEMORY_

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "misc.h"

/**************************************************************/
/* Data structures and constants                              */
/**************************************************************/

#define memory__DYNMAXSIZE   1024      /* At most 1024 Bytes are allocated over the memory
					  module, larger requests are directly mapped to system
					  calls. */

struct {
  POINTER free;
  POINTER new;
  POINTER page;
  int     size;
  int     offset;
} memory_ARRAY[memory__DYNMAXSIZE];

extern unsigned int memory_NEWBYTES;
extern unsigned int memory_FREEDBYTES;
extern const unsigned int memory_MAGICMALLOC;
extern const unsigned int memory_MAGICFREE;
extern const unsigned int memory_ALIGN;

/**************************************************************/
/* Debug Information                                          */
/**************************************************************/

extern const unsigned int memory_MARKSIZE ;

/**************************************************************/
/* Inline Functions                                           */
/**************************************************************/

static __inline__ unsigned int memory_DemandedBytes(void)
{
  return memory_NEWBYTES;
}


static __inline__ unsigned int memory_UsedBytes(void)
{
  return memory_NEWBYTES-memory_FREEDBYTES;
}


#ifdef PURIFY

static __inline__ void memory_Free(POINTER Freepointer, unsigned int Size)
{
  free((char*)Freepointer);
}

#else

static __inline__ void memory_Free(POINTER Freepointer, unsigned int Size) { 
#ifdef DEBUG
    if ( *((int *)(Freepointer) - 2) == memory_MAGICFREE) {
	printf("memory: Pointer %p has already been freed \n", Freepointer);
	misc_DumpCore();
    }
    if ( *((int *)(Freepointer) - 2) != memory_MAGICMALLOC) {
	printf("memory: Pointer %p was not allocated using memory_Malloc,\n",
	       Freepointer);
	printf("        or the memory block was corrupted.\n");
	misc_DumpCore();
    }
    if ( *((int *)(Freepointer) - 1) != Size) {
	printf("memory: Pointer %p was allocated for a block of size %d\n",
	       Freepointer, *((int *)(Freepointer) - 1));
	printf("        but was freed for a block of size %d\n", Size);
	misc_DumpCore();
    }
    *((int *)(Freepointer) - 2) = memory_MAGICFREE;
#endif
  if (Size >= memory__DYNMAXSIZE) {
    memory_FREEDBYTES       += Size + memory_MARKSIZE;
#ifdef DEBUG
    free((int *)(Freepointer)-2);
#else
    free(Freepointer);
#endif
  }
  else {
    memory_FREEDBYTES       += memory_ARRAY[Size].size;
    *(POINTER *)Freepointer = memory_ARRAY[Size].free;
    memory_ARRAY[Size].free = Freepointer;
  }
}

#endif


/**************************************************************/
/* Functions                                                  */
/**************************************************************/

void    memory_Init(int);
void    memory_Restrict(int);
POINTER memory_Malloc(unsigned int);
void    memory_Print(void);
void    memory_FPrint(FILE*);
void    memory_PrintDetailed(void);

#endif
