/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *             SPASS OPTIONS HANDLING                     * */
/* *                                                        * */
/* *  $Module:   OPTIONS                                    * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.61 $                                        * */
/* $State: Exp $                                            * */
/* $Date:                              * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: options.c,v $ */

/***************************************************************
 ***************************************************************

		 COPYRIGHT NOTICE:

                 This file contains code that               
                 has been copied with minor modifications   
                 from the 'getopt' module in the            
                 GNU gcc library 2.0. The copyright for
                 this code is claimed by
		 
		 Copyright 1991 Regents of the 
		 University of California.
		 All rights reserved.

		 The copying and modification of the
		 original code is in accordance 
		 with the copyright conditions for the      
                 GNU gcc library, which are listed below.           


    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions
    are met:

    Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.

    Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.

    All advertising materials mentioning features or use of this software
    must display the following acknowledgement:

    This product includes software developed by the University of
    California, Berkeley and its contributors.

    Neither the name of the University nor the names of its contributors
    may be used to endorse or promote products derived from this software
    without specific prior written permission.

    This software is provided by the regents and contributors ``as is'' and
    any express or implied warranties, including, but not limited to, the
    implied warranties of merchantability and fitness for a particular purpose
    are disclaimed.  in no event shall the regents or contributors be liable
    for any direct, indirect, incidental, special, exemplary, or consequential
    damages (including, but not limited to, procurement of substitute goods
    or services; loss of use, data, or profits; or business interruption)
    however caused and on any theory of liability, whether in contract, strict
    liability, or tort (including negligence or otherwise) arising in any way
    out of the use of this software, even if advised of the possibility of
    such damage.

************************************************************************
************************************************************************/

#include "options.h"

/**************************************************************/
/* Local variables  and types                                 */
/**************************************************************/

/* Array with options information for short options */
/* All abbreviated long options must be listed here */
static char* opts_Short = ":dt:m:p:";

/* Array with option information, set by opts_Init */
static struct OPTION opts_Long[flag_MAXFLAG+opts_MaxOpts+1];

char *opts_OPTNames[opts_MaxOpts];

/* Array with long options that have equivalent short options */
/* Zero character as abbreviation marks end of array */

typedef struct  { char* name;  char abbrev;} ABBREVS;
static ABBREVS opts_Abbreviations[] = { 
  { "TimeLimit", 't' }, 
  {"Memory", 'm'}, 
  {"DocProof", 'p'}, 
  { "",0 } } ;


static LIST opts_PARAMETERS;

/*************************************************************
  Local variables and types from the former getopt module code. 
**************************************************************/


/* Describe how to deal with options that follow non-option ARGV-elements.

   If the caller did not specify anything,
   the default is REQUIRE_ORDER if the environment variable
   POSIXLY_CORRECT is defined, PERMUTE otherwise.

   REQUIRE_ORDER means don't recognize them as options;
   stop option processing when the first non-option is seen.
   This is what Unix does.
   This mode of operation is selected by either setting the environment
   variable POSIXLY_CORRECT, or using `+' as the first character
   of the list of option characters.

   PERMUTE is the default.  We permute the contents of ARGV as we scan,
   so that eventually all the non-options are at the end.  This allows options
   to be given in any order, even with programs that were not written to
   expect this.

   RETURN_IN_ORDER is an option available to programs that were written
   to expect options and other ARGV-elements in any order and that care about
   the ordering of the two.  We describe each non-option ARGV-element
   as if it were the argument of an option with character code 1.
   Using `-' as the first character of the list of option characters
   selects this mode of operation.

   The special argument `--' forces an end of option-scanning regardless
   of the value of `opts_Ordering'.  In the case of RETURN_IN_ORDER, only
   `--' can cause `getopt' to return -1 with `opts_Ind' != ARGC.  */

static enum
{
  REQUIRE_ORDER, PERMUTE, RETURN_IN_ORDER
} opts_Ordering;

/* Value of POSIXLY_CORRECT environment variable.  */
static char *opts_PosixlyCorrect;

static int opts_FirstNonOpt;
static int opts_LastNonOpt;

/* Bash 2.0 gives us an environment variable containing flags
   indicating ARGV elements that should not be considered arguments.  */

static int opts_NonOptionFlagslen;


/* For communication from `getopt' to the caller.
   When `getopt' finds an option that takes an argument,
   the argument value is returned here.
   Also, when `opts_Ordering' is RETURN_IN_ORDER,
   each non-option ARGV-element is returned here.  */

static char *opts_Arg = NULL;

/* Index in ARGV of the next element to be scanned.
   This is used for communication to and from the caller
   and for communication between successive calls to `getopt'.

   On entry to `getopt', zero means this is the first call; initialize.

   When `getopt' returns -1, this is the index of the first of the
   non-option elements that the caller should itself scan.

   Otherwise, `optind' communicates from one call to the next
   how much of ARGV has been scanned so far.  */

/* 1003.2 says this must be 1 before any call.  */
static int opts_Ind = 1;

/* Formerly, initialization of getopt depended on opts_Ind==0, which
   causes problems with re-calling getopt as programs generally don't
   know that. */

static int opts_GetOptInitialized = 0;

/* The next char to be scanned in the option-element
   in which the last option character we returned was found.
   This allows us to pick up the scan where we left off.

   If this is zero, or a null string, it means resume the scan
   by advancing to the next ARGV-element.  */

static char *opts_NextChar;

/* Callers store zero here to inhibit the error message
   for unrecognized options.  */

static int opts_Err = 1;

/* Set to an option character which was unrecognized.
   This must be initialized on some systems to avoid linking in the
   system's own getopt implementation.  */

static int opts_Opt = '?';

/**************************************************************/
/* Forwarding functions                                       */
/**************************************************************/

static int str_ToInt(char* s, int* np);
static BOOL opts_IsAbbreviation(int c);
static BOOL opts_SetFlag(const char* FlagName, char* ValueString);
static char* opts_Abbreviates(int c);
static BOOL opts_AddParam(const char* FlagName, char* ValueString);

static int opts_GetOptLongOnly(int argc,
			    char *const *argv,
			    const char *options,
			    const struct OPTION *long_options,
			    int *opt_index);





/**************************************************************/
/* Functions                                                  */
/**************************************************************/



void opts_Init()
/**************************************************************
  INPUT:   None
  RETURNS: Nothing 
  EFFECT:  Initialize option module: converts SPASS flag 
           information to opts_Long information.
	   Set opts_Err for getopt module.
  MEMORY:  opts_Long array is initialized	   
***************************************************************/
{
  int i,j;

  opts_OPTNames[opts_RemoveFile] = "rf";

  for (i=0; i < flag_MAXFLAG; i++) {
    opts_Long[i].name = flag_Name(i);
    opts_Long[i].has_arg = opts_OptionalArgument;
    opts_Long[i].flag = 0;
    opts_Long[i].val= 0;  }

  /* options set by hand */
  for (j=0; j < opts_MaxOpts; j++) {
     opts_Long[i+j].name = opts_OPTNames[j];
     opts_Long[i+j].has_arg = opts_NoArgument; 
     opts_Long[i+j].flag = 0;
     opts_Long[i+j].val = 0;
  }

  i += j;

 /* set last field all zero (required by getopt) */
  opts_Long[i].name = (char *)NULL;
  opts_Long[i].has_arg = 0;
  opts_Long[i].flag = 0;
  opts_Long[i].val = 0;

  /* let getopt generate error messages */
  opts_Err = 1;
}

void opts_Free()
/**************************************************************
  INPUT:   None
  RETURNS: Nothing 
  EFFECT:  Initialize option module: converts SPASS flag 
           information to opts_Long information.
	   Set opts_Err for getopt module.
  MEMORY:  opts_Long array is initialized	   
***************************************************************/
{
  list_DeleteWithElement(opts_PARAMETERS,list_Delete);
  
}

void opts_PrintNames()
/**************************************************************
  INPUT: None 
  RETURNS: Nothing
  EFFECT:  Prints all options in three rows
  MEMORY:  None
***************************************************************/
{
  int i,j;
  
  for (i=0; i < flag_MAXFLAG; i=i+4) { 
    for (j =0; j <=3; j++) {
      if (i+j < flag_MAXFLAG)
	printf("%-15s ", flag_Name(i+j)); } 
    printf("\n");
  }
}

BOOL opts_Read(int argc, char* argv[])
/**************************************************************
  INPUT:   Program parameter data 
  RETURNS: TRUE iff options are correctly specified 
  EFFECT:  None
  MEMORY:  Builds up 'Parameters list'; does NOT influence
           SPASS flags (see opts_Transfer)
***************************************************************/
{
  int OptIndex,c;
  BOOL Ok;

  Ok = TRUE;

  while (Ok && (c = opts_GetOptLongOnly(argc, argv, opts_Short, 
			  opts_Long, &OptIndex)) != -1)

    {
      if (c == '?') { 

	/**** unknown option ****/ 

	/* fprintf(stderr, "\nERROR, unknown option %c\n", opts_Opt); */
	return(FALSE); }

      else if (c == 0)

	/**** its a long option ****/

	{  if (opts_Arg == NULL)
	    /* if no argument specified, interpret as flag */
	    Ok = opts_AddParam(opts_Long[OptIndex].name,opts_DEFAULTOPTARG);
	  else
	    Ok = opts_AddParam(opts_Long[OptIndex].name,opts_Arg);
	}

      else { 

	/**** its a short option ****/

	/* handle missing arguments: check if they are optional */
	if (c == ':') { 

	  /* arguments of all abbreviations are optional,  */
	  /* missing arguments interpreted as 1            */

	  if (opts_IsAbbreviation(opts_Opt))
	    Ok = opts_AddParam(opts_Abbreviates(opts_Opt),opts_DEFAULTOPTARG);  
	  /* here we could deal with other short options
	     with required but unspecified arguments */
	  else return(FALSE); }



	/* now we have short options without required arguments, or
	   short options with required arguments, and the
	   arguments are present. 

	   One further special case: if a short option has an optional
	   argument, but this argument is a '--', then take the
	   default argument value
	   */

	else {
	  /* check for abbreviations */
	  if (opts_IsAbbreviation(c)) {
	    if (strcmp(opts_Arg,opts_ENDMARKER) == 0)
	      Ok = opts_AddParam(opts_Abbreviates(c),opts_DEFAULTOPTARG);
	    else Ok = opts_AddParam(opts_Abbreviates(c),opts_Arg); }
	  else {
	    printf("\nfound option '%c'",c);
	    if (opts_Arg) 
	      printf("with argument %s !\n",  opts_Arg);
	    else printf(".\n"); }
	}

      }
    }
      
  return(Ok);
}
BOOL opts_IsSet(OPTS_OPTS Option)
/**************************************************************
  INPUT:   An option name
  RETURNS:  TRUE iff option has been set in command line
  EFFECT:  None
  MEMORY:  None
***************************************************************/
{
  LIST Scan;
  LIST Pair;
  BOOL found;
  

  found = FALSE;

  for (Scan = opts_PARAMETERS;  
       (!found && !list_Empty(Scan)); Scan = list_Cdr(Scan)) {
    Pair = list_Car(Scan);
    if (strcmp(opts_OPTNames[Option], list_Car(Pair)) == 0)
      found = TRUE; }

  return(found);
 }



void opts_Transfer()
/**************************************************************
  INPUT:   None  
  RETURNS: Nothing
  EFFECT:  Transfer options from 'opts_PARAMETERS' list into SPASS flags  
  MEMORY:  see EFFECT
***************************************************************/
{
  LIST Scan;
  LIST Pair;
  
  Scan = opts_PARAMETERS;

  while (!list_Empty(Scan))
    {
      Pair = list_Car(Scan);
      opts_SetFlag(list_First(Pair), list_Second(Pair));
      
      Scan = list_Cdr(Scan);
    }
}

static BOOL opts_AddParam(const char* FlagName, char* ValueString)
/**************************************************************
  INPUT:   A flag name and a string with its assigned value 
  RETURNS: TRUE iff value is an integer string
  EFFECT:  Add (FlagName, ValueString) tupel to 'opts_PARAMETERS' list
  MEMORY:  see EFFECT
***************************************************************/

{ 
  int Value;
  LIST Pair;
  BOOL Ok;

  /* check if argument is integer */
  Ok = str_ToInt(ValueString, &Value);
  /* build name/value tupel */
  if (Ok) { 
    Pair = list_Cons((char*)FlagName, list_List(ValueString));
    opts_PARAMETERS = list_Cons(Pair, opts_PARAMETERS); }
  else 
    fprintf(stderr, "error: found '%s' where integer required by option '%s'.\n", ValueString, FlagName); 
  return(Ok);
}


static BOOL opts_SetFlag(const char* FlagName, char* ValueString)
/**************************************************************
  INPUT:   A flag name and a string with its assigned value  
  RETURNS: TRUE iff value is an integer string
  EFFECT:  Sets SPASS flag 'FlagName' to value specified by
           'ValueString'
  MEMORY:  None
***************************************************************/
{
  int Id;
  int Value;
  BOOL Ok;

  Id = flag_Id(FlagName);
  
  Ok = str_ToInt(ValueString, &Value);
 
 if (Ok)
    flag_SetFlagValue(Id, Value);
  else 
    fprintf(stderr, "\n error, found %s where integer argument required.\n", ValueString); 
  
  return(Ok);
}

static BOOL opts_IsAbbreviation(int c)
/**************************************************************
  INPUT:   A character
  RETURNS: TRUE iff 'c' is an abbreviation of a long option,
           as specified by array 'opts_Abbreviations'
  EFFECT:  None
  MEMORY:  None
***************************************************************/

{ 
  int i;
 
  for (i = 0; opts_Abbreviations[i].abbrev != 0 && opts_Abbreviations[i].abbrev != c; i++) ;
  
  return(opts_Abbreviations[i].abbrev != 0);
}
static char* opts_Abbreviates(int c)
/**************************************************************
  INPUT:   A character
  RETURNS: Long option abbreviated by 'c'
  EFFECT:  None
  MEMORY:  None
***************************************************************/

{ 
  int i;
 
  for (i = 0; (opts_Abbreviations[i].abbrev != 0) && (opts_Abbreviations[i].abbrev != c); i++) ;
  
  return(opts_Abbreviations[i].name);
}


static int str_ToInt(char* s, int* np)
/**************************************************************
  INPUT: a string and an integer by reference            
  RETURNS: TRUE iff string represents some integer       
  EFFECTS: np takes value of represented integer         
**************************************************************/
{
    int i, sign, n;

    i = 0;
    sign = 1;
    if (s[0] == '+' || s[0] == '-') {
	if (s[0] == '-')
	    sign = -1;
	i = 1;
	}
    if (s[i] == '\0')
	return(0);
    else {
	n = 0;
	for( ; s[i] >= '0' && s[i] <= '9'; i++)
	    n = n * 10 + s[i] - '0';
	*np = n * sign;
	return(s[i] == '\0');
	}
}


int opts_Indicator() 
/**************************************************************
  INPUT:   None
  RETURNS: Integer variable indicating position of next argument
  EFFECT:  None
  MEMORY:  None
***************************************************************/
{ return(opts_Ind);}



static void opts_Exchange (char **argv)
/**************************************************************
  INPUT:   Reference to string
  RETURNS: Nothing
  EFFECT AND MEMORY: See below
***************************************************************/

/* Exchange two adjacent subsequences of ARGV.
   One subsequence is elements [opts_FirstNonOpt,opts_LastNonOpt)
   which contains all the non-options that have been skipped so far.
   The other is elements [opts_LastNonOpt,opts_Ind), which contains all
   the options processed since those non-options were skipped.

   `opts_FirstNonOpt' and `opts_LastNonOpt' are relocated so that they describe
   the new indices of the non-options in ARGV after they are moved.  */


{
  int bottom = opts_FirstNonOpt;
  int middle = opts_LastNonOpt;
  int top = opts_Ind;
  char *tem;

  /* Exchange the shorter segment with the far end of the longer segment.
     That puts the shorter segment into the right place.
     It leaves the longer segment in the right place overall,
     but it consists of two parts that need to be swapped next.  */

  while (top > middle && middle > bottom)
    {
      if (top - middle > middle - bottom)
	{
	  /* Bottom segment is the short one.  */
	  int len = middle - bottom;
	  register int i;

	  /* Swap it with the top part of the top segment.  */
	  for (i = 0; i < len; i++)
	    {
	      tem = argv[bottom + i];
	      argv[bottom + i] = argv[top - (middle - bottom) + i];
	      argv[top - (middle - bottom) + i] = tem;
	    }
	  /* Exclude the moved bottom segment from further swapping.  */
	  top -= len;
	}
      else
	{
	  /* Top segment is the short one.  */
	  int len = top - middle;
	  register int i;

	  /* Swap it with the bottom part of the bottom segment.  */
	  for (i = 0; i < len; i++)
	    {
	      tem = argv[bottom + i];
	      argv[bottom + i] = argv[middle + i];
	      argv[middle + i] = tem;
	    }
	  /* Exclude the moved top segment from further swapping.  */
	  bottom += len;
	}
    }

  /* Update records for the slots the non-options now occupy.  */

  opts_FirstNonOpt += (opts_Ind - opts_LastNonOpt);
  opts_LastNonOpt = opts_Ind;
}

static const char * opts_GetOptInitialize (int argc, char *const *argv, const char *optstring)
/**************************************************************
  INPUT:   The command line arguments ('argc', 'argv') and
           a string describing options ('optstring')
  RETURNS: a possibly modified 'optstring'
  EFFECT:  Several static variables related to options
           processing are influenced (see below).
  MEMORY:  None
***************************************************************/

{
  /* Start processing options with ARGV-element 1 (since ARGV-element 0
     is the program name); the sequence of previously skipped
     non-option ARGV-elements is empty.  */

  opts_FirstNonOpt = opts_LastNonOpt = opts_Ind = 1;

  opts_NextChar = NULL;

  opts_PosixlyCorrect = getenv ("POSIXLY_CORRECT");

  /* Determine how to handle the ordering of options and nonoptions.  */

  if (optstring[0] == '-')
    {
      opts_Ordering = RETURN_IN_ORDER;
      ++optstring;
    }
  else if (optstring[0] == '+')
    {
      opts_Ordering = REQUIRE_ORDER;
      ++optstring;
    }
  else if (opts_PosixlyCorrect != NULL)
    opts_Ordering = REQUIRE_ORDER;
  else
    opts_Ordering = PERMUTE;

    opts_NonOptionFlagslen = 0;

  return optstring;
}

int opts_GetOptInternal (int argc,
		    char *const *argv,
		    const char *optstring,
		    const struct OPTION *longopts,
		    int *longind,
		    int long_only)

/**************************************************************
  INPUT:   An array of pointers to arguments (strings),
           and format information. See below for extensive
	   description
  RETURNS: -1 only if there are no more options, otherwise
           a value identifying a read option. See below.
  EFFECT:  Affects statics opts_Ind, opts_Arg, opts_Opt,
           opts_NextChar and argument array. See below.
  MEMORY:  See below. 
***************************************************************/
/* Scan elements of ARGV (whose length is ARGC) for option characters
   given in OPTSTRING.

   If an element of ARGV starts with '-', and is not exactly "-" or "--",
   then it is an option element.  The characters of this element
   (aside from the initial '-') are option characters.  If `getopt'
   is called repeatedly, it returns successively each of the option characters
   from each of the option elements.

   If `getopt' finds another option character, it returns that character,
   updating `opts_Ind' and `opts_NextChar' so that the next call to `getopt' can
   resume the scan with the following option character or ARGV-element.

   If there are no more option characters, `getopt' returns -1.
   Then `opts_Ind' is the index in ARGV of the first ARGV-element
   that is not an option.  (The ARGV-elements have been permuted
   so that those that are not options now come last.)

   OPTSTRING is a string containing the legitimate option characters.
   If an option character is seen that is not listed in OPTSTRING,
   return '?' after printing an error message.  If you set `opts_Err' to
   zero, the error message is suppressed but we still return '?'.

   If a char in OPTSTRING is followed by a colon, that means it wants an arg,
   so the following text in the same ARGV-element, or the text of the following
   ARGV-element, is returned in `opts_Arg'.  Two colons mean an option that
   wants an optional arg; if there is text in the current ARGV-element,
   it is returned in `opts_Arg', otherwise `opts_Arg' is set to zero.

   If OPTSTRING starts with `-' or `+', it requests different methods of
   handling the non-option ARGV-elements.
   See the comments about RETURN_IN_ORDER and REQUIRE_ORDER, above.

   Long-named options begin with `--' instead of `-'.
   Their names may be abbreviated as long as the abbreviation is unique
   or is an exact match for some defined option.  If they have an
   argument, it follows the option name in the same ARGV-element, separated
   from the option name by a `=', or else the in next ARGV-element.
   When `getopt' finds a long-named option, it returns 0 if that option's
   `flag' field is nonzero, the value of the option's `val' field
   if the `flag' field is zero.

   The elements of ARGV aren't really const, because we permute them.
   But we pretend they're const in the prototype to be compatible
   with other systems.

   LONGOPTS is a vector of `struct OPTION' terminated by an
   element containing a name which is zero.

   LONGIND returns the index in LONGOPT of the long-named option found.
   It is only valid when a long-named option has been found by the most
   recent call.

   If LONG_ONLY is nonzero, '-' as well as '--' can introduce
   long-named options.  */


{
  opts_Arg = NULL;

  if (!opts_GetOptInitialized || opts_Ind == 0)
    {
      optstring = opts_GetOptInitialize (argc, argv, optstring);
      opts_Ind = 1;		/* Don't scan ARGV[0], the program name.  */
      opts_GetOptInitialized = 1;
    }

  /* Test whether ARGV[opts_Ind] points to a non-option argument.
     Either it does not have option syntax, or there is an environment flag
     from the shell indicating it is not an option.  The later information
     is only used when the used in the GNU libc.  */


  if (opts_NextChar == NULL || *opts_NextChar == '\0')
    {
      /* Advance to the next ARGV-element.  */

      /* Give OPTS_FIRSTNONOPT & LAST_NONOPT rational values if OPTS_IND has been
	 moved back by the user (who may also have changed the arguments).  */
      if (opts_LastNonOpt > opts_Ind)
	opts_LastNonOpt = opts_Ind;
      if (opts_FirstNonOpt > opts_Ind)
	opts_FirstNonOpt = opts_Ind;

      if (opts_Ordering == PERMUTE)
	{
	  /* If we have just processed some options following some non-options,
	     exchange them so that the options come first.  */

	  if (opts_FirstNonOpt != opts_LastNonOpt && opts_LastNonOpt != opts_Ind)
	    opts_Exchange ((char **) argv);
	  else if (opts_LastNonOpt != opts_Ind)
	    opts_FirstNonOpt = opts_Ind;

	  /* Skip any additional non-options
	     and extend the range of non-options previously skipped.  */

	  while (opts_Ind < argc &&  (argv[opts_Ind][0] != '-' || argv[opts_Ind][1] == '\0'))
	    opts_Ind++;
	  opts_LastNonOpt = opts_Ind;
	}

      /* The special ARGV-element `--' means premature end of options.
	 Skip it like a null option,
	 then exchange with previous non-options as if it were an option,
	 then skip everything else like a non-option.  */

      if (opts_Ind != argc && !strcmp (argv[opts_Ind], "--"))
	{
	  opts_Ind++;

	  if (opts_FirstNonOpt != opts_LastNonOpt && opts_LastNonOpt != opts_Ind)
	    opts_Exchange ((char **) argv);
	  else if (opts_FirstNonOpt == opts_LastNonOpt)
	    opts_FirstNonOpt = opts_Ind;
	  opts_LastNonOpt = argc;

	  opts_Ind = argc;
	}

      /* If we have done all the ARGV-elements, stop the scan
	 and back over any non-options that we skipped and permuted.  */

      if (opts_Ind == argc)
	{
	  /* Set the next-arg-index to point at the non-options
	     that we previously skipped, so the caller will digest them.  */
	  if (opts_FirstNonOpt != opts_LastNonOpt)
	    opts_Ind = opts_FirstNonOpt;
	  return -1;
	}

      /* If we have come to a non-option and did not permute it,
	 either stop the scan or describe it to the caller and pass it by.  */

      if ( (argv[opts_Ind][0] != '-' || argv[opts_Ind][1] == '\0'))
	{
	  if (opts_Ordering == REQUIRE_ORDER)
	    return -1;
	  opts_Arg = argv[opts_Ind++];
	  return 1;
	}

      /* We have found another option-ARGV-element.
	 Skip the initial punctuation.  */

      opts_NextChar = (argv[opts_Ind] + 1
		  + (longopts != NULL && argv[opts_Ind][1] == '-'));
    }

  /* Decode the current option-ARGV-element.  */

  /* Check whether the ARGV-element is a long option.

     If long_only and the ARGV-element has the form "-f", where f is
     a valid short option, don't consider it an abbreviated form of
     a long option that starts with f.  Otherwise there would be no
     way to give the -f short option.

     On the other hand, if there's a long option "fubar" and
     the ARGV-element is "-fu", do consider that an abbreviation of
     the long option, just like "--fu", and not "-f" with arg "u".

     This distinction seems to be the most useful approach.  */

  if (longopts != NULL
      && (argv[opts_Ind][1] == '-'
	  || (long_only && (argv[opts_Ind][2] || !strchr (optstring, argv[opts_Ind][1])))))
    {
      char *nameend;
      const struct OPTION *p;
      const struct OPTION *pfound = NULL;
      int exact = 0;
      int ambig = 0;
      int indfound = -1;
      int option_index;

      for (nameend = opts_NextChar; *nameend && *nameend != '='; nameend++)
	/* Do nothing.  */ ;

      /* Test all long options for either exact match
	 or abbreviated matches.  */
      for (p = longopts, option_index = 0; p->name; p++, option_index++)
	if (!strncmp (p->name, opts_NextChar, nameend - opts_NextChar))
	  {
	    if ((unsigned int) (nameend - opts_NextChar)
		== (unsigned int) strlen (p->name))
	      {
		/* Exact match found.  */
		pfound = p;
		indfound = option_index;
		exact = 1;
		break;
	      }
	    else if (pfound == NULL)
	      {
		/* First nonexact match found.  */
		pfound = p;
		indfound = option_index;
	      }
	    else
	      /* Second or later nonexact match found.  */
	      ambig = 1;
	  }

      if (ambig && !exact)
	{
	  if (opts_Err)
	    fprintf (stderr, ("%s: option `%s' is ambiguous\n"),
		     argv[0], argv[opts_Ind]);
	  opts_NextChar += strlen (opts_NextChar);
	  opts_Ind++;
	  opts_Opt = 0;
	  return '?';
	}

      if (pfound != NULL)
	{
	  option_index = indfound;
	  opts_Ind++;
	  if (*nameend)
	    {
	      /* Don't test has_arg with >, because some C compilers don't
		 allow it to be used on enums.  */
	      if (pfound->has_arg)
		opts_Arg = nameend + 1;
	      else
		{
		  if (opts_Err)
		   if (argv[opts_Ind - 1][1] == '-')
		    /* --option */
		    fprintf (stderr,
		     ("%s: option `--%s' doesn't allow an argument\n"),
		     argv[0], pfound->name);
		   else
		    /* +option or -option */
		    fprintf (stderr,
		     ("%s: option `%c%s' doesn't allow an argument\n"),
		     argv[0], argv[opts_Ind - 1][0], pfound->name);

		  opts_NextChar += strlen (opts_NextChar);

		  opts_Opt = pfound->val;
		  return '?';
		}
	    }
	  else if (pfound->has_arg == 1)
	    {
	      if (opts_Ind < argc)
		opts_Arg = argv[opts_Ind++];
	      else
		{
		  if (opts_Err)
		    fprintf (stderr,
			   ("%s: option `%s' requires an argument\n"),
			   argv[0], argv[opts_Ind - 1]);
		  opts_NextChar += strlen (opts_NextChar);
		  opts_Opt = pfound->val;
		  return optstring[0] == ':' ? ':' : '?';
		}
	    }
	  opts_NextChar += strlen (opts_NextChar);
	  if (longind != NULL)
	    *longind = option_index;
	  if (pfound->flag)
	    {
	      *(pfound->flag) = pfound->val;
	      return 0;
	    }
	  return pfound->val;
	}

      /* Can't find it as a long option.  If this is not opts_GetOptLong_only,
	 or the option starts with '--' or is not a valid short
	 option, then it's an error.
	 Otherwise interpret it as a short option.  */
      if (!long_only || argv[opts_Ind][1] == '-'
	  || strchr (optstring, *opts_NextChar) == NULL)
	{
	  if (opts_Err)
	    {
	      if (argv[opts_Ind][1] == '-')
		/* --option */
		fprintf (stderr, ("%s: unrecognized option `--%s'\n"),
			 argv[0], opts_NextChar);
	      else
		/* +option or -option */
		fprintf (stderr, ("%s: unrecognized option `%c%s'\n"),
			 argv[0], argv[opts_Ind][0], opts_NextChar);
	    }
	  opts_NextChar = (char *) "";
	  opts_Ind++;
	  opts_Opt = 0;
	  return '?';
	}
    }

  /* Look at and handle the next short option-character.  */

  {
    char c = *opts_NextChar++;
    char *temp = strchr (optstring, c);

    /* Increment `opts_Ind' when we start to process its last character.  */
    if (*opts_NextChar == '\0')
      ++opts_Ind;

    if (temp == NULL || c == ':')
      {
	if (opts_Err)
	  {
	    if (opts_PosixlyCorrect)
	      /* 1003.2 specifies the format of this message.  */
	      fprintf (stderr, ("%s: illegal option -- %c\n"),
		       argv[0], c);
	    else
	      fprintf (stderr, ("%s: invalid option -- %c\n"),
		       argv[0], c);
	  }
	opts_Opt = c;
	return '?';
      }
    /* Convenience. Treat POSIX -W foo same as long option --foo */
    if (temp[0] == 'W' && temp[1] == ';')
      {
	char *nameend;
	const struct OPTION *p;
	const struct OPTION *pfound = NULL;
	int exact = 0;
	int ambig = 0;
	int indfound = 0;
	int option_index;

	/* This is an option that requires an argument.  */
	if (*opts_NextChar != '\0')
	  {
	    opts_Arg = opts_NextChar;
	    /* If we end this ARGV-element by taking the rest as an arg,
	       we must advance to the next element now.  */
	    opts_Ind++;
	  }
	else if (opts_Ind == argc)
	  {
	    if (opts_Err)
	      {
		/* 1003.2 specifies the format of this message.  */
		fprintf (stderr, ("%s: option requires an argument -- %c\n"),
			 argv[0], c);
	      }
	    opts_Opt = c;
	    if (optstring[0] == ':')
	      c = ':';
	    else
	      c = '?';
	    return c;
	  }
	else
	  /* We already incremented `opts_Ind' once;
	     increment it again when taking next ARGV-elt as argument.  */
	  opts_Arg = argv[opts_Ind++];

	/* opts_Arg is now the argument, see if it's in the
	   table of longopts.  */

	for (opts_NextChar = nameend = opts_Arg; *nameend && *nameend != '='; nameend++)
	  /* Do nothing.  */ ;

	/* Test all long options for either exact match
	   or abbreviated matches.  */
	for (p = longopts, option_index = 0; p->name; p++, option_index++)
	  if (!strncmp (p->name, opts_NextChar, nameend - opts_NextChar))
	    {
	      if ((unsigned int) (nameend - opts_NextChar) == strlen (p->name))
		{
		  /* Exact match found.  */
		  pfound = p;
		  indfound = option_index;
		  exact = 1;
		  break;
		}
	      else if (pfound == NULL)
		{
		  /* First nonexact match found.  */
		  pfound = p;
		  indfound = option_index;
		}
	      else
		/* Second or later nonexact match found.  */
		ambig = 1;
	    }
	if (ambig && !exact)
	  {
	    if (opts_Err)
	      fprintf (stderr, ("%s: option `-W %s' is ambiguous\n"),
		       argv[0], argv[opts_Ind]);
	    opts_NextChar += strlen (opts_NextChar);
	    opts_Ind++;
	    return '?';
	  }
	if (pfound != NULL)
	  {
	    option_index = indfound;
	    if (*nameend)
	      {
		/* Don't test has_arg with >, because some C compilers don't
		   allow it to be used on enums.  */
		if (pfound->has_arg)
		  opts_Arg = nameend + 1;
		else
		  {
		    if (opts_Err)
		      fprintf (stderr, ("\
%s: option `-W %s' doesn't allow an argument\n"),
			       argv[0], pfound->name);

		    opts_NextChar += strlen (opts_NextChar);
		    return '?';
		  }
	      }
	    else if (pfound->has_arg == 1)
	      {
		if (opts_Ind < argc)
		  opts_Arg = argv[opts_Ind++];
		else
		  {
		    if (opts_Err)
		      fprintf (stderr,
			       ("%s: option `%s' requires an argument\n"),
			       argv[0], argv[opts_Ind - 1]);
		    opts_NextChar += strlen (opts_NextChar);
		    return optstring[0] == ':' ? ':' : '?';
		  }
	      }
	    opts_NextChar += strlen (opts_NextChar);
	    if (longind != NULL)
	      *longind = option_index;
	    if (pfound->flag)
	      {
		*(pfound->flag) = pfound->val;
		return 0;
	      }
	    return pfound->val;
	  }
	  opts_NextChar = NULL;
	  return 'W';	/* Let the application handle it.   */
      }
    if (temp[1] == ':')
      {
	if (temp[2] == ':')
	  {
	    /* This is an option that accepts an argument optionally.  */
	    if (*opts_NextChar != '\0')
	      {
		opts_Arg = opts_NextChar;
		opts_Ind++;
	      }
	    else
	      opts_Arg = NULL;
	    opts_NextChar = NULL;
	  }
	else
	  {
	    /* This is an option that requires an argument.  */
	    if (*opts_NextChar != '\0')
	      {
		opts_Arg = opts_NextChar;
		/* If we end this ARGV-element by taking the rest as an arg,
		   we must advance to the next element now.  */
		opts_Ind++;
	      }
	    else if (opts_Ind == argc)
	      {
		if (opts_Err)
		  {
		    /* 1003.2 specifies the format of this message.  */
		    fprintf (stderr,
			   ("%s: option requires an argument -- %c\n"),
			   argv[0], c);
		  }
		opts_Opt = c;
		if (optstring[0] == ':')
		  c = ':';
		else
		  c = '?';
	      }
	    else
	      /* We already incremented `opts_Ind' once;
		 increment it again when taking next ARGV-elt as argument.  */
	      opts_Arg = argv[opts_Ind++];
	    opts_NextChar = NULL;
	  }
      }
    return c;
  }
}


/* Like opts_GetOptLong, but '-' as well as '--' can indicate a long option.
   If an option that starts with '-' (not '--') doesn't match a long option,
   but does match a short option, it is parsed as a short option
   instead.  */

static int opts_GetOptLongOnly(int argc,
			    char *const *argv,
			    const char *options,
			    const struct OPTION *long_options,
			    int *opt_index)
/**************************************************************
  FUNCTIONALITY: See opts_GetOptInternal
***************************************************************/
{
  return opts_GetOptInternal (argc, argv, options, long_options, opt_index, 1);
}

