;===========================================================
;
;                          resio.scm
;    
;                 Input-output and auxiliary funs
;               of the resolution proof searcher.
;  
;  Other files: resbase.scm
;              (basic operations, incl subst. unif. and subsumption)
;               resmain.scm 
;
;====================================================================


;===================================================================
;
;              creating the input clause set
;
;===================================================================

; make-input-clauses takes a raw NUMERIFIED clause set

; and builds an input clause list 

;(define (make-input-clauses cllist)
;  (set! *kept-cl-nr* 0)
;  (set! cllist (numerify-clause-list cllist))
;  (fill-predsymb-bit-vector!)
;  (set! cllist (map make-input-clause cllist)) 
;  cllist)

; make-input-clause takes an undecorated numerified clause
; and returns a decorated input clause

(define (make-input-clause cl)

  ;(newline)(display "make-input-clause entered with: ")(newline)
  ;(print-pure-clause cl)(newline)

  
  (set! cl (remove-duplicates cl))
  (let ((litinfolst (make-input-litinfolst cl))
        (tmp '()) )

    ;(newline)(display "litinfolst made: ")(display litinfolst)(newline)

    (set-x-vars-selfref-touched! *max-nr-of-vars-in-clause*)
    (set! cl (map (lambda(x)(subst-deco-term x #t)) cl)) 

    (set! *used-demodulators* '())
    (set! tmp (sort-clause (map cons litinfolst cl) #t))
    (set! cl  (map cdr tmp))
    (set! litinfolst (map car tmp))

    ;(newline)(display "clause after sorting: ")(newline)
    ;(print-pure-clause cl)(newline)
    ;(newline)(display "litinfolst after sorting: ")(display litinfolst)
    ;(newline)

    (set! *clause-var-nr* 0)
    (if (pure-clause-may-cont-vars? cl)
	(begin 
	  (set! cl (copy-list cl))
	  ;create a list of lists (lit litrepvars litvars othervars follvars):
	  (make-varinfo-lst! cl (list '())) ; cl is destr replaced to be varinfo-lst
	  (set! *new-var-names* '())
	  (make-fulldeco-clause! cl) )) ; cl is destr replaced to be a clause

    ;(newline)(debug-print-pure-clause cl)(display *clause-var-nr*) (newline)

    ; cl is now the finished pure clause.
    ; *clause-var-nr* is now the max var nr in the final pure clause
    ; start making the info-part.
    (let ((strengthint 0)
	  (len 0)
	  (predsymbsint 0)
	  (weight 0) 
	  (clinfo '()) )
      (do ((clpart cl (cdr clpart)))
	  ((null? clpart))
	(let* ((lit (car clpart))
	       (fun (car lit)) )
	  (if (countable-pred? fun) ; answer-preds may be uncountable
	      (begin
		(if (zero? strengthint)
		    (set! strengthint (logand *strength-mask* fun))
		    (set! strengthint (sum-fun-strengths fun strengthint)) )
		(if (not (zero? (logand (predsymb-bit fun) predsymbsint)))
		    (set! predsymbsint 
			  (logior *cl-rep-predsymbs-mask* predsymbsint) ))
		(set! predsymbsint (logior (predsymb-bit fun) predsymbsint))
		(set! len (+ 1 len))
		(set! weight (+ weight (lit-weight lit))) ))))
      (if (> len *cl-maxlen*) (set! len *cl-maxlen*))
      (if (> *clause-var-nr* *cl-maxvar*)(set! *clause-var-nr* *cl-maxvar*))
      (set! strengthint
	    (logior (logior (shiftleft len *cl-len-shift*)
			    (shiftleft *clause-var-nr* *cl-maxvar-shift*) )
		    strengthint))
      (set! *kept-cl-nr* (+ 1 *kept-cl-nr*))
      (if (or *order-strategy-flag*
	      *name-order-strategy-flag*
              *firstlitpref-order-strategy-flag*
	      *posweight-order-strategy-flag*
	      *ordered-paramodulation-flag*)
	  (set! litinfolst
		(map cddr (order-prohibit-litinfos 
			    (map (lambda(x)(cons '() (cons '() x)))
				 litinfolst)
			    cl))))
      
      (set! clinfo
	    (list strengthint
		  predsymbsint
		  litinfolst
                  weight
		  *kept-cl-nr*
		  (list 'input *kept-cl-nr*) ))
      
      (set! tmp (cons clinfo cl))
      (if *dynamic-demodulation-flag*
 	(attempt-to-make-a-demodulator tmp) )
      tmp)))    


(define (remove-duplicates lst)
  (cond 
   ((null? lst) '())
   ((member (car lst) (cdr lst)) 
      (remove-duplicates (cdr lst)) )
   (else
      (cons (car lst) (remove-duplicates (cdr lst))) )))

(define (make-input-litinfolst cl)
  (cond
   ((not *hyper-strategy-flag*)	  
      (make-input-litinfolist-aux cl) )
   ((positive-cl? cl)
      (make-input-litinfolist-aux cl) )
   ; nucleus
   (else
      (let ((nr 1)
	    (res '()) )
	(do ((clpart cl (cdr clpart)))
	    ((null? clpart) (reverse res))
	  (if (neg-fun? (caar clpart))
	      (begin (set! res (cons (list (put-litinfo-ordernr
					    nr *litinf-noparainto-mask*))
				     res))
		     (set! nr (+ 1 nr)) )
	      (set! res (cons (make-pos-lit-litinfo (car clpart))
			      res))))))))


(define (make-input-litinfolist-aux cl)
  (map make-pos-lit-litinfo cl) )

(define (make-pos-lit-litinfo x)
  (cond ((answer-lit? x) 
	   (list *answer-litinf*) )
	((and (eq? *eq-pred-nr* (pure-fun (car x)))
	      (equal? (eq-lit-arg1 x) (eq-lit-arg2 x)) )
	   ; eq reflexivity axiom
	   (list (logior *litinf-noparainto-mask*
			 (logior *litinf-noparafromleft-mask*
				 *litinf-noparafromright-mask*))))
	((and (composite? (cadr x))
	      (eq? *eq-fun-nr* (pure-fun (caadr x)))
	      (equal? (eq-lit-arg1 (cadr x)) (eq-lit-arg2 (cadr x))) )
	   ; eq-fun reflexivity axiom
	   (list (logior *litinf-noparainto-mask*
			 (logior *litinf-noparafromleft-mask*
				 *litinf-noparafromright-mask*))))
	(else
	   ; everything allowed
	   (list 0) )))

; make-varinfo-lst! takes a pure clause and replaces it 
; destructively with a list 
; (lit litrepvars litvars othervars follvars)
; or for the ground literals:
; (lit)
; it returns the lists of all lists of vars in clauses.
; the second arg is the list of lists of vars in previous clauses
;
; answer-literals are treated specially: to the rest of the clause
; it seems that the answer-literal is ground.

(define (make-varinfo-lst! cl prevars)
  (if (null? cl)
      '()
      (let* ((lit (car cl))
	     (fun (car lit))
	     (lcvars
	       (if (fun-may-cont-vars? fun) 
		   (collect-vars-composite! lit)
		   (cons '() '()) ))
	     (follvars '()) )
	(if (answer-pred? fun)
	    (set! follvars
	          (make-varinfo-lst! (cdr cl) (cons '() prevars)) )
	    (set! follvars
	          (make-varinfo-lst! (cdr cl) (cons (cdr lcvars) prevars)) ))
	(if (fun-may-cont-vars? fun)
	    (set-car! cl 
		      (list lit
			    (car lcvars)
			    (cdr lcvars)
			    (append prevars follvars)
			    follvars))
	    (set-car! cl (list lit)) )
	(if (answer-pred? fun)
	    (cons '() follvars) 
	    (cons (cdr lcvars) follvars) ))))

; make-fulldeco-clause! takes a list of elements in the form
; (lit litrepvars litvars othervars follvars)
; and replaces its elements destructively by the fully
; decorated literals.

(define (make-fulldeco-clause! reslst)

  ;(newline)(display "make-fulldeco-clause! entered with: ")(newline)
  ;(display reslst)(newline)
  ;(display *clause-var-nr*)(newline)
  ;(display *new-var-names*)(newline)(newline)
  ;(for-each (lambda(x)(debug-print-lit (car x))(newline)
  ;		      (display (cdr x))(newline) )
  ;	    reslst)

  (if (not (null? reslst))
      (let* ((el (car reslst)))
	(if (term-may-cont-vars? (car el))
	    (decorate-vars-lit! 
	      (car el) ;lit
	      (cadr el) ;repeating vars in the given literal
	      (caddr el) ;all vars in the given literal
	      (caddr (cdr el)) ;list of var lists of the other literals
	      (caddr (cddr el)) )) ;list of var lists of the foll literals
	(set-car! reslst (car el)) ; el has been destructively changed
	(make-fulldeco-clause! (cdr reslst)) )))

; fill-predsymb-bit-vector uses the *pred-num-table* created by
; numerify-literal during numerify-clause-list. 
; It is an alist ((puresymbol . nr) ...) where nr-s start from
; 1 (this is assumed!)
; if the ans-pred and equality pred are present, they are always taken
; into the first 13 bit-representable predsymbs.
;
; It also sets the value for *ans-pred-nr* and *eq-pred-nr*.

(define *ans-pred-nr* 0)
(define *eq-pred-nr* 0)
(define *eq-fun-nr* 0)
(define *case-fun-nr* 0)
(define *if-fun-nr* 0)

(define *subst-fun-nr* 0)
(define *subst-prim-fun-nr* 0)
(define *lambda-fun-nr* 0)
(define *lambda-dummypar-nr* 0)

(define *int-outmost-symb-labelfun* '())
(define *int-outmost-label-fun-nr* 0)

(define (pos-eq-pred? fn)
  (eq? *eq-pred-nr* fn) )

(define (pos-eq-fun? fn)
  (eq? *eq-fun-nr* fn) )


(define (fill-predsymb-bit-vector!) 
  (let ((a (assq *ans-symbol* *pred-num-table*))
	(e (assq *eq-symbol* *pred-num-table*)) 
	(ef (assq *eq-fun-symbol* *fun-num-table*)) 
        (cf (assq *case-fun-symbol* *fun-num-table*)) 
	(ifs (assq *if-fun-symbol* *fun-num-table*))
	(intlabel 
	  (if *int-logic-flag*
	      (assq *int-outmost-symb-labelfun* *pred-num-table*)
	      #f)))
    (set! *ans-pred-nr* 0)
    (set! *eq-pred-nr* 0)
    (set! *eq-fun-nr* 0)
    (set! *case-fun-nr* 0)
    (set! *if-fun-nr* 0)
    (set! *int-outmost-label-fun-nr* 0)
    (if e
	(begin 
	  (set! *pred-num-table*
		(cons e (remove-eq e *pred-num-table*)) )
	  (set! *eq-pred-nr* (cdr e)) ))
    (if a
	(begin
	  (set! *pred-num-table*
		(cons a (remove-eq a *pred-num-table*)) )
	  (set! *ans-pred-nr* (cdr a)) ))
    (if ef (set! *eq-fun-nr* (cdr ef))) 
    (if cf (set! *case-fun-nr* (cdr cf)))
    (if ifs (set! *if-fun-nr* (cdr ifs)))
    (if intlabel (set! *int-outmost-label-fun-nr* (cdr intlabel)))
    (do ((part *pred-num-table* (cdr part))
	 (n 1 (if (< n *cl-predsymbs-pos-max*)
		  (shiftleft n 1) 
		  *cl-pred-overflow-mask*)))
	((null? part))
      (vector-set! *predsymb-bit-vector* 
		       (shiftright (cdar part) *fnsymb-shift*)
		       n)
      (vector-set! *predsymb-bit-vector* 
		       (shiftright (logior *sgn-mask* (cdar part))
				   *fnsymb-shift*)
		       (if (eq? n *cl-pred-overflow-mask*)
			   *cl-pred-overflow-mask*
			   (shiftleft n *cl-predsymbs-neg-shift*) )))))


;===================================================================
;
;              strategies and strategical flags
;
;===================================================================


;;; NB! Currently *ordered-factor-flag* works OK only for input sos,
;;;     otherwise it is like always #t

(define (turn-on-demodulation)
  (set! *demodulate-flag* #t)
  (set! *dynamic-demodulation-flag* #t) 
  ;(set! *new-demodulators-flag* #f)
  (set! *new-demodulators-flag* #t)
  (set! *demodulate-constants-flag* #f) 
  (set! *ordered-paramodulation-flag* #t) )


(define (turn-off-demodulation)
  (set! *demodulate-flag* #f)
  (set! *dynamic-demodulation-flag* #f) 
  (set! *new-demodulators-flag* #f)
  (set! *demodulate-constants-flag* #f) )
  ; (set! *ordered-paramodulation-flag* #f) 


(define (init-binres-strat)
  (set! *unit-strategy-flag* #f)
  (set! *user-criterias-post-flag* #f)
  (set! *factors-stratflag* #t)
  (set! *ordered-factor-flag* #f) 
  (set! *name-order-strategy-flag* #f)
  (set! *firstlitpref-order-strategy-flag* #f)
  (set! *remove-identical-lit-stratflag* #t)
  (set! *paramodulation-allowed-flag* #t)
  (set! *hyper-strategy-flag* #f)
  (set! *type-theory-stratflag* #f)
  (set! *para-pos-unit-strategy* #f) )

(define (init-binary-hyper-sos-strat)
   (init-binres-strat)
   (set! *binary-hyper-sos-strategy* #t) )

(define (init-unit-strat)
  (init-binres-strat)
  (set! *unit-strategy-flag* #t) )

(define (init-posweight-ordered-unit-strat)
  (init-unit-strat)
  (set! *posweight-order-strategy-flag* #t) 
  (set! *ordered-factor-flag* #f) )

(define (init-ordered-binres-strat)
  (init-binres-strat)
  (set! *order-strategy-flag* #t) 
  (set! *name-order-strategy-flag* #f) 
  (set! *ordered-factor-flag* #f) )

(define (init-name-ordered-binres-strat)
  (init-binres-strat)
  (set! *order-strategy-flag* #f) 
  (set! *name-order-strategy-flag* #t) 
  (set! *ordered-factor-flag* #f) )

;(define (init-ordered-binres-0-strat)
;  (init-binres-strat)
;  (set! *order-strategy-flag* #t) 
;  (set! *name-order-strategy-flag* #f) 
;  (set! *ordered-factor-flag* #f) )

;(define (init-name-ordered-binres-0-strat)
;  (init-binres-strat)
;  (set! *order-strategy-flag* #f) 
;  (set! *name-order-strategy-flag* #t) 
;  (set! *ordered-factor-flag* #f) )

(define (init-weight-ordered-binres-strat)
  (init-binres-strat)
  (set! *firstlitpref-order-strategy-flag* #t) 
  (set! *ordered-factor-flag* #f) )

(define (init-posweight-ordered-binres-strat)
  (init-binres-strat)
  (set! *posweight-order-strategy-flag* #t) 
  (set! *ordered-factor-flag* #f) )


(define (init-hyper-strat)
  (set! *unit-strategy-flag* #f)
  (set! *user-criterias-post-flag* #f)
  (set! *factors-stratflag* #t)
  (set! *ordered-factor-flag* #f)
  (set! *name-order-strategy-flag* #f)
  (set! *remove-identical-lit-stratflag* #t)
  (set! *paramodulation-allowed-flag* #t)
  (set! *hyper-strategy-flag* #t)
  (set! *type-theory-stratflag* #f)
  (set! *para-pos-unit-strategy* #t) )

(define (init-hyper-wos-strat)
   (init-hyper-strat)
   (set! *hyper-wos-strategy-flag* #t) )


(define (init-ordered-hyper-strat)
  (init-hyper-strat)
  (set! *order-strategy-flag* #t) )

(define (init-typet-strat)
  (set! *user-criterias-post-flag* #f)
  (set! *factors-stratflag* #f)
  (set! *remove-identical-lit-stratflag* #f)
  (set! *paramodulation-allowed-flag* #t)
  (set! *hyper-strategy-flag* #t)
  (set! *type-theory-stratflag* #t)
  (set! *para-pos-unit-strategy* #t) )

(define (init-synth-strat)
  (set! *user-criterias-post-flag* #f)
  (set! *factors-stratflag* #t)
  (set! *ordered-factor-flag* #f) 
  (set! *paramodulation-allowed-flag* #t)
  (set! *remove-identical-lit-stratflag* #t)
  (set! *hyper-strategy-flag* #f)
  (set! *type-theory-stratflag* #f)
  (set! *para-pos-unit-strategy* #f) 
  (set! *classical-synth-stratflag* #t) )

(define (init-binary-unit-depth-backdemod)
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #t)
  (set! *order-equality-lex-strat* #f)
  (set! *sos-preference-ratio* 0) )

(define (init-binary-unit-depth)
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-lex-strat* #f)
  (set! *sos-preference-ratio* 0) )

(define (init-binary-unit-depth-sospref)
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-lex-strat* #f)
  (set! *sos-preference-ratio* 0) )

(define (init-binary-unit-lex) 
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-lex-strat* #t) )

(define (init-binary-unit-lex-backdemod)
  (init-binres-strat)
  (turn-on-demodulation)
  (set! *unit-equality-strat* #t)
  (set! *back-demodulate-flag* #t)
  (set! *order-equality-lex-strat* #t) )


; - - - - - - -- - - - - - - strategical flags - - - - - - - - - -

(define *split-strategy-flag* #f) 
(define *unit-equality-strat* #f)
(define *order-equality-lex-strat* #f) ; use depth- or lex-ordering
(define *sos-strategy-flag* #t) ; locally used flag: determines
    ; whether sos strategy is used in a pass
(define *keep-all-old-clauses-flag* #f) ; if #t, all clauses from old
       ; passes are kept for history
(define *auto-mode-flag* #f) ; if #t, automatic strategy selection

(define *stop-res-real-time* 0) ; time to stop res. If = 0, then
                                ; no stopping 

(define *max-kept-clauses-nr* 20000) ; the maximal number of kept clauses

(define *display-info-level* 50) ; gradation:
   ; 20 : derivation found/not
   ; 30 : history shown
   ; 35 : statistics shown, 
   ; 38 : strategy changes shown
   ; 40 : input clauses in final form
   ; 50 : given clauses
   ; 60 : kept clauses
   ; 63 ; back subsumption 
   ; 65 : cutoff literals and clauses
   ; 70 : raw clauses
   ; 80 : removal reasons
   ; 90 : usable clauses
   
(define *display-info-char-flag* #f) ; special chars are output to trace search
(define *display-debug-info-flag* #f) ; debug forms of clauses also shown

(define *remove-all-tautologies-stratflag* #t)

(define *hyper-wos-strategy-flag* #f) ;
(define *binary-hyper-sos-strategy* #f) ; first binary sos, then hyper
(define *hyper-contradiction-strategy* #f) ; iff #t, then kept clauses are
  ; immediately checked for full unit clash contradiction
(define *unit-clash-flag* #f) ; #t only during hyper-contradiction check

(define *factors-stratflag* #f)
(define *back-subsume-stratflag* #t)
(define *havealready-timecounter* 0) ; counts the times havealready called:
                                     ; used for avoiding double work
(define *remove-identical-lit-stratflag* #t)
(define *paramodulation-allowed-flag* #t)   ; for user control only
(define *paramodulation-strategy-flag* #t)  ; actually used during search
(define *paramodulate-from-vars-stratflag* #f)
(define *para-pos-unit-strategy* #f)
(define *unit-strategy-flag* #f)
(define *hyper-strategy-flag* #f)
(define *order-strategy-flag* #f)
(define *name-order-strategy-flag* #f)
(define *firstlitpref-order-strategy-flag* #f)
(define *poslitpref-order-strategy-flag* #f)
(define *posweight-order-strategy-flag* #f)
(define *ordered-factor-flag* #f)
(define *neg-hyper-strategy-flag* #f)
(define *type-theory-stratflag* #f)
(define *int-logic-flag* #f)
(define *int-use-binres-flag* #f)
(define *classical-synth-stratflag* #f)
(define *int-use-hyperres-flag* #f)

(define *discr-tree-subsume-stratflag* #t)

(define *discr-depth-first* #t) ;;; a search-guiding flag for discr-tree
                                ;;; unit subsumption. May be #t or #f.

(define *ordered-paramodulation-flag* #t) ;;; indicates whether
                              ;;; positive equalities are ordered

(define *dynamic-demodulation-flag* #f) ;;; indicates whether 
                              ;;; dynamic demodulation is attempted
                              ;;; (this requires also the *ordered-param...
                              ;;; to be #t
(define *demodulate-lit-flag* #f) ;;; indicates whether lit demodulation
                              ;;;should be
                              ;;; attempted on initial and derived clauses
(define *back-demodulate-flag* #f)
(define *demodulate-flag* #f) ;;; indicates whether term demodulation should be
                              ;;; attempted on initial and derived clauses
(define *demodulate-constants-flag* #f) ;;; demodulation applied
       ;;; to constants? If *demodulate-flag* = #f, it must be #f too
(define *new-demodulators-flag* #f) ;;; demodulation to be applied
       ;;; to unchanged terms (ie ground)?
       ;;; If *demodulate-flag* = #f, it must be #f too


(define *user-criterias-post-flag* #f) ; set by the prover itself
(define *user-criterias-pre-flag* #f) ; set by the prover itself

(define *clause-usenr-limits* '())
(define *clause-usedepth-limits* '())
(define *clause-usestruct-limits* '())
(define *clause-keepsize-limit* 0)
(define *clause-keepdepth-limit* 0)
(define *clause-uselength-limit* 0)
(define *clause-keeplength-limit* 0)

(define *special-insertion-place* '()) ; used in back-demodulation

(define (init-one-pass-strategical-flags)
  (set! *sos-preference-ratio* 5)
  (set! *split-strategy-flag* #f) 
  (set! *sos-strategy-flag* #t)
  (set! *binary-hyper-sos-strategy* #f)
  (set! *hyper-wos-strategy-flag* #f)
  (set! *remove-all-tautologies-stratflag* #t)
  (set! *hyper-contradiction-strategy* #f)
  (set! *unit-clash-flag* #f) 
  (set! *factors-stratflag* #f)
  (set! *back-subsume-stratflag* #t)
  (set! *remove-identical-lit-stratflag* #t)
  (set! *paramodulation-allowed-flag* #t)   
  (set! *paramodulation-strategy-flag* #t)  
  (set! *paramodulate-from-vars-stratflag* #f)
  (set! *para-pos-unit-strategy* #f)
  (set! *unit-strategy-flag* #f)
  (set! *hyper-strategy-flag* #f)
  (set! *order-strategy-flag* #f)
  ; (set! *order-equality-lex-strat* #f)
  (set! *name-order-strategy-flag* #f)
  (set! *firstlitpref-order-strategy-flag* #f)
  (set! *poslitpref-order-strategy-flag* #f)
  (set! *posweight-order-strategy-flag* #f)
  (set! *ordered-factor-flag* #f)
  (set! *neg-hyper-strategy-flag* #f)
  (set! *type-theory-stratflag* #f)
  (set! *int-logic-flag* #f)
  (set! *int-use-binres-flag* #f)
  (set! *classical-synth-stratflag* #f)
  (set! *int-use-hyperres-flag* #f)  
  (set! *discr-tree-subsume-stratflag* #t)
  (set! *discr-depth-first* #t) 
  (set! *ordered-paramodulation-flag* #t) 
  ; (set! *dynamic-demodulation-flag* #f)
  ; (set! *demodulate-lit-flag* #f) 
  ; (set! *demodulate-flag* #f)
  ; (set! *demodulate-constants-flag* #f)
  ; (set! *new-demodulators-flag* #f) 

  (set! *unit-equality-strat* #f)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-lex-strat* #f)

  (set! *user-criterias-post-flag* #f) 
  (set! *user-criterias-pre-flag* #f) 
  (set! *clause-keepdepth-limit* 0)
  (set! *clause-keeplength-limit* 0)
  (set! *clause-uselength-limit* 0)
  (set! *clause-usenr-limits* '())
  (set! *special-insertion-place* '())
  (set! *clause-usedepth-limits* '())
  (set! *clause-usestruct-limits* '()) )
  

; - - - - - - - - - -- - - - strategic functions - - - - - - -  -




(define *answer-litinf* (logior *litinf-noresolve-mask* 
                                (logior *litinf-noparainto-mask*
                                        *litinf-isanswer-mask*)))

(define (empty-cl? cl)
  (if *int-logic-flag*
      (or (null? cl)
	  (and (null? (cdr cl))
   	       (eq? *int-outmost-label-fun-nr* (pure-fun (caar cl))) ))
      (let ((resflag #t))
	(do ((clpart cl (cdr clpart)))
	    ((or (null? clpart) (not resflag))  resflag)
	  (if (not (answer-pred? (caar clpart))) (set! resflag #f)) ))))
 

(define (unit-cl? cl)
  (if *classical-synth-stratflag*
      (null? (cdr cl)) 
      (let ((count 0))
	(do ((clpart cl (cdr clpart)))
	    ((or (null? clpart) (eq? count 2))  (not (eq? count 2)))
	  (if (not (answer-pred? (caar clpart)))
	      (if (eq? count 0) (set! count 1) (set! count 2)) )))))

(define (first-nonans-lit cl)
  (do ((clpart cl (cdr clpart)))
      ((not (answer-pred? (caar clpart)))
         (car clpart) )))

(define (member-nonans-lit cl)
  (do ((clpart cl (cdr clpart)))
      ((not (answer-pred? (caar clpart)))
         clpart) ))

(define (member-ans-lit cl)
  (do ((clpart cl (cdr clpart)))
      ((answer-pred? (caar clpart))
         clpart) ))

(define (remove-ans-literals cl)
  (let* ((res (cons '() '()))
	 (resend res) )
    (do ((clpart cl (cdr clpart)))
	((null? clpart)
	   (cdr res) )
      (if (not (answer-pred? (caar clpart)))
	  (begin
	    (set-cdr! resend (cons (car clpart) '()))
	    (set! resend (cdr resend)) )))))
	   
(define (ground-fullcl?	fullcl)
  (not (zero? (logand (caar fullcl) *cl-ground-flag*))) )

(define (positive-cl? cl)
  (let ((resflag #t))
    (do ((clpart cl (cdr clpart)))
	((or (null? clpart) (not resflag))  resflag)
      (if (neg-fun? (caar clpart)) (set! resflag #f)) )))

(define (negative-cl? cl)
  (let ((resflag #t))
    (do ((clpart cl (cdr clpart)))
	((or (null? clpart) (not resflag))  resflag)
      (if (not (neg-fun? (caar clpart))) (set! resflag #f)) )))

(define (mixed-cl? cl)
  (let ((negflag #f)
	(posflag #f) )
    (do ((clpart cl (cdr clpart)))
	((or (null? clpart) 
	     (and negflag posflag) )
	  (and negflag posflag) )
      (if (neg-fun? (caar clpart))
	  (set! negflag #t)
	  (set! posflag #t) ))))

(define (has-single-hyper-premiss? cl)
  (let ((count 0))
    (do ((clpart cl (cdr clpart)))
	((or (null? clpart) (eq? count 2))  (not (eq? count 2)))
      (if (if *neg-hyper-strategy-flag*
	      (and (pos-fun? (caar clpart)) (not (answer-pred? (caar clpart))))
	      (neg-fun? (caar clpart)) )
	  (if (eq? count 0) (set! count 1) (set! count 2)) ))))

(define (first-hyper-premiss-lit cl)
  (do ((clpart cl (cdr clpart)))
      ((if *neg-hyper-strategy-flag*
	   (and (pos-fun? (caar clpart)) (not (answer-pred? (caar clpart))))
	   (neg-fun? (caar clpart)) )	   
         (car clpart) )))

(define (first-neg-lit cl)
  (do ((clpart cl (cdr clpart)))
      ((neg-fun? (caar clpart))
         (car clpart) )))

(define (contains-eqlit-cl? cl)
  (do ((clpart cl (cdr clpart)))
      ((or (null? clpart) (pos-eq-pred? (pure-fun (caar clpart))))
         (not (null? clpart)) )))

(define (contains-eqorneg-cl? cl)
  (do ((clpart cl (cdr clpart)))
      ((or (null? clpart)
           (neg-fun? (caar clpart))
           (pos-eq-pred? (pure-fun (caar clpart))) )
         (not (null? clpart)) )))


(define (commutative-fun? fun) ;;; !!!!!!!! TMP !!!!!!!!! tmp always #f
  #f) 

(define (reflexive-fun? fun)
  #f)

(define (removable-tautology-pr? litinfcl-pr)
  (if (and *remove-all-tautologies-stratflag*
	   (tautology-pure-cl? (cdr litinfcl-pr)) )
      (begin 
	(cond ((>= *display-info-level* 80)
	         (begin (newline)
		    (display "the derived clause removed as a tautology") ))
	      (*display-info-char-flag*
		   (display "T") ))
	#t)
      #f))
	

(define (clause-pass-user-limits-post? purecl)
  (and (or (zero? *clause-keepsize-limit*) (pass-keepsize-limit? purecl))
       (or (zero? *clause-keepdepth-limit*) (pass-keepdepth-limit? purecl))
       (or (zero? *clause-keeplength-limit*)(pass-keeplength-limit? purecl)) ))

(define (pass-keepdepth-limit? purecl)
  (every? (lambda(lit)
	    (<= (shiftright (logand *term-depth-mask* (car lit)) 
			   *term-depth-shift*)
		*clause-keepdepth-limit*))
	  purecl))

(define (pass-keepsize-limit? purecl)
  (let ((sum 0))
    (do ((part purecl (cdr part)))
	((or (null? part)
	     (> sum *clause-keepsize-limit*) )
	  (<= sum *clause-keepsize-limit*) )
      (set! sum (+ sum (shiftright (logand *term-size-mask* (caar part)) 
				   *term-size-shift*))) )))

(define (pass-keeplength-limit? purecl)
  (<= (length purecl) *clause-keeplength-limit*) )


(define (clause-pass-user-crit-post? inflist purecl hist)
  (and (or (null? *clause-usenr-limits*) (hist-pass-usenr-limits? hist))
       (or (null? *clause-usedepth-limits*) (hist-pass-usedepth-limits? hist))
       (or (null? *clause-usestruct-limits*)
	   (hist-pass-usestruct-limits? hist)) ))

(define (hist-pass-usenr-limits? hist) 
  (let ((countlst (map (lambda(x)(list (car x) 0 (cadr x)))
		       *clause-usenr-limits*)))
    (hist-count-check countlst hist) ))

(define (hist-count-check countlst hist)
  (if (pair? hist)
      (let ((constr (car hist)))
	(cond
	 ((eq? constr 'hyper)
	  (and (hist-count-check countlst (caddr hist))
	       (every? (lambda(x)(hist-count-check countlst x))
		       (cdddr hist) )))
	 ((eq? constr 'literal)
	  (hist-count-check countlst (cadr hist)) )
	 ((eq? constr 'input)
	  (set! constr (assoc-memq (cadr hist) countlst))
	  (if constr 
	      (if (eq? (cadr constr) (caddr constr)) 
		  #f
		  (begin (set-car! (cdr constr) (+ 1 (cadr constr)))
			 #t))
	      #t))
	 (else #t) ))
      #t))


(define (hist-pass-usedepth-limits? hist) 
   (let ((constr (car hist))
	 (given '())
	 (limit1 '()) 
	 (limit2 '()) )
    (cond
     ((eq? constr 'para) 
       (set! limit1 (caddr hist))
       (set! limit2 (cadddr hist)) 
       (cond
	 ((and (eq? 'input (car limit1))
	       (begin
		 (set! given (assq (cadr limit1) *clause-usedepth-limits*))
		 given)
	       (greater-clause-occ-depth? (cadr limit1) (cdr given) limit2) )
	   #f)
	 ((and (eq? 'input (car limit2))
	       (begin
		 (set! given (assq (cadr limit2) *clause-usedepth-limits*))
		 given)
	       (greater-clause-occ-depth? (cadr limit1) (cdr given) limit1) )
	  #f)
	 (else #t) ))
     ((eq? constr 'hyper)
         (set! given (assq (caddr hist) *clause-usedepth-limits*))
	 (if (not given)
	     #t
	     (not (greater-clause-occ-depth?
		    (caddr hist) (cdr given) )
		  ....))))))
	     


; *clause-usestruct-limits* 

(define (hist-pass-usestruct-limits? hist) 
  (let ((constr (car hist))
	(res #f)
	(one-found #f)
	(tmp '()) )
    (cond
     ((eq? constr 'hyper)
        (set! tmp (cadr (caddr hist))) ; tmp := nr of the nucleon 
        (do ((place (member-if (lambda(x)(eq? (car x) tmp))
			       *clause-usestruct-limits*)
		    (member-if (lambda(x)(eq? (car x) tmp))
			       (cdr place) )))
	    ((or (null? place)(not place) res) 
	        (if res #t (not one-found)) )
	  (set! one-found #t)
	  (if (hist-pass-usestruct-limits-hyper? hist (car place))
	      (set! res #t) )))
     (else #t) )))

(define (hist-pass-usestruct-limits-hyper? hist limitcase)
  (let ((nr (cadr (caddr hist)))
	(res #t) 
	(tmp '()) )
    (do ((els (cdddr hist) (cdr els)))
	((or (null? els) (not res)) res)
      (if (and (pair? (car els))
	       (begin (set! tmp (get-hyper-hist-el-limitnr (car els)))
		      (not (some? (lambda(x)
				   (or (eq? x tmp)
				       (and (pair? x) (eq? (car x) tmp)) ))
				 (cdr limitcase) ))))
	  (set! res #f) ))))

(define (get-hyper-hist-el-limitnr hist)
  (cond 
    ((eq? 'literal (car hist))
        (get-hyper-hist-el-limitnr (cadr hist)) )
    ((eq? 'input (car hist))
        (cadr hist) )
    ((or (eq? 'hyper (car hist)) (eq? 'fact (car hist)))
        (get-hyper-hist-el-limitnr (caddr hist)) )
    (else 0) ))




;-------------------------------------------------------------------
;
;                    init nonstatistical globals
;
;-------------------------------------------------------------------


(define (init-globals)
  (init-one-pass-globals)
  (set! *unit-equality-strat* #f)
  (set! *back-demodulate-flag* #f)
  (set! *order-equality-lex-strat* #f)
  (set! *use-real-time* #f)
  (set! *keep-all-old-clauses-flag* #f)
  (set! *timelimit* #f)
  (set! *foundflag* #f)
  (set! *given-checkpoint-keptnrs* '())
  (set! *checkpoint-keptnrs* '())

  (set! *sos-strategy-flag* #t)
  (set! *ans-prohibited-funsymbs* '())
  (set! *ans-prohibited-constsymbs* '())
  (set! *ans-prohibited-predsymbs* '())
  
  (set! *ans-prohibited-funnrs* '())
  (set! *ans-prohibited-constnrs* '())
  (set! *ans-prohibited-prednrs* '())

  (set! *discr-tree-list* '())
  (set! *unit-alist-list* '())
  (set! *sos-clauses-list-list* '())
  (set! *usable-clauses-list-list* '())
  (set! *lex-symbol-orderlst* '())

  (set! *unit-alist* '()) ; alist of all unit clauses, with assoc-keys being
  ; the pure predicate symbs of units.

  (set! *clause-keepsize-limit* 0)
  (set! *clause-keepdepth-limit* 0)
  (set! *clause-keeplength-limit* 0)  
  (set! *clause-uselength-limit* 0)  

  (set! *prolog-style-variables* #f)
  (set! *const-num-table* '())
  (set! *const-num* 0)
  (set! *min-const-num-sub1* 100000) 
  (set! *fun-num-table* '())
  (set! *fun-num* 0)
  (set! *min-fun-num-sub1* 0) 
  (set! *max-nr-of-funs* 255) 
  (set! *pred-num-table* '())
  (set! *pred-num* 0)
  (set! *min-pred-num-sub1* 0) 
  (set! *max-nr-of-preds* 127)
  (set! *var-num-table* '()) 
  (set! *var-num* 0)
  (set! *min-var-num-sub1* 0) 
  (set! *max-nr-of-vars-in-clause* 255) )

(define *one-pass-kept-clauses-firstnr* 1) ; the first kept clausnr of a pass

(define (init-one-pass-globals)
  (set! *hyper-contradiction-strategy* #f)
  (set! *clause-usenr-limits* '())
  (set! *clause-usedepth-limits* '())
  (set! *clause-usestruct-limits* '())
  (set! *one-pass-kept-clauses-firstnr* (+ 1 *kept-clauses-nr*))

  (set!  *foundflag* #f)      ; set to #t iff the empty clause is found
  (set!  *answer-cl* '())     ; set to a pure answer-clause if empty is found
  (set! *answer-history* '()) ; set to history of the answer clause
  (set! *too-many-vars-flag* #f) ; set when such a clause is generated

  (set! *lex-symbol-orderlst* '())
  (set! *sos-clauses-list* '())
  (set! *sos-clauses-list-end* '())   ; last pair of *sos-clauses-list
  (set! *ans-input-sos* '()) ; sos before removing answer-literals 
  (set! *usable-clauses-list* '())
  (set! *usable-clauses-list-end* '()) ; last pair of *usable-clauses-list*
  (set! *subsumption-list* (list (list 0)) ) 
  (set! *unit-alist* '()) ; alist of all unit clauses, with assoc-keys being
  ; the pure predicate symbs of units.
  (set! *discr-tree* (list '()) )
  (set! *nonunit-discr-tree* (list '()) )  
  (set! *discr-tree-demod-terms* (list '()) ) ; a special discrimination tree
  ; for demodulating terms  
  (set! *discr-tree-demod-atoms* (list '()))  ; a special discrimination tree
  ; for demodulating atoms

  (set! *sos-clauses-list-list* '())
  (set! *usable-clauses-list-list* '()) )

     
  
; - - - - - - - - - - - - - - - - - - - - - 


		    
    
;-------------------------------------------------------------------
;
;                        statistics
;
;-------------------------------------------------------------------

(define *derived-clauses-nr* 0)	
(define *kept-clauses-nr* 0)
(define *kept-mid-nucleus-nr* 0)
(define *kept-new-demodulators-nr* 0)

(define *forward-subsumed-unit-nr* 0)
(define *forward-subsumed-double-nr* 0)
(define *forward-subsumed-overdouble-nr* 0)

(define *forward-subsumed-nonunit-nr* 0)

(define *back-subsumed-nr* 0)

(define *simple-unit-cutoff-nr* 0)
(define *full-unit-cutoff-nr* 0)
(define *double-unit-cutoff-nr* 0)

(define *proof-start-real-time* 0)
(define *proof-start-run-time* 0)

(define *subs-tried-nr* 0)
(define *subs-length-fail-nr* 0)
(define *subs-strength-fail-nr* 0)
(define *subs-aux-strength-fail-nr* 0)
(define *subs-gclpreds-fail-nr* 0)
(define *subs-funlist-fail-nr* 0)
(define *subs-reg-subs-tried-nr* 0)
(define *subs-reg-subs-fail-nr* 0)

(define (init-stats)
  (set! *given-clause-nr* 0)

  (set! *derived-clauses-nr* 0)	
  (set! *kept-clauses-nr* 0)
  (set! *last-usable-cl-nr* 0)
  (set! *kept-mid-nucleus-nr* 0)
  (set! *kept-new-demodulators-nr* 0)
  
  (set! *forward-subsumed-unit-nr* 0)
  (set! *forward-subsumed-nonunit-nr* 0)
  (set! *forward-subsumed-double-nr* 0)
  (set! *forward-subsumed-overdouble-nr* 0)

  (set! *back-subsumed-nr* 0)
  (set! *havealready-timecounter* 0)
  (set! *simple-unit-cutoff-nr* 0)
  (set! *full-unit-cutoff-nr* 0)
  (set! *double-unit-cutoff-nr* 0)

  (set! *proof-start-real-time* (get-internal-real-time))
  (set! *proof-start-run-time* (get-internal-run-time))

  (set! *subs-tried-nr* 0)
  (set! *subs-length-fail-nr* 0)
  (set! *subs-strength-fail-nr* 0)
  (set! *subs-aux-strength-fail-nr* 0)
  (set! *subs-gclpreds-fail-nr* 0)
  (set! *subs-funlist-fail-nr* 0)
  (set! *subs-reg-subs-tried-nr* 0)
  (set! *subs-reg-subs-fail-nr* 0) )

(define (show-stats)
  (show-main-stats)
  (show-subsumption-stats) )

(define (show-main-stats)
  (newline)
  (display "statistics: ---------------------------")(newline)
  (display " given clauses:    ") (display *given-clause-nr*)(newline)
  (display " derived clauses:   ") (display *derived-clauses-nr*)(newline)
  (display " kept clauses:      ") (display *kept-clauses-nr*)(newline)
  (display " kept mid-nuclei:   ") (display *kept-mid-nucleus-nr*)(newline)
  (display " kept new demods:   ") (display *kept-new-demodulators-nr*)
  (newline)
  
  (display " forw unit-subs:    ") (display *forward-subsumed-unit-nr*)
  (newline)
  (display " forw double-subs: ") (display *forward-subsumed-double-nr*)
  (newline)
  (display " forw overdouble-subs: ") 
  (display *forward-subsumed-overdouble-nr*)
  (newline)
  (display " backward subs:     ") (display *back-subsumed-nr*)(newline)
  
  (display " fast unit cutoff:  ") (display *simple-unit-cutoff-nr*)(newline)
  (display " full unit cutoff:  ") (display *full-unit-cutoff-nr*)(newline)
  (display " dbl  unit cutoff:  ") (display *double-unit-cutoff-nr*)(newline)

  (display " real runtime    :  ")
  (display-time-seconds (- (get-internal-real-time) *proof-start-real-time*))
  (newline)
  (display " process. runtime:  ")
  (display-time-seconds (- (get-internal-run-time) *proof-start-run-time*))
  (newline) )

(define (show-subsumption-stats)   
  (display "specific non-discr-tree subsumption statistics: ")(newline)
  (display " tried:           ")(display *subs-tried-nr*)(newline)
  (display " length fails:    ")(display *subs-length-fail-nr*)(newline)
  (display " strength fails:  ")(display *subs-strength-fail-nr*)(newline)
  (display " predlist fails:  ")(display *subs-gclpreds-fail-nr*)(newline)
  (display " aux str. fails:  ")(display *subs-aux-strength-fail-nr*)(newline)
  (display " by-lit fails:    ")(display *subs-funlist-fail-nr*)(newline)
  (display " full subs tried: ")(display *subs-reg-subs-tried-nr*)(newline)
  (display " full subs fail:  ")(display *subs-reg-subs-fail-nr*)(newline) )


(define (display-time-seconds units)
  (let* ((full (quotient units internal-time-units-per-second))
	 (rest (remainder units internal-time-units-per-second)) )
	 ; (rest100s (quotient (* rest 10) 6)) )
    (display full)(display #\.)(display rest) ))

;=====================================================================
;
;
;                          I/O algorithms
;
;
;====================================================================



;--------------------------------------------------------------------
;
;                   top-level procedures
;
;-------------------------------------------------------------------

(define (fprove filename)
  (init-globals)
  (set! *auto-mode-flag* #t)  ; default in auto mode
  (set! *display-info-level* 40)
  (preprocess-otter-file filename) 
  (let* ((tmpname *otter-tmpfile*)
               ; (string-append
	       ;	    *otter-tmpfile-prefix*
	       ;	    filename *otter-tmpfile-suffix*))
	 (rawlst (read-processed-otter-file tmpname)) 
	 (clauselst (read-otter-list-clauses rawlst)) 
	 (readylst '()) 
	 (res #f) )
       (set! *ans-prohibited-constsymbs* '())

    (set! *ans-prohibited-predsymbs* '())
    (set! *ans-prohibited-funsymbs* '())
    (set! readylst (sort-out-otter-list-clauses clauselst))
    ;(newline)
    ;(display readylst)
    ;(newline)

    ;(display "readylst: ")(newline)
    ;(pretty-print readylst) 

    (set! res (prove readylst 99999999 50))
    (if res
	(if (>= *display-info-level* 30)
	    (begin (newline)
		   (if (not (null? *answer-cl*))
		       (begin (display "answer-clause: ")
			      (print-pure-clause *answer-cl*) (newline) ))
		   (show-history *answer-history*) #t)
	    #t)
	(if (>= *display-info-level* 30)
	    (if (eq? *answer-history* 'not-enough-time)
		(begin (newline)
		  (display "time limit exhausted: proof search terminated.")
		  (newline)
		  #f)
		(begin (newline)
		  (display "search space exhausted.")
		  (newline)
		  #f))
	    #f))
    (if (>= *display-info-level* 35) (show-stats))
    (if res #t #f) ))



(define (use-input-settings raw-input)
  (let ((setlist (assq 'set raw-input))
	(clearlist (assq 'clear raw-input))
	(assignlist (assq 'assign raw-input)) )
    (if (not setlist) (set! setlist '()))
    (if (not clearlist) (set! clearlist '()))
    (if (not assignlist) (set! assignlist '()))
    (for-each
      (lambda(el)
	(cond 
	 ((eq? el 'auto)
	    (set! *auto-mode-flag* #t) )
	 ((eq? el 'hyper_res)
	    (set! *auto-mode-flag* #f)
	    (init-hyper-strat) ) 
         ((eq? el 'hyper_order_res)
	    (set! *auto-mode-flag* #f)
	    (init-ordered-hyper-strat) )
	 ((eq? el 'binary_res)
            (set! *auto-mode-flag* #f)
	    (init-binres-strat) )
         ((eq? el 'binary_order_res)
            (set! *auto-mode-flag* #f)
	    (init-ordered-binres-strat) )
	 ((eq? el 'binary_nameorder_res)
            (set! *auto-mode-flag* #f)
	    (init-name-ordered-binres-strat) )
	 ((eq? el 'binary_weightorder_res)
            (set! *auto-mode-flag* #f)
	    (init-weight-ordered-binres-strat) ) 
	 ((eq? el 'binary_posweightorder_res)
            (set! *auto-mode-flag* #f)
	    (init-posweight-ordered-binres-strat) )
         ((eq? el 'binary_hypersos_res)
            (set! *auto-mode-flag* #f)
	    (init-binary-hyper-sos-strat) )
	 ((eq? el 'binary_unit_res)
            (set! *auto-mode-flag* #f)
	    (init-unit-strat) )
	 ((eq? el 'binary_posweightunit_res)
            (set! *auto-mode-flag* #f)
	    (init-posweight-ordered-unit-strat) )
	 ((eq? el 'binary_synth_res)
            (set! *auto-mode-flag* #f)
	    (init-synth-strat) )
	 ((eq? el 'eq_order_lex)
  	    (set! *order-equality-lex-strat* #t) )
	 ((eq? el 'eq_order_depth)
  	    (set! *order-equality-lex-strat* #f) )
         ((eq? el 'check_hypercontra)
	    (set! *hyper-contradiction-strategy* #t) )    
         ((eq? el 'demodulate)
	    (turn-on-demodulation) )
	 ((eq? el 'back_demodulate)
	    (turn-on-demodulation) 
	    (set! *back-demodulate-flag* #t) )
	 ((eq? el 'split)
	   (set! *split-strategy-flag* #t) )
	 ((eq? el 'wall_time)
	    (set! *use-real-time* #t) )
	 ((eq? 'print_tracechar el)
	    (set!  *display-info-char-flag* #t) )
	 ((eq? 'print_debug el)
	    (set! *display-debug-info-flag* #t) )
	 ((eq? el 'prolog_style_variables)
	    (set!  *prolog-style-variables* #t) )))

      setlist)
    (for-each 
      (lambda(el)
	(cond 
	 ((not (pair? el)))
	 ((not (pair? (cdr el))))
	 ((eq? el paramodulate)
	   (set! *paramodulation-allowed-flag* #f) )
	 ((eq? el 'demodulate)
	    (turn-off-demodulation) )))
      clearlist)
    (for-each 
       (lambda(el)
	 (cond 
	  ((not (pair? el)))
	  ((not (pair? (cdr el))))
	  ((eq? 'checkpoints (car el))
	     (set! *given-checkpoint-keptnrs* (cdadr el)) )
	  ((eq? 'max_kept (car el))
	     (set! *max-kept-clauses-nr* (cadr el)) )
	  ((eq? 'max_seconds (car el))
	     ; (set! *total-time-given* #t)
	     (set! *timelimit* (number->string (cadr el))) ) 
	  ((eq? 'print_level (car el))
	     (set! *display-info-level* (cadr el)) )
          ((eq? 'max_literals (car el))
	     (set! *clause-keeplength-limit* (cadr el)) )
	  ((eq? 'max_weight (car el))
	     (set! *clause-keepsize-limit* (cadr el)) )
	  ((eq? 'max_depth (car el))
	     (set! *clause-keepdepth-limit* (cadr el)) )))
       assignlist) ))
    


;--------------------------------------------------------------------
;
;        Reading and parsing otter-syntax clauses from file
;
;--------------------------------------------------------------------


(define (mtest)
  (preprocess-otter-file "jama.in")
  (let* ((lst (read-processed-otter-file "jama.in.tmp")))
    (let* ((res (read-otter-list-clauses lst)))
      (newline)
      (for-each (lambda(x)(display x)(newline)) res) )))


(define (sort-out-otter-list-clauses otterlst)
  (let* ((soslist '())
	 (usablelist '())
	 (axiomslist '())
	 (demodlist '())
	 (commandlist '()) 
         (setlist '())
	 (clearlist '())
         (assignlist '())  
	 (cl '()) 
	 (curlist '()) 
	 (eol-reached #t) 
	 (res '()) )    
    (do ((part otterlst (cdr part)))
	((null? part)
	  ; (newline)
	  ; (display otterlst)
	  ; (newline)
	  ; (display soslist)
	  ; (newline)
	  ; (display demodlist) 
	  ; (quit)  
	   (if (not (null? soslist))
	       (set! res 
		     (cons (cons 'sos (reverse (prop-conv soslist))) res)) )
	   (if (not (null? usablelist))
	       (set! res 
		 (cons (cons 'usable (reverse (prop-conv usablelist))) res)) )
	   (if (not (null? axiomslist))
	       (set! res (cons (cons 'axioms (reverse axiomslist)) res)) )
	   (if (not (null? demodlist))
	       (set! res (cons (cons 'demodulators (reverse demodlist)) res)) )
	   (if (not (null? setlist))
	       (set! res (cons (cons 'set setlist) res)) )
	   (if (not (null? clearlist))
	       (set! res (cons (cons 'clear clearlist) res)) )
	   (if (not (null? assignlist))
	       (set! res (cons (cons 'assign assignlist) res)) )
	   (if (not (null? commandlist))
	       (set! res (cons (cons 'command commandlist) res)) )
	   res)
      (set! cl (car part))
      ; (display cl)(newline)
      (cond 
       ((not (pair? cl))
	  (report-error "syntax error in input file" cl) )
       ((eq? (car cl) 'end_of_list)
	  (set! eol-reached #t) 
	  (set! curlist #f) )
       ((not (pair? (car cl)))
	  (report-error "syntax error in input file" cl) )
       ((eq? 'prohibit_ans_pred (caar cl))
	  (set! *ans-prohibited-predsymbs* (cadar cl)) )
       ((eq? 'prohibit_ans_fun (caar cl))
	  (set! *ans-prohibited-funsymbs* (cadar cl)) )
       ((eq? 'prohibit_ans_const (caar cl))
	  (set! *ans-prohibited-constsymbs* (cadar cl)) )
       ((and (null? (cdr cl)) 
	     eol-reached
	     (eq? 'list (caar cl))
	     (eq? 'sos (cadar cl)) )
	  (set! eol-reached #f)
	  (set! curlist 'sos) )
       ((and (null? (cdr cl)) 
	     eol-reached
	     (eq? 'list (caar cl))
	     (eq? 'usable (cadar cl)) )
	  (set! eol-reached #f)
	  (set! curlist 'usable) )
       ((and (null? (cdr cl))
	     eol-reached
	     (eq? 'list (caar cl))
	     (eq? 'axioms (cadar cl)) )
	  (set! eol-reached #f)
	  (set! curlist 'axioms) )
       ((and (null? (cdr cl)) 
	     eol-reached
	     (eq? 'list (caar cl))
	     (eq? 'demodulators (cadar cl)) )
	  (set! eol-reached #f)
	  (set! curlist 'demodulators) )
       ((eq? curlist 'sos)
	  (set! soslist (cons cl soslist)) )
       ((eq? curlist 'usable)
	  (set! usablelist (cons cl usablelist)) )
       ((eq? curlist 'axioms)
	  (set! axiomslist (cons cl axiomslist)) )
       ((eq? curlist 'demodulators)
	  (set! demodlist (cons cl demodlist)) )      
     
       ((eq? 'set (caar cl))
	  (if (not (pair? (cdar cl)))
	      (report-error "syntax error in input file" cl) )
	  (set! setlist (cons (cadar cl) setlist)) )
       ((eq? 'clear (caar cl))
	  (if (not (pair? (cdar cl)))
	      (report-error "syntax error in input file" cl) )
	  (set! clearlist (cons (cadar cl) clearlist)) )
       ((eq? 'assign (caar cl))
	  (if (or (not (pair? (cdar cl)))
		  (not (pair? (cddar cl))) )
	      (report-error "syntax error in input file" cl) )
	  (set! assignlist (cons (cdar cl) assignlist)) )
       (else 
	  (set! commandlist (cons cl commandlist)) )))))

(define (prop-conv lst)
  (map (lambda(cl)
	 (map (lambda(lit)
		(if (pair? lit)
		    lit
		    (list lit 0) ))
	      cl))
       lst))

; read-otter-list-clauses makes a list of lisp-syntax clauses from
; the list of clauses otterlst read in by read-processed-otter-file

(define (read-otter-list-clauses otterlst)
  (let* ((part otterlst)
	 (clause-rest '()) 
	 (res '()) )
    (do ()
	((null? part) (reverse res))
      (set! clause-rest (parse-otter-clause part))
      (set! res (cons (car clause-rest) res)) 
      (set! part (cdr clause-rest)) )))
		

; parse-otter-clause parses a clause from an input list.
; it returns a pair (parsed-clause . rest)

(define *otter-or-symbol* (string->symbol "|"))

(define (parse-otter-clause lst)
  (let ((part lst)
	(res '()) 
	(endflag #f)
	(tmp1 '()))
    (do ()
	(endflag (cons (reverse res) part))
      (if (and (pair? part) (pair? (cdr part)) (pair? (cddr part))
	       (or (eq? '= (cadr part)) 
		   (eq? '!= (cadr part)) ))
	  (begin
	    (if (eq? '!= (cadr part))
		(set! tmp1 (string->symbol
	          (string-append "-" (symbol->string *eq-symbol*)) ))
		(set! tmp1 *eq-symbol*) )
	    (set! part (append (list tmp1 
				     (list (car part) '! (caddr part)) )
			       (cdddr part) )) ))
      ;(if (and (pair? (car part)) (memq '= (car part)))
      ;	  (begin (set! tmp1 (prepare-read-eqlit (car part) #t))
      ;		 (set! part (append tmp1 (cdr part))) ))
      ; (if (and (pair? (car part)) (memq '!= (car part)))
      ;	  (begin (set! tmp1 (prepare-read-eqlit (car part) #f))
      ;		 (set! part (append tmp1 (cdr part))) ))
      (cond ((eq? *otter-dot-repl* (car part))
	     (set! endflag #t) )
            ((eq? *otter-or-symbol* (car part)))
	    ((or (symbol? (car part)) (number? (car part)))
	     (cond ((null? (cdr part))
		    (report-error "syntax error in input file") ) 
		   ((eq? *otter-dot-repl* (cadr part))
		    (set! res (cons (car part) res)) )
		   ((eq? *otter-or-symbol* (cadr part))
		    (if (null? (cddr part)) 
			(report-error "syntax error in input file") 
			(set! res (cons (car part) res)) ))		   
		   ((not (pair? (cadr part)))
		    (report-error "syntax error in input file") )
		   (else
		    (let ((tmp (parse-otter-explist (cadr part))))
		      (set! res (cons (cons (car part) tmp) res)) 
		      (set! part (cdr part)) ))))
	    ((pair? (car part))
	     (report-error "syntax error in input file") ))
      (if (not (null? part)) (set! part (cdr part))) )))



(define (prepare-read-eqsymb eqlit posflag)
  (let* ((eqsymb (if posflag '= '!=))
	 (neweqsymb (if posflag 
			*eq-symbol* 
			(string->symbol
			 (string-append "-" (symbol->string *eq-symbol*)) ))) 
	 (tmp (replace-one-eq eqsymb '! eqlit)) )
    (list neweqsymb tmp) ))

(define (prepare-read-eqlit eqlit posflag)
  (let* ((eqsymb (if posflag '= '!=))
	 (neweqsymb (if posflag 
			*eq-symbol* 
			(string->symbol
			 (string-append "-" (symbol->string *eq-symbol*)) ))) 
	 (tmp (replace-one-eq eqsymb '! eqlit)) )
    (list neweqsymb tmp) ))

; parse-otter-exp parses an argument list of a functional term.
; it returns the input arglist in a parsed form.

(define (parse-otter-explist exp)
  (let ((part exp)
	(res '()) )
    (do ()
	((null? part) (reverse res))
      (cond ((eq? *otter-comma-repl* (car part)))
	    ((or (symbol? (car part)) (number? (car part)))
	     (cond ((null? (cdr part))
		    (set! res (cons (car part) res)) )
		   ((eq? *otter-comma-repl* (cadr part))
		    (if (null? (cddr part)) 
			(report-error "syntax error in input file") 
			(set! res (cons (car part) res)) ))
		   ((not (pair? (cadr part)))
		    (report-error "syntax error in input file") )
		   (else
		    (let ((tmp (parse-otter-explist (cadr part))))
		      (set! res (cons (cons (car part) tmp) res)) 
		      (set! part (cdr part)) ))))
	    ((pair? (car part))
	     (set! res (cons (parse-otter-explist (car part)) res)) ))
    (if (not (null? part)) (set! part (cdr part))) )))



; preprocess-typet-file reads a file char-by-char and outputs
; a file with the following changes:
;   . is separated by extra spaces and replaced by ...
;   % is replaced by ;
;   ',' is separated by extra spaces and replaced by !
;   | is separated by extra spaces
;   => is surrounded by extra spaces
;   upper-case chars are prefixed by ^


(define *otter-uppercase-prefix* #\^)
(define *otter-comma-repl* '!)
(define *otter-dot-repl* '...)

(define *otter-tmpfile* "/home/tanel/gandalf/gandalf.tmp")
(define *otter-tmpfile-suffix* ".tmp")
(define *otter-tmpfile-prefix* "/users/cs/tammet/gandalf/")
;(define *otter-tmpfile-prefix* "/users/cs/tammet/gandalf/prb.tmp")

(define *otter-outputfile-suffix* ".res")

(define (preprocess-otter-file filename)
  (let ((iport (open-input-file filename)))
    (if iport
      (let ((oport (open-output-file *otter-tmpfile*)))
                         ; (string-append
			 ;   *otter-tmpfile-prefix*
			 ;  filename *otter-tmpfile-suffix*) )))
        (if oport
          (do ((char (read-char iport) (read-char iport)))
	      ((eof-object? char)
                 (close-output-port oport) )
            (cond
	      ((eq? char #\,)
                 (write-char #\space oport)
                 (write *otter-comma-repl* oport)
		 (write-char #\space oport) ) 
	      ((eq? char #\%)
                 (write-char #\; oport)  )
              ((eq? char #\.)
                 (write-char #\space oport)
                 (write *otter-dot-repl* oport)
		 (write-char #\space oport) ) 
              ((eq? char #\|)
                 (write-char #\space oport)
                 (write-char #\| oport)
		 (write-char #\space oport) )
	      ((and (eq? char #\!) (eq? (peek-char iport) #\=))
	         (read-char iport)
		 (write-char #\space oport)
		 (write-char #\! oport)
		 (write-char #\= oport)
		 (write-char #\space oport) )
	      ((eq? char #\=)
                 (write-char #\space oport)
                 (write-char #\= oport) 
                 (if (eq? (peek-char iport) #\>)
                     (begin (write-char #\> oport)
                            (write-char #\space oport)
                            (read-char iport) )
                     (write-char #\space oport) ))
	      ((and (char-alphabetic? char) (char-upper-case? char))
	         (write-char *otter-uppercase-prefix* oport)
		 (write-char char oport) )
              (else
                 (write-char char oport) )))
          (begin (close-input-port iport)
                 (report-error "could not open file: " *otter-tmpfile*)
                   ; (string-append 
		   ; *otter-tmpfile-prefix*
		   ; filename *otter-tmpfile-suffix*) )
                 #f)))
      (begin (report-error "could not open file: " filename)
             #f))))


; read-processed-otter-file reads a preprocessed otter-file
; and outputs a list of all the read objects (in the original order).
;

(define (read-processed-otter-file filename)
  (let ((iport (open-input-file filename))
        (lst '())
        (lst-end '()) )
    (set! lst (cons '() '()))
    (set! lst-end lst)
    (do ((obj (read iport) (read iport)))
        ((eof-object? obj)
          (close-input-port iport)
          (set! lst (cdr lst))
          lst)
      (set-cdr! lst-end (cons obj '()))
      (set! lst-end (cdr lst-end)) )))


; ---------------- reading clauses ----------------------------------

(define *ans-symbol* 'ans)
(define *eq-symbol* 'equal) ; 'eq
(define *eq-fun-symbol* 'eq)
(define *case-fun-symbol* 'case)
(define *if-fun-symbol* 'if)

(define *inhabited-symbol* 'inh)
(define *subst-fun-symbol* 'idsubst)
(define *subst-prim-fun-symbol* 'idsubst-prim)
(define *id-refl-fun-symbol* 'idrefl)
(define *lambda-fun-symbol* 'lambda)
(define *lambda-dummypar-symbol* '*)
(define *typet-speclit*
  (list *ans-symbol*
	(list *lambda-fun-symbol*
	      (list *subst-prim-fun-symbol* *lambda-dummypar-symbol*)
	      (list *subst-fun-symbol* *lambda-dummypar-symbol*) )))
(define *classicsynth-speclit*
  (list *ans-symbol* (list *if-fun-symbol* 1 2 3)) )


(define *demodulator-list-symbol* 'demodulators)
(define *lit-demodulator-list-symbol* 'lit-demodulators)
(define *sos-list-symbol* 'sos)
(define *usable-list-symbol* 'usable)

(define *ans-prohibited-funsymbs* '())
(define *ans-prohibited-constsymbs* '())
(define *ans-prohibited-predsymbs* '())

(define *ans-prohibited-funnrs* '())
(define *ans-prohibited-constnrs* '())
(define *ans-prohibited-prednrs* '())



; ---------------- numerifying clauses and terms ---------------------

; *prolog-style-variables* determines the var/nonvar criteria
;   for input symbols

(define *prolog-style-variables* #f)

;;; In the following constant, function and predicate symbols
;;; of a clause set have each a different numeration
;;; (since function symbs cannot be constant symbs cannot be pred symbs
;;; due to their position in the clause or a term)
;;; vars are enumerated separately for each clause

; - - - - - - - - - - - - - - - - - - - - - -      

; *const-num-table*: ((foo . N) (bar . M) ...)
;     for all constant symbs as first els of the pair

(define *const-num-table* '())

; *const-num* is initially set to a smallest possible function number - 1

(define *const-num* 0)
(define *min-const-num-sub1* 100000) ; the smallest fun num - 1 
                                     ; can be made as small as 65535

; - - - - - - - - - - - - - - - - - - - - - -      

; *fun-num-table*: ((foo . N) (bar . M) ...)
;     for all function symbs as first els of the pair

(define *fun-num-table* '())

; *fun-num* is initially set to a smallest possible function number - 1

(define *fun-num* 0)
(define *min-fun-num-sub1* 0) ; the smallest fun num - 1 
(define *max-nr-of-funs* 255) ; can be made bigger by 1 bit 

; - - - - - - - - - - - - - - - - - - - - - -      

; *pred-num-table*: ((foo . N) (bar . M) ...)
;     for all pure pred symbs (NO -) as first els of the pair
     
(define *pred-num-table* '())

; *pred-num* is initially set to a smallest possible function number - 1

(define *pred-num* 0)
(define *min-pred-num-sub1* 0) ; the smallest fun num - 1 
(define *max-nr-of-preds* 127)

; - - - - - - - - - - - - - - - - - - - - - -      

; *var-num* is for each new clause set to a smallest possible var number - 1

; *var-num-table*: ((xv . N) (u . M) ...)
;     for all var symbs IN THE SINGLE CLAUSE as first els of the pair

(define *var-num-table* '()) ; new for each single clause

(define *var-num* 0)
(define *min-var-num-sub1* 0) 

; the following def is given at the beginning of the file
; (define *max-nr-of-vars-in-clause* 255) ; can be made three bits bigger

; - - - - - - - - - - - - - - - - - - - - - -      

;;; in the following the clause is a standard list like:
;;; ((p 1 (f x) x) (-p x y (f (f y))))

(define (numerify-clause-list clauses)
  (init-numerify-clause-list) 
  (map numerify-clause clauses) )
 


(define (init-numerify-clause-list)
  (set! *const-num-table* '())
  (set! *fun-num-table* '())
  (set! *pred-num-table* '())
  (set! *const-num* (- *minconst* 1))
  (set! *fun-num* *min-fun-num-sub1*)
  (set! *pred-num* *min-pred-num-sub1*) 
  (set! *ans-prohibited-funnrs* '())
  (set! *ans-prohibited-prednrs* '())
  (set! *ans-prohibited-constnrs* '())
  (if *type-theory-stratflag* (numerify-literal *typet-speclit*)) 
  (if *classical-synth-stratflag*
      (begin (numerify-literal *classicsynth-speclit*)
	     (set! *ans-prohibited-prednrs*
		   (map  (lambda(x) (car (numerify-literal (list x))))
			 *ans-prohibited-predsymbs*))
	     (set! *ans-prohibited-funnrs* 
		   (map  (lambda(x) (car (numerify-term (list x))))
			 *ans-prohibited-funsymbs*))
	     (set! *ans-prohibited-constnrs* 
		   (map numerify-term *ans-prohibited-constsymbs*) ))))
 
(define (numerify-clause clause)
  (set! *var-num-table* '())
  (set! *var-num* *min-var-num-sub1*)
  (map numerify-literal clause) )

(define (numerify-literal lit)
  (let* ((pred (car lit))
	 (predstr (symbol->string pred))
	 (polarity (if (char=? #\- (string-ref predstr 0)) #f #t)) ; #t or #f
	 (purepred (if polarity pred 
           (string->symbol (substring predstr 1 (string-length predstr))) ))
	 (place (assq purepred *pred-num-table*))
	 (resnr
	    (if place
		(polarize-pred-nr (cdr place) polarity)
		(begin (set! *pred-num* (+ 1 *pred-num*))
		       (if (> *pred-num* *max-nr-of-preds*) 
	       (report-error "Too many predicate symbs in the clause set" #t) )
		       (set! *pred-num-table*
			     (cons (cons purepred 
				     (shiftleft *pred-num* *fnsymb-shift*) )
				   *pred-num-table*))
		       (polarize-pred-nr 
			   (shiftleft *pred-num* *fnsymb-shift*)
			   polarity)))))
    (cons resnr (map numerify-term (cdr lit))) ))


(define (polarize-pred-nr nr boolsgn)
  (if boolsgn 
      nr
      (logior *sgn-mask* nr) ))

(define (numerify-term term)
  (cond
    ((pair? term)
       (let* ((fun (car term))
	      (nr 0)
	      (place (assq fun *fun-num-table*)))
	 (if place
	     (set! nr (cdr place))
	     (begin (set! *fun-num* (+ 1 *fun-num*))
		    (if (> *fun-num* *max-nr-of-funs*)
	(report-error "Too many function symbs in the clause set" #t) )
		    (set! *fun-num-table*
			  (cons (cons fun 
				      (shiftleft *fun-num* *fnsymb-shift*) )
				*fun-num-table*))
		    (set! nr (shiftleft *fun-num* *fnsymb-shift*)) ))
	 (cons nr (map numerify-term (cdr term))) ))
    ((var-symbol? term) ; term is a variable
       (let ((place (assq term *var-num-table*)))
	 (if place
	     (cdr place)
	     (begin (set! *var-num* (+ 1 *var-num*))
		    (if (> *var-num* *max-nr-of-vars-in-clause*) 
	(report-error "Too many vars in the encountered clause" #t) )
		    (set! *var-num-table*
			  (cons (cons term *var-num*)
			        *var-num-table*))
		    *var-num*))))
   (else ; term is a constant symb 
       (let ((place (assq term *const-num-table*)))
	 (if place
	     (cdr place)
	     (begin (set! *const-num* (+ 1 *const-num*))
		    (set! *const-num-table*
			  (cons (cons term *const-num*)
			        *const-num-table*))
		    *const-num*))))))


(define (var-symbol? symb)
  (and (symbol? symb)
       (let ((char1 (string-ref (symbol->string symb) 0)))
	 (if *prolog-style-variables*
	     (eqv? char1 #\^)
	     (member char1 '(#\x #\y #\z #\u #\v #\w)) ))))


; ------------------------------------------------------------------
;                          funs for printing
; -------------------------------------------------------------------


(define *list-output* #f) ; #t: (p (f x) 1)  #f: p(f(x) 1)

; *comma-output* has effect only if *list-output* is #f

(define *comma-output* #t) ; #t: p(f(x),1)    #f: p(f(x) 1)
                 
; print-full-clause restores the original input order of
; negative literals in nucleons (for hyperresolution only!)
; These are moreover placed before positive literals.

(define (print-full-clause fullcl)
  (if (and *hyper-strategy-flag*
	   (if *neg-hyper-strategy-flag*
	       (not (negative-cl? (cdr fullcl)))
	       (not (positive-cl? (cdr fullcl))) ))
      (let ((new (map (lambda(x)'()) (cdr fullcl)))
	    (nr 0)
	    (tmp '())
	    (otherlst '()) )
	(for-each (lambda(inf lit)
		    (set! nr (get-litinfo-ordernr (car inf)))
		    (if (not (eq? nr 0))
			(set-car! (nth-cdr (- nr 1) new) lit)
			(set! otherlst (cons lit otherlst)) ))
		  (caddar fullcl) ; litinfolist
		  (cdr fullcl) ) ; pure cl
	(if (not (null? otherlst))
	    (set! new (map (lambda(lit)
			     (if (null? lit) 
				 (begin (set! tmp (car otherlst))
					(set! otherlst (cdr otherlst)) 
					tmp)
				 lit))
			   new)))
	(print-pure-clause new) )
      (print-pure-clause (cdr fullcl)) ))

(define (print-clause fullcl)
  (print-pure-clause (cdr fullcl)) )
          
(define (print-pure-clause pureclause)
  (cond
   ((null? pureclause)
      (if *list-output*
	  (display '())
	  (display "EMPTY CLAUSE.") ))
  (*list-output* 
     (display #\()
     (do ((part pureclause (cdr part)))
	 ((null? part))
       (print-lit (car part))
       (if (not (null? (cdr part))) (display #\space)) )
     (display #\)) )
  (else
     (do ((part pureclause (cdr part)))
	 ((null? part))
       (print-lit (car part))
       (if (not (null? (cdr part))) 
	   (display " | ") ))
     (display #\.) )))


(define (print-lit lit)
  (if *list-output* 
      (begin
	(display #\()
	(print-pred (car lit))
	(if (not (null? (cdr lit))) (display #\space)) 
	(do ((part (cdr lit) (cdr part)))
	    ((null? part))
	  (print-term (car part))
	  (if (not (null? (cdr part))) (display #\space)) )
	(display #\)) )
      (begin
	(print-pred (car lit))
	(display #\()
	(do ((part (cdr lit) (cdr part)))
	    ((null? part))
	  (print-term (car part))
	  (if (not (null? (cdr part))) 
	      (if *comma-output*
		  (display #\,)
		  (display #\space)) ))
	(display #\)) )))


(define (print-term term)
  (cond 
    ((composite? term)
       (if *list-output* 
	   (begin
	     (display #\()
	     (print-fun (car term))
	     (if (not (null? (cdr term))) (display #\space)) 
	     (if (eq? *if-fun-nr* (logand *fnsymb-mask* (car term)))
		 (begin (print-lit (cadr term))
			(display #\space)
			(set! term (cdr term)) ))
	     (do ((part (cdr term) (cdr part)))
		 ((null? part))
	       (print-term (car part))
	       (if (not (null? (cdr part))) (display #\space)) )
	     (display #\)) )
	   (begin
	     (print-fun (car term))
	     (display #\()
	     (if (eq? *if-fun-nr* (logand *fnsymb-mask* (car term)))
		 (begin (print-lit (cadr term))
			(if *comma-output*
			    (display #\,)
			    (display #\space) )
			(set! term (cdr term)) ))
	     (do ((part (cdr term) (cdr part)))
		 ((null? part))
	       (print-term (car part))
	       (if (not (null? (cdr part))) 
		   (if *comma-output*
		       (display #\,)
		       (display #\space)) ))
	     (display #\)) )))
    ((var? term)
       (print-var term) )
    (else
       (print-const term) )))

(define (print-pred prednum)
    (if (not-zero? (logand *sgn-mask* prednum)) (display #\-))
    (display (car (rassq (logand *predsymb-mask* prednum)
			 *pred-num-table*))))

(define (print-fun funnum)
  (display (car (rassq (logand *fnsymb-mask* funnum) *fun-num-table*))) )

(define (print-const constnum)
  (display (car (rassq constnum *const-num-table*))) )

; *nr-var-table* is for printing only and has no connections to the
; *var-nr-table* used for numerifying vars in input clauses

(define *nr-var-table* 
  '((1 . x)(2 . y)(3 . z)(4 . u)(5 . v)(6 . w)) ) 

(define (print-var varnr)
  (let ((yflg #f))
    (if (not (zero? (logand *y-var-mask* varnr)))
	(begin (set! yflg #t)(set! varnr (logxor *y-var-mask* varnr))) ) 
    (set! varnr (logand *var-nr-mask* varnr))
    (let ((try (assq varnr *nr-var-table*)))
      (if yflg (display #\'))
      (if try 
	  (display (cdr try))
	  (begin (display #\x) (display (- varnr 6))) ))))


(define (print-list-as-term lst)
  (if (and (list? lst)
	   (not (null? lst)) )
      (begin
	(display (car lst))
	(display "(")
	(do ((part (cdr lst) (cdr part)))
	    ((null? part)
	       (display ")") )
	  (print-list-as-term (car part))
	  (if (not (null? (cdr part)))
	      (display ",")) ))
      (display lst) ))
	  
	

;----------------------- making discr-tree readable ------------------
; *discr-tree* is a discrimination tree for literals.
; It includes all unit-clause literals and also all ground literals
; occurring in nonunit clauses. The top level contains only
; POSITIVE predsymbs.
;
; structure:
;
; each node-level is built of three sublists for funs, consts and vars:
; (funs consts . vars), and in case there are no consts nor vars
; it is just (funs).
;
; level_i:  ( ((f1 level_i+1) (f2 level_i+1) ...)    funs
;             ((c1 level_i+1) (c2 level_i+1) ...)    consts
;             (x1 level_i+1) (x2 level_i+1) ... )  vars  
;           
; the last level on each branch is a leaf:
; (positive-lit negative-lit full-unit-cl)
; where the nonexistant ones of the first two are $f and the
; full-unit-cl may be either negative or positive or nonexistent.
;
; leaf is the standard tree level.    
; (positive-lit negative-lit full-unit-cl)
; where the nonexistant ones of the first two are '() and the
; full-unit-cl may be either negative or positive or nonexistent.

(define (show-discr-tree tree litflag)
  (pretty-print
    (if (and (pair? tree) (not (and (null? (car tree)) (null? (cdr tree)))))
	(if (tree-debug-leaf? tree)
	    (make-readable-discr-leaf tree)
	    (make-readable-discr-tree tree litflag) )
	tree)))
  
(define (make-readable-discr-tree tree litflag)
  (let 
    ((el1 
       (map 
	(lambda(x)
	   (let* ((tmp '())
		  (funnum (car x))
		  (fn (if litflag
			 (car (rassq funnum ;(logand *predsymb-mask* funnum)
				     *pred-num-table*))
			 (car (rassq funnum ;(logand *fnsymb-mask* funnum)
				     *fun-num-table*)) )))
	     (if (tree-debug-leaf? (cdr x))
		 (set! tmp (make-readable-discr-leaf (cdr x)))
		 (set! tmp (make-readable-discr-tree (cdr x) #f)) )
	     (cons fn tmp) ))
	(car tree) ))
     (el2 
      (if (and (not (null? (cdr tree)))
	       (pair? (cdr tree)) )
        (map 
	 (lambda(x)
	   (let* ((tmp '())
		  (funnum (car x))
		  (fn (car (rassq funnum *const-num-table*))) )
	     (if (tree-debug-leaf? (cdr x))
		 (set! tmp (make-readable-discr-leaf (cdr x)))
		 (set! tmp (make-readable-discr-tree (cdr x) #f)) )
	     (cons fn tmp) ))
	 (cadr tree) )
	'()))
     (rst (if (not (null? (cdr tree)))
	      (map (lambda(x)
		     (let* ((tmp '())
			    (funnum (car x))
			    (fn (string->symbol
				  (string-append "v"
	                            (number->string (logand *var-nr-mask*
							    funnum))))))
		       (if (tree-debug-leaf? (cdr x))
			   (set! tmp (make-readable-discr-leaf (cdr x)))
			   (set! tmp (make-readable-discr-tree (cdr x) #f)) )
		       (cons fn tmp) ))
		   (cddr tree) )
	      '())) )
    (cons el1
	  (if (or (not (null? el2))
		  (not (null? rst)) )
	      (cons el2 rst)
	      '() ))))

		   

(define (tree-debug-leaf? tree)
  (or (null? tree)
      (and (pair? (car tree)) ; first el in tree (leaf)
	   (not (pair? (caar tree))) )
      (and (pair? (cdr tree))
	   (pair? (cadr tree))
	   (not (pair? (caadr tree))) )
      (and (pair? (cdr tree))
	   (pair? (cddr tree))
	   (pair? (car (caddr tree))) )))

(define (make-readable-discr-leaf leaf)
  (if *nonunit-discr-flag* 
      leaf
      (let ((el1 (if (pair? (car leaf))
		     (make-readable-lit (car leaf))
		     '()))
	    (el2 (if (and (not (null? (cdr leaf)))
			  (pair? (cadr leaf)) )
		     (make-readable-lit (cadr leaf))
		     '()))
	    (el3 (if (and (not (null? (cdr leaf)))
			  (not (null? (cddr leaf))) )
		 (make-readable-fullcl (caddr leaf))
		 '() )))
	(cons el1
	      (if (or (not (null? el2))
		      (not (null? el3)) )
		  (cons el2
		    (if (not (null? el3))
			(list el3)
			'() ))
		  '() )))))


(define (make-readable-fullcl fullcl)
  (cons (list (caar fullcl) (cadar fullcl)) ; (car fullcl) for full one
	(map make-readable-lit (cdr fullcl)) ))

(define (make-readable-cl cl)
  (map make-readable-lit cl) )

(define (make-readable-lit lit)
 (let ((funnum (car lit)))
  (cons  (if (not-zero? (logand *sgn-mask* funnum)) 
	     (string->symbol
	      (string-append "-"
	         (symbol->string (car (rassq (logand *predsymb-mask* funnum)
					     *pred-num-table*)))))
	      (car (rassq (logand *predsymb-mask* funnum)
			  *pred-num-table*)))
	(if (eq? *lambda-fun-nr* (pure-fun funnum))
	    (cons (make-readable-term (cadr lit))
		  (map make-readable-term (cddr lit)) )
	    (map make-readable-term (cdr lit)) ))))


(define (make-readable-term term)
  (if (pair? term)
      (cons (car (rassq (logand *fnsymb-mask* (car term))
			*fun-num-table*))
	    (if (eq? *lambda-fun-nr* (pure-fun (car term)))
		(cons (make-readable-term (cadr term))
		      (map make-readable-term (cddr term)) )
		(map make-readable-term (cdr term)) ))
      (if (var? term)
	  (string->symbol 
	    (string-append "v" (number->string  (logand *var-nr-mask* term))))
	  (car (rassq term *const-num-table*)) )))


; ---------------------- debugging-oriented printers -------------------


(define (debug-print-clause cl)
  (newline)
  (let* ((info (car cl))
	 (pure (cdr cl))
         (num (car info)) )	      
    (display (vector 
	      (shiftright (logand *cl-maxvar-mask* num) *cl-maxvar-shift*)
	      (shiftright (logand *cl-len-mask* num) *cl-len-shift*)
	      (shiftright (logand *ground-mask* num) 16) 
	      (shiftright (logand *term-depth-mask* num) 11)
	      (shiftright (logand *term-size-mask* num) 5)
	      (logand *const-number-mask* num) ))
    (display #\space)
    (display (number->string (cadr info) 2))
    (display #\space)
    (newline)
    (display (map (lambda(el)
                     (if (not (pair? el))
                         el   
                         (cons (vector 
                                (get-litinfo-classnr (car el))
				(get-litinfo-ordernr (car el))
				(zero? (logand *litinf-reductionrulelit-mask*
					       (car el)))
				(zero? (logand *litinf-rightimplrulelit-mask*
					       (car el)))
				(zero? (logand *litinf-leftorrulelit-mask*
					       (car el)))
				(zero? (logand *litinf-rulelit-mask*
                                               (car el)))
                                (zero? (logand *litinf-noparafromright-mask*
                                               (car el)))
                                (zero? (logand *litinf-noparafromleft-mask*
                                               (car el)))
                                (zero? (logand *litinf-noparainto-mask*
                                               (car el)))
                                (zero? (logand *litinf-isanswer-mask*
                                               (car el)))
                                (zero? (logand *litinf-nofactor-mask*
                                               (car el)))
                                (zero? (logand *litinf-noresolve-mask*
                                               (car el))))
                               (cdr el) )))                
                  (caddr info) )) 
    (newline)              
    (display (cdddr info)) 
    (debug-print-pure-clause pure) ))


(define (debug-print-pure-clause cl)
  (newline)
  (display #\{)
  (for-each (lambda(x)(debug-print-lit x)(display #\space)) cl) 
  (display #\})
  (newline) )

(define (debug-print-lit term)
  (display #\[)
  (debug-print-pred (car term))
  (if (not (null? (cdr term))) (display #\space)) 
  (do ((part (cdr term) (cdr part)))
      ((null? part))
    (debug-print-term (car part))
    (if (not (null? (cdr part))) (display #\space)) )
  (display #\]) )

(define (debug-print-term term)
  (cond 
    ((composite? term)
       (display #\()
       (debug-print-fun (car term))
       (if (not (null? (cdr term))) (display #\space)) 
       (do ((part (cdr term) (cdr part)))
	   ((null? part))
	 (debug-print-term (car part))
	 (if (not (null? (cdr part))) (display #\space)) )
       (display #\)) )
    ((var? term)
       (debug-print-var term) )
    (else
       (debug-print-const term) )))


(define (debug-print-pred funnum)
  (if (not-zero? (logand *sgn-mask* funnum)) (display #\-))
  (display 
   (vector (car (rassq (logand *predsymb-mask* funnum) *pred-num-table*))
	   (shiftright (logand *maxvar-mask* funnum) 18)
	   (shiftright (logand *repeated-vars-mask* funnum) 17)
	   (shiftright (logand *ground-mask* funnum) 16) 
	   (shiftright (logand *term-depth-mask* funnum) 11)
	   (shiftright (logand *term-size-mask* funnum) 5)
	   (logand *const-number-mask* funnum) )))

(define (debug-print-fun funnum)
  (display 
   (vector (car (rassq (logand *fnsymb-mask* funnum) *fun-num-table*))
	   (shiftright (logand *maxvar-mask* funnum) 18)
	   (shiftright (logand *repeated-vars-mask* funnum) 17)
	   (shiftright (logand *ground-mask* funnum) 16) 
	   (shiftright (logand *term-depth-mask* funnum) 11)
	   (shiftright (logand *term-size-mask* funnum) 5)
	   (logand *const-number-mask* funnum) )))


(define (debug-print-const constnum)
  (if (eq? constnum *unassigned*)
      (display "*unassigned*")
      (display 
       (car (rassq constnum *const-num-table*)) )))

(define (debug-print-var varnr)
  (let ((yflg #f))
    (if (not (zero? (logand *y-var-mask* varnr)))
	(begin (set! yflg #t)(set! varnr (logxor *y-var-mask* varnr))) )
    (display
     (string-append 
      (number->string
       (shiftright (logand *single-occ-lit-var-mask* varnr) 15) )
      (number->string
       (shiftright (logand *no-occ-otherlit-var-mask* varnr) 14) )
      (number->string 
       (shiftright (logand *no-occ-folllit-var-mask* varnr) 13) )
      (let ((try (assq (logand *var-nr-mask* varnr)
		       *nr-var-table*)))
	(if try 
	    (if yflg
		(string-append "'" (symbol->string (cdr try)))
		(symbol->string (cdr try)) )
	    (if yflg
		(string-append "'x" (number->string 
				     (- (logand  *var-nr-mask* varnr)
					6)))
		(string-append "x" (number->string 
				    (- (logand  *var-nr-mask* varnr)
				       6))))))))))


(define (debug-print-xvars var)
  (newline)(display "xvars and immediate refs and touched/not:")(newline)
  (do ()
      ((zero? var))
    (debug-print-var var)(display " : ")
    (debug-print-term (var-ref var))(display " : ")
    (display (touched-var? var))(newline)
    ; (if (not (eq? (var-ref var) *unassigned*)) (break))
    (set! var (- var 1)) ))


;(define (debug-print-xvars-spec var)
;  ;(newline)(display "xvars and immediate refs and touched/not:")(newline)
;  (do ()
;      ((zero? var))
;    ;(debug-print-var var)(display " : ")
;    ;(debug-print-term (var-ref var))(display " : ")
;    ;(display (touched-var? var))(newline)
;    (if (not (eq? (var-ref var) *unassigned*)) (break))
;    (set! var (- var 1)) ))

(define (debug-print-yvars var)
  (newline)(display "yvars and immediate refs and touched/not:")(newline)
  (set! var (logior *y-var-mask* var)) ; it is first made an yvar
  (do ()
      ((eq? *max-x-var* var)) ; a zero for yvars
    (debug-print-var var)(display " : ")
    (debug-print-term (var-ref var))(display " : ")
    (display (touched-var? var))(newline)
    (set! var (- var 1)) ))

(define (debug-print-zvars var)
  (newline)(display "zvars and immediate refs:")(newline)
  (set! var (logior *z-var-mask* var)) ; it is first made an yvar
  (do ()
      ((eq? *max-y-var* var)) ; a zero for zvars
    (debug-print-var var)(display " : ")
    (debug-print-term (var-ref var))(newline)
    (set! var (- var 1)) ))

(define (report-error str . obj)
  (newline)
  (display "ERROR: ")(display str)(display " ")
  (for-each display obj)
  (newline)
  (quit) )


; NB! Argument display-info-level not used

(define *internal-time-units-per-second* 100)
(define *total-time-start* 0)
(define *total-time-given* #f)
(define *one-pass-time* 0)
(define *strat-list* '())

(define *given-checkpoint-keptnrs* '())
(define *checkpoint-keptnrs* '())
(define *last-got-time* 0)
(define *use-real-time* #f)

(define  (get-time)
  (if *use-real-time*
      (get-internal-real-time) 
      (get-internal-run-time) ))

;(define (get-check-time-aux x)
;  (display ": ")  (display x)(display " ")
;  (display *kept-cl-nr*) (display "; ")
;  (get-check-time) )

(define (get-check-time x)
  (let ((tmp 0))
    (if (null? *given-checkpoint-keptnrs*)
	(begin 
	  (set! *last-got-time* (get-time)) 
	  *last-got-time* )
	(if (and (>= *kept-cl-nr* (car *given-checkpoint-keptnrs*))
		 (eqv? x (cadr *given-checkpoint-keptnrs*)) )
	    (begin
	      ; (newline)
	      ; (display "given checkpoint used: ")
	      ; (display *given-checkpoint-keptnrs*)
	      ; (newline)
	      (set! *last-got-time* (caddr *given-checkpoint-keptnrs*))
	      (set! *given-checkpoint-keptnrs* 
		    (cdddr *given-checkpoint-keptnrs*) )
	      *last-got-time*)
	    (begin
	      (set! *last-got-time* 0)
	      0)))))

(define  (store-time-checkpoint! x)
   (set! *checkpoint-keptnrs* 
	 (cons  *last-got-time* 
		(cons x
		      (cons *kept-cl-nr*
			    *checkpoint-keptnrs*)))) )
   ; (newline)
   ; (display "!!!!!!!! ")
   ; (display *checkpoint-keptnrs*)
   ; (display " !!!!!!!! ")
   ; (newline))
      

(define (prove raw-input stop-real-time display-info-level)
  (init-stats) 
  (init-globals)
  (set! *internal-time-units-per-second*  internal-time-units-per-second)  
  (use-input-settings raw-input)
  (set! *total-time-start* (get-time))
  (if *timelimit*
      (set! *total-time-given* 
	    (*  *internal-time-units-per-second*
		(string->number *timelimit*) ))
      (set! *total-time-given* #f) )
  

  ; - - - - - - initialisation done - - - - - - - 
  ; (turn-on-demodulation)
  ; (set! *paramodulation-strategy-flag* #t)
  (set! *overdouble-discr-flag* #f)
  (set! *nonunit-discr-flag* #f)
  (let* ((compl-lit-demodulators '())
 	 (demodulators (assq *demodulator-list-symbol* raw-input))
	 (lit-demodulators (assq *lit-demodulator-list-symbol* raw-input))
         (sos (assq *sos-list-symbol* raw-input)) 
	 (usable (assq *usable-list-symbol* raw-input)) 
	 (tmp '()) )
    (set! *stop-res-real-time* stop-real-time)
    ;(set! *display-info-level* display-info-level)
    ;(set! *display-info-char-flag* #f)
    ;(set! *display-debug-info-flag* #f)
    ;(if *timelimit*
	;(begin
	;  (set! *timelimit* (string->number *timelimit*)) ))
    (if demodulators 
        (begin (set! demodulators (cdr demodulators))
               (set! *demodulate-flag* #t)
               (set! *new-demodulators-flag* #t)
               (set! *demodulate-constants-flag* #t)
	       (if *display-debug-info-flag*
		   (begin
		     (newline)(display "input demodulators: ")(newline)
		     (for-each (lambda(x)(display x)(newline))
			       demodulators)))) )
   ;     (begin (set! *demodulate-flag* #f)
   ;            (set! *new-demodulators-flag* #f)
   ;            (set! *demodulate-constants-flag* #f) )
    (if lit-demodulators 
	(begin (set! lit-demodulators (cdr lit-demodulators))
               (set! *demodulate-lit-flag* #t)
               (set! *new-demodulators-flag* #t) ;;;; !!!!! ?????? !!!!!!
	       (if *display-debug-info-flag*
		   (begin
		     (newline)(display "input literal demodulators: ")(newline)
		     (for-each (lambda(x)(display x)(newline))
			       lit-demodulators) )))
        (begin (set! *demodulate-lit-flag* #f)) )
    (if sos
       (begin (set! sos (cdr sos))
	      (if *display-debug-info-flag*
		  (begin
		    (newline)(display "input sos: ")(newline)
		    (for-each (lambda(x)(display x)(newline)) sos) )))) 
    (if usable
       (begin (set! usable (cdr usable))
	      (if *display-debug-info-flag*
		  (begin
		    (newline)(display "input usable: ")(newline)
		    (for-each (lambda(x)(display x)(newline)) usable) ))))

    ; if sos is empty, make usable sos.

    (cond 
     ((and sos (not usable)))
     ((and usable (not sos))
        (set! sos usable) 
	(set! usable #f)))

    (set! *sos-clauses-list* '())
    (set! *usable-clauses-list* '())
    (set! *subsumption-list* (list (list 0))) ; a unique '((0))
    (set! *unit-alist* '())
    (set! *discr-tree* (list '())) ; a unique (list '())
    (set! *nonunit-discr-tree* (list '())) ; a unique (list '()) 
    (set! *discr-tree-demod-terms* (list '()))
    (set! *discr-tree-demod-atoms* (list '()))
    (set! *kept-cl-nr* 0)    

    (init-numerify-clause-list)
    (if demodulators
        (begin
          (if *type-theory-stratflag* 
              (set! demodulators  
                    (map (lambda(x) (add-typet-demod-ans x))
                         demodulators)))
          (set! demodulators (map numerify-clause demodulators)) ))


    (if lit-demodulators
	(set! lit-demodulators (map numerify-clause lit-demodulators)) )
    (if usable (set! usable (map numerify-clause usable)))
    (if sos (set! sos (map numerify-clause sos))) 
    
    ;if usable is empty (all the stuff in sos),
    ; move all but fully negative clauses to usable

    (if (and (not usable)
	     *auto-mode-flag*)
	(begin
	  (set! tmp '())
	  (set! usable '())
	  (do ((part sos (cdr part)))
	      ((null? part)
	         (set! sos (reverse tmp)) )
	    (if (negative-cl? (car part))
		(set! tmp (cons (car part) tmp))
		(set! usable (cons (car part) usable)) ))))	  

    (if *type-theory-stratflag*
      (begin
	(set! *subst-fun-nr* (cdr (assq *subst-fun-symbol* *fun-num-table*)))
	(set! *subst-prim-fun-nr* 
	      (cdr (assq *subst-prim-fun-symbol* *fun-num-table*)) )
	(set! *lambda-fun-nr* (cdr (assq *lambda-fun-symbol* *fun-num-table*)))
	(set! *lambda-dummypar-nr* 
	      (cdr (assq *lambda-dummypar-symbol* *const-num-table*)) )))


    ;(newline)(newline)
    ;(display *pred-num-table*)
    ;(if (>= *display-info-level* 40)
    ;	(begin
    ;	  (newline)
    ;	  (for-each (lambda(x)(display x)(newline))demodulators)
    ;	  (for-each (lambda(x)(display x)(newline))sos) ))

    (fill-predsymb-bit-vector!) ; data for this is built in numerify-clause

    ;(newline)(newline)
    ;(display *pred-num-table*)
    ;(for-each (lambda(x)(display x)(newline))sos)

   ; - - - - - - - - - clauses are numerified - - - - - - - - - - 
    
    ;(set! tmp
    ;	  (prove-make-initial-clauses 
    ;	   demodulators lit-demodulators compl-lit-demodulators
    ;	   usable sos) )
    ; (set! usable (car tmp))
    ;(set! sos (cadr tmp)) 
    ;(analyse-input usable sos)
    ;(if *auto-mode-flag* (select-initial-strategy))
    (set! *strat-list* '((hyper 3)(binary 4)))
    (prove-all-passes 
      usable sos  demodulators lit-demodulators compl-lit-demodulators)))


; prove-make-initial-clauses takes numerified clause lists
; and builds a full input set of clauses, returning a list
; (usable sos)


(define (prove-make-initial-clauses usable sos
	  demodulators lit-demodulators compl-lit-demodulators)
    ; usable unit equality clauses created of demodulators are not
    ; demodulated themselves.
    (if demodulators 
        (begin
         (set! demodulators (map make-input-clause demodulators))
	 (for-each modify-to-demod-inf! demodulators)
         (for-each intern-fullclause demodulators)
         (for-each 
           (lambda(x)
             (set! *cur-clause* 
		   (cons (car x) (map rename-x-to-z-term (cdr x))) )
             (set! *cur-lit* (cadr *cur-clause*))
             (insert-discr-tree!
               *discr-tree-demod-terms* (eq-lit-arg1 *cur-lit*) #f #t #t))
           (if *type-theory-stratflag*
               (map remove-typet-demod-ans demodulators)
               demodulators) )))
  ; now lit-demodulators are going to be term-demodulated while being built.
    (if lit-demodulators 
        (begin
	 ; make complementary clauses 
	 (set! compl-lit-demodulators
	       (map (lambda(x)
		      (map (lambda(y)
			     (cons (logxor *sgn-mask* (car y)) (cdr y)))
			   x))
		    lit-demodulators))
         (set! lit-demodulators (map make-input-clause lit-demodulators))
	 (set! compl-lit-demodulators
	       (map make-input-clause compl-lit-demodulators))
         (for-each intern-fullclause lit-demodulators)
	 (for-each intern-fullclause compl-lit-demodulators)
         (for-each 
           (lambda(x)
	     ; rename for z-vars
             (set! *cur-clause* 
		   (cons (car x) (map rename-x-to-z-term (cdr x))) )
	     ; put the positive lit first
	     (if (neg-fun? (caadr *cur-clause*))
		 (set! *cur-clause*
		       (list (car *cur-clause*)
			     (caddr *cur-clause*) (cadr *cur-clause*) )))
             (set! *cur-lit* (cadr *cur-clause*))
             (insert-discr-tree!
               *discr-tree-demod-atoms* *cur-lit* #f #t #t))
           lit-demodulators)))
    ; now sos is going to be demodulated while being built. 
    ; Even constants and ground terms are demodulated.              
    (if *display-debug-info-flag*
	(begin
	  (newline)
	  (display "starting to make-input-clause on sos:")
	  (newline)(newline)
	  (for-each (lambda(x)(display x)(newline)
			   (debug-print-pure-clause x)(newline))
		    sos)))
    (if sos 
        (begin (if usable 
		   (set! usable (map make-input-clause usable))
		   (set! usable '()) )
	       ; clauses with nr less than *last-usable-cl-nr*
	       ; are in the "usable" set - for set of support strategy
	       (set! *last-usable-cl-nr* *kept-cl-nr*)
               (set! sos (map make-input-clause sos))

	       (if (and (>= *display-info-level* 40)
			*display-debug-info-flag*)
		   (begin
		     (newline)
		     (display "finished make-input-clause:")
		     (newline)(newline)
		     (for-each (lambda(x)(display x)(newline)
				      (debug-print-clause x)(newline))
			       (append usable sos))) )
	       (if *type-theory-stratflag*
	       	   (begin 
	       	     (set! *ans-input-sos* (append demodulators sos))
	             (set! sos (remove-ans-from-clauses sos)) ))
	       (for-each intern-fullclause usable)
               (for-each intern-fullclause sos) )
	(begin (newline)
	       (display "Error: no sos-list given.")
	       (newline)
	       (quit) ))	       
    ; now there is no need any more for demodulating constants or 
    ; assuming new demodulators.         
    ;(newline)
    ;(display "finished intern-fullclause on sos:")
    ;(newline)(newline)
    ;(for-each (lambda(x)(display x)(newline)
	;	     (debug-print-clause x)(newline))
	 ;     sos)
    ; (set! *new-demodulators-flag* #f)
    (set! *demodulate-constants-flag* #f)
    (if (not demodulators) (set! demodulators '()))
    (if (not sos) (set! sos '()))
    (if (not lit-demodulators) (set! lit-demodulators '()))
    (if (not compl-lit-demodulators) (set! compl-lit-demodulators '()))
;    (set! *sos-clauses-list*
;	  (append demodulators lit-demodulators compl-lit-demodulators sos))
    ; the following is fine for hyperresolution:
    (if *type-theory-stratflag*
	(set! *sos-clauses-list*
	      (append (remove-ans-from-clauses demodulators)
		      compl-lit-demodulators usable sos))
	; create the initial clause lists
	(begin 
	  (set! *usable-clauses-list* usable)
	  (set! *sos-clauses-list*
		; (append demodulators compl-lit-demodulators sos) 
		sos) ))
    (list usable sos demodulators lit-demodulators compl-lit-demodulators) )


(define (prove-show-input-clauses usable sos 
	   demodulators lit-demodulators compl-lit-demodulators)
    (if (and (>= *display-info-level* 40) (not (null? demodulators)))
	(begin
	  (newline)(display "demodulators: ")(newline)
	  (for-each (lambda(x)(print-pure-clause (cdr x))(newline))
		    demodulators)
	  (if *display-debug-info-flag*
	      (begin
		(newline)(display "demodulators in debug form: ")(newline)
		(for-each (lambda(x)(debug-print-clause x)(newline))
			  demodulators) ))))
    (if (and (>= *display-info-level* 40)  (not (null? lit-demodulators)))
	(begin
	  (newline)(display "literal demodulators: ")(newline)
	  (for-each (lambda(x)(print-pure-clause (cdr x))(newline))
		    lit-demodulators)
	  (if *display-debug-info-flag*
	      (begin
		(newline)(display "literal demodulators in debug form: ")
		(newline)
		(for-each (lambda(x)(debug-print-clause x)(newline))
			  lit-demodulators) ))))
    (if (and (>= *display-info-level* 40)  (not (null? usable)))
	(begin
	  (newline)(display "clauses in usable: ")(newline)
	  (for-each (lambda(x)(print-pure-clause (cdr x))(newline))
		    usable)
	  (if *display-debug-info-flag*
	      (begin
		(newline)(display "usable in debug form: ")
		(newline)
		(for-each (lambda(x)(debug-print-clause x)(newline))
			  sos) ))))
    (if (and (>= *display-info-level* 40) (not (null? sos)))
	(begin
	  (newline)(display "clauses in sos: ")(newline)
	  (for-each (lambda(x)(print-pure-clause (cdr x))(newline))
		    sos)
	  (if *display-debug-info-flag*
	      (begin
		(newline)(display "sos in debug form: ")
		(newline)
		(for-each (lambda(x)(debug-print-clause x)(newline))
			  sos) )))) )


(define *no-passes-started* #t)

(define *all-passes-time* 0)

; prove-all-passes loops over several prover passes. 
; it takes numerified clause lists as input.

(define (prove-all-passes usable sos
	   demodulators lit-demodulators compl-lit-demodulators)

  (let ((tmp '()))
    (set! *no-passes-started* #t)
    (set! *strat-list* 
	  '((hyper 5 #t)
	    (binary 10 #t)
	    ; (binary-hyper-sos 20 #t) ; *binary-hyper-sos-strategy*
	    ; (hyper-order 10 #f)
	    (binary-unit 10 #t)
	    (binary-unit 10 #f)
	    (binary-order 10 #f)
	    (binary-nameorder 10 #f) 

	    (hyper-order 30 #f)
	    (binary 30 #t)
	    (binary 50 #f) ))    
    (cond
     ((not *auto-mode-flag*) 
         (if *total-time-given*
	     (set! *one-pass-time* *total-time-given*)
	     (set! *one-pass-time* 1000000) )	 	 
         (set! tmp
	       (prove-one-pass usable sos
	         demodulators lit-demodulators compl-lit-demodulators))
	 (get-check-time 20)
	 (store-time-checkpoint! 20) 
	 tmp)
     (else     ; automatic case 
      (if *total-time-given*
	  (set! *all-passes-time* 
		(quotient *total-time-given* 
			  *internal-time-units-per-second*))
	  (set! *all-passes-time* 180) )      
      ; this first strategy is not actually used - just a dummy
      (set! *strat-list* (list '(binary 5 #t)))
      ; loop over strategies, doing one pass for each
      (do ()
	  ((or *foundflag*
	       (null? *strat-list*)
	       (and *total-time-given*
		    (> (- (get-check-time 10) *total-time-start*)
		       *total-time-given*)))

	   (if (and (not *foundflag*)
		    (not (null? *strat-list*)) )
	       (store-time-checkpoint! 10) )
	   (get-check-time 30)
	   (store-time-checkpoint! 30)
	   *foundflag*)
	; clear strategical flags
	(init-one-pass-strategical-flags)
	; take a first strat
	(pick-strat-from-list)	
	; do one pass
	(prove-one-pass usable sos
	   demodulators lit-demodulators compl-lit-demodulators) )))))

(define (pick-strat-from-list) 	
  (let ((el (car *strat-list*)))
  ; pick a strategy from the given list
    (cond 
     ((eq? (car el) 'binary-hyper-sos)
      (init-binary-hyper-sos-strat) )
     ((eq? (car el) 'hyper)
      (init-hyper-strat) )
     ((eq? (car el) 'hyper-order)
      (init-ordered-hyper-strat) )
     ((eq? (car el) 'hyper-wos)
      (init-hyper-wos-strat) )
     ((eq? (car el) 'binary)
      (init-binres-strat) ) 
     ((eq? (car el) 'binary-unit)
      (init-unit-strat) )  
     ((eq? (car el) 'binary-order)
      (init-ordered-binres-strat) )
     ((eq? (car el) 'binary-nameorder)
      (init-name-ordered-binres-strat) )
     ((eq? (car el) 'binary-unit-depth)
      (init-binary-unit-depth) )
     ((eq? (car el) 'binary-unit-depth-sospref)      
      (init-binary-unit-depth-sospref) )       
     ((eq? (car el) 'binary-unit-depth-backdemod)
      (init-binary-unit-depth-backdemod) )
     ((eq? (car el) 'binary-unit-lex-backdemod)
      (init-binary-unit-lex-backdemod) )
     ((eq? (car el) 'binary-unit-lex)
       (init-binary-unit-lex) )
     (else
      (init-binres-strat) ))
    
    ; pick time for a pass	
    (set! el (cdr el))
    (set! *one-pass-time* (* *internal-time-units-per-second*
			     (car el) ))
    ; pick sos strategy used/not
    (set! el (cdr el))
    (set! *sos-strategy-flag* (car el))
    ; pick maxdepth and maxlength
    (set! el (cdr el))
    (if (not (null? el))
	(begin (set! *clause-keepdepth-limit* (car el))
	       (set! el (cdr el)) ))
    (if (not (null? el))
	(begin (set! *clause-keeplength-limit* (car el))
	       (set! el (cdr el)) ))

    ; remove first el from stratlist
    (set! *strat-list* (cdr *strat-list*)) ))

; prove-one-pass does one concrete prover pass with a certain
; strategy

(define (prove-one-pass usable sos 
	   demodulators lit-demodulators compl-lit-demodulators)
  (let ((tmp '())
	(nr 0) )
    (init-one-pass-globals)
  
    (set! tmp
	  (prove-make-initial-clauses usable sos
		demodulators lit-demodulators compl-lit-demodulators))
    (set! usable (car tmp))
    (set! tmp (cdr tmp))
    (set! sos (car tmp))
    (set! tmp (cdr tmp))
    (set! demodulators (car tmp))
    (set! tmp (cdr tmp))
    (set! lit-demodulators (car tmp))
    (set! tmp (cdr tmp))
    (set! compl-lit-demodulators (car tmp))
    (set! tmp (cdr tmp))
    (if *no-passes-started*
	(begin
	  (prove-show-input-clauses usable sos 
            demodulators lit-demodulators compl-lit-demodulators)
	  (analyse-input usable sos)
	  (if *auto-mode-flag* 
	      (begin
		(select-initial-strategy)
		(init-one-pass-strategical-flags)
		(pick-strat-from-list) ))
	  (set! *no-passes-started* #f) ))

;     (prove-show-input-clauses usable sos 
;            demodulators lit-demodulators compl-lit-demodulators)

    ; if not sos-strategy, move all usable to sos
    (if (not *sos-strategy-flag*) 
	(begin
	  (set! sos (append usable sos)) 
	  (set! usable '()) ))

    ; if hyper-strategy, make correct input-litinfo for nucleons
    (if *hyper-strategy-flag*
	(for-each 
	 (lambda(cl)
	   (if (not (positive-cl? (cdr cl)))
	       (set-car! (cddr (car cl))
			 (make-input-litinfolst (cdr cl)) )))
	 (append usable sos) )) 

    ; for hyperres put positive clauses to sos, rest to usable
    (if (and *hyper-strategy-flag*
	     (not *sos-strategy-flag*) )
	(do ((part sos (cdr part)))
	    ((null? part)
	       (set! usable (reverse tmp)) )
	  (if (and (pair? (car part))
		   (positive-cl? (cdar part)) )
	      (set! sos (cons (car part) sos))
	      (set! tmp (cons (car part) tmp)) )))	       
    
   
;  (prove-show-input-clauses usable sos 
;            demodulators lit-demodulators compl-lit-demodulators)
    ; if ordered strategy, make correct litinfos for all input clauses
    (if (or *order-strategy-flag* 
	    *name-order-strategy-flag*
	    *firstlitpref-order-strategy-flag*
	    *posweight-order-strategy-flag*
	    *ordered-paramodulation-flag*)
	(for-each
	  (lambda(cl)
	    (set-car! (cddr (car cl))
		      (map cddr (order-prohibit-litinfos
				  (map (lambda(x)(cons 0 (cons 0 x)))
				       (caddr (car cl)) )
				  (cdr cl) ))))
	  (append usable sos) ))

    (show-one-pass-strategy)
    (if *type-theory-stratflag* (enumerate-sos-literals sos))  
    (set! *sos-clauses-list* sos)
    (set! *usable-clauses-list* usable)

    (set! tmp (given-clause-slow-loop))
    (set! *discr-tree-list* (cons *discr-tree* *discr-tree-list*))
    (set! *unit-alist-list* (cons *unit-alist* *unit-alist-list*))
    (if *keep-all-old-clauses-flag*
	(begin
	  (set! *sos-clauses-list-list* 
		(cons *sos-clauses-list* *sos-clauses-list-list*) )
	  (set! *usable-clauses-list-list* 
		(cons *usable-clauses-list* *usable-clauses-list-list*) )))
    ; if half of clause kept resource used, throw away old units
    (set! nr 0)
    (for-each (lambda(alist)
		(for-each 
		 (lambda(el) (set! nr (+ (length (cdr el)) nr)))
		 alist))
	      *unit-alist-list*)
    (if (and 
	 (not *keep-all-old-clauses-flag*)
	 (> (* 2 nr) *max-kept-clauses-nr*) )
	(begin
	  (if (>= *display-info-level* 38)
	      (begin (newline)
		     (display "old unit clauses discarded")
		     (newline) ))
	  (set! *discr-tree-list* '())
	  (set! *unit-alist-list* '())
	  (set! *sos-clauses-list-list* '())
	  (set! *usable-clauses-list-list* '()) ))

    tmp))


(define (show-one-pass-strategy)
  (if (> *display-info-level* 38)
      (begin
	(cond 
	 (*hyper-strategy-flag*
	   (newline)(display "using hyperresolution"))
	 (else
	   (newline)(display "using binary resolution")) )
	(newline)
	(if *order-strategy-flag*
	    (begin
	      (display "using term-depth-order strategy") 
	      (newline) ))
	(if *name-order-strategy-flag*
	    (begin
	      (display "using pred-name-order strategy") 
	      (newline) ))
	(if *sos-strategy-flag*
	    (begin
	      (display "using sos strategy") 
	      (newline) ))
	    ; (display "not using sos strategy") 
	; (newline)
	(if *unit-strategy-flag*
	    (begin
	      (display "using unit strategy") 
	      (newline) ))
	(if *dynamic-demodulation-flag*
	    (begin
	      (display "using dynamic demodulation")
	      (newline) ))
	(if *unit-equality-strat*
	    (begin
	      (display "using unit equality specials")
	      (newline) ))
	(if *back-demodulate-flag*
	    (begin
	      (display "using back demodulation")
	      (newline) ))
	(if *order-equality-lex-strat* 
	    (begin
	      (display "using lex for equality ordering")
	      (newline) ))
	(if (not (zero? *clause-keeplength-limit*))
	    (begin
	      (display "clause length limited to ") 
	      (display *clause-keeplength-limit*)
	      (newline) ))
	(if (not (zero? *clause-keepdepth-limit*))
	    (begin
	      (display "clause depth limited to ") 
	      (display *clause-keepdepth-limit*)
	      (newline) ))	
	(if (not (equal? *sos-preference-ratio* 5))
	    (begin
	      (display "initial sos preference ratio is ") 
	      (display *sos-preference-ratio*)
	      (newline) ))
	(newline) )))


(define (eq-lit-in-sos? sos)
  (some? (lambda(cl)
	   (some? (lambda(lit)
		     (or (pos-eq-pred? (pure-fun (car lit)))
			 (and *type-theory-stratflag*
			      (and (composite? (cadr lit))
				   (pos-eq-fun? (pure-fun (caadr lit))) ))))
		  (cdr cl) ))
	 sos))


; add-typet-demod-ans adds a special ans-lit containing
; an application of idrefl constructor to the left eq arg;
;
; In addition, it wraps the literal into the special inhabited-predicate.

(define (add-typet-demod-ans cl)
  (let* ((eqlit (car cl))
         (eqtype (eq-lit-type eqlit))
         (eqarg1 (eq-lit-arg1 eqlit)) )
    (cons (list *ans-symbol* (list *id-refl-fun-symbol* eqtype eqarg1))
          (list (list *inhabited-symbol* (car cl))) )))

; NB! remove-typet-demod-ans does not correct the term size and term
; depth info of the resulting clause, although both decrease!!!

(define (remove-typet-demod-ans cl)
 (let* ((litlst 
	 (remove-if (lambda (y)
		      (not (zero? (logand *litinf-isanswer-mask*
					  (car y) ))))
		    (caddr (car cl)) ))
	(newcl (remove-ans-literals (cdr cl))) 
	(newlit (cadar newcl)) )
   (cons (cons (car (car cl))
	       (cons (cadr (car cl))
		     (cons litlst (cdddr (car cl))) ))
	 (list newlit) )))


; remove-ans-from-clauses removes ans-literals and corresponding
; litinfolst elements from all clauses in the given clause list

(define (remove-ans-from-clauses cllist)
  (map (lambda(x)
	 (let ((litlst 
		 (remove-if (lambda (y)
                              (not (zero? (logand *litinf-isanswer-mask*
                                                 (car y) ))))
			    (caddr (car x)) ))
	       (cl (remove-ans-literals (cdr x))) )
	   (cons (cons (car (car x))
		       (cons (cadr (car x))
			     (cons litlst (cdddr (car x))) ))
		 cl)))
       cllist))


; enumerate-sos-literals takes sos and enumerates all
; its literals, ie adds an unique number to the 
; info-part of a literal.

(define *enumerate-sos-literals-nr* 0)

(define (enumerate-sos-literals sos)
  (set! *enumerate-sos-literals-nr* 0)
  (for-each
   (lambda(cl)
     (let ((litinfolst (caddar cl)))
       (set-car!
	  (cddar cl)
	  (map (lambda(litinf)
		 (set! *enumerate-sos-literals-nr*
		       (+ 1 *enumerate-sos-literals-nr*) )
		 (append litinf (list *enumerate-sos-literals-nr*)) )
	       litinfolst))))
   sos))


;=================================================================
;
;                  strategy selection
;
;=================================================================


(define *input-cl-nr* 0)
(define *input-unitcl-nr* 0)
(define *input-negcl-nr* 0)
(define *input-poscl-nr* 0)
(define *input-mixcl-nr* 0)
(define *input-horncl-nr* 0)
(define *input-invhorncl-nr* 0)
(define *input-negdhorncl-nr* 0)
(define *input-wellorderedcl-nr* 0)
(define *input-somewhat-ordered-nr* 0)
(define *input-wellhyper-nr* 0)
(define *input-maxlen-nr* 0)
(define *input-maxdepth-nr* 0)
(define *input-contains-eq* #f)
(define *input-negun-d* 0)
(define *input-posun-d* 0)

(define (analyse-input usable sos)
  (set! *input-cl-nr* 0)
  (set! *input-unitcl-nr* 0)
  (set! *input-negcl-nr* 0)
  (set! *input-poscl-nr* 0)
  (set! *input-mixcl-nr* 0)
  (set! *input-horncl-nr* 0)
  (set! *input-invhorncl-nr* 0)
  (set! *input-negdhorncl-nr* 0)
  (set! *input-wellorderedcl-nr* 0)
  (set! *input-somewhat-ordered-nr* 0)
  (set! *input-wellhyper-nr* 0)
  (set! *input-maxlen-nr* 0)
  (set! *input-maxdepth-nr* 0)
  (set! *input-contains-eq* #f)
  (set! *input-negun-d* 0)
  (set! *input-posun-d* 0)
  (if (or (eq-lit-in-sos? sos)
	  (eq-lit-in-sos? usable) )
      (set! *input-contains-eq* #t)
      (set! *input-contains-eq* #f) )
  (let ((tmp '())
	(cl '()) 
	(clinf '())
	(purecl '()) 
	(litinfolst '())
	(tmplitinfolst '())
	(prohibitlitnr 0)
	(newlitinf '())
	(litnr 0)
        (ndepth 0)
        (pdepth 0)
	(depth 0)
	(poslitnr 0)
	(neglitnr 0) )
    (do ((clauselstpart (append usable sos) (cdr clauselstpart)))
	((null? clauselstpart)) 
      (set! cl (car clauselstpart))
      (set! clinf (car cl))
      (set! purecl (cdr cl))
      (set! litinfolst (caddr clinf))

     ; (newline)(display cl)(newline)

      (set! tmplitinfolst (map (lambda(x)(cons 0 (cons 0 x))) litinfolst))
      ; propositional stuff
      (set! *input-cl-nr* (+ 1 *input-cl-nr*))
      (set! litnr 0)
      (set! neglitnr 0)
      (set! poslitnr 0)
      (set! ndepth 0)
      (set! pdepth 0)
      (for-each (lambda(lit)
		  (set! litnr (+ 1 litnr))
		  (if (neg-fun? (car lit))
		      (begin
                         (set! neglitnr (+ 1 neglitnr))
                         (if (> (fun-real-depth (car lit)) ndepth)
                             (set! ndepth (fun-real-depth (car lit))) ))
		      (begin 
                         (set! poslitnr (+ 1 poslitnr))
                         (if (> (fun-real-depth (car lit)) pdepth)
                             (set! pdepth (fun-real-depth (car lit))) ))))
		purecl)
      (if (> ndepth pdepth)
	  (set! depth ndepth)
	  (set! depth pdepth) )
      (if (> depth *input-maxdepth-nr*)
	  (set! *input-maxdepth-nr* depth) )
      (if (> litnr *input-maxlen-nr*)
	  (set! *input-maxlen-nr* litnr) )
      (if (eq? litnr 1)
	  (begin
            (if (and (neg-fun? (caar purecl))
                     (> (fun-real-depth (caar purecl)) *input-negun-d*) )
                (set! *input-negun-d* (fun-real-depth (caar purecl))) )
            (if (and (pos-fun? (caar purecl))
                     (> (fun-real-depth (caar purecl)) *input-posun-d*) )
                (set! *input-posun-d* (fun-real-depth (caar purecl))) ) 
            (set! *input-unitcl-nr* (+ 1 *input-unitcl-nr*)) ))
      (cond 
        ((zero? poslitnr)
	  (set! *input-negcl-nr* (+ 1 *input-negcl-nr*)) )
	((zero? neglitnr)
	  (set! *input-poscl-nr* (+ 1 *input-poscl-nr*)) )
	(else
	  (set! *input-mixcl-nr* (+ 1 *input-mixcl-nr*)) ))
      (if (and (> litnr 1) (<= poslitnr 1))
	  (begin 
             (set! *input-horncl-nr* (+ 1 *input-horncl-nr*))
             (if (> pdepth ndepth)
                 (set! *input-negdhorncl-nr* (+ 1 *input-negdhorncl-nr*)) )))
      (if (and (> litnr 1)(<= neglitnr 1))
	  (set! *input-invhorncl-nr* (+ 1 *input-invhorncl-nr*)) )
      (if (and (> litnr 1) 
	       (>= (* 2 neglitnr) poslitnr) )
	  (set! *input-wellhyper-nr* (+ 1 *input-wellhyper-nr*)) )
      ; ordering stuff
      (set! prohibitlitnr 0)
      (if (not (null? (cdr purecl)))
	  (begin
	    (set! tmp *order-strategy-flag*)
	    (set! *order-strategy-flag* #t)
	    (set! newlitinf (vdepth-order-litinfo-aux 
			      tmplitinfolst purecl))
	    (set! *order-strategy-flag* tmp)
	    (for-each 
	     (lambda(inf)
	       (if (not (litinf-binrespara? (caddr inf)))
		   (begin
		     (set! prohibitlitnr (+ 1 prohibitlitnr)) )))
	     newlitinf) 
	    (if (>= (* 2 prohibitlitnr) litnr)
		(set! *input-wellorderedcl-nr*
		      (+ 1 *input-wellorderedcl-nr*) ))
	    (if (> prohibitlitnr 0)
		(set! *input-somewhat-ordered-nr* 
		      (+ 1 *input-somewhat-ordered-nr*) ))
	    ;(newline)
	    ;(print-pure-clause purecl)
	    ;(newline)
	    ;(display newlitinf)
	    ;(display "litnr ")(display litnr)
	    ;(display "prohibit ")(display prohibitlitnr)
	    ;(newline) 

	    )))))


; select-initial-strategy selects the initial strategy, based on
; the data about input clause set computed by the analyse-input.

(define (select-initial-strategy)
  (let ((res '())
	(tmp '()) )
    (set! res (copy-tree (select-initial-strategy-aux)))    
    (set! res 
	  (map (lambda(x)
		 (cons (car x)
		       (cons (quotient (* *all-passes-time* (cadr x)) 100)
			     (cddr x)) ))
	       res))
    (set! tmp (last-pair res))
    (set-car! (cdr (car tmp)) '100000)
    (if (>= *display-info-level* 30)
	(begin (newline)(display "strategies selected: ")(newline)
	       (display res) (newline) ))
    (set! *strat-list* res)
    (cond 
      ((and *input-contains-eq* *paramodulation-allowed-flag*)
         (if (>= *display-info-level* 35)
	   (begin (newline)(display "using equality")(newline)) )
	 (set! *paramodulation-strategy-flag* #t)
	 (turn-on-demodulation) )
      (else
         (if (>= *display-info-level* 35)
	     (begin (newline)(display "not using equality")(newline)) )
	 (set! *paramodulation-strategy-flag* #f) 
	 (turn-off-demodulation) ))
    res))


(define (select-initial-strategy-aux)
  (let ((tmp '())
	(res '()) 
	(maxtime 150) 
	(incrlen 0) )
    (if (>= *display-info-level* 50)
	(begin
	  (newline)
	  (display "========= stats concerning the input clauses: =========")
	  (newline)
	  (display "*input-cl-nr* ")(display *input-cl-nr*)(newline)
	  (display "*input-unitcl-nr* ")(display *input-unitcl-nr*)(newline)
	  (display "*input-negcl-nr* ")(display *input-negcl-nr*)(newline)
	  (display "*input-poscl-nr* ")(display *input-poscl-nr*)(newline)
	  (display "*input-mixcl-nr* ")(display *input-mixcl-nr*)(newline)
	  (display "*input-horncl-nr* ")(display *input-horncl-nr*)(newline)
	  (display "*input-invhorncl-nr* ")
	  (display *input-invhorncl-nr*)(newline)
          (display "*input-negdhorncl-nr* ")
          (display *input-negdhorncl-nr*)(newline)
	  (display "*input-wellorderedcl-nr* ")
	  (display  *input-wellorderedcl-nr*)(newline)
	  (display "*input-wellhyper-nr* ")
	  (display  *input-wellhyper-nr*)(newline)
	  (display "*input-somewhat-ordered-nr* ")
	  (display  *input-somewhat-ordered-nr*)(newline)
	  (display "*input-maxlen-nr* ")(display *input-maxlen-nr*)(newline)
          (display "*input-maxdepth-nr* ")(display *input-maxdepth-nr*)
	            (newline)
	  (display "*input-contains-eq* ")
	  (display *input-contains-eq*)(newline)
	  (display "=======================")
	  (newline) ))
    ; selection starts
      ;(hyper 5 #t)
      ;(binary 10 #t)
      ;(binary-hyper-sos 20 #t) ; *binary-hyper-sos-strategy*
      ;(hyper-order 10 #f)
      ;(binary-unit 10 #t)
      ;(binary-unit 10 #f)
      ;(binary-order 10 #f)
      ;(binary-nameorder 10 #f)       
      ;(hyper-order 30 #f)
      ;(binary 30 #t)
      ;(binary 50 #f) 


    (set! incrlen (- (* 2 *input-maxlen-nr*) 1))
    (cond 
     ; pure unit => select binary
     ((= *input-cl-nr* *input-unitcl-nr*)
       ;(if (>= *display-info-level* 35)
       ;    (begin (newline)(display "using binary resolution")(newline)) ) 
       (set! res
	     `((binary-unit-depth-backdemod 10 #f ,*input-maxdepth-nr* 1)
	       (binary-unit-depth-sospref 20 #t)
	       (binary-unit-depth-backdemod 30 #f)	       
	       (binary-unit-lex-backdemod 40 #f) )))
     ; pure horn 
     ((= *input-horncl-nr* (- *input-cl-nr* *input-unitcl-nr*))
        (cond
           ((and (<= (* 9  *input-horncl-nr*)  
                     (* 10 *input-negdhorncl-nr*) )
                 (not *input-contains-eq*) )
               ;(if (>= *display-info-level* 35)
               ;    (begin (newline)
               ;           (display "using name-ordered binary resolution")
               ;           (newline) ))

	       ; horn clauses and backward hyper might be a good idea
	       (cond
		((> *input-cl-nr* 100)
		   ; heavy stress on sos
		   (set! res 
		     `((hyper 5 #t)
		       (binary-unit 15 #t)
		       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 30 #t) 
		       (hyper 10 #f)
		       (binary-nameorder 10 #f ,*input-maxdepth-nr* ,incrlen)
		       (binary-nameorder 5 #f)       
		       (binary-unit 15 #f)) ))
		((> *input-cl-nr* 20)
		   ; stress sos
		   (set! res 
		     `((hyper 2 #t)
		       (binary-unit 10 #t)
		       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 20 #t) 
		       (hyper 10 #f)
                       (binary-nameorder 10 #f, *input-maxdepth-nr* ,incrlen)
		       (binary-nameorder 5 #f)       
		       (binary-unit 28 #f)) ))
		   ; do not stress sos
		(else 
		   (set! res 
		     `((binary 3 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 7 #t)
		       (hyper 20 #f)
                       (binary-nameorder 20 #f, *input-maxdepth-nr* ,incrlen)
		       (binary-nameorder 10 #f)       
		       (binary-unit 40 #f)) ))))
               ;(init-name-ordered-binres-strat) 
           ((and (= *input-negcl-nr* 1)
                 (> *input-negun-d* 0)
                 (> *input-negun-d* *input-posun-d*) )
               ;(if (>= *display-info-level* 35)
               ;    (begin (newline)(display "using unit resolution")
               ;           (newline) ))  
	       ; negative unit deeper than positive unit: use more unit strat
	       (cond
		 ((> *input-cl-nr* 100)
		   ; heavy stress sos
		   (set! res 
		     `((hyper 5 #t)
		       (binary-unit 15 #t)
		       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 30 #t) 
		       (hyper 20 #f)
		       (binary-unit 20 #f)) ))
		 ((> *input-cl-nr* 20)
		   ; stress sos
		   (set! res 
		     `((hyper 2 #t)
		       (binary-unit 10 #t)
		       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 20 #t) 
		       (hyper 20 #f)
		       (binary-order 8 #f ,*input-maxdepth-nr* ,incrlen)
		       (binary-unit 30 #f)) ))
		(else   ; do not stress sos
		   (set! res 
		     `((hyper 30 #f)
		       (binary 3 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 7 #t)
		       (binary-nameorder 5 #f)
		       (binary-order 5 #f ,*input-maxdepth-nr* ,incrlen)
		       (binary-unit 55 #f)) ))))
               ;(init-unit-strat) 
            (else
               ;(if (>= *display-info-level* 35)
	       ;    (begin (newline)(display "using hyperresolution")
               ;           (newline) ))
	       ; hyperresolution preferred
  	      (cond
	        ((> *input-cl-nr* 100)
		   ; heavy stress sos
		   (set! res 
		     `((hyper 5 #t)
		       (binary-unit 15 #t)
		       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 35 #t)
		       (binary-unit 15 #f)
		       (hyper 20 #f) )))
	        ((> *input-cl-nr* 20)
		   ; stress sos
		   (set! res 
		     `((hyper 5 #t)
		       (binary-unit 10 #t)
		       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 20 #t)
		       (binary-order 5 #f ,*input-maxdepth-nr* ,incrlen)
		       (binary-unit 15 #f)
		       (hyper 35 #f) )))
	        (else
		   ; do not stress sos
		   (set! res 
		     `((binary 10 #t)
		       (binary-unit 30 #f)
		       (binary-order 5 #f ,*input-maxdepth-nr* ,incrlen)
		       (hyper 55 #f)) ))))))
               ; (init-hyper-strat) 

     ; horn part ended

     ; pure inverse horn => prefer nameorder and unit
     ((= *input-invhorncl-nr* (- *input-cl-nr* *input-unitcl-nr*))
       ;(if (>= *display-info-level* 35)
       ;    (begin (newline)(display "using unit resolution")(newline)) )
          (cond
	        ((> *input-cl-nr* 100)
		   ; heavy stress sos
		   (set! res 
		     `((hyper 5 #t)
		       (binary-unit 15 #t)
		       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 30 #t)
		       (binary-unit 25 #f)
		       (binary-nameorder 15 
			  #f ,*input-maxdepth-nr* ,incrlen) )))
	        ((> *input-cl-nr* 20)
		   ; stress sos
		   (set! res 
		     `((hyper 2 #t)
		       (binary-unit 10 #t)
		       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 20 #t)
		       (binary-unit 30 #f)
		       (binary-order 5 #f)
		       (binary-nameorder 10 #f ,*input-maxdepth-nr* ,incrlen)
		       (binary-nameorder 13 #f) )))
	        (else
		   ; do not stress sos
		   (set! res 
		     `((binary 10 #t)
		       (binary-unit 30 #f)
		       (hyper 10 #f)
		       (binary-order 5 #f)
		       (binary-nameorder 20 #f ,*input-maxdepth-nr* ,incrlen)
		       (binary-nameorder 25 #f) )))))

       ; (init-unit-strat) 

     ; 8/10 of nonunit are well-ordered => use also ordered binary
     ((>= (* 10 *input-wellorderedcl-nr*)
	  (* 8 (- *input-cl-nr* *input-unitcl-nr*)))
       ;(if (>= *display-info-level* 35)
       ;   (begin (newline)(display "using ordered binary resolution")
       ;		  (newline)) )
            (cond
	        ((> *input-cl-nr* 100)
		   ; heavy stress sos
		   (set! res 
		     `((hyper 5 #t)
		       (binary-unit 15 #t)
		       (binary 15 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 35 #t)
		       (binary-nameorder 10 #f ,*input-maxdepth-nr* ,incrlen) 
		       (binary-order 20 #f) )))
	        ((> *input-cl-nr* 20)
		   ; stress sos
		   (set! res 
		     `((hyper 2 #t)
		       (binary-unit 10 #t)
		       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 20 #t)
		       (binary-unit 10 #f)
		       (binary-order 10 #f ,*input-maxdepth-nr* ,incrlen)
		       (binary-order 10 #f)
		       (binary-nameorder 10 #f) )))
	        (else
		   ; do not stress sos
		   (set! res 
		     `((binary 10 #t)
		       (binary-unit 20 #f)
		       (hyper 20 #f)
		       (binary-order 10 #f ,*input-maxdepth-nr* ,incrlen)
		       (binary-order 30 #f)
		       (binary-nameorder 10 #f) )))) )

     ; 3/4 of nonunit are horn => prefer ordered hyper
     ((>= (* 4 *input-horncl-nr*)
	  (* 3 (- *input-cl-nr* *input-unitcl-nr*)))
        ;(if (>= *display-info-level* 35)
	;   (begin (newline)(display "using ordered hyperresolution")
	;	  (newline)) )
        (cond
	        ((> *input-cl-nr* 100)
		   ; heavy stress sos
		   (set! res 
		     `((hyper 5 #t)
		       (binary-unit 15 #t)
		       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 30 #t)
		       (binary-unit 15 #f)
		       (hyper-order 25 #f) )))
	        ((> *input-cl-nr* 20)
		   ; stress sos
		   (set! res 
		     `((hyper 5 #t)
		       (binary-unit 10 #t)
		       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 20 #t)
		       (binary-order 5 #f ,*input-maxdepth-nr* ,incrlen)
		       (binary-unit 15 #f)
		       (hyper-order 35 #f) )))
	        (else
		   ; do not stress sos
		   (set! res 
		     `((binary 10 #t)
		       (binary-unit 30 #f)
		       (binary-order 10 #f ,*input-maxdepth-nr* ,incrlen)
		       (binary-order 10 #f)
		       (hyper-order 40 #f)) ))) )
               ; (init-hyper-strat) 

     (else
        ;default strategy
         (cond
	        ((> *input-cl-nr* 100)
		   ; heavy stress sos
		   (set! res 
		     `((hyper 5 #t)
		       (binary-unit 15 #t)
		       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 25 #t)
		       (binary-unit 10 #f)
		       (binary-order 15 #f)
		       (hyper-order 20 #f) )))
	        ((> *input-cl-nr* 20)
		   ; stress sos
		   (set! res 
		     `((hyper 2 #t)
		       (binary-unit 5 #t)
		       (binary 10 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 23 #t)
		       (binary-unit 15 #f)
		       (binary-order 7  #f)
		       (binary-nameorder 8 #f ,*input-maxdepth-nr* ,incrlen)
		       (binary-nameorder 10 #f)
		       (hyper-order 20 #f) )))
	        (else
		   ; do not stress sos
		   (set! res 
		     `((binary 5 #t ,*input-maxdepth-nr* ,incrlen)
		       (binary 15 #t)
		       (binary-unit 15 #f)
		       (binary-order 8 #f ,*input-maxdepth-nr* ,incrlen)
		       (binary-order 12 #f)		       
		       (binary-nameorder 15 #f)
		       (hyper-order 30 #f)) ))) ))
    res))
               ; (init-hyper-strat) 


     ; 2/3 of nonunit are well-ordered => prefer ordered binary
     ;((>= (* 3 *input-wellorderedcl-nr*)
     ;	  (* 2 (- *input-cl-nr* *input-unitcl-nr*)))
     ;  (if (>= *display-info-level* 35)
     ; 	   (begin (newline)(display "using ordered binary resolution")
     ;		  (newline)) )
     ;  (init-ordered-binres-strat) )
     ; all nonunit are somewhat ordered => ordered binary res
     ;((= *input-somewhat-ordered-nr* (- *input-cl-nr* *input-unitcl-nr*))
     ;   (if (>= *display-info-level* 35)
     ;	   (begin (newline)(display "using ordered binary resolution")
     ;		  (newline)) )
     ;  (init-ordered-binres-strat) )
     ; at least two unit clauses exist and 
     ; 3/4 of nonunit are well-hyper => select ordered hyper
     ;((and 
     ;  (>= *input-unitcl-nr* 2) 
     ;  (>= (* 4 *input-wellhyper-nr*)
     ;	   (* 3 (- *input-cl-nr* *input-unitcl-nr*)) ))
     ; (if (>= *display-info-level* 35)
     ;	  (begin (newline)(display "using ordered hyperresolution")
     ;		 (newline)) )
     ; (init-ordered-hyper-strat) )
     ; default with equality
     ;(*input-contains-eq*
     ;   (if (>= *display-info-level* 35)
     ;      (begin (newline)(display "using ordered binary resolution")
     ;             (newline)) )
     ;   (init-ordered-binres-strat) ) 
     ; default without equality
     ;(else 
     ;   (if (>= *display-info-level* 35)
     ;	   (begin (newline)(display "using name-ordered binary resolution")
     ;		  (newline)) )
     ;   (init-name-ordered-binres-strat) ))


       
;==================================================================
;
;               temporary debugging defs
;=================================================================


;(trace subsumed-pure-clause-aux? ground-lit-equal?
; fun-can-subsume? fun-may-cont-vars? restore-gcl-vars
;  clear-gcl-vars later-occurrences-var? subsumed-term-list-ordinary?)

;(trace subsumed-pure-clause-aux? fun-can-subsume? havealready-clause?
; funlist-can-subsume-aux?)

;(trace bigger-strength-aux? funlist-can-subsume-aux?
; subsumed-pure-clause-aux? fun-can-subsume? >-fun-ordering-aux?)

;(trace make-kept-clause full-unit-cutoff! full-unit-cutoff-aux!
; has-some-local-only-vars? has-only-local-only-vars? resolve-unify-cutoff
;  resolve-unify-cutoff-aux)

;(trace subsumed-clause?)
;(trace subsumed-term-list-ordinary?)

;(trace back-subsume)
;(trace havealready-clause?)
;(trace resolve-unify-ordinary)
;(trace resolve-unify-ordinary-aux)
;(trace resolve-unify-cutoff)
;(trace factor-unify-ordinary)
;(trace factor-unify-ordinary-aux)
;(trace subst-deco-cl)

;(trace subst-deco-term)
;(trace subst-deco-weight-term)

;(trace unit-handle-cl!)
;(trace make-clause-info)

;(trace make-kept-clause)
;(trace make-varinfo-lst!)
;(trace make-fulldeco-clause!)
;(trace make-kept-litinflist)

;(trace full-unit-cutoff!)
;(trace handle-derived-inf-cl-pair)
;(trace unit-cl?)
;(trace sort-clause)
;(trace factor)
;(trace factor-aux)


(define (break)
  (let ((flag #f))
    (do ()
	(flag)
      (newline)
      (display "|")
      (let ((tmp (eval (read))))
	(if (eq? tmp 'quit)
	    (set! flag #t)
	    (display tmp) ))
      (newline) )))

;(define (up) (load "resma.scm"))

;(require 'pretty-print)


;(define (clear-x-vars! var)
;  (do ()
;      ((zero? var)
;        (newline)(debug-print-xvars-spec 6)(newline))
;    (set-var-ref! var *unassigned*)
;    (set! var (- var 1)) ))

;(define (clear-x-vars-and-touched! var)
;  (do ()
;      ((zero? var)
;       (newline)(debug-print-xvars-spec 6)(newline))
;    (set-var-ref! var *unassigned*)
;    (set-var-touched! var #f)
;    (set! var (- var 1)) ))

; (define *qp-width* 1500) 

(define (hprove x y z)(init-hyper-strat)(prove x y z))

(define (hfprove x y)(init-hyper-strat)(fprove x))

(define (init)
  (init-hyper-strat) )
;  (set! *neg-hyper-strategy-flag* #t)) 


(define (go)
  (init-binres-strat)
  ; (init-hyper-strat)
  ; (init-ordered-binres-strat)
  ;(trace handle-derived-inf-cl-pair)
  ;(trace handle-derived-inf-cl-pair-aux)
  ;(trace insert-clause-discr-tree!)
  ;(trace order-prohibit-litinfos)
  ;(trace vdepth-order-litinfo-eq-only)
  ; (trace order-equality-litinfo)
  ;(trace vdepth-order-litinfo-aux)
  ; (trace neg-fun?)

;  (set! *qp-width* 1500) 
 ;(set! *qp-width* 200) 
  
  ;(set! *display-debug-info-flag* #t)

  ;(fprove "problems/steam.prb" 50) )
  
  ; (fprove "t2.prb" 50) )

  (fprove "SYN/SYN353-1+eq_stfp.in") )

  ; (fprove "problems/steam.prb" 50) )

  ; (fprove "otter/GRP003-1+rm_noeq.in" 50) )

  ; (fprove "otter/GRP008-1+rm_eq_stfp.in" 50) )

; (fprove "problems/ch18n5.prb" 200) )


;=======================================================================
;
;                      batch version
;
;======================================================================

(define (abs-path-part path)
  (substring path
             0
             (- (string-length path) (string-length "gandalf"))  ))

(define (batch)
 (display (program-arguments))
 (fprove "tmp2.in") )


(define (batch)
 (if (null? (cdr (program-arguments)))
    (begin (display "No argument file name given: going interactive.")
	   (newline)
	   (abort)
	   (quit))
    (begin ; (newline)
      (if (null? (cddr (program-arguments)))
	  (begin 
	    (set! *timelimit* #f)
	    (fprove (cadr (program-arguments))) )
	  (begin 
	    (set! *timelimit* 
		  (car (cddr (program-arguments))))
	    (fprove (cadr (program-arguments))) ))
      (newline)
      (quit) )))

(define (show-old-units)
  (for-each 
   (lambda(x)
     (display "--------------------------")
     (newline)
     (for-each 
       (lambda (sublist)
	(for-each 
	 (lambda (el)
	   (print-clause el)(newline) )
	 (cdr sublist) ))
	x))
   *unit-alist-list*))

;(display "starting to run batch")
;(newline)

;(batch)

;(if (pair? x) (print-clause x))(newline))  *usable-clauses-list*)

; (for-each (lambda(x)(if (pair? x) (print-clause x))(newline))  *usable-clauses-list*)

; (for-each (lambda(x)(if (pair? x) (print-clause x))(newline))  *sos-clauses-list*)

;(trace check-discr-tree-part)
;(trace check-nonunit-discr-tree)
;(trace check-discr-tree-leaf)


;============================================================

;(set!  *prolog-style-variables* #t)
;(init-hyper-strat)

; (set! *timelimit* "10")

; (set! *qp-width* 1500) 

;(trace hyper-contradiction?)
;(trace given-clause-slow-loop)
;(trace usable-clauses-fast-loop)
;(trace two-clause-binres-loop)
;(trace one-step-for-hyper)
;(trace resolve-unify-ordinary)
;(trace make-poslit-preferred!)
;(trace mixed-cl?)
;(trace make-all-resolvable!)

; (fprove "problems/luk3.prb" #f 60)
; (fprove "hyptest1.prb" #f 120)
; (fprove "problems/luk2.prb" #f 60)
; (fprove "problems/steam.in" #f 60)
;derived clause after 2-nd order deco:
;-p(i(x,i(i(i(i(y,z),i(i(u,v),i(v,i(u,v)))),i(w,i(i(u,v),i(v,i(u,v))))),i(x1,i(i(i(a,b),a),a))))) | -p(x1) | -p(x).
;(trace resolve-unify-cutoff)
;"CALLED" term-may-cont-vars? (2099296 24836 24834)
;"RETURNED" term-may-cont-vars? #t
;"CALLED" occur-check? 24580 (2099296 24836 24834)
; "CALLED" term-may-cont-vars? (2099296 24836 24834)
; "RETURNED" term-may-cont-vars? #t
; "CALLED" occur-check? 24580 (2099296 24836 24834)

;(trace otter-show-history-aux-aux)
;(trace otter-show-history-aux)
;(trace show-history-clnr)
;(trace show-history-find-clausepl)
;(set! *qp-width* 1500)


;==================================================================
;
;                    EOF of resio.scm
;
; It is necessary to also have resbase.scm and resmain.scm
;
;========================== END ===================================

