/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *               SPLITTING OF CLAUSES                     * */
/* *                                                        * */
/* *  $Module:   SPLIT                                      * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.59 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/03 12:29:08 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: rules-split.c,v $ */

/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "rules-split.h"

static LIST  split_DeleteClausesDependingOnLevelFromList(PROOFSEARCH,LIST, int);
static LIST  split_RemoveUnnecessarySplits(PROOFSEARCH, CLAUSE);

/**************************************************************/
/* Functions                                                  */
/**************************************************************/


LIST split_Backtrack(PROOFSEARCH PS, CLAUSE EmptyClause) 
/**************************************************************
  INPUT:   An empty clause and a proofsearch object
  RETURNS: A list of clauses deleted in the backtracked split levels
  MEMORY:  Backtracks the top of the split stack wrt the empty clause's level
***************************************************************/
{
  SPLIT ActBacktrackSplit = (SPLIT)NULL;
  LIST  RecoverList, ReinsertClauses;
  int   Backtracklevel;

  RecoverList    = split_RemoveUnnecessarySplits(PS, EmptyClause);
  Backtracklevel = clause_SplitLevel(EmptyClause);
  
  while (!prfs_SplitIsTopLevel(PS)
	 && (prfs_ValidLevel(PS) > Backtracklevel)) {
      ActBacktrackSplit = prfs_SplitStackTop(PS);
      prfs_SplitStackPop(PS);
      if (prfs_GetFatherClause(ActBacktrackSplit) != (CLAUSE)NULL) {
	RecoverList = list_Cons(prfs_GetFatherClause(ActBacktrackSplit), RecoverList);
	prfs_SetFatherClause(ActBacktrackSplit, NULL);
      }
      RecoverList = list_Nconc(split_DeletedClauses(ActBacktrackSplit), RecoverList);
      clause_DeleteClauseList(split_BlockedClauses(ActBacktrackSplit));
      split_FreeOneNode(ActBacktrackSplit);
      prfs_DecValidLevel(PS);
  }
  
  while (!prfs_SplitIsTopLevel(PS)
	 && list_Empty(split_BlockedClauses(prfs_SplitStackTop(PS)))) {
      ActBacktrackSplit = prfs_SplitStackTop(PS);
      prfs_SplitStackPop(PS);
      if (prfs_GetFatherClause(ActBacktrackSplit) != (CLAUSE)NULL)
	RecoverList = list_Cons(prfs_GetFatherClause(ActBacktrackSplit), RecoverList);
      RecoverList = list_Nconc(split_DeletedClauses(ActBacktrackSplit), RecoverList);
      split_FreeOneNode(ActBacktrackSplit);
      prfs_DecValidLevel(PS);
  }
  
  if (!prfs_SplitIsTopLevel(PS)) {
      LIST Scan;
      int SplitMinus1;

      SplitMinus1       = prfs_ValidLevel(PS) - 1;
      ActBacktrackSplit = prfs_SplitStackTop(PS);
      RecoverList       = list_Nconc(split_DeletedClauses(ActBacktrackSplit), RecoverList);
      split_SetDeletedClauses(ActBacktrackSplit, list_Nil());

      RecoverList     = split_DeleteInvalidClausesFromList(PS, SplitMinus1, RecoverList);
      ReinsertClauses = split_BlockedClauses(ActBacktrackSplit);

      for (Scan = ReinsertClauses; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
	  clause_NewNumber((CLAUSE) list_Car(Scan));
	  clause_AddParentClause((CLAUSE) list_Car(Scan), clause_Number(EmptyClause));
	  clause_AddParentLiteral((CLAUSE) list_Car(Scan), 0);
      }

      RecoverList     = list_Nconc(ReinsertClauses , RecoverList);
      split_SetBlockedClauses(ActBacktrackSplit, list_Nil());
      
      /* Then, delete clauses from current level (Hack) */
      prfs_DecValidLevel(PS);
      prfs_MoveInvalidClausesDocProof(PS);
      prfs_IncValidLevel(PS);
  } else {
    LIST Scan;
    /*Don't delete clauses from current level (split_IsToplevel())*/
    prfs_MoveInvalidClausesDocProof(PS);
    for (Scan=RecoverList;!list_Empty(Scan);Scan=list_Cdr(Scan))
      prfs_InsertDocProofClause(PS,list_Car(Scan));
    list_Delete(RecoverList);
    RecoverList = list_Nil();
  }
  prfs_PutLastBacktrackLevel(PS, prfs_ValidLevel(PS));
  
  return RecoverList;
}


LIST split_DeleteClausesDependingOnLevelFromList(PROOFSEARCH Search, LIST ClauseList, int Level) {
/**************************************************************
  INPUT:   A list of clauses and a split level
  EFFECT:  Deletes all clauses depending on split level from list
  RETURNS: The updated list
***************************************************************/
  LIST   Scan;
  CLAUSE Clause;

  for(Scan = ClauseList; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Clause = list_Car(Scan);   
    if(clause_DependsOnSplitLevel(Clause, Level)) {
      prfs_InsertDocProofClause(Search,Clause);
      list_Rplaca(Scan, NULL);
    }
  }
  return list_PointerDeleteElement(ClauseList, NULL);
}
 

LIST split_DeleteClausesDependingOnLevelFromSet(PROOFSEARCH PS, LIST ClauseList, int SplitLevel)
/**************************************************************
  INPUT:   A list of clauses, a PROOFSEARCH object, a split level and a pointer to a list
  EFFECT:  Deletes all clauses depending on split level from list and index
  RETURNS: A list of blocked clauses from splits, for which a clause in the list
           was deleted
***************************************************************/
{
  LIST   Scan, DelList, Recover;
  CLAUSE Clause;
  SPLIT  Reinsert;

  DelList = list_Nil();
  Recover = list_Nil();
  for(Scan = ClauseList; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Clause = list_Car(Scan);
    if(clause_DependsOnSplitLevel(Clause, SplitLevel)) {
      Reinsert = prfs_GetSplitOfLevel(clause_SplitLevel(Clause), PS);
      if (split_DeletedClauses(Reinsert) != list_Nil()) {
	Recover = list_Nconc(split_DeletedClauses(Reinsert), Recover);
	split_SetDeletedClauses(Reinsert, list_Nil());
      }
      DelList = list_Cons(Clause, DelList);
    }
  }
  for(Scan = DelList; !list_Empty(Scan); Scan =list_Cdr(Scan)) {
    Clause = list_Car(Scan);
    if (clause_GetFlag(Clause, WORKEDOFF))
      prfs_MoveWorkedOffDocProof(PS,Clause);
    else
      prfs_MoveUsableDocProof(PS,Clause);
  }
  list_Delete(DelList);
  return Recover;
}



LIST split_DeleteInvalidClausesFromList(PROOFSEARCH Search, int Level, LIST ClauseList)
/**************************************************************
  INPUT:   An int as level and a list of clauses.
  RETURNS: The list where invalid clauses wrt 'Level' are deleted.
***************************************************************/
{
  LIST   Scan;
  CLAUSE Clause;

  /*printf("\nDiese Liste soll von ungueltigen (Level > %d) "
    "befreit werden: \n",Level);fflush(stdout);
    clause_ListPrint(ClauseList);*/

  for(Scan = ClauseList; !list_Empty(Scan); Scan = list_Cdr(Scan)) {
    Clause = list_Car(Scan);
    if(!split_IsClauseValid(Clause,Level)) {
      prfs_InsertDocProofClause(Search,Clause);
      list_Rplaca(Scan, NULL);
    }
  }
  return list_PointerDeleteElement(ClauseList, NULL);
}


LIST split_RemoveUnnecessarySplits(PROOFSEARCH PS, CLAUSE EmptyClause)
/**************************************************************
  INPUT:   An empty clause and proofsearch object
  EFFECT:  Clears all splits that were not necessary to derive the empty clause up to the last
           Backtrack level
  RETURNS: A list of recovered clauses
***************************************************************/
{
  LIST Scan;
  LIST Recover;
  LIST Deleted;
  LIST ScanStack;

  int SplitLevel;
  int LastBacktrackLevel;
  SPLIT Split;

  Scan               = prfs_SplitStack(PS);
  SplitLevel         = prfs_ValidLevel(PS);
  LastBacktrackLevel = prfs_LastBacktrackLevel(PS);
  Recover            = list_Nil();

  while (SplitLevel > LastBacktrackLevel) {
    if (split_IsDirty(list_Car(Scan))) {
      if (!clause_DependsOnSplitLevel(EmptyClause, SplitLevel)) {
	Split = list_Car(Scan);

	clause_DeleteClauseList(split_BlockedClauses(Split));
	split_SetBlockedClauses(Split, list_Nil());

	Recover = list_Nconc(split_DeletedClauses(Split), Recover);
	if (prfs_GetFatherClause(Split) != (CLAUSE)NULL) {
	  Recover = list_Cons(prfs_GetFatherClause(Split),Recover);
	  prfs_SetFatherClause(Split,NULL);
	}
	split_SetDeletedClauses(Split, list_Nil());
	Recover = split_DeleteClausesDependingOnLevelFromList(PS,Recover, SplitLevel);

	ScanStack = prfs_SplitStack(PS);
	while (!list_StackEmpty(ScanStack) &&  split_Level(list_Car(ScanStack)) > LastBacktrackLevel) {
	  Deleted = split_DeletedClauses(list_Car(ScanStack));
	  Deleted = split_DeleteClausesDependingOnLevelFromList(PS,Deleted, SplitLevel);
	  split_SetDeletedClauses(list_Car(ScanStack), Deleted);
	  ScanStack = list_Cdr(ScanStack);
	}

	Recover = list_Nconc(Recover, split_DeleteClausesDependingOnLevelFromSet(PS, prfs_UsableClauses(PS), SplitLevel));
	Recover = list_Nconc(Recover, split_DeleteClausesDependingOnLevelFromSet(PS, prfs_WorkedOffClauses(PS), SplitLevel));

	split_CleanedUp(Split);
      }
    }

    SplitLevel--;
    Scan = list_Cdr(Scan);
  }
  return Recover;
}


void split_DeleteClauseAtLevel(PROOFSEARCH PS, CLAUSE Clause, int Level)
/**************************************************************
  INPUT:   A clause, a level and a proofsearch object
  RETURNS: None, 'Clause' is deleted from its list and made unshared.
***************************************************************/
{

  SPLIT Split;

  if (clause_GetFlag(Clause,WORKEDOFF)) 
    prfs_ExtractWorkedOff(PS, Clause);
  else 
    prfs_ExtractUsable(PS, Clause);

  Split = prfs_GetSplitOfLevel(Level, PS);
  split_SetDeletedClauses(Split, list_Cons(Clause, split_DeletedClauses(Split)));
}


void split_KeepClauseAtLevel(PROOFSEARCH PS, CLAUSE Clause, int Level)
/**************************************************************
  INPUT:   A clause and a level as int.
  RETURNS: None.
  MEMORY:  A copy of clause is made and kept within the split stack.
***************************************************************/
{
  SPLIT  Split;

  Split = prfs_GetSplitOfLevel(Level, PS);
  split_SetDeletedClauses(Split,list_Cons(Clause, split_DeletedClauses(Split)));
}


LIST split_ExtractEmptyClauses(LIST Clauses, LIST* EmptyClauses)
/**************************************************************
  INPUT:   A list of clauses and a pointer to a list of empty clauses.
  RETURNS: <Clauses> without all empty clauses where the empty clauses
           are moved to <EmptyClauses>
  MEMORY:  Destructive on <Clauses>.
***************************************************************/
{
  LIST   Scan;
  CLAUSE Clause;

  for(Scan=Clauses;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
    Clause = (CLAUSE)list_Car(Scan);
    if (clause_IsEmptyClause(Clause)) {
      *EmptyClauses = list_Cons(Clause,*EmptyClauses);
      list_Rplaca(Scan,NULL);
    }
  }
  Clauses = list_PointerDeleteElement(Clauses,NULL);

  return Clauses;
}

CLAUSE split_SmallestSplitLevelClause(LIST Clauses)
/**************************************************************
  INPUT:   A non-empty list of clauses.
  RETURNS: The clause with the smallest split level.
***************************************************************/
{
  CLAUSE Result;

  Result  = (CLAUSE)list_Car(Clauses);
  Clauses = list_Cdr(Clauses);
  
  while (!list_Empty(Clauses)) {
    if (clause_SplitLevel(Result) > clause_SplitLevel(list_Car(Clauses)))
      Result  = list_Car(Clauses);
    Clauses = list_Cdr(Clauses);
  }

  return Result;
}

