        /****************************************************************
        *                                                               *
        *                          FINDER 3.0                           *
        *                                                               *
        *           (C) 1993 Australian National University             *
        *                                                               *
        *                      All rights reserved                      *
        *                                                               *
        * The information in this software is subject to change without *
        * notice and should not be construed as a commitment by the     *
        * Australian National University. The Australian National Uni-  *
        * versity makes no representations about the suitability of     *
        * this software for any purpose. It is supplied "as is" without *
        * express or implied warranty.  If the software is modified in  *
        * a manner creating derivative copyright rights, appropriate    *
        * legends may be placed on the derivative work in addition to   *
        * that set forth above.                                         *
        *                                                               *
        * Permission to use, copy, modify and distribute this software  *
        * and its documentation for any purpose and without fee is      *
        * hereby granted, provided that both the above copyright notice *
        * and this permission notice appear in all copies and sup-      *
        * porting documentation, and that the name of the Australian    *
        * National University not be used in advertising or publicity   *
        * pertaining to distribution of the software without specific,  *
        * written prior permission.                                     *
        *                                                               *
        ****************************************************************/

#include "FINDER.h"


/*
*	The first function is called from main right at the start.
*	The name says it all.
*/

void initialize()
{
    int i, j;
    functor f;
    kind k;

/*
*	First zero all the array parts of array elements which are
*	to be null iff this element is undefined.
*/
    for ( i = 0; i < LEXMAX; i++ ) lexicon[i].word[0] = 0;
    for ( i = 0; i < ST_MAX; i++ ) sort[i].sname = 0;
    for ( i = 0; i < FN_MAX; i++ ) fn[i].fname = 0;
    for ( i = 0; i < WF_MAX; i++ ) form[i].sym = 0;
    for ( i = 0; i < CL_MAX; i++ ) theory[i].prem = theory[i].conc = 0;
    theory[0].nextclause = 0;
    FN = 0;
/*
*	Next set up the pre-defined types bool and int, and the 
*	pre-defined relation and function symbols =, <, > and +.
*/
    k = newkind( "bool" );
    k->siz = k->sizmin = k->sizmax = 2;
    canon( k, "false" );  canon( k, "true" );
    f = newfunctor( "=", Func, 2 );
    f->val_type = k;	f->scope = 10;	f->table = *id;
    f = newfunctor( "<", Func, 2 );
    f->val_type = k;	f->scope = 10;	f->table = *less;
    f = newfunctor( ">", Func, 2 );
    f->val_type = k;	f->scope = 10;	f->table = *greater;	
    k = newkind( "int" );
    k->sizmin = k->sizmax = SZ-1;
    f = newfunctor( "+", Func, 2 );  /* The dummy typeless addition */
    f = newfunctor( "E!", Func, 1 );  /* The dummy typeless "exists" */
    f = newfunctor( "LEN", Func, 1 );  /* The dummy typeless length */
/*
*	Now the parameters have default values as follows
*/
    setting.max_models = 0;
    setting.max_backtracks = 0;
    setting.test_point = 100;
    setting.latest_priority = 1;
    setting.pretest_limit = 3;
    setting.batches = S_MAXIMAL;
    setting.extra_batches = 5;
    setting.maxref = MAX_REF_SIZE;
    setting.test_maxref = MAX_REF_SIZE;
    setting.maxtime = 0;
    setting.topsoff = true;
    setting.vars_over_wt = true;
    setting.report_breaks = true;
    setting.noisy = false;
    setting.timestamps = false;
    setting.Nosecs = false;
    setting.invert_change = false;
    setting.pm.verbosity = pfmodels | pbstats;
    setting.pm.pretty = true;
/*
*	And the addresses of the test functions are placed in the
*	structures to be passed to transref
*/
    init_trin( &tr_parameters );
    tr_parameters.Test = model;
    tr_parameters.Onerefs = setposs;
    tr_parameters.Smallrefs = do_smallrefs;

    init_trin( &ttr_par );
    ttr_par.Test = model_t;
    ttr_par.Onerefs = single_t;
    ttr_par.Smallrefs = tuple_t;
/*
*	That completes the initialisation of the special structures.
*	Next there are various integers, booleans and characters.
*/
    for ( i = 0; i < SZ; i++ ) for ( j = 0; j < SZ; j++ ) {
	id[i][j] = (i == j);
	less[i][j] = (i > j);       /* Looks wrong but is right */
	greater[i][j] = (i < j);    /* because SZto reverses i and j */
    }
    zero = 0;
    tot = good = 0;
    for ( i = 0; i < V_LENGTH; i++ ) {
	tr_parameters.forbidden[i] = false;
	thisvector[i] = lastgood[i] = 0;
    }
    for ( i = 0; i < ST_MAX; i++ ) lastsizes[i] = 0;
    lastvlength = 0;
    line_input = 0;
    single_model = false;
    cc_badguy = false;
    thmax = 0;
    TH = theory;
    for ( i = 0; i < VARMAX; i++ ) varlex[i][0] = 0;
    dummy_symbol = (functor) malloc(sizeof(FUNCTOR));
    dummy_symbol->fname = 0;
    first_newval = true;
    set_SZto();
}



/*
*	This function adds a canonical name nn to the enumeration
*	of the objects of sort k, setting sizmin and sizmax.
*/

void canon(kind k, char nn[])
{
    strls *lpp;
    functor f;
    int i;

    f = newfunctor( nn, Const, 0 );
    f->val_type = k;
    k->siz = 1;
    for (lpp = &(k->canonical); *lpp; lpp = &((*lpp)->nextname) )
	k->siz++;
    *(f->table) = k->siz-1;
    k->len[k->siz-1] = strlen(f->fname);
    *lpp = (strls) malloc(sizeof(STRLS));
    (*lpp)->name = f->fname;
    (*lpp)->nextname = 0;
    k->sizmin = k->sizmax = k->siz;
}



/*
*	Lex_pos scans the lexicon for the given string s, returning
*	its position in the lexicon if it exists there and if it is
*	a sort identifier iff kind_expected is true.  The two 
*	following functions call lex_pos.
*
*/

int lex_pos( char s[], boolean kind_expected )
{
    int i;

    for ( i = 0; lexicon[i].word[0]; i++ )
	if (!strcmp(s,lexicon[i].word) && lexicon[i].iskind == kind_expected)
	    return i;
    return -1;
}



/*
*	If the sort identifier with name s already exists, return a 
*	pointer to it.  Otherwise add it to the list, put its name
*	in the lexicon and initialize it to the default values.
*/

kind newkind( char s[] )
{
    int i;
    kind k;
    functor f;
    char string[LEXLENGTH];

    if ( lex_pos(s,false) != -1 )
	inerror(s,"doubly defined as a sort and a functor");
    if ( (i = lex_pos(s,true)) != -1 )
	return( sort+lexicon[i].offset );
    for ( i = 0; lexicon[i].word[0]; i++ );
    for ( k = sort; k->sname; k++ );
    strcpy(lexicon[i].word,s);
    lexicon[i].offset = k-sort;
    lexicon[i].iskind = true;
    k->sname = lexicon[i].word;
    k->sizmin = 1;  k->sizmax = SZ-1;
    k->canonical = 0;
    strcpy(string,"+_"); strcpy(string+2,s);
    f = newfunctor( string, Func, 2 );
    f->total = false;
    f->arg_type[0] = k;  f->arg_type[1] = Integer;  f->val_type = k;  
    f->table = *(k->sum);
    strcpy(string,"E!_"); strcpy(string+3,s);
    f = newfunctor( string, Func, 1 );
    f->arg_type[0] = k;  f->arg_type[1] = 0;  f->val_type = Bool;  
    f->table = k->Ex;
    strcpy(string,"LEN_"); strcpy(string+4,s);
    f = newfunctor( string, Func, 1 );
    f->arg_type[0] = k;  f->arg_type[1] = 0;  f->val_type = Integer;  
    f->table = k->len;
    for ( i = 0; i < SZ; i++ ) k->len[i] = (i < 10? 1: 2);
    k->piped = pipe_open;
    return k;
}




/*
*	This is very much the same thing, for functors instead of 
*	kinds.  In this case the category and adicity are passed 
*	as extra parameters.  The table (matrix) is allocated at
*	this point and is of a size appropriate to the adicity.
*/

functor newfunctor( char s[], syntype t, int a )
{
    int i;
    functor f, f2;

    if (a > ARGMAX)
	inerror(s,"given too many argument places");
    if ( lex_pos(s,true) != -1 )
	inerror(s,"doubly defined as a sort and a functor");
    if ( (i = lex_pos(s,false)) != -1 )
	return( fn+lexicon[i].offset );
    for ( i = 0; lexicon[i].word[0]; i++ );
    for ( f = fn; f->fname; f++ );
    strcpy(lexicon[i].word,s);
    lexicon[i].offset = f-fn;
    lexicon[i].iskind = false;
    f->fname = lexicon[i].word;
    f->category = t;
    if ( a == 2 ) f->scope = 1;
    else f->scope = 0;
    f->adicity = a;
    if ( t == Var ) inerror(s,"both a variable and a functor");
    else if ( *s == '=' ) f->table = *id;
    else if ( *s == '<' ) f->table = *less;
    else if ( *s == '>' ) f->table = *greater;
    else if ( *s != '+' )
	f->table = (int *) malloc(sizeof(int) * SZto[f->adicity]);
    f->total = true;
    f->injective = f->leftin = f->rightin = false;
    f->surjective = f->leftsur = f->rightsur = false;
    f->comm = false;
    f->nop = false;
    f->cut_it = false;
    f->piped = pipe_open;
    for (i=0; i<ARGMAX; i++) f->arg_type[i] = 0;
    f->val_type = 0;
    f->pmode = null;
    f->nextfunct = 0;
    if ( t != Const && !strchr("+<>=",*s) )
	if ( *s != 'E' || s[1] != '!' )
	    if ( *s != 'L' || s[1] != 'E' || s[2] != 'N' ) {
		if ( !FN ) FN = f;
		else {
		    for ( f2 = FN; f2->nextfunct; f2 = f2->nextfunct );
		    f2->nextfunct = f;
		}
	    }
    return f;
}
