        /****************************************************************
        *                                                               *
        *                          FINDER 3.0                           *
        *                                                               *
        *           (C) 1993 Australian National University             *
        *                                                               *
        *                      All rights reserved                      *
        *                                                               *
        * The information in this software is subject to change without *
        * notice and should not be construed as a commitment by the     *
        * Australian National University. The Australian National Uni-  *
        * versity makes no representations about the suitability of     *
        * this software for any purpose. It is supplied "as is" without *
        * express or implied warranty.  If the software is modified in  *
        * a manner creating derivative copyright rights, appropriate    *
        * legends may be placed on the derivative work in addition to   *
        * that set forth above.                                         *
        *                                                               *
        * Permission to use, copy, modify and distribute this software  *
        * and its documentation for any purpose and without fee is      *
        * hereby granted, provided that both the above copyright notice *
        * and this permission notice appear in all copies and sup-      *
        * porting documentation, and that the name of the Australian    *
        * National University not be used in advertising or publicity   *
        * pertaining to distribution of the software without specific,  *
        * written prior permission.                                     *
        *                                                               *
        ****************************************************************/


#include "FINDER.h"



static char	job_line[JL_MAX];	/* Line of input text		*/
static int	thisline = 0;		/* Current line number		*/
static FILE	*source;		/* Current input stream		*/



/*
*	First we have a bunch of small functions to perform input in 
*	meaningful symbols rather than character by character.
*
*	get_a_char returns the next character from the input line,
*	skipping over comments and refilling the input line if it 
*	is exhausted.  The input line may be job_line (if input is
*	coming from stdin) or command_line (if input is given using
*	the -c command line option.
*
*	Getchar is the same except that it skips white space.
*
*	getword fills its string parameter with the next meaningful
*	symbol from the current input source.
*
*	get_int reads an unsigned integer from the current input source.
*
*	get_st reads top level key words and turns them into the codes
*	for the corresponding statement types.
*/

char get_a_char()
{
    if ( line_input ) return command_line[input_point++];
    while ( job_line[input_point] == '%' || !job_line[input_point] ) {
	fgets(job_line,250,source);
	input_point = 0;
	thisline++;
    }	
    return job_line[input_point++];
}


char Getchar()
{
    char c;

    while ( isspace(c = get_a_char()) ) ;
    return c;
}


void getword( char w[] )
{
    int i, coff;

    *w = Getchar();
    if ( strchr("=><+,:.%)(}{",*w) ) w[1] = 0;
    else {
	for ( i = 1; !(isspace(w[i] = get_a_char())); i++ )
	    if ( strchr("=<+,.:%)(}{",w[i]) || 
		( w[i] == '>' && w[i-1] != '-' )) {
		input_point--;
		break;
	    }
	w[i] = 0;
    }
}


boolean isint( char *s )
{
    do if ( !isdigit(*s) ) return false;
    while ( *(++s) );
    return true;
}


int get_int()
{
    char s[LEXLENGTH];
    int rc, i;

    do getword(s); while ( *s == ':' );
    rc = 0;
    for ( i = 0; s[i]; i++ ) {
	if ( !isdigit(s[i]) )
	    inerror(s,"found where integer expected");
	rc *= 10;
	rc += (s[i] - '0');
    }
    return rc;
}


statement_type get_st()
{
    char wd[20];

    getword(wd);
    if ( !strcmp(wd,".") || !strcmp(wd,"end") ) return nul;
    if ( !strcmp(wd,"sort") ) return srt;
    if ( !strcmp(wd,"function") ) return fcn;
    if ( !strcmp(wd,"constant") ) return fcn0;
    if ( !strcmp(wd,"clause") ) return cls;
    if ( !strcmp(wd,"setting") ) return prm;
    if ( !strcmp(wd,"include") ) return inc;
    inwarning
	(wd,"found where key word (sort/function/clause/setting/end) expected");
}




/*
*	Here is the main function for this file.  It should be quite
*	self-explanatory.
*/

void read_job( FILE *infil )
{
    *job_line = 0;
    input_point = 0;
    source = infil;
    for (;;)
	switch( get_st() ) {
	  case nul: return;
	  case srt: get_item(get_sort,0,0); break;
	  case fcn: get_item(get_func,0,0); break;
	  case fcn0: get_item(get_const,0,0); break;
	  case cls: get_item(getclause,0,0); break;
	  case prm: get_item(get_set,0,0); break;
	  case inc: get_incl(); break;
	}
}


/*
*	Get_incl causes a file to be read as though it were
*	included at the current point. Note that it must end
*	with the word `end'. The file name is enclosed in 
*	(round) parentheses. Since the routine Getchar() 
*	is used, the name may have white space such as line
*	breaks inserted in it: these will be ignored.
*/

void get_incl()
{
    FILE *oldsource, *infil;
    char include_filename[128], oldjobline[256];
    int oldinputpoint, i;

    getword( include_filename );
    if ( *include_filename != '(' )
	inerror("Included filename must be in parentheses","");

    i = 0;
    while ( (include_filename[i] = Getchar()) != ')' ) i++;
    include_filename[i] = '\0';

    oldsource = source;
    strcpy(oldjobline, job_line);
    oldinputpoint = input_point;
    if ( !(infil = fopen(include_filename,"r")) )
	inerror("Cannot open file",include_filename);
    read_job( infil );
    fclose(source);
    source = oldsource;
    strcpy(job_line, oldjobline);
    input_point = oldinputpoint;
}



void addclause(char s[])
{
    char as[LEXLENGTH];

    strcpy(job_line,s);
    input_point = 0;
    getword(as);
    Get_clause(as,0);
}


/*
*	Get_item calls the appropriate function (*item)() to get either
*	one item of the given species (e.g. one clause) or a set of them.
*	A set, naturally, is specified by listing its members within a 
*	pair of braces.
*
*	Some items like clauses have an optional integer parameter given
*	in parentheses after the key word, so get_item must search for 
*	this first.
*
*	Functions called by get_item have all four of the parameters it 
*	issues, but generally don't use most of them.  Although the above
*	function read_job does not make use of the parameters k and f, 
*	functions (*item)() may do so when they themselves call get_item 
*	for their sub-items.
*/

void get_item( void (*item)(), kind k, functor f )
{
    char s[LEXLENGTH];
    int x = 0;

    getword(s);
    if ( *s == '(' ) {
	x = get_int();
	getword(s);
	if ( *s != ')' ) inerror("Missing right parenthesis","");
	getword(s);
    }
		
    if ( *s == '{' ) {
    	getword(s);
	while ( *s != '}' ) {
	    if ( *s != '.' ) (*item)(k,f,s,x);
	    getword(s);
	}
    }
    else if ( *s != '.' ) (*item)(k,f,s,x);
}




/*
*	This is structured programming at its glorious best!
*/

void get_sort( kind k, functor f, char s[], int x )
{	
    get_item( sort_spec, newkind(s), 0 );
}




/*
*	Get specifications (if any) for the kind k generated or found
*	by newkind as called from get_sort above.  Recall that newkind
*	actually returns an old kind if it can!
*/

void sort_spec( kind k, functor f, char s[], int x )
{
    char specs[LEXLENGTH];
    int canon_count;

    if ( !strcmp(s,"cardinality") || !strcmp(s,"card") ) {
	if ( k->canonical ) {
	    inwarning(k->sname,"an enumerated sort - can't set cardinality");
	    return;
	}
	getword(specs);
	switch (*specs) {
	  case '=': k->sizmax = k->sizmin = get_int(); return;
	  case '<': k->sizmax = get_int()-1;
	    if ( k->sizmax < k->sizmin ) k->sizmin = k->sizmax;
	    return;
	  case '>': k->sizmin = get_int() + 1;
	    if ( k->sizmax < k->sizmin ) k->sizmax = k->sizmin;
	    return;
	}
	inerror("","Statement of cardinality needs =, < or >");		
    }
    if (!strcmp(s,"enum") ) {
	de_enumerate( k );
	for ( canon_count = 0; canon_count < SZ; canon_count++ ) {
	    getword(specs);
	    if ( *specs == ':' ) getword(specs);
	    canon( k, specs );
	    getword(specs);
	    if ( *specs == '.' ) return;
	    if ( *specs != ',' )
		inerror(specs,"found where comma expected");
	}
    }
    inwarning(s,"found where key word \"cardinality\" or \"enum\" expected");
}



/*
*	De_enumerate removes the existing enumeration if any.
*/

void de_enumerate( kind k )
{
    free_canon( k->canonical );
    k->canonical = 0;
}

void free_canon( strls lp )
{
    if ( !lp ) return;
    free_canon( lp->nextname );
    free( lp->nextname );
    lp->nextname = 0;
}




/*
*	Getting a new function symbol is more complicated than getting
*	a new sort symbol because the symbol itself is followed by its
*	type declaration.  Newfunctor also returns an old functor if 
*	there is one with the given name.  It sets the print mode to 
*	the default value null (which it needs to do because it usually
*	sets up variables and the like) so this function resets it to
*	the other default full.
*
*	Parameters k, f and x are dummies for get_func.  The string s
*	holds the name of the function.
*/

void get_func( kind k, functor f, char s[], int x )
{
    int i;
    char gfs[LEXLENGTH], typestrings[ARGMAX+1][LEXLENGTH];
    int arity;
    boolean got_argtypes, got_valtype;

    for ( i = 0; i <= ARGMAX; i++ ) typestrings[i][0] = '\0';
    arity = 0;
    got_argtypes = got_valtype = false;
    getword(gfs);
    if ( *gfs == ':' ) getword(gfs);
    do {
	if (arity > ARGMAX)
	    inerror("Adicity too great for function",f->fname);
	strcpy(typestrings[arity++],gfs);
	if (got_argtypes) {
	    got_valtype = true;
	    arity--;
	}
	else {
	    getword(gfs);
	    if (!strchr(".,-",*gfs))
		inerror(gfs,"found where punctuation ( , -> or . ) expected");
	    switch(*gfs) {
	      case '.':
		arity--;
		got_valtype = true;
		break;
	      case ',': 
		if (got_argtypes)
		    inerror("Incorrect value type for",f->fname);
		getword(gfs);
		break;
	      case '-':
		if ( gfs[1] != '>' )
		    inerror(gfs,"found where \"->\" expected");
		got_argtypes = true;
		getword(gfs);
		break;
	    }
	}
    }
    while (!got_valtype);
    f = newfunctor( s, Func, arity);
    f->val_type = set_up_type(typestrings[arity]);
    for (i=0; i<arity; i++)
	f->arg_type[i] = set_up_type(typestrings[i]);
    f->pmode = full;
    get_item( func_spec, 0, f );
}


/*
*	Get_const is just as above, but for 0-adics only.
*/

void get_const( kind k, functor f, char s[], int x )
{
    int i;
    char gfs[LEXLENGTH];

    f = newfunctor( s, Func, 0 );
    do	getword(gfs);
    while ( *gfs == ':' );
    f->val_type = set_up_type(gfs);
    f->pmode = full;
    get_item( func_spec, 0, f );
}



/*
*	Set_up_type is just a macro for get_func.
*/

kind set_up_type( char s[] )
{
    int i;

    for ( i = 0; sort[i].sname; i++ )
	if ( !strcmp(sort[i].sname,s) )
	    return( sort+i );
    inwarning("Non-existent sort",s);
}




/*
*	Various properties of the function can be specified in the 
*	section following its declaration.  Note that k and x are 
*	not used in this function, but f should point to a functor 
*	and s should hold the key word of the specification.
*/

void func_spec( kind k, functor f, char s[], int x )
{
    char specs[LEXLENGTH];
    int i, j;

    if ( !strcmp(s,"total") ) f->total = true;
    else if ( !strcmp(s,"partial") ) f->total = false;
    else if ( !strcmp(s,"injective") ) f->injective = true;
    else if ( !strcmp(s,"left-injective") || !strcmp(s,"column-injective") )
	f->leftin = (f->adicity == 2);
    else if ( !strcmp(s,"right-injective") || !strcmp(s,"row-injective") )
	f->rightin = (f->adicity == 2);
    else if ( !strcmp(s,"surjective") ) f->surjective = true;
    else if ( !strcmp(s,"left-surjective") || !strcmp(s,"column-surjective") )
	f->leftsur = (f->adicity == 2);
    else if ( !strcmp(s,"right-surjective") || !strcmp(s,"row-surjective") )
	f->rightsur = (f->adicity == 2);
    else if ( !strcmp(s,"bijective") )
	f->surjective = f->injective = true;
    else if ( !strcmp(s,"left-bijective") || !strcmp(s,"column-bijective") )
	f->leftsur = f->leftin = (f->adicity == 2);
    else if ( !strcmp(s,"right-bijective") || !strcmp(s,"row-bijective") )
	f->rightsur = f->rightin = (f->adicity == 2);
    else if ( !strcmp(s,"commutative") ) f->comm = (f->adicity == 2);
    else if ( !strcmp(s,"cut") || *s == '!' ) f->cut_it = true;
    else if ( !strcmp(s,"no-priority") ) f->nop = true;
    else if ( !strcmp(s,"print") ) {
	getword(specs);
	if ( *specs == ':' ) getword(specs);
	if ( !strcmp(specs,"none") ) f->pmode = null;
	else if ( !strcmp(specs,"brief") ) f->pmode = brief;
	else if ( !strcmp(specs,"full") ) f->pmode = full;
	else inwarning(specs,"found where print mode expected");
	return;
    }
    else if ( !strcmp(s,"scope") ) {
	f->scope = get_int();
	if ( f->scope > 10 ) f->scope = 10;
	return;
    }
    else inwarning(s,"found where function specification expected");
}


void getclause( kind k, functor f, char s[], int x)
{
    clause c;

    Get_clause( s, x);
    add_to_theory();
}


void Get_clause( char s[], int x )
{
    wls *ins;
    wff tempwff;
    boolean gcp_done = false;
    int checkwffs();

    latest_clause.priority = x;
    tempwff = 0;
    begin_parse();
    do {
	if ( *s != '-' ) {
	    tempwff = parse( s );
	    ins = gcp_done? &(latest_clause.conc): 
		&(latest_clause.prem);
	}
	if ( !strchr(",-.",*s) )
	    inerror(s,"found in clause where punctuation expected");
	switch( *s ) {
	  case '.':	gcp_done = true;
	    ins = &(latest_clause.conc);
	    break;
	  case '-':	if ( s[1] != '>' )
	      inerror(s,"found where \"->\" expected");
	    gcp_done = true;
	  case ',':	getword(s);
	}
	if ( tempwff ) {
	    while ( *ins ) ins = &((*ins)->nxt);
	    *ins =
		(wls) malloc(sizeof(WLS));
	    (*ins)->w = tempwff;
	    (*ins)->nxt = 0;
	}
    }
    while ( *s != '.' );
    typecheck();
    end_parse();
    get_weight();
}




void get_set( kind k, functor f, char s[], int x )
{
    void verb_spec();

    if ( !strcmp(s,"solutions") ) setting.max_models = get_int();
    else if ( !strcmp(s,"backtracks") ) setting.max_backtracks = get_int();
    else if ( !strcmp(s,"test-after") ) setting.test_point = get_int();
    else if ( !strcmp(s,"current-priority") )
	setting.latest_priority = get_int();
    else if ( !strcmp(s,"stack") ) {
	getword(s);
	if ( *s == ':' ) getword(s);
	if ( !strcmp(s,"itsy") ) setting.batches = S_ITSY;
	else if ( !strcmp(s,"bitsy") ) setting.batches = S_BITSY;
	else if ( !strcmp(s,"tiny") ) setting.batches = S_TINY;
	else if ( !strcmp(s,"small") ) setting.batches = S_SMALL;
	else if ( !strcmp(s,"medium") ) setting.batches = S_MEDIUM;
	else if ( !strcmp(s,"large") ) setting.batches = S_LARGE;
	else if ( !strcmp(s,"LARGE") ) setting.batches = S_VERYLARGE;
	else if ( !strcmp(s,"huge") ) setting.batches = S_HUGE;
	else if ( !strcmp(s,"maximal") ) setting.batches = S_MAXIMAL;
	else if ( isint(s) ) sscanf(s,"%d",&(setting.batches));
	else inwarning(s,"found where stack size expected");
	return;
    }
    else if ( !strcmp(s,"stack-margin") ) setting.extra_batches = get_int();
    else if ( !strcmp(s,"verbosity") ) get_item( verb_spec, 0, 0 );
    else if ( !strcmp(s,"ref-length") ) setting.maxref = get_int();
    else if ( !strcmp(s,"pre-test") ) setting.pretest_limit = get_int();
    else if ( !strcmp(s,"test-ref-length") ) setting.test_maxref = get_int();
    else if ( !strcmp(s,"time-limit") ) setting.maxtime = get_int();
    
    else if ( !strcmp(s,"primary") ) setting.Nosecs = true;
    else if ( !strcmp(s,"secondary") ) setting.Nosecs = false;
    
    else if ( !strcmp(s,"noisy") ) setting.noisy = true;
    else if ( !strcmp(s,"quiet") ) setting.noisy = false;
    
    else if ( !strcmp(s,"time-stamp") ) setting.timestamps = true;
    else if ( !strcmp(s,"no-time-stamp") ) setting.timestamps = false;
    
    else if ( !strcmp(s,"report-breaks") ) setting.report_breaks = true;
    else if ( !strcmp(s,"silent-breaks") ) setting.report_breaks = false;
    
    else if ( !strcmp(s,"tops-off") ) setting.topsoff = true;
    else if ( !strcmp(s,"tops-on") ) setting.topsoff = false;
    
    else if ( !strcmp(s,"reverse-test") ) setting.invert_change = true;
    else if ( !strcmp(s,"obverse-test") ) setting.invert_change = false;
    
    else if ( !strcmp(s,"test-by-vars") ) setting.vars_over_wt = true;
    else if ( !strcmp(s,"test-by-weight") ) setting.vars_over_wt = false;
    
    else if ( !strcmp(s,"pretty") ) setting.pm.pretty = true;
    else if ( !strcmp(s,"ugly") ) {
	setting.pm.pretty = false;
	setting.pm.verbosity = pugly;
    }
    else inwarning("Unrecognosed setting option",s);
}





void verb_spec( kind k, functor f, char s[], int x )
{
    unsigned vb;

    vb = setting.pm.verbosity;
    if ( !strcmp(s,"models") ) vb = verb_spec_spec(vb,pmodels);
    else if ( !strcmp(s,"job") ) vb = verb_spec_spec(vb,pjob);
    else if ( !strcmp(s,"stats") ) vb = verb_spec_spec(vb,pstats);
    else {
	inwarning("Unrecognosed verbosity option",s);
	verb_spec_spec(0,0);
    }
    setting.pm.verbosity = vb;
}


unsigned verb_spec_spec( unsigned vb, unsigned mask )
{
    char s[LEXLENGTH];

    getword(s);
    if ( *s == ':' ) getword(s);
    vb &= ~mask;
    if ( !strcmp(s,"none") ) return( vb );
    if ( !strcmp(s,"brief") ) return( vb | (mask & (mask >> 1)) );
    if ( !strcmp(s,"full") ) return( vb | (mask & (mask << 1)) );
    inwarning(s, "found where level (none, brief or full) expected");
    return vb;
}




boolean got_partial()
{
    char s[LEXLENGTH];

    source = stdin;
    if ( !piping ) return false;
    getword(s);
    if ( !strcmp(s,"end") ) return false;
    if ( strcmp(s,"model") ) {
	fprintf(stderr,"\n\"%s\" found where \"model\" expected\n",s);
	return false;
    }
    while ( got_function() );
    return true;
}




boolean got_function()
{
    char s[LEXLENGTH];
    int i, j;
    functor f;
    kind k;

    getword(s);
    if ( *s == '.' ) return false;
    if ( !strcmp(s,"size") ) {
	getword(s);
	if ( (i = lex_pos(s,true)) != -1 )
	    k = sort+lexicon[i].offset;
	else skipout("Incorrect sort name transfer",1);
	k->siz = get_int();
	return true;
    }
    if ( (i = lex_pos(s,false)) != -1 )
	f = fn+lexicon[i].offset;
    else skipout("Incorrect function name transfer",1);
    switch( f->adicity ) {
      case 0:	*f->table = get_int();
	break;
      case 1:	for ( i = 0; i < f->arg_type[0]->siz; i++ )
	  *(f->table+i) = get_int();
	*(f->table+i) = f->val_type->siz;
	break;
      case 2:	for ( i = 0; i <= f->arg_type[0]->siz; i++ )
	  for ( j = 0; j <= f->arg_type[1]->siz; j++ )
	      if ( i < f->arg_type[0]->siz && j < f->arg_type[1]->siz )
		  *(f->table+(SZ*i)+j) = get_int();
	      else
		  *(f->table+(SZ*i)+j) = f->val_type->siz;
    }
    return true;
}





void inerror(char *s1, char *s2)
{
    fprintf(stderr,"\nInput error on line %d: %s\n", thisline, job_line);
    fprintf(stderr,"\n%s %s\n\n",s1,s2);
    exit(thisline);
}


void inwarning(char *s1, char *s2)
{
    fprintf(stderr,
	    "\nWarning: possible error on line %d: %s\n", thisline, job_line);
    fprintf(stderr,"%s %s\n\n\n",s1,s2);
}
