/*
 *              Otter 3.0 -- Automated Deduction System
 *
 *                   Argonne National Laboratory
 *
 */

#define IN_MAIN  /* so that global vars in header.h will not be external.
                  * In other files these are declared as 'extern' variables    */

#include "header.h"

 int main(argc, argv)
     int argc;
     char **argv;
{
    struct clause *giv_cl;     /* given_clause: the driving 
                                * variable of the whole show.*/

    int errors, status, level, first_of_next_level;

    char *str;                 /* String to hold reason for stopping*/
    FILE *xlog_fp;             /* log file for Xwindows system (I think)*/
    
    non_portable_init();       /* init platform specific features -    
                                * fcn in file 'nonport.c'*/
    print_banner(argc, argv);  /* 1st thing we see - fcn in' main.c'*/
    init();                    /* initialize global variables -
                                * fcn in 'misc.c'*/

/*-------------------------------------------------------------------------*
 * Scott takes the following command line arguments:                       *
 * d                 perform dynamic semantic resolution.                  *
 * f <fileneme>      takes filename as the FINDER input file               *
 * w {n}             uses n as the weight to add to false clauses.         *
 * s {d}             uses d as a scale to apply to true clauses.           *
 * if w is given then f must also be supplied.                             *
 *-------------------------------------------------------------------------*/

#ifdef SCOTT
    CLOCK_START(FINDER_TIME)    /* Macro for time handling in 'macros.h*/

    /* Boot up FINDER for SCOTT - fcn in scott.c
     * Gets the finder input file and options, also makes a call to
     * init_finder to set up an initial model based on the clauses
     * (if any) and declarations in that file. */
        finder_init(argc, argv);

    CLOCK_STOP(FINDER_TIME)
#endif

    /* Otter input parser - fcn in 'misc.c'
     * includes a call to model the input clauses*/
        read_all_input(argc, argv); 

    errors = Stats[INPUT_ERRORS];
    if (errors != 0) {
    fprintf(stderr, "\n%d input errors were found.\007\n\n", errors);
    printf("%d input errors were found.\n", errors);
    exit(INPUT_ERROR_EXIT);
    }
    else {
    status = check_stop();

    /* get the first given clause - fcn in 'clauses.c' */
    giv_cl = extract_given_clause();

#ifdef SCOTT
    CLOCK_START(FINDER_TIME)
    test_if_dsr(giv_cl);    /* Test new clause if the DSR option is set */
    CLOCK_STOP(FINDER_TIME)
#endif

    level = 0;
    first_of_next_level = 0;

    if (Flags[LOG_FOR_X_SHOW].val)
        xlog_fp = init_log_for_x_show();

/* --------------------- MAIN LOOP STARTS HERE --------------------- */

    printf("\n=========== start of search ===========\n"); fflush(stdout);

    while (giv_cl != NULL && status == KEEP_SEARCHING) {

        if (Flags[SOS_QUEUE].val && giv_cl->id >= first_of_next_level) {
        level++;
        first_of_next_level = next_cl_num();
        printf("\nStarting on level %d, last kept clause of level %d is %d.\n\n",level, level-1, first_of_next_level-1);
        fprintf(stderr, "\nStarting on level %d, last kept clause of level %d is %d.\007\n\n", level, level-1, first_of_next_level-1);
        }

        if (Flags[LOG_FOR_X_SHOW].val)
        log_for_x_show(xlog_fp);

        Stats[CL_GIVEN]++; /* why here?? */

        if (Flags[PRINT_GIVEN].val) {
        printf("\ngiven clause #%ld: ", Stats[CL_GIVEN]);
#ifdef SCOTT
		if (giv_cl->true_in_model) printf("TRUE ");
#endif
        printf("(wt=%d) ", giv_cl->weight);
        print_clause(stdout, giv_cl); fflush(stdout);
        }

#if defined(DOS_GCC) || defined(TURBO_C) || defined(THINK_C)
        /* if DOS or Macintosh, so user knows something is happening */
            fprintf(stderr, ".");
            if (Stats[CL_GIVEN] % 50 == 0)
        fprintf(stderr, " %ld clauses given.\n", Stats[CL_GIVEN]);
#endif

        /* Make references for faster literal clashes - fcn in 'index.c'*/
        index_lits_clash(giv_cl);

        append_cl(Usable, giv_cl); /* Hmmm... does the coppy of giv_cl get 
                                    * literals indexed for clashing too?*/

        /********* Main proceedure *********/
        /* Apply inference rules with giv_cl as focus
         * plus preprocessing and post processing etc, integration 
         * includes calls to FINDER to model infered clauses
         * - fcn in 'process.c'.*/
        infer_and_process(giv_cl);

        /* Enter interactive mode after most recent inferences have been 
         * completed and before next FO loop begins*/
        if (Parms[INTERRUPT_GIVEN].val > 0 &&
        Stats[CL_GIVEN] % Parms[INTERRUPT_GIVEN].val == 0) {
        fprintf(stderr, "\007\n%d clauses have been given.\n", Stats[CL_GIVEN]);
        interact();
        }

        /* Test if any search limits have been reached - fcn in 'misc.c'*/
        status = check_stop();

        if (status == KEEP_SEARCHING) {

        /* Change weight limits if required*/
        if (Parms[CHANGE_LIMIT_AFTER].val == Stats[CL_GIVEN]) {
            int new_limit;
            new_limit = Parms[NEW_MAX_WEIGHT].val;
            Parms[MAX_WEIGHT].val = new_limit;
            printf("\nreducing weight limit to %d.\n", new_limit);
            }

        /* get next given clause then start loop again */
        giv_cl = extract_given_clause();

#ifdef SCOTT
        CLOCK_START(FINDER_TIME)
        test_if_dsr(giv_cl);  /* Test new clause if the DSR option is set */
        CLOCK_STOP(FINDER_TIME)
#endif
        } /* If KEEP_SEARCHING */

        if (status == KEEP_SEARCHING && giv_cl && Parms[REPORT].val > 0)
        report(); /* report current search stats - fcn in 'misc.c' */

        }  /* end of main loop */

/* --------------------- MAIN LOOP ENDS HERE --------------------- */

    /* print the reason the search ended */

    if (status == KEEP_SEARCHING) {
        status = SOS_EMPTY_EXIT;
        fprintf(stderr, "\nSearch stopped because sos empty.\007\n\n");
        printf("\nSearch stopped because sos empty.\n");
        }
    else {
        switch (status) {
          case MAX_GIVEN_EXIT: str = "max_given"; break;
          case MAX_GEN_EXIT: str = "max_gen"; break;
          case MAX_KEPT_EXIT: str = "max_kept"; break;
          case MAX_SECONDS_EXIT: str = "max_seconds"; break;
          default: str = "???"; break;
        }

        fprintf(stderr, "\nSearch stopped by %s option.\007\n\n", str);
        printf("\nSearch stopped by %s option.\n", str);
        }

    cleanup();
    exit(status);
    }

}  /* main */

/*************
 *
 *    void print_banner(argc, argv)
 *
 *************/

void print_banner(argc, argv)
     int argc;
     char **argv;
{
    int i;
    printf("----- Otter 3.0.3, Aug 1994 -----\n");
    printf("The job was started by %s on %s, %s",
        username(), hostname(),  /* these two defined in 'nonport.c' */
        get_time());

    printf("The command was \"");
         /* for each word in the command line..
          * (Note this skips anything after a pipe, etc)
          * print a space separated list, terminated with a period.*/
    for(i = 0; i < argc; i++) 
        printf("%s%s", argv[i], (i < argc-1 ? " " : ""));
    printf("\".\n\n");
}  /* print_banner */
