
        /****************************************************************
        *                                                               *
        *                          FINDER 3.0                           *
        *                                                               *
        *           (C) 1993 Australian National University             *
        *                                                               *
        *                      All rights reserved                      *
        *                                                               *
        * The information in this software is subject to change without *
        * notice and should not be construed as a commitment by the     *
        * Australian National University. The Australian National Uni-  *
        * versity makes no representations about the suitability of     *
        * this software for any purpose. It is supplied "as is" without *
        * express or implied warranty.  If the software is modified in  *
        * a manner creating derivative copyright rights, appropriate    *
        * legends may be placed on the derivative work in addition to   *
        * that set forth above.                                         *
        *                                                               *
        * Permission to use, copy, modify and distribute this software  *
        * and its documentation for any purpose and without fee is      *
        * hereby granted, provided that both the above copyright notice *
        * and this permission notice appear in all copies and sup-      *
        * porting documentation, and that the name of the Australian    *
        * National University not be used in advertising or publicity   *
        * pertaining to distribution of the software without specific,  *
        * written prior permission.                                     *
        *                                                               *
        ****************************************************************/


#include "FINDER.h"

/*
*	Local external variables for communication
*/

static boolean clause_is_valid;
static clause this_clause;




int tisgood( clause c )
{
    int i, j;

    this_clause = c;
    ttr_par.vlength = tset_cc();
    ttr_par.done = false;
    ttr_par.report_breaks = false;
    ttr_par.verb = 1;
    clause_is_valid = true;
    transref( &ttr_par );
    untset_cc();
    return clause_is_valid;
}



int tset_cc()
{
    wls wf;
    int tsmax = 0;

    for ( wf = this_clause->varlist; wf; wf = wf->nxt )
	wf->w->wcc = tsmax++;
    for ( wf = this_clause->testlist; wf; wf = wf->nxt )
	wf->w->wcc = tsmax++;
/*
** dump_wcc();
*/
    return tsmax;
}


void untset_cc()
{
    wls wf;

    for ( wf = this_clause->varlist; wf; wf = wf->nxt )
	wf->w->wcc = -1;
    for ( wf = this_clause->testlist; wf; wf = wf->nxt )
	wf->w->wcc = -1;
}


int model_t( unsigned info[], TRS *T )
{
    wls wf;
    wff badwf = 0;
    int i;
    int offset;

    if ( clause_is_valid ) {
	for ( wf = this_clause->varlist; wf; wf = wf->nxt )
	    wf->w->val = info[wf->w->wcc];
	
	for ( wf = this_clause->testlist; wf; wf = wf->nxt ) {
	    offset = 0;
	    for (i=0; i<wf->w->sym->adicity; i++)
		offset += (*(wf->w->sbval[i]) * SZto[i]);
	    wf->w->val = *(wf->w->matrix + offset);
	    if ( wf->w->val != info[wf->w->wcc] ) {
		badwf = wf->w;
		wf->w->val = info[wf->w->wcc];
	    }
	}

	if ( badwf ) {
	    Ref( badwf->wcc, T );
	    for (i=0; i<badwf->sym->adicity; i++)
		if ( badwf->sb[i]->wcc >= 0 )
		    Ref( badwf->sb[i]->wcc, T );
	}
	else clause_is_valid = false;
    }
    return false;		/* Sends 0-refutation if search is done */
}



int single_t( unsigned info[], TRS *T )
{
    wls wf;

    for ( wf = this_clause->prem; wf; wf = wf->nxt )
        info[wf->w->wcc] &= ~((unsigned)1);
    for ( wf = this_clause->conc; wf; wf = wf->nxt )
        info[wf->w->wcc] &= (unsigned)1;
    for ( wf = this_clause->varlist; wf; wf = wf->nxt )
        info[wf->w->wcc] &= wf->w->sym->val_type->sizmask;
    for ( wf = this_clause->testlist; wf; wf = wf->nxt ) {
	info[wf->w->wcc] &= wf->w->sym->val_type->sizmask;
	if ( !wf->w->sym->adicity )
	    info[wf->w->wcc] &= (1 << *(wf->w->matrix));
    }
}
 
 
 
int tuple_t( unsigned info[], TRS *T )
{
    wls wf;
    wff w;
    int i, j, k;
    int cells[PRE_MAX], vals[PRE_MAX];
    boolean valency[PRE_MAX];
    int subvalues[ARGMAX];

    for ( wf = this_clause->testlist; wf; wf = wf->nxt ) {
	w = wf->w;
	for ( i = 0; i < w->sym->adicity; i++ )
	    subvalues[i] = 0;
	do {
	    j = matrix_offset(w->sym->adicity,subvalues);
	    k = *(w->matrix + j);
	    cells[0] = w->wcc;  vals[0] = k; valency[0] = true;
	    j = 1;
	    for (i=0; i<w->sym->adicity; i++)
		if (w->sb[i]->sym->category != Const) {
		    cells[j] = w->sb[i]->wcc;
		    vals[j] = subvalues[i];
		    valency[j++] = false;
		}
	    cells[j] = -1;
	    new_small_ref( cells, vals, valency, T );
	}
	while (next_instance(w,subvalues));
    }
}



dump_wcc()
{
    wls wf;
    int i;

    for ( wf = this_clause->varlist; wf; wf = wf->nxt )
	printf("%-6d %s\n", wf->w->wcc, wf->w->sym->fname);
    for ( wf = this_clause->testlist; wf; wf = wf->nxt ) {
	printf("%-6d %s", wf->w->wcc, wf->w->sym->fname);
	for (i=0; i<wf->w->sym->adicity; i++)
	    printf(" %d", wf->w->sb[i]->wcc);
	printf("\n");
    }
}


boolean next_instance(wff w, int subvalues[])
{
    int i;

    for (i=0; i<w->sym->adicity; i++)
	if (w->sb[i]->sym->category != Const) {
	    if (subvalues[i] == w->sym->arg_type[i]->siz-1)
		subvalues[i] = 0;
	    else {
		subvalues[i]++;
		return true;
	    }
	}
    return false;
}
