        /****************************************************************
        *                                                               *
        *                          FINDER 3.0                           *
        *                                                               *
        *           (C) 1993 Australian National University             *
        *                                                               *
        *                      All rights reserved                      *
        *                                                               *
        * The information in this software is subject to change without *
        * notice and should not be construed as a commitment by the     *
        * Australian National University. The Australian National Uni-  *
        * versity makes no representations about the suitability of     *
        * this software for any purpose. It is supplied "as is" without *
        * express or implied warranty.  If the software is modified in  *
        * a manner creating derivative copyright rights, appropriate    *
        * legends may be placed on the derivative work in addition to   *
        * that set forth above.                                         *
        *                                                               *
        * Permission to use, copy, modify and distribute this software  *
        * and its documentation for any purpose and without fee is      *
        * hereby granted, provided that both the above copyright notice *
        * and this permission notice appear in all copies and sup-      *
        * porting documentation, and that the name of the Australian    *
        * National University not be used in advertising or publicity   *
        * pertaining to distribution of the software without specific,  *
        * written prior permission.                                     *
        *                                                               *
        ****************************************************************/



#include "FINDER.h"



/*
*	These two are as trivial as they come.
*/

void print_the_job()
{
    fprint_the_job( stdout );
}

void print_job()
{
    fprint_job( stdout );
}




/*
*	This is the same as fprint_job except that the job is printed in full
*	regardless of the verbosity settings. It is not called from within
*	FINDER.
*/

void fprint_the_job( FILE *fil )
{
    unsigned tmp;

    tmp = setting.pm.verbosity;
    setting.pm.verbosity |= pfjob;
    fprint_job( fil );
    setting.pm.verbosity = tmp;
}



/*
*	Send the job description to fil, in FINDER input format.
*/

void fprint_job( FILE *fil )
{
    if ( !setting.pm.verbosity & pjob ) return;
    print_sorts(fil);
    print_functions(fil);
    if ( setting.pm.verbosity & pfjob ) {
	print_clauses(fil);
	print_setting(fil);
    }
    fprintf(fil,"\nend\n");
}



/*
*	First obvious subroutine of fprint_job.
*/

void print_sorts( FILE *fil )
{
    int i;

    fprintf(fil,"sort {\n");
    for ( i = 0; sort[i].sname; i++ ) {
	if ( i ) fprintf(fil,"\n");
	print_one_sort(fil,sort+i);
    }
    fprintf(fil,"}\n");
}



/*
*	Exactly what it says.
*/

void print_one_sort( FILE *fil, kind k )
{
    strls ep;

    if ( k->piped ) fprintf(fil,"\t%s {  %c piped\n\t\t", k->sname, '%');
    else fprintf(fil,"\t%s {\n\t\t", k->sname);
    if ( k->canonical ) {
	fprintf(fil,"enum: ");
	for ( ep = k->canonical; ep; ep = ep->nextname ) {
	    if ( ep != k->canonical ) fprintf(fil,", ");
	    fprintf(fil,ep->name);
	}
	fprintf(fil,".");
    }
    else if ( k->sizmin == k->sizmax )
	fprintf(fil,"cardinality = %d", k->sizmin);
    else fprintf(fil,"cardinality > %d\n\t\tcardinality < %d",
		 k->sizmin-1, k->sizmax+1);
    fprintf(fil,"\n\t}\n");
}



/*
*	Second obvious subroutine of fprint_job.
*/

void print_functions( FILE *fil )
{
    int i;
    functor f;

    fprintf(fil,"\nfunction {\n");
    i = 0;
    for ( f = FN; f; f = f->nextfunct )
	if ( setting.pm.pretty || f->pmode ) {
	    if ( ++i > 1 ) fprintf(fil,"\n");
	    print_one_function( fil, f );
	}
    fprintf(fil,"}\n");
}



/*
*	Describing a function symbol is more complex than describing a sort,
*	but conceptually it is no deeper.
*/

void print_one_function( FILE *fil, functor f )
{
    int i;

    fprintf(fil,"\t%s: ", f->fname);
    for ( i = 0; i < f->adicity; i++ ) {
	if ( i ) printf(", ");
	if ( f->arg_type[i] )
	    fprintf(fil,f->arg_type[i]->sname);
	else fprintf(fil,"??");
    }
    if ( f->adicity ) fprintf(fil," -> ");
    fprintf(fil,"%s", f->val_type->sname);
    if ( !f->adicity ) fprintf(fil,".");
    fprintf(fil," {\n\t", f->val_type->sname);
    if ( f->total ) fprintf(fil,"\ttotal\n\t");
    else fprintf(fil,"\tpartial\n\t");
    if ( f->injective && f->surjective ) fprintf(fil,"\tbijective\n\t");
    else {
	if ( f->injective ) fprintf(fil,"\tinjective\n\t");
	if ( f->surjective ) fprintf(fil,"\tsurjective\n\t");
    }	
    if ( f->leftin && f->leftsur ) fprintf(fil,"\tcolumn-bijective\n\t");
    else {
	if ( f->leftin ) fprintf(fil,"\tcolumn-injective\n\t");
	if ( f->leftsur ) fprintf(fil,"\tcolumn-surjective\n\t");
    }
    if ( f->rightin && f->rightsur ) fprintf(fil,"\trow-bijective\n\t");
    else {
	if ( f->rightin ) fprintf(fil,"\trow-injective\n\t");
	if ( f->rightsur ) fprintf(fil,"\trow-surjective\n\t");
    }
    if ( f->comm ) fprintf(fil,"\tcommutative\n\t");
    if ( f->cut_it ) fprintf(fil,"\tcut\n\t");
    if ( f->nop ) fprintf(fil,"\tno-priority\n\t");
    fprintf(fil,"\tscope: %d\n\t", f->scope);
    if ( f->pmode == null ) fprintf(fil,"\tprint: none\n\t");
    else if ( f->pmode == brief ) 
	fprintf(fil,"\tprint: brief\n\t");
    else	fprintf(fil,"\tprint: full\n\t");
    fprintf(fil,"}\n");
}




/*
*	Third obvious subroutine of fprint_job.
*/

void print_clauses( FILE *fil )
{
    clause c;

    fprintf(fil,"\nclause {\n");
    for ( c = TH; c; c = c->nextclause ) {
	fprintf(fil,"\t");
	print_one_clause(fil,c);
	if ( c->weight == 1 && c->nextclause && 
	    c->nextclause->weight > 1 ) fprintf(fil,"\n");
	if ( c->weight == 2 && c->nextclause && 
	    c->nextclause->weight > 2 ) fprintf(fil,"\n");
    }
    fprintf(fil,"}\n");
}



/*
*	Printing a clause is simple. Add the weight, priority and position.
*/

void print_one_clause( FILE *fil, clause c )
{
    wls wf;

    for ( wf = c->prem; wf; wf = wf->nxt ) {
	if ( wf != c->prem ) fprintf(fil,", ");
	outwff(fil,wf->w);
    }
    if ( !c->prem ) printf("true");
    fprintf(fil,"  ->  ");
    for ( wf = c->conc; wf; wf = wf->nxt ) {
	if ( wf != c->conc ) fprintf(fil,", ");
	outwff(fil,wf->w);
    }
    if ( !c->conc ) printf("false");
    fprintf(fil,".  %c (%d:%d:%d:%d)\n", 
	    '%', c->varcount, c->weight, c->priority, c-theory);
}



/*
*	Recursively print a formula in Polish.
*/

void outwff( FILE *fil, wff w )
{
    int i;

    if ( !w ) return;
    fprintf(fil,"%s",w->sym->fname);
    if ( w->sym->adicity ) {
	fprintf(fil,"(");
	for (i=0; i<w->sym->adicity; i++) {
	    if (i) fprintf(fil,",");
	    outwff(fil,w->sb[i]);
	}
	fprintf(fil,")");
    }
}




/*
*	Fourth obvious subroutine of fprint_job.
*/

void print_setting( FILE *fil )
{
    fprintf(fil,"\nsetting {\n");
    prints_int( fil, "solutions", setting.max_models );
    prints_int( fil, "backtracks", setting.max_backtracks );
    prints_int( fil, "test-after", setting.test_point );
    prints_int( fil, "current-priority", setting.latest_priority );
    prints_int( fil, "ref-length", setting.maxref );
    prints_int( fil, "test-ref-length", setting.test_maxref );
    prints_int( fil, "time-limit", setting.maxtime );
    prints_int( fil, "pre-test", setting.pretest_limit );
    prints_bool( fil, "primary", "secondary", setting.Nosecs );
    prints_bool( fil, "reverse-test", "obverse-test", setting.invert_change );
    prints_bool( fil, "pretty", "ugly", setting.pm.pretty );
    prints_bool( fil, "noisy", "quiet", setting.noisy );
    prints_bool( fil, "time-stamp", "no-time-stamp", setting.timestamps );
    prints_bool( fil, "tops-off", "tops-on", setting.topsoff );
    prints_bool( fil, "report-breaks", "silent-breaks", setting.report_breaks );
    prints_bool( fil, "test-by-vars", "test-by-weight", setting.vars_over_wt );
    fprintf(fil,"\tstack: ");
    if ( setting.batches == S_ITSY ) fprintf(fil,"itsy\n");
    if ( setting.batches == S_BITSY ) fprintf(fil,"bitsy\n");
    if ( setting.batches == S_TINY ) fprintf(fil,"tiny\n");
    else if ( setting.batches == S_SMALL ) fprintf(fil,"small\n");
    else if ( setting.batches == S_MEDIUM ) fprintf(fil,"medium\n");
    else if ( setting.batches == S_LARGE ) fprintf(fil,"large\n");
    else if ( setting.batches == S_VERYLARGE ) fprintf(fil,"LARGE\n");
    else if ( setting.batches == S_HUGE ) fprintf(fil,"huge\n");
    else if ( setting.batches == S_MAXIMAL ) fprintf(fil,"maximal\n");
    else fprintf(fil,"%d\n", setting.batches);
    fprintf(fil,"\tverbosity {\n\t");
    prints_verb( fil, "models", pmodels );
    prints_verb( fil, "job", pjob );
    prints_verb( fil, "stats", pstats );
    fprintf(fil,"}\n}\n");
}




/*
*	There now follow the subroutines called by print_set.
*/

void prints_int( FILE *fil, char *s, int x )
{
    if ( x ) fprintf(fil,"\t%s %d\n", s, x);
}


void prints_bool( FILE *fil, char *strue, char *sfalse, boolean b )
{
    fprintf(fil,"\t%s\n", b? strue: sfalse);
}


void prints_verb( FILE *fil, char *s, unsigned mask )
{
    unsigned u;

    u = setting.pm.verbosity & mask;
    if ( !u ) fprintf(fil,"\t%s: none\n\t", s);
    else if ( u & (mask >> 1) ) fprintf(fil,"\t%s: brief\n\t", s);
    else if ( u & (mask << 1) ) fprintf(fil,"\t%s: full\n\t", s);
    else skipout("Strange verbosity setting",1);
}


/*
=============================================================================
*/


/*
*	Print_model is called when there is a new model to report. It checks
*	for error conditions such as reporting the same model twice and 
*	as a side effect maintains the vector last_good. It also updates the
*	global counter good.
*
*	There are two formats, pretty and ugly (human-readable and machine-
*	readable). In addition, there are two print modes. Brief means print
*	only when there is a change to report, and Full means print no matter 
*	what. There is also None, which is self-explanatory. There is a 
*	general verbosity setting and a local one for each function. The 
*	lower of these two over-rides the higher.
*
*	Before the real thing starts, there are some trivia as in the 
*	above job-printing case.
*/

boolean print_the_model()
{
    return fprint_the_model( stdout);
}

boolean fprint_the_model( FILE *fil )
{
    unsigned tmp;
    boolean rc;

    tmp = setting.pm.verbosity;
    setting.pm.verbosity |= pfmodels;
    rc = fprint_model( fil );
    setting.pm.verbosity = tmp;
    return rc;
}

boolean print_model()
{
    return fprint_model( stdout );
}

boolean fprint_model( FILE *fil )
{
    kind k;
    functor f;
    int i, j;
    int timenow;
    boolean Full, Brief, header = true;
    functor un_vec_pos();

    if ( (setting.pm.verbosity & pmodels) && !setting.pm.pretty )
	fprintf(fil,"model\n");
    
    if ( firstmodel ) {
	firstmodel = false;
	if ( (setting.pm.verbosity & pmodels) && !setting.pm.pretty ) {
	    for ( k = sort+1; k->sname; k++ )
		fprintf(fil,"size %s %d\n", k->sname, k->siz);
	}
    }

    if ( setting.noisy ) fprintf(stderr,"\007");
    if ( !(setting.pm.verbosity & pmodels) ) return true;
    for ( f = FN; f; f = f->nextfunct ) if ( !negative || f->piped ) {
	Full = (f->pmode==full && setting.pm.verbosity & pfmodels);
	Brief = ((f->pmode==full && setting.pm.verbosity & pbmodels)
		 || (f->pmode==brief && setting.pm.verbosity & pmodels));
	if ( Full || ( Brief && (f->piped || f->fcc <= first_change) )) {
	    if ( header && setting.pm.pretty ) {
		if ( setting.timestamps ) {
		    timenow = timestamp();
		    fprintf(fil,
			    "\nModel %d (%3.2f seconds)\n", good,
			    (timenow*1.0)/60);
		}
		else fprintf(fil,"\nModel %d\n", good);
		header = false;
	    }
	    if (f->adicity > 2)
		fprintf(fil,
			"\n>> %d-dimensional tables not yet printable",
			f->adicity);
	    else switch( f->adicity ) {
	      case 0: zeroprint( fil, f );
		break;
	      case 1: oneprint( fil, f );
		break;
	      case 2: twoprint( fil, f );
	    }
	    fprintf(fil,"\n");
	}
    }
    if ( !setting.pm.pretty ) fprintf(fil,".\n");
    else if ( !header ) fprintf(fil,"\n");
    return true;
}



/*
*	Print a zero-adic function.
*/

void zeroprint( FILE *fil, functor f )
{
    int i;
    strls lp;

    if ( setting.pm.pretty ) {
	fprintf(fil,"\n%s = ", f->fname);
	printvalue( fil, f->val_type, *(f->table), 0 );
    }
    else fprintf(fil,"%s %d", f->fname, *(f->table));
}



/*
*	Print a monadic function.
*/

void oneprint( FILE *fil, functor f )
{
    int i, j, len, arglen, lastlen;
    strls lp;

    if ( setting.pm.pretty ) {
	len = 2;
	i = canonlength( f->arg_type[0] );
	if ( i > len ) len = i;
	i = canonlength( f->val_type );
	if ( i > len ) len = i;
	arglen = strlen(f->fname);
	i = strlen(f->arg_type[0]->sname);
	if ( i > arglen ) arglen = i;
	lastlen = f->arg_type[0]->len[f->arg_type[0]->siz-1];
	i = *(f->table + f->arg_type[0]->siz-1);
	if ( f->val_type->len[i] > lastlen )
	    lastlen = f->val_type->len[i];
	if ( lastlen < 2 ) lastlen = 2;

	fprintf(fil,"\n");
	for ( i = strlen(f->arg_type[0]->sname); i < arglen; i++ )
	    fprintf(fil," ");
	fprintf(fil,"%s | ", f->arg_type[0]->sname);
	for ( i = 0; i < f->arg_type[0]->siz; i++ )
	    printvalue( fil, f->arg_type[0], i, len );
	fprintf(fil,"\n");
	for ( i = 0; i <= arglen; i++ ) fprintf(fil,"-");
	fprintf(fil,"+-");
	for ( j = 1; j < f->arg_type[0]->siz; j++ )
	    for ( i = 0; i <= len; i++ ) fprintf(fil,"-");
	for ( i = 0; i < lastlen; i++ ) fprintf(fil,"-");
	fprintf(fil,"\n");
	for ( i = strlen(f->fname); i < arglen; i++ ) fprintf(fil," ");
	fprintf(fil,"%s | ", f->fname);
	for ( i = 0; i < f->arg_type[0]->siz; i++ )
	    printvalue( fil, f->val_type, *(f->table + i), len );
    }
    else {
	fprintf(fil,"%s ", f->fname);
	for ( i = 0; i < f->arg_type[0]->siz; i++ )
	    fprintf(fil," %d", *(f->table + i));
    }
}




/*
*	Much the same, but with a dyadic function.
*/

void twoprint( FILE *fil, functor f )
{
    int i, j, k, m, len, lenl, lastlen;
    strls lp;

    if ( setting.pm.pretty ) {
	lenl = strlen(f->fname);
	if ( lenl < 2 ) lenl = 2;
	i = canonlength( f->arg_type[0] );
	if ( i > lenl ) lenl = i;
	len = 2;
	len = canonlength( f->arg_type[1] );
	i = canonlength( f->val_type );
	if ( i > len ) len = i;
	lastlen = 2;
	k = f->arg_type[1]->siz-1;
	for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
	    j = f->val_type->len[*(f->table + SZ*k + i)];
	    if ( j > lastlen ) lastlen = j;
	}
	if ( f->arg_type[1]->len[f->arg_type[1]->siz-1] > lastlen )
	    lastlen = f->arg_type[1]->len[f->arg_type[1]->siz-1];
	
	fprintf(fil,"\n");
	for ( i = strlen(f->fname); i < lenl; i++ ) fprintf(fil," ");
	fprintf(fil,"%s | ", f->fname);
	for ( i = 0; i < f->arg_type[1]->siz; i++ )
	    printvalue( fil, f->arg_type[1], i, len );
	fprintf(fil,"\n");
	for ( i = 0; i <= lenl; i++ ) fprintf(fil,"-");
	fprintf(fil,"+-");
	for ( j = 1; j < f->arg_type[1]->siz; j++ )
	    for ( i = 0; i <= len; i++ ) fprintf(fil,"-");
	for ( i = 0; i < lastlen; i++ ) fprintf(fil,"-");
	for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
	    fprintf(fil,"\n");
	    printarg( fil, f->arg_type[0], i, lenl );
	    fprintf(fil," | ");
	    for ( j = 0; j < f->arg_type[1]->siz; j++ )
		printvalue( fil, f->val_type, *(f->table + SZ*j + i), len );
	}
    }
    else {
	fprintf(fil,"%s ", f->fname);
	for ( i = 0; i < f->arg_type[0]->siz; i++ )
	    for ( j = 0; j < f->arg_type[1]->siz; j++ )
		fprintf(fil," %d", *(f->table + SZ*i + j));
    }
}



/*
*	Return the length of the longest canonical name of sort k.
*/

int canonlength( kind k )
{
    int i = 0, j;

    for ( j = 0; j < k->siz; j++ )
	if ( k->len[j] > i ) i = k->len[j];
    return i;
}




/*
*	Print the canonical name of object x of kind k, padded on the right 
*	with spaces up to the length len.
*/

void printvalue( FILE *fil, kind k, int x, int len )
{
    strls lp;
    int i, j;

    if ( x == k->siz ) j = 0;
    else if ( k->canonical && setting.pm.pretty ) {
	lp = k->canonical;
	for ( i = 0; i < x; i++ ) lp = lp->nextname;
	fprintf(fil,lp->name);
	j = strlen(lp->name);
    }
    else {
	j = 2;
	fprintf(fil,"%2d",x);
    }
    for ( i = j; i <= len; i++ ) fprintf(fil," ");
    fflush(stdout);
}




/*
*	Rather similar, for printing the arguments of the function.
*/

void printarg( FILE *fil, kind k, int x, int len )
{
    strls lp;
    int i, j;

    if ( x == k->siz ) j = 0;
    else if ( k->canonical ) {
	lp = k->canonical;
	for ( i = 0; i < x; i++ ) lp = lp->nextname;
	j = strlen(lp->name);
    }
    else j = 2;
    for ( i = j; i < len; i++ ) fprintf(fil," ");

    if ( x == k->siz ) return;
    if ( k->canonical ) fprintf(fil,lp->name);
    else fprintf(fil,"%2d",x);
    fflush(stdout);
}



/*
============================================================================
*/


/*
*	When the FINDER job has been completed, some statistics on the run
*	are printed out. These are determined by the search controller used.
*/

void printstats()
{
    int i;
    void dump_report();

    printf("\n\n\n");
    dump_report( &tr_parameters );
    printf("\n");
    dump_report( &ttr_par );
    printf("FINDER time: %3.2f seconds\n\n",((end_time-begin_time)*1.0) / 60);
}



void dump_report( trin tr )
{
    int i;
    logical t;

    for ( i = 0; i < MAX_REPORT && (t = tr->reps[i].report_type); i++ ) {
	if ( i == SOLUTIONS )
	    printf("%s %d\n", tr->reps[i].s_report, tr->reps[i].i_report);
	else if ( (t & INT_REPORT) && tr->reps[i].i_report )
	    printf("%s %d\n", tr->reps[i].s_report, tr->reps[i].i_report);
	if ( t & DO_REPORT )
	    (*tr->reps[i].do_report)( tr );
	if ( (t & FLOAT_REPORT) && tr->reps[i].f_report )
	    printf("%s %3.2f\n",tr->reps[i].s_report,tr->reps[i].f_report);
    }
}





void ref_size_report()
{
    int i, j, k, m;
    int ctotal[RRB], rtotal[RRA], totaltotal;

    totaltotal = 0;
    for ( i = 0; i < RRB; i++ ) {
	ctotal[i] = 0;
	for ( j = 0; j < RRA; j++ )
	    ctotal[i] += tr_parameters.ref_record[j][i];
    }
    for ( i = 0; i < RRA; i++ ) {
	rtotal[i] = 0;
	for ( j = 0; j < RRB; j++ )
	    rtotal[i] += tr_parameters.ref_record[i][j];
	totaltotal += rtotal[i];
    }

    for ( k = RRA-1; k >= 0; k-- ) {
	j = 0;
	for ( i = 0; i < RRB; i++ )
	    if ( tr_parameters.ref_record[k][i] ) j = 1;
	if ( j ) break;
    }
    if ( k < 0 ) return;
    for ( m = RRB-1; m >= 0; m-- ) {
	j = 0;
	for ( i = 0; i <= k; i++ )
	    if ( tr_parameters.ref_record[i][m] ) j = 1;
	if ( j ) break;
    }
    if ( !m ) return;
    dump_refs( k+1, m+1 );
}




void report_branches( trin tr )
{
    int i, tsz;

    for ( tsz = SZ-1; tsz; tsz-- ) if ( tr->branches[tsz] ) break;
    if ( tsz ) {
	printf("\nBranching: ");
	for ( i = 1; i <= tsz; i++ )
	    printf(" %d: %-8d",i,tr->branches[i]);
	printf("\n\n");
    }
}


/*
*	The distribution matrix of the refutations is dumped to file
*	FINDER.refs. First the numbers of rows and columns are printed
*	along with the maximum number in the grid.
*/

void dump_refs( int rows, int columns )
{
    FILE *ref_file;
    int i, j, ccmax = 0;

    for ( i = rows-1; i >= 0; i-- )
	for ( j = 0; j < columns; j++ )
	    if ( tr_parameters.ref_record[i][j] > ccmax )
		ccmax = tr_parameters.ref_record[i][j];
    if ( !(ref_file = fopen("FINDER.refs","w")) ) return;
    fprintf(ref_file,"%d %d %d\n", rows, columns, ccmax);

    for ( i = rows-1; i >= 0; i-- ) {
	for ( j = 0; j < columns; j++ )
	    fprintf(ref_file, " %d", tr_parameters.ref_record[i][j]);
	fprintf(ref_file, "\n");
    }
    fclose(ref_file);
}
