/*
*			parser.c
*/

        /****************************************************************
        *                                                               *
        *                          FINDER 3.0                           *
        *                                                               *
        *           (C) 1993 Australian National University             *
        *                                                               *
        *                      All rights reserved                      *
        *                                                               *
        * The information in this software is subject to change without *
        * notice and should not be construed as a commitment by the     *
        * Australian National University. The Australian National Uni-  *
        * versity makes no representations about the suitability of     *
        * this software for any purpose. It is supplied "as is" without *
        * express or implied warranty.  If the software is modified in  *
        * a manner creating derivative copyright rights, appropriate    *
        * legends may be placed on the derivative work in addition to   *
        * that set forth above.                                         *
        *                                                               *
        * Permission to use, copy, modify and distribute this software  *
        * and its documentation for any purpose and without fee is      *
        * hereby granted, provided that both the above copyright notice *
        * and this permission notice appear in all copies and sup-      *
        * porting documentation, and that the name of the Australian    *
        * National University not be used in advertising or publicity   *
        * pertaining to distribution of the software without specific,  *
        * written prior permission.                                     *
        *                                                               *
        ****************************************************************/


#include "FINDER.h"
#define PWF_MAX 1024

/*
*	We need another data type for semi-parsed formulas
*/

typedef struct pswff {
    functor main;
    wff actual;
    struct pswff *leftsub, *rightsub;
} PSW, *psw;

void pw_merge( psw *pp, psw qq, functor mf );
wff pf_transcribe( psw pp );

static WFF pwf[PWF_MAX];
static boolean tads;
static PSW parse_wffs[PWF_MAX];
static int psw_next;

/*
*	This just zeros the temporary wff stack and the
*	various lists of the current clause. It is called
*	from Get_clause before parsing starts.
*/

void begin_parse()
{
    int i;

    for ( i = 0; i < PWF_MAX; i++ ) {
	pwf[i].sym = 0;
	pwf[i].val = -1;
    }
    free_wff_list(latest_clause.prem);
    latest_clause.prem = 0;
    free_wff_list(latest_clause.conc);
    latest_clause.conc = 0;
    free_wff_list(latest_clause.varlist);
    latest_clause.varlist = 0;
    free_wff_list(latest_clause.testlist);
    latest_clause.testlist = 0;
    psw_next = 0;
}


/*
*	This one is pretty trivial.
*/

void free_wff_list( wls ptr )
{
    if ( !ptr ) return;
    free_wff_list( ptr->nxt );
    free( ptr );
}


/*
*	The formula as string of symbols is a general wff (gwf)
*	which consists of one or more simple wffs (swf) separated
*	by binary functors.  A swf is either a functor followed 
*	by the appropriate number of swf to form its arguments or
*	a gwf in parentheses.
*
*	Hence we need two interacting parse functions, one for 
*	gwf and one for swf.  In each case the string parameter 
*	holds the first symbol of the sequence to be parsed. After 
*	the parse it holds the symbol which follows the formula.
*/



wff parse( char s[] )
{
    psw pp, qq;
    functor mf;
    int os;

    pp = parse_wffs + psw_next;
    if (++psw_next >= PWF_MAX)
	inerror("Apparently too many subformulae parsed","");
    pp->leftsub = pp->rightsub = 0;
    pp->main = 0;
    pp->actual = sparse(s);
    getword(s);
    while ( !strchr(".,)-",*s) ) {
	os = lex_pos(s,false);
	if ( os < 0 )
	    inerror(s,"found where a function symbol was expected");
	mf = fn+lexicon[os].offset;
	if ( mf->adicity != 2 )
	    inerror(s,"found where a binary operator was expected");
	qq = (psw) malloc(sizeof(PSW));
	qq->leftsub = qq->rightsub = 0;
	qq->main = 0;
	getword(s);
	qq->actual = sparse(s);
	pw_merge( &pp, qq, mf );
	getword(s);
    }
    return pf_transcribe(pp);
} /* END parse */




/*
*	Pw_merge puts the next functor and subformula into the tree of 
*	partially parsed subformulas. Where the tree so far is, say
*		pp
*	       /  \
*	      L    R
*		  / \
*		Rl   Rr
*	the result of merging (...mf qq) with it will be either
*		pp
*	       /  \
*	      L    R
*		  / \
*		 Rl  mf
*		    / \
*		   Rr  qq
*	or else
*		mf
*	       /  \
*	      pp   qq
*	     / \
*	    L   R
*	       / \
*	     Rl   Rr
*/

void pw_merge( psw *pp, psw qq, functor mf)
{
    psw rr;

    if ( !(*pp)->main || mf->scope >= (*pp)->main->scope ) {
	rr = (psw) malloc(sizeof(PSW));
	rr->leftsub = *pp;
	rr->rightsub = qq;
	rr->main = mf;
	rr->actual = 0;
	*pp = rr;
	return;
    }
    pw_merge( &((*pp)->rightsub), qq, mf );
    return;
} /* END pw_merge */



/*
*	Pf_transcribe is part of getting the semi-parsed inside of a pair of
*	parentheses into fully transcribed form. It returns the fully parsed 
*	subformula whose partially parsed version is rooted at pp.
*/

wff pf_transcribe( psw pp )
{
    wff subforms[2];

    if ( pp->actual ) return pp->actual;
    subforms[0] = pf_transcribe(pp->leftsub);
    subforms[1] = pf_transcribe(pp->rightsub);
    return locate( pp->main, subforms );
} /* END pf_transcribe */




/*
*	Sparse (subformula parse) is the other half of the back-and-forth 
*	recursive routine for parsing the next subformula in a string.
*	This deals with the cases other than that of a parenthesised 
*	subformula.
*/

wff sparse( char s[] )
{
    wff sp, spsb[ARGMAX];
    functor f;
    int i, os;

    if ( *s == '(' ) {
	getword(s);
	sp = parse(s);
	if ( *s != ')' )
	    inerror(s,"found where a right parenthesis was expected");
	return sp;
    }
    os = lex_pos(s,false);
    if ( os >= 0 ) f = fn+lexicon[os].offset;
    else {
	if ( isalpha(*s) )
	    return locate( newvar(s,0), spsb );
	else if ( isdigit(*s) ) {
	    for ( i = 0; s[i]; i++ ) if ( !isdigit(s[i]) )
		inerror(s,"starts numerically but isn't a number");
	    f = newfunctor( s, Const, 0 );
	    f->val_type = Integer;
	    *(f->table) = *s - '0';
	    if ( s[1] ) {
		*(f->table) *= 10;
		*(f->table) += (s[1] - '0');
	    }
	    return locate( f, spsb );
	}
	else inerror(s,"found where a function symbol was expected");
    }
    for (i=0; i<f->adicity; i++) {
	getword(s);
	if (f->adicity>1) {
	    if (i==0) {
		if ( *s != '(' )
		    inerror(s,"found where a left parenthesis was expected");
		getword(s);
	    }
	    else {
		if ( *s != ',' )
		inerror(s,"found where a comma was expected");
		getword(s);
	    }
	}
	spsb[i] = sparse(s);
    }
    if (f->adicity>1) {
	getword(s);
	if ( *s != ')' )
	    inerror(s,"found where a right parenthesis was expected");
    }
    return locate( f, spsb );
} /* END sparse */



/*
*	Locate returns a pointer to the (sub)formula as eventually parsed.
*	Note that the formula is located in the "parse wff stack" pwf, not
*	in form[].
*/

wff locate( functor mn, wff subform[] )
{
    int i, j;
    wls wf;
    boolean b;

    for ( i = 0; pwf[i].sym; i++ )
	if ( pwf[i].sym == mn ) {
	    b = true;
	    for (j=0; j<mn->adicity; j++)
		if (subform[j] != pwf[i].sb[j])
		    b = false;
	    if (b)
		return pwf+i;
	}
    if ( i >= PWF_MAX ) inerror("Clause is too long","");
    if ( mn->category == Const ) pwf[i].val = *(mn->table);
    pwf[i].sym = mn;
    for (j=0; j<ARGMAX; j++) {
	pwf[i].sb[j] = (j<mn->adicity? subform[j]: 0);
	pwf[i].sbval[j] = (j<mn->adicity? &(subform[j]->val): &zero);
    }
    pwf[i].matrix = mn->table;
    pwf[i].wcc = -1;
    return pwf+i;
} /* END locate */



/*
*	Checking for agreement of types, especially for variables and for
*	typically ambiguous symbols like + and =, is a messy business.
*	Here it is.
*/

void typecheck()
{
    wls wf;

    do {
	tads = false;
	for ( wf = latest_clause.prem;
	     wf; wf = wf->nxt ) type_adds(wf->w);
	for ( wf = latest_clause.conc;
	     wf; wf = wf->nxt ) type_adds(wf->w);
    }
    while ( tads );

    for ( wf = latest_clause.prem;
	 wf; wf = wf->nxt ) type_complete( wf->w );
    for ( wf = latest_clause.conc;
	 wf; wf = wf->nxt ) type_complete( wf->w );

    for ( wf = latest_clause.prem;
	 wf; wf = wf->nxt ) type_consistent( wf->w );
    for ( wf = latest_clause.conc;
	 wf; wf = wf->nxt ) type_consistent( wf->w );
}



/*
*	This checks that every symbol in the current formula has a
*	value type. Some (like `=') have no argument types.
*/

void type_complete( wff w )
{
    int i;

    if ( !w ) return;
    if ( !w->sym->val_type ) 
	inerror("Type missing for",w->sym->fname);
    for (i=0; i<w->sym->adicity; i++)
	type_complete( w->sb[i] );
}



/*
*	Not every symbol is born with a type. This is the function for 
*	adding types to symbols without. There are various cases, for
*	sort-relative symbols like + and E!, and then for variables.
*	Types have to be deduced from the context.
*/

void type_adds( wff w )
{
    int i;

    if ( !w ) return;
    if ( !strcmp(w->sym->fname,"+") ) AT_plus( w );
    if ( !strcmp(w->sym->fname,"E!") ) AT_up( w, "E!_" );
    if ( !strcmp(w->sym->fname,"LEN") ) AT_up( w, "LEN_" );

    for (i=0; i<w->sym->adicity; i++)
	if ( w->sym->arg_type[i] && !w->sb[i]->sym->val_type )
	    add_type(w->sb[i],w->sym->arg_type[i]);

    if (w->sym->adicity == 2 && strchr("=<>",w->sym->fname[0])) {
	if ( w->sb[0]->sym->val_type && !w->sb[1]->sym->val_type)
	    add_type( w->sb[1], w->sb[0]->sym->val_type );
	if ( w->sb[1]->sym->val_type && !w->sb[0]->sym->val_type)
	    add_type( w->sb[0], w->sb[1]->sym->val_type );
    }

    for (i=0; i<w->sym->adicity; i++)
	type_adds(w->sb[i]);
}



/*
*	In formula (A + B) the + is of the same type as the A.
*/

void AT_plus( wff w )
{
    if ( w->sb[0]->sym->val_type )
	add_type( w, w->sb[0]->sym->val_type );
}



/*
*	In formula LEN(A) or E!(A), the type of the predicate is determined
*	by that of its argument.
*/

void AT_up( wff w, char *s )
{
    char string[LEXLENGTH];

    if ( w->sb[0]->sym->val_type ) {
	strcpy( string, s );
	strcpy( string+strlen(s), w->sb[0]->sym->val_type->sname );
	w->sym = fn+lexicon[lex_pos(string,false)].offset;
	w->matrix = w->sym->table;
	tads = true;
    }
}



/*
*	This is for adding types because of the argument types of their 
*	superiors. Variables get their types this way.
*/

void add_type( wff w, kind k )
{
    char string[LEXLENGTH];

    if ( !strcmp(w->sym->fname,"E!") ) return;
    if ( !strcmp(w->sym->fname,"LEN") ) return;
    if ( w->sym->fname[0] == '+' ) {
	*string = '+'; string[1] = '_';
	strcpy(string+2,k->sname);
	w->sym = fn+lexicon[lex_pos(string,false)].offset;
	w->matrix = w->sym->table;
    }
    else if ( w->sym->category == Var )
	w->sym = newvar(w->sym->fname,k);
    else inerror("Missing type for",w->sym->fname);
    tads = true;
}




/*
*	After everything has a type, there may be errors. This function 
*	goes recursively through a formula looking for type mismatches.
*/

void type_consistent( wff w )
{
    int i;

    if (strchr("=<>",w->sym->fname[0])) {
	if ( w->sb[0]->sym->val_type != w->sb[1]->sym->val_type )
	    inerror("Type mismatch: arguments of",w->sym->fname);
    }
    else
	for (i=0; i<w->sym->adicity; i++)
	    if ( w->sym->arg_type[i] != w->sb[i]->sym->val_type )
		inerror("Type mismatch with argument of",w->sym->fname);

    for (i=0; i<w->sym->adicity; i++)
	type_consistent( w->sb[i] );
}



/*
*	And now for something completely different.
*	This function simply delivers a fresh variable if one is required.
*/

functor newvar( char s[], kind k )
{
    wls wf;
    functor f;
    int i,j;

    for ( f = fn; f->fname; f++ )
	if ( !strcmp(s,f->fname) && f->val_type == k )
	    return f;
    for ( i = 0; varlex[i][0]; i++ )
	if ( !strcmp(s,varlex[i]) ) break;
    if ( !varlex[i][0] ) strcpy(varlex[i],s);
    f->fname = varlex[i];
    f->category = Var;
    f->scope = 0;
    f->adicity = 0;
    f->table = 0;
    f->total = true;
    f->injective = false;
    f->cut_it = false;
    for (j=0; j<ARGMAX; j++)
	f->arg_type[j] = 0;
    f->val_type = k;
    f->pmode = null;
    return f;
}



/*
*	When the current clause has been parsed, this is called to set up
*	the various lists (varlist and testlist) and to set the pointers 
*	to subformula values correctly.
*/

void end_parse()
{
    wls wpt, wpv;	/* Ends of testlist and varlist */
    int i;

    wpt = wpv = 0;
    for ( i = 0; pwf[i].sym; i++ )
	if ( pwf[i].sym->category == Var ) {
	    if ( !wpv ) {
		wpv = (wls) malloc(sizeof(WLS));
		latest_clause.varlist = wpv;
	    }
	    else {
		wpv->nxt = (wls) malloc(sizeof(WLS));
		wpv = wpv->nxt;
	    }
	    wpv->w = pwf+i;
	    wpv->nxt = 0;
	}
	else if ( pwf[i].sym->category != Const ) {
	    if ( !wpt ) {
		wpt = (wls) malloc(sizeof(WLS));
		latest_clause.testlist = wpt;
	    }
	    else {
		wpt->nxt = (wls) malloc(sizeof(WLS));
		wpt = wpt->nxt;
	    }
	    wpt->w = pwf+i;
	    wpt->nxt = 0;
	}
    for ( wpt = latest_clause.prem; wpt; wpt = wpt->nxt )
	set_valptrs( wpt->w );
    for ( wpt = latest_clause.conc; wpt; wpt = wpt->nxt )
	set_valptrs( wpt->w );

    if ( latest_clause.conc &&
	!(latest_clause.conc->nxt) &&
	*(latest_clause.conc->w->sym->fname) == '=' &&
	is_open(latest_clause.conc->w->sb[0]->sym) &&
	!occurs_inside(latest_clause.conc->w->sb[0],
		       latest_clause.conc->w->sb[1]) &&
	!occurs_in_list(latest_clause.conc->w->sb[0],
			latest_clause.prem) )
	positive_last();
}


/*
*	This sets the subformula value pointers.
*/

void set_valptrs( wff wf )
{
    int i;

    if ( !wf ) return;
    if ( wf->val >= 0 ) return;
    for (i=0; i<wf->sym->adicity; i++)
	set_valptrs( wf->sb[i] );
    wf->matrix = wf->sym->table;
    for (i=0; i<ARGMAX; i++)
	wf->sbval[i] = (i<wf->sym->adicity? &(wf->sb[i]->val): &zero);
}




/*
*	Certain Horn clauses can be treated specially with one entry in
*	the list of refutation points counted as "positive". The positive
*	point must be moved to the end of the testlist so that any values
*	needed to evaluate it (which need not be in its subformulae) will
*	have been calculated before it is encountered.
*
*	This is rather ugly code, but inefficiency here is of no great
*	significance since the testlist is always fairly short.
*/

void positive_last()
{
    wls lc, cls, wp;	/* latest clause conclusion, its left sub, a ptr */

    for ( lc = latest_clause.testlist; lc->nxt; lc = lc->nxt );
    for (cls = latest_clause.testlist;
	 cls->w != lc->w->sb[0];
	 cls = cls->nxt);
    if ( cls->nxt == lc ) return;
    if ( cls == latest_clause.testlist )
	latest_clause.testlist = cls->nxt;
    else {
	for ( wp = latest_clause.testlist; wp->nxt != cls; wp = wp->nxt );
	wp->nxt = cls->nxt;
    }
    cls->nxt = lc;
    for ( wp = latest_clause.testlist; wp->nxt != lc; wp = wp->nxt );
    wp->nxt = cls;
}




/*
*	If we decide to keep a clause, it has to be moved from the temporary 
*	stack pwf to the permanent one form. The next two functions are parts
*	of that process.
*/

void add_to_theory()
{
    wls wptr;
    int i;

    for ( wptr = latest_clause.prem; wptr; wptr = wptr->nxt )
	wptr->w = corr(wptr->w);
    for ( wptr = latest_clause.conc; wptr; wptr = wptr->nxt )
	wptr->w = corr(wptr->w);
    for ( wptr = latest_clause.varlist; wptr; wptr = wptr->nxt )
	wptr->w = corr(wptr->w);
    for ( wptr = latest_clause.testlist; wptr; wptr = wptr->nxt )
	wptr->w = corr(wptr->w);

    if ( thmax == CL_MAX ) skipout("Too many clauses (whew!)",1);
    theory[thmax].prem = latest_clause.prem;
    theory[thmax].conc = latest_clause.conc;
    theory[thmax].varlist = latest_clause.varlist;
    theory[thmax].testlist = latest_clause.testlist;
    theory[thmax].weight = latest_clause.weight;
    theory[thmax].varcount = latest_clause.varcount;
    theory[thmax].priority = latest_clause.priority;
    theory[thmax].lastclause = 0;
    if ( thmax ) {
	theory[thmax].nextclause = TH;
	TH->lastclause = theory+thmax;
	TH = theory+thmax;
	normalize();
    }
    else theory[thmax].nextclause = 0;
    thmax++;

    latest_clause.prem = 0;
    latest_clause.conc = 0;
    latest_clause.varlist = 0;
    latest_clause.testlist = 0;
}



/*
*	Corr recursively copies the wff from one structure to the other,
*	identifying subformulas that already occur as it goes.
*/

wff corr( wff wf )
{
    wff wfp;
    int i;
    boolean b;

    if ( !wf ) return 0;
    for (i=0; i<wf->sym->adicity; i++)
	wf->sb[i] = corr( wf->sb[i] );
    for ( wfp = form; wfp->sym; wfp++ ) {
	b = (wfp->sym == wf->sym);
	for (i=0; i<wf->sym->adicity; i++)
	    if (wfp->sb[i] != wf->sb[i])
		b = false;
	if (b)
	    return wfp;
    }
    if ( wfp-form >= WF_MAX ) skipout("Too many formulas parsed",1);
    wfp->sym = wf->sym;
    for (i=0; i<ARGMAX; i++) {
	wfp->sb[i] = (i<wf->sym->adicity? wf->sb[i]: 0);
	wfp->sbval[i] = (wfp->sb[i]? &(wfp->sb[i]->val): &zero);
    }
    wfp->matrix = wf->sym->table;
    if ( wf->sym->category == Const ) wfp->val = wf->val;
    wfp->wcc = -1;
    return wfp;
}
