#include "FINDER.h"

        /****************************************************************
        *                                                               *
        *                          FINDER 3.0                           *
        *                                                               *
        *           (C) 1993 Australian National University             *
        *                                                               *
        *                      All rights reserved                      *
        *                                                               *
        * The information in this software is subject to change without *
        * notice and should not be construed as a commitment by the     *
        * Australian National University. The Australian National Uni-  *
        * versity makes no representations about the suitability of     *
        * this software for any purpose. It is supplied "as is" without *
        * express or implied warranty.  If the software is modified in  *
        * a manner creating derivative copyright rights, appropriate    *
        * legends may be placed on the derivative work in addition to   *
        * that set forth above.                                         *
        *                                                               *
        * Permission to use, copy, modify and distribute this software  *
        * and its documentation for any purpose and without fee is      *
        * hereby granted, provided that both the above copyright notice *
        * and this permission notice appear in all copies and sup-      *
        * porting documentation, and that the name of the Australian    *
        * National University not be used in advertising or publicity   *
        * pertaining to distribution of the software without specific,  *
        * written prior permission.                                     *
        *                                                               *
        ****************************************************************/


int vec_pos(functor f, int argument[])
{
    int i;
    int offset=0;

    if ( f->comm ) {
	if ( argument[0] < argument[1] )
	    return (f->fcc + ((argument[1]*(argument[1]+1))/2) + argument[0]);
	return (f->fcc + ((argument[0]*(argument[0]+1))/2) + argument[1]);
    }

    for (i=0; i<f->adicity; i++)
	offset += (argument[i] * f->fats[i]);
    return f->fcc + offset;
}


int matrix_offset(int adic, int argument[])
{
    int i;
    int offset=0;

    for (i=0; i<adic; i++)
	offset += (argument[i] * SZto[i]);
    return offset;
}



functor un_vec_pos( int x, int argument[] )
{
    functor f;
    int i, j;
    int temp;

    for ( f = FN; f; f = f->nextfunct )
	if ( !f->piped && f->fcc <= x ) {
	    if (f->comm) {
		for (argument[0] = 0;
		     argument[0] <= f->arg_type[0]->siz; 
		     argument[0]++)
		    for (argument[1] = 0; 
			 argument[1] <= argument[0]; 
			 argument[1]++)
			if (vec_pos(f,argument) == x)
			    return f;
		skipout("Bad vector position reported",SKIP);
	    }
	    temp = x - f->fcc;
	    for (i=f->adicity-1; i>=0; i--) {
		for (argument[i] = 0;
		     (argument[i]+1)*f->fats[i] < x;
		     argument[i]++);
		x -= argument[i];
	    }
	    if (!x) return f;
	    skipout("Bad vector position reported",SKIP);
	}
}



/*
*	This simply counts the non-constant symbols in the latest
*	clause (merging duplicates) and assigns the result to its 
*	"weight" field. It also counts the constituent variables 
*	for its "varcount".
*/


void get_weight()
{
    wls wf;
    int wt;
    functor f;
    boolean piped_content = false;

    wt = 0;
    for ( wf = latest_clause.testlist; wf; wf = wf->nxt ) {
	f = wf->w->sym;
	if ( f->piped ) piped_content = true;	
	if ( is_open(f) && f->category != Const )
	    wt++;
    }
    if ( !(latest_clause.weight = wt) && !piped_content )
	inerror("Clause contains no function symbol","");

    latest_clause.varcount = 0;
    for ( wf = latest_clause.varlist; wf; wf = wf->nxt )
	latest_clause.varcount++;
}



/*
*	We have just added a new clause to the start of the theory
*	list and set TH to point at it. So move it into the middle
*	of the list until it is in its correct position, then move
*	TH to point back at the start of the list. Note that this 
*	version of normalize assumes that the clauses are added one
*	by one, and that the pre-existing list is already normal.
*/

void normalize()
{
    while ( is_after(TH, TH->nextclause) ) {
	TH->nextclause->lastclause = TH->lastclause;
	if ( TH->lastclause )
	    TH->lastclause->nextclause = TH->nextclause;
	TH->lastclause = TH->nextclause;
	TH->nextclause = TH->lastclause->nextclause;
	TH->lastclause->nextclause = TH;
	if ( TH->nextclause )
	    TH->nextclause->lastclause = TH;
    }
    while ( TH->lastclause ) TH = TH->lastclause;
}




/*
*	Clause c1 is after clause c2 if its weight, variable count 
*	and priority say so. This is where the order of clauses is 
*	determined. Note that the latest clause is not included.
*
*	A clause of pre-testable weight beats one of greater weight
*	irrespective of any other conditions.
*
*	If both are pre-testable the lighter wins, then the one with 
*	fewer variables, then the greater priority, then the 
*	accession order.
*
*	If neither is pre-testable the greater priority wins, then 
*	weight followed by variables (or variables followed by 
*	weight if vars_over_wt is set), then accession order.
*/

boolean is_after( clause c1, clause c2 )
{
    if ( !c1 || !c2 ) return false;

    if ( c2->weight < setting.pretest_limit ) {
	if ( c1->weight > c2->weight ) return true;
	if ( c1->weight < c2->weight ) return false;
	if ( c1->varcount > c2->varcount ) return true;
	if ( c1->varcount < c2->varcount ) return false;
	if ( c1->priority > c2->priority ) return true;
	if ( c1->priority < c2->priority ) return false;
	return ( c1 > c2 );
    }
    if ( c1->weight < setting.pretest_limit ) return false;
    
    if ( c1->priority < c2->priority ) return true;
    if ( c1->priority > c2->priority ) return false;

    if ( setting.vars_over_wt ) {
	if ( c1->varcount > c2->varcount ) return true;
	if ( c1->varcount < c2->varcount ) return false;
	if ( c1->weight > c2->weight ) return true;
	if ( c1->weight < c2->weight ) return false;
    }
    else {
	if ( c1->weight > c2->weight ) return true;
	if ( c1->weight < c2->weight ) return false;
	if ( c1->varcount > c2->varcount ) return true;
	if ( c1->varcount < c2->varcount ) return false;
    }
    return ( c1 > c2 );
}




void set_up_cc()
{
    tr_parameters.vlength = 0;
    set_up_for(FN);
    if ( !(setting.pm.verbosity & pstats) )
	tr_parameters.verb = 0;
    else if ( setting.pm.verbosity & pbstats )
	tr_parameters.verb = 1;
    else if ( setting.pm.verbosity & pfstats )
	tr_parameters.verb = 2;
    tr_parameters.Nosecs = setting.Nosecs;
    tr_parameters.topsoff = setting.topsoff;
    tr_parameters.report_breaks = setting.report_breaks;
    tr_parameters.batches = setting.batches;
    tr_parameters.extra_batches = setting.extra_batches;
    tr_parameters.maxref = setting.maxref;
    tr_parameters.maxbak = setting.max_backtracks;
    ttr_par.maxref = setting.test_maxref;
    set_sum();
    set_E();
    set_cc();
}



void set_SZto()
{
    int i;

    SZto[0] = 1;
    for (i=1; i<ARGMAX; i++) SZto[i] = SZto[i-1]*SZ;
}




void set_up_for( functor f )
{
    int sz;
    int i, j;

    if ( !f ) return;
    set_up_for( f->nextfunct );
    if ( f->piped ) return;
    f->fcc = tr_parameters.vlength;
    if (tr_parameters.vlength >= V_LENGTH) return;
    if (!f->adicity)
	tr_parameters.forbidden[tr_parameters.vlength++] = f->cut_it;
    else
	rec_set_up_for(f,f->adicity-1);
}



void rec_set_up_for(functor f, int x)
{
    int i;

    for (i=0; i<=f->arg_type[x]->siz; i++) {
	if (x)
	    rec_set_up_for(f,x-1);
	else
	    tr_parameters.forbidden[tr_parameters.vlength++] = f->cut_it;
	if (tr_parameters.vlength >= V_LENGTH) return;
    }
}



void set_sum()
{
    int i, j, k;

    for ( k = 0; sort[k].sname; k++ ) {
	for ( i = 0; i < SZ; i++ )
	    for ( j = 0; j < SZ; j++ )
		sort[k].sum[i][j] = i+j < sort[k].siz? i+j: sort[k].siz;
	sort[k].sizmask = 0;
	for ( i = 0; i < sort[k].siz; i++ )
	    sort[k].sizmask |= (1 << i);
    }
}



void set_E()
{
    int i, k;

    for ( k = 0; sort[k].sname; k++ )
	for ( i = 0; i < SZ; i++ )
	    sort[k].Ex[i] = i < sort[k].siz? true: false;
}



void set_fats(functor f)
{
    int i;

    if (f->adicity)
	f->fats[f->adicity-1] = 1;
    for (i=f->adicity-2; i>= 0; i--)
	f->fats[i] = f->fats[i+1] * (f->arg_type[i+1]->siz + 1);
}



void set_cc()
{
    functor f;
    int i, j;

    for ( f = FN; f; f = f->nextfunct ) {
	if ( f->comm && f->arg_type[0]->siz != f->arg_type[1]->siz )
	    skipout("Unequal argument domains for commutative function",1);
	set_fats(f);
    }
#ifdef DEBUG
    for ( f = FN; f; f = f->nextfunct ) if ( !f->piped ) print_set( f );
#endif
}


print_set( functor f )
{
    int i, j;
    int argument[ARGMAX];

    printf("\nIndex for %s",f->fname);
    if ( !f->adicity ) {
	printf(" %d\n", vec_pos(f,argument));
	return;
    }
    puts("");
    if ( f->adicity == 1 ) {
    	for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
	    argument[0] = i;
	    printf("%4d",vec_pos(f,argument));
	}
	argument[0] = i;
	printf("%5d\n",vec_pos(f,argument));
	return;
    }
    for ( i = 0; i < f->arg_type[0]->siz; i++ ) {
	argument[0] = i;
	for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
	    argument[1] = j;
	    printf("%4d",vec_pos(f,argument));
	}
	argument[1] = j;
	printf("%5d",vec_pos(f,argument));
	puts("");
    }
    puts("");
    argument[0] = i;
    for ( j = 0; j < f->arg_type[1]->siz; j++ ) {
	argument[1] = j;
	printf("%4d",vec_pos(f,argument));
    }
    argument[1] = j;
    printf("%5d",vec_pos(f,argument));
    puts("");
}


/*
writeref(int x, int y, int z)
{
    int xx, yy;
    functor ff;
    char cc;

    if (z) cc = '#'; else cc = '=';
    ff = un_vec_pos(x,&xx,&yy);
    switch(ff->adicity) {
      case 0: printf("  %s%c%d", ff->fname, cc, y); break;
      case 1: printf("  %s(%d)%c%d", ff->fname, yy, cc, y); break;
      case 2: printf("  %s(%d,%d)%c%d", ff->fname, xx, yy, cc, y); break;
    }
}
*/
