        /****************************************************************
        *                                                               *
        *                          FINDER 3.0                           *
        *                                                               *
        *           (C) 1993 Australian National University             *
        *                                                               *
        *                      All rights reserved                      *
        *                                                               *
        * The information in this software is subject to change without *
        * notice and should not be construed as a commitment by the     *
        * Australian National University. The Australian National Uni-  *
        * versity makes no representations about the suitability of     *
        * this software for any purpose. It is supplied "as is" without *
        * express or implied warranty.  If the software is modified in  *
        * a manner creating derivative copyright rights, appropriate    *
        * legends may be placed on the derivative work in addition to   *
        * that set forth above.                                         *
        *                                                               *
        * Permission to use, copy, modify and distribute this software  *
        * and its documentation for any purpose and without fee is      *
        * hereby granted, provided that both the above copyright notice *
        * and this permission notice appear in all copies and sup-      *
        * porting documentation, and that the name of the Australian    *
        * National University not be used in advertising or publicity   *
        * pertaining to distribution of the software without specific,  *
        * written prior permission.                                     *
        *                                                               *
        ****************************************************************/


#include "FINDER.h"


/*
*	Setposs is called from transref to trim the search space by removing
*	impossible values from the vector of sets of possibilities. Note that
*	according to TRANSREF this function should only subtract values, never
*	add them.
*
*	First it sets upper bounds according to the cardinalities of the 
*	domains, then it looks to clauses of weight 1 to take out further
*	impossible values. Clauses which are of local weight 1 but not of 
*	absolute weight 1 will get treated a little later in do_twos.
*
*	Clauses of weight 0 (which must include piped stuff) are dealt with
*	at this point, and need never again be checked.
*/

int setposs( unsigned info[], TRS *T )
{
    functor f;
    clause c;

    set_priority( FN, T );

    for ( f = FN; f; f = f->nextfunct ) if ( !f->piped )
	set_possval( info, f, T );

    for ( c = TH; c && (c->weight < 2); c = c->nextclause )
	if ( !c->weight ) {
	    if ( !setposs_test(info,c,0) )
		return 0;
	}
	else if ( !good_one(info,c) ) return 0;
    return 1;
}



/*
*	The set of possible values is trimmed to exclude any which are bigger
*	than the highest within the domain of the value sort. Special allowance
*	has to be made for the dummy value in cases of dummy arguments.
*/

void set_possval( unsigned info[], functor f, TRS *T )
{
    int i, j, sz;
    unsigned mask;
    int argument[ARGMAX];
    boolean undefined;

    sz = f->total? f->val_type->siz: f->val_type->siz+1;
    mask = 0;
    for ( i = 0; i < sz; i++ )
	mask |= (1 << i);
    for (i=0; i<f->adicity; i++)
	argument[i] = 0;
    do {
	undefined = false;
	for (i=0; i<f->adicity; i++)
	    if (argument[i] == f->arg_type[i]->siz)
		undefined = true;
	if (undefined)
	    info[vec_pos(f,argument)] &= (1 << f->val_type->siz);
	else
	    info[vec_pos(f,argument)] &= mask;
    }
    while (another_argument(f,argument));
}


boolean another_argument(functor f, int argument[])
{
    int i;

    for (i=0; i<f->adicity; i++) {
	if (argument[i] < f->arg_type[i]->siz) {
	    argument[i]++;
	    return true;
	}
	else argument[i] = 0;
    }
    return false;
}




int set_priority( functor f, TRS *T )
{
    int i, j;
    int x;
    int argument[ARGMAX];

    if ( !f ) return 0;
    x = set_priority( f->nextfunct, T );
    if ( x > 0 && f->nop ) x--;
    for (i=0; i<f->adicity; i++) argument[i] = 0;
    do
	co_priority(vec_pos(f,argument),x,T);
    while (another_argument(f,argument));
    return x+1;
}





boolean good_one( unsigned info[], clause c )
{		
    wls tf, wf;
    int sz;

    for ( tf = c->testlist; tf && !is_open(tf->w->sym); tf = tf->nxt );
    sz = tf->w->sym->val_type->siz+1;
    for ( tf->w->val = 0; tf->w->val < sz; tf->w->val++ )
	if ( !setposs_test(info, c, tf) ) return false;
    return true;
}




boolean setposs_test( unsigned info[], clause c, wls tf )
{
    wls wf, check;
    int vp;
    boolean failure;
    int offset;
    int i;
    int subvalues[ARGMAX];

    for ( wf = c->varlist; wf; wf = wf->nxt )
	wf->w->val = wf->w->sym->val_type->siz-1;
    do {
	for ( wf = c->testlist; wf; wf = wf->nxt )
	    if ( wf != tf ) {
		offset = 0;
		for (i=0; i<wf->w->sym->adicity; i++)
		    offset += (*(wf->w->sbval[i]) * SZto[i]);
		wf->w->val = *(wf->w->matrix + offset);
	    }
	failure = true;
	for ( check = c->prem; check; check = check->nxt )
	    if ( !check->w->val ) failure = false;
	for ( check = c->conc; check; check = check->nxt )
	    if ( check->w->val ) failure = false;
	if ( failure ) {
	    if ( !tf ) return false;
	    for (i=0; i<tf->w->sym->adicity; i++)
		subvalues[i] = *(tf->w->sbval[i]);
	    vp = vec_pos(tf->w->sym, subvalues);
	    info[vp] &= ~(1 << tf->w->val );
	    if ( !info[vp] ) return false;
	}
    }
    while ( next_assignment(c) );
    return true;
}
