        /****************************************************************
        *                                                               *
        *                          FINDER 3.0                           *
        *                                                               *
        *           (C) 1993 Australian National University             *
        *                                                               *
        *                      All rights reserved                      *
        *                                                               *
        * The information in this software is subject to change without *
        * notice and should not be construed as a commitment by the     *
        * Australian National University. The Australian National Uni-  *
        * versity makes no representations about the suitability of     *
        * this software for any purpose. It is supplied "as is" without *
        * express or implied warranty.  If the software is modified in  *
        * a manner creating derivative copyright rights, appropriate    *
        * legends may be placed on the derivative work in addition to   *
        * that set forth above.                                         *
        *                                                               *
        * Permission to use, copy, modify and distribute this software  *
        * and its documentation for any purpose and without fee is      *
        * hereby granted, provided that both the above copyright notice *
        * and this permission notice appear in all copies and sup-      *
        * porting documentation, and that the name of the Australian    *
        * National University not be used in advertising or publicity   *
        * pertaining to distribution of the software without specific,  *
        * written prior permission.                                     *
        *                                                               *
        ****************************************************************/


#include "FINDER.h"


/*
*	Model is the test routine called from transref. It has to decode the
*	vector info[] of values into testable data structures, carry out the 
*	testing of any clauses of weight greater than 2, use the external 
*	function Ref to record a refutation if one be found, and then send 
*	the model to the printup routine if it passes all tests. The value 
*	returned is true (1) if all tests are passed and false (0) if some 
*	refutation occurs.
*
*	Things are a little different if the _negative flag is set. Then we
*	want to keep structures that come through the whole generate-and-
*	test procedure without ever passing the tests. So instead of calling
*	the print function, model records that a clause was found to fail.
*/

int model( unsigned info[], TRS *T )
{
    clause c;
    int i, j;
    boolean cc_is_good;

    tot++;
    vect_into_model(info);

    if ( negative && !clauses_fail ) return 0;
    if ( !T ) {
	cc_is_good = isgood(&latest_clause);
	if ( negative ) clauses_fail = !cc_is_good;
	return (cc_badguy != cc_is_good);
    }

    if ( latest_clause.prem || latest_clause.conc ) {
	for ( c = TH; c; c = c->nextclause )
	    if (c->weight > setting.pretest_limit
		&& is_after( &latest_clause, c ))
		if ( !isgood(c) ) {
		    record_refutation( c, T );
		    return 0;
		}
	cc_is_good = isgood(&latest_clause);
	if ( !cc_is_good && !cc_badguy ) {
	    record_refutation( &latest_clause, T );
	    return 0;
	}
	for ( c = TH; c; c = c->nextclause )
	    if (c->weight > setting.pretest_limit
		&& is_after( c, &latest_clause ))
		if ( !isgood(c) ) {
		    record_refutation( c, T );
		    return 0;
		}
    }
    else for ( c = TH; c; c = c->nextclause )
	if ( c->weight > setting.pretest_limit )
	    if ( !isgood(c) ) {
		record_refutation( c, T ); 
		return 0;
	    }

    if ( negative ) clauses_fail = 0;
    else if ( cc_badguy && cc_is_good ) return 1;
    else {
	good++;
	if ( T->firstchange < 0 || T->firstchange >= tr_parameters.vlength )
	    first_change = tr_parameters.vlength;
	else first_change = T->inorder[T->firstchange];
	set_lastgood();
	print_model();
	if ( (good && good == setting.max_models) || single_model )
	    job_done( &tr_parameters, "limit on number of solutions" );
    }
    return 1;
}



/*
*	The vector info[] contains the values assigned to functions, but it
*	has to be translated into testable data structures.
*
*	The array thisvector[] is filled with a copy of info[] for purposes 
*	of deciding what should be printed, and of recording the current 
*	model where FINDER is used as a TESTER.
*
*	Note that piped functions do not feature in the vector, though they 
*	are in the list of testable functions.
*/

void vect_into_model( unsigned info[] )
{
    functor f;
    int sz, i, j;
    int argument[ARGMAX];
    int offset;

    for ( i = 0; i < tr_parameters.vlength; i++ )
	thisvector[i] = info[i];
    for ( f = FN; f; f = f->nextfunct ) if ( !f->piped ) {
	for (i=0; i<f->adicity; i++) argument[i] = 0;
	do {
	    offset = 0;
	    for (i=0; i<f->adicity; i++) offset += (argument[i] * SZto[i]);
		*(f->table+offset) = info[vec_pos(f,argument)];
	}
	while (another_argument(f,argument));
    }
}



/*
*	For supertester purposes, the program needs to remember the most
*	recent good model. This is stored in the vector lastgood and in
*	lastsizes and the old_fcc fields of the function symbols.
*/

void set_lastgood()
{
    int i, j;
    functor f;

    for ( i = 0; i < tr_parameters.vlength; i++ )
	lastgood[i] = thisvector[i];
    for ( i = 0; sort[i].sname; i++ ) lastsizes[i] = sort[i].siz;
    for ( f = FN; f; f = f->nextfunct ) f->old_fcc = f->fcc;
    lastvlength = tr_parameters.vlength;
}




/*
*	For each assignment of values to the variables in c do
*		Test c
*	End for
*/

boolean isgood( clause c )
{
    wls wf;
    int vcount = 0;

    if ( c->varcount > VCMAX ) return tisgood(c);

    for ( wf = c->varlist; wf; wf = wf->nxt )
	wf->w->val = (setting.invert_change? 0: wf->w->sym->val_type->siz-1 );
    do
	if ( clause_fails(c) )
	    return false;
    while ( next_assignment(c) );
    return true;
}



/*
*	This is the actual test. The value of each subformula is calculated
*	and assigned to its _val field. Then the premises and conclusions 
*	of the clause are checked to see whether it has been refuted.
*/

boolean clause_fails( clause c )
{
    wls wf;
    boolean failure;
    int i;
    int offset;

    for ( wf = c->testlist; wf; wf = wf->nxt ) {
	offset = 0;
	for (i=0; i<wf->w->sym->adicity; i++)
	    offset += (*(wf->w->sbval[i]) * SZto[i]);
	wf->w->val = *(wf->w->matrix + offset);
    }
    failure = true;
    for ( wf = c->prem; wf; wf = wf->nxt )
	if ( !wf->w->val ) failure = false;
    for ( wf = c->conc; wf; wf = wf->nxt )
	if ( wf->w->val ) failure = false;
    return failure;
}




/*
*	Next_assignment gets the next_assignment of values to the variables.
*/

boolean next_assignment( clause c )
{
    wls wf;

    for ( wf = c->varlist; wf; wf = wf->nxt )
	if ( c->weight <= setting.pretest_limit  || !setting.invert_change ) {
	    if ( !wf->w->val ) wf->w->val = wf->w->sym->val_type->siz-1;
	    else {
		wf->w->val--;
		return true;
	    }
	}
	else {
	    if ( wf->w->val == wf->w->sym->val_type->siz-1 ) wf->w->val = 0;
	    else {
		wf->w->val++;
		return true;
	    }
	}
    return false;
}




/*
*	When a refutation is found, it must be communicated back to transref.
*/

void record_refutation( clause c, TRS *T )
{
    wls wf;
    int argument[ARGMAX];
    int i;

    for ( wf = c->testlist; wf; wf = wf->nxt )
	if ( is_open(wf->w->sym) ) {
	    for (i=0; i<wf->w->sym->adicity; i++)
		argument[i] = *(wf->w->sbval[i]);
	    Ref( vec_pos(wf->w->sym,argument), T );
	}
}
