#include "stree.h"

_Stack  insertBTStack;
_Stack   matchBTStack;
_Stack reverseBTStack;
_Stack   unifyBTStack;

_Stack      refStack;
_Stack demodRefStack;

_Stack rewritestack;
_Stack stacknew;
_Stack fatherStack;

/*
Helping LOCAL MACROS
*/

#define SameVars(v1, v2) (HeapIsRef(v1) And (v1 == v2))
#define SameStruct(v1, v2) (HeapTag(a1) == HeapTag(a2))
#define SameHeads(a1, a2)        \
(HeapTag(a1) == HeapTag(a2) ?    \
  (HeapIsRef(a1) ?               \
    (HeapAddr(a1)==HeapAddr(a2)):\
       TRUE):                    \
  FALSE)


/* 
S P E C I A L    R E T R I E V A L S    F O R   M A I N T E N A N C E
*/


/* 
   Matching allowing only internal variables (of set) 
   to get instantiated, pushing in stacknew the instantiations.
   Recursive.
*/ 
_Bool TermFlatMatchPush( _HeapAddr query, _HeapAddr set)
Proc
  Var
    _Bool match;
    int arity;
  EndVar
  HeapDeref( query, query );
  If query == set Then
    Return(TRUE);
  ElseIf HeapIsRef(set) Then
    Return(FALSE);
  ElseIf HeapIsIntRef(set) Then
    StackPush( stacknew, set );
    HeapSetAddr( set, query );
    Return(TRUE);
  ElseIf HeapTag(set) == HeapTag(query) Then
    arity = HeapTagArity(set);
    match = TRUE;
    While match And arity>0 Do
      match = TermFlatMatchPush( HeapIthArg(query,arity), 
                        HeapAddr(HeapIthArg(set,  arity)) );
      arity--;
    EndWhile
    Return(match);
  Else
    Return(FALSE);
  EndIf
EndProc


_Bool SubstFlatMatchPush(_Subst subst)
Proc
  Var
    _Subst    s;
    _Bool match;
  EndVar
  s = subst;
  match = TRUE;
  While s And match Do
    match = TermFlatMatchPush(HeapAddr(s->lhs),s->rhs);
    s = s->next;
  EndWhile
  Return match;
EndProc


/* 
   Matching allowing only internal variables (of set) to get instantiated.
   Returns the pointer to the found leaf (if found) or NULL (otherwise).
*/

_Pointer *STreeFind(_STree stree, _HeapAddr term)
Proc
  Var
    _STNode node;
    _Bool success;
  EndVar
  If stree == NULL Then
    Return(NULL);
  EndIf
  StackFlush( insertBTStack );
  StackFlush(stacknew);
  StackPush( insertBTStack, StackHeight(stacknew) );
  HeapSetAddr(stree->subst->lhs, term);
  node = stree;
  While TRUE Do
    success = SubstFlatMatchPush(node->subst);
    While (Not success) And (node->brother) Do
      node =  node->brother;
      RestoreRefsUntilHeight( stacknew, (int)StackTop(insertBTStack) );
      success = SubstFlatMatchPush(node->subst);
    EndWhile
    If success Then
      If node->brother  Then 
        StackPush( insertBTStack, node->brother );
      Else
        StackPop( insertBTStack );
      EndIf
      If node->firstSon == NULL Then
        RestoreRefsUntilEmpty(stacknew);
        Return &(node->info);
      Else
        StackPush( insertBTStack, StackHeight(stacknew) );
        node = node->firstSon;
      EndIf
    Else
      StackPop( insertBTStack );
      If StackIsEmpty( insertBTStack ) Then
        RestoreRefsUntilEmpty(stacknew);
        Return(NULL);
      Else
        node = (_STNode) StackTopPop( insertBTStack );
        RestoreRefsUntilHeight( stacknew, (int)StackTop(insertBTStack) );
      EndIf
    EndIf
  EndWhile
EndProc


/*  
I N S E R T I O N   R O U T I N E S  
*/


void STreeInitStacks()
Proc
  StackFlush(  insertBTStack );
  StackFlush(   matchBTStack );
  StackFlush( reverseBTStack );
  StackFlush(   unifyBTStack );

  StackFlush( stacknew );
  StackFlush( fatherStack );
EndProc


void SubstSetSelfRef( _Subst s)
Proc
  While s Do
      HeapSetSelfRef(s->lhs);
      s = s->next;
  EndWhile
EndProc


_Bool SubstSharePart( _Subst subst)
Proc
  Var
    _Subst  s;
    _Bool part;
  EndVar
  s = subst;
  part = FALSE;
  While s And (Not part) Do
      part = SameHeads(HeapAddr(s->lhs),s->rhs);
      s = s->next;
  EndWhile
  Return part;
EndProc

void termmscg( _HeapAddr new, _HeapAddr old, _Subst* restold )
Proc
  Var
    _HeapAddr drefold, drefnew;
    _Subst substaux;
    int arity;
  EndVar
  HeapDeref( new, drefnew );
  HeapDeref( old, drefold );
  If drefnew != drefold Then
    If HeapIsIntRef(drefold) Then
      StackPush( stacknew, drefold );
      HeapSetAddr( drefold, drefnew );
    ElseIf SameHeads(drefnew,drefold) Then
      arity = HeapTagArity(drefold);
      While arity>0 Do
        termmscg( HeapIthArg(drefnew,arity), HeapIthArg(drefold,arity), 
                                                            restold );
        arity--;
      EndWhile
    Else
      HeapSetIntRef(old);
      HeapSetAddr( old, drefnew );
      StackPush( stacknew, old );
      substaux = (_Subst) MemoryAlloc( sizeof(struct _SubstNode));
      substaux->lhs = old; 
      substaux->rhs = drefold;
      substaux->next = *restold;
      *restold = substaux;
    EndIf
  EndIf
EndProc


/* 
   restold  is returned
   restnew  is pushed in stacknew
   *old     is converted into commonpart
*/

_Subst SubstMSCG(_Subst *old)
Proc
  Var
    _Subst s, restold, prev, nextnode;
  EndVar
  s = *old;
  prev = NULL;
  restold = NULL;
  While s Do
    nextnode = s->next;
    If SameHeads(HeapAddr(s->lhs),s->rhs) Then
      termmscg(s->lhs,s->rhs,&restold);
      prev = s;
      HeapSetSelfRef(s->lhs);
    Else
      If prev Then
        prev->next = s->next;
      Else
        *old = (*old)->next;
      EndIf
      s->next = restold;
      restold = s;
      StackPush( stacknew, s->lhs );
    EndIf
    s = nextnode;
  EndWhile
  Return restold;
EndProc


_Subst BuildNewSubst()
Proc
Var
  _Subst newsubst,s;
  _HeapAddr lhs,rhs;
EndVar
  newsubst = NULL;
  While Not StackIsEmpty( stacknew ) Do
    lhs = StackTopPop( stacknew );
    If Not HeapIsSelfRef(lhs) Then
      s = (_Subst) MemoryAlloc( sizeof( struct _SubstNode));
      s->lhs = lhs;
      HeapDeref(lhs,rhs);
      HeapSetSelfRef(lhs);
      s->rhs = HeapCopyNormalTerm(rhs);
      s->next = newsubst;
      newsubst = s;
    EndIf
  EndWhile
  Return(newsubst);
EndProc


/* Inserts a term */

_Pointer *STreeInsert( _STree *tree, _HeapAddr term )
Proc
Var
  _STNode newNode, node, node1;
  _Subst restold;
  _Bool part, match;
  int height;
EndVar

If *tree == NULL Then
  *tree = (_STNode) MemoryAlloc(sizeof(struct _STreeNode));
  (*tree)->subst = (_Subst) MemoryAlloc(sizeof(struct _SubstNode));
  HeapCreateIntRef((*tree)->subst->lhs);
  HeapSetSelfRef((*tree)->subst->lhs);
  (*tree)->subst->rhs = HeapCopyNormalTerm(term);
  (*tree)->subst->next = NULL;
  (*tree)->firstSon = NULL;
  (*tree)->brother = NULL;
  (*tree)->info = NULL;
  Return( &(*tree)->info );
Else
  HeapSetAddr((*tree)->subst->lhs, term);
  StackPush( stacknew, (*tree)->subst->lhs);
  node = (*tree);
  While TRUE Do
    node1 = node;
    height = StackHeight(stacknew);
    match = SubstFlatMatchPush(node->subst);
    While (Not match) And (node->brother) Do
      node =  node->brother;
      RestoreRefsUntilHeight( stacknew, height );
      match = SubstFlatMatchPush(node->subst);
    EndWhile
    If match Then               /* ----- Total matching found */
      SubstSetSelfRef(node->subst);
      If node->firstSon == NULL Then 
#if DEBUG
        printf("Term already existed\n");
#endif
        Return( &node->info );
      Else
        node = node->firstSon;
      EndIf
    Else
      RestoreRefsUntilHeight( stacknew, height );
      node = node1;
      part = SubstSharePart(node->subst);
      While (Not part) And (node->brother) Do
        node =  node->brother;
        part = SubstSharePart(node->subst);
      EndWhile
      If part Then              /* ----- Partial matching found */
        restold = SubstMSCG(&node->subst);
        node1   = (_STNode) MemoryAlloc(sizeof(struct _STreeNode));
        newNode = (_STNode) MemoryAlloc(sizeof(struct _STreeNode));
        node1->firstSon = node->firstSon;
        node1->brother  = newNode;
        node1->info     = node->info;
        node1->subst    = restold;
        node->firstSon  = node1;
        node->info      = NULL;
      Else                       /* ----- not even partial matching found */
        node->brother = (_STNode) MemoryAlloc(sizeof(struct _STreeNode));
        newNode = node->brother;
      EndIf
      newNode->brother  = NULL;
      newNode->firstSon = NULL;
      newNode->info     = NULL;
      newNode->subst    = BuildNewSubst();
      Return( &newNode->info );
    EndIf
  EndWhile
EndIf
EndProc


/*  
D E L E T I O N   R O U T I N E S  
*/

void STreeDeletionMatch(  _STree stree,  _HeapAddr term,
                         _STNode *node,  _STNode *father)
Proc
  Var
    _Bool success;
  EndVar

  StackFlush( insertBTStack );
  StackFlush( fatherStack );
  StackFlush( stacknew );

  HeapSetAddr( stree->subst->lhs, term );
  StackPush( insertBTStack, StackHeight(stacknew) );
  StackPush( fatherStack, NULL );

  *father = NULL;
  *node = stree;
  success = TRUE;

  While TRUE Do

    success = SubstFlatMatchPush((*node)->subst);
    While (Not success) And ((*node)->brother ) Do
      *node =  (*node)->brother;
      RestoreRefsUntilHeight( stacknew,  (int) StackTop(insertBTStack) );
      success = SubstFlatMatchPush((*node)->subst);
    EndWhile

    If success Then
      *father = (_STNode) StackTop(fatherStack);
      If (*node)->brother  Then 
        StackPush( insertBTStack, (*node)->brother );
      Else
        StackPop( insertBTStack );
        StackPop( fatherStack );
      EndIf
      If (*node)->firstSon == NULL Then
        RestoreRefsUntilEmpty(stacknew);
        Return;
      Else
        StackPush( insertBTStack, StackHeight(stacknew) );
        StackPush( fatherStack, *node );
        (*node) = (*node)->firstSon;
      EndIf
    Else
      StackPop( insertBTStack );
      StackPop( fatherStack );
      If StackIsEmpty( insertBTStack ) Then
        RestoreRefsUntilEmpty(stacknew);
        *node = NULL;
        Return;
      Else
        *node = (_STNode) StackTopPop( insertBTStack );
        RestoreRefsUntilHeight( stacknew,  (int) StackTop(insertBTStack) );
      EndIf
    EndIf

  EndWhile
EndProc



void SubstApply( _Subst subst)
Proc
  While (subst ) Do
    HeapSetAddr(subst->lhs, subst->rhs);
    subst = subst->next;
  EndWhile
EndProc


void TermSetIntVar2Ref(_HeapAddr from)
Proc
  Var
    int arity;
    _HeapAddr term;
  EndVar
  If HeapIsIntRef(from) Then
    If Not HeapIsSelfRef(from) Then
      HeapSetRef(from);
    EndIf
  Else
    term = HeapAddr(from);
    If HeapIsSymbol(term) Then
      arity = HeapTagArity(term);
      While arity Do
        TermSetIntVar2Ref(HeapIthArg(term,arity));
        arity--;
      EndWhile
    EndIf
  EndIf
EndProc

void TermSetIntVar2RefTop(_HeapAddr term) 
Proc 
  Var 
    int arity; 
  EndVar 
  If HeapIsSymbol(term) Then 
    arity = HeapTagArity(term); 
    While arity Do 
      TermSetIntVar2Ref(HeapIthArg(term,arity)); 
       arity--; 
    EndWhile 
  EndIf 
EndProc

/* Frees the substNodes and the rhs-terms asociated with them */

void SubstFree( _Subst subst)
Proc
  Var
    _Subst node;
  EndVar
  While subst  Do
    node = subst;
    subst = subst->next;
    HeapFreeNormalTerm(node->rhs);
    MemoryFree((_Ptr)node,sizeof(struct _SubstNode));
  EndWhile
EndProc

/* Fusion of 2 substs, if there are lhs-subterms in sonSubst that are
subterms of fatherSubst then its substitution is composed */


void SubstFussion( _Subst* fatherSubst, _Subst sonSubst)
Proc
  Var
    _Subst node;
  EndVar
  SubstApply(sonSubst);
  node = *fatherSubst;
  While node Do
    TermSetIntVar2RefTop(node->rhs);
    node = node->next;
  EndWhile
  While sonSubst  Do
    node = sonSubst;
    sonSubst = sonSubst->next;
    If HeapIsIntRef(node->lhs) Then
      HeapSetSelfRef(node->lhs);
      node->next = *fatherSubst;
      *fatherSubst = node;
    Else
      MemoryFree((_Ptr)node,sizeof(struct _SubstNode));
    EndIf
  EndWhile
EndProc

void SubstPrint(_Subst l);

_Pointer STreeDelete( _STree* tree, _HeapAddr term)
Proc
  Var
    _STNode node, father, first, aux;
    _Bool firstSon, posibleFussion, found;
    _Pointer info;
  EndVar

  father = NULL, node = NULL;
  STreeDeletionMatch(*tree, term, &node, &father);
  If node == NULL Then
    info = NULL;
    printf("term:");HeapWriteTerm(term);printf(" isn't into index\n");
STreePrint(*tree);
exit(0);
  EndIf
  info = node->info;
  firstSon = FALSE;
/*                                En caso de ser 1st. Son, aqui lo deja fuera
                                  manejando los apuntadores */
  If father == NULL Then
    posibleFussion = FALSE;
    first = *tree;
    If node == first Then
      *tree = node->brother;
      firstSon = TRUE;
    EndIf
  Else
    posibleFussion = TRUE;
    first = father->firstSon;
    If node == first Then
      father->firstSon = node->brother;
      firstSon = TRUE;
    EndIf
  EndIf
/*                                No es 1st. Son, lo busca y cortocircuita
                                  al hermano anterior con el siguiente */
  If Not firstSon Then
    aux = first->brother;
    found = FALSE;
    While (aux) And (Not found) Do
      If aux == node Then
        found = TRUE;
        first->brother = node->brother;
      EndIf
      first = aux;
      aux = aux->brother;
    EndWhile
  EndIf
/*                                Libera el nodo */
  SubstFree(node->subst);
  MemoryFree((_Ptr)node,sizeof(struct _STreeNode));

/*                                Fusiona si es necesario y asigna a father los
                                  valores de firstSon, desapareciendo este */

  If (posibleFussion) And (father->firstSon->brother == NULL) Then
    SubstFussion(&(father->subst),father->firstSon->subst);
    father->info       = father->firstSon->info;
    aux = father->firstSon;
    father->firstSon   = father->firstSon->firstSon;
    MemoryFree((_Ptr)aux, sizeof(struct _STreeNode));
  EndIf
  Return info;
EndProc


/*  
O U T P U T   R O U T I N E S  
*/

void SubstPrint(_Subst l)
Proc
  If l == NULL Do
    printf("(EmptySubts)");
  EndIf
  While l  Do
    printf("[");
    HeapWriteTerm(l->lhs);
    printf(":");
    HeapWriteTerm(l->rhs);
    l = l->next;
    printf("]-");
  EndWhile
  printf("-|");
EndProc

void STreeBrothersPrint(_STNode node, int depth)
Proc
  Var
    int i;
  EndVar
  For i = 0; i <= depth; i++ Do
    printf("  ");
  EndFor
printf("(%p)",node);
  If node->firstSon == NULL Then
    SubstPrint(node->subst);
printf("#%p#\n",node->info);
  Else
    SubstPrint(node->subst);
    printf("\n");
    STreeBrothersPrint(node->firstSon, depth + 1);
  EndIf
  If node->brother  Then
    STreeBrothersPrint(node->brother, depth);
  EndIf
EndProc

void STreePrint(_STree stree)
Proc
printf("\n-------------------- B e g i n   T R E E -------------------------\n");
  If stree == NULL Then
    printf("Empty STree\n");
  Else
    STreeBrothersPrint(stree, 0);
  EndIf
printf("----------------------- E n d   T R E E --------------------------\n");
EndProc

/* show the terms inserted on the tree */

/*
void STreeRetrieveTerms( _STree stree)
Proc
  Var
    _STNode node;
    _Bool finished;
  EndVar
  StackFlush( insertBTStack ); 
  node = stree;
  finished = FALSE;
  While Not finished Do
    SubstApply(node->subst);
    If node->brother  Then 
      StackPush( insertBTStack, node->brother );
    EndIf
    If node->firstSon == NULL Then
      HeapWriteTerm(stree->subst->lhs);printf(".\n");
      If StackIsEmpty( insertBTStack ) Then
        finished = TRUE;
      Else
        node = (_STNode) StackTopPop( insertBTStack );
      EndIf
    Else
      node = node->firstSon;
    EndIf
  EndWhile
EndProc
*/

/*--------------------------STree  R E T R I E V E S-------------------------*/

#define bankQuery   (0)
#define bankDemod   (1)
#define bankSub     (2)
#define bankOld     (3)
#define bankTheorem (4)

/* 
R E V E R S E   M A T C H (Search instances of _Term into _Set) 
*/

_Bool TermReverseMatchIV( _HeapAddr query, _HeapAddr set)
Proc
  Var
    int arity;
    _Bool match;
  EndVar
  HeapDeref(query, query);
  HeapDeref(set,set);
  If query == set Then
    Return(TRUE);
  ElseIf HeapIsIntRef(set) Then 
    StackPush(refStack,set);
    HeapSetAddr(set, query);
    Return(TRUE);
  ElseIf HeapIsIntRef(query) Then 
    StackPush(refStack,query);
    HeapSetAddr(query,set);
    Return(TRUE);
  ElseIf HeapIsRef(query) And VBBankNumber(query) != bankOld Then
    StackPush(refStack,query);
    HeapSetAddr(query, set);
    Return(TRUE);
  ElseIf SameHeads(set,query) Then
    arity = HeapTagArity(set);
    match = TRUE;
    While arity > 0 And match Do
      match = TermReverseMatchIV( HeapIthArg(query, arity), 
                                  HeapIthArg(set, arity) );
      arity--;
    EndWhile
    Return match;
  Else
    Return(FALSE);
  EndIf
EndProc


_Bool TermReverseMatch1IV( _HeapAddr query, _HeapAddr set)
Proc
  Var
    int arity;
    _Bool match;
  EndVar
  HeapDeref(query,query);
  HeapDeref(set,set);
  If query == set Then
    Return(TRUE);
  ElseIf HeapIsIntRef(query) Then 
    StackPush(refStack,query);
    HeapSetAddr(query,set);
    Return(TRUE);
  ElseIf HeapIsRef(query) And VBBankNumber(query) !=  bankOld Then
    StackPush(refStack, query);
    HeapSetAddr(query, set);
    Return(TRUE);
  ElseIf SameHeads(set,query) Then
    arity = HeapTagArity(set);
    match = TRUE;
    While arity > 0 And match Do
      match = TermReverseMatchIV( HeapIthArg(query, arity), 
                                  HeapIthArg(set, arity) );
      arity--;
    EndWhile
    Return match;
  Else
    Return(FALSE);
  EndIf
EndProc


_Bool SubstReverseMatch(_Subst subst)
Proc
  Var
    _Subst    s;
    _Bool match;
  EndVar
  s = subst;
  match = TRUE;
  While s And match Do
    match = TermReverseMatch1IV( s->lhs, s->rhs );
    s = s->next;
  EndWhile
  Return match;
EndProc


_Pointer STreeFirstReverseMatch(_STree stree, _HeapAddr term)
Proc
  Var
    _STNode node;
    _Bool finished, success;
  EndVar

  If stree == NULL Then
    Return(NULL);
  EndIf

  StackFlush(insertBTStack);
  StackFlush(refStack);
  StackPush( insertBTStack,  StackHeight(refStack) );
  HeapSetAddr(stree->subst->lhs, term);

  node = stree;
  success = TRUE;
  finished = FALSE;

  While Not finished Do
    success = SubstReverseMatch(node->subst);
    While (Not success) And (node->brother ) Do
      node =  node->brother;
      RestoreRefsUntilHeight( refStack, (int) StackTop(insertBTStack) );
      success = SubstReverseMatch(node->subst);
    EndWhile
    If success Then
      If node->brother  Then 
        StackPush( insertBTStack, node->brother );
      Else
        StackPop(insertBTStack);
      EndIf
      If node->firstSon == NULL Then
        finished = TRUE;
      Else
        StackPush( insertBTStack, StackHeight(refStack) );
        node = node->firstSon;
      EndIf
    Else
      StackPop( insertBTStack );
      If StackIsEmpty( insertBTStack ) Then
        RestoreRefsUntilEmpty( refStack );
        Return(NULL);
      Else
        node = (_STNode) StackTopPop( insertBTStack );
        RestoreRefsUntilHeight( refStack, (int) StackTop(insertBTStack) );
      EndIf
    EndIf
  EndWhile
  Return(node->info);
EndProc


_Pointer STreeNextReverseMatch()
Proc
  Var
    _STNode node;
    _Bool finished, success;
  EndVar

  If StackIsEmpty( insertBTStack ) Then
    RestoreRefsUntilEmpty(refStack);
    Return(NULL);
  Else
    node = (_STNode)StackTopPop(insertBTStack);
  EndIf

  RestoreRefsUntilHeight( refStack, (int) StackTop(insertBTStack) );
  success = TRUE;
  finished = FALSE;

  While Not finished Do
    success = SubstReverseMatch(node->subst);
    While (Not success) And (node->brother ) Do
      node =  node->brother;
      RestoreRefsUntilHeight( refStack, (int) StackTop(insertBTStack) );
      success = SubstReverseMatch(node->subst);
    EndWhile
    If success Then
      If node->brother  Then 
        StackPush( insertBTStack, node->brother );
      Else
        StackPop( insertBTStack );
      EndIf
      If node->firstSon == NULL Then
        finished = TRUE;
      Else
        StackPush( insertBTStack, StackHeight(refStack) );
        node = node->firstSon;
      EndIf
    Else
      StackPop( insertBTStack );
      If StackIsEmpty( insertBTStack ) Then
        RestoreRefsUntilEmpty( refStack );
        Return(NULL);
      Else
        node = (_STNode) StackTopPop(insertBTStack);
        RestoreRefsUntilHeight( refStack, (int) StackTop(insertBTStack) );
      EndIf
    EndIf
  EndWhile
  Return(node->info);
EndProc




/* 
M A T C H (Search more general expresions of _Term into _Set) 
*/

_Bool TermMatch(_HeapAddr query, _HeapAddr set)
Proc
  Var
    int arity;
    _Bool match;
  EndVar
  HeapDeref(query, query);
  HeapDeref(set,set);
  If query == set Then
    Return(TRUE);
  ElseIf HeapIsSelfRef(set) And VBBankNumber(set) ==  bankDemod Then
    StackPush(demodRefStack,set);
    HeapSetAddr(set, query);
    Return(TRUE);
  ElseIf SameHeads(set,query) Then
    arity = HeapTagArity(set);
    match = TRUE;
    While arity > 0 And match Do
      match = TermMatch( HeapIthArg(query, arity), HeapIthArg(set, arity) );
      arity--;
    EndWhile
    Return match;
  Else
    Return(FALSE);
  EndIf
EndProc


_Bool TermMatchIV(_HeapAddr query, _HeapAddr set)
Proc
  Var
    int arity;
    _Bool match;
  EndVar
  HeapDeref(query,query);
  HeapDeref(set,set);
  If query == set Then
    Return(TRUE);
  ElseIf HeapIsSelfRef(set) Then
    If (VBBankNumber(set) ==  bankDemod) Or HeapIsIntRef(set) Then
      StackPush(demodRefStack,set);
      HeapSetAddr(set, query);
      Return(TRUE);
    Else
      Return(FALSE);
    EndIf
  ElseIf HeapTag(set) == HeapTag(query) Then
    arity = HeapTagArity(set);
    match = TRUE;
    While arity > 0 And match Do
      match = TermMatchIV( HeapIthArg(query, arity), HeapIthArg(set, arity) );
      arity--;
    EndWhile
    Return match;
  Else
    Return(FALSE);
  EndIf
EndProc

_Bool SubstMatch(_Subst subst)
Proc
  Var
    _Bool match;
  EndVar
  match = TRUE;
  While subst  And match Do
    match = TermMatchIV( subst->lhs, subst->rhs);
    subst = subst->next;
  EndWhile
  Return match;
EndProc

_Pointer STreeFirstMatch( _STree stree, _HeapAddr term)
Proc
  Var
    _STNode node;
    _Bool finished, success;
  EndVar

  If stree == NULL Then
    Return(NULL);
  EndIf

  StackFlush( matchBTStack );
  StackFlush( demodRefStack );
  StackPush( matchBTStack, 0 );
  HeapSetAddr(stree->subst->lhs, term);

  node = stree;
  success = TRUE;
  finished = FALSE;

  While Not finished Do

    success = SubstMatch(node->subst);
    While (Not success) And (node->brother ) Do
      node =  node->brother;
      RestoreRefsUntilHeight(demodRefStack,(int)StackTop(matchBTStack) );
      success = SubstMatch(node->subst);
    EndWhile

    If success Then
      If node->brother  Then 
        StackPush( matchBTStack, node->brother );
      Else
        StackPop( matchBTStack );
      EndIf
      If node->firstSon == NULL Then
        finished = TRUE;
      Else
        StackPush( matchBTStack, (int)StackHeight(demodRefStack) );
        node = node->firstSon;
      EndIf
    Else
      StackPop( matchBTStack );
      If StackIsEmpty( matchBTStack ) Then
        RestoreRefsUntilEmpty(demodRefStack);
        Return(NULL);
      Else
        node = (_STNode) StackTopPop( matchBTStack );
        RestoreRefsUntilHeight(demodRefStack,(int)StackTop(matchBTStack));
      EndIf
    EndIf

  EndWhile
  Return(node->info);
EndProc


_Pointer STreeNextMatch()
Proc
  Var
    _STNode node;
    _Bool finished, success;
  EndVar

  If StackIsEmpty( matchBTStack ) Then
    RestoreRefsUntilEmpty(demodRefStack);
    Return(NULL);
  EndIf

  node = (_STNode)StackTopPop(matchBTStack);
  RestoreRefsUntilHeight(demodRefStack,(int)StackTop(matchBTStack) );
  success = TRUE;
  finished = FALSE;

  While Not finished Do

    success = SubstMatch(node->subst);
    While (Not success) And (node->brother ) Do
      node =  node->brother;
      RestoreRefsUntilHeight(demodRefStack,(int)StackTop(matchBTStack) );
      success = SubstMatch(node->subst);
    EndWhile

    If success Then
      If node->brother  Then 
        StackPush( matchBTStack, node->brother );
      Else
        StackPop( matchBTStack );
      EndIf
      If node->firstSon == NULL Then
        finished = TRUE;
      Else
        StackPush( matchBTStack, (int)StackHeight(demodRefStack) );
        node = node->firstSon;
      EndIf
    Else
      StackPop( matchBTStack );
      If StackIsEmpty( matchBTStack ) Then
        RestoreRefsUntilEmpty(demodRefStack);
        Return(NULL);
      Else
        node = (_STNode) StackTopPop( matchBTStack );
        RestoreRefsUntilHeight(demodRefStack,(int)StackTop(matchBTStack));
      EndIf
    EndIf

  EndWhile
  Return(node->info);
EndProc


/*
U N I F I C A T I O N   (Search unifiers of _Term into _Set) 
*/

/* Note: The argument terms must be unferentied */
_Bool TermContainsVar(_HeapAddr term, _HeapAddr var)
Proc
  Var
    _HeapAddr ti, PDLterm[100];
    int arity, sp;
  EndVar
  arity = HeapTagArity(term);
  sp=0;
  While arity > 0 Do
    HeapDeref(HeapIthArg(term,arity--), ti);
    If ti == var Then
      Return(TRUE);
    ElseIf HeapIsSymbol(ti) Then
      sp++;
      PDLterm[sp] = ti;
    EndIf
  EndWhile
  While sp Do
    term = PDLterm[sp--];
    arity = HeapTagArity(term);
    While arity > 0 Do
      HeapDeref(HeapIthArg(term,arity--), ti);
      If ti == var Then
        Return(TRUE);
      ElseIf HeapIsSymbol(ti) Then
        sp++;
        PDLterm[sp] = ti;
      EndIf
    EndWhile
  EndWhile
  Return(FALSE);
EndProc


/*
   Unification allowing all internal and external variables (of both set 
   and query) to get instantiated, pushing in refstack the instantiations.
   Recursive.
*/ 
_Bool TermUnify( _HeapAddr query, _HeapAddr set)
Proc
  Var
    _Bool unify;
    int arity;
  EndVar
  HeapDeref(query, query);
  HeapDeref(set, set);
  If query == set Then
    Return(TRUE);
  ElseIf HeapIsRef(set) Then
    If TermContainsVar(query,set) Then
      Return(FALSE);
    Else
      StackPush(refStack, set);
      HeapSetAddr(set, query);
      Return(TRUE);
    EndIf
  ElseIf HeapIsRef(query) Then
    If TermContainsVar(set,query) Then
      Return(FALSE);
    Else
      StackPush(refStack, query);
      HeapSetAddr(query, set);
      Return(TRUE);
    EndIf
  ElseIf HeapTag(set) == HeapTag(query) Then
    arity = HeapTagArity(set);
    unify = TRUE;
    While unify And arity>0 Do
      unify = TermUnify( HeapIthArg(query,arity), HeapIthArg(set,arity) );
      arity--;
    EndWhile
    Return(unify);
  Else
    Return(FALSE);
  EndIf
EndProc


_Bool TermUnifyIV( _HeapAddr query, _HeapAddr set)
Proc
  Var
    _Bool unify;
    int arity;
  EndVar
  HeapDeref(query, query);
  HeapDeref(set, set);
  If query == set Then
    Return(TRUE);
  ElseIf HeapIsIntRef(set) Then
    If TermContainsVar(query,set) Then
      Return(FALSE);
    Else
      StackPush(refStack, set);
      HeapSetAddr(set, query);
      Return(TRUE);
    EndIf
  ElseIf HeapIsIntRef(query) Then
    If TermContainsVar(set,query) Then
      Return(FALSE);
    Else
      StackPush(refStack, query);
      HeapSetAddr(query, set);
      Return(TRUE);
    EndIf
  ElseIf HeapIsRef(set) Then
    If TermContainsVar(query,set) Then
      Return(FALSE);
    Else
      StackPush(refStack, set);
      HeapSetAddr(set, query);
      Return(TRUE);
    EndIf
  ElseIf HeapIsRef(query) Then
    If TermContainsVar(set,query) Then
      Return(FALSE);
    Else
      StackPush(refStack, query);
      HeapSetAddr(query, set);
      Return(TRUE);
    EndIf
  ElseIf HeapTag(set) == HeapTag(query) Then
    arity = HeapTagArity(set);
    unify = TRUE;
    While unify And arity>0 Do
      unify = TermUnifyIV( HeapIthArg(query,arity), HeapIthArg(set,arity) );
      arity--;
    EndWhile
    Return(unify);
  Else
    Return(FALSE);
  EndIf
EndProc



_Bool SubstUnify( _Subst subst)
Proc
  Var
    _Bool unify;
  EndVar
  unify = TRUE;
  While subst  And unify Do
    unify = TermUnifyIV( subst->lhs, subst->rhs);
    subst = subst->next;
  EndWhile
  Return unify;
EndProc



_Pointer STreeFirstUnify( _STree stree, _HeapAddr term)
Proc
  Var
    _STNode node;
    _Bool finished, success;
  EndVar

  If stree == NULL Then
    Return(NULL);
  EndIf

  StackFlush(unifyBTStack);
  StackFlush(refStack);
  HeapSetAddr(stree->subst->lhs, term);
  StackPush( unifyBTStack, StackHeight(refStack) );

  node = stree;
  success = TRUE;
  finished = FALSE;

  While Not finished Do

    success = SubstUnify(node->subst);
    While (Not success) And (node->brother ) Do
      node =  node->brother;
      RestoreRefsUntilHeight( refStack,  (int) StackTop(unifyBTStack) );
      success = SubstUnify(node->subst);
    EndWhile

    If success Then
      If node->brother  Then 
        StackPush( unifyBTStack, node->brother );
      Else
        StackPop( unifyBTStack );
      EndIf
      If node->firstSon == NULL Then
        finished = TRUE;
      Else
        StackPush( unifyBTStack, StackHeight(refStack) );
        node = node->firstSon;
      EndIf
    Else
      StackPop( unifyBTStack );
      If StackIsEmpty( unifyBTStack ) Then
        RestoreRefsUntilEmpty( refStack );
        Return(NULL);
      Else
        node = (_STNode) StackTopPop( unifyBTStack );
        RestoreRefsUntilHeight( refStack,  (int) StackTop(unifyBTStack) );
      EndIf
    EndIf

  EndWhile
  Return(node->info);
EndProc



_Pointer STreeNextUnify()
Proc
  Var
    _STNode node;
    _Bool finished, success;
  EndVar

  If StackIsEmpty( unifyBTStack ) Then
    RestoreRefsUntilEmpty( refStack );
    Return(NULL);
  EndIf

  node = (_STNode)StackTopPop( unifyBTStack );
  RestoreRefsUntilHeight( refStack,  (int) StackTop(unifyBTStack) );
  success = TRUE;
  finished = FALSE;

  While Not finished Do

    success = SubstUnify(node->subst);
    While (Not success) And (node->brother ) Do
      node =  node->brother;
      RestoreRefsUntilHeight( refStack,  (int) StackTop(unifyBTStack) );
      success = SubstUnify(node->subst);
    EndWhile

    If success Then
      If node->brother Then 
        StackPush( unifyBTStack, node->brother );
      Else
        StackPop( unifyBTStack );
      EndIf
      If node->firstSon == NULL Then
        finished = TRUE;
      Else
        StackPush( unifyBTStack, StackHeight(refStack) );
        node = node->firstSon;
      EndIf
    Else
      StackPop( unifyBTStack );
      If StackIsEmpty( unifyBTStack ) Then
        RestoreRefsUntilEmpty( refStack );
        Return(NULL);
      Else
        node = (_STNode) StackTopPop( unifyBTStack );
        RestoreRefsUntilHeight( refStack, (int) StackTop(unifyBTStack) );
      EndIf
    EndIf

  EndWhile
  Return(node->info);
EndProc




