#ifndef STREE_H
#define STREE_H

#include "heap.h"
#include "stack.h"

typedef struct _SubstNode *_Subst;
struct _SubstNode{
   _HeapAddr lhs;
   _HeapAddr rhs;
   _Subst next;
   };
typedef struct _SubstNode _SubstNode;

typedef enum {
              NoOperation,
              Matching,
              Instance,
              Unification 
   } _OperationType;

typedef struct _STreeNode *_STNode;
struct _STreeNode{
   _STNode  firstSon;
   _STNode  brother;
   _Subst   subst;
   _Pointer info;
   };
typedef _STNode _STree;


/* 
Maintenance Functions Declaration
*/
extern void        STreeInitStacks( void );
extern _Pointer    *STreeInsert( _STree* tree, _HeapAddr term);
extern _Pointer    STreeDelete( _STree* tree, _HeapAddr term);
extern _Pointer    *STreeFind( _STree stree, _HeapAddr term);

/* 
Retrieve Functions Declaration
*/
extern _Pointer    STreeFirstMatch(        _STree tree, _HeapAddr term);
extern _Pointer    STreeFirstReverseMatch( _STree tree, _HeapAddr term);
extern _Pointer    STreeFirstUnify(        _STree tree, _HeapAddr term);
extern _Pointer    STreeNextMatch( void );
extern _Pointer    STreeNextReverseMatch( void );
extern _Pointer    STreeNextUnify( void );

extern _Bool TermMatch(        _HeapAddr query, _HeapAddr set );
extern _Bool TermReverseMatch( _HeapAddr query, _HeapAddr set );
extern _Bool TermUnify(        _HeapAddr query, _HeapAddr set );

/* 
IO Functions Declaration
*/
extern _HeapAddr   TermCopyForInsert(_HeapAddr term);
extern void        STreeRetrieveTerms( _STree stree);
extern void        STreePrint( _STree tree);/*
extern void        STreeCheck( _STree stree, _Ident message);
*/

/*
Helping MACROS
*/
#define RestoreRefsUntilHeight(stack,height) \
Proc \
  While StackHeight(stack) > height Do \
    HeapSetSelfRef( StackTop( stack ) ); \
    StackPop( stack ); \
  EndWhile \
EndProc

#define RestoreRefsUntilEmpty(stack) \
Proc \
  While Not StackIsEmpty(stack) Do \
    HeapSetSelfRef( StackTop( stack ) ); \
    StackPop( stack ); \
  EndWhile \
EndProc

#endif
