/* -------------------------------------------------------------*/
/*      clause.h          data structures for TGTP
/*
/*      Copyright Newborn Software 1996
/*
/* -------------------------------------------------------------*/
 
/*--------------------------------------------------------------*/ 
/*-----------    data structure for a literal header  ----------*/    
/*--------------------------------------------------------------*/ 
 
typedef struct lt {
	struct lt   *l_next;          /* ^ next literal in chain      */
	long        l_X;                     /* literal's hashcode    */
	long        l_Y;   
	long        l_id;}             /* [NEGMASK+] PID + ARGS       */
	literal;
 
/*--------------------------------------------------------------*/ 
/*---------------------    l_id field      ---------------------*/
/*--------------------------------------------------------------*/ 
 
#define HHITMASK    0x40000000L           /* hash resolvent     */
 
/*--------------------------------------------------------------*/ 
/*-------------   data structure for a clause header   ---------*/    
/*--------------------------------------------------------------*/ 
 
typedef struct cl {
	long        c_id;                      /* clause id           */
	struct cl   *c_end;             /* pointer past end of clause */
	long        c_bits;                            /* see below   */
	long        c_X;                        /* clause hash code   */
	long        c_Y;                        /* more hash code     */
	struct cl   *c_lpar,*c_rpar;               /* our parents     */
	long        c_indices;    /* index of resolvent of 1st clause */
	literal     *c_lit;               /* pointer to literal chain */
	int         c_nlit;                   /* # literals in clause */
	long        c_nv;}          /* # distinct variables in clause */
	clause;
 
/*--------------------------------------------------------------*/ 
/*---------------------    c_bits field    ---------------------*/
/*--------------------------------------------------------------*/ 
 
#define UNIT        0x00000001L          /* 1 = unit clause     */
#define MERGE       0x00000004L           /* 1 = merge clause   */
#define BASE        0x00000008L             /* 1 = base clause  */
#define MA          0x00000020L      /* 1 = merge clauses above */
#define VID         0x00000040L /* var names start at 0 or 200  */
#define ACTIVE      0x00000080L  /* 1 = cls active after elim   */
#define NC          0x00000100L   /* denotes negated conclusion */
#define LIST        0x00000400L  /*  list in output of proof    */
#define EQCLS       0x00000800L  /* denotes eq. cls, fxn or lit */
 
#define LEFT_INDEX_MASK   0x00000fc0L   /* mask for left index  */
#define RIGHT_INDEX_MASK  0x0000003fL   /* mask for right index */
#define LEFT_INDEX_SHIFT  6L
 
/*--------------------------------------------------------------*/ 
/*------------     actual predicate code definitions  ----------*/
/*--------------------------------------------------------------*/ 
 
#define IDMASK      0x3FFFFFE0L        /* pid,vid,fid masks     */
#define NIDMASK     0xc000001FL
#define ARGMASK     0x0000001FL           /* 0..31 arguments    */
#define FCNMASK     0x80000000L          /* 1=item is function  */
#define NEGMASK     0x80000000L /* in preds, indicates negation */
#define IAMASK      0x3FFFFFFFL           /* i.e. ~NEGMASK      */
#define ARGSHIFT    0x00000005L 
#define CONMASK     0x8000001FL            /* constant mask     */ 
 
/*--------------------------------------------------------------*/ 
/* ----------- pointers to clause stack ----------------------- */
/*--------------------------------------------------------------*/ 
 
typedef struct {
	long        d_ec;          /* top clause at this level        */
	clause      *d_rc;         /* root clause (first clause ptr)  */
	long        d_tc;        /* clause under search at this level */
	long        d_bc;}        /* bottom clause at this level      */
	dnode;
 
/*--------------------------------------------------------------*/ 
/* ------------ clause hash table data structures --------------*/
/*--------------------------------------------------------------*/ 
 
typedef struct hn {
	long        h_X;                 /* checksum of clause found  */
	long        h_Y;                 /* checksum of clause found  */
	int         h_in;                 /* iteration number         */
	int         h_ply;}               /* must be at same ply      */
	hnode;
 
    
/*--------------------------------------------------------------*/ 
/* -------------- random number table entry---------------------*/
/*--------------------------------------------------------------*/ 
 
typedef struct rn {
	long        r_hash1;
	long        r_hash2;}
	rnode;
 
 
/*--------------------------------------------------------------*/ 
/*---- data structures used by affc.c---------------------------*/
/*--------------------------------------------------------------*/ 
 
typedef struct cn {
	struct      cn *c_next,*c_prev;
	long        c_pred;} 
	cnode;
 
typedef struct pf {
	struct      pf *p_llink,*p_rlink;
	long        p_bits;
	char        *p_name;                    /* tree linkage stuff  */
	long        p_id;                          /* unique ID        */
	long        p_nargs;}                  /* # arguments of obj   */
	pfnode;
 
#define P_RTHREAD   0x00000001L          /* threaded tree stuff  */
#define P_LTHREAD   0x00000002L
#define P_PRED      0x00000004L    /* pred. bit (1=predicate)    */
#define P_FUNC      0x00000000L           /* 0 = function        */
#define P_TYPE      0x00000004L           /* type mask           */
 
