/*----------------------------------------------------------------*/
/*      search.c
/*
/*      Copyright Newborn Software 1996
/*
/*      functions:
/*      install_clause()        ......................#1
/*      int calc_inferences()   ......................#2
/*      search_tree()           ......................#3
/*      advance_to_successor()  ......................#4
/*      simple_subsume2()       ......................#5
/*----------------------------------------------------------------*/
 
#include <stdio.h>
 
#include "const.h"
#include "clause.h"

/*----------------------------------------------------------------*/
/* externally defined functions                                   */
/*----------------------------------------------------------------*/
 
extern fatal();                     /* can be found in tgtp.c     */
extern long time_now();
 
extern test_equality();             /* can be found in ....gen.c  */
extern repair_literal();
extern repair_clause();
 
extern binary_resolvents();      /* infer.c:   inference rules    */
extern binary_factors();
 
extern print_clause();           /* tdump.c: print clause fxn     */
extern print_continuation();
extern halted_search_options();

/*----------------------------------------------------------------*/
/* the next two functions are not used in this version.           */
/*----------------------------------------------------------------*/

extern interrupt_options();
extern kbd_hit();
 
extern is_clause_in_ht();                          /*   hash.c    */
extern long add_clause_to_ht();
extern is_unit_in_ht();
extern is_comp_in_ht();
extern long calc_alt_hashcodes();
extern normalize_literal();
extern calc_lit_hc();
extern calc_clause_hashcode();
 
extern on_hash_code_list();                          /* pass2.c    */
extern save_line();
extern alt_on_hash_code_list();

/*-----------------------------------------------------------------*/
/* externally defined constants                                    */
/*-----------------------------------------------------------------*/
 
extern FILE *outfile;
 
extern long constants_in_clause;
 
extern long base, ply, nextc, iter_depth, nc;         /* tgtp.c     */
extern long org_base;
extern int flag_kbd_hit;

extern long search_start_time;
extern long search_time;
extern long equal_pred;
extern int phase;
extern int number_constants;
extern int number_const_hash;
 
extern int o_extend, o_hash, in_elim;      
extern int o_print, o_step, o_cr;
extern int o_merge, o_subsume;
extern int x_stop;      
extern int o_maxtime;
extern int o_debug;
extern long s_nod;                       /* # nodes in tree        */
extern long s_res, s_fac;
extern long s_subsume;                   /* count of s-subsumed cl */
 
extern long hash_code_list_active;               /* pass2.c        */
extern long proof_length;
extern long lit_in_hash;
 
/*-------------------------------------------------------------------*/
/* clist is a list of ptrs to clauses that lie in clause_space.      */
/* During the search each group of pointers belongs to a particular  */
/* depth. Since the clauses at each depth are ordered by their score */
/* the list is not ordered (i.e. the last clist item does not        */
/* necessarily point to the last clause). However, a group of cls    */
/* at a particular depth are consecutive. Pointers to the clauses    */
/* at each level are in the array depth[].                           */
/*-------------------------------------------------------------------*/
 
extern long *clause_space[];            /* where clauses are stored  */
extern clause *clist[];                 /* pointers to clauses       */
extern dnode depth[];            /* pointers to pointers of clauses  */

/*-------------------------------------------------------------------*/
/* internally defined functions                                      */
/*-------------------------------------------------------------------*/
 
int install_clause();
int calc_inferences();
int search_tree();
int advance_to_successor();
int simple_subsume2();

/*-------------------------------------------------------------------*/
/* internally defined constant                                       */
/*-------------------------------------------------------------------*/
 
long magic_ptr;                 /* ptr to unit clause at end of line */
 
/*-----------------------------------------------------------------*/
/*      install_clause(c)             ......................#1
/*
/*      install_clause() is called after the clause has been
/*      generated by generate_clause().  It is pointed to by c.
/*      Install_clause() has the job of deciding whether to
/*      actually list the clause with the other clauses at that
/*      level in the tree.
/*
/*      install_clause adds a clause to the group of clauses
/*      at ply. We assume that this has been properly
/*      initialized. We insert the new clause so that all clauses
/*      at this ply are sorted by the number of literals.
/*      The clause is located at clist[nextc].
/*
/*          clist:      .
/*
/*              [       ] <- d_bc
/*              [       ]
/*              [       ] <- d_tc
/*              [   c   ] <- nextc
/*
/*      called by:  binary_resolvents()         infer.c
/*                  binary_factors()            infer.c
/*                  bres_replace()              elim2.c
/*                  bres_link()                 elim2.c
/*
/*      calls:      simple_subsume2()           search.c
/*                  test_equality()             gen.c
/*                  repair_literal()            gen.c
/*                  add_clause_to_ht()          hash.c
/*                  is_clause_in_ht()           hash.c
/*                  is_comp_in_ht()             hash.c
/*                  calc_clause_hashcode()      hash.c
/*                  calc_alt_hashcodes()        hash.c
/*                  on_hash_code_list()         pass2.c
/*                  alt_on_hash_code_list()     pass2.c
/*                  save_line()                 pass2.c
/*                  print_continuation()        tdump.c
/*
/*      returns     1 if the clause was indeed installed,
/*                  0 if the clause was not installed
/*------------------------------------------------------------------*/
 
int install_clause(c)
clause *c;
{
register long  i, k;
register int comp_lit;
register literal *l;
register dnode *d;
long l_index;
long *lptr, aptr;
int found, deep_in_tree;
 
d = &depth[ply];                      /* insert at current ply     */
 
comp_lit = 0;   

/*------------------------------------------------------*/
/* if the null clause has been found at this node,      */
/* then lit_in_hash will be 2.  in that case, no        */
/* more clauses should be generated at this node.       */
/*------------------------------------------------------*/

if(lit_in_hash >= 2L) { 
   return(0);}

/*--------------------------------------------------------------*/
/* eliminate clauses in extend search if they have no constants */
/* and if there is one or more constants in the theorem.        */
/* this is a risky heuristic that seems to work.                */
/*--------------------------------------------------------------*/

if(phase > 0 && constants_in_clause == 0L && ply > iter_depth + 1L) {
   if(c->c_nlit != 0 && number_constants != 0) {
      return(0);}}

/*--------------------------------------------------------------*/
/* if the merge option is on and the transposition table option */
/* is on and the clause is not a merge ancestor and the cluase  */
/* is in the extended-search part of the tree, and if the clause*/
/* is found in the clause_hash_table, don't search it           */
/*--------------------------------------------------------------*/

if(ply >= iter_depth && o_merge != 0 && o_hash == 1) {
   deep_in_tree = 1;}
else {
   deep_in_tree = 0;}
 
if(deep_in_tree) {
   if(is_clause_in_ht(c) != -1)  { 
      return(0);}}

/*--------------------------------------------------------------*/
/* subsumption test:  if the clause is s-subsumed by either a   */
/* base clause or an ancestor clause, delete it from the list   */
/*--------------------------------------------------------------*/
	
if(in_elim != 0) {
   for(i = base; i >= 0L; i--) {            /* try base clauses */
      if(clist[i]->c_nlit > c->c_nlit) {
         if(i < org_base - nc + 1L) {break;}}
      if(clist[i]->c_bits & ACTIVE) {
         if(simple_subsume2(clist[i],c))  {
            return(0);}}}
		
   for(i = ply - 1L; i > 0L; i--) {             /* try anc. cls */
      if(clist[depth[i].d_tc]->c_nlit <= c->c_nlit) {
	       if(simple_subsume2(clist[depth[i].d_tc],c)) {
	          return(0);}}}}
	
/*----------------------------------------------------------------*/
/* examine clause to see if one of its literals is the equality   */
/* predicate. if so, test for Equal(x,x) in clause and if found   */
/* delete the clause. otherwise if not, see if a literal can be   */
/* eliminated because the hash code of its complement is on       */
/* the unit clause hash table.                                    */
/*----------------------------------------------------------------*/

if(c->c_nlit > 1 && phase > 0 && lit_in_hash != 2L) {
   l_index = 0L;
   comp_lit = 0;        
   for(l = c->c_lit ; l != 0L ; l = l->l_next) {        
      l_index++;
      if(l->l_id == (equal_pred | FCNMASK)&&(l->l_id & HHITMASK)==0L) {
	     if(test_equality(l)) { 
	        comp_lit = (int) l_index;   /* remove this literal */
	        break;}}}
   if(comp_lit == 0) {
      comp_lit = is_comp_in_ht(c);}
 
/*-----------------------------------------------------*/
/* comp found, its index in cl = comp_lit; delete lit. */       
/*-----------------------------------------------------*/
		
   if(comp_lit != 0) {
      repair_literal(c,comp_lit);
      if(deep_in_tree) {
	     if(is_clause_in_ht(c) != -1) {
	        return(0);}}}}      
						
/*----------------------------------------------------------------*/
/* examine clause to see if the complement of one of its literals */
/* is in the clause hash table                                    */
/* if complement of literal is in hash table, eliminate literal,  */
/* replace literal with its hash code for later reevaluation      */
/*----------------------------------------------------------------*/
 
while(comp_lit!=0 && c->c_nlit>1 && phase > 0 && lit_in_hash!=2L) {
   comp_lit = is_comp_in_ht(c);
   if(comp_lit != 0) {
      repair_literal(c,comp_lit);
      if(deep_in_tree) {        
	    if(is_clause_in_ht(c) != -1) {
	       return(0);}}}}   
        
/*---------------------------------------------------------------*/
/* if there is only one lit in the clause execute the following  */
/*---------------------------------------------------------------*/
if(c->c_nlit == 1 && phase > 0 && lit_in_hash != 2L) {

/*------------------------------------------*/
/* find first real literal                  */
/*------------------------------------------*/
 
   for(l = c->c_lit; (l->l_id & HHITMASK) != 0L ; l = l->l_next) {;}

   if((l->l_id) == equal_pred) {
      lptr = &l->l_id;
      for(aptr = (*lptr++ & ARGMASK) ; aptr > 1L ; aptr--) {
	 if(*lptr & FCNMASK) {
	    aptr += *lptr++ & ARGMASK;}
         else {
	    lptr++;}}}

   if(c->c_nv != 0L) { 
      normalize_literal(l,c);} 
   if(c->c_nv != 0L && l != c->c_lit) {
	    calc_lit_hc(l,c);
	    calc_clause_hashcode(c);}
 
   lit_in_hash = add_clause_to_ht(l->l_X,l->l_Y,1);
   if(lit_in_hash != 2L) {

      lit_in_hash = calc_alt_hashcodes(l,c);}
   if((hash_code_list_active == 0L) && (lit_in_hash == 2L)) {
 
/*---------------------------------------*/
/* in phase 1:if contradiction found.... */
/*---------------------------------------*/
 
      c->c_bits |= MERGE;}        /* make a merge clause */
			
   if(hash_code_list_active != 0L) {  
 
/*----------------------------------------------*/
/* during phase 2: when looking for complements */
/*----------------------------------------------*/
 
      found = on_hash_code_list(l->l_X,l->l_Y);
      if(found == 0) {
         found = found + alt_on_hash_code_list(l,c);}

      if(found) { 
         save_line(c); 
         if(hash_code_list_active == 0L) {
         lit_in_hash = 2L;}}}} 

/*----------------------------------------------------------*/
/* determine where the next clause will be placed           */
/*----------------------------------------------------------*/

d->d_tc = nextc++;
 
clist[nextc] = c->c_end;

/*----------------------------------------------------------*/
/* We insert the new clause so that high scores are nearer  */
/* the end (considered first).                              */
/* However, we skip this part of the code while in elim     */
/*----------------------------------------------------------*/
 
if(in_elim == 1) {
   for(k = d->d_tc; k >= d->d_bc; k--) {
      if(k == d->d_bc) {
         break;}
      if(clist[k-1L]->c_nlit >= c->c_nlit) {
         break;}
      clist[k] = clist[k-1L];}}
else k = d->d_tc;
 
/*-----------------------------------------------------------*/
/* If too many clauses are generated, quit search            */
/* The maximum number is CSPACE_CLAUSES                      */
/* This constant is defined in const.h                       */
/*-----------------------------------------------------------*/
 
if(k > (long) CSPACE_CLAUSES - 2L) {
   printf("\nToo many clauses(%ld) \n",k);
   for(i = 0L; i <= ply; i++) {
      d = &depth[i];
      printf("\nPly %ld clauses: %ld ",i, d->d_tc);}
   print_continuation();
   getchar();
   fatal("\nQuitting search\n");}
 
if((long) c->c_end > (long) d->d_rc) {
   d->d_rc = c->c_end;}
 
/*------------------------------------------------------------*/
/* if either the left or right parent is in the NC, set this  */
/* flag in c_bits                                             */
/*------------------------------------------------------------*/

if(phase != 0) { 
if((c->c_lpar->c_bits & NC) != 0L) {
   c->c_bits |= NC;}
else if(c->c_rpar != 0L) { 
   if((c->c_rpar->c_bits & NC) != 0L) {
      c->c_bits |= NC;}}}
 
/*------------------------------------------------------------*/
/* place clause on clist and add it to hash table             */
/*------------------------------------------------------------*/

clist[k] = c;
add_clause_to_ht(c->c_X,c->c_Y,c->c_nlit);

if(lit_in_hash == 2L) { 
   magic_ptr = k;}

return(1);
}
  
/*-----------------------------------------------------------------*/
/*      calc_inferences()                ......................#2
/*
/*      calc_inferences will generate all inferences of a clause
/*      with all base clauses and all ancestor clauses (depending
/*      on the options) at the current ply. We return the number
/*      of resolvents generated.
/*
/*      called by:  search_tree()           search.c
/*
/*      calls:      binary_resolvents()     infer.c
/*                  binary_factors()        infer.c
/*                  is_unit_in_ht()         hash.c
/*                  simple_subsume2()       search.c
/*
/*      returns:    0,nc
/*-----------------------------------------------------------------*/
 
int calc_inferences()
{
register clause *c;
register long  i, j;
register long opt;
clause *c1;
dnode *d;
int number_of_inferences;
int number_of_factors;
long iter_horizon, merge_depth, nc_depth;

number_of_inferences = 0;
number_of_factors = 0;

/*--------------------------------------------------------------*/
/* if the extended search option is off and ply > iter_depth    */
/* don't generate any clauses at this node                      */
/*--------------------------------------------------------------*/
 
if(o_extend == 0) {
   if(ply > iter_depth) {
      return(0);}}
		
/*--------------------------------*/
/* c = clause to expand           */
/*--------------------------------*/
 
c = clist[depth[ply - 1L].d_tc];        
 
/*--------------------------------*/
/* check for literal subsumption  */
/*--------------------------------*/
 
if(c->c_nlit > 1) {
   if(is_unit_in_ht(c) > 0) {
      return(0);}}
 
/*---------------------------------*/
/* adjust search horizon taking    */
/* into account whether the node   */
/* is a merge or a nc descendent   */
/*---------------------------------*/
 
if(ply > 1L) {
   merge_depth = (c->c_bits & MERGE) >> 2;
   nc_depth = (c->c_bits & NC) >> 8;
   iter_horizon = iter_depth + (merge_depth | nc_depth);}
else {
   iter_horizon = iter_depth;}
 
opt = (ply <= iter_horizon) ?
	  ((long) o_merge * MERGE) : UNIT;
	
/*--------------------------------------------------*/
/* first, we generate all factors of this clause    */
/*--------------------------------------------------*/

if(c->c_bits & ACTIVE) {
   number_of_factors = binary_factors(c);
   number_of_inferences += number_of_factors;}

/*--------------------------------------------------*/
/* if c is a base clause generate all resolvents    */
/* only with clauses with small clause id's         */
/* otherwise generate resolvents with all base      */
/* clauses                                          */
/*--------------------------------------------------*/
 
if(c->c_bits & BASE) {
   for(i = c->c_id - 2L ; i >= 0L ; i--) { 
   if(lit_in_hash == 2L) {break;}
      number_of_inferences += binary_resolvents(c,clist[i],opt);}}
	
else {
   for(i = 0L; i  <= base; i++) {
      if(lit_in_hash == 2L) {break;}
      number_of_inferences += binary_resolvents(c,clist[i],opt);}}
 
/*----------------------------------------------------------*/
/* now that the base resolvents are in, we can generate     */
/* the resolvents of this clause with its ancestors         */
/*----------------------------------------------------------*/
	 
for(i = ply - 2L; i > 0L; i--) {
   if(lit_in_hash == 2L) {break;}
   number_of_inferences += 
	     binary_resolvents(c,clist[depth[i].d_tc],opt);}
 
d = &depth[ply];    

/*--------------------------------------------------*/
/* delete s-subsumed clauses by other clauses       */
/* at this node                                     */
/*--------------------------------------------------*/
 
for(i = d->d_bc ; i < d->d_tc ; i++) {
   if(clist[i]->c_bits & ACTIVE) {
      for(j = i + 1L; j <= d->d_tc; j++) {
	     if(clist[j]->c_bits & ACTIVE) {
	        if(simple_subsume2(clist[j],clist[i])) {
	           clist[i]->c_bits = (clist[i]->c_bits&(~ACTIVE));
	           break;}}}}}

/*---------------------------------------------------*/
/* if in extended search and clause c has n literals */
/* where n is greater than 2, and the number of      */
/* inferences generated is only one (so no units     */
/* can be produced, terminate search...........      */
/*---------------------------------------------------*/

if(ply > iter_depth) {
   if(c->c_nlit > 2) {
      if(number_of_inferences  == 1) { 
         c1 = clist[d->d_bc];
            if(c1->c_bits & ACTIVE) {
               if(c1->c_nlit >= c->c_nlit - 1) {
                  c1->c_bits = c1->c_bits & (~ACTIVE);}}}}}

return(number_of_inferences);
}                                                       
 
/*----------------------------------------------------------------*/
/*      search_tree()                  ......................#3
/*
/*      search will return TRUE if we have generated the
/*      empty clause during our explorations of the tree.
/*      Note: we assume that there are at least two base clauses
/*
/*      called by:  prove_by_iterative_deepening()    tgtp.c
/*
/*      calls:      calc_inferences()                 search.c
/*                  advance_to_successor()            search.c
/*                  print_clause()                    tdump.c
/*                  interrupt_options() [not used!!]  tdump.c
/*
/*      returns:    1 if proof found on current iteration
/*                  0 otherwise
/*----------------------------------------------------------------*/
 
int search_tree()
{
int adv_val, number_of_clauses;
register dnode *d;
 
depth[0].d_tc = base;               /* initial generating clause  */
depth[0].d_ec = base;
depth[0].d_bc = 0L;
x_stop = 0;
 
nextc = base + 1L;
 
clist[nextc] = depth[0].d_rc = clist[base]->c_end;
	
for(ply = 1L;;) {
   d = &depth[ply];             /* current depth node   */
		
/*------------------------------------------------------*/
/* we generate all possible resolvents and factors      */
/* with the current ply's generating clause             */
/*                                                      */
/* x_stop is normally 0, but can be set to 1 during     */
/* the search by the user when searching the tree       */
/* in an interactive mode                               */
/*------------------------------------------------------*/
 
   number_of_clauses = 0;
   if(x_stop == 0) {            
      nextc  = depth[ply - 1L].d_ec + 1L;
      clist[nextc] = depth[ply - 1L].d_rc;
      d->d_rc = 0L;
      d->d_tc = d->d_bc = nextc;
      number_of_clauses = calc_inferences();

      if(number_of_clauses == 0) {
         d->d_tc = d->d_bc;}
      d->d_ec = depth[ply - 1L].d_ec + number_of_clauses;}
   x_stop = 0;
	
   if(number_of_clauses == 0) {
      while(adv_val = advance_to_successor()) {
	 
/*-------------------------------------------------*/
/* if adv_val=2, proof found, quit search          */
/* if adv_val=1 and search is at root => no proof  */
/* if adv_val=0, exit while and search this branch */
/*               deeper                            */
/*-------------------------------------------------*/  
         if(adv_val == 2) {
	          lit_in_hash = 0L; 
	          depth[ply].d_tc = magic_ptr;
	          return(1);}
	       else if(adv_val == 3) {
	          if(x_stop == 2) {
	             return(0);}
	          return(3);}
	       else if(ply-- <= 1L) {
	          return(0);}                 
	       nextc = depth[ply].d_tc;               
	       clist[nextc] = depth[ply - 1].d_rc;}
      continue;}
 
/*--------------------------------------------------*/
/* continue deeper into tree                        */
/*--------------------------------------------------*/

   s_nod++;
   if(clist[d->d_tc]->c_rpar == 0L) {s_fac++;}
   else {s_res++;}

/*--------------------------------------------------*/
/* every 512 nodes searched, check the time to see  */
/* whether to stop. Don't stop in phase 2.          */
/*--------------------------------------------------*/
 
   if(s_nod % 512L == 0L && phase < 2) {
      if(time_now() - search_start_time >= (long) o_maxtime) {
	       return(2);}} 
   if((int) ply <= o_print)  {
      o_cr = 1;
      print_clause(clist[d->d_tc]);
      o_cr = 0;
      if((int) ply <= o_step) {
	       if(halted_search_options() == 1) {
	          return(3);}}
      if(x_stop == 2) {
         return(0);}}
 
/*--------------------------------------------------*/
/* the following 6 lines of code are not used in    */
/* this version. they have to do with single        */
/* character input (non-buffered input)             */
/*--------------------------------------------------*/

/*
   else {
      if(kbd_hit()) {
         if(interrupt_options() == 1) {
            flag_kbd_hit = 0;
	          return(3);}
         flag_kbd_hit = 0;}}
*/
		
/*--------------------------------------------------*/
/* since the empty clause has the highest score     */
/* we check the top clause of the current ply       */
/* to see if it is IT.                              */
/*--------------------------------------------------*/
	
   if(clist[d->d_tc]->c_nlit == 0) {
      return(1);}
		
   if(clist[d->d_tc]->c_nlit == 1) {
      if(lit_in_hash == 2L) { 
	       depth[ply].d_tc = magic_ptr;
	       lit_in_hash = 0L;      /* complement in hash table  */
	       return(1);}}           /* proof found               */
 
   ply++;}                            /* inc. search depth         */
}                                                       
 
/*-----------------------------------------------------------------*/
/*      advance_to_successor()          .....................#4
/*
/*      advance will modify a depth node so that we can consider
/*      the next sub-AFF tree. We return TRUE if we cannot
/*      advance the pointer and a backup is in order
/*
/*      p  = ply - 1L
/*
/*      called by:  search_tree()       search.c
/*
/*      calls:      print_clause()      tdump.c
/*
/*      returns:    0 => more branches to consider at this node
/*                  1 => no more branches to consider
/*                  2 => proof found either as a null clause
/*                       or a complement in hash table
/*------------------------------------------------------------------*/
 
int advance_to_successor()
{
register dnode *d;
int hso;

d = &depth[ply - 1L];
	
if(lit_in_hash == 2L) {
   depth[ply].d_tc = magic_ptr;
   if(ply != 1L) {
      if(clist[d->d_tc]->c_rpar == 0L) s_fac++; else s_res++;
      if((int) ply - 1 <= o_print) {
	       ply--; 
         o_cr = 1;
	       print_clause(clist[d->d_tc]);         
         o_cr = 0;
	       ply++;}}
   return(2);}

if(d->d_tc == d->d_bc) {
   return(1);}

d->d_tc--;                  /* decrement pointer to next clause     */
 
if(ply != 1L) {
   if(clist[d->d_tc]->c_rpar == 0L) {s_fac++;} else {s_res++;}
   if((int) ply - 1 <= o_print) {
      ply--; 
      o_cr = 1;
      print_clause(clist[d->d_tc]);         
      o_cr = 0;
      if((int) ply <= o_step) {
         hso = halted_search_options();}
      ply++;
      if(hso == 1) {
         return(3);}
      if(x_stop == 2) {
         return(3);}}
		
   if(clist[d->d_tc]->c_nlit == 0) {
      return(2);}}
 
return(0);
}
 
/*------------------------------------------------------------------*/
/*      simple_subsume2(c1,c2)         ......................#5
/*
/*      called by:      install_clause()    search.c
/*                      calc_inferences()   search.c
/*
/*      calls:          nobody
/*
/*      Returns "1" if c1 subsumes in the simplest way c2,
/*      ie., s-subsumes (test to see if each literal in c1
/*      appears in c2. Actually hash codes of literals are checked)
/*------------------------------------------------------------------*/
 
int simple_subsume2(c1,c2)
clause *c1,*c2;
{
register literal *l1,*l2;      /* pointers to literals in c1, c2    */
 
for(l1 = c1->c_lit ; l1 != 0L ; l1 = l1->l_next) { /* scan c1 lit s */
   if((l1->l_id & HHITMASK) == 0L) {
 
/*--------------------------------------*/
/* test to see if hashcode of l1 is     */
/* the same as the hashcode of some     */
/* literal in c2. if so, set lit_found  */
/* equal to 1                           */
/*--------------------------------------*/
 
      for(l2 = c2->c_lit ; l2 != 0L ; l2 = l2->l_next) {
	     if((l2->l_id & HHITMASK) == 0L) {
	        if(l1->l_X == l2->l_X && l1->l_Y == l2->l_Y) {
	           break;}}}
 
/*--------------------------------------*/
/* if no literal in l2 has hashcode of  */
/* l1, return with "no s-subsumption    */
/*--------------------------------------*/
 
      if(l2 == 0L)  {
	     return(0);}}}
 
if(o_subsume == 0) {
   return(0);} /* if no simple subsumption strat.,don't elim. */

s_subsume++; 
return(1);
}
 
 
 
