/**************************************************************/
/* ********************************************************** */
/* *                                                        * */
/* *              TOP MODULE OF SPASS                       * */
/* *                                                        * */
/* *  $Module:   TOP                                        * */ 
/* *                                                        * */
/* *                                                        * */
/* $Revision: 0.71 $                                        * */
/* $State: Exp $                                            * */
/* $Date: 1997/07/03 12:33:29 $                             * */
/* $Author: weidenb $                                       * */
/* *                                                        * */
/* *             Contact:                                   * */
/* *             Christoph Weidenbach                       * */
/* *             MPI fuer Informatik                        * */
/* *             Im Stadtwald                               * */
/* *             66123 Saarbruecken                         * */
/* *             Email: weidenb@mpi-sb.mpg.de               * */
/* *             Germany                                    * */
/* *                                                        * */
/* ********************************************************** */
/**************************************************************/


/* $RCSfile: top.c,v $ */

/*** MAINLOOP *************************************************/


/**************************************************************/
/* Includes                                                   */
/**************************************************************/

#include "dfgparser.h"
#include "rules-inf.h"
#include "rules-sort.h"
#include "rules-split.h"
#include "rules-red.h"
#include "analyze.h"
#include "clock.h"
#include "options.h"
#include "context.h"
#include "cnf.h"
#include "search.h"

char *top_InputFile;

#if (defined(SPASS_SIGNALS))
#include <signal.h>

void top_SigHandler(int Signal)
{
    if (Signal == SIGSEGV || Signal == SIGBUS) {
        printf("\nSPASS is going to crash. This is probably caused by a\n");
        printf("bug in SPASS.  Please send copy of the input file(s) to\n");
        printf("weidenb@mpi-sb.mpg.de, and someone will try to fix the problem.\n\n");
    }

    if (opts_IsSet(opts_RemoveFile))
      remove(top_InputFile);
}
#endif



/**************************************************************/
/* Clause Selection Functions                                 */
/**************************************************************/


static CLAUSE top_SelectClauseDepth(LIST List)
{
  CLAUSE Result;
  NAT    Vars,NewVars,Weight,Depth,NewDepth;

  Result = (CLAUSE)list_Car(List);
  Depth  = clause_Depth(Result);
  Weight = clause_Weight(Result);
  Vars   = clause_NumberOfVarOccs(Result);
  List   = list_Cdr(List);

  while (!list_Empty(List)) {
    NewDepth =  clause_Depth(list_Car(List));
    if (NewDepth <= Depth)
      if (NewDepth < Depth || clause_Weight(list_Car(List)) < Weight) {
	Depth  = NewDepth;
	Result = (CLAUSE)list_Car(List);
	Weight = clause_Weight(Result);
	Vars   = clause_NumberOfVarOccs(list_Car(List));
      }
      else
	if (clause_Weight(list_Car(List)) == Weight) {
	  NewVars = clause_NumberOfVarOccs(list_Car(List));
	  if (flag_GetFlagValue(flag_PREFVAR)) {
	    if (Vars < NewVars) {
	      Depth  = NewDepth;
	      Result = (CLAUSE)list_Car(List);
	      Weight = clause_Weight(Result);
	      Vars   = NewVars;
	    }
	  }
	  else
	    if (Vars > NewVars) {
	      Depth  = NewDepth;
	      Result = (CLAUSE)list_Car(List);
	      Weight = clause_Weight(Result);
	      Vars   = NewVars;
	    }
	}
    List = list_Cdr(List);
  }

  return Result;
}

static CLAUSE top_SelectMinimalWeightClause(LIST List)
{
  CLAUSE Result;
  NAT    Vars,NewVars,Weight;

  Result = (CLAUSE)list_Car(List);
  Weight = clause_Weight(Result);
  Vars   = clause_NumberOfVarOccs(Result);
  List   = list_Cdr(List);

  while (!list_Empty(List)) {
    if (clause_Weight(list_Car(List)) == Weight) {
      NewVars = clause_NumberOfVarOccs(list_Car(List));
      if (flag_GetFlagValue(flag_PREFVAR)) {
	if (Vars < NewVars) {
	  Result = (CLAUSE)list_Car(List);
	  Weight = clause_Weight(Result);
	  Vars   = NewVars;
	}
      }
      else
	if (Vars > NewVars) {
	  Result = (CLAUSE)list_Car(List);
	  Weight = clause_Weight(Result);
	  Vars   = NewVars;
	}
    }
    else
      return Result;
    List = list_Cdr(List);
  }
  return Result;
}


/*static CLAUSE top_SelectClauseDepth(LIST List)
{
  CLAUSE Result;
  int    Min, Depth;

  Result = (CLAUSE)list_Car(List);
  Depth  = clause_Depth(Result);
  Min    = Depth * clause_Weight(Result);
  List   = list_Cdr(List);

  if (Depth == 0)
    return Result;

  while (!list_Empty(List)) {
    Depth =  clause_Depth(list_Car(List));
    if (Min > Depth * clause_Weight(list_Car(List))) {
      Result = list_Car(List);
      if (Depth == 0)
	return Result;
      Min    = clause_Depth(Result) * clause_Weight(Result);
    }
    List = list_Cdr(List);
  }

  return Result;
}*/



/**************************************************************/
/* Main Function                                              */
/**************************************************************/

void main(int argc, char* argv[])
{
  LIST         InputClauses,Usables,WorkedOffs,Derivables,EmptyClauses,
               SubsortClauses,Scan,Axioms,Conjectures;
  CLAUSE       GivenClause, SplitClause, HighestLevelEmptyClause;
  int          Counter,Backtracked,Loops,Splits;
  NAT          Derived;
  PROOFSEARCH  Search;

#if (defined(SPASS_SIGNALS))
  signal(SIGINT, top_SigHandler);
  signal(SIGSEGV, top_SigHandler);
  signal(SIGBUS, top_SigHandler);
#endif


  clock_Init();
  clock_StartCounter(clock_OVERALL);
  memory_Init(-1);
  symbol_Init(TRUE);
  stack_Init();
  hash_Init();
  sort_Init();
  term_Init();
  fol_Init(TRUE);
  cont_Init();
  unify_Init();
  flag_Init();
  subs_Init();
  clause_Init();
  red_Init();
  dp_Init();
  opts_Init();


  if (!opts_Read(argc, argv)) 
    return;

  if (argc != opts_Indicator()+1) {                 /* No input file given */
    printf("\n\t          SPASS Version 0.55");
    printf("\n\t       Usage: SPASS [options] <input-file> \n\n");
    printf("Possible options:\n\n");
    opts_PrintNames(); 
    return;
  }

  Axioms = Conjectures = list_Nil();
  top_InputFile = argv[opts_Indicator()];
  clock_StartCounter(clock_INPUT);
  Derivables = dfg_DFGParser(top_InputFile, &Axioms, &Conjectures);
  clock_StopPassedTime(clock_INPUT);

  opts_Transfer();
  if (opts_IsSet(opts_RemoveFile))
    remove(top_InputFile);
  opts_Free();

  clock_StartCounter(clock_CNF);
  if (list_Empty(Derivables)) {
    InputClauses = EmptyClauses = list_Nil();
    cnf_Flotter(Axioms,Conjectures,&InputClauses,&EmptyClauses);
    Derivables = list_Nconc(InputClauses,EmptyClauses);
    InputClauses = EmptyClauses = list_Nil();
  }
  else {
    list_DeleteWithElement(Axioms,term_Delete);
    list_DeleteWithElement(Conjectures,term_Delete);
  }
  clock_StopPassedTime(clock_CNF);


  Search                  = prfs_Create(flag_GetFlagValue(flag_DOCPROOF));
  HighestLevelEmptyClause = (CLAUSE)NULL;
  InputClauses            = list_Nil();
  Usables                 = list_Nil();
  WorkedOffs              = list_Nil();
  EmptyClauses            = list_Nil();
  SubsortClauses          = list_Nil();
  Counter                 = 1;
  Derived                 = 0;
  Backtracked             = 0;
  Loops                   = flag_GetFlagValue(flag_LOOPS);


  memory_Restrict(flag_GetFlagValue(flag_MEMORY));
  prfs_PutSplitCounter(Search,flag_GetFlagValue(flag_SPLITS));

  ana_AnalyzeProblem(Derivables); /* Important ! Recomputes Weight ! */

  if (flag_GetFlagValue(flag_AUTO)) {
    ana_AutoConfiguration(Derivables);
  }

  if (flag_GetFlagValue(flag_PPROBLEM)) {
    printf("\nInput Problem:\n");
    clause_ListPrint(Derivables);
    ana_Print();
  }

  if (flag_GetFlagValue(flag_SORTS) != 0) {
    BOOL Strong;
    Strong     = (flag_GetFlagValue(flag_SORTS) == 2);
    for(Scan=Derivables;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
      GivenClause = (CLAUSE)list_Car(Scan);
      clause_SetSortConstraint(GivenClause,Strong);
    }
  }

  if (flag_GetFlagValue(flag_RINPUT))
    EmptyClauses = red_SatInput(Search, Derivables, flag_GetFlagValue(flag_DOCPROOF),&Derived);
  else {
    for(Scan=Derivables;!list_Empty(Scan);Scan=list_Cdr(Scan))
      prfs_InsertUsableClause(Search, list_Car(Scan));
    list_Delete(Derivables);
  }
  Derivables = list_Nil();

  if (ana_SortRestrictions()) {
    if (flag_GetFlagValue(flag_RSST))
      prfs_PutStaticSortTheory(Search,sort_ApproxStaticSortTheory(prfs_UsableClauses(Search)));
    if (flag_GetFlagValue(flag_RDST))
      prfs_PutApproximatedDynamicSortTheory(Search,
					    sort_ApproxDynamicSortTheory(prfs_UsableClauses(Search)));
    prfs_PutDynamicSortTheory(Search,sort_TheoryCreate());
  }

  if (flag_GetFlagValue(flag_PPROBLEM)) {
    printf("\nProcessed Problem:\n");
    clause_ListPrint(prfs_UsableClauses(Search));
  }

  Derivables  = list_Nil();
  SplitClause = (CLAUSE)NULL;
  Splits      = flag_GetFlagValue(flag_SPLITS);

  while ((list_Empty(EmptyClauses) || !prfs_SplitIsTopLevel(Search)) && 
	 Loops != 0 &&
	 (!list_Empty(prfs_UsableClauses(Search)) || !list_Empty(EmptyClauses)) &&
	 (flag_GetFlagValue(flag_TIMELIMIT) < 0 || 
	  flag_GetFlagValue(flag_TIMELIMIT) > clock_GetSeconds(clock_OVERALL))) {
      
    if (!list_Empty(EmptyClauses)) {

      clock_StartCounter(clock_BACKTRACK);
      Derivables = split_Backtrack(Search,HighestLevelEmptyClause);
      clock_StopAddPassedTime(clock_BACKTRACK);
      Backtracked += list_Length(Derivables);
      
      if (prfs_SplitIsTopLevel(Search)) {
	Derivables = list_Nconc(EmptyClauses,Derivables);
      }
      else {
	LIST Scan2;
	for(Scan2 = EmptyClauses; Scan2 != list_Nil(); Scan2 = list_Cdr(Scan2))
	  if ((CLAUSE) list_Car(Scan2) != HighestLevelEmptyClause)
	    clause_Delete((CLAUSE) list_Car(Scan2));
	prfs_InsertDocProofClause(Search, HighestLevelEmptyClause);
	list_Delete(EmptyClauses);
	if (flag_GetFlagValue(flag_DOCSPLIT))
	  printf("\n\t Split Backtracking level %d:",prfs_ValidLevel(Search));
      }
      EmptyClauses            = list_Nil();
      HighestLevelEmptyClause = (CLAUSE)NULL;
    }
    else {			/* not (EMPTY_CLAUSE). */

#if DEBUG
  if (!prfs_Check(Search)) {
    fputs("\n\nIllegal state of search in mainloop.\n", stderr);
    misc_DumpCore();
  }
#endif 

      if ((SplitClause == (CLAUSE)NULL) && (Counter%flag_GetFlagValue(flag_WDRATIO)) == 0)
	GivenClause = top_SelectClauseDepth(prfs_UsableClauses(Search));
      else
	GivenClause = top_SelectMinimalWeightClause(prfs_UsableClauses(Search));
      Counter++;
      
      if (flag_GetFlagValue(flag_FULLRED)) 
	Derivables = list_Nil();
      else {
	clock_StartCounter(clock_REDUCTION);
	Derivables = red_BackReduction(Search,GivenClause,flag_GetFlagValue(flag_DOCPROOF),red_ALL);
	clock_StopAddPassedTime(clock_REDUCTION);
      }
      
      
      if (Splits != 0 && (SplitClause = prfs_PerformSplitting(Search,GivenClause)) != (CLAUSE)NULL) {
	Splits--;
	Derivables = list_Cons(SplitClause,Derivables);
      }
      else { 
	clause_SelectLiteral(GivenClause);
	prfs_MoveUsableWorkedOff(Search,GivenClause);
	clock_StartCounter(clock_INFERENCE);
	Derivables = list_Nconc(Derivables,inf_DerivableClauses(Search,GivenClause));
	clock_StopAddPassedTime(clock_INFERENCE);
      }

      if (flag_GetFlagValue(flag_PGIVEN)) {
	printf("\n\tGiven clause: ");
	clause_Print(GivenClause); 
	fflush(stdout);
      }            

      Derived   += list_Length(Derivables);
    }

    if (flag_GetFlagValue(flag_PDER))
      for(Scan=Derivables;!list_Empty(Scan);Scan=list_Cdr(Scan)) {
	printf("\nDerived: "); 
	clause_Print(list_Car(Scan));
      }

    Derivables = split_ExtractEmptyClauses(Derivables,&EmptyClauses);

    if (flag_GetFlagValue(flag_FULLRED)) { 
      clock_StartCounter(clock_REDUCTION);
      EmptyClauses = list_Nconc(EmptyClauses,
				red_CompleteReductionOnDerivedClauses(Search, Derivables, 
					       flag_GetFlagValue(flag_DOCPROOF), &Derived));
      clock_StopAddPassedTime(clock_REDUCTION);
    }
    else {
      clock_StartCounter(clock_REDUCTION);
      EmptyClauses = list_Nconc(EmptyClauses,
	red_ReductionOnDerivedClauses(Search, Derivables, 
				      flag_GetFlagValue(flag_DOCPROOF), red_WORKEDOFF));
      clock_StopAddPassedTime(clock_REDUCTION);
    }
      

    if (!list_Empty(EmptyClauses)) {
      HighestLevelEmptyClause = split_SmallestSplitLevelClause(EmptyClauses);
      if (flag_GetFlagValue(flag_PEMPTYCLAUSE)) {
	printf("\nEmptyClause: ");
	clause_Print(HighestLevelEmptyClause);
      }
    }
    Loops--;
    /*prfs_Print(Search);*/
  }

  printf("\nProblem: %s ", top_InputFile);

  if (flag_GetFlagValue(flag_TIMELIMIT) > 0 &&
      flag_GetFlagValue(flag_TIMELIMIT) <= clock_GetSeconds(clock_OVERALL))
    printf("\nSPASS beiseite: Ran out of time.");
  else
    if (Loops == 0)
      printf("\nSPASS beiseite: Maximal number of loops exceeded.");
    else
      printf("\nSPASS beiseite: %s found.",(list_Empty(EmptyClauses) ? "Completion" : "Proof"));

  if (flag_GetFlagValue(flag_PSTATISTIC)) {
    clock_StopPassedTime(clock_OVERALL);
    printf("\nSPASS derived %d clauses, backtracked %d clauses "
	   "and kept %d clauses.", Derived,Backtracked,prfs_KeptClauses(Search));
    printf("\nSPASS allocated %d KBytes.", memory_DemandedBytes()/1024);
    printf("\nSPASS spent\t"); clock_PrintTime(clock_OVERALL);
    printf(" on the problem.");
    printf("\n\t\t"); clock_PrintTime(clock_INPUT);
    printf(" for the input.");
    printf("\n\t\t"); clock_PrintTime(clock_CNF);
    printf(" for the FLOTTER CNF translation.");
    printf("\n\t\t"); clock_PrintTime(clock_INFERENCE);
    printf(" for inferences.");
    printf("\n\t\t"); clock_PrintTime(clock_BACKTRACK);
    printf(" for the backtracking.");
    printf("\n\t\t"); clock_PrintTime(clock_REDUCTION);
    printf(" for the reduction.\n");
  }

  if (list_Empty(EmptyClauses) && flag_GetFlagValue(flag_FPMODEL)) {
    FILE *Output;
    char name[100];
    sprintf(name,"%s.model",top_InputFile);
    Output = misc_OpenFile(name,"w");
    fprintf(Output, "begin_problem(Completion_by_SPASS).\n\n");
    dfg_FPrintDescription(Output);
    fprintf(Output, "list_of_symbols.\n");
    fol_PrintDFGSignature(Output);
    fprintf(Output, "end_of_list.\n\n");
    clause_FPrintCnfDFG(Output, list_Nil(), prfs_WorkedOffClauses(Search));
    clause_FPrintCnfDFG(Output, list_Nil(), prfs_UsableClauses(Search));
    fprintf(Output, "list_of_settings(SPASS).\n{*\n");
    fol_FPrintPrecedence(Output);
    fprintf(Output, "\n*}\nend_of_list.\n\n");
    fprintf(Output, "end_problem.\n\n");
    misc_CloseFile(Output,name);
    printf("\nModel printed to: %s\n",name);
  }

  if (flag_GetFlagValue(flag_DOCPROOF))
    if (list_Empty(EmptyClauses)) {
      if (Loops != 0) {
	printf("\n\n The saturated set of worked-off clauses is :\n");
	clause_ListPrint(prfs_WorkedOffClauses(Search));
      }
    }
    else
      dp_PrintProof(Search, EmptyClauses,top_InputFile);

  clause_DeleteClauseList(EmptyClauses);
  fflush(stdout);

  prfs_Delete(Search);

  /*symbol_PrintAll();*/
  sort_Free();
  unify_Free();
  if (flag_GetFlagValue(flag_PFLAGS)) {
    printf("\n");
    flag_Print();
  }
  cont_Free();
  /*symbol_Dump();*/
  fol_Free();
  symbol_FreeAllSymbols();
  dfg_Free();
#ifdef DEBUG
  memory_Print();
#endif
  printf("\n");
}
