-TITLE THINKER - a little theorem prover
*FRAME
* 
*   PROGRAM NAME:  THINKER 
* 
*   AUTHOR: F. J. Pelletier 
*	    Dan Wilson 
* 
*   DATE: 6 April 1979 
* 
*   PROGRAM PURPOSE:  Using heuristic techniques, theorems in 
*   propositional calculus are shown in a complete proof 
*   format.  When the program gets lost it asks for guidance. 
* 
*   I/O units: 
*      SCARDS - All input. 
*      SPRINT - Terminal output. 
*      -TH.OUT - FORMAT output.
* 
*   CHANGE HISTORY - 
*   DCW   80 02 07 - support the APL character set for 9700 
*   DCW   80 02 07 - fix the definition of a valid formula 
*   DCW   80 02 08 - the FINDCB cheats - it does not set  R and S 
*		     to null as it should - fix 
*   DCW   80 02 08 - above fix causes problems - change at POOP 
*		     to do a failure branch to INDIRECT - 
*		     ramifications as to halt uncertain ... OK 
*   FJP   80 02 10 - fix some of the proof techniques so that all 
*		     most everything (such as (p+~~~p) can be 
*		     shown. 
*   DCW   80 02 10 - attempt to add a tidy function which deletes 
*		     unused antecedent lines - this does not work 
*		     correctly - it deletes to many things 
*   DCW   80 05 25 - renumber the file - VERSION 2 begins 
*   CNH   86 06 23 - add a tidy function which deletes unused
*		     antecedent lines 
*   CNH   86 07 04 - fix the identity problem by stating all
*		     consequences of identities when they are
*		     entered.
*   CNH   86 07 07 - add a function to delete unnecessary show 
*		     lines.
*   CNH   86 07 16 - fix the problem with identities in the
*		     premises.	NOTE:  there may be still a
*		     problem in findid if PLINE is nil.
*   CNH   86 07 16 - renumber
*   CRM   87 04 29 - changed io routines so we can run on both mts and unix
* 
*   AVAILABILITY: To obtain a copy contact one of the above 
*		  authors in PELL or ZEUS respectively. 
* 
-STITL DATA - Definitions and Variables
*     No limit on number of statements should speed things up a bit.
      &STLIMIT = -1
*     Turn off profiling for better efficiency
      &PROFILE = 0
-STITL PROFILING information retrieval.
* lets try redefining define, so that we can profile things
      func_table = TABLE(30)
      DATA("func_stats(location,call,return,freturn,nreturn,time)")
      func_table<"**MAIN"> = 0
-STITL Function: DEFINE(dstr,lab)
*FRAME
* Function: DEFINE 
* Date: June 1983 
* Last modification: June, 1983
* Purpose: 
*	Redifines the system routine DEFINE. This will allow
*	statistics on each DEFINEd function to be collected.
* 
* Calling sequence: 
*	DEFINE(dstr,lab) 
* Parameters: 
*	dstr  - Data string containing the normal DEFINE parmeter.
*	lab   - Entry point for the defined function (if null then
*		it is assumed to be the function name).
* 
* On return: 
*	RETURN	- DEFINE succeeded
*	FRETURN - DEFINE failed.
* 
* Local variables: 
*	lastno	- Statement number of origin of DEFINE call.
*	fname	- Name the the function being DEFINEd.
* 
      DEFINE('profDEFINE(dstr,lab)lastno,fname')    :(endprofDEFINE)
profDEFINE
+     lastno = &LASTNO
      dstr BREAK('(') $ fname
      func_table<fname> = lastno
       DEFINE(dstr,lab) 			:S(RETURN)F(FRETURN)
endprofDEFINE
-STITL Function: profile_on()
*FRAME
* Function: profile_on
* Date: June 1983 
* Last modification: June, 1983
* Purpose: 
*     "profile_on" will turn the profiling mechanism on by seting
*     up a TRACE on each of the entries in the table "func_table".
* 
* Calling sequence: 
*	profile_on()
* Parameters: none
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: none
* 
       DEFINE("profile_on()")			    :(endprofile_on)
profile_on
+     func_array = CONVERT(func_table,"ARRAY")
* run thru and turn function tracing on for each
      i = 1
pn_loop
+     fname = func_array<i,1>			      :F(pn_endloop)
      LEQ(fname,"**MAIN")				 :S(pn_main)
      TRACE(fname,"CALL",i,.call_tracer)
      TRACE(fname,"RETURN",i,.return_tracer)
pn_cont
+     func_array<i,2> = func_stats(func_array<i,2>,0,0,0,0)
      i = i + 1 					  :(pn_loop)
pn_main
+     G_TRACEBACK = STACK(i)				  :(pn_cont)
pn_endloop
+     CPUTIME(0,0,goo)
      G_TICKER = 0.0D+0
      PUTLINE(" ** WARNING ** &TRACE is curently less then 100"
+	       LT(&TRACE,100))
+							   :(RETURN)
endprofile_on
-STITL Function: profile_off()
*FRAME
* Function: profile_off
* Date: June 1983 
* Last modification: June, 1983
* Purpose: 
* "profile_off()" will deactivate profiling by calling STOPTR for
* every element in func_array.
* 
* Calling sequence: 
*	profile_off()
* Parameters: none
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables:
*	i - index into func_array.
*	fname - name of each function.
*
      DEFINE("profile_off()i,fname")		   :(endprofile_off)
profile_off
+     IDENT(func_array) 				  :S(RETURN)
      i = 0
po_loop
+     i = i + 1
      fname = func_array<i,1>				  :F(RETURN)
      LEQ(fname,"**MAIN")				 :S(po_loop)
      STOPTR(fname,"CALL")
      STOPTR(fname,"RETURN")				  :(po_loop)
endprofile_off
-STITL Function: print_profile()
*FRAME
* Function: print_profile
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* print_profile is called by CMD_DISPLAY to print the profiling
* information.
* 
* Calling sequence: 
*	print_profile()
* Parameters: 
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: 
*	i - Index into func_array.
*	fname - Function name entry in func_array.
*	stats - DATA (func_stats) entry in func_array.
*
       DEFINE("print_profile()i,fname,stats")	 :(endprint_profile)
print_profile
      IDENT(func_array) 			    :S(pp_noprofile)
      i = 1
      PUTLINE( RPAD("function",20) "	 calls" "   returns"
+	 "  freturns" "  nreturns" "   time" )
pp_loop
      fname = func_array<i,1>			      :F(pp_endloop)
      stats = func_array<i,2>
      PUTLINE( RPAD(fname,20)
+	 LPAD(call(stats),10)
+	 LPAD(return(stats),10)
+	 LPAD(freturn(stats),10)
+	 LPAD(nreturn(stats),10)
+	 "   " CONVERT(time(stats) * 1000,"REAL") )
      i = i + 1 					  :(pp_loop)
pp_endloop
							   :(RETURN)
pp_noprofile
      PUTLINE(" There is no profile data to print")	   :(RETURN)
endprint_profile
-STITL Function: profile_histogram(callsize)
*FRAME
* Function: profile_histogram
* Date: June 1983 
* Last modification: June, 1983
* Purpose: 
*	This function outputs the histograms associated with
*	the information collected by the PROFILING mechenism.
*
* 
* Calling sequence: 
*     profiling_histogram(callsize)
* Parameters: 
*     callsize - ?
* 
* On return: 
*	RETURN	- normal termination.
* 
* Local variables: 
*     i - counter
*     fname - Name of function whose information is currently being
*	      formatted in the histogram.
*     tcals - numeric total of all the "call" fields in func_array.
*     maxcalls - the maximum value for any of the "call fields in
*		 func_array.
* 
       DEFINE("profile_histogram(callsize)i,fname,tcalls,maxcalls")
+					     :(endprofile_histogram)
profile_histogram
+     IDENT(func_array) 			    :S(ph_noprofile)
* first pass thru to count up the number of all calls
      i = 1
      maxcalls = 0
ph_loop
+     tcalls = tcalls + call(func_array<i,2>)	      :F(ph_endloop)
      maxcalls = GT(call(func_array<i,2>),maxcalls) call(func_array<i,2>)
      i = i + 1 					  :(ph_loop)
ph_endloop
+     EQ(tcalls,0)				    :S(ph_noprofile)
      rtcalls = CONVERT(tcalls,"REAL")
      rmaxcalls = CONVERT(maxcalls,"REAL")
      barsize = callsize / rmaxcalls
* second pass, print the charts
* but first the headings
      PUTLINE( LPAD("Function call/return histogram",50) )
      PUTLINE( RPAD("function",20) "count    %" )
      i = 1
ph_loop2
+     fname = func_array<i,1>			     :F(ph_endloop2)
      stats = func_array<i,2>
      PUTLINE( RPAD(fname,20) LPAD(call(stats),5)
+	 LPAD( CONVERT((call(stats) / rtcalls) * 100,"INTEGER"),5)
+	 "  |" DUPL("R", return(stats) * barsize)
+	 DUPL("F", freturn(stats) * barsize)
+	 DUPL("N", nreturn(stats) * barsize) )
      i = i + 1 					 :(ph_loop2)
ph_endloop2
+							   :(RETURN)
ph_noprofile
      PUTLINE(" There is no profile data to print")	   :(RETURN)
endprofile_histogram
-STITL Function: call_tracer(fname,tag)
*FRAME
* Function: call_tracer
* Date: June 1983 
* Last modification: June, 1983
* Purpose: 
*	This routine collects profiling information when the
*	the variable "PROFILING" is set to on. It works by seting
*	TRACE calls for each function.
* 
* Calling sequence: 
*	(This is called by TRACE, and not by the user).
* Parameters: 
*     fname - the NAME of the function which was being traced.
*     tag - Whatever was attached to the tag field of TRACE call.
*	    (in this case it is an index into func_array).
* 
* On return: 
*	RETURN	- normal termination
* 
* Local variables: 
*     lastno - The statement number of the line where the call
*	       took place.
*     t - CPU time used.
*     l - V field of the DATA structure pointed to by G_TRACEBACK.
* 
       DEFINE("call_tracer(fname,tag)lastno,t,l")  :(endcall_tracer)
call_tracer
+     lastno = &LASTNO
      t = CPUTIME(15,0,goo)
      l = V(G_TRACEBACK)
      G_TRACEBACK = STACK(tag,G_TRACEBACK)
      call(func_array<tag,2>) = call(func_array<tag,2>) + 1
      time(func_array<l,2>) = time(func_array<l,2>) +
+	 (t - G_TICKER)
      G_TICKER = CPUTIME(15,0,goo)
+							   :(RETURN)
endcall_tracer
-STITL Function: return_tracer(fname,tag)
*FRAME
* Function: return_tracer
* Date: June 1983 
* Last modification: June, 1983
* Purpose: 
* return_tracer is identical to call_tracer except that it keeps
* track of different statitistics. 
* 
* Calling sequence: 
*	(This is called by TRACE, and not by the user).
* Parameters: 
*     fname - the NAME of the function which was being traced.
*     tag - Whatever was attached to the tag field of TRACE call.
*	    (in this case it is an index into func_array).
* 
* On return: 
*	RETURN	- normal termination
* 
* Local variables: 
*     lastno - The statement number of the line where the call
*	       took place.
*     t - CPU time used.
*     l - V field of the DATA structure pointed to by G_TRACEBACK.
* 
*
*FRAME
*
       DEFINE("return_tracer(fname,tag)lastno,t,l")
+						 :(endreturn_tracer)
return_tracer
+     lastno = &LASTNO
      t = CPUTIME(15,0,goo)
      l = V(G_TRACEBACK)
      G_TRACEBACK = N(G_TRACEBACK)
      time(func_array<l,2>) = time(func_array<l,2>) +
+	 (t - G_TICKER)
      return(func_array<tag,2>) = LEQ(&RTNTYPE,"RETURN")
+	 return(func_array<tag,2>) + 1			  :S(retten)
      freturn(func_array<tag,2>) = LEQ(&RTNTYPE,"FRETURN")
+	 freturn(func_array<tag,2>) + 1 		  :S(retten)
      nreturn(func_array<tag,2>) = LEQ(&RTNTYPE,"NRETURN")
+	 nreturn(func_array<tag,2>) + 1 		  :S(retten)
retten
+     G_TICKER = CPUTIME(15,0,goo)			   :(RETURN)
endreturn_tracer
-STITL Parameter checking for SPITBOL
*FRAME
* Function: pDEFINE 
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*  This package will check parameter for specified routines
*  (using pDEFINE) in THINKER. See the file TINK:PDF.DOC for
*  more details on how to use pDEFINE.
* 
* Calling sequence: 
*	pDEFINE(pD_header,pD_entry,pD_legalpars,pD_legalrets,
*      +	pD_size)
* Parameters: 
*	pD_header - Normal header string used as first parameter
*		    to DEFINE.
*	pD_entry - Entry point for function when it is called.
*	pD_legalpars - Legal paramters. Each field is seperated by
*		       commas "," and each alternative value for
*		       a field is seperated by vertical bars "|".
*	pD_legalrets - Possible data types for return.
*	pD_size - Initilisation size for table of function (only
*		  used on first call to pDEFINE).
* 
* On return: 
*	RETURN	- All went well.
*	FRETURN - DEFINE failed.
* 
* Local variables: 
*	pD_name - Name of function to be declared.
*	pD_pars - Parameters of that function.
*	pD_i - Index into the array.
*	pD_npars - Number of parameters in function.
*	pD_calls - Array of legal parameter types for function call.
*	pD_types - Temporary variable containing the current TYPE.
*	pD_returns - Legal return types for function RETURN.
*
* Global variables :
*      g_split - Character used to split the parameter fields.
*      g_or - Character used to seperate alternatives within a field
*      g_functionsize - Default size for function table.
*      g_pDEFINEflag - Indicates if pDEFINE has already been called.
*      g_functable - Table of functions.
*      g_functions - 2D array equivalent to g_functable.
*      g_parcheckflag - Indicates if the package is set to ON, OFF
*			or ONRETURN.
* 
* First declare all the global variables we will need
* 
*      OUTPUT('SERCOM','SERCOM','(1X,32000A1)')
      g_split	     = ','
      g_or	     = '|'
      g_functionsize = 50
      g_pDEFINEflag  = 'Uninitialised'
      DATA('PARDATA(LEGALCALL,LEGALRET,PAR_NO)')
*
      DEFINE('pDEFINE(pD_header,pD_entry,pD_legalpars,pD_legalrets,'
+	     'pD_size)pD_name,pD_pars,pD_i,pD_npars,' 
+	     'pD_calls,pD_types,pD_returns')
+						       :(ENDpDEFINE)
pDEFINE
* 
* If this is the first call then initialize g_functable to given
* size or default value.
*
      LEQ(g_pDEFINEflag,'Initialised')		       :S(pD_crunch)
      g_functionsize = DIFFER(pD_size) pD_size
      g_functable = TABLE(g_functionsize)
      g_pDEFINEflag = 'Initialised'
*
* Take out all spaces.
*
pD_crunch
      pD_head ' ' = null			       :S(pD_crunch)
      pD_legalpars ' ' = null			       :S(pD_crunch)
      pD_legalrets ' ' = null			       :S(pD_crunch)
*
* Call DEFINE and start parsing.
*
      DEFINE(pD_header,pD_entry)
      pD_header BREAK('(') . pD_name
      pD_header BREAK('(') (BREAK(')') ')') . pD_pars
*
* If no parameters then skip all parsing.
*
      pD_npars = LEQ(pD_pars,'()') 0		       :S(pD_nopars)
      pD_npars = 1
pD_loop1
      pD_pars ',' = null				  :F(pD_010)
      pD_npars = pD_npars + 1				 :(pD_loop1)
pD_010
      pD_calls = ARRAY(pD_npars)
      pD_i = 0
*
* Get each list and insert it into the array.
*
pD_loop2
	 pD_i = pD_i + 1
	 LE(pD_i,pD_npars)				 :F(pD_020)
	 pD_legalpars = pD_legalpars g_split
	 pD_calls<pD_i> = g_or
pD_loop3
	    IDENT(pD_legalpars) 			 :S(pD_020)
	    pD_legalpars BREAK(g_split g_or) . pD_types
+	      LEN(1) . pD_chr = null 
	    pD_types = IDENT(pD_types) 'ANY' 
	    pD_calls<pD_i> = pD_calls<pD_i> pD_types g_or
	    LEQ(pD_chr,g_or)		     :S(pD_loop3)F(pD_loop2)
pD_020
* 
* Now fill in remaining parameters with ANY's
*
pD_loop4
      pD_i = LE(pD_i,pD_npars) pD_i + 1 		  :F(pD_030)
      pD_calls<pD_i> = g_or 'ANY' g_or			 :(pD_loop4)
pD_030
pD_nopars
*
* Now get the legal return types.
*
	 pD_types = null
	 pD_returns = g_or
	 pD_legalrets = pD_legalrets g_or
pD_loop5
	 pD_legalrets BREAK(g_or) . pD_types g_or = null
	 pD_types = IDENT(pD_types) 'ANY'
	 pD_returns = pD_returns pD_types g_or
	 DIFFER(pD_legalrets)				:S(pD_loop5)
pD_040
*
* Now lets set up the data structure.
*
      g_functable<pD_name> = PARDATA(pD_calls,pD_returns,pD_npars)
+							   :(RETURN)
ENDpDEFINE
-STITL Turning the system on and off.
*
* PARCHECK is called to turn the parameter checking on and off.
* 
      g_parcheckflag = 'OFF'
      DEFINE('PARCHECK(PC_switch)PC_i')
+						      :(ENDPARCHECK)
PARCHECK
      PC_switch = REPLACE(PC_switch,'onf','ONF')
      LEQ(PC_switch,'ON')			      :S(PC_oncheck)
      LEQ(PC_switch,'ONRETURN') 		      :S(PC_oncheck)
      LEQ(PC_switch,'OFF')		   :S(PC_offcheck)F(FRETURN)
*
PC_offcheck 
*
* Cannot turn PARCHECK OFF twice in a row.
*
      LEQ(g_parcheckflag,'OFF') 			 :S(FRETURN)
      PC_i = 0
PC_loop1
	 PC_i = PC_i + 1
	 g_functions<PC_i,2>			       :F(PC_return)
	 STOPTR(g_functions<PC_i,1>,'CALL')
	 STOPTR(g_functions<PC_i,1>,'RETURN')		 :(PC_loop1)
*
PC_oncheck
      g_parcheckflag 'ON'			       :S(PC_return)
*
* OK lets turn on the tracing.
*
      g_functions = CONVERT(g_functable,"ARRAY")
      PC_i = 0
PC_loop2
      PC_i = PC_i + 1
      g_functions<PC_i,2>			       :F(PC_return)
      IDENT(g_functions<PC_i,2>)		       :S(PC_return)
      TRACE(g_functions<PC_i,1>,'CALL',PC_i,'CALLCHECK')
      TRACE(g_functions<PC_i,1>,'RETURN',PC_i,'RETCHECK')
+							 :(PC_loop2)
PC_return
      g_parcheckflag = PC_switch			   :(RETURN)
ENDPARCHECK
-STITL Exit routine for CALL parameter checking.
*
* CALLCHECK will annalise each of the arguments and verify
* that it's DATATYPE is defined in call to pDEFINE.
* 
      DEFINE('CALLCHECK(CC_name,CC_tag)CC_npars,CC_partypes,CC_i,'
+	     'CC_type,CC_arg,CC_legtypes,CC_where')  :(ENDCALLCHECK)
CALLCHECK
+     CC_where	  = &LASTNO
      CC_name	  = CONVERT(CC_name,'STRING')
      CC_npars	  = PAR_NO(g_functions<CC_tag,2>)
      EQ(CC_npars,0)					  :S(CC_010)
      CC_partypes = LEGALCALL(g_functions<CC_tag,2>)
      CC_i	  = 0
CC_loop1
	 CC_i	 = LT(CC_i,CC_npars) CC_i + 1		  :F(CC_010)
	 CC_partypes<CC_i> g_or 'ANY' g_or		:S(CC_loop1)
	 CC_arg  = ARG(CC_name,CC_i)
	 CC_type = IDENT(CC_arg) 'NULL' 		  :S(CC_005)
	 CC_type = DATATYPE($CC_arg)
	 CC_type = LEQ(CC_type,'STRING') IDENT($CC_arg) 'NULL'
CC_005
	 CC_partypes<CC_i> g_or CC_type g_or		:S(CC_loop1)
*
* Check for FINTEGER and FREAL
*
	 IDENT(CC_arg)				       :S(CC_badpar)
	 CC_partypes<CC_i> g_or 'FINTEGER' g_or 	:F(CC_FREAL)
	 CONVERT($CC_arg,'INTEGER')			:S(CC_loop1)
CC_FREAL
	 CC_partypes<CC_i> g_or 'FREAL' g_or	       :F(CC_badpar)
	 CONVERT($CC_arg,'REAL')	    :S(CC_loop1)F(CC_badpar)
*
CC_010							   :(RETURN)
CC_badpar
      CC_partypes<CC_i> LEN(1) RTAB(1) . CC_legtypes
      SERCOM = '** ERROR ** Parameter no. ' CC_i ' of "' CC_name '"'
      SERCOM = '   was of type ' CC_type ' while it expected'
      SERCOM = '   a parameter of type ' CC_legtypes '. Call'
      SERCOM = '   originated from statement no. ' CC_where ' **'
      LEQ(g_parcheckflag,'ONRETURN')			:S(CC_loop1)
      DEBUG()					  :S(CC_loop1)F(END)
ENDCALLCHECK
-STITL Exit routine for RETURN check.
*
* RETCHECK will check the return values of the functions.

      DEFINE('RETCHECK(RC_name,RC_tag)RC_partypes,RC_type,'
+	     'RC_arg,RC_legtypes,RC_where')	      :(ENDRETCHECK)
RETCHECK
+     RC_where = &LASTNO
      LEQ(&RTNTYPE,'RETURN')				  :F(RETURN)
      RC_name  = CONVERT(RC_name,'STRING')
      RC_partypes = LEGALRET(g_functions<RC_tag,2>)
      RC_partypes g_or 'ANY' g_or			  :S(RETURN)
      RC_type = DATATYPE($RC_name)
      RC_type = LEQ(RC_type,'STRING') IDENT($RC_name) 'NULL'
      RC_partypes g_or RC_type g_or			  :S(RETURN)
*
* Check for FINTEGER and FREAL
*
      RC_partypes g_or 'FINTEGER' g_or			:F(RC_FREAL)
      CONVERT($RC_name,'INTEGER')			  :S(RETURN)
RC_FREAL
      RC_partypes g_or 'FREAL' g_or		       :F(RC_badpar)
      CONVERT($RC_name,'REAL')				  :S(RETURN)
RC_badpar
      RC_partypes LEN(1) RTAB(1) . RC_legtypes
      SERCOM = '** ERROR ** ' RC_type ' is an illegal return type'
      SERCOM = '   from "' RC_name '".	Type ' RC_legtypes ' was'
      SERCOM = '   expected. Error occured at statement ' RC_where ' **'
      LEQ(g_parcheckflag,'ONRETURN')			  :S(RETURN)
      DEBUG()					    :S(RETURN)F(END)
ENDRETCHECK
-STITL DATA structures used in THINKER.
*FRAME
*  ANTE is used to build the antecedent line entries. 
*  The antecedent lines are then placed in the table 
*  G_ANTELINES using the ANTE data structure. 
*  The fields of ANTE are as follows: 
*    PRLINE - an integer which indicates which line of the 
*	      proof matrix the antecedent line occurs in. 
*    SHOWLV - indicates the level of nesting of SHOW statements 
*    ANTELINK - will point to another ANTE structure if a 
*		particular antecedent line has occured more than 
*		once.  Antecedent lines are stacked in a FIFO 
*		ordering. 
-SPACE
      DATA('ANTE(PRLINE,SHOWLV,ANTELINK)')
-SPACE 2
*FRAME
*  The TEMP data structure is used to hold information about 
*  the antecedent line templates.  They are again used as entries 
*  in the G_ANTELINES table.  The fields of TEMP are as follows: 
*    TOKEN - replacing the @ with TOKEN will result in the original
*	     line. 
*    TEMPLINK - TEMP structures are stacked in the same manner that
*		ANTE structures are stacked, TEMPLINK a pointer to 
*		another TEMP stucture (or NULL). 
-SPACE
      DATA('TEMP(TOKEN,TEMPLINK)')
-SPACE 2
*FRAME
*  The GOAL data type will be used to stack goals and intermediate 
*  subgoals.  The GOAL records will be maintained in a FIFO stack 
*  which the global G_GOALS will point to.  The fields of GOAL are as
*  follows: 
*    PRLIN - the line in the proof matrix where this line occured 
*    LEVEL - the level of nesting of GOAL ( also the depth of the 
*	     stack. 
*    GOALLINK - a link to the next record in the stack 
*    WHAT - the goal 
*    GFVTPLATE - table of fv templates for this goal 
*    FINDEIPT, FINDEIPT2 used by FINDEI 
*    IDENTRAN - the table of identity transitive groups
*    FINDIDPT,FINDIDPT2 - used by FINDID
*    IDENTVARS - stack of variables for which there are identity
*		 transformations
-SPACE
      DATA('GOAL(PRLIN,LEVEL,WHAT,GFVTPLATE,FINDEIPT,FINDEIPT2'
+	 ',IDENTRAN,FINDIDPT,FINDIDPT2,IDENTVARS,GOALLINK)')
-SPACE
*FRAME
*  DESP is used to hold those lines that have already been
*  tried in desperation subproofs that have been aborted
-SPACE
      DATA('DESP(WHAT,NEXTLINK)')
-SPACE
*FRAME
*  SIMPLE is used to hold all of the antecedent lines 
*  which have a given major connective.  SIMPLE 
*  is used to maintain a cirular doubly-linked list. 
*  There is one circular list for each of the major connectives. 
*  Each circle has a header node with the first two fields 
*  set to null. 
*    LINE - the antecendent line corresponding to this 
*	    entry. 
*    PLINE - the number of the line in the proof matrix 
*    FLINK - the forward link 
*    BLINK - the backward link 
-SPACE
      DATA('SIMPLE(LINE,PLINE,FLINK,BLINK)')
-SPACE
*FRAME
*  STACK is simply used by a variety of things to build 
*  a simple stack with one value. 
*    V - the entry on the stack 
*    N - the link to the next entry 
      DATA('STACK(V,N)')
*FRAME
*  EQUIVCLASS is used to hold information about the equivalence
*  classes which have been discovered. It is used to hold membership
*  information for individual variables.
*    ECREP - the replacement
*    ECLAST - the last line that this particular replacement was
*	      done on
*    ECREASON - list of justifications for this replacement.
     DATA('EQUIVCLASS(ECREP,ECLAST,ECREASON)')
*FRAME
*  CRM -the following data types used be further down the program.
*  They were moved during the porting of the mts version to unix.
      DATA('CMD_FORMAT(COMMAND,PARAMETERS)')
      DATA('DISPLAY_FORMAT(VARIABLE,FORMAT)')
      DATA('SET_FORMAT(NAME,L_BOND,U_BOND)')
-TITLE THINKER - initialization functions
-STITL Function: INITER()
      pDEFINE('INITER()',,'NULL','NULL',70)
+							:(ENDINITER)
*FRAME
* Function: INITER 
* Date: Sun Jul 25 17:14:12 1982 
* Last modification: June, 1983
* Purpose: 
*  INITER establishes all of the global variables. 
*  It must be called before each proof is started. 
* 
* Calling sequence: 
*	INITER() 
* Parameters: None 
* On return: 
*	RETURN - Normal termination.
* 
* Local variables: None 
* 
-SPACE 2
*FRAME
*  G_ANTELINES is a table which is used to hold antecedent lines 
*  and templates for antecedent lines.	The data records which 
*  will occur in G_ANTELINES are TEMP and ANTE. 
-SPACE 2
*FRAME
*  G_PRMAT is an array which will be used to hold the lines of the 
*  proof as they are generated.  The first column of the array 
*  will hold the line, the second column of the array will hold 
*  the reason for the line 
*  The fifth entry is the ancestor list for this line and 
*  the sixth entry is the prohibit list for this line 
*  CHANGE:   June 20, 1986   CNH
*  A seventh field was added to the matrix
*  this field will denote whether the line is 'required'
*  or not during the CONDENSE phase.  A 'T' is placed
*  if the line is required and it is left blank otherwise
-SPACE 2
INITER
      G_PRMAT = ARRAY('-' P_MAXPREM ':' P_PROOFSIZE ',7')
      G_ANTELINES = TABLE(P_ANTES)
      G_PROHIBLIST = TABLE(50)
* CHANGE June 1997 counting statements disabled for efficiency.
*      G_STCNT = &STCOUNT
      G_STTIME = TIME()
*FRAME
*  HOLD  is used to hold a subproof while exploring other
*  subproofs for possible completion in the parallelism
*  code.	
*  ADDED   Sept 5, 1986   CNH
*  CHANGE June 1997 parallelism disabled.
*      HOLD = ARRAY('500,2')
*FRAME
*  initialize L_SIZE  (size of the long proof) to 0
      L_SIZE = 0
-SPACE 2
*FRAME
*  G_CLIST, G_VLIST, and G_FVLIST are occurance tables for 
*  all pconstants in G_ANTELINES, all pvariables that have ever 
*  been used and all free pvariables in G_ANTELINES repectively 
      G_CLIST = TABLE(25)
      G_VLIST = TABLE(25)
      G_FVLIST = TABLE(25)
      G_PREMVLIST = TABLE(25)
      G_PREMFVLIST = TABLE(25)
* CHANGES July 16, 1986  CNH  PCONS
* add a table, an array, a datatype and a global variable
* for use entering premise consequences
      ISPC = TABLE(P_ANTES)
      PREMCON = ARRAY(P_ANTES)
      DATA('PC(PCENTRY,PCCOM)')
      PC_CUR = 0
*FRAME
*  G_VARIDENT is a table containing all of the variables and contants
*  that are antecedent in the proof.  It has hanging from each of its
*  nodes a list of pointers to the proof lines in which each of the 
*  terms has occured.
      G_VARIDENT = TABLE(P_VARSIZE + 40)
*FRAME
*  Initialize all of the circularly linked lists so that the 
*  head of each points to itself 
      G_NSTACK = SIMPLE()
      BLINK(G_NSTACK) = G_NSTACK
      FLINK(G_NSTACK) = G_NSTACK
-SPACE
      G_BCSTACK = SIMPLE()
      BLINK(G_BCSTACK) = G_BCSTACK
      FLINK(G_BCSTACK) = G_BCSTACK
-SPACE
      G_ANDSTACK = SIMPLE()
      BLINK(G_ANDSTACK) = G_ANDSTACK
      FLINK(G_ANDSTACK) = G_ANDSTACK
-SPACE
      G_ORSTACK = SIMPLE()
      BLINK(G_ORSTACK) = G_ORSTACK
      FLINK(G_ORSTACK) = G_ORSTACK
-SPACE
      G_IMPLSTACK = SIMPLE()
      BLINK(G_IMPLSTACK) = G_IMPLSTACK
      FLINK(G_IMPLSTACK) = G_IMPLSTACK
-SPACE
      G_EQSTACK = SIMPLE()
      BLINK(G_EQSTACK) = G_EQSTACK
      FLINK(G_EQSTACK) = G_EQSTACK
-SPACE
      G_UQSTACK = SIMPLE()
      BLINK(G_UQSTACK) = G_UQSTACK
      FLINK(G_UQSTACK) = G_UQSTACK
-SPACE
      G_IDSTACK = SIMPLE()
      BLINK(G_IDSTACK) = G_IDSTACK
      FLINK(G_IDSTACK) = G_IDSTACK
*  Set the goal stack pointer to null
      G_GOALS =
*FRAME
*  Clean house so we get nice neat DUMP(2)s 
      COLLECT(1)
-SPACE 2
*FRAME
*  G_CURLEVEL contains the current nesting level of goals. 
      G_CURLEVEL = 0
*FRAME
*  G_CURLINE contains the current line in the proof matrix 
      G_CURLINE  = 0
-SPACE 2
*FRAME
*  G_LASTFREEVAR used by getnewvar to get a fresh variable 
      G_LASTFREEVAR = 0 				   :(RETURN)
ENDINITER
-TITLE THINKER - antecedent line manipulations
-STITL Function: ADDANTE(AA_PRLN,AA_SHLV,AA_LINE,AA_PF,AA_VMARK)
*FRAME
* Function: ADDANTE 
* Date: Sun Jul 25 17:16:08 1982 
* Last modification: June, 1983
* Purpose: 
*  ADDANTE will add an antecedent line to the table of 
*  antecedent lines.  Then use the G_PF pattern to stuff away 
*  all the pvariables, pconstants and free pvariables in the 
*  formula. 
*  Then add free variable templates for the formula. 
*  After this is done, check for a main binary connective 
*  or a global quantifier.  In either of these cases add templates 
*  for the formula. 
*  Then, add the formula to the appropriate main connective ring. 
* 
* Calling sequence: 
*	ADDANTE(AA_PRLN,AA_SHLV,AA_LINE,AA_PF,AA_VMARK) 
* Parameters: 
*	AA_PRLN - is the line in the proof matrix for this line 
*	AA_SHLV - is the level of show lines which have occured 
*		  (recursion level)
*	AA_LINE - Formula being added.
*	AA_PF - the name of the pattern that is used to put 
*		away information about pterms 
*	AA_VMARK - is non-null if locations of free variables in
*		   this formula are to be recored
* On return: 
*	RETURN - Normal termination
* 
* Local variables: 
*	AA_Z - The actual antecedent line.
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
* 
      pDEFINE('ADDANTE(AA_PRLN,AA_SHLV,AA_LINE,AA_PF,AA_VMARK)'
+		       'AA_Z,G_ROP,G_OP,G_LOP',,
+		       'INTEGER,INTEGER,STRING,NAME,INTEGER|NULL','NULL')
+						       :(ENDADDANTE)
ADDANTE
+     AA_Z  =  G_ANTELINES<AA_LINE>
      G_ANTELINES<AA_LINE>  =  ANTE(AA_PRLN,AA_SHLV,AA_Z)
      AA_LINE $AA_PF					:F(CONCHECK)
      ADDFVTPLATE(AA_LINE,.ADDTPLATE,AA_VMARK)
      AA_LINE G_BSPLIT				       :F(AA_CHQADD)
      ADDTPLATE('(' G_LOP G_OP '@)',G_ROP)
      ADDTPLATE('(@' G_OP G_ROP ')',G_LOP)		:(AA_ADDSTK)
AA_CHQADD
+     AA_LINE G_QSPLIT				       :F(AA_ADDSTK)
      ADDTPLATE(MAKEQTPLATE(AA_LINE),G_VAR)
* only add the line to the rings if the line is not antecedent
AA_ADDSTK
+     DIFFER(AA_Z)					  :S(RETURN)
      AA_LINE G_SPLIT					  :S(AA_NOTIDENT)
      AA_LINE G_IDSPLIT 			  :S(AA_ADDID)F(RETURN)
AA_NOTIDENT
      LEQ(G_OP,'~')					 :S(AA_ADDN)
      LEQ(G_OP,'=')					:S(AA_ADDBC)
      LEQ(G_OP,'&')				       :S(AA_ADDAND)
      LEQ(G_OP,'+')					:S(AA_ADDOR)
      LEQ(G_OP,'>')				      :S(AA_ADDIMPL)
      LEQ(G_OP,'A')					:S(AA_ADDUQ)
      LEQ(G_OP,'E')					:S(AA_ADDEQ)
AA_ADDN
+     CSTACK(G_NSTACK,AA_LINE,AA_PRLN)			   :(RETURN)
AA_ADDBC
+     CSTACK(G_BCSTACK,AA_LINE,AA_PRLN) 		   :(RETURN)
AA_ADDAND
+     CSTACK(G_ANDSTACK,AA_LINE,AA_PRLN)		   :(RETURN)
AA_ADDOR
+     CSTACK(G_ORSTACK,AA_LINE,AA_PRLN) 		   :(RETURN)
AA_ADDIMPL
+     CSTACK(G_IMPLSTACK,AA_LINE,AA_PRLN)		   :(RETURN)
AA_ADDUQ
+     CSTACK(G_UQSTACK,AA_LINE,AA_PRLN) 		   :(RETURN)
AA_ADDEQ
+     CSTACK(G_EQSTACK,AA_LINE,AA_PRLN) 		   :(RETURN)
AA_ADDID
+     CSTACK(G_IDSTACK,AA_LINE,AA_PRLN) 		   :(RETURN)
ENDADDANTE
-STITL Function: DELANTE(DA_LINE)
*FRAME
* Function: DELANTE 
* Date: Sun Jul 25 17:17:57 1982 
* Last modification: June, 1983
* Purpose: 
*  DELANTE will remove the latest occurance of an antecedent line 
*  from the table.  Delete the templates for this line, and also 
*  delete the entry in the ring-list.  Delete the free pvariable 
*  and pconstant stuff. 
* 
* Calling sequence: 
*	DELANTE(DA_LINE) 
* Parameters: 
*	DA_LINE - Formula to be deleted from the antecedent lines.
* On return: 
*	RETURN - Normat termination.
* 
* Local variables: 
*	DA_Z - The pertinent entry in G_ANTELINES.
*	DA_ZZ - The LINK field of G_ANTELINES.
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
* 
      pDEFINE('DELANTE(DA_LINE)DA_Z,DA_ZZ,G_LOP,G_OP,G_ROP',,
+	      'STRING','NULL') 
+						       :(ENDDELANTE)
DELANTE
+     DA_Z = G_ANTELINES<DA_LINE>
      DA_ZZ = ANTELINK(DA_Z)
      G_ANTELINES<DA_LINE> = DA_ZZ
      DA_LINE G_DF					:F(CONCHECK)
      DELFVTPLATE(DA_LINE)
      DA_LINE G_BSPLIT					  :F(DA_CHQ)
      DELTPLATE('(' G_LOP G_OP '@)')
      DELTPLATE('(@' G_OP G_ROP ')')			   :(DA_STK)
DA_CHQ
+     DA_LINE G_QSPLIT					  :F(DA_STK)
      DELTPLATE(MAKEQTPLATE(DA_LINE))
* only delete the ring connective if this is the last line
* (matches up with ADDSTK)
DA_STK
+     DIFFER(DA_ZZ)					  :S(RETURN)
      DA_LINE G_SPLIT					  :S(DA_NOTIDENT)
      DA_LINE G_IDSPLIT 				  :S(DA_ID)F(RETURN)
DA_NOTIDENT
      LEQ(G_OP,'~')					    :S(DA_N)
      LEQ(G_OP,'&')					  :S(DA_AND)
      LEQ(G_OP,'+')					   :S(DA_OR)
      LEQ(G_OP,'=')					   :S(DA_BC)
      LEQ(G_OP,'>')					 :S(DA_IMPL)
      LEQ(G_OP,'A')					   :S(DA_UQ)
      LEQ(G_OP,'E')					   :S(DA_EQ)
DA_N
+     DELCSTACK(G_NSTACK)				   :(RETURN)
DA_AND
+     DELCSTACK(G_ANDSTACK)				   :(RETURN)
DA_OR
+     DELCSTACK(G_ORSTACK)				   :(RETURN)
DA_BC
+     DELCSTACK(G_BCSTACK)				   :(RETURN)
DA_IMPL
+     DELCSTACK(G_IMPLSTACK)				   :(RETURN)
DA_UQ
+     DELCSTACK(G_UQSTACK)				   :(RETURN)
DA_EQ
+     DELCSTACK(G_EQSTACK)				   :(RETURN)
DA_ID
+     DELCSTACK(G_IDSTACK)				   :(RETURN)
ENDDELANTE
-STITL Function: FINDANTE(FA_LINE)
*FRAME
* Function: FINDANTE 
* Date: Sun Jul 25 17:18:44 1982 
* Last modification: June, 1983
* Purpose: 
*  FINDANTE is a function which checks to see if a given line 
*  has occured in the table.  If it has it returns the record, 
*  if it hasn't it returns failing. 
* 
* Calling sequence: 
*	result = FINDANTE(FA_LINE) 
* Parameters: 
*	FA_LINE - Line for wich we are looking in the table.
* On return: 
*	RETURN - Returns as a value the record for the antecedent
*		 line in the table.
*	FRETURN - The line has not occured.
* 
      pDEFINE('FINDANTE(FA_LINE)',,'STRING','ANTE')
+						      :(ENDFINDANTE)
FINDANTE
      FINDANTE = G_ANTELINES<FA_LINE>
* CHANGES July 16, 1986  CNH PCONS
* put an additional check for the line to be a premise
* consequence, in which case add the appropriate lines
* then recheck the array for the line
* CHANGE June 1997 If there are no antecedent lines then fail.
      IDENT(FINDANTE)					:F(RETURN)
      ENTERCONSEQ(ISPC<FA_LINE>)
      FINDANTE = G_ANTELINES<FA_LINE>
      IDENT(FINDANTE)				:F(RETURN)S(FRETURN)
ENDFINDANTE
-TITLE THINKER - template manipulation
-STITL Function: ADDTPLATE(AT_LINE,AT_REPLACE)
*FRAME
* Function: ADDTPLATE 
* Date: Sun Jul 25 17:16:52 1982 
* Last modification: June, 1983
* Purpose: 
*  ADDTPLATE will add a template to the table of antecedent lines. 
* 
* Calling sequence: 
*	ADDTPLATE(AT_LINE,AT_REPLACE) 
* Parameters: 
*	AT_LINE - is the line to add, with the @ - signs in place 
*	AT_REPLACE - is the substitution for the @ - sign 
* On return: 
*	RETURN - Normal termination.
* 
* Description: 
* 
*  AT_LINE - is the line to add, with the @ - signs in place 
*  AT_REPLACE - is the substitution for the @ - sign 
      pDEFINE('ADDTPLATE(AT_LINE,AT_REPLACE)',,
+	      'STRING,STRING','NULL')
+						     :(ENDADDTPLATE)
ADDTPLATE
+     G_ANTELINES<AT_LINE> = TEMP(AT_REPLACE,G_ANTELINES<AT_LINE>)
+							   :(RETURN)
ENDADDTPLATE
-STITL Function: DELTPLATE(DT_LINE)
*FRAME
* Function: DELTPLATE 
* Date: Sun Jul 25 17:18:32 1982 
* Last modification: June, 1983
* Purpose: 
*  DELTPLATE will remove the latest occurace of a template from the
*  table. 
* 
* Calling sequence: 
*	DELTPLATE(DT_LINE) 
* Parameters: 
*	DT_LINE - the template to be removed ( with @ - signs) 
* On return: 
*	RETURN - Normal termination.
* 
* Local variables: None 
* 
      pDEFINE('DELTPLATE(DT_LINE)',,'STRING','NULL')
+						     :(ENDDELTPLATE)
DELTPLATE
+     G_ANTELINES<DT_LINE> = TEMPLINK(G_ANTELINES<DT_LINE>)
+							   :(RETURN)
ENDDELTPLATE
-STITL Function: FINDTEMP(FT_LINE)
*FRAME
* Function: FINDTEMP 
* Date: Sun Jul 25 17:19:03 1982 
* Last modification: June, 1983
* Purpose: 
*  FINDTEMP is a function to find if a template has occured 
*  in the antecedent line table.  It fails if a suitable template 
*  does not exist, if it succeeds it returns the antecedent lines 
*  it would match as a string. 
* 
* Calling sequence: 
*	result = FINDTEMP(FT_LINE) 
* Parameters: 
*	FT_LINE - the template to search for ( with @ - signs ) 
* On return: 
*	RETURN - A string representing the antecedent line that
*		 matches the template.
*	FRETURN - No similar template has yet occrued
* 
* Local variables: 
*	FT_Z - Pertinent entry in G_ANTELINES. 
* 
      pDEFINE('FINDTEMP(FT_LINE)FT_Z',,'STRING','STRING')
+						      :(ENDFINDTEMP)
FINDTEMP
+     FT_Z = G_ANTELINES<FT_LINE> 
      IDENT(FT_Z)					 :S(FRETURN)
FT_TEMPREP
+     FT_LINE '@' = TOKEN(FT_Z) 		      :S(FT_TEMPREP)
      FINDTEMP	 = FT_LINE				   :(RETURN)
ENDFINDTEMP
-STITL Function: ADDFVTPLATE(AF_FORM,AF_ADDFUNC,AF_VMARK)
*FRAME
* Function: ADDFVTPLATE 
* Date: Sun Jul 25 17:17:16 1982 
* Last modification: May, 1984
* Purpose: 
*  ADDFVTPLATE will add templates for all the free variables 
*  and constants in the formula AF_FORM. 
*  Now we also remember all of the lines which each free variable and 
*  constant occurs in.
* 
* Calling sequence: 
*	ADDFVTPLATE(AF_FORM,AF_ADDFUNC,AF_VMARK) 
* Parameters: 
*	AF_FORM - the formula in which free variables are to be
*		  replaced by @ - signs.
*	AF_ADDFUNC - the function to use to add the template
*	AF_VMARK - non-null if the free variables in this function are
*		   to be remembered in G_IDENTVARS
* On return: 
*	RETURN - Normal termination.
* 
* Local variables: 
*	G_FVL - Free varible list (TABLE)
*	AF_FVLA - Array form of G_FVL
*	AF_TEMP - Template returned form MAKEQTPLATE
*	AF_FREEVAR - I'th element of AF_FVLA
*	AF_I - Index to AF_FVLA
*
      pDEFINE('ADDFVTPLATE(AF_FORM,AF_ADDFUNC,AF_VMARK)'
+	      'G_FVL,AF_FVLA,AF_TEMP,AF_FREEVAR,AF_I',,
+	      'STRING,NAME|STRING,INTEGER|NULL','NULL')
+						   :(ENDADDFVTPLATE)
ADDFVTPLATE
+     G_FVL = TABLE(11)
*      OUTPUT = "ADDFVTPLATE(" AF_FORM "," AF_ADDFUNC "," AF_VMARK ")"
      AF_FORM (G_FF RPOS(0))				:F(CONCHECK)
      AF_FVLA = CONVERT(G_FVL,'ARRAY')			  :F(RETURN)
      IDENT(AF_FVLA)					  :S(RETURN)
      AF_I = 1
AF_LOOP
+     AF_FREEVAR = AF_FVLA<AF_I,1>			  :F(RETURN)
      IDENT(AF_VMARK)					  :S(AF_NOTANTE)
*  CHANGES July 22, 1986   CNH change G_curline to aa_prln
      G_VARIDENT<AF_FREEVAR> = STACK(AA_PRLN,G_VARIDENT<AF_FREEVAR>)
AF_NOTANTE
      AF_TEMP = MAKEQTPLATE('(A' AF_FREEVAR ')' AF_FORM)
      AF_TEMP '(A@)' =
      APPLY(AF_ADDFUNC,AF_TEMP,AF_FREEVAR)
      AF_I = AF_I + 1					  :(AF_LOOP)
ENDADDFVTPLATE
-STITL Function: DELFVTPLATE(DF_FORM)
*FRAME
* Function: DELFVTPLATE 
* Date: Sun Jul 25 17:17:33 1982 
* Last modification: May, 1984
* Purpose: 
*  DELFVTPLATE is a function which will delete templates 
*  for free variables and constants for the formula DF_FORM 
*  Delete the information in G_VARIDENT added by ADDFVTPLATE
* 
* Calling sequence: 
*	DELFVTPLATE(DF_FORM) 
* Parameters: 
*	DF_FORM - The formula for which the free variable 
*		  templates are to be deleted.
* On return: 
*	RETURN - Normal termination.
* 
* Local variables: 
*	G_FVL - Free variable list (TABLE)
*	DF_FVLA - Array form of G_FVL
*	DF_TEMP - Template returned from MAKEQTPLATE
*	DF_FREEVAR - I'th element of DF_FVLA
*	DF_I - Index to DF_FVLA
* 
      pDEFINE('DELFVTPLATE(DF_FORM)'
+	      'G_FVL,DF_FVLA,DF_TEMP,DF_FREEVAR,DF_I',,
+	       'STRING','NULL')
+						   :(ENDDELFVTPLATE)
DELFVTPLATE
+     G_FVL = TABLE(11)
      DF_FORM G_FF
      DF_FVLA = CONVERT(G_FVL,'ARRAY')			  :F(RETURN)
      IDENT(DF_FVLA)					  :S(RETURN)
      DF_I = 1
DF_LOOP
+     DF_FREEVAR = DF_FVLA<DF_I,1>			  :F(RETURN)
      G_VARIDENT<DF_FREEVAR> = N(G_VARIDENT<DF_FREEVAR>)
      DF_TEMP = MAKEQTPLATE('(A' DF_FREEVAR ')' DF_FORM)
      DF_TEMP '(A@)' =
      DELTPLATE(DF_TEMP,DF_FREEVAR)
      DF_I = DF_I + 1					  :(DF_LOOP)
ENDDELFVTPLATE
-STITL Function: MAKEQTPLATE(MQ_FORM)
*FRAME
* Function: MAKEQTPLATE 
* Date: Sun Jul 25 17:14:55 1982 
* Last modification: May, 1984
* Purpose: 
*  MAKEQTPLATE is a function which makes templates for quantified 
*  formulae 
* 
* Calling sequence: 
*	result = MAKEQTPLATE(MQ_FORM) 
* Parameters: 
*	MQ_FORM - the formula for which the template is to be made.
* On return: 
*	RETURN - Quantified template: ie. (Ex)Fx => (E@)F@
* 
* Local variables: 
*	G_MV - Copy of G_VAR
*	MQ_ROPL - Lenght of G_ROP
*	MQ_P - Value field of stack G_POSSTACK.
*	G_POSSTACK - Stack (?)
*	MQ_I - SUBSTRING index into the string G_ROP
*	G_VAR - Quantifier variable from G_SPLIT
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
* 
* Description: 
* 
      pDEFINE('MAKEQTPLATE(MQ_FORM)'
+	      'G_MV,MQ_ROPL,MQ_P,G_POSSTACK,MQ_I,G_VAR,'
+	      'G_ROP,G_LOP,G_OP',,'STRING','STRING')
+						   :(ENDMAKEQTPLATE)
MAKEQTPLATE
+     MQ_FORM G_QSPLIT					:F(CONCHECK)
      MQ_ROPL = SIZE(G_ROP) 
      G_MV = G_VAR
      G_ROP (G_CF RPOS(0))				:F(CONCHECK)
      MQ_I = 0
MQ_LOOP
      IDENT(G_POSSTACK) 				 :S(MQ_MAKE)
      MQ_P = V(G_POSSTACK)
      EQ(MQ_P,2)					:S(MQ_AT0)
      G_ROP = SUBSTR(G_ROP,1,MQ_P - 2) '@' 
+	      SUBSTR(G_ROP,MQ_P + 1,MQ_ROPL - MQ_I - MQ_P) :(MQ_CONT)
MQ_AT0
+     G_ROP = '@' SUBSTR(G_ROP,MQ_P + 1,MQ_ROPL - MQ_I - MQ_P)
MQ_CONT
      MQ_I = MQ_I + 1
      G_POSSTACK = N(G_POSSTACK)			  :(MQ_LOOP)
MQ_MAKE
+     MAKEQTPLATE = '(' G_OP '@)' G_ROP
							   :(RETURN)
ENDMAKEQTPLATE
-STITL Function: ADDFVGOALTPLATE(AV_FORM,AV_FV)
*FRAME
* Function: ADDFVGOALTPLATE 
* Date: Sun Jul 25 17:22:15 1982 
* Last modification: June, 1983
* Purpose: 
*  In the template table the current goal add an entry at 
*  location form with value AV_FV 
* 
* Calling sequence: 
*	ADDFVGOALTPLATE(AV_FORM,AV_FV) 
* Parameters: 
*	AV_FORM - Template to be added.
*	AV_FV - Replacement variable for @.
* On return: 
*	RETURN - Normal termination
* 
* Local variables: None 
* 
* Description: 
* 
      pDEFINE('ADDFVGOALTPLATE(AV_FORM,AV_FV)',,
+	     'STRING,STRING','NULL')
+					       :(ENDADDFVGOALTPLATE)
ADDFVGOALTPLATE
+     GFVTPLATE(G_GOALS)<AV_FORM> = AV_FV		   :(RETURN)
ENDADDFVGOALTPLATE
-TITLE THINKER - goal stack manipulation
-STITL Function: ADDGOAL(AG_LINE)
*FRAME
* Function: ADDGOAL 
* Date: Sun Jul 25 17:19:42 1982 
* Last modification: June, 1983
* Purpose: 
*  ADDGOAL - first see if the goal to be added is already 
*  active.  If it is there is trouble, and we return failing. 
*  If not increment G_CURLEVEL, adds the line to the proof matrix 
*  as a SHOW line, and then push the goal onto the goal stack 
* 
* Calling sequence: 
*	ADDGOAL(AG_LINE) 
* Parameters: 
*	AG_LINE - The line to be added to the goal stack.
*	AG_ANS - the ancestor of the goal 
* On return: 
*	RETURN - Normal termination.
*	FRETURN - AG_LINE is already a goal.
* 
* Local variables: 
*	AG_EIPT - Result from FINDEIPT
*	AG_EIPT2 - Result from FINDEIPT2
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
*	G_VAR - Quantifier variable given by G_SPLIT
* 
      pDEFINE('ADDGOAL(AG_LINE)'
+      'AG_EIPT,AG_EIPT2,AG_IDTRAN,AG_IDPT,AG_IDPT2,G_LOP,G_ROP,G_VAR,G_OP',,
+	      'STRING,INTEGER|NULL','NULL')
+						       :(ENDADDGOAL)
ADDGOAL
+     TESTGOAL(AG_LINE) 				 :F(FRETURN)
      G_CURLEVEL = G_CURLEVEL + 1
*  changes Aug 1, 1986	CNH  addproof to addprii
      ADDPRII(AG_LINE,'*')
      DIFFER(G_GOALS)					:F(AG_FIRST)
      AG_EIPT = FINDEIPT(G_GOALS)
      AG_EIPT2 =  FINDEIPT2(G_GOALS)
      AG_IDTRAN = COPY(IDENTRAN(G_GOALS))
      AG_IDPT = FINDIDPT(G_GOALS)
      AG_IDPT2 = FINDIDPT2(G_GOALS)
AG_DOIT
+     G_GOALS = GOAL(G_CURLINE,G_CURLEVEL,AG_LINE,TABLE(5),
+		     AG_EIPT,AG_EIPT2,AG_IDTRAN,AG_IDPT,AG_IDPT2,,G_GOALS)
      ADDFVTPLATE(AG_LINE,.ADDFVGOALTPLATE)		   :(RETURN)
AG_FIRST
+     AG_EIPT = BLINK(G_EQSTACK)
      AG_EIPT2 = G_EQSTACK
      AG_IDPT = BLINK(G_IDSTACK)
      AG_IDPT2 = G_IDSTACK
      AG_IDTRAN = TABLE(11)				   :(AG_DOIT)
ENDADDGOAL
-STITL Function: GETGOAL()
*FRAME
* Function: GETGOAL 
* Date: Sun Jul 25 17:20:23 1982 
* Last modification: June, 1983
* Purpose: 
*  GETGOAL simply returns the top goal on the stack, it does not 
*  change the stack. 
* 
* Calling sequence: 
*	result = GETGOAL() 
* Parameters: None 
* On return: 
*	RETURN - Returns a string containing the current goal.
*	FRETURN - No current goal (?)
* 
* Local variables: None 
* 
* Description: 
      pDEFINE('GETGOAL()',,,'STRING')		       :(ENDGETGOAL)
GETGOAL
+     GETGOAL = WHAT(G_GOALS)
      IDENT(GETGOAL)				:S(FRETURN)F(RETURN)
ENDGETGOAL
-STITL Function: TESTGOAL(TG_LINE)
*FRAME
* Function: TESTGOAL 
* Date: Sun Jul 25 17:20:01 1982 
* Last modification: June, 1983
* Purpose: 
*  TESTGOAL determines whether the TG_LINE passed to it has 
*  is a currently active goal. 
* 
* Calling sequence: 
*	TESTGOAL(TG_LINE) 
* Parameters: 
*	TG_LINE - The line to test for.
* On return: 
*	RETURN - Goal has never occured.
*	FRETURN - Goal has occured before.
* 
* Local variables: 
*	TG_T - point which starts at top of stack and works down.
* 
      pDEFINE('TESTGOAL(TG_LINE)TG_T',,'STRING','NULL')
+						      :(ENDTESTGOAL)
TESTGOAL
      IDENT(G_GOALS)				      :S(RETURN)
*   I also FailRETURN if TG_LINE is the negation of an identity since
*   has no useful purpose of being set as a goal.
*   CHANGED Jan 20, 1987
      TG_LINE POS(0) G_UNOP . TG_OP BAL . TG_RIG RPOS(0)  :F(TG_OK)
   TG_RIG POS(0) BREAK(G_IDOP) . TG_LOP G_IDOP . TG_OP REM . TG_ROP :F(TG_OK)
      LEQ(TG_LOP,TG_ROP)			      :S(FRETURN)
TG_OK
+     TG_T = G_GOALS
TG_LOOP
+     IDENT(TG_T)					  :S(RETURN)
      LEQ(WHAT(TG_T),TG_LINE)				 :S(FRETURN)
      TG_T = GOALLINK(TG_T)				  :(TG_LOOP)
ENDTESTGOAL
-STITL Function: GETUQGOAL()
*FRAME
* Function: GETUQGOAL 
* Date: Sun Jul 25 17:25:55 1982 
* Last modification: June, 1983
* Purpose: 
* GET THE BOUND VARIBLE FROM THE MOST RECENTLY UNIVERSALLY 
* QUANTIFIED SHOW LINE, OR FAIL IF THERE ARE NONE. 
* 
* Calling sequence: 
*	result = GETUQGOAL() 
* Parameters: None 
* On return: 
*	RETURN - Returns a string containing the bounded variable
*		 from the most recent univesally quantified show
*		 line.
*	FRETURN - No variable found.
* 
* Local variables: 
*	GU_GOALPT - Pointer for the goal stack
*	GU_GL - Goal in stack to be split
* 
      pDEFINE('GETUQGOAL()GU_GOALPT,GU_GL',,,'STRING')
+						     :(ENDGETUQGOAL)
GETUQGOAL
+     GU_GOALPT = G_GOALS
GU_LOOP
+     IDENT(GU_GOALPT)					 :S(FRETURN)
      GU_GL = WHAT(GU_GOALPT)
      GU_GL G_QSPLIT					 :F(GU_NEXT)
      LEQ(G_OP,'A')					 :F(GU_NEXT)
      GETUQGOAL = G_VAR 				   :(RETURN)
GU_NEXT
+     GU_GOALPT = GOALLINK(GU_GOALPT)			  :(GU_LOOP)
ENDGETUQGOAL
-TITLE THINKER - proof matrix manipulations
-STITL Function: ADDPRII(AP_LINE,AP_REASON,P_NUM)
*FRAME
* Function: ADDPRII 
* Date: Sun Jul 25 17:26:12 1982 
* Last modification: June, 1983
* Purpose: 
*  ADDPRII will add a line to the proof matrix.  If the line 
*  is already there at the same level then don't add it again. 
*  The first thing that ADDPRII checks is to see if any of 
*  the justifications specify negative lines, if they do then 
*  a recursive addproof gets done on the premise 
* 
* Calling sequence: 
*	ADDPRII(AP_LINE,AP_REASON,P_NUM) 
* Parameters: 
*	AP_LINE - the line to add 
*	AP_REASON - the justification for the line 
*  changes Aug 12, 1986  CNH  add a parameter of the premise number
*	P_NUM if line is a premise then this is the prem number
* On return: 
*	RETURN - Normal termination.
* 
* Local variables: 
*	AP_Z - Result of FINDANTE on line to be added.
* 
*  CNH	3rd parameter to ADDPRII removed
      pDEFINE('ADDPRII(AP_LINE,AP_REASON,P_NUM)ANC,'
+	      'AP_Z,AP_N,AP_ST,G_ROP,G_OP,G_LOP,G_VAR,AP_IDENT,AP_NORM',,
+	     'STRING,STRING,INTEGER|NULL,INTEGER|NULL,INTEGER|NULL',
+	      'INTEGER')
+						      :(ENDADDPRII)
ADDPRII
* * *
* Normalization consistency check
* * *
*    OUTPUT = 'IN ADDPRII  LINE: ' AP_LINE
      AP_NORM = AP_LINE
AP_NORMLOOP
+     AP_NORM (G_BT '-' G_BT) . AP_IDENT = NULL 	:F(AP_NONORM)
      AP_IDENT G_IDSPLIT
      LGT(G_LOP,G_ROP)					:F(AP_NORMLOOP)
      OUTPUT = '***** INCONSISTENCY check in ADDPRII'
      OUTPUT = '***** non-normalized identity in proof line'
      OUTPUT = '***** Line: ' AP_LINE
      OUTPUT = '***** Reason: ' AP_REASON		:(CONCHECK)
AP_NONORM
+     AP_REASON BREAK('-') $ AP_ST '-' SPAN('0123456789') $ AP_N =
+	 AP_ST ADDPRII(G_PREM<AP_N>,'PREM', 0 - AP_N)  :S(ADDPRII)
*  CNH	3rd parameter to ADDPRII removed
* CNH PCONS JULY 16, 1986
* change call to findante to prevent infinite loop
      AP_Z = G_ANTELINES<AP_LINE>
      IDENT(AP_Z)					:S(AP_ADDIT)
*     AP_Z = FINDANTE(AP_LINE)				:F(AP_ADDIT)
      ADDPRII = PRLINE(AP_Z)
      EQ(SHOWLV(AP_Z),G_CURLEVEL)		       :S(AP_RETURN)
AP_ADDIT
+     G_CURLINE = GT(P_PROOFSIZE,G_CURLINE) G_CURLINE + 1 :S(AP_NOTOOBIG)
      OUTPUT = 'Too many lines in the proof.'             :(CONCHECK)
AP_NOTOOBIG
+     ADDPRII = G_CURLINE
      G_PRMAT<G_CURLINE,1> = AP_LINE
      G_PRMAT<G_CURLINE,2> = AP_REASON
      G_PRMAT<G_CURLINE,3> = G_CURLEVEL
*  CHANGES June 30, 1986   CNH
*  I took out the lines which refered to ANS1,ANS2 and ANS3
* check to see if this is a universally quanitfied line
* changes Aug 8, 1986  CNH
* add ancestor list
*  changes Aug 12, 1986  CNH  if it is a premise then add P_NUM
      AP_REASON 'PREM'				      :F(AP_ANC)
*     OUTPUT = 'ADDING PREMISE NUMBER ' G_CURLINE ',' P_NUM
      G_PRMAT<G_CURLINE,5> = STACK(P_NUM,'')	      :(AP_ANSDONE)
AP_ANC
      AP_REASON '*'				      :S(AP_ANSDONE)
      AP_REASON SPAN('0123456789') . ANC =	      :F(AP_ANSDONE)
*     OUTPUT = 'IN ADDPRII ABOUT TO ADDALIST ANC: ' ANC
      ADDALIST(.G_PRMAT<G_CURLINE,5>,ANC)	      :(AP_ANC)
AP_ANSDONE
      AP_LINE G_QSPLIT				  :F(AP_ANSCOMPLETE)
      LEQ(G_OP,'A')				  :F(AP_ANSCOMPLETE)
      G_PRMAT<G_CURLINE,5> = STACK(G_CURLINE,G_PRMAT<G_CURLINE,5>)
      G_PRMAT<G_CURLINE,6> = ''
AP_ANSCOMPLETE
+     LEQ(AP_REASON,'*')			      :F(AP_NOTGOAL)
* goal, the real level is one less
      G_PRMAT<G_CURLINE,3> = G_CURLEVEL - 1		 :(AP_ADDED)
AP_NOTGOAL
+     ADDANTE(G_CURLINE,G_CURLEVEL,AP_LINE,.G_PF,1)	   :(AP_ADDED)
AP_ADDED
+     EQ(P_CHECKPRINT,0)			       :S(AP_RETURN)
      EQ( (G_CURLINE / P_CHECKPRINT) * P_CHECKPRINT, G_CURLINE)
+						       :F(AP_RETURN)
*    OUTPUT = 'ABOUT TO CALL NDUMPROOF	G_CURLINE: ' G_CURLINE
      NDUMPPROOF(G_CURLINE - P_CHECKPRINT + 1)		:(AP_RETURN)
AP_RETURN
* CRM April 29,2987 stuff for unix
	IDENT(G_OS,G_UNIX)				:S(RETURN)
      CHKATN()						  :F(RETURN)
      PUTLINE(" *** ATTENTION! Now entering Debug mode ***")
      DEBUG()					    :S(RETURN)F(END)
ENDADDPRII
-STITL function: ADDPROOF
*FRAME
* Function: ADDPROOF
* Date: Wed Jul 2, 1986
* Last modification: July 2 1986
* Purpose:
*   ADDPROOF is an intermediate step to the original
*  ADDPROOF (now called ADDPRII) so that any FORMLINE containing
*  an identity can be further processed.  It recursively
*  adds any identities that can now be stated.
*
*  Calling Sequence:
*     ADDPROOF(FORMLINE,reason)
*  Parameters:
*     LINE:  FORMLINE to add
*     reason: reason for adding FORMLINE
*
*  RETURN - normal termination
*  Local variables:
*    aid_pres:	LINE number of the FORMLINE added 
*    aid_ptr:	pointer into the IDSTACK
*    aid_fst:	first variable in the identity
*    aid_snd:	second variable in the identity
*    aid_r:    first variable in identity
*    aid_l:    second variable in identity
*    aid_form:	    the formula to add
* CHANGES JULY 16, 1986  CNH PCONS
* add a local variable for use if the entry in the
* idstack is a premconsequence (i.e. no number)
*   aid_pp : linenumber of identity
* Changes Aug 1, 1986  CNH
* this routine will return only if the present show is complete
* ie if onestep succeeds or succeeded at a deeper level.
      pDEFINE('ADDPROOF(FORMLINE,REASON)AID_PRES,AID_PTR,AID_FST' 
+	  ',AID_SND,AID_R,AID_L,AID_PP,AID_FORM')	:(ENDADDPROOF)
ADDPROOF
+     AID_PRES = G_CURLINE + 1
*  add the line into the proof
      ADDPRII(FORMLINE,REASON)
      GT(AID_PRES ,G_CURLINE)			    :S(FRETURN)
*     OUTPUT = 'JUST CALLED ADDPRII'
*     OUTPUT = 'ABOUT TO CALL ONESTEP ON ' FORMLINE
*  Changes Aug 1, 1986	CNH  call onestep here
      ONESTEP(FORMLINE) 				   :S(RETURN)
*     OUTPUT = 'ONESTEP FAILED '
*  Changes Aug 1, 1986	CNH change all returns to Freturns in Addproof
*   check if it is an identity 
* CHANGES ID July 4, 1986 CNH
* changed pattern
      FORMLINE POS(0) LEN(2) . AID_FST G_IDOP 
+     LEN(2) . AID_SND					:F(FRETURN)
      LEQ(AID_FST,AID_SND)				:S(FRETURN)
*  CHANGES July 30, 1986  CNH
*  put a check in for the line to be a SHOW line
*  if it is then return
      G_PRMAT<G_CURLINE,2> '*'				 :S(FRETURN)
      AID_PTR = FLINK(FLINK(G_IDSTACK))
AID_LOOP
+     IDENT(AID_PTR,G_IDSTACK)				:S(FRETURN)
* CHANGES ID July 4, 1986 CNH
* changed pattern
      LINE(AID_PTR) POS(0) LEN(2) . AID_L G_IDOP
+     LEN(2) . AID_R					:F(FRETURN)
* make aid_pp be the line number of the identity
      AID_PP = PLINE(AID_PTR)
      DIFFER(AID_PP)					:S(AID_CONT)
*  the pline was nil so that means the line we are looking
*  for is a premise consequence so we add the necessary lines
*  in now.
      ENTERCONSEQ(ISPC<LINE(AID_PTR)>)
      PP = FINDANTE(LINE(AID_PTR))
AID_CONT
*  we must check all possible matches for the current
*  identity and the line from the circular list
      AID_REASON = G_CURLINE ',' AID_PP ',LL'
      LEQ(AID_L,AID_FST)				:F(AID_CASE2)
      LLT(AID_SND,AID_R)				:F(AID_C11)
      AID_FORM = AID_SND '-' AID_R			:(AIDEND)
AID_C11
      AID_FORM = AID_R '-' AID_SND			:(AIDEND)
AID_CASE2
+     LEQ(AID_R,AID_FST)				:F(AID_CASE3)
      AID_FORM = AID_L '-' AID_SND			:(AIDEND)
AID_CASE3
+     LEQ(AID_L,AID_SND)				:F(AID_CASE4)
      AID_FORM = AID_FST '-' AID_R			:(AIDEND)
AID_CASE4
+     LEQ(AID_R,AID_SND)				:F(AIDFAIL)
      LLT(AID_L,AID_FST)				:F(AID_C41)
      AID_FORM = AID_L '-' AID_FST			:(AIDEND)
AID_C41
      AID_FORM = AID_FST '-' AID_L			:(AIDEND)
*  it didn't match so we repeat with the next item
*  on the circular list
AIDFAIL
+     AID_PTR = FLINK(AID_PTR)				:(AID_LOOP)
AIDEND
*  it matched so we add the new formula to the proof
*  CHANGES July 30, 1986  CNH
*  if the line is already an antecedent line then don't
*  add it in again.
      FINDANTE(AID_FORM)				   :S(FRETURN)
*  Changes Aug 1, 1986	CNH  if addproof succeeds then we are done
      ADDPROOF(AID_FORM, AID_REASON)			   :S(RETURN)
      AID_PTR = FLINK(AID_PTR)				:(AID_LOOP)
ENDADDPROOF
-STITL function: PRINTPROOF()
*FRAME
* Function: PRINTPROOF 
* Date: Sun Jul 25 17:26:39 1982 
* Last modification: June, 1983
* Purpose: 
*  PRINTPROOF prints out the proof matrix in its final formatted 
*  form. 
* 
* Calling sequence: 
*	PRINTPROOF() 
* Parameters: None 
* On return: 
*	RETURN - Normal termination.
* 
* Local variables: 
*	PR_C - Index into the proof matrix
*	PR_F - **** NOT USED ****
*	PR_INDSTR - Indent string to be DUPLed for ever level.
*	PR_CURRENTIND - DUPLed PR_INDSTR
*	PR_CURRENTLEVEL - Curent SHOW level (recursion level)
*	PR_SHOWSTR - **** NOT USED ****
*   CHANGES Aug 13, 1986  CNH  add local var L_SIZE
	L_SIZE - size off the proof before condensing
* 
* Description: 
      pDEFINE('PRINTPROOF(RR)PR_C,PR_F,PR_INDSTR,'
+	      'PR_CURRENTIND,PR_CURRENTLEVEL,PR_SHOWSTR',,
+	      ,'NULL')
+						    :(ENDPRINTPROOF)
PRINTPROOF
+     PR_C = 0
      PR_CURRENTLEVEL = 0
      PR_INDSTR = '|' DUPL(' ',P_INDENT)
      size = G_CURLINE
*  CHANGES Aug 13, 1986  CNH  initialize L_SIZE
      L_SIZE = G_CURLINE
      DIFFER(RR)				      :F(PR_LOOP)
*  CHANGES Aug 13, 1986  CNH  check the flag p_format
      GT(P_FORMAT,0)				      :F(PR_CON)
      CMD_PRDUMP()
PR_CON
      P_CONDENSE 0                                    :F(PR_LOOP)
      CONDENSE()
PR_LOOP
+     PR_C = PR_C + 1
* see if we should decrease in levels
      LT(G_PRMAT<PR_C,3>,PR_CURRENTLEVEL)		 :F(PR_EASY)
      PR_CURRENTLEVEL = G_PRMAT<PR_C,3>
      PR_CURRENTIND = DUPL(PR_INDSTR,PR_CURRENTLEVEL)
*  CHANGE:  June 20, 1986   CNH
*  The check to see if done was changed from
*  IDENT(G_PRMAT<PR_C,1>)   to GT(PR_C,size)
*  where size is initialized to G_CURLINE (also added
*  at this time)  since the size of the proof
*  will be changed by CONDENSE
PR_EASY
+     GT(PR_C,size)				    :S(PR_DONE)
* see if this is a show line
*  CHANGE June 20, 1986    CNH
*  I changed the check for a show line from LEQ(G_PRMAT<PR_C,2>,'*')
*  Since I add the list of the critical lines for the show
*  into the field, hence it is no longer LEQ to '*'
      G_PRMAT<PR_C,2> '*'			  :F(PR_NOTSHOW)
* this is a show line
      PUTLINE( LPAD(PR_C,6) '.	'
+	 RPAD(PR_CURRENTIND '*show '
+	 PUTFORM(G_PRMAT<PR_C,1>,P_FORMTYPE) '	',P_LINELEN))
      PR_CURRENTIND = PR_CURRENTIND PR_INDSTR
      PR_CURRENTLEVEL = PR_CURRENTLEVEL + 1		  :(PR_LOOP)
PR_NOTSHOW
+     PUTLINE( LPAD(PR_C,6) '.	'
+	 RPAD(PR_CURRENTIND PUTFORM(G_PRMAT<PR_C,1>,P_FORMTYPE)
+	 '  ',P_LINELEN) G_PRMAT<PR_C,2> )		  :(PR_LOOP)
PR_DONE
+     PUTLINE(EQ(PR_C,1) "	No proof has yet been calculated.")
+							  :S(RETURN)
*  CHANGES Aug 13, 1986  CNH  don't put stats for long
      RR 'T'					      :S(RETURN)
      G_TOTALTIME = TIME() - G_STTIME
* CHANGE June 1997 counting statements disabled for efficiency
*      G_TOTALCNT = &STCOUNT - G_STCNT
*      PUTLINE("-Statements executed: " G_TOTALCNT)
      PUTLINE( "CPU time for proof: " G_PTIME " msec. ")
      PUTLINE( "CPU time required: " G_TOTALTIME " msec. ")
*  CHANGES Aug 13, 1986  CNH  print # formulas discarded
      PUTLINE("Number of statements discarded: " L_SIZE - size)
							   :(RETURN)
ENDPRINTPROOF
-STITL Function:  CONDENSE
* Function:  CONDENSE
*  Date:  Friday June 20, 1986
*  Last Modification:  June 20, 1986
*  Purpose:
*   This function was added as part of the CHANGES by CNH to condense
*   the final proof so as to contain only the required lines
*   It calls REQUIRED to determine which lines in the proof are
*   needed and COPYREQ to copy those lines which are required to
*   The top of the matrix.
*
*  Calling sequence
*     CONDENSE()
*  Parameters:	None
*  On return:
*     RETURN - Normal Termination.
*
*  Local variables: 
*     None
*
   pDEFINE('CONDENSE()PRES,LINENUM')			:(ENDCONDENSE)
CONDENSE
*  set upt the arrays for keeping track of the new numbering
*  initialize a few variables
      DIGIT = '0123456789'
      NEWNUM = ARRAY(size)
      NEWNUM2 = ARRAY(size)
      PRES = 1
      LINENUM = 1
      NUMBER = SPAN(DIGIT)
*  call the routines to do the work
      REQUIRED(1)
      COPYREQ() 
      ELIMSHOW(size,0) 
      size = LT(LINENUM - 1,size) LINENUM - 1 
*  update the size as changed by elimshow
      G_CURLINE = size					   :(RETURN)
ENDCONDENSE
-STITL Function: REQUIRED
*  Function: REQUIRED
*  Date:  Friday June 20, 1986
*  This is part of the CHANGES made by CNH to condense the final proof
*  This recursively checks which lines are required by the proof
* 
*  Calling Sequence:
*     REQUIRED(linenum)
*
*  Parameters:
*      linenum - the number of the line that is required
*		 and hence who's parents are required
*
*  Local variables:
*     critlist - a string of zero, one or two parent's line numbers
*     crit - one parents line number
*
     pDEFINE('REQUIRED(LINENUM)CRITLIST,CRIT')		:(ENDREQUIRED)
REQUIRED 
*  if it is already required then nothing need be done
      G_PRMAT<LINENUM,7> 'T'				:S(RETURN)
      G_PRMAT<LINENUM,7> = 'T'
*  make it be required then make each of its parents required
      CRITLIST = G_PRMAT<LINENUM,2>
R_GET CRITLIST NUMBER . CRIT =				:F(RETURN)
      REQUIRED(CRIT)					:(R_GET)
ENDREQUIRED
-STITL Function: COPYREQ
*  Function: COPYREQ
*  Date:  Friday June 20, 1986
*  This is part of the CHANGES made by CNH to condense the proof
*  Purpose:
*    This function copies only those lines which are required back
*    into the matrix overwriting the old lines.  It only copies
*    the first three fields as these are all that is needed to
*    to print the proof.  It changes the global variable size
*    to be the new size of the proof.
*
*  Parameters:	None
*  Local variables:
*     old - a pointer in the matrix of each line in the old proof
*     new - a pointer in the matrix for the new proof (new <= old)
*     crit1 - line number of a parent
*     crit2 - line number of a parent
*     compart - used when changing the referenced line numbers
*
    pDEFINE('COPYREQ()OLD,NEW,CRIT1,CRIT2,COMPART') :(ENDCOPYREQ)
COPYREQ
*  increase the counter, update size to be the new size if we are done
	OLD = OLD + 1
      size = GT(OLD,size) NEW				:S(RETURN)
*  if it isn't required then go to the next  line
      G_PRMAT<OLD,7> 'T' =				:F(COPYREQ)
*  it is required so increase the counter, copy it to its
*  new place
      NEW = NEW + 1
      G_PRMAT<NEW,1> = G_PRMAT<OLD,1>
      G_PRMAT<NEW,3> = G_PRMAT<OLD,3>
      G_PRMAT<NEW,2> = G_PRMAT<OLD,2>
*  if it is a show line we don't need to update the reference
      G_PRMAT<NEW,2> '*'				:S(CFIXED)
CBACK
*  update the references
      G_PRMAT<NEW,2> NUMBER . CRIT1 BREAK(DIGIT) . COMPART
+	NUMBER . CRIT2 = NEWNUM<CRIT1> COMPART NEWNUM<CRIT2>
+							:S(CFIXED)
      G_PRMAT<NEW,2> NUMBER . CRIT1 = NEWNUM<CRIT1> 
CFIXED
*  update the newnum array for this line
      NEWNUM<OLD> = NEW 				:(COPYREQ)
ENDCOPYREQ
-STITL Function:  Elimshow()
*FRAME
* Function:  Elimshow
* Date: July 7, 1986
* Last Modification:  July 13, 1986
* Author:   CNH
* Purpose:   This function is the second pass of the condenser.  Its 
*   purpose is to remove unnecessary SHOWs.  A show is considered 
*   unnecessary if it is not the first show, the last line is exactly
*   what was being shown,  the line following the show statements is
*   NOT an assumption line.  If the show is unnecessary then the
*   show line is removed, the lines within the show are raised one
*   level, the reference lines are changed accordingly and any reference
*   to the show is set to be the last line in the show.
* Calling Sequence:    ELIMSHOW()
* Parameters:	LAST : line number of the last line in the show
*     NUMEL : the number of levels of show been removed at this point
* On return:	     normal termination
* Local variables:  SH : line number of the old show line
*   CRIT, CRIT1 :  used to get the critical lines to change them
*  CHANGES elimshow July 7, 1986  CNH
      pDEFINE('ELIMSHOW(LAST,NUMEL)SH,CRIT,CRIT1')	:(ENDELIMSHOW)
ELIMSHOW
EL_LOOP
*  first we check all the conditions for this line being an
*  unnecessary SHOW line. 
      GT(PRES,LAST)					:S(RETURN)
      G_PRMAT<PRES,2> '*'				:F(EL_FAIL)
      EQ(PRES,1)					:S(EL_FAIL)
      G_PRMAT<PRES,2> NUMBER . CRIT =			:F(EL_FAIL)
      G_PRMAT<PRES,2> NUMBER . CRIT =			:S(EL_FAIL)
      CRIT = NEWNUM<CRIT>
      G_PRMAT<PRES,1> G_PRMAT<CRIT,1>			:F(EL_FAIL)
      G_PRMAT<CRIT,1> G_PRMAT<PRES,1>			:F(EL_FAIL)
      G_PRMAT<PRES + 1,2> 'ASSUME'			:S(EL_FAIL)
*  it passed all the tests so it must be an unnecessary SHOW
*  line.  We remember the line we are on (so we can update
*  the newnum2 array)  then recursively call elimshow.
      SH = PRES
      PRES = PRES + 1
      ELIMSHOW(CRIT,NUMEL + 1)
      NEWNUM2<SH> = NEWNUM2<CRIT>			:(EL_LOOP)
EL_FAIL
*  it  failed a test so we simply copy it into its new place
*  we update the reference field to be the new number
      G_PRMAT<LINENUM,1> = G_PRMAT<PRES,1>
      G_PRMAT<LINENUM,2> = G_PRMAT<PRES,2>
      G_PRMAT<LINENUM,3> = G_PRMAT<PRES,3> - NUMEL
      G_PRMAT<LINENUM,2> NUMBER . CRIT BREAK(DIGIT) . COMPART
+	NUMBER . CRIT1 = NEWNUM2<CRIT> COMPART 
+	NEWNUM2<CRIT1>					:S(EL_TWO)
      G_PRMAT<LINENUM,2> NUMBER . CRIT =
+	NEWNUM2<CRIT>
EL_TWO
+     NEWNUM2<PRES> = LINENUM
      PRES = PRES + 1
      LINENUM = LINENUM + 1				:(EL_LOOP)
ENDELIMSHOW
-STITL Function: SHOWN()
*FRAME
* Function: SHOWN 
* Date: Sun Jul 25 17:20:35 1982 
* Last modification: June, 1983
* Purpose: 
*  SHOWN simply removes the top goal, and then fixes up the 
*  proof matrix so that it is nicely indented.	It also 
*  deletes all of the antecedent lines which have been boxed 
*  and adds the line shown to the antecedent line table. 
*  CHANGES:   June 20,1986   CNH 
*  parameter critlines	 to shown   this parameter contains
*  a list of the numbers of the lines which are critical
*  to the show.  These are placed in the 2nd (annotation)
*  field of the matrix
* CHANGES Aug 27, 1986	CNH Parallelism
* add a second parameter to be non null if the show line is
* to not to be added as an antecedent line (i.e. if it being
* used by the parallelism code)
* 
* Calling sequence: 
*	SHOWN(critlines,noadd)
* Parameters: critlines see CHANGES note
* On return: 
*	RETURN - Normal termination.
* 
* Local variables: 
*	SH_G - Pointer in GOAL stack.
*	SH_PRC - Index into proof matrix.
*	SH_TOP - Maximum value for SH_PRC.
* 
      pDEFINE('SHOWN(CRITLINES,NOADD)SH_G,SH_PRC,SH_TOP',,,'NULL')
+					   :(ENDSHOWN)
SHOWN
+     SH_G = G_GOALS
*     OUTPUT = 'IN SHOWN LINE: ' G_CURLINE
*     OUTPUT = 'CRITLINES = ' CRITLINES
      G_CURLEVEL = G_CURLEVEL - 1
      G_GOALS = GOALLINK(G_GOALS)
      SH_TOP  = PRLIN(SH_G)
      SH_PRC = G_CURLINE
      IDENT(SH_TOP,SH_PRC)				:S(SH_NOEXT)
      G_PRMAT<SH_TOP,2> = G_PRMAT<SH_TOP,2> CRITLINES
*     OUTPUT = 'SH_TOP = ' SH_TOP
*     OUTPUT = 'G_PRMAT<SH_TOP> = ' G_PRMAT<SH_TOP,2>
SH_INDENT
+     DIFFER(G_PRMAT<SH_PRC,4>) 			:S(SH_NODEL)
      DELANTE(G_PRMAT<SH_PRC,1>)
* and if this is an existential, then undue the plist
      IDENT(G_PRMAT<SH_PRC,6>)				:S(SH_NOEXT)
      LEQ(DATATYPE(G_PRMAT<SH_PRC,6>),'STRING') 	:F(SH_NOEXT)
* this is a live one
      PROHIBIT(G_DELETE,SH_PRC)
SH_NOEXT
* CHANGES Aug 27, 1986	CNH  4th field fix
* CHANGES Oct 11, 1986	CNH remove 4th field fix
      G_PRMAT<SH_PRC,4> = '1'
SH_NODEL
+     SH_PRC = SH_PRC - 1
      GT(SH_PRC,SH_TOP) 			       :S(SH_INDENT)
      DIFFER(NOADD) GT(SH_TOP,SH_PRC)		      :S(RETURN)
      ADDANTE(SH_PRC,G_CURLEVEL,G_PRMAT<SH_PRC,1>,.G_PF,1) 
      DIFFER(NOADD)			      :S(SH_INDENT)F(RETURN)
ENDSHOWN
-TITLE THINKER - main connective ring manipulations
-STITL Function: CSTACK(CS_HEAD,CS_LINE,CS_PRLN)
*FRAME
* Function: CSTACK 
* Date: Sun Jul 25 17:16:35 1982 
* Last modification: June, 1983
* Purpose: 
*  Add a formula to the ring-list for this stack. Add 
*  at the forward link from the head of the list. 
* 
* Calling sequence: 
*	CSTACK(CS_HEAD,CS_LINE,CS_PRLN) 
* Parameters: 
*	CS_HEAD - Pointer to doubly linked list CS_LINE is to be
*		  added to.
*	CS_LINE - Line to be added.
*	CS_PRLN - Line# at which it occured in the proof matrix.
* On return: 
*	RETURN - Normal termination.
* 
* Local variables: 
*	CS_Z - New node to be added to ring.
* 
      pDEFINE('CSTACK(CS_HEAD,CS_LINE,CS_PRLN)CS_Z',,
+	      'SIMPLE,STRING,INTEGER','NULL')
+							:(ENDCSTACK)
CSTACK
+     CS_Z = SIMPLE(CS_LINE,CS_PRLN,FLINK(CS_HEAD),CS_HEAD)
      BLINK(FLINK(CS_HEAD)) = CS_Z
      FLINK(CS_HEAD) = CS_Z				   :(RETURN)
ENDCSTACK
-STITL Function: DELCSTACK(DS_HEAD)
*FRAME
* Function: DELCSTACK 
* Date: Sun Jul 25 17:18:13 1982 
* Last modification: June, 1983
* Purpose: 
*  DELSTACK deletes the first forward entry from the 
*  ring-list specified. 
* 
* Calling sequence: 
*	DELCSTACK(DS_HEAD) 
* Parameters: 
*	DS_HEAD - Pointer in doubly linked circular list in which
*		  deletion is to occure.
* On return: 
*	RETURN - Normat termination.
* 
* Local variables: None 
* 
* Description: 
* 
*  DS_HEAD - the list in which the deletion is to occur 
      pDEFINE('DELCSTACK(DS_HEAD)',,'SIMPLE','NULL')
+						     :(ENDDELCSTACK)
DELCSTACK
+     BLINK(FLINK(FLINK(DS_HEAD))) = DS_HEAD
      FLINK(DS_HEAD) = FLINK(FLINK(DS_HEAD))		   :(RETURN)
ENDDELCSTACK
-TITLE THINKER - variable manipulations
-STITL Function: FINDVARS()
*FRAME
* Function: FINDVARS 
* Date: Sun Jul 25 17:25:00 1982 
* Last modification: June, 1983
* Purpose: 
*  Find all of the universally quantified goals.  Get the 
*  quantified variable from all of those goals.  Delete 
*  from this list of variables all variables that have 
*  occured free in the antecedent lines. 
* 
* Calling sequence: 
*	result = FINDVARS() 
* Parameters: None 
* On return: 
*	RETURN - Table of all universally quantified goals.
*	FRETURN - No variables found (?)
* 
* Local variables: 
*	FV_GOTONE - (?)
*	FV_I - Index into FV_FVLA
*	FV_CLA - Array form of G_CLIST
*	FV_FVLA - Array form of free varible list
*	FV_G - Copy of G_GOALS.
*	FV_GL - WHAT field of FV_G
*	G_VAR - Quantifier variable from G_SPLIT
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
*	FV_GOTUQ - (?) 
* 
* Description: 
      pDEFINE('FINDVARS()FV_GOTONE,FV_I,FV_CLA,FV_FVLA,'
+	      'FV_G,FV_GL,G_VAR,G_LOP,G_ROP,G_OP,FV_GOTUQ',,
+	      ,'TABLE')
+						      :(ENDFINDVARS)
FINDVARS
+     FINDVARS = TABLE(11)
      FV_G = G_GOALS
FV_LOOP
+     IDENT(FV_G)					 :S(FV_DONE)
      FV_GL = WHAT(FV_G)
      FV_GL G_QSPLIT					 :F(FV_NEXT)
      G_OP 'A'						 :F(FV_NEXT)
      FV_GOTUQ = '1'
      IDENT(G_FVLIST<G_VAR>)				 :F(FV_NEXT)
      FV_GOTONE = '1'
      FINDVARS<G_VAR> = '1'
FV_NEXT
+     FV_G = GOALLINK(FV_G)				  :(FV_LOOP)
FV_DONE
+     IDENT(FV_GOTONE)					  :F(RETURN)
      IDENT(FV_GOTUQ)					 :F(FRETURN)
* see if there are any free variables or constants
      FV_FVLA = CONVERT(G_FVLIST,'ARRAY')		 :F(FV_CONS)
      FV_I = 1
FV_VARSLOOP
+     FV_FVLA<FV_I,2>					 :F(FV_CONS)
      IDENT(FV_FVLA<FV_I,2>)				 :F(FRETURN)
      FV_I = FV_I + 1				      :(FV_VARSLOOP)
FV_CONS
+     FV_CLA = CONVERT(G_CLIST,'ARRAY') 		  :F(FV_NEW)
      FV_I = 1
FV_CONSLOOP
+     FV_CLA<FV_I,2>					  :F(FV_NEW)
      IDENT(FV_CLA<FV_I,2>)				 :F(FRETURN)
      FV_I = FV_I + 1				      :(FV_CONSLOOP)
FV_NEW
+     FINDVARS<GETNEWVAR()> = '1'		:F(FRETURN)S(RETURN)
ENDFINDVARS
-STITL Function: GETNEWVAR()
*FRAME
* Function: GETNEWVAR 
* Date: Sun Jul 25 17:19:24 1982 
* Last modification: June, 1983
* Purpose: 
*  GETNEWVAR - get a variable that has never been used in the 
*  the proof before 
* 
* Calling sequence: 
*	result = GETNEWVAR() 
* Parameters: None 
* On return: 
*	RETURN - Returns an unsed variable.
*	FRETURN - No variable is avaiable.
* 
* Local variables: None.
* 
      pDEFINE('GETNEWVAR()',,,'STRING') 	     :(ENDGETNEWVAR)
GETNEWVAR
      GETNEWVAR = GETVARNAME(G_LASTFREEVAR)		 :F(FRETURN)
      G_LASTFREEVAR = G_LASTFREEVAR + 1
      IDENT(G_VLIST<GETNEWVAR>) 	      :F(GETNEWVAR)
      IDENT(G_PREMVLIST<GETNEWVAR>)           :F(GETNEWVAR)S(RETURN)
ENDGETNEWVAR
-STITL Function: GETVARNAME(GVN_NUMBER)
*FRAME
* Function: GETVARNAME
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*	This function will return the name of the variable
*	associated with the inputparamter. Valid input is
*	from 0 (first variable) to G_VARSIZE - 1 (last one)
* 
* Calling sequence: 
*	return = GETVARNAME(GVN_NUMBER)
* Parameters: 
*	GVN_NUMBER - Number associated with variable.
* 
* On return: 
*	RETURN	- Name of variable found.
*	FRETURN - Number of variable out of range.
* 
* Local variables: 
*	GVN_N1 - Fist digit(s) of number.
*	GVN_N2 - Last digit of number.
* 
      pDEFINE('GETVARNAME(GVN_NUMBER)GVN_N1,GVN_N2',,
+	      'INTEGER','STRING')		    :(ENDGETVARNAME)
GETVARNAME
      LT(GVN_NUMBER,G_VARSIZE)				 :F(FRETURN)
      GVN_NUMBER RTAB(1) . GVN_N1 REM . GVN_N2		:F(CONCHECK)
      GETVARNAME = SUBSTR(G_VARORDER,GVN_N1 + 1,1) 9 - GVN_N2
+							   :(RETURN)
ENDGETVARNAME
-STITL Function: FREE(FRE_VAR)
*FRAME
* Function: FREE 
* Date: Sun Jul 25 17:22:55 1982 
* Last modification: June, 1983
* Purpose: 
*  The question is has var occured free in a line which 
*  is antecedent to the current goal 
* 
* Calling sequence: 
*	FREE(FRE_VAR) 
* Parameters: 
*	FRE_VAR - Variable to be tested. 
* On return: 
*	RETURN - FRE_VAR has occured free in a line antecedent to
*		 the current goal.
*	FRETURN - FRE_VAR was not free.
* 
* Local variables: 
*	FRE_FVL - Name of free variable.
* 
      pDEFINE('FREE(FRE_VAR)FRE_FVL',,'STRING','NULL')	  :(ENDFREE)
FREE
+     FRE_FVL = G_FVLIST<FRE_VAR>
FRE_LOOP
+     IDENT(FRE_FVL)					 :S(FRETURN)
      LT(V(FRE_FVL),G_CURLEVEL) 			  :S(RETURN)
      FRE_FVL = N(FRE_FVL)				 :(FRE_LOOP)
ENDFREE
-STITL Function: LISTEMPTY(LE_VLIST)
*FRAME
* Function: LISTEMPTY 
* Date: Sun Jul 25 17:25:36 1982 
* Last modification: June, 1983
* Purpose: 
*  Listempty takes a "variable list" and determines if it is empty
*
*  ********* THIS FUNCTION is NEVER CALLED ***********
*
* 
* Calling sequence: 
*	LISTEMPTY(LE_VLIST) 
* Parameters: 
*	LE_VLIST - Variable list to be tested
* On return: 
*	RETURN - LE_VLIST is empty.
*	FRETURN - LE_VLIST is not empty.
* 
* Local variables: 
*	LE_ALIST - ARRAY form of LE_VLIST
*	LE_I - Index into the array LE_ALIST
* 
      pDEFINE('LISTEMPTY(LE_VLIST)LE_ALIST,LE_I',,
+	     'NULL|TABLE','NULL')
+						     :(ENDLISTEMPTY)
LISTEMPTY
+     IDENT(LE_VLIST)					  :S(RETURN)
      LE_ALIST = CONVERT(LE_VLIST,'ARRAY')		  :F(RETURN)
      LE_I = 1
LE_LELOOP
+     ?LE_ALIST<LE_I,2> 				  :F(RETURN)
      IDENT(LE_ALIST<LE_I,2>)				 :F(FRETURN)
      LE_I = LE_I + 1					:(LE_LELOOP)
ENDLISTEMPTY
-TITLE THINKER - antecedent line simplifications
*FRAME
* The following routines all search using the ring-lists for 
* various rules to see what may be added to the proof matrix. 
* Often while searching around a ring, another entry will be 
* added to the ring which should also be searched.  If this 
* is required then the ring is searched in the direction of 
* the BLINKS so that new material added at the start of the 
* ring will be searched before the starting point is reached. 
*  In all of the routines, if a line is found that fits the 
* required rule it is added, and if a contradiction can be 
* found by one level, then SHOWN is used and a successful 
* return is made.  Otherwise a failure return is made. 
-STITLE Function:  FINDANDS(FD_X)
*FRAME
* Function:  FINDANDS
* Date:  Wed Aug 13, 1986
* Author:  CNH
* Purpose:	
*  To simplify the (P&Q) pointed to by FD_X if possible
* Calling Sequence:
*  FINDANDS(FD_X)
* Parameters:
*  FD_X : pointer into the doubly linked circular list
* On return:
*  RETURN - The show has been closed
*  FRETURN - Otherwise
* Local variables:
*  G_ROP - right operator as given by G_SPLIT
*  G_OP  - Main operator as given by G_SPLIT
*  G_LOP - left operand as given by F_SPLIT
*
      pDEFINE('FINDANDS(FD_X)G_LOP,G_OP,G_ROP',,,'NULL')
+						      :(ENDFINDANDS)
FINDANDS
      LINE(FD_X) G_BSPLIT			      :F(CONCHECK)
      FINDANTE(G_LOP)				      :S(FD_RIGHT)
      ADDPROOF(G_LOP,PLINE(FD_X) ',S')		      :S(RETURN)
FD_RIGHT
      FINDANTE(G_ROP)				      :S(FRETURN)
      ADDPROOF(G_ROP,PLINE(FD_X) ',S')		      :S(RETURN)F(FRETURN)
ENDFINDANDS
-STITLE Function:  FINDBCS(FB_X)
*FRAME
* Function:  FINDBCS
* Date:  Wed Aug 13, 1986
* Author:  CNH
* Purpose:	
*  To simplify the (P=Q) pointed to by FD_X if possible
* Calling Sequence:
*  FINDBCS(FB_X)
* Parameters:
*  FB_X : pointer into the doubly linked circular list
* On return:
*  RETURN - The show has been closed
*  FRETURN - Otherwise
* Local variables:
*  FB_NEW - new implication
*  G_ROP - right operator as given by G_SPLIT
*  G_OP  - Main operator as given by G_SPLIT
*  G_LOP - left operand as given by F_SPLIT
      pDEFINE('FINDBCS(FB_X)FB_NEW,G_LOP,G_OP,G_ROP',,,'NULL')
+						      :(ENDFINDBCS)
FINDBCS
      LINE(FB_X) G_BSPLIT			      :F(CONCHECK)
      FB_NEW = '(' G_LOP '>' G_ROP ')'
      FINDANTE(FB_NEW)				      :S(FB_RIGHT)
      ADDPROOF(FB_NEW, PLINE(FB_X) ',BC')	      :S(RETURN)
FB_RIGHT
      FB_NEW = '(' G_ROP '>' G_LOP ')'
      FINDANTE(FB_NEW)				      :S(FRETURN)
      ADDPROOF(FB_NEW, PLINE(FB_X) ',BC')	      :S(RETURN)F(FRETURN)
ENDFINDBCS
-STITL Function: FINDEI()
*FRAME
* Function: FINDEI 
* Date: Sun Jul 25 17:23:54 1982 
* Last modification: June, 1983
* Purpose: 
*  find an existentially quantified formula that hasn't yet 
*  been findei()'ed immediately under this goal.  Then EI 
*  to a new variable.  If there are no new lines or now 
*  new variables fail. 
* 
* Calling sequence: 
*	FINDEI() 
* Parameters: None 
* On return: 
*	RETURN - Call to ONESTEP succeeded 
*	FRETURN - Otherwise. 
* 
* Local variables: 
*	FE_NEWVAR - New variable returned from GETNEWVAR.
*	FE_TEMPL - Template returned from MAKEQTPLATE.
* 
      pDEFINE('FINDEI()FE_NEWVAR,FE_TEMPL',,,'NULL')	:(ENDFINDEI)
FINDEI
+     IDENT(G_EQSTACK,FINDEIPT(G_GOALS))		   :F(FE_DO)
      IDENT(G_EQSTACK,BLINK(FINDEIPT2(G_GOALS)))	 :S(FRETURN)
      FINDEIPT(G_GOALS) = BLINK(FINDEIPT2(G_GOALS))
FE_DO
+     FE_NEWVAR = GETNEWVAR()				 :F(FRETURN)
      FE_TEMPL = MAKEQTPLATE(LINE(FINDEIPT(G_GOALS)))
      FE_TEMPL '(E@)' = 				:F(CONCHECK)
FE_NEWVAR
+     FE_TEMPL '@' = FE_NEWVAR			       :S(FE_NEWVAR)
      FE_TEMPL = NORMALIZE(FE_TEMPL)
      ADDPROOF(FE_TEMPL,PLINE(FINDEIPT(G_GOALS)) ',EI') 
*  changes Aug 6, 1986	CNH  take out return so will prohibit
*  changes Aug 5, 1986	CNH put return on success
*  CNH	3rd parameter to ADDPROOF removed
* mark the variable as prohibited for this lines ancestors
*     OUTPUT = 'JUST ABOUT TO CALL PROHIBIT WITH VAR: ' FE_NEWVAR
      PROHIBIT(G_ADD,PLINE(FINDEIPT(G_GOALS)),FE_NEWVAR)
      FINDEIPT2(G_GOALS) = FINDEIPT(G_GOALS)
      FINDEIPT(G_GOALS) = BLINK(FINDEIPT(G_GOALS))	   :(RETURN)
ENDFINDEI
-STITL Function: FINDID()
*FRAME
* Function: FINDID
* Date: Sat Jul 20 16:12:34 1983
* Last modification: 
* Purpose: 
*  To find all identities, find equivalence classes for them, replace each
*  occurance of each variable in an equivalence class with the lexically
*  smallest variable in that class in ALL antecedent lines.
* 
* 
* Calling sequence: 
*	FINDID() 
* Parameters: None 
* On return: 
*	RETURN - ONESTEP succeeded.
*	FRETURN - otherwise.
* 
* Local variables: 
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
*	NEGLINE - index into g_prmat
*	PCPP - used in the case that the line ref is  blank
* 
*  CHANGES ID July 2, 1986
*  Add REFLINE as a local variable
      pDEFINE('FINDID()REFLINE,FI_Z,FI_X,FI_NORM,G_LOP,G_ROP,G_OP,PCPP,NEGLINE',,,'NULL')
+						       :(ENDFINDID)
FINDID
      EQ(P_NOID)				    :S(FRETURN)
      FI_X = FINDIDPT(G_GOALS)
      IDENT(G_IDSTACK,FI_X)				:F(FI_00)
      IDENT(G_IDSTACK,BLINK(FINDIDPT2(G_GOALS)))	:S(FRETURN)
      FI_X = BLINK(FINDIDPT2(G_GOALS))
FI_00
      FI_TRAN = IDENTRAN(G_GOALS)
FI_LOOP1
      IDENT(FI_X,G_IDSTACK)				:S(FI_DONE1)
* get the variable names
      LINE(FI_X) G_IDSPLIT				:F(CONCHECK)
* is this an a=a ?
      LEQ(G_LOP,G_ROP)					:S(FI_CONT1)
      FI_V1 = G_LOP
      FI_V2 = G_ROP
* pick off translations if they exist, and if they don't then
* we pretend we translate to ourselves.
     FI_EQV1 = FI_V1
     FI_EQV2 = FI_V2
     IDENT(FI_TRAN<FI_V1>)				:S(FI_05)
     FI_EQV1 = ECREP(FI_TRAN<FI_V1>)
FI_05
     IDENT(FI_TRAN<FI_V2>)				:S(FI_10)
     FI_EQV2 = ECREP(FI_TRAN<FI_V2>)
FI_10
* now sort so that FI_EQV1 < FI_EQV2 
     LEQ(FI_EQV1,FI_EQV2)				:S(FI_CONT1)
     LLT(FI_EQV1,FI_EQV2)				:S(FI_15)
     FI_TEMP = FI_EQV1
     FI_EQV1 = FI_EQV2
     FI_EQV2 = FI_TEMP
FI_15
* If fi_v1 is no the name of the equivalence class being processed
* then add an entry of it in the translation table if the entry is 
* not there alreay.  If we do add the entry then this is the first
* time we have seen this variable so add it to the stack of variables
* that have equivalance replacements to be preformed.
     LEQ(FI_V1,FI_EQV1) 				:S(FI_20)
     DIFFER(FI_TRAN<FI_V1>)				:S(FI_20)
* add the entry
      PCPP = PLINE(FI_X)
      IDENT(PCPP)					   :F(PC1_F)
      PCPP = PRLINE(FINDANTE(LINE(FI_X)))
*  CHANGE THE REFERENCE FROM PLINE(FI_X) TO PCPP
PC1_F FI_TRAN<FI_V1> = EQUIVCLASS(FI_EQV1,0,PCPP)
     IDENTVARS(G_GOALS) = STACK(FI_V1,IDENTVARS(G_GOALS))
FI_20
* Now do the same for fi_v2
     LEQ(FI_V2,FI_EQV1) 				:S(FI_25)
     DIFFER(FI_TRAN<FI_V2>)				:S(FI_25)
* add the entry
*  CHANGE THE REFERENCE FROM PLINE(FI_X) TO PCPP
     FI_TRAN<FI_V2> = EQUIVCLASS(FI_EQV1,0,PCPP)
     IDENTVARS(G_GOALS) = STACK(FI_V2,IDENTVARS(G_GOALS))
FI_25
* Now do the same for fi_eqv2
     LEQ(FI_EQV2,FI_EQV1)				:S(FI_30)
     DIFFER(FI_TRAN<FI_EQV2>)				:S(FI_30)
* add the entry
*  CHANGE THE REFERENCE FROM PLINE(FI_X) TO PCPP
     FI_TRAN<FI_EQV2> = EQUIVCLASS(FI_EQV1,0,PCPP)
     IDENTVARS(G_GOALS) = STACK(FI_EQV2,IDENTVARS(G_GOALS))
FI_30
* Now run through the replacement table and replace all occurances
* of eqv2 as a replacement with eqv1s.	Note that we have to copy the
* entries so that we do not effect the levels above us.
     FI_P = IDENTVARS(G_GOALS)
FI_LOOP2
     IDENT(FI_P)					:S(FI_DONE2)
     FI_V = V(FI_P)
     LEQ(ECREP(FI_TRAN<FI_V>),FI_EQV2)			:F(FI_CONT2)
* make a copy
     FI_TRAN<FI_V> = COPY(FI_TRAN<FI_V>)
     ECREP(FI_TRAN<FI_V>) = FI_EQV1
      PCPP = PLINE(FI_X)
      IDENT(PCPP)					   :F(PC2_F)
      PCPP = PRLINE(FINDANTE(LINE(FI_X)))
*  CHANGE THE REFERENCE FROM PLINE(FI_X) TO PCPP
PC2_F ECREASON(FI_TRAN<FI_V>) = ECREASON(FI_TRAN<FI_V>) ',' PCPP
*Mods made by Randy
*     OUTPUT = "Entering Randy's Mods"
     FI_NORM = NORMALIZE(FI_V '-' FI_EQV1)
     FINDANTE(FI_NORM)					:S(FI_CONT2)
*     OUTPUT = 'Adding proof line ...' FI_NORM
*  CHANGES ID July 2, 1986   CNH
*  I believe the line that was in puts the identity in the proof
*  CHANGES ID July 2, 1986  CNH
*  REFLINE has the line number of the identity
      REFLINE = G_CURLINE + 1
      ADDPROOF(FI_NORM, PRLINE(FINDANTE(FI_NORM)) ',R') :S(RETURN)
*  changes Aug 5, 1986	CNH put return on success
      REFLINE = GT(REFLINE,G_CURLINE) PRLINE(FINDANTE(FI_NORM))
*    ADDPROOF(FI_NORM,ECREASON(FI_TRAN<FI_V>) ',LL')
*  CNH	3rd parameter to ADDPROOF removed
*End mods
FI_CONT2
     FI_P = N(FI_P)					:(FI_LOOP2)
FI_DONE2
FI_CONT1
     FI_X = BLINK(FI_X) 				:(FI_LOOP1)
FI_DONE1
     FINDIDPT2 = FI_X
     FINDIDPT(G_GOALS) = BLINK(FI_X)
* Ok, we have an up to date replacement table, now all we have to do
* is make all of the identity replacements that we can manage.
     FI_P = IDENTVARS(G_GOALS)
FI_LOOP3
     IDENT(FI_P)					:S(FI_DONE3)
     FI_V = V(FI_P)
     FI_REP = ECREP(FI_TRAN<FI_V>)
     FI_LAST = ECLAST(FI_TRAN<FI_V>)
     FI_REASON = ECREASON(FI_TRAN<FI_V>)
* Now, find each effected antecedent line
     FI_THISVAR = G_VARIDENT<FI_V>
FI_LOOP4
     IDENT(FI_THISVAR)					:S(FI_DONE4)
*  CHANGES July 22, 1986  CNH add check for a negative line number
     V(FI_THISVAR) '-' REM . NEGLINE			  :F(PC_2C)
     FI_NEW = MAKEQTPLATE('(A' FI_V ')' G_PREM<NEGLINE>)  :(PC_1C)
PC_2C
+    LT(V(FI_THISVAR),FI_LAST)				:S(FI_DONE4)
     FI_NEW = MAKEQTPLATE('(A' FI_V ')' G_PRMAT<V(FI_THISVAR),1>)
PC_1C FI_NEW '(A@)' = ''
FI_LOOP5
+    FI_NEW '@' = FI_REP				:S(FI_LOOP5)
     FI_NORM = NORMALIZE(FI_NEW)
     FINDANTE(FI_NORM)					:S(FI_CONT4)
* CHANGES ID JULY 4, 1986 CNH
* reiterate the line FI_REP '-' FI_V
*  changes Aug 5, 1986	CNH put return on success
      ADDPROOF(FI_REP '-' FI_V, PRLINE(FINDANTE(FI_REP '-' FI_V)) ',R')
+						      :S(RETURN)
      REFLINE = PRLINE(FINDANTE(FI_REP '-' FI_V))
*  CHANGES ID July 2, 1986
*  Change the first line reference  from FI_REASON
     ADDPROOF(FI_NORM, REFLINE
+	',' V(FI_THISVAR) ',LL')		      :S(RETURN)
*  changes Aug 5, 1986	CNH put return on success
*  CNH	3rd parameter to ADDPROOF removed
FI_CONT4
     FI_THISVAR = N(FI_THISVAR) 			:(FI_LOOP4)
FI_DONE4
* Mark this variable as replaced up to this point
     EQ(FI_LAST,0)					:S(FI_35)
     FI_TRAN<FI_V> = COPY(FI_TRAN<FI_V>)
FI_35
     ECLAST(FI_TRAN<FI_V>) = G_CURLINE
     FI_P = N(FI_P)					:(FI_LOOP3)
FI_DONE3
							:(FRETURN)
ENDFINDID
-STITLE Function: REFINST(FORMULA)
*FRAME
* Function: REFINST
* Date: Tue Jun 12 10:58:10 1984
* Written by: Randy Kopach
* Last modification: Tue Jun 12 10:58:10 1984
* Purpose:
*  The function simply checks if the formula is an instance
* of "a-a".
*
* Calling sequence:
*	REFINST(FORMULA)
* Parameters: 
*	FORMULA -	The formula to be checked
* On return:
*	RETURN -	Formula passed is an instance of "a-a"
*	FRETURN -	Formula passed is not an instance of "a-a"
*
* Local variables: 
*	G_LOP - 	left side of operator
*	G_OP -		binary operator
*	G_ROP - 	right side of operator
      pDEFINE('REFINST(FORMULA)G_LOP,G_OP,G_ROP',,'STRING','NULL')
+							:(ENDREFINST)
REFINST
      EQ(P_NOID)				    :S(FRETURN)
      FORMULA G_IDSPLIT 				:F(FRETURN)
      IDENT(G_LOP,G_ROP)				:S(RETURN)F(FRETURN)
ENDREFINST
-STITLE Function: EGREF(FORMULA)
*FRAME
* Function: EGREF
* Date: Tue Jun 12 10:26:28 1984
* Written by: Randy Kopach
* Last modification: Tue Jun 12 10:26:28 1984
* Purpose:
*  This routine will see whether or not the formula passed
* can be gotten from an identity a-a by rule EG.  That is,
* it looks to see if P has one of the forms:
*	    (Ex)x-a	(Ex)a-x     (Ex)x-x
* where "a" represents any constant and "x" represents any
* variable.
*
* Calling sequence:
*	EGREF(FORMULA)
* Parameters: 
*	FORMULA -	The formula that EGREF is to act on
* On return:
*	RETURN -	Formula P can be gotten from an identity
*			and rule EG
*	FRETURN -	Formula P failed
*
* Local variables: 
*	REST -	rest of formula after quantifier removed
*	G_OP -	binary operator
*	G_LOP - formula to the left of operator
*	G_ROP - formula to the right of operator
*	G_VAR - quantified variable
*		(for further information on the variables, 
*		 check the G_QSPLIT and G_IDSPLIT patterns)
      pDEFINE('EGREF(FORMULA)REST,G_OP,G_LOP,G_ROP,G_VAR',,'STRING','NULL')
+							:(ENDEGREF)
EGREF FORMULA G_QSPLIT					:F(FRETURN)
      EQ(P_NOID)				    :S(FRETURN)
      G_OP 'E'						:F(FRETURN)
      REST = G_ROP
      REST G_IDSPLIT					:F(FRETURN)
      IDENT(G_LOP,G_VAR)				:S(RETURN)
      IDENT(G_ROP,G_VAR)				:S(RETURN)F(FRETURN)
ENDEGREF
-STITLE Function: VARSUB(EQN,BUILD)
* * *
*  EQN_LIST is a small datatype used to pruduce linked lists of
*  equations (and info on them) for use by the VARSUB function
*  and the routines that call it.
* * *
      DATA('EQN_LIST(EQUATION,V_OLD,V_NEW,NEXT)')
      DATA('ANTE_PTRS(FORM_PTR,IDENT_PTR)')
*FRAME
* Function: VARSUB
* Date: Mon Jun  4 13:52:55 1984
* Written by: Randy Kopach
* Last modification: Mon Jun  4 13:52:55 1984
* Purpose:
* VARSUB will take a given formula, variable, and identity, and
* produce all possible formulas with identity substitutions for
* the given variable.  For example, if we had the formula:
*		       ((a1&a2)+a1)
* and we knew that b1=a1, then the spitbol call:
*     PTR = VARSUB("((a1&a2)+a1)",NULL)
* would produce a linked list containing:
*		       ((a1&a2)+a1)
*		       ((a1&a2)+b1)
*		       ((b1&a2)+a1)
*		       ((b1&a2)+b1)
*
* Calling sequence:
*	result = VARSUB(EQN,BUILD)
*	   Also, the variable VAR should contain the original variable
*	and the variable NEW should contain the replacement variable.
*	Note also that the variable PTR is used to build the linked
*	list on.  This may seem like a sin against scope, but it's
*	only done to reduce the overhead of passing them as parameters
*	recursively.
* Parameters: 
*	EQN -	Equation to have identity substitution carried out on.
*	BUILD - Left part of formula as variable substitution progresses.
*		Note that the BUILD parameter should be left null since
*		it is only used during recursive invokations of VARSUB
*		by VARSUB itself.
* On return:
*	RETURN -	Global variable PTR has a node tacked on to it
*			yielding a new formula with identity substitution.
*
* Local variables: 
*	L_EQN - Left side of formula subsection being scanned
*	R_EQN - Remainder of formula subsection being scanned
*	P -	Position of identity variable in formula string
      pDEFINE('VARSUB(EQN,BUILD)L_EQN,R_EQN,P,Q',,
+			       'STRING|NULL,STRING|NULL','NULL')
+							:(ENDVARSUB)
VARSUB
+     Q = 0
V_TOP EQN @P VAR REM $ R_EQN				:F(V_BOTTOM)
      EQ(P + Q ,0)					:S(V_NULL_STRING)
      L_EQN = SUBSTR(EQN,1,P)				:(V_TCONT)
V_NULL_STRING
+     L_EQN = NULL
V_TCONT
+     IDENT(G_FVLOC<SIZE(BUILD) + P + Q>)		:S(V_BOUNDVAR)
      VARSUB(R_EQN,(BUILD L_EQN VAR))
      VARSUB(R_EQN,(BUILD L_EQN NEW))			:(RETURN)
V_BOUNDVAR
+     L_EQN = L_EQN VAR
      EQN = R_EQN					:(V_TOP)
V_BOTTOM
+     NEXT(PTR) = EQN_LIST((BUILD EQN),VAR,NEW,NULL)
      PTR = NEXT(PTR)					:(RETURN)
ENDVARSUB
-STITLE Function: TESTID(FORMULA,IDENTITY)
*FRAME
* Function: TESTID
* Date: Wed Jun 13 11:20:41 1984
* Written by: Randy Kopach
* Last modification: Wed Jun 13 11:20:41 1984
* Purpose:
*   TESTID will look in the antecedent line table to find
* a formula just like the one passed except that the identity
* variable has been substituted into it.  If a matching
* antecedent line is found, the function returns the antecedent
* line pointer.  If no antecedent line is found, the TESTID
* function fails.
*
* Calling sequence:
*	result = TESTID(FORMULA,IDENTITY)
* Parameters: 
*	FORMULA -	formula to be tested
*	IDENTITY -	identity formula
* On return:
*	RETURN -	returns pointer to antecedent line
*	FRETURN -	no antecedent line found
*
* Local variables: 
*	PTR -		Used for linked list of formulae
*	START_PTR -	Used for linked list of formulae
*	TEMP_PTR -	Used for linked list of formulae
*	TEMP_FORMULA -	Holds modified version of formula (all occurances
*			of one identity variable are converted to the other)
*	G_LOP - 	Left side of identity formula
*	G_ROP - 	Right side of identity formuala
*	G_OP -		identity operator
*	VAR -		variable (in formula) to be substituted
*	NEW -		replacement variable for above
      pDEFINE('TESTID(FORMULA,IDENTITY)PTR,START_PTR,TEMP_PTR,FORM_OLD,'
+			     'TEMP_FORMULA,G_LOP,G_ROP,G_OP,VAR,NEW',,
+			     'STRING,STRING','ANTE')
+							:(ENDTESTID)
TESTID
      EQ(P_NOID)				    :S(FRETURN)
      EQ(SIZE(IDENTITY),5)				:F(FRETURN)
      IDENTITY G_IDSPLIT				:F(FRETURN)
      EQ(SIZE(G_LOP),2) 				:F(FRETURN)
      EQ(SIZE(G_ROP),2) 				:F(FRETURN)
      DIFFER(G_LOP,G_ROP)				:F(FRETURN)
      FORMULA (G_LOP | G_ROP)				:F(FRETURN)
      FORM_OLD = FORMULA
* * *
*    First we'll convert all occurances of one of the identity
* variables to the other variable.  Then ensures that we generate
* all possible combinations of the variables in the formula list.
* Then we'll produce a linked list of formulae with variable
* substitution based on the identity relation passed to us.
* * *
      TEMP_FORMULA = FORMULA
T_CVT TEMP_FORMULA G_ROP = G_LOP				:S(T_CVT)
      START_PTR = EQN_LIST(TEMP_FORMULA,NULL)
      PTR = START_PTR
      VAR = G_LOP
      NEW = G_ROP
      G_FVLOC = TABLE(10)
*     OUTPUT = '%%% TESTID exucuting G_LF with ' EQUATION(START_PTR)
*     OUTPUT = '%%% and VAR = ' VAR ' , NEW = ' NEW
      EQUATION(START_PTR) G_LF
      VARSUB(EQUATION(START_PTR),NULL)
* * *
*   Now that we have a list of formulae, we can search the
* ANTELINES table to see if it already exists in the proof.
* If so, TESTID returns the antecedent line pointer that
* FINDANTE gives it.  If no formulae exist in the antecedent
* line table, the TESTID function fails.
* * *
      PTR = NEXT(START_PTR)
T_LINK
+     IDENT(FORMULA,EQUATION(PTR))		      :S(T_SKIP)
      TESTID = FINDANTE(EQUATION(PTR))			:S(RETURN)
T_SKIP	 PTR = NEXT(PTR)
      DIFFER(PTR,NULL)					:S(T_LINK)F(FRETURN)
ENDTESTID
-STITLE Function: IDTEST(FORM_1,FORM_2)
*FRAME
* Function: IDTEST
* Date: Thu Jun 14 11:04:28 1984
* Written by: Randy Kopach
* Last modification: Thu Jun 14 11:04:28 1984
* Purpose:
*   The IDTEST function will check whether two formulae are
* justifiable except that one needs to have a variable substituted
* in to get the other (ie. Is there an identity that transforms
* one into the other).	If such an identity is found, IDTEST searches
* the ANTELINE tables to see if the identity is there.	If so, IDTEST
* returns the ANTELINES pointer.  If not, it keeps trying until it
* has exausted all variable substitutions (at which point it fails).
*
* Calling sequence:
*	result = IDTEST(FORM_1,FORM_2)
* Parameters: 
*	FORM_1 -	a legal formula
*	FORM_2 -	a legal formula
* On return:
*	RETURN -	returns the antecedent line pointer for the
*			identity that justifies the two formulae.
*	FRETURN -	IDTEST failed to find an identity for the formulae
*
* Local variables: 
*	PTR -		Used for linked list of formulae
*	START1_PTR -	Used for linked list of formulae
*	START2_PTR -	Used for linked list of formulae
*	VAR -		variable (in formula) to be substituted
*	NEW -		replacement variable for above
*	F1_TMP -	scratch storage
*	F2_TMP -	scratch storage
*	F1_VAR -	scratch storage
*	F2_VAR -	scratch storage
*	F1_STORE -	scratch storage
*	F2_STORE -	scratch storage
      pDEFINE('IDTEST(FORM_1,FORM_2)PTR,START_PTR,VAR,NEW,NEWFORM_1,'
+				  'F1_TMP,F2_TMP,F1_VAR,F2_VAR,'
+				  'F1_STORE,F2_STORE,ID_NORM',,
+				  'STRING,STRING','ANTE')
+							:(ENDIDTEST)
IDTEST
      EQ(P_NOID)				    :S(FRETURN)
* * *
*   First, let's extract all the variables out of the equations
* * *
      F1_TMP = FORM_1
I_SCAN1
+     F1_TMP G_BT $ VAR = NULL				:F(I_NXT)
      F1_VAR VAR					:S(I_SCAN1)
      F1_VAR = F1_VAR VAR				:(I_SCAN1)
I_NXT
+     F2_TMP = FORM_2
I_SCAN2
+     F2_TMP G_BT $ VAR = NULL				:F(I_FOUNDALL)
      F2_VAR VAR					:S(I_SCAN2)
      F2_VAR = F2_VAR VAR				:(I_SCAN2)
I_FOUNDALL
* * *
*   Now let's check if we're wasting our time.	If not, then we
*   can proceed with variable substitutions on the equations.
* * *
      IDENT(F1_TMP,F2_TMP)				:F(FRETURN)
      DIFFER(F1_VAR,NULL)				:F(FRETURN)
      DIFFER(F2_VAR,NULL)				:F(FRETURN)
      F1_STORE = F1_VAR
      F2_STORE = F2_VAR
      START_PTR = EQN_LIST(FORM_1,NULL,NULL,NULL)
* * *
*  The main loop will take all occurances of a variable in
*  formula 2, and substitute them into every variable in
*  formula 1.  If ever we can produce formula 2 again with
*  such a substitution, we've found an identity.
* * *
I_LOOP1
      PTR = START_PTR
      F1_VAR = F1_STORE
      F2_VAR G_BT $ NEW = NULL				:F(FRETURN)
I_LOOP2
      F1_VAR G_BT $ VAR = NULL				:F(I_LOOP1)
      IDENT(VAR,NEW)					:S(I_LOOP2)
      NEWFORM_1 = FORM_1
I_SUB NEWFORM_1 NEW = VAR				:S(I_SUB)
      G_FVLOC = TABLE(10)
*     OUTPUT = '%%% IDTEST exucuting G_LF with ' EQUATION(START_PTR)
*     OUTPUT = '%%% and VAR = ' VAR ' , NEW = ' NEW
      NEWFORM_1 G_LF
      VARSUB(NEWFORM_1,NULL)
      PTR = START_PTR
I_SCAN
      IDENT(FORM_2,EQUATION(PTR))			:S(MATCHED)
I_CONT
      IDENT(NEXT(PTR),NULL)				:S(I_LOOP1)
      PTR = NEXT(PTR)					:(I_SCAN)
MATCHED
* * *
*   GREAT.  We found an identity between VAR and NEW that makes the
* two equations identical.  Now we can check if there already exists
* such an identity.  If so, return victorious, and if not, continue
* where we left off.
* * *
      ID_NORM = NORMALIZE(VAR '-' NEW)
*  CHANGES ID July 4, 1986
*  Check if FINDANTE(ID_NORM) first
      FINDANTE(ID_NORM) 				:F(I_CONT)
*  CHANGES ID July 2, 1986   CNH
*  repeat the identity at this level
      F1_TMP = G_CURLINE + 1
      ADDPRII(ID_NORM, PRLINE(FINDANTE(ID_NORM)) ',R')
*  changes Aug 5, 1986	CNH change addproof to addprii
      F1_TMP = GT(F1_TMP,G_CURLINE) PRLINE(FINDANTE(ID_NORM))
* CHANGE ID July 2, 1986 CNH
*  change from FINDANTE(ID_NORM)
      IDTEST = F1_TMP					:(RETURN)
ENDIDTEST
-STITLE Function: SIMPLEID(FORMULA)
*FRAME
* Function: SIMPLEID
* Date: Mon Jun 18 11:25:03 1984
* Written by: Randy Kopach
* Last modification: Mon Jun 18 11:25:03 1984
* Purpose:
*   SIMPLEID will go through the identity ring looking for those
* identities in which one of the variables occurs in the formula.
* If found, it calls TESTID with the formula and identity.  If
* TESTID succeeds, SIMPLEID will return the antecedent line pointer
* for the formula.  Otherwise it keeps going through the identity
* ring.  If not antecedent line formula is found after completely
* searching the identity ring, SIMPLEID fails.
*
* Calling sequence:
*	result = SIMPLEID(FORMULA)
* Parameters: 
*	FORMULA -	The formula to use
* On return:
*	RETURN -	Equivelent antecedent line formula found
*	FRETURN -	No equivelent antecedent line formula found
*
* Local variables: 
*	PTR -		Pointer to scan identity ring
*	G_LOP - 	left side of identity formula
*	G_OP -		identity operator
*	G_ROP - 	right side of identity formula
      pDEFINE('SIMPLEID(FORMULA)PTR,G_LOP,G_OP,G_ROP',,'STRING','ANTE_PTRS')
+							:(ENDSIMPLEID)
SIMPLEID
      EQ(P_NOID)				    :S(FRETURN)
      PTR = FINDIDPT(G_GOALS)
      IDENT(G_IDSTACK,PTR)			      :F(SI_SKIP)
      IDENT(G_IDSTACK,BLINK(FINDIDPT2(G_GOALS)))      :S(FRETURN)
      PTR = BLINK(FINDIDPT2(G_GOALS))
SI_SKIP
      SIMPLEID = ANTE_PTRS(NULL,NULL)
SI_SCAN
      LINE(PTR) G_IDSPLIT		     :F(FRETURN)
      FORMULA G_LOP					:S(SI_TEST)
      FORMULA G_ROP					:S(SI_TEST)
SI_CONT
+     PTR = FLINK(PTR)
      DIFFER(PTR,G_IDSTACK)				:S(SI_SCAN)F(FRETURN)
SI_TEST
      FORM_PTR(SIMPLEID) = TESTID(FORMULA,LINE(PTR))	:F(SI_CONT)
      IDENT_PTR(SIMPLEID) = FINDANTE(LINE(PTR)) 	  :(RETURN)
ENDSIMPLEID
-STITL Function: FINDNEGID()
      pDEFINE('FINDNEGID()NEG_PTR,G_OP,G_ROP,FORM,VARLIST,FNI_VAR,FNI_NEW,'
+			 'C_LIST,V_LIST,PV_LIST,A_INDEX,GLOBAL_VARS,'
+			 'IDENT,C_OK,V_OK,PV_OK',,,'NULL')
+							:(ENDFINDNEGID)
FINDNEGID
      EQ(P_NOID)				    :S(FRETURN)
*     OUTPUT = 'FINDNEGID called...'
      NEG_PTR = BLINK(G_NSTACK)
*     OUTPUT = 'CONVERTING ARRAYS...' 
      C_LIST = CONVERT(G_CLIST,'ARRAY') 		:F(FNI_V)
      C_OK = G_TRUE
FNI_V V_LIST = CONVERT(G_VLIST,'ARRAY') 		:F(FNI_P)
      V_OK = G_TRUE
FNI_P PV_LIST = CONVERT(G_PREMVLIST,'ARRAY')		:F(FNI_DONEARRAYS)
      PV_OK = G_TRUE
FNI_DONEARRAYS
*     OUTPUT = 'DONE CONVERTING ARRAYS...'
*  OUTPUT = 'C_OK V_OK PV_OK ' C_OK ',' V_OK ',' PV_OK
      A_INDEX = 1
      GLOBAL_VARS = NULL
      IDENT(C_OK)					:S(END_FNI_C_LOOP)
FNI_C_LOOP
+     GLOBAL_VARS = GLOBAL_VARS C_LIST<A_INDEX,1>	:F(END_FNI_C_LOOP)
*  OUTPUT = 'C ' C_LIST<A_INDEX,1> C_LIST<A_INDEX,2>
*  OUTPUT = 'A_INDEX' A_INDEX
      A_INDEX = A_INDEX + 1				:(FNI_C_LOOP)
END_FNI_C_LOOP
+     IDENT(V_OK)					:S(END_FNI_V_LOOP)
      A_INDEX = 1
FNI_V_LOOP
+     GLOBAL_VARS = GLOBAL_VARS V_LIST<A_INDEX,1>	:F(END_FNI_V_LOOP)
*  OUTPUT = 'V ' V_LIST<A_INDEX,1> V_LIST<A_INDEX,2>
*  OUTPUT = 'A_INDEX' A_INDEX
      A_INDEX = A_INDEX + 1				:(FNI_V_LOOP)
END_FNI_V_LOOP
      IDENT(PV_OK)					:S(FNI_LOOP)
      A_INDEX = 1
FNI_PV_LOOP
+     GLOBAL_VARS = GLOBAL_VARS PV_LIST<A_INDEX,1>	:F(FNI_LOOP)
*  OUTPUT = 'PV ' PV_LIST<A_INDEX,1> PV_LIST<A_INDEX,2>
*  OUTPUT = 'A_INDEX' A_INDEX
      A_INDEX = A_INDEX + 1				:(FNI_PV_LOOP)
FNI_LOOP
*     OUTPUT = 'DONE INDEXING VARIABLE LISTS...'
      IDENT(NEG_PTR,G_NSTACK)				:S(FRETURN)
*     OUTPUT = 'LINE(NEG_PTR) = ' LINE(NEG_PTR)
      LINE(NEG_PTR) G_USPLIT				:F(CONCHECK)
      FORM = G_ROP
      VARLIST = NULL
*     OUTPUT = 'STRIPPING VARIABLES...'
FNI_STRIP
+     G_ROP G_BT $ FNI_VAR = NULL			:F(FNI_GOTVARS)
      VARLIST FNI_VAR					:S(FNI_STRIP)
      VARLIST = VARLIST FNI_VAR 			:(FNI_STRIP)
FNI_GOTVARS
FNI_SUB_LOOP1
*     OUTPUT = 'IN LOOP1.  VARLIST = ' VARLIST
      VARLIST G_BT $ FNI_VAR = NULL			:F(FNI_NEXT)
*  OUTPUT = 'GLOBAL_VARS ' GLOBAL_VARS
      NEWVARS = GLOBAL_VARS
FNI_SUB_LOOP2
*     OUTPUT = 'IN LOOP2. NEWVARS = ' NEWVARS
      NEWVARS G_BT $ FNI_NEW = NULL			:F(FNI_SUB_LOOP1)
*     OUTPUT = 'TESTID being called with ' FORM ' and ' FNI_VAR '-' FNI_NEW
      LINE_PTR = TESTID(FORM,FNI_VAR '-' FNI_NEW)	:F(FNI_SUB_LOOP2)
*     OUTPUT = 'TESTID found the identity ' FNI_VAR '-' FNI_NEW ' to work...'
      IDENT = '~' FNI_VAR '-' FNI_NEW
      IDENT = NORMALIZE(IDENT)
      FINDANTE(IDENT)					:S(FNI_SUB_LOOP2)
*  changes Aug 5, 1986	CNH incorporate onestep into addproof
      ADDPROOF(IDENT,PLINE(NEG_PTR) ',' PRLINE(LINE_PTR) ',NEGID')
+						:S(RETURN)F(FNI_SUB_LOOP2)
FNI_NEXT
*     OUTPUT = 'ON TO THE NEXT FORMULA...'
      NEG_PTR = BLINK(NEG_PTR)				:(FNI_LOOP)
ENDFINDNEGID
-STITL Function: NORMALIZE(FORMULA)
*FRAME
* Function: NORMALIZE
* Date: Fri Jul 20 13:30:57 1984
* Written by: Randy Kopach
* Last modification: Fri Jul 20 13:30:57 1984
* Purpose: This function will lexically normalize any identity
*  relationships in a formula and return the new version of
*  the formula as a string.
*
* Calling sequence:
*	result = NORMALIZE(FORMULA)
* Parameters: 
*	FORMULA -	formula to be normalized
* On return:
*	RETURN -	normalized formula (of course)
*
* Local variables: 
*	G_LOP - 	left term of identity
*	G_OP -		identity operator
*	G_ROP - 	right term of identity
*	NR_IDENT -	identity relation
*	NR_P -		starting string position of identity
*	NR_APPEND -	formula manipulation variable
*
      pDEFINE('NORMALIZE(FORMULA)G_LOP,G_OP,G_ROP,NR_IDENT,NR_P,NR_APPEND',,
+				'STRING','STRING')	:(ENDNORMALIZE)
NORMALIZE
+     FORMULA @NR_P (G_BT '-' G_BT) . NR_IDENT		:F(NR_DONE)
      NR_IDENT G_IDSPLIT				:F(NR_BOOM)
      LGT(G_LOP,G_ROP)					:S(NR_REVERSE)
      FORMULA LEN(NR_P + 5) $ NR_APPEND = NULL		:F(NR_BOOM)
      NORMALIZE = NORMALIZE NR_APPEND			:(NORMALIZE)
NR_REVERSE
+     FORMULA LEN(NR_P) $ NR_APPEND LEN(5) = NULL	:F(NR_BOOM)
      NORMALIZE = NORMALIZE NR_APPEND G_ROP '-' G_LOP	:(NORMALIZE)
NR_DONE
      NORMALIZE = NORMALIZE FORMULA			:(RETURN)
NR_BOOM
+     OUTPUT = 'Internal error in NORMALIZE routine'	:(CONCHECK)
ENDNORMALIZE
-STITL Function: FINDMPS()
*FRAME
* Function: FINDMPS 
* Date: Sun Jul 25 17:30:00 1982 
* Last modification: July, 1984
* Purpose: 
*  Try to apply modus ponens 
*   i.e.  (P>Q) , P -> Q 
*   also looks for:
*	  (a-a>P), a-a -> P
*	  ((Ex)x-a>P), (Ex)x-a -> P
* 
* Calling sequence: 
*	FINDMPS() 
* Parameters: None 
* On return: 
*	RETURN - ONESTEP succeeded.
*	FRETURN - otherwise.
* 
* Local variables: 
*	FM_Z - Return from findante on left operator.
*	FM_X - Back -link in ring.
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
* 
      pDEFINE('FINDMPS()FM_Z,FM_X,G_LOP,G_ROP,G_OP',,,'NULL')
+						       :(ENDFINDMPS)
FINDMPS
+     FM_X   = BLINK(G_IMPLSTACK)
FM_LOOP
+     IDENT(G_IMPLSTACK,FM_X)				 :S(FRETURN)
      LINE(FM_X) G_BSPLIT				:F(CONCHECK)
* * *
*  Check for (a-a>P)
* * *
      REFINST(G_LOP)					:F(FM_NOPAIR)
      FINDANTE(G_ROP)					:S(FM_CONT)
      ADDPROOF(G_ROP, 'REFL,' PLINE(FM_X) ',MP')      :F(FM_CONT)S(RETURN)
*  changes Aug 5, 1986	CNH incorporate onestep into addproof
*  CNH	3rd parameter to ADDPROOF removed
* * *
*  Check for ((Ex)x-a>P) or ((Ex)a-x>P) or ((Ex)x-x>P)
* * *
FM_NOPAIR
+     EGREF(G_LOP)					:F(FM_NOIDENT)
      FINDANTE(G_ROP)					:S(FM_CONT)
      ADDPROOF(G_ROP, 'REFL,EG,' PLINE(FM_X) ',MP')   :F(FM_CONT)S(RETURN)
*  changes Aug 5, 1986	CNH incorporate onestep into addproof
*  CNH	3rd parameter to ADDPROOF removed
* * *
*  Finally, check for (P>Q)
* * *
FM_NOIDENT
+     FM_Z = FINDANTE(G_LOP)				 :F(FM_CONT)
      FINDANTE(G_ROP)					 :S(FM_CONT)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(G_ROP,PLINE(FM_X) ',' PRLINE(FM_Z) ',MP')
+						:F(FM_CONT)S(RETURN)
*  changes Aug 5, 1986	CNH incorporate onestep into addproof
FM_CONT
+     FM_X = BLINK(FM_X)				  :(FM_LOOP)
ENDFINDMPS
-STITL Function: FINDMTP()
*FRAME
* Function: FINDMTP 
* Date: Sun Jul 25 17:30:36 1982 
* Last modification: July, 1984
* Purpose: 
*  Try to apply modus tollens ponens 
*    i.e. (P+Q) , ~P -> Q 
*    OR   (P+Q) , ~Q -> P 
*    OR   (P+~Q), Q -> P 
*    OR   (~P+Q), P -> Q 
*    OR   (~a-a+P), ~a-a -> P
*    OR   (P+~a-a), ~a-a -> P
*    OR   (~(Ex)x-a+P), (Ex)x-a -> P
*    OR   (P+(Ex)x-a), (Ex)x-a -> P
* 
* Calling sequence: 
*	FINDMTP() 
* Parameters: None 
* On return: 
*	RETURN - ONESTEP succeeded
*	FRETURN - Otherwise.
* 
* Local variables: 
*	FP_Z - Return from FINDANTE onOP.
*	FP_X - Back-link in ring.
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
*	FP_LOP1 - First left op.
*	FP_ROP1 - First right op.
* 
      pDEFINE('FINDMTP()FP_Z,FP_X,G_LOP,G_OP,G_ROP,'
+	      'FP_LOP1,FP_ROP1',,,'NULL')
+						       :(ENDFINDMTP)
FINDMTP
+     FP_X = BLINK(G_ORSTACK)
FP_LOOP
+     IDENT(FP_X,G_ORSTACK)				 :S(FRETURN)
      LINE(FP_X) G_BSPLIT				:F(CONCHECK)
      FP_Z = FINDANTE('~' G_LOP)			 :F(FP_TRY2)
      FINDANTE(G_ROP)					 :S(FP_TRY2)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(G_ROP,PRLINE(FP_Z) ',' PLINE(FP_X) ',MTP')
+						:F(FP_TRY2)S(RETURN)
*  changes Aug 5, 1986	CNH incorporate onestep into addproof
FP_TRY2
+     FP_Z = FINDANTE('~' G_ROP)			 :F(FP_TRY3)
      FINDANTE(G_LOP)					 :S(FP_TRY3)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(G_LOP,PRLINE(FP_Z) ',' PLINE(FP_X) ',MTP')
+						:F(FP_CONT)S(RETURN)
*  changes Aug 5, 1986	CNH incorporate onestep into addproof
FP_TRY3
+     FP_LOP1 = G_LOP
      FP_ROP1 = G_ROP
      FP_LOP1 G_USPLIT					 :F(FP_TRY4)
      FP_Z = FINDANTE(G_ROP)				 :F(FP_TRY4)
      FINDANTE(FP_ROP1) 				 :S(FP_TRY4)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(FP_ROP1,PRLINE(FP_Z) ',' PLINE(FP_X) ',MTP')
+						:F(FP_CONT)S(RETURN)
*  changes Aug 5, 1986	CNH incorporate onestep into addproof
FP_TRY4
+     FP_ROP1 G_USPLIT					 :F(FP_TRY5)
      FP_Z = FINDANTE(G_ROP)				 :F(FP_TRY5)
      FINDANTE(FP_LOP1) 				 :S(FP_TRY5)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(FP_LOP1,PRLINE(FP_Z) ',' PLINE(FP_X) ',MTP')
+						:F(FP_CONT)S(RETURN)
*  changes Aug 5, 1986	CNH incorporate onestep into addproof
FP_TRY5
+     FP_LOP1 G_IDSPLIT 				:F(FP_TRY6)
      G_LOP G_USPLIT					:F(FP_TRY6)
      REFINST(SUBSTR(FP_LOP1,2,SIZE(FP_LOP1) - 1))	:F(FP_TRY6)
      FINDANTE(FP_ROP1) 				:S(FP_TRY6)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(FP_ROP1, 'REFL,' PLINE(FP_X) ',MTP')
+						 :F(FP_CONT)S(RETURN)
*  changes Aug 5, 1986	CNH incorporate onestep into addproof
FP_TRY6
+     FP_ROP1 G_IDSPLIT 				:F(FP_TRY7)
      G_LOP G_USPLIT					:F(FP_TRY7)
      REFINST(SUBSTR(FP_ROP1,2,SIZE(FP_ROP1) - 1))	:F(FP_TRY7)
      FINDANTE(FP_LOP1) 				:S(FP_TRY7)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(FP_LOP1, 'REFL,' PLINE(FP_X) ',MTP')
+						 :F(FP_CONT)S(RETURN)
*  changes Aug 5, 1986	CNH incorporate onestep into addproof
FP_TRY7
+     FP_LOP1 G_USPLIT					:F(FP_TRY8)
      EGREF(SUBSTR(FP_LOP1,2,SIZE(FP_LOP1) - 1))	:F(FP_TRY8)
      FINDANTE(FP_ROP1) 				:S(FP_TRY8)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(FP_ROP1, 'REFL,EG,' PLINE(FP_X) ',MTP')
+						 :F(FP_CONT)S(RETURN)
*  changes Aug 5, 1986	CNH incorporate onestep into addproof
FP_TRY8
+     FP_ROP1 G_USPLIT					:F(FP_CONT)
      EGREF(SUBSTR(FP_ROP1,2,SIZE(FP_ROP1) - 1))	:F(FP_CONT)
      FINDANTE(FP_LOP1) 				:S(FP_CONT)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(FP_LOP1, 'REFL,EG,' PLINE(FP_X) ',MTP')
+						 :F(FP_CONT)S(RETURN)
*  changes Aug 5, 1986	CNH incorporate onestep into addproof
FP_CONT
+     FP_X = BLINK(FP_X)				  :(FP_LOOP)
ENDFINDMTP
-STITL Function: FINDMTS()
*FRAME
* Function: FINDMTS 
* Date: Sun Jul 25 17:30:16 1982 
* Last modification: July, 1984
* Purpose: 
*  Try to apply Modus tollens 
*   ie (P>Q) , ~Q -> ~P 
* 
* Calling sequence: 
*	FINDMTS() 
* Parameters: None 
* On return: 
*	RETURN - ONESTEP succeeded
*	FRETURN - Otherwise.
* 
* Local variables: 
*	FS_Z - Return from FINDANTE on OP.
*	FS_X - Back-link in ring.
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
* 
* Description: 
      pDEFINE('FINDMTS()FS_Z,FS_X,G_LOP,G_OP,G_ROP,FS_LOP1,FS_ROP1',,,'NULL')
+						       :(ENDFINDMTS)
FINDMTS
+     FS_X   = BLINK(G_IMPLSTACK)
FS_LOOP
+     IDENT(FS_X,G_IMPLSTACK)				 :S(FRETURN)
      LINE(FS_X) G_BSPLIT				:F(CONCHECK)
      FS_Z = FINDANTE('~' G_ROP)			 :F(FS_TRYID)
      FINDANTE('~' G_LOP)				 :S(FS_TRYID)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF('~' G_LOP,PRLINE(FS_Z) ',' PLINE(FS_X) ',MT')
+						 :F(FS_CONT)S(RETURN)
*  changes Aug 5, 1986	CNH incorporate onestep into addproof
FS_TRYID
+     FS_LOP1 = G_LOP
      FS_ROP1 = G_ROP
      FS_ROP1 G_IDSPLIT 				:F(FS_CONT)
      G_LOP G_USPLIT					:F(FS_CONT)
      REFINST(SUBSTR(FS_ROP1,2,SIZE(FS_ROP1) - 1))	:F(FS_CONT)
      FINDANTE('~' FS_LOP1)				:S(FS_CONT)
*  changes Aug 5, 1986	Third parameter to addproof removed
*  changes Aug 5, 1986	CNH incorporate onestep into addproof
      ADDPROOF('~' FS_LOP1, 'REFL,' PLINE(FS_X) ',MT')
+						 :F(FS_CONT)S(RETURN)
FS_CONT
+     FS_X = BLINK(FS_X)				  :(FS_LOOP)
ENDFINDMTS
-STITLE Function:  FINDNEGS(FN_X)
*FRAME
* Function:  FINDNEGS
* Date:  Wed Aug 13, 1986
* Author:  CNH
* Purpose:	
*  To simplify the double negation pointed to by FN_X if possible
* Calling Sequence:
*  FINDNEGS(FN_X)
* Parameters:
*  FN_X : pointer into the doubly linked circular list
* On return:
*  RETURN - The show has been closed
*  FRETURN - Otherwise
* Local variables:
* G_ROP - right operand as given by B_SPLIT
* G_OP	- Main operator as given by B_SPLIT
* G_LOP - Left Operand as given by B_SPLIT
*
      pDEFINE('FINDNEGS(FN_X)G_ROP,G_OP,G_LOP',,,'NULL')
+						      :(ENDFINDNEGS)
FINDNEGS
      LINE(FN_X) FENCE '~~' REM . G_ROP 	      :F(FRETURN)
      FINDANTE(G_ROP)				      :S(FRETURN)
      ADDPROOF(G_ROP,PLINE(FN_X) ',DN') 	      :S(RETURN)F(FRETURN)
ENDFINDNEGS
-STITLE Function:  FINDQNS(FQ_X)
*FRAME
* Function:  FINDQNS
* Date:  Wed Aug 13, 1986
* Author:  CNH
* Purpose:	
*  To simplify ~(Ax)@ to (Ex)~@ and ~(Ex)@ to (Ax)~@
* Calling Sequence:
*  FINDQNS(FQ_X)
* Parameters:
*  FQ_X : pointer into the doubly linked circular list
* On return:
*  RETURN - The show has been closed
*  FRETURN - Otherwise
* Local variables:
*  G_ROP - right operator as given by G_SPLIT
*  G_OP  - Main operator as given by G_SPLIT
*  G_LOP - left operand as given by F_SPLIT
*  FQ_L - Line field of data structure
*  FQ_NO - New operator with replacement
*  FQ_NL - New line created by switch
*
      pDEFINE('FINDQNS(FQ_X)G_ROP,G_OP,G_LOP,FQ_L,FQ_NO,FQ_NL',,,
+     'NULL')					      :(ENDFINDQNS)
FINDQNS
      FQ_L = LINE(FQ_X)
      FQ_L POS(0) '~' = 			      :F(CONCHECK)
      FQ_L G_QSPLIT				      :F(FRETURN)
      FQ_NO = REPLACE(G_OP,'AE','EA')
      FQ_NL = '(' FQ_NO G_VAR ')~' G_ROP
      FINDANTE(FQ_NL)				      :S(FRETURN)
      ADDPROOF(FQ_NL,PLINE(FQ_X) ',QN') 	      :S(RETURN)F(FRETURN)
ENDFINDQNS
-TITLE THINKER - look for patterns in antecedent lines
*FRAME
* Each of these routines looks around one of the rings 
* for a special pattern of some sort.  If it finds that 
* pattern it returns the line.	Often they will pre-check 
* to see if a given line has been used as a goal before. 
-STITL Function: SEARCHARROW()
*FRAME
* Function: SEARCHARROW 
* Date: Sun Jul 25 17:31:35 1982 
* Last modification: June, 1983
* Purpose: 
*  Look for (P>Q) where Q has never occured. 
* 
* Calling sequence: 
*	result = SEARCHARROW() 
* Parameters: None 
* On return: 
*	RETURN - Appropriate line was found and returned.
*	FRETURN - No line found. 
* 
* Local variables: 
*	SA_X - Back link in circular list.
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
* 
      pDEFINE('SEARCHARROW()SA_X,G_LOP,G_OP,G_ROP',,,'STRING')
+						   :(ENDSEARCHARROW)
SEARCHARROW
+     SA_X   = BLINK(G_IMPLSTACK)
SA_LOOP
+     IDENT(SA_X,G_IMPLSTACK)				 :S(FRETURN)
      SEARCHARROW = LINE(SA_X)
      SEARCHARROW G_BSPLIT				:F(CONCHECK)
      FINDANTE(G_ROP)					 :S(SA_CONT)
      TESTGOAL(G_LOP)				:F(SA_CONT)S(RETURN)
SA_CONT
+     SA_X = BLINK(SA_X)				  :(SA_LOOP)
ENDSEARCHARROW
-STITL Function: SEARCHWEDGE()
*FRAME
* Function: SEARCHWEDGE 
* Date: Sun Jul 25 17:31:56 1982 
* Last modification: June, 1983
* Purpose: 
* LOOK FOR (P+Q) WHERE NEITHER P OR Q IS AN ANTECEDENT LINE 
* FOR ONE OF P OR Q HAS NOT OCCURED AS A GOAL. 
* 
* Calling sequence: 
*	result = SEARCHWEDGE() 
* Parameters: None 
* On return: 
*	RETURN - Appropriate line was found and returned.
*	FRETURN - No line found.
* 
* Local variables: 
*	SW_X - Back link in circular list
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
* 
      pDEFINE('SEARCHWEDGE()SW_X,G_LOP,G_OP,G_ROP',,,'STRING')
+						   :(ENDSEARCHWEDGE)
SEARCHWEDGE
+     SW_X = BLINK(G_ORSTACK)
SW_LOOP
+     IDENT(SW_X,G_ORSTACK)				 :S(FRETURN)
      SEARCHWEDGE = LINE(SW_X)
      SEARCHWEDGE G_BSPLIT				:F(CONCHECK)
      ~FINDANTE(G_LOP) ~FINDANTE(G_ROP) 		 :F(SW_CONT)
      TESTGOAL(G_LOP)					  :S(SW_ADD)
      TESTGOAL(G_ROP)					 :F(SW_CONT)
SW_ADD
+							   :(RETURN)
SW_CONT
+     SW_X = BLINK(SW_X)				  :(SW_LOOP)
ENDSEARCHWEDGE
-STITL Function: SEARCHCONTRA()
*FRAME
* Function: SEARCHCONTRA 
* Date: Sun Jul 25 17:27:43 1982 
* Last modification: May, 1984
* Purpose: 
*  Try and find a P and ~P in the antecedent lines.
*  Also look for ~x1-x1
* 
* Calling sequence: 
*	SEARCHCONTRA() 
* Parameters: None 
* On return: 
*	RETURN - Contradiction has been shown.
*	FRETURN - Nothing has been gained.
* 
* Local variables: 
*	SC_X - Back link in circular list.
*	SC_L - Temporary variable containing current line.
*	SC_Z - Return from FINDANTE.
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
* 
*  CHANGES:  June 20, 1986  CNH
*  Two new local variables are added CRIT1,CRIT2  they are
*  explained in the next notation
  pDEFINE('SEARCHCONTRA()SC_X,SC_L,SC_Z,G_LOP,G_ROP,G_OP,CRIT1,CRIT2',,,
+	      'STRING|NULL')			  :(ENDSEARCHCONTRA)
SEARCHCONTRA
+     SC_X = BLINK(G_NSTACK)
SC_LOOP
+     IDENT(SC_X,G_NSTACK)				 :S(FRETURN)
      SC_L = LINE(SC_X)
      SC_L G_USPLIT					:F(CONCHECK)
      SC_Z = FINDANTE(G_ROP)				 :F(SC_CHKID)
*  CHANGE:  June 20, 1986   CNH
*  crit1,crit2 are used to determine the critical
*  lines for the SHOW
*  crit1 is the line number of the first part
*  of the contradiction.  It is either earlier
*  in the SHOW or is added in (either PLINE(SC_X)
*  or the present line + 1
      CRIT1 = G_CURLINE + 1
*  changes Aug 5, 1986	CNH addproof to addprii
      ADDPRII(SC_L,PLINE(SC_X) ',R')
      CRIT1 = GT(CRIT1,G_CURLINE) PLINE(SC_X)
*  crit2 is the line number of the second part
*  of the contradiction.  It is either earlier
*  in the SHOW or is added in (either PRLINE(SC_Z)
*  or the present line + 1
      CRIT2 = G_CURLINE + 1
      ADDPRII(G_ROP,PRLINE(SC_Z) ',R')
*  changes Aug 5, 1986	CNH change addproof to addprii
      CRIT2 = GT(CRIT2,G_CURLINE) PRLINE(SC_Z)
      SHOWN(CRIT1 ',' CRIT2)				     :(RETURN)
SC_CHKID
      SC_ROP = G_ROP
      SC_ROP G_IDSPLIT					  :F(SC_CONT)
      LEQ(G_LOP,G_ROP)					  :F(SC_CONT)
* oh a ~x-x
      CRIT1 = G_CURLINE + 1
      ADDPRII(SC_L,PLINE(SC_X) ',R')
*  changes Aug 5, 1986	CNH change addproof to addprii
      CRIT1 = GT(CRIT1,G_CURLINE) PLINE(SC_X)
      ADDPRII(SC_ROP,'REFL')
*  changes Aug 5, 1986	CNH change addproof to addprii
      SHOWN(G_CURLINE ',' CRIT1)			      :(RETURN)
SC_CONT
+     SC_X = BLINK(SC_X)				  :(SC_LOOP)
ENDSEARCHCONTRA
-STITL Function: SEARCHNEG(sn_x)
*FRAME
* Function: SEARCHNEG(sn_x)
* Date: March 23, 1987
* Author:  CNH
* Purpose:
*   This is a new (and improved?) searchneg to work with the	
*   paralleliser.  This routine will go throught the negation
*   ring looking for the first line of the form ~(p+q) ~(p&q)
*   ~(p=q) or ~(p>q) where the unnegated form can be set as a
*   goal unless the line is ~(p+q) in which case where either
*   p (or q) can be set as a goal or p (or q) is an antecedent
*   line.  Note:  It starts the search one position back from
*   the location it is given.
* Calling sequence: 
*	result = SEARCHNEG(sn_x)
* Parameters: sn_x   a pointer into the negation circle
* On return: 
*	RETURN - Appropriate line was found and returned.
*	FRETURN - No line was found.
* 
* Local variables: 
*	SN_Z - Temporary right operator of line.
*	SN_L - Temporary variable containing current line.
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
* 
      pDEFINE('SEARCHNEG(SN_X)SN_Z,SN_L,G_LOP,G_OP,G_ROP',,,
+	      'STRING') 			     :(ENDSEARCHNEG)
SEARCHNEG
SN_LOOP
      $SN_X = BLINK($SN_X)
      IDENT(G_NSTACK,$SN_X)				 :S(FRETURN)
      SN_L = LINE($SN_X)
      SN_L G_USPLIT					:F(SN_LOOP)
      SN_Z = G_ROP					:F(SN_LOOP)
      SN_Z G_BSPLIT					:F(SN_LOOP)
      G_OP '+'						:S(SN_OR)
      G_OP '>'						:S(SN_OTHER)
      G_OP '='						:S(SN_OTHER)
      G_OP '&'						:S(SN_OTHER)F(SN_LOOP)
SN_OR
      TESTGOAL(G_LOP)					:S(SN_RET)
      TESTGOAL(G_ROP)					:S(SN_RET)
      FINDANTE(G_ROP)					:S(SN_RET)
      FINDANTE(G_LOP)					:S(SN_RET)F(SN_LOOP)
SN_OTHER
      TESTGOAL(SN_Z)					:F(SN_LOOP)
SN_RET
      SEARCHNEG = SN_L					:(RETURN)
ENDSEARCHNEG
-TITLE THINKER - quantifier support functions
-STITL Function: FINDUI(FU_VARLIST)
*FRAME
* Function: FINDUI 
* Date: Sun Jul 25 17:24:18 1982 
* Last modification: June, 1983
* Purpose: 
*  This function is used by FINDUIA and FINDUIB. 
*  Varlist is a table.	Each universally quantified antecedent 
*  line is UIed using each variable in turn in varlist. 
*  A variable in varlist is considered to be used if the 
*  entry for that variable is NONnull. 
* 
* Calling sequence: 
*	FINDUI(FU_VARLIST) 
* Parameters: 
*	FU_VARLIST - Table of varibles.
* On return: 
*	RETURN - Universal Instatiation successfull
*	FRETURN - Nothing was gained.
* 
* Local variables: 
*	FU_VARLISTA - Array form of FU_VARLIST
*	FU_X - Back link in circular list.
*	FU_L - Temporary variable containing current line.
*	FU_TEMPL - Return from MAKEQTPLATE.
*	FU_I - General index.
* 
* Description: 
      pDEFINE('FINDUI(FU_VARLIST)'
+	      'FU_VARLISTA,FU_X,FU_L,FU_TEMPL,FU_I',,
+	      'TABLE','NULL')				:(ENDFINDUI)
FINDUI
      FU_VARLISTA = CONVERT(FU_VARLIST,'ARRAY') 	 :F(FRETURN)
*     OUTPUT = 'IN FINDUI'
      FU_X = BLINK(G_UQSTACK)
FU_LOOP
      IDENT(FU_X,G_UQSTACK)				 :S(FRETURN)
      FU_L = LINE(FU_X)
*     OUTPUT = 'IN FU_LOOP  LINE: ' FU_L
      FU_TEMPL = MAKEQTPLATE(FU_L)
      FU_TEMPL '(A@)' = 				:F(CONCHECK)
      FU_I = 0
FU_VARLOOP
      FU_I = FU_I + 1
*     OUTPUT = 'IN FU_VARLOOP FU_I: ' FU_I
      FU_L = FU_TEMPL
      FU_NEWVAR = FU_VARLISTA<FU_I,1>			 :F(FU_CONT)
*     OUTPUT = 'STILL FU_VARLOOP FU_NEWVAR: ' FU_NEWVAR
      IDENT(FU_VARLISTA<FU_I,2>)		      :S(FU_VARLOOP)
* check to see if we should be prohibiting this formula
      ?G_PRMAT<PLINE(FU_X),6>				:F(CONCHECK)
*     OUTPUT = 'JUST ABOUT TO CHECK G_PRMAT<*,6>' PLINE(FU_X) ',' G_PRMAT<PLINE(FU_X),6>
      IDENT(G_PRMAT<PLINE(FU_X),6>)		       :S(FU_VARREP)
*     OUTPUT = 'JUST ABOUT TO CHECK IF GT 0: ' G_PROHIBLIST<FU_NEWVAR>
      GT(G_PROHIBLIST<FU_NEWVAR>,0)		      :S(FU_VARLOOP)
FU_VARREP
      FU_L '@' = FU_NEWVAR			       :S(FU_VARREP)
      FU_L = NORMALIZE(FU_L)
*     OUTPUT = 'NEW LINE? ' FU_L
      FINDANTE(FU_L)				      :S(FU_VARLOOP)
*     OUTPUT = 'DIDNT FIND ALREADY SO ADD IT'
*  changes Aug 5, 1986	CNH TAKE out duplicate addproofs
*  two addproofs were used having differing 5th parameters
*  but since that parameter is now gone only one call is
*  required.  The call the onestep is incorporated in
*  addproof
      FU_L QSPLIT					:F(FU_NOTUI)
      LEQ(G_VAR,'A')					:F(FU_NOTUI)
      ADDPROOF(FU_L,PLINE(FU_X) ',UI')		 :S(RETURN)F(FU_VARLOOP)
*  CNH	3rd parameter to ADDPROOF removed
*							 :(FU_DOONE)
FU_NOTUI
     ADDPROOF(FU_L,PLINE(FU_X) ',UI')
+					     :S(RETURN)F(FU_VARLOOP)
*  CNH	3rd parameter to ADDPROOF removed
*FU_DOONE
FU_CONT
+     FU_X = BLINK(FU_X)				  :(FU_LOOP)
ENDFINDUI
-STITL Function: FINDUIA()
*FRAME
* Function: FINDUIA 
* Date: Sun Jul 25 17:24:41 1982 
* Last modification: June, 1983
* Purpose: 
*  To try to make Universal Instantiation work with different 
*  variable lists.
* 
* Calling sequence: 
*	FINDUIA() 
* Parameters: None 
* On return: 
*	RETURN - FINDUI succeeded with one of Free varible list,
*		 free variables in premises, or CLIST
*	FRETURN - FINDUI failed.
* 
* Local variables: None 
* 
      pDEFINE('FINDUIA()',,,'NULL')		       :(ENDFINDUIA)
FINDUIA
+     FINDUI(G_FVLIST)					  :S(RETURN)
      FINDUI(G_PREMFVLIST)				  :S(RETURN)
      FINDUI(G_CLIST)				:S(RETURN)F(FRETURN)
ENDFINDUIA
-STITL Function: MAKEEG(ME_FORM)
*FRAME
* Function: MAKEEG 
* Date: Sun Jul 25 17:23:16 1982 
* Last modification: June, 1983
* Purpose: 
*  ME_FORM is an existentially quantified formula, look to 
*  se if there is an instance of it in antelines 
* 
* Calling sequence: 
*	result = MAKEEG(ME_FORM) 
* Parameters: 
*	ME_FORM - Formula on which to perform Existential
*		  Generalization.
* On return: 
*	RETURN - EG worked. Returns antecedent line.
*	FRETURN - EG failed.
* 
* Local variables: 
*	ME_TPLATE - Return from MAKEQTPLATE.
*	ME_LINE - Return from FINDTEMP.
* 
* Description: 
      pDEFINE('MAKEEG(ME_FORM)ME_TPLATE,ME_LINE',,
+	      'STRING','STRING')			:(ENDMAKEEG)
MAKEEG
+     ME_TPLATE = MAKEQTPLATE(ME_FORM)
      ME_TPLATE '(E@)' =				:F(CONCHECK)
      ME_LINE = FINDTEMP(ME_TPLATE)			 :F(FRETURN)
      MAKEEG = FINDANTE(ME_LINE)			   :(RETURN)
ENDMAKEEG
-STITL Function: MAKEUI(MU_FORM)
*FRAME
* Function: MAKEUI 
* Date: Sun Jul 25 17:23:38 1982 
* Last modification: June, 1983
* Purpose: 
* form is a run of the mill formula.  MAKEUI suceeds and 
* returns a formula if form could come from an antecedent 
* line which is universally quantified by UI. 
* 
* Calling sequence: 
*	result = MAKEUI(MU_FORM) 
* Parameters: 
*	MU_FORM - Formula on which to perform Universal
*		  Instantiation.
* On return: 
*	RETURN - UI worked. Relevent antecedent line returned.
*	FRETURN - UI failed.
* 
* Local variables: 
*	G_FVL - Free variable table.
*	MU_FVLA - Array form of G_FVL
*	MU_TEMPL - Return from MAKEQTPLATE.
*	MU_FREEVAR - I'th element of MU_FVLA
*	MU_I - Index into MU_FVLA
*	MU_LINE - Return from FINDTEMP.
* 
* Description: 
      pDEFINE('MAKEUI(MU_FORM)'
+	      'G_FVL,MU_FVLA,MU_TEMPL,MU_FREEVAR,MU_I,MU_LINE',,
+	      'STRING','STRING')			:(ENDMAKEUI)
MAKEUI
+     G_FVL = TABLE(11)
      MU_FORM G_FF
      MU_FVLA = CONVERT(G_FVL,'ARRAY')			 :F(FRETURN)
      MU_I = 0
MU_LOOP
+     MU_I = MU_I + 1
      MU_FREEVAR = MU_FVLA<MU_I,1>			 :F(FRETURN)
      MU_TEMPL = MAKEQTPLATE('(A' MU_FREEVAR ')' MU_FORM)
      MU_LINE = FINDTEMP(MU_TEMPL)			 :F(MU_LOOP)
      MAKEUI = FINDANTE(MU_LINE)			   :(RETURN)
ENDMAKEUI
-STITL Function: TESTEG()
*FRAME
* Function: TESTEG 
* Date: Sun Jul 25 17:21:50 1982 
* Last modification: June, 1983
* Purpose: 
*  The question to be answered is can the current goal be 
*  obtained from the last line added by EG_  That is, is the 
*  last goal a generalization of the last line? 
* 
* Calling sequence: 
*	TESTEG() 
* Parameters: None 
* On return: 
*	RETURN - If EG can be applied 
*	FRETURN - If EG cannot be applied 
* 
* Local variables: 
*	TE_LINE - Return from FINDTEMP.
*	TE_TPLATE - Return from MAKEQTPLATE.
* 
      pDEFINE('TESTEG()TE_LINE,TE_TPLATE',,,'NULL')	:(ENDTESTEG)
TESTEG
+     TE_TPLATE = MAKEQTPLATE(GETGOAL())
      TE_TPLATE '(E@)' =				:F(CONCHECK)
      TE_LINE = FINDTEMP(TE_TPLATE)			 :F(FRETURN)
      LEQ(G_PRMAT<G_CURLINE,1>,TE_LINE) 	:S(RETURN)F(FRETURN)
ENDTESTEG
-STITL Function: TESTUI()
*FRAME
* Function: TESTUI 
* Date: Sun Jul 25 17:20:53 1982 
* Last modification: June, 1983
* Purpose: 
*  The question to be answered is can the last line added (which is
*  universally quantified) yield the last goal by UI. 
*  That is, is the last goal an instanance of the last line. 
* 
* Calling sequence: 
*	TESTUI() 
* Parameters: None 
* On return: 
*	RETURN - If UI can be applied 
*	FRETURN - If UI cannot be applied 
* 
* Local variables: 
*	TU_TPLATE - Return from MAKEQTPLATE.
* 
      pDEFINE('TESTUI()TU_TPLATE',,,'NULL')		:(ENDTESTUI)
TESTUI
+     TU_TPLATE = MAKEQTPLATE(G_PRMAT<G_CURLINE,1>)
      TU_TPLATE '(A@)' =				:F(CONCHECK)
      DIFFER(GFVTPLATE(G_GOALS)<TU_TPLATE>)	:F(FRETURN)S(RETURN)
ENDTESTUI
-TITLE THINKER - special patterns support routines
-STITL Function: CFF(CF_TEST)
*FRAME
* Function: CFF 
* Date: Sun Jul 25 17:11:12 1982 
* Last modification: June, 1983
* Purpose: 
*  CFF is used by the C patterns. 
*  If the string passed in CF_TEST matches the variable being looked 
*  for in G_MV, then further occurances of that variable after 
*  the quantifier are not to be counted and therefore the 
*  basic G_BF pattern is returned so information is not pushed 
*  on the stack.  Otherwise the G_CF pattern is returned so 
*  that further pushing would occur if neccessary. 
* 
* Calling sequence: 
*	result = CFF(CF_TEST) 
* Parameters: 
*	CF_TEST - Test string to be matched with G_MV.
* On return: 
*	RETURN - Expression to be returned.
* 
      pDEFINE('CFF(CF_TEST)',,'STRING','EXPRESSION')	   :(ENDCFF)
CFF   LEQ(CF_TEST,G_MV) 				:S(CF_FOUND)
      CFF = *G_CF					   :(RETURN)
CF_FOUND
+     CFF = *G_BF					   :(RETURN)
ENDCFF
-STITL Function: PFV(PF_VAL,PF_LIST,PF_CHECK,PF_FUNC)
*FRAME
* Function: PFV 
* Date: Sun Jul 25 17:13:22 1982 
* Last modification: June, 1983
* Purpose: 
*  PFV is a function which is used to update an occurance count 
*  for a pvariable iff it is a free pvariable.	A pvariable is not 
*  free if the stack pointed to by check contains that pvariable 
*  in its list.  The stack will be build by PPQ. 
* 
* Calling sequence: 
*	result = PFV(PF_VAL,PF_LIST,PF_CHECK,PF_FUNC) 
* Parameters: 
*	PF_VAL - the pvariable name (string) 
*	PF_LIST - the occurance table 
*	PF_CHECK - if present a stack of STACK datatypes where each 
*		   V entry represents a pvariable which is NOT free 
*	PF_FUNC - the function code (added to the occurance value
*		  if PF_CHECK)
* On return: 
*	NRETURN - Returns then name GARBAGE.
* 
* Local variables: None.
* 
      pDEFINE('PFV(PF_VAL,PF_LIST,PF_CHECK,PF_FUNC)',,
+	      'STRING,TABLE,STACK|NULL,INTEGER','NULL')    :(ENDPFV)
PFV   PFV = .G_GARBAGE
PF_LOOP
+     IDENT(PF_CHECK)					   :S(PF_DO)
      LEQ(PF_VAL,V(PF_CHECK))				 :S(NRETURN)
      PF_CHECK = N(PF_CHECK)				  :(PF_LOOP)
PF_DO
+     EQ(PF_FUNC,G_ADD) 				  :F(PF_DEL)
      PF_LIST<PF_VAL> = STACK(G_CURLEVEL,PF_LIST<PF_VAL>) :(NRETURN)
PF_DEL
+     PF_LIST<PF_VAL> = N(PF_LIST<PF_VAL>)		  :(NRETURN)
ENDPFV
-STITL Function: PFL(PL_VAL,PL_LIST,PL_LOC,PL_CHECK,PL_FUNC)
      pDEFINE('PFL(PL_VAL,PL_LIST,PL_LOC,PL_CHECK,PL_FUNC)',,
+		'STRING,TABLE,INTEGER,STACK|NULL,INTEGER','NULL')
+							:(ENDPFL)
PFL   PFL = .G_GARBAGE
PL_LOOP
+     IDENT(PL_CHECK)					:S(PL_DO)
      LEQ(PL_VAL,V(PL_CHECK))				:S(NRETURN)
      PL_CHECK = N(PL_CHECK)				:(PL_LOOP)
PL_DO
+     EQ(PL_FUNC,G_ADD) 				:F(PL_DEL)
*     OUTPUT = '##### adding ' PL_VAL ' at location ' PL_LOC
      PL_LIST<PL_LOC> = PL_VAL				:(NRETURN)
PL_DEL
+     PL_LIST<PL_LOC> = NULL				:(NRETURN)
ENDPFL
-STITL Function: PPC(PC_VAL,PC_LIST,PC_FUNC)
*FRAME
* Function: PPC 
* Date: Sun Jul 25 17:12:42 1982 
* Last modification: June, 1983
* Purpose: 
*  Either increment or decrement the value found at 
*  PC_LIST<PC_VAL> depending on the value of PC_FUNC. 
*  Always return the NAME of G_GARBAGE. 
* 
* Calling sequence: 
*	result = PPC(PC_VAL,PC_LIST,PC_FUNC) 
* Parameters: 
*	PC_VAL - the pterm name (string) 
*	PC_LIST - the occurance table 
*	PC_FUNC - the function code (added to the occurance value) 
* On return: 
*	NRETURN - the name of G_GARBAGE 
* 
* Local variables: 
*	PC_TV - the PC_VAL element of PC_LIST.
* 
      pDEFINE('PPC(PC_VAL,PC_LIST,PC_FUNC)PC_TV',,
+	      'STRING,TABLE,INTEGER','NULL')		   :(ENDPPC)
PPC   PC_TV = PC_LIST<PC_VAL>
      PPC = .G_GARBAGE
      PC_LIST<PC_VAL> = PC_TV + PC_FUNC
      PC_LIST<PC_VAL> = EQ(PC_LIST<PC_VAL>,0)		  :(NRETURN)
ENDPPC
-STITL Function: PPQ(PF_STCK,PF_FUNC)
*FRAME
* Function: PPQ 
* Date: Sun Jul 25 17:13:46 1982 
* Last modification: June, 1983
* Purpose: 
*  PPQ is a function used to push/pop pvariables on a stack to 
*  indicate that they are not/are free pvariables. 
*  This information is used by PFV to determine if a variable 
*  is free. If an add function is specified a STACK is add 
*  to the top of the stack and the V field is returned as a name. 
*  If a delete function is specified, the top element of the 
*  stack is deleted and the name of G_GARBAGE is returned. 
* 
* Calling sequence: 
*	result = PPQ(PF_STCK,PF_FUNC) 
* Parameters: 
*	PF_STCK - pointer to the top of the stack 
*	PF_FUNC - function to preform (add or delete) 
* On return: 
*	NRETURN - if func is add then the name of the V field of the
*		     element on the top of the stack 
*		else the name of G_GARBAGE 
* 
* Local variables: None 
* 
* 
*  RETURN 
      pDEFINE('PPQ(PF_STCK,PF_FUNC)',,'NAME,INTEGER','NULL')
+							   :(ENDPPQ)
PPQ
+     EQ(PF_FUNC,G_ADD) 				:S(PF_PQADD)
      $PF_STCK = N($PF_STCK)
      PPQ = .G_GARBAGE					  :(NRETURN)
PF_PQADD
+     $PF_STCK = STACK(,$PF_STCK)
      PPQ = .V($PF_STCK)				  :(NRETURN)
ENDPPQ
-STITL Function: PUSHPOS()
*FRAME
* Function: PUSHPOS 
* Date: Sun Jul 25 17:11:51 1982 
* Last modification: June, 1983
* Purpose: 
*  PUSHPOS is used by the G_CF patterns to push the current 
*  cursor position on the stack with root G_POSSTACK. 
*  The V conponent of the new stack element is returned 
*  as a NAME. 
* 
* Calling sequence: 
*	result = PUSHPOS() 
* Parameters: None 
* On return: 
*	NRETURN - the name of the V field on the top of the stack 
* 
* Local variables: None 
* 
      pDEFINE('PUSHPOS()',,,'NULL')		       :(ENDPUSHPOS)
PUSHPOS
+     G_POSSTACK = STACK(,G_POSSTACK)
      PUSHPOS = .V(G_POSSTACK)				  :(NRETURN)
ENDPUSHPOS
-TITLE THINKER - the brains functions
-STITL Function: ONESTEP(OS_P)
*FRAME
* Function: ONESTEP 
* Date: Sun Jul 25 17:34:33 1982 
* Last modification: June, 1983
* Purpose: 
*   Given a formula X, use it to find a 1-step proof of the current 
*   goal.  This includes finding a contradiction with X.  Relevant 
*   rules are DN, MP, MT, G_BC, MTP, QN, UI, EG, UG and maybe 
*   embedded uses of DN.  If the proof is found, it will add the
*   required.
* 
* Calling sequence: 
*	ONESTEP(OS_P) 
* Parameters: 
*	OS_P - Input formula for which a proof is to be found.
* On return: 
*	RETURN - Current goal is shown!
*	FRETURN - ONESTEP failed.
* 
* Local variables: 
*  CHANGE June 23, 1986   CNH
*  Two new local variables for use as parameters to SHOWN
*	CRIT1 - first critical line in simple contradiction
*	CRIT2 - second critical line in simple contradiction
*	OS_Q - Current goal
*	OS_R - Return from FINDANTE.
*	OS_RL - Negated OS_P.
*	OS_VAR1 - Copy of G_VAR.
*  CHANGE:  June 20, 1986   CNH
*	The local variable CRIT is added
*	CRIT  -  The critical line for the SHOW 
*	OS_LOP1 - Copy of G_LOP
*	OS_ROP1 - Copy of G_ROP
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
*	G_VAR - Quantifier variable from G_SPLIT
*	OS_CUR - current line number
* 
* CHANGE ID July 2, 1986
* Add LINEREF as a local variable
      pDEFINE('ONESTEP(OS_P)LINEREF,OS_Q,OS_R,OS_RL,OS_VAR1,CRIT,'
+	      'OS_LOP1,OS_ROP1,G_LOP,G_ROP,G_OP,G_VAR,OS_CUR,'
+	      'L_CONJUNCT,R_CONJUNCT',,
+	      'STRING','NULL')			       :(ENDONESTEP)
ONESTEP
+     DIFFER(G_GOALS)					 :F(RETURN)
      OS_CUR = G_CURLINE
*     OUTPUT = 'IN ONESTEP LINE NUMBER ' G_CURLINE
      OS_Q = GETGOAL()					 :F(FRETURN)
      LEQ(OS_P,OS_Q)					 :S(OS_SHOW)
      REFINST(OS_Q)				 :S(OS_REFL)
      OS_RL = '~' OS_P
      OS_R = FINDANTE(OS_RL)				 :S(OS_CONT)
*  CHANGES Aug 18, 1986 CNH  check for ~P before adding lines
      OS_R = IDTEST(OS_P,OS_Q)			 :S(OS_IDTEST)
      OS_Q G_SPLIT					    :F(OS_1)
      LEQ(G_OP,'A')					    :F(OS_2)
      FREE(G_VAR)					    :S(OS_1)
      LEQ(G_ROP,OS_P)				  :S(OS_SHOW)F(OS_1)
OS_2
+     LEQ(G_OP,'E')					    :F(OS_1)
      EGREF(OS_Q)				 :S(OS_EGREF)
      TESTEG()						   :S(OS_EG)
OS_1
+     OS_P G_SPLIT					    :F(OS_4)
      LEQ(G_OP,'~')					    :F(OS_3)
      OS_RL = G_ROP
*  CHANGES  Aug 18, 1986  CNH  check for ~a-a
      REFINST(OS_RL)				:S(OS_REF2)
      OS_R = FINDANTE(G_ROP)				 :S(OS_CONT)
      OS_R = FINDANTE(G_ROP)				 :S(OS_CONT)
      OS_RL G_SPLIT					    :F(OS_3)
      LEQ(G_OP,'&')					   :F(OS_1a)
      L_CONJUNCT = FINDANTE(G_ROP)			    :F(OS_3)
      R_CONJUNCT = FINDANTE(G_LOP)	       :F(OS_3)S(OS_NEWBOX1)
OS_1a
      LEQ(G_OP,'+')					   :F(OS_1b)
      L_CONJUNCT = FINDANTE(G_ROP)		      :S(OS_NEWBOX2)
      L_CONJUNCT = FINDANTE(G_LOP)	       :S(OS_NEWBOX2)F(OS_3)
OS_1b
      LEQ(G_OP,'=')					    :F(OS_3)
      L_CONJUNCT = FINDANTE(G_LOP '>' G_ROP)		    :F(OS_3)
      R_CONJUNCT = FINDANTE(G_ROP '>' G_LOP)	      :S(OS_NEWBOX3)
OS_3
+     OS_P G_SPLIT
      LEQ(G_OP,'A')					    :F(OS_4)
      TESTUI()						   :S(OS_UI)
OS_4
+     OS_R = FINDANTE('(' OS_P '>' OS_Q ')')		  :S(OS_MP1)
      OS_R = FINDANTE('(' OS_P '>~~' OS_Q ')')		  :S(OS_MP3)
      OS_R = FINDANTE('(' OS_P '=' OS_Q ')')		  :S(OS_BC1)
      OS_R = FINDANTE('(' OS_Q '=' OS_P ')')		  :S(OS_BC1)
      OS_R = FINDANTE('(' OS_Q '=~~' OS_P ')')		  :S(OS_BC2)
      OS_R = FINDANTE('(~~' OS_P '=' OS_Q ')')		  :S(OS_BC2)
      OS_R = FINDANTE('(~~' OS_Q '=' OS_P ')')		  :S(OS_BC3)
      OS_R = FINDANTE('(' OS_P '=~~' OS_Q ')')		  :S(OS_BC3)
      OS_R = FINDANTE('(~~' OS_P '=~~' OS_Q ')')	  :S(OS_BC4)
      OS_R = FINDANTE('(~~' OS_Q '=~~' OS_P ')')	  :S(OS_BC4)
      OS_R = FINDANTE('(' OS_Q '+~' OS_P ')')		  :S(OS_MTP)
      OS_R = FINDANTE('(~' OS_P '+' OS_Q ')')		  :S(OS_MTP)
      OS_R = FINDANTE('(~' OS_Q '>~' OS_P ')')		  :S(OS_MT3)
      OS_P G_SPLIT					    :F(OS_7)
      LEQ(G_OP,'~')					    :F(OS_7)
      OS_R = FINDANTE('(~' OS_Q '>' G_ROP ')')		  :S(OS_MT4)
      OS_R = FINDANTE('(' G_ROP '+' OS_Q ')')		  :S(OS_MTP)
      OS_R = FINDANTE('(' OS_Q '+' G_ROP ')')		  :S(OS_MTP)
      G_ROP G_SPLIT					    :F(OS_7)
      LEQ(G_OP,'~')					    :F(OS_5)
      LEQ(G_ROP,OS_Q)					 :S(OS_DN)
      OS_R = FINDANTE('(' G_ROP '>' OS_Q ')')		:S(OS_MP2)
      OS_R = FINDANTE('(' G_ROP '>~~' OS_Q ')') 	:S(OS_MP4)
      OS_R = FINDANTE('(' G_ROP '=' OS_Q ')')		:S(OS_BC5)
      OS_R = FINDANTE('(' OS_Q '=' G_ROP ')')		:S(OS_BC5)
OS_5
+     LEQ(G_OP,'A')					    :F(OS_6)
      OS_VAR1 = G_VAR
      OS_ROP1 = G_ROP
      OS_Q G_SPLIT					    :F(OS_7)
      LEQ(G_OP,'E')					    :F(OS_7)
      LEQ(OS_VAR1,G_VAR)				    :F(OS_7)
      G_ROP G_SPLIT					    :F(OS_7)
      LEQ(G_OP,'~')					    :F(OS_7)
      LEQ(G_ROP,OS_ROP1)			  :S(OS_QN)F(OS_7)
OS_6
+     LEQ(G_OP,'E')					    :F(OS_7)
      OS_VAR1 = G_VAR
      OS_ROP1 = G_ROP
      OS_Q G_SPLIT					    :F(OS_7)
      LEQ(G_OP,'A')					    :F(OS_7)
      LEQ(OS_VAR1,G_VAR)				    :F(OS_7)
      G_ROP G_SPLIT					    :F(OS_7)
      LEQ(G_OP,'~')					    :F(OS_7)
      LEQ(G_ROP,OS_ROP1)				 :S(OS_QN)
OS_7
+     OS_Q G_SPLIT					   :F(OS_13)
      LEQ(G_OP,'~')					    :F(OS_8)
      OS_R = FINDANTE('(' G_ROP '>~' OS_P ')')		:S(OS_MT3)
      G_ROP G_SPLIT					    :F(OS_9)
      LEQ(G_OP,'~')					    :F(OS_9)
      LEQ(OS_P,G_ROP)					 :S(OS_DN)
OS_9
+     OS_Q G_SPLIT
      OS_ROP1 = G_ROP
      OS_P G_SPLIT					    :F(OS_8)
      LEQ(G_OP,'~')					   :F(OS_12)
      OS_R = FINDANTE('(' OS_ROP1 '>' G_ROP ')')	:S(OS_MT3)
OS_12
+     LEQ(G_OP,'>')					   :F(OS_11)
      LEQ(G_LOP,OS_ROP1)				   :F(OS_11)
      OS_R = FINDANTE('~' G_ROP)			:S(OS_MT2)
      G_ROP G_SPLIT					   :F(OS_11)
      LEQ(G_OP,'~')					   :F(OS_11)
      OS_R = FINDANTE(G_ROP)				:S(OS_MT1)
OS_11
+     OS_Q G_SPLIT
      G_ROP G_SPLIT					    :F(OS_8)
      LEQ(G_OP,'E')					   :F(OS_10)
      OS_VAR1 = G_VAR
      OS_ROP1 = G_ROP
      OS_P G_SPLIT					    :F(OS_8)
      LEQ(G_OP,'A')					    :F(OS_8)
      LEQ(OS_VAR1,G_VAR)				    :F(OS_8)
      G_ROP G_SPLIT					    :F(OS_8)
      LEQ(G_OP,'~')					    :F(OS_8)
      LEQ(OS_ROP1,G_ROP)			  :S(OS_QN)F(OS_8)
OS_10
+     LEQ(G_OP,'A')					    :F(OS_8)
      OS_VAR1 = G_VAR
      OS_ROP1 = G_ROP
      OS_P G_SPLIT					    :F(OS_8)
      LEQ(G_OP,'E')					    :F(OS_8)
      LEQ(OS_VAR1,G_VAR)				    :F(OS_8)
      G_ROP G_SPLIT					    :F(OS_8)
      LEQ(G_OP,'~')					    :F(OS_8)
      LEQ(OS_ROP1,G_ROP)				 :S(OS_QN)
OS_8
+     OS_Q G_SPLIT
      LEQ(G_OP,'>')					   :F(OS_13)
      OS_ROP1 = G_ROP
      OS_LOP1 = G_LOP
      REFINST(G_ROP)				 :S(OS_REFINSTA)
      EGREF(G_ROP)				 :S(OS_EGREFA)
      OS_P G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'=')					   :F(OS_13)
      LEQ(G_LOP,OS_LOP1)				   :F(OS_14)
      LEQ(G_ROP,OS_ROP1)				 :S(OS_BC)
OS_14
+     LEQ(G_LOP,OS_ROP1)				   :F(OS_13)
      LEQ(G_ROP,OS_LOP1)				 :S(OS_BC)
OS_13
+     OS_P G_IDSPLIT					   :F(FRETURN)
      LEQ(G_OP,'-')				 :F(OS_13A)
*     OUTPUT = 'CALLING TESTID'
      OS_R = TESTID(OS_Q,OS_P)			 :S(OS_TESTID)
*     OUTPUT = 'TESTID FAILED'
OS_13A
+     LEQ(G_OP,'>')					   :F(OS_15)
      LEQ(G_ROP,OS_Q)					   :F(OS_16)
      OS_R = FINDANTE(G_LOP)				:S(OS_MP1)
      OS_R = FINDANTE('~~' G_LOP)			:S(OS_MP2)
OS_16
+     OS_LOP1 = G_LOP
      OS_ROP1 = G_ROP
      G_LOP G_SPLIT					   :F(OS_17)
      LEQ(G_OP,'~')					   :F(OS_17)
      LEQ(G_ROP,OS_Q)					   :F(OS_17)
      OS_R = FINDANTE('~' OS_ROP1)			:S(OS_MT4)
      OS_ROP1 G_SPLIT					   :F(OS_17)
      LEQ(G_OP,'~')					   :F(OS_17)
      OS_R = FINDANTE(G_ROP)				:S(OS_MT3)
OS_17
+     OS_ROP1 G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'~')					 :F(FRETURN)
      G_ROP G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'~')					 :F(FRETURN)
      LEQ(G_ROP,OS_Q)					 :F(FRETURN)
      OS_R = FINDANTE(OS_LOP1)				:S(OS_MP3)
      OS_R = FINDANTE('~~' OS_LOP1)			:S(OS_MP4)
OS_15
+     OS_P G_SPLIT
      LEQ(G_OP,'+')					   :F(OS_20)
      LEQ(OS_Q,G_ROP)					   :F(OS_18)
      OS_R = FINDANTE('~' G_LOP)			:S(OS_MTP)
OS_18
+     LEQ(OS_Q,G_LOP)					   :F(OS_19)
      OS_R = FINDANTE('~' G_ROP)			:S(OS_MTP)
OS_19
+     LEQ(OS_Q,G_ROP)					   :F(OS_21)
      OS_LOP1 = G_LOP
      OS_ROP1 = G_ROP
      G_LOP G_SPLIT					   :F(OS_21)
      LEQ(G_OP,'~')					   :F(OS_21)
      OS_R = FINDANTE(G_ROP)				:S(OS_MTP)
OS_21
+     LEQ(OS_Q,OS_LOP1) 				 :F(FRETURN)
*  CHANGES July 23, 1986   CNH
*  change the line from OSIT   :f(freturn) to the next 3
*  typographical error?
      OS_LOP1 = G_LOP
      OS_ROP1 = G_ROP
      G_ROP G_SPLIT					:F(FRETURN)
      LEQ(G_OP,'~')					 :F(FRETURN)
      OS_R = FINDANTE(G_ROP)		      :S(OS_MTP)F(FRETURN)
OS_20
+     OS_P G_SPLIT
      LEQ(G_OP,'=')					 :F(FRETURN)
      LEQ(OS_Q,G_LOP)					   :F(OS_22)
      OS_R = FINDANTE(G_ROP)				:S(OS_BC6)
      OS_R = FINDANTE('~~' G_ROP)			:S(OS_BC7)
OS_22
+     LEQ(OS_Q,G_ROP)					   :F(OS_23)
      OS_R = FINDANTE(G_LOP)				:S(OS_BC6)
      OS_R = FINDANTE('~~' G_LOP)			:S(OS_BC7)
      OS_LOP1 = G_LOP
      OS_ROP1 = G_ROP
      G_LOP G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'~')					 :F(FRETURN)
      G_ROP G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'~')					 :F(FRETURN)
      OS_R = FINDANTE(G_ROP)		      :S(OS_BC7)F(FRETURN)
OS_23
+     LEQ(OS_Q,OS_LOP1) 				   :F(OS_24)
      OS_ROP1 G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'~')					 :F(FRETURN)
      G_ROP G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'~')					 :F(FRETURN)
      OS_R = FINDANTE(G_ROP)		      :S(OS_BC7)F(FRETURN)
OS_24
+     OS_ROP1 G_SPLIT					   :F(OS_25)
      LEQ(G_OP,'~')					   :F(OS_25)
      G_ROP G_SPLIT					   :F(OS_25)
      LEQ(G_OP,'~')					   :F(OS_25)
      LEQ(OS_Q,G_ROP)					   :F(OS_25)
      OS_R = FINDANTE(OS_LOP1)				:S(OS_BC8)
      OS_LOP1 G_SPLIT					   :F(OS_25)
      LEQ(G_OP,'~')					   :F(OS_25)
      G_ROP G_SPLIT					   :F(OS_25)
      LEQ(G_OP,'~')					   :F(OS_25)
      OS_R = FINDANTE(G_ROP)				:S(OS_BC7)
OS_25
+     OS_LOP1 G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'~')					 :F(FRETURN)
      G_ROP G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'~')					 :F(FRETURN)
      LEQ(OS_Q,G_ROP)					 :F(FRETURN)
      OS_R = FINDANTE(OS_ROP1)				:S(OS_BC8)
      OS_LOP1 G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'~')					 :F(FRETURN)
      G_ROP G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'~')					 :F(FRETURN)
      OS_R = FINDANTE(G_ROP)		      :F(FRETURN)S(OS_BC9)
*FRAME
* Now having checked all the ways to box + cancel, we 
* are ready to add them into the proof matrix and 
* annotate them correctly and box. 
*
*  CHANGE June 23, 1986   CNH
*  OS_CONT calls SHOWN itself since it must pass the two line
*  numbers of the contradictory lines.	One of these is the
*  present line (before calling addproof) the other is
*  either the last line or PRLINE(OS_R) (if no repeat is done)
*  changes Aug 5, 1986	CNH change addproof to addprii
*  since each of these lines will finish the show we need
*  only use addprii
OS_CONT
+     CRIT1 = G_CURLINE
      CRIT2 = G_CURLINE + 1
      ADDPRII(OS_RL, PRLINE(OS_R) ',R') 
*  changes Aug 5, 1986	CNH change addproof to addprii
      CRIT2 = GT(CRIT2,G_CURLINE) PRLINE(OS_R)
      SHOWN(CRIT1 ',' CRIT2)				  :(RETURN)
OS_EG
+     ADDPRII(OS_Q, G_CURLINE ',EG')			 :(OS_SHOW)
OS_UI
+     ADDPRII(OS_Q, G_CURLINE ',UI')			 :(OS_SHOW)
OS_MP1
+     ADDPRII(OS_Q, G_CURLINE ',' PRLINE(OS_R) ',MP')	 :(OS_SHOW)
OS_MP2
+     ADDPRII(OS_Q, G_CURLINE ',DN,' PRLINE(OS_R) ',MP') :(OS_SHOW)
OS_MP3
+     ADDPRII(OS_Q, G_CURLINE ',' PRLINE(OS_R) ',MP,DN') :(OS_SHOW)
OS_MP4
+     ADDPRII(OS_Q, G_CURLINE ',DN,' PRLINE(OS_R) ',MP,DN')
+							  :(OS_SHOW)
OS_BC
+     ADDPRII(OS_Q, G_CURLINE ',BC')			 :(OS_SHOW)
OS_BC1
+     ADDPRII(OS_Q, PRLINE(OS_R) ',BC,' G_CURLINE ',MP')
+							  :(OS_SHOW)
OS_BC2
+     ADDPRII(OS_Q, PRLINE(OS_R) ',BC,' G_CURLINE ',DN,MP')
+							  :(OS_SHOW)
OS_BC3
+     ADDPRII(OS_Q, PRLINE(OS_R) ',BC,' G_CURLINE ',MP,DN')
+							  :(OS_SHOW)
OS_BC4
+     ADDPRII(OS_Q, PRLINE(OS_R) ',BC,' G_CURLINE ',DN,MP,DN')
+							  :(OS_SHOW)
OS_BC5
+     ADDPRII(OS_Q, G_CURLINE ',DN,' PRLINE(OS_R) ',BC,MP')
+							  :(OS_SHOW)
OS_BC6
+     ADDPRII(OS_Q, G_CURLINE ',BC,' PRLINE(OS_R) ',MP')
+							  :(OS_SHOW)
OS_BC7
+     ADDPRII(OS_Q, G_CURLINE ',BC,' PRLINE(OS_R) ',DN,MP')
+							  :(OS_SHOW)
OS_BC8
+     ADDPRII(OS_Q, G_CURLINE ',BC,' PRLINE(OS_R) ',MP,DN')
+							  :(OS_SHOW)
OS_BC9
+     ADDPRII(OS_Q, G_CURLINE ',BC,' PRLINE(OS_R) ',DN,MP,DN')
+							  :(OS_SHOW)
OS_MT1
+     ADDPRII(OS_Q, PRLINE(OS_R) ',DN,' G_CURLINE ',MT') :(OS_SHOW)
OS_MT2
+     ADDPRII(OS_Q, G_CURLINE ',' PRLINE(OS_R) ',MT')	 :(OS_SHOW)
OS_MT3
+     ADDPRII(OS_Q, G_CURLINE ',DN,' PRLINE(OS_R) ',MT,DN')
+							  :(OS_SHOW)
OS_MT4
+     ADDPRII(OS_Q, G_CURLINE ',' PRLINE(OS_R) ',MT,DN') :(OS_SHOW)
OS_MTP
+     ADDPRII(OS_Q, G_CURLINE ',' PRLINE(OS_R) ',MTP')	 :(OS_SHOW)
OS_DN
+     ADDPRII(OS_Q, G_CURLINE ',DN')			 :(OS_SHOW)
OS_QN
+     ADDPRII(OS_Q, G_CURLINE ',QN')			 :(OS_SHOW)
OS_REFL
+     ADDPRII(OS_Q, 'REFL')			:(OS_SHOW)
* CHANGES Aug 18, 1986	CNH  check for ~a-a
OS_REF2
      ADDPRII(OS_RL, 'REFL')			:(OS_SHOW)
OS_EGREF
+     ADDPRII(OS_Q, 'REFL,EG')			:(OS_SHOW)
OS_IDTEST
*  CHANGES ID July 2, 1986   CNH
*  the references are to the line with OS_P (OS_CUR) and the
*  line number OS_R
      ADDPRII(OS_Q, OS_CUR ',' OS_R ',LL')  :(OS_SHOW)
OS_TESTID
      OS_CUR = OS_CUR + 2
OS_TELOOP
*  CHANGES July 29, 1986   CNH 
*  set OS_CUR to be the line number of OS_P looping until
*  it is at the right line.
*  since consequences of the identity could have been added
      OS_CUR = OS_CUR - 1
      LEQ(OS_P,G_PRMAT<OS_CUR,1>)			 :F(OS_TELOOP)
      LE(G_CURLEVEL,G_PRMAT<OS_CUR,3>)			 :F(OS_TELOOP)
      ADDPRII(OS_Q, OS_CUR ',' PRLINE(OS_R) ',LL')  :(OS_SHOW)
OS_REFINSTA
+     ADDPRII(G_ROP, 'REFL')			:(OS_SHOW)
OS_EGREFA
+     ADDPRII(G_ROP, 'REFL,EG') 			 :(OS_SHOW)
OS_NEWBOX1
+     ADDPRII(OS_RL, PRLINE(L_CONJUNCT) ',' PRLINE(R_CONJUNCT) ',ADJ')
+							  :(OS_SHOW)
OS_NEWBOX2
+     ADDPRII(OS_RL, PRLINE(L_CONJUNCT) ',ADD') 	 :(OS_SHOW)
OS_NEWBOX3
+     ADDPRII(OS_RL, PRLINE(L_CONJUNCT) ',' PRLINE(R_CONJUNCT) ',CB')
+							  :(OS_SHOW)
*  CHANGE:  June 20, 1986  CNH
*  we must determine what the critical line is for the SHOW
*  and pass that to SHOWN as the parameter.  The critical
*  line is the last line in the show that is only one
*  level deeper than the top of the show.
OS_SHOW
+     CRIT = G_CURLINE
OS_SHOW2
+     CRIT = DIFFER(G_PRMAT<PRLIN(G_GOALS),3> + 1,
+		    G_PRMAT<CRIT,3>)  CRIT - 1	       :S(OS_SHOW2)
*  CHANGE:  June 24, 1986
*  contradictions have been slipping through onestep
*  I believe it is when the last line contradicts the line
*  to be proved???
*  the line 3852:   SHOWN(CRIT) 	 :(RETURN)
      CRIT1 = CRIT
      CRIT = IDENT(OS_P,'~' G_PRMAT<CRIT1,1>) CRIT ' ' 
+     PRLINE(FINDANTE(OS_P))
      CRIT = IDENT('~' OS_P,G_PRMAT<CRIT1,1>) CRIT ' ' 
+     PRLINE(FINDATE(OS_P))
      SHOWN(CRIT)					   :(RETURN)
ENDONESTEP
-STITL Function: SIMPLEPROOF(SP_Q)
*FRAME
* Function: SIMPLEPROOF 
* Date: Sun Jul 25 17:32:32 1982 
* Last modification: June, 1983
* Purpose: 
* SIMPLEPROOF(X) tries to find a one-line proof of X.  X is not 
* assumed to be the current goal.  If the proof is found, it will 
* ADDPROOF(X) with appropriate annotations, but not box.  This 
* function does not find contradictions. 
* 
* Calling sequence: 
*	SIMPLEPROOF(SP_Q) 
* Parameters: 
*	SP_Q - Goal for SIMPLEPROOF <not same as GETGOAL()>
* On return: 
*	RETURN - SP_Q has been proven
*	FRETURN - SIMPLEPROOF has failed.
* 
* Local variables: 
*	SP_P - Return from FINDTEMP.
*	SP_R - Return from FINDANTE.
*	SP_S -	  "	"      "
*	SP_VAR1 - Copy of G_VAR.
*	G_VAR - Quantifier variable from G_SPLIT
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
* 
* Description: 
      pDEFINE('SIMPLEPROOF(SP_Q)SP_P,SP_R,SP_S,SP_VAR1,'
+	      'G_OP,G_LOP,G_ROP,G_VAR',,
+	      'STRING','NULL')			   :(ENDSIMPLEPROOF)
SIMPLEPROOF
*  changes Aug 5, 1986	CNH  ADDPROOF
*  it is necessary for simpleproof to return a value of
*  'T' or 'F' or succeed.  If it succeeds then it found a
*  proof of the given line, if it fails and the value
*  is 'T' then it closed the show, otherwise it failed.
*  all calls to addproof followed by a :(RETURN) are
*  now changed to F(RETURN)S(SP_ONE)
*     OUTPUT = 'IN SIMPLE SP_VAL NOW F'
      SP_VAL = 'F'
*     OUTPUT = 'IN SIMPLEPROOF LINE NUMBER ' G_CURLINE
      SP_P = FINDANTE(SP_Q)				    :F(SP_0)
*  changes Aug 5, 1986	CNH
*  initialize simpleproof
      ADDPROOF(SP_Q, PRLINE(SP_P) ',R') 	
+						 :F(RETURN)S(SP_ONE)
*  CNH	3rd parameter to ADDPROOF removed
SP_0 
+     REFINST(SP_Q)				      :F(SP_0A)
      ADDPROOF(SP_Q, 'REFL')			 
+						 :F(RETURN)S(SP_ONE)
SP_0A
+     EGREF(SP_Q)				      :F(SP_0B)
      ADDPROOF(SP_Q, 'REFL,EG') 		 
+						 :F(RETURN)S(SP_ONE)
SP_0B
+     SP_P = SIMPLEID(SP_Q)			 :F(SP_1)
*      OUTPUT = 'SIMPLEID succeeded for ' SP_Q
      ADDPROOF(SP_Q,PRLINE(FORM_PTR(SP_P)) ',' 
+		    PRLINE(IDENT_PTR(SP_P)) ',LL')		
+						 :F(RETURN)S(SP_ONE)
*  CNH	3rd parameter to ADDPROOF removed
SP_1
+     SP_P = FINDTEMP('(' SP_Q '&@)')			    :S(SP_2)
      SP_P = FINDTEMP('(@&' SP_Q ')')			    :F(SP_3)
SP_2
+     SP_R = FINDANTE(SP_P)
      ADDPROOF(SP_Q, PRLINE(SP_R) ',S') 	
+						 :F(RETURN)S(SP_ONE)
*  CNH	3rd parameter to ADDPROOF removed
SP_3
+     SP_P = FINDANTE('~~' SP_Q)			    :S(SP_4)
      SP_Q G_SPLIT					    :F(SP_5)
      LEQ(G_OP,'~')					    :F(SP_5)
      G_ROP G_SPLIT					    :F(SP_5)
      LEQ(G_OP,'~')					    :F(SP_5)
      SP_P = FINDANTE(G_ROP)				    :F(SP_5)
SP_4
+     ADDPROOF(SP_Q, PRLINE(SP_P) ',DN')	
+						 :F(RETURN)S(SP_ONE)
*  CNH	3rd parameter to ADDPROOF removed
SP_5
+     SP_Q G_SPLIT					   :F(SP_10)
      LEQ(G_OP,'=')					    :F(SP_6)
      SP_P = FINDANTE('(' G_LOP '>' G_ROP ')')		    :F(SP_6)
      SP_R = FINDANTE('(' G_ROP '>' G_LOP ')')		    :F(SP_6)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(SP_Q, PRLINE(SP_P) ',' PRLINE(SP_R) ',CB')
+						 :F(RETURN)S(SP_ONE)
SP_6
+     LEQ(G_OP,'+')					    :F(SP_7)
      SP_P = FINDANTE(G_LOP)				    :S(SP_8)
      SP_P = FINDANTE(G_ROP)				    :F(SP_7)
SP_8
+     ADDPROOF(SP_Q, PRLINE(SP_P) ',ADD')	
+						 :F(RETURN)S(SP_ONE)
*  CNH	3rd parameter to ADDPROOF removed
SP_7
+     LEQ(G_OP,'&')					    :F(SP_9)
      SP_P = FINDANTE(G_LOP)				    :F(SP_9)
      SP_R = FINDANTE(G_ROP)				    :F(SP_9)
      ADDPROOF(SP_Q, PRLINE(SP_P) ',' PRLINE(SP_R) ',ADJ')
+						 :F(RETURN)S(SP_ONE)
*  CNH	3rd parameter to ADDPROOF removed
SP_9
+     LEQ(G_OP,'>')					   :F(SP_10)
      SP_P = FINDANTE('(' G_LOP '=' G_ROP ')')		   :S(SP_11)
      SP_P = FINDANTE('(' G_ROP '=' G_LOP ')')		   :F(SP_10)
SP_11
+     ADDPROOF(SP_Q, PRLINE(SP_P) ',BC')	
+						 :F(RETURN)S(SP_ONE)
*  CNH	3rd parameter to ADDPROOF removed
SP_10
+     SP_P = FINDTEMP('(@>' SP_Q ')')			   :F(SP_12)
      SP_R = FINDANTE(SP_P)
      SP_P G_SPLIT
      SP_S = FINDANTE(G_LOP)				   :F(SP_12)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(SP_Q, PRLINE(SP_R) ',' PRLINE(SP_S) ',MP')
+						 :F(RETURN)S(SP_ONE)
SP_12
+     SP_P = FINDTEMP('(~' SP_Q '>@)')			   :F(SP_13)
      SP_R = FINDANTE(SP_P)
      SP_P G_SPLIT
      SP_S = FINDANTE('~' G_ROP)			   :S(SP_14)
      G_ROP G_SPLIT					   :F(SP_13)
      LEQ(G_OP,'~')					   :F(SP_13)
      SP_S = FINDANTE(G_ROP)				   :S(SP_14)
SP_13
+     SP_Q G_SPLIT					   :F(SP_15)
      LEQ(G_OP,'~')					   :F(SP_15)
      SP_P = FINDTEMP('(' G_ROP '>@)')			   :F(SP_15)
      SP_R = FINDANTE(SP_P)
      SP_P G_SPLIT
      SP_S = FINDANTE('~' G_ROP)			   :S(SP_14)
      G_ROP G_SPLIT					   :F(SP_15)
      LEQ(G_OP,'~')					   :F(SP_15)
      SP_S = FINDANTE(G_ROP)				   :F(SP_15)
SP_14
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(SP_Q, PRLINE(SP_R) ',' PRLINE(SP_S) ',MT')
+						 :F(RETURN)S(SP_ONE)
SP_15
+     SP_P = FINDTEMP('(' SP_Q '+@)')			   :F(SP_16)
      SP_R = FINDANTE(SP_P)
      SP_P G_SPLIT
      SP_S = FINDANTE('~' G_ROP)			   :S(SP_17)
      G_ROP G_SPLIT					   :F(SP_16)
      LEQ(G_OP,'~')					   :F(SP_16)
      SP_S = FINDANTE(G_ROP)				   :S(SP_17)
SP_16
+     SP_P = FINDTEMP('(@+' SP_Q ')')			   :F(SP_18)
      SP_R = FINDANTE(SP_P)
      SP_P G_SPLIT
      SP_S = FINDANTE('~' G_LOP)			   :S(SP_17)
      G_LOP G_SPLIT					   :F(SP_18)
      LEQ(G_OP,'~')					   :F(SP_18)
      SP_S = FINDANTE(G_ROP)				   :F(SP_18)
SP_17
+     ADDPROOF(SP_Q, PRLINE(SP_R) ',' PRLINE(SP_S) ',MTP')	
+						 :F(RETURN)S(SP_ONE)
*  CNH	3rd parameter to ADDPROOF removed
SP_18
+     SP_Q G_SPLIT					   :F(SP_19)
      LEQ(G_OP,'E')					   :F(SP_19)
      SP_S = MAKEEG(SP_Q)				   :F(SP_19)
      ADDPROOF(SP_Q, PRLINE(SP_S) ',EG')	
+						 :F(RETURN)S(SP_ONE)
*  CNH	3rd parameter to ADDPROOF removed
SP_19
+     SP_S = MAKEUI(SP_Q)				   :F(SP_20)
      ADDPROOF(SP_Q, PRLINE(SP_S) ',UI')	
+						 :F(RETURN)S(SP_ONE)
*  CNH	3rd parameter to ADDPROOF removed
SP_20
+     SP_Q G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'~')					   :F(SP_21)
      G_ROP G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'A')					   :F(SP_22)
      SP_R = FINDANTE('(E' G_VAR ')~' G_ROP)		 :F(FRETURN)
      ADDPROOF(SP_Q, PRLINE(SP_R) ',QN')	
+						 :F(RETURN)S(SP_ONE)
*  CNH	3rd parameter to ADDPROOF removed
SP_22
+     LEQ(G_OP,'E')					 :F(FRETURN)
      SP_R = FINDANTE('(A' G_VAR ')~' G_ROP)		 :F(FRETURN)
      ADDPROOF(SP_Q, PRLINE(SP_R) ',QN')	
+						 :F(RETURN)S(SP_ONE)
*  CNH	3rd parameter to ADDPROOF removed
SP_21
+     LEQ(G_OP,'A')					   :F(SP_23)
      SP_VAR1 = G_VAR
      G_ROP G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'~')					 :F(FRETURN)
      SP_R = FINDANTE('~(E' SP_VAR1 ')' G_ROP)		 :F(FRETURN)
      ADDPROOF(SP_Q, PRLINE(SP_R) ',QN')	
+						 :F(RETURN)S(SP_ONE)
*  CNH	3rd parameter to ADDPROOF removed
SP_23
+     LEQ(G_OP,'E')					 :F(FRETURN)
      SP_VAR1 = G_VAR
      G_ROP G_SPLIT					 :F(FRETURN)
      LEQ(G_OP,'~')					 :F(FRETURN)
      SP_R = FINDANTE('~(A' SP_VAR1 ')' G_ROP)		 :F(FRETURN)
      ADDPROOF(SP_Q, PRLINE(SP_R) ',QN')	
+						 :F(RETURN)S(SP_ONE)
*  CNH	3rd parameter to ADDPROOF removed
*  changes Aug 5, 1986	CNH
*  add the routine to set the value for simpleproof
SP_ONE
*     OUTPUT = 'SP_VAL NOW T'
      SP_VAL = 'T'			    :(FRETURN)
ENDSIMPLEPROOF
-STITL Function: PROOF(DO_WHAT)
*FRAME
* Function: PROOF 
* Date: Sun Jul 25 17:37:48 1982 
* Last modification: June, 1983 
* Purpose: 
* PROOF() is the overall proof supervisor which directs the strategy
* to be used in proving the current goal.  It tries various 
* combinations of direct proof using SEARCHCONTRA() and SIMPLEPROOF;
* then it SPLITS the goal (if possible) and makes the appropriate 
* assumptions when called for.	After various straightforward 
* attempts to prove the goal, it moves to (label) INDIRECT and 
* attempts to find a contradiction by applying every rule that is 
* applicable to the G_ANTELINES.  If even that fails, it tries a few
* backup strategies -- TRYANYTHING and DESPARATION. 
* 
* Calling sequence: 
*	result = PROOF(DO_WHAT) 
* Parameters: DO_WHAT	used by parallelism code
*	    if it is null (default)  then do all 
*	    if it is 'F'  then just do the first part
*	    if it is 'S'  then just do the second part
* On return: 
*	RETURN - Normal termination
*	FRETURN - Formula not provable.
* 
* Local variables: 
*	PO_RL - Return from FINDANTE.
*	PO_S - Copy of G_ROP.
*	PO_T - Return from FINDANTE.
*	PO_Q - Return from GETGAOL.
*	PO_R - Return from SERCHNEG.
*  CHANGE June 20, 1986 CNH
*  A new local variable CRIT is added for the parameter
*  to SHOWN
*	CRIT - number of critical line for SHOWN
*  CHANGE June 23, 1986 CNH
*  Two new variables are added CRIT1, CRIT2 as parameters
*  To SHOWN when it is a contradiction
*	CRIT1 - first line number as parameter to SHOWN
*	CRIT2 - second line number as parameter to SHOWN
*	PO_X - Return from other "SERCH" functions.
*	PO_LOP1 - Copy of G_LOP.
*	PO_ROP1 - Copy of G_ROP.
*	PO_NEWLIST - Return from FINDVARS.
*	PO_OLDCUR - Copy of G_CURRLINE.
*	PO_CMDRC - Return code from command parser.
*	G_VAR - Quantifier variable from G_SPLIT
*	G_ROP - Right operand as given by G_SPLIT
*	G_OP - Main operator given by G_SPLIT
*	G_LOP - Left operand as given by G_SPLIT
      pDEFINE('PROOF(DO_WHAT)PO_RL,PO_S,PO_T,PO_Q,PO_R,CRIT,'
+	 'CRIT1,CRIT2,PO_X,PO_LOP1,PO_ROP1,PO_NEWLIST,PO_OLDCUR,'
+	 'PO_CMDRC,G_LOP,G_ROP,G_OP,G_VAR,ALLDONE,'
+	 'PTR_NEG,PTR_NEGS,PTR_BCS,PTR_ANDS,PTR_QNS,'
+	 'HOLD_SIZE,TRY_SIZE,START,PTR_WEDGE,PTR_ARROW,'
+	 'PO_WH,COUNT,HOLD,LASTFREE,SAV_VLIST',,,'NULL')
+							 :(ENDPROOF)
PROOF
* CNT CHECK IF PARALLEL IS SET
      GT(P_PARALLEL,0)					 :S(PO_PON)
      DO_WHAT = null
PO_PON
* CHANGES Aug 21, 1986	CNH  parallelism  check parameter
      DO_WHAT 'S'					  :S(PO_DESPARATION)
*  changes Aug 5, 1986	CNH
*  after each call to simpleproof if it is not a success
*  then we check if it is a 'T' (hence closed the show)
      PO_Q = GETGOAL()					  :F(RETURN)
      SEARCHCONTRA()					  :S(RETURN)
      SIMPLEPROOF(PO_Q) 				  :S(PO_BOX)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
      PO_Q G_SPLIT					   :F(PO_10)
      LEQ(G_OP,'A')					    :F(PO_1)
      FREE(G_VAR)					   :S(PO_10)
      SIMPLEPROOF(G_ROP)				  :S(PO_BOX)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
      ADDGOAL(G_ROP)			    :F(PO_10)
*  CNH	2nd parameter to ADDGOAL removed
      PROOF()					:S(PO_BOX)F(FRETURN)
PO_1
*	OUTPUT = "I am at PO_1"
*        OUTPUT = "G_LOP = " G_LOP
*        OUTPUT = "G_ROP = " G_ROP
      LEQ(G_OP,'>')					    :F(PO_2)
      SIMPLEPROOF(G_ROP)				  :S(PO_BOX)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(G_LOP, 'ASSUME')				:S(RETURN)
*  changes Aug 5, 1986	CNH  onestep incorporated into addproof
      SIMPLEPROOF(G_ROP)				  :S(PO_BOX)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
      ADDGOAL(G_ROP)		      :F(PO_FINDS)
*  CNH	2nd parameter to ADDGOAL removed
      PROOF()					:S(PO_BOX)F(FRETURN)
PO_2
*	OUTPUT = "I am at PO_2"
      LEQ(G_OP,'&')					    :F(PO_5)
      SIMPLEPROOF(G_LOP)				    :S(PO_3)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
      ADDGOAL(G_LOP)			     :F(PO_3)
*  CNH	2nd parameter to ADDGOAL removed
      PROOF()						 :F(FRETURN)
PO_3
*	OUTPUT = "I am at PO_3"
      SIMPLEPROOF(PO_Q) 				  :S(PO_BOX)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
      SIMPLEPROOF(G_ROP)				    :S(PO_4)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
      ADDGOAL(G_ROP)		      :F(PO_FINDS)
*  CNH	2nd parameter to ADDGOAL removed
      PROOF()						 :F(FRETURN)
PO_4
*	OUTPUT = "I am at PO_4"
      SIMPLEPROOF(PO_Q) 		    :S(PO_BOX)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
*  changes Aug 6, 1986	CNH  fail condition moved to after sp_val check
      SP_VAL 'T'				 :S(RETURN)F(PO_FINDS)
PO_5
*	OUTPUT = "I am at PO_5"
      LEQ(G_OP,'=')					    :F(PO_8)
      SIMPLEPROOF('(' G_ROP '>' G_LOP ')')		    :S(PO_6)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
      ADDGOAL('(' G_ROP '>' G_LOP ')')	     :F(PO_6)
*  CNH	2nd parameter to ADDGOAL removed
      PROOF()						 :F(FRETURN)
PO_6
*	OUTPUT = "I am at PO_6"
      SIMPLEPROOF(PO_Q) 				  :S(PO_BOX)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
      SIMPLEPROOF('(' G_LOP '>' G_ROP ')')		    :S(PO_7)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
*  CNH	2nd parameter to ADDGOAL removed
      ADDGOAL('(' G_LOP '>' G_ROP ')')
+						     :F(PO_FINDS)
      PROOF()						 :F(FRETURN)
PO_7
*	OUTPUT = "I am at PO_7"
      SIMPLEPROOF(PO_Q) 		    :S(PO_BOX)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
*  changes Aug 6, 1986	CNH  fail check put after sp_val check
      SP_VAL 'T'				 :S(RETURN)F(PO_FINDS)
PO_8
*	OUTPUT = "I am at PO_8"
      PO_Q G_SPLIT					   :F(PO_10)
      LEQ(G_OP,'~')					   :F(PO_10)
      ADDPROOF(G_ROP, 'ASSUME')
+						 :S(RETURN)
*  CNH	3rd parameter to ADDPROOF removed
*  changes Aug 5, 1986	CNH  onestep incorporated into addproof
      IND = 'ON'				      :(PO_FINDS)
PO_10
*	OUTPUT = "I am at PO_10"
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF('~' PO_Q, 'ASSUME')
+						 :S(RETURN)
*  changes Aug 5, 1986	CNH  onestep incorporated into addproof
      IND = 'ON'
PO_FINDS
*	OUTPUT = "I am at PO_FINDS"
      PTR_NEGS = G_NSTACK
      PTR_BCS = G_BCSTACK
      PTR_ANDS = G_ANDSTACK
      PTR_QNS = G_NSTACK
PO_INDIRECT
*	OUTPUT = "I am at PO_INDIRECT"
      ALLDONE = 'F'
      PO_OLDCUR = G_CURLINE
      SEARCHCONTRA()				      :S(RETURN)
PO_1LOOP
*	OUTPUT = "I am at PO_1LOOP"
      ALLDONE 'T'				      :S(PO_1DONE)
      ALLDONE = 'T'
PO_1NEG
*	OUTPUT = "I am at PO_1NEG"
      IDENT(FLINK(G_NSTACK),PTR_NEGS)		      :S(PO_1BCS)
      ALLDONE = 'F'
      PTR_NEGS = BLINK(PTR_NEGS)
      FINDNEGS(PTR_NEGS)			     :S(RETURN)
PO_1BCS
*	OUTPUT = "I am at PO_1BCS"
      IDENT(FLINK(G_BCSTACK),PTR_BCS)		      :S(PO_1ANDS)
      ALLDONE = 'F'
      PTR_BCS = BLINK(PTR_BCS)
      FINDBCS(PTR_BCS)				     :S(RETURN)
PO_1ANDS
*	OUTPUT = "I am at PO_1ANDS"
      IDENT(FLINK(G_ANDSTACK),PTR_ANDS) 	      :S(PO_1LOOP)
      ALLDONE = 'F'
      PTR_ANDS = BLINK(PTR_ANDS)
      FINDANDS(PTR_ANDS)			     :S(RETURN)F(PO_1LOOP)
PO_1DONE
*	OUTPUT = "I am at PO_1DONE"
      FINDMPS() 				      :S(RETURN)
      FINDMTS() 				       :S(RETURN)
PO_2NEG
*	OUTPUT = "I am at PO_2NEG"
      IDENT(FLINK(G_NSTACK),PTR_NEGS)		      :S(PO_2DONE)
      PTR_NEGS = BLINK(PTR_NEGS)
      FINDNEGS(PTR_NEGS)			     :S(RETURN)F(PO_2NEG)
PO_2DONE
*	OUTPUT = "I am at PO_2DONE"
      FINDMTP() 				      :S(RETURN)
*        OUTPUT = "FINDMTP failed"
      FINDALLEI()				      :S(RETURN)
PO_3QNS
*	OUTPUT = "I am at PO_3QNS" 
*        XPXPROHIB = CONVERT(G_PROHIBLIST,'ARRAY')     :F(PO_XPXP)
*        OUTPUT = 'Prohiblist is:'
*        OUTPUT = XPXPROHIB                            :(PO_XPXP1)
*PO_XPXP OUTPUT = 'PROHIBLIST is empty'
*PO_XPXP1
      IDENT(FLINK(G_NSTACK),PTR_QNS)		      :S(PO_3DONE)
      PTR_QNS = BLINK(PTR_QNS)
      FINDQNS(PTR_QNS)				     :S(RETURN)F(PO_3QNS)
PO_3DONE
*	OUTPUT = "I am at PO_3DONE"
      SIMPLEPROOF(PO_Q) 				  :S(PO_BOX)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
      FINDUIA() 					  :S(RETURN)
      FINDID()					      :S(RETURN)
      FINDNEGID()					  :S(RETURN)
      FINDUIPROHIBCNT() 				  :S(RETURN)
      EQ(G_CURLINE,PO_OLDCUR)			     :F(PO_INDIRECT)
      SIMPLEPROOF(PO_Q) 				  :S(PO_BOX)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
PO_EIS
*	OUTPUT = "I am at PO_EIS"
*  changes Aug 6, 1986	CNH  check if show closed
      P_CURLEV = G_CURLEVEL
      FINDEI()						 :F(PO_NEWS)
      GT(P_CURLEV,G_CURLEVEL)			 :S(RETURN)
      FINDUIA() 					  :S(RETURN)
      SIMPLEPROOF(PO_Q) 				  :S(PO_BOX)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
PO_1INDIRECT
*	OUTPUT = "I am at PO_1INDIRECT"
      ALLDONE = 'F'
      PO_OLDCUR = G_CURLINE
      SEARCHCONTRA()				      :S(RETURN)
PO_11LOOP
*	OUTPUT = "I am at PO_11LOOP"
      ALLDONE 'T'				      :S(PO_11DONE)
      ALLDONE = 'T'
PO_11NEG
*	OUTPUT = "I am at PO_11NEG"
      IDENT(FLINK(G_NSTACK),PTR_NEGS)		      :S(PO_11BCS)
      ALLDONE = 'F'
      PTR_NEGS = BLINK(PTR_NEGS)
      FINDNEGS(PTR_NEGS)			     :S(RETURN)
PO_11BCS
*	OUTPUT = "I am at PO_11BCS"
      IDENT(FLINK(G_BCSTACK),PTR_BCS)		      :S(PO_11ANDS)
      ALLDONE = 'F'
      PTR_BCS = BLINK(PTR_BCS)
      FINDBCS(PTR_BCS)				     :S(RETURN)
PO_11ANDS
*	OUTPUT = "I am at PO_11ANDS"
      IDENT(FLINK(G_ANDSTACK),PTR_ANDS) 	      :S(PO_11LOOP)
      ALLDONE = 'F'
      PTR_ANDS = BLINK(PTR_ANDS)
      FINDANDS(PTR_ANDS)			     :S(RETURN)F(PO_11LOOP)
PO_11DONE
*	OUTPUT = "I am at PO_11DONE"
      FINDMPS() 				      :S(RETURN)
      FINDMTS() 				      :S(RETURN)
PO_12NEG
*	OUTPUT = "I am at PO_12NEG"
      IDENT(FLINK(G_NSTACK),PTR_NEGS)		      :S(PO_12DONE)
      PTR_NEGS = BLINK(PTR_NEGS)
      FINDNEGS(PTR_NEGS)			     :S(RETURN)F(PO_12NEG)
PO_12DONE
*	OUTPUT = "I am at PO_12DONE"
      FINDMTP() 				      :S(RETURN)
PO_13QNS
*	OUTPUT = "I am at PO_13QNS"
      IDENT(FLINK(G_NSTACK),PTR_QNS)		      :S(PO_13DONE)
      PTR_QNS = BLINK(PTR_QNS)
      FINDQNS(PTR_QNS)				    :S(RETURN)F(PO_13QNS)
PO_13DONE
*	OUTPUT = "I am at PO_13DONE"
      FINDUIA() 					  :S(RETURN)
      FINDID()					      :S(RETURN)
* June 1997 PO_1INDIRECT changed to PO_INDIRECT
      EQ(PO_OLDCUR,G_CURLINE)		   :F(PO_INDIRECT)S(PO_EIS)
PO_NEWS
*	OUTPUT = "I am at PO_NEWS"
      IDENT(PO_NEWLIST) 			  :F(PO_DESPARATION)
      PO_NEWLIST = FINDVARS()			  :F(PO_DESPARATION)
      FINDUI(PO_NEWLIST)				  :S(RETURN)
      SIMPLEPROOF(PO_Q) 				  :S(PO_BOX)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
PO_2INDIRECT
*	OUTPUT = "I am at PO_2INDIRECT"
      ALLDONE = 'F'
      PO_OLDCUR = G_CURLINE
      SEARCHCONTRA()				      :S(RETURN)
PO_2LOOP
*	OUTPUT = "I am at PO_2LOOP"
      ALLDONE 'T'				      :S(PO_21DONE)
      ALLDONE = 'T'
PO_21NEG
*	OUTPUT = "I am at PO_21NEG"
      IDENT(FLINK(G_NSTACK),PTR_NEGS)		      :S(PO_21BCS)
      ALLDONE = 'F'
      PTR_NEGS = BLINK(PTR_NEGS)
      FINDNEGS(PTR_NEGS)			     :S(RETURN)
PO_21BCS
*	OUTPUT = "I am at PO_21BCS"
      IDENT(FLINK(G_BCSTACK),PTR_BCS)		      :S(PO_21ANDS)
      ALLDONE = 'F'
      PTR_BCS = BLINK(PTR_BCS)
      FINDBCS(PTR_BCS)				     :S(RETURN)
PO_21ANDS
*	OUTPUT = "I am at PO_21ANDS"
      IDENT(FLINK(G_ANDSTACK),PTR_ANDS) 	      :S(PO_2LOOP)
      ALLDONE = 'F'
      PTR_ANDS = BLINK(PTR_ANDS)
      FINDANDS(PTR_ANDS)			     :S(RETURN)F(PO_2LOOP)
PO_21DONE
*	OUTPUT = "I am at PO_21DONE"
      FINDMPS() 				      :S(RETURN)
      FINDMTS() 				      :S(RETURN)
PO_22NEG
*	OUTPUT = "I am at PO_22NEG"
      IDENT(FLINK(G_NSTACK),PTR_NEGS)		      :S(PO_22DONE)
      PTR_NEGS = BLINK(PTR_NEGS)
      FINDNEGS(PTR_NEGS)			     :S(RETURN)F(PO_22NEG)
PO_22DONE
*	OUTPUT = "I am at PO_22DONE"
      FINDMTP() 				      :S(RETURN)
PO_23QNS
*	OUTPUT = "I am at PO_23QNS"
      IDENT(FLINK(G_NSTACK),PTR_QNS)		      :S(PO_23DONE)
      PTR_QNS = BLINK(PTR_QNS)
      FINDQNS(PTR_QNS)				     :S(RETURN)F(PO_23QNS)
PO_23DONE
*	OUTPUT = "I am at PO_23DONE"
      FINDUIA() 					  :S(RETURN)
      FINDID()					      :S(RETURN)
*  CHANGES July 23, 1986  CNH
*  add a call to findnegid
      FINDNEGID()					:S(RETURN)
      EQ(PO_OLDCUR,G_CURLINE)		   :F(PO_2INDIRECT)S(PO_EIS)
PO_DESPARATION
*	OUTPUT = "I am at PO_DESPARATION"
      DO_WHAT 'F'					  :S(FRETURN)
      DESP_LIST =  DESP()
*     OUTPUT = 'IN DESPARATION'
      HOLD = ARRAY('500,2')
* CHANGES Aug 21, 1986	CNH  parallelism  check parameter
      PTR_WEDGE = BLINK(G_ORSTACK)
      PTR_ARROW = BLINK(G_IMPLSTACK)
      PTR_NEG = G_NSTACK
      HOLD_SIZE = 0
PO_TRYANYTHING
*	OUTPUT = "I am at PO_TRYANYTHING"
      TRY = 'ON'
      PO_R = SEARCHNEG(.PTR_NEG)			 :F(TRY_WEDGE)
      PO_RL = FINDANTE(PO_R)
      PO_R G_SPLIT
      PO_S = G_ROP
      PO_S G_SPLIT
      PO_ROP1 = G_ROP
      PO_LOP1 = G_LOP
      G_OP  '>' | '=' | '&'				  :F(PO_ORS)
      PO_WH = 5
      SIMPLEPROOF(PO_S) 				  :S(PO_100)
*     OUTPUT = 'CALLED SIMPLEPROOF FAILED SP_VAL: ' SP_VAL
      SP_VAL 'T'				 :S(RETURN)
      START = G_CURLINE + 1
      LASTFREE = G_LASTFREEVAR
* CNT SAVE THE G_VLIST
      SAV_VLIST = CONVERT(G_VLIST,'ARRAY')
      SAV_VLIST = CONVERT(SAV_VLIST,'TABLE')
      ADDGOAL(PO_S)					:F(CONCHECK)
*  CNH	2nd parameter to ADDGOAL removed
      PROOF('F')					:F(COPY_HOLD)
PO_100
*	OUTPUT = "I am at PO_100 with goal" PO_S
      SEARCHCONTRA()				:F(FRETURN)S(RETURN)
PO_ORS
*	OUTPUT = "I am at PO_ORS"
      LEQ(G_OP,'+')					:F(TRY_WEDGE)
      PO_T = FINDANTE(G_LOP)				   :S(PO_99)
      PO_T = FINDANTE(G_ROP)				   :F(PO_98)
*  CHANGES  June 23, 1986   CNH
*  The cases PO_99, PO_98 and PO_97 were changed to pass
*  SHOWN two parameters since they are contradictions
PO_99
*	OUTPUT = "I am at PO_99"
      CRIT1 = G_CURLINE + 1
      ADDPRII(PO_S,PRLINE(PO_T) ',ADD')
*  changes Aug 5, 1986	CNH  addproof to addprii
*  CNH	3rd parameter to ADDPROOF removed
      CRIT2 = G_CURLINE + 1
      ADDPRII(PO_R,PRLINE(PO_RL) ',R')
*  changes Aug 5, 1986	CNH  addproof to addprii
*  CNH	3rd parameter to ADDPROOF removed
      CRIT2 = GT(CRIT2,G_CURLINE) PRLINE(PO_RL)
      SHOWN(CRIT1 ' , ' CRIT2)				   :(RETURN)
PO_98
*	OUTPUT = "I am at PO_98"
      START = G_CURLINE + 1
      LASTFREE = G_LASTFREEVAR
* CNT SAVE THE G_VLIST
      SAV_VLIST = CONVERT(G_VLIST,'ARRAY')
      SAV_VLIST = CONVERT(SAV_VLIST,'TABLE')
      ADDGOAL(G_LOP)			     :F(PO_97)
      PO_WH = 6
*  CNH	2nd parameter to ADDGOAL removed
      PROOF('F')					:F(COPY_HOLD)
      CRIT1 = G_CURLINE + 1
      ADDPRII(PO_R, PRLINE(PO_RL) ',R')
*  changes Aug 5, 1986	CNH  addproof to addprii
*  CNH	3rd parameter to ADDPROOF removed
      CRIT1 = GT(CRIT1,G_CURLINE) PRLINE(PO_RL)
      CRIT2 = G_CURLINE + 1
      PO_T = FINDANTE(PO_LOP1)
      ADDPRII(PO_S,PRLINE(PO_T) ',ADD')
*  changes Aug 5, 1986	CNH  addproof to addprii
*  CNH	3rd parameter to ADDPROOF removed
      SHOWN(CRIT1 ' , ' CRIT2)				   :(RETURN)
PO_97
*	OUTPUT = "I am at PO_97"
      START = G_CURLINE + 1
      LASTFREE = G_LASTFREEVAR
* CNT SAVE THE G_VLIST
      SAV_VLIST = CONVERT(G_VLIST,'ARRAY')
      SAV_VLIST = CONVERT(SAV_VLIST,'TABLE')
      ADDGOAL(G_ROP)					:F(TRY_WEDGE)
      PO_WH = 7
*  CNH	2nd parameter to ADDGOAL removed
      PROOF('F')					:F(COPY_HOLD)
      CRIT1 = G_CURLINE + 1
      ADDPRII(PO_R, PRLINE(PO_RL) ',R')
*  changes Aug 5, 1986	CNH  addproof to addprii
*  CNH	3rd parameter to ADDPROOF removed
      CRIT1 = GT(CRIT1,G_CURLINE) PRLINE(PO_RL)
      CRIT2 = G_CURLINE + 1
      PO_T = FINDANTE(PO_ROP1)
      ADDPRII(PO_S,PRLINE(PO_T) ',ADD')
*  changes Aug 5, 1986	CNH  addproof to addprii
*  CNH	3rd parameter to ADDPROOF removed
      SHOWN(CRIT1 ' , ' CRIT2)				   :(RETURN)
TRY_WEDGE
*	OUTPUT = "I am at PO_WEDGE"
      IDENT(PTR_WEDGE,G_ORSTACK)			  :S(TRY_ARROW)
      PO_WH = 1
      PO_X = LINE(PTR_WEDGE)
      PTR_WEDGE = BLINK(PTR_WEDGE)
      PO_X G_SPLIT					  :F(CONCHECK)
      ~FINDANTE(G_ROP) ~FINDANTE(G_LOP) 		  :F(TRY_WEDGE)
      LASTFREE = G_LASTFREEVAR
* CNT SAVE THE G_VLIST
      SAV_VLIST = CONVERT(G_VLIST,'ARRAY')
      SAV_VLIST = CONVERT(SAV_VLIST,'TABLE')
      START = G_CURLINE + 1
      EGREF(G_ROP)					  :S(PO_89)
      CHECK_DESP('~' G_LOP)			      :F(PO_89)
      ADDGOAL('~' G_LOP)				  :F(PO_89)
      PROOF('F')			    :S(PO_INDIRECT)F(COPY_HOLD)
PO_89
*	OUTPUT = "I am at PO_89"
      PO_WH = 2 					:(TRY_WEDGE)
      EGREF(G_LOP)					  :S(TRY_WEDGE)
      CHECK_DESP('~' G_ROP)			      :F(TRY_WEDGE)
      ADDGOAL('~' G_ROP)				  :F(TRY_WEDGE)
      PROOF('F')			    :S(PO_INDIRECT)F(COPY_HOLD)
TRY_ARROW
*	OUTPUT = "I am at TRY_ARROW"
      IDENT(PTR_ARROW,G_IMPLSTACK)			  :S(ENTER_HOLD)
      PO_WH = 3
      PO_X = LINE(PTR_ARROW)
      PTR_ARROW = BLINK(PTR_ARROW)
      PO_X G_SPLIT					  :F(CONCHECK)
      ~FINDANTE(G_ROP) ~FINDANTE('~' G_LOP)		  :F(TRY_ARROW)
      LASTFREE = G_LASTFREEVAR
* CNT SAVE THE G_VLIST
      SAV_VLIST = CONVERT(G_VLIST,'ARRAY')
      SAV_VLIST = CONVERT(SAV_VLIST,'TABLE')
      START = G_CURLINE + 1
      EGREF(G_ROP)					  :S(PO_90)
      CHECK_DESP(G_LOP) 			  :F(PO_90)
      ADDGOAL(G_LOP)					  :F(PO_90)
      PROOF('F')			    :S(PO_INDIRECT)F(COPY_HOLD)
PO_90
*	OUTPUT = "I am at PO_90"
      PO_WH = 4 					:(TRY_ARROW)
      G_LOP '~' =					  :F(A_OK)
      EGREF(G_LOP)					  :S(TRY_ARROW)
A_OK
*	OUTPUT = "I am at A_OK"
      CHECK_DESP('~' G_ROP)			      :F(TRY_ARROW)
      ADDGOAL('~' G_ROP)				  :F(TRY_ARROW)
      PROOF('F')			    :S(PO_INDIRECT)F(COPY_HOLD)
COPY_HOLD
* CNT CHECK PARALLEL
      EQ(P_PARALLEL,0)					 :S(FRETURN)
*	OUTPUT = "I am at COPY_HOLD"
      COUNT = 1
      TRY_SIZE = G_CURLINE + 1 - START
*     LE(TRY_SIZE,1)				      :S(PO_NOSHOW)
      SHOWN(,'T')					 :(PODIDSHOW)
PO_NOSHOW
*	OUTPUT = "I am at PO_NOSHOW"
      G_CURLEVEL = G_CURLEVEL - 1
PODIDSHOW
*	OUTPUT = "I am at PODIDSHOW"
      G_CURLINE = START - 1
      G_LASTFREEVAR = LASTFREE
*CNT OUTPUT
*      OUTPUT = 'G_LASTFREEVAR is now ' G_LASTFREEVAR
* CNT RECALL THE G_VLIST
      G_VLIST = CONVERT(SAV_VLIST,'ARRAY')
      G_VLIST = CONVERT(G_VLIST,'TABLE')
      GT(TRY_SIZE, HOLD_SIZE)				  :F(PO_WHICH)
CH_LOOP
*	OUTPUT = "I am at CH_LOOP"
      GT(COUNT, TRY_SIZE)				  :S(PO_WHICH)
      HOLD_SIZE = COUNT
      HOLD<COUNT,1> = G_PRMAT<START + COUNT - 1, 1>
      HOLD<COUNT,1> = G_PRMAT<START + COUNT - 1, 1>
      HOLD<COUNT,2> = G_PRMAT<START + COUNT - 1, 2>
      COUNT = COUNT + 1 				  :(CH_LOOP)
PO_WHICH
*	OUTPUT = "I am at PO_WHICH"
      COUNT = START + TRY_SIZE - 1
      G_CURLEVEL = G_PRMAT<START,3>
CL_LOOP
*	OUTPUT = "I am at CL_LOOP"
      LT(COUNT, START)				      :S(CHOOSE)
      LEQ(G_PRMAT<COUNT,1>, WHAT(G_GOALS))	      :F(CL_CONT)
      G_GOALS = GOALLINK(G_GOALS)
CL_CONT
*	OUTPUT = "I am at CL_CONT"
      G_PRMAT<COUNT,1> = 
      G_PRMAT<COUNT,2> = 
      G_PRMAT<COUNT,3> = 
      G_PRMAT<COUNT,4> = 
      G_PRMAT<COUNT,5> = 
      G_PRMAT<COUNT,6> = 
      G_PRMAT<COUNT,7> = 
      COUNT = COUNT - 1 			      :(CL_LOOP)
CHOOSE
*	OUTPUT = "I am at CHOOSE"
      EQ(PO_WH,5)					:S(PO_TRYANYTHING)
      EQ(PO_WH,6)					:S(PO_97)
      EQ(PO_WH,7)					:S(PO_TRYANYTHING)
      EQ(PO_WH,1)					  :S(PO_89)
      EQ(PO_WH,2)					  :S(TRY_WEDGE)
      EQ(PO_WH,3)					  :S(PO_90)
      EQ(PO_WH,4)				   :S(TRY_ARROW)F(CONCHECK)
ENTER_HOLD
*      OUTPUT = "I am at ENTER_HOLD"
      EQ(HOLD_SIZE,0)				      :S(PO_HELP)
*      OUTPUT = "HOLD_SIZE <> 0"
      OFFSET = G_CURLINE 
      ADDSHOW(1,HOLD_SIZE)
      PROOF('S')			    :F(FRETURN)S(PO_INDIRECT)
PO_BOX
*  CHANGE June 20, 1986  CNH
*  A parameter to SHOWN is given, being the last line in
*  The show (not including those already boxed in)
      CRIT = G_CURLINE
PO_BOX2
*  CHANGES July 29, 1986  CNH
*  put a check for null g_goals
      IDENT(G_GOALS)					 :S(RETURN)
      CRIT = DIFFER(G_PRMAT<PRLIN(G_GOALS),3> + 1,
+		    G_PRMAT<CRIT,3>)  CRIT - 1	       :S(PO_BOX2)
      SHOWN(CRIT)					   :(RETURN)
PO_HELP
*	OUTPUT = "I am at PO_HELP"
*  Let's try to see if maybe there's something on someone's
*  prohibited list.  If there is, we'll branch back to indirect;
*  otherwise I guess we're stuck.
      PO_OLDCUR = G_CURLINE
      FINDUIPROHIBALL() 				  :S(RETURN)
      EQ(PO_OLDCUR,G_CURLINE)			     :F(PO_INDIRECT)
*
*  Control arrives here when we don't know what to do.
*  Involk the command parser with the list of legal commands
*  from the G_STUCKCMDS table. If the parser returns a null
*  we must assume that the last command was supposed to be
*  a "STOP" instead of an (EOF). A return code of "STOP"
*  will quit this proof and jump back to the main command parser.
*  Finally if the RC is "CONT" THINKER will attempt to continue
*  the proof. SETATN is turned on and off here so as not
*  to have the attention trap set during user input.
*
PO_STUCK
*	OUTPUT = "I am at PO_STUCK"
      IDENT(G_OS,G_MTS) SETATN('OFF')
      PUTLINE(" *** PROBLEM ***")
      PUTLINE(" To err is human,")
      PUTLINE(" To forgive divine.")
      PUTLINE(" I am lost,")
*  CHANGES June 30, 1986   CNH
*  The last line was missing
      PUTLINE(" What's the next line? ")
      PO_CMDRC = CMDPARSE(G_STUCKCMDS,'addline','?')
      IDENT(G_OS,G_MTS) SETATN("ON")
      LEQ(PO_CMDRC,"STOP")				 :S(FRETURN)
      LEQ(PO_CMDRC,"CONT")			     :S(PO_INDIRECT)
*  changes Aug 5, 1986	CNH
*  add a check for DONE  (if the line added closed the show)
      LEQ(PO_CMDRC,"DONE")			 :S(RETURN)
PO_NOCONT
*	OUTPUT = "I am at PO_NOCONT"
      PROOF(LEQ(PO_CMDRC,"PROVE"))	   :F(FRETURN)S(PO_INDIRECT)
      PUTLINE(IDENT(PO_CMDRC) "  Assuming a STOP command")
+							 :S(FRETURN)
      PUTLINE(' Illigal return code from CMDPARSE: "' PO_CMDRC '"')
+							  :(FRETURN)
ENDPROOF
-STITLE
* Function: CHECK_DESP
* Date: Jan 21, 1987
* Author: Claire Harrison
* Purpose:  to check if the line has already been tried as 
*     the goal for a subproof that has been aborted.
* Calling sequence:
*     CHECK_DESP(line)
* Parameters:
*     line   - the line to check for priors
*
* On Return:
*     RETURN - didn't find the line
*     FRETURN - the line was found
*
      pDEFINE('CHECK_DESP(D_LINE)D_L')		      :(ENDCHECK_DESP)
CHECK_DESP
      D_L = DESP_LIST
CHD_LOOP
      IDENT(D_L)				      :S(CD_ADDIT)
      LEQ(WHAT(D_L),D_LINE)			      :S(FRETURN)
      D_L = NEXTLINK(D_L)			      :(CHD_LOOP)
CD_ADDIT
      DESP_LIST = DESP(D_LINE,DESP_LIST)	      :(RETURN)
ENDCHECK_DESP
* CHANGES   CNH  Sept 21, 1986	  add the function ADDSHOW  (needs doc)
      pDEFINE('ADDSHOW(START,END)LINE,SHEND,NUM1,NUM2,ANNOT') :(ENDADDSHOW)
ADDSHOW
*     OUTPUT = 'IN ADDSHOW START, END: ' START ',' END
*     OUTPUT = 'ADDING GOAL: ' HOLD<START,1>
      ADDGOAL(HOLD<START,1>)			      :F(CONCHECK)
      LINE = START + 1
AS_LOOP
      GT(LINE,END)				      :S(RETURN)
      HOLD<LINE,2> '*'				      :F(AS_ADDLINE)
      ANNOT = HOLD<LINE,2>
      NUM1 = 
      NUM2 =
      ANNOT SPAN('0123456789') . NUM1
      ANNOT SPAN('0123456789') . NUM2
      EQ(NUM1)					      :S(AS_NOEND)
      NUM1 = GT(NUM1) NUM1  - OFFSET
      NUM2 = GT(NUM2) NUM2  - OFFSET
      SHEND = NUM1
      SHEND = GT(NUM2, NUM1) NUM2
      ADDSHOW(LINE,SHEND)
      SHOWN(NUM1 ',' NUM2,'T')
      LINE = G_CURLINE + 1 - OFFSET		      :(AS_LOOP)
AS_ADDLINE
      ADDPRII(HOLD<LINE,1>,HOLD<LINE,2>)
      LINE = LINE + 1				      :(AS_LOOP)
AS_NOEND
      ADDSHOW(LINE,END) 			      :(RETURN)
ENDADDSHOW
-TITLE THINKER - a little theorem prover
-STITLE PATTERNS - formula manipulations patterns
*FRAME
* 
*	     GLOBAL PATTERNS FOR FORMULAS 
* 
      G_UNOP = '~'
      G_BINOP = ANY('+&>=')
      G_IDOP = '-'
      G_QUANTOP = 'A' | 'E'
*FRAME
*  G_BSPLIT when applied to a formula will split it at its 
*  main connective, placing the right operand in G_ROP, the left 
*  operand in G_LOP and the operator in G_OP.  Note that it expects
*  fully bracketed ( I still can't spell it Jeff ) expressions. 
      G_BSPLIT = POS(0) '(' BAL . G_LOP G_BINOP . G_OP BAL . G_ROP ')' RPOS(0)
*FRAME
*  G_USPLIT will split off a unary operator into G_OP and its operand
*  into G_ROP 
      G_USPLIT = POS(0) G_UNOP . G_OP BAL . G_ROP RPOS(0)
*FRAME
*  G_QSPLIT will split off a quantifier in G_OP (the quantifier) 
*  G_VAR (the quantified variable) and G_ROP (the rest of the formula)
      G_QSPLIT = POS(0) '(' G_QUANTOP $ G_OP BREAK(')') $ G_VAR ')'
+	 REM $ G_ROP RPOS(0)
*FRAME
*  G_IDSPLIT will split an identity formula into G_LOP and G_ROP
      G_IDSPLIT = POS(0) BREAK(G_IDOP) . G_LOP G_IDOP . G_OP REM . G_ROP
*FRAME
*  G_SPLIT will first try for binary the unary then quantified 
      G_SPLIT  =  G_BSPLIT | G_USPLIT | G_QSPLIT
-EJECT
*FRAME
*  A series of patterns to munge with formulas. 
*  They are all recursive and some are just damn ugly 
*  (but quite neat).  More explaination for each is contained 
*  within individual descriptions, but they all are very 
*  similar and all use the same basic pattern for variable 
*  names (a single letter prefix for each group and a single 
*  letter suffix for each application variable). The suffixes 
*  and their meanings are: 
*    ?C - this will match a constant 
*    ?V - this will match a variable 
*    ?T - this will match a term (variable or constant) 
*    ?A - an array to match the parameters to a predicate 
*	  where the index indicates how many parameters to 
*	  match. 
*    ?P - this will match a predicate and it's parameters 
*    ?Q - this will match a quantifer and variable enclosed 
*	  in brackets (ie (Ax1) or (Ey1) ) 
*    ?F - this will match a valid formula (this will usually 
*	  be the only pattern used in the rest of the program) 
* NOTE:
*  Before these paterns can be used the variable G_VARNAMES must 
*  be set and the function INITPAT() called.
*
      DEFINE("INITPAT()")			       :(ENDINITPAT)
INITPAT
-SPACE
*  First some basics
      G_N = ANY('0123456789')
-EJECT
*
*  BBBB
*  B   B
*  BBBB   patterns - match valid formulas
*  B   B
*  BBBB
*
*  The pattern G_BF will succeed if matched against a valid
*  formula and fail otherwise
*   eg	 X (G_BF RPOS(0)) :S(FORMOK)F(FORMBAD)
*
      G_BV = ANY(G_VARNAMES) G_N
      G_BC = ANY('abcd') G_N
      G_BT = G_BC | G_BV
      G_BA = ARRAY('0:9')
      G_BA<0> = '()'
      G_BA<1> = '(' G_BT ')'
      G_BA<2> = '(' G_BT "," G_BT  ')'
      G_BA<3> = '(' G_BT "," G_BT "," G_BT  ')'
      G_BA<4> = '(' G_BT "," G_BT "," G_BT "," G_BT  ')'
      G_BA<5> = '(' G_BT "," G_BT "," G_BT "," G_BT "," G_BT  ')'
      G_BP = G_PREDCHARS G_N "'" G_N $ G_NT *G_BA<G_NT>
      G_BQ = '(' G_QUANTOP G_BV ')'
      G_BF = G_BP		    |
+	 G_BT G_IDOP G_BT	    |
+	 G_UNOP *G_BF		    |
+	 G_BQ *G_BF		    |
+	 '(' *G_BF G_BINOP *G_BF ')'
-EJECT
*
*   CCC
*  C   C
*  C	   Patterns - for finding cases of unquantified pvariables
*  C   C
*   CCC
*
*  The pattern G_CF may be used to find the locations within a
*  string of unquantified instances of a certain pvariable.
*  When an occurence is found of a pvariable, the position of
*  the LAST character in the pvariable name is pushed onto the stack
*  which has root G_POSSTACK.  The pvariable to be found is determined
*  by the contents of the variable G_MV.
*  TO USE
*   1) set G_MV to the string matching the variable to be found
*   2) set G_POSSTACK to null
*   3) match G_CF to the formula in question
*   4) G_POSSTACK will now point to a stack of STACK datatypes
*      in which the V field of each STACK points to the last
*      character in an unquantified instance of the pvariable in
*      question
*   EG
*      FORM = "(P1'1(x1)>(Ax1)P1'1(x1))"
*      G_MV = "x1"
*      G_POSSTACK =
*      FORM G_CF
*   Result
*     G_POSSTACK --> STACK(8,)
*
*   WARNING - do not try to match against an invalid formula -
*	      the results are unpredictable
      G_CC = *G_MV @*PUSHPOS() | G_BC
      G_CV = *G_MV @*PUSHPOS() | G_BV
      G_CT = G_CC | G_CV
      G_CP = G_PREDCHARS G_N "'" G_N $ G_NT *G_CA<G_NT>
      G_CA = ARRAY('0:5')
      G_CA<0> = '()'
      G_CA<1> = '(' G_CT ')'
      G_CA<2> = '(' G_CT "," G_CT  ')'
      G_CA<3> = '(' G_CT "," G_CT "," G_CT  ')'
      G_CA<4> = '(' G_CT "," G_CT "," G_CT "," G_CT  ')'
      G_CA<5> = '(' G_CT "," G_CT "," G_CT "," G_CT "," G_CT  ')'
      G_CQ = '(' G_QUANTOP G_BV $ G_QV ')'
      G_CF = G_CP		     |
+	 G_CT G_IDOP G_CT	     |
+	 G_UNOP *G_CF		     |
+	 '(' *G_CF G_BINOP *G_CF ')' |
+	 G_CQ *CFF(G_QV)
-EJECT
*
*  PPPP
*  P   P
*  PPPP   Patterns - put away data about pterms
*  P
*  P
*
*  The patterns G_PF may be used to store all sorts of neat and
*  useful information about the occurances of pterms with a formula
*  It will store the following information in occurance tables
*  (an occurance table is simply a spitbol table which has as
*  an index a pvariable or pconstant and as contents a number
*  which usually indicates how many occurances of that pterm
*  have taken place) as follows:
*    1) for each constant the entry in G_CLIST is incremented
*    2) for each variable the entry in G_VLIST is incremented
*    3) for each free variable the entry in G_FVLIST is incremented
*
*  TO USE
*    1) match G_PF against the formula in question
*  EXAMPLE
*    FORM G_PF
*
*  WARNING - do not match against invalid formulas - results are
*	     unpredictable
      G_PC = G_BC $ G_C $ *PPC(G_C,G_CLIST,G_ADD)
      G_PV = G_BV $ G_V $ *PPC(G_V,G_VLIST,G_ADD)
+	 $ *PFV(G_V,G_FVLIST,G_QLIST,G_ADD)
      G_PT = G_PC | G_PV
      G_PA = ARRAY('0:6')
      G_PA<0> = '()'
      G_PA<1> = '(' G_PT ')'
      G_PA<2> = '(' G_PT "," G_PT  ')'
      G_PA<3> = '(' G_PT "," G_PT "," G_PT  ')'
      G_PA<4> = '(' G_PT "," G_PT "," G_PT "," G_PT  ')'
      G_PA<5> = '(' G_PT "," G_PT "," G_PT "," G_PT "," G_PT  ')'
      G_PQ = '(' G_QUANTOP G_BV $ G_Q $ *PPQ(.G_QLIST,G_ADD) ')'
      G_PP = G_PREDCHARS G_N "'" G_N $ G_NT *G_PA<G_NT>
      G_PF = G_PP		     |
+	 G_PT G_IDOP G_PT	     |
+	 G_UNOP *G_PF		     |
+	 '(' *G_PF G_BINOP *G_PF ')' |
+	 G_PQ *G_PF $ *PPQ(.G_QLIST,G_DELETE)
-EJECT
*
*  DDDD
*  D   D
*  D   D    Patterns - Delete information about pterms
*  D   D
*  DDDD
*
*  The pattern G_DF may be used to delete much of the information
*  stored by the G_PF pattern.	The occurance tables are modified
*  as follows:
*    1) for each constant found the G_CLIST entry is decremented
*    2) for each free variable found the G_FVLIST entry is decremented
*  TO USE
*    1) match G_DF against the formula in question
*  EXAMPLE
*    FORM G_DF
*
*  WARNING - do not match G_DF against an invalid formula - results
*	     are unpreditable
*
      G_DC = G_BC $ G_C $ *PPC(G_C,G_CLIST,G_DELETE)
      G_DV = G_BV $ G_V $ *PFV(G_V,G_FVLIST,G_QLIST,G_DELETE)
      G_DT = G_DC | G_DV
      G_DA = ARRAY('0:6')
      G_DA<0> = '()'
      G_DA<1> = '(' G_DT ')'
      G_DA<2> = '(' G_DT "," G_DT  ')'
      G_DA<3> = '(' G_DT "," G_DT "," G_DT  ')'
      G_DA<4> = '(' G_DT "," G_DT "," G_DT "," G_DT  ')'
      G_DA<5> = '(' G_DT "," G_DT "," G_DT "," G_DT "," G_DT  ')'
      G_DQ = '(' G_QUANTOP G_BV $ G_Q $ *PPQ(.G_QLIST,G_ADD) ')'
      G_DP = G_PREDCHARS G_N "'" G_N $ G_NT *G_DA<G_NT>
      G_DF = G_DP		     |
+	 G_DT G_IDOP G_DT	     |
+	 G_UNOP *G_DF		     |
+	 '(' *G_DF G_BINOP *G_DF ')' |
+	 G_DQ *G_DF $ *PPQ(.G_QLIST,G_DELETE)
-EJECT
*
*  FFFFF
*  F
*  FFFFF  Patterns - find free pvariables
*  F
*  F
*
*  The pattern G_FF may be used to find all occurances of
*  free pvariables within a formula.  The names of the
*  pvariables found are stored within the occurance table
*  G_FVL.
*  TO USE:
*    1) Initalize G_FVL to a table
*    2) match G_FF against the formula in question
*    3) G_FVL will now contain a count of the occurances of
*	all free pvariables and constants in the formula.
*  EXAMPLE:
*     FORM = (Az1)P1'5(z0,z1,z2,z2,z3)
*     G_FVL = TABLE(11)
*     FORM G_FF
*  RESULT
*     G_FVL<'z0'> = 1
*     G_FVL<'z2'> = 2
*     G_FVL<'z3'> = 1
*
*  WARNING - as always don't match G_FF against an invalid formula
      G_FT = (G_BC | G_BV) $ G_V $ *PFV(G_V,G_FVL,G_QL,G_ADD)
      G_FA = ARRAY('0:6')
      G_FA<0> = '()'
      G_FA<1> = '(' G_FT ')'
      G_FA<2> = '(' G_FT "," G_FT  ')'
      G_FA<3> = '(' G_FT "," G_FT "," G_FT  ')'
      G_FA<4> = '(' G_FT "," G_FT "," G_FT "," G_FT  ')'
      G_FA<5> = '(' G_FT "," G_FT "," G_FT "," G_FT "," G_FT  ')'
      G_FQ = '(' G_QUANTOP G_BV $ G_Q $ *PPQ(.G_QL,G_ADD) ')'
      G_FP = G_PREDCHARS G_N "'" G_N $ G_NT *G_FA<G_NT>
      G_FF = G_FP		     |
+	 G_FT G_IDOP G_FT	     |
+	 G_UNOP *G_FF		     |
+	 '(' *G_FF G_BINOP *G_FF ')' |
+	 G_FQ *G_FF $ *PPQ(.G_QL,G_DELETE)
-EJECT
*
*  L
*  L
*  L	  Patterns - locate free pvariables
*  L
*  LLLLL
*
*  The pattern G_LF may be used to find all occurances of
*  free pvariables within a formula.  The names of the
*  pvariables found are stored within the occurance table
*  G_FVL.
*  TO USE:
*    1) Initalize G_FVLOC to a table
*    2) match G_LF against the formula in question
*    3) G_FVLOC will now contain a count of the occurances of
*	all free pvariables and constants in the formula.
*  EXAMPLE:
*     FORM = (Az1)P1'5(z0,z1,z2,z2,z3)
*     G_FVLOC = TABLE(11)
*     FORM G_LF
*  RESULT
*     G_FVL<'z0'> = 1
*     G_FVL<'z2'> = 2
*     G_FVL<'z3'> = 1
*
*  WARNING - as always don't match G_LF against an invalid formula
      G_LT = @G_LOC (G_BC | G_BV) $ G_V $ *PFL(G_V,G_FVLOC,G_LOC,G_QL,G_ADD)
      G_LA = ARRAY('0:6')
      G_LA<0> = '()'
      G_LA<1> = '(' G_LT ')'
      G_LA<2> = '(' G_LT "," G_LT  ')'
      G_LA<3> = '(' G_LT "," G_LT "," G_LT  ')'
      G_LA<4> = '(' G_LT "," G_LT "," G_LT "," G_LT  ')'
      G_LA<5> = '(' G_LT "," G_LT "," G_LT "," G_LT "," G_LT  ')'
      G_LQ = '(' G_QUANTOP G_BV $ G_Q $ *PPQ(.G_QL,G_ADD) ')'
      G_LP = G_PREDCHARS G_N "'" G_N $ G_NT *G_LA<G_NT>
      G_LF = G_LP		     |
+	 G_LT G_IDOP G_LT	     |
+	 G_UNOP *G_LF		     |
+	 '(' *G_LF G_BINOP *G_LF ')' |
+	 G_LQ *G_LF $ *PPQ(.G_QL,G_DELETE)
-STITLE Routine to put away data about pterms for hypotheses
*
*  H   H
*  H   H
*  HHHHH  Patterns - put away data about pterms for hypotheses
*  H   H
*  H   H
*
*  The patterns G_HF is just the same at the pattern G_HF, except
*  that it records the information for hypotheses (premeses).
*  It will store the following information in occurance tables
*    1) for each constant the entry in G_CLIST is incremented
*    2) for each variable the entry in G_PREMVLIST is incremented
*    3) for each free variable the entry in G_PREMFVLIST is made
*
*  TO USE
*    1) match G_HF against the formula in question
*  EXAMPLE
*    FORM G_HF
*
*  WARNING - do not match against invalid formulas - results are
*	     unpredictable
      G_HC = G_BC $ G_C $ *PPC(G_C,G_CLIST,G_ADD)
      G_HV = G_BV $ G_V $ *PPC(G_V,G_PREMVLIST,G_ADD)
+	 $ *PFV(G_V,G_PREMFVLIST,G_QLIST,G_ADD)
      G_HT = G_HC | G_HV
      G_HA = ARRAY('0:6')
      G_HA<0> = '()'
      G_HA<1> = '(' G_HT ')'
      G_HA<2> = '(' G_HT "," G_HT  ')'
      G_HA<3> = '(' G_HT "," G_HT "," G_HT  ')'
      G_HA<4> = '(' G_HT "," G_HT "," G_HT "," G_HT  ')'
      G_HA<5> = '(' G_HT "," G_HT "," G_HT "," G_HT "," G_HT  ')'
      G_HQ = '(' G_QUANTOP G_BV $ G_Q $ *PPQ(.G_QLIST,G_ADD) ')'
      G_HP = G_PREDCHARS G_N "'" G_N $ G_NT *G_HA<G_NT>
      G_HF = G_HP		     |
+	 G_HT G_IDOP G_HT	     |
+	 G_UNOP *G_HF		     |
+	 '(' *G_HF G_BINOP *G_HF ')' |
+	 G_HQ *G_HF $ *PPQ(.G_QLIST,G_DELETE)
+							   :(RETURN)
ENDINITPAT
*FRAME
-STITLE Routine to Display the contents of a variable
*FRAME
* 
* Function Name:  DISPLAY 
* 
* Purpose:  To display the contents of a variable. 
*	       This routine displays the contents of any 
*	       variable (arrays, tables and fields). 
* 
* Calling Sequence: 
* 
*	    DISPLAY( VARIABLE, NAME, FLAG, LEVEL )	 :F(OOPS) 
* 
*	      VARIABLE : The variable to be displayed. 
*	      NAME : The name of the variable to be displayed. 
*	      FLAG : Flag indicating if the variable is to be 
*			 searched to it's maximum depth (handy 
*			 for printing out trees). 
*		       Null -> print top level only. 
*		       n -> Print only n levels down. 
*		       'R' -> print all levels. 
*	      LEVEL : The indentation level.  Should be either 
*			  omitted or set to zero.  Used in 
*			  recursive calls. 
* 
* Note:  If you discover a way of causing this routine to generate 
*	    a spitbol run-time error (other than by passing it 
*	    illegal NAME, FLAG or LEVEL values), please let me know.
* 
* Warning:  An attempt to display SETEXIT(.label) causes an error. 
*	    This appears to be a bug in SETEXIT that I can't seem 
*	    to avoid or prevent. 
* 
* Designer:  Daniel Boulet, UQV. 
* 
* Date:  August 7, 1981. 
* 
       DEFINE('DISPLAY(VARIABLE,NAME,FLAG,LEVEL)'
+	 'MYTABLE,TYPE,ARINDEX,DESCRIPTOR,N,VALUE,OUTLIST,'
+	 'INTYPE,NUMBER,SVALUE,ERRLIMIT')	       :(ENDDISPLAY)
DISPLAY
      LEVEL = LEVEL + 1
*      FLAG = 1 IDENT(FLAG)
      FLAG = 999999999 
*      LEQ(FLAG,'R')
      FLAG = FLAG - 1
      LEGALTYPES = FENCE ('STRING' | 'INTEGER' | 'REAL' | 'ARRAY' |
+		   'TABLE' | 'PATTERN' | 'NAME' | 'EXPRESSION' | 
+		   'CODE') RPOS(0)
*
* If this is the very first call to DISPLAY, create the data type
*    used to hold array descriptors.
*
      IDENT(display.inited)			:F(DISPLAY.NOTFIRST)
      DATA('ARRAYDESC(LOWER,STRIDE,INDEX,NEXT)')
      display.inited = 'yes'
DISPLAY.NOTFIRST
*
* Create a table to keep tract of unique structure numbers.
*   (The table is created at the top level call and deleted
*    at the end of that call).
*
      IDENT(display.number)			:F(DISPLAY.GOTTABLE)
      display.number = TABLE(100)
      display.unique = TABLE(100)
      item.last = 0
      MYTABLE = 'yes'
DISPLAY.GOTTABLE
*
* Convert name to printable form.
*
      ERRLIMIT = &ERRLIMIT
      &ERRLIMIT = 1
      ( NAME ' ' )				  :S(DISPLAY.NAMEOK)
      &ERRLIMIT = ERRLIMIT
      NAME = GETVALUE( NAME )
DISPLAY.NAMEOK
      &ERRLIMIT = ERRLIMIT
      DATATYPE(NAME)
*
* Get the type of the item and process it accordingly.
*
      TYPE = DATATYPE( VARIABLE )
      LEQ( TYPE, 'ARRAY' )			   :S(DISPLAY.ARRAY)
      LEQ( TYPE, 'TABLE' )			   :S(DISPLAY.TABLE)
      LEQ( TYPE, 'PATTERN' )			 :S(DISPLAY.PATTERN)
      TYPE LEGALTYPES				:F(DISPLAY.DATATYPE)
      SVALUE = CONVERT( VARIABLE, 'STRING' )	   :S(DISPLAY.VALUE)
						 :(DISPLAY.DATATYPE)
*
* Simple value:
*
DISPLAY.VALUE
*OUTPUT = '*** value'
      DISPLAYOUT( NAME ' = ' GETVALUE(VARIABLE), LEVEL )
+						   :(DISPLAY.RETURN)
*
* Pattern:
*
DISPLAY.PATTERN
*OUTPUT = '*** pattern'
      DISPLAYOUT( NAME ' = PATTERN #' ITEMNUMBER( VARIABLE ), LEVEL )
+						   :(DISPLAY.RETURN)
*
* Array:
*
DISPLAY.ARRAY
*OUTPUT = '*** array'
      IDENT( display.unique<VARIABLE> ) 	  :F(DISPLAY.RETURN)
      display.unique<VARIABLE> = 'done'
      DISPLAYOUT( NAME ' = ARRAY #' ITEMNUMBER(VARIABLE)
+	 '(' PROTOTYPE(VARIABLE) ')', LEVEL )
      GT( FLAG, 0 )				  :F(DISPLAY.RETURN)
      DESCRIPTOR = BUILDDESC( PROTOTYPE(VARIABLE) )
      INDEX(DESCRIPTOR) = INDEX(DESCRIPTOR) - 1
DISPLAY.AR.LOOP
      ARINDEX = ARNEXT( DESCRIPTOR )		  :F(DISPLAY.RETURN)
      VALUE = EVAL(  'VARIABLE<' ARINDEX '>' )
      DISPLAY( VALUE, NAME '<' ARINDEX '>', FLAG, LEVEL )
      N = N + 1 				  :(DISPLAY.AR.LOOP)
*
* Table:
*
DISPLAY.TABLE
*OUTPUT = '*** table'
      IDENT( display.unique<VARIABLE> ) 	  :F(DISPLAY.RETURN)
      display.unique<VARIABLE> = 'done'
      DISPLAYOUT( NAME ' = TABLE #' ITEMNUMBER(VARIABLE)
+	 '(' PROTOTYPE(VARIABLE) ')', LEVEL )
      GT( FLAG, 0 )				  :F(DISPLAY.RETURN)
      VALUE = CONVERT( VARIABLE, 'ARRAY' )	  :F(DISPLAY.RETURN)
      N = 1
DISPLAY.TB.LOOP
      VALUE< N, 1 >				  :F(DISPLAY.RETURN)
      DISPLAY( VALUE<N,2>, NAME '<' GETVALUE(VALUE<N,1>) '>', FLAG, LEVEL )
      N = N + 1 				  :(DISPLAY.TB.LOOP)
*
* Programmer defined data type.
*
DISPLAY.DATATYPE
*OUTPUT = '*** data type'
      IDENT( display.unique<VARIABLE> ) 	  :F(DISPLAY.RETURN)
      display.unique<VARIABLE> = 'done'
      OUTLIST = NULL
      N = 1
      INTYPE = NULL
      NUMBER = ' #' ITEMNUMBER(VARIABLE)
DISPLAY.PD.LOOP1
      FNAME = FIELD( TYPE, N )			:F(DISPLAY.PD.EXIT1)
      VALUE = EVAL( FNAME '(VARIABLE)' )
      DATATYPE(VALUE) POS(0) ( 'ARRAY' | 'TABLE' ) RPOS(0)
+					     :S(DISPLAY.PD.DATATYPE)
      DATATYPE(VALUE) LEGALTYPES	    :S(DISPLAY.PD.OTHER)
DISPLAY.PD.DATATYPE
      INTYPE = 'yes'
DISPLAY.PD.OTHER
      SVALUE = GETVALUE( VALUE )
      OUTLIST = OUTLIST ', ' SVALUE
      N = N + 1 				 :(DISPLAY.PD.LOOP1)
DISPLAY.PD.EXIT1
      OUTLIST FENCE ', ' = NULL
      DISPLAYOUT( NAME ' = ' TYPE NUMBER '( ' OUTLIST ' )',
+	 LEVEL )
      IDENT( INTYPE )				  :S(DISPLAY.RETURN)
*
* Data type had atleast one field that was either an ARRAY, TABLE
*   or programmer defined data type.
*
      GT(FLAG,0)				  :F(DISPLAY.RETURN)
      N = 1
DISPLAY.PD.LOOP2
      FNAME = FIELD( TYPE, N )			:F(DISPLAY.PD.EXIT2)
      VALUE = EVAL( FNAME '(VARIABLE)' )
      DATATYPE(VALUE) POS(0) ( 'ARRAY' | 'TABLE' ) RPOS(0)
+					    :S(DISPLAY.PD.DATATYPE2)
      DATATYPE(VALUE) LEGALTYPES	      :S(DISPLAY.PD.NEXT)
DISPLAY.PD.DATATYPE2
      DISPLAY( VALUE, FNAME '(' TYPE NUMBER ')', FLAG, LEVEL )
DISPLAY.PD.NEXT
      N = N + 1 				 :(DISPLAY.PD.LOOP2)
DISPLAY.PD.EXIT2
						   :(DISPLAY.RETURN)
*
* Done.  Delete the table (if we allocated it) and return.
*
DISPLAY.RETURN
      IDENT( MYTABLE )					  :S(RETURN)
      display.number = NULL
      display.unique = NULL				   :(RETURN)
ENDDISPLAY
-SPACE 5
*FRAME
* 
* Routine to return the value of a variable. 
* 
* The little dance with &ERRLIMIT is necessary to ensure we don't 
*    try to convert the name of a created variable into a string 
*    (which will go boom if we try). 
* 
       DEFINE('GETVALUE(VALUE)ERRLIMIT')	    :(SKIP.GETVALUE)
GETVALUE
      GETVALUE = 'NULL' IDENT(VALUE)			  :S(RETURN)
      LEQ(DATATYPE(VALUE),'NAME')		:F(GETVALUE.NOTNAME)
      ERRLIMIT = &ERRLIMIT
      &ERRLIMIT = 1
      ( NAME ' ' )				 :S(GETVALUE.NAMEOK)
      &ERRLIMIT = ERRLIMIT			:(GETVALUE.DATATYPE)
GETVALUE.NAMEOK
      &ERRLIMIT = ERRLIMIT
GETVALUE.NOTNAME
      LEQ(DATATYPE(VALUE),'STRING')		  :F(GETVALUE.OTHER)
      GETVALUE = VALUE
      GETVALUE BREAK('"') REM = "'" VALUE "'"		  :S(RETURN)
      GETVALUE = '"' VALUE '"'				   :(RETURN)
GETVALUE.OTHER
      DATATYPE(VALUE) LEGALTYPES	   :F(GETVALUE.DATATYPE)
      DATATYPE(VALUE) POS(0) ( 'ARRAY' | 'TABLE' | 'PATTERN' ) RPOS(0)
+					       :S(GETVALUE.DATATYPE)
      GETVALUE = CONVERT( VALUE, 'STRING' )		   :(RETURN)
GETVALUE.DATATYPE
      GETVALUE = DATATYPE( VALUE ) ' #' ITEMNUMBER( VALUE )
							   :(RETURN)
SKIP.GETVALUE
-SPACE 5
*FRAME
* 
* Routine to return the item number of a data type. 
* 
       DEFINE('ITEMNUMBER(DATATYPE)')		  :(SKIP.ITEMNUMBER)
ITEMNUMBER
      ITEMNUMBER = display.number< DATATYPE >
      IDENT(ITEMNUMBER) 				  :F(RETURN)
      item.last = item.last + 1
      display.number<DATATYPE> = item.last
      ITEMNUMBER = item.last				   :(RETURN)
SKIP.ITEMNUMBER
-SPACE 5
*FRAME
* 
* Routine to build an array descriptor linked list. 
* 
       DEFINE('BUILDDESC(PROTO)LOWER,UPPER,DIM')   :(SKIP.BUILDDESC)
BUILDDESC
      IDENT(PROTO)					  :S(RETURN)
      PROTO POS(0) ( BREAK(',') | REM ) . DIM ( ',' | NULL ) = NULL
      LOWER = 1
      UPPER = DIM
      DIM BREAK(':') . LOWER ':' REM . UPPER
      BUILDDESC = ARRAYDESC( LOWER, UPPER - LOWER + 1, LOWER,
+	 BUILDDESC( PROTO ) )				   :(RETURN)
SKIP.BUILDDESC
-SPACE 5
*FRAME
* 
* Routine to build an index list for an array element. 
* 
       DEFINE('ARNEXT(DESC,N)PTR')		      :(SKIP.ARNEXT)
ARNEXT
      PTR = DESC
ARNEXT.INCR
      INDEX(PTR) = INDEX(PTR) + 1
      LT( INDEX(PTR), LOWER(PTR) + STRIDE(PTR) )    :S(ARNEXT.BUILD)
      INDEX(PTR) = LOWER(PTR)
      PTR = NEXT(PTR)
      IDENT( PTR )			   :S(FRETURN)F(ARNEXT.INCR)
ARNEXT.BUILD
      PTR = DESC
ARNEXT.BUILD.LOOP
      ARNEXT = ARNEXT ',' INDEX(PTR)
      PTR = NEXT(PTR)
      IDENT( PTR )			       :F(ARNEXT.BUILD.LOOP)
      ARNEXT FENCE ',' = NULL
*     OUTPUT = '*** ARNEXT = ' ARNEXT
							   :(RETURN)
SKIP.ARNEXT
-SPACE 5
*FRAME
* 
* Display output routine. 
* 
       DEFINE('DISPLAYOUT(LINE,LEVEL)') 	    :(ENDDISPLAYOUT)
DISPLAYOUT
      PUTLINE('SPRINT', ' ' DUPL( '| ', LEVEL - 1 ) LINE )  :(RETURN)
ENDDISPLAYOUT
-STITL Function: ENTERCONSEQ(EC_NUM)
*FRAME
* Function: Enterconseq
* Date: July 16, 1986
* Author:  CNH
* Last modification: July 1986
* Purpose:
* to add lines from the premcon array into the proof
* it adds them recursively by first adding the two
* lines it is dependant on.  If the line is a premise
* it does nothing since these are added automatically
* by Addprii.  It uses addprii not addproof since
* all possible combinations of the identities is already
* known.  It deletes the entry in the g_idstack for
* each line since they will be added automatically
* by addprii (with a valid line number)
*
* Calling Sequence:
* ENTERCONSEQ(EC_NUM)
*
* Return:  normal termination
*
* Paramters:  EC_NUM the index into the array of the line 
*
* Local variables:
* c1,c2 : the indexes into the array of the parents
*	  and later their line number in the proof
* ec_line: the line to be added
* ec_ref:  the annotation for the line to be added
* ec_bit:  a part of the annotation
* ptr:	   a pointer in the idstack so the entry
*	   can be deleted
      pDEFINE('ENTERCONSEQ(EC_NUM)C1,C2,EC_LINE,EC_REF'
+     ',EC_BIT,PTR')				   :(ENDENTERCONSEQ)
ENTERCONSEQ
      NUMBER = SPAN('0123456789')
      DIFFER(PREMCON<EC_NUM>)				:F(RETURN)
      EC_LINE = PCENTRY(PREMCON<EC_NUM>)
      EC_REF = PCCOM(PREMCON<EC_NUM>)
      EC_REF 'PREM'					:S(RETURN)
      PTR = G_IDSTACK
EC_LOOP
+     PTR = DIFFER(FLINK(PTR), G_IDSTACK) FLINK(PTR)	:F(EC_ENDLOOP)
      DIFFER(LINE(PTR),EC_LINE) 			:S(EC_LOOP)
      DELCSTACK(BLINK(PTR))
EC_ENDLOOP
+     EC_REF NUMBER . C1 BREAK('0123456789') NUMBER . C2
      ENTERCONSEQ(C1)
      C1 = PRLINE(FINDANTE(PCENTRY(PREMCON<C1>)))
      ENTERCONSEQ(C2)
      C2 = PRLINE(FINDANTE(PCENTRY(PREMCON<C2>)))
      EC_REF NUMBER BREAK('0123456789') . EC_BIT NUMBER
+     = C1 EC_BIT C2
      ADDPRII(EC_LINE,EC_REF)				:(RETURN)
ENDENTERCONSEQ
-STITL Function: ADDCONSEQ(line,ref)
*FRAME
* Function: ADDCONSEQ
* Date: July 16, 1986
* Last modification: July 1986
* Author: CNH
* Purpose:
* this procedure is similar to the addproof in that it 
* recursively finds all identities based on the given
* identity and previously known stuff. It doesn't add
* them directly to the proof as this would violate the
* added-when-needed law for premises, instead it puts
* them into a structure to remember them.  Two structures
* are needed since the information must be retrievable
* both by line and reference number. the first is a
* table called ISPC which is referenced by the line
* and contains the reference number for the second
* structure.  The second is an array called
* PREMCON and contains the line and the annotation
* It also adds the lines onto the idstack even though
* they do not have a linenumber yet. (they are not yet
* entered into the proof).  It uses a utility procedure
* to actually add them in to various structures.
*
* Calling sequence:
* ADDCONSEQ(line,ref)
*
* Parameters:
*  line : the line to add
*  ref : its annotation
*
* RETURN:  normal termination
*
* local variables:
* pc_pres : reference number of the line
* pc_ptr : pointer into the idstack
* pc_fst,pc_snd,pc_l,pc_r : parts of identities
* pc_reason: annotation for line to be added
      pDEFINE('ADDCONSEQ(PC_LINE,PC_REF)PC_PRES,PC_PTR,PC_FST'
+     ',PC_SND,PC_L,PC_R,PC_REASON')		  :(ENDADDCONSEQ)
ADDCONSEQ
*  check if it is an identity
      PC_LINE POS(0) LEN(2) . PC_FST G_IDOP
+     LEN(2) . PC_SND					:F(RETURN)
      LEQ(PC_FST,PC_SND)				:S(RETURN)
*  it is so we add it to the various arrays
      ADDPC(PC_LINE,PC_REF)				:F(RETURN)
      PC_PRES = ISPC<PC_LINE>
*  go through the identity stack adding any consequences
*  of this identity to the various places
      PC_PTR = FLINK(FLINK(G_IDSTACK))
PC_LOOP
*  check if we have done all the items in the stack
      IDENT(PC_PTR,G_IDSTACK)				:S(RETURN)
      LINE(PC_PTR) POS(0) LEN(2) . PC_L G_IDOP
+     LEN(2) . PC_R					:F(RETURN)
      PC_REASON = PC_PRES ',' ISPC<LINE(PC_PTR)>
+     ',LL'
*  check each possible match of the line and the
*  identity on the stack
      LEQ(PC_L,PC_FST)					:F(PC_CASE2)
      LLT(PC_SND,PC_R)					:F(PC_C11)
      ADDCONSEQ(PC_SND '-' PC_R, PC_REASON)		:(PCEND)
PC_C11
+     ADDCONSEQ(PC_R '-' PC_SND, PC_REASON)		:(PCEND)
PC_CASE2
+     LEQ(PC_R,PC_FST)					:F(PC_CASE3)
      ADDCONSEQ(PC_L '-' PC_SND, PC_REASON)		:(PCEND)
PC_CASE3
+     LEQ(PC_L,PC_SND)					:F(PC_CASE4)
      ADDCONSEQ(PC_FST '-' PC_R, PC_REASON)		:(PCEND)
PC_CASE4
+     LEQ(PC_R,PC_SND)					:F(PCEND)
      LLT(PC_L,PC_FST)					:F(PC_C41)
      ADDCONSEQ(PC_L '-' PC_FST, PC_REASON)		:(PCEND)
PC_C41
+     ADDCONSEQ(PC_FST '-' PC_L, PC_REASON)
PCEND
*  repeat for the next item in the idstack
      PC_PTR = FLINK(PC_PTR)				:(PC_LOOP)
ENDADDCONSEQ
-STITLE Function: ADDPC(p_line,p_ref)
*FRAME
* Function: ADDPC
* Date: July 16, 1986
* Author: CNH
* Last modification: July 1986
* Purpose:
* add the line, ref to the table,array,linked lists
* Calling sequence:
* ADDPC(line,ref)
* Parameters:
* line: line to add
* ref:	reference for the line
* return: fails iff the line has already been added
* local variables:   NONE
      pDEFINE('ADDPC(P_LINE,P_REF)')			:(ENDADDPC)
ADDPC
*  check if it is already on the table
*  ispc is a table indexed by the line and containing
*  a reference number into premcon.  premcon is an array
*  containing the line and the annotation.
      DIFFER(ISPC<P_LINE>)				:S(FRETURN)
      PC_CUR = PC_CUR + 1
      ISPC<P_LINE> = PC_CUR
      PREMCON<PC_CUR> = PC(P_LINE,P_REF)
*  unless it is a premise we add it onto the cstack
*  premises are already added by doprem
      REF 'PREM'					:(RETURN)
      CSTACK(G_IDSTACK,P_LINE)				:(RETURN)
ENDADDPC
-STITL Function: DOPREM(DP_FORM,DP_OP)
*FRAME
* Function: DOPREM
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* do the operation specified by DP_OP on the matrix
* of premeses
* Currently valid operations:
*	'INIT' - create G_PREM
*	'ADD'  - add a new premise
*	'ENTER' - enter the current set of premises
*		  into the working structure
*	'ADDENTER' - ADD a new premise and ENTER it into the table
*		     immediately.
* 
* Calling sequence: 
*	DOPREM(DP_FORM,DP_OP)
* Parameters: 
*	DP_FORM = Formula to be added to Premise table.
*	DP_OP -  Operation to be done on DP_FORM.
* 
* On return: 
*	RETURN	- Normal termination.
*	FRETURN - Illegal DP_OP parameter
* 
* Local variables: 
*	DP_I - General index.
* 
      pDEFINE('DOPREM(DP_FORM,DP_OP)DP_I',,,'NULL')	:(ENDDOPREM)
DOPREM
+     LEQ(DP_OP,'INIT') 				 :S(DP_INIT)
      LEQ(DP_OP,'ADD')					  :S(DP_ADD)
      LEQ(DP_OP,'ADDENTER')			     :S(DP_ADDENTER)
      LEQ(DP_OP,'ENTER')		      :S(DP_ENTER)F(FRETURN)
DP_INIT
+     G_PREM = ARRAY(P_MAXPREM)
      G_CURPREM = 0					   :(RETURN)
DP_ADD
+     GT(G_CURPREM,P_MAXPREM)				 :S(FRETURN)
      G_CURPREM = G_CURPREM + 1
      DP_FORM = NORMALIZE(DP_FORM)
      G_PREM<G_CURPREM> = DP_FORM			   :(RETURN)
DP_ENTER
+     DP_I = 0
DP_LOOP1
+     DP_I = LT(DP_I,G_CURPREM) DP_I + 1	     :F(DP_ENDLOOP1)
* CHANGES July 22, 1986  CNH add 1 as third parameter
      ADDANTE(-DP_I,-1,G_PREM<DP_I>,.G_HF,1)
* CHANGES July 7, 1986	CNH   PREM
*  call ADDCONSEQ on the premise line 
      ADDCONSEQ(G_PREM<DP_I>, 'PREM')
      G_PREM<DP_I> G_QSPLIT				:F(DP_LOOP1)
      LEQ(G_OP,'A')					:F(DP_LOOP1)
      G_PRMAT<-DP_I,5> = STACK(-DP_I,'')		:(DP_LOOP1)
DP_ENDLOOP1
+							   :(RETURN)
DP_ADDENTER
+     GT(G_CURPREM,P_MAXPREM)				 :S(FRETURN)
      G_CURPREM = G_CURPREM + 1
      DP_FORM = NORMALIZE(DP_FORM)
      G_PREM<G_CURPREM> = DP_FORM
* CHANGES July 22, 1986  CNH add 1 as third parameter
      ADDANTE(-G_CURPREM,-1,DP_FORM,.G_HF,1)
* CHANGES July 7, 1986	CNH   PREM
*  call ADDCONSEQ on the premise line
      ADDCONSEQ(DP_FORM, 'PREM')
      DP_FORM G_QSPLIT					:F(RETURN)
      LEQ(G_OP,'A')					:F(RETURN)
      G_PRMAT<-G_CURPREM,5> = STACK(-G_CURPREM,'')	:(RETURN)
ENDDOPREM
-STITL Function: FINDALLEI()
*FRAME
* Function: FINDALLEI
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*      Calles FINDEI succesively untill they are all found.
* 
* Calling sequence: 
*	FINDALLEI()
* Parameters: None.
* 
* On return: 
*	RETURN	-Normal termination.
* 
* Local variables: None
*  Changes Aug 1, 1986	CNH
*  add a local variable curlev to be the show level before addproof
*  is called so I can tell if the show is closed
* 
      pDEFINE('FINDALLEI()CURLEV',,,'NULL')	     :(ENDFINDALLEI)
FINDALLEI
* changes Aug 1, 1986  CNH
* now three cases:  findei fails so freturn
*	 findei succeeds, still at same show level so continue
*	 findei succeeds, not at same show level so return
      CURLEV = G_CURLEVEL
      FINDEI()					      :F(FRETURN)
      GT(CURLEV,G_CURLEVEL)			      :S(RETURN)F(FINDALLEI)
*     FINDEI()				      :S(FINDALLEI)F(RETURN)
ENDFINDALLEI
-STITL Function: ADDALIST(AL_WHERE,AL_FROM)
*FRAME
* Function: ADDALIST
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* Copy to the ancestor list pointed to by AL_WHERE all of
* the ancestors of AL_FROM.
* 
* Calling sequence: 
*	ADDALIST(AL_WHERE,AL_FROM)
* Parameters: 
*	AL_WHERE - The insertion point passed by name.
*	AL_FROM - The proof matrix line number of where we are
*		  comming from.
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: 
*	AL_PTR - Pointer to the Proof matrix.
* 
      pDEFINE('ADDALIST(AL_WHERE,AL_FROM)AL_PTR',,
+	      'NAME,INTEGER','NULL')		      :(ENDADDALIST)
ADDALIST
+     AL_PTR = G_PRMAT<AL_FROM,5>
AL_LOOP
+     IDENT(AL_PTR)					  :S(RETURN)
      $AL_WHERE = STACK(V(AL_PTR),$AL_WHERE)
      AL_PTR = N(AL_PTR)				  :(AL_LOOP)
ENDADDALIST
-STITL Function: PROHIBIT(PH_OP,PH_TO,PH_VAR)
*FRAME
* Function: PROHIBIT
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* add/delete to/from the prohibit list specified by the alist of
* PH_TO.  Note that the line PH_TO should be an existentially
* quantified line
* 
* Calling sequence: 
*	PROHIBIT(PH_OP,PH_TO,PH_VAR)
* Parameters: 
*	PH_OP - Operation to be performed on line pointed to by
*		PH_TO.
*	PH_TO - Index into proof matrix.
*	PH_VAR - Name of variable in prohibitlist.
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: 
*	PH_ALIST - Pointer into linked list pointed to by 5th 
*		   column of proof matrix.
* 
      pDEFINE('PROHIBIT(PH_OP,PH_TO,PH_VAR)PH_ALIST',,
+	      'INTEGER,INTEGER,STRING|NULL','NULL')   :(ENDPROHIBIT)
PROHIBIT
*     OUTPUT = 'IN PROHIBIT OP, TO, VAR: ' PH_OP ',' PH_TO ',' PH_VAR
      EQ(PH_OP,G_DELETE)				  :F(PH_ADD)
      PH_VAR = G_PRMAT<PH_TO,6> 			  :(PH_DOIT)
PH_ADD
+     G_PRMAT<PH_TO,6> = PH_VAR
PH_DOIT
+     PH_ALIST = G_PRMAT<PH_TO,5>
PH_LOOP
+     IDENT(PH_ALIST)					  :S(RETURN)
      G_PRMAT<V(PH_ALIST),6> = IDENT(G_PRMAT<V(PH_ALIST),6>) + 0
*     OUTPUT = 'PROHIBIT: LINE, VAL: ' V(PH_ALIST) ',' G_PRMAT<V(PH_ALIST),6>
      G_PROHIBLIST<PH_VAR> = G_PROHIBLIST<PH_VAR> +
+	 PH_OP
      PH_ALIST = N(PH_ALIST)				  :(PH_LOOP)
ENDPROHIBIT
-STITL Function: FINDUIPROHIBCNT()
*FRAME
* Function: FINDUIPROHIBCNT
* Date: Wed Dec 12 20:08:56 1984
* Last modification: Wed Dec 12 20:09:10 1984
* Purpose: 
* UI each universally quantified formula to each
* variable in the prohibited list.  Allow this to happen a controlled
* number of times (determined by P_FINDUICNT)
* 
* Calling sequence: 
* Parameters: 
* 
* On return: 
*	RETURN	- ONESTEP succeeded
*	FRETURN - Otherewise
* 
* Local variables: 
*	FUC_X - Back link in circular linked list.
*	FUC_VARLISTA - Array form of G_PROHIBITLIST
*	FUC_L - Line field in prihibit list
*	FUC_TEMPL - Return from MAKEQTPLATE.
*	FUC_I - Index into FUC_VARLISTA.
*	FUC_A - ANTE corresponding to FUC_L
* 
      pDEFINE('FINDUIPROHIBCNT()FUC_X,FUC_VARLISTA,'
+	      'FUC_L,FUC_TEMPL,FUC_I,FUC_A',,,'NULL')
+						  :(ENDFINDUIPROHIBCNT)
FINDUIPROHIBCNT
+     FUC_VARLISTA = CONVERT(G_PROHIBLIST,'ARRAY')	 :F(FRETURN)
*     OUTPUT = 'PROHIBCNT called' 
      FUC_X = BLINK(G_UQSTACK)
FUC_LOOP
+     IDENT(FUC_X,G_UQSTACK)				 :S(FRETURN)
      FUC_L = LINE(FUC_X)
* check to see how many times that we have called finduiprohibcnt on
* this particular line
      FUC_A = FINDANTE(FUC_L)				:F(CONCHECK)
      IDENT(G_PRMAT<PRLINE(FUC_A),6>)			:S(FUC_UICONT)
      LT(G_PRMAT<PRLINE(FUC_A),6>,P_FINDUICNT)		:F(FUC_UICONT)
      G_PRMAT<PRLINE(FUC_A),6> = G_PRMAT<PRLINE(FUC_A),6> + 1
      FUC_TEMPL = MAKEQTPLATE(FUC_L)
      FUC_TEMPL '(A@)' =
      FUC_I = 0
FUC_UIVARLOOP
+     FUC_I = FUC_I + 1
      FUC_L = FUC_TEMPL
      FUC_NEWVAR = FUC_VARLISTA<FUC_I,1>	      :F(FUC_UICONT)
      IDENT(FUC_VARLISTA<FUC_I,2>)		   :S(FUC_UIVARLOOP)
FUC_UIVARREP
+     FUC_L '@' = FUC_NEWVAR			    :S(FUC_UIVARREP)
      FUC_L = NORMALIZE(FUC_L)
      FINDANTE(FUC_L)				   :S(FUC_UIVARLOOP)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(FUC_L,PLINE(FUC_X) ',UI')
+					  :S(RETURN)F(FUC_UIVARLOOP)
*  changes Aug 5, 1986	CNH  onestep incorporated into addproof
FUC_UICONT
+     FUC_X = BLINK(FUC_X)				 :(FUC_LOOP)
ENDFINDUIPROHIBCNT
-STITL Function: FINDUIPROHIBALL()
*FRAME
* Function: FINDUIPROHIBALL
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* UI each universally quantified formula to each
* variable in its OWN prohibited list
* 
* Calling sequence: 
*	return = XXXXXXXXXXXXXXXX 
* Parameters: 
* 
* On return: 
*	RETURN	- ONESTEP succeeded
*	FRETURN - Otherewise
* 
* Local variables: 
*	FUP_X - Back link in circular linked list.
*	FUP_VARLISTA - Array form of G_PROHIBITLIST
*	FUP_L - Line field in prihibit list
*	FUP_TEMPL - Return from MAKEQTPLATE.
*	FUP_I - Index into FUP_VARLISTA.
* 
      pDEFINE('FINDUIPROHIBALL()FUP_X,FUP_VARLISTA,'
+	      'FUP_L,FUP_TEMPL,FUP_I',,,'NULL')
+						  :(ENDFINDUIPROHIBALL)
FINDUIPROHIBALL
+     FUP_VARLISTA = CONVERT(G_PROHIBLIST,'ARRAY')	 :F(FRETURN)
      FUP_X = BLINK(G_UQSTACK)
FUP_LOOP
+     IDENT(FUP_X,G_UQSTACK)				 :S(FRETURN)
      FUP_L = LINE(FUP_X)
      FUP_TEMPL = MAKEQTPLATE(FUP_L)
      FUP_TEMPL '(A@)' =
      FUP_I = 0
FUP_UIVARLOOP
+     FUP_I = FUP_I + 1
      FUP_L = FUP_TEMPL
      FUP_NEWVAR = FUP_VARLISTA<FUP_I,1>	      :F(FUP_UICONT)
      IDENT(FUP_VARLISTA<FUP_I,2>)		   :S(FUP_UIVARLOOP)
FUP_UIVARREP
+     FUP_L '@' = FUP_NEWVAR			    :S(FUP_UIVARREP)
      FINDANTE(FUP_L)				   :S(FUP_UIVARLOOP)
*  CNH	3rd parameter to ADDPROOF removed
      ADDPROOF(FUP_L,PLINE(FUP_X) ',UI')
+					  :S(RETURN)F(FUP_UIVARLOOP)
*  changes Aug 5, 1986	CNH  onestep incorporated into addproof
FUP_UICONT
+     FUP_X = BLINK(FUP_X)				 :(FUP_LOOP)
ENDFINDUIPROHIBALL
-STITL CONCHECK : Error destination.
*
********************************************************************
* CONCHECK is a failure poit from anywhere's in he program.*
********************************************************************
*
						      :(ENDCONCHECK)
CONCHECK  G_WHERE = &LASTNO 
      PUTLINE(" CONCHECK was reached from statement # " G_WHERE) 
      PUTLINE(" *** Entering terminal DEBUG mode ***")
      DEBUG()						      :(END)
ENDCONCHECK
-STITL Function: GETFORM(GL_LINE,GL_TYPE)
*
      G_PROPFORM = ANY("pqrs") |
+	 G_UNOP *G_PROPFORM    |
+	 '(' *G_PROPFORM G_BINOP *G_PROPFORM ')'
*
*FRAME
* Function: GETFORM 
* Date: June 1983 
* Last modification: June 1983
* Purpose:
*	To check out a new line to see if it is legal. If
*	it is then return its type in the TYPE paramter. If
*	it is PROPOSITIONAL then add in the correct super-
*	and subscripts (0'0) to make it syntatically correct.
*
* 
* Calling sequence: 
*	return = GETFORM(GL_LINE,GL_TYPE)
* Parameters: 
*	GL_LINE = Line to be tested.
*	GL_TYPE = Pass be name paramter to return type.
* 
* On return: 
*	RETURN	- GL_LINE was a formula. The modified line is
*		  returned.
*	FRETURN - GL_LINE was not a formula.
* 
* Local variables: None. 
* 
       DEFINE('GETFORM(GL_LINE,GL_TYPE)')	       :(ENDGETFORM)
* for this to work &ANCHOR    MUST   be 0
GETFORM
+     GETFORM = GL_LINE
GF_BLANKDEL
+     GETFORM ' ' =				     :S(GF_BLANKDEL)
      GETFORM POS(0) G_BF RPOS(0)			:S(GF_PRED0)
      GETFORM POS(0) G_PROPFORM RPOS(0)       :S(GF_PROP0)F(FRETURN)
GF_PRED0
+     $GL_TYPE = "PREDICATE-0"				   :(RETURN)
GF_PROP0
+     $GL_TYPE = "PROPOSITIONAL-0"
* replace simple variables with 0-order predicates.
GF_R1
+     GETFORM 'p' = "P0'0()"				   :S(GF_R1)
GF_R2
+     GETFORM 'q' = "Q0'0()"				   :S(GF_R2)
GF_R3
+     GETFORM 'r' = "R0'0()"				   :S(GF_R3)
GF_R4
+     GETFORM 's' = "S0'0()"			  :S(GF_R4)F(RETURN)
ENDGETFORM
-STITL Function: PUTFORM(PT_LINE,PT_TYPE)
*FRAME
* Function: PUTFORM 
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*	To put a PROPOSITIONAL or PREDICATE formulat into standard
*	form.
* 
* Calling sequence: 
*	return = PUTFORM(PT_LINE,PT_TYPE)
* Parameters: 
*	PT_LINE - Line to be converted
*	PT_TYPE - the TYPE of that line.
* 
* On return: 
*	RETURN	- Normal termination. Altered line is returned.
*	FRETURN - PT_TYPE had an illegal value.
* 
* Local variables: None.
* 
       DEFINE('PUTFORM(PT_LINE,PT_TYPE)')	       :(ENDPUTFORM)
PUTFORM
      PT_TYPE 'PREDICATE-0'				:S(PT_PRED0)
      PT_TYPE 'PROPOSITIONAL-0' 	      :S(PT_PROP0)F(FRETURN)
PT_PRED0
      PUTFORM = PT_LINE 				   :(RETURN)
PT_PROP0
PT_0R1
      PT_LINE "P0'0()" = "p"				  :S(PT_0R1)
PT_0R2
      PT_LINE "Q0'0()" = "q"				  :S(PT_0R2)
PT_0R3
      PT_LINE "R0'0()" = "r"				  :S(PT_0R3)
PT_0R4
      PT_LINE "S0'0()" = "s"				  :S(PT_0R4)
      PUTFORM = PT_LINE 				   :(RETURN)
ENDPUTFORM
-STITL Function: DEBUG()
*FRAME
* Function: DEBUG
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*	This routine will enter debug mode. It is the routine
*	called if the attention trap was triped.
*	It is also a usefull debuging tool, just
*	add the line DEBUG() anywhere to get a look at variables...
*	The dance done with SETATN is so that there is no attention
*	set in Command mode. It is reset when exiting DEBUG. The
*	only other place it is set is in CMD_PROVE (only accessible
*	from MAIN COMMAND MODE - "|"). If runing in BATCH mode DEBUG
*	cannot be used (it simply FRETURNS imediately).
* Changes
*	CRM April 29,1987 changes to get it to work on UNIX
*		needs more work
*
* Calling sequence: 
*	DEBUG()
* Parameters: None.
* 
* On return: 
*	RETURN	- All went well (EOF or CONTINUE)
*	FRETURN - User ENTERED GIVEUP or STOP.
* 
* Local variables: 
*	DB_CMDRC - Return code from the command parser.
* 
*
*
       DEFINE('DEBUG()DB_CMDRC')			 :(ENDDEBUG)
DEBUG
* * * TERMINAL()				      :F(DB_NOTTERM)
      IDENT(G_OS,G_MTS) SETATN('OFF')
      DB_CMDRC = CMDPARSE(G_DEBUGCMDS,'display','@')
      IDENT(G_OS,G_MTS) SETATN('ON')
      IDENT(DB_CMDRC)					  :S(RETURN)
      LEQ(DB_CMDRC,'CONT')				  :S(RETURN)
      LEQ(DB_CMDRC,'STOP')				 :S(FRETURN)
      PUTLINE(' ** Illegal RC from CMDPARSE : "' DB_CMDRC '" **')
+							   :(RETURN)
DB_NOTTERM
+     IDENT(G_OS,G_MTS) CLRATN()			  :(FRETURN)
ENDDEBUG
-STITL Function: CMDPARSE(DP_ALLOWED,CP_DEFAULT,CP_PFX,[CP_FNAME])
*FRAME
* Function: CMDPARSE
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* CMDPARSE will parse user input. If the desired command is in
* the command table (1st parameter) then the appropriate function
* is called. If the user dosen't specify any command then the 
* default command is used (2nd parameter). The prefix is set to
* whatever the caller wants (3rd parameter) and is restored at
* the end. CMDPARSE returns null if the user enters an EOF line.
* If a function returns something other then null then CMDPARSE
* immediately terminates and returns that result. If the function
* returns null CMDPARSE continues to get user input. To apply a
* command CMDPARSE add the prefix "CMD_" and calls the resulting
* function (thus all command functions start with CMD_).
* 
* Calling sequence: 
*	return = CMDPARSE(CP_ALLOWED,CP_DEFAULT,CP_PFX,[CP_FNAME])
* Parameters: 
*	CP_ALLOWED - Table of allowable commands.
*	CP_DEFAULT - Default command if user omits one.
*	CP_PFX - Prefix character to be used.
*	CP_FNAME - Optional file name to read input from
* 
* On return: 
*	RETURN	- Normal termination. Result of command is returned.
* 
* Local variables: 
*	CP_IN - User input line.
*	CP_OLDPFX - Previous prefix.
*	CP_APPLYRC - Result from applying the command (Return code).
*	CP_PARAM - PARAMETERS field of current entry in command
*		   table.
*	CP_DATA - Line of DATA type CMD_FORMAT.
*
	DEFINE('CMDPARSE(CP_ALLOWED,CP_DEFAULT,CP_PFX,CP_FNAME)'
+	'CP_IN,CP_WORK,CP_OLDPFX,CP_APPLYRC,CP_PARAM,CP_DATA')
+							:(ENDCMDPARSE)
CMDPARSE
+	CP_OLDPFX = PREFIX(CP_PFX)
CP_LOOP
	INDEX = DIFFER(P_SILENT,'ON') G_PREFIX
	CP_IN = GETLINE(CP_FNAME)			:F(CP_RETURN)
	IDENT(CP_IN)					:S(CP_LOOP)
	LEQ(SUBSTR(CP_IN,1,1),G_ESCAPEPFX)		:S(CP_ESCAPE)
*
* ok, we have a command, first check to see if we're in DEBUG mode
* If so then see if we need to add the DEFAULT command by checking
* the display table. Otherwise the DEFAULT command is added if the
* line is a valid formula.
*
* * *	GETFORM(CP_IN,.G_GARBAGE)			:S(CP_DEFAULT)
	LEQ(CP_PFX,'@') DIFFER(G_VALIDDISPLAY<CP_IN>)	:F(CP_NODEFAULT)
*
* ok, we have a formula, now turn this into a command
*
CP_DEFAULT
	CP_IN = CP_DEFAULT ' ' CP_IN
CP_NODEFAULT
* now, pick off the command
	CP_IN POS(0) BREAK(" ") $ CP_CMD SPAN(" ") =	:S(CP_GOTCMD)
	CP_CMD = CP_IN
	CP_IN =
CP_GOTCMD
+	IDENT(CP_ALLOWED<CP_CMD>)			:S(CP_BADCMD)
	CP_DATA = CP_ALLOWED<CP_CMD>
	CP_CMD = COMMAND(CP_DATA)
	CP_PARAM = PARAMETERS(CP_DATA)
* ok so we do a case statement on the command name by applying
* the command name as a function. CP_IN is passed as an argument.
	CP_APPLYRC = APPLY("CMD_" CP_CMD,CP_IN,CP_PARAM)
+							:F(CP_LOOP)
	IDENT(CP_APPLYRC)				:S(CP_LOOP)
	CMDPARSE = CP_APPLYRC				:(CP_RETURN)
CP_BADCMD
*
* Check if it's in any of the other command tables. If so then
* CP_NOTALLOWED.
*
	DIFFER(G_MAINCMDS<CP_CMD>)			:S(CP_NOTALLOWED)
	DIFFER(G_DEBUGCMDS<CP_CMD>)			:S(CP_NOTALLOWED)
	ERROR(" Command not known")			:(CP_LOOP)
CP_NOTALLOWED
+	ERROR(" Command not allowed in this mode.",
+		O_ERROR)				:(CP_LOOP)
CP_ESCAPE
* This is an MTS or UNIX command, let's execute it.
	CP_IN G_ESCAPEPFX = null
	IDENT(G_OS,G_UNIX) HOST(1,'/bin/csh -c ' CP_IN) :S(CP_LOOP)
	IDENT(G_OS,G_MTS)  SCMD(CP_IN,SIZE(CP_IN))	:(CP_LOOP)
CP_RETURN
+	PREFIX(CP_OLDPFX)				:(RETURN)
ENDCMDPARSE
-STITL Function: CMD_PROVE(CP_IN)
*FRAME
* Function: CMD_PROVE
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*	This routine will call the theorem prover with the 
*	the function given in the input parameter. 
*	If the formula is invalid a message is printed.
*	If no formula is given CMD_PROVE will prompt you for one.
*	This is where SETATN is turned off and on.
*
* 
* Calling sequence: 
*	CMD_PROVE(CP_IN)
* Parameters: 
*	CP_IN - Formula to be passed to PROOF().
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: 
*	CP_PR_LINE - User input line (CP_IN was null).
*	CP_PR_PREF - Old prefix (when geting a formula).
* 
       DEFINE("CMD_PROVE(CP_IN)CP_PR_LINE,CP_PR_PREF")
+						     :(ENDCMD_PROVE)
CMD_PROVE
      DIFFER(CP_IN)				      :S(CP_PR_CONT)
      CP_PR_PREF = PREFIX('S')
      CP_IN = GETLINE() 			      :F(CP_PR_DONE)
      PREFIX(CP_PR_PREF)
CP_PR_CONT
      CP_PR_LINE = GETFORM(CP_IN,.P_FORMTYPE)	   :F(CP_PR_BADFORM)
      IDENT(G_OS,G_MTS) SETATN('ON')
      PUTLINE("-PROVING: "  CP_PR_LINE )
* If there is an identity set P_NOID to 1
      CP_PR_LINE G_IDOP 			      :F(CP_PNOTID)
      P_NOID = 1
CP_PNOTID
      LEQ(P_PROFILE,"ON") profile_on()		      :S(CP_PR_PROF)
      func_array = null
CP_PR_PROF
      INITER()
      CP_PR_LINE = NORMALIZE(CP_PR_LINE)
      ADDGOAL(CP_PR_LINE)
      DOPREM(,'ENTER')
      PROOF()					 :S(CP_PR_PROVEN)
      profile_off()
      DOPREM(,'INIT')
      PUTLINE(" THINKER cannot prove that")	     :(CP_PR_RETURN)
CP_PR_PROVEN
      G_PTIME = TIME() - G_STTIME
      DOPREM(,'INIT')
*      PUTLINE('1')
*  CHANGES Aug 13, 1986  CNH  add a parameter to printproof
      PRINTPROOF('P')
      profile_off()				     :(CP_PR_RETURN)
CP_PR_BADFORM
      PUTLINE(" Invalid formula")		     :(CP_PR_RETURN)
CP_PR_DONE
      PREFIX(CP_PR_PREF)
CP_PR_RETURN
      IDENT(G_OS,G_MTS) SETATN('OFF')		   :(RETURN)
ENDCMD_PROVE
-STITL Function: CMD_PREMISE(CP_IN)
*FRAME
* Function: CMD_PREMISE
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*	This function will call DOPREM to add a premise to the
*	premise table. If no formula is given then the user is
*	prompted for one. An invalid formula will print an error
*	message. If the premise is added from STUCK or DEBUG modes
*	and it's type is not identical to the TYPE of the formula
*	being proven another error message is printed. 
* 
* Calling sequence: 
*	CMD_PREMISE(CP_IN)
* Parameters: 
*	CP_IN - Formula to be added as a premise.
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: 
*	CP_PM_LINE - Return from GETFORM.
*	CP_PM_FLAG - Indicates if CP_IN was initialy null.
*	CP_PM_PREF - Old prefix (restored at end).
*	CP_PM_TYPE - Type given by GETFORM.
* 
       DEFINE("CMD_PREMISE(CP_IN)CP_PM_LINE,CP_PM_FLAG,CP_PM_PREF,"
+	      "CP_PM_TYPE")			   :(ENDCMD_PREMISE)
CMD_PREMISE
+     CP_PM_PREF = null
      CP_PM_FLAG = DIFFER(CP_IN) 1		      :S(CP_PM_CONT)
CP_PM_LOOP
+     IDENT(CP_PM_FLAG) 			      :F(CP_PM_DONE)
      CP_PM_PREF = IDENT(CP_PM_PREF) PREFIX("H")
      CP_IN = GETLINE() 			      :F(CP_PM_DONE)
      IDENT(CP_IN)				      :S(CP_PM_DONE)
CP_PM_CONT
+     CP_PM_LINE = GETFORM(CP_IN,.CP_PM_TYPE)	   :F(CP_PM_BADFORM)
      DIFFER(CP_PM_TYPE,P_FORMTYPE) DIFFER(CP_PFX,"|")
+						   :S(CP_PM_BADTYPE)
      PUTLINE( ' adding premise "' CP_PM_LINE '"')
* CNH Feb 21 1987   If there is an identity set P_NOID to 1
      CP_PM_LINE G_IDOP 			      :F(CP_NOTID)
      P_NOID = 1
CP_NOTID
*
* in normal command mode simply ADD else ADDENTER.
*
      CP_PFX ( '|' | '<' )			     :F(CP_PM_OTHER)
      DOPREM(CP_PM_LINE, "ADD")      :S(CP_PM_LOOP)F(CP_PM_OVERFLOW)
CP_PM_OTHER
      DOPREM(CP_PM_LINE,'ADDENTER')		       :(CP_PM_LOOP)
CP_PM_BADFORM
+     PUTLINE(" Invalid formula for PREMISE")	       :(CP_PM_LOOP)
CP_PM_OVERFLOW
+     PUTLINE(" Number of premises added exceeds current maximum of "
+	      P_MAXPREM ".")			       :(CP_PM_LOOP)
CP_PM_DONE
      PREFIX(DIFFER(CP_PM_PREF) CP_PM_PREF)		   :(RETURN)
ENDCMD_PREMISE
-STITL Function: CMD_DISPLAY(CP_IN,CP_TABLE)
*FRAME
* Function: XXXXXXX 
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* Allows for the display of multiple things.  The valid
* things to display are kept in the table G_VALIDDISPLAY
* The index in the table is the displayable item, the
* entry is of type DISPLAY_FORMAT.  If the FORMAT item
* of the entry is null then the VARIABLE item is assumed
* to be the name of a function to be called to do the
* display.  If the FORMAT entry is non-null, then
* the '&' symbol within the string is replaced with the
* value of the variable specifed by the VARIABLE field
* 
* Calling sequence: 
*	CMD_DISPLAY(CP_IN,CP_TABLE)
* Parameters: 
*	CP_IN - Key variable to be displayed. Must be in table. 
*	CP_TABLE - Table from which the variable may be displayed.
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: 
*	CP_DI_ITEM - Item to be displayed. 
*	CP_DI_NAME - Value of table entry (internal variable name).
*	CP_DI_FORMAT - Format field of display table entry.
* 
       DEFINE('CMD_DISPLAY(CP_IN,CP_TABLE)'
+	      'CP_DI_ITEM,CP_DI_NAME,CP_DI_FORMAT')
+						   :(ENDCMD_DISPLAY)
CMD_DISPLAY
+     IDENT(CP_IN)				   :S(CP_DI_NOITEMS)
      CP_IN = REPLACE(CP_IN,"ABCDEFGHIJKLMNOPQRSTUVWXYZ",
+			    "abcdefghijklmnopqrstuvwxyz")
      CP_IN = CP_IN ' '
CP_DI_LOOP
+     CP_IN BREAK(" ") $ CP_DI_ITEM SPAN(" ") =       :F(CP_DI_DONE)
      CP_DI_NAME = CP_TABLE<CP_DI_ITEM>
      IDENT(CP_DI_NAME) 			   :S(CP_DI_BADNAME)
      IDENT(FORMAT(CP_DI_NAME)) 		  :S(CP_DI_FUNCTION)
* ok, it is just a variable name
      P_FTRACE = &FTRACE; P_TRACE = &TRACE
      CP_DI_FORMAT = FORMAT(CP_DI_NAME)
      CP_DI_FORMAT '&' = $VARIABLE(CP_DI_NAME)
      PUTLINE(CP_DI_FORMAT)			       :(CP_DI_LOOP)
CP_DI_FUNCTION
+     APPLY(VARIABLE(CP_DI_NAME))		       :(CP_DI_LOOP)
CP_DI_BADNAME
+     PUTLINE(" " CP_DI_ITEM " is unknown and not displayable")
						       :(CP_DI_LOOP)
CP_DI_NOITEMS
+     PUTLINE(" DISPLAY requires arguments")		   :(RETURN)
CP_DI_DONE
+							   :(RETURN)
ENDCMD_DISPLAY
-SPACE 5
*
*FRAME
* This a valid display variable. If displayed the function
* "prfile_histogram" is called.
*
       DEFINE('CP_PRINT_HISTO()')		:(ENDCP_PRINT_HISTO)
CP_PRINT_HISTO
      profile_histogram(P_HISTSIZE)			   :(RETURN)
ENDCP_PRINT_HISTO
-SPACE 5
*
* FRAME
* PRINTPREM will dump the contents of G_PREM thus giving all present
* premicies. It is also called from CMD_DISPLAY as a valid display.
*
       DEFINE("PRINTPREM()PP_I,PP_FG")		     :(ENDPRINTPREM)
PRINTPREM
      PP_I = 0
      PP_FG = "EMPTY"
      PUTLINE("- Active premises are :")
PP_LOOP
      PP_I = PP_I + 1
      G_PREM<PP_I>					 :F(PP_DONE)
      PUTLINE("        - " DIFFER(G_PREM<PP_I>) G_PREM<PP_I>) :F(PP_LOOP)
      PP_FG = "NOTEMPTY"				  :(PP_LOOP)
PP_DONE
      LEQ(PP_FG,"EMPTY")				  :F(RETURN)
      PUTLINE(" 	No premises presently active")	  :(RETURN)
ENDPRINTPREM
-SPACE 5
*
*FRAME
* This is another DISPLAYable function. It is called if the user
* decides to display SHORTPROOF. It simply dumps the proof matrix
* without too much formating.
*
       DEFINE('NDUMPPROOF(DU_START)DU_C,DU_F,DU_INDSTR,DU_CURRENTIND,DU_CURRENTLEVEL,DU_SHOWSTR')
+						    :(ENDNDUMPPROOF)
NDUMPPROOF
      DU_C = DIFFER(DU_START) DU_START - 1
      DU_CURRENTLEVEL = G_PRMAT<DU_START,3>
      DU_INDSTR = '|' DUPL(' ',P_INDENT)
      DU_CURRENTIND = DUPL(DU_INDSTR,DU_CURRENTLEVEL)
DU_LOOP
      DU_C = DU_C + 1
* CHANGES Aug 27, 1986	CNH  change end condition
      GT(DU_C,G_CURLINE)			      :S(DU_DONE)
      IDENT(G_PRMAT<DU_C,1>)				 :S(DU_DONE)
      GT(DU_C,P_PROOFSIZE)				 :S(DU_DONE)
* see if we should decrease in levels
DU_LOOP2
      LT(G_PRMAT<DU_C,3>,DU_CURRENTLEVEL)		 :F(DU_EASY)
* mark the ends of the boxes
      PUTLINE(' 	' DU_CURRENTIND)
      DU_CURRENTLEVEL = DU_CURRENTLEVEL - 1
      DU_CURRENTIND = DUPL(DU_INDSTR,DU_CURRENTLEVEL)	 :(DU_LOOP2)
DU_EASY
* see if this is a show line
      LEQ(G_PRMAT<DU_C,2>,'*')			      :F(DU_NOTSHOW)
* this is a show line
      PUTLINE( LPAD(DU_C,6) '.	'
+	 RPAD(DU_CURRENTIND '*show '
+	 PUTFORM(G_PRMAT<DU_C,1>,P_FORMTYPE) '	',P_LINELEN))
      DU_CURRENTIND = DU_CURRENTIND DU_INDSTR
      DU_CURRENTLEVEL = DU_CURRENTLEVEL + 1		  :(DU_LOOP)
DU_NOTSHOW
      PUTLINE( LPAD(DU_C,6) '.	'
+	 RPAD(DU_CURRENTIND PUTFORM(G_PRMAT<DU_C,1>,P_FORMTYPE)
+	 '  ',P_LINELEN) G_PRMAT<DU_C,2> )		  :(DU_LOOP)
DU_DONE
      PUTLINE(EQ(DU_C,1) "	No proof has yet been calculated.")
+							   :(RETURN)
ENDNDUMPPROOF
-STITL Function: CMD_SET(CP_IN,CP_TABLE)
*FRAME
* Function: CMD_SET 
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* Set will allow you to manipulate all the Key Variables accessed
* by the program THINKER. It will set a variable to the specified
* integer is the appropriate table entries are specified. Otherwise
* it will try to execute special case code for that particular
* set via a :($...)
* 
* Calling sequence: 
*	CMD_SET(CP_IN,CP_TABLE)
* 
* Input parameters:
*	CP_IN - Variable(s) to be set with new value
*		seperated by "=".
*	CP_TABLE - Table containing all setable variables.
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables:
*	CP_SE_ITEM - Varible to be set (with value)
*	CP_SE_NAME - Name of that varible
*	CP_SE_VALUE - Value to be given to it.
* 
       DEFINE("CMD_SET(CP_IN,CP_TABLE)"
+	      "CP_SE_ITEM,CP_SE_NAME,CP_SE_VALUE")
+						       :(ENDCMD_SET)
CMD_SET
      IDENT(CP_IN)				   :S(CP_SE_NOITEMS)
      CP_IN = REPLACE(CP_IN,"ABCDEFGHIJKLMNOPQRSTUVWXYZ",
+			    "abcdefghijklmnopqrstuvwxyz") 
      CP_IN = CP_IN ' '
CP_SE_LOOP
      CP_IN BREAK(" ") $ CP_SE_ITEM SPAN(" ") = null  :F(CP_SE_DONE)
      CP_SE_ITEM BREAK("=") $ CP_SE_NAME "="
+		 REM $ CP_SE_VALUE
+						   :F(CP_SE_BADFORM)
* The valid left hand sides for SET are in G_VALIDSET
      CP_SE_ITEM = CP_TABLE<CP_SE_NAME>
      IDENT(CP_SE_ITEM) 			   :S(CP_SE_BADNAME)
*
* Now as the upper and lower bounds are null we must do a 
* special case to set the variable. Otherwise we simply call
* the routine GETINT to set the appropriate variable.
*
      IDENT(U_BOND(CP_SE_ITEM) L_BOND(CP_SE_ITEM), null)
+		     :S($("CP_SE_" NAME(CP_SE_ITEM)))F(CP_SE_INTSET)
CP_SE_PROFILE
      PUTLINE(" ** WARNING ** &TRACE is curently less then 100"
+		LT(&TRACE,100) LEQ(CP_SE_VALUE,'on'))
      P_PROFILE  = LEQ(CP_SE_VALUE,'on') 'ON'	      :S(CP_SE_LOOP)
      LEQ(P_PROFILE,'OFF')			      :S(CP_SE_LOOP)
      P_PROFILE  = LEQ(CP_SE_VALUE,'off') 'OFF' profile_off()
+						      :S(CP_SE_LOOP)
      PUTLINE(" Value required for PROFILING is ON or OFF")
+						       :(CP_SE_LOOP)
*
CP_SE_P_MAXPREM
+     GE(CP_SE_VALUE,5) 			   :F(CP_SE_PREMERR)
      LE(CP_SE_VALUE,500)			   :F(CP_SE_PREMERR)
      P_MAXPREM = CP_SE_VALUE

DOPREM(,'INIT')				       :(CP_SE_LOOP)
CP_SE_PREMERR
+     PUTLINE(" ** maxprem must be assigned an integer "
+		 "between 5 and 500 ** ")	       :(CP_SE_LOOP)
CP_SE_TESTPARS
      P_TESTPARS = LEQ(CP_SE_VALUE,'on') 'ON'	      :S(CP_SE_CALL)
      P_TESTPARS = LEQ(CP_SE_VALUE,'onreturn') 'ONRETURN'
+						      :S(CP_SE_CALL)
      P_TESTPARS = LEQ(CP_SE_VALUE,'off') 'OFF'       :S(CP_SE_CALL)
      PUTLINE(" TESTPARS must be set to ON, ONRETURN, or OFF")
+						       :(CP_SE_LOOP)
*
CP_SE_VARSIZE
      G_VARSIZE = GETINT(CP_SE_VALUE,1,G_MAXVARSIZE) + 9
+						  :F(CP_SE_BADVSIZE)
      G_VARSIZE RPOS(1) REM = '0'
      G_VARNAMES = SUBSTR(G_VARORDER,1,G_VARSIZE / 10) 
      INITPAT() 				       :(CP_SE_LOOP)
CP_SE_BADVSIZE
      PUTLINE(" VARNUM must be an positive integer less then "
+	       G_MAXVARSIZE ".")		       :(CP_SE_LOOP)
*
CP_SE_CALL
      PARCHECK(P_TESTPARS)		   :F(CONCHECK)S(CP_SE_LOOP)
* 
CP_SE_INTSET
* must be carefull with &FTRACE and P_FTRACE.
      P_FTRACE = &FTRACE
      P_TRACE = &TRACE
      $(NAME(CP_SE_ITEM)) = GETINT(CP_SE_VALUE,
+	     L_BOND(CP_SE_ITEM),U_BOND(CP_SE_ITEM))    :F(CP_BADSET)
      &FTRACE = P_FTRACE
      &TRACE = P_TRACE				       :(CP_SE_LOOP)
CP_BADSET
      PUTLINE(' **' CP_SE_NAME ' must be assigned an integer between '
+     L_BOND(CP_SE_ITEM) ' and '  U_BOND(CP_SE_ITEM) '**')
+						       :(CP_SE_LOOP)
********************************************************************
CP_SE_BADFORM
      PUTLINE(" " CP_SE_ITEM " does not contain an '='")
+						       :(CP_SE_LOOP)
CP_SE_BADNAME
      PUTLINE(" " CP_SE_NAME " is unknown and not SETable")
+						       :(CP_SE_LOOP)
CP_SE_NOITEMS
      PUTLINE(" SET requires arguements")		   :(RETURN)
CP_SE_DONE
							   :(RETURN)
ENDCMD_SET
-STITL Function: GETINT(GI_VALUE,GI_LWB,GI_UPB)
*FRAME
* Function: GETINT 
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* GETINT will convert the first input parameter to an integer
* and return that value if it is between the upper and lower bonds
* specified by the 2nd and 3rd arguments.
* 
* Calling sequence: 
*	return = GETINT(GI_VALUE,GI_LWB,GI_UPB)
* Parameters: 
*	GI_VALUE - Integer value to be set.
*	GI_LWB - Minimum value for that integer.
*	GI_UPB _ Maximum value for that integer.
* 
* On return: 
*	RETURN	- All went well, integer returned.
*	FRETURN - GI_VALUE not an integer or not withing bounds
* 
* Local variables: None.
* 
       DEFINE("GETINT(GI_VALUE,GI_LWB,GI_UPB)") 	:(ENDGETINT)
GETINT
      GETINT = CONVERT(GI_VALUE,'INTEGER')		 :F(FRETURN)
      GE(GETINT,GI_LWB) LE(GETINT,GI_UPB)		 :F(FRETURN)
							   :(RETURN)
ENDGETINT
-STITL Function: CMD_PRDUMP()
*FRAME
* Function: CMD_PRDUMP
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* write out the proof matrix in an easy machine readable form
* so that we can format it outside of thinker. Output goes to
* the file -TH.OUT
* 
* Calling sequence: 
*	CMD_PRDUMP()
* Parameters: None.
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: 
*	i - Index into proof matrix.
*	fdub - MTS file/device block for file -TH.OUT
*
       DEFINE('CMD_PRDUMP()i,fdub')			:(ENDPRDUMP)
CMD_PRDUMP
      IDENT(G_OS,G_UNIX)				:S(RETURN)
      IDENT(G_PRMAT)				       :S(prv_nomat)
      fdub = SGETFD('-th.out(*l+1) ')
*  CHANGES Aug 18, 1986  CNH  add # statements discarded
*  CHANGES Aug 19, 1986  CNH  if long then don't include L_SIZE
      discard = DIFFER(L_SIZE,G_CURLINE) + L_SIZE - size
 SWRITE(fdub,P_FORMTYPE ':' G_TOTALTIME ',' G_PTIME ',' G_TOTALCNT ',' discard)
      i = 1
pd_loop
*  CHANGES June 23, 1986   CNH
*  I changed the end check from DIFFER(G_PRMAT<i,1>)
*  To checking the size so it will work with CONDENSE
      GT(i,G_CURLINE)					      :S(pr_done)
      SWRITE(fdub,G_PRMAT<i,1> '!' G_PRMAT<i,2> '!' G_PRMAT<i,3>)
      i = i + 1 					  :(pd_loop)
pr_done
      SWRITE(fdub,"-end-of-proof")			   :(RETURN)
ENDPRDUMP
-STITL Function: CMD_VDISPLAY(CP_IN)
*FRAME
* Function: CMD_VDISPLAY
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* VDISPLAY uses SPIT.DISPLAY to show the contents of any variable
* in THINKER. It uses internal variable names.
* 
* Calling sequence: 
*	CMD_VDISPLAY(CP_IN)
* Parameters: 
*	CP_IN - Internal varible name to be displayed.
* 
* On return: 
*	RETURN	- normal termination.
* 
* Local variables: 
*	CP_VD_VAR - Variable to be displayed.
*
       DEFINE('CMD_VDISPLAY(CP_IN)CP_VD_VAR')	  :(ENDCMD_VDISPLAY)
CMD_VDISPLAY
      PUTLINE(" ** VDISPLAY must have a parameter ** "
+	      IDENT(CP_IN))				  :S(RETRUN)
      CP_IN = CP_IN ' '
CP_VD_LOOP
      CP_IN BREAK(" ") . CP_VD_VAR SPAN(" ") = null	  :F(RETURN)
      DISPLAY($CP_VD_VAR,CP_VD_VAR,"R",0)	       :(CP_VD_LOOP)
ENDCMD_VDISPLAY
-STITL Function: CMD_EVAL(CP_IN)
*FRAME
* Function: CMD_EVAL
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* CMD_EVAL will take the remainder of the line after the EVAL
* and try to evaluate it as a spitbol program line. A dance
* with SETEXIT is done to avoid possible execution errors in EVAL.
* 
* Calling sequence: 
*	CMD_EVAL(CP_IN)
* Parameters: 
*	CP_IN - String to be evaluated.
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: 
*	CE_RESULT - Result from evaluation.
*	CE_OLDEXIT - Previous SETEXIT command.
*
       DEFINE("CMD_EVAL(CP_IN)CE_RESULT,CE_OLDEXIT")  :(ENDCMD_EVAL)
CMD_EVAL
      CE_OLDEXIT = SETEXIT(.CE_PGNTRP)
      G_ERRLIMIT = &ERRLIMIT
      &ERRLIMIT = 1
      CE_RESULT = EVAL(CP_IN)				 :F(CE_FAIL)
      PUTLINE(' EVAL returned : "' CE_RESULT '".')	:(CE_RETURN)
CE_FAIL 
      PUTLINE(' EVAL has FAILED. ')			:(CE_RETURN)
CE_RETURN
      &ERRLIMIT = G_ERRLIMIT
      SETEXIT(CE_OLDEXIT)				   :(RETURN)
CE_PGNTRP
      G_WHERE = &LASTNO
      PUTLINE(' *** ERROR in statement #' G_WHERE ' with code ' &ERRTYPE '.')
							:(CE_RETURN)
ENDCMD_EVAL
-STITL Function: CMD_MTS()
*FRAME
* Function: CMD_MTS 
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* CMD_MTS will simply call the MTS subroutine to allow the user
* to execute MTS commands. He can issue a $RESTART to get back
* to THINKER.
* 
* Calling sequence: 
*	CMD_MTS()
* Parameters: None.
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: None.
*
       DEFINE("CMD_MTS()")			       :(ENDCMD_MTS)
CMD_MTS
      PUTLINE("- ** You may issue a $RESTART to continue **")
      MTS()						   :(RETURN)
ENDCMD_MTS
-STITL Function: CMD_UNIX()
*FRAME
* Function: CMD_UNIX
* Date: July 1983 
* Last modification: July 1983
* Purpose: 
*	Will cause an immediate jump to UNIX command mode.
* 
* Calling sequence: 
*	CMD_UNIX()
* Parameters: None.
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: None.
*
	DEFINE('CMD_UNIX()')				:(ENDCMD_UNIX)
CMD_UNIX
	PUTLINE(' ')
	PUTLINE(' *** Entering UNIX. Press ^D (EOF) to return ***')
	HOST(1,'csh')
	PUTLINE(' ')					:(RETURN)
ENDCMD_UNIX
-STITL Function: CMD_ADDLINE(CP_IN)
*FRAME
* Function: CMD_ADDLINE
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*  CMD_ADDLINE will add a line to the proof and tell THINKER to
*  CONTinue to try to solve this problem. It first checks to see 
*  that the formula is legal, then it calls addline and finaly
*  returns the value "CONT" if all went well.
* 
* Calling sequence: 
*	return = CMD_ADDLINE(CP_IN)
* Parameters: 
*	CP_IN - Line to be added. Must be legal formula of 
*		correct type.
* 
* On return: 
*	RETURN	- Line added and CONT returned.
*	FRETURN - Line not added. Invalid formula or type.
* 
* Local variables: 
*	AL_TYPE - Type of the line to be added.
*
       DEFINE("CMD_ADDLINE(AL_FORMULA)AL_TYPE")    :(ENDCMD_ADDLINE)
CMD_ADDLINE
      AL_FORMULA = GETFORM(AL_FORMULA,.AL_TYPE)       :F(AL_BADFORM)
      DIFFER(AL_TYPE,P_FORMTYPE)		      :S(AL_BADTYPE)
*  changes Aug 5, 1986	CNH
*  if the line added closes the show go to cmd_adone
      ADDPROOF(AL_FORMULA,"GOD says so")	 :S(CMD_ADONE)
      CMD_ADDLINE = "CONT"				   :(RETURN)
*  changes Aug 5, 1986	CNH
*  add the routine to set the code to DONE
CMD_ADONE
+     CMD_ADDLINE = "DONE"			 :(RETURN)
AL_BADFORM
      PUTLINE(" >> NOT a proper formulat for ADDLINE <<") :(FRETURN)
AL_BADTYPE
      PUTLINE(" >> Formula is not of the correct type <<")
      PUTLINE("    ADDLINE expected a " P_FORMTYPE " formula") 
							  :(FRETURN)
ENDCMD_ADDLINE
-STITL Function: CMD_ADDGOAL(CP_IN)
*FRAME
* Function: CMD_ADDGOAL
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* CMD_ADDGOAL will add a new goal to the proof and ask THINKER
* to try to prove this goal. It first checks to see 
* that the formula is legal, then it calls ADDGOAL and finaly
* returns the value "PROVE" if all went well.
* 
* Calling sequence: 
*	return = CMD_ADDGOAL(CP_IN)
* Parameters: 
* 
* On return: 
*	RETURN	- GOAL added. Return the value "PROVE"
*	FRETURN - GOAL not added. Invalid formula or type.
* 
* Local variables: 
*	AG_TYPE - Type of the line being added.
*
       DEFINE("CMD_ADDGOAL(AG_FORMULA)AG_TYPE")    :(ENDCMD_ADDGOAL)
CMD_ADDGOAL
      AG_FORMULA = GETFORM(AG_FORMULA,.AG_TYPE)       :F(AG_BADFORM)
      DIFFER(AG_TYPE,P_FORMTYPE)		      :S(AG_BADTYPE)
      ADDGOAL(AG_FORMULA)
      CMD_ADDGOAL = "PROVE"				   :(RETURN)
AG_BADFORM
      PUTLINE(" >> NOT a proper formulat for ADDGOAL <<") :(FRETURN)
AG_BADTYPE
      PUTLINE(" >> Formula is not of the correct type <<")
      PUTLINE("    ADDGOAL expected a " P_FORMTYPE " formula") 
							  :(FRETURN)
ENDCMD_ADDGOAL
-STITL Function: CMD_DEBUG()
*FRAME
* Function: CMD_DEBUG
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*	Will cause an immediate jump to DEBUG mode.
* 
* Calling sequence: 
*	return = CMD_DEBUG()
* Parameters: None.
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: 
*	CDB_RC - Return code from command parser.
*
       DEFINE("CMD_DEBUG()CDB_RC")		     :(ENDCMD_DEBUG)
CMD_DEBUG
      CDB_RC = CMDPARSE(G_DEBUGCMDS,'display','@')
      CMD_DEBUG = LEQ(CDB_RC,'STOP') 'STOP'		   :(RETURN)
ENDCMD_DEBUG
-STITL Function: CMD_DUMP()
*FRAME
* Function: CMD_DUMP
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*	Will cause a dump of all natural variables.
* 
* Calling sequence: 
*	CMD_DUMP()
* Parameters: None.
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: None.
*
       DEFINE("CMD_DUMP()")			      :(ENDCMD_DUMP)
CMD_DUMP
      DUMP(1)						   :(RETURN)
ENDCMD_DUMP
-STITL Function: CMD_COMMANDS(G_GARBAGE,CP_CMDS)
*FRAME
* Function: CMD_COMMANDS
* Date: June 1983 
* Last modification: June 1983
* Purpose:
*	Will dump the command table in an orderly fashion.
* 
* Calling sequence: 
*	CMD_COMMANDS(G_GARBAGE,CP_CMDS)
* Parameters: 
*	G_GARBAGE - Ignored.
*	CP_CMDS - The table of commands to be dumped.
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: 
*	CM_ARRAY - Array form of command table.
*	CM_I - Index into command table.
* 
*
       DEFINE("CMD_COMMANDS(G_GARBAGE,CM_CMDS)CM_ARRAY,CM_I")
+						  :(ENDCMD_COMMANDS)
CMD_COMMANDS
      CM_ARRAY = CONVERT(CM_CMDS,"ARRAY")
      CM_I = 1
      PUTLINE("   Legal COMMANDS are :")
      PUTLINE(" ")
      PUTLINE("       - " CM_ARRAY<1,1>)
CM_LOOP
+     CM_I = CM_I + 1
      CM_ARRAY<CM_I,1>					  :F(RETURN)
      PUTLINE("       - " DIFFER(COMMAND(CM_ARRAY<CM_I,2>),
+				 COMMAND(CM_ARRAY<CM_I - 1,2>))
+			  CM_ARRAY<CM_I,1>)
+							  :(CM_LOOP)
ENDCMD_COMMANDS
-STITL Function: CMD_DISPLAYS(G_GARBAGE,CD_DSPS)
*FRAME
* Function: CMD_DISPLAYS
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*	Will dump out all the valid display variables.
* 
* Calling sequence: 
*	CMD_DISPLAY(G_GARBAGE,CD_DSPS)
* Parameters: 
*	G_GARBAGE - Ignored.
*	CD_DSPS - Table of legal variables to display.
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: 
*	CD_ARRAY - Array form of CD_DSPS
*	CD_I - Index variable into CD_ARRAY.
* 
       DEFINE("CMD_DISPLAYS(G_GARBAGE,CD_DSPS)CD_ARRAY,CD_I")
+						  :(ENDCMD_DISPLAYS)
CMD_DISPLAYS 
+     CD_ARRAY = CONVERT(CD_DSPS,"ARRAY")
      CD_I = 0
      PUTLINE("   Variables which may be DISPLAYed are :")
      PUTLINE(" ")
CD_LOOP
+     CD_I = CD_I + 1
      CD_ARRAY<CD_I,1>					  :F(RETURN)
      PUTLINE("       - " CD_ARRAY<CD_I,1>)		  :(CD_LOOP)
ENDCMD_DISPLAYS 
-STITL Function: CMD_SETS(G_GARBAGE,CS_SETS)
*FRAME
* Function: CMD_SETS
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*	To see all the variables which may be set and what the upper
*	and lower bounds of these are.
* 
* Calling sequence: 
*	CMD_SETS(G_GARBAGE,CS_SETS)
* Parameters: 
*	G_GARBAGE - Ignored
*	CS_SETS - Table of variables which may be set.
*
* 
* On return: 
*	RETURN	- Normal termination.
* 
* Local variables: 
*	CS_ARRAY - Array form of CS_SETS.
*	CS_I - Index into CS_SETS.
* 
       DEFINE("CMD_SETS(G_GARBAGE,CS_SETS)CS_ARRAY,CS_I")
+						      :(ENDCMD_SETS)
CMD_SETS
+     CS_ARRAY = CONVERT(CS_SETS,"ARRAY")
      CS_I = 0
      PUTLINE("   Variables which may be SET are :")
      PUTLINE(" ")
CS_LOOP
+     CS_I = CS_I + 1
      CS_ARRAY<CS_I,1>					  :F(RETURN)
      IDENT(L_BOND(CS_ARRAY<CS_I,2>),null)	       :S(CS_NOBOND)
      PUTLINE("       - " RPAD(CS_ARRAY<CS_I,1>,15) 
+	 RPAD(' Lower bound=' L_BOND(CS_ARRAY<CS_I,2>) ',',20)
+	      ' Upper bound=' U_BOND(CS_ARRAY<CS_I,2>))   :(CS_LOOP)
CS_NOBOND
+     PUTLINE("       - " CS_ARRAY<CS_I,1>)		  :(CS_LOOP)
ENDCMD_SETS 
-STITL Function: CMD_STOP()
*FRAME
* Function: STOP
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*	This function stops the current level of command parsing
*	by returning a "STOP"
* 
* Calling sequence: 
*	result = CMD_STOP()
* Parameters: None.
* 
* On return: 
*	RETURN	- Normal termination. Returns "STOP"
* 
* Local variables: 
*
       DEFINE("CMD_STOP()")				  :(ENDSTOP)
CMD_STOP CMD_STOP = 'STOP'				   :(RETURN)
ENDSTOP
-STITL Function: CMD_CONTINUE()
*FRAME
* Function: CMD_CONTINUE
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*	This routine exits the current level of command parsing
*	and tells THINKER to continue processing by returning "CONT"
* 
* Calling sequence: 
*	return = CMD_CONTINUE()
* Parameters: None.
* 
* On return: 
*	RETURN	- Normal termination. Returns "CONT"
* 
* Local variables: None.
*
       DEFINE("CMD_CONTINUE()") 		  :(ENDCMD_CONTINUE)
CMD_CONTINUE
      CMD_CONTINUE = "CONT"				   :(RETURN)
ENDCMD_CONTINUE
-STITL COMMAND tables.
*
*FRAME
* The command parser is table driven. These are the tables used
* by the parser.
*
*FRAME
* This is the command table for all variables which may be legally
* displayed.
*
      G_VALIDDISPLAY = TABLE(20)
      G_VALIDDISPLAY<"proof">	  = DISPLAY_FORMAT("PRINTPROOF")
      G_VALIDDISPLAY<"shortproof"> = DISPLAY_FORMAT("NDUMPPROOF")
      G_VALIDDISPLAY<"premises">  = DISPLAY_FORMAT("PRINTPREM")
      G_VALIDDISPLAY<"profiling"> = DISPLAY_FORMAT("P_PROFILE",
+	 " Profiling is currently &.")
      G_VALIDDISPLAY<"condense"> = DISPLAY_FORMAT("P_CONDENSE",
+	 " Condensing is currently set to &.")
      G_VALIDDISPLAY<"testpars">  = DISPLAY_FORMAT("P_TESTPARS",
+	 " The SPITBOL parameter checker is currently set to &.")
      G_VALIDDISPLAY<"profile">   = DISPLAY_FORMAT("print_profile")
      G_VALIDDISPLAY<"histogram"> = DISPLAY_FORMAT("CP_PRINT_HISTO")
      G_VALIDDISPLAY<"varnum">	  = DISPLAY_FORMAT("G_VARSIZE",
+	 " There are presently & variables available to THINKER.")
      G_VALIDDISPLAY<"linelen">   = DISPLAY_FORMAT("P_LINELEN",
+	 " Proof line length is & characters.")
      G_VALIDDISPLAY<"indent">	  = DISPLAY_FORMAT("P_INDENT",
+	 " Proof indenting set at & spaces.")
      G_VALIDDISPLAY<"checkprint"> = DISPLAY_FORMAT("P_CHECKPRINT",
+	  " Check print is set to &.")
      G_VALIDDISPLAY<"histsize">  = DISPLAY_FORMAT("P_HISTSIZE",
+	 " Histogram size is &.")
      G_VALIDDISPLAY<"proofsize"> = DISPLAY_FORMAT("P_PROOFSIZE",
+	  " Maximum proof size is currently set to & lines.")
      G_VALIDDISPLAY<"maxprem">   = DISPLAY_FORMAT("P_MAXPREM",
+	  " No more then & premises are allowable for one proof.")
      G_VALIDDISPLAY<"antes">	  = DISPLAY_FORMAT("P_ANTES",
+	  " Size of antecedent line table is &.")
      G_VALIDDISPLAY<"prohibitdepth"> = DISPLAY_FORMAT("P_FINDUICNT",
+	  " Prohibitdepth is set at &.")
      G_VALIDDISPLAY<"ftrace">	  = DISPLAY_FORMAT("P_FTRACE",
+	  " FTRACE has a current value of &.")
      G_VALIDDISPLAY<"trace">	  = DISPLAY_FORMAT("P_TRACE",
+	  " TRACE is currently set ot &.")
      G_VALIDDISPLAY<"variables"> = DISPLAY_FORMAT("G_VARNAMES",
+	  " Valid variable prefixes are '&'.")
*
* This is the command table for the variables which may be set.
*
      G_VALIDSET = TABLE(20)
      G_VALIDSET<"profiling">  = SET_FORMAT("PROFILE",null,null)
      G_VALIDSET<"testpars">   = SET_FORMAT("TESTPARS",null,null)
      G_VALIDSET<"varnum">     = SET_FORMAT("VARSIZE",null,null)
      G_VALIDSET<"trace">      = SET_FORMAT("P_TRACE",0,999999999)
      G_VALIDSET<"ftrace">     = SET_FORMAT("P_FTRACE",0,999999999)
      G_VALIDSET<"histsize">   = SET_FORMAT("P_HISTSIZE",1,50)
      G_VALIDSET<"indent">     = SET_FORMAT("P_INDENT",0,3)
      G_VALIDSET<"linelen">    = SET_FORMAT("P_LINELEN",20,132)
      G_VALIDSET<"checkprint"> = SET_FORMAT("P_CHECKPRINT",0,100)
      G_VALIDSET<"condense"> = SET_FORMAT("P_CONDENSE",0,1)
      G_VALIDSET<"proofsize">  = SET_FORMAT("P_PROOFSIZE",100,1000000)
      G_VALIDSET<"antes">      = SET_FORMAT("P_ANTES",100,1000000)
      G_VALIDSET<"maxprem">    = SET_FORMAT("P_MAXPREM",null,null)
      G_VALIDSET<"prohibitdepth"> = SET_FORMAT("P_FINDUICNT",0,20)
*  CHANGES Aug 13, 1986  add the set for formatting long
      G_VALIDSET<"formatlong">	= SET_FORMAT("P_FORMAT",0,1)
* CNT ADD P_PARALLEL
      G_VALIDSET<"parallel"> = SET_FORMAT("P_PARALLEL",0,1)
*
* Set commands allowed in DEBUG and STUCK modes.
*
      G_VALSET2 = TABLE(10)
      G_VALSET2<"testpars">    = SET_FORMAT("TESTPARS",null,null)
      G_VALSET2<"ftrace">      = SET_FORMAT("P_FTRACE",0,999999999)
      G_VALSET2<"trace">       = SET_FORMAT("P_TRACE",0,999999999)
      G_VALSET2<"indent">      = SET_FORMAT("P_INDENT",0,3)
      G_VALSET2<"linelen">     = SET_FORMAT("P_LINELEN",20,132)
      G_VALSET2<"checkprint">  = SET_FORMAT("P_CHECKPRINT",0,100)
      G_VALSET2<"condense">  = SET_FORMAT("P_CONDENSE",0,1)
*
* This is the table used by the command parser for all legal command
* accessible from the main command mode.
*
    G_MAINCMDS = TABLE(20)
    G_MAINCMDS<"prove">      = CMD_FORMAT("PROVE") 
    G_MAINCMDS<"premise">    = CMD_FORMAT("PREMISE") 
    G_MAINCMDS<"prem">	     = CMD_FORMAT("PREMISE") 
    G_MAINCMDS<"display">    = CMD_FORMAT("DISPLAY",G_VALIDDISPLAY) 
    G_MAINCMDS<"d">	     = CMD_FORMAT("DISPLAY",G_VALIDDISPLAY) 
    G_MAINCMDS<"dis">	     = CMD_FORMAT("DISPLAY",G_VALIDDISPLAY) 
    G_MAINCMDS<"vdisplay">   = CMD_FORMAT("VDISPLAY") 
    G_MAINCMDS<"v">	     = CMD_FORMAT("VDISPLAY") 
    G_MAINCMDS<"eval">	     = CMD_FORMAT("EVAL") 
    G_MAINCMDS<"set">	     = CMD_FORMAT("SET",G_VALIDSET) 
    G_MAINCMDS<"s">	     = CMD_FORMAT("SET",G_VALIDSET) 
    G_MAINCMDS<"format">     = CMD_FORMAT("PRDUMP") 
    G_MAINCMDS<"?commands">  = CMD_FORMAT("COMMANDS",G_MAINCMDS) 
    G_MAINCMDS<"?c">	     = CMD_FORMAT("COMMANDS",G_MAINCMDS) 
    G_MAINCMDS<"?c">	     = CMD_FORMAT("COMMANDS",G_MAINCMDS) 
    G_MAINCMDS<"help">	     = CMD_FORMAT("COMMANDS",G_MAINCMDS) 
    G_MAINCMDS<"?display">   = CMD_FORMAT("DISPLAYS",G_VALIDDISPLAY)
    G_MAINCMDS<"?d">	     = CMD_FORMAT("DISPLAYS",G_VALIDDISPLAY)
    G_MAINCMDS<"?set">	     = CMD_FORMAT("SETS",G_VALIDSET) 
    G_MAINCMDS<"?s">	     = CMD_FORMAT("SETS",G_VALIDSET) 
    G_MAINCMDS<"debug">      = CMD_FORMAT("DEBUG",G_DEBUGCMDS) 
    G_MAINCMDS<"mts">	     = CMD_FORMAT("MTS") 
    G_MAINCMDS<"stop">	     = CMD_FORMAT("STOP") 
    G_MAINCMDS<"end">	     = CMD_FORMAT("STOP") 
    G_MAINCMDS<"quit">	     = CMD_FORMAT("STOP") 
    G_MAINCMDS<"bye">	     = CMD_FORMAT("STOP") 
*
* This is the command table for leagal commands which THINKER
* can use when it is stuck.
*
    G_STUCKCMDS = TABLE(20) 
    G_STUCKCMDS<"addline">   = CMD_FORMAT("ADDLINE") 
    G_STUCKCMDS<"al">	     = CMD_FORMAT("ADDLINE") 
    G_STUCKCMDS<"addgoal">   = CMD_FORMAT("ADDGOAL") 
    G_STUCKCMDS<"ag">	     = CMD_FORMAT("ADDGOAL") 
    G_STUCKCMDS<"show">      = CMD_FORMAT("ADDGOAL") 
    G_STUCKCMDS<"display">   = CMD_FORMAT("DISPLAY",G_VALIDDISPLAY)
    G_STUCKCMDS<"dis">	     = CMD_FORMAT("DISPLAY",G_VALIDDISPLAY)
    G_STUCKCMDS<"d">	     = CMD_FORMAT("DISPLAY",G_VALIDDISPLAY)
    G_STUCKCMDS<"set">	     = CMD_FORMAT("SET",G_VALSET2) 
    G_STUCKCMDS<"s">	     = CMD_FORMAT("SET",G_VALSET2) 
    G_STUCKCMDS<"vdisplay">  = CMD_FORMAT("VDISPLAY") 
    G_STUCKCMDS<"v">	     = CMD_FORMAT("VDISPLAY") 
    G_STUCKCMDS<"eval">      = CMD_FORMAT("EVAL") 
    G_STUCKCMDS<"premise">   = CMD_FORMAT("PREMISE") 
    G_STUCKCMDS<"prem">      = CMD_FORMAT("PREMISE") 
    G_STUCKCMDS<"?commands"> = CMD_FORMAT("COMMANDS",G_STUCKCMDS) 
    G_STUCKCMDS<"?c">	     = CMD_FORMAT("COMMANDS",G_STUCKCMDS) 
    G_STUCKCMDS<"help">      = CMD_FORMAT("COMMANDS",G_STUCKCMDS) 
    G_STUCKCMDS<"?display">  = CMD_FORMAT("DISPLAYS",G_VALIDDISPLAY)
    G_STUCKCMDS<"?d">	     = CMD_FORMAT("DISPLAYS",G_VALIDDISPLAY)
    G_STUCKCMDS<"?set">      = CMD_FORMAT("SETS",G_VALSET2) 
    G_STUCKCMDS<"?s">	     = CMD_FORMAT("SETS",G_VALSET2) 
    G_STUCKCMDS<"debug">     = CMD_FORMAT("DEBUG",G_DEBUGCMDS) 
    G_STUCKCMDS<"continue">  = CMD_FORMAT("CONTINUE") 
    G_STUCKCMDS<"c">	     = CMD_FORMAT("CONTINUE") 
    G_STUCKCMDS<"giveup">    = CMD_FORMAT("STOP") 
    G_STUCKCMDS<"stop">      = CMD_FORMAT("STOP") 
    G_STUCKCMDS<"quit">      = CMD_FORMAT("STOP") 
    G_STUCKCMDS<"end">	     = CMD_FORMAT("STOP") 
*
* This is the command table to be used in DEBUG mode.
*
    G_DEBUGCMDS = TABLE(10) 
    G_DEBUGCMDS<"display">   = CMD_FORMAT("DISPLAY",G_VALIDDISPLAY)
    G_DEBUGCMDS<"dis">	     = CMD_FORMAT("DISPLAY",G_VALIDDISPLAY)
    G_DEBUGCMDS<"d">	     = CMD_FORMAT("DISPLAY",G_VALIDDISPLAY)
    G_DEBUGCMDS<"set">	     = CMD_FORMAT("SET",G_VALSET2) 
    G_DEBUGCMDS<"s">	     = CMD_FORMAT("SET",G_VALSET2) 
    G_DEBUGCMDS<"vdisplay">  = CMD_FORMAT("VDISPLAY") 
    G_DEBUGCMDS<"v">	     = CMD_FORMAT("VDISPLAY") 
    G_DEBUGCMDS<"dump">      = CMD_FORMAT("DUMP") 
    G_DEBUGCMDS<"eval">      = CMD_FORMAT("EVAL") 
    G_DEBUGCMDS<"premise">   = CMD_FORMAT("PREMISE") 
    G_DEBUGCMDS<"prem">      = CMD_FORMAT("PREMISE") 
    G_DEBUGCMDS<"?commands"> = CMD_FORMAT("COMMANDS",G_DEBUGCMDS) 
    G_DEBUGCMDS<"?c">	     = CMD_FORMAT("COMMANDS",G_DEBUGCMDS) 
    G_DEBUGCMDS<"help">      = CMD_FORMAT("COMMANDS",G_DEBUGCMDS) 
    G_DEBUGCMDS<"?display">  = CMD_FORMAT("DISPLAYS",G_VALIDDISPLAY)
    G_DEBUGCMDS<"?d">	     = CMD_FORMAT("DISPLAYS",G_VALIDDISPLAY)
    G_DEBUGCMDS<"?set">      = CMD_FORMAT("SETS",G_VALSET2) 
    G_DEBUGCMDS<"?s">	     = CMD_FORMAT("SETS",G_VALSET2) 
    G_DEBUGCMDS<"continue">  = CMD_FORMAT("CONTINUE") 
    G_DEBUGCMDS<"c">	     = CMD_FORMAT("CONTINUE") 
    G_DEBUGCMDS<"giveup">    = CMD_FORMAT("STOP") 
    G_DEBUGCMDS<"stop">      = CMD_FORMAT("STOP") 
    G_DEBUGCMDS<"end">	     = CMD_FORMAT("STOP") 
    G_DEBUGCMDS<"quit">      = CMD_FORMAT("STOP") 
-TITLE		GPSG -- System Initialization
-STITLE 	System I/O Initialization
* * *
*   The I/O initialization code must be executed before the
*   other I/O routines are defined.  This is so that the
*   routines are defined to execute the proper OS dependent
*   code to do the I/O.  The initialization should only be done
*   once for a given program run, and on UNIX, can be done only
*   in the compile by using the EXIT(3) command after everything
*   is defined.
* * *
	G_MTS		= 'MTS'
	G_UNIX		= 'UNIX'
*
	G_READ		= 'READ'
	G_WRITE 	= 'WRITE'
	G_APPEND	= 'APPEND'
	G_TEMP		= 'TMP'
	G_MTSTEMP	= 'MTMP'
	G_UNIXTMP	= 'UTMP'
*
	G_IN		= 'STDIN'
	G_OUT		= 'STDOUT'
	G_ERR		= 'STDERR'
*
	G_PREFIX	= '|'
	G_CMDPFX	= '|'
	G_DEBUGFPX	= '@'
	G_MTSESCAPEPFX	= '$'
	G_UNIXESCAPEPGX = '!' 
*   First, we need to find out if we are running on MTS or UNIX.
*   To do this, we look at &ALPHABET to see if the string 'IJ'
*   is there.  If so, we are on an ASCII machine (namely a VAX
*   running UNIX).  If not, then we are on an Amdahl 580 running MTS.
* * *
	
	&ALPHABET 'IJ'					:S(IO_INIT_UNIX)
* * *
*   Say hello to MTS
* * *
	G_OS = G_MTS
* * *	LOAD('SREAD(STRING,INTEGER,INTEGER)STRING','SNO.:UNSP.SPITLIB')
* * *	LOAD('SWRITE(STRING,STRING,INTEGER,INTEGER)INTEGER','SNO.:UNSP.SPITLIB')
* * *	LOAD('SCMD(STRING)','SNO.:UNSP.SPITLIB')
* * *	LOAD('SMTSCMD(STRING)','SNO.:UNSP.SPITLIB')
* * *	LOAD('SGETFD(STRING)STRING','SNO.:UNSP.SPITLIB')
* * *	LOAD('SCREPLY()','SNO.:UNSP.SPITLIB')
	&FULLSCAN = 1
	&ANCHOR = 0
	LOAD('SGETFD(STRING)STRING')
	LOAD('SFREEFD(STRING)')
	LOAD('SREAD(STRING,INTEGER,INTEGER)STRING')
	LOAD('SWRITE(STRING,STRING,INTEGER,INTEGER)STRING')
	LOAD('SCMDNOE(STRING)STRING','GAME:SPITLIB ')
	LOAD('TOFDAY()STRING','GAME:SPITLIB ')
	LOAD('DATE()STRING','GAME:SPITLIB ')
	LOAD('SLOCK(STRING,INTEGER,INTEGER)INTEGER','GAME:SPITLIB ')
	LOAD('SUNLK(STRING)INTEGER','GAME:SPITLIB ')
	LOAD('SEMPTY(STRING)INTEGER','GAME:SPITLIB ')
	LOAD('SCREATE(STRING,INTEGER,INTEGER,INTEGER,'
+		'INTEGER)INTEGER','GAME:SPITLIB ')
	LOAD('SETUNIT(STRING,STRING)STRING','GAME:SPITLIB ')
	LOAD('SCMD(STRING)STRING')
	LOAD('TIME(INTEGER,INTEGER,INTEGER)DREAL')
	LOAD('MTS()INTEGER','SPLS:SPITLIB')
	LOAD('PREFIX(STRING)STRING','GAME:SPITLIB')
	LOAD('TERMINAL()','GAME:SPITLIB')
	LOAD('SETATN(STRING)','TINK:TH.A.SUBS')
	LOAD('CHKATN()','*DUMMY* ')
	LOAD('CLRATN()','*DUMMY* ')
	IO_MODE = TABLE(20)
	IO_FDUBS = TABLE(20)				:(ENDIOINIT)
IO_INIT_UNIX
* * *
*   UNIX file I/O must be handled somewhat differently...
* * *
*   IO_CHANNEL is an information table that can be indexed in 2 ways.
*   Indexing by the file name gives the I/O Channel number.
*   Indexing by the I/O channel number gives the associated file name.
*
*   Indexing IO_MODE by the file name gives the mode.
*	MODE = G_READ or G_WRITE or G_APPEND
*
*   IO_DEVICE is a table whose elements are directly
*   associated with the file name.  Indexing IO_DEVICE<IO_CHANNEL<fname>>
*   (where 'fname' is a string) will input or output 
*   to the attached file.
* * *
*   The variable IO_DEVICE has been converted to an array. This was
*   done to solve the problom of associating files with elements of
*   tables. There is now a limit of 101 files that can be opened.
*

	G_OS = G_UNIX
	&ANCHOR = 0
	IO_CHANNEL = TABLE(200)
	IO_MODE = TABLE(100)
	IO_DEVICE = ARRAY('0:100')
	IO_CHANNEL<0> = G_IN
	IO_CHANNEL<1> = G_OUT
	IO_CHANNEL<2> = G_ERR
	IO_CHANNEL<G_IN> = 0
	IO_CHANNEL<G_OUT> = 1
	IO_CHANNEL<G_ERR> = 2
	IO_MODE<G_IN> = G_READ
	IO_MODE<G_OUT> = G_WRITE
	IO_MODE<G_ERR> = G_WRITE
	INPUT(.IO_DEVICE<IO_CHANNEL<G_IN>>,0," -f00")
	OUTPUT(.IO_DEVICE<IO_CHANNEL<G_OUT>>,1," -f01")
	OUTPUT(.IO_DEVICE<IO_CHANNEL<G_ERR>>,2," -f02")
ENDIOINIT
-TITLE		GPSG -- System Level I/O Routines
-STITLE 	Function: OPEN(FNAME,MODE,FTYPE)
*FRAME
* Function: OPEN
* Date: Thu May 30 13:35:37 1985
* Written by: Randy Kopach
* Last modification: Thu May 30 13:35:37 1985
* Purpose:  Opens a file for reading or writing.  There
*  are two entry points in the program (one for MTS, the
*  other for UNIX).
*
* Calling sequence:
*	result = OPEN(FNAME,MODE,FTYPE)
* Parameters: 
*	FNAME - File name to be opened
*	MODE -	either G_READ, G_WRITE, or G_APPEND (default is G_READ)
*	FTYPE - by default, a permanent file will be used.
*		A temporary file can be specified in one of 3 ways:
*			G_MTSTEMP --- Use temporary file only if on MTS
*			G_UNIXTEMP -- Use temporary file only if on UNIX
*			G_TEMP ------ Use temporary file always
* On return:
*	RETURN -	Open succeeded
*	FRETURN -	Either an OS error or parameter error occured.
*			Check STDERR for message.
*
* Local variables: 
*	CHANNEL -	I/O channel number
*	HEADER -	NULL (for permanent file)
*			either "/tmp/" (UNIX) or "-" (MTS) for temporary file
*
	DEFINE('OPEN(FNAME,MODE,FTYPE)CHANNEL,HEADER',G_OS '_OPEN')
+							:(ENDOPEN)
MTS_OPEN
	FTYPE POS(0) (G_MTSTEMP | G_TEMP)		:F(IO_OP_MPERM)
	HEADER = '-'
IO_OP_MPERM
+	IDENT(IO_FDUBS<FNAME>,NULL)			:S(IO_OP_NOFDUB)
	SFREEFD(IO_FDUBS<FNAME>)
IO_OP_NOFDUB
+	IO_FDUBS<FNAME> = SGETFD(HEADER FNAME ' ')	:S(RETURN)
	ERROR('OPEN: Error getting FDUB for file ' HEADER FNAME)
+							:(FRETURN)
UNIX_OPEN
* * *
*   We want to open a file in the UNIX OS.  First, we find
*   an I/O Channel number that is free for use.
* * *
	FTYPE POS(0) (G_UNIXTEMP | G_TEMP)		:F(IO_OP_UPERM)
	HEADER = '/tmp/'
IO_OP_UPERM
+	CHANNEL = 5
IO_OP_CHANNEL
+	IDENT(IO_CHANNEL<CHANNEL>,NULL) 		:S(IO_OP_GOT_CHANNEL)
	CHANNEL = LT(CHANNEL,99) CHANNEL + 1		:S(IO_OP_CHANNEL)
	ERROR('OPEN: No channels free') 		:(FRETURN)
IO_OP_GOT_CHANNEL
+	IO_CHANNEL<CHANNEL> = FNAME
	IO_CHANNEL<FNAME> = CHANNEL
	MODE (POS(0) (G_READ | NULL) RPOS(0))		:F(IO_OP_WRITE)
* * *
*   Open for reading
* * *
	FILES ' ' FNAME ' '				:F(IO_OP_ERROR3)
	IO_MODE<FNAME> = G_READ
	INPUT(.IO_DEVICE<IO_CHANNEL<FNAME>>,CHANNEL,HEADER FNAME ' ')
+							:F(IO_OP_ERROR1)
+							 S(RETURN)
IO_OP_WRITE
+	IDENT(MODE,G_WRITE)				:F(IO_OP_APPEND)
* * *
*   Open for writing
* * *
	IO_MODE<FNAME> = G_WRITE
	OUTPUT(.IO_DEVICE<IO_CHANNEL<FNAME>>,CHANNEL,HEADER FNAME ' ')
+							:F(IO_OP_ERROR1)
+							 S(RETURN)
IO_OP_APPEND
+	IDENT(MODE,G_APPEND)				:F(IO_OP_ERROR2)
* * *
*   Open to append to end of file
* * *
	IO_MODE<FNAME> = G_APPEND
	OUTPUT(.IO_DEVICE<IO_CHANNEL<FNAME>>,CHANNEL,HEADER FNAME ' -a')
+							:S(RETURN)
IO_OP_ERROR1
+	IO_CHANNEL<CHANNEL> = NULL
	IO_CHANNEL<FNAME> = NULL
	ERROR('OPEN: Error in variable association.  OPEN failed')
+							:(FRETURN)
IO_OP_ERROR2
+	ERROR('OPEN: Illegal mode specified in call: "' MODE '"')
+							:(FRETURN)
IO_OP_ERROR3
+	ERROR('OPEN: Attempt to open nonexistent file') :(FRETURN)
ENDOPEN
-STITLE 	Function: CLOSE(FNAME)
*FRAME
* Function: CLOSE
* Date: Thu May 30 13:56:43 1985
* Written by: Randy Kopach
* Last modification: Thu May 30 13:56:43 1985
* Purpose: Closes a file previously opened by OPEN
*	
*
* Calling sequence:
*	result = CLOSE(FNAME)
* Parameters: 
*	FNAME - File name to be closed
* On return:
*	RETURN -	Always occurs.
*	FRETURN -	Never occurs.
*
* Author's note:
*	Most SPITBOL programmers find it somewhat asthetically
*	pleasing to have at least one routine somewhere
*	that always succeeds.  Well, this is it.
*
*	"And on the sixth day, the programmer looked down, and
*	 saw that it was good.	And on that day he went skiing."
*
	DEFINE('CLOSE(FNAME,FTYPE)HEADER',G_OS '_CLOSE')
+							:(ENDCLOSE)
MTS_CLOSE
* * *
*   Closing the file on MTS.  Default file type is PERMANENT
* * *
	FTYPE POS(0) (G_MTSTEMP | G_TEMP)		:F(IO_CL_MPERM)
	HEADER = '-'
IO_CL_MPERM
+	SFREEFD(IO_FDUBS<FNAME>)
	IO_FDUBS<FNAME> = NULL			:(RETURN)
UNIX_CLOSE
* * *
*   We want to close a file in the UNIX OS.  We need to detach
*   the variable associated with the file, then close the file.
*   Following this, we must update our data structures so that
*   we know the channel is free and the file is not accessable.
* * *
	ENDFILE(IO_CHANNEL<FNAME>)
	DETACH(IO_DEVICE<IO_CHANNEL<FNAME>>)
	IO_CHANNEL<IO_CHANNEL<FNAME>> = NULL
	IO_CHANNEL<FNAME> = NULL
	IO_MODE<FNAME> = NULL				:(RETURN)
ENDCLOSE
-STITLE 	Function: READ(FNAME)
*FRAME
* Function: READ
* Date: Thu May 30 13:54:04 1985
* Written by: Randy Kopach
* Last modification: Tue Jul 16 10:34:04 1985
* Purpose:  Reads a line from a file
*	
*
* Calling sequence:
*	result = READ(FNAME)
* Parameters: 
*	FNAME - File to read a line from
* On return:
*	RETURN -	Returns line read in
*	FRETURN -	EOF or file not open to read.
*			Check STDERR for error message.
*
	DEFINE('READ(FNAME)',G_OS '_READ')		:(ENDREAD)
MTS_READ
+	IDENT(FNAME,G_IN)				:S(IO_RE_STDIN)
	IDENT(IO_MODE<FNAME>,NULL)			:S(IO_RE_ERROR1)
	IDENT(IO_MODE<FNAME>,G_READ)			:F(IO_RE_ERROR2)
	READ = SREAD(IO_FDUBS<FNAME>)			:S(RETURN)F(FRETURN)
IO_RE_STDIN
+	READ = INPUT					:S(RETURN)F(FRETURN)
UNIX_READ
+	IDENT(IO_MODE<FNAME>,NULL)			:S(IO_RE_ERROR1)
	IDENT(IO_MODE<FNAME>,G_READ)			:F(IO_RE_ERROR2)
	READ = IO_DEVICE<IO_CHANNEL<FNAME>>		:S(RETURN)F(FRETURN)
IO_RE_ERROR1
+	ERROR('READ: File ' FNAME ' not open')		:(FRETURN)
IO_RE_ERROR2
+	ERROR('READ: File ' FNAME ' not open for read') :(FRETURN)
ENDREAD
-STITLE 	Function: WRITE(FNAME,LINE)
*FRAME
* Function: WRITE
* Date: Thu May 30 13:47:28 1985
* Written by: Randy Kopach
* Last modification: Thu May 30 13:47:28 1985
* Purpose: Writes line to specified file.
*
* Calling sequence:
*	result = WRITE(FNAME,LINE)
* Parameters: 
*	FNAME - File name to be written to
*	LINE -	Line to write to file
* On return:
*	RETURN -	No problems with write
*	FRETURN -	Either file is not open to write or append,
*			or an OS error occured. Check STDERR for message.
*
	DEFINE('WRITE(FNAME,LINE)',G_OS '_WRITE')	:(ENDWRITE)
MTS_WRITE
+	IDENT(FNAME,G_OUT)				:S(IO_WR_STDOUT)
	IDENT(FNAME,G_ERR)				:S(IO_WR_STDERR)
	IDENT(IO_MODE<FNAME>,NULL)			:S(IO_WR_ERROR1)
	DIFFER(IO_MODE<FNAME>,G_READ)			:F(IO_WR_ERROR2)
	SWRITE(IO_FDUBS<FNAME>,LINE)			:S(RETURN)F(FRETURN)
IO_WR_STDOUT
+	OUTPUT = LINE					:S(RETURN)F(FRETURN)
IO_WR_STDERR
+	OUTPUT = LINE					:S(RETURN)F(FRETURN)
UNIX_WRITE
+	IDENT(IO_MODE<FNAME>,NULL)			:S(IO_WR_ERROR1)
	DIFFER(IO_MODE<FNAME>,G_READ)			:F(IO_WR_ERROR2)
	IO_DEVICE<IO_CHANNEL<FNAME>> = LINE		:S(RETURN)F(FRETURN)
IO_WR_ERROR1
+	ERROR('READ: File ' FNAME ' not open')		:(FRETURN)
IO_WR_ERROR2
+	ERROR('READ: File ' FNAME ' not open for write')
+							:(FRETURN)
ENDWRITE
-STITLE 	Function: PREFIX(PFX_CHR) (UNIX only)
	IDENT(G_OS,G_MTS)				:S(ENDPREFIX)
	DEFINE('PREFIX(PFX_CHR)')			:(ENDPREFIX)
PREFIX	G_PREFIX ' ' = NULL
	PREFIX = G_PREFIX
	G_PREFIX = PFX_CHR ' '				:(RETURN)
ENDPREFIX
-TITLE		GPSG -- High Level I/O Routines
-STITL		Function: GETLINE([FNAME])
*FRAME
* Function: GETLINE 
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
* GETLINE will get the user input, translate the command to lower
* case, and take out leading and trailing spaces. It also allows
* lines to be continued using the continuation character "/". The
* prefix character is then temporarily changed.
* 
* Calling sequence: 
*	return = GETLINE([FNAME])
* Parameters:
*	GL_NAME - File name to read from (defaults to G_IN)
* 
* On return: 
*	RETURN	- Normal termination.
*	FRETURN - End-of-file reached.
* 
* Local variables: 
*	GL_LINE - User input line.
*	GL_RET - Retained part of line (when "/" is used)
*	GL_PREF - Old prefix character (when "/" is used)
*	GL_GL_UC - Upper case form of each word in line.
*
	DEFINE('GETLINE(GL_FNAME)GL_LINE,GL_RET,GL_PREF,GL_UC')
+							:(ENDGETLINE)
GETLINE
+	GL_RET = null
	GL_FNAME = IDENT(GL_FNAME,NULL) G_IN
GL_LOOP
	GL_LINE = READ(GL_FNAME)			:F(GL_EOF)
* CHANGE JUNE 1997 Lines starting with '#' are comments.
        GL_LINE POS(0) '#'				:S(GETLINE)
	GL_LINE = TRIM(GL_LINE)
	GL_LINE RTAB(1) '/'				:F(GL_DONE)
	GL_PREF = PREFIX(IDENT(GL_PREF) "/")
	(GL_RET GL_LINE) RTAB(1) . GL_RET		:(GL_LOOP)
-SPACE
GL_DONE
+	GETLINE = GL_RET TRIM(GL_LINE)
	GETLINE POS(0) SPAN(' ') = null
*
* if the first word is a valid command in LC then translate it.
*
	(GETLINE " ") BREAK(" ") . GL_UC
	GL_LINE = REPLACE(GL_UC, "ABCDEFGHIJKLMNOPQRSTUVWXYZ",
+	"abcdefghijklmnopqrstuvwxyz")
	DIFFER(G_MAINCMDS<GL_LINE>)			:S(GL_CMDLC)
	DIFFER(G_DEBUGCMDS<GL_LINE>)			:F(GL_NOCMD)S(GL_CMDLC)
GL_CMDLC
+	GETLINE GL_UC = GL_LINE 
GL_NOCMD
+	PREFIX(DIFFER(GL_PREF) GL_PREF) 		:(RETURN)
GL_EOF	WRITE(G_OUT,' ')				:(FRETURN)
ENDGETLINE
-STITL		Function: PUTLINE(PL_LINE)
*FRAME
* Function: PUTLINE 
* Date: June 1983 
* Last modification: June 1983
* Purpose: 
*	To print a line to the user's terminal.
* 
* Calling sequence: 
*	PUTLINE(PL_LINE) 
* Parameters: 
*	PL_LINE - Line to be printed.
* 
* On return: 
*	RETURN	- Normal termination.
*	FRETURN - Nothing attached to SPRINT?
* 
* Local variables: None.
* 
	DEFINE('PUTLINE(PL_LINE)')			:(ENDPUTLINE)
PUTLINE
+	INDEX	 = G_PREFIX
	WRITE(G_OUT,PL_LINE)				:S(RETURN)F(FRETURN)
ENDPUTLINE
-STITL Function ERROR(LINE)
*FRAME
* Function ERROR(LINE)
* Purpose:
*	Output an error message. On Unix the message goes to stderr.
*	On mts ?????????????
* Parameters:
*	LINE -line to be outputted
* On return:
*	RETURN -output succeeded
*	FRETURN -output failed
*
	DEFINE('ERROR(LINE)')				:(ENDERROR)
ERROR	WRITE(G_ERR,'*** ' LINE)			:S(RETURN)F(RETURN)
ENDERROR
*
* Default values for the variables are:
*
* CHANGE JUNE 97 JF P_TRACE defaulted to 0 to speed things up a bit.
*                   P_INDENT defaulted to 2 for readability.
*                   P_VARSIZE defaulted to 40.
*                   P_PROOFSIZE defaulted to 100000
*                   P_ANTES defaulted to 100001
*                   P_HISTSIZE defaulted to 0
*                   P_CONDENSE added
*
      P_PROFILE      = 'OFF'
      P_TESTPARS     = 'OFF'
      P_FTRACE	     = 0	 ;  &FTRACE	   = P_FTRACE
      P_TRACE	     = 0	 ;  &TRACE	   = P_TRACE
      P_HISTSIZE     = 0
      P_INDENT	     = 2
      P_LINELEN      = 60
      P_CHECKPRINT   = 0
      P_PROOFSIZE    = 20000
      P_ANTES	     = 20001
      P_CONDENSE     = 0
      P_MAXPREM      = 50
      P_FINDUICNT    = 3
      G_VARSIZE      = 220
      G_VARORDER     = 'xyzwuvefghijklmnopqrst'
      G_PREDCHARS    = ANY('BCDFGHIJKLMNOPQRSTUVWXYZ')
      G_VARNAMES     = SUBSTR(G_VARORDER,1,G_VARSIZE / 10)
      G_MAXVARSIZE   = SIZE(G_VARORDER) * 10
* CNH Feb 21 1987    use P_NOID to check use of identity.
      P_NOID = 0
      INITPAT()
* G_ADD and G_DELETE are constants used for adding 1 or subtracting 1
* from occurance tables, and are also used as flags to indicate
* whether to push or pop from a stack of occurances.
      G_ADD = 1
      G_DELETE = -1
	G_PREFIX = '|'
*  CHANGES July 30, 1986   CNH
*  initialize G_TRUE
     G_TRUE = 'T'
*  CHANGES Aug 13, 1986  CNH initialize p_form to not format long
      P_FORMAT = 0
* CNT ADD P_PARALLEL
      P_PARALLEL = 0
*
* G_VERSION stands for <account (1=DANL, 2=TINK)>.<number>
*
      G_VERSION = G_OS ' :TINK.2.4'
*
* If we are on UNIX then produce an object file 
*
*	IDENT(G_OS,G_UNIX)				:F(MAIN_MTS)
*	EXIT(3)
*	OUTPUT(.INDEX,3, ' -f1 -w -c')
MAIN_MTS
      DOPREM(,'INIT')
      PUTLINE("	     THINKER version: " G_VERSION)
      PUTLINE()
      CMDPARSE(G_MAINCMDS,"prove","|")			      :(END)
END
